# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import time
import datetime
import Queue
import uuid
import json
import collections

from sanet_common.sanetdefs.constants import DEFAULT_PRIORITY
from sanet_common.sanetdefs.constants import _CSTATE_TO_CODENAME, ELEMENT_KIND_STR

from entables.constants import *

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class Context(object):

	def __init__(self):
		self.registered_matches = MatchDefinitionList()
		self.registered_targets = TargetDefinitionList()
		self.registered_listeners = ListenerDefinitionList()
		self.registered_tagproviders = TagproviderDefinitionList()

		self.listenerthreads = {}

		self.targetthreads = {}

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class TemplateMemoryObject(object):

	def __init__(self, name, data):
		self.data = data
		self.name = name
 
#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class TagproviderDefinitionList(dict):

	def add_tagprovider(self, tagproviderdef):
		self[tagproviderdef.name] = tagproviderdef

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class ListenerDefinitionList(dict):

	def add_listener(self, name, listener_class):
		self[name] = listener_class

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#
    
class TargetDefinitionList(dict):

	def add_target(self, targetdef):
		self[targetdef.name] = targetdef

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class MatchDefinitionList(dict):

	def add_match(self, matchdef):
		self[matchdef.get_name()] = matchdef

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

    
class MatchMemoryObject(object):

	def __init__(self):
		self.name = ""
		self.options = {} #dict of "name" : "value" 

	def _str_(self):
		return self.name

	def build(self, arguments):
		pass

	def set_options(self, options):
		self.options = options

	def verify(self,context, alarm):
		options = self.options
		handler = context.registered_matches[self.name]
		return handler.match(options, alarm)

	def objectdump(self):
		matchoptionsdump = {}
		for option in self.options.keys():
			matchoptionsdump[option] = self.options[option]

		return {
			"name"   : self.name,
			"options": matchoptionsdump,
		}

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class TargetMemoryObject(object):

	def __init__(self, name="", options=None, table="", chain=""):
		self.name    = name
		self.options = options or {}
		self.table   = table
		self.chain   = chain

	def set_options(self, options):
		self.options = options

	def objectdump(self):
		return {
			"name"   : self.name,
			"options": dict( self.options.items() ),
			"table"  : self.table,
			"chain"  : self.chain,
		}

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class TableMemoryObject(object):

	def __init__(self):
		self.name = ""
		self.priority = 0
		self.chains = {} # dict of "name": ChainMemoryObject

	def objectdump(self):
		table = {}
		for chain in self.chains.keys():
			table[chain] = self.chains[chain].objectdump()
		return table
    
#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#
    
class ChainMemoryObject(object):

	def __init__(self):
		self.name = ""
		self.rules = [] #list of RuleMemoryObject
		self.policy = TargetMemoryObject()

	def objectdump(self):

		return {
			'rules' : [ rule.objectdump() for rule in self.rules ],
			'policy': self.policy.objectdump(),
		}
 
#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class RuleMemoryObject(object):

	def __init__(self, targetobj, uuid=None, table="", chain="", position=None, stop_if_match=False, disabled=RULE_STATE_ACTIVE):

		self.uuid  = uuid
		self.table = table
		self.chain = chain

		self.position = position

		self.stop_if_match = stop_if_match

		# 0 = Not disabled
		# 1 = Administratively Disabled
		# >2 = Disabled because of errors
		self.disabled = disabled

		self.matchset = [] # list of MatchMemoryObject

		self.target = targetobj

		# TODO: validate rule_def
		# FIXME: Ogni RULE deve avere un match di default che restituisce sempre 1
		# cosi le regole "entable -j SMS" possono rappresentare un default su cui
		# tutti gli allarmi ricadono

	def add_match(self,matchobj):
		self.matchset.append(matchobj)

	def del_match(self,name):
		pass

	def objectdump(self):
		rule = {}
		rule.update(self.__dict__)
		rule['matchset'] = []
		for match in self.matchset:
			rule['matchset'].append(match.objectdump())
		rule['target'] = self.target.objectdump()
		return rule

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class TenantMemoryObject(object):

	def __init__(self,uuid, name):

		self.uuid = uuid
		self.name = name

		self.tables = collections.OrderedDict()
		self.templates = {}
		self.active_table = TableMemoryObject() 

	def objectdump(self):

		tenant = {}
		tenant["tables"] = {}
		tenant["templates"] = {}

		tables = self.tables.keys()

		for table in self.tables.keys():
			tenant[table] = self.tables[table].objectdump() 

		return tenant

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class AlarmMemoryObject(object):

	def __init__(self, tenant_uuid=None, auuid=None, priority=None, status=None, data=None,  stages=None, rulestack=None, hold_time=0, hold_wakeuptime=0, hold_handler=None, merge_group_key=None):

		self.tenant_uuid = tenant_uuid
		self.uuid        = auuid

		self.status      = status
		self.priority    = priority if priority is not None else DEFAULT_PRIORITY

		self.data        = data
		self.stages      = stages

		self.rulestack  = rulestack

		self.hold_time       = hold_time
		self.hold_wakeuptime = hold_wakeuptime
		self.hold_handler    = hold_handler

		self.merge_group_key = merge_group_key

		if self.uuid == None:
			self.uuid = uuid.uuid4().hex

	# Decode Constants
	#def decode_constants(self):    
	#    self.data["result"]["status"] = _CSTATE_TO_CODENAME[self.data["result"]["status"]]
	#    self.data["element"]["type"] = ELEMENT_KIND_STR[self.data["element"]["type"]]

	#def add_stage(self, stage):
	#	self.stages.append(stage)
	
	def has_rulestack(self):
		return (self.rulestack is not None)
	
	def get_rulestack(self):
		return self.rulestack
		#return self.rulestack if self.rulestack is not None else []

	def set_rulestack(self, rulestack):
		self.rulestack = rulestack


	def set_hold_state(self, hold_time, hold_wakeuptime, hold_handler=None):
		
		self.set_verbstatus("Changed status to HOLD")
		
		self.status          = ALARM_STATUS_HOLD
		
		self.hold_time       = hold_time
		self.hold_wakeuptime = hold_wakeuptime
		self.hold_handler    = hold_handler
		
	def reset_hold_data(self):
		self.hold_time       = 0
		self.hold_wakeuptime = 0
		self.hold_handler    = None

	def set_verbstatus(self, s):
		#if not self.data.has_key('verbstatus'):
		#	self.data['verbstatus'] = []
		#self.data['verbstatus'].append("%s: %s" % (time.time(), s))
		if self.stages == None:
			self.stages = []
		
		ts = time.time()
		
		d = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H:%M:%S.%f')
			
		self.stages.append("%s (%s): %s" % (d, ts, s))

			
		
	def get_verbstatus(self):
		#return self.data['verbstatus']
		return self.stages
