# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import json
import time
import datetime

from sanet_common.sanetdefs.timing import NoHolidayCalendarHandle

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

_DAYS = {
	'monday'   : 1,
	'tuesday'  : 2,
	'wednesday': 3,
	'thursday' : 4,
	'friday'   : 5,
	'saturday' : 6,
	'sunday'   : 7,
}

#NEXT_TS_MIN_DELTA = datetime.timedelta(seconds=60 * 5)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def calculate_until_ts( start_ts, until_format, calendar_handler=None):
	"""
	Calcola un timestamp sulla base del timestamp di partenza "start_ts" e 
	le specifiche temporali descritte dalla stringa "until_format":
		
		1) yyyy-mm-gg [ hh:mm [ :ss ]  ]
		
		2) ext-work-day <hh>:<mm> [ <working days> ]
		
		3) <day name>          [ hh:ss ]
		   <day week number>   [ hh:ss ]
	"""

	#
	# yyyy-mm-gg [ hh:mm [ :ss ]  ]
	#

	date = __parse_isostring_to_date(until_format)
	if date != None:
		#print "FORMAT yyy-mm-gg [ hh:ss ]"
		return int(date.strftime("%s"))

	#
	# next-working-day <hh>:<mm> [ <working days> ]
	#
	date = __parse_next_working_day(start_ts, until_format, calendar_handler=calendar_handler)
	if date != None:
		return int(date.strftime("%s"))


	#
	# <day name>          [ hh:ss ]
	# <day week number>   [ hh:ss ]
	#
	date = __parse_day_name_number_and_hour(start_ts, until_format)

	return int(date.strftime("%s"))

#------------------------------------------------------------------------------#

def __parse_isostring_to_date(string):

	try:
		return datetime.datetime.strptime(string, '%Y-%m-%d')
	except ValueError:
		pass

	try:
		return datetime.datetime.strptime(string, '%Y-%m-%d %H:%M')
	except ValueError:
		pass

	try:
		return datetime.datetime.strptime(string, '%Y-%m-%d %H:%M:%S')
	except ValueError:
		pass

	return None

#------------------------------------------------------------------------------#

def __parse_next_working_day(start_ts, until_format, calendar_handler=None):

	if calendar_handler is None:
		calendar_handler = NoHolidayCalendarHandle()

	if not until_format.startswith('next-work-day'):
		return None

	#
	# Parsing dei parametri
	#
	#    next-working-day <hh>:<mm> [ <working days> ]
	#
	parts = until_format.replace('next-work-day','').strip().split(' ', 1)

	if len(parts) == 1:
		hour         = parts[0]
		working_days = None
	else:
		hour         = parts[0]
		working_days = [int(d) for d in parts[1].split(",")]

	h, m = hour.split(':')
	h, m = int(h), int(m)

	if not working_days:
		working_days = [1,2,3,4,5]

	#
	# Primo giorno "next" da oggi
	#
	start_date = datetime.datetime.fromtimestamp(start_ts)

	target_date = datetime.datetime(start_date.year, start_date.month, start_date.day)
	target_date += datetime.timedelta(days=1)

	#
	# Andiamo avanti finche' non incontriamo un giorno lavorativo valido
	#
	while True:

		is_h = calendar_handler.isHoliday(target_date.year, target_date.month, target_date.day)

		#print target_date, target_date.weekday()+1, is_h

		if not is_h:

			wday = target_date.weekday()+1
			if wday in working_days:
				break

		target_date += datetime.timedelta(days=1)

	#
	# Giorno valido + orario specificato
	#
	d = datetime.datetime(target_date.year, target_date.month, target_date.day) + datetime.timedelta(hours=h, minutes=m)

	return d

#------------------------------------------------------------------------------#

def __parse_day_name_number_and_hour(start_ts, until_format):

	d, h, m = __parse_holduntil(until_format)

	date = __calculate_next_date( datetime.datetime.fromtimestamp(start_ts), d, h, m)

	return date


def __calculate_next_date( start_date, d, h, m):

	start_date = start_date.replace(hour=0, minute=0, second=0, microsecond=0)
	start_day  = start_date.weekday() + 1  # days start from 1, not 0

	# la data di inizio cade in un giorno della settimana successivo a quello indicato,
	# quindi dobbiamo passare alla settimana successiva.
	if start_day >= d:
		days_to_go = (7 - start_day) + d
	else:
		days_to_go = d - start_day

	delta    = datetime.timedelta(days=days_to_go, hours=h, minutes=m)

	return start_date + delta


def __parse_holduntil(s):

	parts = s.split(' ')
	if len(parts) == 1:
		day  = parts[0]
		hour = None
	elif len(parts) == 2:
		day  = parts[0]
		hour = parts[1]
	else:
		raise Exception("Invalid syntax: %s" % (s))

	# Day
	day_num = None
	try:
		day_num = _DAYS[day.lower()]
	except Exception, e:
		try:
			day_num = int(day)
		except Exception, e:
			pass

	if day_num == None:
		raise Exception("Invalid syntax: %s" % (s))

	if day_num not in [1,2,3,4,5,6,7]:
		raise Exception("Invalid day: %s" % (day))

	d = day_num

	h, m = _parse_hour(hour)

	return d, h, m

def _parse_hour(hour):
	"""
	Converte un stringa con sintassi <hh>:<mm> nella tupla numerica corrispondente (<hh>, <mm>).
	
	Se "hour" e' None restituisce la tupla (0, 0)
	"""

	# Hour
	if hour != None:
		h,m = hour.split(':')
		h = int(h)
		m = int(m)
		if not ((0<=h) and (h<=23)):
			raise Exception("Invalid hour: %s" % (h))
		if not ((0<=m) and (h<=59)):
			raise Exception("Invalid hour: %s" % (m))
	else:
		h = 0
		m = 0

	return h, m

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def calculate_next_ts(start_ts, until_next_format):

	try:
		h, m = _parse_hour(until_next_format)

		start_date = datetime.datetime.fromtimestamp(start_ts)

		next_date = start_date

		next_date = next_date.replace(hour=h, minute=m, second=0, microsecond=0)
		#print "NEXT DATE", next_date
		next_date_ts = int(next_date.strftime("%s"))

		# L'orario e' precedente all'ora di inizio, quindi
		# dobbiamo passare all'ora nel giorno successivo.

		if (next_date_ts < start_ts):
			next_date = start_date + datetime.timedelta(days=1)
			next_date = next_date.replace(hour=h, minute=m, second=0, microsecond=0 )
			#print "NEXT DATE -> NEXT", next_date
			next_date_ts = int(next_date.strftime("%s"))

		return next_date_ts
	except Exception, e:
		print e
		raise e

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

if __name__ == "__main__":

	import sys
	import time

	from sanet_common.timeutils import timestamp_to_isostr

	now = time.time()

	ts = calculate_next_ts(now, "8:00")
	print ts, timestamp_to_isostr(ts)

	print "--------------------------------------------"
	print timestamp_to_isostr( calculate_until_ts(now, "2015-04-17") )

	print "--------------------------------------------"
	print timestamp_to_isostr( calculate_until_ts(now, "2015-04-17 01:00") )

	print "--------------------------------------------"
	print timestamp_to_isostr( calculate_until_ts(now, "friday 01:00") )

	print "--------------------------------------------"
	print timestamp_to_isostr( calculate_until_ts(now, "3 01:00") )

	print "--------------------------------------------"
	print timestamp_to_isostr( calculate_until_ts(now, "next-work-day 01:00") )

