# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

.. _entables-matches-attributes:

Attributes
==========

Premessa: struttura degli allarmi in memoria
------------------------------------------------------------------------

Gli allarmi gestiti da entables sono memorizzati utilizzando una struttura a dizionario *chiave*/*valore*.

**NOTA**: La struttura completa di un allarme puo' cambiare nel tempo.

Esempio: 

::
	
	{
		'condition': {
			'classification': '',
			'expr': u"int(readFile('/tmp/prova.txt'))==1",
			'maxtries': 3,
			'name': 'c',
			'path': 'n1;;dg;c',
			'priority': 31,
			'priority_level': 'medium',
			'statuslastchange': 1414161050,
			'uuid': 'cb90f8d535ff4d9fbd149dae2908db1a'
		},
		'datasources': [],
	}

Per identificare una particolare *chiave*/*sottochiave* si utilizza il *path*. Il *path* e' una stringa che
codifica il livello di annidamento della chiave, ed ha la seguente sintassi:

::
	
	<key1>:<sub-key>:<sub-key>:......


Esempi di path e valori individuati:

::
	
	condition:maxtries          ---> 3
	condition:statuslastchange  ---> 1414161050


Utilizzo del match
------------------------------------------------------------------------

Il match *attributes* permette di fare match sui *valori* associati ad una qualunque delle *chiavi* del pacchetto dati dell'allarme.

Gli parametri del match vengono specificati con la sintassi:

::
	
	-m attr --attr_field_name <path>  --attr_field_value <valore>

I parametri sono:

	+--------------------------+---------------------------------------------------------------------------------------------------------------------------------------------+
	| Parametro                | Descrizione                                                                                                                                 |
	+==========================+=============================================================================================================================================+
	| attr_field_name          | Stringa che codifica il *path* di una chiave nel messaggio.                                                                                 |
	+--------------------------+---------------------------------------------------------------------------------------------------------------------------------------------+
	| attr_field_value         | Valore della chiave specificata da *attr_field_name*                                                                                        |
	+--------------------------+---------------------------------------------------------------------------------------------------------------------------------------------+
	| attr_field_value_between | Stringa che specifica i valori *min* e *max* (separati da spazio) ammessi per il valore della chiave *attr_field_name*                      |
	+--------------------------+---------------------------------------------------------------------------------------------------------------------------------------------+

Esempi: Questi sono possibili utilizzi del parametro *field_name* e *field_name*:

::
	
	--attr_field_name "condition:maxtries"            --attr_field_value  3        --> MATCH
	--attr_field_name "condition:statuslastchange"    --attr_field_value  234234   --> NO MATCH


Uso di shortcut generici
+++++++++++++++++++++++++++

Un allarme di sanet ha una struttura complessa che puo' anche variare nel tempo. 

Per facilitare il match su alcune *chiavi importanti* dell'allarme, sono disponibili dei parametri *shortcut* che permettono di semplificare la scrittura del match.

Quando si usa uno shortcut bisogna usare la seguente sintassi:

::
	
	--attr_<nome shortcut>  <valore>

Esempio: Stesso match usando o meno lo shortcut *path* :

::
	
	NORMALE    --attr_field_name  condition:path  --attr_field_value  'localhost;;icmp-reachability;reach'
	
	SHORTCUT   --attr_path                                            'localhost;;icmp-reachability;reach'

.. warning:: non e' possibile usare alcune opzioni di match (es: attr_field_between) quando si usano gli shortcut

Specificare valori usando regular expression
+++++++++++++++++++++++++++++++++++++++++++++

Per alcune *chiavi* (e corrispondendi *shortcut*) del messaggio e' possibile specificare un valore attraverso *espressioni regolari*.

Esempio:

::
	
	--attr_field_name  condition:path  --attr_field_value   '^sw.*'     \__ equivalneti
	--attr_path                                             '^sw.*'     /



Tabella riassuntiva degli shortcut disponiboli
++++++++++++++++++++++++++++++++++++++++++++++++

	+--------------------------------------+----------------+----------------------------------------------------------------------------------------------------+-----------------+
	| Reale *path* del campo nel messaggio | Accetta RegExp | Descrizione                                                                                        | shortcut        |
	+======================================+================+====================================================================================================+=================+
	| alarmtype                            |                | Stringa con il tipo di allarme. "normal" generati da Sanet, "pushed" generati da *push_alarm*      | alarmtype       |
	+--------------------------------------+----------------+----------------------------------------------------------------------------------------------------+-----------------+
	| condition:path                       | si             | Stringa con il path della condition che ha generato l'allarme                                      | path            |
	+--------------------------------------+----------------+----------------------------------------------------------------------------------------------------+-----------------+
	| element:name                         | si             | Nome dell'elemento (nome nodo, nome interfaccia, ecc...)                                           | name            |
	+--------------------------------------+----------------+----------------------------------------------------------------------------------------------------+-----------------+
	| result:status                        |                | Stato corrente della condition "UP", "DN", "FA"                                                    | status          |
	+--------------------------------------+----------------+----------------------------------------------------------------------------------------------------+-----------------+
	| condition:priority                   |                | Priorita della condition. Valori ammessi: none, 1...99                                             | priority        |
	+--------------------------------------+----------------+----------------------------------------------------------------------------------------------------+-----------------+
	| condition:priority                   |                | Priorita' minima della condition (in numero o livello di priorita') (es. 10, 'low', ecc. )         | min_priority    |
	+--------------------------------------+----------------+----------------------------------------------------------------------------------------------------+-----------------+
	| condition:priority_level             |                | Livello di priorita della condition (es. "critical")                                               | priority_level  |
	+--------------------------------------+----------------+----------------------------------------------------------------------------------------------------+-----------------+
	| mail:to                              |                | Destinario email dell'allarme (Attributi "email" e/o "upemail" della condition dell'allarme)       | mail_to         |
	+--------------------------------------+----------------+----------------------------------------------------------------------------------------------------+-----------------+
	| mail:subject                         |                | Soggetto email dell'allarme (Attributi "msg_upsubj" o "msg_downsubj" della condition dell'allarme) | mail_subject    |
	+--------------------------------------+----------------+----------------------------------------------------------------------------------------------------+-----------------+
	| mail:body                            |                | Corpo email dell'allarme (Attributi "msg_upbody" o "msg_downbody" della condition dell'allarme)    | mail_body       |
	+--------------------------------------+----------------+----------------------------------------------------------------------------------------------------+-----------------+

.. _entables-matches-attributes-digest:

Campi/shortcut per i messaggi digest
++++++++++++++++++++++++++++++++++++++

	+--------------------------------------+----------------+----------------------------------------------------------------------------------------------------+-----------------+
	| Reale *path* del campo nel messaggio | Accetta RegExp | Descrizione                                                                                        | shortcut        |
	+======================================+================+====================================================================================================+=================+
	| digestinfo:downs                     |                | Numero di condition down pendenti nel digest.                                                      | digest_downs    |
	+--------------------------------------+----------------+----------------------------------------------------------------------------------------------------+-----------------+
	| isdigest",                           |                | Se l'allarme e' un digest o meno (1 digest, 0 normale)                                             | isdigest        |
	+--------------------------------------+----------------+----------------------------------------------------------------------------------------------------+-----------------+



Vari esempi di utilizzo
-------------------------

::
	
	entables -t notify -A DEFAULT -m attr --attr_priority                                                       1       -j SMTP
	                                                                                                          
	entables -t notify -A DEFAULT -m attr --attr_priority                                                       none    -j SYSLOG --syslog_msg "ERRORE"
	                                                                                                          
	entables -t notify -A DEFAULT -m attr --attr_min_priority                                                   88      -j SMTP
	                                                                                                          
	entables -t notify -A DEFAULT -m attr --attr_min_priority                                                   low     -j SMTP
	                                                                                                          
	entables -t notify -A DEFAULT -m attr --attr_priority_level                                                 medium  -j DROP
	
	entables -t notify -A DEAFULT -m attr --attr_field_name        digestinfo:downs   --attr_field_value        0       -j DROP \_ equivalenti
	entables -t notify -A DEAFULT -m attr --attr_digestinfo_downs                                               0       -j DROP /
	
	
	entables -t notify -A DEAFULT -m attr --attr_field_name        condition:priority  --attr_field_value_range "42 51" -j DROP 


"""

from sanet_common.sanetdefs.constants import *


from core.models import Tenant, PriorityLabel
from core.api.base import db_set_current_tenant_schema

from entables.constants import TABLE_NOTIFY

from entables.matches import BaseMatch
from entables.targets import Option

"""
Match on alarm attributes

{
	'condition': {
		'classification': '',
		'expr': u"int(readFile('/tmp/prova.txt'))==1",
		'maxtries': 3,
		'name': 'c',
		'path': 'n1;;dg;c',
		'priority': 31,
		'priority_level': 'medium',
		'statuslastchange': 1414161050,
		'uuid': 'cb90f8d535ff4d9fbd149dae2908db1a'
	},
	'datasources': [],
	'element': {
		'kind': 1,
		'name': 'n1',
		'nodedescr': '',
		'parent_id': None,
		'parent_name': None,
		'path': 'n1',
		'type': 1,
		'uuid': 'ed2a3f23574d42c5a5821a35bc71340a'
	},
	'event': {
		'uuid': '5f3caa5047aa4b55b5d4d0325cfa0b76'
	},
	'flow-information': [],
	'info': {
		'sysdescr': None, 'syslocation': None, 'sysname': None
	},
	'linked': {
		'iface': None, 'node': None
	},
	'mail': {
		'body': '',
		'headers': {
			'X-SANET-Priority': 31,
			'X-SANET-Target-Id': 'n1;;dg;c',
			'X-SANET-Target-Status': 'DN'
		},
		'subject': '',
		'to': 'netwarning'
	},
	'params': {},
	'realemail': '',
	'result': {'laststatus': 20,
	'status': 10,
	'verbstatus': u" int(readFile('/tmp/prova.txt'))==1 expanded to (0 == 1)"},
	'snmp_values': [],
	'tags': [ 't1:/t1', 't1:/t1'],
	'tags_name': {
		't1': ['t1']
	},
	'tenantid': '7f4d9bb05cf44ce1ac432de2c3384677',
	'tenantname': 'site',
	'ts': 1414161050,
	'url': None,
	'uuid': 'dce1453dc37240a790677f9f0a898887'
}
"""

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

REAL_ATTR_PATH = {
	"alarmtype"     : "alarmtype",

	"path"          : "condition:path",
	
	"min_priority"  : "condition:priority",
	"priority"      : "condition:priority",
	"priority_level": "condition:priority_level",

	"name"          : "element:name",
	"type"          : "element:type",

	"status"        : "result:status",
	
	"mail_to"       : "mail:to",
	"mail_subject"  : "mail:subject",
	"mail_body"     : "mail:body",
	
	"digest_downs"  : "digestinfo:downs",
	"isdigest"      : "isdigest",
}

ATTR_DEFAULTS = {
	"isdigest" : 0,
}

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

from sanet_common.sanetdefs.constants import CSTATE_CODENAMES

class Match(BaseMatch):
	
	def __init__(self):
		super(Match, self).__init__()
		
		self.name    = "attr"
		
		self.tables  = [ TABLE_NOTIFY ]
		
		self.options  = [
			Option("alarmtype"     ),
			
			Option("path"          ),
			
			Option("min_priority"  ),
			Option("priority_level"),
			Option("priority"      ),

			Option("name"          ),
			Option("type"          ),
			
			Option("status"        , choices=['']+[c for c in CSTATE_CODENAMES]),
			
			
			Option("mail_to"       ),
			Option("mail_subject"  ),
			Option("mail_body"     ),

			Option("digest_downs"  ),

			Option("isdigest"      ),

			Option("field_name"          ),
			Option("field_value"         ),
			Option("field_value_between" ),
		]

	def validate_match_options(self, options):
		
		super(Match, self).validate_match_options(options)

		alarmtype = options.get('alarmtype', None)
		if alarmtype is not None and not alarmtype:
			raise Exception("Invalid value for alarmtype: %r" % (alarmtype))

		
		# Condition path
		path = options.get('path', None)
		if path is not None and not path:
			raise Exception("Invalid value for element path: %r" % (path))

		name = options.get('name', None)
		if name is not None and not name:
			raise Exception("Invalid value for element name: %r" % (name))
	    
		etype = options.get('type', None)
		if etype is not None:
			if etype not in ELEMENT_KINDS_STR:
				raise Exception("Invalid element kind: %s" % (etype))
	    
		status = options.get('status', None)
		if status is not None:
			if status not in CSTATE_CODENAMES:
				raise Exception("Invalid status code: %s" % (status))
	    
		min_priority   = options.get('min_priority', None)
		if min_priority is not None:
			if min_priority.isdigit():
				try:
					min_priority = int(min_priority)
					
					if not (PRIORITY_MIN <= min_priority and min_priority <= PRIORITY_MAX):
						raise ValueError(min_priority)
				except ValueError, ex:
					raise Exception("Invalid priority value: %s" % (min_priority) )
			else:
				try:
					l = PriorityLabel.objects.get(label=min_priority)
					return True
				except PriorityLabel.DoesNotExist, e:
					raise ValueError("Invalid priority label: %s" % (min_priority))
				
		priority_level = options.get('priority_level', None)
		if priority_level is not None:
			if priority_level.strip() == "":
				raise Exception("Invalid priority label: '%s'" % (priority_level))

		priority   = options.get('priority', None)
		if priority is not None:
			
			if priority == "none":
				return True

			try:
				priority = int(priority)
					
				if not (PRIORITY_MIN <= priority and priority <= PRIORITY_MAX):
					raise ValueError(priority)
			except ValueError, ex:
				raise Exception("Invalid priority value: %s" % (priority) )

		isdigest   = options.get('isdigest', None)
		if isdigest is not None:
			if not isdigest.isdigit():
				raise Exception("Invalid value: %s" % (isdigest) )
			if int(isdigest) not in [0,1]:
				raise Exception("Invalid value: %s" % (isdigest) )



		field_name = options.get('field_name', None)
		if field_name is not None:
			if field_name.strip() == "":
				raise Exception("Invalid value for field_name: %r" % (field_name))

		field_value = options.get('field_value', None)
		if field_value is not None:
			if field_value.strip() == "":
				raise Exception("Invalid value for field_value: %r" % (field_value))

		field_value_between = options.get('field_value_between', None)
		if field_value_between is not None:
			if field_value_between.strip() == "" or len(field_value_between.split(" ")) != 2:
				raise Exception("Invalid value for field_value_between: %r" % (field_value_between))

		if field_value and field_value_between:
			raise Exception("Can not specify both field_name_value and field_value_between")
			
		if field_name or field_value:
			if (field_name and not (field_value or field_value_between)) or (not field_name and (field_value or field_value_between)):
				raise Exception("Speficy both field_name and field_value or field_value_between")

		return True

	def match(self, context, alarmobj, options):
		
		CHECK_FIELDS_WITH_REGEXPR = [
			'name', 
			'path',
		]

		log = context.log

		alarm = alarmobj.data

		# Dobbiamo accedere ai livelli di priority del tenant. Aggiorniamo lo schema per essere sicuri di essere posizionati sullo schema corretto.
		tenant = Tenant.objects.get(uuid=alarmobj.tenant_uuid)
		db_set_current_tenant_schema(tenant.name)

		# Every option uses the generic path method
		# but specifies the path.

		log.debug("Checking match")

		options_to_match = self._prepare_options_to_match(log, options)
		if options_to_match is None:
			log.error("Saved options are not valid! Match not verified! Please verify your DB consistency")

		log.debug("Effective match options: %s", options_to_match)


		for option_name, option_value in options_to_match.items():

			# Queste opzioni vengono controllate indirettamente.
			if option_name in ['field_value_between', 'field_value']:
				continue

			try:
				log.debug("Checking attr %s = %r", option_name, option_value)
				
				if option_name == "field_name":
					
					selected_field_name  = option_value
					
					selected_field_value         = options_to_match.get('field_value'        , None)
					selected_field_value_between = options_to_match.get('field_value_between', None)
					
					#
					# Find value in alarm dictionary
					#
					real_field_value = None
					real_field_path  = option_value
					real_field_name  = None
					try:
						segment = alarm
						field_sequence = real_field_path.split(":")
						for field in field_sequence:
							segment = segment[field]

						real_field_name  = field_sequence[-1]
						real_field_value = segment

					except KeyError, e:
						log.debug("Attribute '%s' not found. (%s:%s)", option_name, type(e), str(e))
						return False
					
					#
					# Match value
					#
					if selected_field_value:

						if str(real_field_value) != str(selected_field_value):
							if real_field_name in CHECK_FIELDS_WITH_REGEXPR:
								try:
									if not re.search(selected_field_value, real_field_value):
										log.debug("Field '%s' with value '%s' do not match specified value '%s'", selected_field_name, real_field_value, selected_field_value)
										return False
								except Exception, e:
									# Wrong regexpr?
									return False
							else:
								log.debug("Field '%s' has value '%s' != '%s'", selected_field_name, real_field_value, selected_field_value)
								return False
								
					elif selected_field_value_between:

						log.debug("Match with value: %s", selected_field_value_between)

						min_val, max_val = selected_field_value_between.split(' ')
						min_val = float(min_val)
						max_val = float(max_val)

						log.debug("   max value: %s", max_val)
						log.debug("   min value: %s", min_val)
						log.debug("   value    : %s", real_field_value)

						if not ((min_val <= float(real_field_value)) and (float(real_field_value) <= max_val)):
							return False
					else:
						return False

				else:

					# Find field in alarm message
					real_path = REAL_ATTR_PATH.get(option_name, None)
					if real_path is None:
						log.debug("Option '%s' not recognized!", option_name)
						continue
					
					real_field_value = None
					real_field_value = None
					try:
						segment = alarm
						field_sequence = real_path.split(":")
						for field in field_sequence:
							segment = segment[field]
							
						real_field_value = segment
						real_field_name  = field_sequence[-1]
						
					except KeyError, e:
						log.debug("Attribute '%s' not found. (%s)", option_name, e)
						
						if ATTR_DEFAULTS.has_key(option_name):
							real_field_value = ATTR_DEFAULTS[option_name]
							log.debug("Using default value for attribute '%s': %s", option_name,real_field_value )
						else:
							return False

					log.debug("Matching attribute value %r with option value %r", segment, option_value)
				
					if option_name == 'min_priority':
						try:
							p = int(real_field_value)
						except ValueError, e:
							matched = False
							
						matched = p >= option_value
						
					elif option_name == 'priority':
						
						if option_value == "none":
							option_value = PRIORITY_NONE
							
						try:
							matched = (int(option_value) == real_field_value)
						except ValueError, e:
							matched = False
					else:
						matched = (option_value == real_field_value) or (str(option_value) == str(real_field_value))
						
						if not matched and option_name in CHECK_FIELDS_WITH_REGEXPR:
							try:
								matched = matched or re.search(option_value, real_field_value)
							except Exception, e:
								pass
					
					if not matched:
						log.debug("Attribute '%s' not mached with '%s'", option_name, real_field_value)
						return False

					log.debug("Attribute '%s' matched with '%s'", option_name, segment)

			except Exception, e:
				log.error("%s: ERROR: %s", self.name, str(e))
				return False

		alarmobj.set_verbstatus("\t match '%s' matched: %s" % (self.name, options))
		
		return True


	def _prepare_options_to_match(self, log, options):
		
		real_options = {}
		
		for option_name, option_value in options.items():

			try:
				if option_value is None:
					continue

				if option_name == 'status':
					option_value = CSTATE_FROM_CODENAME(option_value)
					if not option_value:
						raise Exception("Saved option '%s' with invalid value '%s" % (option_name, option_value))
				elif option_name == 'min_priority':
					try:
						if option_value.isdigit():
							option_value = int(option_value)
						else:
							l = PriorityLabel.objects.get(label=option_value)
							option_value = l.min_level
					except PriorityLabel.DoesNotExist, e:
						raise Exception("Saved option '%s' with invalid priority level '%s'" % (option_name, option_value))
					except ValueError, e:
						raise Exception("Saved option '%s' with invalid value '%s'" % (option_name, option_value))
						
				real_options[ option_name ] = option_value
			except Exception, e:
				log.error("Error while peparing for match: %s:%s" % (type(e), str(e)) )

		return real_options
		
		
		
