# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

Status
======

Questo match verifica lo stato della condition che ha generato l'allarme:

	-m status --status_<attributo> <valore>

Gli attributi su cui e' possibile fare match sono:

	+---------------+---------------+------------------+---------------------------------------------------------------------------------------------------+
	| Parametro     | Type          | Default          | Descrizione                                                                                       |
	+===============+===============+==================+===================================================================================================+
	| curr          | str           |                  | stato corrente                                                                                    |
	+---------------+---------------+------------------+---------------------------------------------------------------------------------------------------+
	| prev          | str           |                  | stato precedente                                                                                  |
	+---------------+---------------+------------------+---------------------------------------------------------------------------------------------------+
	| changedin     | numero        |                  | Numero di secondi dall'ultima variazione                                                          |
	+---------------+---------------+------------------+---------------------------------------------------------------------------------------------------+
	
Esempi
------

Ignora tutti gli allarmi down piu' frequenti di 3 minuti: ::

	entables -A DEFAULT -m status  --status_curr DN --status_prev UP,FA --status_changein 180 -j DROP

"""

import time
import datetime

from sanet_common.sanetdefs.constants import *

from core.models import Tenant, Condition
from core.api.base import db_set_current_tenant_schema

from entables.constants import TABLE_NOTIFY

from entables.targets import Option
from entables.matches import BaseMatch


#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#


class Match(BaseMatch):
	
	def __init__(self):
		super(Match, self).__init__()
		
		self.name    = "status"
		
		self.tables  = [ TABLE_NOTIFY ]
		
		self.options = [
			Option("curr"     , type=str, required=False),
			Option("prev"     , type=str, required=False),
			Option("changedin", type=int),
		]

	def check_match_arguments(self, options_dict):
		
		super(BaseMatch, self).check_match_arguments(options_dict)
		
		curr      = options_dict.get('curr', '*not specified*')
		if curr not in CSTATE_CODENAMES: raise Exception("Invalid state code: %s" % (curr))
		
		prev      = options_dict.get('prev', '*not specified*')
		if prev not in CSTATE_CODENAMES: raise Exception("Invalid state code: %s" % (prev))
		
		changedin = options_dict['changedin']
	
		return True

	def match(self, context, alarmobj, options):

		NOT_DEFINED = ''

		log = context.log
		
		alarmobj.set_verbstatus("Match %s" % (self.name))

		alarm = alarmobj.data

		now = time.time()
		
		try:
			curr      = CSTATE_FROM_CODENAME(options.get('curr',None)) or NOT_DEFINED
			prev      = CSTATE_FROM_CODENAME(options.get('prev',None)) or NOT_DEFINED
			changedin = int(options['changedin'])

			cond_id           = alarm['condition']['uuid']
			cond_curr         = alarm['result']['status']
			cond_prev         = alarm['result']['laststatus']
			cond_lastchange   = alarm['condition']['statuslastchange']


			
			tenant_id = alarm['tenantid']
			tenant = Tenant.objects.get(uuid=tenant_id)
			db_set_current_tenant_schema(tenant.name)
			
			c = Condition.objects.get(tenant=tenant, uuid=cond_id)

			time_diff = now - cond_lastchange

			msg = "Matching %s(curr=%s,prev=%s,changedin=%s) with (curr=%s,prev=%s,lastchange=%s). Diff %s" % (
				self.name,
				curr,
				prev,
				changedin,
				cond_curr,      
				cond_prev,      
				cond_lastchange,
				time_diff
			)
			log.info(msg)
			
			alarmobj.set_verbstatus("\t%s" % (msg))

			if curr != NOT_DEFINED:
				if curr != cond_curr:
					return False
			
			if prev != NOT_DEFINED:
				if prev != cond_prev:
					return False
			
			
			
			if time_diff >= changedin:
				return False
				
				
			return True

		except KeyError, e:
			msg = "Unable to get info from alarm: %s" % (str(e))
			log.error(msg)

		except Exception, e:
			msg = "Unable to perform match: %s" % (str(e))
			log.error(msg)

		return False

