# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""
TAGS
====

Questo match permette di fare match sui tag associati all'elemento (nodo,interfaccia,ecc.) che ha generato l'allarme.

Gli attributi su cui fare match vengono specificati con la sintassi:  ::

	-m tags --tags_<attributo> <valore>

Gli attributi su cui e' possibile fare match sono:

	+-----------+--------------------------------------------------------------------------------------------------------+
	| Parametro | Descrizione                                                                                            |
	+===========+========================================================================================================+
	| path      | Regexpr per fare match sui tag associati all'allarme                                                   |
	+-----------+--------------------------------------------------------------------------------------------------------+
	
NOTA: Il path in formato stringa di un tag contiene anche il nome del tagtree. Esempio: ::

	geo:/Visione Globale
	geo:/Visione Globale/Pontecchio
	
	servizi:/Servizi IT/Web
	servizi:/Servizi IT/Posta

Esempi
------

Esempio di match su un tag preciso: ::

	entables -t notify -A DEFAULT -m tags --tags_path '^allarmi:/Allarmi/Tutti$' -j DROP SMTP


"""

import re
from sanet_common.sanetdefs.constants import *

from entables.constants import TABLE_NOTIFY

from entables.matches import BaseMatch
from entables.targets import Option

"""
Match on alarm attributes

{
	'condition': {
		'name': 'check',
		'path': 'localhost;;debug;check',
		'priority': -1,
		'uuid': '7b87d0fc87a340bba6a5e1d9cf03a76d'
	},
	'datasources': [],
	'element': {
		'name': 'localhost',
		'parent_id': None,
		'parent_name': None,
		'type': 'node',
		'uuid': 'c3af80abb3104a2cad031b90e30efbdb'
	},
	'event': {
		'uuid': '897277a37ff94a168249eef4e591dcc7'},
		'flow-information': [{'agent_id': '1ca186d87c04458f86eaa09fca57f08c'}],
		'mail_data': {
			'body': '',
			'headers': {'X-SANET-Priority': -1,
			'X-SANET-Target-Id': 'localhost;;debug;check',
			'X-SANET-Target-Status': 'UP'},
			'subject': '',
		'to': ''
	},
	'result': {
		'laststatus': 10, 
		'status': 'UP', 
		'verbastatus': ''
	},
	'snmp_values': [],
	'tags': [],
	'tenantid': '10a279a6ce5c4438a98f9c33c0b77cde',
	'tenantlongname': '',
	'tenantname': 'site',
	'uuid': 'a64462832a2f4911b701b258c5515a40'
}
"""

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class Match(BaseMatch):
	
	def __init__(self):
		super(Match, self).__init__()
		
		self.name    = "tags"
		
		self.tables  = [ 
			TABLE_NOTIFY 
		]
		
		self.options  = [
			Option("path"  , default="geo:/.*"),
		]

	def check_match_arguments(self, options):

		super(BaseMatch, self).check_match_arguments(options)
		
		path_expr = options.get('path', None)
		if path_expr:
			try:
				re.compile(path_expr) 
			except Exception, e:
				raise Exception("Invalid regexp for path: %s" % (path_expr))
		return True

	def match(self, context, alarmobj, options):
		
		log = context.log
		
		log.debug("TAG OPTIONS: %s", options)

		path_expr = options.get('path', None)
		if path_expr is None:
			return False

		paths = self.get_tags(alarmobj)
		log.debug("alarm tag paths '%s'", paths)
			
		for tag_path in self.get_tags(alarmobj):
			if re.match(path_expr, tag_path):
				msg = "Tag '%s' matched with '%s'" % (tag_path, path_expr)
				log.debug(msg)
				alarmobj.set_verbstatus(msg)
				return True
				
		return False
		
	def get_tags(self, alarmobj):

		return alarmobj.data.get('tags', [])
		

