# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

.. _entables-matches-times:

Times
==========

Questo match permette di fare match sull'orario di generazione dell'evento che ha prodotto l'allarme.

Gli attributi su cui fare match vengono specificati con la sintassi:  

	::
		
		-m times --times_<attributo> <valore>

Gli attributi su cui e' possibile fare match sono:

	+-----------+-------------+------------------+---------------------------------------------------------------------------------------------------------+
	| Parametro | Type        | Default          |  Descrizione                                                                                            |
	+===========+=============+==================+=========================================================================================================+
	| from      | stringa     |                  |  Data di inizio intervallo temporale di interesse (ISO)                                                 |
	+-----------+-------------+------------------+---------------------------------------------------------------------------------------------------------+
	| to        | numero      |                  |  Data di fine intervallo temporale di interesse (ISO)                                                   |
	+-----------+-------------+------------------+---------------------------------------------------------------------------------------------------------+
	| range     | stringa     |                  |  Stringa range temporali (ore del giorno/giorni della settimana) di interesse (es: 8-20/1-3)            |
	+-----------+-------------+------------------+---------------------------------------------------------------------------------------------------------+




Esempi
------

Matcha tutti i messaggi nella fascia oraria dalle 8:00 alle 18:00 dal lunedi' al venerdi': 

	::
		
		entables  -A DEFAULT  -m times --times_range  '8:00-18:00/1-5'  -j .....


Formato parametro range
-----------------------------------

La stringa per specificare il range temporale di interesse deve rispettare la sintassi per i range temporali usata: :ref:`times`.

Esempi di formati
~~~~~~~~~~~~~~~~~

	+----------------------------------------------------------+------------------------------------------------------------------------------------------------+
	| Stringa                                                  | Descrizione                                                                                    |
	+==========================================================+================================================================================================+
	| all                                                      | Tutti i giorni                                                                                 |
	+----------------------------------------------------------+------------------------------------------------------------------------------------------------+
	| none                                                     | Tutti i giorni                                                                                 |
	+----------------------------------------------------------+------------------------------------------------------------------------------------------------+
	| 7-20/1                                                   | Tutti i lunedi, tra le 7:00 e le 20:00                                                         |
	+----------------------------------------------------------+------------------------------------------------------------------------------------------------+
	| 8:00-13:00,14:00-18:00/1-5                               | Tutti i giorni da lunedi (1) a venerdi (5), tral le 8:00 e le 13:00  e tra le 14:00 e le 18:00.|
	+----------------------------------------------------------+------------------------------------------------------------------------------------------------+

Match inverso
-------------------------

E' possibile aggiungere il carattere *!* come prefisso al valore del *range* per indicare che si vuole fare match sul range temporale *inverso*.

.. warning:: l'utilizzo del carattere *!* produce l'inverso di **tutto** il range temporale specificato.


Esempi:

	::
		
		!all                    not (tutti i giorni) -> nessun giorno
		!none                   not (tutti i giorni) -> nessun giorno    
		!8:00-18:00/1-5         not (8:00-18:00/1-5) -> dal 1 al 5 dalle 00:00 alle 8:00 e dalle 18:00 alle 24 + tutto sabato e domenica

.. danger:: il carattere (*!*) e' un carattere speciale di alcune shell. Per assicurarsi che non dia errori di interpretazione specificare sempre tutto il range temporale tra apici singoli (*'*).
	::
		
		"!8:00-18:00/1-5"       NO
		                          
		'!8:00-18:00/1-5'       SI

"""


import time
import datetime
import types
import sys

from sanet_common.sanetdefs.timing import match_now

from core.timing import init_holidays_converter

from entables.constants import TABLE_NOTIFY

from entables.matches import BaseMatch
from entables.targets import Option

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def __isostring_to_date(string):

	try:
		return datetime.datetime.strptime(string,'%Y-%m-%d')
	except ValueError:
		pass

	try:
		return datetime.datetime.strptime(string,'%Y-%m-%d %H:%M')
	except ValueError:
		pass

	try:
		return datetime.datetime.strptime(string,'%Y-%m-%d %H:%M:%S')
	except ValueError:
		pass

	return None

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class Match(BaseMatch):

	def __init__(self):

		super(Match, self).__init__()

		self.name    = "times"

		self.tables  = [ TABLE_NOTIFY ]

		self.options = [
			Option("range", default=""),
			Option("from" , default=""),
			Option("to"   , default=""),
		]

	def get_time_range_and_context(self, time_range):

		parts = time_range.split(" ")
		if len(parts) == 1:
			time_range = parts[0].strip()
			context    = None
			return time_range, context
		elif len(parts) == 2:
			time_range = parts[0].strip()
			context    = parts[1].strip()
			return time_range, context
		else:
			raise ValueError("Invalid syntax for time range (<time range> <context>): '%s'." % (time_range) )


	def validate_match_options(self, options):

		super(Match, self).validate_match_options(options)

		now_test    = datetime.datetime.now()
		now_test_ts = time.mktime(now_test.timetuple())

		time_range = options.get('range', None)
		if time_range: # != None and != ""
			try:

				if time_range.startswith('!'):
					time_range = time_range[1:]

				time_range, time_context = self.get_time_range_and_context(time_range)

				# IMPORTANT: we are only checking sintax. We do not need to handler special holidays here.
				match_now(time_range, now_test_ts)
			except Exception, e:
				raise Exception("Invalid 'range' format: %s" % (e))

		from_d = None
		to_d   = None

		from_date = options.get('from', None)
		if from_date: # != None and != ""
			from_d = __isostring_to_date(from_date)
			if from_d == None:
				raise Exception("Invalid 'from' format")

		to_date = options.get('to', None)
		if to_date: 
			to_d = __isostring_to_date(to_date)
			if to_d == None:
				raise Exception("Invalid 'to' format")

		if from_d and to_d:
			if to_d < from_d:
				raise Exception("'from' date > 'to' date")

		return True    


	def match(self, context, alarmobj, options):

		log = context.log

		alarm = alarmobj.data

		now    = datetime.datetime.now()
		now_ts = time.mktime(now.timetuple())

		matched = True

		log.debug("Checking match %s", self.name)

		time_range = options.get('range', None)
		if time_range:
			log.debug("Checking 'range' option: %s", time_range)


			if time_range.startswith('!'):
				invert_result = True
				time_range = time_range[1:]
			else:
				invert_result = False

			holidays_converter = init_holidays_converter()          # FIX: init once when creating Match object

			time_range, time_context = self.get_time_range_and_context(time_range)

			r = match_now(time_range, now_ts, calendar_handler=holidays_converter, time_context=time_context, log=log)

			log.debug("range result: %s",r)

			if invert_result:
				r = not r

			log.debug("range result2: %s",r)

			matched = matched and r 

		from_date= options.get('from', None)
		if from_date:    
			log.debug("Checking 'from' option: %s", from_date)
			from_date = __isostring_to_date(options['from'])
			matched  = matched and (now >= from_date)

		to_date= options.get('to', None)
		if to_date:    
			log.debug("Checking 'to' option: %s", to_date)
			from_date = __isostring_to_date(to_date)
			matched  = matched and (now <= to_date)

		return matched


