# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import argparse

import daemon
import signal

import pprint
import logging
import lockfile
import threading
import sys
import os

from django.db import connection

from entables.constants import *
from entables.tablesstore.djangostore import RuleDB, AlarmQueue
from entables.modulesmanager import *
from entables.datastructures import *
from entables.server.threads import *

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def load_modules(log, configuration, context):

	listeners_modules = [ module_name for module_name, config in configuration['listeners'] ]
	load_listener_modules(log, listeners_modules, context.registered_listeners)

	load_match_modules(log, configuration['matches'], context.registered_matches)

	load_target_modules(log, configuration['targets'], context.registered_targets)    

	load_tagprovider_modules(log, configuration['tagproviders'], context.registered_tagproviders)

	return 

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def spawn_listeners(log, configuration, context):

	log.debug("Launching listeners...")

	config_dict = dict(configuration["listeners"])
    
	for listenername, listeneclass in context.registered_listeners.items():
        
		listerer_config = config_dict[listenername]

		log.info("Spawning listener '%s' with options '%s'", listenername, ','.join(['%s=%s' % (k,v) for k,v in  listerer_config.items()]) )
		    
		listenerthread = listeneclass(listerer_config, context)

		context.listenerthreads[listenername] = listenerthread
		context.listenerthreads[listenername].start()

	return 

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

stop = 0

def entablesd_mainloop(log, configuration, runoptions):
	''' this function call the manage queue
	'''

	global entablesd_context
	global stop

	log.debug("Creating context")
	entablesd_context = create_context(log, runoptions)

	log.debug("Loading modules")
	load_modules(log, configuration, entablesd_context)

	log.debug("Loading rules")
	load_rules(log, entablesd_context)

	#
	# Launch internal threads
	#
	log.debug("Launching routing thread")    
	entablesd_context.routingthread = RoutingThread(entablesd_context, entablesd_context.ruledb)
	entablesd_context.routingthread.start()    

	log.debug("Launching Holding thread")    
	entablesd_context.holdthread  = HoldingThread(entablesd_context)
	entablesd_context.holdthread.start()    

	log.debug("Launching Server thread")    
	entablesd_context.serverthread  = ServerThread(entablesd_context, dbrefresh_interval= configuration['refresh_interval'])
	entablesd_context.serverthread.start()    

	log.debug("Launching ctl thread")    
	entablesd_context.ctlthread = CtlThread(entablesd_context)
	entablesd_context.ctlthread.start()    

	spawn_listeners(log, configuration, entablesd_context)

	#
	# Wait for stop signal
	#
	def sighup_handler(signum, frame):
		global stop
		print "SIGNAL"
		stop = 1

	signal.signal(signal.SIGHUP , sighup_handler)
	while stop == 0:
		time.sleep(10)

	#
	# Stop threads
	#
	log.info("Stopping threads...")
	entablesd_context.ctlthread.stop()
	entablesd_context.routingthread.stop()
	entablesd_context.holdthread.stop()
	entablesd_context.serverthread.stop()
		
	log.info("Stopping waiting stopping threads...")
	entablesd_context.ctlthread.join()
	entablesd_context.routingthread.join()
	entablesd_context.holdthread.join()    
	entablesd_context.serverthread.join()

	log.info("entablesd EXIT")

	return


def load_rules(log, context):
	''' load rules into memory objects from the database
	'''

	log.debug("loading rules form db")
	context.ruledb.load_rules()

	#signal.signal(signal.SIGHUP, context.ruledb.reload_rules)


def create_context(log, arguments):
	''' create the context object
	'''

	log.debug("Init context")

	context = Context()

	# Rules database
	context.ruledb = RuleDB()

	# Main Queue
	# create object
	context.permaqueue = AlarmQueue()

	# load alarms into object from database
	context.permaqueue.load_alarms()


	# Logging
	context.log = log

	return context



def init_log(arguments):

	log = logging.getLogger("entablesd")


	if arguments.debug:
		formatter = logging.Formatter('%(asctime)s %(levelname)s entablesd(%(process)d)/%(threadName)s %(filename)s:%(funcName)s:%(lineno)d - %(message)s')
	else:
		formatter = logging.Formatter('%(asctime)s %(levelname)s entablesd(%(process)d)/%(threadName)s - %(message)s')


	level = arguments.loglevel
	if level != None:
		level = int(level)
	else:
		if arguments.debug: level = logging.DEBUG
		else              : level = logging.INFO
	log.setLevel( level )


	if arguments.foreground:
		hdlr = logging.StreamHandler()
	else:
		logfile_name = arguments.logfile or "/var/tmp/entablesd_log"

		# Update pid file
		dir_name = os.path.dirname(logfile_name)
		try:
			if not os.path.isdir(dir_name):
				print "Creating %s" % (dir_name)
				os.makedirs(dir_name)
		except OSError, e:
			print "Error while checking %s: %s" % (dir_name, e)

		#hdlr = logging.FileHandler(logfile_name)
		hdlr = logging.handlers.WatchedFileHandler(logfile_name)

	hdlr.setFormatter( formatter )
	hdlr.setLevel( level)

	log.addHandler(hdlr)

	log.info("INFO LEVEL: %s", level)

	return log
     
     
def write_pid_file(pidfile):

	if pidfile == None:
		raise Exception("Pid file not specified")
	
	f = open(pidfile, "w+")
	f.write("%s\n" % (os.getpid()))
	f.close()

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def entablesd_run(configuration, argv_options):
    
	log = init_log(argv_options)

	# Start event loop

	if argv_options.foreground == True:
		log.info("started in foreground")

		entablesd_mainloop(log, configuration, argv_options)
	else:    
		# Preserve all file descriptors (for now)

		if not argv_options.pidfile:
			raise Exception("Pid file not specified")

		if not argv_options.logfile:
			raise Exception("Log file not specified")

		
		# Update pid file
		pid_dir_name = os.path.dirname(argv_options.pidfile)
		try:
			if not os.path.isdir(pid_dir_name):
				log.info("Creating run directory: %s", pid_dir_name)
				os.makedirs(pid_dir_name)
		except OSError, e:
			log.error("Error while checking %s: %s", pid_dir_name, e)

		
        
		log.info("Entables starting in deamon mode")

		daemon_context = daemon.DaemonContext(
			files_preserve = [ fd for fd in range(1,13) ],
			pidfile        = lockfile.FileLock(argv_options.pidfile),
		)

		with daemon_context:

			log.info("Writing pid")
			
			pidfile = open(argv_options.pidfile, "w+")
			pidfile.write("%s\n" % (str(os.getpid())))
			pidfile.flush()
			pidfile.close()


			log.info("DAEMON STARTED")

			log.debug("Closing connections from db")
			connection.close()

			entablesd_mainloop(log, configuration, argv_options)


#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#


def entablesd_parse_commandline(argv):

	cmdline_parser = entablesd_create_cmdline_parser()
	
	parsed = cmdline_parser.parse_known_args(argv)

	return parsed[0]
	
def entablesd_create_cmdline_parser():

	main_parser = argparse.ArgumentParser(description='Manage EventNotify Tables')
	main_parser.add_argument("-d","--debug"     , action="store_true", help='Enable debug mode') 
	main_parser.add_argument("-f","--foreground", action="store_true", help='Launch program in foreground. Do not daemonize the server.')
	main_parser.add_argument("-l","--logfile"   , default=None       , help='Log file')
	main_parser.add_argument("-L","--loglevel"  , action="store"     , default=None, help='Log level')
	main_parser.add_argument("-p","--pidfile"   , action="store"     , help='Pid File')
	
	return main_parser

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def _entabled_hack_keyboard_interrupt():
    
	# Fork. The parent process waits for a Ctrl+C. The child process 
	# performs the business logic.
	childpid = os.fork()

	if childpid <> 0:
		# MAIN PROCESS
		try:
			# wait for child termination
			os.wait()
		except KeyboardInterrupt:
			# I put the capital B in KeyBoardInterrupt so I can
			# tell when the Watcher gets the SIGINT
			#log.debug('POLLER: KeyBoardInterrupt! killing child process.')
			print "Ctrl+C"
			try:
				os.killpg(os.getpgrp(), signal.SIGKILL)
			except OSError: 
				pass

		sys.exit()
	else:
		# CHILD PROCESS: ignore SIGINT
		signal.signal(signal.SIGINT, signal.SIG_IGN)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def entablesd_main(configuration, argv):

	parsed_argv = entablesd_parse_commandline(argv)

	# HACK to kill everything with Ctrl+C when running in foreground
	if parsed_argv.foreground:
		_entabled_hack_keyboard_interrupt()

	entablesd_run(configuration, parsed_argv)
    
    

   
