# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""
                   |                                                                          
                   |                                                                          
                   |                         +--------------+                                 
                   |                         |              |.................                                 
                   |                         |              |                .                    
       [ ]------------------------------------->   ctl      |                .                    
  entables-ctl     |                         |    thread    |                .                    
                   |                         |      |       |                .                    
                   |                         |      |       |          +-----------+        
       [ ]<-----------------------------------------+       |          |  CONTEXT  |        
  reply queue      |                         |              |          +-----------+        
                   |                         +--------------+             . . . .                   
                   |            ........................................... . . .                                                            
                   |            .                         ................... . ...................                                   
                   |            .                         .                   .                   .
                   |            .                         .                   .                   .
                   |   +----------+          +--------------+            +-------------+        +------------------------------+
                   |   |          |          |              |            |             |        |                              |
                   |   | listener |          |    server    |            |   routing   |        |           hold               |
     REDIS         |   |  thread  |          |    thread    |            |   thread    |        |           thread             |
                   |   |          |          |              |            |             |        |                              |
                   |   |          |          |   +------+   |            |   +-----+   |        |   +-----+       +-------+    |
      [ ]     +------------------------------->   A A A  -+---------------->  A A A -------------->  A A A ----->  A       -------------+--+
entables-alarms    |   |          |      +-->|   +------+ | |       +--> |   +-----+   |        |   +-----+       +-------+    |        |  |
                   |   |          |      |   |    queue   | |       |    |   queue     |        |  scheduler       queue       |        |  |
                   |   |          |      |   |            | |       |    |             |        |                              |        |  |
                   |   +----------+      |   +------------|-+       |    +-------------+        +------------------------------+        |  |
                   |                     |                |         |                                                                   |  |
                   |                     |                |         +-------------------------------------------------------------------+  |
                   |                     +----------------|--------------------------------------------------------------------------------+
                   |                                      |                                  
                   |                                      |                                  
                   |                                      +-------------> ARCHIVE (DB)       
                   |                                               
                   |                    

"""

import os
import time
import threading
import Queue
import copy
import traceback

from django.db import connection

USE_REDIS_IPC = os.environ.get('SANET_USE_REDISIPC', None)
if USE_REDIS_IPC:
	from sanet_common.ipc.redisipc import RedisIPCConnection as IPCConnection
else:
	from sanet_common.ipc          import IPCConnection
	
from sanet_common.ipc import IPCConnectionError, IPCException
from sanet_common.ipc import encode_request, decode_response
from sanet_common.ipc import decode_request, encode_response_with_id, RESPONSE_ERROR, RESPONSE_SUCCESS
from sanet_common.timeutils import timestamp_to_isostr

from entables.constants import *
from entables.datastructures import *

from entables.tablesstore.djangostore import RuleDB

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def entablesd_ctl(broker, command, arguments):
	
	from sanet_common.ipc import encode_request, decode_response
		
	request = encode_request(command, *arguments)

	if broker is None:
		broker = "sanet/sanet@localhost"
	
	host = None
	port = None
	user = None
	password = None
	
	auth, addr = broker.split('@')
	user,password = auth.split('/')
	addr = addr.split(':')
	host = addr[0]
	if len(addr) > 1:
		port = int(addr[1])
	
	connect_options = {}
	if port     != None: connect_options['port']     = port
	if user     != None: connect_options['username'] = user
	if password != None: connect_options['password'] = password
	
	c = IPCConnection( host, **connect_options)
	
	c.settimeout(3)
	
	c.open()
	
	c.bind("temp-sanetdctl-pid-%s" % (os.getpid()), persistent=False)

	c.send_msg( 'entables-ctl' , request, blocking=True)
	
	addr, data = c.receive_msg(ack_msg=True)
	if data is None:
		raise Exception("Timeout")
	c.close()
	response = decode_response(data)
	
	if response['status'] == RESPONSE_SUCCESS:
		return response['content']
	else:
		raise Exception("Error from remote host: %s" % (response['content']))

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class CtlThread(threading.Thread):
		
	DEFAULT_BROKER_SETTINGS = {
		"address"        : ('127.0.0.1', None),
		"username"       : 'sanet',
		"password"       : 'sanet',
	}
	
	STOPSIGNAL    = "STOP"
	
	def __init__(self, context, broker_settings=None):
		
		super(CtlThread, self).__init__(name="server")
		
		if (hasattr(self, "daemon")):
			self.daemon = True
		else:
			self.setDaemon(True)

		self.context = context
		self.logger  = context.log

		
		self.broker_settings = dict(self.DEFAULT_BROKER_SETTINGS) 
		if broker_settings:
			self.log.info("Using settings: %s", broker_settings)
			self.broker_settings.update(broker_settings)

		self.start_time = 0
		self._stop      = 0


	def stop(self):
		self._stop = 1
		self.signal_stop()

	def run(self):
		
		self.start_time = time.time()
		
		self.ctl_loop()
		
	def ctl_loop(self):
	
		self.logger.info("[ctl] Start...")

		while not self._stop:

			try:
				self.logger.info("[ctl] Creating connection")

				self.queue_ctl = self.create_connection()
				self.queue_ctl.open()
				self.queue_ctl.bind('entables-ctl', persistent=False)

				while not self._stop:
					self.receive_command()

				self.queue_ctl.close()

			except IPCConnectionError, e:
				self.logger.error("[ctl] Connection lost")
				
			finally:
				try:
					if self.queue_ctl:
						self.queue_ctl.close()
				except Exception, e:
					pass

		self.logger.info("[ctl] Exit")
		
	def create_connection(self):

		connect_options = {}
		
		host, port = self.broker_settings['address']

		if port is not None:
			connect_options['port'] = port
			
		connect_options['username'] = self.broker_settings['username']
		connect_options['password'] = self.broker_settings['password']

		self.logger.info("Opening connection with server %s" % (host) )
		self.logger.debug("Options:")
		for k,v in connect_options.items():
			self.logger.debug("\t%s = %s", k, v)

		c = IPCConnection(host, **connect_options)

		return c
		
	def receive_command(self):

		try:
			(addr, msg) = self.queue_ctl.receive_msg(timeout=1, ack_msg=True)
			if msg == None:
				return

			request = decode_request(msg)
			
			request_id = request['id']
			
			self.logger.log(9,"[ctl] Received ctl from %s", addr)
			self.logger.log(9,"[ctl]   id     = %s" % (request_id,))
			self.logger.log(9,"[ctl]   cmd    = %s" % (request['type'],))
			self.logger.log(9,"[ctl]   params = %s" % (request['params'],))


			h_name = 'cmd_' + request['type']
			if not hasattr(self, h_name):
				self.logger.warning("[ctl] Received unknown command: %s, %s", request['type'], request['params'])
				response = encode_response_with_id( RESPONSE_ERROR, request_id, "unknown command")
			else:
				handler = getattr(self, h_name) 
				try:
					response = handler(addr, request_id, request['params'])
				except IPCException, e:
					raise e
				except Exception, e:
					msg = str(e) + traceback.format_exc()
					self.logger.error("[ctl] error while processing request %s: %s", request['type'], msg)
					response = encode_response_with_id( RESPONSE_ERROR, request_id, msg )

			if response and addr != None:
				try:
					self.queue_ctl.send_msg( addr, response, blocking=False)
				except IPCConnection, e:
					raise e
				except Exception, e:
					self.logger.warning("Error while sending response to %s:%s" % (addr, str(e)))
					
		except IPCConnectionError, e:
			self.logger.error("[ctl] Error on connection: %s" % (str(e)))
			raise e

		except IPCException, e:
			self.logger.error("[ctl] Error on connection: %s" % (str(e)))
			raise e
					
		except Exception, e:
			self.logger.error("[ctl] Error while processing message: %s (%s - %s)", e, type(e), dir(e))
			self.logger.error("TYPE %s", type(e))
			self.logger.error("DATA %s", dir(e))

	def uptime(self):
		seconds = int( time.time() - self.start_time)
		return "%s m %s s" % ((seconds / 60), (seconds % 60))
		


	def cmd_help(self, addr, request_id, params):
		"""
		List available commands
		"""
		
		response = []
		for k in dir(self):
			if k.startswith('cmd_'):
				doc = getattr(self, k).__doc__
				if doc is None:
					doc = ''
				else:
					doc = doc.strip()
				response.append(
					"%s - %s" % (k[4:], doc)
				)
				
		response = '\n'.join(response)

		response = encode_response_with_id( RESPONSE_SUCCESS, request_id, response )

		return response			

		
		

	def cmd_info(self, addr, request_id, params):
		"""
		Get internal status data
		"""

		self.logger.info("[ctl] Info received")

		rows = []

		rows.append( "Thread server : %s" % (self.context.serverthread.isAlive()  ) )
		rows.append( "Thread routing: %s" % (self.context.routingthread.isAlive() ) )
		rows.append( "Thread hold   : %s" % (self.context.holdthread.isAlive()    ) )
		
		rows += [ "Uptime: %s" % (self.uptime()) ]
		rows += [ "Configuration" ]
		rows += [ "\tlast reload: %s" % (time.ctime(self.context.ruledb.load_time)) ]
		
		rows += [ "Server info" ]
		rows += [ "\t%s" % (s) for s in self.context.serverthread.get_infos() ]
				
		rows += [ "Routing info" ]
		rows += [ "\t%s" % (s) for s in self.context.routingthread.get_infos()]
		
		rows += [ "Holding info" ]
		rows += [ "\t%s" % (s) for s in self.context.holdthread.get_infos() ]
		
		rows += [ "Target counters" ]
		for name, target_handler in self.context.registered_targets.items():
			info = target_handler.get_text_info()
			if info is not None:
				rows += [ "\t%20s %s" % (name, info) ]

		response = '\n'.join(rows)
		response += '\n'
		
		response = encode_response_with_id( RESPONSE_SUCCESS, request_id, response )

		return response

	def cmd_set_log_level(self, addr, request_id, params):
		"""
		Set log level
		"""
		
		level = int(params[0])

		self.logger.info("[ctl] Setting log level to %s", level)

		self.logger.setLevel(level)
		for handler in self.logger.handlers:
			handler.setLevel( level )
		
		response = encode_response_with_id( RESPONSE_SUCCESS, request_id, "ok" )
		
		return response


	def cmd_hold_queue(self, addr, request_id, params):
		"""
		Get hold queue elements
		"""
		
		
		if len(params) > 1:
			num = int(params[0])
		else:
			num = 30

		self.logger.info("[ctl] hold_queue %s", params)

		data = self.context.holdthread.get_hold_queue_info(num)
			
		data = '\n'.join(data)
		
		response = encode_response_with_id( RESPONSE_SUCCESS, request_id, data )
		
		return response


#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class ServerThread(threading.Thread):
	
	STOPSIGNAL    = "STOP"
	NEW_ALARM     = "NEW_ALARM"
	ARCHIVE_ALARM = "ARCHIVE_ALARM"
	
	def __init__(self, context, dbrefresh_interval=30):

		super(ServerThread, self).__init__(name="server")

		if (hasattr(self, "daemon")):
			self.daemon = True
		else:
			self.setDaemon(True)

		self.context = context
		self.log     = context.log
		
		self.dbrefresh_interval = dbrefresh_interval

		self.__msg_queue = Queue.Queue(-1)

		self.status = 'init'
		self._stop = 0
		
		self.current_configuration_time    = 0   # timestamp dell'ultima volta che ho ricaricato la configurazione
		self.current_configuration_elapsed = -1  # durata del refresh dell'ultima configurazione
		
		self.counter_received_alarms = 0
		self.counter_rescheduled_alarms = 0
		
	def stop(self):
		self._stop = 1
		self.signal_stop()

	def run(self):
		# Override thread method
		self.main_loop()

	#--------------------------------------------------------------#

	# Metodi chiamati dall'esterno per segnalare al thread cosa fare
	def signal_stop(self):
		self.__notify_event(self.STOPSIGNAL)

	def start_process_new_alarm(self, alarm_uuid):
		self.__notify_event( {'event_type': self.NEW_ALARM    , 'uuid' : alarm_uuid} )
		
	def archive_alarm(self, alarm):
		self.__notify_event( {'event_type': self.ARCHIVE_ALARM, 'alarm': alarm     } )

	def __notify_event(self, msg):
		self.__msg_queue.put_nowait(msg)

	#--------------------------------------------------------------#

	def reload_pending_alarms(self):

		log     = self.log
		context = self.context

		alarmlist = context.permaqueue.getall()

		log.info("Reprise alarm queeue: num alarms %s", len(alarmlist) )
	    
		for alarm in alarmlist:
			    
			log.info("Reloaded alarm %s, status: %s", alarm.uuid, ALARM_STATUS_STR(alarm.status) )
		    
			if alarm.status == ALARM_STATUS_ROUTING:
				log.info("Re-routing alarm %s", alarm.uuid)
				context.routingthread.route_new_alarm( alarm.uuid )
				
			elif alarm.status == ALARM_STATUS_HOLD:
				log.info("Re-holding alarm %s: wakeup time: %s %s", alarm.uuid, alarm.hold_wakeuptime, timestamp_to_isostr(alarm.hold_wakeuptime))
				context.holdthread.schedule_new_held_alarm(alarm.hold_wakeuptime, alarm.uuid)
				
			else:
				log.info("Archiving alarm %s, status %s", alarm.uuid, ALARM_STATUS_STR(alarm.status))
				context.permaqueue.archive(alarm.tenant_uuid, alarm.uuid)
		

	def refresh_rules(self):

		log     = self.log
		context = self.context
		
		try:
			log.info("Reloading rules DB...")
			
			t1 = time.time()
			
			new_db = RuleDB()
			new_db.load_rules()
			
			context.ruledb = new_db
			
			t2 = time.time()
			elapsed = t2 - t1
			
			self.current_configuration_time = t2
			self.current_configuration_elapsed = elapsed

			log.info("DB reloaded in: %s s", elapsed)
			
		except Exception, e:
			log.critical("Error while reloading rules from db: %s:%s", type(e), str(e))

	def main_loop(self):
		
		self.status = 'started'

		log     = self.log
		context = self.context

		self.status = 'reloading pending alarms'

		self.reload_pending_alarms()

		self.log.info( "SERVER start")
		
		while not self._stop:

			event = None

			self.status = 'waiting'
			
			try:
				log.log(1, "MAIN THREAD Waiting for events")
				event = self.__msg_queue.get(True, self.dbrefresh_interval)
			except Queue.Empty, er:
				pass


			if event == None: # refresh interval elapsed
				self.status = 'loading configuration'
				self.refresh_rules()
				continue
		
			if event == self.STOPSIGNAL:
				break

			log.debug("Processing event: %s", event)

			if event['event_type'] == self.NEW_ALARM:

				self.status = 'routing new alarm'
				
				uuid = event['uuid']

				log.info("Launching routing for alarm %s", uuid)
				context.permaqueue.change_status(uuid, ALARM_STATUS_ROUTING)

				context.routingthread.route_new_alarm( uuid )
				
				self.counter_received_alarms += 1
			
			elif event['event_type'] == self.ARCHIVE_ALARM:

				self.status = 'archiving alarm'

				alarm = event['alarm']
				
				log.info("Archiving request for alarm %s", alarm.uuid)
				context.permaqueue.archive(alarm.tenant_uuid, alarm.uuid)
				
				self.counter_rescheduled_alarms += 1
			else:
				log.critical("UNKNOWN event type: %s", event)

		self.log.info( "SERVER exit")

		self.status = 'exiting'

		return

	def get_infos(self):
		
		r = []
		
		
		r += [ "Status                      : %s" % (self.status) ]
		r += [ "Queued alarms from listeners: %s" % (self.__msg_queue.qsize()) ]
		r += [ "Total received    alarms    : %s" % (self.counter_received_alarms) ]
		r += [ "Total rescheduled alarms    : %s" % (self.counter_rescheduled_alarms) ]

		r += [ "Configuration: Last update: %s. Took: %s " % (self.current_configuration_time, self.current_configuration_elapsed) ]
		
		return r
		


#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def ERASE_STACK(s):
	while s:
		s.pop()


class RoutingThread(threading.Thread):
	
	STOPSIGNAL = "stop"

	def __init__(self,context,ruledb):

		super(RoutingThread, self).__init__(name="alarmrouter")

		if (hasattr(self, "daemon")):
			self.daemon = True
		else:
			self.setDaemon(True)

		self.ruledb = ruledb
		self.context = context
		
		self.log = context.log

		# Inter Threads communication
		self._msgqueue = Queue.Queue(-1)

		self.active_frame = {}
		
		self._stop = 0
		
		self.counter_routed_alarms = 0
		
	def stop(self):
		self._stop = 1
		self._msgqueue.put(self.STOPSIGNAL)


	def run(self):
		# Override thread method
		self.route_loop()


	def route_new_alarm(self, alarmid):#, routingstagenum):
	
		self.log.debug("Request for routing alarm: %s", alarmid)
		msg = {
			'alarmid'		 : alarmid,
		}
		self._msgqueue.put(msg)
		return


	def route_loop(self):
		
		self.log.info( "ROUTE started")

		while not self._stop:

			message = None
			try:
				self.log.log(1, "ROUTE receiving")
				message = self._msgqueue.get(True, 1)
			except Queue.Empty, e:
				# sleep ?
				continue
		
				
			if message == self.STOPSIGNAL:
				continue


			try:
				alarm_uuid = message['alarmid']

				self.log.info( "Routing alarm: %s", str(alarm_uuid))

				self.process_single_alarm(alarm_uuid)
				
				self.counter_routed_alarms += 1
			except Exception, e:
				self.log.error("Unknown error while routing: %s:%s" % (type(e), str(e)))
		
		self.log.info( "ROUTE exit")
		
		return

	def process_single_alarm(self, auuid):
		
		log = self.log

		try:
			alarm = self.context.permaqueue.get(auuid)
		except Exception, e:
			msg = "Unknown error: %s:%s" % (type(e), str(e))
			self.log.error("Error while processing alarm %s: %s", auuid, msg)
			return

		try:
		#if 1:
			self.log.debug( "Alarm rulestack: %s", alarm.rulestack)
			
			# Questo alarme non ha rulestack quindi dovrebbe essere nuovo
			if not alarm.has_rulestack():
				
				alarm.set_verbstatus("Routing started")
				
				self.log.debug( "requested routing for alarm")
				
				rulestack = self.create_initial_rulestack_for_alarm(alarm.tenant_uuid)
				alarm.set_rulestack(rulestack)
			else:
				alarm.set_verbstatus("Re-routing")
				
					
			self.route(alarm)
			
		except Exception, e:
			msg = "Unknown error: %s:%s" % (type(e), str(e))
			self.log.error("Error while processing alarm %s: %s", auuid, msg)
			
			#trace = traceback.format_stack()
			#self.log.error(trace)
			
			s = traceback.format_exc()
			print s
			self.log.error(s)

		return


	def route(self, alarm):


		outcome	= ROUTING_OUTCOME_DROP

		rulestack = alarm.get_rulestack()


		while True:
			
			if len(rulestack) == 0: # completed?
				self.log.debug("Alarm %s, rulestack empty. Exit routing loop!", alarm.uuid)
				break

			rulestackelement = rulestack.pop()

			self.log.debug("Alarm %s, current rule: %s", alarm.uuid, rulestackelement)

			#
			active_frame = self.set_active_frame(alarm, rulestackelement)
			
			outcome = self.examine_active_frame(alarm, active_frame)

			self.log.debug("Alarm %s  current rule outcome is: %s", alarm.uuid, ROUTING_OUTCOME_STR[outcome])

			if outcome == ROUTING_OUTCOME_CONTINUE:
				pass
			
			elif outcome == ROUTING_OUTCOME_ERROR:
				# Fine routing. Per sicurezza cancelliamo il suo rulestack.
				ERASE_STACK(rulestack)
				break

			elif outcome == ROUTING_OUTCOME_DROP:
				# Fine routing. Per sicurezza cancelliamo il suo rulestack.
				ERASE_STACK(rulestack)
				break
				
				
			elif outcome in [ ROUTING_OUTCOME_HOLD, ROUTING_OUTCOME_HOLDNEXT ]:
				# FIX: La manipolazione del rulestack sarebbe meglio non farla in questa funzione ma 
				#      in call_target. Per ora funziona.
				#rulestack.append(rulestackelement)
				break

			elif outcome == ROUTING_OUTCOME_MERGED:
				# Fine routing. Per sicurezza cancelliamo il suo rulestack.
				ERASE_STACK(rulestack)
				break



			# end default chain == end table, see if we can move to the next
			elif outcome == ROUTING_OUTCOME_RETURN:
				# nothing to be done, RUETURN means nothing was appended to stack
				pass

			elif outcome == ROUTING_OUTCOME_ENDTABLE:
				
				table_name = active_frame['tableobj'].name

				self.log.debug("ENDTABLE on: %s", table_name)

				self.log.debug("Visiting rulestack (len %s)", len(rulestack))
				while len(rulestack) > 0:
					frame = rulestack[-1]
					self.log.debug("\tstack entry: %s", frame)
					if table_name not in frame:
						break
					rulestack.pop()
					self.log.debug("\t\tpopped")

				self.log.debug("Final rulestack (len %s)", len(rulestack) )
				for f in rulestack:
					self.log.debug("\t%s", f)
					
			#
			# FIX:
			#

			# SAVE CURRENT RULESTACK ON DB

		self.log.debug("Routing for alarm: %s ended with: %s , rulestack: %s", alarm.uuid, ROUTING_OUTCOME_STR[outcome], rulestack)

		ready_for_archive = False
		
		# routing completed?
		if outcome in [ ROUTING_OUTCOME_CONTINUE , ROUTING_OUTCOME_ENDTABLE ]:
			if len(rulestack) == 0:
				
				msg = "Alarm %s ended routing  successfully." % (alarm.uuid)
				alarm.set_verbstatus(msg)

				self.log.info(msg)
				self.context.permaqueue.change_status(alarm.uuid, ALARM_STATUS_COMPLETED)
			else:
				msg = "Alarm %s STRANGE ROUTING END. STACKRULE NOT EMPTY: %s" % (alarm.uuid, rulestack)
				alarm.set_verbstatus(msg)
				
				self.log.info(msg)
				self.context.permaqueue.change_status(alarm.uuid, ALARM_STATUS_ERROR)

			ready_for_archive = True

		elif outcome == ROUTING_OUTCOME_DROP:
			msg = "Alarm ended routing with DROP! Archiving"
			alarm.set_verbstatus(msg)			
			
			self.log.info(msg)
			self.context.permaqueue.change_status(alarm.uuid, ALARM_STATUS_DROPPED)
			ready_for_archive = True
			
		elif outcome == ROUTING_OUTCOME_ERROR:
			msg = "Alarm ended routing with error! Archiving"
			alarm.set_verbstatus(msg)			

			self.log.info(msg)
			self.context.permaqueue.change_status(alarm.uuid, ALARM_STATUS_ERROR)
			ready_for_archive = True
			
		elif (outcome in [ROUTING_OUTCOME_HOLD, ROUTING_OUTCOME_HOLDNEXT]):
			msg = "Alarm need hold (until %s)" % (alarm.hold_wakeuptime)
			alarm.set_verbstatus(msg)			

			self.context.permaqueue.change_status(alarm.uuid, ALARM_STATUS_HOLD)

			self.log.info(msg)

			hold_wakeup_time = alarm.hold_wakeuptime
			self.context.holdthread.schedule_new_held_alarm(hold_wakeup_time, alarm.uuid)

		elif outcome == ROUTING_OUTCOME_MERGED:
			msg = "Alarm %s merged." % (alarm.uuid)
			alarm.set_verbstatus(msg)

			self.log.info(msg)
			
			self.context.permaqueue.change_status(alarm.uuid, ALARM_STATUS_MERGED)
			ready_for_archive = True
			
		else:
			msg = "Alarm %s ended routing with invalid outcome %s! Archiving with error" % ( alarm.uuid, ROUTING_OUTCOME_STR[outcome] )
			alarm.set_verbstatus(msg)
			
			self.log.info(msg)

			self.context.permaqueue.change_status(alarm.uuid, ALARM_STATUS_ERROR)
			ready_for_archive = True

		if ready_for_archive:
			msg = "Alarm %s now ready for archiving" % (alarm.uuid)
			alarm.set_verbstatus(msg)
			
			self.log.info(msg)
			self.context.serverthread.archive_alarm(alarm)
			
			for l in alarm.get_verbstatus():
				self.log.debug("   -> %s", l)

		return


	def examine_active_frame(self, alarm, active_frame):

		self.log.debug("**************************************** STEP ") 

		# Examine rule if present
		if active_frame['ruleobj'] is not None:

			rule_outcome = self.examine_active_frame_rule(alarm, active_frame)			
			self.log.debug("examine_active_frame_rule() outcome: %s", ROUTING_OUTCOME_STR[rule_outcome])
			return rule_outcome

		# No more rules, alarm hit bottom of the chain, call policy target
		elif active_frame['policytargetobj'] is not None:

			ruleobj   = None
			targetobj = active_frame['chainobj'].policy
			self.log.debug("alarm hit policy")
			target_outcome = self.call_target(alarm, ruleobj, targetobj, active_frame, is_policy=True)
			return target_outcome

		# should never reach this point
		raise Exception()

	def examine_active_frame_rule(self, alarm, active_frame):

		rulestack = alarm.get_rulestack()
		
		ruleobj = active_frame['ruleobj']
		
		if ruleobj.disabled != RULE_STATE_ACTIVE:
			self.log.debug("Current rule disabled. Continue.")

			next_frame = [ active_frame['tenantobj'].uuid, active_frame['tableobj'].name, active_frame['chainobj'].name, active_frame['rulenumber'] + 1 ]
			rulestack.append(next_frame)

			return ROUTING_OUTCOME_CONTINUE

		self.log.debug("Checking matches......")

		#
		# Check matches
		#
		if not self.matches_all(alarm, ruleobj.matchset):

			self.log.debug("Rule does not match. Continue.")

			next_frame = [ active_frame['tenantobj'].uuid, active_frame['tableobj'].name, active_frame['chainobj'].name, active_frame['rulenumber'] + 1 ]
			rulestack.append(next_frame)

			return ROUTING_OUTCOME_CONTINUE

		# 
		# Call Target
		#
		self.log.debug("Executing target......")

		targetobj = ruleobj.target

		outcome = self.call_target(alarm, ruleobj, targetobj, active_frame)

		if outcome == ROUTING_OUTCOME_CONTINUE:
			if ruleobj.stop_if_match:
				alarm.set_verbstatus('stop if match = True on last rule')
				outcome = ROUTING_OUTCOME_ENDTABLE

		return outcome


	def matches_all(self, alarm, matchset):

		for matchobj in matchset:

			self.log.debug ("match check: "+ matchobj.name)
			
			handler = self.context.registered_matches.get(matchobj.name, None)
			
			if handler is None:
				self.log.warning("match module '%s' used inside rule but not loaded!!", matchobj.name)
				return False

			options = matchobj.options
			self.log.debug("match options" + str(options))

			# FIXME: if engine is made threaded (for tenants) match must handle concurrency on permanent data
			if not handler.match(self.context, alarm, options):
				self.log.debug(matchobj.name + " has not matched")
				return False
			else:
				self.log.debug(matchobj.name+" has matched")

		self.log.debug ("match check: all matched")
		
		return True


	def call_target(self, alarm, ruleobj, targetobj, active_frame, is_policy=False):

		rulestack = alarm.get_rulestack()

		tenant_config = active_frame['tenantobj']

		table_name    = active_frame['tableobj'].name
		active_frame_table = tenant_config.tables[table_name]

		target_name = targetobj.name
		target_name = target_name.lower()
		self.log.debug("target is: %s", target_name)

		# Target is a jump to a chain
		if target_name in active_frame_table.chains.keys():

			alarm.set_verbstatus("Jumping to chain: %s"  % (target_name))

			# Quando facciamo un salto ad una catena dobbiamo ritornare sulla regola successiva a quella
			# che ha generato il salto, altrimenti entriamo in loop
			future_position_on_chain_return = [active_frame['tenantobj'].uuid, active_frame['tableobj'].name, active_frame['chainobj'].name, active_frame['rulenumber'] + 1]
			rulestack.append(future_position_on_chain_return)			

			# Saltiamo alla prima regola della catena
			chain_name = target_name
			rulestack.append([active_frame['tenantobj'].uuid, active_frame_table.name , chain_name, RULESTACK_POS_FIRSTRULE])

			self.log.debug("**************** JUMP ***************")
			for r in rulestack:
				self.log.debug("\t\t %s", r)

			return ROUTING_OUTCOME_CONTINUE

		# Target is a plugin module
		elif target_name in self.context.registered_targets.keys():

			rule_outcome = self.perform_dispatch(alarm, ruleobj, targetobj, active_frame)

			# Dopo aver eseguito il target con  CONTINUE o HOLDNEXT dobbiamo settare il rulestack
			# per partire dalla regola successiva a quella appena eseguita
			if (rule_outcome == ROUTING_OUTCOME_CONTINUE and not is_policy) or \
			   (rule_outcome == ROUTING_OUTCOME_HOLDNEXT):

				next_frame = [ active_frame['tenantobj'].uuid, active_frame['tableobj'].name, active_frame['chainobj'].name, active_frame['rulenumber'] + 1 ]
				rulestack.append(next_frame)

			return rule_outcome
		else: 
			# Target is not known, this shouldn't happen, do nothing but log
			self.log.error ("target unknown or not loaded: " + target_name)

			# TODO: Cosa facciamo in caso di errore ? cosa ritorno ?

		return ROUTING_OUTCOME_ERROR


	def perform_dispatch(self, alarm, ruleobj, targetobj, active_frame):

		self.log.debug("*** CALL TARGET: %s", targetobj.name) 

		target_handler = self.context.registered_targets[targetobj.name]
		
		#
		# CALL
		#
		outcome = target_handler.dispatch(self.context, alarm, ruleobj, targetobj.options)

		alarm.set_verbstatus('\t%s returned %s' % (target_handler.get_name(), ROUTING_OUTCOME_STR[outcome]))
		
		return outcome


	def set_active_frame(self, alarm, rulestackelement):
		''' 
		populate active frame dict with pointers to RUleDB from rulestack elements
		'''
		tenant_id, tablename, chainname, rulenumber = rulestackelement
		
		active_frame = self.active_frame
		
		# FIX: gestire casi bastardi:
		#      - il tenant viene cancellato    -> Il messaggio va cestinato perche' non ho dove salvarlo -> LOG CRITICAL + INVIO PER EMAIL
		#      - se la chain  viene cancellata -> Possibile se il messaggio
		#
		#
		
		#tenantobj = self.ruledb.tenants[tenant_id]
		tenantobj = self.context.ruledb.tenants[tenant_id]
		tableobj  = tenantobj.tables[tablename]
		chainobj  = tableobj.chains[chainname]

		active_frame['rulenumber'] = rulenumber
		active_frame['tenantobj'] = tenantobj
		active_frame['tableobj'] = tableobj
		active_frame['chainobj'] = chainobj
		
		if rulenumber == RULESTACK_POS_POLICY:  # Policy di default
			
			targetobj = chainobj.policy
			active_frame['policytargetobj'] = targetobj
			active_frame['ruleobj'] = None
			
		elif rulenumber <= len(chainobj.rules): # REGOLA
		
			ruleobj = chainobj.rules[rulenumber-1]
			active_frame['ruleobj'] = ruleobj
			active_frame['policytargetobj'] = None
			
		else:
			# Se il rulenumber punta ad una posizione non valida ci sono due casi:
			# - Stiamo tornando da un salto e ritorniamo in fondo alla catena che avevamo lasciato
			# - A causa di un aggiornamento il numero di regole presenti sulla catena
			#   e' cambiato.
			# In entrambi i casi passiamo alla policy di default
			rulenumber = RULESTACK_POS_POLICY

			targetobj = chainobj.policy
			active_frame['policytargetobj'] = targetobj
			active_frame['ruleobj'] = None
			
		alarm.set_verbstatus("Rule posisiont: %s:%s:%s:%s" % (tenant_id, tablename, chainname, rulenumber) )

		return active_frame


	def create_initial_rulestack_for_alarm(self, tenant_uuid):

		stack = []

		#tenant = self.ruledb.get_tenant(tenant_uuid)
		tenant = self.context.ruledb.get_tenant(tenant_uuid)

		tables = tenant.tables

		for tablename, tableconfig in tables.items():

			chain_config = tableconfig.chains[DEFAULT_CHAIN]

			stack.insert(0, [tenant.uuid, tablename, DEFAULT_CHAIN, RULESTACK_POS_FIRSTRULE] )

		return stack

	def get_infos(self):

		r = []
		r += [ "Routed alarms: %s" % (self.counter_routed_alarms) ]

		return r

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

from sanet_poller.lib.threadutils import PollScheduler

class HoldingScheduler(PollScheduler):
	"""
	Struttura dati che mantiene in memoria lo scheduling degli allarmi in stato
	HOLD.
	
	Per ora usiamo lo scheduler del poller. Vedremo se sara' necessario
	implementarne uno nuovo con funzionalita' differenti.
	"""
	def __init__(self, schedule_callback):
		super(HoldingScheduler, self).__init__(schedule_callback)



class HoldingThread(threading.Thread):
	"""
	                              HoldingThread (class)
	     +--------------------------------------------------------------------+                                                                 
	     |                                                                    |
	     |     schedule    .............      unhold_queu       ...........   |
	     |   ------------  . scheduler .      |---------|       . holding .   | 
	     |    a4 a7 a3     .  thread   . --->   a1        --->  . thread  .------------------------> ROUTING THREAD
	     |   ------------  .............      |---------|       ...........   | route_new_alarm()
	     |        ^                                                  |        | archive_alarm()
	     |        +--------------------------------------------------+        |
	     |                                                                    |
	     +--------------------------------------------------------------------+                                                                 
	"""

	def __init__(self, context):

		super(HoldingThread, self).__init__(name="server")

		if (hasattr(self, "daemon")):
			self.daemon = True
		else:
			self.setDaemon(True)

		self.context = context

		self.log = context.log

		self.scheduler    = None
		self.unhold_queue = Queue.Queue(-1)

		self._stop = 0

		self.counter_held_alarms = 0
		self.counter_woken_alarms = 0

	def stop(self):
		self._stop = 1

	def run(self):

		self._start_scheduler()

		self._hold_loop()

	#------------------------------------------------------------------------------#
	#                                                                              #
	#------------------------------------------------------------------------------#

	def _start_scheduler(self):
		self.log.debug("Starting hold scheduler")
		self.scheduler = HoldingScheduler(self.__scheduled_callback)  
		self.scheduler.start()

	def __scheduled_callback(self, scheduled_handler, scheduled_params):
		# Just Move alarm to the unhold queue
		self.unhold_queue.put_nowait( (scheduled_handler, scheduled_params) )

	#------------------------------------------------------------------------------#
	#                                                                              #
	#------------------------------------------------------------------------------#

	def schedule_new_held_alarm(self, wakeup_time, alarm_uuid):
		"""
		Main method used by other entables thread (routing thread).
		"""

		self.log.info("Scheduling alarm %s wakeup at %s", alarm_uuid, timestamp_to_isostr(wakeup_time))

		params = ( alarm_uuid, ) 
		self.scheduler.scheduleNewEvent(wakeup_time, self.wakeup_held_message, params)
		self.counter_held_alarms += 1

		return

	#------------------------------------------------------------------------------#
	#                                                                              #
	#------------------------------------------------------------------------------#

	def _hold_loop(self):

		self.log.info( "HOLD start")

		#
		# Close all pending DB connections from this thread
		#
		connection.close()

		while not self._stop:
			# Potremmo ciclare ogni X secondi per vedere se nel DB
			# c'e' un allarme in hold da svegliare.
			# Questo lavoro lo fa gia' lo scheduler in memoria quindi
			# possiamo evitare.

			# Questo thread potrebbe fare altre cose? Non mi pare.
			# Per ora fa quasi tutto lo scheduler.

			#time.sleep(1)

			try:
				scheduled_handler, scheduled_params = self.unhold_queue.get(True, 1) # block waiting

				self.log.info( "Alarm to unhold")

				scheduled_handler(*scheduled_params)

			except Queue.Empty, er:
				continue
			
			except Exception, er:
				self.log.error("Unknown error while handling unhold alarm: %s:%s -> %s", type(e), str(e), traceback.format_exc())

		self.log.info( "HOLD exit")

	#------------------------------------------------------------------------------#
	#                                                                              #
	#------------------------------------------------------------------------------#

	def wakeup_held_message(self, alarm_uuid):
		"""
		Metodo principale eseguito per gestire gli allarmi da sbloccare.
		"""

		try:
			self._wakeup_held_message(alarm_uuid)

			self.counter_woken_alarms += 1

		except Exception, e:
			self.log.error("Unexpected error while waking up alarm %s: %s", alarm_uuid, str(e))

		#
		# Close all pending DB connections from this thread
		#
		connection.close()

		return

	def _wakeup_held_message(self, alarm_uuid):

		now = int(time.time())

		self.log.info("Un-holding alarm: %s", alarm_uuid)

		#
		# Load alarm 
		#
		alarm = self.context.permaqueue.get(alarm_uuid)

		self.log.info("    hold time  : %s %s", alarm.hold_time       , timestamp_to_isostr(alarm.hold_time       ))
		self.log.info("    wakeup time: %s %s", alarm.hold_wakeuptime , timestamp_to_isostr(alarm.hold_wakeuptime ))
		self.log.info("            now: %s %s", now                   , timestamp_to_isostr(now                   ))
		self.log.info("        handler: %s", alarm.hold_handler)

		
		try:
			outcome = ROUTING_OUTCOME_CONTINUE

			#
			# Le operazioni di un-hold sono gestite da un handler (target)?
			#
			if alarm.hold_handler != None:

				handler = self.context.registered_targets.get(alarm.hold_handler,None)
				if handler == None:
					raise Exception("Invalid target handler: %s" % (alarm.hold_handler))

				outcome = handler.on_unhold_alarm(self.context, alarm)

				self.log.info("Alarm %s on_unhold outcome = %s." % (alarm_uuid, outcome))

			#
			# L'allarme deve restare in stato HOLD? (REHOLD)
			#
			if outcome in [ROUTING_OUTCOME_HOLD, ROUTING_OUTCOME_HOLDNEXT]:

				self.context.permaqueue.change_status(alarm.uuid, ALARM_STATUS_HOLD)

				wakeup_time = alarm.hold_wakeuptime

				self.log.info("Alarm %s RE-HOLDING until %s.", alarm_uuid, wakeup_time)

				self.schedule_new_held_alarm(wakeup_time, alarm_uuid)

				self.log.info("Scheduled.")
				
			else:

				#
				# Fine effettiva dello stato HOLD
				#
				alarm.reset_hold_data()

				# Assicuriamoci di aver salvato lo stato di questo 
				self.context.permaqueue.save(alarm_uuid)

				#
				# Continua?
				#
				if outcome == ROUTING_OUTCOME_CONTINUE:
					# Quando un allarme deve essere "svegliato" dallo stato hold 
					# deve semplicemente ripartire il routing.
					# Avvisiamo il routing thread di riprendere l'allarme fermo.
					self.context.routingthread.route_new_alarm( alarm_uuid )
				else:
					#
					# Salviamo lo stato dropped
					#
					msg = "Alarm unhold ended with DROP! Archiving"
					alarm.set_verbstatus(msg)

					self.log.info(msg)
					self.context.permaqueue.change_status(alarm.uuid, ALARM_STATUS_DROPPED)

					#
					# Archiviamo l'allarme definitivamente
					#
					msg = "Alarm %s now ready for archiving" % (alarm.uuid)
					alarm.set_verbstatus(msg)

					self.log.info(msg)
					self.context.serverthread.archive_alarm(alarm)

		except Exception, e:
			msg = "Error while unholding alarm %s: %s." % (alarm_uuid, str(e))
			self.log.error(msg)

			alarm.set_verbstatus(msg)
			self.context.permaqueue.change_status(alarm_uuid, ALARM_STATUS_ERROR)

			self.context.serverthread.archive_alarm(alarm)

		return

	def get_infos(self):

		r = []

		r = [ "Held alarms: %s"      % self.counter_held_alarms ]
		r = [ "Scheduled alarms: %s" % (self.scheduler.get_schedule_size())]
		r = [ "Woken alarms: %s"     % self.counter_woken_alarms ]

		return r


	def get_hold_queue_info(self, size=30):

		r = []

		queue_data = self.scheduler.getScheduledEvents(size)
		for starttime, handler, handler_params in queue_data:

			alarm_uuid = handler_params[0]

			entry = "%s %s" % (timestamp_to_isostr(starttime), alarm_uuid)

			r.append(entry)

		return r
