# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

.. _entables-tagproviders-digest:

Digest
======

**IMPORTANTE**: questo tagprovider ha senso solo se usato per processare 
messaggi digest prodotti utilizzando il target :ref:`entables-targets-digest`.

Questo tag processa i dati memorizzati l'attributo "digest_data" associato 
ad un messaggio, se presente, e permette di usare tag per espandere il *digest* in formato testuale.

I simboli definiti da questo tag sono definiti nella seguente tabella.

.. note:: Poiche' questo e' un tag fornito dal sistema tutti i tag definiti sono accessibili anche tramite degli shortcut: 


	+-----------------------------+---------------------------+---------------------------------------------------------------------------------------------------------------------+
	| TAG                         | Shortcut                  | Descrizione                                                                                                         |
	+=============================+===========================+=====================================================================================================================+
	| {{ digest.textbrief }}      | {{ digesttextbrief }}     | Stringa di testo contenente una versione compressa del digest con le informazioni base                              |
	+-----------------------------+---------------------------+---------------------------------------------------------------------------------------------------------------------+
	| {{ digest.textfull }}       | {{ digesttextfull }}      | Stringa di testo contenente il digest formattato in maniera molto verbosa.                                          |
	+-----------------------------+---------------------------+---------------------------------------------------------------------------------------------------------------------+

Esempio di utilizzo
-------------------

Ritarda tutti gli allarmi fino alle 18:00 di oggi (o del giorno dopo)

::
	
	entables -t notify -A DEFAULT -j DIGEST --digest_holduntil_next 18:00
	
	entables -t notify -A DEFAULT -j SMTP --smtp_rcpt_to '{{ mail.to }}'      
	                                      --smtp_subject '{{ mail.subject }}' 
	                                      --smtp_message '{{ mail.body }} - {{ digest.textbrief }} {{ digest.textfull }}' 



Esempio digesttextbrief
~~~~~~~~~~~~~~~~~~~~~~~

::
	
	TENANT:site
	DOWN:1


Esempio digesttextfull
~~~~~~~~~~~~~~~~~~~~~~

::
	
	============================== START DIGEST ==============================
	
	DIGEST BRIEF
	
	Tenant           :  site
	First alarm time :  2015-01-16 16:58:43
	First alarm path :  localhost;;y;check
	Total alarms seen:  6
	
	DOWN CONDITIONS (total 1)
	
	2015-01-16 16:58:47 , localhost                      , localhost;;x;check             ,                                , status=DN, changes = 2 , http://1.1.15.160:8000/web/tenantpage/7f4d9bb05cf44ce1ac432de2c3384677/node/b78ff3df8fb64a8e9c18e070a0added4
	
	OTHER  (total 1)
	
	2015-01-16 16:58:49 , localhost                      , localhost;;y;check             ,                                , status=UP, changes = 4 , http://1.1.15.160:8000/web/tenantpage/7f4d9bb05cf44ce1ac432de2c3384677/node/b78ff3df8fb64a8e9c18e070a0added4
	============================== END DIGEST ==============================



Digest personalizzabile
------------------------

E' possibile creare digest completamente personalizzati sfruttando una speciale sintassi:



::
	
	{% for c in digest.others %}
		... testo ripetuto per ogni condition down ...
	{% endfor %}
	
	{% for c in digest.downs  %}
		... testo ripetuto per ogni condition non down ...
	{% endfor %}
	

All'inteno del testo si possono usare i seguenti tag:

	+-------------------------+------------------------------------------------------------+
	| TAG                     | DESCRIPTION                                                |
	+=========================+============================================================+
	| {{ c.ts }}              | Data della condition                                       |
	+-------------------------+------------------------------------------------------------+
	| {{ c.element }}         | Nome elemento associato all'allarme                        |
	+-------------------------+------------------------------------------------------------+
	| {{ c.condition }}       | Path della condition dell'allarme                          |
	+-------------------------+------------------------------------------------------------+
	| {{ c.title }}           | Titolo della condition                                     |
	+-------------------------+------------------------------------------------------------+
	| {{ c.state }}           | Stato della condition (DN, UP, ecc.)                       |
	+-------------------------+------------------------------------------------------------+
	| {{ c.num_changes }}     | Numero di variazioni aggregate nel digest                  |
	+-------------------------+------------------------------------------------------------+
	| {{ c.url }}             | URL alla pagina di sanet                                   |
	+-------------------------+------------------------------------------------------------+

Sono anche disponibili i seguenti tag:

::
	+----------------------------+------------------------------------------------------------+
	| TAG                        | DESCRIPTION                                                |
	+============================+============================================================+
	| {{ digest.tenant }}        |  Nome del tenant                                           |
	+----------------------------+------------------------------------------------------------+
	| {{ digest.firsttime }}     |  Orario del primo allarme processato dal digest            |
	+----------------------------+------------------------------------------------------------+
	| {{ digest.firstpath }}     |  Path della condition associata al primo allarme           |
	+----------------------------+------------------------------------------------------------+
	| {{ digest.total }}         |  Numero totale di allarmi aggretagi dal digest             |
	+----------------------------+------------------------------------------------------------+

Esempio di messaggio di digest completamente personalizzato:

::
	
	GENERAL SECTION
	----------------
	
	Tenant            = {{ digest.tenant    }}
	First alarm time  = {{ digest.firsttime }}
	First alarm path  = {{ digest.firstpath }}
	Total alarms seen = {{ digest.total     }}
	
	DOWN CONDITIONS 
	---------------
	
	Down Count        = {{ digest.downcount }}
	
	{% for c in  digest.downs %}
		ts        = {{ c.ts          }} 
		element   = {{ c.element     }}
		condition = {{ c.condition   }}
		title     = {{ c.title       }}
		state     = {{ c.state       }}
		changes   = {{ c.num_changes }}
		url       = {{ c.url         }}
	
	{% endfor %}
	
	OTHERS
	---------------
	
	{% for c in digest.others %}
		ts        = {{ c.ts          }}
		element   = {{ c.element     }}
		condition = {{ c.condition   }}
		title     = {{ c.title       }}
		state     = {{ c.state       }}
		changes   = {{ c.num_changes }}
		url       = {{ c.url         }}
	
	{% endfor  %}



















Sintassi deprecata
---------------------

.. danger:: La seguente sintassi e' disponibile per retro-compatibilita' e verra' rimossa. Si consiglia di aggiornare i template utilizzati e utilizzare la sintassi descritta nel paragrafo precedente.

I tag:

	+----------------------------------+---------------------------+---------------------------------------------------------------------------------------------------------------------+
	| TAG                              | Shortcut                  | Descrizione                                                                                                         |
	+==================================+===========================+=====================================================================================================================+
	| ${tags['digest:digesttextbrief']}| ${digesttextbrief}        | Stringa di testo contenente una versione compressa del digest con le informazioni base                              |
	+----------------------------------+---------------------------+---------------------------------------------------------------------------------------------------------------------+
	| ${tags['digest:digesttextbrief']}| ${digesttextfull}         | Stringa di testo contenente il digest formattato in maniera molto verbosa.                                          |
	+----------------------------------+---------------------------+---------------------------------------------------------------------------------------------------------------------+

Esempio di utilizzo
-------------------

Ritarda tutti gli allarmi fino alle 18:00 di oggi (o del giorno dopo)

::
	
	entables -t notify -A DEFAULT -j DIGEST --digest_holduntil_next 18:00
	
	entables -t notify -A DEFAULT -j SMTP --smtp_rcpt_to '${mail_to}'      
	                                      --smtp_subject '${mail_subject}' 
	                                      --smtp_message '${mail_body} - ${digesttextbrief} ${digesttextfull}' 





Digest personalizzabile
------------------------

E' possibile creare digest completamente personalizzati sfruttando una speciale sintassi:



::
	
	% for c in tags['digest:digestothers']:
		... testo ripetuto per ogni condition down ...
	% endfor 
	
	% for c in tags['digest:digestdowns']:
		... testo ripetuto per ogni condition non down ...
	% endfor 

All'inteno del testo si possono usare i seguenti tag:

	+---------------------+------------------------------------------------------------+
	| TAG                 | DESCRIPTION                                                |
	+=====================+============================================================+
	| ${c['ts']}          | Data della condition                                       |
	+---------------------+------------------------------------------------------------+
	| ${c['element']}     | Nome elemento associato all'allarme                        |
	+---------------------+------------------------------------------------------------+
	| ${c['condition']}   | Path della condition dell'allarme                          |
	+---------------------+------------------------------------------------------------+
	| ${c['title']}       | Titolo della condition                                     |
	+---------------------+------------------------------------------------------------+
	| ${c['state']}       | Stato della condition (DN, UP, ecc.)                       |
	+---------------------+------------------------------------------------------------+
	| ${c['num_changes']} | Numero di variazioni aggregate nel digest                  |
	+---------------------+------------------------------------------------------------+
	| ${c['url']}         | URL alla pagina di sanet                                   |
	+---------------------+------------------------------------------------------------+

Sono anche disponibili i seguenti tag:

::
	+---------------------+------------------------------------------------------------+
	| TAG                 | DESCRIPTION                                                |
	+=====================+============================================================+
	| ${digesttenant}     |  Nome del tenant                                           |
	+---------------------+------------------------------------------------------------+
	| ${digestfirsttime}  |  Orario del primo allarme processato dal digest            |
	+---------------------+------------------------------------------------------------+
	| ${digestfirstpath}  |  Path della condition associata al primo allarme           |
	+---------------------+------------------------------------------------------------+
	| ${digesttotal}      |  Numero totale di allarmi aggretagi dal digest             |
	+---------------------+------------------------------------------------------------+

Esempio di messaggio di digest completamente personalizzato:

::
	
	GENERAL SECTION
	----------------
	
	Tenant            = ${digesttenant}
	First alarm time  = ${digestfirsttime}
	First alarm path  = ${digestfirstpath}
	Total alarms seen = ${digesttotal}
	
	DOWN CONDITIONS 
	---------------
	
	Down Count        = ${digestdowncount}
	
	% for c in tags['digest:digestdowns']:
		ts        = ${c['ts']} 
		element   = ${c['element']}
		condition = ${c['condition']}
		title     = ${c['title']}
		state     = ${c['state']}
		changes   = ${c['num_changes']}
		url       = ${c['url']}
		
	% endfor 
	
	OTHERS
	---------------
	
	% for c in tags['digest:digestothers']:
		
		ts        = ${c['ts']} 
		element   = ${c['element']}
		condition = ${c['condition']}
		title     = ${c['title']}
		state     = ${c['state']}
		changes   = ${c['num_changes']}
		url       = ${c['url']}
		
	% endfor 








"""

import StringIO
import time
import pprint
import traceback
from socket import gethostname

from sanet_common.sanetdefs.constants import CSTATE_MACRO_DOWN
from sanet_common.timeutils  import timestamp_to_isostr

from django.conf import settings

from core.api.cache import init_cache_manager_from_conf

from entables.tagproviders import BaseTagProvider

from entables.targets.digest import get_infos_from_digest_data

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

NOT_DEFINED = '**N/A**'

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#
    
class TagProvider(BaseTagProvider):

	def __init__(self):
		super(TagProvider, self).__init__()
		self.name = 'digest'


	def get_provider_symbol(self, alarm, log=None):

		digest_data = alarm.get('digest_data', None)
		if digest_data is None:
			digest_data = {}
			digest_data['downs']                = []
			digest_data['others']               = []
			digest_data['errors']                     = []
			return digest_data
		
		symbols = dict( self.__gettags(alarm, None)  )

		for name, value in symbols.items():
			if name.startswith('digest'):
				name = name[6:]
				symbols[name] = value
		
		symbols['changescount'] = alarm.get('digestchangescount', NOT_DEFINED)

		return symbols


	# >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
	# DEPRECATO >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
	# >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

	VALID_SYMBOLS = [
		'digesttextbrief',
		'digesttextfull' ,
		'digestchangescount',

		#'conditions_downcount',
		#'conditions_down',
		#'conditions_other',
		#'errors',
		'digesttenant',
		'digesttotal',
		'digestfirsttime',
		'digestfirstpath',
		'digestdowncount',
		'digestdowns',
		'digestothers',
		'errors',
	]

	VALID_SHORTCUTS = [
		'digesttextbrief',
		'digesttextfull' ,
		'digestchangescount',

		#'conditions_downcount',
		#'conditions_down',
		#'conditions_other',
		#'errors',
		'digesttenant',
		'digesttotal',
		'digestfirsttime',
		'digestfirstpath',
		'digestdowncount',
		'digestdowns',
		'digestothers',
		'errors',
	]


	def has_symbol(self, alarm, symbol):

		if symbol in self.VALID_SYMBOLS:

			#if symbol == 'digestchangescount':
			#	return True
			#else:
			#	digest_data = alarm.get('digest_data', None)
			#	if digest_data is not None:
			#		return True
			return True	

		return False

	def get_symbol(self, alarm, symbol):
		
		if symbol == 'digestchangescount':
			return alarm.get('digestchangescount', NOT_DEFINED)
		else:
			tags = self.__gettags(alarm, symbol)

			try:
				return tags[symbol]
			except KeyError, e:
				raise Exception("Symbold %s not defined", str(e))

	def get_shortcut_symbol(self, alarm, symbol):

		if symbol == 'digestchangescount':
			return alarm.get('digestchangescount', NOT_DEFINED)
		else:
			tags = self.__gettags(alarm, symbol)

			try:
				return tags[symbol]
			except KeyError, e:
				raise Exception("Symbold %s not defined", str(e))


	def __gettags(self, alarm, symbol):

		#
		# FIX: CACHE THESE TAGS
		#
		tags = {}

		default_msg = "*** no digest data to process ***"

		try:
			if alarm is None:
				#return tags
					
				tags['digesttextbrief']            = default_msg
				tags['digesttextfull' ]            = default_msg
				tags['digesttenant']               = None
				tags['digesttotal']                = 0
				tags['digestfirsttime']            = None
				tags['digestfirstpath']            = None
				tags['digestdowncount']            = 0
				tags['digestdowns']                = []
				tags['digestothers']               = []
				tags['errors']                     = []
				
				tags['digestconditions_downcount'] = -1
				tags['digestconditions_down']      = []
				tags['digestconditions_other']     = []
				tags['digesterrors']               = []
				return tags
				

			digest_data = alarm.get('digest_data', None)
			if digest_data is not None:
				digest = self._create_plain_text(alarm, digest_data)

				tags['digesttextbrief']            = digest['brief']
				tags['digesttextfull' ]            = digest['full' ]
				tags['digesttenant']               = digest['tenant'              ]
				tags['digesttotal']                = digest['total'               ]
				tags['digestfirsttime']            = digest['first_alarm_time'    ]
				tags['digestfirstpath']            = digest['first_alarm_path'    ]
				tags['digestdowncount']            = digest['conditions_downcount']
				tags['digestdowns']                = digest['conditions_down']
				tags['digestothers']               = digest['conditions_other']
				tags['errors']                     = digest['errors']
			else:
				tags['digesttextbrief']            = default_msg
				tags['digesttextfull' ]            = default_msg
				tags['digesttenant']               = None
				tags['digesttotal']                = 0
				tags['digestfirsttime']            = None
				tags['digestfirstpath']            = None
				tags['digestdowncount']            = 0
				tags['digestdowns']                = []
				tags['digestothers']               = []
				tags['errors']                     = []
				
				tags['digestconditions_downcount'] = -1
				tags['digestconditions_down']      = []
				tags['digestconditions_other']     = []
				tags['digesterrors']               = []


		except Exception, e:
			print e
			msg = "* Error processing digest: %s: %s *" % (str(e), traceback.format_exc())

			tags['digesttextbrief']            = msg
			tags['digesttextfull' ]            = msg
			tags['digesttenant']               = None
			tags['digesttotal']                = 0
			tags['digestfirsttime']            = None
			tags['digestfirstpath']            = None
			tags['digestdowncount']            = 0
			tags['digestdowns']                = []
			tags['digestothers']               = []
			tags['errors']                     = []
			

			tags['digestconditions_downcount'] = -1
			tags['digestconditions_down']      = []
			tags['digestconditions_other']     = []
			tags['digesterrors']               = []

		return tags


	def _create_plain_text(self, main_alarm, digest_data):
		
		from core.models import Condition, Tenant
		from core.api.base import db_set_current_tenant_schema
		import webui.models  # for Urls

		#
		# Move to tenant
		#
		tenant_id = main_alarm['tenantid']
		tenant    = Tenant.objects.get(uuid=tenant_id)
		db_set_current_tenant_schema(tenant.name)

		cache_manager = init_cache_manager_from_conf()

		#
		# Load alarms data from digest data
		#
		conditions_changes_by_path , errors = get_infos_from_digest_data( main_alarm, digest_data )

		#
		# Produce informations used by digest full/briefs
		#
		first_ts  = main_alarm['ts']

		downs_count       = Condition.objects.select_related('element', 'element__parent').filter(path__in=conditions_changes_by_path.keys()).filter(status__in=CSTATE_MACRO_DOWN).count()
		down_conditions   = Condition.objects.select_related('element', 'element__parent').filter(path__in=conditions_changes_by_path.keys()).filter(status__in=CSTATE_MACRO_DOWN).order_by('path')
		other_conditions  = Condition.objects.select_related('element', 'element__parent').filter(path__in=conditions_changes_by_path.keys()).exclude(status__in=CSTATE_MACRO_DOWN).order_by('path')

		digest_symbols = {}
		digest_symbols['tenant'              ] = tenant.name
		digest_symbols['total'               ] = len(digest_data['alarms'])
		digest_symbols['first_alarm_time'    ] = timestamp_to_isostr(first_ts)
		digest_symbols['first_alarm_path'    ] = main_alarm['condition']['path']
		digest_symbols['conditions_downcount'] = downs_count
		digest_symbols['conditions_down'     ] = self.get_conditions_data_for_output(down_conditions , conditions_changes_by_path, cache_manager)
		digest_symbols['conditions_other'    ] = self.get_conditions_data_for_output(other_conditions, conditions_changes_by_path, cache_manager)
		digest_symbols['errors'              ] = errors
		
		#
		# Calculate brief text
		#
		text = self.get_brief_digest_text(digest_data, digest_symbols)
		digest_symbols['brief'] = text

		#
		# Calculate full text
		#
		text = self.get_full_digest_text(digest_data, digest_symbols)
		digest_symbols['full'] = text

		return digest_symbols


	def get_full_digest_text(self, digest_data, digest_symbols):

		output = StringIO.StringIO()

		print >>output, ""
		print >>output, "============================== START DIGEST =============================="
		print >>output, ""
		print >>output, "DIGEST BRIEF"
		print >>output, ""
		print >>output, "Tenant           : ", digest_symbols['tenant']
		print >>output, "First alarm time : ", digest_symbols['first_alarm_time']
		print >>output, "First alarm path : ", digest_symbols['first_alarm_path']
		print >>output, "Total alarms seen: ", len(digest_data['alarms'])

		conditions = digest_symbols['conditions_down']
		print >>output, ""
		print >>output, "DOWN CONDITIONS (total %s)" % (len(conditions))
		print >>output, ""
		if conditions:
			self._print_conditions(output, conditions)
		else:
			print >>output, "  *no pending down conditions*"

		conditions = digest_symbols['conditions_other']
		if conditions:
			print >>output, ""
			print >>output, "OTHER  (total %s)" % (len(conditions))
			print >>output, ""
			self._print_conditions(output, conditions)

		errors = digest_symbols['errors']
		if errors:
			print >>output, ""
			print >>output, "ERRORS"
			print >>output, ""
			for e in errors:
				print >>output, "%s" % (e)

		print >>output, "============================== END DIGEST =============================="
		text = output.getvalue()
		output.close()

		return text

	def get_brief_digest_text(self, digest_data, digest_symbols):


		downs_count = digest_symbols['conditions_downcount']
		errors      = digest_symbols['errors']

		output = StringIO.StringIO()

		print >>output, "TENANT:%s"    % (digest_symbols['tenant'])
		print >>output, "DOWN:%s"      % (digest_symbols['conditions_downcount'])
		if errors:
			print >>output, "ERRORS: %s" % (len(digest_symbols['errors']))

		text = output.getvalue()
		output.close()

		return text

	def _print_conditions(self, output, conditions):
		for c in conditions:
			print >>output, "%s , %-30s , %-30s , %-30s , status=%s, changes = %s , %s" % (
				c['ts'          ],
				c['element'     ],
				c['condition'   ],
				c['title'       ],
				c['state'       ],
				c['num_changes' ],
				c['url'         ],
			)

	def get_conditions_data_for_output(self, conditions, conditions_changes_by_path, cache_manager):

		r = []

		for c in conditions:
			url  = c.element.get_node().global_page_url
			base = settings.ALARM_URL_BASE % { 'hostname': gethostname() }
			r.append({
				'ts'          : timestamp_to_isostr(c.statuslastchange),
				'element'     : c.element.path,
				'condition'   : c.path,
				'title'       : c.expanded_title(cache_manager),
				'state'       : c.get_state(),
				'num_changes' : conditions_changes_by_path[c.path],
				'url'         : base + url
			})

		return r
