"""

Target di sistema
#################

CONTINUE
==========

Non esegue nessuna operazione.  

::
	
	-J CONTINUE

Esempi
------

::
	
	entables   -A DEFAULT  -m ......  -J CONTINUE

RETURN
==========

L'allarme termina la traversata della catena corrente, tornando eventualmente alla catena precedente in caso avesse effettuato un *jump* (o piu' *jump*).

**NOTA**: non viene applicata la regola di default della catena corrente.

::
	
	-J RETURN

Esempi
------

::
	
	entables   -A CATENA_H24  -m ......  -J RETURN


ENDTABLE
==========

L'allarme termina **immediatamente** la sua traversata della tabella corrente (senza che vengano processate regole di default delle chain che sta attraversando):

::
	
	-J ENDATABLE

Esempi
------

::
	
	entables   -A DEFAULT  -m ......  -J ENDTABLE


DROP
==========

Ferma definitvamente un allarme.  

::
	
	-J DROP

Esempi
------

Tutti gli allarmi nel weekend vengono ignorati:

::
	
	entables   -A DEFAULT  -m times --times_range  0-24/6-7  -J DROP

JUMP
==========

Il comando di *jump* non viene indicato esplicitamente. Per effettuare un jump ad una *catena* bisogna indicare come *target*
il nome della catena:

::
	
	-J COMMON_RULES

Esempi
------

Tutti gli allarmi prodotto nei giorni del weekend vengono mandati ad una catena chiamata *WEEKENDS*:

::
	
	entables   -A DEFAULT  -m times --times_range  0-24/6-7  -J WEEKENDS

"""

import time

from entables.constants import ROUTING_OUTCOME_CONTINUE
from entables.constants import ROUTING_OUTCOME_RETURN  
from entables.constants import ROUTING_OUTCOME_DROP    
from entables.constants import ROUTING_OUTCOME_ENDTABLE
from entables.constants import ROUTING_OUTCOME_HOLD    

from entables.templatemanager import calculate_options_value

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class Option(object):

	def __init__(self, name, type=str, required=False, default=None, shortname=None,  choices=None, help=None, multi=False):
		self.name      = name
		self.shortname = shortname
		self.type      = type
		self.choices   = choices
		self.required  = required
		self.default   = default
		self.help      = help
		self.multi     = multi
		
	def has_default(self):
		return (self.default is not None)
		
	def add_parser_options(self, prefix, argument_parser):

		names = []
		args = {}
		
		cmdline_name = '--%s_%s' % (prefix, self.name)
		#print "--------", cmdline_name
		
		names.append(cmdline_name)

		if self.shortname: names.append('-%s' % (self.shortname))

		args['type'] = self.type
		args['required'] = self.required
		
		if self.help    is not None: args['help']    = self.help			
		if self.choices is not None: args['choices'] = self.choices
		#if self.default is not None: args['default'] = self.default
			
		argument_parser.add_argument(*names,**args)
		
#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class BaseTargetHandler(object):

	def __init__(self):
		self.name    = None
		self.options = []
		self.tables  = None

		self.hitcounter = 0
		return

	def init(self):
		return

	def get_text_info(self):
		return "hit:" + (str(self.hitcounter))

	def get_name(self):
		return self.name

	def get_options(self):
		return dict([ (o.name, o) for o in self.options])

	def get_options_default(self):
		effective_options = {}
		for n,o in self.get_options().items():
			if o.has_default():
				effective_options[n] = o.default
		return effective_options

	def validate_target_options(self, options_dict):

		# Base validation: check only required options
		for optiondef in self.options:
			if optiondef.required:
				if options_dict.get(optiondef.name, None) is None:
					raise Exception("Missing required option: %s", optiondef.name)

	def expand_options(self, context, alarm, options):
		return self.expand_options_from_alarmdata(context, alarm.data, options)

	def expand_options_from_alarmdata(self, context, alarm_data, options):

		options_defaults = self.get_options_default()

		tag_providers = context.registered_tagproviders
		templates     = context.ruledb.tenants[alarm_data['tenantid']].templates
		
		context.log.debug("tag providers: %s", tag_providers)
		
		context.log.debug("handler        : %s", self.name)
		context.log.debug("default options: %s", options_defaults)
		context.log.debug("target options : %s", options)

		expanded_options = calculate_options_value(context.log, alarm_data, options, options_defaults, tag_providers, templates)
		
		return expanded_options

	def check_rule(self, ruleobj):
		return True

	def dispatch(self, context, alarm, ruleobj, options):
		self.hitcounter += 1

	def on_unhold_alarm(self, context, alarm):
		"""
		Questo metodo viene chiamato da entables quando un allarme "bloccato" dal target deve essere 
		fatto ripartire (unhold),
		
		Il metodo deve restituire come outcome:
		
			* ROUTING_OUTCOME_CONTINUE  quando l'operazione di unhold va a buon fine
			* ROUTING_OUTCOME_DROP      se l'allarme non deve ripartire dopo l'unhold.
		"""

		context.log.debug("%s: UNHOLD CALLBACK on %s", self.name, alarm.uuid)

		return ROUTING_OUTCOME_CONTINUE

	def clean(self):
		return

	#
	# For parsing
	#
	def add_args_to_parser(self, argument_parser):

		prefix = self.name

		for option_name, option in self.get_options().items():
			option.add_parser_options(prefix, argument_parser)
		return

	def build_from_parsed_args(self, args):
		
		defaults = self.get_options_default()
		
		options = {}
		prefix = self.name + "_"
		for optionname, optionvalue in args.items():
			
			if optionvalue is None:
				continue
				
			if optionname.startswith(prefix):
				optionname = optionname[len(prefix):]
			
			# Skip defaults
			#if defaults.has_key(optionname):
			#	if defaults[optionname] == optionvalue:
			#		continue
				
			options[optionname] = optionvalue
		return options

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class DROPTarget(BaseTargetHandler):

	def __init__(self):
		super(DROPTarget, self).__init__()

		self.name    = 'drop'
		
	def dispatch(self, context, alarm, ruleobj, options):
		
		super(DROPTarget, self).dispatch(context, alarm, ruleobj, options)
		
		alarm.set_verbstatus("Target %s" % (self.name))

		return ROUTING_OUTCOME_DROP     

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

DEFAULT_HOLD_TIME = 60 # 60 seconds

class HOLDTarget(BaseTargetHandler):

	def __init__(self):
		super(HOLDTarget, self).__init__()

		self.name = 'hold'

	def dispatch(self, context, alarm, ruleobj, options):
		
		super(HOLDTarget, self).dispatch(context, alarm, ruleobj, options)
		
		log = context.log
		
		alarm.set_verbstatus("Target %s", self.name)
		
		start_ts = time.time()

		unhold_ts = start_ts + DEFAULT_HOLD_TIME
		
		alarm.set_verbstatus("\tSetting unholdtime to: %s" % (unhold_ts))
		alarm.set_hold_state( start_ts, unhold_ts)
		
		outcome = ROUTING_OUTCOME_HOLD
		

		return ROUTING_OUTCOME_HOLD

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class CONTINUETarget(BaseTargetHandler):

	def __init__(self):
		super(CONTINUETarget, self).__init__()

		self.name = 'continue'

	def dispatch(self, context, alarm, ruleobj, options):

		super(CONTINUETarget, self).dispatch(context, alarm, ruleobj, options)

		return ROUTING_OUTCOME_CONTINUE

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class RETURNTarget(BaseTargetHandler):

	def __init__(self):
		super(RETURNTarget, self).__init__()

		self.name = 'return'

	def dispatch(self, context, alarm, ruleobj, options):

		super(RETURNTarget, self).dispatch(context, alarm, ruleobj, options)

		return ROUTING_OUTCOME_RETURN

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class ENDTABLETarget(BaseTargetHandler):

	def __init__(self):
		super(ENDTABLETarget, self).__init__()

		self.name = 'endtable'

	def dispatch(self, context, alarm, ruleobj, options):

		super(ENDTABLETarget, self).dispatch(context, alarm, ruleobj, options)

		return ROUTING_OUTCOME_ENDTABLE

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#


def load_system_targets(targetlist):

	SYSTEM_TARGET_HANDLERS = [
		DROPTarget,
		HOLDTarget,
		CONTINUETarget,
		RETURNTarget,
		ENDTABLETarget,
	]

	for system_target_class in SYSTEM_TARGET_HANDLERS:
		handler_class = system_target_class()
		targetlist.add_target(handler_class)  
        




