# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

.. _entables-targets-delay:

DELAY
==========

Trattiene un allarme nel sistema fino al termine indicato.

I parametri questo target vanno specificati secondo la sintassi: 

::
	
	-J DELAY --delay_<parametro> <valore> [ --delay_<parametro> <valore> ... ]



Parametri:

	+-----------------+---------+----------+-------------------------------------------------------------------------------------------------------------------------+
	| Parametro       | Type    | Default  | Descrizione                                                                                                             |
	+=================+=========+==========+=========================================================================================================================+
	| holdfor         | numero  |          | Numero di secondi da attendere prima di sbloccare il messaggio                                                          |
	+-----------------+---------+----------+-------------------------------------------------------------------------------------------------------------------------+
	| holduntil       | stringa |          | Data di scadenza di quando sbloccare il messaggio.                                                                      |
	+-----------------+---------+----------+-------------------------------------------------------------------------------------------------------------------------+
	| holduntil_next  | stringa |          | Prossimo orario valido (del giorno stesso o giorno successivo)                                                          |
	+-----------------+---------+----------+-------------------------------------------------------------------------------------------------------------------------+
	| drop_if_up      | numero  | 0        | Quando vale 1, l'allarme di DOWN non riparte se la condition associata all'allarme non e' in stato DOWN (o similiari)   |
	+-----------------+---------+----------+-------------------------------------------------------------------------------------------------------------------------+
	| drop_if_changed | numero  | 0        | Quando vale 1, l'allarme non riparte se la condition associata all'allarme ha subito una qualunque variazione di stato  |
	+-----------------+---------+----------+-------------------------------------------------------------------------------------------------------------------------+
 
Obsoleti:

	+----------------+---------+----------------------------+--------------------------------------------------------------------------+
	| Parametro      | Type    | Default                    | Descrizione                                                              |
	+================+======================================+==========================================================================+
	| seconds        |         |                            | DEPRECATO: vedi "holdfor"                                                |
	+----------------+---------+----------------------------+--------------------------------------------------------------------------+
	| until          |         |                            | DEPRECATO: vedi "holduntil"                                              |
	+----------------+---------+----------------------------+--------------------------------------------------------------------------+
	| until_next     |         |                            | DEPRECATO: vedi "holduntil_next"                                         |
	+----------------+---------+----------------------------+--------------------------------------------------------------------------+


.. _entables-targets-delay-parameters:

Formato e semantica dei paremetri holdfor/holduntil/holduntil_next
--------------------------------------------------------------------------------------

.. _entables-targets-delay-parameter-holduntil:

Formati validi per parametro *holduntil*
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Il parametro *holduntil* seleziona il prossimo giorno **SEMPRE SUCCESSIVO** a quello corrente. 

Questo significa che se adesso sono le 06:30 di lunedi' (monday) 1/1/2016 e un allarme viene processato adesso, se il parametro *holduntil* viene specificato con valore:

::
	
	monday 7:00

il digest viene ritardato/genarto il prossimo lunedi (*monday*) **SUCCESSIVO**, ovvero 8/1/2016.



Formato 1
+++++++++++

::
	
	yyyy-mm-gg [ hh:mm [ :ss ]  ]          Es: 2013-12-23 12:34:56 , 2013-12-23 12:34 , 2013-12-23

Formato 2
+++++++++++

::
	
	<day name       > [ hh:ss ]            Es: monday 12:34 , monday
	<day week number> [ hh:ss ]            Es: 1 12:34, 1 

I giorni della settimana:

::
	
	monday           1
	tuesday          2
	wednesday        3
	thursday         4
	friday           5
	saturday         6
	sunday           7

Formato 3
+++++++++++

::
	
	next-work-day hh:ss [ <working days> ]

Calcola il primo giorno+orario lavorativo immediatamente successivo al momento di elaborazione.
Il giorno lavorativo viene scelto "saltando" i giorni festivi configurati in Sanet.

E' possibile specificare quali giorni della settimana sono da considerare giorni lavorativi attraverso 
il parametro <working days>.

Esempio:

::
	
	next-work-day 8:00                  # il giorno lavorativo piu' vicino (festivita'+ sabato + domeniche esclusi)
	next-work-day 8:00 1,2,3,4,5        # idem

	next-work-day 8:00 1,2,3,4,6,7      # il giorno lavorativo piu' vicino (festivita' + venerdi' di preghiera esclusi)


Se la data di esecuzione ha superato il giorno indicato, viene considerato il giorno della settimana successiva.






.. _entables-targets-delay-parameter-holduntil_next:

Formati validi per parametro *holduntil_next* 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

::
	
	hh:ss                                  Es: 8:00 

Esempi
------

I messaggi nel weekend vengono ritardati fino al lunedi' mattina: 

::
	
	entables   -A DEFAULT  -m times --times_range  0-24/6-7                 -j DELAY --delay_holdfor        120
	
	entables   -A DEFAULT  -m times --times_range  0-24/6-7                 -j DELAY --delay_holduntil      'monday 08:00'
	
	entables   -A DEFAULT  -m times --times_range  00:00-8:00,18:00-24/1-5  -J DELAY --delay_holduntil_next '08:00'
	
	entables   -A DEFAULT  -m times --times_range  00:00-8:00,18:00-24/1-5  -J DELAY --delay_holduntil_next '08:00' --delay_drop_if_up 1

"""

import time
import datetime

from sanet_common.sanetdefs.constants import CSTATE_MACRO_DOWN, CSTATE_DN, CSTATES_UNCHECKABLE
from sanet_common.timeutils import timestamp_to_isostr

from core.timing import init_holidays_converter
from core.models import Tenant, Condition, Event

from entables.constants import *

from entables.targets import BaseTargetHandler
from entables.targets import Option

from entables.constants import TABLE_NOTIFY
from entables.constants import ROUTING_OUTCOME_CONTINUE, ROUTING_OUTCOME_DROP

from entables.lib.timing import calculate_until_ts, calculate_next_ts

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class BaseHoldTarget(BaseTargetHandler):
	"""
	Classe base per gestire target che hanno necessita' di supportare i parametri:
		* holdfor
		* holduntil
		* holduntil_next
		* holidays_context
	"""
	def __init__(self):
		super(BaseHoldTarget, self).__init__()

		self.options = [
			Option("holdfor"         , type=int, required=False), 
			Option("holduntil"       , type=str, required=False),
			Option("holduntil_next"  , type=str, required=False),

			Option("holidays_context", type=str, required=False),
		]

	def validate_target_options(self, options_dict):

		super(BaseHoldTarget, self).validate_target_options(options_dict)

		seconds          = options_dict.get('holdfor'         , options_dict.get('seconds'         , None))
		until            = options_dict.get('holduntil'       , options_dict.get('until'           , None))
		until_next       = options_dict.get('holduntil_next'  , options_dict.get('until_next'      , None))
		holidays_context = options_dict.get('holidays_context', None)

		if (seconds is None and until is None and until_next is None):
			raise Exception("Missing parameter: seconds, until, until_next")

		v = 0
		v += 1 if seconds    is not None else 0
		v += 1 if until      is not None else 0
		v += 1 if until_next is not None else 0
		if v > 1:
			raise Exception("Incompatible arguments: seconds (%s), until (%s), until_next (%s)" % (seconds, until, until_next))

		try:
			if seconds != None:
				int(seconds)
		except Exception, e:
			raise Exception("Invalid value: %s" % (seconds))

		try:
			if until != None:
				calculate_until_ts(0, until)
		except Exception, e:
			raise Exception("Invalid value: %s" % (until))

		try:
			if until_next != None:
				calculate_next_ts(0, until_next)
		except Exception, e:
			raise Exception("Invalid value: %s" % (until_next))

		if holidays_context is not None:
			if holidays_context == "":
				raise ValueError("Invalid holidays context: %s" % (holidays_context))

		return True


	def get_unhold_timestamp(self, log, dispatch_time, options):

		start_ts = dispatch_time

		seconds          = options.get("holdfor"         , options.get("seconds"         , None))
		holduntil        = options.get("holduntil"       , options.get("until"           , None))
		holduntil_next   = options.get("holduntil_next"  , options.get("until_next"      , None))
		holidays_context = options.get('holidays_context', None)

		#
		# Load calendar handler from Sanet
		#
		calendar_handler = init_holidays_converter()

		if holidays_context:
			calendar_handler.set_default_time_context(holidays_context)

		#
		# Calculate 
		#
		log.debug("   Now is         = %s %s", start_ts, time.ctime(start_ts))
		log.debug("   holdfor        = %s"   , holduntil)
		log.debug("   holduntil      = %s"   , holduntil)
		log.debug("   holduntil_next = %s"   , holduntil)

		if seconds is not None:
			seconds   = int(seconds)
			unhold_ts = dispatch_time + seconds

		elif holduntil is not None:
			unhold_ts = calculate_until_ts(dispatch_time, holduntil, calendar_handler=calendar_handler)

		elif holduntil_next is not None:
			unhold_ts = calculate_next_ts(dispatch_time, holduntil_next)

		else:
			raise Exception("Unable to calculate required delay. No options! BUG!")

		log.debug("   unhold at = %s"   , unhold_ts)

		return unhold_ts

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class Target(BaseHoldTarget):

	ALARM_DELAY_DATA_FIELD = "_delay_data"

	def __init__(self):
		super(Target, self).__init__()

		self.name = "delay"

		self.tables = [ 
			TABLE_NOTIFY, 
		]

		self.options += [
			Option("drop_if_up"      , type=str, required=False),
			Option("drop_if_changed" , type=str, required=False),
		]

	def init(self):
		return

	def validate_target_options(self, options_dict):

		super(Target, self).validate_target_options(options_dict)

		drop_if_up       = options_dict.get('drop_if_up', None)
		if drop_if_up is not None:
			try:
				drop_if_up = int(drop_if_up)
			except Exception, e:
				raise ValueError("Invalid value: %s" % (drop_if_up))

		drop_if_changed       = options_dict.get('drop_if_changed', None)
		if drop_if_up is not None:
			try:
				drop_if_changed = int(drop_drop_if_changedif_up)
			except Exception, e:
				raise ValueError("Invalid value: %s" % (drop_if_changed))

		if drop_if_up is not None and drop_if_changed is not None:
			raise ValueError("Can not specify drop_if_up and drop_if_down together")

		return True

	def dispatch(self, context, alarm, ruleobj, target_options):

		log = context.log

		options = self.expand_options(context, alarm, target_options)

		alarm.set_verbstatus("Target %s: %s" % ( self.name, options))

		#
		# GET ALARM DATA
		#
		alarmdata = alarm.data

		#import pprint
		#pprint.pprint(alarmdata)

		try:
			# FIX: NO, l'orario di dispatch e' l'orario di produzione dell'allarme. Non quando l'allarme viene processato! O no??! 
			dispatch_ts = time.time()

			#unhold_ts = self._get_unhold_timestamp(log, dispatch_ts, options)
			unhold_ts = self.get_unhold_timestamp(log, dispatch_ts, options)

			log.info("TARGET DELAY alarm %s  until %s (%s)", alarm.uuid, timestamp_to_isostr(unhold_ts), unhold_ts, )

			unhold_handler = self.name

			alarm.set_hold_state( dispatch_ts, unhold_ts, hold_handler=unhold_handler )

			condition_status = alarmdata.get('result',{}).get('status', None)
			log.info("Alarm status: %s", condition_status )

			drop_if_up      = int(options.get("drop_if_up"     , "0"))
			drop_if_changed = int(options.get("drop_if_changed", "0"))

			#
			# E' un allarme di DOWN e drop_if_down?
			#
			if drop_if_up:
				if condition_status == CSTATE_DN:
					# Remember if this alarm needs drop on "unhold"
					alarmdata[ self.ALARM_DELAY_DATA_FIELD ] = {
						'drop_if_up': drop_if_up
					}

			elif drop_if_changed:
				# Remember if this alarm needs drop on "unhold"
				alarmdata[ self.ALARM_DELAY_DATA_FIELD ] = {
					'drop_if_changed': drop_if_changed
				}

			return ROUTING_OUTCOME_HOLDNEXT

		except Exception, e:
			log.debug("%s", str(e))

			alarm.set_verbstatus("Unable to execute target %s: %s" % ( self.name, str(e)))

			return ROUTING_OUTCOME_ERROR

	def on_unhold_alarm(self, context, alarm):

		log = context.log

		outcome = super(Target,self).on_unhold_alarm(context, alarm)

		if outcome != ROUTING_OUTCOME_CONTINUE:
			return outcome

		drop_on_unhold = False

		log.info("DELAY: unhold alarm %s",  alarm.uuid)

		alarmdata = alarm.data

		if alarmdata.has_key(self.ALARM_DELAY_DATA_FIELD):

			drop_if_up      = alarmdata[ self.ALARM_DELAY_DATA_FIELD ].get('drop_if_up'     , None)
			drop_if_changed = alarmdata[ self.ALARM_DELAY_DATA_FIELD ].get('drop_if_changed', None)

			log.info("DELAY:  drop_if_up     : %s", drop_if_up)
			log.info("DELAY:  drop_if_changed: %s", drop_if_changed)

			# Is this alarm linked to a condition?
			condition_uuid = alarmdata.get("condition", {}).get('uuid'      , None)
			log.info("DELAY:  condition : %s", condition_uuid)

			condition = None
			if condition_uuid is not None:

				# Check condition state
				try:

					tenant = Tenant.objects.get(uuid=alarm.tenant_uuid)
					tenant.db_set_current_schema()

					condition = Condition.objects.get(tenant=tenant, uuid=condition_uuid)

					log.info("DELAY: Alarm %s. Condition %s, current status=%s", alarm.uuid, condition.path, condition.status)

					if drop_if_up:
						if not condition.status in CSTATE_MACRO_DOWN:
							log.info("DELAY: UP -> drop")

							outcome = ROUTING_OUTCOME_DROP

					elif drop_if_changed:
						
						#
						# Cerchiamo tutti gli eventi/log creati dopo il momento di generazione di questo allarme
						#
						alarm_ts = alarmdata['ts']
						event_id = alarmdata['event']['uuid']
						
						events = Event.objects
						events = events.filter(tenant=tenant, condition_path=condition.path)
						events = events.filter(timestamp__gt=alarm_ts).exclude(uuid=event_id)
						events = events.exclude(status__in=CSTATES_UNCHECKABLE) # le transizioni verso stati uncheckable influenzano la logica? per ora no.
						
						num_changes = events.count()

						log.info("DELAY: changes: %s", num_changes)
						if num_changes > 0:
							log.info("DELAY: changed -> drop")

							outcome = ROUTING_OUTCOME_DROP

				except Condition.DoesNotExist, e:
					# NOTA: La condition non esiste (e' stata cancellata?). Nel dubbio, l'allarme che ha generato
					#       meglio non dropparlo.
					log.warning("DELAY: Alarm %s: condition %s not found", alarm.uuid, condition_uuid)

				except Exception, e:
					# NOTA: Anomalie impreviste. Nel dubbio facciamo continuare l'allarme.
					log.error("DELAY: Alarm %s: Unknown error while fetching condition %s: %s:%s", alarm.uuid, condition_uuid, type(e), str(e))

			#
			# Remove DELAY metadata from the alarm before continue
			#
			del alarmdata[self.ALARM_DELAY_DATA_FIELD]

		return outcome
