# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

.. _entables-targets-digest:

DIGEST
==========

Questo target viene utilizzato per fermare uno o piu' allarmi che fanno match sulla regola. I messaggi intercettati vengono
processati per produrre un "digest".

	* Il primo allarme che fa match scatena il calcolo di un nuovo "digest", che altro non e' che un raggruppamento di allarmi.
	* Il primo allarme viene fermato e tenuto bloccato fino alla scadenza indicata.

Tutti gli allarmi successivi al primo che fanno match sulla stessa regola nello stesso range temporale, se vengono processati correttamente, vengono *fermati* e salvati in stato "merged".

Allarmi ignorati dal digest
---------------------------

Se l'allarme da processare e' relativo al cambio di stato di una condition (UP->DN, DN->UP), il target *digest* gestisce in maniera particolare
i messaggi di *UP*:

	**REGOLA**: se l'allarme e' un *UP*, viene trattenuto (merge) solo se il nel digest e' presente anche l'allarme del DOWN precedente. 


Scadenza del digest e ripresa del routing
-----------------------------------------

Alla scadenza del termine indicato, il primo allarme riparte con il routing arricchito dai seguenti campi:

	+-------------+-----------------------------------------------------------------------------------------------------------------------------------------------------------------------+
	| Attributo   | Descrizione                                                                                                                                                           |
	+=============+=======================================================================================================================================================================+
	| isdigest    | 1 l'allarme e' un digest, 0 l'allarme non e' un digest (allarme ignorato)                                                                                             |
	+-------------+-----------------------------------------------------------------------------------------------------------------------------------------------------------------------+
	| digest_data | conviene informazioni su tutti i messaggi che fanno parte dello stesso digest calcolato fino a quel momento.                                                          |
	+-------------+-----------------------------------------------------------------------------------------------------------------------------------------------------------------------+
	| digestinfo  | Campo composto con i seguenti attributi:  'downs' = numero di allarmi down ancora in corso. 'conditions': dizionario con numer i di allarmi raggruppati per condition |
	+-------------+-----------------------------------------------------------------------------------------------------------------------------------------------------------------------+

Per utilizzare le informazioni di questi attributi  dovrebbero essere usati:

	* i match disponibili del match *attr* (vedi documentazione :ref:`entables-matches-attributes-digest` )
	* i tag del tagprovider *digest* (vedi documentazione :ref:`entables-tagproviders-digest`)


Parametri
---------

I parametri questo target vanno specificati secondo la sintassi: 

::
	
	-j digest --digest_<parametro> <valore>

Elenco:

	+---------------+-----------+--------------------------+--------------------------------------------------------------------------+
	| Parametro     | Type      | Default                  | Descrizione                                                              |
	+===============+===========+==========================+==========================================================================+
	| holdfor       | numero    |                          | Numero di secondi da attendere prima di sbloccare il messaggio           |
	+---------------+-----------+--------------------------+--------------------------------------------------------------------------+
	| holduntil     | stringa   |                          | Data di scadenza di quando sbloccare il digest.                          |
	+---------------+-----------+--------------------------+--------------------------------------------------------------------------+
	| holduntil_next| stringa   |                          | Prossimo orario valido (del giorno stesso o giorno successivo)           |
	+---------------+-----------+--------------------------+--------------------------------------------------------------------------+

I parametri sono equivalenti al target :ref:`entables-targets-delay`. 

Si rimanda alla sezione :ref:`entables-targets-delay-parameters` per i dettagli sulla sintassi e semantica dei parametri.


Esempi
------

Esempi di regole per creare un digest: Tutti gli allarmi tra le 00:00 e le 24:00 del weekend vengono aggregati e il digest viene inviato alle 7:00 del lunedi seguente:

::
	
	entables   -N OUTOFTIME                                  -j DROP
	entables   -A OUTOFTIME -m times --times_range  0-24/6-7 -j DIGEST --digest_holduntil "monday 07:00:00"
	entables   -A OUTOFTIME                                  -j SMTP   --smtp_rcpt_to "info@labs.it" --smpt_subject "Summary"  --smtp_message "${digesttextbrief}"  
	
	entables   -A DEFAULT  -m times --times_range  0-24/6-7  -J OUTOFTIME
	entables   -A DEFAULT                                    -j SMTP

"""

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

# Regole per testare gli allarmi UP con DOWN fuori range (vedi ticket #1333)
#
# ./bin/entables -F DEFAULT
# ./bin/entables -A DEFAULT -j SMTP --smtp_rcpt_to '${mail_to}' --smtp_message '${mail_body} - ${rawalarm}' --smtp_xheaders '${mail_xheaders}' --smtp_subject '${mail_subject}'
# 
# 
# export NEXT_DATE=$(date -d@$(expr $(date +%s) + 120) +"%H:%M")
# ./bin/entables -F DEFAULT
# ./bin/entables -A DEFAULT -j DIGEST --digest_holduntil_next "$NEXT_DATE"
# ./bin/entables -A DEFAULT -m attr --attr_isdigest 1       -j SMTP   --smtp_rcpt_to "root" --smtp_subject "Summary"  --smtp_message '${digesttextbrief} - ${digesttextfull}'
# ./bin/entables -A DEFAULT -m attr --attr_isdigest 0       -j SMTP   --smtp_rcpt_to "root" --smtp_subject "Normal"   --smtp_message '${rawalarm}'

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

import json
import time
import datetime

from sanet_common.sanetdefs.constants import CSTATE_UP, CSTATE_DN

from core.models import Tenant
from core.api.base import *

from entables.constants import *

from entables.models import TargetData
from entables.targets import Option

from entables.constants import TABLE_NOTIFY

from entables.lib.timing import *

from entables.targets.delay import BaseHoldTarget

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class Target(BaseHoldTarget):

	def __init__(self):
		super(Target, self).__init__()

		self.name = "digest"

		self.tables = [ 
			TABLE_NOTIFY, 
		]

		self.options += [
			Option("holdfor"         , type=int, required=False), 
			Option("holduntil"       , type=str, required=False), 
			Option("holduntil_next"  , type=str, required=False),

			Option("holidays_context", type=str, required=False),
		]

	def init(self):
		return


	def dispatch(self, context, alarm, ruleobj, target_options):

		log = context.log

		options = self.expand_options(context, alarm, target_options)

		alarmdata = alarm.data

		alarm.set_verbstatus("Target %s" %  (self.name))

		log.info("Alarm: %s %s", alarmdata['uuid'], alarmdata.get('condition', {'path':'*alarm with no condition*'}).get('path', '*NO PATH! STRANGE!*') )

		try:
			tenant = Tenant.objects.get(uuid=alarm.tenant_uuid)

			start_ts = time.time()

			unhold_ts = self.get_unhold_timestamp(log, start_ts,  options)

			digest_id = "entables.target.digest.%s.%s" % (ruleobj.uuid, int(unhold_ts))

			#
			# Se l'allarme e' un UP (di una condition) dopo un DOWN generato fuori dall'orario del digest
			# (quindi il down non compare nel digest), l'allarme passa.
			#
			# Richiesta: Ticket #1333
			#
			is_up = alarmdata.has_key("condition") and alarmdata.get('result',{}).get('status', None) == CSTATE_UP
			log.info("is UP: %s", is_up)
			if is_up:

				# Vediamo se c'e' un digest gia' in corso e se la condition dell'allarme compare
				# gia' negli allarmi "digeriti" fino ad ora.

				ignore = True

				tdata, created = self._load_digest_data(log, tenant, digest_id, create_if_new=False)
				if tdata is not None:

					digest_data = tdata.get_data()
					if digest_data is not None:

						log.debug("Checking digested alarms: %s", digest_data['alarms'])

						from entables.models import Alarm, AlarmArchive

						for alarm_id in digest_data['alarms']:

							log.debug("\talarm: %s", alarm_id)

							db_alarm = None

							# Retrive data of the alarm (active or archived/merged)
							try:
								db_alarm = AlarmArchive.objects.get(uuid=alarm_id)
								log.error("Found in archived alarms")
							except AlarmArchive.DoesNotExist, e:
								try:
									db_alarm = Alarm.objects.get(uuid=alarm_id)
									log.error("Found in current alarms")
								except Alarm.DoesNotExist, e:
									pass

							if db_alarm is None:
								log.error("Digested alarm %s not found in db", alarm_id)
								continue

							db_alarm_data = db_alarm.get_data()
							log.debug("\t\tpath: %s", alarmdata.get('condition',{}).get('path', None))

							if alarmdata['condition']['path'] == db_alarm_data['condition']['path']:
								if db_alarm_data.get('result',{}).get('status', None) == CSTATE_DN:
									log.debug("\t\tCondition related to alarm found in digest. Do not ignore the current alarm")
									ignore = False
									break

					else:
						log.warning("Digest exists, but no data. Strange")
						ignore = True
				else:
					log.info("No digest created yet")
					ignore = True

				if ignore:
					log.info("Alarm ignored. Finish")
					return ROUTING_OUTCOME_CONTINUE

			#
			# L'allarme deve essere raggruppato in un digest.
			#
			created   = False
			tdata, created = self._load_digest_data(log, tenant, digest_id, create_if_new=True)

			log.info("Digest %s created %s", digest_id, created)

			digest_data = tdata.get_data()
			if digest_data == None:
				digest_data = { 'alarms': [] }
			digest_data['alarms'].append(alarm.uuid)
			tdata.set_data(digest_data)

			self._save_digest_data(log, tenant, digest_id, tdata)


			alarm.merge_group_key = digest_id

			if created:
				unhold_handler = self.name

				log.debug(" HOLDING")
				alarm.set_hold_state( start_ts, unhold_ts, hold_handler=unhold_handler )

				log.debug("CURRENT RULESTACK: %s", alarm.rulestack)

				outcome = ROUTING_OUTCOME_HOLDNEXT
			else:
				log.debug(" MERGED")
				alarm.status = ALARM_STATUS_MERGED

				outcome = ROUTING_OUTCOME_MERGED

			return outcome

		except Exception, e:
			digest_id = None
			msg = "Error while handling digest data for alarm %s: %s" % (alarm.uuid, str(e))
			log.error("%s: %s", self.name, msg )

			alarm.set_verbstatus(msg)

			alarm.status = ALARM_STATUS_ERROR
			outcome      = ROUTING_OUTCOME_ERROR

		return outcome


	def on_unhold_alarm(self, context, alarm):
		super(Target,self).on_unhold_alarm(context, alarm)

		log = context.log

		tenant = Tenant.objects.get(uuid=alarm.tenant_uuid)
		db_set_current_tenant_schema(tenant.name)

		digest_id  = alarm.merge_group_key

		tdata, created = self._load_digest_data(log, tenant, digest_id)

		data = tdata.get_data()

		log.debug("DIGEST TO UNHOLD: %s: %s", digest_id, data)

		# Calcoliamo informazioni sul digest
		conditions_changes_by_path , errors = get_infos_from_digest_data(alarm.data, data)

		downs_count = Condition.objects.select_related('element', 'element__parent').filter(path__in=conditions_changes_by_path.keys()).filter(status__in=CSTATE_MACRO_DOWN).count()

		# Salviamo informazioni nel messaggio per poter fare match
		# vari ed eventuali sul digest
		alarm.data['isdigest'   ] = 1
		alarm.data['digest_data'] = data
		alarm.data['digestinfo'] = {
			'downs'     : downs_count,
			'conditions': len( conditions_changes_by_path.keys() ),
		}

		log.debug("Removing digest data from db")
		tdata.delete()

		return ROUTING_OUTCOME_CONTINUE

	def _load_digest_data(self, log, tenant, digest_id, create_if_new=False):
		log.debug("Loading digest data: %s (create if new=%s)", digest_id, create_if_new)
		db_set_current_tenant_schema(tenant.name)

		if create_if_new:
			log.debug("Loading digest data: %s", digest_id)
			tdata, created = TargetData.objects.get_or_create(tenant=tenant, key=digest_id)
		else:
			try:
				tdata = TargetData.objects.get(tenant=tenant, key=digest_id)
				created = False
			except TargetData.DoesNotExist, e:
				tdata   = None
				created = False

		return tdata, created

	def _save_digest_data(self, log, tenant, digest_id, tdata):
		db_set_current_tenant_schema(tenant.name)
		log.debug("Saving digest data: %s", digest_id)
		tdata.save()
		return



def get_infos_from_digest_data(main_alarm, digest_data):

		from entables.models import AlarmArchive
		from core.models import Condition, Tenant

		errors = []

		conditions_changes_by_path = dict()

		for alarm_id in digest_data['alarms']:

			print alarm_id

			try:
				if alarm_id == main_alarm['uuid']:
					alarm = main_alarm
				else:
					db_alarm = AlarmArchive.objects.get(uuid=alarm_id)
					alarm = db_alarm.get_data()
			except Exception, e:
				msg = str(e)
				errors.append(msg)
				continue

			try:
				condition_path = alarm['condition']['path']
				conditions_changes_by_path[ condition_path ] = conditions_changes_by_path.get( condition_path, conditions_changes_by_path.setdefault(condition_path,0) ) + 1

			except KeyError, e:
				msg = str(e)
				errors.append(msg)
				continue

		return conditions_changes_by_path, errors


def get_alarms_by_condition_from_digest(main_alarm, digest_data):

	from entables.models import AlarmArchive
	from core.models import Condition, Tenant

	errors = []

	conditions_alarms_by_path = dict()

	for alarm_id in digest_data['alarms']:

		try:
			if alarm_id == main_alarm.data['uuid']:
				db_alarm = main_alarm
				alarm_data = main_alarm.data
			else:
				db_alarm = AlarmArchive.objects.get(uuid=alarm_id)
				alarm_data = db_alarm.get_data()

			condition_path = alarm_data['condition']['path']

			events = conditions_alarms_by_path.get( condition_path, conditions_alarms_by_path.setdefault(condition_path, [] ) )
			events.append(alarm_data) 

		except KeyError, e:
			msg = str(e)
			errors.append(msg)
			continue

		except Exception, e:
			msg = str(e)
			errors.append(msg)
			continue

	#
	# Gli allarmi del digest li ordiniamo in ordine temporale
	#
	from operator import itemgetter

	for path in conditions_alarms_by_path:

		events = conditions_alarms_by_path[path]

		events = sorted(events, key=itemgetter('ts'))

		conditions_alarms_by_path[path] = events

	return conditions_alarms_by_path, errors

