# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

.. _target-JSON_HTTP:

JSON_HTTP
========================================================================

Questo target permette l'invio dei dati di un allarme codificati in formato JSON tramite protocollo HTTP ad una URL remota:

I parametri questo target vanno specificati secondo la sintassi: 

::
	
	-J JSON_HTTP --json_http_<parametro> <valore> [ --smtp_<parametro> <valore> ... ]

Parametri principali:

	+---------------+---------------+-------------------------------+-----------------------------------------------------------------------------------------------------------------------------------------------+
	| Parametro     | Type          | Default                       | Descrizione                                                                                                                                   |
	+===============+===============+===============================+===============================================================================================================================================+
	| url           | stringa       |                               | Url della richiesta                                                                                                                           |
	+---------------+---------------+-------------------------------+-----------------------------------------------------------------------------------------------------------------------------------------------+
	| method        | stringa       | post                          | Metodo HTTP: GET o POST                                                                                                                       |
	+---------------+---------------+-------------------------------+-----------------------------------------------------------------------------------------------------------------------------------------------+
	| timeout       | stringa       | 5                             | Request timeout                                                                                                                               |
	+---------------+---------------+-------------------------------+-----------------------------------------------------------------------------------------------------------------------------------------------+
	| headers       | stringa       |                               | Stringa con header HTTP addizionali da aggiungere alla richiesta.                                                                             |
	+---------------+---------------+-------------------------------+-----------------------------------------------------------------------------------------------------------------------------------------------+
	| rawdata       | intero        | 0                             | Quando vale 1, i dati vengono inviati in formato JSON con Content-type: application/json. I parametri 'params' e 'alarmparm' vengon ignorati. |
	+---------------+---------------+-------------------------------+-----------------------------------------------------------------------------------------------------------------------------------------------+

Parametri per GET/POST normali:

	+---------------+---------------+-------------------------------+--------------------------------------------------------------------------+
	| Parametro     | Type          | Default                       | Descrizione                                                              |
	+===============+===============+===============================+==========================================================================+
	| params        | stringa       |                               | Query string parametri da inviare insieme ai dati                        |
	+---------------+---------------+-------------------------------+--------------------------------------------------------------------------+
	| alarmparam    | stringa       | alarm                         | Nome variabile che conterra i dati dell'allarme codificati in JSON       |
	+---------------+---------------+-------------------------------+--------------------------------------------------------------------------+

Esempi
------------------

In base a questa regola:

        ::
                
                entables -t notify -A DEFAULT \
                         -j JSON_HTTP \
                         --json_http_url        http://127.0.0.1:8000/test \
                         --json_http_method     POST \
                         --json_http_params     'param1=value1&param2=value2 with space' \
                         --json_http_alarmparam data

Viene generata una post con i seguenti dati:

        ::
                
                param1=value1
                param2=valu2 with space
                data={....alarm json.....}


Struttura dati JSON
-----------------------

La struttura dati (codificata in formato JSON e' la seguente):

        ::
                
                                                                                                      Presente?     NULL?  Descrizione
                {
                     "uuid"      : "7eb499dd9fb54f0a87afd351d0c4ce56"                                               NO     identificativo univoco dell'allarme.
                     "ts"        : 1511001368,                                                                      NO     timestamp UTC di generazione dell'allarme.
                     "element": {                                                                                   
                         "uuid"       : "e173f52135b845a5bf1792e33ad508ac"                                          NO     identificativo univoco dell'elemento associato all'allarme.
                         "path"       : "dm10lx03:rootfs",                                                          NO     identificativo univoco mnemonico dell'elemento associato all'allarme.
                         "name"       : "rootfs",                                                                   NO     nome dell'elemento.
                         "description": "Root FS",                                                                  NO     descrizione dell'elemento.
                         "type"       : 3,                                                                          NO     tipo dell'elemento (1=nodo,2=interfaccia,3=storage,4=service,5=device).
                         "parent_id"  : "a12937d61c874890914182b6d25ebbb8",                                         SI     identificativo dell'elemento padre (l'uuid del nodo).
                         "parent_name": "dm10lx03",                                                                 SI     identificativo univoco dell'elemento padre (il nome del nodo).
                     },
                     "condition": {                                                                      
                         "uuid"            : "050db78a671e48358ca266a92e9c9682"                                     NO     identificativo univoco della condition che ha generato l'allarme.
                         "path"            : "dm10lx03:rootfs;storage;storage;check-storageperc",                   NO     identificativo mnemonico.
                         "name"            : "check-storageperc",                                                   NO     nome della condition.
                         "description"     : "Check occupazione filesystem root",                                   SI     descrizione della condition.
                         "classification"  : "system.storage.usage",                                                NO     classificazione della condition.
                         "priority"        : 60,                                                                    NO     importanza/priorita' associata alla condition.
                         "priority_level"  : "medium",                                                              NO     classe di importanza/priorita' associata alla condition.
                         "statuslastchange": 1511001368,                                                            NO     timestamp UTC della variazione di stato precedente della condition.
                     },
                     "datasources": {                                                                 OPZIONALE            serie dati utilizzate per generare l'allarme.
                         ...
                         "available": {                                                                             NO     nome della serie.
                             "id"   : "09298c8603974662bf067f451bca0670",                                           NO     identificativo univoco della serie.
                             "value": True                                                                          NO     valore della serie al momento dell'allarme.
                         },
                         ...
                     },
                     "result": {                                                                                           risultati del check.
                         "status"    : 10,                                                                          NO     stato della verifica della condition.
                         "laststatus": 20,                                                                          NO     stato precedente.
                     },
                     "mail": {                                                                                             Dati della mail generata da Sanet.
                         "subject": "Disco pieno"                                                                   NO     Subject 
                         "body"   : "......."                                                                       NO     Body 
                     },
                     "info": {                                                                        OPZIONALE            informazioni dell'elemento.
                         "ip4s"       : ['10.0.25.201'],                                                            NO     IP del nodo. Puo' essere [].
                     },
                     "tags": [                                                                                      NO     classificazioni/tag assegnati alla condizione o all'elemento associati all'allarme. Puo' essere [].
                         "dmh:/DMH/Sistemi Critici/3.Labs"
                     ],
                     "cluster": {                                                                     OPZIONALE     NO     opt. cluster al quale l'elemento e' associato.
                         "name"       : ['VMWare-Produzione'],                                                      NO
                     },
                     "meta": {                                                                                             opt. informazioni addizionali dell'elemento.
                         ...
                         "referenti" : [                                                              OPZIONALE            referenti associati all'elemento. Puo' essere [].
                            'Sergio Rizzi <sergio.rizzi@labs.it>',                         
                            'Tiziano Fogli <tiziano.fogli@labs.it>'
                         ]
                         ...
                     },
                }
"""

import types
import json
import datetime
import traceback

from urllib import urlencode
import urllib2 
from urlparse import parse_qs
import ssl
import socket

from core.models   import Tenant, MonitoredElement, Condition
from core.api.base import db_set_current_tenant_schema
from core.api.cache import init_cache_manager_from_conf

from tags.models   import ElementTag, Tag, TagMetaData

from entables.constants import TABLE_NOTIFY

from entables.constants import ROUTING_OUTCOME_CONTINUE

from entables.targets import BaseTargetHandler
from entables.targets import Option

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class Target(BaseTargetHandler):

	def __init__(self):
		super(Target, self).__init__()

		self.name = 'json_http'

		self.tables = [
			TABLE_NOTIFY,
		]

		self.options    = [
			Option("url"          , type=str, required=True),
			Option("method"       , type=str, default="post"),

			Option("timeout"      , type=int, default="5"),

			Option("headers"      , type=str),
			Option("rawdata"      , type=int, default="0"),

			Option("params"       , type=str),
			Option("alarmparam"   , type=str, default="alarm"),
		]

		self.counters = {
			'sent' : 0,
			'error': 0,
		}

	def init(self):
		return

	def get_text_info(self):
		text = super(Target, self).get_text_info()
		text += ',' + ','.join([ "%s:%s" % (c,v) for c,v in self.counters.items()])
		return text

	def validate_target_options(self, options_dict):

		super(Target, self).validate_target_options(options_dict)

		defaults = self.get_options_default()

		if not options_dict.has_key('url'):
			raise Exception("Parameter 'url' is required")

		if options_dict.has_key('method'):
			if (options_dict['method'].lower() not in ['get', 'post']): raise Exception("Invalid method: '%s'" % ( options_dict['method'] ))

		if options_dict.has_key('params'):
			try:
				d = parse_qs( options_dict['params'] )
			except Exception, e:
				raise Exception("Invalid encoded params: '%s'" % ( options_dict['params'] ))

		if options_dict.has_key('alarmparam'):
			if not options_dict['alarmparam']:
				raise Exception("Invalid value for parameter 'alarmparam': '%s'" % ( options_dict['alarmparam'] ))

		if options_dict.has_key('timeout'):
			try:
				v = int(options_dict['timeout'])
			except Exception, e:
				raise Exception("Invalid value for parameter 'timeout': '%s'" % ( options_dict['timeout'] ))


		if options_dict.has_key('rawdata'):
			try:
				v = int(options_dict['rawdata'])
			except Exception, e:
				raise Exception("Invalid value for parameter 'rawdata': '%s'" % ( options_dict['rawdata'] ))

		if options_dict.has_key('headers'):
			try:
				headers = options_dict['headers']
				for line in headers.split("\n"):
					key, value = line.split(": ", 1)
			except Exception, e:
				raise Exception("Invalid value for parameter 'headers': '%s'" % ( options_dict['headers'] ))


		if options_dict.has_key('rawdata') and (options_dict.has_key('alarmparam') or options_dict.has_key('params')):
			raise Exception("Can not use option 'rawdata' with 'alarmparam' or 'params'")

		return True

	def dispatch(self, context, alarm, ruleobj, target_options):

		context.log.debug("JSON_HTTP")

		super(Target, self).dispatch(context, alarm, ruleobj, target_options)

		options = self.expand_options(context, alarm, target_options)

		alarm.set_verbstatus("Target %s" % (self.name,))

		try:
			alarmdata = alarm.data
			self.process_alarm_and_dispatch(context, alarm, alarmdata, ruleobj, options)

		except Exception, e:
			msg = str(e)
			context.log.error(msg)
			alarm.set_verbstatus(msg)

		return ROUTING_OUTCOME_CONTINUE	

	def process_alarm_and_dispatch(self, context, alarm, alarmdata, ruleobj, options):

		url = options['url']

		#
		# Handle params
		#
		headers           = options.get('headers', None)

		timeout           = int(options['timeout'])

		rawdata           = int(options.get('rawdata', 0))

		alarmparam_name   = options.get('alarmparam', None)

		if options.has_key('params'):
			additional_params = parse_qs( options['params'] )
		else:
			additional_params = {}

		for name, value in additional_params.items():
			if isinstance(value, types.ListType):
				if len(value) == 1:
					additional_params[name] = value[0]

		#
		# Create JSON 
		#
		output_alarm = self.create_output_alarm_json( alarm, alarmdata )

		#
		# Prepare request
		#

		if rawdata:
			post_data = json.dumps(output_alarm)
		else:

			if not alarmparam_name:
				alarmparam_name = 'alarm'

			post_data = {}

			post_data.update( additional_params )

			post_data[ alarmparam_name ] = json.dumps(output_alarm)

			post_data = urlencode(post_data)

		#
		# DO REQUEST
		#
		try:
			req = urllib2.Request(url, post_data)

			if rawdata:
				req.add_header('Content-type' , 'application/json')

			# Additional headers
			if headers:
				for header in headers.split("\n"):
					key, value = header.split(": ", 1)
					req.add_header(key, value)

			# HACK: ignore problems with certificates.
			#gcontext = ssl.SSLContext(ssl.PROTOCOL_TLSv1)
			gcontext = ssl._create_unverified_context()

			response = urllib2.urlopen(req, context=gcontext, timeout=timeout)

			#
			# READ RESPONSE
			#
			code          = response.getcode()
			response_data = response.read()

			context.log.debug("Http response code: %s", code)
			context.log.debug("Http response data: %s", response_data)

		except socket.timeout, e:
			# For Python 2.7
			context.log.error("json_http: request timeout to %s", url)

		except Exception, e:
			# For Python 2.7
			context.log.error("json_http: unknown error: %s:%s", type(e), str(e))

		return ROUTING_OUTCOME_CONTINUE

	def create_output_alarm_json(self, alarm, alarmdata):

		cache_manager = init_cache_manager_from_conf()

		"""
		RAW DATA"
			
			{
				u'alarmtype': u'normal',
				u'condition': {
					u'classification': u'',
					u'expr': u"int(readFile('/tmp/prova.txt')) == (int(time()) % 2)",
					u'maxtries': 1,
					u'name': u'c',
					u'path': u'localhost;;dgfile;c',
					u'priority': 10,
					u'priority_level': u'low',
					u'statuslastchange': 1528721792.806303,
					u'uuid': u'7386a7df94b949bd845240ec86fd005b'
				},
				u'datasources': {
				},
				 u'element': {
					u'uuid': u'c61ed1a27be04db698624959a0ce4a63'
					u'path': u'localhost',
					u'name': u'localhost',
					u'kind': 1,
					u'nodedescr': u'',
					u'parent_id': None,
					u'parent_name': None,
					u'type': 1,
				},
				 u'event': {
					u'uuid': u'7f73e563aeb54c63b2d8253a5bdedb83'
				},
				u'flow-information': [],
				u'info': {
					u'ip4s': [
						u'127.0.0.1'
					],
					u'sysdescr': u'Linux debian9 4.9.0-3-amd64 #1 SMP Debian 4.9.30-2+deb9u5 (2017-09-19) x86_64',
					u'syslocation': u'debian9',
					u'sysname': u'debian9'
				},
				u'linked': {
					u'iface': None, 
					u'node': None
				},
				u'mail': {
					u'body': u'UP',
					u'headers': {
						u'X-SANET-Priority': 10,
						u'X-SANET-Target-Id': u'localhost;;dgfile;c',
						u'X-SANET-Target-Status': u'UP'
					},
					u'subject': u'UP',
					u'to': u'netwarning'
				},
				u'params': {},
				u'pushed_data': None,
				u'result': {
					u'laststatus': 10,
					u'status': 100,
					u'verbstatus': u'Status changed by administrator'
				},
				u'saved_variables': None,
				u'tags': [u'prova:/prova/2'],
				u'tags_name': {
					u'prova': [u'2']
				},
				u'tenantid': u'c620e819a12249df8771d5919556f280',
				u'tenantname': u'site',
				u'ts': 1528721792,
				u'url': u'https://debian9/sanet3/web/tenantpage/c620e819a12249df8771d5919556f280/node/c61ed1a27be04db698624959a0ce4a63#techincal',
				u'uuid': u'7e8c8d7a8d0c4f259a546c584277912d'
			}

		FORMATO CONCORDATO


		"""

		me = MonitoredElement.objects.get(uuid=alarmdata['element'  ]['uuid'])
		c  = Condition.objects.get(uuid=alarmdata['condition']['uuid'])

		return_data = {
			"uuid"      : alarmdata['uuid'],
			"ts"        : alarmdata['ts'],                                                  #    *timestamp UTC**di generazione dell'allarme*
		}

		#
		# ELEMENT
		#
		return_data.update({
			"element"    : alarmdata.get('element', {}), #{
			"description": me.description,
			#	 "uuid"       : "e173f52135b845a5bf1792e33ad508ac"                        #*identificativo univoco dell'elemento associato all'allarme*
			#	 "path"       : "dm10lx03:rootfs",                                        #*identificativo univoco mnemonico dell'elemento associato all'allarme*
			#	 "name"       : "rootfs",                                                 #*nome dell'elemento*
			#	 "description": "Root FS",                                                #*descrizione**dell'elemento*
			#	 "type"       : 3,                                                        #*tipo dell'elemento (1=nodo,2=interfaccia,3=storage,4=service,5=device)*
			#	 "parent_id"  : "a12937d61c874890914182b6d25ebbb8",                       #*identificativo dell'elemento padre (l'uuid del nodo)*
			#	 "parent_name": "dm10lx03",                                               #*identificativo univoco dell'elemento padre (il nome del nodo)*
			# }
		})

		#
		# DATASOURCES
		#
		return_data.update({
			"datasources": alarmdata.get('datasources',[]) # {                                                                 #*opt. serie dati utilizzate per generare l'allarme*
			#	 "available": {                                                           #*nome della serie*
			#		 "id"   : "09298c8603974662bf067f451bca0670",                     #*identificativo univoco della serie*
			#		 "value": True                                                    #*valore della serie al momento dell'allarme*
			#	 },
			#	 "installed": {
			#		 "id"   : "d3d0f7a84891485b8558c8fdb87e3a20",
			#		 "value": 2199020105728
			#	 },
			#	 "used": {
			#		 "id"   : "460abc40cfac456e8026ed7e13588771",
			#		 "value": 2155118809088
			#	 }
			#}
		})


		return_data.update({
			"condition"  : alarmdata.get('condition',{}), # {
			#"description": c.description,
			#	 "uuid"            : "050db78a671e48358ca266a92e9c9682"                   #*identificativo univoco della condition che ha generato l'allarme*
			#	 "path"            : "dm10lx03:rootfs;storage;storage;check-storageperc", #*identificativo mnemonico*
			#	 "name"            : "check-storageperc",                                 #*nome della **condition*
			#	 "description"     : "Check occupazione filesystem root",                 #*descrizione**della condition*
			#	 "classification"  : "system.storage.usage",                              #*classificazione della **condition*
			#	 "priority"        : 60,                                                  #*importanza/priorita' associata alla **condition*
			#	 "priority_level"  : "medium",                                            #*classe di importanza/priorita' associata alla **condition*
			#	 "statuslastchange": 1511001368,                                          #*timestamp ***UTC**della *variazione di stato precedente della **condition*
			#},
		})

		try:
			c = Condition.objects.get(uuid=return_data['condition']['uuid'])

			return_data['condition']['description'] = c.expanded_title(cache_manager=cache_manager)
		except Exception, e:
			return_data['condition']['description'] = None

		#
		# RESULT
		#
		return_data.update({
			"result": alarmdata['result'] #{                                                                      *risultati del check*
			#	 "status"    : 10,                                                        *stato della verifica della condition*
			#	 "laststatus": 20,                                                        *stato precedente*
			# },
		})

		#
		# MAIL
		#
		return_data.update({
			"mail": alarmdata['mail']
		})

		#
		# INFOS
		#
		return_data.update({
			"info": {                                                                        #*informazioni dell'elemento*
				"ip4s"       : alarmdata.get('info',{}).get('ip4s',[]), # ['10.0.25.201'],
			},
		})

		#
		# TAGS
		#
		return_data.update({
			"tags": alarmdata.get('tags',[]) #[                                                                        #*opt. **classificazioni/tag assegnati alla condizione o all'elemento associati all'allarme*
			#	"dmh:/DMH/Sistemi Critici/3.Labs"
			#],
		})

		#
		# CLUSTER
		#
		if me.parent:
			me = me.parent

		cluster = me.cluster
		if cluster:
			return_data.update({
				"cluster": cluster.name #{                                                                     #***opt. *cluster al quale l'elemento e' associato*
				#	#"name"       :  ,
				#}
			})

		#
		# META
		#
		tag_metas = self.extract_meta(alarm, alarmdata)
		return_data.update({
			"meta": tag_metas #{                                                                        #****opt. *informazioni addizionali dell'elemento*
			#	"referenti" : ['Sergio Rizzi <sergio.rizzi@labs.it>\nTiziano Fogli <tiziano.fogli@labs.it>'], #*referenti associati all'elemento*
			#},
		})

		#import pprint
		#pprint.pprint(return_data)

		return return_data



	def extract_meta(self, alarm, alarmdata):

		ids = [
			alarmdata['element'  ]['uuid'],
			alarmdata['condition']['uuid'],
		]

		tenant = Tenant.objects.get(uuid=alarm.tenant_uuid)
		db_set_current_tenant_schema(tenant.name)

		tagged = ElementTag.objects.filter(elementid__in=ids)

		tagged = tagged.all()

		tags = [et.tag for et in tagged]

		metas = TagMetaData.objects.filter(tag__in=tags)
		metas = metas.all()

		by_name = {}
		for md in metas:
			l = by_name.get( md.name, by_name.setdefault( md.name, [] ))
			l.append(md.value)

		vals = by_name

		return vals


