# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

SETATTR
=======================

Aggiunge al pacchetto dati di un allarme un campo valorizzato in base ai parametri specificati

I parametri questo target vanno specificati secondo la sintassi: 

::
	
	-J SETATTR --setattr_field_name <nome>  --setattr_<parametro opzionale> <valore>

Parametri obbligatori:

	+---------------+---------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
	| Parametro     | Type    | Descrizione                                                                                                                                                                  |
	+===============+=========+==============================================================================================================================================================================+
	| field_name    | stringa | Nome del nuovo campo da aggiungere al messaggio                                                                                                                              |
	+---------------+---------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+

E' possibile specificare **uno solo** dei seguenti parametri opzionali:

	+---------------+---------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
	| Parametro     | Type    | Descrizione                                                                                                                                                                                                |
	+===============+=========+============================================================================================================================================================================================================+
	| value         | stringa | Salva nel messaggio un nuovo campo *field_name* con il valore indicato (stringa libera)                                                                                                                    |
	+---------------+---------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
	| treetags      | stringa | Salva nel messaggio un nuovo campo *field_name* con tutti i nomi dei tag relativi all'allarme appartenenti al tree indicato                                                                                |
	+---------------+---------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
	| tagmeta       | stringa | Salva nel messaggio un nuovo campo *field_name* con la stringa contenente tutti i metadati con nome *tagmeta* estratto da tutti i tag associatil al nodo/condition dell'allarme. *tagmeta* puo' essere '*' |
	+---------------+---------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+

Esempi
------

Esempio 1: Aggiungiamo al messaggio un campo *custom* con un valore (per essere processato in qualche altra regola) 

::
	
	entables -t notify -A DEFAULT   -J SETATTR --setattr_field_name 'foo'        --setattr_value '12345' 


Esempio 2: I messaggi nel weekend vengono ritardati fino al lunedi' mattina: 

::
	
	entables -t notify -A DEFAULT   -J SETATTR --setattr_field_name 'foo'        --setattr_treetags 'geo' 


Esempio 3: salva un campo "indirizzi" con il valore del metadato "referenti" dei tag associati al nodo/condition dell'allarme:

::
	
	entables -t notify -A DEFAULT   -j SETATTR --setattr_field_name  indirizzi   --setattr_tagmeta referenti
	                                
	entables -t notify -A DEFAULT   -j SMTP    --smtp_subject '${mail_subject}' --smtp_message '${mail_body}'    --smtp_rcpt_to '${tags['"'"'alarmpath:indirizzi'"'"']}'

"""

import time
import datetime


from entables.constants import *

from entables.targets import BaseTargetHandler
from entables.targets import Option

from entables.constants import TABLE_MANGLE

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class Target(BaseTargetHandler):

	def __init__(self):
		super(Target, self).__init__()
		
		self.name = "setattr"
		
		self.tables = [ 
			TABLE_MANGLE
		]

		self.options = [
			Option("field_name", type=str, required=True), 
			Option("value"     , type=str, required=False),
			Option("treetags"  , type=str, required=False),
			Option("tagmeta"   , type=str, required=False),
		]


	def init(self):
		return

	def validate_target_options(self, options_dict):

		super(Target, self).validate_target_options(options_dict)
		
		field_name = options_dict.get('field_name', None) or None

		value      = options_dict.get('value'     , None)          # puo' essere stringa vuota (None = parametro non specificato a riga di comando)
		treetags   = options_dict.get('treetags'  , None) or None
		tagmeta    = options_dict.get('tagmeta'   , None) or None
		
		if field_name is None:
			raise Exception("Missing parameter: field_name")
		
		if value is not None:
			if (treetags or tagmeta):
				raise Exception("Can not specify 'value' option with 'treetrags' or 'tagmeta'")
		else:
			if (treetags is None and tagmeta is None):
				raise Exception("Missing parameter: treetags or tagmeta")

			if (treetags and tagmeta):
				raise Exception("Specify only one parameter: treetags or tagmeta")
		
		return True

	def dispatch(self, context, alarm, ruleobj, target_options):

		options = self.expand_options(context, alarm, target_options)

		log = context.log

		log.debug("SETATTR DISPATCH")

		alarm.set_verbstatus("Target %s: %s" % ( self.name, options))
		
		alarmdata = alarm.data

		try:
			field_name = options['field_name']

			value = options.get('value', None)
			if value is not None:
				alarmdata[ field_name ] = value
			
			treetags_name = options.get('treetags', None)
			if treetags_name is not None:
				names = alarmdata['tags_name'].get(treetags_name, [])
				alarmdata[ field_name ] = ','.join(names)

			tagmeta_name = options.get("tagmeta", None)
			if tagmeta_name is not None:

				log.debug("Reading metadata...")

				log.debug("\tElement   id: %s", alarmdata['element'  ]['uuid'])
				log.debug("\tCondition id: %s", alarmdata['condition']['uuid'])

				ids = [ 
					alarmdata['element'  ]['uuid'], 
					alarmdata['condition']['uuid']
				]

				from core.models import Tenant
				from core.api.base import db_set_current_tenant_schema
				from tags.models import ElementTag, Tag, TagMetaData

				tenant = Tenant.objects.get(uuid=alarm.tenant_uuid)
				db_set_current_tenant_schema(tenant.name)

				tagged = ElementTag.objects.filter(elementid__in=ids)
				#if tagmeta_name != "*":
				#	tagged = tagged.filter(tree__name=tagmeta_name)
				tagged = tagged.all()

				log.debug("\tTagged elements: %s", len(tagged))

				tags = [et.tag for et in tagged]
				log.debug("\tTags %s", tags)
					
				metas = TagMetaData.objects.filter(tag__in=tags)
				if tagmeta_name != '*':
					metas = metas.filter(name=tagmeta_name)
				metas = metas.all()
				
				vals = []
				for md in metas:
					log.debug("\t\tMeta %s = %s", md.name, md.value)
					vals.append(md.value)

				alarmdata[ field_name ] = '\n'.join(vals)
				
			
			log.debug("Set field '%s' = %s", field_name, alarmdata[ field_name ])

			return ROUTING_OUTCOME_CONTINUE
			
		except Exception, e:
			log.debug("%s", str(e))
			
			alarm.set_verbstatus("Unable to execute target %s: %s:%s" % ( self.name, type(e), str(e)))
			
			return ROUTING_OUTCOME_CONTINUE
