# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""


.. _entables-targets-silence:

SILENCE
==========

Ritarda un uno o piu' allarmi fino all'orario indicato. 

Di tutti gli allarmi generati dalla stessa condition e che vengono ritardati allo stesso istante temporale, viene fatto "ripartire" solo l'ultimo.
Questo comportamento e' leggermente modificabile attraverso parametri speciali ( :ref:`entables-targets-silence-special-params` ).

I parametri questo target vanno specificati secondo la sintassi: 

::
	
	-J SILENCE --silence_<parametro> <valore> [ --silence_<parametro> <valore> ... ]


Parametri:

	+-----------------+---------+----------+-------------------------------------------------------------------------------------------------------------------------+
	| Parametro       | Type    | Default  | Descrizione                                                                                                             |
	+=================+=========+==========+=========================================================================================================================+
	| holdfor         | numero  |          | Numero di secondi da attendere prima di sbloccare il messaggio                                                          |
	+-----------------+---------+----------+-------------------------------------------------------------------------------------------------------------------------+
	| holduntil       | stringa |          | Data di scadenza di quando sbloccare il messaggio.                                                                      |
	+-----------------+---------+----------+-------------------------------------------------------------------------------------------------------------------------+
	| holduntil_next  | stringa |          | Prossimo orario valido (del giorno stesso o giorno successivo)                                                          |
	+-----------------+---------+----------+-------------------------------------------------------------------------------------------------------------------------+

I parametri sono equivalenti al target :ref:`entables-targets-delay`. 

Si rimanda alla sezione :ref:`entables-targets-delay-parameters` per i dettagli sulla sintassi e semantica dei parametri.


.. _entables-targets-silence-special-params:

Parametri speciali
---------------------

La logica di filtro del target puo' essere alterata attraverso due parametri:

	+-----------------+---------+----------+------------------------------------------------------------------------------------------------------------------------------------------------------+
	| Parametro       | Type    | Default  | Descrizione                                                                                                                                          |
	+=================+=========+==========+======================================================================================================================================================+
	| drop_last_up    | numero  | 0        | Quando vale 1, gli allarmi che dovrebbero ripartire vengono DROPPATI se sono relativi a stati UP.                                                    |
	+-----------------+---------+----------+------------------------------------------------------------------------------------------------------------------------------------------------------+
	| notify_first_up | numero  | 0        | Quando vale 1, di tutti gli allarmi di una stessa condition "intercettati", non viene gestito (e quindi non viene ritardato) il primo UP.            |
	+-----------------+---------+----------+------------------------------------------------------------------------------------------------------------------------------------------------------+

Il parametro *notify_first_up* serve per decidere come gestire allarmi UP relativi a condition che sono andate DOWN **prima** che si cominciasse a filtrare gli allarmi con una regola di SILENCE.

Esempio:

::
    
         DOWN        UP      DOWN    UP      DOWN       
    -----v-----|-----^-------v-------^-------v---------------------|--------------->
             00:00                                                6:00
    
               |---------------holduntil_next 6:00---------------->

In questa situazione, il primo allarme UP viene "intercettato" e ignorato se non viene specificato altrimenti (valorizzando ad 1 il parametro *notify_first_up*).

.. warning:: Se viene intercettato un solo allarme UP per una determinata condition, se non si utilizza il parametro *notify_first_up* si rischia di non ricvere nessun allarme.



Il parametro *drop_last_up* server per decidere di *ignorare* situazioni di *guasto* che sono gia' *rientrate*.

Esempio:

::
    
                     DOWN    UP      DOWN    UP    DOWN   UP
     ----------|-----v-------^-------v-------^-----v------^--------|--------------->
             00:00                                                6:00
     
               |---------------holduntil_next 6:00---------------->


In questa situazione, alle 6:00 non vogliamo ricevere l'ultimo allarme (UP) perche' la segnalazione non e' *significativa*.

.. warning:: Se viene intercettato un solo allarme UP per una determinata condition, utilizzando il parametro *drop_last_up* si rischia di non ricvere nessun allarme.



Esempi
------

I messaggi nel weekend vengono ritardati fino al lunedi' mattina: 

::
	
	entables   -A DEFAULT                                                   -j SILENCE --silence_holduntil      'monday 08:00'
	entables   -A DEFAULT                                                   -j SILENCE --silence_holduntil      'monday 08:00'     --silence_drop_last_up 1       --silence_notify_first_up 1
	
	entables   -A DEFAULT  -m tags  --tags_path sedi_remote.*               -j SILENCE --silence_holduntil      'monday 08:00'
	
	entables   -A DEFAULT  -m times --times_range  0-24/6-7                 -j SILENCE --silence_holduntil      'monday 08:00'
	entables   -A DEFAULT  -m times --times_range  00:00-8:00,18:00-24/1-5  -J SILENCE --silence_holduntil_next '08:00'
	entables   -A DEFAULT  -m times --times_range  00:00-8:00,18:00-24/1-5  -J SILENCE --silence_holduntil_next '08:00' --delay_drop_if_up 1

"""

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

"""

TEST
==================v======================================================



Testo drop last up
----------------------

entables -F DEFAULT
./scripts/debugging/launch_entables.sh 

sanet-manage force_dg_status  'xha;;icmp-reachability'         --status DN



NEXT=$(date -d 'today + 1 minutes' '+%H:%M')
echo "NEXT:" $NEXT
entables -F DEFAULT
entables -A DEFAULT  -j SILENCE --silence_holduntil_next $NEXT --silence_drop_last_up 1
entables -A DEFAULT  -j SYSLOG  --syslog_msg '${alarmts} ${alarmid} ${path} ${status}'
./scripts/debugging/launch_entables.sh 

sanet-manage force_dg_status  'xha;;icmp-reachability'         --status DN
sanet-manage force_dg_status  'xha;;icmp-reachability'         --status UP



Test notify_first_up
--------------------------

entables -F DEFAULT
./scripts/debugging/launch_entables.sh 

sanet-manage force_dg_status  'xha;;icmp-reachability'         --status DN
sanet-manage force_dg_status  'localhost;;icmp-reachability'   --status DN





NEXT=$(date -d 'today + 1 minutes' '+%H:%M')
echo "NEXT:" $NEXT
entables -F DEFAULT
entables -A DEFAULT  -j SILENCE --silence_holduntil_next $NEXT --silence_notify_first_up 1
entables -A DEFAULT  -j SYSLOG  --syslog_msg '${alarmts} ${alarmid} ${path} ${status}'
./scripts/debugging/launch_entables.sh 

sanet-manage force_dg_status  'xha;;icmp-reachability'         --status UP
sanet-manage force_dg_status  'xha;;icmp-reachability'         --status DN
sanet-manage force_dg_status  'xha;;icmp-reachability'         --status UP

sanet-manage force_dg_status  'localhost;;icmp-reachability'   --status UP
sanet-manage force_dg_status  'localhost;;icmp-reachability'   --status DN



"""

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

import json
import time
import datetime

from sanet_common.sanetdefs.constants import CSTATE_MACRO_UP
from sanet_common.timeutils import timestamp_to_isostr

from core.models import Tenant
from core.api.base import *

from entables.constants import *
from entables.constants import TABLE_NOTIFY

from entables.models import TargetData
from entables.targets import Option

from entables.targets.delay import BaseHoldTarget

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class Target(BaseHoldTarget):

	def __init__(self):
		super(Target, self).__init__()

		self.name = "silence"

		self.tables = [
			TABLE_NOTIFY, 
		]

		self.options += [
			Option("drop_last_up"      , type=str, required=False),
			Option("notify_first_up"   , type=str, required=False),
		]

	def init(self):
		return

	def validate_target_options(self, options_dict):

		super(Target, self).validate_target_options(options_dict)

		drop_last_up       = options_dict.get('drop_last_up'   , None)
		notify_first_up    = options_dict.get('notify_first_up', None)

		if drop_last_up is not None:
			try:
				drop_last_up = int(drop_last_up)
			except Exception, e:
				raise ValueError("Invalid value: %s" % (drop_last_up))

		if notify_first_up is not None:
			try:
				notify_first_up = int(notify_first_up)
			except Exception, e:
				raise ValueError("Invalid value: %s" % (notify_first_up))

		return True

	def dispatch(self, context, alarm, ruleobj, target_options):

		log = context.log

		options = self.expand_options(context, alarm, target_options)

		drop_last_up    = int(options.get("drop_last_up"   , "0"))
		notify_first_up = int(options.get("notify_first_up", "0"))


		alarmdata = alarm.data

		alarm.set_verbstatus("Target %s" %  (self.name))

		log.debug("[SILENCE] Alarm: %s %s", alarmdata['uuid'], alarmdata.get('condition', {'path':'*alarm with no condition*'}).get('path', '*NO PATH! STRANGE!*') )

		condition_path = alarmdata.get('condition', {'path':None}).get('path', None)
		alarm_status   = alarmdata.has_key("condition") and alarmdata.get('result',{}).get('status', None)


		try:
			tenant = Tenant.objects.get(uuid=alarm.tenant_uuid)

			#
			# Calcoliamo l'istante nel futuro dove dobbiamo "risvegliare" l'allarme
			#
			dispatch_ts = time.time()
			unhold_ts = self.get_unhold_timestamp(log, dispatch_ts,  options)
			
			# Questo e' l'identificativo "logico" dell'istante nel futuro "calcolato"
			target_data_id = "entables.target.silence.%s.%s" % (ruleobj.uuid, int(unhold_ts))

			#
			# Dobbiamo tenere traccia di tutti gli allarmi intercettati e ritardati tutti allo stesso istante temporale.
			# 
			# Queste informazioni vengono salvate su DB tramite un record TargetData identificato da "target_data_id".
			#
			# Per ogni condition manteniamo una lista entry con queste informazioni minime:
			#
			#     * uuid allarme
			#     * stato dell'allarme
			#
			# IN questo modo siamo in grado di elaborare la sequenza di allarmi senza andarla a rileggere dal DB.
			#

			entry = { 
				"alarm_id"    : alarm.uuid, 
				"alarm_status": alarm_status 
			}


			# Create persistent data on DB if missing
			tdata, created = self._load_digest_data(log, tenant, target_data_id, create_if_new=True)

			log.debug("[SILENCE]  Target data %s created %s", target_data_id, created)

			target_data = tdata.get_data()
			if target_data == None:
				target_data = { 
					'alarms_by_condition': {} 
				}

			# Update 
			alarms_sequence_list = target_data['alarms_by_condition'].get( condition_path, target_data['alarms_by_condition'].setdefault(  condition_path, [] ) )
			
			alarms_sequence_list.append( entry ) 

			tdata.set_data(target_data)

			# Saving
			log.debug("[SILENCE] Saving digest data: %s", target_data_id)
			tdata.save()

			#
			# La sequenza degli allarmi "visti" e' aggiornata. Dobbiamo lasciar comunque passare il primo UP visto per questa condition?
			#
			if alarm_status in CSTATE_MACRO_UP:
				# E' il primo allarme della condition ed e' un UP e devo lasciarlo passare?
				if len(alarms_sequence_list) == 1  and  notify_first_up:
					log.debug("[SILENCE] Alarm %s is first UP! Notify.", alarm.uuid)
					return ROUTING_OUTCOME_CONTINUE

			#
			# L'allarme va ritardato. Quando si svegliera' dobbiamo sapere: 
			#
			#  - a quale "gruppo" di allarmi intercettati appartiene
			#  - Se deve essere droppato nel caso non sia l'ultimo allarme visto per la propria condition.
			#
			alarmdata[ '_silence' ] = {
				'target_data_id': target_data_id,
				'drop_last_up'    : drop_last_up
			}

			#
			# HOLD
			#
			log.log(19, "[SILENCE] alarm %s hold until %s (%s)", alarm.uuid, timestamp_to_isostr(unhold_ts), unhold_ts )

			unhold_handler = self.name
			alarm.set_hold_state( dispatch_ts, unhold_ts, hold_handler=unhold_handler )

			return ROUTING_OUTCOME_HOLDNEXT

		except Exception, e:
			digest_id = None
			msg = "Error while handling digest data for alarm %s: %s %s" % (alarm.uuid, type(e), str(e))
			log.error("[SILENCE] %s", msg )

			alarm.set_verbstatus(msg)

			alarm.status = ALARM_STATUS_ERROR
			outcome      = ROUTING_OUTCOME_ERROR

		return outcome


	def on_unhold_alarm(self, context, alarm):

		# Questa chiamata dovrebbe sempre restituire ROUTING_OUTCOME_CONTINUE, ma 
		# formalmente e' piu' corretto verificare anche la "logica" della classe padre.
		outcome = super(Target,self).on_unhold_alarm(context, alarm)
		if outcome != ROUTING_OUTCOME_CONTINUE:
			return outcome


		log = context.log

		#
		# Spostiamoci sul tenant giusto
		#
		tenant = Tenant.objects.get(uuid=alarm.tenant_uuid)
		tenant.db_set_current_schema()


		log.info("[SILENCE] unhold alarm %s",  alarm.uuid)

		alarmdata = alarm.data

		#
		# Dati / flag associati all'allarme
		#
		condition_path = alarmdata.get('condition', {'path':None}).get('path', None)
		alarm_status   = alarmdata.has_key("condition") and alarmdata.get('result',{}).get('status', None)
		target_data_id = alarmdata[ '_silence' ]['target_data_id']
		drop_last_up   = alarmdata[ '_silence' ]['drop_last_up']

		#
		# Carichiamo i dati di tutti gli allarmi aggregati dal target per questo orario
		#

		log.debug("[SILENCE] target_data_id %s",  target_data_id)

		tdata, created = self._load_digest_data(log, tenant, target_data_id)

		data = tdata.get_data()

		condition_alarms = data['alarms_by_condition'].get(condition_path, None)

		if condition_alarms is not None:

			for entry in condition_alarms:
				log.debug("[SILENCE] Alarm %s %s (%s)", entry['alarm_id'], entry['alarm_status'], (alarm.uuid == entry['alarm_id']) )

			am_i_last = alarm.uuid == condition_alarms[-1]['alarm_id']
			log.debug("[SILENCE] is the last alarm? (%s) : %s", alarm.uuid, am_i_last)


			if am_i_last:
				if alarm_status in CSTATE_MACRO_UP and drop_last_up:
						outcome = ROUTING_OUTCOME_DROP
			else:
				outcome = ROUTING_OUTCOME_DROP
		else:
			log.error("[SILENCE] : Alarm %s related to %s but no data about found!", alarm.uuid, target_data_id)

		#
		#
		# FIX: tdata non viene mai cancellato dal DB. Come facciamo?
		#
		#

		#
		# Remove silence data
		#
		del alarmdata[ '_silence' ]

		return outcome

	def _load_digest_data(self, log, tenant, digest_id, create_if_new=False):
		log.debug("[SILENCE] Loading digest data: %s (create if new=%s)", digest_id, create_if_new)
		db_set_current_tenant_schema(tenant.name)

		if create_if_new:
			log.debug("[SILENCE] Loading digest data: %s", digest_id)
			tdata, created = TargetData.objects.get_or_create(tenant=tenant, key=digest_id)
		else:
			try:
				tdata = TargetData.objects.get(tenant=tenant, key=digest_id)
				created = False
			except TargetData.DoesNotExist, e:
				tdata   = None
				created = False

		return tdata, created
