# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

.. _target-smtp:

========================================================================
SMTP
========================================================================


Questo target permette l'invio di un messaggio tramite protocollo SMTP.

I parametri questo target vanno specificati secondo la sintassi: 

	::
		
		-J SMTP --smtp_<parametro> <valore> [ --smtp_<parametro> <valore> ... ]

Parametri:

	+---------------+---------------+-------------------------------+--------------------------------------------------------------------------+
	| Parametro     | Type          | Default                       | Descrizione                                                              |
	+===============+===============+===============================+==========================================================================+
	| smtp_host     | stringa       | localhost                     | Indirizzo o hostname del'MTA.                                            |
	+---------------+---------------+-------------------------------+--------------------------------------------------------------------------+
	| smtp_port     | numero        | 25                            | Porta TCP dell'MTA.                                                      |
	+---------------+---------------+-------------------------------+--------------------------------------------------------------------------+
	| envelope_from | email         | monitoring-daemon@localhost   | Return-path SMTP del messaggio.                                          |
	+---------------+---------------+-------------------------------+--------------------------------------------------------------------------+
	| header_from   | email         | monitoring-daemon@$localhost  | Indirizzo che comaparira nel campo "From:"                               |
	+---------------+---------------+-------------------------------+--------------------------------------------------------------------------+
	| rcpt_to       | email         | root@localhost                | Destinatario/i del messaggio (separati da virgole)                       |
	+---------------+---------------+-------------------------------+--------------------------------------------------------------------------+
	| rcpt_cc       | email         |                               | Destinatario/i in CC (separati da virgole)                               |
	+---------------+---------------+-------------------------------+--------------------------------------------------------------------------+
	| rcpt_bcc      | email         |                               | Destinatario/i in BCC (separati da virgole)                              |
	+---------------+---------------+-------------------------------+--------------------------------------------------------------------------+
	| message       | stringa       | Message from ${hostname}      | Corpo del mesasaggio                                                     |
	+---------------+---------------+-------------------------------+--------------------------------------------------------------------------+
	| subject       | stringa       | ${rawalarm}                   | Soggetto del messaggio                                                   |
	+---------------+---------------+-------------------------------+--------------------------------------------------------------------------+
	| xheaders      | stringa       |                               | Stringa con Extra-headers SMTP (es. *X-Message-Foo: bar* )               |
	+---------------+---------------+-------------------------------+--------------------------------------------------------------------------+

Destinatari multipli
--------------------

IL parametro *rcpt_to* puo' contenere destinatari multipli separati da virgola (','). 

	::
		
		entables -A DEFAULT -j SMTP --smtp_rcpt_to noc@labs.it,netwarning,root

**IMPORTANTE**: se la stringa memorizzata in *rcpt_to* contiene caratteri di *new-line* (\n), questi vengono sostituiti con la virgola (',').


.. _target-SMPT-invalid-recipients:

Destinatari invalidi
--------------------

Il target **NON INVIA** email quando il parametro *rcpt_to* conviene le seguenti stringhe:

	* *stringa vuota*

**IMPORTANTE**: l'allarme non viene inviato ma continua ad essere processato dalle regole successive.

**IMPORTANTE**: l'allarme non viene inviato e quindi non produce nessun tipo di LOG nel server MTA utilizzato.

Esempi
------

Invia tutti i messaggi a  noc@labs.it: 

	::
		
		entables -A DEFAULT -j SMTP --smtp_rcpt_to noc@labs.it  --smtp_rcpt_bcc hiddennoc@labs.it  --smpt_xheaders 'X-test: email di prova'

"""
import traceback

import email.utils
from smtplib import SMTP, SMTPRecipientsRefused, SMTPHeloError, SMTPSenderRefused, SMTPDataError, SMTPConnectError

import datetime
import pprint

from entables.constants import TABLE_NOTIFY

from entables.constants import ROUTING_OUTCOME_CONTINUE

from entables.datastructures import TargetMemoryObject

from entables.targets import BaseTargetHandler
from entables.targets import Option

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class SafeSMTPError(Exception):
	pass

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def create_default_smtp_target():
	
	# Questa e' la regola di default per mandare messaggi usando
	# la configurazione delle condition di sanet.
	#
	#   sanet conf (msg_upbody/msg_upsubject/ecc.) -> tags processing -> opzioni target

	return TargetMemoryObject(
		name = 'smtp',
		options = {
			'rcpt_to' : '${mail_to}',
			'subject' : '${mail_subject}',
			'message' : '${mail_body}',
			'xheaders': '${mail_xheaders}',
		}
	)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#


def STR_EMAIL(s):

	s = str(s)
	# FIX: does not work with   'foo@google.com/bar'
	if (s == email.utils.formataddr(email.utils.parseaddr(s))):
		return s
	else:
		raise Exception("Invalid email format: %s" % (s) )


def _cut(s):
	if len(s) > 10:
		return '%s...' % (s[0:7])
	return s
	
#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#
from time import gmtime, strftime

class Target(BaseTargetHandler):

	def __init__(self):
		super(Target, self).__init__()

		self.name = 'smtp'

		self.tables = [
			TABLE_NOTIFY,
		]

		self.options    = [
			Option("smtp_host"    ,           default="localhost"),
			Option("smtp_port"    , type=int, default="25"),
			Option("envelope_from", type=str, default="monitoring-daemon@localhost"),
			Option("header_from"  , type=str, default="monitoring-daemon@localhost"),
			Option("rcpt_to"      , type=str, default="root@localhost"),
			Option("rcpt_cc"      , type=str),
			Option("rcpt_bcc"     , type=str),
			Option("subject"      ,           default="Message from ${hostname}"),
			Option("message"      ,           default="${rawalarm}"),
			Option("xheaders"     ,           default=""),
		]
		
		self.counters = {
			'sent' : 0,
			'error': 0,
		}

	def init(self):
		return
		
	def get_text_info(self):
		text = super(Target, self).get_text_info()
		text += ',' + ','.join([ "%s:%s" % (c,v) for c,v in self.counters.items()])
		return text
		
		
	def validate_target_options(self, options_dict):
		
		super(Target, self).validate_target_options(options_dict)
		
		defaults = self.get_options_default()
		
		if options_dict.has_key('smtp_host'):
			if (options_dict['smtp_host'] == ""): raise Exception("Invalid smtp host: '%s'" % ( options_dict['smtp_host'] ))
		
		return True

	def dispatch(self, context, alarm, ruleobj, target_options):

		context.log.debug("SMPT DISPATCH")

		super(Target, self).dispatch(context, alarm, ruleobj, target_options)
		
		options = self.expand_options(context, alarm, target_options)

		alarm.set_verbstatus("Target %s" % (self.name,))

		try:
			alarmdata = alarm.data
			self.process_alarm_and_dispatch(context, alarm, alarmdata, ruleobj, options)

		except SafeSMTPError, e:
			msg = str(e)
			context.log.error(msg)
			alarm.set_verbstatus(msg)
			
		return ROUTING_OUTCOME_CONTINUE	




	def process_alarm_and_dispatch(self, context, alarm, alarmdata, ruleobj, options):

		try:
			date  = strftime("%a, %d %b %Y %H:%M:%S +0000", gmtime())

			smtp_host     = options['smtp_host']
			smtp_port     = int(options['smtp_port'])
			envelope_from = options["envelope_from"]
			xheaders      = options['xheaders'].strip()
			header_from   = options["header_from"]
			rcpt_to       = options["rcpt_to"]
			rcpt_cc       = options.get("rcpt_cc", None)
			rcpt_bcc      = options.get("rcpt_bcc", None)
			subject       = options['subject']
			message       = options['message']

			# IMPORTANTE: Per evitare facili errori, se nella stringa dei recipient compaiono \n li trasformiamo in ',' in modo da trasformare rcpt_to in un elenco di indirizzi separato da virgole.
			rcpt_to       =  rcpt_to.replace('\n', ',')
			
			msg = "New email at %s:%s : envelope='%s', rcpt_to='%s', rcpt_cc='%s', rcpt_bcc='%s', header_from='%s', subject='%s', body='%s')." % (
				smtp_host, 
				smtp_port, 
				envelope_from, 
				rcpt_to, 
				rcpt_cc,
				rcpt_bcc,
				header_from,  
				subject, 
				_cut(message
			))
			
			if not rcpt_to:
				msg += "Trying to send message to null recipient. Skipping."
				raise SafeSMTPError(msg)

			# >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
			# IMPORTANT: Sanet2 behaviour removed by ticket #722
			#
			#if rcpt_to == "none":
			#	msg += "Recipient address to address '%s'. Skipping"
			#	raise SafeSMTPError(msg)
			#<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

			if alarm is not None:
				alarm.set_verbstatus("\tSending at %s:%s : envelope='%s', rcpt_to='%s', rcpt_cc='%s', rcpt_bcc='%s', header_from='%s', subject='%s', body='%s')" % (
					smtp_host, 
					smtp_port, 
					envelope_from, 
					rcpt_to, 
					rcpt_cc, 
					rcpt_bcc, 
					header_from,  
					subject, 
					_cut(message)
				))

			xentablesheaders  = []
			#xentablesheaders += [ 'X-Entables-Alarmid: %s' % (alarm.uuid) ]
			xentablesheaders += [ 'X-Entables-Alarmid: %s' % (alarmdata['uuid']) ]
			xentablesheaders += [ 'X-Entables-Target: %s' % (self.name) ]
			xentablesheaders += [ 'X-Entables-Date: %s' % (date) ]
			xentablesheaders  = '\n'.join( xentablesheaders )
			
			headers = []
			headers += ["From: %s"      % ( header_from )]
			headers += ["To: %s"        % ( rcpt_to )]
			
			if rcpt_cc:
				headers += ["CC: %s" % ( rcpt_cc )]
			
			headers += ["Subject: %s"   % ( subject)]
			#headers += ["Date: %s"  % ( date )]
			headers = '\n'.join( headers )

			smtp = SMTP()
			smtp.set_debuglevel(0)

			smtp.connect(smtp_host, smtp_port)

			# I destinatari deveno essere memorizzati in una lista. Splittare sul carattere "," dovrebbe bastare.
			# La libreria interpreta bene anche destinatari con sintassi "Foo ham" <foo@bar>
			recipients =  rcpt_to
			recipients =  recipients.split(',')
			
			if rcpt_cc:
				recipients += rcpt_cc.split(',')
			if rcpt_bcc:
				recipients += rcpt_bcc.split(',')
			
			message = xentablesheaders + '\n'
			if xheaders:
				message += xheaders + '\n'
			if headers: 
				message += headers + '\n'
			
			# Add a "\n" and force smtplib to treat message as email body
			message += "\n" + options["message"]
			
			#context.log.debug(message)

			context.log.info("SMTP %s:%s : envelope='%s', rcpt_to='%s', rcpt_cc='%s', rcpt_bcc='%s', header_from='%s', subject='%s', body='%s...')" % (
				smtp_host, 
				smtp_port, 
				envelope_from, 
				rcpt_to, 
				rcpt_cc,
				rcpt_bcc,
				header_from , 
				subject, 
				message[0:10]
			))
			
			#
			# FIX: gestire caratteri non ascii in maniera migliore
			#
			# Esempio: 
			#     msg = MIMEText(message, _charset="UTF-8")
			#            msg['Subject'] = Header(subject, "utf-8")
			#
			message = message.encode('ascii', errors='replace')

			smtp.sendmail( envelope_from, recipients, message)

			smtp.quit()
			
			#return ROUTING_OUTCOME_CONTINUE	
			
		except SMTPConnectError, er:
			msg = "SMTP Connection error: %s" % (er)
			#context.log.error(msg)
			#alarm.set_verbstatus(msg)
			raise SafeSMTPError(msg)

		except SMTPRecipientsRefused, er:
			msg = "SMTP Recipient refused ERROR: %s" % (er)
			#context.log.error(msg)
			#alarm.set_verbstatus(msg)
			raise SafeSMTPError(msg)
			
		except SMTPHeloError, er:
			msg = "SMTP Helo Error ERROR: %s" % (er)
			#context.log.error(msg)
			#alarm.set_verbstatus(msg)
			raise SafeSMTPError(msg)
			
		except SMTPSenderRefused, er:
			msg = "SMTP Sender refused ERROR: %s" % (er)
			#context.log.error(msg)
			#alarm.set_verbstatus(msg)
			raise SafeSMTPError(msg)
			
		except SMTPDataError, er :
			msg = "SMTP Data ERROR: %s" % (er)
			#context.log.error(msg)
			#alarm.set_verbstatus(msg)
			raise SafeSMTPError(msg)
		
		except SafeSMTPError,e:
			raise e
			
		except Exception, e:
			msg = "Unknown error while dispatching allarm via SMTP: %s:%s: %s" % (type(e), str(e), traceback.format_exc())
			#context.log.error(msg)
			#alarm.set_verbstatus(msg)
			raise SafeSMTPError(msg)

		return ROUTING_OUTCOME_CONTINUE	

