# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

.. _target-SMTPDIGEST:


SMTPDIGEST
===========

Questo target permette di inviare tramite protocollo SMTP tutti gli allarmi DOWN piu' recenti ed ancora attivi presenti in in un DIGEST.

Logica del target
------------------

Il target verifica per tutti gli allarmi aggregati quali allarmi di DOWN sono ancora *validi*  (La condition
associata e' ancora DOWN) ed *PER OGNIUONO* invia un messaggio di mail per l'ultimo DOWN intercettato in base ai parametri specificati.

Parametri
----------

I parametri questo target vanno specificati secondo la sintassi: 

::
	
	-J SMTPDIGEST --smtpdigest_<parametro> <valore> [ --smtpdigest_<parametro> <valore> ... ]

Parametri disponibili sono gli stessi del target :ref:`_target-SMTP`. Si rimanda alla documentazione del digest
per dettagl sull'utilizzo dei parametri.


**IMPORTANTE**: Se nei parametri del target vengono usati dei *tag* ( ${mail_to}, ${mail_subject}. ${alarmid}, ecc..)
questi vengono sostituiti di volta in volta con i dati dell'allarme di DOWN attivo e inviato.


Tag speciali utilizzabili nei parametri del digest
-------------------------------------------------------

E' possibile inviare informazioni aggiuntive sull'allarme di DOWN utilizzando dei *tag* appositi:

	+-------------------------+---------------------------------------------------------------------------------------+
	| Tag                     | Descrizione                                                                           |
	+=========================+=======================================================================================+
	| ${digestchangescount}   | Numero di variazioni aggregate dal digest relative alla stessa condition dell'allarme |
	+-------------------------+---------------------------------------------------------------------------------------+


Invio semplice del digest
--------------------------

Aggrega tutti gli allarmi fino alle 8:00 del prossimo lunedi'. Alle 8:00 di lunedi' mattina tutti gli allarmi di DOW

::
	
	entables -A DEFAULT -j DIGEST     --digest_holduntil "monday 8:00"
	...
	...
	entables -A DEFAULT -m attr --attr_digest_downs 0     -j DROP 
	entables -A DEFAULT                                   -j SMTPDIGEST --smtpdigest_rcpt_to '${mail_to}'  --smtpdigest_subject '${mail_subject}' --smtpdigest_message '${mail_body}' 

Invio arricchito con informazioni dal digest
----------------------------------------------

Come il precedente, ma nel subject e corpo del messaggio aggiungiamo qualche informazione per far capire che il messaggio di DOWN era associato ad un digest.

::
	
	entables -A DEFAULT -j DIGEST     --digest_holduntil "monday 8:00"
	...
	...
	entables -A DEFAULT -m attr --attr_digest_downs 0    -j DROP 
	entables -A DEFAULT                                  -j SMTPDIGEST --smtpdigest_rcpt_to ' ${mail_to}' --smtpdigest_subject '[FOR DIGEST] [ ${digestchangescount} changes ] ${mail_subject}'  --smtpdigest_message '${mail_body}' 

"""


import pprint

from entables.constants import ROUTING_OUTCOME_CONTINUE

from entables.targets.digest import get_alarms_by_condition_from_digest


from entables.targets.smtp import SafeSMTPError
from entables.targets.smtp import Target as SmtpTargetHandler


#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

import pprint 
from sanet_common.sanetdefs.constants import CSTATE_MACRO_DOWN

class Target(SmtpTargetHandler):

	def __init__(self):

		super(Target, self).__init__()
		self.name = 'smtpdigest'

	def dispatch(self, context, alarm, ruleobj, target_options):

		#options = self.expand_options(context, alarm, target_options)

		try:
			context.log.debug("SMPTDIGEST DISPATCH")

			main_alarm = alarm.data
			
			context.log.debug("Main alarm: %s", main_alarm['uuid'])

			digest_data = main_alarm.get('digest_data', None)
			if digest_data is None:
				context.log.debug("No digest data")


			from core.models import Condition, Tenant
			from core.api.base import db_set_current_tenant_schema
			import webui.models

			tenant_id = main_alarm['tenantid']
			tenant    = Tenant.objects.get(uuid=tenant_id)
			db_set_current_tenant_schema(tenant.name)

			context.log.debug("Loading alarms by condition")
			alarms_by_condition_path, errors = get_alarms_by_condition_from_digest(alarm, digest_data)

			context.log.debug("Got %s conditions to analize", len(alarms_by_condition_path.keys()) )

			#
			# BRIEF VERSION
			#
			
			context.log.debug("Loading conditions status")
			down_conditions = Condition.objects.select_related('element', 'element__parent').filter(path__in=alarms_by_condition_path.keys()).filter(status__in=CSTATE_MACRO_DOWN)
			
			context.log.debug("Loaded %s DOWN conditions", len(down_conditions))
			for c in down_conditions:

				context.log.debug("Processing condition %s", c.path)

				alarms = alarms_by_condition_path.get( c.path, [] )
				if not alarms:
					continue

				#
				# IMPORTANTE: Sel la condition e' DOWN, l'ultimo allarme per condition dovrebbe essere il DOWN.
				#
				
				last_down_alarm_data = alarms[-1]

				last_down_alarm_data['digestchangescount'] = len(alarms)
				
				context.log.debug("Sending old alarm %s with changes %s", last_down_alarm_data['uuid'], last_down_alarm_data['digestchangescount'])
				
				options2 = self.expand_options_from_alarmdata(context, last_down_alarm_data, target_options)
				
				self.process_alarm_and_dispatch(context, None, last_down_alarm_data, ruleobj, options2)
				
		except SafeSMTPError, e:
			msg = str(e)
			context.log.error(msg)
			alarm.set_verbstatus(msg)

		except Exception, e:
			msg = "Unknown error while processing allarm via SMTPDIGEST: %s:%s" % (type(e), str(e))
			context.log.error(msg)
			alarm.set_verbstatus(msg)

		return ROUTING_OUTCOME_CONTINUE	


