 
# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.


"""

========================================================================
:mod:`ExprFactory`
========================================================================

Exceptions
--------------------------------------------------------------------------------

.. autoexception:: BaseExprParserException

.. autoexception:: ParserException

.. autoexception:: ExprParserAnalysisError

.. autoexception:: LexicalError

.. autoexception:: ParsingError

.. autoexception:: RuntimeExprParserError

.. autoexception:: ExecutionError

.. autoexception:: NotEvaluatedError


Expression Factory
--------------------------------------------------------------------------------

.. autoclass:: BaseExprFactory
	:members:

Expression Tree objects
--------------------------------------------------------------------------------

Questa e' la gerarchia delle classi di tipo :py:class:`ExprTree`:

::
    
     ExprTree
       |
       |-- IdentifierExprTree         Variabili
       |
       |-- FullOIDExprTree            Valori SNMP (vecchi # e nuovi @)
       |
       |-- LiteralExprTree
       |     |-- StringExprTree       'xxxx'
       |     |-- DecimalExprTree      numeri interi
       |     |-- FloatExprTree        numeri virgola mobile
       |     `-- BoolExprTree         True o False
       |
       `-- OpExprTree
             |
             |-- CallOpExprTree       Chiamata a funzione (es. foo() )
             |
             |-- IfExprTree           if <a> then <b> else <c>
             |-- TryExprTree          try <expr> catch [ <error name> ] ( [ <expr> ] )
             |
             |-- RangeOpExprTree      a[i], a[i:j]
             |-- OrOpExprTree         a || b
             |-- AndOpExprTree        a && b
             |-- NotOpExprTree        !a
             |-- BitXorOpExprTree     a ^ b
             |-- BitAndOpExprTree     a & b
             |-- BitOrOpExprTree      a | b
             |-- EqOpExprTree         a = b
             |-- NotEqOpExprTree      a != b
             |-- RegExprEqOpExprTree  a =~ b
             |-- LtOpExprTree         a < b
             |-- GtOpExprTree         a > b
             |-- LEqOpExprTree        a <= b
             |-- GEqOpExprTree        a >= b
             |-- MinusOpExprTree      -a
             |-- PlusOpExprTree       a + b
             |-- SubOpExprTree        a - b
             |-- MultiOpExprTree      a * b
             |-- DivOpExprTree        a / b
             |-- ModOpExprTree        a % b
             `-- PowerOpExprTree      a ** b 

Classi base
~~~~~~~~~~~~~~~~

.. autoclass:: ExprTree
	:members:


.. autoclass:: LiteralExprTree
	:members:

.. autoclass:: OpExprTree
	:members:

Costrutti semplici
~~~~~~~~~~~~~~~~~~~~

.. autoclass:: OIDLiteralExprTree
	:members:

.. autoclass:: BoolExprTree
	:members:

.. autoclass:: StringLiteralExprTree
	:members:

.. autoclass:: IntegerLiteralExprTree
	:members:

.. autoclass:: FloatLiteralExprTree
	:members:

Costrutti Variabili e Identificatori
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. autoclass:: IdentifierExprTree
	:members:

.. autoclass:: DataIdentifierExprTree
	:members:

Alberi SNMP
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. autoclass:: SnmpValueExprTree
	:members:


Operatori
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. autoclass:: RangeOpExprTree
	:members:

.. autoclass:: CallOpExprTree
	:members:

.. autoclass:: OrOpExprTree
	:members:

.. autoclass:: AndOpExprTree
	:members:

.. autoclass:: NotOpExprTree
	:members:

.. autoclass:: BitOneOpExprTree
	:members:

.. autoclass:: BitOrOpExprTree
	:members:

.. autoclass:: BitXOrOpExprTree
	:members:

.. autoclass:: BitAndOpExprTree
	:members:

.. autoclass:: EqOpExprTree
	:members:

.. autoclass:: NotEqOpExprTree
	:members:

.. autoclass:: RegExprEqOpExprTree
	:members:

.. autoclass:: LtOpExprTree
	:members:

.. autoclass:: GtOpExprTree
	:members:

.. autoclass:: LEqOpExprTree
	:members:

.. autoclass:: GEqOpExprTree
	:members:

.. autoclass:: ConcatOIDExprTree
	:members:

.. autoclass:: AddOpExprTree
	:members:

.. autoclass:: AddAddOpExprTree
	:members:

.. autoclass:: SubOpExprTree
	:members:

.. autoclass:: SubSubOpExprTree
	:members:

.. autoclass:: MinusOpExprTree
	:members:

.. autoclass:: MultiOpExprTree
	:members:

.. autoclass:: DivOpExprTree
	:members:

.. autoclass:: ModOpExprTree
	:members:

.. autoclass:: PowerOpExprTree
	:members:


Costrutti complessi
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. autoclass:: IfExprTree
	:members:

.. autoclass:: TryExprTree
	:members:

"""

import re
import types
import inspect

from .Types import LazyValueType
from .Types import regexp_eq

#------------------------------------------------------------------------------#
# DEBUG                                                                        #
#------------------------------------------------------------------------------#

def debugline(str):
	#print str
	pass

def debug(str):
	#print str,
	pass

#------------------------------------------------------------------------------#
# Exceptions                                                                   #
#------------------------------------------------------------------------------#

class BaseExprParserException(Exception):
	"""
	Base exception for all the exceptions raised by all module's code
	"""
	pass

class ParserException(BaseExprParserException):
	"""
	Base exception.
	"""
	pass

class ExprParserAnalysisError(ParserException):
	"""
	Base exception raised by Lexical Analyzer or Parser.
	"""
	pass

class LexicalError(ExprParserAnalysisError):
	"""
	Simple Exception raised when the parser finds a lexical error inside the parsed expression
	"""
	def __init__(self, lexerException):
		self.token = lexerException.getUnexpectedType()
		self.line  = lexerException.line
		self.pos   = lexerException.charPositionInLine

	def __str__(self):
		return "Error in input at %s:%s. Found '%s'. " % (str(self.line),str(self.pos),str(self.token))

class ParsingError(ExprParserAnalysisError):
	"""
	Simple Exception raised when the parser detects a parsing error.
	"""
	def __init__(self, exception):
		self.line = str(exception.line)
		self.pos  = str(exception.charPositionInLine)

	def __str__(self):
		return "Parsing error at position (%s:%s)" % (self.line,self.pos)


class RuntimeExprParserError(ParserException):
	"""
	Base exception raised while evaluating an :py:class:`~.ExprTree`.
	"""
	pass

class ExecutionError(RuntimeExprParserError):
	"""
	This base exception is raised while evaluating an expression.
	
		* A symbol (function or variable) is missing inside the execution context dictionary.
		* An invalid operation is performed (index out of range?)
		* ecc.
	"""
	pass

class NotEvaluatedError(RuntimeExprParserError):
	"""
	This exception is raised when the value of an expression 
	tree can't be evaluate.
	
	All sub-expression must be evaluated before raising the exception!
	"""
	pass

#------------------------------------------------------------------------------#
# Definizioni dei nodi dell'albero di valutazione
#------------------------------------------------------------------------------#

def trunc_long_str_value(s):
	if len(s) > 200:
		return s[0:100] + "........" + s[-100:]
	return s

#------------------------------------------------------------------------------#
# CLASSI                                                                       #
#------------------------------------------------------------------------------#

class NamedArgument():
	"""
	Classe di comodo per codificare i parametri "nominali" nelle chiamate a funzione.
	
		<name>=<expr>
	"""

	def __init__(self, name, expr_tree):
		self.name      = name
		self.expr_tree = expr_tree

	def __str__(self):
		return "%s=%s" % (self.name, self.expr_tree)

#------------------------------------------------------------------------------#

class ExprTree(object):
	"""
	Classe base per rappresentare un nodo dell'albero espressione del linguaggio.
	
	Il valore calcolato dall'albero espressione dal metodo :py:meth:`~.ExprTree.evaluate` deve essere 
	memorizzato nell'attributo:
	
	::
		
		self.value
	
	"""
	def __init__(self):
		
		# The evaluated value of this expression
		self.value = None

	def reset(self):
		# Cleanup method
		pass

	def evaluate(self, context):
		"""
		Calculate the value of the the expression tree.
		
		If this object rappresent a complex expression with sub-expressions,
		this method calculate the value of all the sub-expression trees and then
		computes the final result.
		
		The 'context' parameter rappresents the execution context of the expression. 
		
		It must be of type 'dict' and it must store the values for all the symbols
		(variables and functions) used inside the expression tree.
		
		The 'context' object can be used by user functions to save and retrieve
		additional informations about the computation (es. statistic counters, ecc....)

		:param  context: A mapping object.
		:type   context: dict
		
		:returns: Tthe result of the evaluation.
		:rtype: Depends on the result.
		
		:exceptions:
	
			:class:`ExecutionError`     
				If the expression or some sub-expression references a
				missing symbol.

			:class:`NotEvaluatedError`  
				The value of the expression can not be calculated
				because of some special condition occurred.
				
				All sub-expression are evaluated before this
				expression is raised.
			
				User defined functions contained in the 'context'
				can raise this exception when an error occurs.
		"""
		raise NotImplementedError()

	def evaluated(self):
		"""
		Return a string rappresentation of the evaluated expression.
		"""
		#
		# Tipi "Lazy"
		#
		if isinstance(self.value, LazyValueType):
			# Gli oggetti SnmpValueType sono lazy e quando vengono "calcolati" possono generare una eccezione.
			# Se tentiamo di convertirli in stringa risolleverebbero l'eccezione quindi
			# intercettiamo 
			try:
				return self._to_readable_string( str(self.value) )
			except Exception, e:
				return "**%s**" % (type(e))

		#
		# Tipi normali
		#
		if isinstance(self.value, types.StringType) or isinstance(self.value, types.UnicodeType):
			v = str(self.value) #avoid ascii encoding problems
			v = self._to_readable_string( v )
			v = trunc_long_str_value(v) 
			return v

		elif isinstance(self.value, types.FloatType):
			# show only interger part when the float
			i = int(self.value)
			if i != self.value: 
				return str(self.value)
			else: 
				return str(i)

		elif isinstance(self.value, (types.BooleanType, types.IntType, types.LongType)):
			return str(self.value)

		else:
			#
			# Complex types (array, dict, ecc.)
			#
			return repr(self.value)

	def _to_readable_string(self, s):
		"""
		Produce una versione _SEMPRE_ human-readable della stringa "s" se "s" contiene
		caratteri non "stampabili"
		"""
		# Compatibile utf8?
		try:
			s.decode('utf-8')
			return s
		except Exception, e:
			try:
				s.decode('ascii')
				return s
			except Exception, e:
				pass

		try:
			s = str(s)
		except Exception, e:
			pass

		# Contiene byte casuali.
		s2 = repr(s)
		return s2


	def coerce_type(self, v1, v2):
		"""
		This method tries to understand if *v1* or *v2* are strings and if they can be converted in numbers.
		
		.. note:: SNMP values are sometimes strings instead of numbers, be we want to compare them! It's useful trying to convert strings to numbers when a *number Vs string* comparison is performed
		
		:param  v1: A value
		:type   v1: any
		
		:param  v2: A value
		:type   v2: any
		
		:returns: A tuple with *v1* and *v2* converted to the same type.
		:rtype: Tuple.
		
		:exceptions:
			
			:class:`ValueError`
				if an error occurs when converting *v1* or *v2*.
		""" 

		if type(v1) == type(v2):
			return (v1, v2)

		if isinstance(v1, types.StringType) or isinstance(v1, types.UnicodeType):
			if isinstance(v2, types.IntType):  return (int(v1),v2)
			if isinstance(v2, types.LongType): return (int(v1),v2)
			if isinstance(v2, types.FloatType): return (float(v1),v2)

		if isinstance(v2, types.StringType) or isinstance(v2, types.UnicodeType):
			if isinstance(v1, types.IntType)  : return (v1, int(v2))
			if isinstance(v1, types.LongType) : return (v1, int(v2))
			if isinstance(v1, types.FloatType): return (v1, float(v2))

		return (v1, v2)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class RootExprTree(ExprTree):
	"""
	Classe usata come radice di qualunque albero espressione.
	
	La radice ha il compito di elaborare il valore dell'albero espressione
	e compiere gli aggiustamenti finali prima di restituire il valore
	definitivo dell'espresione al chiamante.
	"""

	def __init__(self, expr):
		super(ExprTree, self).__init__()

		# Albero espressione vero e proprio.
		self.expr = expr

	def __str__(self):
		return (str(self.expr))

	def reset(self):
		self.value = None
		self.expr.reset()

	def evaluate(self, context):

		v = self.expr.evaluate(context)

		# L'albero espressione lavora anche con tipi SnmpValueType
		# che sono "lazy". Per assicurarci di restituire il valore
		# "reale" dell'espressione, assicuriamoci che l'oggetto
		# SnmpValuteType legga il valore SNMP via rete.
		
		if isinstance(v, LazyValueType):
			self.value = v.getvalue()
		else:
			self.value = v

		return self.value

	def evaluated(self):
		return self.expr.evaluated()

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

# Classe padre per tutti i valori costanti
#
class LiteralExprTree(ExprTree):
	"""
	TODO
	"""
	def __init__(self):
		super(LiteralExprTree, self).__init__()

	def evaluate(self, context):
		return self.value

#
# Classe padre di tutti gli alberi espressione
#
class OpExprTree(ExprTree):
	"""
	TODO
	"""
	pass

#------------------------------------------------------------------------------#
# n.n.n.n......n                                                               #
#------------------------------------------------------------------------------#

class OIDLiteralExprTree(LiteralExprTree):
	"""
	TODO
	"""

 	def __init__(self, strValue):
		super(OIDLiteralExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
 		self.value = strValue

 	def __str__(self):
 		return self.value

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class BoolExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, boolValue):
		super(BoolExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.value = boolValue

	def __str__(self):
		return ('Bool(%s)' % str(self.value))

	def evaluate(self, context):
		return bool(self.value)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class StringLiteralExprTree(LiteralExprTree):
	"""
	TODO
	"""

	def __init__(self, strValue, escape):
		super(StringLiteralExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )

		if escape == True:
			self.value = self.unescape( strValue[1:len(strValue) - 1] )
		else:
			self.value = strValue[1:len(strValue) - 1]

	def __str__(self):
		return ("Str(%s)" % self.value)

	def unescape(self, s):
		last=''
		s2=''
		for c in s:
			if last == '\\':
				last = ''
				if   c=='b' : c="\b"
				elif c=='t' : c="\t"
				elif c=='n' : c="\n"
				elif c=='f' : c="\f"
				elif c=='r' : c="\r"
				elif c=='\"': c="\""
				elif c=='\'': c="'"
				else:         c=="\\"
				s2 = s2 + c
			else:
				if c=='\\': last = c
				else:       s2 = s2 + c
		return s2

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class IntegerLiteralExprTree(LiteralExprTree):
	"""
	TODO
	"""

	def __init__(self, decimalValue):
		super(IntegerLiteralExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.value = decimalValue

	def __str__(self):
		return ("Int(%s)" % self.value)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class FloatLiteralExprTree(LiteralExprTree):
	"""
	TODO
	"""

	def __init__(self, floatValue):
		super(FloatLiteralExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.value = floatValue

	def __str__(self):
		return ("Float(%s)" % self.value)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class IdentifierExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, strId):
		super(IdentifierExprTree, self).__init__()
		
		debugline( "init " + str(self.__class__) )
		self.id = strId

	def __str__(self):
		return ('Id(%s)' % self.id)

	def reset(self):
		self.value = None

	def evaluate(self, context):
		if self.id in context:
			self.value = context[self.id]
			return self.value
		else:
			raise ExecutionError("Symbol '%s' is not defined in the context!" % (self.id) )


# { <name> @ }
# { <name> # }
class DataIdentifierExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, dataId):
		super(DataIdentifierExprTree, self).__init__()
		debugline( "init " + str(self.__class__) )
		self.id    = dataId[1:-2]   # <name>
		self.dtype = dataId[-2:-1]  # @ , #

	def __str__(self):
		return 'Data(%s%s)' % (self.id, self.dtype)

	def reset(self):
		self.value = None

	def evaluate(self, context):

		self.value = self.calculateDataValue(context)
		return self.value

	def calculateDataValue(self, context):
		raise NotImplemented()

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

# Classe astratta!
class SnmpValueExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, exprOID, eCommunity, charType, eHost, ePort):
		super(SnmpValueExprTree, self).__init__()
		
		debugline( "init " + str(self.__class__) )
		self.exprOID   = exprOID
		self.community = eCommunity
		self.oidType   = charType
		self.host      = eHost
		self.port      = ePort

	def __str__(self):
		s = str(self.exprOID) ;
		if self.community != None:
			s = s + ":" + str(self.community);
		s = s + self.oidType;
		if self.host != None:
			s = s + str(self.host);
			if self.port != None:
				s = s + ":" + str(self.port);

		return ('{%s}' % s)

	def reset(self):
		self.value = None

		self.exprOID.reset()
		
		if self.community != None:
			self.community.reset()
			
		if self.host != None:
			self.host.reset()
			if self.port != None:
				self.port.reset()

	def evaluate(self, context):
		error = None
		o,c,t,h,p   = None,None,None,None,None

		try:
			o = self.exprOID.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		if self.community != None:
			try:
				c = self.community.evaluate(context)
			except NotEvaluatedError, er:
				error = er

		t = self.oidType;

		if self.host != None:
			try:
				h = self.host.evaluate(context)
			except NotEvaluatedError, er:
				error = er

		if self.port != None:
			try:
				p = self.port.evaluate(context)
			except NotEvaluatedError, er:
				error = er

		if error:
			# FIX: add informations about all errors!
			raise error

		self.value = self.calculateSnmpValue(context, o, c, t, h, p)
		return self.value

	def calculateSnmpValue(self, context, oid,community,ctype,host,port):
		# Esempio di come restituire un valore ma non bloccare 
		# la valutazione dell'espressione.
		# return (False, None)
		raise NotImplementedError()

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class RangeOpExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, table_expr, range_list):
		super(RangeOpExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )

		self.table_expr  = table_expr
		self.range_list  = range_list

	def __str__(self):
		s = str(self.table_expr) + "["

		if len(self.range_list) == 1:
			s += str(self.range_list[0])

		if len(self.range_list) == 2:
			if self.range_list[0] != None:
				s += str(self.range_list[0])
			s += ";"
			if self.range_list[1] != None:
				s += str(self.range_list[1])

		s += "]"
		return s

	def reset(self):
		self.value = None

		self.table_expr.reset()

		for expr_tree in self.range_list:
			expr_tree.reset()

	def evaluate(self, context):
		error = None
		i = None
		j = None

		try:
			t = self.table_expr.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		if len(self.range_list) == 1:
			try:
				i  = self.range_list[0].evaluate(context)
			except NotEvaluatedError, er:
				error = er

			if error:
				raise error

			try:
				self.value = t[i]
			except IndexError, e:
				raise ExecutionError("%i is outside range for value: %r" % (i, t))

			return self.value

		elif len(self.range_list) == 2:
			if self.range_list[0] != None:
				try:
					i  = self.range_list[0].evaluate(context)
				except NotEvaluatedError, er:
					error = er

			if self.range_list[1] != None:
				try:
					j  = self.range_list[1].evaluate(context)
				except NotEvaluatedError, er:
					error = er

			if error:
				raise error

			self.value = t[i:j]

			return self.value

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class CallOpExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, funTree, argsList):
		super(CallOpExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )

		self.fun = funTree
		self.args  = argsList # Lista [] di ExprTree 

	def __str__(self):

		s = "(" + str(self.fun) + "("
		for e in self.args:
			s = s + " " + str(e)
		s = s + "))"
		return s

	def reset(self):
		self.value = None

		self.fun.reset()

		for expr_tree in self.args:
			expr_tree.reset()

	def evaluate(self, context):
		error = None
		try:
			function_ptr = self.fun.evaluate(context)

			if not (isinstance(function_ptr, types.FunctionType) or
			        isinstance(function_ptr, types.LambdaType)):
				raise TypeError("Type for '%s' is not a function! (%s)" % (str(self.fun), type(function_ptr)) )

		except NotEvaluatedError, er:
			error = er

		positional_args, named_args = self.__adjust_args_sequence(self.args)

		positional_evaluated_args = [ ]
		for expr in positional_args:
			try:
				v = expr.evaluate(context)
				positional_evaluated_args.append(v)
			except NotEvaluatedError, er:
				error = er
		if error:
			raise error

		named_evaluated_args = {}
		for arg_name, expr_tree in named_args.items():
			try:
				v = expr_tree.evaluate(context)
				named_evaluated_args[ arg_name ] = v
			except NotEvaluatedError, er:
				error = er

		if error:
			raise error

		self.value = function_ptr(context, *positional_evaluated_args, **named_evaluated_args)

		return self.value

	def __adjust_args_sequence(self, argsList):

		seen_named_arg = False

		positional_args = []
		named_args      = {}

		for arg in argsList:

			if isinstance(arg, NamedArgument):  # Named argument [ name, expr_tree ]

				if not seen_named_arg:
					seen_named_arg = True

				named_args[ arg.name ] = arg.expr_tree

			else: # Positional argument
				if seen_named_arg:
					raise ExecutionError("Wrong argument sequence")

				positional_args.append( arg )

		return positional_args, named_args

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class OrOpExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, left, right):
		super(OrOpExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.left = left
		self.right = right

	def __str__(self):
		return "(%s or %s)" % (self.left, self.right)

	def reset(self):
		self.value = None

		self.left.reset()
		self.right.reset()

	def evaluate(self, context):
		error = None
		try:
			v1 = self.left.evaluate(context)
			#v1 = v1 & True
		except NotEvaluatedError, er:
			error = er

		try:
			v2 = self.right.evaluate(context)
			#v2 = v2 & True
		except NotEvaluatedError, er:
			error = er

		if error:
			raise error

		self.value = v1 or v2
		return  self.value

	def evaluated(self):
		return "(%s or %s)" % (self.left.evaluated(), self.right.evaluated())

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class AndOpExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, left, right):
		super(AndOpExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.left = left
		self.right = right

	def __str__(self):
		return "(%s and %s)" % (self.left, self.right)

	def reset(self):
		self.value = None

		self.left.reset()
		self.right.reset()

	def evaluate(self, context):
		error = None

		try:
			v1 = self.left.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		try:
			v2 = self.right.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		if error:
			raise error

		self.value = v1 and v2
		return self.value

	def evaluated(self):
		return "(%s and %s)" % (self.left.evaluated(), self.right.evaluated())

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class NotOpExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, expr):
		super(NotOpExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.expr = expr

	def __str__(self):
		return "!" + str(self.expr)

	def reset(self):
		self.value = None

		self.left.reset()
		self.right.reset()

	def evaluate(self, context):
		v1 = self.expr.evaluate(context)
		# Only one sub-expression don't need to check for multiple NotEvaluatedError()

		self.value = not v1
		return self.value

	def evaluated(self):
		return "!" + str(self.expr.evaluated())

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class BitOneOpExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, expr):
		super(BitOneOpExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.expr = expr

	def __str__(self):
		return "~" + str(self.expr)

	def reset(self):
		self.value = None

		self.left.reset()
		self.right.reset()

	def evaluate(self, context):
		v1 = self.expr.evaluate(context)
		self.value = (~ v1)
		return self.value

	def evaluated(self):
		return "~" + self.expr.evaluated()

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class BitOrOpExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, left, right):
		super(BitOrOpExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.left = left
		self.right = right

	def __str__(self):
		return "(%s | %s)" % (self.left, self.right)

	def reset(self):
		self.value = None

		self.left.reset()
		self.right.reset()

	def evaluate(self, context):
		error = None

		try:
			v1 = self.left.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		try:
			v2 = self.right.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		if error:
			raise error

		(v1, v2) = self.coerce_type(v1, v2)

		self.value = (v1 | v2)
		return self.value

	def evaluated(self):
		return "(%s | %s)" % (self.left.evaluated(), self.right.evaluated())

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class BitXOrOpExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, left, right):
		super(BitXOrOpExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.left = left
		self.right = right

	def __str__(self):
		return "(%s ^ %s)" % (self.left,self.right)

	def reset(self):
		self.value = None

		self.left.reset()
		self.right.reset()

	def evaluate(self, context):
		error = None

		try:
			v1 = self.left.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		try:
			v2 = self.right.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		if error:
			raise error

		(v1, v2) = self.coerce_type(v1, v2)

		self.value = (v1 ^ v2)
		return self.value

	def evaluated(self):
		return "(%s ^ %s)" % (self.left.evaluated(), self.right.evaluated())

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class BitAndOpExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, left, right):
		super(BitAndOpExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.left = left
		self.right = right

	def __str__(self):
		return "(%s & %s)" % (self.left, self.right)

	def reset(self):
		self.value = None

		self.left.reset()
		self.right.reset()

	def evaluate(self, context):
		error = None

		try:
			v1 = self.left.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		try:
			v2 = self.right.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		if error:
			raise error

		(v1, v2) = self.coerce_type(v1, v2)

		self.value = v1 & v2
		return self.value

	def evaluated(self):
		return "(%s & %s)" % (self.left.evaluated(), self.right.evaluated())

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class EqOpExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, left, right):
		super(EqOpExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.left = left
		self.right = right

	def __str__(self):
		return "(%s == %s)" % (self.left, self.right)

	def reset(self):
		self.value = None

		self.left.reset()
		self.right.reset()

	def evaluate(self, context):
		error = None

		try:
			v1 = self.left.evaluate(context) 
		except NotEvaluatedError, er:
			error = er

		try:
			v2 = self.right.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		if error:
			raise error

		(v1, v2) = self.coerce_type(v1, v2)

		self.value = (v1 == v2)
		return self.value

	def evaluated(self):
		return "(%s == %s)" % (self.left.evaluated(), self.right.evaluated())

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class NotEqOpExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, left, right):
		super(NotEqOpExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.left = left
		self.right = right

	def __str__(self):
		return "(%s != %s)" % (self.left, self.right)

	def reset(self):
		self.value = None

		self.left.reset()
		self.right.reset()

	def evaluate(self, context):
		error = None

		try:
			v1 = self.left.evaluate(context) 
		except NotEvaluatedError, er:
			error = er

		try:
			v2 = self.right.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		if error:
			raise error

		(v1, v2) = self.coerce_type(v1, v2)

		self.value = (v1 != v2)
		return self.value

	def evaluated(self):
		return "(%s != %s)" % (self.left.evaluated(), self.right.evaluated())

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class RegExprEqOpExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, left, right):
		super(RegExprEqOpExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.left = left
		self.right = right

	def __str__(self):
		return "(%s ~= %s)" % (self.left, self.right)

	def reset(self):
		self.value = None

		self.left.reset()
		self.right.reset()

	def evaluate(self, context):
		error = None

		try:
			v1 = self.left.evaluate(context) 
		except NotEvaluatedError, er:
			error = er

		try:
			v2 = self.right.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		if error:
			raise error

		# FIX: Should we use our data types (NumberType, StrType, SnmpValueType, ecc...) instead 
		#      of using a mix of different types taken from python (int, float, ecc..) and ours
		#      (SnmpValueType)? 
		#      In this way we will always call the __regexpr_eq__() operator instead of
		#      controll the operands' type.

		# We don't know what is the value for LazyValueType object.
		# Let it calculate the result for us.
		if isinstance(v1, LazyValueType):
			self.value = v1.__regexpr_eq__(v2)
		elif isinstance(v2, LazyValueType):
			self.value = v2.__rregexpr_eq__(v1)
		else:
			# For simple values we can call the function
			self.value = regexp_eq(v1, v2)

		return self.value

	def evaluated(self):
		return "(%s =~ %s)" % (self.left.evaluated(), self.right.evaluated())

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class LtOpExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, left, right):
		super(LtOpExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.left = left
		self.right = right

	def __str__(self):
		return "(%s < %s)" % (self.left, self.right)

	def reset(self):
		self.value = None

		self.left.reset()
		self.right.reset()

	def evaluate(self, context):
		error = None

		try:
			v1 = self.left.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		try:
			v2 = self.right.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		if error:
			raise error

		(v1, v2) = self.coerce_type(v1, v2)

		self.value =  (v1 < v2)
		return self.value

	def evaluated(self):
		return "(%s < %s)" % (self.left.evaluated(), self.right.evaluated())

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class GtOpExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, left, right):
		super(GtOpExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.left = left
		self.right = right

	def __str__(self):
		return "(%s > %s) " % (self.left, self.right)

	def reset(self):
		self.value = None

		self.left.reset()
		self.right.reset()

	def evaluate(self, context):
		error = None

		try:
			v1 = self.left.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		try:
			v2 = self.right.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		if error:
			raise error

		(v1, v2) = self.coerce_type(v1, v2)

		self.value = (v1 > v2)
		return self.value

	def evaluated(self):
		return "(%s > %s)" % (self.left.evaluated(),  self.right.evaluated())

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class LEqOpExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, left, right):
		super(LEqOpExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.left = left
		self.right = right

	def __str__(self):
		return "(%s <= %s)" % (self.left, self.right)

	def reset(self):
		self.value = None

		self.left.reset()
		self.right.reset()

	def evaluate(self, context):
		error = None

		try:
			v1 = self.left.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		try:
			v2 = self.right.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		if error:
			raise error

		(v1, v2) = self.coerce_type(v1, v2)

		self.value = (v1 <= v2)
		return self.value

	def evaluated(self):
		return "(%s <= %s)" % (self.left.evaluated(), self.right.evaluated())

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class GEqOpExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, left, right):
		super(GEqOpExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.left = left
		self.right = right

	def __str__(self):
		return "(%s >= %s)" % (self.left, self.right)

	def reset(self):
		self.value = None

		self.left.reset()
		self.right.reset()

	def evaluate(self, context):
		error = None

		try:
			v1 = self.left.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		try:
			v2 = self.right.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		if error:
			raise error

		(v1, v2) = self.coerce_type(v1, v2)

		self.value = (v1 >= v2)
		return self.value

	def evaluated(self):
		return "(%s >= %s)" % (self.left.evaluated(), self.right.evaluated())

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class ConcatOIDExprTree(ExprTree):
	"""
	TODO
	"""
	
	def __init__(self, leftExpr, rightExpr):
		super(ConcatOIDExprTree, self).__init__()
		
		debugline( "init " + str(self.__class__) )
		self.left  = leftExpr
		self.right = rightExpr
			
	def __str__(self):
		return "%s.%s" % (self.left, self.right)
	
	def reset(self):
		self.value = None
		
		self.left.reset()
		self.right.reset()
	
	def evaluate(self, context):
		error = None
		try:
			v1 = self.left.evaluate(context)
		except NotEvaluatedError, er:
			error = er
		try:
			v2 = self.right.evaluate(context)
		except NotEvaluatedError, er:
			error = er
		
		if error:
			raise error
		
		# Special string conversion for floats:
		# 2.0  ->  '2'
		# 2.3  ->  '2.3'
		if isinstance(v1, types.FloatType) and (round(v1) == v1):
			s1 = str(int(v1))
		else:
			s1 = str(v1)
		
		if isinstance(v2, types.FloatType) and (round(v2) == v2):
			s2 = str(int(v2))
		else:
			s2 = str(v2)
		
		
		self.value = (s1 + "." + s2)
		return self.value

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class AddOpExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, left, right):
		super(AddOpExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.left = left
		self.right = right

	def __str__(self):
		return "(%s + %s)" % (self.left, self.right)

	def reset(self):
		self.value = None

		self.left.reset()
		self.right.reset()

	def evaluate(self, context):
		error = None

		try:
			v1 = self.left.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		try:
			v2 = self.right.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		if error:
			raise error

		self.value = (v1 + v2)
		return self.value

	def evaluated(self):
		return "(%s + %s)" % (self.left.evaluated(), self.right.evaluated())

class AddAddOpExprTree(ExprTree):
	"""
	TODO
	"""
	def __init__(self, expr):
		super(AddAddOpExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.expr = expr

	def __str__(self):
		return "++%s" % self.expr

	def reset(self):
		self.value = None

		self.expr.reset()

	def evaluate(self, context):
		v1 = self.expr.evaluate(context)
		self.value = v1 + 1
		return self.value

	def evaluated(self):
		return "++%s" % self.expr.evaluated()

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class SubOpExprTree(ExprTree):
	"""
	TODO
	"""
	def __init__(self, left, right):
		super(SubOpExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.left = left
		self.right = right

	def __str__(self):
		return "(%s - %s)" % (self.left, self.right)

	def reset(self):
		self.value = None

		self.left.reset()
		self.right.reset()

	def evaluate(self, context):
		error = None

		try:
			v1 = self.left.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		try:
			v2 = self.right.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		if error:
			raise error

		self.value = (v1 - v2)
		return self.value

	def evaluated(self):
		return "(%s - %s)" % (self.left.evaluated(), self.right.evaluated()) 

class SubSubOpExprTree(ExprTree):
	"""
	TODO
	"""
	def __init__(self, expr):
		super(SubSubOpExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.expr = expr

	def __str__(self):
		return "--%s" % self.expr

	def reset(self):
		self.value = None

		self.expr.reset()

	def evaluate(self, context):
		v1 = self.expr.evaluate(context)
		self.value = v1 - 1
		return self.value

	def evaluated(self):
		return "--%s" % (self.left.evaluated())

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class MinusOpExprTree(ExprTree):
	"""
	TODO
	"""
	def __init__(self, expr):
		super(MinusOpExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.expr = expr

	def __str__(self):
		return "-%s" % (self.expr)

	def reset(self):
		self.value = None

		self.expr.reset()

	def evaluate(self, context):
		v1 = self.expr.evaluate(context)
		self.value = (-v1)
		return self.value

	def evaluated(self):
		return "-%s" % (self.expr.evaluated())

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#
class MultiOpExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, left, right):
		super(MultiOpExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.left = left
		self.right = right

	def __str__(self):
		return "(%s * %s)" % (self.left, self.right)

	def reset(self):
		self.value = None

		self.left.reset()
		self.right.reset()

	def evaluate(self, context):
		error = None

		try:
			v1 = self.left.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		try:
			v2 = self.right.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		if error:
			raise error

		self.value = (v1 * v2)

		return self.value

	def evaluated(self):
		return "(%s * %s)" % (self.left.evaluated(), self.right.evaluated())

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#
class DivOpExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, left, right):
		super(DivOpExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.left = left
		self.right = right

	def __str__(self):
		return "(%s / %s)" % (self.left, self.right)

	def reset(self):
		self.value = None

		self.left.reset()
		self.right.reset()

	def evaluate(self, context):
		error = None
		try:
			v1 = self.left.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		try:
			v2 = self.right.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		if error:
			raise error

		# Show 
		v2 = float(v2)
		if v2 == 0:
			raise ZeroDivisionError("division %s / %s" % (v1,v2))

		self.value = (float(v1) / float(v2))
		return self.value

	def evaluated(self):
		return "(%s / %s)" % (self.left.evaluated(), self.right.evaluated())

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class ModOpExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, left, right):
		super(ModOpExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.left = left
		self.right = right

	def __str__(self):
		return "(%s mod %s)" % (self.left, self.right)

	def reset(self):
		self.value = None

		self.left.reset()
		self.right.reset()

	def evaluate(self, context):
		error = None
		try:
			v1 = self.left.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		try:
			v2 = self.right.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		if error:
			raise error

		self.value = (int(v1) % int(v2))
		return self.value

	def evaluated(self):
		return "(%s mod %s)" % (self.left.evaluated(), self.right.evaluated())

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#
class PowerOpExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, left, right):
		super(PowerOpExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )
		self.left = left
		self.right = right

	def __str__(self):
		return "(%s ** %s)" % (self.left, self.right)

	def reset(self):
		self.value = None

		self.left.reset()
		self.right.reset()

	def evaluate(self, context):
		error = None
		try:
			v1  = self.left.evaluate(context)
		except NotEvaluatedError, er:
			error = er
		try:
			v2  = self.right.evaluate(context)
		except NotEvaluatedError, er:
			error = er

		if error:
			raise error

		self.value = (v1 ** v2)
		return self.value

	def evaluated(self):
		return "(%s ** %s)" % (self.left.evaluated(), self.right.evaluated())

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class TryExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, expr, catch_list):
		super(TryExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )

		self.expr = expr
		self.catch_list = catch_list

		self.catched_expr = None

	def __str__(self):

		s = "try %s " % (self.expr)
		for (e, expr) in self.catch_list:
			s += 'catch %s ( %s )' % (e, expr)
		return s

	def reset(self):
		self.value = None

		self.expr.reset()

		if self.catched_expr:
			ex, expr = self.catched_expr
			expr.reset()

			self.catched_expr = None

		return

	def evaluate(self, context):

		error = None
		exception = None

		try:
			value  = self.expr.evaluate(context)

			if isinstance(value, LazyValueType):
				value = value.getvalue()

		except NotEvaluatedError, er:
			error = er
		except Exception, ex:
			exception = ex

		if error:
			raise error

		if exception:
			
			
			valid_catchable_exceptions_names = [ ex_class.__name__ for ex_class in list(inspect.getmro(exception.__class__)) ]

			for catch_name, catch_expr in self.catch_list:

				if catch_name is not None:
					if catch_name in valid_catchable_exceptions_names:
						self.catched_expr = (exception, catch_name, catch_expr)
						break
				else:
					self.catched_expr = (exception, None, catch_expr)

			if self.catched_expr is None:
				raise exception

			try:
				exception, catch_name, catch_expr = self.catched_expr

				if catch_expr is not None:

					context['$_exception_'] = exception

					value2 = catch_expr.evaluate(context)

					if isinstance(value2, LazyValueType):
						value2 = value2.getvalue()
				else:
					value2 = "%s:%s" % (exception.__class__.__name__, str(exception))

				value = value2

			except NotEvaluatedError, ex:
				error = ex

			if error:
				raise error

		self.value = value

		return self.value


	def evaluated(self):
		if self.catched_expr is not None:

			exception, catch_name, catched_expr = self.catched_expr

			s1 = self.expr.evaluated()
			s2 = catch_name or exception.__class__.__name__
			s3 = catched_expr.evaluated() if catched_expr is not None else str(exception)

			s = "(tried (%s) catched %s -> (%s))" % (s1, s2, s3)

			return s
		else:
			return self.expr.evaluated()

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class IfExprTree(ExprTree):
	"""
	TODO
	"""

	def __init__(self, cond_expr, then_expr, else_expr):
		super(IfExprTree, self).__init__()

		debugline( "init " + str(self.__class__) )

		self.cond_expr = cond_expr
		self.then_expr = then_expr
		self.else_expr = else_expr
		
		self._evaluated_expr = None  # pointer to then_epxr or else_expr.

	def __str__(self):
		s = "(if %s then %s else %s)" % (self.cond_expr, self.then_expr, self.else_expr)
		return s

	def reset(self):
		self.value = None

		self.cond_expr.reset()
		self.then_expr.reset()
		self.else_expr.reset()

	def evaluate(self, context):

		cond_value = self._eval_expr(context, self.cond_expr)

		#if cond_value:
		#	self._evaluated_expr = self.then_expr
		#
		#	value = self._eval_expr(context, self.then_expr)
		#
		#else:
		#	self._evaluated_expr = self.else_expr
		#
		#	value = self._eval_expr(context, self.else_expr)


		# Choose
		if cond_value:
			self._evaluated_expr = self.then_expr
		else:
			self._evaluated_expr = self.else_expr

		# Evalutate
		value = self._eval_expr(context, self._evaluated_expr)


		return value
	
	def _eval_expr(self, context, expr_tree):
		value     = None
		error     = None

		try:
			value  = expr_tree.evaluate(context)

			if isinstance(value, LazyValueType):
				value = value.getvalue()

		except NotEvaluatedError, er:
			error = er

		if error:
			raise error

		return value

	def evaluated(self):
		return "(if %s then %s)" % (self.cond_expr.evaluated(), self._evaluated_expr.evaluated())


#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class BaseExprFactory:
	"""
	This class is used to create an :class:`ExprTree`
	object rappresenting the parsed expression.
	
	Every sub-expression is converted :class:`ExprTree` object by calling the
	corrisponding method of	this class.
	"""

	def __init__(self):
		pass

	def createRoot(self, exprTree):
		"""
		Returns a :class:`RootExprTree` object.
		"""
		return RootExprTree(exprTree)

	def createIdentifier(self, strId):
		"""
		Returns a :class:`IdentifierExprTree` object.
		"""
		return IdentifierExprTree(strId)

	def createDataIdentifier(self, dataId):
		"""
		Returns a :class:`DataIdentifierExprTree` object.
		"""
		return DataIdentifierExprTree(dataId)

	def createSnmpValue(self, exprOID, eCommunity, charType, eHost, ePort):
		"""
		Returns a :class:`SnmpValueExprTree` object.
		"""
		return SnmpValueExprTree(exprOID, eCommunity, charType, eHost, ePort)

	def createOID(self, strValue):
		"""
		Returns a :class:`OIDLiteralExprTree` object.
		"""
		return OIDLiteralExprTree(strValue)

	def createBool(self, boolValue):
		"""
		Returns a :class:`BoolExprTree` object.
		"""
		return BoolExprTree(boolValue)

	def createString(self, strValue, escape):
		"""
		Returns a :class:`StringLiteralExprTree` object.
		"""
		return StringLiteralExprTree(strValue, escape)

	def createInteger(self, decimalValue):
		"""
		Returns a :class:`IntegerLiteralExprTree` object.
		"""
		return IntegerLiteralExprTree(decimalValue)

	def createFloat(self, floatValue):
		"""
		Returns a :class:`FloatLiteralExprTree` object.
		"""
		return FloatLiteralExprTree(floatValue)

	def createChar(self, charValue):
		"""
		Returns a :class:`CharLiteralExprTree` object.
		"""
		return CharLiteralExprTree(charValue)

	def createConcatOID(self, leftExpr, rightExpr):
		"""
		Returns a :class:`ConcatOIDExprTree` object.
		"""
		return ConcatOIDExprTree(leftExpr, rightExpr)

	def createCallOp(self, funTree, argsList):
		"""
		Returns a :class:`CallOpExprTree` object.
		"""
		return CallOpExprTree(funTree, argsList)

	def createNamedArgument(self, arg_name, arg_expr):
		return NamedArgument(str(arg_name), arg_expr)

	def createRangeOp(self, tableTree, indexlist):
		"""
		Returns a :class:`RangeOpExprTree` object.
		"""
		return RangeOpExprTree(tableTree, indexlist)

	def createOrOp(self, left, right):
		"""
		Returns a :class:`OrOpExprTree` object.
		"""
		return OrOpExprTree(left, right)

	def createAndOp(self, left, right):
		"""
		Returns a :class:`AndOpExprTree` object.
		"""
		return AndOpExprTree(left, right)

	def createNotOp(self, expr):
		"""
		Returns a :class:`NotOpExprTree` object.
		"""
		return NotOpExprTree(expr)

	def createBitOneOp(self, expr):
		"""
		Returns a :class:`BitOneOpExprTree` object.
		"""
		return BitOneOpExprTree(expr)

	def createBitOrOp(self, left, right):
		"""
		Returns a :class:`BitOrOpExprTree` object.
		"""
		return BitOrOpExprTree(left, right)

	def createBitXOrOp(self, left, right):
		"""
		Returns a :class:`BitXOrOpExprTree` object.
		"""
		return BitXOrOpExprTree(left, right)

	def createBitAndOp(self, left, right):
		"""
		Returns a :class:`BitAndOpExprTree` object.
		"""
		return BitAndOpExprTree(left, right)
	
	def createEqOp(self, left, right):
		"""
		Returns a :class:`EqOpExprTree` object.
		"""
		return EqOpExprTree(left, right)

	def createNotEqOp(self, left, right):
		"""
		Returns a :class:`NotEqOpExprTree` object.
		"""
		return NotEqOpExprTree(left, right)

	def createRegExpEqOp(self, left, right):
		"""
		Returns a :class:`RegExprEqOpExprTree` object.
		"""
		return RegExprEqOpExprTree(left, right)

	def createLtOp(self, left, right):
		"""
		Returns a :class:`LtOpExprTree` object.
		"""
		return LtOpExprTree(left, right)

	def createGtOp(self, left, right):
		"""
		Returns a :class:`GtOpExprTree` object.
		"""
		return GtOpExprTree(left, right)

	def createLEqOp(self, left, right):
		"""
		Returns a :class:`LEqOpExprTree` object.
		"""
		return LEqOpExprTree(left, right)

	def createGEqOp(self, left, right):
		"""
		Returns a :class:`GEqOpExprTree` object.
		"""
		return GEqOpExprTree(left, right)

	def createAddOp(self, left, right):
		"""
		Returns a :class:`AddOpExprTree` object.
		"""
		return AddOpExprTree(left, right)

	def createAddAddOp(self, expr):
		"""
		Returns a :class:`AddAddOpExprTree` object.
		"""
		return AddAddOpExprTree(expr)

	def createSubOp(self, left, right):
		"""
		Returns a :class:`SubOpExprTree` object.
		"""
		return SubOpExprTree(left, right)

	def createSubSubOp(self, expr):
		"""
		Returns a :class:`SubSubOpExprTree` object.
		"""
		return SubSubOpExprTree(expr)

	def createMinusOp(self, expr):
		"""
		Returns a :class:`MinusOpExprTree` object.
		"""
		return MinusOpExprTree(expr)

	def createMultiOp(self, left, right):
		"""
		Returns a :class:`MultiOpExprTree` object.
		"""
		return MultiOpExprTree(left, right)

	def createDivOp(self, left, right):
		"""
		Returns a :class:`DivOpExprTree` object.
		"""
		return DivOpExprTree(left, right)

	def createModOp(self, left, right):
		"""
		Returns a :class:`ModOpExprTree` object.
		"""
		return ModOpExprTree(left, right)

	def createPowerOp(self, left, right):
		"""
		Returns a :class:`PowerOpExprTree` object.
		"""
		return PowerOpExprTree(left, right)

	def createTry(self, expr, catch_list):
		"""
		Returns a :class:`TryExprTree` object.
		"""
		return TryExprTree(expr, catch_list)

	def createIf(self, expr1, expr2, expr3):
		"""
		Returns a :class:`TryExprTree` object.
		"""
		return IfExprTree(expr1, expr2, expr3)

