 
/* Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
 *
 * This file is part of SANET
 * SANET is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, version 3 of the License
 *
 * SANET is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with SANET. If not, see <http://www.gnu.org/licenses/>.
 */

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*                                                                           */
/* SANET Poller Scripting Language                                           */
/*                                                                           */
/* History:                                                                  */
/*                                                                           */
/*    2008-02-12 - dbilli - v0.1 (tested with ANTLR v3.1.1)                  */
/*    2017-07-01 - dbilli - v1.2 (tested with ANTLR v3.1.1)                  */
/*                                                                           */
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/

grammar ExprScript;

options {
    backtrack=true; 
    language=Python;    
}

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*                                                                           */
/* PollerLangLexer's code                                                    */
/*                                                                           */
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*
 *   Imports and globals symbols
 */
@lexer::header {

from .ExprFactory import LexicalError

}

/*
 *   Code for the PollerLangLexer.__init__().
 */
@lexer::init {
    pass
}

/*
 *   PollerLangLexer's methods.
 */
@lexer::members {

#
# Implements a simple handler for lexical exceptions.
# Stop lexical analisys as soon as it discovers an error.
#
def reportError(self, e):
    raise LexicalError(e)

#
# Debug methods
#
def debug(self, str):
#	print str,
	pass

#
# The next token belogs to the previews OID token
# if is stars with a '.' and a number
#
#   previews oid      next token
#   1.2.3.4.5.6.7  +  .121.1.3.2.3
#                     ^^
#
# FIX: This is an hack function for the OID_SEQUENCE token!!
#
def checkOID(self):
	return (chr(self.input.LA(1)) == '.') and (chr(self.input.LA(2)) in "0123456789")
    	
}

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*                                                                           */
/* PollerLangParser's code                                                   */
/*                                                                           */
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/

/*
 *   Imports and globals symbols
 */
@header { 

#
# Import parsing classes (Tree, Exceptions, ecc....)
#
from .ExprFactory import BaseExprFactory
from .ExprFactory import ParsingError

}

/* 
 *   Code for the PollerLangParser.__init__().
 */
@init {
    self.setExprFactory( None )
}

/*
 *   PollerLangParser's methods.
 */
@members {
#
#
# Implements a simple handler for parsing exceptions.
# Stop parsing as soon as it discovers an error.
#
def reportError(self, e):
    raise ParsingError(e);	 
	
#
# Debug methods
#	
def debug(self, str):
#	print str,
	pass

def debugline(self, str):
#	print str
	pass

def initParsing(self):
	if self.f == None:
		self.setExprFactory( BaseExprFactory() )
	
def setExprFactory(self, factory):
	self.f = factory

}


/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*                                                                           */
/* TOKENS                                                                    */
/*                                                                           */
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/

//NEWLINE: ('\r'? '\n')+                                  { $channel=HIDDEN } ;

WS: (' ' | '\r' | '\t' | '\u000C' | '\n')+              { self.debug("' '") } ;

//SINGLE_COMMENT:	'//' ~('\r' | '\n')* ('\r'? '\n')+      { self.skip(); } ;

//MULTI_COMMENT                                           options { greedy = false; }
//	: '/*' .* '*/'  ('\r'? '\n')*                   { self.skip(); } ;

/*---------------------------------------------------------------------------*/
/* Syntax elements                                                           */
/*---------------------------------------------------------------------------*/

//
// Boolean constants
//
TRUE : 'True'                                           { self.debug("BOOL(true)") } ;
FALSE: 'False'                                          { self.debug("BOOL(false)") } ;
AND  : 'and'                                            { self.debug("and") } ;
OR   : 'or'                                             { self.debug("or") } ;
NOT  : 'not'                                            { self.debug("not") } ;
EQ   : 'eq'                                             { self.debug("eq") };


//
// Expression's elements
//
LEFT_PAREN: '('                                         { self.debug("'"+self.getText()+"'") };
RIGHT_PAREN: ')'                                        { self.debug("'"+self.getText()+"'") };
LEFT_SPAREN: '['                                        { self.debug("'"+self.getText()+"'") };
RIGHT_SPAREN: ']'                                       { self.debug("'"+self.getText()+"'") };
COMMA: ','                                              { self.debug("'"+self.getText()+"'") };
SEMICOLON: ';'                                          { self.debug("'"+self.getText()+"'") };

// OID's syntax elements
COLUMN: ':'						{ self.debug("'"+self.getText()+"'") };
TYPE_NEW: '@'                                           { self.debug("'"+self.getText()+"'") };
TYPE_OLD: '#'                                           { self.debug("'"+self.getText()+"'") };

// Unescaped String prefix
UNESCAPE_STR: 'r'                                       { self.debug("UNESCPAE") } ;


//
// Operators
//

ASSIGN: '='                                             { self.debug("'"+self.getText()+"'") };

// OIDs concat operator
DOT	: '.'                                           { self.debug("'"+self.getText()+"'") };

PLUS_PLUS: '++'                                         { self.debug("'"+self.getText()+"'") };
MINUS_MINUS: '--'                                       { self.debug("'"+self.getText()+"'") };

POWER: '**'                                             { self.debug("power") };

PLUS: '+'                                               { self.debug("'"+self.getText()+"'") };
MINUS: '-'                                              { self.debug("'"+self.getText()+"'") };
MULTI: '*'                                              { self.debug("'"+self.getText()+"'") };
DIV: '/'                                                { self.debug("'"+self.getText()+"'") };
MOD: '%'                                                { self.debug("'"+self.getText()+"'") };


EQUAL : '=='                                            { self.debug("'"+self.getText()+"'") };
NEQUAL: '!='                                            { self.debug("'"+self.getText()+"'") };

REGEXPQUAL: '=~'                                        { self.debug("'"+self.getText()+"'") };

GT  : '>'                                               { self.debug("'"+self.getText()+"'") };
LT  : '<'                                               { self.debug("'"+self.getText()+"'") };
GEQ : '>='                                              { self.debug("'"+self.getText()+"'") };
LEQ : '<='                                              { self.debug("'"+self.getText()+"'") };

LOGIC_NOT : '!'                                         { self.debug("'"+self.getText()+"'") };
LOGIC_AND : '&&'                                        { self.debug("'"+self.getText()+"'") };
LOGIC_OR  : '||'                                        { self.debug("'"+self.getText()+"'") };

BITAND : '&'                                            { self.debug("'"+self.getText()+"'") };
BITOR  : '|'                                            { self.debug("'"+self.getText()+"'") };
BITXOR : '^'                                            { self.debug("'"+self.getText()+"'") };

//One's complement operator
BITONE : '~'                                            { self.debug("'"+self.getText()+"'") };

IF: 'if'                                                { self.debug("'"+self.getText()+"'") };
THEN: 'then'                                            { self.debug("'"+self.getText()+"'") };
ELSE: 'else'                                            { self.debug("'"+self.getText()+"'") };

TRY  : 'try'                                            { self.debug("'"+self.getText()+"'") };
CATCH: 'catch'                                          { self.debug("'"+self.getText()+"'") };

/*---------------------------------------------------------------------------*/
//  Language Types
/*---------------------------------------------------------------------------*/

//
// SNMP OIDs: sequence of 3 or more numbers separated by dot.
//
// NOTE: ANTLR's lexer and parser are greedy by default. The 'greedy=false' 
// options doesn't works (BUG!?!).
//
// Check for OID sequence by hand with the checkOID() method.
//
OID_SEQUENCE
    : DIGIT+ '.' DIGIT+ '.' DIGIT+ 
     ( { self.checkOID()==True }? => '.' DIGIT+)*
                                                              { self.debug("OID(" + self.getText() + ")") }
    ;
    
//
// Literals
//
//CHAR: '\'' ( EscapeSequence | ~( '\'' | '\\' ) ) '\''        { self.debug("CHAR("   + self.getText() + ")") };
//STRING:  '"' ( EscapeSequence | ~( '\\' | '"' ) )* '"'       { self.debug("STRING(" + self.getText() + ")") };
//RSTRING:  'r' '"' ( EscapeSequence | ~( '\\' | '"' ) )* '"'  { self.debug("STRING(" + self.getText() + ")") };

STRING
    :   ('\'') ( EscapeSequence | ~( '\'' | '\\' ) )* ('\'')   { self.debug("STRING(" + self.getText() + ")") }
    |   ('"' ) ( EscapeSequence | ~( '"'  | '\\' ) )* ('"')    { self.debug("STRING(" + self.getText() + ")") }
    ;
    
RSTRING  
    : 'r' ('\'') ( ESC | ~( '\'' | '\\' ) )* ('\'')            { self.debug("STRING(" + self.getText() + ")") }
    | 'r' ('"') ( ESC | ~( '\"'  | '\\' ) )* ('"')             { self.debug("STRING(" + self.getText() + ")") }
    ;


fragment EscapeSequence
    :   '\\' ('b'|'t'|'n'|'f'|'r'|'\"'|'\''|'\\')
    |   OctalEscape
    ;
fragment OctalEscape
    :   '\\' ('0'..'3') ('0'..'7') ('0'..'7')
    |   '\\' ('0'..'7') ('0'..'7')
    |   '\\' ('0'..'7')
    ;
    
fragment ESC
    :   '\\' .
    ;
//
// Variable identifiers ($a, $a1, $_a,  $a_a123)
//
VAR_ID: '\$' (ALPHA|'_') ((ALPHA|'_') | DIGIT)*          { self.debug("VAR(" + self.getText() + ")") } ;

//
// Function identifiers
//
FUNCTION_ID: (ALPHA|'_') ((ALPHA|'_') | DIGIT)*          { self.debug("ID(" + self.getText() + ")") } ;
    
fragment ALPHA: ( 'A'..'Z' | 'a'..'z');
fragment DIGIT: '0'..'9';
    
//
// Numbers
//
HEX: '0' ('x'|'X') HexDigit+                            { self.debug("HEX(" + self.getText() + ")") };
fragment HexDigit : ('0'..'9'|'a'..'f'|'A'..'F') ;

DEC: ('0' | '1'..'9' '0'..'9'*)                         { self.debug("DEC(" + self.getText() + ")") };

OCTAL: '0' ('0'..'7')+                                  { self.debug("OCTAL(" + self.getText() + ")") };

FLOAT: 
//     ('0'..'9')+ '.' ('0'..'9')* Exponent?              { self.debug("FLOAT(" + self.getText() + ")") }
     ('0'..'9')+ '.' ('0'..'9')+ Exponent?              { self.debug("FLOAT(" + self.getText() + ")") }
//     | '.' ('0'..'9')+ Exponent?                        { self.debug("FLOAT(" + self.getText() + ")") }
     | ('0'..'9')+ Exponent?                            { self.debug("FLOAT(" + self.getText() + ")") }
     ;
fragment Exponent : ('e'|'E') ('+'|'-')? ('0'..'9')+ ;


//
// Special identifiers
//
DATA_ID: '{' (ALPHA|'_'|'-') ((ALPHA|'_'|'-') | DIGIT | '.')*  ('@'|'#') '}'  { self.debug("DATA(" + self.getText() + ")") } ;

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*                                                                           */
/* GRAMMAR SYNTAX                                                            */
/*                                                                           */
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Initiali RULE                                                             */
/*---------------------------------------------------------------------------*/
start  returns [Node]
    : { self.initParsing() }
      WS* ce=cond_expr WS* EOF                                  { Node = self.f.createRoot(ce) }
    ;

/*---------------------------------------------------------------------------*/
// a, b, c, d, ...
/*---------------------------------------------------------------------------*/
argument_expr_list returns [Node]
    : ce1=argument                                     { r = [ce1] }
      (WS* COMMA WS*                                    
       ce2=argument                                    { r.append(ce2) }
      )*
                                                        { Node = r }
    ;

/*
argument returns [Node]
    : ce1=cond_expr                                      { r    = ce1 }
                                                         { Node = r }
    ;
*/


argument returns [Node]
    : 
      
      (options {k=2;}: 
       vi=FUNCTION_ID                                    { arg_name = $vi.text }
       ASSIGN
       ce2=cond_expr                                     { r        = self.f.createNamedArgument(arg_name, ce2) }
      )
                                                         { Node = r }
        |                                                 
	ce1=cond_expr                                    { Node = ce1 }                                                         
    ;

    
/*---------------------------------------------------------------------------*/
// : , a: ,  a:b
/*---------------------------------------------------------------------------*/
//range_expr returns [Node]
//    : COMMA                                             { Node = [None,None] }
//    | s1=calc_expr                                      { r = [s1, None] }
//      (WS* 
//       COMMA 
//       (WS* e1=calc_expr                                { r = [s1,e1] }
//       )?
//      )                                                 { Node = r }
//    ;
    
/*---------------------------------------------------------------------------*/
// a + b
// a - b
/*---------------------------------------------------------------------------*/
add_expr returns [Node]
    : me1=mult_expr		                        { r = me1 }
      (WS* 
       ( PLUS   WS*  me2=mult_expr                      { r = self.f.createAddOp(r,me2) }
       | MINUS  WS*  me2=mult_expr                      { r = self.f.createSubOp(r,me2) }
       )
      )*
                                                        { Node = r }
    ;

/*---------------------------------------------------------------------------*/
// a * b
// a / c
// a % c
/*---------------------------------------------------------------------------*/
mult_expr returns [Node]
    : ue1=power_expr                                    { r = ue1 }
      (WS*
       ( MULTI WS* ue2=power_expr                       { r = self.f.createMultiOp(r,ue2) }
       | DIV   WS* ue2=power_expr                       { r = self.f.createDivOp(r,ue2) }
       | MOD   WS* ue2=power_expr                       { r = self.f.createModOp(r,ue2) }
       )
      )*                       
                                                        { Node = r }
;
/*---------------------------------------------------------------------------*/
//  expr ** expr
/*---------------------------------------------------------------------------*/

power_expr returns [Node]
    : ue1=unary_expr                                    { r = ue1 }
      (WS* POWER WS* ue2=unary_expr                     { r = self.f.createPowerOp(r,ue2) }
      )*                       
                                                        { Node = r }
;
/*---------------------------------------------------------------------------*/
//  id , id()
//  + expr 
//  - expr 
//  ~ expr
//  ! expr
/*---------------------------------------------------------------------------*/

unary_expr returns [Node]
    : 
      ie=if_expr                                        { Node = ie }
    | te=try_expr                                       { Node = te }
    | ce=calc_expr                                      { Node = ce }
    | se=snmp_value_expr                                { Node = se }
    ;

calc_expr returns [Node]
    : pe=postfix_expr                                   { Node = pe }
    | PLUS_PLUS  WS* ue=unary_expr                      { Node = self.f.createAddAddOp(ue) }
    | MINUS_MINUS WS* ue=unary_expr                     { Node = self.f.createSubSubOp(ue) }
    | PLUS WS* ue=unary_expr                            { Node = ue }
    | MINUS WS* ue=unary_expr                           { Node = self.f.createMinusOp(ue) }
    | (LOGIC_NOT|NOT) WS* ue=unary_expr                 { Node = self.f.createNotOp(ue) }
    | BITONE WS* ue=unary_expr                          { Node = self.f.createBitOneOp(ue) }
    ;


//
// IF <expr> THEN <expr ELSE <expr>
//
if_expr returns [Node]
    : 
      IF
      WS*
      e1=cond_expr2
      WS* 
      THEN
      WS*
      e2=cond_expr2 
      WS*
      ELSE
      WS* 
      e3=cond_expr2                                         { Node = self.f.createIf(e1, e2, e3) }
    ;


//
//  try <expr> catch [ <id> ] ( <expr> ) [ catch <id> ( <expr> ) ....]
//
try_expr returns [Node]
    : 
      TRY 
      WS* 
      e=cond_expr 
      WS* 
      cl=catch_list                                                  { Node = self.f.createTry(e, cl) }
    ;

catch_list returns [Node]
    :
      CATCH WS* v1=FUNCTION_ID? WS* '(' WS* e1=cond_expr? WS* ')'    { r = [ ( $v1.text if $v1 is not None else None ,e1) ] }
      (WS*                                                           
       CATCH                                                         { v2, e2 = None, None # Reset v2 and e2 before using them in the rule }
             WS* v2=FUNCTION_ID? WS* '(' WS* e2=cond_expr? WS* ')'   { r.append( ($v2.text if $v2 is not None else None,e2) ) }
      )*
                                                                     { Node = r }
    ;












/*---------------------------------------------------------------------------*/
// OID Expressions
/*---------------------------------------------------------------------------*/
// These expressions rappresent SNMP OID values: 
//
//    <oid expression>  [':' <community-expr>] '@' [ <host-expr> [':' <port-expr>] ] 
//
// where <oid expression> is:
//
//    n.n.n '.' <exp-1> '.' <exp-2> '.' <exp-n>
//
/*---------------------------------------------------------------------------*/
snmp_value_expr returns [Node]
    : se1=snmp_single_expr                              { r = se1 }
     (WS*
      ie1=range_expr                                    { r = self.f.createRangeOp(r, ie1) }
     )*
	                                                { Node = r }
    ;
    
snmp_single_expr returns [Node]
    : {r1=None
       r2=None
       r3=None
       r4=None
      }
      oc1=oid_compositon                                { r1 = oc1 }
      (COLUMN pe1=postfix_expr                          { r2 = pe1 }           
      )?    		                 
      (WS* (t=TYPE_OLD | t=TYPE_NEW))                   { ty = $t.text }
      (pe2=postfix_expr                                 { r3 = pe2 }
       (COLUMN pe3=postfix_expr                         { r4 = pe3 }
       )?                                     
      )? 	       	                                { r = self.f.createSnmpValue(r1,r2,ty,r3,r4)}
	                                                { Node = r }
    ;

//
//  The DOT ('.') operator is a bit tricky: expressions are converted 
//  into strings and concatenated.
//
//  This means that if i concatenate OID  1.2.3.4  with the float 
//  literal 5.6  the result is  1.2.3.4.5.6
//
oid_compositon returns [Node]
    : os1=OID_SEQUENCE                                  { r = self.f.createOID($os1.text) }
      (WS* DOT WS*                                                  
       ( pe1=postfix_expr                               { r2 = pe1 }
       | os2=OID_SEQUENCE                               { r2 = self.f.createOID($os2.text)  }
       )                                                { r =  self.f.createConcatOID(r, r2)  }
      )*                                                      
                                                        { Node = r }
    ;
		
/*---------------------------------------------------------------------------*/
// funzione()
// funzione(a,b,...)
// $id()
// $id(a,b,...)
/*---------------------------------------------------------------------------*/
postfix_expr returns [Node]
    : 
      // Il risultato della valutazione di una espressione o una variabile
      // puo' essere un puntatore a funzione e quindi deve poter essere
      // gestito come tale!
      pe=primary_expr                                   { r = pe }
      (WS*
        (ce1=call_expr                                  { r = self.f.createCallOp(r, ce1) }
        |ie1=range_expr                                 { r = self.f.createRangeOp(r, ie1) }
        )
      )*                           
                                                        { Node = r }
      | 
      // Gli identificatori di funzione devono sempre essere sequiti
      // almeno dalla chiamata!
      fi=FUNCTION_ID                                    { r = self.f.createIdentifier($fi.text) }
      WS*
      ce1=call_expr                                     { r = self.f.createCallOp(r, ce1) }
      (WS*
        (ce1=call_expr                                  { r = self.f.createCallOp(r, ce1) }
       |ie1=range_expr                                  { r = self.f.createRangeOp(r, ie1) }
        )
      )*      
      //                                                  { Node = r }
      //fi=FUNCTION_ID                                    { first_match = 1
      //                                                    r = self.f.createString('"'+$fi.text+'"',False)
      //                                                  }	
      //(WS*
      //   (ce1=call_expr                                 { if first_match == 1:
      //                                                        r = self.f.createIdentifier($fi.text) 
      //                                                        first_match = 0
      //                                                    r = self.f.createCallOp(r, ce1)
      //                                                  }
      //   |ie1=range_expr                                { r = self.f.createRangeOp(r, ie1) }
      //   )
      //)*      
                                                         { Node = r }
    ;

//  ( ), (a,b,c,...)
call_expr returns [Node]
    : LEFT_PAREN WS* RIGHT_PAREN                         { Node = [] }
      |
      LEFT_PAREN WS* 
      ae1=argument_expr_list 
      WS* RIGHT_PAREN                                    { Node = ae1 }
    ;
    
// [i],   [i;j]
range_expr returns [range]
    : LEFT_SPAREN WS*                                   
      ue1=unary_expr?                                    { range = [ue1] }
      (WS* SEMICOLON WS*
        (                                                { range = range + [None] }
         ue2=unary_expr?                                 { range[1]= ue2 }
        )
      )?   
      WS* RIGHT_SPAREN                                 
    ;
    
/*---------------------------------------------------------------------------*/
/* Main expression 
/*---------------------------------------------------------------------------*/
cond_expr  returns [Node]
    : 
      oe=cond_expr2                                     { Node = oe }
    ;


cond_expr2  returns [Node]
    : 
      oe=or_expr                                        { Node = oe }
    ;

/*---------------------------------------------------------------------------*/
//  a || b
/*---------------------------------------------------------------------------*/
or_expr returns [Node]
    : ae1=and_expr                                      { r = ae1 }
      (WS* 
       (LOGIC_OR| OR ) 
       WS* ae2=and_expr                                 { r = self.f.createOrOp(r, ae2) }
      )*
                                                        { Node = r }
    ;

/*---------------------------------------------------------------------------*/
// a && b
/*---------------------------------------------------------------------------*/
and_expr returns [Node]
    : ioe1=inclusive_or_expr                            { r = ioe1 }
      (WS* 
      (LOGIC_AND | AND) 
       WS* ioe2=inclusive_or_expr                       { r = self.f.createAndOp(r, ioe2) }
      )*
                                                        { Node = r }
    ;

/*---------------------------------------------------------------------------*/
// a | b
/*---------------------------------------------------------------------------*/
inclusive_or_expr returns [Node]
    : ioe1=exclusive_or_expr                            { r = ioe1 }
      (WS* BITOR WS* ioe2=exclusive_or_expr             { r = self.f.createBitOrOp(r, ioe2) }
      )*
                                                        { Node = r }
    ;

/*---------------------------------------------------------------------------*/
// a ^ b
/*---------------------------------------------------------------------------*/
exclusive_or_expr returns [Node]
    : eae1=exclusive_and_expr                           { r = eae1 }
      (WS* BITXOR WS* eae2=exclusive_and_expr           { r = self.f.createBitXOrOp(r, eae2) }
      )* 
                                                        { Node = r }
    ;

/*---------------------------------------------------------------------------*/
// a & b
/*---------------------------------------------------------------------------*/
exclusive_and_expr returns [Node]
    : ee1=equal_expr                                    { r = ee1 }
      (WS* BITAND WS* ee2=equal_expr                    { r = self.f.createBitAndOp(r, ee2) }
      )* 
                                                        { Node = r }
    ;

/*---------------------------------------------------------------------------*/
// a == b
// a != b
// a ~= b   
/*---------------------------------------------------------------------------*/
equal_expr returns [Node]
    : re1=relational_expr                               { r = re1 }
      (WS*
        ( (EQUAL|EQ) WS* re2=relational_expr            { r = self.f.createEqOp(r, re2) }
        | NEQUAL WS* re2=relational_expr                { r = self.f.createNotEqOp(r, re2) }
        | REGEXPQUAL WS* re2=relational_expr            { r = self.f.createRegExpEqOp(r, re2) }
        )
      )*
                                                        { Node = r }
    ;

/*---------------------------------------------------------------------------*/
// a < b
// a > b
// a <= b
// a >= b
/*---------------------------------------------------------------------------*/
relational_expr returns [Node]
    : ae1=add_expr                                      { r = ae1 }
    (WS* 
      ( LT  WS* ae2=add_expr                            { r = self.f.createLtOp(r, ae2) }
      | GT  WS* ae2=add_expr                            { r = self.f.createGtOp(r, ae2) }
      | LEQ WS* ae2=add_expr                            { r = self.f.createLEqOp(r, ae2) }
      | GEQ WS* ae2=add_expr                            { r = self.f.createGEqOp(r, ae2) }
      )
    )*
                                                        { Node = r }
    ;

/*---------------------------------------------------------------------------*/
//  variabile
//  costante    (1, 1.0, 'c', "foo", ecc...)
//  ( ... )     espressione tra parentesi
/*---------------------------------------------------------------------------*/
primary_expr returns [Node]
    : 
      vi=VAR_ID                                         { Node = self.f.createIdentifier($vi.text)     }
    | di=DATA_ID                                        { Node = self.f.createDataIdentifier($di.text) }
    | c=constant                                        { Node = c }
    | '(' WS* ce=cond_expr WS* ')'                      { Node = ce }
    ;

constant returns [Node]
    :   
      c=TRUE                                            { Node = self.f.createBool(True) }
    | c=FALSE                                           { Node = self.f.createBool(False) }
    | c=DEC                                             { Node = self.f.createInteger(int($c.text,10)) }
    | c=OCTAL                                           { Node = self.f.createInteger(int($c.text, 8)) }
    | c=HEX                                             { Node = self.f.createInteger(int($c.text,16)) }
    | c=FLOAT                                           { Node = self.f.createFloat(float($c.text)) }
    | c=STRING                                          { Node = self.f.createString($c.text,True) }
    | c=RSTRING                                         { Node = self.f.createString($c.text[1:],False) }
//    | c=CHAR                                            { Node = self.f.createChar($c.text) }
//    | c=OID_SEQUENCE                                    { Node = self.f.createOID($c.text) }
    ;
