 
# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

========================================================================
:mod:`Types`
========================================================================

Introduzione
--------------------------------------------------------------------

Nel linguaggio del poller possono essere definite espressioni SNMP in varie forme:

::
	
	1.3.6.1.2.1.2.2.1.14@
	1.3.6.1.2.1.2.2.1.14:$community@$node:$port

In alcuni casi queste espressioni rappresentano un valore SNMP da estrarre con 
GET. Il altri contesti l'espressione rappresenta una tabella di valori da estrarre
con operazioni di WALK. 

Vediamo alcuni esempi:

::
	
	1.3.6.1.2.1.2.2.1.14:$community@$node:$port - 3

Rappresenta una espressione dove bisogna estrarre un dato in GET

::
	
	byRegexpUnique(1.3.6.1.2.1.25.2.3.1.3:$community@$node:$snmpport, '^foo$')

In questo caso, l'espressione SNMP deve restituire una tabella di WALK

Per risolvere questa ambiguita', le espressioni SNMP restituiscono sempre come valore un oggetto di tipo :py:class:`~.SnmpValueType`.

Se questo oggetto viene utilizzato all'interno di espressioni aritmentiche/logiche, il valore viene calcolato effettuando una GET.

L'utilizzo di operatori che lavorano su oggetti iterabili (for, []) permette di accedere al valore calcolato tramite WALK().

Una volta che il valore dell'oggetto e' stato calcolato (o con GET o con WALK in base all'operatore usato), questo rimane costante per tutto il tempo di vita dell'oggetto.

NOTA: Per rendere l'oggetto sufficentmente generico, le funzioni di effettuare la GET() o WALK() vengono passate come parametri.


Classi
--------------------------------------------------------------------

.. autoclass:: LazyValueType
	:members:
   
.. autoclass:: SnmpValueType
	:members:

"""

import sys
import types
import re

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def regexp_eq(v1, v2):

	if not (isinstance(v1, types.StringType) or isinstance(v1, types.UnicodeType)):
		raise TypeError("Type for %s (%s) is not a valid string for =~ operator!" % (v1, type(v1)) )

	if not (isinstance(v2, types.StringType) or isinstance(v2, types.UnicodeType)):
		raise TypeError("Type for %s (%s) is not a valid string for =~ operator!" % (v2, type(v2)) )

	data = ''.join(v1)

	try:
		re_flags = re.DOTALL | re.MULTILINE
		r = re.compile(v2, re_flags)

		m = r.search(data)
		if m != None:
			return True
		else:
			return False

	except re.error, er:
		raise ValueError("~= parameter %s is not a valid regexp -> %s" % (v2, er))

	except TypeError, er:
		raise TypeError("~= invalid operand type %s" % (er))

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class LazyValueType(object):
	"""
	Abstract class for all lazy-values returned by :py:meth:`~.ExprTree.evalute`.
	"""

	def __init__(self):
		pass

	def getvalue(self):
		"""
		This method must (calculate and) return the *effective*/*real* value for this object.
		"""
		raise Exception("Not implemented")

	#  me =~ other
	def __regexpr_eq__(self, other):
		"""
		Method called when using operator the '=~' on this object.
		
		::
			
			<this object> =~ <value>
		"""
		raise Exception("Not implemented")

	#  other =~ me
	def __rregexpr_eq__(self, other):
		"""
		Method called when using operator the '=~' on this object.
		
		::
			
			<value> =~ <this object>
		"""
		raise Exception("Not implemented")

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class SnmpValueType(LazyValueType):
	"""
	Base class for all SNMP values calculated by SNMP-expression.
	
	The real SNMP value is calculated by :py:meth:`effective_snmpget` and  :py:meth:`effective_snmpwalk` 
	
	This espression:
	
	::
	    
	                                                                      
	       1.3.6.1.2.1.1.1.0:$community@$node:$port                       
	                                                                      
	
	Is converted in this tree:
	
	::
	    
	                                                                                                        
	                                                   RootExprTree()                                       
	                                                         |                                              
	                                                         |                                              
	                                                 SnmpValueExprTree()                                    
	                                                     |  |  |  |                                         
	                                                     |  |  |  |                                         
	         OIDLiteralExprTree("1.3.6.1.2.1.1.1.0") ----+  |  |  +---- IdentifierExprTree("$port")         
	                                                        |  |                                            
	                   IdentifierExprTree("$community") ----+  +---- IdentifierExprTree("$node")            
	                                                                                                        
	
	These are the execution steps
	
	::
	    
	                                                                                                                    
	    STEP 1                                                                                                          
	                                                         |                                                          
	                                                         | .evaluate(context)                                              
	                                                         |                                                          
	                                                   RootExprTree()                                                   
	                                                         |                                                          
	                                                         | .evaluate(context)                                              
	                                                         |                                                          
	                                                 SnmpValueExprTree()                                                
	                                                     |  |  |  |                                                     
	                                                     |  |  |  |  .evaluate(context)                                        
	                                                     |  |  |  |                                                     
	         OIDLiteralExprTree("1.3.6.1.2.1.1.1.0") ----+  |  |  +---- IdentifierExprTree("$port")                     
	                                                        |  |                                                        
	                   IdentifierExprTree("$community") ----+  +---- IdentifierExprTree("$node")                        
	                                                                                                                    
	                                                                                                                    
	    STEP 2                                                                                                          
	                                                         |                                                          
	                                                         | .evaluate(context)                                       
	                                                         |                                                          
	                                                   RootExprTree()                                                   
	                                                         |                                                          
	                                                         | .evaluate(context)                                       
	                                                         |                                                          
	                                                 SnmpValueExprTree()                                                
	                                                     |  |  |  |                                                     
	                                                     |  |  |  |                                                     
	                                                     |  |  |  |                                                     
	                                   "1.3.6.1.2.1.1.1.0"  |  |  161                                                   
	                                                        |  |                                                        
	                                                 "public"  "localhost"                                              
	                                                                                                                    
	                                                                                                                    
	    STEP 3                                                                                                          
	                                                         |                                                          
	                                                         | .evaluate(context)                                       
	                                                         |                                                          
	                                                   RootExprTree()                                                   
	                                                         |                                                          
	                                                         |                                                          
	                          SnmpValueType(context, "1.3.6.1.2.1.1.1.0", "public", "localhost", 161)                   
	                                                                                                                    
	                                                                                                                    
	    STEP 4                                                                                                          
	                                                         |                                                          
	                                                         | .evaluate(context)                                       
	                                                         |                                                          
	                                                   RootExprTree()                                                   
	                                                         |                                                          
	                                                         | .getvalue()                                              
	                                                         |                                                          
	                          SnmpValueType(context, "1.3.6.1.2.1.1.1.0", "public", "localhost", 161)      LAZY VALUE   
	                                                                                                                    
	                                                                                                                    
	    STEP 5                                                                                                          
	                                                         |                                                          
	                                                         | .evaluate(context)                                       
	                                                         |                                                          
	                                                   RootExprTree()                                                   
	                                                         |                                                          
	                                                         | .getvalue()                                              
	                                                         |                                                          
	                          SnmpValueType(context, "1.3.6.1.2.1.1.1.0", "public", "localhost", 161)                   
	                                                                                                                    
	    STEP 6                                                                                                          
	                                                                                                                    
	                                                         |                                                          
	                                                         | .evaluate(context)                                       
	                                                         |                                                          
	                                                   RootExprTree()                                                   
	                                                         |                                                          
	                                                         |                                                          
	                          "Linux debiantest 3.2.0-3-486 #1 Thu Jun 28 08:08:24 UTC 2012 i686"                       
	    STEP 6                                                                                                          
	                                                                                                                    
	                                                                                                                    
	                          "Linux debiantest 3.2.0-3-486 #1 Thu Jun 28 08:08:24 UTC 2012 i686"                       
	"""

	def __init__(self, context, oid, community, host, port):

		super(LazyValueType, self).__init__()

		# The calculated value of this snmp expression.
		self.value = None

		self.context      = context

		self.host         = host

		self.oid          = oid
		
		self.sv           = 1
		self.community    = community
		self.port         = port

	def coerce_type(self, v1, v2):

		if type(v1) == type(v2):
			return (v1, v2)

		if isinstance(v1, types.StringType) or isinstance(v1, types.UnicodeType):
			if isinstance(v2, types.IntType)  : return (int(v1)  , v2)
			if isinstance(v2, types.LongType) : return (int(v1)  , v2)
			if isinstance(v2, types.FloatType): return (float(v1), v2)

		if isinstance(v2, types.StringType) or isinstance(v2, types.UnicodeType):
			if isinstance(v1, types.IntType)  : return (v1,   int(v2))
			if isinstance(v1, types.LongType) : return (v1,   int(v2))
			if isinstance(v1, types.FloatType): return (v1, float(v2))

		return (v1, v2)

	#--------------------------------------------------------------------------#
	#                                                                          #
	#--------------------------------------------------------------------------#

	def effective_snmpget(self): 
		"""
		This method MUST calculate the real SNMP value for this object.
		
		The value is supposed to be calculated with a SNMP GET action.
		"""
		raise NotImplemented("")
	
	def effective_snmpwalk(self):
		"""
		This method MUST calculate the real SNMP value for this object.
		
		The value is supposed to be calculated with a SNMP WALK action.
		"""
		raise NotImplemented("")

	#--------------------------------------------------------------------------#
	#                                                                          #
	#--------------------------------------------------------------------------#

	def getvalue(self):
		return self.__getvalue()

	def __getvalue(self):
		if self.value == None:
			self.value = self.effective_snmpget()
		return self.value

	def __gettable(self):
		if self.value == None:
			self.value = self.effective_snmpwalk()
		return self.value

	#--------------------------------------------------------------------------#
	# Operatori che effettuano la GET                                          #
	#--------------------------------------------------------------------------#

	def __repr__(self):                      return "SnmpValueType(%s)" % (self.oid)
	def __str__(self):                       return str  (self.__getvalue())
	def __int__(self):                       return int  (self.__getvalue())
	def __float__(self):                     return float(self.__getvalue())
	def __bool__(self):                      return bool (self.__getvalue())
	def __complex__(self):                   return complex(self.__getvalue()) #needed?

	#--------------------------------------------------------------------------#

	def __eq__(self, other):
		v = self.__getvalue()
		(v, other) = self.coerce_type(v, other)
		return v == other

	def __ne__(self, other):                 
		v = self.__getvalue()
		(v, other) = self.coerce_type(v, other)
		return v != other

	def __lt__(self, other):                 
		v = self.__getvalue()
		(v, other) = self.coerce_type(v, other)
		return v < other

	def __gt__(self, other):                 
		v = self.__getvalue()
		(v, other) = self.coerce_type(v, other)
		return v > other

	def __le__(self, other):                 
		v = self.__getvalue()
		(v, other) = self.coerce_type(v, other)
		return v <= other

	def __ge__(self, other):                 
		v = self.__getvalue()
		(v, other) = self.coerce_type(v, other)
		return v >= other

	def __nonzero__(self):
		# NOTE: This is required for right evaluation otherwise the
		# expression:
		# 
		#       ! 1.3.6.1.2.1.1.10@ 
		#
		# would trigger the __len__ operator and the WALK, and this is
		# wrong because it's silly to apply boolean operators to
		# WALK tables.
		#    
		# See: http://docs.python.org/reference/datamodel.html#object.__nonzero__
		#
		v = self.__getvalue()
		if v:
			return True
		else:
			return False

	#--------------------------------------------------------------------------#
	def __neg__(self):                       return (-self.__getvalue())

	def __pos__(self):                       return (+self.__getvalue())

	def __add__(self, other):                return self.__getvalue() + other
	def __radd__(self, other):               return other + self.__getvalue()

	def __sub__(self, other):                return self.__getvalue() - other
	def __rsub__(self, other):               return other - self.__getvalue()

	def __mul__(self, other):                return self.__getvalue() * other
	def __rmul__(self, other):               return other * self.__getvalue()

	def __div__(self, other):                return self.__getvalue() / other
	def __rdiv__(self, other):               return other / self.__getvalue()

	def __truediv__(self, other):            return self.__getvalue() / other
	def __rtruediv__(self, other):           return other / self.__getvalue()

	def __floordiv__(self, other):           return self.__getvalue() // other
	def __rfloordiv__(other):                return other // self.__getvalue()

	def __mod__(self, other):                return self.__getvalue() % other
	def __rmod__(self, other):               return other % self.__getvalue()

	def __pow__(self, othr, modulo=None):    return self.__getvalue() ** other
	def __rpow__(self, other, modulo=None):  return other ** self.__getvalue()

	#--------------------------------------------------------------------------#

	#  me =~ other
	def __regexpr_eq__(self, other):
		myval = self.__getvalue()

		# We are insterested in the 'value' of the other operand,
		# not in the operand itself.
		if isinstance(other, SnmpValueType):
			other = other.getvalue()

		return regexp_eq(myval, other)

	#  other =~ me
	def __rregexpr_eq__(self, other):

		myval = self.__getvalue()

		# We are insterested in the 'value' of the other operand,
		# not in the operand itself.
		if isinstance(other, SnmpValueType):
			other = other.getvalue()

		return regexp_eq(other, myval)

	#--------------------------------------------------------------------------#

	def __lshift__(self, other):             return self.__getvalue() << other
	def __rlshift__(self, other):            return other << self.__getvalue()

	def __rshift__(self, other):             return self.__getvalue() >> other
	def __rrshift__(self, other):            return other >> self.__getvalue()

	def __and__(self, other):                return self.__getvalue() & other
	def __rand__(self, other):               return  other & self.__getvalue()

	def __or__(self, other):                 return self.__getvalue() | other
	def __ror__(self, other):                return other | self.__getvalue()

	def __xor__(self, other):                return self.__getvalue() ^ other
	def __rxor__(self, other):               return other ^ self.__getvalue()

	def __invert__(self):                    return (~self.__getvalue())

	#--------------------------------------------------------------------------#

	def __abs__(self):                       return abs(self.__getvalue())
	def __round__(self, n=None):             return round(self.__getvalue())
	def __divmod__(self, other):             return divmod(self.__getvalue(), other)

	#--------------------------------------------------------------------------#
	# Operatori che effettuano la WALK                                         #
	#--------------------------------------------------------------------------#

	def __len__(self):                       return len(self.__gettable())

	def __getitem__(self, i):                return self.__gettable()[i]

	def __iter__(self):                      return iter(self.__gettable())
