# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""
################################################################################
Expression Parser 
################################################################################

Premesse: requisiti per lo sviluppo
========================================================================

Questo modulo si basa su codice generato automaticamente da ANTLR.

Per utilzzare ANTRL e' necessario aver installato i seguenti tool:

* JDK >= 1.6.0


Introduzione
========================================================================

Il linguaggio utilizzato da Sanet per effettuare i suoi controlli e' un linguaggio  
"imperativo" creato ad-hoc per permettere di interpretare *espressioni* (eseguendo 
operazioni) e calcolare un valore.


Differenze con altri linguaggi interpretati
---------------------------------------------

E' stato deciso di implementare un linguaggio interpretato ah-hoc perche' alcuni casi particolari
non sono facilmente gestibili *utilizzando*/*interpretando* espressioni in linguaggi
gia' esistenti (python, lua, ecc.). 

In particolare alcune operazioni non hanno la stessa semantica che avrebbero
altri linguaggi. Prendiamo ad esempio la seguente espressione

::
	
	(1.2.3@ > 1.2.3#) and (1.2.4@ > 1.2.4#)

Questa espressione deve confrontare il valore attuale (al momento dell'esecuzione) di due OID (1.2.3  e 1.2.4) con 
il valore che le stesse OID avevano ad una precedente esecuzione dell'espressione.

Ci sono due ordini di problemi:

	1) Nessun linguaggio supporta nativamente la gestione di variabili con valori "correnti" e "passati"
	
	2) L'operatore AND in molti linguaggi e' "cortocirtuitato". L'espressione nell'esempio
	   rischia di eseguire solo la parte *sinistra* dell'AND e mai la seconda. Questo
	   potrebbe creare problemi in alcune situazioni, anche in relazione al punto (1).



Sintassi del linguaggio
========================================================================

Il linguaggio supporta i seguenti costrutti sintattici:

Stringhe
---------
::
	
	"....."           "hello world"
	'.....'           'hello world'
	r"....."         r"hello world"
	r"....."         r'hello world"

Numeri Decimali
------------------

::
	
	<nnn>             123456
	-<nnn>           -123456

Numeri Esadecimali
---------------------------

::
	
	0x<xxxx>          0x0001

Numeri in vigola mobile

::
	
	 <nnn>.<nnn>      3.1
	-<nnn>.<nnn>     -3.14

Costanti booleane:
---------------------------

::
	
	True
	False
    
Variabili semplici
---------------------------
    
::
	
	$<name>           $foo, $_foo

Operatori Logici
---------------------------

::
	
	<expr> || < expr>      True || False ,  (2 > 3) || ( 3 < 4)
	<expr> or  <expr>      True or False ,  (2 > 3) or ( 3 < 4)
	
	<expr> &&  <expr>      True &&  False ,  (2 > 3) &&  ( 3 < 4)
	<expr> and <expr>      True and False ,  (2 > 3) and ( 3 < 4)
	
	!<expr>                !True , !(1 > 2)

Operatori su Bit
---------------------------
::
	
	<expr> ^ <expr>        0x0001 ^ 0x0010
	<expr> & <expr>        0x0001 & 0x1000
	<expr> | <expr>        0x0001 | 0x1000
        
Operatori di confronto
---------------------------
::
	
	<expr>  ==  <expr>
	<expr>  !=  <expr>
	
	<expr>  <   <expr>
	<expr>  >   <expr>
	<expr>  <=  <expr>
	<expr>  >=  <expr>
	
	<expr>  =~  <expr>


Operatori matematici:
---------------------------
::
	
	-a
	
	a + b
	a - b
	a * b
	a / b
	
	a % b
	a ** b 

Operatore di indicizzazione:
------------------------------------------------------
::
	
	<expr> [ <expr> ]                  a[1], a["key1"]
	<expr> [ <expr> ; <expr> ]         a[1;2], a[;2], a[;]

Chiamata a funzione
---------------------------
::
	
	<symbol> (  <expr>, ... , <symbol> = <expr> ... )       -> foo(1,2,3, param1=10, param2="hello")


Espressioni SNMP       
---------------------------
::
	
	<expr oid> [ . <expr> ... ] [ : <expr community> ] @ [ <expr host> [ : <expr port> ] ]
	<expr oid> [ . <expr> ... ] [ : <expr community> ] # [ <expr host> [ : <expr port> ] ]
	
	1.3.6.1.2.1.1.1.0@
	1.3.6.1.2.1.1.1.0#
	1.3.6.1.2.1.1.1.0:"public"@
	1.3.6.1.2.1.1.1.0:"public"@"localhost"
	1.3.6.1.2.1.1.1.0:"public"@"localhost":161

Variabili avanzate
---------------------------

::
	
	{ <symbol> @ }
	{ <symbol> # }


Costrutto condizionale (IF)
------------------------------------------------------

::
	
	if <cond> then <expr> else <expr>

Costrutto per gestione delle eccezioni
------------------------------------------------------

::
	
	try <expr> catch [ <exception type> ] ( [ <expr> ] )
	
	try 1 catch Exception ()
	
	try 1 catch Exception ( "fail" )
	
	try 1/0 catch ZeroDivisionError ( 0 ) catch Exception ( -1 ) 


Tipi di valore calcolati dal linguaggio
========================================================================

Tipi base
---------------------------------------------

I valori calcolati dal linguaggio possono essere di un qualunque tipo supportato dal linguaggio python, quind:

	* Booleani 
	* Numeri interi
	* Numeri in virgola mobilie
	* Stringhe di caratteri
	* Vettori (tuple) di valori.
	* Dizionari

.. warning:: aver utilizzato come tipi di dato quelli supportati da python significa che creare un interprete di questo linguaggio con altre tecnologie (C, C++, Java, C#, ecc.) potrebbe comportare dei problemi di "porting" notevoli.

Tipi lazy
---------------------------------------------

Per alcune operazioni il valore di una sotto-espressione non puo' essere calcolato immediatamente.

In questo caso il valore restituito da :py:meth:`~.ExprTree.evaluate` puo' restituire oggetti di tipo :py:class:`~.LazyValueType`.

Al momento gli unici tipi previsti dalla libreria per gestire valori lazy sono quelli derivati dalla classe :py:class:`~.SnmpValueType`.


Interpretazione e valutazione del linguaggio: gli ExprTree
========================================================================

La valutazione di una espressione si divide in due fasi:

	1) Analizzare la stringa che contiene l'espressione e creazione di una struttura dati per poter elaborare l'espressione.
	2) Visita ed elaborazione (valutazione) dell'albero espressione.

Una espressione e' logicamente rappresentata in memoria da una struttura dati ad albero (chiamato anche *albero sintattico*).

Esempio:

::
    
                                                                                                     
          STRINGA                 ALBERO ESPRESSIONE                                                 
                                                                                                     
                                          [-]                                                        
      "1 + 2 * 10 - $a"    -->           /   \                                                       
                                      [+]     [a]                                                    
                                     /   \                                                           
                                  [1]     [*]                                                        
                                         /   \                                                       
                                      [2]     [10]                                                   
                                                                                                     

I nodi e foglie dell'albero rappresentano:

	* Operatori logici / matematici
	* Operatori di chiamata a funzione
	* Riferimenti a variabili
	* Costanti (numeri, stringhe, booleani)

La classe base per codificare un nodo dell'albero e' la classe :py:class:`~.ExprTree`.

Tutti gli elementi del linguaggio (operatori, variabili, ecc.) sono rappresentati da sotto-classi di :py:class:`~.ExprTree`. 

Un oggetto di tipo :py:class:`~.ExprTree`, in base al tipo di elemento del linguaggio che rappresenta, deve gestire uno, due o piu' sotto-alberi espressione.

Esempio: espressione e corrispondente struttura :py:class:`~.ExprTree`:

::
    
                                                                                            
                                                                                            
                        1 + 2 * 10                                                          
                                                                                            
                                                                                            
                      RootExprTree()                                                        
                            |                                                               
                            |                                                               
                     PlusOpExprTree()                                                       
                         /      \                                                           
                        /        \                                                          
      DecimalExprTree(1)          MultiOpExprTree()                                         
                                      /       \                                             
                                     /         \                                            
                    DecimalExprTree(2)          DecimalExprTree(10)                         
                                                                                            

Questo albero viene costruito programmaticamente istanziando gli oggetti nel seguente modo:

::
    
                                                                                         
    tree = RootExprTree(                                                                 
               PlusOpExprTree(                                                           
                   MultiOpExprTree(                                                      
                       DecimalExprTree(2) ,                                              
                       DecimalExprTree(10),                                              
                   ),                                                                    
                   DecimalExprTree(1)                                                    
               )                                                                         
           )                                                                             
                                                                                         


Creazione di alberi espressione
========================================================================

Utilizzo di oggetti ExprFactory
-----------------------------------------------------------------------

Invece di istanziare direttamente oggetti :py:class:`~.ExprTree` per creare gli alberi espressione, 
vengono utilizzati dei *factory objects* di tipo :py:class:`~.BaseExprFactory`.

Un :py:class:`~.BaseExprFactory` dichiara un metodo per creare tutti gli oggetti 
:py:class:`~.ExprTree` richiesti dal linguaggio per generare un albero espressione completo.

Esempio: Data l'espressione:

::
    
                                                                                
    1 + $a                                                                      
                                                                                

Queste sono le istruzioni per creare l' :py:class:`~.ExprTree` corrispondente utilizzando un :py:class:`~.BaseExprFactory`:

::
    
                                                                
    factory = BaseExprFactory()                                 
                                                                
    tree = factory.createRoot(                                  
               factory.createAddAddOp(                          
                   factory.createInteger(1),                    
                   factory.createIdentifier('$a')               
               )                                                
           )                                                    
                                                                


L'utilizzo di :py:class:`~.BaseExprFactory` permette di disaccoppiare due operazioni:

	* Analisi lessicale/Parsing di una stringa
	* Creazione alberi :py:class:`~.ExprTree`.

In questo modo il codice che si occupa del *parsing* non deve sapere quali
classi :py:class:`~.ExprTree` devono essere istanziate e in che modo, ma semplicemente
chiamare metodi di un oggetto di tipo :py:class:`~.BaseExprFactory`.


Parsing delle espressioni (ANTLR3)
-----------------------------------------------------------------------

Per effettuare il parsing di una stringa (che contiene una espressione valida)
viene utilizzata la libreria/tool ANTLR 3.

ANTLR3 e' un tool per creare automaticante un parser (analizzatore lessicale + parser)
a partire da un file che specifica:

	* I token del linguaggio
	* Le regole di produzione della grammatica
	* Le azioni da compiere ogni volta che una regola di produzione viene riconosciuta.

Il file ANTLR che specifica la grammatica e':

::
    
                                       
    ExprScript.g                       
                                       

Questo file contiene la definizione dei *token* e *regole di produzione* necessarie ad ANTLR per produrre i file:

::
    
                                                                                                     
    ExprScriptLexer.py         Analizzatore Lessicale (token)
    ExprScriptParser.py        Parser del linguaggio.

Le azioni specificate nel file ExprScript.g  creano alberi espressione (:py:class:`~.ExprTree`)
*indirettamente* delegando la creazione degli oggetti ExprTree a oggetti
di tipo :py:class:`~.ExprTree`.


Rigenerare i moduli di parsing
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Per rigenerare i file *ExprScriptLexer.py* e *ExprScriptParser.py* bisogna
eseguire le seguenti operazioni:

::
	
	cd {{SVN_ROOT}}/sanet-common/sanet_common/exprparser
	../../dev-scripts/create_lexer_parser.sh


Valutazione di un albero espressione
========================================================================

Valutazione ricorsiva
------------------------------------------------------------------------

La valutazione di un albero espressione avviene invocando il metodo :py:meth:`ExprTree.evaluate`.

Questo metodo deve ricorsivamente invocare lo stesso metodo su tutti i sotto-alberi
corrispondenti e restituire il valore calcolato dall'albero espressione.

Spetta ad ogni classe di ExpTree sapere in che modo gestire i sotto-alberi.

Dato questo albero:

::
    
                                                                                                     
    factory = BaseExprFactory()
    
    tree = factory.createRoot(
               factory.createAddAddOp(
                   factory.createInteger(1),
                   factory.createIdentifier('$a')
               )
           )
    
    context = {}
    
    tree.evalute(context)
    

La valutazione dell'albero si traduce in una visita ricorsiva "in ordine"
che provoca la chiamata del metodo evaluate() su tutti i nodi dell'albero.

::
    
                          CALLER                                         
                             |                                     
                             | .evaluate()                         
                             |                                     
                        RootExprTree()                             
                             |                                     
                             | .evaluate()                         
                             |                                     
                       PlusOpExprTree()                            
                          /      \                                 
            .evaluate()  /        \  .evaluate()                   
                        /          \                               
      DecimalExprTree(1)            IdentifierExprTree('$a').      
                                                                   

Contesto di esecuzione
------------------------------------------------------------------------

Nelle espressioni e' possibile fare riferimento a simboli identificati da un nome.
Questi simboli rappresentano:

	* Funzioni da chiamare
	* Variabili semplici
	* Variabili speciali ( DataIdentifier )

Il valore associato ad un simbolo puo' essere:

	* Un valore costante (numero, stringa, ecc)
	* Un riferimento a funzione

Un *contesto di esecuzione* **DEVE ESSERE una struttura dati che permette il mapping tra
un nome ed un valore**. Nella pratica un qualunque oggetto derivato di tipo dizionario *dict*.

Un *contesto* di esecuzione viene passato in input al metodo :py:meth:`~.ExpTree.evaluate`
per permettere a tutto l'albero espressione di poter gestire i simboli correttamente.

Esempio: La seguente espressione:

::
    
                                        
    1 + $a                              
                                        

Viene codificata attraverso il seguente albero:

::
    
                                                        
    factory = BaseExprFactory()                         
                                                        
    tree = factory.createRoot(                          
               factory.createAddAddOp(                  
                   factory.createInteger(1),            
                   factory.createIdentifier('$a')       
               )                                        
           )                                            
                                                        

Che corrisponde al seguente albero:

::
    
                                                               
                        RootExprTree()                         
                             |                                 
                             |                                 
                       PlusOpExprTree()                        
                          /      \                             
                         /        \                            
       DecimalExprTree(1)          IdentifierExprTree('$a').   
                                                               

Per valutare l'albero bisogna invocare il metodo evalute() della radice dell'albero 
passando un contesto adatto:

::
    
                                        
    context = { '$a': 3 }               
                                        
    v = tree.evaluate( context )        
                                        
    print v  -> 3                       
                                        


Logica di valutazione di un ExprTree
------------------------------------------------------------------------

Quando viene chiamato il metodo :py:meth:`~.ExprTree.evaluate` sulla radice dell'albero
ogni singolo nodo dell'albero gestisce la valuazione del suo sotto-albero (o sotto-alberi)
(in base alla semantica prevista dal linguaggio per il costrutto sintattico
che quel nodo rappresenta), e restituisce un valore.


Valutazione di variabili: 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

L'accesso al valore di una variabile e' gestita dall'oggetto :py:class:`~.IdentifierExprTree`.

.. important:: I nodi di tipo :py:class:`~.IdentifierExprTree` non controllo che il *simbolo* di cui restituire il valore sia per forza "$<symbolo>". E' il parser/analizzatore lessicale che impedisce di scrivere espressioni invalide, tipo "1 + a" (invalida) anziche' "1 + $a" (valida).

Valutazione di DataIdentifier
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

L'accesso al valore di una variabile e' gestita dall'oggetto :py:class:`~.DataIdentifierExprTree`.

Valutazione di espressioni SNMP
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Le sotto-espressioni SNMP vengono gestite dagli oggetti :py:class:`~.SnmpValueExprTree`.

Questi oggetti dovrebbero restituire valori di tipo: :py:class:`~.SnmpValueType`.




Gestione delle eccezioni
=======================================================================================

Gerarchia delle eccezioni
----------------------------

Queste sono le eccezioni che possono essere sollevate:

::
    
    BaseExprParserException
    |
    `-- ParserException
        |
        |-- ExprParserAnalysisError
        |   |-- LexicalError
        |   `-- ParsingError
        |
        `-- RuntimeExprParserError
            |-- ExecutionError
            `-- NotEvaluatedError

Le eccezioni derivate da :py:exc:`~.ExprParserAnalysisError` vengono sollevate solo
in fase di analisi lessicale / parsing di una stringa.

Le eccezioni derivate da :py:exc:`~.RuntimeExprParserError` vengono sollevate quando
viene eseguito un :py:meth:`~.ExprTree.evalute` su un albero espressione






L'esecuzione del metodo :py:meth:`~.ExprTree.evaluate` potrebbe generare un qualunque
tipo di eccezione.

E' revisto che il metodo :py:meth:`~.ExprTree.evaluate` possa restituire almeno eccezioni 
di tipo :py:exc:`~.RuntimeExprParserError`.

Esistono due sotto-classi:

	* :py:exc:`~.ExecutionError` : Questo errore viene sollevato in caso di errori a runtime che il parser e' in grado di riconoscere.
	
	* :py:exc:`~.NotEvaluatedError`: Questa eccezione viene sollevata per gestire la valutazione dell'espressione in presenza di *valori correnti non calcolabili*. Vedi :ref:`valori-non-calcolabili`.




.. _valori-non-calcolabili:

Gestione di valori correnti non calcolabili
-------------------------------------------------------

Consideriamo la seguente espressione:

::
	
	(1.3.6.1.2.1.1.1.0@ != 1.3.6.1.2.1.1.1.0#)  and  (1.3.6.1.2.1.1.2.0@ != 1.3.6.1.2.1.1.2.0#)

Viene confrontato il valore di due OID con il loro valore calcolato all'esecuzione precedente.

Si traduce nel sguente albero

::
	
	                                         and                                                          
	                                         | |                                                          
	                     +-------------------+ +---------------------+                                    
	                     |                                           |                                    
	                    !=                                          !=                                    
	                   /  \                                        /  \                                   
	(1.3.6.1.2.1.1.1.0@    1.3.6.1.2.1.1.1.0#    1.3.6.1.2.1.1.2.0@    1.3.6.1.2.1.1.2.0#                 
	                                                                                                      
	         A                     B                    C                        D                        
	                                                                                                      

Per poter effettuare correttamente il calcolo, la prima volta che viene eseguita, l'espressione deve
calcolare il valore delle sotto-espressioni SNMP (A) e (C).

Il valore delle sotto-espressione (B) e (D), chiamato **valore precedente**, non puo' essere calcolato la prima volta (perche' non ci sono
state esecuzioni precedenti), ma questo non deve impedire di calcolare *ALMENO* il **valore corrente** di (A) e (B).

Per gestire correttamente questa situaizone, e' previsto che il metodo :py:meth:`~.ExprTree.evaluate` 
(o qualunque pezzo di codice chiamato indirettamente dal metodo), quando si trova in una situazione analoga a questa
di **valori precedenti** non calcolabili , debba sollevare una eccezione :py:exc:`~.NotEvaluatedError`.

Un qualunque nodo :py:class:`~.ExprTree` deve sempre eseguire il metodo :py:meth:`~.ExprTree.evaluate` 
su tutti i sotto-alberi figli. 
Se un sotto-albero solleva una eccezione  :py:exc:`~.NotEvaluatedError`
il nodo padre deve ritornare a sua volta la stessa eccezione, ma solo dopo aver chiamato
il metodo :py:meth:`~.ExprTree.evaluate` di tutti i nodi figli.


Valutazione costrutto IF
========================================================================

Il costrutto "IF THEN ELSE" del linguaggio e' gestito dalla classe :py:class:`~.IfExprTree`.


Valutazione costrutto try/catch
========================================================================

Il costrutto "Try/Catch" del linguaggio e' gestito dalla classe :py:class:`~.TryExprTree`.

Le "eccezioni" intercettabili nel linguaggio hanno lo stesso "nome" delle eccezioni
python generate durante la valutazione dell'espressione.

Questa scelta di rendere visibile nel linguaggio le eccezioni con lo stesso nome
che hanno in python e' una scelta di comodo. 

Questo permette di non essere costretti a sapere in anticipo tutte le possibili 
eccezioni sollevabili da una espressione interpretata.

.. warning:: questa gestione dei nomi delle eccezioni rende molto difficile il porting del linguaggio in un altro framework che non sia python.



Moduli
========================================================================

.. automodule:: sanet_common.exprparser.ExprFactory

.. automodule:: sanet_common.exprparser.Types

.. automodule:: sanet_common.exprparser.utils


"""

__copyright__  = "Copyright 2017, Laboratori Guglielmo Marconi S.p.A."
__version__    = "1.0.0"
__author__     = "Diego Billi"
__credits__    = ["Michele Bergonzoni", "Luca Ferroni"]
__license__    = "AGPL v3"
__maintainer__ = "Diego Billi"
__email__      = "sviluppo@labs.it"
__status__     = "Production"

__all__ = ["ExprFactory", "ExprScriptParser", "ExprScriptLexer", "Types", "utils"]

#------------------------------------------------------------------------------#
# TESTING CODE                                                                 #
#------------------------------------------------------------------------------#

if __name__ == "__main__":

	import os
	import sys

	from sanet_common.exprparser.ExprFactory      import LexicalError, ParsingError, ExecutionError
	from sanet_common.exprparser.ExprFactory      import SnmpValueExprTree
	from sanet_common.exprparser.ExprFactory      import BaseExprFactory
	from sanet_common.exprparser.utils            import parseString

	class MySnmpValueExprTree(SnmpValueExprTree):

		def __init__(self, exprOID, eCommunity, charType, eHost, ePort):
			super(MySnmpValueExprTree, self).__init__(exprOID, eCommunity, charType, eHost, ePort)

		def calculateSnmpValue(self, context, oid,community,ctype,host,port):

			community = community if community is not None else 'public'
			port      = port      if port      is not None else 121
			host      = host      if host      is not None else 'localhost'

			if ctype == '@':
				v = 1
			else:
				v = -1

			s = oid 
			if community != "":
				s = s + ":" + community
			s = s + ctype
			if host != "":
				s = s + host
			if port != "":
				s = s + ':' + str(port)

			print "............. VALORE SNMP PER: " + str(s) + " = " + str(v)
			return v

	class MyFactory(BaseExprFactory):
		def createSnmpValue(self, exprOID, eCommunity, charType, eHost, ePort):
			return MySnmpValueExprTree(exprOID, eCommunity, charType, eHost, ePort)


	def newSymbol(context, name,value):
		symbolTable[name] = value
		return symbolTable[name]

	def f(context, a):
		return int(a)
	
	def g(context, a1,a2,a3):
		return a1 + a2 + a3
	
	def h(context, *args):
		for a in args:
			print "parametro: ", str(a)

	def getHex(context):
		return hex

	def isReachable(context, agent):
		print agent
		print "\n...verifica raggiungibilita per: " + agent + "... TRUE!"
		return True

	def byRegexpUnique(context, oid,regexpr):
		#return "6.6.6"
		return oid + str(regexpr)

	def test_str(n, string, verify):
		print "Test %s" % (n)
		print "\t", "STR  :", string
		exprTree2 = parseStringWithFactory(string, MyFactory() )
		print "\t", "TREE :", exprTree2
		val = exprTree2.evaluate(symbolTable)
		print "\t", "VALUE:", val, "(type: " + str(type(val)) + ")"
		print "\t", "VER  :", verify

		if val != verify:
			raise Exception("ERROR")

	def interactive_test(symbolTable):
		print "******************************************************"
		print "*********** SnmpExprParser Package Test **************"
		print "******************************************************"
		print "Defined symbols:"
		for (k,v) in symbolTable.iteritems():
			print k + " -> " + str(type(v))
		print "******************************************************"
		print "*************** PRESS RETURN TO EXIT *****************"
		print "******************************************************"

		while True:

			try:
				string = raw_input('test parser>>> ')
				if string == '':
					break
				print 

				exprTree2 = parseString(string, factory=MyFactory() )
				print "--- ALBERO FINALE 2---"
				print exprTree2
				print "--- VALUTAZIONE ESPRESSIONE 2--- "
				val = exprTree2.evaluate(symbolTable)
				print "RISULTATO2: ", val, "(type: " + str(type(val)) + ")"

			except LexicalError, ex:
				print "ERRORE LESSICALE: ", str(ex)

			except ParsingError, ex:
				print "ERRORE PARSING: ", str(ex)

			except ExecutionError, ex:
				print "ERRORE ESECUZIONE:", str(ex)
		
	def stdin_test(symbolTable):

		input_str = ''

		while True:
			try:
				string = raw_input('')
				input_str +=  string + '\n'
			except EOFError, ex:
				break

		input_str = input_str.strip()

		print "INPUT: " + input_str

		try:
			exprTree2 = parseString(string, factory=MyFactory() )
			print "--- ALBERO FINALE---"
			print exprTree2
			print "--- VALUTAZIONE ESPRESSIONE --- "
			val = exprTree2.evaluate(symbolTable)
			print "RISULTATO2: ", val, "(type: " + str(type(val)) + ")"

		except LexicalError, ex:
			print "ERRORE LESSICALE: ", str(ex)

		except ParsingError, ex:
			print "ERRORE PARSING: ", str(ex)

		except ExecutionError, ex:
			print "ERRORE ESECUZIONE:", str(ex)

	#
	# Define some symbols
	#
	symbolTable                  = {}

	symbolTable['newSymbol']     = newSymbol

	symbolTable['$i']            = 1			# variabile
	symbolTable['$c']            = 'a'			# variabile
	symbolTable['$s']            = "mondo"			# variabile
	symbolTable['$c']            = '123 \\"mondo\\" 456 \'mondo\''# variabile
	symbolTable['$f']            = 3.14			# variabile

	symbolTable['f']             = f			# funzione
	symbolTable['g']             = g			# funzione
	symbolTable['h']             = h			# funzione

	symbolTable['hex']           = hex
	symbolTable['$hex']          = hex
	symbolTable['getHex']        = getHex

	symbolTable['isReachable']   = isReachable		# funzionea
	symbolTable['byRegexpUnique'] = byRegexpUnique

	interactive = True
	if len(sys.argv) == 2:
		if sys.argv[1] == "multiline":
			interactive_test(symbolTable)
		else:
			print "Wrong parameters!"
			print ""
			print "Syntax: python  __init__.py  [parameters]"
			print ""
			print "  multiline\t\tRead input without prompting."
			sys.exit(-1)
	else:
		stdin_test(symbolTable)
