# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.
"""
########################################################################
String formatting utilities
########################################################################
"""

import re
import types

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def number_to_str(v):
	"""
	if *v* has no decimal digits the function returns its integer decimal rappresentation,
	otherwise it returns the floating point string rappresentation.
	"""
	
	if not isinstance(v, types.FloatType):
		return v

	v2 = round(v)
	if v2 != v:
		return str(v)

	return str(int(v2))

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def calc_num_prefix(num, base=1000):
	"""
	Calcola la scala di misura da applicare al "num" calcolata data la "base" dell'unita' di misura.
	
	Restituisce la tupla
	
		( <num in scala>, <base usata>, <esponente>, <prefisso> )
	
	Esempio:
		
		calc_num_prefix(1)                     ->   (1.0  , 1000,  0, ' ')
		calc_num_prefix(1000)                  ->   (1.0  , 1000,  1, 'k')
		calc_num_prefix(9999)                  ->   (9.999, 1000,  1, 'k')
		calc_num_prefix(1000000000)            ->   (1.0  , 1000,  3, 'G')
		calc_num_prefix(0.1)                   ->   (100.0, 1000, -1, 'm')
		calc_num_prefix(0.01)                  ->   (10.0 , 1000, -1, 'm')
		calc_num_prefix(0.0001)                ->   (100.0, 1000, -2, 'u')

		calc_num_prefix(1024, base=1024)       ->   (1.0  , 1024,  1, 'k')
	"""
	
	if num is None:
		raise ValueError()
	
	if num == 0.0:
		return (0, base, 0, ' ')
	
	num = float(num)
	
	#           1000                 1024                  e    i   
	#                                                          
	# P         1000000000000000     1125899906842624     15    5   
	# T         1000000000000        1099511627776        12    4   
	# G         1000000000           1073741824            9    3   
	# M         1000000              1048576               6    2   
	# k         1000                 1024                  3    1   
	#           1                    1                     0    0   
	# m         0.001                0.0009765625         -3   -1   
	# u         0.000001             9.53674316406e-07    -6   -2   
	# n         0.000000001          9.31322574615e-10    -9   -3   
	# p         0.000000000001       9.31322574615e-10   -15   -4

	#         0123456789
	PREFIX = 'fpum kMGTPE'

	s = abs(num) >= 1

	i = 0
	if s:
		r = base
		while not (-r < num < r):
			num /= base
			i += 1

		p = PREFIX[ 4+i ]
		return (num, base, i, p)
	else:
		r = 1
		while -r < num < r:
			num *= base
			i += 1

		p = PREFIX[ 4-i ]
		return (num, base, -i, p)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def _atoi(text):
    return int(text) if text.isdigit() else text

def natural_keys(text):
	'''
	        alist=[
	            "something1",
	            "something12",
	            "something17",
	            "something2",
	            "something25",
	            "something29"
	        ]
	        
	        alist.sort(key=natural_keys)
	        print(alist)
	'''

	# http://nedbatchelder.com/blog/200712/human_sorting.html
	# (See Toothy's implementation in the comments)

	return [ _atoi(c) for c in re.split('(\d+)', text) ]
