# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

.. _sanet-common-ipc:

################################################################################
Sanet IPC API
################################################################################

Inroduzione
================================================================================

Questo modulo permette la comunicazione tra processi (IPC) basata su *code* 
attraverso un *broker* di comunicazione.

Un processo che vuole inviare dei dati (*sender*) deve:
	* Creare una connessione con il *broker*
	* Inviare dati ad una *coda*

Un processso che vuole ricevere dati (*receiver*) deve:
	* Creare una connessione con il *broker*
	* Leggere i dati su una *coda*.

Gestione delle code
--------------------------------------------------------------------------------

Una coda e' una struttura dati che permette l'inserimento e la rimozione
di elementi con politica FIFO (fist-in fist-out).

::
    
         |                                                          |
         +----------------------------------------------------------+
           +-------+      +----------------+ +-------+ +---------+  
    --->   | dataN | ...  | data3          | | data2 | |  data1  |   --->
           +-------+      +----------------+ +-------+ +---------+
         +----------------------------------------------------------+
         |                                                          |

Valgono le seguenti regole:

	* Le *code* sono identificate da una stringa alfanumerica di lunghezza arbitraria.

	* Ogni elemento della coda e' una stringa di byte di lunghezza arbitraria.

	* Non esiste un limite alla dimensione di una singola coda



Eccezioni
================================================================================

.. automodule:: sanet_common.ipc.exceptions

Backend 
================================================================================

La libreria di IPC e' pensata per supportare diversi *backend* (*broker*).

Al momento e' implementato solo il backend :class:`~.RedisIPCConnection`.

.. automodule:: sanet_common.ipc.redisipc


Utility
================================================================================

Factory
--------------------------------------------------------------------------------

Creare una connessione con il broker richiede numerosi parametri.

Un processo che vuole creare tante connessioni velocemente puo'
implementare una :class:`~.AbstractIPCFactory`.

.. autoclass:: sanet_common.ipc.AbstractIPCFactory

Formato messaggi
--------------------------------------------------------------------------------

Per semplificare l'implementazione di protocolli di comunicazione, sono
state scritte queste funzioni per 

.. autofunction:: sanet_common.ipc.encode_request

.. autofunction:: sanet_common.ipc.encode_request_with_id

.. autofunction:: sanet_common.ipc.decode_request

.. autofunction:: sanet_common.ipc.encode_response

.. autofunction:: sanet_common.ipc.encode_response_with_id

.. autofunction:: sanet_common.ipc.decode_response


"""

import os
import sys
import socket

import uuid
import json

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

from sanet_common.ipc.exceptions import *

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

from sanet_common.ipc.redisipc import RedisIPCConnection as IPCConnection

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class AbstractIPCFactory(object):
	"""
	Classe base di comodo per avere un factory model per creare connessioni
	invece che passare in giro puntatori a funzioni (o metodi)
	"""
	
	def __init__(self, logger):
		"""
		TODO
		"""
		self.logger = logger

	def create_connection(self):
		"""
		TODO
		"""
		raise NotImplementedException()

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

RESPONSE_SUCCESS = 1
RESPONSE_ERROR   = 0

def encode_request( mtype, *mparams):
	"""
	TODO
	"""
	msg = {
		'id'    : None,
		'type'  : mtype,
		'params': mparams,
	}
	return json.dumps(msg)

def encode_request_with_id( mtype, mid, *mparams):
	"""
	TODO
	"""
	
	msg = {
		'id'    : mid,
		'type'  : mtype,
		'params': mparams,
	}
	return json.dumps(msg)

def decode_request( request):
	"""
	TODO
	"""
	return json.loads(request)




def encode_response( rtype, content):
	"""
	TODO
	"""
	msg = {
		'id'      : None,
		'status'  : rtype,
		'content' : content,
	}
	return json.dumps(msg)

def encode_response_with_id( rtype, mid, content):
	"""
	TODO
	"""
	msg = {
		'status': rtype,
		'id'     : mid,
		'content' : content,
	}
	return json.dumps(msg)

def decode_response( response ):
	"""
	TODO
	"""
	return json.loads(response)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#



if __name__ == "__main__":

	broker = sys.argv[1]
	username = sys.argv[2]
	password = sys.argv[3]

	sender   = IPCConnection(broker, username=username, password=password)
	receiver = IPCConnection(broker, username=username, password=password)

	sender.open()
	sender.bind("s")

	receiver.open()
	receiver.bind("r")
	
	print "--------------------------------------------------------"
	sender.send_msg("r", "Hello world")


	print "--------------------------------------------------------"
	(a, d) = receiver.receive_msg()
	print "from", a, 'data', d
	
	
	receiver.send_msg(a, d + " ho ho ho")
	
	print "--------------------------------------------------------"
	(a, d) = sender.receive_msg()

	print "from", a, 'data', d
	
	receiver.close()
	sender.close()



