# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""
################################################################################
Redis Backend
################################################################################

Introduzione
==============

Questo backend utilizza le *liste* di REDIS per implementare *code* IPC.

In Redis tutte le code sono identificate da chiavi con un prefisso "QUEUE_". Vedi :fun:`REDIS_QUEUE_ADDRESS`.

Per instaurare una connessione, leggere e scrivere code bisogna istanziare oggetti :class:`~.RedisIPCConnection`.

.. autoclass: sanet_common.ipc.redisipc.RedisIPCConnection

Dababase
===============

Per gestire le code viene selezionato il database numero 3.



"""

import os

import redis
from redis.exceptions import RedisError


from sanet_common.ipc.exceptions import IPCConnectionError, IPCException

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

#
# Per evitare di mischiare IPC e dati in cache di altro tipo di default usiamo un database
# redis diverso.
#
_USE_REDIS_IPC_DB = 3

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

NO_ADDRESS = ''

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

def REDIS_QUEUE_ADDRESS(queue_name):
	"""
	
	"""
	if queue_name is None:
		raise Exception("Invalid address: %s" % (queue_name))

	return 'QUEUE_%s' % (queue_name)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

class RedisIPCConnection(object):

	def __init__(self, host, port=None, username=None, password=None, secure_connection=True, db=_USE_REDIS_IPC_DB):

		self.host = host
		self.port = port
		self.db   = db

		self._redisconnection = None

		# FIX: should pass this parameter in the costructor?
		self.socket_timeout = 15

		self.send_timeout = None
		self.recv_timeout = None


		self.bound_address        = None
		self._redis_bound_address = None

	def settimeout(self, timeout, send=True, receive=True):
		if send:
			self.send_timeout   = timeout
		if receive:
			self.recv_timeout   = timeout

	def close(self):
		try:
			self._pool.disconnect()
		except RedisError, e:
			raise IPCException("Error while closing connection: %s:%s" % (type(e),e))

	def open(self):
		args = []
		kwargs = { 
			'db'  : self.db,
			'host': self.host
		}
		if self.port: kwargs['port'] = int(self.port)

		#kwargs['socket_timeout'] = self.socket_timeout

		try:
			self._pool = redis.ConnectionPool(*args, **kwargs)
			self._redisconnection = redis.Redis(connection_pool=self._pool)
			
			# FIX: rewrite with .ping() ?
			if self._redisconnection.echo("connected") != "connected":
				raise IPCConnectionError("Invalid connect to redis")
			
		except RedisError, e:
			raise IPCConnectionError("Unable to connect to redis: %s:%s" % (type(e),e))


	def bind(self, address, persistent=False):

		self.bound_address        = address
		self._redis_bound_address = REDIS_QUEUE_ADDRESS(address)

		return

	def send_msg(self, address, data, durable=False, blocking=True, timeout=None):

		if timeout == None:
			timeout = self.send_timeout

		address = REDIS_QUEUE_ADDRESS(address)

		reply_to_address = self.bound_address or NO_ADDRESS

		try:
			self._redisconnection.rpush(address, reply_to_address + '|' + data)
		except RedisError, e:
			raise IPCException("Unable to send message: %s:%s" % (type(e),e))


	def available(self):
		try:
			return self._redisconnection.llen(self._redis_bound_address)
		except RedisError, e:
			raise IPCException("Unable to get size for '%s': %s" % (self._redis_bound_address))

	def receive_msg(self, timeout=None, ack_msg=True):

		if timeout is None:
			timeout = self.recv_timeout

		try:
			data = self._redisconnection.blpop(self._redis_bound_address, timeout)

			# Timeout
			if data is None:
				return None, None

			# Molto strano
			redis_key, data = data
			if not data:
				return None, None

			# <sender address> "|" data
			address, data = data.split('|', 1)
			if address == NO_ADDRESS :
				address = None

			return address, data

		except RedisError, e:
			raise IPCException("Unable to receive message: %s:%s" % (type(e),e))

	def ack_pending(self):
		pass

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

class RedisIPCConnectionPubSub(object):
	"""
	Alternative version using Publish/Subscribe APIs.
	"""

	def __init__(self, host, port=None, username=None, password=None, secure_connection=True, db=_USE_REDIS_IPC_DB):

		self.host = host
		self.port = port
		self.db   = db

		self._redisconnection = None

		# FIX: should pass this parameter in the costructor?
		self.socket_timeout = 15

		self.send_timeout = None
		self.recv_timeout = None


		self.persistent           = False
		self.bound_address        = None
		self._redis_bound_address = None

	def settimeout(self, timeout, send=True, receive=True):
		if send:
			self.send_timeout   = timeout
		if receive:
			self.recv_timeout   = timeout

	def close(self):
		try:
			self._pool.disconnect()
		except RedisError, e:
			raise IPCException("Error while closing connection: %s:%s" % (type(e),e))

	def open(self):
		args = []
		kwargs = { 
			'db'  : self.db,
			'host': self.host
		}
		if self.port: kwargs['port'] = int(self.port)

		kwargs['socket_timeout'] = self.socket_timeout

		try:
			self._pool = redis.ConnectionPool(*args, **kwargs)
			self._redisconnection = redis.Redis(connection_pool=self._pool)
			
			# FIX: rewrite with .ping() ?
			if self._redisconnection.echo("connected") != "connected":
				raise IPCConnectionError("Invalid connect to redis")
			
		except RedisError, e:
			raise IPCConnectionError("Unable to connect to redis: %s:%s" % (type(e),e))


	def bind(self, address, persistent=False):

		self.bound_address        = address
		self._redis_bound_address = REDIS_QUEUE_ADDRESS(address)

		self.persistent  = persistent
		if not self.persistent:
			
			try:
				self._pubsub = self._redisconnection.pubsub()
				
				#print __file__, "bind", self._redis_bound_address
				self._pubsub.subscribe([self._redis_bound_address])
			except Exception, e:
				print __file__, "ERRRRRRRRRRRRRRRRROR", type(e), str(e)
				raise e

		return

	def send_msg(self, address, data, durable=False, blocking=True, timeout=None):

		if timeout == None:
			timeout = self.send_timeout

		address = REDIS_QUEUE_ADDRESS(address)

		reply_to_address = self.bound_address or NO_ADDRESS

		try:

			if self.persistent:
				self._redisconnection.rpush(address, reply_to_address + '|' + data)
			else:
				print __file__, "publish", address
				self._redisconnection.publish(address, reply_to_address + '|' + data)

		except RedisError, e:
			raise IPCException("Unable to send message: %s:%s" % (type(e),e))


	def available(self):
		try:
			if self.persistent:
				return self._redisconnection.llen(self._redis_bound_address)
			else:
				return 0
		except RedisError, e:
			raise IPCException("Unable to get size for '%s': %s" % (self._redis_bound_address))

	def receive_msg(self, timeout=None, ack_msg=True):

		if timeout is None:
			timeout = self.recv_timeout

		try:

			#print __file__, "receive_msg", self.persistent, self._redis_bound_address

			if self.persistent:
				data = self._redisconnection.blpop(self._redis_bound_address, timeout)

				# Timeout
				if data is None:
					return None, None

				# Molto strano
				redis_key, data = data
				if not data:
					return None, None

			else:
				print __file__, "get_message", timeout
				#timeout = timeout if timeout is not None else 99999999
				while 1:
					data = self._pubsub.get_message(ignore_subscribe_messages=False, timeout=timeout )
					#print "XXXXXXX", data
					if data is not None:
						if data['type'] == 'subscribe':
							continue
					break

				#print __file__, "receive_msg", "data=", data
				if data is None:
					return None, None

				data = data['data']
				#if not data:
				#	return None, None

			# <sender address> "|" data
			address, data = data.split('|', 1)
			if address == NO_ADDRESS :
				address = None

			return address, data



		except RedisError, e:
			raise IPCException("Unable to receive message: %s:%s" % (type(e),e))

	def ack_pending(self):
		pass

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

if __name__ == "__main__":

	import time
	import sys

	broker = sys.argv[1]
	username = sys.argv[2]
	password = sys.argv[3]

	sender   = RedisIPCConnection(broker, username=username, password=password, db=5)
	receiver = RedisIPCConnection(broker, username=username, password=password, db=5)

	sender.open()
	sender.bind("s")

	receiver.open()
	receiver.bind("r")
	
	print "--------------------------------------------------------"
	sender.send_msg("r", "Hello world")


	print "--------------------------------------------------------"
	(a, d) = receiver.receive_msg()
	print "from", a, 'data', d
	
	
	print "--------------------------------------------------------"

	print "Sending to ", a
	receiver.send_msg(a, d + " ho ho ho")
	
	time.sleep(5)
	
	print "--------------------------------------------------------"
	(a, d) = sender.receive_msg()

	print "from", a, 'data', d
	
	receiver.close()
	sender.close()
                
