# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

################################################################################
Mailing Utilities
################################################################################

Questo modulo contiene oggetti e funzioni per gestire l'invio di posta elettronica in Sanet.

Per inviare una mail bisogna istanziare un oggetto di tipo :class:`BaseMailHandler`.

E' possibile caricare dinamicamente una classe usando la funzione :func:`load_mail_handler_class`.

Classi
--------

.. autoclass:: sanet_common.mailing.BaseMailHandler
	:members:

Funzioni
----------

.. autofunction:: sanet_common.mailing.load_mail_handler_class

Moduli
-----------

.. automodule:: sanet_common.mailing.sendmail

.. automodule:: sanet_common.mailing.djangosmtp

"""

from sanet_common.pythonutils import load_symbol

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

class BaseMailHandler(object):
	"""
	Classe astratta per inviare email.
	"""

	def __init__(self, logger, settings):
		"""
		Costruttore dell'oggetto.
		
		:param  logger: Un *Logger* python.
		:type logger:  logging.Logger

		:param  settings: Dizionario con parametri di configurazione specifici per l'oggetto.
		:type settings:  dict
		"""

		self.logger   = logger
		self.settings = settings

	def send(self, frm, to, subject, body, extra_headers_dict=None, attachments=None):
		"""
		Invia una mail al destinatario *to*, con soggetto *subject* e corpo del messaggio *body*.
		
		Ogni chiave del dizionario *extra_headers_dict* viene aggiunta come header addizionale alla mail.
		
		.. warning:: Non viene fatto nessun controllo sulla validita' dei header opzionali che vengono specificati.
		
		Esempio: 
		
			::
				
				handler = BaseMailHandler(logger, {})
				
				handler.send("foo", "bar", "subject", "body", { 
					"Header1": "hello",
					"Header2": "world",
					"X-Header3": "the quick brown fox jumps over the lazy dog.",
				})
				
			La mail ricevuta sara:
			
			::
				
				From XXXXXXXXXX  Thu Jul 13 12:17:15 2017
				From: foo@localhost.localdomain.it
				To: bar@localhost.localdomain.it
				Subject: subject
				Return-Path: XXXXXXXX
				.,,
				X-MailHandler: DjangoMailHandler
				X-Header3: the quick brown fox jumps over the lazy dog.
				Header1: hello
				Header2: world
				
				body
		
		:param  to: Indirizzo email del mittente. Puo' essere un indirizzo email *locale* ( senza "@<dominio>").
		:type to:  string
		
		:param  subject: Soggetto della mail
		:type subject:  string
		
		:param  body: Corpo della mail
		:type body: string
		
		:param  extra_headers_dict: Header extra aggiuntivi.
		:type extra_headers_dict:  dict

		:param  attachments: Lista di file da allegare. La lista e' compostae da tuple (<filepath>, <name>, <mime_type>).
		:type attachments:  list
		
		:returns: True se la mail e' stata inviata, False altrimenti.
		:rtype: bool
		
		:exception: 
		
			Viene sollevata una qualunque eccezione se si verficano malfunzionamenti non previsti.
		"""

		raise Exception("Not Implemented")

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

def load_mail_handler_class(handler_module_name):
	"""
	Questo metodo importa la classe specificata dal path *handler_module_name*.
	
	La classe deve essere di tipo :class:`BaseMailHandler`, o il metodo sollevera' una eccezione.
	
	Esempio:
	
		::
			
			handler_class = load_mail_handler_class('sanet_common.mailing.djangosmtp.DjangoMailHandler')
			
			handler = handler_class(logger, { } )
	
	
	:param  handler_module_name: Stringa del path completo della classe da importare.
	:type handler_module_name:  string
	
	:returns: A class object
	
	:exception: 
	
		Raise an exception if an error occours.
	"""

	handler_class = load_symbol( handler_module_name )

	if not issubclass(handler_class, BaseMailHandler):
		raise Exception("Invalid type for '%s': %s" % (handler_module_name, handler_class.__name__))

	return handler_class
