# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.
"""
################################################################################
Django SMTP  backend
################################################################################

Introduzione
-------------

Questo modulo contiene classi di tipo :class:`BaseMailHandler` che utilizzano
le api Django per l'invio delle email.

Le email vengono inviate tramite protocollo SMTP collegandosi al server/porta specificati.

Questa classe uo' essere istanziata ed utilizzata solo all'interno di un'applicazione django.


::
	
	      DjangoMailHandler (BaseMailHandler)
	                      |
	______________________|________________________
	                      | 
	  Django              |
	                      v
	  django.core.mail.backends.smtp.EmailBackend

Esempio di utilizzo
-----------------------

Esempio di utilizzo del modulo:

	::
		
		import logging
		
		os.environ.setdefault("DJANGO_SETTINGS_MODULE", "conf.settings")
		
		import sanet_common.mailing.djangosmtp DjangoMailHandler
		
		logger = logging.getLogger("log")
		
		h = DjangoMailHandler(
			logger, 
			{ 
				"HOST"               : "localhost",
				"PORT"               : 25,
				"ENVELOPE_SENDER"    : "noc@labs.it",
			}
		)
		
		attachments = [ 
			('/tmp/file_1234512313.pdf', 'file.pdf' , 'appliacation/pdf'),
			('/tmp/file_1234512313.jpg', 'image.jpg', 'image/jpg'       ),
		]
		
		h.send("foo@localhost", "bar@localhost", "Dummy subject", "Dummy body", {'X-Extra': 'extra header'}, attachments)

Classi
-------------

.. autoclass:: DjangoMailHandler
	:private-members:
	:special-members:
	:members:

"""



import os
import socket
import smtplib

from django.core import mail
from django.core.mail import EmailMessage

from sanet_common.mailing import BaseMailHandler

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

class DjangoMailHandler(BaseMailHandler):
	"""
	Questa classe permette l'invio di email utilizzando le api di Django.
	"""

	BACKEND = 'django.core.mail.backends.smtp.EmailBackend'

	def __init__(self, logger, settings):
		"""
		Costruttore dell'oggetto.
		
		La classe deve essere istanziata passando in input al costruttore:
		
			* Un oggetto *logger*
			* Un dizionario con i parametri di configurazione
		
		Il dizionario di configurazione deve contenere le seguenti chiavi:
		
		::
			
			{ 
				"HOST"               : <smtp hostname>,
				"PORT"               : <smtp port>,
				"ENVELOPE_SENDER"    : <envelope sender email address>,
				
				"DEFAULT_DOMAIN_NAME": <domain string>,
			}
		
		Il parametro DEFAULT_DOMAIN_NAME e' **opzionale**. Se non specificato
		verra' utilizzato come dominio il dominio ricavato dalla chiamata a
		:py:func:`socket.getfqdn`.
		
		
		:param  logger: A python logger.
		:type logger:  logging.Logger

		:param  settings: A dictionary with specific configuraion data for the objects.
		:type settings:  dict
		"""
		super(DjangoMailHandler, self).__init__(logger, settings)

		self.logger.debug("Init %s handler", self.__class__.__name__)

	def send(self, frm, to, subject, body, extra_headers={}, attachments=[]):
		"""
		Vedi :py:meth:`~.BaseMailHandler.send`.
		"""

		self.logger.log(9, "[notify] send(%s, %s, %s, %s, %s)", frm, to, subject, body, extra_headers)

		server_address  = self.settings['HOST']
		server_port     = self.settings['PORT']
		default_domain  = self.settings.get('DEFAULT_DOMAIN_NAME', None) or socket.getfqdn()

		envelope_sender = self.settings['ENVELOPE_SENDER']

		if not server_address or not server_port or not envelope_sender:
			self.logger.fatal("[notify] SNMP server address (%s) or port (%s) or sender (%s) not configured!", server_address, server_port, envelope_sender)
			ret = -1

		# handle multiple email addresses 
		to = to.split(',')

		recipients = []
		for t in to:
			t = t.strip()

			# Remember that email addresses for targets do not contain the domain
			if t.find('@') < 0:
				t = '%s@%s' % (t, default_domain)

			recipients.append(t)

		if frm.find('@') < 0:
			frm = '%s@%s' % (frm, default_domain)

		extra_headers['X-MailHandler'] = self.__class__.__name__

		# If we do not specify the 'From' header, the envelope_sender will be
		# used.
		extra_headers['From'] = frm
		
		#
		# Attachments
		#
		email_attachments = []
		for filename, attach_name, mime_type in attachments:

			try:

				self.logger.debug("[notify] Loading attacment...")

				f = open(filename,"r")
				data = f.read()
				f.close()

				self.logger.debug("[notify] Loaded %s bytes (%s)", len(data), mime_type)

				attachment = [attach_name, data, mime_type] 

				email_attachments.append( attachment )

			except Exception, e:
				raise Exception("Unable to load attachment file '%s'. File does not exists" % (filename))

		#
		# Send
		#
		self.logger.debug("[notify] (smtp) from '%s' to '%s' using server <%s:%s> : %s", frm, recipients, server_address, server_port, subject )

		try:
			# FIX: the backend uses smptlib but do not specify a timeout. How to set a timeout?

			#
			# FIX: EmailMessage should handle unicode in body and subject properly, right?
			#

			connection_params = {
				'host' : server_address,
				'port' : server_port,
				'fail_silently': False,
			}

			self.logger.debug("[notify] Creating connection: %s %s", DjangoMailHandler.BACKEND, str(connection_params))

			connection =  mail.get_connection(backend=DjangoMailHandler.BACKEND, **connection_params)

			# django's shortcut
			email = EmailMessage(subject, 
				body, 	
				envelope_sender,
				recipients,
				[], #bcc
				headers = extra_headers,

				attachments = email_attachments,

				connection = connection,
			)
     
			self.logger.debug("[notify] Sending...")
			email.send(fail_silently=False)

			self.logger.info("[notify] (smtp) MAIL DELIVERED")

			return True

		except smtplib.SMTPException, es:
			self.logger.error("[notify] (smtp) ERROR  TO %s: %s", to, es)
			return False
		except socket.error, es:
			self.logger.error("[notify] (smtp) ERROR  TO %s: %s", to, es)
			return False


if __name__ == "__main__":
	
	import os
	import sys
	import logging
	
	os.environ.setdefault("DJANGO_SETTINGS_MODULE", "conf.settings")

	from sanet_common.mailing import load_mail_handler_class

	logger = logging.getLogger("prova")

	log_format = '%(asctime)s %(levelname)s> %(message)s'
	logger.setLevel(10)
	hdlr = logging.StreamHandler()
	hdlr.setFormatter( logging.Formatter(log_format) )
	hdlr.setLevel( 10 )
	logger.addHandler(hdlr)

	#from sanet_common.mailing.sendmail import SendMailShellHandler as handler
	#h = handler(logger, { "command": "/usr/local/bin/sendmail -f noc@labs.it -t" } )

	#from sanet_common.mailing.djangosmtp import DjangoMailHandler as handler
	
	handler = load_mail_handler_class('sanet_common.mailing.djangosmtp.DjangoMailHandler')
	
	h = handler(logger, { 
		"HOST"               : "localhost",
		"PORT"               : 25,
		
		#"DEFAULT_DOMAIN_NAME": "localdomain",
		
		"ENVELOPE_SENDER"    : "noc@labs.it",
	})

	frm     = sys.argv[1]
	to      = sys.argv[2]
	subject = sys.argv[3]
	body    = sys.argv[4]

	h.send(frm, to, subject, body, {
		"Header1": "hello",
		"Header2": "world",
		'X-Extra': 'extra header',
		"X-Header3": "the quick brown fox jumps over the lazy dog.",
	})
	
