# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""
################################################################################
Sendmail backend
################################################################################


Introduzione
-------------

Questo modulo contiene classi di tipo :class:`BaseMailHandler` che utilizzano
i comandi di sistema (*sendmail*) per invaire email.

Questa classe uo' essere istanziata ed utilizzata solo all'interno di un'applicazione django.


::
	
	      SendMailShellHandler (BaseMailHandler)
	                      |
	______________________|________________________
	                      | 
	  OS                  |
	                      v
	         SHELL COMMAND: /bin/sendmail ?

Esempio di utilizzo
-----------------------

	::
		
		import os
		import sys
		import logging
		
		from sanet_common.mailing import load_mail_handler_class
		
		logger = logging.getLogger("log")
		
		handler = load_mail_handler_class('sanet_common.mailing.sendmail.SendMailShellHandler')
		
		h = handler(logger, { 
			"COMMAND"       : "/usr/sbin/sendmail",
			"ENVELOP_SENDER": "noc@localhost.localdomain",
		})
		
		h.send("foo", "bar", "subject", "body", {'X-Extra': 'extra header'})

Classi
-------------

.. autoclass:: SendMailShellHandler
	:private-members:
	:special-members:
	:members:
"""

import subprocess

from sanet_common.mailing import BaseMailHandler

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

class SendMailShellHandler(BaseMailHandler):
	"""
	Questa classe permette l'invio di email utilizzando i comandi di sistema.
	"""

	def __init__(self, logger, settings):
		"""
		Costruttore dell'oggetto.
		
		La classe deve essere istanziata passando in input al costruttore:
		
			* Un oggetto *logger*
			* Un dizionario con i parametri di configurazione
		
		Il dizionario di configurazione deve contenere le seguenti chiavi:
		
		::
			
			{ 
				"COMMAND"            : <path sendmail command>,
				"ENVELOPE_SENDER"    : <envelope sender email address>,
			}
		
		:param  logger: A python logger.
		:type logger:  logging.Logger
		
		:param  settings: A dictionary with specific configuraion data for the objects.
		:type settings:  dict
		"""
		super(SendMailShellHandler, self).__init__(logger, settings)

		self.logger.info("Init sendmail handler")

	def send(self, frm, to, subject, body, extra_headers_dict={}, attachments=[]):
		"""
		Vedi :py:meth:`~.BaseMailHandler.send`.
		
		:exceptions: Solleva una eccezione se il comando specificato da *COMMAND* non puo' essere eseguito.
		
		"""
		
		
		#
		# FIX: attachments NOT SUPPORTED YET
		#

		self.logger.debug("send(%s, %s, %s, %s, %s)", to, subject, body, extra_headers_dict, frm)

		extra_headers_dict['X-MailHandler'] = str(self.__class__.__name__)

		#
		# Prepare input for sendmail process
		#
		msg = ''
		if extra_headers_dict:
			msg += ''.join([ "%s: %s\n" % (k,v) for (k,v) in extra_headers_dict.items() ])

		msg +=  "From: %s\n" \
			"To: %s\n" \
			"Subject: %s\n" \
			"\n" \
			"%s" \
			"\n" % (frm, to, subject, body)

		#
		# Launch sendmail
		#
		sendmail_params = [ self.settings["COMMAND"] , "-f", self.settings["ENVELOP_SENDER"], "-t" ]

		self.logger.info("[notify] using: %s", sendmail_params  )

		#
		#
		#
		p = None

		try:
			self.logger.debug("[notify] Launching sendmail command")

			# Lauch sendmail
			p = subprocess.Popen(sendmail_params, stdin=subprocess.PIPE)

			self.logger.debug("[notify] Sending data")
			p.communicate(msg) # Sending stdin

			self.logger.debug("[notify] Sending waiting")
			p.wait()

			self.logger.debug("[notify] Return code: %s", p.returncode)

			if p.returncode == 0:
				return True
			else:
				return False

		except OSError, e:
			if p != None:
				self.logger.error("[notify] Error while launching sendmail command. Return code %s -> %s", p.returncode, e)
			else:
				self.logger.error("[notify] Error while launching sendmail command: %s", e)

			return False

if __name__ == "__main__":

	import os
	import sys
	import logging

	from sanet_common.mailing import load_mail_handler_class

	logger = logging.getLogger("prova")

	log_format = '%(asctime)s %(levelname)s> %(message)s'
	logger.setLevel(10)
	hdlr = logging.StreamHandler()
	hdlr.setFormatter( logging.Formatter(log_format) )
	hdlr.setLevel( 10 )
	logger.addHandler(hdlr)

	handler = load_mail_handler_class('sanet_common.mailing.sendmail.SendMailShellHandler')

	h = handler(logger, { 
		"COMMAND"       : "/usr/sbin/sendmail",
		"ENVELOP_SENDER": "noc@labs.it",
	})

	frm     = sys.argv[1]
	to      = sys.argv[2]
	subject = sys.argv[3]
	body    = sys.argv[4]

	h.send(frm, to, subject, body, {'X-Extra': 'extra header'})
