#!/usr/bin/python

# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it> 
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""
################################################################################
Network utilities
################################################################################

Introduction
================================================================================

TODO

.. automodule:: sanet_common.netutils.ber

.. automodule:: sanet_common.netutils.snmptrap

"""

import os
import sys
import random
import re
import string
import threading
import socket

import time
from time   import time   as time_fun

import array

import socket
from socket import ntohs

import select
from select import select as select_fun

import struct
from struct import unpack as unpack_fun

import ipaddr

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

IPVERSION = 4

IP_IHL_MIN = 5

IP_FLAG_DF = 0x4000
IP_FLAG_MF = 0x2000


#--------------------------------------------------------------------------------------------
# Ip utility

masks_by_bits={
	0:'0.0.0.0',
	1:'128.0.0.0',
	2:'192.0.0.0',
	3:'224.0.0.0',
	4:'240.0.0.0',
	5:'248.0.0.0',
	6:'252.0.0.0',
	7:'254.0.0.0',
	8:'255.0.0.0',
	9:'255.128.0.0',
	10:'255.192.0.0',
	11:'255.224.0.0',
	12:'255.240.0.0',
	13:'255.248.0.0',
	14:'255.252.0.0',
	15:'255.254.0.0',
	16:'255.255.0.0',
	17:'255.255.128.0',
	18:'255.255.192.0',
	19:'255.255.224.0',
	20:'255.255.240.0',
	21:'255.255.248.0',
	22:'255.255.252.0',
	23:'255.255.254.0',
	24:'255.255.255.0',
	25:'255.255.255.128',
	26:'255.255.255.192',
	27:'255.255.255.224',
	28:'255.255.255.240',
	29:'255.255.255.248',
	30:'255.255.255.252',
	31:'255.255.255.254',
	32:'255.255.255.255',
}

bits_by_masks={
	'0.0.0.0': 0,
	'128.0.0.0': 1,
	'192.0.0.0': 2,
	'224.0.0.0': 3,
	'240.0.0.0': 4,
	'248.0.0.0': 5,
	'252.0.0.0': 6,
	'254.0.0.0': 7,
	'255.0.0.0': 8,
	'255.128.0.0': 9,
	'255.192.0.0': 10,
	'255.224.0.0': 11,
	'255.240.0.0': 12,
	'255.248.0.0': 13,
	'255.252.0.0': 14,
	'255.254.0.0': 15,
	'255.255.0.0': 16,
	'255.255.128.0': 17,
	'255.255.192.0': 18,
	'255.255.224.0': 19,
	'255.255.240.0': 20,
	'255.255.248.0': 21,
	'255.255.252.0': 22,
	'255.255.254.0': 23,
	'255.255.255.0': 24,
	'255.255.255.128': 25,
	'255.255.255.192': 26,
	'255.255.255.224': 27,
	'255.255.255.240': 28,
	'255.255.255.248': 29,
	'255.255.255.252': 30,
	'255.255.255.254': 31,
	'255.255.255.255': 32,
}

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def base256_list2num(l):
	exp = 0
	num = 0
	while l:
		num = num + l.pop() * (256**exp)
		exp = exp + 1
	return num

def base256_list2ipv4(l):
	return str(ipaddr.IPv4Address(base256_list2num(l)))

def base256_list2ipv6(l):
	return str(ipaddr.IPv6Address(base256_list2num(l)))

def base256_list2ip(l):
	if len(l) > 4:
		return str(ipaddr.IPv6Address(base256_list2num(l)))
	else:
		return str(ipaddr.IPv4Address(base256_list2num(l)))

def base256_list2blob(l):
        # "180.164.227.100.112.176" -> 'B4A4E36470B0'
        return ''.join([ '%02x' % int(a) for a in l.split(".") ]).upper()

def ascii_list2str(l):
	rv=''
	for i in l:
		rv = rv + chr(i)
	return rv

def ascii_str2list(s):
	rv=[]
	for c in s:
		rv.append(ord(c))
	return rv

def mask2bits(mask):
	return bits_by_masks.get(mask, None)

def bits2mask(bits):
	return mask_by_bits.get(bits, None)

def valid_ip(address):
	try: 
		ipaddr.IPv4Network(address)
		return True
	except:
		return False

def valid_partial_ip(addr):

	# (partial) sintax only
	if not re.match(r"^(\d{1,3})(\.(\d{1,3})(\.(\d{1,3})(\.(\d{1,3}))?)?)?$", addr):
		return False

	# Valid numbers
	try:
		# a.b.c.d
		# a.b.c
		# a.b
		# a
		addr += ".0" * (4 - len(addr.split("."))) # 127 -> 127.0.0.0

		# NOTE: inet_aton(999) returns a strange result. Check inverse too.
		n = socket.inet_aton(addr)
		a = socket.inet_ntoa(n)
		return True
	except Exception, e:
		#print e
		pass
	
	return False

def addr_ntoa(n_addr):
	try:
		addr_ntoa = socket.inet_ntoa(struct.pack('!L',n_addr))
	except socket.error:
		raise #ValueError, "La stringa %s non e' un valido indirizzo IP"	 % addr

	return addr_ntoa
inet_ntoa = addr_ntoa

def addr_aton(addr):
	try:
		addr_aton = struct.unpack('!L', socket.inet_aton(addr))[0]
	except socket.error:
		raise ValueError, "La stringa %s non e' un valido indirizzo IP"	 % addr

	return addr_aton
	
inet_aton = addr_aton
		
def get_aton_bounds_from_cidr(base, bits):
	aton_base = inet_aton(base)
	full_aton_mask = 2**32 - 1
	aton_mask =  2**(32 - bits) - 1

	aton_haddr = aton_base | aton_mask
	aton_laddr = aton_base & (full_aton_mask - aton_mask)

	return (aton_laddr, aton_haddr)

def ip_is_inside_inet_aton_range(ip, laddr_aton, haddr_aton):
	ip_aton = addr_aton(ip)
	if (ip_aton <= haddr_aton) and (ip_aton >= laddr_aton):
		return True
	else:
		return False

def ip_is_inside_inet_aton_ranges(ip, aton_ranges):
	ip_aton = addr_aton(ip)
	for laddr_aton, haddr_aton in aton_ranges:
		if (ip_aton <= haddr_aton) and (ip_aton >= laddr_aton):
			return True
	return False

def ip_is_inside_range(ip, laddr, haddr):
	laddr_aton = addr_aton(laddr)
	haddr_aton = addr_aton(haddr)
		
	return ip_is_inside_inet_aton_range(ip, laddr_aton, haddr_aton)

def explode_ip_range(laddr, haddr):
	laddr_aton = addr_aton(laddr)
	haddr_aton = addr_aton(haddr)
	
	ip_addrs = map(lambda x : addr_ntoa(x), range(laddr_aton, haddr_aton))
	return ip_addrs

private_ip_ranges = [
	('192.168.0.0','192.168.255.255'), 
	('172.16.0.0','172.31.255.255'), 
	('10.0.0.0','10.255.255.255')
]

exp_ip_ranges = [
	('224.0.0.0','255.255.255.255')
]
aton_private_ranges = map(lambda r : (addr_aton(r[0]), addr_aton(r[1])) , private_ip_ranges)
aton_exp_ip_ranges = map(lambda r : (addr_aton(r[0]), addr_aton(r[1])) , exp_ip_ranges)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#


def resolve_ip_addr(hostname, version, stable_order=True):
	"""
	Risolve "hostname" e restituisce gli indirizzi IP in base alla versione IP specificata da "version": 4 per IPv4 e 6 per IPv6.
	
	Il parametro "do_sort" indica se gli indirizzi restituiti devono essere restituiti in ordine "stabile" o "casuale".
	
	Se "do_dort" e' True, gli indirizzi vengono ordinati in base all'ordinamento crescente nativo del formato binario dell'indirizzo IP (network order o bigendian.
	(gli indirizzi vengono ordinati confrontanto i byte dal piu' significativo al meno significativo)
	"""

	result = []

	r = socket.getaddrinfo(hostname, None)
	if version == 4:
		family = 2
	else:
		family = 10

	#
	# OUTPUT 
	#
	#    (family, socktype, proto, canonname, sockaddr)
	#
	
	# Sample result, with IPv6 converted by pprint():
	#[( 2, 1,  6, '', ('137.204.2.228', 0)),
	# ( 2, 2, 17, '', ('137.204.2.228', 0)),
	# ( 2, 3,  0, '', ('137.204.2.228', 0)),
	# (10, 1,  6, '', ('2001:760:2e00:affe::a001', 0, 0, 0)),
	# (10, 2, 17, '', ('2001:760:2e00:affe::a001', 0, 0, 0)),
	# (10, 3,  0, '', ('2001:760:2e00:affe::a001', 0, 0, 0))]
	if r:
		for i in r:
			if i[0] == family and i[2] == 0:
				result.append(i[4][0])
			# Currently this does not work:
			#elif i[0] == 10 and i[2] == 0:
			#	self.v6.append(i[4][0])

	if stable_order:
		# order by network order 
		sortable_tuples = [ (socket.inet_pton(family, ip_str), ip_str) for ip_str in result                    ]
		result          = [ t[1]                                       for t      in sorted( sortable_tuples ) ]

	return result

class _resolve_ip(threading.Thread):
	def __init__(self, name, version, debug=False):
		super(_resolve_ip, self).__init__()
		self.debug = debug
		self.name = name
		self.exception = None
		self.result=None
		self.version=version
	
	def run ( self ):
		try:
			if self.debug: print "__resolve_ip(%s) run" % self.name
			self.result = resolve_ip_addr(self.name, self.version)
			if self.debug: print "__resolve_ip(%s) got : %r" % (self.name, self.result)
			
		except Exception, es:
			if self.debug: print "__resolve_ip(%s): got exception = %s" % (self.name, str(es))
			self.exception = es

def resolve_ip(name, version, debug=False, timeout=2):
	t=_resolve_ip(name, version, debug=debug)
	t.start()
	
	if timeout >= 0:
		t.join(timeout)
	else:
		t.join()
	
	if t.isAlive():
		raise socket.timeout

	if t.exception:
		raise t.exception
	
	return t.result

def resolve_ipv4(name, debug=False, timeout=2):
	return resolve_ip(name, 4, debug=debug, timeout=timeout)

def resolve_ipv6(name, debug=False, timeout=2):
	return resolve_ip(name, 6, debug=debug, timeout=timeout)

def aaaa(addr):
	public_srv_addrs = []
	c = 0
	for aton_r in aton_private_ranges:
		if ip_is_inside_inet_aton_range(addr, aton_r[0], aton_r[1]):
			break
		else:
			c += 1
	if c == len(aton_private_ranges):
		public_srv_addrs.append(addr)
	return public_srv_addrs

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def cksum(s):
	if len(s) & 1:
		s = s + '\0'
	words = array.array('h', s)
	sum = 0
	for word in words:
		sum = sum + (word & 0xffff)
	hi = sum >> 16
	lo = sum & 0xffff
	sum = hi + lo
	sum = sum + (sum >> 16)
	return (~sum) & 0xffff




ICMP_ERR_UNREACHABLE = 3
ICMP_CODE_PKT_TOO_BIG = 4 


def debug(s):
	#print(s)
	pass

blob_re=re.compile(r'^([0-9A-Fa-f]{12})$')
colon_re=re.compile(r'^(([0-9A-Fa-f]{2}:){5}[0-9A-Fa-f]{2})$')
dash_re=re.compile(r'^(([0-9A-Fa-f]{2}-){5}[0-9A-Fa-f]{2})$')
colon12_re=re.compile(r'^(([0-9A-Fa-f]{1,2}:){5}[0-9A-Fa-f]{1,2})$')
cisco_re=re.compile(r'^(([0-9A-Fa-f]{4}\.){2}[0-9A-Fa-f]{4})$')
hp_re=re.compile(r'^(([0-9A-Fa-f]{6}-)[0-9A-Fa-f]{6})$')
base256_re=re.compile(r'^(([0-9]{1,3}\.){5}[0-9]{1,3})$')
spaced_re=re.compile(r'^(([0-9A-Fa-f]{1,2} ){5}[0-9A-Fa-f]{1,2})$')

def clean_mac(mac):
	"""Convert various MAC address string formats into blob: no separator, hex uppercase. Returns None if cannot convert."""
	m = blob_re.search(mac)
	if m:
		return string.upper(m.groups()[0])
	if colon_re.search(mac):
		return string.upper(''.join(mac.split(':')))
	if dash_re.search(mac):
		return string.upper(''.join(mac.split('-')))
	if cisco_re.search(mac):
		return string.upper(''.join(mac.split('.')))
	if hp_re.search(mac):
		return string.upper(''.join(mac.split('-')))
	if colon12_re.search(mac):
		blob=''
		for i in mac.split(':'):
			blob += "%02X" % int(i,16)
		return blob
	if base256_re.search(mac):
		return base256_list2blob(mac)
	if spaced_re.search(mac):
		blob=''
		for i in mac.split(' '):
			blob += "%02X" % int(i,16)
		return blob
	if len(mac) == 6:
		return ''.join(map(lambda x: '%02X' % ord(x), mac))
	return None

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class Ping(object):

	def __init__(self, host='127.0.0.1', timeout=2.0, count=3, size=1472, delay=0.0, callback=None, do_connect=True, do_frag=True, bind_address=None):
		
		size = int(size)
		self.host = host
		self.timeout=timeout
		self.count=count
		self.size=size
		self.do_frag=do_frag
		self.test_text='THE QUICK BROWN FOX JUMPS OVER THE LAZY DOG 1234567890 '
		self.test_text_len=len(self.test_text)
		self.id = random.randint(0,200)
		self.seq = random.randint(0,200)

		self.delay=delay

		self.callback = callback

		self.do_connect = do_connect

		self.error = None

		try:
			self.s = None
			
			#
			# Send ICMP packets, make the kernel do the dirty job about IP headers.
			#
			self.s = socket.socket(socket.AF_INET, socket.SOCK_RAW, socket.IPPROTO_ICMP)
			
			# Include IP headers
			#self.s.setsockopt(socket.IPPROTO_IP, socket.IP_HDRINCL, 1)
			
			if bind_address != None:
				try:
					if self.callback: self.callback("Binding source address to %s" % (bind_address))
					self.s.bind( (bind_address, 0) )
				except socket.error, e:
					raise Exception("Unable to bind icmp socket to address %s (error: %s )" % ( bind_address, e) )
				except Exception, e:
					raise Exception("Unable to bind icmp socket to address %s (error: %s )" % (bind_address, e) )
		
			
			# Bind to the destination host in order to filter incoming responces 
			# WARNING: if we connect() the kernel will intercept
			#          the incoming ICMP errors and transform them in 
			#          socket errors. 
			if self.do_connect:
				self.s.connect( (self.host,0) )
			
			if not self.do_frag:
				#
				# Don't frag outgoing packets
				#
				# Linux doesn't support the IP_DONTFRAG options. Use IP_MTU_DISCOVER instead (man 7 ip).
				#
				# Python's socket module does not define the required symbols!!!
				#
				
				# Taken from: linux/in.h
				IP_MTU_DISCOVER   = 10
							# man INFO                      /  linux/in.h  COMMENTS
				IP_PMTUDISC_DONT  = 0   # Never do Path MTU Discovery   /  Never send DF frames
				IP_PMTUDISC_WANT  = 1   # Use per-route settings.       /  Use per route hints
				IP_PMTUDISC_DO    = 2   # Always do Path MTU Discovery  /  Always DF
				IP_PMTUDISC_PROBE = 3   # Set DF but ignore Path MTU    /  Ignore dst pmtu        <---
			
				#
				# With IP_PMTUDISC_PROBE we send IP packets with DF flag  without
				# checking kernel's Path MTU infos, but somehow it isn't supported on
				# some linux kernels ( =< 2.6 ?). Use IP_PMTUDISC_DO instead.
				#
				# This might work as like as IP_DONTFRAG options.
				#
				# We should receive ICMP errors "Destination unreachable too big" (ICMP type 3, 
				# code 4) when the packet size is too big for the destiantion MTU.
				#
				# NOTE: there are tricky differences between the man page and 
				# comments in linux headers about the options' semanthic!!
				#
				ok = False
				try:
					self.s.setsockopt(socket.IPPROTO_IP, IP_MTU_DISCOVER, IP_PMTUDISC_PROBE)
					ok = True
					if self.callback: self.callback("Using IP_PMTUDISC_PROBE")
				except:
					pass
				
				if not ok:
					try:
						self.s.setsockopt(socket.IPPROTO_IP, IP_MTU_DISCOVER, IP_PMTUDISC_DO)
						if self.callback: self.callback("Using IP_PMTUDISC_DO")
					except:
						pass
				

			if True: # It seems these filtering speeds up the recvfrom() 
				
				# bits/socket.h
				SOL_RAW             = 255		

				# linux/icmp.h
				ICMP_FILTER         = 1

				ICMP_ECHOREPLY      =  0   # Echo Reply			
				ICMP_DEST_UNREACH   =  3   # Destination Unreachable
				ICMP_SOURCE_QUENCH  =  4   # Source Quench	
				ICMP_REDIRECT       =  5   # Redirect (change route)
				ICMP_ECHO           =  8   # Echo Request		
				ICMP_TIME_EXCEEDED  = 11   # Time Exceeded	
				ICMP_PARAMETERPROB  = 12   # Parameter Problem	
				ICMP_TIMESTAMP      = 13   # Timestamp Request	
				ICMP_TIMESTAMPREPLY = 14   # Timestamp Reply	
				ICMP_INFO_REQUEST   = 15   # Information Request
				ICMP_INFO_REPLY	    = 16   # Information Reply	
				ICMP_ADDRESS        = 17   # Address Mask Request
				ICMP_ADDRESSREPLY   = 18   # Address Mask Reply
				
				try:
					data = ~((1<<ICMP_SOURCE_QUENCH)|
						 (1<<ICMP_DEST_UNREACH)|
						 (1<<ICMP_TIME_EXCEEDED)|
						 (1<<ICMP_PARAMETERPROB)|
						 (1<<ICMP_REDIRECT)|
						 (1<<ICMP_ECHOREPLY));	
					
					self.s.setsockopt(SOL_RAW, ICMP_FILTER, data)
				except Exception, e:
					#print e
					pass

			self.closed = False
		except socket.error, e:
			# Handle better
			raise e
		
		self.payload=''
		self.packet=''
		self.packet_len=0

	def test_pattern(self, n):
		rv='%d %f ' % (os.getpid(), time.time())
		if len(rv) >= n:
			return(rv[0:n])
	
		base=len(rv)
		for i in range(len(rv),n):
			rv = rv + self.test_text[(i + base)%self.test_text_len]
		return rv

	def new_payload(self):
		self.payload=self.test_pattern(self.size)

	@staticmethod
	def _in_cksum(packet):
		"""From Lars Strand"""
	
		if len(packet) & 1:
			packet = packet + '\0'
	
		words = array.array('h', packet)
		sum = 0
	
		for word in words:
			sum += (word & 0xffff)
	
		hi = sum >> 16
		lo = sum & 0xffff
		sum = hi + lo
		sum = sum + (sum >> 16)
	
		return (~sum) & 0xffff # return ones complement
	
	def new_packet(self):
		
		self.new_payload()
		
		# ICMP HEADER
		#
		#    0                   1                   2                   3
		#    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
		#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
		#   |     Type      |     Code      |          Checksum             |
		#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
		#   |           Identifier          |        Sequence Number        |
		#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
		#   |     Data ...
		#   +-+-+-+-+-
		
		icmp_type = 8  # echo request
		icmp_code = 0
		
		csum     = 0
		self.seq = (self.seq + 1 ) % 256
		icmp_header = struct.pack('bbHHh', icmp_type, icmp_code, csum, self.id, self.seq)
		
		packet = icmp_header + self.payload
		csum   = self._in_cksum(packet)
		icmp_header = struct.pack('bbHHh', icmp_type, icmp_code, csum, self.id, self.seq)
		
		self.packet = icmp_header + self.payload
		self.packet_len = len(self.packet)
		
	def send_packet(self, n):
		self.error = None
		try:
			self.s.sendto(self.packet, (self.host, 1))
		except socket.error, e:
			self.error = "ping #%s: %s (errno=%s)" % (n, e.args[1], e.args[0])

	def receive_answer(self, n):
		
		self.error = None
		
		now   = time_fun()
		begin = now
		limit = begin + self.timeout
		
		# Move these symbols to local variables in order to speed up
		# symbols lookup in the loop.
		sockfd     = self.s
		
		try:
			
			while True:
				
				now   = time_fun()
				
				if now > limit:
					break
					
				iwtd, owtd, ewtd = select_fun([sockfd], [], [], limit - now)
				when = time_fun()
				if iwtd:
					data, addr = sockfd.recvfrom(self.packet_len + 20)
					addr = addr[0]
					
					
					#
					# Parse received packet
					# NOTE: AF_INET + SOCKET_RAW always include IP headers in received packets
					#
					
					# IP HEADER
					#ip4_header = data[0:20]
					#(v_ihl, tos, totlen, id, flags_offset, ttl, proto, cksum, src, dst) = unpack_fun("!BBHHHBBH4s4s", ip4_header)
					#src = socket.inet_ntoa(src)
					
					# ICMP HEADER
					icmp_header = data[20:28]
					r_type, r_code, r_csum, r_id, r_seq, = unpack_fun("bbHHH", icmp_header)
					r_payload = data[28:]
					
					#if self.callback: self.callback("ICMP from %r, type %d, code %d, id %d, seq %d, payload_len %s" % (addr, r_type, r_code, r_id, r_seq, len(r_payload)))
					
					#
					# Ignore unrecognized ICMP packets. FIX: use socket option ICMP_FILTER!?
					#
					
					if r_type == 0:
						
						if addr != self.host:
							#if self.callback: self.callback('wrong addr %r (want %r)' % (addr, self.host))
							continue
											
						if r_code != 0: # always 0 with ICMP Echo replys
							#if self.callback: self.callback('wrong reply code')
							continue
							
						if r_id != self.id:
							#if self.callback: self.callback('wrong reply id')
							continue
							
						if r_seq != self.seq:
							#if self.callback: self.callback('wrong reply seq')
							continue
							
						# Check response payload
						lp = len(self.payload)
						
						lr = len(r_payload)
						
						if lr != lp:
							#if self.callback: self.callback('payload different in length (%s,%s)' % (lp,lr) )
							pass
							
						min_len = min(lr, lp)
						if r_payload[0:min_len] != self.payload[0:min_len]:
							#if self.callback: self.callback('wrong payload')
							continue
							
							
					elif r_type == ICMP_ERR_UNREACHABLE:
						
						if r_code == ICMP_CODE_PKT_TOO_BIG:
							
							#
							# Analize the original packet seen from the remote gateway
							#
							ip4_header = r_payload[0:20]
							(v_ihl_2, tos_2, totlen_2, id_2, flags_offset_2, ttl_2, proto_2, cksum_2, src_2, dst_2) = unpack_fun("!BBHHHBBH4s4s", ip4_header)
							
							if proto_2 != 1: # ICMP
								#if self.callback: self.callback('ICMP response doesn\'t contain our packet (IPv4+proto(%s))' % (proto_2) )
								continue
								
							ip4_header_len = (v_ihl_2 & 0x0F) * 4
							r_payload = r_payload[ip4_header_len + 8:]  #skip ip header + icmp header

							# 
							# Compare payload
							#
							min_len = min( len(r_payload) , len(self.payload))
							
							if r_payload[0:min_len] != self.payload[0:min_len]:
								#if self.callback: self.callback('wrong payload')
								continue
												
							# It was our packet!	
							next_mtu = ntohs(r_seq)
							self.error = 'packet too big from %s (next MTU %s)' % (addr, next_mtu)
							break
						else:
							#if self.callback: self.callback('wrong unreach code')
							continue
					else:
						#if self.callback: self.callback('wrong type')
						continue
					
					return when
					
				#now = time.time()
			
		except socket.error, e:
			self.error = str(e)

		# Timeout or Destination unreachable
		return None

	def ping(self):
		probe_rtts = []
		probe_infos = []
		
		for n in xrange(0, self.count):
			
			self.new_packet()

			if self.callback: self.callback("sending icmp #%d len=%s" % (n, len(self.packet)))

			elapsed = None
			info    = ''
			
			self.send_packet(n)
			if self.error != None:
				#if self.callback: self.callback(self.error)
				info = self.error
			else:
				begin = time_fun()
				when = self.receive_answer(n)
				if when != None:
					elapsed = when - begin
					
					# The ICMP response is valid!
					#if self.callback: self.callback('icmp reply #%s received' % n)
				else:
	
					if self.error != None:  # timeout
						if self.callback: self.callback(self.error)

						probe_infos.append(self.error)
						info = self.error
					else:	
						info = 'timeout'
						if self.callback: self.callback(info)

					#probe_rtts.append(None)
					
					
			probe_rtts.append( elapsed )
			probe_infos.append( info )

			time.sleep(self.delay)

		return (probe_rtts, probe_infos)

	def close(self):
		self.s.close()
		self.s = None

	def __del__(self):
		if self.s != None:
			self.s.close()





if __name__ == "__main__":

	example_addresses = ['10.0.0.10', '192.168.0.1', '232.10.11.20']

	for addr in example_addresses:
		print "Indirizzo %s " % addr
		if len(aaaa(addr)):
			print "pubblico"
		else:
			print "privato"

	example_macs = ['000102aA0e0F', '00:01:02:aa:0e:0F', '0:1:02:Aa:E:F', '000102-aa0E0f', '00-01-02-Aa-0e-0f', '0001.02AA.0e0f']
	for addr in example_macs:
		print "%s convertito in: %s" % (addr, clean_mac(addr))
