# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""
Queue File
==================

La classe QueueFile permette di scrivere leggere/dati con politica FIFO.

Il file contiene:
    - un header di metainformazioni fisso
    - una sequenza di blocchi di dati
        
        <header> <data1> <data2> <data3> ...

L'header ha un struttura fissa con i seguenti campi:

        +-----------+-------------+--------------+
        | read_pos  |  write_pos  | stored_bytes |
        +-----------+-------------+--------------+


    BYTE POS     LEN     TIPO      NOME            DESCRIZIONE
              
    0            8       double    read_pos
    8            8       double    write_pos
    16           8       double    stored_bytes
            TOT  24
            
Il blocco dati e' composto da un header fisso e una parte dati di lunghezza variabile
specificata nell'header

        +-----------+-------------+--------------
        | timestamp |     len     |    ......
        +-----------+-------------+--------------

    BYTE POS     LEN     TIPO      NOME            DESCRIZIONE
              
    0            8       double    read_pos
    8            4       integer   write_pos
                 ...               stored_bytes

Gestione lettura e scrittura
===============================

Nuovi blocchi dati vengono sempre aggiunti in fonto al file.

IL parametro read_pos indica la posizione del prossimo blocco da leggere.


Poiche' non sappiamo se le write() di python sono "atomiche" teniamo 
sempre salvata la posizione dove scrivere il prossimo blocco dati: write_pos.

Se write_pos non corrisponde alla fine del file significa che dei dati erano stati scritti
ma il processo e' morto per qualche motivo e potrebbero non essere stati scritti correttamente
tutti.

Quando (read_pos == write_pos) allora non ci sono dati da leggere. In questo caso il file puo' essere azzerato troncando il contenuto


                       read pos                                 write_pos
                       |                                        |
                       v                                        v
     HHHHH hhDDDDD hhDD hhDDDDDDDD hhDDDD hhDDDD hhDDDDDD hhDDDD xxxxxxxxxxxxxxxxxxx
                                                                 \___partial data__/      
    |-------------------------------------------------------------------------------|
   BOF                                                                             EOF

"""
import os
import struct
import time

DEBUG = 0
    
#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

FILE_HEADER_FORMAT = "ddd"
FILE_HEADER_SIZE = len(struct.pack(FILE_HEADER_FORMAT, 
	0.0, # read_pos
	0.0, # write_pos
	0.0, # stored_bytes
))

BLOCK_HEADER_FORMAT = "!dL"
BLOCK_HEADER_SIZE   = len(struct.pack(BLOCK_HEADER_FORMAT, 
	0.0,   # timestamp
	0,     # data_len
))


def CREATE_BLOCK(data):
	return struct.pack(BLOCK_HEADER_FORMAT, time.time(), len(data)) + data


#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class QueueFileException(Exception):
	pass

class QueueFileBadFormatException(QueueFileException):
	pass

class QueueFileMaxSizeReachedException(QueueFileException):
	pass

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class QueueFile(object):

	def __init__(self, filepath, maxsize):
		self.filepath = filepath
		self.maxsize  = maxsize

		self.f            = None

		self.read_pos     = None
		self.write_pos    = None
		self.stored_bytes = None

		self.uncommitted_read_pos = None
		self.uncommitted_read_len = None

		self.had_uncommited_writes = False

		self._reset_meta()

	def _reset_meta(self):
		
		#self.last_id      = 0
		self.read_pos     = FILE_HEADER_SIZE
		self.write_pos    = FILE_HEADER_SIZE
		self.stored_bytes = 0
		
		self.uncommitted_read_pos = self.read_pos
		self.uncommitted_read_len = 0
	
	def get_info(self):
		return {
			#'last_id'              : #self.last_id,
			'read_pos'             : self.read_pos,
			'write_pos'            : self.write_pos,
			'stored_bytes'         : self.stored_bytes,
			'uncommitted_read_pos' : self.uncommitted_read_pos,
			'uncommitted_read_len' : self.uncommitted_read_len,
		}

	def isEmpty(self):
		return self.stored_bytes == 0

	def __move_to_bof(self):
		self.f.seek(0, os.SEEK_SET)

	def __move_to_read_pos(self):
		self.f.seek(self.read_pos, os.SEEK_SET)

	def __move_to_uncommited_pos(self):
		self.f.seek(self.uncommitted_read_pos, os.SEEK_SET)

	def __move_to_eof(self):
		self.f.seek(0, os.SEEK_END)
		
	def __flush(self):
		self.f.flush()
		os.fsync(self.f.fileno())

	def __read_header_block(self):

		if DEBUG: print __file__, "__read_header_block"

		self.__move_to_bof()

		if DEBUG: print __file__, "__read_header_block", self.f.tell()

		buff = self.f.read(FILE_HEADER_SIZE)
		if len(buff) != FILE_HEADER_SIZE:
			raise QueueFileBadFormatException("Invalid file format")

		data = struct.unpack(FILE_HEADER_FORMAT, buff)

		if DEBUG: print __file__, "__read_header_block", data

		self.read_pos     = data[0]
		self.write_pos    = data[1]
		self.stored_bytes = data[2]

		if DEBUG: print __file__, "__read_header_block", "now at=", self.f.tell()

		return

	def __write_header_block(self):
		#
		# Move to beginning of file
		#
		self.f.seek(0, 0)
		
		#
		# Write header
		#
		data = struct.pack(FILE_HEADER_FORMAT, 
			self.read_pos, 
			self.write_pos,
			self.stored_bytes,
		)
		self.f.write(data)

		#
		# FLUSH
		#
		self.__move_to_eof()

		return

	def open(self):

		if os.path.exists(self.filepath):
			self.f = open(self.filepath, "r+")
		else:
			self.f = open(self.filepath, "w+")


		
		#
		# NEW FILE or wrong file?
		#

		# Calculate file size
		self.f.seek(0, os.SEEK_END)
		filesize = self.f.tell()
		
		if DEBUG: print __file__, "open", "size=", filesize

		
		if filesize == 0:
			self.__write_header_block()
		elif filesize < FILE_HEADER_SIZE:
			raise QueueFileBadFormatException("Invalid file format")

		#
		# REAT META
		#
		self.__read_header_block()


		if DEBUG: print "FILE SIZE     :", filesize
		if DEBUG: print "INIT READ POS :", self.read_pos 
		if DEBUG: print "INIT WRITE POS:", self.write_pos
		if DEBUG: print "STORED BYTES  :", self.stored_bytes,

		if self.write_pos < filesize:

			self.had_uncommited_writes = True

			print "WARNING: uncommitted data"
			self.f.truncate(self.write_pos)

		self.__move_to_read_pos()

		return

	def close(self):
		self.f.close()

	def flush(self):
		self.__flush()
		
	def writeBlock(self, data):

		#
		# Reached limit?
		#
		used = self.stored_bytes

		if self.maxsize > 0:
			if len(data) > (self.maxsize - used):
				raise QueueFileMaxSizeReachedException("FULL datalen=%s  used=%s  max=%s" % (len(data), used, self.maxsize))

		#
		#
		#
		self.__move_to_eof()

		#
		# Write block
		#
		data_id  = time.time()
		data_len = len(data)
		
		if DEBUG: print __file__, "writeBlock", data_id, data_len

		buff = struct.pack(BLOCK_HEADER_FORMAT, data_id, data_len) + data

		self.f.write(buff)

		#
		# Update internal data
		#
		self.stored_bytes += data_len
		self.write_pos     = self.f.tell()

		#
		# Update header
		#
		self.__write_header_block()

		self.flush()

		
		if DEBUG: print __file__, "writeBlock", "now 2 at", self.f.tell()

		return data_id


	def readBlock(self):
		
		if DEBUG: print __file__, "readBlock", self.read_pos, self.uncommitted_read_pos

		self.__move_to_uncommited_pos()

		#
		# Read block header
		#
		buff = self.f.read(BLOCK_HEADER_SIZE)

		# EOF?
		if buff == "":
			return None

		#
		# PARSE HEADER
		#
		if len(buff) != BLOCK_HEADER_SIZE:
			raise QueueFileBadFormatException("Invalid block header len (%s)" % (len(buff)))

		if DEBUG: print __file__, "readBlock", "read header block", len(buff), repr(buff)

		try:
			data_id, data_len = struct.unpack(BLOCK_HEADER_FORMAT, buff)
		except Exception, e:
			raise QueueFileBadFormatException("Invalid block header format (%s)" % (str(e)))
		
		if DEBUG: print __file__, "readBlock", data_id, data_len
		
		#
		# Read block data
		#
		data = self.f.read(data_len)

		if len(data) != data_len:
			raise QueueFileBadFormatException("Invalid block header data (len %s. read %s)" % (data_len, len(data)))

		if DEBUG: print __file__, "readBlock", 'data len=', len(data), repr(data)

		#
		# Update 
		#
		self.uncommitted_read_pos  = self.f.tell()
		self.uncommitted_read_len += data_len
		return data


	def commit(self):

		self.read_pos      = self.uncommitted_read_pos
		self.stored_bytes -= self.uncommitted_read_len
		
		self.uncommitted_read_len = 0
		
		if self.read_pos == self.write_pos:

			self._reset_meta()

			self.f.truncate(self.write_pos)

		if DEBUG: print __file__, "readBlock", 'new read pos', self.read_pos

		self.__write_header_block()

		self.f.flush()

		return

	def rollback(self):

		self.uncommitted_read_pos  = self.read_pos
		self.uncommitted_read_len  = 0




if __name__ == "__main__":
	
	

	"""
	import sys

	filename = sys.argv[1]

	maxsize  = int(sys.argv[2])
	data     = sys.argv[3].encode('ascii')
	
	
	f = QueueFile(filename, maxsize)
	
	f.open()
	
	print "LAST_ID ", f.last_id
	print "READ POS", f.read_pos
	print "TELL    ", f.f.tell()
	
	print "------------------------"
	print "READING                 "
	print "------------------------"

	d = f.readBlock()
	while d:
		print d
		d = f.readBlock()

	f.commit()

	print "------------------------"
	print "WRITING                 "
	print "------------------------"
	
	for n in xrange(100):
		f.writeBlock( "%100s" % (data + "_" + str(n)) )

	print "LAST_ID ", f.last_id
	print "READ POS", f.read_pos
	print "TELL    ", f.f.tell()

	print "------------------------"
	print "READING                 "
	print "------------------------"

	data = f.readBlock()
	while data:
		print data
		data = f.readBlock()


	f.commit()

	f.flush()
	f.close()
	"""
	



	"""

	BLOCK_HEADER_FORMAT = "!dL"
	BLOCK_HEADER_SIZE   = len(struct.pack(BLOCK_HEADER_FORMAT, 
		0.0, 
		0,
	))


	def CREATE_HEADER_BLOCK(t, length):
		return struct.pack(BLOCK_HEADER_FORMAT, t, length)

	def CREATE_BLOCK(data):
		return CREATE_HEADER_BLOCK(1, len(data)) + data




	import sys

	filename = sys.argv[1]
	data     = sys.argv[2].encode('ascii')

	f_w = open(filename, "a+")
	f_r = open(filename, "r+")

	print f_w.tell()
	print f_r.tell()

	print "-" * 80

	f_w.write( CREATE_BLOCK(data) )
	f_w.flush()

	print "W TELL", f_w.tell()
	print "R TELL", f_r.tell()

	print "-" * 80

	while 1:
		
		header_pos = f_r.tell()
		
		buff = f_r.read(BLOCK_HEADER_SIZE)
		if not buff:
			break
		
		header = struct.unpack(BLOCK_HEADER_FORMAT, buff)
		data_type, data_len = header

		data = f_r.read(data_len)

		print "LEN %5s  TYPE %5s  DATA %5s %s " % ( data_len, data_type, len(data), repr(data))

		last_pos = f_r.tell()

		new_header = CREATE_HEADER_BLOCK(0, data_len)
		f_r.seek(header_pos, 0)
		f_r.write(new_header)
		f_r.flush()
		
		f_r.seek(last_pos)


		
	print "W TELL", f_w.tell()
	print "R TELL", f_r.tell()
	
	#f_r.truncate(0)
	#f_r.flush()

	f_r.seek(0, 0)
	f_w.seek(0, 0)

	print "-" * 80

	f_w.write( CREATE_BLOCK(data+data) )
	f_w.flush()

	print "W TELL", f_w.tell()
	print "R TELL", f_r.tell()

	print "-" * 80

	f_w.close()
	f_r.close()
	"""
	
	pass
