# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""
################################################################################
Timing utilities
################################################################################


Timerange string format
================================================

::
	

"""

import time
import re
import datetime

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

FIRST_WDAY = 1

LAST_WDAY = 7

#----------------------------------------------------------------------#

HOURS_SPEC    = r"^(\d+)(?:\:(\d+))?\-(\d+)(?:\:(\d+))?$"    # hh[:mm]-hh[:mm]

DAYS_H        = r"H"
DAYS_SPEC     = r"^([1-7]p?)(?:\-([1-7]p?))?$"               # <day>[-<day>]  , <day>p[-<day>p]

TIME_SPEC     = r"(?P<hspec>[^\/]+)?(?:\/(?P<wspec>.+))?"    # hours / days

TIMENAME_SPEC = r"[a-zA-Z][a-zA-Z0-9_-]*"

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

class AbstractCalendarHandler(object):
	"""
	Classe astratta per gestire un calendario (delle festivita'.)
	"""

	DEFAULT_CONTEXT = None

	EMPTY_CALENDAR = set()

	def set_default_time_context(self, time_context):
		"""
		La gestione delle festivita' puo' cambiare in base a diversi criteri
		(geografici, settoriali, ecc.)
		
		Tutte le festivita' sono associate ad un contesto "temporale" identificato da un nome.
		
		Il contesto temporale di default utilizzato quando non specificato e' 'DEFAULT' 
		"""
		raise NotImplemented()

	def isHoliday(self, year, month, mday, time_context=None):
		"""
		Restituisce True se il giorno (year, month, mday) e' un giorno
		festivo in baso al calendario delle festivita' valido per il 
		contesto temporale in corso.
		
		Se "time_context" e' specificato, il contesto temporale da utilizzare viene
		forzato a "time_context".
		Se "time_context" non e' specificato viene utilizzato il contesto "DEFAULT".
		"""
		raise NotImplemented()

	def isPreHoliday(self, year, month, mday, time_context=None):
		"""
		Restituisce True se il giorno (year, month, mday) e' un giorno
		festivo in baso al calendario delle festivita' valido per il 
		contesto "time_context".
		
		Se "time_context" e' specificato, il contesto temporale da utilizzare viene
		forzato a "time_context".
		Se "time_context" non e' specificato viene utilizzato il contesto "DEFAULT".
		"""
		raise NotImplemented()




	def dateToWDay(year, month, mday, time_context=None):
		""" 
		
		DEPRECATO DEPRECATO DEPRECATO DEPRECATO DEPRECATO
		
		Date to Week Day
		
		Calcola se il giorno identificato dalla tupla year-month-mday
		corrisponde ad un giorno "festivo" e restituisce il giorno
		della settimana corrispondente.
		Restituisce None se il giorno non gestito come festivo
		"""
		raise NotImplemented()

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

class NoHolidayCalendarHandle(AbstractCalendarHandler):
	"""
	Classe di comodo per avere un "converter" di default che non gestisce nessuna festivita'.
	"""
	def __init__(self):
		super(NoHolidayCalendarHandle, self).__init__()

	def set_default_time_context(self, time_context):
		return

	def dateToWDay(year, month, mday, time_context=None):
		return None

	def isHoliday(self, year, month, mday, time_context=None):
		return False

	def isPreHoliday(self, year, month, mday, time_context=None):
		return False

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

class SimpleCalendarHandler(AbstractCalendarHandler):
	"""
	Classe per gestire calendari memorizzati in un dizionario.
	
	La configurazione di tutti i calendari deve essere codificata tramite un dizionario
	strutturato in questo modo:
	
		{
			None: set([ 
				'MMDD',
				'YYYYMMDD',
			]),
			'<context1>': set([
				'MMDD',
				'YYYYMMDD',
			]),
			'<context2>': set([
				'MMDD',
				'YYYYMMDD',
			]),
		}
	"""

	def __init__(self, dict_calendar, forced_wday ):
		super(SimpleCalendarHandler, self).__init__()

		self._default_time_context = self.DEFAULT_CONTEXT
		self.calendar              = dict_calendar
		self.forced_wday           = forced_wday

	def set_default_time_context(self, time_context):
		self._default_time_context = time_context

	def dateToWDay(self, year, month, mday, time_context=None):

		if time_context is None:
			time_context = self._default_time_context

		calendar = self.calendar.get(time_context, self.EMPTY_CALENDAR)

		d  = '%02d%02d'     % (month, mday)
		d2 = '%04d%02d%02d' % (year, month, mday)

		if (d in calendar) or (d2 in calendar):
			return self.forced_wday

		return None

	def isHoliday(self, year, month, mday, time_context=None):

		if time_context is None:
			time_context = self._default_time_context

		calendar = self.calendar.get(time_context, self.EMPTY_CALENDAR)

		next_day = self.__isostring_to_date("%s-%s-%s" % (year, month, mday))

		d1 = next_day.strftime("%Y%m")
		d2 = next_day.strftime("%Y%m%d")

		if (d1 in calendar) or (d2 in calendar):
			return True

		return False

	def isPreHoliday(self, year, month, mday, time_context=None):

		if time_context is None:
			time_context = self._default_time_context

		calendar = self.calendar.get(time_context, self.EMPTY_CALENDAR)
		
		next_day = self.__get_next_day(year, month, mday)

		d1 = next_day.strftime("%Y%m")
		d2 = next_day.strftime("%Y%m%d")

		if (d1 in calendar) or (d2 in calendar):
			return True

		return False


	def __isostring_to_date(self, string):

		try:
			return datetime.datetime.strptime(string,'%Y-%m-%d')
		except ValueError:
			pass

		try:
			return datetime.datetime.strptime(string,'%Y-%m-%d %H:%M')
		except ValueError:
			pass

		try:
			return datetime.datetime.strptime(string,'%Y-%m-%d %H:%M:%S')
		except ValueError:
			pass

		return None

	def __get_next_day(self, year, month, mday):
		d = self.__isostring_to_date("%s-%s-%s" % (year, month, mday))
		d = d + datetime.timedelta(days=1)
		return d

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

def match_now(timespec, now,  calendar_handler=None, timespec_names={}, log=None, time_context=None):
	"""
	Check if given timestamp "now" matches the range expressed "timespec", //if it's a working day based on the "calendar_handler" object (instance of AbstractHolidayConverter subclass)???//
	if it's  // boh? // "timespec_names" dictionary.
	
	
	::
		
		time range:
		
			all
			none
			<time> [ "|" <time> ...]
		
		<time>:
			<time name>  
		
			[ <hours> [ "," <hours>...] ] "/" [ <days> [ "," <days>... ] ]
		
		<hours>:
			<hh> [ ":" <mm> ] "-" <hh> [ ":" <mm> ]
		
		<days>:
			"H"
			<d>  [ "-" <d>  ]
			<d>p [ "-" <d>p ]
		
		<hh>: 
			'00' .. '23'
		
		<mm>: 
			'00' .. '59'
		
		<d>:  
			'1' .. '7'
			
					1 = Monday
					2 = Tuesday
					3 = Wednesday
					4 = Thursday
					5 = Friday
					6 = Saturday
					7 = Sunday
	
	:param str timespec:             A string representing one or more time ranges or a short name mapped to timespec_names dictionary
	:param int now:                  The time stamp to check
	:param object holidays_converte: An instance of an AbstractHolidayConverter subclass
	:param dict timespec_names:      A dictionary representing different time intervals associated to shortnames //boh?//
	:param logger log:               A logger object
	"""
	if timespec == "" or timespec == "all" or timespec == "none":
		return True

	if calendar_handler is None:
		calendar_handler = NoHolidayCalendarHandle()

	#                                 sec  week  dayOfYear  dayLightSaving
	(year, month, mday, hour, minutes, d , wday,    e    ,        f       ) = time.localtime(now)

	if log: log.debug("now=%s", now)
	if log: log.debug("year=%s, month=%s, mday=%s, hour=%s, minutes=%s, d=%s, wday=%s, e=%s, f=%s)", year, month, mday, hour, minutes, d, wday, e, f)

	# days            Mon Tue Wed Thu Fri Sat Sun
	# localtime()     0   1   2   3   4   5   6
	# us              1   2   3   4   5   6   7

	# day of the week [1-7]
	wday = wday + 1

	if log: log.debug("wday: %s", wday)

	# A times spec is of the form
	#
	#       all
	#       none
	#       time [ | time ...]
	#
	# where 'time' is:
	#
	#       <time name>  
	#      
	#       [hours[,hours...]] / [days[,days...]]
	#
	# where 'hours' is:
	#
	# 	<hh>[:<mm>]-<hh>[:<mm>]
	#
	# where 'days':
	#
	#       H
	#       <d>  [ - <d> ]
	#       <d>p [ - <d>p ]
	#
	# hours are 24-hours, day 1 is monday

	if calendar_handler != None:
		# Is a special holiday? Forse it to the desired week of the day.
		hwday = calendar_handler.dateToWDay(year, month, mday)
		if hwday != None:
			wday = hwday

	if log: log.debug("Effective wday: %s", wday)

	# time [| time ]
	times = re.split(r"\|", timespec)
	for t in times:

		if log: log.debug("Testing: %s", t)

		t = t.strip()

		if not t:
			raise Exception("Invalid time spec: '%s' in '%s'" % (t, timespec))

		# if t is a shortcut, map it to it's interval in timespec_names dictionary
		if re.match(TIMENAME_SPEC, t):
			if not timespec_names.has_key(t):
				raise Exception("Invalid time name: %s. Valid timenames: %s" % (t, ','.join(timespec_names.keys())))
			t = timespec_names[t]

		m = re.match(TIME_SPEC, t)

		# Remove Nones
		hspec = m.group('hspec') or ''
		wspec = m.group('wspec') or ''

		if log: log.debug("\thours = %s", hspec)
		if log: log.debug("\tdays  = %s", wspec)

		if hspec == '':
			match  = True  
		else:
			match  = False

		# [ hours [ , hours ... ] ]
		hours = re.split(',', hspec)
		for tinterv in hours:

			if log: log.debug("\ttesting: %s", tinterv)

			m = re.match( HOURS_SPEC , tinterv)
			if m  == None:
				raise Exception("Invalid hour spec: '%s' in '%s'" % (tinterv, t))

			(bh,bm, eh,em) = m.groups()
			if bh == None: bh = 0
			if bm == None: bm = 0
			if eh == None: eh = 0
			if em == None: em = 0

			intbeg = 60 * int(bh) + int(bm)
			intend = 60 * int(eh) + int(em)

			if intbeg > 1439: intbeg = 0	# 1439 is the number of minutes in 24 hours, minus one
			if intend > 1439: intend = 1439

			now_mins = 60 * hour + minutes;
			if intbeg <= intend:
				if ((intbeg <= now_mins) and (now_mins <= intend)):
					match = True
					break
			else:
				match = True
				break

		if match != True:
			continue

		if wspec == '':
			return True;

		#
		#       H
		#       <d>  [ - <d> ]
		#       <d>p [ - <d>p ]
		#
		wintervals = re.split(',', wspec)
		for winterv in wintervals:

			day_matched = False

			#
			# H ?
			#
			m = re.match( DAYS_H , winterv )
			if m is not None:
				if calendar_handler is not None:
					if log: log.debug("Checking calendar: %s %s %s %s", year, month, mday, time_context)
					if calendar_handler.isHoliday(year, month, mday, time_context=time_context):
						if log: log.debug("\t isHoliday")
						day_matched = True
				else:
					if log: log.debug("No calendar handler")

			else:
				#
				#  <d>  [ - <d> ]
				#  <d>p [ - <d>p ]
				#
				m = re.match( DAYS_SPEC , winterv)
				if m == None:
					#log.debug("Invalid format for day interval '%s' in '%s'. ", winterv, t)
					break

				groups = m.groups()
				if groups == None:
					#log.debug("Invalid format for day interval '%s' in '%s' ", winterv, t)
					break

				(intbeg, intend) = groups
				if intend == None:
					intend = intbeg


				#
				# Is <d>p [ - <d>p ] ?
				#
				check_pre_holiday = False

				if intbeg[-1] == 'p': 
					check_pre_holiday = True
					intbeg = intbeg[0:1]

				if intend[-1] == 'p': 
					check_pre_holiday = True
					intend = intend[0:1]

				intbeg = int(intbeg)
				intend = int(intend)

				if log: log.debug("\t pre holiday = %s", check_pre_holiday)

				if int(intbeg)  > LAST_WDAY: intbeg = FIRST_WDAY
				if int(intend)  > LAST_WDAY: intend = LAST_WDAY

				if intbeg <= intend:
					if ((intbeg <= wday) and (wday <= intend)):

						if not check_pre_holiday:
							day_matched = True
						else:
							if calendar_handler:
								if calendar_handler.isPreHoliday(year, month, mday, time_context=time_context):
									day_matched = True

			if day_matched:
				return True

	return False
