# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it> 
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""
########################################################################
Terminal utilities
########################################################################

Questo modulo contiene funzioni per interagire con terminali a caratteri (TTY) in ambienti *unix-like*.

Formattazione dell'output
=========================================

E' possibile gestire la stampa/formattazione di stringhe colorate.

Per formattare le stringhe vengono utilizzati gli `ANSI Escape Code <https://en.wikipedia.org/wiki/ANSI_escape_code>`_

Esempio:

::
	
	from sanet_common.termutils import *
	
	print "Terminal size: ", term_size()
	
	print "Hello", term_color(TERM_COLOR_RED, "world")


Colori disponibili
----------------------------

I colori utilizzabili vengono specificati attraverso le seguenti costanti.

Elenco dei colori:

	::
		
		TERM_COLOR_DEFAULT        
		
		TERM_COLOR_BLACK          
		
		TERM_COLOR_RED            
		TERM_COLOR_GREEN          
		TERM_COLOR_YELLOW         
		TERM_COLOR_BLUE           
		TERM_COLOR_MAGENTA        
		TERM_COLOR_CYAN           
		TERM_COLOR_DARK_GRAY      
		TERM_COLOR_BLACK          
		TERM_COLOR_WHITE          
		
		TERM_COLOR_LIGHT_GRAY     
		TERM_COLOR_LIGHT_RED      
		TERM_COLOR_LIGHT_GREEN    
		TERM_COLOR_LIGHT_YELLOW   
		TERM_COLOR_LIGHT_BLUE     
		TERM_COLOR_LIGHT_MAGENTA  
		TERM_COLOR_LIGHT_CYAN     


Funzioni Base
---------------------

.. autofunction:: term_color

.. autofunction:: term_bold

Utilities
---------------------

.. autofunction:: term_red

.. autofunction:: term_green

Informazioni sul terminale
==============================

.. autofunction:: term_size

"""

import sys

import fcntl, termios, struct

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

_TERM_COLOR = "\033[%sm"

TERM_COLOR_DEFAULT        = 39

TERM_COLOR_BLACK          = 30

TERM_COLOR_RED            = 31
TERM_COLOR_GREEN          = 32
TERM_COLOR_YELLOW         = 33
TERM_COLOR_BLUE           = 34
TERM_COLOR_MAGENTA        = 35
TERM_COLOR_CYAN           = 36
TERM_COLOR_DARK_GRAY      = 37
TERM_COLOR_BLACK          = 30
TERM_COLOR_LIGHT_GRAY     = 90
TERM_COLOR_LIGHT_RED      = 91
TERM_COLOR_LIGHT_GREEN    = 92
TERM_COLOR_LIGHT_YELLOW   = 93
TERM_COLOR_LIGHT_BLUE     = 94
TERM_COLOR_LIGHT_MAGENTA  = 95
TERM_COLOR_LIGHT_CYAN     = 96
TERM_COLOR_WHITE          = 97


TERM_COLOR_ENDC   = _TERM_COLOR % (0)

TERM_BOLD         = "1;36"

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

#
# Primitive Base.
#

def _term_color(color, msg):
	return msg

def _term_bold(msg):
	return msg


try:
	#
	# Se lo stdout del processo e' collegato ad un terminale TTY allora
	# le primitive di colorazione possono funzionare.
	#
	if sys.stdout.isatty():
		def _term_color(color, msg):
			return _TERM_COLOR % (color) + msg + TERM_COLOR_ENDC

		def _term_bold(msg):
			return _TERM_COLOR % (TERM_BOLD) + msg + TERM_COLOR_ENDC
except Exception, e:
	pass

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def term_color(color, msg):
	"""
	Adds to the string *msg* special terminal characters. The returned
	string can be printed on stdout and will be colored with the color
	specified by the given *color* code.
	
	:returns: A terminal colored string.
	:rtype: string
	"""
	return _term_color(color, msg)

def term_bold(msg):
	"""
	Adds to the string *msg* special terminal characters in order
	to have printable **bold** string.
	
	:returns: A terminal bolded string.
	:rtype: string
	"""
	return _term_bold(msg)

def term_red(msg):
	"""
	Restituisce una stringa "rossa" stampabile su terminale.
	
	:returns: A terminal colored string.
	:rtype: string
	"""
	return term_color(TERM_COLOR_RED, msg)

def term_green(msg):
	"""
	Restituisce una stringa "verde" stampabile su terminale.
	
	:returns: A terminal colored string.
	:rtype: string
	"""
	return term_color(TERM_COLOR_GREEN, msg)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def term_size():
	"""
	Returns the size of the current terminal.
	
	:returns: A tuple (<width>, <height>).
	:rtype: integer
	
	:exceptions:
	
		:class:`IOError` if the size of the terminal can not be calculated.
	"""
	#h, w, hp, wp = struct.unpack('HHHH', fcntl.ioctl(0, termios.TIOCGWINSZ, struct.pack('HHHH', 0, 0, 0, 0)))
	h, w, hp, wp = struct.unpack('HHHH', fcntl.ioctl(1, termios.TIOCGWINSZ, struct.pack('HHHH', 0, 0, 0, 0)))
	return w, h
