# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""
########################################################################
Time utilities
########################################################################
"""

import time, datetime

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S'

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def timestamp_to_isostr(timestamp):
	"""
	Converts the given timestamp value into a date/time ISO string.
	
	:param  timestamp:  Current execution context.
	:type   timestamp:  integer
	
	:returns: ISO stirng
	:rtype: string
	
	:exception: 
	        None
	"""

	try:
		return datetime.datetime.fromtimestamp(timestamp).strftime(DATETIME_FORMAT)
	except TypeError:
		return ''

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def utc_mktime(utc_tuple):
	"""Returns number of seconds elapsed since epoch

	Note that no timezone are taken into consideration.

	utc tuple must be: (year, month, day, hour, minute, second)

	"""
	if len(utc_tuple) == 6:
		utc_tuple += (0, 0, 0)

	return time.mktime(utc_tuple) - time.mktime((1970, 1, 1, 0, 0, 0, 0, 0, 0))

def datetime_to_timestamp(dt):
	"""
	Converts a datetime object to UTC timestamp
	"""
	
	utc_ts = utc_mktime(dt.timetuple());
	utc_ts -= time.daylight * 3600
	
	return int( utc_ts )

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

#def check_if_working_day(day):
#	"""
#	Checks whether given day is a working day
#	:param datetime day: A datetime object representing the given day
#	"""
#	if day.weekday() in range(5):
#		return True
#	else:
#		return False

def get_timestamps_interval(day, timespec):
	"""
	Returns a tuple containing two timestamps representing given day at given hours
	
	:param datetime day: A datetime object representing the given day at midnight (00:00)
	:param str timespec: A string representing the time range.
	
	"""

	# import messi qua per poi spostare tutto il codice in un altro modulo
	from sanet_common.sanetdefs.timing import HOURS_SPEC, DAYS_SPEC, TIME_SPEC
	import re
	
	ret = []
	
	times = re.split(r"\|", timespec)
	for t in times:
		
		# nelle expr i giorni della settimana sono nel range [1-7], mentre nell'oggetto datetime vengono salvate nel range [0-6]
		wday = day.weekday() + 1
		
		t = t.strip()
		
		if not t:
			raise Exception("invalid time spec: '%s', in '%s'" % (t, timespec))
		
		m = re.match(TIME_SPEC, t)
		
		hspec = m.group('hspec') or ''
		wspec = m.group('wspec') or ''
		
		wintervals = re.split(',', wspec)
		for winterv in wintervals:
			
			m = re.match(DAYS_SPEC, winterv)
			if m == None:
				raise Exception("Invalid week spec: '%s' in '%s'" % (winterv, t))
			
			(intbeg, intend) = m.groups()
			
			intbeg = int(intbeg)
			intend = int(intend)
			
			if ( (intbeg > wday) or (intend < wday) ):
				wday = None
				break
		
		if not wday:
			continue
		
		hours = re.split(',', hspec)
		for interv in hours:
			
			m = re.match(HOURS_SPEC, interv)
			if m == None:
				raise Exception("invalid hour spec: '%s' in '%s'" % (interv, t))
			
			#before hh, mm   end hh, mm
			(       bh, bm,      eh, em) = m.groups()
			bh = int(bh) if bh else 0
			bm = int(bm) if bm else 0
			eh = int(eh) if eh else 0
			em = int(em) if em else 0
			
			start = datetime.timedelta(hours=bh, minutes=bm)
			end   = datetime.timedelta(hours=eh, minutes=em)

			start = time.mktime( (day + start).timetuple() )
			end   = time.mktime( (day + end).timetuple() )

			ret.append( (start, end) )

	return ret


def get_working_timestamps_in_range(ts_start, ts_end, timespec):
	"""
	Returns a list of tuples with working days timestamps in working time range
	
	:param timestamp ts_start: A timestamp representing the starting timestamp
	:param timestamp ts_end:   A timestamp representing the ending timestamp
	:param str timespec:       A string representing the time range (syntax described in entables module
	"""

	start       = datetime.datetime.fromtimestamp(ts_start).replace(hour=0, minute=0, second=0)
	end         = datetime.datetime.fromtimestamp(ts_end)

	current_day = start

	day_adder   = datetime.timedelta(days=1)

	ret_value   = []

	while current_day <= end:

		#if check_if_working_day(current_day):
		ret_value += get_timestamps_interval(current_day, timespec)

		current_day = current_day + day_adder

	return ret_value
