# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>. 

"""

.. _sanet-libsnmp:

################################################################################
Sanet SNMP LIbrary
################################################################################

=============================================================================
Introduction
=============================================================================

This library is a wrapper to the NetSNMP's python bindings.

=============================================================================
Exceptions
=============================================================================

Hierarchy:

::
	
	SnmpWrapperException
		|
		+- SnmpWrapperInvalidHost
		|
		+- SnmpTimeout
		|
		`- SnmpWrapperBUGException


.. autoexception:: SnmpWrapperException

.. autoexception:: SnmpWrapperInvalidHost

.. autoexception:: SnmpTimeout

.. autoexception:: SnmpWrapperBUGException

=============================================================================
Classes
=============================================================================

.. autoclass:: SnmpSession

=============================================================================
Functions
=============================================================================

Session API
------------

.. autofunction:: snmp_create_sess

.. autofunction:: snmp_get_s

.. autofunction:: snmp_walk_s

.. autofunction:: snmp_bulkwalk_s

Simple API
-----------

These functions are high-level version of the previous functions.

.. autofunction:: snmp_get

.. autofunction:: snmp_walk

.. autofunction:: snmp_bulkwalk

"""

__copyright__  = "Copyright 2018, Laboratori Guglielmo Marconi S.p.A."
__version__    = "1.0.0"
__author__     = "Diego Billi"
__credits__    = ["Michele Bergonzoni", "Luca Ferroni"]
__license__    = "AGPL v3"
__maintainer__ = "Diego Billi"
__email__      = "nad@labs.it"
__status__     = "Production"

import os
import time

import netsnmp

# HACK: user SNMP BULK for normal WALK. FOR DEBUG ONLY.
_SANET_ENV_FORCE_BULK_WALK = os.environ.get('SANET_BULK_WALK', None)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

#: Default SNMP protocol version used for requests (1).
SNMP_DEFAULT_VERSION   = 1

#: Default SNMP port number used for requests (161).
SNMP_DEFAULT_PORT      = 161

#: Default SNMP community value used for requests (161).
SNMP_DEFAULT_COMMUNITY = 'public'

#: Default timeout for a single request (0.5 seconds).
SNMP_DEFAULT_TIMEOUT   = 0.5       # seconds

#: Default retries for a single request (0).
SNMP_DEFAULT_RETRIES   = 0

SNMP_ONE_SECOND        = 1000000   # micro-seconds

#: Default SNMP BULK request size.
SNMP_DEFAULT_BULK_SIZE = 10        # IMPORTANT: tests show that a bigger bulk size do not work with some SMTP Agents, causing timeouts in requests.

#-----------------------------------------------------------------------------#
# Exceptions                                                                  #
#-----------------------------------------------------------------------------#

class SnmpWrapperException(Exception):
	"""
	Generic exception raised by module's functions when an error occurs.
	This is also the base class for every exception of this module.
	"""
	pass

class SnmpWrapperInvalidHost(SnmpWrapperException):
	"""
	This exception is raised when the SNMP functions can not open
	a SNMP session with the target host. A description of the problem is
	stored inside *msg*.
	The most probable reason is that the name of the target host can not 
	be resolved in a valid IP address.
	"""
	def __init__(self, msg):
		self.msg = msg

	def __str__(self):
		return "Invalid host: %s" % (self.msg)

class SnmpTimeout(SnmpWrapperException):
	"""
	Raised when the Net-SNMP layer returns a timeout error on a GET or
	WALK operation.
	"""
	def __init__(self, msg):
		self.msg = msg

	def __str__(self):
		return self.msg

#
# Special case: the snmp library is bugged and we want to be sure these case will
#               produce an uncheckable result.
#
# HISTORY: Added for ticket #1885.
#
class SnmpWrapperBUGException(SnmpWrapperException):
	"""
	Python's netsnmp wrapper module behaves strangely are returns strange data .
	We raise this type of exception when we detect this cases.
	"""
	pass

#-----------------------------------------------------------------------------#
# Classes                                                                     #
#-----------------------------------------------------------------------------#

class SnmpSession(netsnmp.Session):
	"""
	Stores informations about a SNMP session.
	"""
	def __init__(self, **args):
		super(SnmpSession, self).__init__(**args)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#
#
# FIX: mar dic 22 10:39:30 CET 2009  
# NET-SNMP v5.5 doesn't set some deafult values correctly and
# this couse segmentations faults in the python's bindings.
#	
v3_defaults = {
	 'SecLevel' : 'noAuthNoPriv'
	,'SecName'  : ''
	,'AuthProto': 'MD5'
	,'AuthPass' : ''
}

def __fix_netsnmp_v3_defaults(options):

	if options['Version'] == 3 or options['Version'] == 63:
		for key in v3_defaults:
			if not options.has_key(key):
				options[key] = v3_defaults[key]

	return options

#-----------------------------------------------------------------------------#
# SESSION                                                                     #
#-----------------------------------------------------------------------------#

def snmp_create_sess(host, 
                     port=SNMP_DEFAULT_PORT, 
                     community=SNMP_DEFAULT_COMMUNITY, 
                     timeout=SNMP_DEFAULT_TIMEOUT,
                     retries=SNMP_DEFAULT_RETRIES,
                     **options):
	"""
	Create a returns an :class:`SnmpSession` object.

	:param host: SNMP agent 
	:type host: string
	
	:param port: SNMP port  (Default: :const:`SNMP_DEFAULT_VERSION`. See: :ref:`snmp-wrapper-constants`)
	:type port: integer
	
	:param community: SNMP community  (Default: :const:`SNMP_DEFAULT_COMMUNITY`. See: :ref:`snmp-wrapper-constants`)
	:type community: string
	
	:param timeout: max timeout for SNMP response (Default: :const:`SNMP_DEFAULT_TIMEOUT`. See: :ref:`snmp-wrapper-constants`)
	:type timeout: number
	
	:param retries: (integer) number of retries if no response is received (Default: :const:`SNMP_DEFAULT_RETRIES`. See: :ref:`snmp-wrapper-constants`)
	
	:param options: other optional options
	
	:returns: the SnmpSession objects.
	          
	:rtype: single value, variable type
	
	:exception: 
	        :class:`SnmpWrapperException`
	
	        :class:`SnmpWrapperInvalidHost`
	        
	        See Netsnmp for more exceptions
	"""

	# FIXED: Documentation (README) in package sources is wrong! Net-snmp's python 
	#       bindings don't use 'RemotePort' anymore. Put "<host>:<port>" in DestHost!
	#       2009-07-22
	#options['DestHost']   = host
	#options['RemotePort'] = port
	options['DestHost'] = "%s:%s" % (host,port)

	options['Community']  = community
	options['Timeout']    = int(timeout * SNMP_ONE_SECOND)
	options['Retries']    = retries

	if not options.has_key('Version'):
		options['Version'] = SNMP_DEFAULT_VERSION

	if options['Version'] > 60:
		options['Version'] = options['Version'] - 60
		options['DestHost'] = 'udp6:' + options['DestHost']

	if options['Version'] == 0:
		return None

	options['UseNumeric'] = 1	# don not translate oids in return values!

	options = __fix_netsnmp_v3_defaults(options)

	sess = SnmpSession(**options)
	if not sess.sess_ptr:
		raise SnmpWrapperInvalidHost("Unable to open snmp session with: %s" % options['DestHost'])

	return sess

#-----------------------------------------------------------------------------#
# These functions replace netsnmp.snmpget() and netsnmp.snmpwalk()            #
#-----------------------------------------------------------------------------#

def __my_snmpget_s(sess, *args, **kargs):

	var_list = netsnmp.VarList()
	for arg in args:
		if isinstance(arg, netsnmp.client.Varbind):
			var_list.append(arg)
		else:
			var_list.append(netsnmp.Varbind(arg))

	res = sess.get(var_list)

	return (res, sess.ErrorStr)

#-----------------------------------------------------------------------------#




def __my_snmpwalk_s(sess, *args, **kargs):

	if isinstance(args[0], netsnmp.client.VarList):
		var_list = args[0]
	else:
		var_list = netsnmp.VarList()

		for arg in args:
			if isinstance(arg, netsnmp.client.Varbind):
				var_list.append(arg)
			else:
				var_list.append(netsnmp.Varbind(arg))

	res = sess.walk(var_list)

	#print "__my_snmpwalk res=%s (%s)  ERROR str(%s) num(%s) ind(%s)" % (res, type(res), sess.ErrorStr, sess.ErrorNum, sess.ErrorInd)

	return (res, sess.ErrorStr)



def __my_snmpgetnext_s(sess, *args, **kargs):

	var_list = netsnmp.VarList()
	for arg in args:
		if isinstance(arg, netsnmp.client.Varbind):
			var_list.append(arg)
		else:
			var_list.append(netsnmp.Varbind(arg))

	res = sess.getnext(var_list)

	return (res, sess.ErrorStr)


#def __my_snmpget(*args, **kargs):
#
#	sess = SnmpSession(**kargs)
#	if not sess.sess_ptr:
#		raise SnmpWrapperInvalidHost("Unable to open snmp session with: %s" % kargs['DestHost'])
#
#	return __my_snmpget_s(sess, *args, **kargs)

#def __my_snmpwalk(*args, **kargs):
#
#	sess = SnmpSession(**kargs)
#	if not sess.sess_ptr:
#		raise SnmpWrapperInvalidHost("Unable to open snmp session with: %s" % kargs['DestHost'])

#def __my_snmpgetnext(*args, **kargs):
#
#	sess = SnmpSession(**kargs)
#	if not sess.sess_ptr:
#		raise SnmpWrapperInvalidHost("Unable to open snmp session with: %s" % kargs['DestHost'])
#
#	return __my_snmpgetnext_s(sess, *args, **kargs)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def snmp_get_s(sess, oid):
	"""
	Returns the SNMP value for the OID *oid* from SNMP agent *host*.
	
	:param sess:  Snmp session objects 
	:type sess: SnmpSession
	
	:param oid:  OID 
	:type oid: string
	
	:returns: the SNMP value. Returns **None** if the value can't be retrieved (no such mib) or
	          raise a SnmpTimeout exception.
	          
	:rtype: single value, variable type
	
	:exception: 
	        :class:`SnmpWrapperException`
	
	        See Netsnmp for more exceptions
	"""

	if oid[0].isdigit():
		oid = '.' + oid

	oidvar = netsnmp.Varbind(oid) 

	#
	# GET
	#
	(res, err) = __my_snmpget_s(sess, oidvar) 

	#print __file__, "RES", res
	#print __file__, "ERR", err

	if res == None:
		raise SnmpWrapperBUGException("netsnmp returned %r. Bad OID (%s) or BUG?" % (res, oid) )

	if len(res) != 1:
		raise SnmpWrapperBUGException("netsnmp returned %r. BUG?" % res)

	#
	# WARNING: Sometime we read  ( res = ('',) , err = '(noSuchName) There is no such variable name in this MIB' )
	#

	if res[0] != None and not err:
		if oidvar.type in ("INTEGER","INTEGER32", "COUNTER","COUNTER64","GAUGE","UINTEGER","TICKS"):
			val = int(oidvar.val)
		else:
			val = oidvar.val

		#del(oidvar)

		return val
	else:
		del(oidvar)

		if err != None:
			err = err.lower()

			# This should work for retro-compatibility when
			# NET-SNMP didn't return an error message for no suchname
			if err == '':
				return None

			if err.lower().find('nosuchname') > -1:
				return None

			if err == 'timeout':
				raise SnmpTimeout(err)

			raise SnmpWrapperException("low level error: %s" % (err))

		raise SnmpWrapperException("unknown snmp error")

		# No values for that OID! Maybe a WALK should be done instead!
		#return None

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#
#
#   Varbind.type = 
#
#    OBJECTID   => dotted-decimal (e.g., .1.3.6.1.2.1.1.1)
#    OCTETSTR   => perl scalar containing octets,
#    NETADDR    => dotted-decimal,
#    IPADDR     => dotted-decimal,
#    OPAQUE     => perl scalar containing octets,
#    NULL       => perl scalar containing nothing,

#    INTEGER    => decimal signed integer (or enum),
#    COUNTER    => decimal unsigned integer,
#    COUNTER64  => decimal unsigned integer,
#    GAUGE      => decimal unsigned integer,
#    UINTEGER   => decimal unsigned integer,
#    TICKS      => decimal unsigned integer,

def __do_walk_s(sess, oid):

	# netsnmp wants fully qualified, dotted-decimal, numeric OID 
	if oid[0].isdigit():
		oid2 = '.' + oid
	else:
		oid2 = oid

	varlist = netsnmp.VarList( netsnmp.Varbind(oid2)  )

	(values, err) = __my_snmpwalk_s(sess, varlist) 

	if values == None:
		# Probably the OID is invalid!
		raise SnmpWrapperBUGException("netsnmp returned %r (%s) BUG?" % (res, type(res)) )

	if len(values) == 0:
		if err != None:
			err = err.lower()

			if err == '': 
				pass
			elif err == 'timeout':
				raise SnmpTimeout(err)
			else:
				raise SnmpWrapperException("low level error: %s" % (err))
		else:
			raise SnmpWrapperException("unknown snmp error")

	return (varlist, values)


def snmp_walk_s(sess, oid):
	"""
	Returns an array of tuples (oid, value) for the base OID *oid* from SNMP agent *host*.
	
	:param sess:  Snmp session objects 
	:type sess: SnmpSession
	
	:param oid:  OID 
	:type oid: string
	
	:returns: the SNMP value. Returns **None** if the value can't be retrieved (no such mib) or
	          raise a SnmpTimeout exception.
	          
	:rtype: single value, variable type
	
	:exception: 
	        :class:`SnmpWrapperException`
	
	        :class:`SnmpWrapperInvalidHost`
	        
	        See Netsnmp for more exceptions
	"""

	if _SANET_ENV_FORCE_BULK_WALK:
		v = sess.Version
		if (v is None or v >= 2):
			#print __file__, "BULK"
			#t1 = time.time()
			r = snmp_bulkwalk_s(sess, oid)
			#t2 = time.time()
			#print __file__, "SNMP WALK BULK", t2 - t1, len(r) if r is not None else None
			return r


	res = None

	(varlist, values) = __do_walk_s(sess, oid)

	if values:
		res = []
		for i in xrange(len(varlist)):

			if varlist[i].tag == None:
				# FIX: THIS IS A STRANGE BUG: sometimes the walk returns a list
				#      of tuples (<None> , 1).
				#raise SnmpWrapperBUGException("NET-SNMP WALK BUG for: %s:%s:%s:%s" % (oid, host, port, community))
				raise SnmpWrapperBUGException("NET-SNMP WALK BUG for: %s:%s:%s:%s" % (oid, sess.DestHost, sess.LocalPort, sess.Community))

			key = varlist[i].tag + '.' + str(varlist[i].iid)

			# convert integer values in INT values
			if varlist[i].type in ("INTEGER", "INTEGER32", "COUNTER","COUNTER64","GAUGE","UINTEGER","TICKS"):
				val = int(varlist[i].val)
			else:
				val = varlist[i].val

			res.append( (key, val) )

	return res

def snmp_bulkwalk_s(snmp_sess, 
                    oid, 
                    bulk_size = SNMP_DEFAULT_BULK_SIZE,
                   ):
	"""
	This is the "bulk" version of :function:`snmp_bulkwalk_s`.
	"""
	ret = []

	#
	# STARTING OID
	#
	# netsnmp wants fully qualified, dotted-decimal, numeric OID 
	if oid[0].isdigit():
		oid = '.' + oid

	def _bulk_walk(sess, start_oid):

		tot_vals = []

		current_oid = start_oid
		stop = False
		while not stop:

			vars = netsnmp.VarList(netsnmp.Varbind(current_oid))

			values = sess.getbulk(0, bulk_size, vars)
			err    = sess.ErrorStr

			#print "DEBUG:", len(vars)    
			#print "DEBUG:", len(values)  if values is not None else None
			#print "DEBUG:", repr(vars[0].tag), repr(vars[0].iid)
			#print "DEBUG:", err
			#print "ERROR:", err, vars[0].tag, vars[0].iid

			if values is None:
				if err is not None:
					break
				else:
					raise Exception("SNMP BUG?")

			if len(values) == 0:
				if err != None:
					err = err.lower()

					if err == '': 
						pass
					elif err == 'timeout':
						raise SnmpTimeout(err)
					else:
						raise SnmpWrapperException("low level error: %s" % (err))
				else:
					raise SnmpWrapperException("unknown snmp error")

			for i, var in enumerate(vars):

				oid = var.tag + '.' + var.iid

				if not oid.startswith(start_oid):
					stop = True
					break

				# convert integer values in INT values
				if var.type in ("INTEGER", "INTEGER32", "COUNTER","COUNTER64","GAUGE","UINTEGER","TICKS"):
					val = int(var.val)
				else:
					val = var.val

				tot_vals.append( (oid, val) )

			current_oid = oid

		return tot_vals

	ret = _bulk_walk(snmp_sess, oid)

	if len(ret) == 0:
		return None

	return ret

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def snmp_get(oid, host, port=SNMP_DEFAULT_PORT, 
                        community=SNMP_DEFAULT_COMMUNITY, 
                        timeout=SNMP_DEFAULT_TIMEOUT,
                        retries=SNMP_DEFAULT_RETRIES,
                        **options):
	"""
	Returns the SNMP value for the OID *oid* from SNMP agent *host*.
	
	:param oid:  OID 
	:type oid: string
	
	:param host: SNMP agent 
	:type host: string
	
	:param port: SNMP port  (Default: :const:`SNMP_DEFAULT_VERSION`. See: :ref:`snmp-wrapper-constants`)
	:type port: integer
	
	:param community: SNMP community  (Default: :const:`SNMP_DEFAULT_COMMUNITY`. See: :ref:`snmp-wrapper-constants`)
	:type community: string
	
	:param timeout: max timeout for SNMP response (Default: :const:`SNMP_DEFAULT_TIMEOUT`. See: :ref:`snmp-wrapper-constants`)
	:type timeout: number
	
	:param retries: (integer) number of retries if no response is received (Default: :const:`SNMP_DEFAULT_RETRIES`. See: :ref:`snmp-wrapper-constants`)
	
	:param options: other optional options
	
	:returns: the SNMP value. Returns **None** if the value can't be retrieved (no such mib) or
	          raise a SnmpTimeout exception.
	          
	:rtype: single value, variable type
	
	:exception: 
	        :class:`SnmpWrapperException`
	
	        :class:`SnmpWrapperInvalidHost`
	        
	        See Netsnmp for more exceptions
	"""

	sess = snmp_create_sess(
		host, 
		port=port, 
		community=community, 
		timeout=timeout,
		retries=retries,
		**options
	)

	return snmp_get_s(sess, oid)

def snmp_walk(oid, host, port=SNMP_DEFAULT_PORT, 
                         community=SNMP_DEFAULT_COMMUNITY, 
                         timeout=SNMP_DEFAULT_TIMEOUT,
                         retries=SNMP_DEFAULT_RETRIES,
                         **options):
	"""
	Returns an array of tuples (oid, value) for the base OID *oid* from SNMP agent *host*.
	
	:param oid:  OID 
	:type oid: string
	
	:param host: SNMP agent 
	:type host: string
	
	:param port: SNMP port  (Default: :const:`SNMP_DEFAULT_VERSION`. See: :ref:`snmp-wrapper-constants`)
	:type port: integer
	
	:param community: SNMP community  (Default: :const:`SNMP_DEFAULT_COMMUNITY`. See: :ref:`snmp-wrapper-constants`)
	:type community: string
	
	:param timeout: max timeout for SNMP response (Default: :const:`SNMP_DEFAULT_TIMEOUT`. See: :ref:`snmp-wrapper-constants`)
	:type timeout: number
	
	:param retries: (integer) number of retries if no response is received (Default: :const:`SNMP_DEFAULT_RETRIES`. See: :ref:`snmp-wrapper-constants`)
	
	:param options: other optional options
	
	:returns: the SNMP value. Returns **None** if the value can't be retrieved (no such mib) or
	          raise a SnmpTimeout exception.
	          
	:rtype: single value, variable type
	
	:exception: 
	        :class:`SnmpWrapperException`
	
	        :class:`SnmpWrapperInvalidHost`
	        
	        See Netsnmp for more exceptions
	"""

	sess = snmp_create_sess(
		host, 
		port=port, 
		community=community, 
		timeout=timeout,
		retries=retries,
		**options
	)

	return snmp_walk_s(sess, oid)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def test_snmp_lib_version():
	"""
	This function tests if the installed version of NET-SNMP is patched
	with this:
	
		Description: [python] Segmentation fault on SNMP error or timeout - ID: 2952766
		URL: http://sourceforge.net/tracker/?func=detail&aid=2952766&group_id=12694&atid=112694
	"""

	host    = "google.com"
	timeout = 1   # one second
	retries = 0
	snmpversion = 1

	valid = False

	print "Checking SNMP library..."

	try:
		# Force a timeout with a fake community. A SNMPv1 server should not
		# responde to a request with unknown community.
		val = snmp_get('1.3.6.1.2.1.1.1.0', '127.0.0.1', 161, 'deadbeef', timeout, retries, Version=snmpversion)
	except SnmpTimeout, e:
		print "Timeout: ", e
		print "Timeout error handling seems to be correct"
		valid = True
	except SnmpWrapperException, e:
		print "ERROR: ", e
		print "The NET-SNMP python module doesn't work correct"

	if not valid:
		print "*********************************************"
		print "*********************************************"
		print "*********************************************"
		print "*********************************************"
		print "*** YOU MUST UPDATE YOUR NET-SNMP LIBRARY ***"
		print "*********************************************"
		print "*********************************************"
		print "*********************************************"
		print "*********************************************"
		exit(-1)








#-----------------------------------------------------------------------------#
#-----------------------------------------------------------------------------#
#-----------------------------------------------------------------------------#
#-----------------------------------------------------------------------------#
#-----------------------------------------------------------------------------#
#-----------------------------------------------------------------------------#
#-----------------------------------------------------------------------------#
#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#
#-----------------------------------------------------------------------------#
#-----------------------------------------------------------------------------#
#-----------------------------------------------------------------------------#
#-----------------------------------------------------------------------------#
#-----------------------------------------------------------------------------#
#-----------------------------------------------------------------------------#
#-----------------------------------------------------------------------------#

"""
def snmp_walk_dict(oid, host, port=SNMP_DEFAULT_PORT, 
                         community=SNMP_DEFAULT_COMMUNITY, 
                         timeout=SNMP_DEFAULT_TIMEOUT,
                         retries=SNMP_DEFAULT_RETRIES,
                         **options):
	
	###
	Return.
	
	:param oid:  OID 
	:type oid: string
	
	:param host: SNMP agent 
	:type host: string
	
	:param port: SNMP port  (Default: :const:`SNMP_DEFAULT_VERSION`. See: :ref:`snmp-wrapper-constants`)
	:type port: integer
	
	:param community: SNMP community  (Default: :const:`SNMP_DEFAULT_COMMUNITY`. See: :ref:`snmp-wrapper-constants`)
	:type community: string
	
	:param timeout: max timeout for SNMP response (Default: :const:`SNMP_DEFAULT_TIMEOUT`. See: :ref:`snmp-wrapper-constants`)
	:type timeout: number
	
	:param retries: (integer) number of retries if no response is received (Default: :const:`SNMP_DEFAULT_RETRIES`. See: :ref:`snmp-wrapper-constants`)
	
	:param options: other optional options
	
	:returns: a list of tuple ( <oid string>, <oid value> ). Returns *None* if no response
	          is received before the given timeout.
	:rtype: list.

	:exception: 
	        :class:`SnmpWrapperException`
		
	        :class:`SnmpWrapperInvalidHost`
		
		See Netsnmp for more exceptions
	###
	
	res = None
	
	(varlist, values) = __do_walk(oid,host,port,community,timeout,retries, **options)
	
	if values != None:
		# Now 'varlist' contanis all the oids
		res = {}
		for i in xrange(len(varlist)):
			key = varlist[i].tag + '.' + str(varlist[i].iid)

			# convert integer values in INT values
			if varlist[i].type in ("INTEGER", "COUNTER","COUNTER64","GAUGE","UINTEGER","TICKS"):
				val = int(varlist[i].val)
			else:
				val = varlist[i].val

			res[ key ] = val
	
	return res
"""

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

'''
def __do_getnext(oid, **options):

	oidvar = netsnmp.Varbind(oid) 

	# The a tuple with the value of 'oid'
	(res, err) = __my_snmpgetnext(oidvar, **options) 

	if res == None:
		raise SnmpWrapperBUGException("netsnmp returned %r. Bad OID (%s) or BUG?" % (res, oid) )

	if len(res) != 1:
		raise SnmpWrapperBUGException("netsnmp returned %r. BUG?" % res)

	if res[0] != None:

		res_oid = oidvar.tag+'.'+str(oidvar.iid)

		#print "DEBUG _do_getnext IID  =" , oidvar.iid
		#print "DEBUG _do_getnext TAG  =" , oidvar.tag
		#print "DEBUG _do_getnext TYPE =", oidvar.type
		#print "DEBUG _do_getnext VAL  =", oidvar.val
		#print "DEBUG _do_getnext", type(res[0])
		if oidvar.type in ("INTEGER","INTEGER32", "COUNTER","COUNTER64","GAUGE","UINTEGER","TICKS"):
			val = int(oidvar.val)
		else:
			val = oidvar.val

		return (res_oid, val)
	else:
		del(oidvar)

		if err != None:
			err = err.lower()

			# This should work for retro-compatibility when
			# NET-SNMP didn't return an error message for no suchname
			if err == '':
				return None

			if err.find('nosuchname') > -1:
				return None

			if err == 'timeout':
				raise SnmpTimeout(err)

		raise SnmpWrapperException("unknown snmp error")

		# No values for that OID! Maybe a WALK should be done instead!
		#return None

def snmp_getnext(oid, host, port=SNMP_DEFAULT_PORT, 
                        community=SNMP_DEFAULT_COMMUNITY, 
                        timeout=SNMP_DEFAULT_TIMEOUT,
                        retries=SNMP_DEFAULT_RETRIES,
                        **options):
	"""
	Returns the SNMP value for the OID *oid* from SNMP agent *host*.
	
	:param oid:  OID 
	:type oid: string
	
	:param host: SNMP agent 
	:type host: string
	
	:param port: SNMP port  (Default: :const:`SNMP_DEFAULT_VERSION`. See: :ref:`snmp-wrapper-constants`)
	:type port: integer
	
	:param community: SNMP community  (Default: :const:`SNMP_DEFAULT_COMMUNITY`. See: :ref:`snmp-wrapper-constants`)
	:type community: string
	
	:param timeout: max timeout for SNMP response (Default: :const:`SNMP_DEFAULT_TIMEOUT`. See: :ref:`snmp-wrapper-constants`)
	:type timeout: number
	
	:param retries: (integer) number of retries if no response is received (Default: :const:`SNMP_DEFAULT_RETRIES`. See: :ref:`snmp-wrapper-constants`)
	
	:param options: other optional options
	
	:returns: the SNMP value. Returns **None** if the value can't be retrieved (no such mib) or
	          raise a SnmpTimeout exception.
	:rtype: single value, variable type
	
	:exception: 
	        :class:`SnmpWrapperException`
	        
	        :class:`SnmpWrapperInvalidHost`
	        
	        See Netsnmp for more exceptions
	"""

	# FIXED: Documentation (README) in package sources is wrong! Net-snmp's python 
	#       bindings don't use 'RemotePort' anymore. Put "<host>:<port>" in DestHost!
	#       2009-07-22
	#options['DestHost']   = host
	#options['RemotePort'] = port	
	options['DestHost'] = "%s:%s" % (host,port)

	options['Community']  = community
	options['Timeout']    = int(timeout * SNMP_ONE_SECOND)
	options['Retries']    = retries

	if not options.has_key('Version'):
		options['Version'] = SNMP_DEFAULT_VERSION

	if options['Version'] > 60:
		options['Version'] = options['Version'] - 60
		options['DestHost'] = 'udp6:' + options['DestHost']

	if options['Version'] == 0:
		return None

	options['UseNumeric'] = 1 # don not translate oids in return values!

	options = __fix_netsnmp_v3_defaults(options)

	# netsnmp wants fully qualified, dotted-decimal, numeric OID 
	if oid[0].isdigit():
		oid = '.' + oid

	return __do_getnext(oid, **options)
'''

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

'''
def snmp_smartwalk(oid, host, port=SNMP_DEFAULT_PORT, 
                        community=SNMP_DEFAULT_COMMUNITY, 
                        timeout=SNMP_DEFAULT_TIMEOUT,
                        retries=SNMP_DEFAULT_RETRIES,
                        stop_callback=None,
                        **options):
                        
	ret = []

	#
	# INIT OPTIONS
	#
	options['DestHost'] = "%s:%s" % (host,port)

	options['Community']  = community
	options['Timeout']    = int(timeout * SNMP_ONE_SECOND)
	options['Retries']    = retries

	if not options.has_key('Version'):
		options['Version'] = SNMP_DEFAULT_VERSION

	if options['Version'] > 60:
		options['Version'] = options['Version'] - 60
		options['DestHost'] = 'udp6:' + options['DestHost']

	if options['Version'] == 0:
		return None

	options['UseNumeric'] = 1    # don not translate oids in return values!

	options = __fix_netsnmp_v3_defaults(options)

	#
	# STARTING OID
	#
	# netsnmp wants fully qualified, dotted-decimal, numeric OID 
	if oid[0].isdigit():
		oid = '.' + oid

	base_oid = oid

	while oid.startswith(base_oid) == True :

		oidvar = netsnmp.Varbind(oid) 
		data = __do_getnext(oid, **options)

		if data == None:
			return None

		(oid, res) = data

		if oid.startswith(base_oid) == False:
			break

		if stop_callback:
			if stop_callback(oid, res) == True:
				break

		ret.append( (oid, res) )

	return ret
'''

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

'''
def snmp_bulkwalk(oid, 
                  host, 
                  port      = SNMP_DEFAULT_PORT, 
                  community = SNMP_DEFAULT_COMMUNITY, 
                  timeout   = SNMP_DEFAULT_TIMEOUT,
                  retries   = SNMP_DEFAULT_RETRIES,
                  bulk_size = SNMP_DEFAULT_BULK_SIZE,
                  **options
                 ):
	"""
	TO DO
	"""

	ret = []

	#
	# INIT OPTIONS
	#
	options['DestHost'] = "%s:%s" % (host,port)

	options['Community']  = community
	options['Timeout']    = int(timeout * SNMP_ONE_SECOND)
	options['Retries']    = retries

	if not options.has_key('Version'):
		options['Version'] = SNMP_DEFAULT_VERSION

	if options['Version'] > 60:
		options['Version'] = options['Version'] - 60
		options['DestHost'] = 'udp6:' + options['DestHost']

	if options['Version'] == 0:
		return None

	options['UseNumeric'] = 1    # don not translate oids in return values!

	options = __fix_netsnmp_v3_defaults(options)

	#
	# STARTING OID
	#
	# netsnmp wants fully qualified, dotted-decimal, numeric OID 
	if oid[0].isdigit():
		oid = '.' + oid


	snmp_sess = netsnmp.Session(**options)
	if not snmp_sess.sess_ptr:
		raise SnmpWrapperInvalidHost("Unable to open snmp session with: %s" % (options['DestHost']))


	def _bulk_walk(sess, start_oid):
	    
		tot_vals = []

		current_oid = start_oid
		stop = False
		while not stop:

			vars = netsnmp.VarList(netsnmp.Varbind(current_oid))

			values = sess.getbulk(0, bulk_size, vars)
			err    = sess.ErrorStr

			#print "DEBUG:", len(vars)    
			#print "DEBUG:", len(values)  if values is not None else None
			#print "DEBUG:", repr(vars[0].tag), repr(vars[0].iid)
			#print "DEBUG:", err
			#print "ERROR:", err, vars[0].tag, vars[0].iid

			if values is None:
				if err is not None:
					break
				else:
					raise Exception("SNMP BUG?")

			if len(values) == 0:
				if err != None:
					err = err.lower()

					if err == '': 
						pass
					elif err == 'timeout':
						raise SnmpTimeout(err)
					else:
						raise SnmpWrapperException("low level error: %s" % (err))
				else:
					raise SnmpWrapperException("unknown snmp error")


			for i, var in enumerate(vars):

				oid = var.tag + '.' + var.iid

				if not oid.startswith(start_oid):
					stop = True
					break

				# convert integer values in INT values
				if var.type in ("INTEGER", "INTEGER32", "COUNTER","COUNTER64","GAUGE","UINTEGER","TICKS"):
					val = int(var.val)
				else:
					val = var.val

				tot_vals.append( (oid, val) )

			current_oid = oid

		return tot_vals

	ret = _bulk_walk(snmp_sess, oid)

	if len(ret) == 0:
		return None

	return ret
'''


