# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.
"""

################################################################################
Agent (modulo principale)
################################################################################

.. contents:: Indice
   :backlinks: none
   :depth: 4

================================================================================
Strutttura e funzionamento dell'agente
================================================================================

Inizializzazione
--------------------------------------------------------------------------------

TODO

Esecuzione
--------------------------------------------------------------------------------

TODO


Verifica dello stato funzionamento interno e moduli caricati dinamicamente
--------------------------------------------------------------------------------

TODO


.. _sanet-poller.agent.configuration_load:

Caricamento delle configurazioni
--------------------------------------------------------------------------------


TODO




.. _sanet-poller.agent.update_data:

Invio dati di monitoraggio al server
--------------------------------------------------------------------------------


L'agente invia i dati di monitoraggio al server costantemente in base alla frequenza dei controlli.

I dati vengono codificati/serializzati in due formati:

	* Stringa JSON
		* Questo e' il formato preferibile quando i dati da inviare non contengono dati "binari".
		
	* Stringa BASE64
		* Quando i dati contengono dati binari non convertibili in JSON viene usato questo formato.
		* i dati vengono prima serializzati tramite libreria Pickle e poi convertiti in BASE64.

Vedere i metodi:

	* :func:`~sanet_poller.agent.BaseAgent.encode_update_events`
	* :func:`~sanet_poller.agent.BaseAgent.decode_update_events`




.. _sanet-poller.agent.commands_and_requests:

Gestione richieste/comandi
--------------------------------------------------------------------------------

TODO


================================================================================
Dettagli modulo
================================================================================

.. autoclass:: BaseAgent
   :members: decode_update_events, encode_update_events

"""
import os
import time
import logging
import threading
import json
import traceback
import socket
import cPickle as pickle

import Queue

from sanet_common.uuids import create_global_uuid

from sanet_common.osutils.osutils import get_proc_memory, get_proc_resident

from sanet_common.ipc import IPCConnection
from sanet_common.ipc import IPCNotFound, IPCConnectionError, IPCException, IPCException
from sanet_common.ipc import RESPONSE_SUCCESS, RESPONSE_ERROR
from sanet_common.ipc import encode_response, decode_response, encode_request, decode_request, encode_request_with_id, encode_response_with_id

from sanet_common.ipc import AbstractIPCFactory


from sanet_common.sanetdefs.ipc import AGENT_TEMP_REPLY_QUEUE, AGENT_CTL_QUEUE


from sanet_poller.lib.exceptions import *

from sanet_poller.lib.threadutils import AgentThread

from sanet_poller.lib.threadutils import get_current_thread, PollThreadPool, PollWorkerThread
from sanet_poller.lib.threadutils import PollScheduler

from sanet_poller.lib.classes import ELEMENT_KIND_DEVICE, ELEMENT_KIND_IFACE, ELEMENT_KIND_NODE, ELEMENT_KIND_SERVICE
from sanet_poller.lib.classes import ELEMENT_KIND_STORAGE, PollerConfiguration, PDevice, PNode, PService, PStorage, PCluster
from sanet_poller.lib.classes import PIface, PDataGroup, PDataSource, PLink, PCondition
from sanet_poller.lib.classes import FakeDataGroupChecker as DataGroupChecker

from sanet_poller.lib.classes import POLLSTATE_IDLE, POLLSTATE_SCHEDULED, POLLSTATE_QUEUED, POLLSTATE_CHECKING, POLLSTATE_COMMITTING, POLLSTATE_STR

from sanet_poller.lib.communication import AgentUpdateEvent

#from sanet_poller.lib.functions import load_system_expr_functions
from sanet_poller.lib.functions import load_system_expr_and_agentstatus_functions

from sanet_poller.lib.utils import get_time

from sanet_poller.agentctl import LOCAL_TCP_ADDRESS, local_tcp_port_from_agent_name

if 'FAKE_CHECKER' in os.environ:
	print 'using fake checker'
	from sanet_poller.lib.classes import FakeDataGroupChecker as EffectiveDataGroupChecker
else:
	from sanet_poller.lib.classes import DataGroupChecker     as EffectiveDataGroupChecker

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

DEFAULT_LOAD_SETTINGS_TIMEOUT = 10  # seconds        

DEFAULT_LOAD_SETTINGS_DELAY   = 5  # seconds

DEFAULT_CONFIG_UPDATE_MINPERIOD = 180

DEFAULT_NUM_PENDING_EVENTS = 30

DEFAULT_CONFIG_UPDATE_TOLERANCE = 2   # seconds

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

import zlib
import base64
import cPickle

def COMPRESS_DATA(data):
	data = base64.b64encode(zlib.compress(cPickle.dumps(data, cPickle.HIGHEST_PROTOCOL)))
	return data

def DECOMPRESS_DATA(data):
	data = cPickle.loads(zlib.decompress(base64.b64decode(data)))
	return data

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class RemoteServerException(Exception):
	"""
	Exception raised when the remote server responds with an error 
	message
	"""
	pass

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

CONFIG_UPDATE_STATUS_NONE    = 0
CONFIG_UPDATE_STATUS_LOADING = 1
CONFIG_UPDATE_STATUS_DONE    = 2
CONFIG_UPDATE_STATUS_ERR     = 3

CONFIG_UPDATE_STATUS_STR = {
	CONFIG_UPDATE_STATUS_NONE    : 'NONE',
	CONFIG_UPDATE_STATUS_LOADING : 'LOADING',
	CONFIG_UPDATE_STATUS_DONE    : 'DONE',
	CONFIG_UPDATE_STATUS_ERR     : 'ERR',
}

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

#
# State of the cached entry
#
CACHESTATE_INVALID   = 0
CACHESTATE_VALID     = 1


class ScheduledDatagroup(object):
	""" 
	Simple structure with informations about loaded and scheduled elements (datagroups)
	"""
	def __init__(self, pdatagroup):

		# Schedule informations
		self.sched_state  = POLLSTATE_IDLE
		self.schedule_time = 0
		self.lastdone = pdatagroup.lastdone

		# Real data
		self.pdatagroup = pdatagroup

		self.__cachestate = CACHESTATE_VALID


	def set_invalid(self):
		self.__cachestate = CACHESTATE_VALID

	def is_valid(self):
		return (self.__cachestate == CACHESTATE_VALID)

	def __str__(self):
		return "<%s:%s>" % (self.pdatagroup.uuid, self.pdatagroup.path)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class AgentWorkerThread(PollWorkerThread):

	def __init__(self, *args, **kwargs):
		super(AgentWorkerThread, self).__init__(*args, **kwargs)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class AgentIPCFactory(AbstractIPCFactory):
	"""
	Gli agenti conoscono la configurazione per comunicare col server centrale.
	
	Questa classe factory serve da wrapper per passare dall'agente per
	creare nuove connessioni.
	"""

	def __init__(self, logger, agent):
		super(AgentIPCFactory, self).__init__(logger)
		self.agent = agent

	def create_connection(self):
		return self.agent.create_connection()

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class BaseAgent(object):
	"""
	TODO
	"""
	
	def __init__(self):
		return

	def setup(self, tenant_name, agent_name, logger, cache_manager, num_working_threads=None, extra_symbols=None, extra_agentstatus_functions=None):

		self.load_settings_timeout = DEFAULT_LOAD_SETTINGS_TIMEOUT
		self.load_settings_delay   = DEFAULT_LOAD_SETTINGS_DELAY

		self.tenant_name = tenant_name
		self.agent_name  = agent_name

		self.tenantid   = None           # Valorizzato nella fase iniziale di comunicazione col server
		self.agentid    = None           # Valorizzato nella fase iniziale di comunicazione col server

		self.num_working_threads = num_working_threads   # Se <> None, la parametrizzazione data dal server viene ignorata

		self.system_expr_functions = {}

		self.extra_symbols               = extra_symbols                if extra_symbols               is not None else {}
		self.extra_agentstatus_functions = extra_agentstatus_functions  if extra_agentstatus_functions is not None else {}

		self.logger = logger
		self.cache_manager = cache_manager

		self.ipc_factory = AgentIPCFactory(self.logger, self)

		self.broker_address = None
		self.broker_auth    = None

		self._stopping = False

		# Working threads
		self.threadpool = None

		# Schedule thread
		self.scheduler  = None

		# all other threads
		self.threads    = {}

		# Fields for configuration update
		self.update_connection = None
		self.update_signal = threading.Event()
		if not self.update_signal.isSet(): self.update_signal.set()

		self.update_status        = CONFIG_UPDATE_STATUS_NONE
		self.update_info          = 'Configuration not loaded yet'
		self.update_lastvalid_ts  = get_time()

		self.config_update_minperiod = DEFAULT_CONFIG_UPDATE_MINPERIOD
		self.configuration = None

		self.scheduled_datagroups = {}

		self.send_queue = None

		self.max_num_pending_event = DEFAULT_NUM_PENDING_EVENTS   #numero di update da mandare al server per ogni invio
		self.pending_events = []

		self._paused = False
		self.resume_event = threading.Event()
		self.resume_event.clear()

	#-----------------------------------------------------------------------------#
	#                                                                             #
	#-----------------------------------------------------------------------------#

	def set_broker(self, address ):
		self.broker_address = address

	def set_auth(self, auth ):
		self.broker_auth = auth

	#-----------------------------------------------------------------------------#
	#                                                                             #
	#-----------------------------------------------------------------------------#

	def start(self):

		self.logger.info("Starting agent %s:%s", self.tenantid, self.agentid)

		# Scarichiamo la configurazione dal server. Se ci sono problemi di connessione
		# restiamo bloccati su questa chiamata finche' il server non da segni di vita.
		self.load_settings_from_server()

		self.init_data()

		self.launch_all()

	#-----------------------------------------------------------------------------#
	#                                                                             #
	#-----------------------------------------------------------------------------#

	def init_data(self):
		self._load_modules()
		self._init_thread_pool(self.num_working_threads)
		self._init_scheduler()
		return

	#-----------------------------------------------------------------------------#

	def launch_all(self):
		try:
			self._launch_command_thread()
			self._launch_config_thread()
			self._launch_event_thread()

			# RIMUOVERE: L'utilizzo dell'helper nell'agente BaseAgent e' obsoleto
			#            - gli agenti locali non vengono mai interessati
			#            - gli agenti remoti lanciano un processo "parallelo"
			#      
			#self._launch_helper_thread()

			self.wait_for_stop()
		except Exception, e:
			self.logger.fatal("ERROR WHILE LAUNCHING: %s", e)

		return

	def wait_for_stop(self):

		#
		# BLock for termination
		#

		self.stop_event = threading.Event()
		self.stop_event.clear()
		self.stop_event.wait()

		#
		# Signal stop to other components/threads
		#
		self.logger.info("Stopping update thread")
		self.update_signal.set()

		self.logger.info("Received stop command")
		self.set_stopping()

		self.logger.info("Stopping scheduler")
		self.scheduler.stop()

		self.logger.info("Stopping threadpool")
		self.threadpool.stop()

		#
		# Wait for threads to terminate
		#
		self.logger.info("Waiting for pending threads")
		for name, t in self.threads.items():
			self.logger.info("Waiting for thread %s...", t.name)
			while t.isAlive():
				t.join(1)

		self.logger.info("All threads stopped")

	def _launch_command_thread(self):
		self.logger.info("Init control thread")
		self._launch_thread('CTL'    , self.handleCommands)
		self._launch_thread('CTL_TCP', self.handleCommands_tcp)

	def _launch_config_thread(self):
		self.logger.info("Init configuration thread")
		self._launch_thread('CONFIG', self.updateConfiguration)

	def _launch_event_thread(self):
		self.logger.info("Init event thread")
		self.send_queue = Queue.Queue(-1)
		self.send_queue_timeout = 1 # second
		self._launch_thread('EVENT' , self.handleEvents)



	# RIMUOVERE: L'utilizzo dell'helper nell'agente BaseAgent e' obsoleto
	#            - gli agenti locali non vengono mai interessati
	#            - gli agenti remoti lanciano un processo "parallelo"
	#def _launch_helper_thread(self):
	#	self.logger.info("Init helper thread")
	#	self._launch_thread('HELP'    , self.handleHelperCommands)

	#-----------------------------------------------------------------------------#
	#                                                                             #
	#-----------------------------------------------------------------------------#

	def signal_pause_to_main_thread(self):
		self.resume_event.clear()
		self._paused = True

	def is_paused(self):
		return self._paused

	def wait_if_paused(self):
		if self._paused:
			self.resume_event.wait()

	def signal_resume_to_main_thread(self):

		#
		# FREE PENDING DATA
		#
		self.scheduler.flush()
		self.scheduled_datagroups = {}

		# Signal resume
		self._paused = False
		self.resume_event.set()

		# Signal update config
		self.update_signal.set()

	#-----------------------------------------------------------------------------#
	#                                                                             #
	#-----------------------------------------------------------------------------#

	def signal_stop_to_main_thread(self):
		self.stop_event.set()

	def set_stopping(self):
		self._stopping = True

	def is_stopping(self):
		return self._stopping

	def _launch_thread(self, name, handler):
		ct = AgentThread(name, handler)
		self.threads[ ct.name ] = ct
		ct.start()

	def _wait_for_all_threads(self):

		for t in self.threads.values():

			self.logger.debug("Waiting for thread %s...", t.name)

			while t.isAlive:
				t.join(1)

			self.logger.debug("Thread %s stopped", t.name)

		self.logger.debug("Exiting")

	#-----------------------------------------------------------------------------#
	#                                                                             #
	#-----------------------------------------------------------------------------#

	def _load_modules(self):

		self.logger.debug("Loading modules...")

		expr_functions, agentstatus_functions = load_system_expr_and_agentstatus_functions()

		self.system_expr_functions         = expr_functions
		self.system_agentstatus_functions  = agentstatus_functions

		self.logger.debug("Expr symbols:")
		for symbol, v in sorted(self.system_expr_functions.items()):
			self.logger.debug('\t%s = %s', symbol, type(v))

		self.logger.debug("Status functions:")
		for symbol, v in sorted(self.system_agentstatus_functions.items()):
			self.logger.debug('\t%s = %s', symbol, type(v))

		return

	#-----------------------------------------------------------------------------#
	#                                                                             #
	#-----------------------------------------------------------------------------#

	def get_internal_status_functions(self):
		return sorted(self.system_agentstatus_functions.items()) + sorted(self.extra_agentstatus_functions.items())

	def verify_internal_status(self, check_name_to_verify=None):
		"""
		Esegue tutte le funzioni di callback di verifica dello stato caricate
		dai moduli "standard" e i moduli esterni.
		
		Restituisce la tupla con due elementi:
		
			* status : booleano. Vale True se tutti i controlli di stato hanno dato esito positivo
			* msgs   : vettore di stringhe. Elenco di stringhe con informazioni di dettaglio per ogni controllo di stato chiamato.
		"""

		self.logger.info("[agent] Verifying internal status...")

		check_to_call = self.get_internal_status_functions()

		status = True
		infos  = {}

		for check_name, check_fun in check_to_call:

			if check_name_to_verify is not None:
				if check_name != check_name_to_verify:
					continue

			self.logger.info("[agent] \t%s...", check_name)

			check_status, check_msg = check_fun(self)

			self.logger.info("[agent] \t\t%s = %s", check_status, check_msg)

			if check_status == False:
				status = False
				infos[ check_name ] = check_msg

		return status, infos

	#-----------------------------------------------------------------------------#
	#                                                                             #
	#-----------------------------------------------------------------------------#

	def create_connection(self):

		connect_options = {}

		self.logger.debug("Broker address: %s", self.broker_address )

		host = self.broker_address[0]
		if len(self.broker_address) > 1: 
			try:
				if self.broker_address[1]:
					connect_options['port']     = int(self.broker_address[1])
			except Exception, e:
				pass

		if self.broker_auth:
			connect_options['username'] = self.broker_auth[0]
			if len(self.broker_auth) > 1:
				connect_options['password'] = self.broker_auth[1]

		self.logger.log(9, "Opening connection with server %s", host )
		self.logger.log(9, "Options:")
		if self.logger.isEnabledFor(9):
			for k,v in connect_options.items():
				self.logger.log(9, "\t%s = %s", k, v)

		c = IPCConnection( host, **connect_options)

		self.logger.debug("Connection created!")

		return c

	#-----------------------------------------------------------------------------#
	#                                                                             #
	#-----------------------------------------------------------------------------#

	def load_settings_from_server(self):

		#
		# FIX: fissare un numero massimo di tentativi per cercare di scaricare la configurazione
		#      dal server.
		#
		while 1:
			try:
				self.logger.info("Loading settings from server...")

				connection = self.create_connection()
				connection.open()

				replay_id = '-'.join([create_global_uuid(), 'initialconfig'])

				replay_address = AGENT_TEMP_REPLY_QUEUE( replay_id )
				connection.bind(replay_address)

				config = self._send_request(connection, 'get_agent_config', self.tenant_name, self.agent_name, timeout=self.load_settings_timeout)

				self._update_settings_from_config(config)

				connection.close()

				break

			except IPCException, er:
				self.logger.critical("IPC error. Unable to load settings from server: %s", er)
			except RemoteServerException, er:
				self.logger.critical("Server error. Unable to load settings from server: %s", er)
			except Exception, er:
				self.logger.critical("Unknow error. Unable to load settings from server: %s", er)

			self.logger.info("Waiting %s seconds before retry to contact the server " % (self.load_settings_delay) )
			time.sleep(self.load_settings_delay)

		return True

	def _update_settings_from_config(self, config):

		self.logger.info("Config from server: %s", config)

		# Set all global settings
		self.tenantid            = config['tenantid']
		self.agentid             = config['agentid']

		self.config_update_minperiod = config.get('config_update_minperiod', DEFAULT_CONFIG_UPDATE_MINPERIOD)

		if self.num_working_threads is None: #ignore config from server if the agent was launched with this option already set
			self.num_working_threads = config['threads']

		return

	#-----------------------------------------------------------------------------#
	#                                                                             #
	#-----------------------------------------------------------------------------#

	def handleCommands(self):

		self.logger.info("[ctl] Start...")

		self.not_acked = 0

		while True:

			if self.is_stopping():
				break

			try:
				self.logger.debug("[ctl] Creating connection")

				self.ctl_connection = self.create_connection()
				self.ctl_connection.open()

				bind_address = AGENT_CTL_QUEUE(self.agentid)
				self.ctl_connection.bind(bind_address, persistent=False)

				while not self.is_stopping():
					self.receive_command()

				self.ctl_connection.ack_pending()

				self.ctl_connection.close()

			except Exception, e:
				#self.logger.error("[ctl] Error on connection: %s:%s -> %s", type(e), str(e), traceback.format_exc() )
				self.logger.error("[ctl] Error on connection: %s:%s", type(e), str(e) )

			finally:
				try:
					self.logger.debug("[ctl] Closing connection")
					if self.ctl_connection:
						self.ctl_connection.close()
				except Exception, e:
					pass

			self.logger.debug("[ctl] Wait before restarting")
			time.sleep(2)

		self.logger.info("[ctl] Exit")

	def receive_command(self):

		try:
			self.logger.log(9, "[ctl] Receiving...")

			(msg_addr, msg) = self.ctl_connection.receive_msg(timeout=1, ack_msg=False)

			self.logger.log(9, "[ctl] received from: %s:%s" % (msg_addr, msg) )

			if (msg == None and self.not_acked) or self.not_acked > 100:
				self.logger.info("[ctl] ack pending...")
				self.ctl_connection.ack_pending()
				self.not_acked = 0

			if msg == None:
				return

			self.logger.debug("[ctl] Received message %s:%s", msg_addr, msg)

			response = self.process_single_request_msg(msg)

			self.logger.debug("[ctl] Sending response to %s: %s", msg_addr, response)

			self.ctl_connection.send_msg( msg_addr, response, blocking=False)

		except IPCNotFound, e:
			self.logger.error("[ctl] Error on connection: %s", e)
			pass

		except IPCException, e:
			self.logger.error("[ctl] Error on connection: %s", e)
			raise e	

		except Exception, e:
			self.logger.warning("[ctl] Error while processing message: %s:%s", e, traceback.format_exc())

		self.not_acked += 1

	def handleCommands_tcp(self):

		try:

			self.logger.info("[ctl-tcp] Start...")

			server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
			server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

			self.server_socket_tcp = server_socket

			local_port = local_tcp_port_from_agent_name(self.agent_name)
			server_socket.bind((LOCAL_TCP_ADDRESS, local_port))

			server_socket.listen(5)

			while True:

				self.logger.info("[ctl-tcp] accepting...")
				client_socket, addr = server_socket.accept()
				self.logger.info("[ctl-tcp] New connection from %s", addr)

				msg = client_socket.recv(4096)

				self.logger.info("[ctl-tcp] Received request: %s", len(msg))
				response = self.process_single_request_msg(msg)

				self.logger.info("[ctl-tcp] Sending response")
				client_socket.send(response)

				self.logger.info("[ctl-tcp] Closing")
				client_socket.close()
		except Exception, e:
			msg = "%s:%s -> %s", type(e), str(e), traceback.format_exc()
			self.logger.error("Generic error in TCP connection: %s", msg)

		self.logger.info("[ctl-tcp] Exit")


	def process_single_request_msg(self, msg):

		request = decode_request(msg)

		request_id     = request['id']
		request_params = request['params']

		a_name = 'cmd_%s' % (str(request['type']))
		if hasattr(self, a_name):
			try:
				handler = getattr(self, a_name)
				response = handler(request_id, request_params)

			except Exception, e:
				response = encode_response_with_id( RESPONSE_ERROR, request_id, "Error on client request: %s:%s" % (str(e), traceback.format_exc()) )
		else:
			self.logger.warning("[ctl] Received unknown command: %s, %s", request['type'], request_params)
			response = encode_response_with_id( RESPONSE_ERROR, request_id, "invalid request from client: %s" % (request['type']) )

		return response

	def cmd_help(self, request_id, params):
		"""
		Return the list of supported commands.
		"""

		commands = []
		for k in dir(self):
			if k.startswith('cmd_'):
				doc = getattr(self, k).__doc__
				if doc is None:
					doc = ''
				else:
					doc = doc.strip()
				commands.append( (k[4:], doc))

		if commands:
			col_width = max([len(row[0]) for row in commands])

		line_format = '\t%-' + str(col_width) + 's  %s'

		commands = [ line_format % row for row in commands ]

		text  = ''
		text += 'Interna commands:\n'
		text += '\n'.join(commands)

		response = encode_response_with_id( RESPONSE_SUCCESS, request_id, text )

		return response

	def cmd_echo(self, request_id, params):
		"""
		echo/reply command
		"""
		self.logger.debug("Received info command")

		data = ' '.join([str(p) for p in params])

		response = encode_response_with_id(RESPONSE_SUCCESS, request_id, data)
		return response

	def cmd_stop(self, request_id, params):

		self.logger.info("Signaling stop to main thread")
		self.signal_stop_to_main_thread()

		response = encode_response_with_id( RESPONSE_SUCCESS, request_id, 'stopping')
		return response

	def cmd_pause(self, request_id, params):

		self.logger.info("Signaling stop to main thread")
		self.signal_pause_to_main_thread()
		response = encode_response_with_id( RESPONSE_SUCCESS, request_id, 'pause')
		return response

	def cmd_resume(self, request_id, params):

		self.logger.info("Signaling resume to main thread")
		self.signal_resume_to_main_thread()
		response = encode_response_with_id( RESPONSE_SUCCESS, request_id, 'resumed')
		return response

	def cmd_update_config(self, request_id, params):
		"""
		Signals the agent to reload monitoring configuration.
		"""

		self.logger.info("Signaling to update")
		#self.signal_stop_to_main_thread()
		self.update_signal.set()
		response = encode_response_with_id( RESPONSE_SUCCESS, request_id, 'updating')
		return response

	def cmd_info(self, request_id, params):
		"""
		Returns human-readable informations about agent's internal status.
		"""

		self.logger.info("Received info command: %s", params)

		text = []

		text += [ self.threadpool.get_info_text() ]

		text += [ 'Scheduled datagroups: %s' % (self.scheduler.get_schedule_size(), ) ]

		text += self.get_queues_status_text()

		text += self.get_configuration_status_text()

		text = '\n'.join(text)

		self.logger.info("Sending response: %s:%s", request_id, text)
		response = encode_response_with_id( RESPONSE_SUCCESS, request_id, text )

		return response

	def cmd_infojson(self, request_id, params):
		"""
		Returns structured informations about agent's internal status (JSON format).
		"""
		try:
			data = self._infojson(request_id, params)
		except Exception, e:
			msg = "Unable to calculate status: %s %s" % (type(e), e)
			self.logger.error(msg)
			data = {
				'critical_problems': { 
					'status' : msg 
				}
			}

		text = json.dumps(data)

		self.logger.info("Sending response: %s %s", request_id, text)
		response = encode_response_with_id( RESPONSE_SUCCESS, request_id, text )

		return response

	def _infojson(self, request_id, params):

		#
		# Calculate data
		#
		infos  = self.get_configuration_status()

		infos2 = self.get_queues_status()

		infos3 = self.threadpool.get_info()

		wqueuelen = infos3.get('wait_queue', 0)

		vmsize = get_proc_memory()
		rmsize = get_proc_resident()

		#
		# Calculate critical problems
		#
		critical_problems = {}
		critical_problems.update( infos ['critical_problems'] )
		critical_problems.update( infos2['critical_problems'] )
		critical_problems.update( infos3['critical_problems'] ) 

		#
		# Prepare response
		#
		data = {
			'nodes'      : infos.get('nodes_count'    , None),
			'ifaces'     : infos.get('ifaces_count'   , None),
			'storages'   : infos.get('storages_count' , None),
			'services'   : infos.get('services_count' , None),
			'devices'    : infos.get('devices_count'  , None),

			'datagroups' : infos.get('datagroups_count' , None),
			'datasources': infos.get('datasources_count', None),
			'conditions' : infos.get('conditions_count' , None),

			'dgrate'     : infos3.get('rate', None),

			'wqueuelen'  : wqueuelen,

			'squeuelen'  : infos2.get('send_queue', None),

			'vmemory'    : int(vmsize),
			'rmemory'    : int(rmsize),

			'critical_problems': critical_problems,
		}
		
		data.update( 
			dict([ ('threadpool.process.%s' % (k), v) for k,v in infos3['process'].items() ])
		)

		return data

	def cmd_verify_status(self, request_id, params):
		"""
		Check status of internal components/plugins.
		"""
		params = params[0:-1]

		self.logger.info("Received info command: %s", params)

		if params:
			check_status_name = params[0] 

			status, msgs = self.verify_internal_status(check_status_name)

			text = '%s - %s' % (status, msgs)
		else:
			text  = ''
			text += "Internals:\n"
			text += '\n'.join([ ("\t"+name) for name, fun in self.get_internal_status_functions() ])

		self.logger.info("Sending response: %s:%s", request_id, text)
		response = encode_response_with_id( RESPONSE_SUCCESS, request_id, text )

		return response

	def cmd_show_schedule(self, request_id, params):
		"""
		Returns the next <n> scheduled checks.
		"""
		now = get_time()
		num = 5

		cmd_params = params[0:-1]
		
		if len(cmd_params) > 0:
			num = cmd_params[0]

		try:
			num = int(num)
		except ValueError, e:
			num = 5

		events = self.scheduler.getScheduledEvents(num)

		data = []
		i = 0
		for (starttime, handler, params) in events:
			check = params[0]
			s = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(starttime))
			t = int(starttime - now)
			data.append( "%d %s (%s)\t%s" % (i, s, t, check) )
			i += 1

		data = '\n'.join(data)

		response = encode_response_with_id(RESPONSE_SUCCESS, request_id, data)
		return response

	def cmd_check_in(self, request_id, params):
		"""
		Reschedule the specified datagroup in the next <n> seconds.
		"""
		cmd_params = params[0:-1]

		try:
			path     = cmd_params[0]
			new_time = time.time() + int(cmd_params[1])
		except IndexError, e:
			raise Exception("Invalid parameters")

		return self._reschedule(request_id, path, new_time)

	def cmd_check_now(self, request_id, params):
		"""
		Force execution for the specified datagroup.
		"""
		cmd_params = params[0:-1]

		try:
			path     = cmd_params[0]
			new_time = None
		except IndexError, e:
			raise Exception("Invalid parameters")

		return self._reschedule(request_id, path, new_time)

	def _reschedule(self, request_id, path, new_time):
		"""
		Cerca lo ScheduledDataGroup per il datagroup "path" e lo rischedula al tempo "next_time".
		
		Se "next_time" e' nullo, il datagroup viene rischedulato per l'esecuzione immediata.
		"""
		if new_time is None:
			new_time = time.time()

		handler        = self.do_check_and_reschedule
		fun_match_data = lambda event_data: (event_data[0].pdatagroup.path == path)   # event_data = <  time  ,  ScheduledDatagroup obj >

		changed = self.scheduler.alterEventScheduleTime( handler, fun_match_data, new_time)

		response = encode_response_with_id(RESPONSE_SUCCESS, request_id, "rescheduled" if changed else "not changed")

		return response

	def cmd_when_is(self, request_id, params):
		"""
		Gives informations about scheduling time for the given datagroup
		"""

		path = params[0]

		found = None
		for sdg in self.scheduled_datagroups.values():
			if sdg.is_valid():
				if sdg.pdatagroup.path == path:
					found = sdg
					break

		if found is None:
			data = 'NOTFOUND'
		else:
			data = POLLSTATE_STR[found.sched_state]

			if found.sched_state == POLLSTATE_SCHEDULED:
				scheduled_time = found.schedule_time
				s = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(scheduled_time))
				data += " %s (%s)" % (s, int(scheduled_time - get_time()))

			#if found.sched_state == POLLSTATE_IDLE:
			#	data = "idle"
			#elif found.sched_state == POLLSTATE_SCHEDULED:
			#	scheduled_time = found.schedule_time
			#	s = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(scheduled_time))
			#	data = "%s (%s)" % (s, int(scheduled_time - get_time()))
			#elif found.sched_state == POLLSTATE_QUEUED:
			#	data = "waiting"
			#elif found.sched_state == POLLSTATE_CHECKING:
			#	data = "checking"
			#elif found.sched_state == POLLSTATE_COMMITTING:
			#	data = "committing"

		response = encode_response_with_id(RESPONSE_SUCCESS, request_id, data)
		return response

	def cmd_clusters(self, request_id, params):
		"""
		Returns internal cluster informations (JSON format).
		"""

		configuration = self.configuration

		data = []
		for uuid, c in configuration.clusters.items():
			data.append(
				#"%s:%s:%s" % (uuid, c.name, c.mtime)
				c.to_dict(),
			)

		data = '\n'.join(data)

		response = encode_response_with_id(RESPONSE_SUCCESS, request_id, data)
		return response

	def cmd_nodes(self, request_id, params):
		"""
		Returns agent's internal data about clusters (JSON format).
		"""
		return self._cmd_elements(request_id, params, [ELEMENT_KIND_NODE])

	def cmd_interfaces(self, request_id, params):
		"""
		Returns agent's internal data about interfaces (JSON format).
		"""
		return self._cmd_elements(request_id, params, [ELEMENT_KIND_IFACE])

	def cmd_storages(self, request_id, params):
		"""
		Returns agent's internal data about storages (JSON format).
		"""
		return self._cmd_elements(request_id, params, [ELEMENT_KIND_STORAGE])

	def cmd_services(self, request_id, params):
		"""
		Returns agent's internal data about services (JSON format).
		"""
		return self._cmd_elements(request_id, params, [ELEMENT_KIND_SERVICE])

	def cmd_devices(self, request_id, params):
		"""
		Returns agent's internal data about devices (JSON format).
		"""
		return self._cmd_elements(request_id, params, [ELEMENT_KIND_DEVICE])

	def cmd_elements(self, request_id, params):
		"""
		Returns internal cluster informations (JSON format).
		"""
		return self._cmd_elements(request_id, params)

	def _cmd_elements(self, request_id, params, filter_types=None):

		configuration = self.configuration
		element_name = None
		cmd_params = params[0:-1]

		try:
			element_name = cmd_params[0]
		except IndexError, e:
			element_name = None
		data = []
		if element_name != None:
			for uuid, e in configuration.elements.items():
				if e.name == element_name:
					data = e.get_dict()
					data = json.dumps(data)
					break
		else:
			data = []
			for uuid, e in configuration.elements.items():
				if filter_types is not None:
					if e.kind not in filter_types:
						continue
				data.append(
					#"%s:%s:%s:%s" % (uuid, ELEMENT_KIND_STR[e.kind], e.name, e.mtime)
					e.get_dict()
				)

			#data = '\n'.join(data)
			data = json.dumps(data)

		response = encode_response_with_id(RESPONSE_SUCCESS, request_id, data)
		return response

	def cmd_links(self, request_id, params):
		"""
		Returns agent's internal data about links (JSON format).
		"""

		configuration = self.configuration

		data = []
		for uuid, l in configuration.links.items():
			data.append(
				"%s:%s:%s:%s:%s:%s" % (uuid, l.name, l.mtime, l.ltype, l.ifaces[0], l.ifaces[1])
			)

		data = '\n'.join(data)

		response = encode_response_with_id(RESPONSE_SUCCESS, request_id, data)
		return response

	def cmd_datagroups(self, request_id, params, filter_types=None):

		configuration = self.configuration

		cmd_params = params[0:-1]
		try:
			path_or_id = cmd_params[0]
		except IndexError, e:
			path_or_id = None

		data = ''

		if path_or_id is not None:
			for uuid, dg in configuration.datagroups.items():
				if dg.path == path_or_id or dg.uuid == path_or_id:

					data = dg.get_dict()

					l = []
					for oid in dg.datasources:
						o = configuration.get_datasource(oid)
						l.append(o.get_dict())
					data['datasources'] = l

					l = []
					for oid in dg.conditions:
						o = configuration.get_condition(oid)
						l.append(o.get_dict())
					data['conditions'] = l

					data = json.dumps(data)
					break
		else:
			data = []
			for uuid, dg in configuration.datagroups.items():
				data.append(
					dg.get_dict()
				)

			data = json.dumps(data)

		response = encode_response_with_id(RESPONSE_SUCCESS, request_id, data)
		return response

	def cmd_datasources(self, request_id, params, filter_types=None):

		configuration = self.configuration

		cmd_params = params[0:-1]
		try:
			path_or_id = cmd_params[0]
		except IndexError, e:
			path_or_id = None

		data = ''

		if path_or_id is not None:
			for uuid, ds in configuration.datasources.items():
				if ds.path == path_or_id or ds.uuid == path_or_id:
					data = ds.get_dict()
					data = json.dumps(data)
					break
		else:
			data = []
			for uuid, ds in configuration.datasources.items():
				data.append(
					ds.get_dict()
				)

			data = json.dumps(data)

		response = encode_response_with_id(RESPONSE_SUCCESS, request_id, data)
		return response

	def cmd_conditions(self, request_id, params, filter_types=None):

		configuration = self.configuration

		cmd_params = params[0:-1]
		try:
			path_or_id = cmd_params[0]
		except IndexError, e:
			path_or_id = None

		data = ''

		if path_or_id is not None:
			for uuid, c in configuration.conditions.items():
				if c.path == path_or_id or c.uuid == path_or_id:
					data = c.get_dict()
					data = json.dumps(data)
					break
		else:
			data = []
			for uuid, c in configuration.conditions.items():
				data.append(
					c.get_dict()
				)

			data = '\n'.join(data)
			data = json.dumps(data)

		response = encode_response_with_id(RESPONSE_SUCCESS, request_id, data)
		return response

	def cmd_set_log_level(self, request_id, params, filter_types=None):
		"""
		Set agent's logging level.
		"""

		num = int(params[0])

		self.logger.setLevel(num)
		for handler in self.logger.handlers:
			handler.setLevel( num )

		response = encode_response_with_id(RESPONSE_SUCCESS, request_id, 'ok')
		return response

	def cmd_set(self, request_id, params, filter_types=None):
		"""
		Set agent's working parameters.
		"""

		what = params[0]

		if what == 'num-pending-events':
			self.max_num_pending_event = int(params[1])
		else:
			raise Exception("Unrecognized option '%s'" % (what))

		response = encode_response_with_id(RESPONSE_SUCCESS, request_id, 'ok')
		return response

	#-----------------------------------------------------------------------------#
	#                                                                             #
	#-----------------------------------------------------------------------------#

	def updateConfiguration(self):

		DEFAULT_CONFIG_UPDATE_RETRY_DELAY = 1

		self.logger.info("[config] Start...")

		while 1:

			if self.is_stopping():
				break

			while 1:

				if self.is_paused():
					break

				now = get_time()

				if self.configuration is not None:

					# Se non siamo al primo giro e non e' ancora scaduto il timeout per aggiornare la configurazione, aspettiamo!
					if (self.config_update_minperiod - (now - self.update_lastvalid_ts)) > 0:

						# busy wait, because we have to wait for the signal or the 'config_update_minperiod'
						self.update_signal.wait(1)
						
						# Qualcuno ha 
						#  a) "forzato l'aggiornamento" 
						#  b) inviato il segnale di stop
						#
						#  --> smettiamo di aspettare
						#
						if not self.update_signal.isSet():
							continue

				success = self.do_load_configuration()

				# Indipendentemente da cosa e' successo, stiamo pronti a ricevere
				# un nuovo segnale di aggiornamento dall'esterno.
				self.update_signal.clear()

				if not success:
					self.logger.info("Error on last update. Wait (%s seconds) and retry", DEFAULT_CONFIG_UPDATE_RETRY_DELAY)
					time.sleep(DEFAULT_CONFIG_UPDATE_RETRY_DELAY)

			self.logger.info("[config] paused. Waiting for resume")
			self.wait_if_paused()

		self.logger.info("[config] exit")

		return

	def do_load_configuration(self):

		valid = False
		error = None

		try:
			t1 = get_time()

			self.update_status = CONFIG_UPDATE_STATUS_LOADING
			self.update_info   = 'opening connection with server...'

			self.logger.info("[config] Checking new configuration.")

			self.logger.info("[config] Connecting to server")
			update_connection = self.create_connection()

			self.logger.info("[config] Connection created, opening...")
			update_connection.open()

			self.logger.info("[config] Connection created, binding...")
			update_connection.bind('temp-sanet-agent-%s-%s-updates' % (self.agentid, create_global_uuid()))

			self.update_connection = update_connection

			self.update_status = CONFIG_UPDATE_STATUS_LOADING
			self.update_info   = 'loading data'

			self._load_all_configuration()

			t2 = get_time()

			self.update_lastvalid_ts = t2
			self.update_status       = CONFIG_UPDATE_STATUS_DONE
			self.update_info         = 'Update took: %s s' % (t2-t1)

			valid = True

		except IPCException, e:
			self.logger.critical("Unable to contact server: %s", str(e))
			error = e

		except RemoteServerException, e:
			self.logger.error("[config] error message by remote server: %s ", str(e), )
			error = e

		except Exception, e:
			self.logger.critical("[config] Unknown error while updateding: %s - %s\nTraceback: %s", type(e), str(e), traceback.format_exc() )
			error = e

		finally:

			if error is not None:
				self.update_status = CONFIG_UPDATE_STATUS_ERR
				self.update_info   = 'Error: %s:%s' % (type(error), str(error))

			try:
				if self.update_connection:
					self.update_connection.close()
			except Exception, e:
				self.logger.error("[config] error while closing update connection: %s %s", type(e), str(e))
				self.update_connection = None

			self.update_connection = None

		return valid


	def get_configuration_status(self):

		critical_problems = {}

		elapsed       = get_time() - self.update_lastvalid_ts
		tolerance     = self.config_update_minperiod * DEFAULT_CONFIG_UPDATE_TOLERANCE
		
		if tolerance <= elapsed:
			critical_problems['configuration_load'] = "Not loading configuration. Last try %s (%ss, tolerance %s)." % (self.update_lastvalid_ts, elapsed, tolerance)

		infos = {
			'lastvalid_ts'     : self.update_lastvalid_ts,
			'config_status'    : self.update_status,
			'update_info'      : self.update_info,
			'critical_problems': critical_problems,
		}

		if self.configuration != None:

			load_ts = self.configuration.get_load_time()

			if load_ts != None:
				next_update = int(load_ts + self.config_update_minperiod - get_time())
			else:
				next_update = None

			infos.update(  self.configuration.get_info() )

			infos['load_time'  ] = load_ts
			infos['next_update'] = next_update

		#
		# Additional checks
		#
		internal_status, internal_infos = self.verify_internal_status()
		if internal_status == False:
			infos['critical_problems'].update( internal_infos )

		return infos

	def get_queues_status(self):

		infos = {
			'critical_problems': {}
		}

		try:
			if self.send_queue is not None:
				infos['send_queue'] = self.send_queue.qsize()
			else:
				infos['send_queue'] = 0
		except Exception, e:
				infos['send_queue'] = None

		infos['pending_queue'] = len(self.pending_events)
		infos['max_pending'  ] = self.max_num_pending_event

		return infos

	def get_queues_status_text(self):

		infos = self.get_queues_status()
		
		text  = []
		text += [ 'Sending Queue: %s' % (infos['send_queue']) ]
		text += [ 'Pending events: %s/%s' % (infos['pending_queue'], infos['max_pending'  ]) ]
		
		return text

	def get_configuration_status_text(self):

		infos = self.get_configuration_status()

		#
		# FORMAT
		#
		text  = []
		text += [ "Config thread status: %s %s (last complete: %s - elapsed %s)" % (
			CONFIG_UPDATE_STATUS_STR[infos['config_status']], 
			infos['update_info'],
			infos['lastvalid_ts'],
			get_time() - infos['lastvalid_ts'],
		)]
		text += [ 'Configuration loaded at: %s (next update in %s s).' % (
			infos.get('load_time', None), infos.get('next_update', None)
		)]

		text += [ 'Clusters   : %s' % (infos.get('clusters_count'   , None) ) ]
		text += [ 'Elements   : %s' % (infos.get('elements_count'   , None) ) ]
		text += [ 'Nodes      : %s' % (infos.get('nodes_count'      , None) ) ]
		text += [ 'Interfaces : %s' % (infos.get('ifaces_count'     , None) ) ]
		text += [ 'Storages   : %s' % (infos.get('storages_count'   , None) ) ]
		text += [ 'Devices    : %s' % (infos.get('devices_count'    , None) ) ]
		text += [ 'Services   : %s' % (infos.get('services_count'   , None) ) ]
		text += [ 'Links      : %s' % (infos.get('links_count'      , None) ) ]
		text += [ 'Datagroups : %s' % (infos.get('datagroups_count' , None) ) ]
		text += [ 'Datasources: %s' % (infos.get('datasources_count', None) ) ]
		text += [ 'Conditions : %s' % (infos.get('conditions_count' , None) ) ]

		if infos['critical_problems']:
			text += ['CRITICAL PROBLEMS']
			for name, info in sorted(infos['critical_problems'].items()):
				text += ['- %s: %s' % (name, info) ]

		return text

	def _load_all_configuration(self):

		t1 = t2 = t3 = t4 = t5 = t6 = t7 = t8 = t9 = t10 = t11 = t12 = t13 = 0

		try:
			old_configuration = self.configuration

			new_config = PollerConfiguration(self.logger, self.tenantid)

			self.logger.info("[config] Loading....")

			t1 = time.time()
			self.update_info   = 'loading clusters...'
			self._update_clusters(new_config)

			t2 = time.time()
			self.update_info   = 'loading nodes...'
			self._update_nodes(new_config)

			t3 = time.time()
			self.update_info   = 'loading ifaces...'
			self._update_ifaces(new_config)

			t4 = time.time()
			self.update_info   = 'loading storages...'
			self._update_storages(new_config)

			t5 = time.time()
			self.update_info   = 'loading services...'
			self._update_services(new_config)

			t6 = time.time()
			self.update_info   = 'loading devices...'
			self._update_devices(new_config)

			t7 = time.time()
			self.update_info   = 'loading links...'
			self._update_links(new_config)

			t8 = time.time()
			self.update_info   = 'loading datasources...'
			self._update_datasources(new_config)

			t9 = time.time()
			self.update_info   = 'loading conditions...'
			self._update_conditions(new_config)

			t10 = time.time()
			self.update_info   = 'loading parameters..'
			self._update_datagroups_parameters(new_config)

			t11 = time.time()

			new_config.reindex_datagroups()

			self.update_info   = 'loading datagroups...'
			datagroup_to_schedule = []
			datagroup_to_schedule = self._update_datagroups(new_config)
			
			t12 = time.time()

			new_config.set_load_time( get_time() )

			self.logger.info("[config] configuration updated. (datagroup to reschedule %s)", len(datagroup_to_schedule))

			#
			# La configurazione scaricata diventa quella "valida"
			#
			self.configuration = new_config

			self.schedule_datagroups(datagroup_to_schedule)

			# Mentre stiamo aggiornando la configurazione il sistema non e'
			# fermo. Assicuriamoci che gli elementi della vecchia configurazione
			# non vengano piu' usati da nessuno.
			#
			# NOTA: Datagroup gia' schedulati devono controllare
			# prima di ogni check' che il datagroup puntato sia ancora
			# valido.
			if old_configuration is not None:
				old_configuration.invalid_all_elements()

			t13 = time.time()

			return
		finally:
			self.logger.info("Config update: Times:")
			self.logger.info("Config update:     cluster     %5.3f" ,  t2 -  t1)
			self.logger.info("Config update:     nodes       %5.3f" ,  t3 -  t2)
			self.logger.info("Config update:     iface       %5.3f" ,  t4 -  t3)
			self.logger.info("Config update:     storage     %5.3f" ,  t5 -  t4)
			self.logger.info("Config update:     service     %5.3f" ,  t6 -  t5)
			self.logger.info("Config update:     devices     %5.3f" ,  t7 -  t6)
			self.logger.info("Config update:     links       %5.3f" ,  t8 -  t7)
			self.logger.info("Config update:     datasources %5.3f" ,  t9 -  t8)
			self.logger.info("Config update:     conditions  %5.3f" , t10 -  t9)
			self.logger.info("Config update:     parameters  %5.3f" , t11 - t10)
			self.logger.info("Config update:     UPDATE      %5.3f" , t12 - t11)
			self.logger.info("Config update:     RESCHEDULE  %5.3f" , t13 - t12)

	def schedule_datagroups(self, datagroup_to_schedule):

		self.logger.debug("NEED TO SCHEDULE %s datagroups", len(datagroup_to_schedule))

		now = get_time()

		for sdg in datagroup_to_schedule:
			self.schedule_one_datagroup(now, sdg)

		return

	def schedule_one_datagroup(self, now, sdg):

		pdatagroup = sdg.pdatagroup

		# Execute the datagroup immediately if a minperiod has passed.
		next_timeout = 0

		if pdatagroup.lastdone is not None:
			delta   = (now - pdatagroup.lastdone)
			next_timeout = pdatagroup.minperiod - delta
			next_timeout = max(next_timeout, 0)
		else:
			delta = 0

		if next_timeout <= 0:
			# We are probably late. Execute immediatly.
			params = ( sdg, ) 
			self._exec_scheduled_job_callback(self.do_check_and_reschedule, params)
			self.logger.debug("SCHEDULED NOW (%s, %s, %s, %s): %s", now, pdatagroup.lastdone, pdatagroup.minperiod, delta,  sdg)
		else:
			# There is enough time until the next check. Reschedule 
			sdg.sched_state   = POLLSTATE_SCHEDULED
			sdg.schedule_time = now + next_timeout
			params = ( sdg, ) 
			self.scheduler.scheduleNewEvent(sdg.schedule_time, self.do_check_and_reschedule, params)
			self.logger.debug("SCHEDULED IN %s: %s", next_timeout, sdg)

	def _send_request(self, connection, rtype, *params, **kwargs):

		DEFAULT_TIMEOUT = 60

		timeout = kwargs.get('timeout', DEFAULT_TIMEOUT)

		self.logger.debug("Sending request: %s", rtype)

		mid = str(get_time())
		msg = encode_request_with_id(rtype, mid, *params)

		# SEND
		connection.send_msg('sanetd-ctl', msg, timeout=timeout)

		# RECEIVE
		(addr, msg) = connection.receive_msg(ack_msg=True, timeout=timeout)
		if (msg == None):
			raise Exception("Timeout while loading configuration for %s" % (rtype))

		response = decode_response(msg)
		if response['id'] != mid:
			raise RemoteServerException("Waiting for response with id %s. Received %s" % (mid, response['id']))

		if response['status'] == RESPONSE_ERROR:
			raise RemoteServerException("Unable to get elements from server: %s" % ( response['content'] ) )

		return response['content']

	def _update_clusters(self, new_config):
		data = self._send_request(self.update_connection, 'get_clusters', self.tenantid, self.agentid)		
		data = DECOMPRESS_DATA(data)
		self._update_clusters_from_data(new_config, data)

	def _update_nodes(self, new_config):
		data = self._send_request(self.update_connection, 'get_nodes', self.tenantid, self.agentid)
		data = DECOMPRESS_DATA(data)
		self._update_nodes_from_data(new_config, data)

	def _update_ifaces(self, new_config):
		data = self._send_request(self.update_connection, 'get_ifaces', self.tenantid, self.agentid)
		data = DECOMPRESS_DATA(data)
		self._update_ifaces_from_data(new_config, data)

	def _update_storages(self, new_config):
		self._update_element(new_config, 'get_storages', PStorage)
	
	def _update_services(self, new_config):
		self._update_element(new_config, 'get_services', PService)
		
	def _update_devices(self, new_config):
		self._update_element(new_config, 'get_devices' , PDevice)

	def _update_element(self, new_config, cmd, element_class):
		data = self._send_request(self.update_connection, cmd, self.tenantid, self.agentid)
		data = DECOMPRESS_DATA(data)
		self._update_element_from_data(new_config, data, element_class)

	def _update_links(self, new_config):
		data = self._send_request(self.update_connection, 'get_links', self.tenantid, self.agentid)
		data = DECOMPRESS_DATA(data)
		self._update_links_from_data(new_config, data)

	def _update_datagroups_parameters(self, new_config):
		data = self._send_request(self.update_connection, 'get_datagroups_params', self.tenantid, self.agentid)
		data = DECOMPRESS_DATA(data)
		self._update_datagroups_parameters_from_data(new_config, data)

	def _update_datagroups(self, new_config):
		data = self._send_request(self.update_connection, 'get_datagroups', self.tenantid, self.agentid)
		data = DECOMPRESS_DATA(data)
		return self._update_datagroups_from_data(new_config, data)

	def _update_datasources(self, new_config):
		data = self._send_request(self.update_connection, 'get_datasources', self.tenantid, self.agentid)
		data = DECOMPRESS_DATA(data)
		self._update_datasources_from_data(new_config, data)

	def _update_conditions(self, new_config):
		data = self._send_request(self.update_connection, 'get_conditions', self.tenantid, self.agentid)
		data = DECOMPRESS_DATA(data)
		self._update_conditions_from_data(new_config, data)

	def _update_clusters_from_data(self, new_config, data):

		self.update_info   = 'updating clusters...'

		self.logger.log(9, "Loading %s clusters", len(data))
		for d in data:
			self.logger.log(9, "Loading cluster %s", d)
			pc = PCluster()
			pc.load_from_dict(d)
			new_config.add_cluster(pc)

	def _update_element_from_data(self, new_config, data, element_class):

		self.update_info   = 'updating elements...'

		self.logger.log(9, "Loading %s elements", len(data))
		for d in data:
			self.logger.log(9, "Loading %s", d)
			e = element_class()
			e.load_from_dict(d)
			new_config.add_element(e)

	def _update_nodes_from_data(self, new_config, data):

		self.update_info   = 'updating nodes...'

		self.logger.log(9, "Loading %s nodes", len(data))
		for d in data:
			self.logger.log(9, "Loading node %s", d)
			pn = PNode()
			pn.load_from_dict(d)
			new_config.add_element(pn)

	def _update_ifaces_from_data(self, new_config, data):

		self.update_info   = 'updating ifaces...'

		self.logger.log(9, "Loading %s ifaces", len(data))
		for d in data:
			self.logger.log(9, "Loading iface %s", d)
			pi = PIface()
			pi.load_from_dict(d)
			new_config.add_element(pi)

	def _update_links_from_data(self, new_config, data):

		self.update_info   = 'updating links...'

		self.logger.log(9, "Loading %s links (%s linked ifaces)", len(data['links']), len(data['linked_ifaces']) )

		for d in data:
			self.logger.log(9, "Loading link %s", d)

			# Grouping ifaces
			iface_by_link_id = {}
			for linkedifaced in data['linked_ifaces']:
				ifaces = iface_by_link_id.get(linkedifaced['link_id'], [])
				ifaces.append(linkedifaced['iface_id'])
				iface_by_link_id[linkedifaced['link_id']] = ifaces

			# Creating links
			for linkd in data['links']:

				try:
					ifaces = iface_by_link_id[ linkd['uuid'] ]
				except KeyError, e:
					self.logger.error("Ifaces for link %s not found!", linkd['uuid'])
					continue

				pl = PLink(
					uuid   = linkd['uuid'],
					name   = linkd['name'],
					ltype  = linkd['ltype'],
					ifaces = ifaces
				)

				try:
					new_config.add_link(pl)
					self.logger.log(9, "Loaded link %s", pl)
				except Exception, e:
					self.logger.error("Error while adding link %s: %s", linkd['uuid'], e  )

	def _update_datagroups_parameters_from_data(self, new_config, data):

		self.update_info   = 'updating parameters...'

		self.logger.log(9, "Loading %s datagroup parameters", len(data))

		params_by_dgid = {}
		for p in data:
			p_list = params_by_dgid.get( p['datagroup_id'], [])
			p_list.append(p)
			params_by_dgid[ p['datagroup_id'] ] = p_list

		for dg_id, params in params_by_dgid.items():
			new_config.add_datagroup_params(dg_id, params)

	def _update_datagroups_from_data(self, new_config, data):

		self.update_info   = 'updating datagroups...'

		self.logger.log(9, "Loading %s datagroups", len(data))

		now = get_time()

		datagroup_to_schedule = []

		for d in data:
			self.logger.log(9, "Loading datagroup %s", d)

			if d['element'] not in new_config.elements:
				self.logger.error("Element %s for datagroup %s is missing in new poller configuration. Skipping.", d['element'], d['path'])
				continue

			pdg = PDataGroup()
			pdg.load_from_dict(d)

			pdg.datasources = new_config.get_datasources_ids_by_dg_id(pdg.uuid)
			pdg.conditions  = new_config.get_conditions_ids_by_dg_id(pdg.uuid)

			new_config.add_datagroup(pdg)

			was_scheduled = False
			valid_for_check = True


			valid_for_check = pdg.valid_for_check()

			if not valid_for_check:
				self.logger.info("Datagroup loaded %s but not valid for check", pdg.path)

			if pdg.minperiod <= 0:
				self.logger.log(9, "Datagroup loaded %s but minperiod=0", pdg.path)
				valid_for_check = False

			#
			# Ora che l'oggetto PDatagroup e' sicuro in memoria 
			# vediamo se deve essere schedulato o meno.
			#
			sdg = self.scheduled_datagroups.get(pdg.uuid, None)
			if sdg is None:
				sdg = ScheduledDatagroup(pdg)
				self.scheduled_datagroups[pdg.uuid] = sdg
			else:
				# Se il PDataGroup era gia' stato Schedulato 
				# aggiorniamo il rifermento in modo che 
				# il nuovo datagroup sia gia' pronto all'uso
				# nel momento in cui lo scheduler lo dovra' far eseguire
				# da uno dei thread.
				sdg.pdatagroup = pdg

				if sdg.sched_state != POLLSTATE_IDLE:
					was_scheduled = True

			if valid_for_check and was_scheduled == False:
				datagroup_to_schedule.append( sdg )

				self.logger.log(9, "loaded for schedule %s:%s", pdg.uuid, pdg.path)

		return datagroup_to_schedule

	def _update_datasources_from_data(self, new_config, data):

		self.update_info   = 'updating datasources...'

		self.logger.log(9, "Loading %s datasources", len(data))

		for d in data:
			#self.logger.info("Loading datasource %s", d)
			pds = PDataSource()
			pds.load_from_dict(d)
			new_config.add_datasource(pds)

			# link this datasource with its datagroup
			#pdg = new_config.get_datagroup(pds.datagroup_id)
			#pdg.datasources.append(pds.uuid)

	def _update_conditions_from_data(self, new_config, data):

		self.update_info   = 'updating conditions...'

		self.logger.log(9, "Loading %s condition", len(data))

		for d in data:
			#self.logger.info("Loading condition %s", d)
			pc = PCondition()
			pc.load_from_dict(d)
			new_config.add_condition(pc)

			# link this condition with its datagroup
			#pdg = new_config.get_datagroup(pc.datagroup_id)
			#pdg.conditions.append(pc.uuid)

	#-----------------------------------------------------------------------------#
	# SCHEDULER                                                                   #
	#-----------------------------------------------------------------------------#

	def _init_scheduler(self):

		self.logger.debug("Init scheduler...")

		self.scheduler  = PollScheduler( self._exec_scheduled_job_callback )

		self.logger.info("Starting scheduler...")
		self.scheduler.start()

	def _exec_scheduled_job_callback(self, scheduled_handler, scheduled_params):

		# Se mi sto fermando semplicemente scarto quello che c'era da fare
		if self.is_stopping():
			return

		# Se mi hanno messo in pausa scarto quello che c'e' da fare
		if self.is_paused():
			return

		scheduled_datagroup = scheduled_params[0]
		pdg = scheduled_datagroup.pdatagroup

		# The scheduled element is invalid or the pointed datagroup is no more valid (old configuration?)
		if not (scheduled_datagroup.is_valid() and pdg.is_valid()):
			# Reference to this object is not lost. Trashed by garbage collector
			try:
				del self.scheduled_datagroups[ pdg.uuid ]
			except KeyError, ex:
				pass
			return

		scheduled_datagroup.sched_state = POLLSTATE_QUEUED

		self.logger.debug('[scheduler] queued %s', scheduled_datagroup)

		self.threadpool.execJob(scheduled_handler, scheduled_params)

		return

	#-----------------------------------------------------------------------------#
	# THREAD POOL                                                                 #
	#-----------------------------------------------------------------------------#

	def _init_thread_pool(self, num_working_threads):

		self.logger.info( "Init thread pool (n.thread=%s)...", num_working_threads)
		self.threadpool = PollThreadPool(self.logger, worker_class=AgentWorkerThread)

		self.logger.info( "Creating workers (n.thread=%s)...", num_working_threads)
		self.threadpool.createWorkers(num_working_threads, pre_run_callback=self._init_working_thread_callback)


	def _init_working_thread_callback(self, thread_obj):

		log = thread_obj.getLogger()
		log.debug("Thread[%s] started (%s)", thread_obj.id, get_current_thread())

	#-----------------------------------------------------------------------------#
	# THREAD WORKER CALLBACK                                                      #
	#-----------------------------------------------------------------------------#

	def do_check_and_reschedule(self, workingthread, queue_delay, scheduled_datagroup):

		# Se mi sto fermando non controlliamo datagroup in attesa di elaborazione
		if self.is_stopping():
			return

		# Se mi hanno messo in pausa non controllo niente
		if self.is_paused():
			return

		t1 = get_time()
		self.logger.log(15, "[checker] executing %s", scheduled_datagroup)

		pdg = scheduled_datagroup.pdatagroup

		if not scheduled_datagroup.is_valid():
			self.logger.log(15, "[checker] scheduled datagroup %s old. skip", scheduled_datagroup)
			return

		scheduled_datagroup.sched_state = POLLSTATE_CHECKING


		# Global symbols for expression
		expr_symbols = {}
		expr_symbols.update( self.system_expr_functions )
		expr_symbols.update( self.extra_symbols )

		self.configuration.set_global_symbols(expr_symbols)

		#
		# Perform check
		#
		reschedule = True

		dgc = EffectiveDataGroupChecker(self.configuration, self.logger, self.cache_manager, self.ipc_factory, pdg)

		dgc.set_agent(self)

		try:
			workingthread.current_work_status = 'check'

			dgc.check( get_time(), scheduled_datagroup.schedule_time )

			workingthread.current_work_status = 'commit'

			scheduled_datagroup.sched_state = POLLSTATE_COMMITTING

			reschedule = self.process_datagroup_update_data(dgc)

		except ResourceNotFoundException, e:
			self.logger.warning("Error while checking: %s : %s", pdg.path, e)
			reschedule = False
			
		except Exception, e:
			self.logger.fatal("Unkown exception while checking datagroup %s : %s:%s -> %s", pdg.path, type(e), e, traceback.format_exc())

		# FIX: This should not be necessary
		dgc.dispose()

		self.logger.log(15, "[checker] finished %s.", scheduled_datagroup)

		if reschedule:
			self.reschedule(scheduled_datagroup)

		self.post_check_cleanup()

		workingthread.current_work_status = ''

		return

	def post_check_cleanup(self):
		# Fix: what to do?
		pass

	def process_datagroup_update_data(self, dgc):

		dg_data = dgc.get_datagroup_all_updates()

		self.send_queue.put_nowait(dg_data)

		return True  # reschedule with no problems

	def reschedule(self, scheduled_datagroup):

		if self.is_stopping():
			return

		pdatagroup = scheduled_datagroup.pdatagroup

		if pdatagroup.minperiod == 0:
			self.logger.info("Datagroup %s not rescheduled. Min-period = 0", pdatagroup.path)
			return

		seconds = pdatagroup.minperiod

		next_time = get_time() + seconds

		scheduled_datagroup.schedule_time = next_time
		scheduled_datagroup.sched_state   = POLLSTATE_SCHEDULED

		self.logger.debug("%s rescheduled at %s s (%s) ", scheduled_datagroup, seconds, time.ctime(next_time))

		params = (scheduled_datagroup,)
		self.scheduler.scheduleNewEvent( next_time , self.do_check_and_reschedule, params )

	#-----------------------------------------------------------------------------#
	#                                                                             #
	#-----------------------------------------------------------------------------#

	@staticmethod
	def encode_update_events(events):
		"""
		Serializza i dati *events* in una stringa di byte.
		"""
		#
		# IMPORTANT: "events" data could contain binary strings (Es: SNMP OID with MAC 
		#            values or binary data).
		#            If we stringify this data with json.dumps(), it tries to handle 
		#            them with UTF8 and it does not work.
		#            Try binary encoding pickle(+base64) encoding as fallback.
		#
		try:
			return json.dumps(events)
		except Exception, e:
			return base64.b64encode(pickle.dumps(events, pickle.HIGHEST_PROTOCOL))

	@staticmethod
	def decode_update_events(encoded_events):
		"""
		Deserializza i dati di aggiornamento *encoded_events* in memoria.
		"""
		#
		# IMPORTANT: see  encode_update_events()
		#
		try:
			return json.loads(encoded_events)
		except Exception, e:
			return pickle.loads(base64.b64decode(encoded_events))

	#-----------------------------------------------------------------------------#
	#                                                                             #
	#-----------------------------------------------------------------------------#

	def handleEvents(self):

		self.logger.info("[event] Start...")

		self.event_connection = None

		while 1: 

			if self.is_stopping():
				self.logger.info("[event] stopping...")
				break

			if self.is_paused():
				break

			try:
				self.logger.info("[event] Creating connection...")
				self.event_connection = self.create_connection()
				self.event_connection.open()

				#
				# The connection is established
				#
				self.process_pending_events()

				self.event_connection.close()

			except IPCConnectionError, e:
				self.logger.error("[event] IPC Connection Error on event: %s", e)

			except IPCException, e:
				self.logger.error("[event] IPC Error on event connection: %s", e)

			except Exception, e:
				self.logger.fatal("[event] Fatal Error while sending data to server: %s:%s -> %s", type(e), str(e), traceback.format_exc())

			finally:
				if self.event_connection is not None:
					try:
						self.logger.debug("[event] Trying to close connection")
						self.event_connection.close()
					except Exception, e:
						self.logger.error("[event] Error when closing connection: %s:%s -> %s", type(e), str(e), traceback.format_exc())

			self.logger.debug("[event] Restarting...")
			time.sleep(1)

		self.logger.info("[event] exiting")

	def process_pending_events(self):

		self.last_send_ts   = 0
		self.pending_events = []

		while 1:
			if self.is_stopping():
				return

			if self.is_paused():
				break

			dg_update_data = None
			try:
				dg_update_data = self.send_queue.get(True, self.send_queue_timeout) # block waiting
			except Queue.Empty, er:
				pass
				
			now = time.time()

			force_send = 0
			if not dg_update_data:
				if len(self.pending_events) == 0:
					self.logger.log(1, "[event] empty queue")
					continue
				else:
					force_send = 1
			
			if (now - self.last_send_ts) > 5.0:
				force_send = 1

			if self.is_stopping():
				self.logger.info("[event] stopping...exit loop")
				break

			if dg_update_data:
				update_event = AgentUpdateEvent( get_time(), self.tenantid, self.agentid, dg_update_data)
				self.pending_events.append(update_event)

			if len(self.pending_events) >= self.max_num_pending_event or force_send == 1:
				
				self.delivery_update_events(self.pending_events)
				
				self.pending_events = []
				self.last_send_ts   = time.time()

		self.logger.info("[event] waiting for resume")
		self.wait_if_paused()

	def delivery_update_events(self, events):

		t1 = time.time()

		# VERY IMPORTANT 
		events = BaseAgent.encode_update_events(events)

		t2 = time.time()

		self.event_connection.send_msg( 'sanetd-events', events, blocking=False)

		t3 = time.time()

		self.logger.log(19, "Send data (%s) took %.3f %.3f %.3f", len(events), t3-t1, t2-t1, t3-t2)

