# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import os
import time
import traceback
import json

from sanet_common.osutils.osutils import get_proc_memory, get_proc_resident

# da fixare
from sanet_poller.agent import BaseAgent, DECOMPRESS_DATA, ScheduledDatagroup
from sanet_poller.lib.classes import POLLSTATE_SCHEDULED, POLLSTATE_QUEUED, POLLSTATE_IDLE, POLLSTATE_ORGANIZING
from sanet_poller.lib.utils import get_time
from sanet_poller.lib.classes import PSchedulingGroup, PDataGroup, PCluster, PNode, PIface
from sanet_poller.lib.classes import PollerConfigurationWithSchedulingGroups
from sanet_poller.lib.exceptions import SchedulingGroupNotFoundException
from sanet_poller.lib.threadutils import PollOrganizer

from sanet_common.ipc import encode_response_with_id
from sanet_common.ipc import RESPONSE_SUCCESS

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def AUTOMATIC_SCHEDULING_GROUP_UUID_NAME(uuid, name):
	return ('_' + name, '_' + uuid)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class BaseAgentAdvancedScheduling(BaseAgent):

	def __init__(self):
		return

	# MODIFICATO
	def setup(self, *args, **kwargs):

		super(BaseAgentAdvancedScheduling, self).setup(*args, **kwargs)

		# parametri di configurazione dal server
		self.scheduling_min_gap      = None
		self.scheduling_max_parallel = None
		self.scheduling_max_deadline = None

		# Organizer thread
		self.organizer = None

	#-----------------------------------------------------------------------------#
	#                                                                             #
	#-----------------------------------------------------------------------------#

	# riscritto perche' e' importante avviare prima l'organizer
	def init_data(self):

		self._load_modules()
		self._init_thread_pool(self.num_working_threads)
		self._init_organizer()
		self._init_scheduler()
		return

	#-----------------------------------------------------------------------------#
	#                                                                             #
	#-----------------------------------------------------------------------------#

	# MODIFICATO
	def _update_settings_from_config(self, config):
		
		super(BaseAgentAdvancedScheduling, self)._update_settings_from_config(config)

		# Parametri di scheduling globali
		self.scheduling_min_gap      = config['scheduling_min_gap'     ]
		self.scheduling_max_parallel = config['scheduling_max_parallel']
		self.scheduling_max_deadline = config['scheduling_max_deadline']

		self.logger.info("Agent global config %s %s %s", self.scheduling_min_gap, self.scheduling_max_parallel, self.scheduling_max_deadline)

		return

	#-----------------------------------------------------------------------------#
	#                                                                             #
	#-----------------------------------------------------------------------------#

	# riscritto perche' e' importante caricare gli scheduling group prima di ogni altra cosa.
	def _load_all_configuration(self):

		t0 = t1 = t2 = t3 = t4 = t5 = t6 = t7 = t8 = t9 = t10 = t11 = t12 = t13 = t14 = 0

		try:
			old_configuration = self.configuration

			#new_config = PollerConfiguration(self.logger, self.tenantid)
			new_config = PollerConfigurationWithSchedulingGroups(self.logger, self.tenantid)

			self.logger.info("[config] Loading....")

			t0 = time.time()
			self.update_info   = 'loading scheduling groups..'
			self._update_scheduling_groups(new_config)

			t1 = time.time()
			self.update_info   = 'loading clusters...'
			self._update_clusters(new_config)

			t2 = time.time()
			self.update_info   = 'loading nodes...'
			self._update_nodes(new_config)

			t3 = time.time()
			self.update_info   = 'loading ifaces...'
			self._update_ifaces(new_config)

			t4 = time.time()
			self.update_info   = 'loading storages...'
			self._update_storages(new_config)

			t5 = time.time()
			self.update_info   = 'loading services...'
			self._update_services(new_config)

			t6 = time.time()
			self.update_info   = 'loading devices...'
			self._update_devices(new_config)

			t7 = time.time()
			self.update_info   = 'loading links...'
			self._update_links(new_config)

			t8 = time.time()
			self.update_info   = 'loading datasources...'
			self._update_datasources(new_config)

			t9 = time.time()
			self.update_info   = 'loading conditions...'
			self._update_conditions(new_config)

			t10 = time.time()
			self.update_info   = 'loading parameters..'
			self._update_datagroups_parameters(new_config)

			t11 = time.time()

			#self.update_info   = 'loading scheduling groups..'
			#self._update_scheduling_groups(new_config)

			t12 = time.time()

			new_config.reindex_datagroups()

			self.update_info   = 'loading datagroups...'

			datagroup_to_schedule = self._update_datagroups(new_config)

			t13 = time.time()

			new_config.set_load_time( get_time() )

			self.logger.info("[config] configuration updated. (datagroup to schedule %s)", len(datagroup_to_schedule))

			#
			# La configurazione scaricata diventa quella "valida"
			#
			self.configuration = new_config

			self.schedule_datagroups(datagroup_to_schedule)

			# Mentre stiamo aggiornando la configurazione il sistema non e'
			# fermo. Assicuriamoci che gli elementi della vecchia configurazione
			# non vengano piu' usati da nessuno.
			#
			# NOTA: Datagroup gia' schedulati devono controllare
			# prima di ogni check' che il datagroup puntato sia ancora
			# valido.
			if old_configuration is not None:
				old_configuration.invalid_all_elements()

			t14 = time.time()

			return
		finally:
			self.logger.log(19, "Config update: Times:")
			self.logger.log(19, "Config update:     sch. groups %5.3f" ,  t1 -  t0)
			self.logger.log(19, "Config update:     cluster     %5.3f" ,  t2 -  t1)
			self.logger.log(19, "Config update:     nodes       %5.3f" ,  t3 -  t2)
			self.logger.log(19, "Config update:     iface       %5.3f" ,  t4 -  t3)
			self.logger.log(19, "Config update:     storage     %5.3f" ,  t5 -  t4)
			self.logger.log(19, "Config update:     service     %5.3f" ,  t6 -  t5)
			self.logger.log(19, "Config update:     devices     %5.3f" ,  t7 -  t6)
			self.logger.log(19, "Config update:     links       %5.3f" ,  t8 -  t7)
			self.logger.log(19, "Config update:     datasources %5.3f" ,  t9 -  t8)
			self.logger.log(19, "Config update:     conditions  %5.3f" , t10 -  t9)
			self.logger.log(19, "Config update:     parameters  %5.3f" , t11 - t10)
			self.logger.log(19, "Config update:     UPDATE      %5.3f" , t13 - t12)
			self.logger.log(19, "Config update:     RESCHEDULE  %5.3f" , t14 - t13)

	# MODIFICATO
	def schedule_datagroups(self, datagroup_to_schedule):

		now = get_time()

		for sdg in datagroup_to_schedule:
			
			if sdg.pdatagroup.scheduling_group is not None:
				
				self.logger.info("[INIT->group] %s", sdg.pdatagroup.path)
				
				pdatagroup = sdg.pdatagroup
				
				# Mai scheudlato prima (NUOVO?)
				if pdatagroup.lastdone is None:
					# Lo schedulo nel gruppo ADESSO (verra' rispettato il mingap)
					next_timeout = now
				else:
					# Lo schedulo nel gruppo in un momento nel futuro (verra' rispettato il mingap)
					# delta   = (now - pdatagroup.lastdone)
					# next_timeout = pdatagroup.minperiod - delta
					# next_timeout = int(max(next_timeout, 0))
					if now - pdatagroup.lastdone < pdatagroup.minperiod:
						next_timeout = now + pdatagroup.minperiod
					else:
						next_timeout = now
				
				params = ( sdg, )
				sdg.pdatagroup.scheduling_group.scheduleJob( 
					next_timeout,
					self.scheduler.scheduleNewEvent,
					self.do_check_and_reschedule, 
					params
				)
				self.logger.debug("SCHEDULED IN %s: %s", next_timeout, sdg)

			else:
				# Scheduling iniziale classico
				super(BaseAgentAdvancedScheduling, self).schedule_one_datagroup(now, sdg)

	# NUOVO
	def _update_scheduling_groups(self, new_config):

		data = self._send_request(self.update_connection, 'get_scheduling_groups', self.tenantid, self.agentid)
		data = DECOMPRESS_DATA(data)
		self._update_scheduling_groups_from_data(new_config, data)


	# NUOVO
	def _update_scheduling_groups_from_data(self, new_config, data):

		self.update_info   = 'updating scheduling groups...'

		self.logger.log(9, "Loading %s scheduling groups", len(data))
		for d in data:
			self.logger.log(9, "Loading scheduling groups %s", d['name'])

			# In fase di aggiornamento della configurazione, gli schedulingroup gia' creati vengono
			# presi e riutilizzati senza essere ricreati ogni volta da zero, modificando soltanto i
			# parametri di scheduling.
			# Questo evita che i dati sullo scheduling corrente vengano "persi" ogni volta che
			# la configurazione viene aggiornata.
			psg = None
			if self.configuration is not None:
				try:
					psg = self.configuration.scheduling_groups[ d['uuid'] ]
					#self.logger.log(9, "\t" + "%s taken from previous configuration", psg.name)

				except KeyError, e:
					psg = PSchedulingGroup(logger=self.logger)

			if psg is None:
				psg = PSchedulingGroup(logger=self.logger)

			psg.load_from_dict(d)
			new_config.add_scheduling_group(psg)


	# NUOVO
	def _create_implicit_scheduling_group(self, new_config, entity):
		"""
		La variabili "entity" puo' essere un 
			- PCluster
			- PElement 
		"""

		# Se l'agente ha dei parametri di scheduling globali bisogna creare
		# scheduling group impliciti a livello di cluster o nodo.
		if self.scheduling_min_gap is not None or self.scheduling_max_parallel is not None:

			self.logger.debug("Creating implicit scheduling group for: %s", entity)

			fake_uuid, fake_name = AUTOMATIC_SCHEDULING_GROUP_UUID_NAME(entity.uuid, entity.name)

			# In fase di aggiornamento della configurazione, gli schedulingroup gia' creati vengono
			# presi e riutilizzati senza essere ricreati ogni volta da zero. 
			# Questo evita che i dati sullo scheduling corrente vengano "persi" ogni volta che
			# la configurazione viene aggiornata.
			psg = None
			if self.configuration is not None:
				try:
					psg = self.configuration.scheduling_groups[ fake_uuid ]
					self.logger.log(9, "\t" + "taken from previous configuration", psg.name)
				except KeyError, e:
					psg = PSchedulingGroup(logger=self.logger)
					psg.uuid         = fake_uuid
					psg.name         = fake_name

			if psg is None:
				psg = PSchedulingGroup(logger=self.logger)
				psg.uuid = fake_uuid
				psg.name = fake_name

			psg.min_gap      = self.scheduling_min_gap
			psg.max_deadline = self.scheduling_max_deadline
			psg.max_parallel = self.scheduling_max_parallel

			self.logger.info("Created implicit schedulinggroup %s:%s", psg.uuid, psg.name)

			new_config.add_scheduling_group(psg)

			entity.scheduling_group_id = fake_uuid


	# MODIFICATO
	def _update_clusters_from_data(self, new_config, data):

		self.update_info   = 'updating clusters...'

		self.logger.log(9, "Loading %s clusters", len(data))
		for d in data:
			self.logger.log(9, "Loading cluster %s", d)

			pc = PCluster()
			pc.load_from_dict(d)

			new_config.add_cluster(pc)

			# Se serve, creiamo uno schedulinggroup implicito a livello di nodo
			if pc.scheduling_group_id is None:
				self._create_implicit_scheduling_group(new_config, pc)


	# MODIFICATO
	def _update_nodes_from_data(self, new_config, data):

		self.update_info   = 'updating nodes...'

		self.logger.log(9, "Loading %s nodes", len(data))
		for d in data:
			self.logger.log(9, "Loading node %s", d)

			pn = PNode()
			pn.load_from_dict(d)

			new_config.add_element(pn)

			# Se serve, creiamo uno PSchedulingGroup implicito a livello di nodo
			#
			# Se il nodo appartiene ad un cluster, il suo PSchedulingGroup e' quello del cluster:
			#    a) e' gia' stato definito configurato esplicitamente uno PSchedulingGroup (configurazione dal server)
			#    b) e' gia' stato creato (!= None) dall'agente uno PSchedulingGroup esplicito quando abbiamo creato il cluster.
			#
			if pn.scheduling_group_id is None:

				if pn.cluster_id is not None:

					pcluster = new_config.get_cluster(pn.cluster_id) # get_cluster() non dovrebbe mai dare errore qui

					pn.scheduling_group_id = pcluster.scheduling_group_id # va bene anche None
				else:
					self._create_implicit_scheduling_group(new_config, pn)

	# MODIFICATO
	def _update_ifaces_from_data(self, new_config, data):

		self.update_info   = 'updating ifaces...'

		self.logger.log(9, "Loading %s ifaces", len(data))
		for d in data:
			self.logger.log(9, "Loading iface %s", d)

			pi = PIface()
			pi.load_from_dict(d)

			new_config.add_element(pi)

			# In assenza di configurazioni esplicite, il PSchedulingGroup group di un sotto-elemento e' lo PSchedulingGroup del nodo padre
			# (creato esplicitamente o implicitamente prima di arrivare qui).
			if pi.scheduling_group_id is None:
				pn = new_config.get_element_by_id(pi.parent_id) # NOn dovrebbe mai andare in errore. Il nodo deve esistere.
				pi.scheduling_group_id = pn.scheduling_group_id

	# MODIFICATO
	def _update_element_from_data(self, new_config, data, element_class):

		self.update_info   = 'updating elements...'

		self.logger.log(9, "Loading %s elements", len(data))
		for d in data:
			self.logger.log(9, "Loading %s", d)
			e = element_class()
			e.load_from_dict(d)
			new_config.add_element(e)

			# In assenza di configurazioni esplicite, il PSchedulingGroup group di un sotto-elemento e' lo PSchedulingGroup del nodo padre
			# (creato esplicitamente o implicitamente prima di arrivare qui).
			if e.scheduling_group_id is None:
				pn = new_config.get_element_by_id(e.parent_id) # NOn dovrebbe mai andare in errore. Il nodo deve esistere.
				e.scheduling_group_id = pn.scheduling_group_id


	# MODIFICATO
	def _update_datagroups_from_data(self, new_config, data):

		self.update_info = 'updating datagroups...'

		self.logger.log(9, "Loading %s datagroups", len(data))

		now = get_time()

		datagroup_to_schedule = []

		for d in data:
			self.logger.log(9, "Loading datagroup %s", d)

			if d['element'] not in new_config.elements:
				self.logger.error("Element %s for datagroup %s is missing in new poller configuration. Skipping.", d['element'], d['path'])
				continue

			pdg = PDataGroup()
			pdg.load_from_dict(d)

			pdg.datasources = new_config.get_datasources_ids_by_dg_id(pdg.uuid)
			pdg.conditions  = new_config.get_conditions_ids_by_dg_id(pdg.uuid)


			#
			# Assegnazione dello PSchedulingGroup corrente
			#

			# Se pdg.scheduling_group_id non c'e' una configuazione di scheduling
			# esplicita ricevuta dal server.
			#
			# Se questo agente ha dei parametri di scheduling, allora 
			# per l'elemento di monitoraggio associato al datagroup esiste un
			# PSchedulingGruop creato dall'agente e dobbiamo usare quello.

			if pdg.scheduling_group_id is None:
				pelement = new_config.get_element_by_id(pdg.element)
				pdg.scheduling_group_id = pelement.scheduling_group_id

			# Il gruppo c'e'. Prendiamo l'oggetto dalla configurazione
			if pdg.scheduling_group_id is not None:
				new_scheduling_group = new_config.get_scheduling_group(pdg.scheduling_group_id)
			else:
				new_scheduling_group = None
			#else:
			#	try:
			#		print "********** NODE FAKE GROUP", 'fake_' + pdg.element, pdg.element
			#		new_scheduling_group = new_config.get_scheduling_group('fake_' + pdg.element)
			#		print "\t", "FOUND"
			#	except SchedulingGroupNotFoundException, er:
			#		try:
			#
			#			if pelement.cluster_id is not None:
			#				print "********** NODE CLUSTER GROUP", 'fake_' + pdg.element, pdg.element
			#				new_scheduling_group = new_config.get_scheduling_group('fake_'+pelement.cluster_id)
			#				print "\t", "FOUND"
			#
			#		except SchedulingGroupNotFoundException, er:
			#			pass

			if self.configuration:
				old_scheduling_group = self.configuration.get_scheduling_group_by_dg_id(pdg.uuid)
				if old_scheduling_group is not None:
					# Lo scheduling corrente e':
					#    a) diverso dal precedente
					#    b) nullo (il datagroup non e' piu' in uno scheduling group)
					if old_scheduling_group.uuid != pdg.scheduling_group_id:
						# Pulisci lo scheduling groupa  cui apparteneva lo PScheduledDatagroup
						old_scheduling_group.remove_sdg_from_scheduled_queue(pdg.uuid)

			pdg.scheduling_group = new_scheduling_group
			
			#
			# FINE
			#

			new_config.add_datagroup(pdg)

			was_scheduled = False

			valid_for_check = pdg.valid_for_check()

			if not valid_for_check:
				self.logger.info("Datagroup loaded %s but not valid for check", pdg.path)

			if pdg.minperiod <= 0:
				self.logger.log(9, "Datagroup loaded %s but minperiod=0", pdg.path)
				valid_for_check = False

			#
			# Ora che l'oggetto PDatagroup e' sicuro in memoria
			# vediamo se deve essere schedulato o meno.
			#
			sdg = self.scheduled_datagroups.get(pdg.uuid, None)

			self.logger.debug("update datagroups from data. sdg: %s (%s)", pdg.path, sdg != None )

			# se il datagroup e' in esecuzione, non ha ancora mandato i dati al server centrale
			# quindi quelli che sto ricevendo non sono aggiornati
			if sdg is None:
				sdg = ScheduledDatagroup(pdg)
				self.scheduled_datagroups[pdg.uuid] = sdg
			else:
				# Esiste gia' un PDatagroup "vecchio" gia' caricato e schedulato (c'e' il suo ScheduledDatagroup).
				# Aggiorniamo il rifermento in modo che il nuovo PDatagroup sia gia' pronto all'uso
				# nel momento in cui lo scheduler lo dovra' far deschedulare il corrispondente ScheduledDatagroud.

				old_minperiod = sdg.pdatagroup.minperiod

				# IMPORTANTE: non spostare questa riga prima del controllo del minperiod
				sdg.pdatagroup = pdg

				if sdg.sched_state != POLLSTATE_IDLE:
					was_scheduled = True

				# IMPORTANTE: Il minperiod e' cambiato ed il vecchio aveva uno scheduling piu' "lento"!
				#             Invece di aspettare il prossimo check per cominciare ad usare 
				#             il nuovo "minperiod", possiamo deschedulare 
				#             e schedulare (che usera' il nuovo minperiod).
				if old_minperiod > pdg.minperiod:

					self.logger.info("Datagroups %s changed minperiod! ", pdg.path)

					# Lo schedule datagroup potrebbe essere:
					# - schedulato (quindi nello scheduler principale)
					# - in attesa di esecuzione nel suo scheduling group
					# - in esecuzione nel thread pool
					#
					# Cerchiamo di rimuoverlo dallo scheduler. 
					fun_match_data = lambda event_data: (event_data[0].pdatagroup.path == sdg.pdatagroup.path)
					removed_element = self.scheduler.removeEvent(self.do_check_and_reschedule, fun_match_data)
					if removed_element is not None:
						self.logger.info("Rescheduling datagroup %s with new minperiod.", pdg.path)
						# L'elemento era nello scheduler. Quindi ora non lo e' piu' 
						# e nessuno lo sta "toccando" quindi possiamo
						# (ri)schedularlo senza rompere le scatole agli altri
						# thread attivi nell'agente ( organizer, threadpool, ecc.)
						sdg.sched_state = POLLSTATE_IDLE
						was_scheduled = False

			if valid_for_check and was_scheduled == False:
				datagroup_to_schedule.append(sdg)

				self.logger.log(9, "loaded for schedule %s:%s", pdg.uuid, pdg.path)

		return datagroup_to_schedule

	#-----------------------------------------------------------------------------#
	# ORGANIZER                                                                   #
	#-----------------------------------------------------------------------------#

	def _init_organizer(self):

		self.logger.debug("Init organizer...")
		self.organizer = PollOrganizer(self.logger, self._exec_job_check_callback )

		self.logger.debug("Starting organizer...")
		self.organizer.start()


	def _exec_job_check_callback(self, handler, params):
		
		try:
			scheduled_datagroup = params[0]

			pdg = scheduled_datagroup.pdatagroup

			# controllare il max_deadline
			if pdg.scheduling_group:
				# Make scheduling group handle concurrency
				#self.logger.log(11, "[organizer -> group] %s", pdg.path)
				pdg.scheduling_group.exec_or_idle(handler, params, self.threadpool.execJob)
			else:
				# Send to thread worker for execution
				#self.logger.log(11, "[organizer -> threadpool] %s", pdg.path)
				self.threadpool.execJob(handler, params)

		except Exception, e:
			self.logger.fatal("%s:%s -> %s", type(e), str(e), traceback.format_exc())

	# MODIFICATO
	def _exec_scheduled_job_callback(self, scheduled_handler, scheduled_params):
		"""
		Handler da passare allo scheduler

		@param scheduled_handler:
		@param scheduled_params: tupla con dentro uno ScheduledDatagroup object
		"""


		# Se mi sto fermando semplicemente scarto quello che c'era da fare
		if self.is_stopping():
			return

		# Se mi hanno messo in pausa scarto quello che c'e' da fare
		if self.is_paused():
			return

		scheduled_datagroup = scheduled_params[0]

		pdg = scheduled_datagroup.pdatagroup

		#self.logger.log(21,'[scheduler] wakeup. time to execute %s', pdg.path)

		# The scheduled element is invalid or the pointed datagroup is no more valid (old configuration?)
		if not (scheduled_datagroup.is_valid() and pdg.is_valid()):
			# Reference to this object is not lost. Trashed by garbage collector
			try:
				psg = pdg.scheduling_group

				if psg:
					psg.remove_sdg_from_scheduled_queue(psg.uuid)

				del self.scheduled_datagroups[ pdg.uuid ]
			except KeyError, ex:
				pass
			return

		scheduled_datagroup.sched_state = POLLSTATE_ORGANIZING

		#self.logger.log(11, '[sched->organizer] queued %s', scheduled_datagroup)

		#self.organizer.inputQueue.put( (scheduled_handler, scheduled_params) )
		self.organizer.newJob( scheduled_handler, scheduled_params )

		return

	#-----------------------------------------------------------------------------#
	# THREAD WORKER CALLBACK                                                      #
	#-----------------------------------------------------------------------------#

	# MODIFICATO
	def reschedule(self, scheduled_datagroup):

		if scheduled_datagroup.pdatagroup.scheduling_group:

			if self.is_stopping():
				return

			#self.logger.log(11, "[checker->group] Rescheduling %s", scheduled_datagroup.pdatagroup.path)

			scheduled_datagroup.pdatagroup.scheduling_group.scheduleJobAndUnlockPending(
				self.scheduler.scheduleNewEvent,
				self.threadpool.execJob,
				self.do_check_and_reschedule,
				(scheduled_datagroup,),
			)
		else:
			super(BaseAgentAdvancedScheduling, self).reschedule(scheduled_datagroup)

		return

	# MODIFICATO
	def cmd_info(self, request_id, params):
		"""
		Returns human-readable informations about agent's internal status.
		"""
		self.logger.info("Received info command: %s", params)

		text = []

		text += [ self.threadpool.get_info_text() ]

		text += [ 'Scheduled datagroups: %s' % (self.scheduler.get_schedule_size(), ) ]

		text += self.get_organizer_status_text()

		text += self.get_queues_status_text()

		text += self.get_configuration_status_text()

		text = '\n'.join(text)

		self.logger.info("Sending response: %s:%s", request_id, text)
		response = encode_response_with_id( RESPONSE_SUCCESS, request_id, text )

		return response

	# MODIFICATO
	def _infojson(self, request_id, params):
		#
		# Calculate data
		#
		infos  = self.get_configuration_status()

		infos2 = self.get_queues_status()

		infos3 = self.threadpool.get_info()

		# Con lo scheduling avanzato ci sono due code di attesa:
		#     1) Scheduluer -----[coda]---->Organizer
		#     2) Organizer  -----[coda]---->ThreadPool
		infos4 = self.get_organizer_status()

		wqueuelen = infos3.get('wait_queue', 0) + infos4.get('queue_size', 0)

		vmsize = get_proc_memory()
		rmsize = get_proc_resident()

		#
		# Calculate critical problems
		#
		critical_problems = {}
		critical_problems.update( infos ['critical_problems'] )
		critical_problems.update( infos2['critical_problems'] )
		critical_problems.update( infos3['critical_problems'] )

		#
		# Prepare response
		#
		data = {
			'nodes'      : infos.get('nodes_count'    , None),
			'ifaces'     : infos.get('ifaces_count'   , None),
			'storages'   : infos.get('storages_count' , None),
			'services'   : infos.get('services_count' , None),
			'devices'    : infos.get('devices_count'  , None),

			'datagroups' : infos.get('datagroups_count' , None),
			'datasources': infos.get('datasources_count', None),
			'conditions' : infos.get('conditions_count' , None),

			'dgrate'     : infos3.get('rate', None),

			'wqueuelen'  : wqueuelen,

			'squeuelen'  : infos2.get('send_queue', None),

			'vmemory'    : int(vmsize),
			'rmemory'    : int(rmsize),

			'critical_problems': critical_problems,
		}

		data.update( 
			dict([ ('threadpool.process.%s' % (k), v) for k,v in infos3['process'].items() ])
		)

		return data

	# NUOVO
	def get_organizer_status_text(self):

		infos = self.get_organizer_status()

		text  = []
		text += [ 
			'Post-schedule Queue: %s (%s/s) (delay %f)' % (infos['queue_size'], infos['run_rate'], infos['run_delay']) 
		]

		return text

	# NUOVO
	def get_organizer_status(self):

		infos = self.organizer.get_info()

		#
		# FIX: ci sono dei problemi nell'organizer che dobbiamo comunicare al watchdog?
		#      Se si, dobbiamo gestirli con 
		#
		#           infos['critical_problems'] = ....
		#
		#      e rivedere come gestirli in   self.cmd_infojson().

		return infos


	## MODIFICATO
	#def cmd_check_now(self, request_id, params):
	#	"""Execute a scheduled datagroup immediately."""
	#	path     = params[0]
	#	return self._reschedule(request_id, path, None)

	# MODIFICATO
	def _reschedule(self, request_id, path, new_time):
		"""
		Cerca lo ScheduledDataGroup per il datagroup "path" e lo rischedula al tempo "next_time".
		
		Se "next_time" e' nullo, il datagroup viene rischedulato per l'esecuzione immediata.
		
		Se il datagroupa appartiene ad un gruppo di scheduling, l'esecuzione immediata 
		ignora i vincoli di esecuzione del gruppo.
		"""
		#
		# Troviamo e rimuoviamo lo ScheduledDataGroup nell scheduler principale
		#
		handler        = self.do_check_and_reschedule
		fun_match_data = lambda event_data: (event_data[0].pdatagroup.path == path)  #  event_data = ( ts , ScheduledDatagroup obj )

		removed_scheduler_event = self.scheduler.removeEvent( handler, fun_match_data)

		# Trovato nello scheduler?
		if removed_scheduler_event:

			# removed_scheduler_event = ( s_handler, (scheduled_datagroup, ) )
			sdg = removed_scheduler_event[1][0]

			self.logger.info("Forced check for %s (next_time=%s)", sdg.pdatagroup.path, new_time)

			# Appartiene ad un gruppo di scheduling?
			psg = sdg.pdatagroup.scheduling_group
			if psg:

				if new_time is not None:
					# Reschedule nel futuro, devo schedulare il momento "giusto" nel gruppo.
					# psg.remove_sdg_from_scheduled_queue(event.pdatagroup.uuid)  <------- viene gia' fatto da  scheduleJob()
					sdg.pdatagroup.scheduling_group.scheduleJob(
						new_time,
						self.scheduler.scheduleNewEvent,
						self.do_check_and_reschedule,
						(sdg,)
					)
				else:
					# FIX: fare un metodo a parte
					psg.queueLock.acquire()
					psg.remove_sdg_from_scheduled_queue(sdg.pdatagroup.uuid)
					psg.executing += 1
					psg.queueLock.release()

					# Esecuzione immediata.
					sdg.schedule_time = time.time()         # per coerenza facciamo come se l'esecuzione fosse stata schedulata proprio per "adesso"
					sdg.sched_state   = POLLSTATE_QUEUED
					self.threadpool.execJob(handler, (sdg,))

			else:
				if new_time is not None:
					# Rescheduling nel futuro. Reinserisco nello scheduler principale con il tempo nuovo.
					sdg.schedule_time = new_time
					sdg.sched_state   = POLLSTATE_SCHEDULED
					self.scheduler.scheduleNewEvent(new_time, self.do_check_and_reschedule, (sdg,))

				else:
					# Esecuzione immediata.
					sdg.schedule_time = time.time()         # per coerenza facciamo come se l'esecuzione fosse stata schedulata proprio per "adesso"
					sdg.sched_state   = POLLSTATE_QUEUED
					self.threadpool.execJob(handler, (sdg,))

		response = encode_response_with_id(RESPONSE_SUCCESS, request_id, "rescheduled" if removed_scheduler_event else "not changed")
		return response

	def cmd_scheduling_group_info(self, request_id, params):
		"""
		Returns the list of current scheduling groups or details about the specified scheduling group.
		"""

		configuration = self.configuration

		cmd_params = params[0:-1]

		if len(cmd_params) > 0:
			sg_name_or_uuid = cmd_params[0]
		else:
			sg_name_or_uuid = None

		if sg_name_or_uuid is None:
			FORMAT = '%-35s %-50s %3s %3s %3s' 
			data = ''
			data += FORMAT % ('UUID','NAME', 'G', 'P', "D")
			data += '\n'
			for uuid, psg in self.configuration.scheduling_groups.items():
				data += FORMAT % (uuid, psg.name, psg.min_gap or '', psg.max_parallel or '', psg.max_deadline or '')
				data += '\n'
		else:
			now = time.time()
			data = "*NOT FOUND*"
			for sg in self.configuration.scheduling_groups.values():

				if sg.name == sg_name_or_uuid or sg.uuid == sg_name_or_uuid:
					data = ''
					data += 'GROUP: %s (UUID: %s)' % (sg.uuid, sg.name)
					data += '\n'
					data += 'min-gap     : %s' % (sg.min_gap)
					data += '\n'
					data += 'max-parallel: %s' % (sg.max_parallel)
					data += '\n'
					data += 'max-deadline: %s' % (sg.max_deadline)
					data += '\n'
					data += 'timeline'
					data += '\n'

					for ts, sdg in sorted(sg.scheduledQueue):
						s = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(ts))
						data += '\t%s %s (%3s seconds)' % (s, sdg.pdatagroup.path, int(ts-now))
						data += '\n'

					sg.queueLock.acquire()

					data += 'Lock queue:\n'
					for handler, params in sg.lockedQueue:
						sdg = params[0]
						data += '\t%s' % (sdg.pdatagroup.path)
						data += '\n'
					sg.queueLock.release()

					break

		response = encode_response_with_id(RESPONSE_SUCCESS, request_id, data)
		return response
