# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import os
import time
import sys
import getopt
import socket

from sanet_common.uuids import create_global_uuid

from sanet_common.ipc import IPCConnection
from sanet_common.ipc import IPCNotFound
from sanet_common.ipc import IPCException

from sanet_common.ipc import RESPONSE_SUCCESS
from sanet_common.ipc import decode_response, encode_request, encode_request_with_id

from sanet_common.sanetdefs.ipc import TEMP_REPLY_QUEUE, AGENT_QUEUE_SUBSYSTEM

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

DEFAULT_AGENT_CTL_TIMEOUT = 3 #seconds

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

LOCAL_TCP_ADDRESS = '127.0.0.1'

def local_tcp_port_from_agent_name(agent_name):
	"""
	Calculate a TCP port number from the agent id string
	"""
	return 6666 + sum([ord(c) for c in agent_name])

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

# Un agente ha piu' componenti (subsystem) che possono restare in ascolto ed accettare
# richieste.

AGENT_SUBSYSTEM_CTL = 'ctl'

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def agent_ctl(broker, agentid, cmd, params, timeout=DEFAULT_AGENT_CTL_TIMEOUT, blocking=False):
	
	return agent_sendreceive(broker, agentid, AGENT_SUBSYSTEM_CTL, cmd, params, timeout=timeout, blocking=blocking)

def agent_sendreceive(broker, agentid, subsystem, cmd, params, timeout=DEFAULT_AGENT_CTL_TIMEOUT, blocking=False):
	
	agent_address = AGENT_QUEUE_SUBSYSTEM(agentid, subsystem)
	
	host = None
	port = None
	user = None
	password = None
	
	auth, addr = broker.split('@')
	user,password = auth.split('/')
	addr = addr.split(':')
	host = addr[0]
	if len(addr) > 1:
		port = int(addr[1])

	connect_options = {}
	if port     != None: connect_options['port']     = port
	if user     != None: connect_options['username'] = user
	if password != None: connect_options['password'] = password
	
	# OPEN
	c = IPCConnection( host, **connect_options)
	c.open()
	
	# BIND
	replyid  = TEMP_REPLY_QUEUE( 'agentctl-%s' % (create_global_uuid()) ) 
	c.bind(replyid)

	requestid = str(time.time())
	
	# SEND
	try:
		request = encode_request_with_id( cmd, requestid, *(params+[ time.time() ]) )
		c.send_msg( agent_address, request, blocking=blocking)
	except IPCNotFound, e:
		raise Exception("Unable to send message to agent %s: %s" % (agentid, e))
	except IPCException, e:
		raise Exception("Generic error while sending data to agent %s: %s" % (agentid, e))

	# RECEIVE
	try:
		response = None
		(msg_addr, msg) = c.receive_msg(ack_msg=True, timeout=timeout)
		if msg is None: # timeout
			raise Exception("Timeout")
			
	except IPCException, e:
		raise Exception("Generic error on receive from agent %s: %s" % (agentid, e))

	# CLOSE
	c.close()
	

	response = decode_response(msg)
	if response['status'] == RESPONSE_SUCCESS:
		if response['id'] != requestid:
			raise Exception("Received response with id %s instead of %s " % (response['id'], requestid))
		return response['content']
	else:
		raise Exception("Error response from server: %s" % (response['content']))
		

	
	return None

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def agent_ctl_local(agent_name, cmd, params, timeout=DEFAULT_AGENT_CTL_TIMEOUT, blocking=False):

	host = LOCAL_TCP_ADDRESS
	port = local_tcp_port_from_agent_name(agent_name)
	
	client_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
	client_socket.connect((host, port))
	
	request = encode_request( cmd, *(params+[ time.time() ]) )
	
	try:
		client_socket.send( request )
	except socket.error, e:
		raise Exception("Unable send message to agent: %s" % (e))

	try:
		response = ''
		while True:
			data = client_socket.recv(1024)
			if not data: # closed connection
				break
			response += data

	except socket.error, e:
		raise Exception("Generic error on receive from agent: %s" % (e))

	client_socket.close()
	
	#if msg is None: # timeout
	#	raise Exception("Timeout")

	response = decode_response(response)
	if response['status'] == RESPONSE_SUCCESS:
		return response['content']
	else:
		raise Exception("Error response from server: %s" % (response['content']))

	return None

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

if __name__ == "__main__":
	
	#-----------------------------------------------------------------------------#
	# DEBUG                                                                       #
	#-----------------------------------------------------------------------------#

	def print_help():
		print ""

	def agent_ctl_main():

		opt_agentid        = None
		opt_broker         = None
		
		argv = sys.argv

		try:
			opts, args = getopt.getopt(sys.argv[1:], "hB:I:")
			for o, a in opts:
				if   o in ("-h"):
					show_help()
					sys.exit(0)
				elif o in ("-I"):
					opt_agentid = a
				elif o in ("-B"):
					opt_broker = a

		except getopt.error, msg:
			print "Error while parsing options: %s" % msg
			print_help()
			sys.exit()
		except ValueError, ex:
			print "Invalid value for input parameter: %s" % ex
			print_help()
			sys.exit()

		cmd    = args[0]
		params = args[1:]

		response = agent_ctl(opt_broker, opt_agentid, cmd, params)

		return

	agent_ctl_main()


