# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

################### check_mssql_server.py ##############################
# Version    : 2.1.0
# Date       : 06/09/2016
# Maintainer : Nagios Enterprises, LLC
# Licence    : GPLv3 (http://www.fsf.org/licenses/gpl.txt)
#
# Author/Maintainers:
#   Nicholas Scott (Original author, Nagios)
#   Jake Omann (Nagios)
#   Scott Wilkerson (Nagios)
#
# Changelog :
#   2.1.0 - Added server cpu usage, memory usage, and connection counters (campenberger)
#   2.0.3 - Remove misleading description of lockwait, removing the word Average (SW)
#   2.0.2 - Fixed issues where the SQL cache hit queries were yielding improper results when done on large systems (CTrahan)
#   2.0.1 - Fixed try/finally statement to accomodate Python 2.4 for legacy systems (NS)
#   2.0.0 - Complete rewrite of the structure, re-evaluated some queries to hopefully make them more portable (CFriese)
#           Updated the way averages are taken, no longer needs tempdb access (NS)
#   1.2.0 - Added ability to specify instances (NS)
#   1.1.0 - Fixed port bug allowing for non default ports (CBTSDon)
#           Added batchreq, sqlcompilations, fullscans, pagelife (Thanks mike from austria)
#           Added mode error checking which caused non-graceful exit (Thanks mike from austria)
#   1.0.2 - Fixed Uptime Counter to be based off of database (NS)
#           Fixed divide by zero error in transpsec (NS)
#
########################################################################

'''
Microsoft Database State Definitions
The following table defines the database states.

+------------------+--------------------------------------------------------------------------------------------------------------------------------------+
| State            | Definition                                                                                                                           |
+------------------+--------------------------------------------------------------------------------------------------------------------------------------+
| ONLINE           | Database is available for access. The primary filegroup is online, although the undo phase of recovery may not have been completed.  |
+------------------+--------------------------------------------------------------------------------------------------------------------------------------+
| OFFLINE          | Database is unavailable. A database becomes offline by explicit user action and remains offline until additional user action is      |
|                  | taken. For example, the database may be taken offline in order to move a file to a new disk. The database is then brought back       |
|                  | online after the move has been completed.                                                                                            |
+------------------+--------------------------------------------------------------------------------------------------------------------------------------+
| RESTORING        | One or more files of the primary filegroup are being restored, or one or more secondary files are being restored offline.            |
|                  | The database is unavailable.                                                                                                         |
+------------------+--------------------------------------------------------------------------------------------------------------------------------------+
| RECOVERING       | Database is being recovered. The recovering process is a transient state; the database will automatically become online if the       |
|                  | recovery succeeds. If the recovery fails, the database will become suspect. The database is unavailable.                             |
+------------------+--------------------------------------------------------------------------------------------------------------------------------------+
| RECOVERY PENDING | SQL Server has encountered a resource-related error during recovery. The database is not damaged, but files may be missing or system |
|                  | resource limitations may be preventing it from starting. The database is unavailable. Additional action by the user is required to   |
|                  | resolve the error and let the recovery process be completed.                                                                         |
+------------------+--------------------------------------------------------------------------------------------------------------------------------------+
| SUSPECT          | At least the primary filegroup is suspect and may be damaged. The database cannot be recovered during startup of SQL Server.         |
|                  | The database is unavailable. Additional action by the user is required to resolve the problem.                                       |
+------------------+--------------------------------------------------------------------------------------------------------------------------------------+
| EMERGENCY        | User has changed the database and set the status to EMERGENCY. The database is in single-user mode and may be repaired or restored.  |
|                  | The database is marked READ_ONLY, logging is disabled, and access is limited to members of the sysadmin fixed server role. EMERGENCY |
|                  | is primarily used for troubleshooting purposes. For example, a database marked as suspect can be set to the EMERGENCY state.         |
|                  | This could permit the system administrator read-only access to the database. Only members of the sysadmin fixed server role can set  |
|                  | a database to the EMERGENCY state.                                                                                                   |
+------------------+--------------------------------------------------------------------------------------------------------------------------------------+
'''

import pymssql
import time
import sys
import os

from sanet_poller.lib.exceptions import ExternalErrorException, UndefinedValueException, PollException


#----------------------------------------------------------------------------------------------------#
#                                                                                                    #
#----------------------------------------------------------------------------------------------------#

DB_STORE_QUERY = """USE "%s"\n\
SELECT\n\
  ISNULL(fg.name, 'TLOGS') AS filegroup,\n\
  dbf.name,\n\
  dbf.is_sparse,\n\
  dbf.size,\n\
  dbf.max_size,\n\
  dbf.growth,\n\
  dbf.is_percent_growth,\n\
  FILEPROPERTY(dbf.NAME,'SpaceUsed') AS used_size,\n\
  dbf.type_desc,\n\
  UPPER(SUBSTRING(dbf.physical_name, 1, 1)) AS filesystem_drive_letter\n\
FROM\n\
  sys.database_files AS dbf\n\
LEFT OUTER JOIN\n\
  sys.filegroups AS fg\n\
ON\n\
  dbf.data_space_id = fg.data_space_id"""

UPTM_QUERY = "SELECT sqlserver_start_time FROM sys.dm_os_sys_info"

BASE_QUERY = "SELECT cntr_value FROM sysperfinfo WHERE counter_name='%s' AND instance_name='';"
INST_QUERY = "SELECT cntr_value FROM sysperfinfo WHERE counter_name='%s' AND instance_name='%s';"
TRAN_QUERY = "SELECT cntr_value, instance_name FROM sysperfinfo WHERE counter_name='%s'"
OBJE_QUERY = "SELECT cntr_value FROM sysperfinfo WHERE counter_name='%s'"
DIVI_QUERY = "SELECT cntr_value FROM sysperfinfo WHERE counter_name LIKE '%s%%' AND instance_name='%s';"
DBO_QUERY = "SELECT name, state, state_desc FROM master.sys.databases;"
CON_QUERY = "SELECT COUNT(dbid) as NumberOfConnections FROM sys.sysprocesses WHERE dbid in (select database_id from sys.databases %s)"
MEM_QUERY = "SELECT 100*(1.0-(available_physical_memory_kb/(total_physical_memory_kb*1.0))) FROM sys.dm_os_sys_memory;"
CPU_QUERY = "SELECT "+\
	"record.value('(./Record/SchedulerMonitorEvent/SystemHealth/ProcessUtilization)[1]', 'int') AS [CPU] "+\
	"FROM ( "+\
		"SELECT[timestamp], CONVERT(XML, record) AS [record] "+\
			"FROM sys.dm_os_ring_buffers WITH ( NOLOCK ) "+\
			"WHERE ring_buffer_type=N'RING_BUFFER_SCHEDULER_MONITOR' "+\
			"AND record LIKE N'%<SystemHealth>%'"+\
	") as x;"

MODES = {

	'time2connect'      : { 'help'      : 'Time to connect to the database.' },
	
	'fullscans'         : { 'help'      : 'Full Scans / Sec',
							'stdout'    : 'Full Scans / Sec is %s/sec',
							'label'     : 'full_scans',
							'query'     : OBJE_QUERY % 'Full Scans/sec',
							'type'      : 'delta',
							},
	
	'transpec'          : { 'help'      : 'Transactions Per Second',
							'stdout'    : 'Transactions Per Second is %s/sec',
							'label'     : 'transactions_per_sec',
							'query'     : TRAN_QUERY % 'Transactions/sec',
							'type'      : 'Tdelta'
							},
	
	'db-free'           : { 'help'      : 'Free Storage Percentage',
							'stdout'    : 'Available space percentage is: %s',
							'label'     : 'db-full-size',
							'type'      : 'store_percentage',
							'query'     : DB_STORE_QUERY
							},

	'databaseonline'   : { 'help'       : 'List of online DBs',
							'stdout'    : 'Online status of DBs are: %s',
							'label'     : 'db-online',
							'type'      : 'state',
							'query'     : DBO_QUERY
						   },



	'connections'       : { 'help'      : 'Number of open connections',
							'stdout'    : 'Number of open connections is %s',
							'label'     : 'connections',
							'type'      : 'database',
							'query'     : CON_QUERY 
							},
	
	#'memory'            : { 'help'      : 'Used server memory',
	#						'stdout'    : 'Server using %s%% of memory',
	#						'label'     : 'memory',
	#						'query'     : MEM_QUERY
	#						},

# Per questo controllo i dati, seppur corretti, non sono in tempo reale come quelli forniti dal sistema operativo
	'cpu'               : { 'help'      : 'Server CPU utilization',
							'stdout'    : 'Current CPU utilization is %s%%',
							'lablel'    : 'cpu',
							'query'     : CPU_QUERY
							},

	'bufferhitratio'    : { 'help'      : 'Buffer Cache Hit Ratio',
							'stdout'    : 'Buffer Cache Hit Ratio is %s%%',
							'label'     : 'buffer_cache_hit_ratio',
							'unit'      : '%',
							'query'     : DIVI_QUERY % ('Buffer cache hit ratio', ''),
							'type'      : 'divide',
							'modifier'  : 100,
							},
	
	'pagelooks'         : { 'help'      : 'Page Lookups Per Second',
							'stdout'    : 'Page Lookups Per Second is %s',
							'label'     : 'page_lookups',
							'query'     : BASE_QUERY % 'Page lookups/sec',
							'type'      : 'delta'
							},
	
	'freepages'         : { 'help'      : 'Free Pages (Cumulative)',
							'stdout'    : 'Free pages is %s',
							'label'     : 'free_pages',
							'type'      : 'standard',
							'query'     : BASE_QUERY % 'Free pages'
							},
							
	'totalpages'        : { 'help'      : 'Total Pages (Cumulative)',
							'stdout'    : 'Total pages is %s',
							'label'     : 'totalpages',
							'type'      : 'standard',
							'query'     : BASE_QUERY % 'Total pages',
							},
							
	'targetpages'       : { 'help'      : 'Target Pages',
							'stdout'    : 'Target pages are %s',
							'label'     : 'target_pages',
							'type'      : 'standard',
							'query'     : BASE_QUERY % 'Target pages',
							},
							
	'databasepages'     : { 'help'      : 'Database Pages',
							'stdout'    : 'Database pages are %s',
							'label'     : 'database_pages',
							'type'      : 'standard',
							'query'     : BASE_QUERY % 'Database pages',
							},
	
	'stolenpages'       : { 'help'      : 'Stolen Pages',
							'stdout'    : 'Stolen pages are %s',
							'label'     : 'stolen_pages',
							'type'      : 'standard',
							'query'     : BASE_QUERY % 'Stolen pages',
							},
	
	'lazywrites'        : { 'help'      : 'Lazy Writes / Sec',
							'stdout'    : 'Lazy Writes / Sec is %s/sec',
							'label'     : 'lazy_writes',
							'query'     : BASE_QUERY % 'Lazy writes/sec',
							'type'      : 'delta'
							},
	
	'readahead'         : { 'help'      : 'Readahead Pages / Sec',
							'stdout'    : 'Readahead Pages / Sec is %s/sec',
							'label'     : 'readaheads',
							'query'     : BASE_QUERY % 'Readahead pages/sec',
							'type'      : 'delta',
							},
							
	
	'pagereads'         : { 'help'      : 'Page Reads / Sec',
							'stdout'    : 'Page Reads / Sec is %s/sec',
							'label'     : 'page_reads',
							'query'     : BASE_QUERY % 'Page reads/sec',
							'type'      : 'delta'
							},
	
	'checkpoints'       : { 'help'      : 'Checkpoint Pages / Sec',
							'stdout'    : 'Checkpoint Pages / Sec is %s/sec',
							'label'     : 'checkpoint_pages',
							'query'     : BASE_QUERY % 'Checkpoint pages/Sec',
							'type'      : 'delta'
							},
							
	
	'pagewrites'        : { 'help'      : 'Page Writes / Sec',
							'stdout'    : 'Page Writes / Sec is %s/sec',
							'label'     : 'page_writes',
							'query'     : BASE_QUERY % 'Page writes/sec',
							'type'      : 'delta',
							},
	
	'lockrequests'      : { 'help'      : 'Lock Requests / Sec',
							'stdout'    : 'Lock Requests / Sec is %s/sec',
							'label'     : 'lock_requests',
							'query'     : INST_QUERY % ('Lock requests/sec', '_Total'),
							'type'      : 'delta',
							},
	
	'locktimeouts'      : { 'help'      : 'Lock Timeouts / Sec',
							'stdout'    : 'Lock Timeouts / Sec is %s/sec',
							'label'     : 'lock_timeouts',
							'query'     : INST_QUERY % ('Lock timeouts/sec', '_Total'),
							'type'      : 'delta',
							},
	
	'deadlocks'         : { 'help'      : 'Deadlocks / Sec',
							'stdout'    : 'Deadlocks / Sec is %s/sec',
							'label'     : 'deadlocks',
							'query'     : INST_QUERY % ('Number of Deadlocks/sec', '_Total'),
							'type'      : 'delta',
							},
	
	'lockwaits'         : { 'help'      : 'Lockwaits / Sec',
							'stdout'    : 'Lockwaits / Sec is %s/sec',
							'label'     : 'lockwaits',
							'query'     : INST_QUERY % ('Lock Waits/sec', '_Total'),
							'type'      : 'delta',
							},
	
	'lockwait'          : { 'help'      : 'Lock Wait Time (ms)',
							'stdout'    : 'Lock Wait Time (ms) is %sms',
							'label'     : 'lockwait',
							'unit'      : 'ms',
							'query'     : INST_QUERY % ('Lock Wait Time (ms)', '_Total'),
							'type'      : 'standard',
							},
	
	'averagewait'       : { 'help'      : 'Average Wait Time (ms)',
							'stdout'    : 'Average Wait Time (ms) is %sms',
							'label'     : 'averagewait',
							'unit'      : 'ms',
							'query'     : DIVI_QUERY % ('Average Wait Time', '_Total'),
							'type'      : 'divide',
							},
	
	'pagesplits'        : { 'help'      : 'Page Splits / Sec',
							'stdout'    : 'Page Splits / Sec is %s/sec',
							'label'     : 'page_splits',
							'query'     : OBJE_QUERY % 'Page Splits/sec',
							'type'      : 'delta',
							},
	
	'cachehit'          : { 'help'      : 'Cache Hit Ratio',
							'stdout'    : 'Cache Hit Ratio is %s%%',
							'label'     : 'cache_hit_ratio',
							'query'     : DIVI_QUERY % ('Cache Hit Ratio', '_Total'),
							'type'      : 'divide',
							'unit'      : '%',
							'modifier'  : 100,
							},
	
	'batchreq'          : { 'help'      : 'Batch Requests / Sec',
							'stdout'    : 'Batch Requests / Sec is %s/sec',
							'label'     : 'batch_requests',
							'query'     : OBJE_QUERY % 'Batch Requests/sec',
							'type'      : 'delta',
							},
	
	'sqlcompilations'   : { 'help'      : 'SQL Compilations / Sec',
							'stdout'    : 'SQL Compilations / Sec is %s/sec',
							'label'     : 'sql_compilations',
							'query'     : OBJE_QUERY % 'SQL Compilations/sec',
							'type'      : 'delta',
							},
	
	'pagelife'          : { 'help'      : 'Page Life Expectancy',
							'stdout'    : 'Page Life Expectancy is %s/sec',
							'label'     : 'page_life_expectancy',
							'query'     : OBJE_QUERY % 'Page life expectancy',
							'type'      : 'standard'
							},
	
	#~ 'debug'             : { 'help'      : 'Used as a debugging tool.',
							#~ 'stdout'    : 'Debugging: ',
							#~ 'label'     : 'debug',
							#~ 'query'     : DIVI_QUERY % ('Average Wait Time', '_Total'),
							#~ 'type'      : 'divide' 
							#~ },
	
	'test'              : { 'help'      : 'Run tests of all queries against the database.' },

}

#----------------------------------------------------------------------------------------------------#
#                                                                                                    #
#----------------------------------------------------------------------------------------------------#

def getDBList(connection):
	"""
	This function is used in the db storage check.

	:param connection: a connection object
	:return: list of db names
	"""
	cur = connection.cursor()
	cur.execute('SELECT name FROM sys.databases')
	
	tmp = cur.fetchall()
	dbList = []
	
	for elem in tmp:
		dbList.append(elem[0])
	
	return dbList


#----------------------------------------------------------------------------------------------------#
#                                                                                                    #
#----------------------------------------------------------------------------------------------------#


class MSSQLQuery(object):

	"""
	Base MS SQL Query Object.
	It is used to perfom check specific queries and to compute their results.
	"""
	
	def __init__(self, query, options, label='', unit='', stdout='', host='', modifier=1, *args, **kwargs):
		self.query = query
		self.label = label
		self.unit = unit
		self.stdout = stdout
		self.options = options
		self.host = host
		self.modifier = modifier
		self.dcache = self.options.dcache
		self.instance = self.options.instance
	
	def run_on_connection(self, connection):
		"""
		Performs a check query.

		:param connection: a connection object
		:return: a tuple representing the result of the query
		"""

		cur = connection.cursor()
		cur.execute(self.query)
		
		row = cur.fetchone()
		
		#
		# FIX: in alcuni casi (riscontrato dopo un aggiornamento automatico di windows con reboot)
		#      la tabella sysperfinfo si svuota e non contiene piu' dati
		#
		if len(row) == 0:
			raise UndefinedValueException('Data unavailable for this check.')
		
		self.query_result = row[0]
	
	def finish(self):
		return self.result
	
	def calculate_result(self):
		"""
		This method is overridden by subclasses. It performs computations on data fetched by query and stores it in self.result
		"""
		if self.query_result is None:
			raise UndefinedValueException()
		
		self.result = float(self.query_result) * self.modifier
	
	def do(self, connection):
		"""
		This is the entry point of every check.

		:param connection: a connection object
		:return: data stored on self.result
		"""
		self.run_on_connection(connection)
		self.calculate_result()
		return self.finish()

#----------------------------------------------------------------------------------------------------#
#                                                                                                    #
#----------------------------------------------------------------------------------------------------#

class MSSQLDatabaseQuery(MSSQLQuery):
	"""
	This class performs queries on a specific database or sum every db values
	"""

	def __init__(self, *args, **kwargs):
		super(MSSQLDatabaseQuery, self).__init__(*args, **kwargs)

	def run_on_connection(self, connection):
		"""
		Performs a check query.

		:param connection: a connection object
		:return: a tuple representing the result of the query
		"""

		if self.instance:
			instance = "WHERE name= '" + self.instance + "'"
			query = self.query % instance
		else:
			query = self.query % ''

		cur = connection.cursor()
		cur.execute(query)

		row = cur.fetchone()

		#
		# FIX: in alcuni casi (riscontrato dopo un aggiornamento automatico di windows con reboot)
		#      la tabella sysperfinfo si svuota e non contiene piu' dati
		#
		if len(row) == 0:
			raise UndefinedValueException('Data unavailable for this check.')

		self.query_result = row[0]


#----------------------------------------------------------------------------------------------------#
#                                                                                                    #
#----------------------------------------------------------------------------------------------------#

class MSSQLDivideQuery(MSSQLQuery):
	"""
	This class performs queries and computes data as percentage.
	"""
	
	def __init__(self, *args, **kwargs):
		super(MSSQLDivideQuery, self).__init__(*args, **kwargs)
	
	def calculate_result(self):
		if self.query_result is None:
			raise UndefinedValueException()
		
		if self.query_result[1] != 0:
			self.result = (float(self.query_result[0]) / float(self.query_result[1])) * self.modifier
		else:
			self.result = float(self.query_result[0]) * self.modifier
	
	def run_on_connection(self, connection):
		cur = connection.cursor()
		cur.execute(self.query)
		self.query_result = [x[0] for x in cur.fetchall()]
		if len(self.query_result) == 0:
			raise UndefinedValueException('Data unavailable for this check.')


#----------------------------------------------------------------------------------------------------#
#                                                                                                    #
#----------------------------------------------------------------------------------------------------#


class MSSQLStorePercentageQuery(MSSQLQuery):
	"""
	This class is specific for the database free storage space check. On MS SQL Server, storage files are divided in two categories: LOG files (used to store transactions logs) and ROW files (used to store data such as tables). These files can be set to grow on need with a maximum file size or with unlimited size (its growth stops when filesystem is full), or they can be set not to grow at all, keeping a fixed file size.

	In the first and third cases this maximum file size is taken to compute a free db store percentage value.
	In the second case MS SQL Server is asked for disk storage data and its free storage data is used to compute a free db store percentage value.
	"""
	
	def __init__(self, *args, **kwargs):
		super(MSSQLStorePercentageQuery, self).__init__(*args, **kwargs)
		self.filesystems = {}
	
	def get_filesystems(self, connection):
		"""
		This method performs a query on MS SQL Server and fetches information about disk storages and saves a dictionary, the keys of represent a filesystem name and the values associated to them their free space, and stores it in self.filesystems.

		:param connection: a connection object
		"""
		cur = connection.cursor()
		cur.execute('exec master.dbo.xp_fixeddrives')
		
		# Filesystems dictionary with their free space expressed in B
		for row in cur.fetchall():
			self.filesystems[row[0]] = row[1] * 1024 * 1024
	
	def calculate_result(self):
		if self.query_result is None:
			raise UndefinedValueException()
		
		self.result = {}
		
		#
		# Rows is a list of tuples, one per row
		#
		for db, rows in self.query_result.items():
			self.result[ db ] = {}
			self.result[ db ][ 'LOG' ] = {}
			self.result[ db ][ 'ROWS' ] = {}
			unlimited = { 'LOG': [], 'ROWS': [] }
			
			if type(rows) == str:
				self.result[ db ] = rows
				continue
			
			for row in rows:
				
				if row[0] not in self.result[ db ][ row[8] ]:
					#
					# Preparing datastructure for given db name, filetype and filegroup
					#         db name  LOG or ROWS   filegroup
					self.result[ db ]  [ row[8] ]    [ row[0] ] = {}
					self.result[ db ][ row[8] ][ row[0] ][ 'used' ] = 0
					self.result[ db ][ row[8] ][ row[0] ][ 'max_size' ] = 0
				
				self.result[ db ][ row[8] ][ row[0] ][ 'used' ] += row[7] * 8 * 1024
				
				#
				# If file growth is unlimited
				#
				if row[4] == -1:
					unlimited[ row[8] ].append( row[0] )
					
					if not 'drivers' in self.result[ db ][ row[8] ][ row[0] ]:
						self.result[ db ][ row[8] ][ row[0] ][ 'drivers' ] = []
						self.result[ db ][ row[8] ][ row[0] ][ 'max_size'] += row[7] * 8 * 1024
						
					if not row[9] in self.result[ db ][ row[8] ][ row[0] ][ 'drivers' ]:
						self.result[ db ][ row[8] ][ row[0] ][ 'drivers' ].append( row[9] )
						self.result[ db ][ row[8] ][ row[0] ][ 'max_size'] += row[7] * 8 * 1024
				else:
					self.result[ db ][ row[8] ][ row[0] ][ 'max_size' ] += row[4] * 8 * 1024
			
			if unlimited[ 'ROWS' ]:
				for group in unlimited[ 'ROWS' ]:
					for drive in self.result[ db ][ 'ROWS' ][ group ][ 'drivers' ]:
						self.result[ db ][ 'ROWS' ][ group ][ 'max_size' ] += self.filesystems[ drive ]
			
			if unlimited[ 'LOG' ]:
				for group in unlimited[ 'LOG' ]:
					for drive in self.result[ db ][ 'LOG' ][ group ][ 'drivers' ]:
						self.result[ db ][ 'LOG' ][ group ][ 'max_size' ] += self.filesystems[ drive ]
	
	#
	#                                                 Index meanings for each row
	#
	#    0         1         2          3       4           5          6                   7            8              9
	# filegroup   name    is_sparse   size   max_size    growth   is_percent_growth    used_size   type_desc    filesystem_drive_letter
	def run_on_connection(self, connection):
		self.get_filesystems(connection)
		
		cur = connection.cursor()
		self.query_result = {}
		
		if self.options.instance:
			dbList = [ self.instance ]
		
		else:
			dbList = getDBList(connection)
		
		for elem in dbList:
			query = self.query % elem
			try:
				cur.execute(query)
			except pymssql.OperationalError, e:
				self.query_result[elem] = str(e[1])
			else:
				self.query_result[elem] = cur.fetchall()


#----------------------------------------------------------------------------------------------------#
#                                                                                                    #
#----------------------------------------------------------------------------------------------------#


class MSSQLStatusQuery(MSSQLQuery):
	"""
	This class is used to check dbs state and returns a list of dictionaries, each representing one specific database. The content of the dictionary is
	{
		'db_name': the name of the db,
		'state': the numeric state of the db,
		'verbose': a string containing short information about db status
	}

	Microsoft Database State Definitions
	------------------------------------

	+------------------+--------------------------------------------------------------------------------------------------------------------------------------+
	| State            | Definition                                                                                                                           |
	+------------------+--------------------------------------------------------------------------------------------------------------------------------------+
	| ONLINE           | Database is available for access. The primary filegroup is online, although the undo phase of recovery may not have been completed.  |
	+------------------+--------------------------------------------------------------------------------------------------------------------------------------+
	| OFFLINE          | Database is unavailable. A database becomes offline by explicit user action and remains offline until additional user action is      |
	|                  | taken. For example, the database may be taken offline in order to move a file to a new disk. The database is then brought back       |
	|                  | online after the move has been completed.                                                                                            |
	+------------------+--------------------------------------------------------------------------------------------------------------------------------------+
	| RESTORING        | One or more files of the primary filegroup are being restored, or one or more secondary files are being restored offline.            |
	|                  | The database is unavailable.                                                                                                         |
	+------------------+--------------------------------------------------------------------------------------------------------------------------------------+
	| RECOVERING       | Database is being recovered. The recovering process is a transient state; the database will automatically become online if the       |
	|                  | recovery succeeds. If the recovery fails, the database will become suspect. The database is unavailable.                             |
	+------------------+--------------------------------------------------------------------------------------------------------------------------------------+
	| RECOVERY PENDING | SQL Server has encountered a resource-related error during recovery. The database is not damaged, but files may be missing or system |
	|                  | resource limitations may be preventing it from starting. The database is unavailable. Additional action by the user is required to   |
	|                  | resolve the error and let the recovery process be completed.                                                                         |
	+------------------+--------------------------------------------------------------------------------------------------------------------------------------+
	| SUSPECT          | At least the primary filegroup is suspect and may be damaged. The database cannot be recovered during startup of SQL Server.         |
	|                  | The database is unavailable. Additional action by the user is required to resolve the problem.                                       |
	+------------------+--------------------------------------------------------------------------------------------------------------------------------------+
	| EMERGENCY        | User has changed the database and set the status to EMERGENCY. The database is in single-user mode and may be repaired or restored.  |
	|                  | The database is marked READ_ONLY, logging is disabled, and access is limited to members of the sysadmin fixed server role. EMERGENCY |
	|                  | is primarily used for troubleshooting purposes. For example, a database marked as suspect can be set to the EMERGENCY state.         |
	|                  | This could permit the system administrator read-only access to the database. Only members of the sysadmin fixed server role can set  |
	|                  | a database to the EMERGENCY state.                                                                                                   |
	+------------------+--------------------------------------------------------------------------------------------------------------------------------------+
	"""
	
	def __init__(self, *args, **kwargs):
		super(MSSQLStatusQuery, self).__init__(*args, **kwargs)
	
	def calculate_result(self):
		if self.query_result is None:
			raise UndefinedValueException()
		
		l = []
		for row in self.query_result:
			if self.instance is not None and self.instance != row[0]:
				continue
			
			l.append({ 'db_name': row[0],
						'state'  : row[1],
						'verbose': 'State of DB "' + row[0] + '" is ' + row[2] })
		
		self.result = l
	
	def run_on_connection(self, connection):
		cur = connection.cursor()
		cur.execute(self.query)
		self.query_result = cur.fetchall()


#----------------------------------------------------------------------------------------------------#
#                                                                                                    #
#----------------------------------------------------------------------------------------------------#


class MSSQLTDeltaQuery(MSSQLQuery):

	"""
	This class is specific for transactions/sec check. It is similar to MSSQLDeltaQuery class.
	"""

	def __init__(self, *args, **kwargs):
		super(MSSQLTDeltaQuery, self).__init__(*args, **kwargs)
		self.uptime = None
	
	def run_on_connection(self, connection):
		cur = connection.cursor()
		
		if self.options.instance:
			self.query += "AND instance_name='%s';" % self.options.instance
		else:
			self.query += "AND instance_name!='_Total' AND instance_name!='mssqlsystemresource';"
		
		cur.execute(self.query)
		self.query_result = cur.fetchall()
	
	def make_pickle_name(self):
		"""
		This method compute a univoc string to use in a cache of data.
		:return: a string
		"""
		tmpname = hash(self.host + self.query)
		self.picklename = 'mssql-%s-%s.tmp' % (self.host, tmpname)
	
	def calculate_result(self):
		if self.query_result is None:
			raise UndefinedValueException()
		
		self.make_pickle_name()
		self.result = {}
		
		last_run = self.dcache.get(self.picklename)
		
		tmp_old = {}
		
		if last_run is None:
			last_run = { 'time' : None, 'value' : None }
		
		if last_run['time']:
			
			old_time = last_run['time']
			new_time = time.time()
			old_vals = last_run['query_result']
			for elem in self.query_result:
				
				if elem[1].strip() in old_vals:
					old_val = old_vals[ elem[1].strip() ]
				else:
					old_val = 0
					
				self.result[ elem[1].strip() ] = (float(elem[0] - old_val ) / float(new_time - old_time)) * self.modifier
		
		else:
			self.result = None
		
		for elem in self.query_result:
			tmp_old[ elem[1].strip() ] = elem[0]
		
		new_run = { 'time' : time.time(), 'query_result' : tmp_old }
		
		self.dcache.set(self.picklename, new_run)
		
		if last_run['time'] < self.uptime:
			raise UndefinedValueException('Server was restarted after last check. Last value is not valid')
		
		if self.result is None:
			raise UndefinedValueException('No value for this check now.')
		
	def check_uptime(self, connection):

		"""
		This method fetches uptime data of MS SQL Server. This information is useful because at every server reboot, the counters used to perfom this kind of checks are set to 0, so data sored in the cache cannot be used anymore.
		It stores the data in self.uptime.

		:param connection: a connection object
		"""
		
		cur = connection.cursor()
		cur.execute(UPTM_QUERY)
		
		dt = cur.fetchone()[0]
		self.uptime = time.mktime(dt.timetuple())
	
	def do(self, connection):
		self.check_uptime(connection)
		self.run_on_connection(connection)
		self.calculate_result()
		return self.finish()

#----------------------------------------------------------------------------------------------------#
#                                                                                                    #
#----------------------------------------------------------------------------------------------------#


class MSSQLDeltaQuery(MSSQLQuery):

	"""
	This class is used to compute delta data.
	"""
	
	def make_pickle_name(self):
		"""
		This method compute a univoc string to use in a cache of data.
		:return: a string
		"""
		tmpname = hash(self.host + self.query)
		self.picklename = 'mssql-%s-%s.tmp' % (self.host, tmpname)
	
	def calculate_result(self):
		
		if self.query_result is None:
			raise UndefinedValueException()
		
		self.make_pickle_name()
		
		last_run = self.dcache.get(self.picklename)
		
		if last_run is None:
			last_run = { 'time' : None, 'value' : None }
		
		if last_run['time']:
			old_time = last_run['time']
			new_time = time.time()
			old_val  = last_run['query_result']
			new_val  = self.query_result
			self.result = (float(new_val - old_val) / float(new_time - old_time)) * self.modifier
		else:
			self.result = None
		
		new_run = { 'time' : time.time(), 'query_result' : self.query_result }
		
		self.dcache.set(self.picklename, new_run)

		if last_run['time'] < self.uptime:
				raise UndefinedValueException('Server was restarted after last check. Last value is not valid')

		if self.result is None:
			raise UndefinedValueException('No value for this check now.')
	
	def check_uptime(self, connection):

		"""
		This method fetches uptime data of MS SQL Server. This information is useful because at every server reboot, the counters used to perfom this kind of checks are set to 0, so data sored in the cache cannot be used anymore.
		It stores the data in self.uptime.

		:param connection: a connection object
		"""
		
		cur = connection.cursor()
		cur.execute(UPTM_QUERY)
		
		dt = cur.fetchone()[0]
		self.uptime = time.mktime(dt.timetuple())
	
	def do(self, connection):
		self.check_uptime(connection)
		self.run_on_connection(connection)
		self.calculate_result()
		return self.finish()



#----------------------------------------------------------------------------------------------------#
#                                                                                                    #
#----------------------------------------------------------------------------------------------------#


def is_within_range(nagstring, value):
	if not nagstring:
		return False
	import re
	import operator
	first_float = r'(?P<first>(-?[0-9]+(\.[0-9]+)?))'
	second_float= r'(?P<second>(-?[0-9]+(\.[0-9]+)?))'
	actions = [ (r'^%s$' % first_float,lambda y: (value > float(y.group('first'))) or (value < 0)),
				(r'^%s:$' % first_float,lambda y: value < float(y.group('first'))),
				(r'^~:%s$' % first_float,lambda y: value > float(y.group('first'))),
				(r'^%s:%s$' % (first_float,second_float), lambda y: (value < float(y.group('first'))) or (value > float(y.group('second')))),
				(r'^@%s:%s$' % (first_float,second_float), lambda y: not((value < float(y.group('first'))) or (value > float(y.group('second')))))]
	for regstr,func in actions:
		res = re.match(regstr,nagstring)
		if res: 
			return func(res)
	raise Exception('Improper warning/critical format.')


#----------------------------------------------------------------------------------------------------#
#                                                                                                    #
#----------------------------------------------------------------------------------------------------#


def connect_db(options):
	host = options.hostname
	host += ":" + str(options.port)
	start = time.time()
	try:
		mssql = pymssql.connect(host = host, user = options.user, password = options.password, database=options.database, tds_version='7.0')
	except Exception, e:
		raise ExternalErrorException(e)
	total = time.time() - start
	return mssql, total, host


#----------------------------------------------------------------------------------------------------#
#                                                                                                    #
#----------------------------------------------------------------------------------------------------#


class Options(object):
	
	def __init__(self, dcache, hostname, user, password=None, port=1433, mode=None, database='master', instance=None):
		self.dcache   = dcache
		self.hostname = hostname
		self.user     = user
		self.password = password
		self.port     = port
		self.mode     = mode
		self.database = database
		self.instance = instance


#----------------------------------------------------------------------------------------------------#
#                                                                                                    #
#----------------------------------------------------------------------------------------------------#


def _mssql_check_server(dcache, hostname, port, user, password, mode, database='master', instance=None):
	
	if not mode or mode == 'time2connect':

		instance = instance or 'master'

		options = Options(dcache=dcache,
		                  hostname=hostname,
		                  port=port,
		                  user=user,
		                  password=password,
		                  mode=mode,
		                  database=instance
        )

		try:
			mssql, total, host = connect_db(options)
		except ExternalErrorException, e:
			raise PollException('Connection error. DB %s could be offline.' % instance)

		return total

	else:
		options =  Options(dcache=dcache,
		                   hostname=hostname,
		                   port=port,
		                   user=user,
		                   password=password,
		                   mode=mode,
		                   database=database,
		                   instance=instance 
        )

		mssql, total, host = connect_db(options)

		return execute_query(mssql, options, host)


#----------------------------------------------------------------------------------------------------#
#                                                                                                    #
#----------------------------------------------------------------------------------------------------#


def execute_query(mssql, options, host=''):

	sql_query = MODES[options.mode]
	sql_query['options'] = options
	sql_query['host'] = host
	query_type = sql_query.get('type')
	if query_type == 'delta':
		mssql_query = MSSQLDeltaQuery(**sql_query)
	elif query_type == 'Tdelta':
		mssql_query = MSSQLTDeltaQuery(**sql_query)
	elif query_type == 'divide':
		mssql_query = MSSQLDivideQuery(**sql_query)
	elif query_type == 'state':
		mssql_query = MSSQLStatusQuery(**sql_query)
	elif query_type == 'store_percentage':
		mssql_query = MSSQLStorePercentageQuery(**sql_query)
	elif query_type == 'database':
		mssql_query = MSSQLDatabaseQuery(**sql_query)
	else:
		mssql_query = MSSQLQuery(**sql_query)
	return mssql_query.do(mssql)


