# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it> 
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import time
import traceback

from sanet_common.ipc import IPCNotFound, IPCConnectionError, IPCException
from sanet_common.ipc import RESPONSE_SUCCESS, RESPONSE_ERROR
from sanet_common.ipc import encode_response_with_id, decode_request

from sanet_common.sanetdefs.ipc import AGENT_NETINFOHELPER_QUEUE

from sanet_common.netutils import resolve_ipv4

from sanet_libsnmp import snmp_get, snmp_walk

from sanet_poller.lib.threadutils import AgentThread

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class NetInfoAgentHelper(object):

	REQUEST_NETINFO_SNMP_GET     = 'netinfo_snmp_get'
	REQUEST_NETINFO_SNMP_WALK    = 'netinfo_snmp_walk'

	REQUEST_NETINFO_RESOLVE_IPv4 = 'netinfo_resolve_ipv4'

	REQUEST_NETINFO_EXECUTE      = 'netinfo_execute'

	#-----------------------------------------------------------------------------#
	#                                                                             #
	#-----------------------------------------------------------------------------#

	def handleHelperCommands(self):

		self.logger.info("[NetinfoHELPER] Start...")

		while True:

			if self.is_stopping():
				break

			try:
				self.logger.info("[NetinfoHELPER] Creating connection")

				# OPEN
				self.helper_connection = self.create_connection()
				self.helper_connection.open()

				# BIND
				self.helper_connection.bind(AGENT_NETINFOHELPER_QUEUE(self.agentid), persistent=False)

				self.logger.info("[NetinfoHELPER] Looping...")

				# LOOP
				while not self.is_stopping():
					self.receive_helper_command()

				self.helper_connection.ack_pending()

				self.helper_connection.close()

			except IPCConnectionError, e:
				self.logger.error("[NetinfoHELPER] Connection lost")

			finally:
				try:
					if self.helper_connection:
						self.helper_connection.close()
				except Exception, e:
					pass

		self.logger.info("[HELPER] Exit")


	def receive_helper_command(self):

		try:
			self.logger.log(9, "[NetinfoHELPER] Receiving...")

			(msg_addr, msg) = self.helper_connection.receive_msg(timeout=1, ack_msg=False)
			if msg == None: # timeout
				return

			if msg is None:
				return

			self.logger.info("[NetinfoHELPER] Received %s:%s", msg_addr, msg)

			self.spawn_thread_for_request(msg_addr, msg)

			#self.queue_ctl.send_msg( msg_addr, response, blocking=False)
		except IPCNotFound, e:
			self.logger.error("[NetinfoHELPER] Error on connection: %s", e)

		except IPCException, e:
			self.logger.error("[NetinfoHELPER] Error on connection: %s", e)
			raise e	

		except Exception, e:
			self.logger.warning("[NetinfoHELPER] Error while processing message: %s", e)

	def spawn_thread_for_request(self, msg_addr, msg):

		name = 'helper %s' % (int(time.time()))

		params = { 'msg_addr': msg_addr, 'msg': msg }

		helper_thread = AgentThread(name, self.single_request_handler, params)
		helper_thread.start()

		return

	def single_request_handler(self, msg_addr, msg):

		try:
			request = decode_request(msg)

			request_id     = request['id']
			request_type   = request['type']
			request_params = request['params']
			
			self.logger.info("[NetinfoHELPER] Processing %s", request)

			if   request_type == self.REQUEST_NETINFO_RESOLVE_IPv4: response_data = self.get_netinf_resolve_ipv4(request_params)

			elif request_type == self.REQUEST_NETINFO_SNMP_GET    : response_data = self.get_netinf_snmp_get(request_params)
			elif request_type == self.REQUEST_NETINFO_SNMP_WALK   : response_data = self.get_netinf_snmp_walk(request_params)

			elif request_type == self.REQUEST_NETINFO_EXECUTE     : response_data = self.get_netinf_execute(request_params)

			else:
				raise Exception("Invalid request: %s" % (request['type']))

                        self.logger.info("encoding response")

			response = encode_response_with_id( RESPONSE_SUCCESS, request_id, response_data )
		except Exception, ex:
			self.logger.error("[NetinfoHELPER] Error while processing request: %s %s %s", type(ex), str(ex), traceback.format_exc())
			response = encode_response_with_id( RESPONSE_ERROR, request_id, "Error while processing request: %s:%s" % (str(ex), traceback.format_exc()) )

		# SEND
		try:
			self.logger.info("[NetinfoHELPER] Sending response to: %s", msg_addr)
			self.helper_connection.send_msg( msg_addr, response, blocking=False)
			self.logger.info("[NetinfoHELPER] Sent")
                except Exception, ex:
		#except IPCException, ex:
			self.logger.error("[HELPER] Error while processing request for %s: %s", msg_addr, str(ex))


	def get_netinf_snmp_get(self, params):
		params = params[0]
		return snmp_get(**params)

	def get_netinf_snmp_walk(self, params):
		params = params[0]
		return snmp_walk(**params)

	def get_netinf_resolve_ipv4(self, params):
		params = params[0]
		hostname = params['hostname']
		return resolve_ipv4(hostname)

	def get_netinf_execute(self, params):
		params = params[0]

		from sanet_poller.lib.evaluation import ExecutionContext
		from sanet_poller.lib.functions.execcmd import exec_launch_program
		
		context         = ExecutionContext()
		log             = self.logger
		program_path    = params['path']
		args            = params['args']
		execute_timeout = params['execute_timeout']

		try:
			output = exec_launch_program(context, log, program_path, args, execute_timeout=execute_timeout)

			reason    = ','.join(context.getExecutionInfos())
			exception = None
			
		except Exception, e:
			reason    = ','.join(context.getExecutionInfos())
			output    = None
			exception = str(e)

		data = {
			"reason"   : reason,
			"output"   : output,
			"exception": exception,
		}

		return data
