# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it> 
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

###########################################################################
Classes
###########################################################################

TODO

"""

import os
import socket
import time
import datetime
import types
import copy
import threading
import datetime
import json
import random
from collections import deque
from operator import attrgetter
import logging

# Init random
random.seed(os.getpid())


from sanet_common.sanetdefs.constants import *
from sanet_common.netutils import resolve_ip_addr

from sanet_poller.lib.exceptions import *

from sanet_poller.lib.evaluation import ExecutionContext
from sanet_poller.lib.evaluation import expreval
from sanet_poller.lib.evaluation import get_cached_expr_tree
from sanet_poller.lib.utils      import crc32
from sanet_poller.lib.utils      import get_time


#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

# FIX: move to constants?
DEPENDSON_STOP_STATES = [CSTATE_DN, CSTATE_FA, CSTATE_TU, CSTATE_TD, CSTATE_IN, CSTATE_DD, CSTATE_DU]


DEFAULT_DELAY_AFTER_RESUME = 1  # seconds

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

#
#                 +---------------------------------------------------+
#                 |                                                   |
#                 v                                                   |
#    IDLE ---> SCHEUDLED ---> QUEUED ---> CHECKING ---> COMMITTING ---+
#                ^   |           ^    
#                |   v           |   
#               WAITING ---------+

#
# These constants tell us what Poller is doing with a cached basecheck.
#
POLLSTATE_IDLE       = 0   # Initial state, not scheduled
POLLSTATE_SCHEDULED  = 1   # Scheduled and waiting for execution
POLLSTATE_ORGANIZING = 2.01# Descheduled and waiting for organizer handling
POLLSTATE_QUEUED     = 2   # Waiting for execution inside main waiting queue (by thread pool)
POLLSTATE_WAITING    = 2.1 # Waiting for execution inside scheduling group queue
POLLSTATE_CHECKING   = 3   # Checking by worker thread
POLLSTATE_COMMITTING = 4   # Preparing update data for server

POLLSTATE_STR = {
	POLLSTATE_IDLE      : 'IDLE',      
	POLLSTATE_SCHEDULED : 'SCHEDULED', 
	POLLSTATE_QUEUED    : 'WAITING',   
	POLLSTATE_CHECKING  : 'CHECKING', 
	POLLSTATE_ORGANIZING: 'ORGANIZING',
	POLLSTATE_COMMITTING: 'COMMIT',
	POLLSTATE_WAITING   : 'WAITING2',
}

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class NotImplementedException(Exception):
	pass

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class Pobject(object):

	#
	# Base objects with utility methods
	#
	def __init__(self):

		# Questo campo serve per assicurare che l'oggetto non venga
		# piu' utilizzato da questo momento in avanti.
		self.___valid = True

	def set_invalid(self):
		self.___valid = False

	def is_valid(self):
		return self.___valid 

	def load_from_dict(self, dict_data):
		for k, v in dict_data.items():
			if not hasattr(self, k): 
				raise Exception("Invalid field '%s' in class '%s'" % (k, self.__class__))
			setattr(self, k, v)

	def get_dict(self):
		return copy.copy(self.__dict__)

#-----------------------------------------------------------------------------#

class PSchedulingGroup(Pobject):
	"""
	This class represents a scheduling group entity inside the agent.
	
	It contains a list of tuples (ts, datagroup) of shceduled timestamps and
	associated poller datagroups entities.
	This because list operations such as sorting, in CPython, are atomic actions.
	
	Also, bisect.insort_left() is a method that inserts elements in an ordered list
	with high performance.
	
	dg_list = [
	            (Scheduled_ts, PDatagroup),
	            (Scheduled_ts1, PDatagroup1),
	            (Scheduled_ts2, PDatagroup2)
	          ]
	
	Only element with index 0 has to be pop'd from list.
	"""

	def __init__(self, logger):
		super(PSchedulingGroup, self).__init__()

		self.logger = logger

		# Data from server
		self.tenant_id    = None
		self.mtime        = None
		self.uuid         = None
		self.name         = None
		self.min_gap      = None
		self.max_deadline = None
		self.max_parallel = None

		#
		# Internal fields
		#
		self.queueLock = threading.Semaphore()

		# Timeline locale degli scheduled datagroup di questo gruppo.
		# Usata per gestire lo sovrapposizioni.
		self.scheduledQueue = deque([])

		# Numero di ScheduledDataGroup che sono attualmente in esecuzione (o nella coda di wait principale)
		#
		# INVARIANTE: Se self.max_paralleld != None  ->  self.executing  <= self.max_parallel
		#                        
		self.executing = 0

		# organizer ---> this ---> agentQueue
		self.lockedQueue = deque([])

		return

	def insert_sdg_in_scheduling_queue(self, queue, initial_ts, sdg):
		"""
		Scheduling algorithm

		:param queue: Scheduling Group queue
		:param initial_ts: theorical next execution time
		:param sdg: Scheduled Datagroup object
		:return: actual next execution time
		"""

		min_gap      = self.min_gap      if self.min_gap is not None else 0

		calculated_ts = initial_ts

		if len(queue) > 0:
			insert_pos    = None
			for idx, (ts, sdg2) in enumerate(queue):
				if calculated_ts <= ts - min_gap:
					insert_pos = idx
					break
				if calculated_ts <= ts + min_gap:
					calculated_ts = ts + min_gap
					continue

			if insert_pos is not None:
				queue.insert(insert_pos, (calculated_ts, sdg))
			else:
				queue.append( (calculated_ts, sdg))
		else:
			queue.append((calculated_ts, sdg))

		return calculated_ts

	def remove_sdg_from_scheduled_queue(self, dg_uuid):
		"""
		Method used to clean scheduledQueue from old datagroups on confguration changes

		@param dg_uuid: dg to remove uuid
		"""
		self.scheduledQueue = [(ts, sdg) for ts, sdg in self.scheduledQueue if sdg.pdatagroup.uuid != dg_uuid]

	def scheduleJob(self, next_time, schedulerHandler, handler, params ):

		scheduled_dg = params[0]

		# IL RESCHEDULING VA FATTO SOLO SE IL MINPERIOD E' != 0 OVVERO IL DATAGROUP NON E' (DIVENTATO) PASSIVO
		if scheduled_dg.pdatagroup.minperiod == 0:
			return 

		self.queueLock.acquire()

		# Aggiornamento dello scheduler locale
		new_scheduling_queue = [ (ts, sdg) for ts, sdg in self.scheduledQueue if sdg != scheduled_dg]
		next_time = self.insert_sdg_in_scheduling_queue(new_scheduling_queue, next_time, scheduled_dg)
		self.scheduledQueue = new_scheduling_queue

		# Inserimento nello scheduler principale
		scheduled_dg.schedule_time = next_time
		scheduled_dg.sched_state   = POLLSTATE_SCHEDULED
		schedulerHandler(next_time, handler, (scheduled_dg,))

		self.queueLock.release()

		return next_time

	def scheduleJobAndUnlockPending(self, schedulerHandler, t_pool_handler, handler, params ):
		"""
		Schedule give datagroup to proper position in time.
		
		The used algorithem scans the ordered list scheduledQueue and stops when
		it finds the right place for scheduling respecting the following rules:
		   * next_time >= time.time() + min_period
		   * next_time >= ts1 + min_gap
		   * ts2 - ts1 >= 2*min_gap
		
		@param handler: handler function to execute datagroup
		@param params: tuple containing a PDataGroup structure (dg, )
		"""

		scheduled_dg = params[0]

		min_gap      = self.min_gap      if self.min_gap is not None else 0
		max_parallel = self.max_parallel
		max_deadline = self.max_deadline

		lastdone     = scheduled_dg.pdatagroup.lastdone

		if lastdone is not None and lastdone > time.time():
			next_time = scheduled_dg.pdatagroup.lastdone + scheduled_dg.pdatagroup.minperiod
		else: 
			next_time = time.time()                      + scheduled_dg.pdatagroup.minperiod

		t1 = time.time()

		self.queueLock.acquire()

		#
		# Aggiornamento dello scheduler locale
		#
		new_scheduling_queue = [ (ts, sdg) for ts, sdg in self.scheduledQueue if sdg != scheduled_dg]

		# IL RESCHEDULING VA FATTO SOLO SE IL MINPERIOD E' != 0 OVVERO IL DATAGROUP NON E' (DIVENTATO) PASSIVO
		if scheduled_dg.pdatagroup.minperiod != 0:
			next_time = self.insert_sdg_in_scheduling_queue(new_scheduling_queue, next_time, scheduled_dg)

			scheduled_dg.schedule_time = next_time
			scheduled_dg.sched_state   = POLLSTATE_SCHEDULED
			schedulerHandler(next_time, handler, (scheduled_dg,))

		self.scheduledQueue = new_scheduling_queue

		#
		# Inserimento nello scheduler principale dell'agente

		#
		# Check scheduled-datagroup pendenti che aspettano di essere eseguiti
		#

		self.executing -= 1

		self.logger.debug('[scheduling group: %s] still in execution: %s', self.name, self.executing)

		lockedQ      = self.lockedQueue
		if len(lockedQ) > 0:
			self.logger.debug('[scheduling group: %s] emptying locked queue. Queue len: %s. Executing: %s', self.name, len(lockedQ), self.executing)

			# Se max_parallel e' impostato devo sbloccarne solo alcuni
			if self.max_parallel:
				max_to_free = max_parallel - self.executing
			else:
				# Ma parallel potrebbe essere stato sconfigurato tra un aggiornamento e l'altro.
				# Poco importa, li sblocco tutti.
				max_to_free = len(lockedQ)

			for i in xrange( max_to_free ):
				handler, params = lockedQ.popleft()

				sdg = params[0]
				dg  = sdg.pdatagroup

				if sdg.schedule_time and time.time() - sdg.schedule_time > max_deadline:
					#
					# Abbiamo superato il limite massimo temporale che ci eravamo prefissati.
					# Da eseguire il prima possibile!
					#

					# Stiamo mettendo in esecuzione un datagroup quindi aumentiamo il counter  (# FIX: test_02)
					self.executing += 1

					sdg.sched_state = POLLSTATE_QUEUED
					t_pool_handler(handler, params)
				else:
					# stiamo rischedulando un gruppo di datagroup bloccati insieme.
					# Per garantire almeno una distanza di min_gap tra lo scheduling l'uno dell'altro usiamo questa formula
					ts = time.time() + min_gap*(i+1)

					self.scheduledQueue = [ (x, y) for x, y in self.scheduledQueue if y != sdg]

					# inserimento efficiente
					next_time = self.insert_sdg_in_scheduling_queue(self.scheduledQueue, ts, sdg)

					scheduled_dg.schedule_time = next_time
					scheduled_dg.sched_state   = POLLSTATE_SCHEDULED
					schedulerHandler(ts, handler, params)

		self.queueLock.release()

		return next_time


	def scheduleNewJob(self, handler, params, schedulerHandler):

		sdg = params[0]

		self.queueLock.acquire()

		#
		# Aggiornamento strutture dati del gruppo
		#
		if len(self.scheduledQueue) > 0:
			# Se la coda di scheduling del gruppo ha un elemento, vieamo l'ultimo
			next_ts = self.scheduledQueue[-1][0] + (self.min_gap if self.min_gap is not None else 0)
		else:
			# elemento quando e' stato schedulato, altrim
			next_ts = time.time()

		# Aggiorniamo la coda locale
		self.scheduledQueue.append( (next_ts, sdg) )

		#
		# Inserimento nello scheduler principale
		#
		sdg.schedule_time = next_ts
		sdg.sched_state   = POLLSTATE_SCHEDULED
		schedulerHandler(next_ts, handler, params)

		self.queueLock.release()

		return next_ts

	def exec_or_idle(self, handler, params, agent_exec_handler):
		"""
		Check whether a datagroup can be executed or not according to max_paral,
		queue it if not the case.

		@param handler: handler function to execute datagroup
		@param params: tuple containing a ScheduledDatagroup structure (dg, )
		@param t_pool_handler: thread pool execJob method (puts scheduled datagroup in worker execution queue)
		"""
		#t1 = time.time()

		sdg = params[0]

		exec_now = True

		self.queueLock.acquire()

		self.logger.log(21, "[scheduling group: %s] exec_or_idle %s  (max_parallel = %s <-> executing %s)", self.name, sdg.pdatagroup.path, self.max_parallel,  self.executing)

		if self.max_parallel:

			# Ho raggiunto il limite di datagroup dello stesso gruppo in esecuzione contemporanemante?
			if self.executing >= self.max_parallel:

				# Non c'e' una deadline configurata, quindi il datagroup non va eseguito.
				if self.max_deadline is None:  # FIX: test_02
					exec_now = False
				else:
					# Se c'e' una deadline impostata, verifichiamo
					# quanto tempo e' passato dall'ultima esecuzione.
					if sdg.pdatagroup.lastdone:

						elapsed = time.time() - sdg.pdatagroup.lastdone

						self.logger.debug("elapsed = %s <-> max_deadline %r", elapsed, self.max_deadline)

						# ho raggiunto il limite del gruppo, ma sto sforando con i tempi di escuzione, quindi la forzo
						if elapsed >= self.max_deadline:
							self.logger.warning("%s (elapsed=%s)  MAX DEADLINE reached (%s)!", sdg.pdatagroup.path, elapsed, self.max_deadline)
							exec_now = True
						else:
							exec_now = False
					else:
						exec_now = False
			else:
				# Esecuzione parallela permessa!
				exec_now = True

		self.logger.debug("[scheduling group: %s] %s exec_now=%s", self.name, sdg.pdatagroup.path, exec_now)

		# Se devo eseguire il job adesso
		if exec_now:

			self.executing += 1 

			params[0].sched_state = POLLSTATE_QUEUED

			self.logger.debug('[scheduling group: %s] queued %s. now in execution: %s', self.name, sdg.pdatagroup.path, self.executing)

			agent_exec_handler(handler, params)
		else:
			params[0].sched_state = POLLSTATE_WAITING
			self.lockedQueue.append((handler, params))

		self.queueLock.release()

#-----------------------------------------------------------------------------#

class PCluster(Pobject):

	def __init__(self):
		super(PCluster, self).__init__()
		self.tenant_id   = None
		self.mtime       = None
		self.uuid        = None

		self.name        = None
		self.description = None

		#self.scheduling_max_deadline = None
		#self.scheduling_max_parallel = None
		#self.scheduling_min_gap      = None

		self.scheduling_group_id = None

#-----------------------------------------------------------------------------#

class PElement(Pobject):

	def __init__(self):
		super(PElement, self).__init__()

		self.tenant_id             = None
		self.mtime                 = None
		self.uuid                  = None

		self.kind                  = None
		self.name                  = None
		self.description           = None
		self.agent_id              = None
		self.is_abstract           = None
		self.parent_id             = None

		self.cluster_id            = None
		self.cluster_role          = None
		self.cluster_distinguisher = None
		self.cluster_xform         = None

		self.cluster_instance      = None

		# Volatile runtime fields
		self.cluster_element_id     = None
		self.cluster_instance_time  = None

		#self.scheduling_max_deadline = None
		#self.scheduling_max_parallel = None
		#self.scheduling_min_gap = None

		self.scheduling_group_id = None

	def __str__(self):
		return "<%s:%s>" % (self.__class__.__name__, self.name)

	def get_expr_symbols(self):
		d = {}
		d['$name'                 ] = self.name

		d['$cluster_distinguisher'] = self.cluster_distinguisher
		d['$cluster_xform'        ] = self.cluster_xform
		return d

	def load_from_dict(self, dict_data):
		super(PElement, self).load_from_dict(dict_data)

		# FIX: questi campi sono solo di comodo per l'output e sono pure valorizzati male
		self.path = self.name  # ! servirebbe usare self.parent_id

#-----------------------------------------------------------------------------#

class PNode(PElement):

	def __init__(self):
		super(PNode, self).__init__()

		self.dns_name         = None
		self.dns6_name        = None
		self.ip4_mgt          = None
		self.ip6_mgt          = None
		self.use_ipv4         = None
		self.timezone         = None
		self.snmp_version     = None
		self.snmp_port        = None
		self.snmp_community   = None
		self.snmp_v3_authdata = None
		self.dft_timeout      = None

		# L'indirizzo IP "effettivo" usato nei check deve essere comunicato al server
		# quindi abbiamo bisogno di due campi per capire se e' cambiato (e quindi
		# da aggiornare sul server).
		# 
		# L'IP "effettivo" non e' "stabile" perche' viene (ri)calcolato ad 
		# ogni DataGroupChecker.check() (che chiama calculate_effective_ip()).
		# 
		# Se il nodo risolve con indirizzi IP diversi, ci sono due problemi:
		# 
		#   1) c'e' una possibile race condition tra due thread che eseguono 
		#      controlli sullo stesso nodo se entrambi risolvono con un IP
		#      diverso al momento del check
		# 
		#      La regola e': l'IP comunicato al server e' quello 
		#                    che l'oggetto DataGroupChecker "calcola" al momento
		#                    del check(). 
		# 
		#                    Questo criterio e' analogo a quanto succede per l'ifindex.
		# 
		#   2) Ad ogni check dobbiamo comunicare un nuovo "effective_ip" da salvare su DB.
		#      Questo problema e' analogo all'aggiornamento dell'ifindex per le interfacce
		#      ma potenzialmente molto piu' frequente.
		# 
		#      Questo problema cerchiamo di gestito lato server.
		# 
		self.effective_ip         = None   # Ultimo valore noto al server
		self.current_effective_ip = None   # campo aggiornato ad ogni DataGroupChecker.check()

		self.agent_runningon  = None

	@property
	def snmp_community_or_auth(self):
		if self.snmp_version == 3:
			return self.snmp_v3_authdata
		else:
			return self.snmp_community

	def calculate_effective_ip(self):

		ip = None

		try:
			if self.use_ipv4:
				if self.ip4_mgt:
					ip = self.ip4_mgt
				else:
					name = self.dns_name  or self.name

					# FIX: in presenza di piu' indirizzi IP e' giusto
					#      risolvere tutti gli indirizzi usando un ordine "stabile"?
					#      Quando facciamo i test di raggiungibilita' tentiamo
					#      con un approccio "shuffle". Forse dovremmo fare lo stesso
					#      qui?

					ips = resolve_ip_addr(name, 4, stable_order=True)
					if ips:
						ip = ips[0]
			else:
				if self.ip6_mgt:
					ip = self.ip6_mgt
				else:
					name = self.dns6_name or self.name

					# FIX: in presenza di piu' indirizzi IP e' giusto
					#      risolvere tutti gli indirizzi usando un ordine "stabile"?
					#      Quando facciamo i test di raggiungibilita' tentiamo
					#      con un approccio "shuffle". Forse dovremmo fare lo stesso
					#      qui?

					ips = resolve_ip_addr(name, 6, stable_order=True)
					if ips:
						ip = ips[0]

			# NOTA: L'IP trovato fin qui non puo' essere None.

		except socket.gaierror, e:
			pass
		except socket.error, e:
			pass

		# 
		# IMPORTANTE: Race condition pericolosa
		#             Se un nodo risolve con tanti indirizzi IP
		#             e cambiano ad ogni controllo
		#             questo campo cambia in continuazione.
		#           
		self.current_effective_ip = ip

		return ip

	def get_expr_symbols(self, prefix=''):

		d = super(PNode, self).get_expr_symbols()

		if self.use_ipv4:
			ip_mgt = self.ip4_mgt   
			name   = self.dns_name  or self.name

			ipv    = 4  
			snmpv  = self.snmp_version
		else:
			ip_mgt = self.ip6_mgt
			name   = self.dns6_name or self.name

			ipv    = 6
			snmpv  = self.snmp_version + 60   # FIX: DON"T LIKE IT. USE $ipversion instead

		#
		# FIX: poiche' $node viene usato per specificare "dove" arrivare, ma se e' un nome DNS
		#      viene risolto ogni volta che deve essere aperta un socket (specialmente
		#      nei controlli SNMP).
		#
		#      Questo causa piu' overhead sul DNS e potenziali incongruenze tra 
		#      l'indirizzo IP effettivo che sanet ha calcolato e quello che alla fine
		#      viene usato nei controlli.
		#
		#      Sarebbe piu' corretto:
		#
		#            $node = ip or self.current_effective_ip or name
		#
		#      in questo modo tutti i controlli che passano da $node, e che non fanno
		#      altre risoluzioni DNS, arrivano al nodo dallo stesso IP.
		#
		#      Questo pero' ADESSO SPACCHERBBE tutti controlli che usano $node concatento ad altre
		#      stringhe, ecc. tipo adiacenze CDP, ecc.
		#
		#      Come risolvere?
		#

		d['$'+prefix+'ipversion'     ] = ipv

		d['$'+prefix+'node'          ] = ip_mgt or name

		d['$'+prefix+'nodename'      ] = self.name

		d['$'+prefix+'snmpversion'   ] = snmpv
		d['$'+prefix+'community'     ] = self.snmp_community
		d['$'+prefix+'snmpport'      ] = self.snmp_port
		d['$'+prefix+'snmpv3authdata'] = self.snmp_v3_authdata

		d['$'+prefix+'timeout'       ] = self.dft_timeout

		d['$'+prefix+'dnsname'       ] = self.dns_name
		d['$'+prefix+'dns6name'      ] = self.dns6_name
		d['$'+prefix+'ip4mgt'        ] = self.ip4_mgt
		d['$'+prefix+'ip6mgt'        ] = self.ip6_mgt

		# Se non siamo riusciti a calcolare un "effective_ip" per il nodo,
		# non deve essere possibile usare $node_effective_ip nelle expr.
		if self.current_effective_ip is not None:
			d['$'+prefix+'node_effective_ip' ] = self.current_effective_ip

		return d

#-----------------------------------------------------------------------------#

class PIface(PElement):

	def __init__(self):
		super(PIface, self).__init__()

		self.distinguisher = None
		self.xform         = None
		self.ifindex       = None
		self.backbone      = None
		self.speed         = None

	def get_expr_symbols(self, prefix=''):
		d = super(PIface, self).get_expr_symbols()

		d['$'+prefix+'iface'        ] = self.name
		d['$'+prefix+'instance'     ] = self.distinguisher
		d['$'+prefix+'distinguisher'] = self.distinguisher
		d['$'+prefix+'backbone'     ] = self.backbone
		d['$'+prefix+'ifspeed'      ] = self.speed

		return d

#-----------------------------------------------------------------------------#

class PStorage(PElement):

	def __init__(self):
		super(PStorage, self).__init__()

		self.distinguisher = None
		self.xform         = None
		self.stindex       = None

	def get_expr_symbols(self, prefix=''):
		d = super(PStorage, self).get_expr_symbols()

		d['$'+prefix+'storage'      ] = self.name
		d['$'+prefix+'distinguisher'] = self.distinguisher

		return d

#-----------------------------------------------------------------------------#

class PService(PElement):

	def __init__(self):
		super(PService, self).__init__()

		self.distinguisher = None
		self.xform         = None
		self.swrunindex    = None

	def get_expr_symbols(self, prefix=''):
		d = super(PService, self).get_expr_symbols()

		d['$'+prefix+'service'        ] = self.name
		d['$'+prefix+'distinguisher'  ] = self.distinguisher

		return d

#-----------------------------------------------------------------------------#

class PDevice(PElement):

	def __init__(self):
		super(PDevice, self).__init__()

		self.distinguisher = None
		self.xform         = None
		self.devindex      = None

	def get_expr_symbols(self, prefix=''):
		d = super(PDevice, self).get_expr_symbols()

		d['$'+prefix+'device'         ] = self.name
		d['$'+prefix+'distinguisher'  ] = self.distinguisher

		return d

#-----------------------------------------------------------------------------#

class PLink(Pobject):

	def __init__(self, uuid, name, ltype, ifaces):
		super(PLink, self).__init__()

		self.tenant_id    = None
		self.mtime        = None

		self.uuid         = uuid

		self.name         = name
		self.ltype        = ltype

		self.ifaces       = ifaces

	def __str__(self):

		return "<%s,%s,%s,%s>" % (self.uuid, self.name, self.ltype, self.ifaces)

#-----------------------------------------------------------------------------#

class PDataGroup(Pobject):

	def __init__(self):
		super(PDataGroup, self).__init__()

		self.mtime             = None

		self.lastdone          = None

		self.datasources       = [] # list of uuids
		self.conditions        = [] # list of uuids

		# FIX: specify all fields
		self.suspend_resume_time = None

		# Reference to the PSchedulingGroup
		self.scheduling_group = None

	def __str__(self):
		return self.uuid

	def get_dict(self):
		#
		# OVERRIDE: self.scheduling_group is not JSON serializable
		#
		c = copy.copy(self.__dict__)
		del c['scheduling_group']  
		return c

	def load_from_dict(self, dict_data):
		for k, v in dict_data.items():
			setattr(self, k, v)

	def get_datasources_ids(self):
		return self.datasources

	def get_conditions_ids(self):
		return self.conditions

	def load_check_state(self, cache):

		key = "%s.checkstate" % (self.uuid)

		old_state = cache.get(key )

		if old_state == None:
			return {}

		return old_state

	def save_check_state(self, timestamp, state, cache):

		old_state = {}
		old_state.update( state )
		old_state['__timestamp__'] = timestamp

		key = "%s.checkstate" % (self.uuid)  #FIX: comporre la key in modo piu' "automatizzato"

		cache.set( key, old_state )

	def get_expr_symbols(self):

		d = {}
		d['$path'      ] = self.path
		d['$shorttries'] = self.shorttries
		d['$timeout'   ] = self.timeout

		return d

	def valid_for_check(self):
		if self.suspended:
			if not self.suspend_resume_time:
				return False
		return True

	def is_suspended(self):
		# FIX: anything else to check with this method?
		return (self.suspended == True)

	def can_be_resumed(self, now):

		if self.suspend_resume_time:
			if self.suspend_resume_time <= now:
				return True

		return False

#-----------------------------------------------------------------------------#

class PDataSource(Pobject):

	def __init__(self):
		super(PDataSource, self).__init__()

		self.tenant_id    = None
		self.mtime        = None
		self.uuid         = None

		#self.lastcheck_total_time = -1

		self.lastchecktotaltime = None

	def load_from_dict(self, dict_data):
		for k, v in dict_data.items():
			setattr(self, k, v)

	def __str__(self):
		return self.path

	def is_valid_value(self, value):
		return isinstance(value, types.FloatType) or isinstance(value, types.IntType) or isinstance(value, types.LongType)

	def is_value_in_range(self, value):

		if self.min_val is not None or self.max_val is not None:
			if self.min_val != None:
				if value < self.min_val:
					return False
			if self.max_val != None:
				if value > self.max_val:
					return False

		return True

	def get_expr_symbols(self):

		d = {}

		d['$path'                  ] = self.path

		d['$_lastchecktotaltime'   ] = self.lastchecktotaltime

		return d

#-----------------------------------------------------------------------------#

class PCondition(Pobject):

	def __init__(self):

		super(PCondition, self).__init__()

		self.tenant_id    = None
		self.mtime        = None
		self.uuid         = None

		#self.lastcheck_total_time = -1

		self.datagroup            = None       # FIX: not used. remove as soon as possible
		
		self.datagroup_id         = None
		self.element_id           = None
		self.path                 = None
		self.name                 = None
		self.title                = None
		self.primary              = None
		self.classification       = None
		self.dependson            = None       # FIX: not used. remove as soon as possible
		self.dependson_id         = None       
		
		self.order                = None
		self.cascade              = None
		
		self.expr                 = None

		self.max_tries            = None
		self.priority             = None
		self.uncheckable_fallback = None
		self.statuschange_action  = None

		self.lastdone             = None
		self.status               = None
		self.verbstatus           = None

		self.statuslast           = None
		self.statuslastchange     = None

		#self.statuslastnotable           = None
		#self.statuslastnotablechange     = None

		self.tries               = None
		self.penalty             = None
		self.dampened            = None

		# See if it is time to review the semantics of email
		self.email        = None
		self.upemail      = None
		self.msg_downsubj = None
		self.msg_downbody = None
		self.msg_upsubj   = None
		self.msg_upbody   = None

		self.lastchecktotaltime = None

	def load_from_dict(self, dict_data):
		for k, v in dict_data.items():
			setattr(self, k, v)

	def __str__(self):
		return self.path


	def get_expr_symbols(self):

		d = {}
		d['$path'                  ] = self.path
		d['$_cname_'               ] = self.name

		d['$_lastchecktotaltime'   ] = self.lastchecktotaltime      # FIX: no more useful. remove ASAP.

		return d

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#
#
#  AbstractPollerConfiguration
#     |
#     +----- AbstractLocalPollerConfiguration
#     |
#     +----- PollerConfiguration
#  

class AbstractPollerConfiguration(object):
	"""
	Oggetto usato dal codice del poller per avere la configurazione corrente
	delle varie entita' del monitoraggio (datagroup, datasource, condition,
	nodi, interfacce, storage, service, device, link, parametri)
	"""

	def __init__(self, logger, tenant_id):

		self.tenant_id = tenant_id

		self.logger = logger
		self.global_symbols = None
		self.load_time = None
		return

	def invalid_all_elements(self):
		return NotImplementedException("invalid_all_elements")

	def get_info(self):
		raise NotImplementedException("get_info")

	def set_load_time(self, ts):
		self.load_time = ts

	def get_load_time(self):
		return self.load_time

	def set_global_symbols(self, symbols):
		self.global_symbols = symbols

	def get_global_symbols(self):
		return self.global_symbols

	def get_tenant_id(self):
		return self.tenant_id

	def get_cluster(self, cid):
		raise NotImplementedException("get_cluster")

	def get_cluster_elements(self, cid):
		raise NotImplementedException("get_cluster_elements")

	def get_datagroup(self, dgid):
		raise NotImplementedException("get_datagroup")

	def get_datagroup_params(self, dgid):
		raise NotImplementedException("get_datagroup_params")

	def get_datasource(self, dsid):
		raise NotImplementedException("get_datasource")

	def get_condition(self, cid):
		raise NotImplementedException("get_condition")

	def get_condition_by_path(self, *paths):
		raise NotImplementedException("get_condition_by_path")

	def get_condition_by_classification(self, classification, node_id, element_id):
		raise NotImplementedException("get_condition_by_classification")

	def get_element_by_id(self, e_id):
		raise NotImplementedException("get_element_by_id")

	def get_element_by_parent_id(self, p_id, kind):
		raise NotImplementedException("get_element_by_id")

	def get_linked_iface_by_id(self, iface_id, link_type=None):
		raise NotImplementedException("get_linked_iface_by_id")

	def get_info(self):
		raise NotImplementedException("get_info")

class AbstractLocalPollerConfiguration(AbstractPollerConfiguration):
	"""
	Alcuni comandi di utility devono eseguire il check di un datagroup
	come se avessero letto i dati "localmente".
	"""

	def __init__(self, logger, tenant_id, cache_manager):

		super(AbstractLocalPollerConfiguration, self).__init__(logger, tenant_id)
		self.cache_manager = cache_manager

		# NOTA: tutte le entita' sono indicizzate per UUID. Non dovrebbero esserci
		#       collisioni! Forse!
		self._cache = {} 

        #--------------------------------------------------------------#

	def get_cluster(self, cid):

		if self._cache.get(cid, None):
			return self._cache[cid]

		self.logger.log(9, "Loading cluster %s", cid)
		try:
			data = self._get_cluster(cid)
			pc = PCluster()
			pc.load_from_dict(data)

			self._cache[cid] = pc

			return pc
		except ClusterException, e:
			raise e

	def _get_cluster(self, cid):
		raise NotImplementedException("get_cluster")

        #--------------------------------------------------------------#

	def get_cluster_elements(self, cid):

		c_id = '%s_elements' % (cid)
		if self._cache.get(c_id, None):
			return self._cache[c_id]

		self.logger.log(9,"Loading cluster elements for cluster %s", cid)

		try:
			elements_ids = self._get_cluster_elements_ids(cid)
			r = []
			for eid in elements_ids:
				element = self.get_element_by_id(eid)
				r.append(element)

			self._cache[c_id] = r

			return r
		except 	ClusterException, e:
			raise e 

	def _get_cluster_elements_ids(self, cid):
		raise NotImplementedException("get_cluster_elements")

	#--------------------------------------------------------------#

	def get_datagroup(self, dgid):

		if self._cache.get(dgid, None):
			return self._cache[dgid]

		self.logger.log(9, "Loading datagroup %s", dgid)
		try:
			data = self._get_datagroup(dgid)
			pdg = PDataGroup()
			pdg.load_from_dict(data)

			self._cache[dgid] = pdg

			return pdg
		except DataGroupNotFoundException, e:
			raise e

	def _get_datagroup(self, dgid):
		raise NotImplementedException("get_datagroup")

	#--------------------------------------------------------------#

	def get_datagroup_params(self, dgid):

		p_id = '%s_params' % (dgid)
		if self._cache.get(p_id, None):
			return self._cache[p_id]

		try:
			data = self._get_datagroup_params(dgid)
			params = [ p for p in data ]

			self._cache[p_id] = params

			return params
		except DataGroupNotFoundException, e:
			return e

	def _get_datagroup_params(self, dgid):
		raise NotImplementedException("get_datagroup_params")

	#--------------------------------------------------------------#

	def get_datasource(self, dsid):

		if self._cache.get(dsid, None):
			return self._cache[dsid]

		self.logger.log(9, "Loading datasource %s", dsid)
		try:
			data = self._get_datasource(dsid)
			pds = PDataSource()
			pds.load_from_dict(data)

			self._cache[dsid] = pds

			return pds
		except DataSourceNotFoundException, e:
			raise e

	def _get_datasource(self, dsid):
		return NotImplementedException("get_datasource")

	#--------------------------------------------------------------#

	def get_condition(self, cid):

		if self._cache.get(cid, None):
			return self._cache[cid]

		try:
			self.logger.log(9, "Loading condition %s", cid)
			data = self._get_condition(cid)
			pc = PCondition()
			pc.load_from_dict( data )

			self._cache[pc.uuid] = pc

			return pc
		except ConditionNotFoundException, e:
			raise e

	def _get_condition(self, cid):
		raise NotImplementedException("get_condition")

	#--------------------------------------------------------------#

	def get_condition_by_path(self, path):

		if self._cache.get(path, None):
			return self._cache[path]

		try:
			self.logger.log(9, "Loading condition %s", path)

			data = self._get_condition_by_path(path)

			pc = PCondition()
			pc.load_from_dict( data )

			self._cache[path] = pc

			return pc
		except ResourceNotFoundException, e:
			raise ConditionNotFoundException("Condition %s not found (%s)" % (path, type(e) ))
		except KeyError, e:
			raise ConditionNotFoundException("Condition %s not found (%s)" % (path, type(e) ))

	def _get_condition_by_path(self, path):
		raise NotImplementedException("get_condition_by_path")

	#--------------------------------------------------------------#

	def get_condition_by_classification(self, classification, node_id, subelement_id):

		ck = "%s_%s_%s" % (classification, node_id, subelement_id)

		if self._cache.get(ck, None):
			return self._cache[ck]

		try:
			self.logger.log(9, "Loading condition %s", classification)

			data = self._get_condition_by_classification(classification, node_id, subelement_id)

			pc = PCondition()
			pc.load_from_dict( data )

			self._cache[ck] = pc

			return pc
		except ResourceNotFoundException, e:
			raise ConditionNotFoundException("Condition %s not found (%s)" % (classification, type(e) ))
		except KeyError, e:
			raise ConditionNotFoundException("Condition %s not found (%s)" % (classification, type(e) ))

	def _get_condition_by_classification(self, path, node_id, subelement_id):
		raise NotImplementedException("get_condition_by_classification")

	#--------------------------------------------------------------#

	def get_element_by_id(self, e_id):

		if self._cache.get(e_id, None):
			return self._cache[e_id]

		try:
			self.logger.log(9, "Loading element %s", e_id)
			e = self._get_element_by_id(e_id)
			if   e['kind'] == ELEMENT_KIND_NODE   : pe = PNode()
			elif e['kind'] == ELEMENT_KIND_IFACE  : pe = PIface()
			elif e['kind'] == ELEMENT_KIND_STORAGE: pe = PStorage()
			elif e['kind'] == ELEMENT_KIND_SERVICE: pe = PService()
			elif e['kind'] == ELEMENT_KIND_DEVICE : pe = PDevice()
			else: raise Exception("INVALID KIND: %s" % (e.kind))
			pe.load_from_dict(e)

			self._cache[e_id] = pe

			return pe
		except ResourceNotFoundException, ex:
			raise ex

	def _get_element_by_id(self, e_id):
		raise NotImplementedException("get_condition_by_path")

        #--------------------------------------------------------------#

	def get_element_by_parent_id(self, p_id, kind):

		elements = []

		try:
			self.logger.log(9, "Loading element with parent %s", p_id)

			loaded_elements = self._get_element_by_parent_id(p_id, kind)
			for e in loaded_elements:
				if   e['kind'] == ELEMENT_KIND_NODE   : pe = PNode()
				elif e['kind'] == ELEMENT_KIND_IFACE  : pe = PIface()
				elif e['kind'] == ELEMENT_KIND_STORAGE: pe = PStorage()
				elif e['kind'] == ELEMENT_KIND_SERVICE: pe = PService()
				elif e['kind'] == ELEMENT_KIND_DEVICE : pe = PDevice()
				else: raise Exception("INVALID KIND: %s" % (e.kind))
				pe.load_from_dict(e)

				elements.append(pe)

			return elements

		except ResourceNotFoundException, ex:
			raise ex

	def _get_element_by_parent_id(self, p_id, kind):
		raise NotImplementedException("get_element_by_id")

	#--------------------------------------------------------------#

	def get_linked_iface_by_id(self, iface_id, link_type = None):

		l_id = '%s_%s' % (iface_id, link_type)
		if self._cache.get(l_id, None):
			return self._cache[l_id]

		try:
			self.logger.log(9, "Loading linked interface by id: %s",  iface_id)
			data = self._get_linked_iface_by_id(iface_id, link_type=link_type)
			if data is not None:

				linked_pe = PIface()
				linked_pe.load_from_dict( data )
			else:
				linked_pe = None

			self._cache[l_id] = linked_pe

			return linked_pe
		except ResourceNotFoundException, e:
			raise e

	def _get_linked_iface_by_id(self, iface_id, link_type = None):
		raise NotImplementedException("get_linked_iface_by_id")

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class PollerConfiguration(AbstractPollerConfiguration):

	def __init__(self, logger, tenant_id):

		super(PollerConfiguration, self).__init__(logger, tenant_id)

		self.ctime = get_time()

		self.last_config_time = None

		# Indexed by id (uuid)
		self.clusters    = {}
		self.clusters_elements = {}

		self.elements    = {}

		self.links       = {}
		self.links_by_iface = {}

		self.datagroups        = {}
		self.datagroups_params = {}

		self.datasources = {}
		self.conditions  = {}

	def get_info(self):

		nodes_count    = len( [ 1 for e in self.elements.values() if e.kind == ELEMENT_KIND_NODE    ] )
		ifaces_count   = len( [ 1 for e in self.elements.values() if e.kind == ELEMENT_KIND_IFACE   ] )
		devices_count  = len( [ 1 for e in self.elements.values() if e.kind == ELEMENT_KIND_DEVICE  ] )
		storages_count = len( [ 1 for e in self.elements.values() if e.kind == ELEMENT_KIND_STORAGE ] )
		services_count = len( [ 1 for e in self.elements.values() if e.kind == ELEMENT_KIND_SERVICE ] )

		d = {
			'load_time'        : self.load_time,
			'clusters_count'   : len(self.clusters),
			'elements_count'   : len(self.elements),
			'nodes_count'      : nodes_count,
			'ifaces_count'     : ifaces_count,
			'storages_count'   : storages_count,
			'devices_count'    : devices_count,
			'services_count'   : services_count,
			'links_count'      : len(self.links),
			'datagroups_count' : len(self.datagroups),
			'datasources_count': len(self.datasources),
			'conditions_count' : len(self.conditions),
		}
		return d

	def invalid_all_elements(self):
		for e in self.clusters.values()   : e.set_invalid()
		for e in self.elements.values()   : e.set_invalid()
		for e in self.links.values()      : e.set_invalid()
		for e in self.datagroups.values() : e.set_invalid()
		for e in self.datasources.values(): e.set_invalid()
		for e in self.conditions.values() : e.set_invalid()

	def add_cluster(self, cluster):
		self.clusters[ cluster.uuid ] = cluster
		self.clusters_elements[ cluster.uuid ] = {}

	def add_element(self, e):
		self.elements [ e.uuid ] = e
		self.add_cluster_element(e)

	def add_cluster_element(self, e):	
		if e.cluster_id == None: return

		c_elements = self.clusters_elements[e.cluster_id]
		c_elements[ e.uuid ] = e

	def add_link(self, link):
		self.links[ link.uuid ] = link

		i1_id = link.ifaces[0]
		i2_id = link.ifaces[1]

		iface_links = self.links_by_iface.get( i1_id , self.links_by_iface.setdefault( i1_id, []) )
		iface_links.append(link)

		iface_links = self.links_by_iface.get( i2_id , self.links_by_iface.setdefault( i2_id, []) )
		iface_links.append(link)

	def add_datagroup(self, dg):
		self.datagroups[ dg.uuid ] = dg

	def add_datagroup_params(self, dgid, params):
		self.datagroups_params[ dgid ] = params

	def add_datasource(self, ds):
		self.datasources[ ds.uuid ] = ds

	def add_condition(self, c):
		self.conditions[ c.uuid ] = c

	def get_cluster(self, cid):
		try:
			return self.clusters[cid]
		except KeyError, e:
			raise ClusterException("cluster %s not found. (not loaded?)" % (cid))

	def get_cluster_elements(self, cid):
		try:
			return self.clusters_elements[ cid ].values()
		except KeyError, e:
			raise ClusterException("cluster %s not found. (not loaded?)" % (cid))

	def get_datagroup(self, dgid):
		try:
			return self.datagroups[dgid]
		except KeyError, e:
			raise DataGroupNotFoundException("DataGroup %s not found" % (dgid))

	def get_datagroup_params(self, dgid):
		r = self.datagroups_params.get(dgid, {})
		return r

	def get_datasource(self, dsid):
		try:
			return self.datasources[dsid]
		except KeyError, e:
			raise DataSourceNotFoundException("DataSource %s not found" % (dsid))

	def get_condition(self, cid):
		try:
			return self.conditions[cid]
		except KeyError, e:
			raise ConditionNotFoundException("Condition %s not found" % (cid))

	def get_scheduling_group(self, sgid):
		try:
			return self.scheduling_groups[sgid]
		except KeyError, e:
			raise SchedulingGroupNotFoundException("Scheduling group %s not found" % (sgid))

	def get_condition_by_path(self, path):

		for c_path, c in self.conditions.items():
			if c_path == path:
				return c

		raise ConditionNotFoundException("Condition %s not found" % (path))

	def get_condition_by_classification(self, classification, node_id, subelement_id):

		selected_c = None

		condition_list = self.conditions.items()

		for c_path, c in condition_list:
			if c.classification == classification and c.element_id == subelement_id:
				selected_c = c
				break

		for c_path, c in condition_list:
			if c.classification == classification and c.element_id == node_id:
				selected_c = c
				break
		if selected_c:
			self.logger.debug("Found condition in configuration")
			return selected_c
		else:
			raise ConditionNotFoundException("Condition %s not found" % (classification))

	def reindex_datagroups(self):

		self._datasource_by_dg_uuid = {}
		for e in  self.datasources.values():
			l = self._datasource_by_dg_uuid.get( e.datagroup_id,  self._datasource_by_dg_uuid.setdefault(e.datagroup_id, [])  )
			l.append( e.uuid )

		self._conditions_by_dg_uuid = {}
		for e in  self.conditions.values():
			l = self._conditions_by_dg_uuid.get( e.datagroup_id,  self._conditions_by_dg_uuid.setdefault(e.datagroup_id, [])  )
			l.append( e.uuid )

	def get_datasources_ids_by_dg_id(self, dg_id):
		l = self._datasource_by_dg_uuid.get(dg_id, [])
		return l

	def get_conditions_ids_by_dg_id(self, dg_id):
		l = self._conditions_by_dg_uuid.get(dg_id, [])
		return l

	def get_element_by_id(self, e_id):
		e = self.elements.get(e_id, None)
		if e == None:
			raise ElementNotFoundException("element not found:%s" % (e_id))
		return e

	def get_element_by_parent_id(self, p_id, kind):
		return [ e for e in self.elements.values() if (e.parent_id == p_id and e.kind == kind) ]

	def get_linked_iface_by_id(self, iface_id, link_type=None):

		links = self.links_by_iface.get(iface_id, None)
		if links == None:
			return None

		# Filter links by type 
		l = None
		if link_type != None:
			for link in links:
				if link.ltype == link_type:
					l = link
					break
		else:
			l = links[0]

		if l is None:
			return None

		# [ <iface1> , <iface2> ]
		if l.ifaces[0] == iface_id:
			iid = l.ifaces[1]
		else:
			iid = l.ifaces[0]

		try:
			return self.elements[iid]
		except KeyError, e:
			raise ResourceNotFoundException("Linked iface %s for %s not found" % (iid, iface_id))


class PollerConfigurationWithSchedulingGroups(PollerConfiguration):

	def __init__(self, logger, tenant_id):

		super(PollerConfigurationWithSchedulingGroups, self).__init__(logger, tenant_id)

		self.scheduling_groups  = {}

		return

	def add_scheduling_group(self, sg):
		self.scheduling_groups[ sg.uuid ] = sg

	def get_scheduling_group(self, sgid):
		try:
			return self.scheduling_groups[sgid]
		except KeyError, e:
			raise SchedulingGroupNotFoundException("Scheduling group %s not found" % (sgid))

	def get_scheduling_group_by_dg_id(self, dg_id):
		"""
		Questo metodo e' chiamato solo in fase di caricamento della nuova configurazione
		per trovare lo scheduling group associato ad un datagroup.
		"""

		# Poiche' e' possibile che "dg_id" non esistesse nella configurazione quando l'ho caricata
		# al giro precedente, questo metodo non deve andare in errore, ma restituire None.
		# - dg_id                     non esiste
		# - pdg.scheduling_group_id   non esiste
		#

		pdg = self.datagroups.get(dg_id, None)  
		if pdg is None:
			return None

		return self.scheduling_groups.get(pdg.scheduling_group_id, None)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def DataGroupUpdate( check_time, uid, fields, mtime_sentinel=None ):
	return {
		'time'  : check_time,

		'uuid'  : uid,
		'smtime': mtime_sentinel, 

		'fields': fields
	}

def DataSourceUpdate( check_time, uid, name, fields, infos, mtime_sentinel=None, nosave=0):
	d = {
		'time'  : check_time,

		'uuid'  : uid,
		'name'  : name,

		'smtime': mtime_sentinel, 

		'fields': fields,

		'infos' : infos,
	}

	if nosave:
		d['nosave'] = 1 

	return d

def ConditionUpdate( check_time, uid, fields, infos, mtime_sentinel=None):
	return {
		'time'  : check_time,

		'uuid'  : uid,
		'smtime': mtime_sentinel, 

		'fields': fields,

		'infos' : infos,
	}

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class DataGroupChecker(object):

	CONDITION_FIELDS_TO_CHECK_FOR_CHANGES = [
		'lastdone',

		'tries',

		'penalty',
		'dampened',

		'status',
		'verbstatus',

		'statuslast',
		'statuslastchange',

		#'statuslastnotable',
		#'statuslastnotablechange',
	]

	def __init__(self, configuration, logger, cache, ipc_factory, pdatagroup):

		self.log           = logger
		self.cache_manager = cache

		self.ipc_factory   = ipc_factory

		self.configuration = configuration

		self.pdatagroup    = pdatagroup

		self._agent         = None

		#
		# Temp data
		#
		self.node                 = None

		#
		# Update data
		#

		# Sequence of updates
		self.update_chunks = []

		# Current updates to elements' fields
		self.update_fields_by_element_id = None
		# Current updates to datagroup fields
		self.datagroup_updates           = None
		# Current updates to datasources' fields
		self.datasources_updates         = None
		# Current updates to conditions' fields
		self.conditions_updates          = None
		# Current SNMP OIDs and Saved values
		self.saved_variables             = None

		self.reset_datagroup_check_updates()

		#
		# Cache 
		#
		self.__related_datasources = None
		self.__related_conditions  = None

		self.exec_context = None #DA CHIEDERE
		
		return

	def set_agent(self, agent):
		self._agent = agent

	def dispose(self):
		del self.log
		del self.cache_manager
		del self.ipc_factory

		del self.configuration

		del self.pdatagroup
		del self.update_chunks
		del self.update_fields_by_element_id
		del self.datagroup_updates
		del self.datasources_updates
		del self.conditions_updates
		del self.saved_variables

		del self.__related_datasources
		del self.__related_conditions


	def get_related_datasources(self):
		if self.__related_datasources == None: # cache
			self.__related_datasources = [ self.configuration.get_datasource(dsid) for dsid in self.pdatagroup.get_datasources_ids() ]

		return self.__related_datasources

	def get_related_conditions(self):
		if self.__related_conditions  == None: # cache
			self.__related_conditions  = [ self.configuration.get_condition(cid)   for cid  in self.pdatagroup.get_conditions_ids()  ]

		return self.__related_conditions

	#
	# Remember 
	#
	def save_datagroup_update(self, update_data):
		self.datagroup_updates = update_data 

	def save_datasource_update(self, id, update_data):
		self.datasources_updates[id] = update_data

	def save_condition_update(self, id, update_data):
		self.conditions_updates[id] = update_data

	def reset_datagroup_check_updates(self):
		self.update_fields_by_element_id = {}
		self.datagroup_updates           = {}
		self.datasources_updates         = {}
		self.conditions_updates          = {}
		self.additional_updates          = None
		self.saved_variables            = {}

	def get_datagroup_check_updates(self, ts):

		if not (self.update_fields_by_element_id or \
		        self.datagroup_updates           or \
		        self.datasources_updates         or \
		        self.conditions_updates          or \
		        self.additional_updates          or \
		        self.saved_variables
		       ):
			return None

		d = {
			'ts'                 : ts,
			'element_id'         : self.pdatagroup.element,
			'datagroup_id'       : self.pdatagroup.uuid,

			'elements'           : self.update_fields_by_element_id,

			'datagroup_updates'  : self.datagroup_updates,
			'datasources_updates': self.datasources_updates,
			'conditions_updates' : self.conditions_updates,
			
			'additional_updates' : self.additional_updates,
			
			'saved_variables'    : self.saved_variables,
		}
		return d

	def push_updates(self, ts):

		# Get all changes up to now
		current_update_chunk = self.get_datagroup_check_updates(ts)
		if current_update_chunk is not None:
			self.update_chunks.append( current_update_chunk )

		# Reset changes 
		self.reset_datagroup_check_updates()

	def reset_updates_history(self):
		self.update_chunks = []

	def get_datagroup_all_updates(self):
		return self.update_chunks

	#
	#
	#

	def get_expr_symbols(self, check_time):

		pdatagroup = self.pdatagroup

		#
		# Get element (and node if necessary)
		#

		element = self.configuration.get_element_by_id( pdatagroup.element )

		node = self.get_element_node( element )
		self.node = node

		self.log.debug("Element: %s", element)
		self.log.debug("Node   : %s", node)

		node.calculate_effective_ip()

		self.log.debug("Current effective IP: %s (last was %s)", node.current_effective_ip, node.effective_ip)

		#
		# Get symbols
		#

		symbols = {}

		# Get element symbols
		symbols['__element__'] = element
		symbols['__node__']    = node

		symbols.update( element.get_expr_symbols() )

		if node != element:
			symbols.update( node.get_expr_symbols() )

		if element.kind == ELEMENT_KIND_IFACE:

			# FIX: interface specific. Move in the PIface object
			# FIX: find the LINK_TYPE_L2
			linked_iface = self.configuration.get_linked_iface_by_id( element.uuid, link_type=LINK_TYPE_L2 )

			if linked_iface != None:
				linked_node = self.get_element_node( linked_iface )

				symbols['__linked_iface__'] = linked_iface
				symbols['__linked_node__' ] = linked_node

				symbols.update( linked_iface.get_expr_symbols(prefix='linked_') )
				symbols.update( linked_node.get_expr_symbols(prefix='linked_') )

			symbols['__iface__'] = element

		elif element.kind == ELEMENT_KIND_STORAGE:
			symbols['__storage__'] = element

		elif element.kind == ELEMENT_KIND_SERVICE:
			symbols['__service__'] = element

		elif element.kind == ELEMENT_KIND_DEVICE:
			symbols['__device__'] = element

		# datagroup symbols
		symbols.update( pdatagroup.get_expr_symbols() )

		return symbols

	def get_datagroup_params(self):

		self.log.debug("Loading free parameters")

		params = self.configuration.get_datagroup_params( self.pdatagroup.uuid )

		symbols = {}
		for p in params:
			self.log.debug("  params '%s' = (%s) %s " % (p['name'],p['vtype'], p['value'])) 
			sym_name  = '$'+p['name']
			sym_value = convert_to_param_type(p['vtype'], p['value'])
			symbols[ sym_name ] = sym_value

		return symbols

	def get_element_node(self, element):
		"""
		Restituisce l'elemento Nodo che contiene l'elemento "element".
		"""
		e = element
		while e.parent_id != None:
			e = self.configuration.get_element_by_id(e.parent_id)

		return e

	#--------------------------------------------------------------------#
	# CHECK                                                              #
	#--------------------------------------------------------------------#

	def createNewExecutionContext(self):
		c = ExecutionContext()

		c['__agent__'] = self._agent

		return c

	def check(self, check_time, scheduled_time=None):
		"""
		Effettua il check completo di un datagroup.
		
		L'eccezione ResourceNotFoundException e' l'unia che non viene intercettata/nascosta
		"""

		ts_dg_check_start = time.time()

		self.log.debug("[checker] checking datagroup %s at %s", self.pdatagroup.path, check_time)

		#
		# Suspended? Needs resume?
		#
		self.log.debug("----------------------- Checking resume ----------------------")
		check_now, check_time = self._resume_if_suspended(check_time, self.pdatagroup)
		if check_now == False:
			return

		#
		# Check dependson for this datagroup
		#
		self.log.debug("----------------------- Checking dependson ----------------------")
		can_check, reason = self.check_datagroup_dependson()
		if not can_check:
			infos = [ reason, ]
			self.update_datagroup_on_dependson(self.pdatagroup, check_time, infos)
			self.push_updates(check_time)
			return

		#
		# Load previous state data
		#
		self.log.debug("----------------------- Start checking ----------------------")

		try:
			self.log.debug("Loading previeous execution state...")
			previous_check_values = self.pdatagroup.load_check_state( self.cache_manager )
		except Exception, e:
			self.log.error("Unable to load previous check state from cache for %s: %s:%s", self.pdatagroup.path, type(e), str(e))
			previous_check_values = {}

		if self.log.isEnabledFor(logging.DEBUG): 
			for k,v in sorted(previous_check_values.items()):   # FIX: Disable unseful code when not in debug mode
				self.log.log(1,"\t%s = %s", k, v)

		#
		# Prepare (almost) all global symbols used inside the expressions
		#
		symbols = {}
		symbols.update( self.get_datagroup_params() )
		symbols.update( self.get_expr_symbols(check_time) )
		symbols.update( self.configuration.get_global_symbols() )

		if self.log.isEnabledFor(1): 
			self.log.log(1, "Global symbols for all checks:")
			for s,v in sorted(symbols.items()):
				self.log.log(1, "\t%s = %s", s, v)

		#
		# Preparing Execution context 
		#
		exec_context = self.createNewExecutionContext()

		exec_context.setLog( self.log )
		exec_context.setCache(self.cache_manager)
		exec_context.setIPCFactory(self.ipc_factory)

		exec_context.setConfiguration( self.configuration )

		exec_context.update( symbols )
		
		# Load previous values
		exec_context.setOldValues( previous_check_values )

		exec_context['__check_time__'] = check_time

		#
		# Evaluate all datasources
		#

		self.log.log(1,"Loading datasources to check")

		pdatasources = self.get_related_datasources()
		if pdatasources:

			cascade_ds    = None
			cascade_error = None

			# Check depending on the order
			pdatasources = sorted(pdatasources, key=attrgetter('order') )

			for pdatasource in pdatasources:

				self.log.debug("-------------------------- Calculating datasource %s ---------------- ", pdatasource.name)

				if cascade_ds == None:

					# Reset execution infos generated by previous datasources, except SNMP values and other execution data.
					exec_context.resetExecInfos()

					# Evaluating Expression
					status = self.check_datasource(pdatasource, exec_context, previous_check_values)

					# IMPORTANT: Datasource is uncheckable; the "ExecInfos" probably explain the reason.
					#            Remember the "ExecInfos" for the future.
					#
					# See: RealDataIdentifierExprTree.calculateDataValue()
					exec_context.saveExecInfoByName(pdatasource.name)

					# Stop evaluation of the next datasources, if necessary.
					if status not in [ DATASOURCE_STATUS_OK, DATASOURCE_STATUS_STRANGE_VAL ]  and pdatasource.cascade == True:
						cascade_ds    = pdatasource
						cascade_error = exec_context.getSavedExecInfo(cascade_ds.name)
				else:
					# Skip datasources on cascade error.
					ctime  = get_time()
					status = DATASOURCE_STATUS_CASCADE 
					value  = None
					info   = "Not evaluated because of cascade error on datasource '%s' (Details: %s)" % (cascade_ds.name, cascade_error)
					exec_context.appendExecutionInfo(info)

					# Save info about this datasource
					exec_context.saveExecInfoByName(pdatasource.name)

					# Prepare update data for the server
					fields = {
						'status'   : status,
						'lastvalue': value,
					}
					update_data = DataSourceUpdate(check_time, pdatasource.uuid, pdatasource.name, fields, [ info, ] , mtime_sentinel=pdatasource.mtime)
					self.save_datasource_update(pdatasource.uuid, update_data)
		else:
			self.log.debug("No datasources to check.")

		#
		# Evaluate all conditions
		#
		self.log.log(1,"Loading conditions to check")

		pconditions = self.get_related_conditions()
		if pconditions:

			primary_cond = None

			cascade_c     = None
			cascade_error = None

			# Evaluate primary conditions first
			pconditions = [ e[3] for e in sorted([(not pc.primary, pc.order, pc.name, pc) for pc in pconditions])  ]

			for pcondition in pconditions:

				self.log.debug("-------------------------- Calculating condition %s (primary %s)---------------- ", pcondition.name, pcondition.primary)

				# No cascade error
				if cascade_c == None:

					# Check depenson on "local" conditions
					if primary_cond is None and pcondition.primary:
						primary_cond = pcondition
					elif primary_cond is not None:
						can_check, reason = self.__check_condition_for_dependson(primary_cond)
						if not can_check:
							infos = [ reason, ]
							self.update_condition_on_dependson(check_time, pcondition, infos)

							self.exec_context = exec_context #DA CHIEDERE

							self.push_updates(check_time)
							continue

					# Reset execution infos generated by previous conditions, except SNMP values and other execution data.
					exec_context.resetExecInfos()

					# Check
					result = self.check_condition(pcondition, exec_context, previous_check_values)

					# Stop evaluation of the next conditions, if necessary.
					if not result:
						#exec_context.saveExecInfoByName(pcondition.name)

						if pcondition.cascade == True:
							cascade_c     = pcondition
							#cascade_error = exec_context.getSavedExecInfo(pcondition.name)
							cascade_error = "Result is %s" % (result)
				else:
					# Skip on cascade error.
					reason   = "Not evaluated because of cascade error on condition '%s' (Details: %s)" % (cascade_c.name, cascade_error)

					infos = [ reason, ]
					self.update_condition_on_dependson(check_time, pcondition, infos)

					self.exec_context = exec_context #DA CHIEDERE

			self.push_updates(check_time)

 		else:
			self.log.debug("No conditions to check.")

		ts_dg_check_end = time.time()

		#
		# Update datagroup
		#

		# FIX: Potrebbe essere passato parecchio tempo 
		#      tra l'inizio del check e questo punto.
		#      Forse sarebbe meglio mettere now() ma rischiamo che un datagroup
		#      molto lento venga schedulato con troppi ritardi?
		self.pdatagroup.lastdone = get_time()

		update_fields = {
			'lastdone'          : self.pdatagroup.lastdone,
			'lastchecktotaltime': ts_dg_check_end - ts_dg_check_start,
		}
		update_data = DataGroupUpdate(self.pdatagroup.lastdone, self.pdatagroup.uuid, update_fields, mtime_sentinel=self.pdatagroup.mtime)
		self.save_datagroup_update(update_data)

		#
		# Saving execution state
		#
		self.log.debug("----------------------- Saving execution state ----------------------")

		current_values = exec_context.getCurrentValues()

		if self.log.isEnabledFor(logging.DEBUG):
			self.log.debug("Last check state")
			for k,v in sorted(previous_check_values.items()):
				self.log.debug("\t%s = %r", k, v)

			self.log.debug("New state")
			for k,v in sorted(current_values.items()):
				self.log.debug("\t%s = %r", k, v)

		#
		# Remember calculated/fetched values for the next check:
		#   
		#   - save SNMP values 
		#   - update Datasource values if not uncheckable
		#   - update lastgoodcheck for Datasources and Conditions
		#
		previous_check_values.update( current_values )
		current_values = previous_check_values

		try:
			self.pdatagroup.save_check_state(get_time(), current_values, self.cache_manager  )
		except Exception, e:
			self.log.critical("Unable to save check state in cache: %s: %s", type(e), str(e))

		self.log.debug("----------------------- Element fields to update ----------------------")

		#
		# Campi modificati durante l'esecuzione delle expr di datasource e condition
		#
		fields_by_element = exec_context.get_updated_fields_by_element_id()

		#
		# Campi calcolati/aggiornati, ma non nelle expr.
		#

		# L'ip effettivo del nodo siamo riuscti a calcolarlo?
		if self.node.current_effective_ip is not None:

			# E' cambiato e dobbiamo comunicarlo al server?
			if self.node.current_effective_ip != self.node.effective_ip:

				self.log.debug("Effective IP: %s (was %s)", self.node.current_effective_ip, self.node.effective_ip)

				node_updated_fields = fields_by_element.get(self.node.uuid, fields_by_element.setdefault(self.node.uuid, {}))

				node_updated_fields['kind']         = 1
				node_updated_fields['effective_ip'] = self.node.current_effective_ip

		if self.log.isEnabledFor(logging.DEBUG):
			for e_id, fields in fields_by_element.items(): 
				self.log.debug("Element: %s", e_id)
				for f_name, v in sorted(fields.items()):
					self.log.debug("\t%s=%r", f_name, v)

		self.update_fields_by_element_id = fields_by_element

		self.log.debug("----------------------- Additional update ----------------------")

		self.additional_updates = fields_by_element = exec_context.get_all_additional_update_data()

		self.log.debug("----------------------- Saved OIDs and Values ----------------------")

		self.saved_variables = self.extract_retreived_snmp_oids_and_saved_values(exec_context)

		# Remember resume updates
		self.log.debug("----------------------- Saving all updates ----------------------")

		self.exec_context = exec_context #DA CHIEDERE

		self.push_updates(check_time)

		return


	def extract_retreived_snmp_oids_and_saved_values(self, exec_context):

		values = exec_context.getCurrentValues()

		#
		# SNMP variables 
		# ---------------
		# 
		#     Keys returned by getCurrentValues() are:
		#     
		#         <oid>#<node>:<port>
		#     
		#     Estract only:
		#     
		#         <oid> = <value>
		#
		#
		# Saved values (by saveval() function)
		# -------------------------------------
		#
		#     Items returned by getCurrentValues() contains
		#     
		#        <saved value>                                = <value> 
		#        <datasource or condition name>               = <value>
		#        <datasource or condition name>.totalch               = <value>
		#        <datasource or condition name>/lastgoodcheck = <timestamp>
		#     
		#     Estract only:
		#     
		#        <saved value> = <value> 
		#

		datasource_condition_names = [ ds.name for ds in self.get_related_datasources() ] + [ c.name for c in self.get_related_conditions() ]

		snmp_variables = {}
		free_variables = {}

		for key, key_value in values.items():

			# SNMP?
			if key.find('#') > -1:
				oid = str(key.split('#')[0])
				snmp_variables[oid]  = key_value
				continue
			else: 
				name = str(key)

				# Key related to a datasource or condition ?
				if name.find('/') > -1:
					name = name.split('/')[0]
				elif name.find('.') > -1:
					name = name.split('.')[0]

				if name in datasource_condition_names:
					continue

				free_variables[name]  = key_value

		return {
			'snmp_variables': snmp_variables,
			'free_variables': free_variables
		}


	def check_datasource(self, pdatasource, exec_context, previous_check_values):

		ts_check_start = time.time() # MUST BE FLOAT! DO NOT USE get_time()

		ctime  = get_time()

		self.log.debug("Evaluating datasource %s", pdatasource)

		# I datasource di tipo COUNTER sono particolari perche' il valore letto dalla rete
		# viene "storicizzato" come rate al secondo (differenza tra il valore attuale e 
		# quello letto in precedenza).
		# Questa operazione viene fatta dal server in fase di salvataggio dati sul db (RRD).
		#
		# A noi pero' serve conoscere valore subito per sapere se un contatore
		# ha superato una soglia. Quindi cerchiamo di calcolare questo rate usando
		# il valore precedente del datasource.
		#
		value        = None  # Value computed for this datasource
		rate         = None  # Counter rate (diffence between current value and value from the last check (if not uncheckable))

		old_status = pdatasource.status
		infos  = []

		exec_context['__datasource__'] = pdatasource

		#
		# Update expr symbols
		#
		ds_symbols = pdatasource.get_expr_symbols()
		exec_context.update( ds_symbols )


		# Old value {<name>.checktotaltime#}
		#checktotaltime = previous_check_values.get(pdatasource.name + "/checktotaltime", None)
		#if checktotaltime is not None:
		#	symbol_name = "%s.checktotaltime" % (pdatasource.name)
		#	exec_context.setOldValue(symbol_name, checktotaltime)
		#	self.log.debug("Setting {%s#} = %s", symbol_name, checktotaltime)


		# Calculate time $delta from last check
		lastgoodcheck = previous_check_values.get(pdatasource.name + "/lastgoodcheck", None)
		if lastgoodcheck != None:
			delta = get_time() - lastgoodcheck
		else:
			delta = None

		exec_context['$delta'] = delta
		self.log.debug("$delta = %s", delta)

		#
		# Perform
		#
		status = DATASOURCE_STATUS_UNCHECKABLE
		try:
			#
			# Compiliamo la expr
			#
			expr = pdatasource.expr

			expr_tree = None

			self.log.log(9,"Loading expr from cache...")
			expr_id   = '%s.expr' % (pdatasource.uuid)
			expr_tag  = str(crc32(expr))

			expr_tree = get_cached_expr_tree(self.log, self.cache_manager, expr, expr_id, expr_tag)

			#self.log.log(9,"str  expression: %s", expr)
			#self.log.log(9,"tree expression: %s", str(expr_tree))

			#
			# EVALUATION
			#
			self.log.log(9,"execeval()")
			value = expreval(self.log, exec_context, expr_tree)
			if value == None:
				infos.append( "Some past variables can't be read or are not present." )

			self.log.log(9,"evaluated()")
			expanded_expr = expr_tree.evaluated() # FIX: IMPORTANTE NON TOGLIERE QUESTA RIGA O NON FUNZIONA IL GIOCO CON LE VARIABILI OLD

			#self.log.log(9,"Calculating infos")
			infos += exec_context.getExecutionInfos()

			#self.log.log(9,"Saving infos")

			eval_descr = "expression %s expanded to %s" % (expr, expanded_expr)
			infos.append( eval_descr)
			self.log.log(9,"%s", eval_descr)

			#self.log.log(9,"Calculating status")

			#
			# Datasource STATUS
			#
			if value == None:
				# Some past variables can't be read or are not present.
				status = DATASOURCE_STATUS_UNCHECKABLE
			else:
				# The value must be numeric
				if pdatasource.is_valid_value(value):

					if pdatasource.gauge == True:
						#
						# Lo stato di un datasource GAURGE si basa sul valore puntuale calcolato.
						#
						if pdatasource.is_value_in_range(value):
							status = DATASOURCE_STATUS_OK
						else:
							status = DATASOURCE_STATUS_STRANGE_VAL
					else:
						# IMPORTANTE: RRD supporta SOLO i valori interi per i COUNTER, quindi
						#             quando mandiamo dati al server cerchiamo di convertire il valore in un INTERO.
						#             Questa operazione la facciamo qui per evitare che il server debba caricare
						#             dal DB informazioni sul datasource (se e' gauge o meno) ogni volta che riceve
						#             dati da salvere.
						value = int(value) # NOTA: qui siamo tranquilli che questo valore sia sempre numerico

						#
						# Nel database un COUNTER viene storicizzato come "rate al secondo" quindi per
						# calcolare lo stato dobbiamo calcolare la differenza tra il valore
						# corrente e il valore precedente e calcolare la derivata
						#
						old_value = previous_check_values.get(pdatasource.name, None)
						if old_value is not None:
							#
							# Se abbiamo il valore precedente allora l'ultimo check e' riuscito a calcolare un valore
							# (no errori e  non mancavano variabili "OLD").
							#
							# IMPORTANTE: Non e' detto che "old_value" sia stato calcolato proprio al precedente CHECK coerente con $delta.
							#             Puo' succedere quando, dopo una serie di check buoni, improvvisamente all'ultimo check manca
							#             un valore "old". $delta sara' riferita al giro subito prima, mentre il valore a due (o N) check prima.
							#             
							#                  TEMPO     t1      t2      t3      t4       t5       t6
							#                  NEW       1       2       3       -        -        5
							#                  OLD       *       1       2       3        3        3       <---- t3
							#                  $delta    t1-*    t2-t1   t3-t2   t4-t3    t5-t4    t6-t5   <-----t5
							#                                                  
							#             E' un caso molto remoto e si verifica solo quando qualcuno cambia una expr a controllo gia'
							#             in esecuzione da un po'.
							#             
							#             Per ora ignoriamo. Vedremo se gestirlo in futuro.

							# calcoliamo il "rate"
							if delta:
								rate = float(value - old_value) / float(delta)

								if pdatasource.is_value_in_range(rate):
									status = DATASOURCE_STATUS_OK
								else:
									status = DATASOURCE_STATUS_STRANGE_VAL
						else:
							# FIX: Forse dovremmo usare DATASOURCE_STATUS_UNKNOWN che non e' stato definito
							status = DATASOURCE_STATUS_STRANGE_VAL
				else: 
					status = DATASOURCE_STATUS_STRANGE_VAL

			self.log.debug("Calculated status = %s", DSSTATE_TO_CODENAME(status))

			# IMPORTANTE: Quando il valore e' NONE l'espressione non e' stata valutata perche'
			#             alcuni valori "old" non sono stati calcolati.
			#             Lo stato del datasource viene considerato UNCHECKABLE, ma
			#             la valutarazione della expr e' considerata "buona" e vale
			#             per il calcolo del $delta a giro succesivo.

			# Remember the last time this datasource produces a valid value (for $delta calculation)
			exec_context.saveCurrentValue(pdatasource.name + "/lastgoodcheck", get_time())

		except TimeoutException, e:
			infos  = [ str(e), ]
			status = DATASOURCE_STATUS_UNCHECKABLE 

		except PollException, e:
			infos  = [ str(e), ]
			status = DATASOURCE_STATUS_UNCHECKABLE 

		except Exception, e:
			infos  = [ str(e), ]
			status = DATASOURCE_STATUS_ERROR

		exec_context.appendExecutionInfo(",".join(infos))

		exec_context.saveExecInfoByName(pdatasource.name)

		self.log.debug("Check result %s, %s, %r, %s", ctime, DSSTATE_TO_CODENAME(status), value, infos)

		#
		# Calculate total execution time
		#
		ts_check_end = time.time()
		
		lastchecktotaltime = ts_check_end - ts_check_start
		
		self.log.debug("Elapsed: %s (%s %s)", lastchecktotaltime,ts_check_end,ts_check_start)

		# Remember the last execution time for the next check
		#exec_context.saveCurrentValue("%s/checktotaltime" % (pdatasource.name) , lastchecktotaltime)

		#
		# Update execution context
		#

		# Update {<name>@} if evaluation was successful
		if status in [ DATASOURCE_STATUS_OK, DATASOURCE_STATUS_STRANGE_VAL ]:

			# Update "old" value ( {<name>#} )for the next check  
			exec_context.saveCurrentValue(pdatasource.name, value)

			# Add datasource value to execution context for other datasource/condition ( {<name>@} )
			exec_context[pdatasource.name] = value


		# Update {<name>.checktotaltime@}
		symbol_name = "%s.checktotaltime" % (pdatasource.name)
		exec_context.saveCurrentValue(symbol_name, lastchecktotaltime)  # {<name>.checktotaltime#} 
		exec_context[symbol_name] = lastchecktotaltime                  # Current value


		#
		# Update datasource fields
		#
		pdatasource.status             = status
		pdatasource.lastchecktotaltime = lastchecktotaltime

		#
		# Prepare data to send to the server
		# 
		fields = {}
		fields['lastdone'          ] = ctime
		fields['lastvalue'         ] = value
		fields['lastchecktotaltime'] = lastchecktotaltime

		if pdatasource.gauge is False:
			# Se il datasource e' un counter, per completezza mandiamo sia
			# il valore calcolato (value) che il rate calcolato (valuec)
			fields['lastrate'] = rate

		if old_status != status:
			fields['status'] = status

		#
		# Prepariamo il blocco dati da mandare al server
		#
		update_data = DataSourceUpdate(ctime, pdatasource.uuid, pdatasource.name, fields, infos, mtime_sentinel=pdatasource.mtime, nosave=not(pdatasource.save_value))

		self.save_datasource_update(pdatasource.uuid, update_data)

		return status


	def check_condition(self, pcondition, exec_context, previous_check_values):
		
		ts_check_start = time.time() # MUST BE FLOAT. DO NOT USE get_time()

		ctime  = get_time()

		self.log.debug("Evaluating condition %s", pcondition.name)

		result = None
		infos  = []

		#
		# Remember fields' current value
		#
		old_field_values = {}
		for fname in DataGroupChecker.CONDITION_FIELDS_TO_CHECK_FOR_CHANGES:
			v = getattr(pcondition, fname)
			self.log.log(1,"Saving fields %s = %s", fname, v)
			old_field_values[ fname ] = v

		#
		# Update expr symbols
		#

		exec_context['__condition__'] = pcondition

		exec_context.update( pcondition.get_expr_symbols() )

		#exec_context['$path']    = pcondition.path
		#exec_context['$_cname_'] = pcondition.name


		# Old value {<name>.checktotaltime#}
		#checktotaltime = previous_check_values.get(pcondition.name + "/checktotaltime", None)
		#if checktotaltime is not None:
		#	symbol_name = "%s.checktotaltime" % (pcondition.name)
		#	exec_context.setOldValue(symbol_name, checktotaltime)
		#	self.log.debug("Setting {%s#} = %s", symbol_name, checktotaltime)


		# Calculate time $delta from last check
		delta = None
		lastgoodcheck = previous_check_values.get(pcondition.name + "/lastgoodcheck", None)
		if lastgoodcheck != None:
			delta = get_time() - lastgoodcheck
		exec_context['$delta'] = delta
		self.log.debug("$delta = %s", delta)


		#
		# Perform
		#
		try:
			#
			# LOAD EXPR
			expr = pcondition.expr

			self.log.log(9, "Loading expr from cache...")
			expr_id   = '%s.expr' % (pcondition.uuid)
			expr_tag  = str(crc32(expr)) 
			expr_tree = get_cached_expr_tree(self.log, self.cache_manager, expr, expr_id, expr_tag)

			#self.log.log(9, "str  expression: %s", expr)
			#self.log.log(9, "tree expression: %s", str(expr_tree))

			#
			# EVAL
			#
			self.log.log(9, "execeval()")

			result = expreval(self.log, exec_context, expr_tree)
			if result == None:
				infos.append( "Some past variables can't be read or are not present." )

			#
			# RESULT
			#
			self.log.log(9, "Check result: %s", result)
			infos += exec_context.getExecutionInfos()

			self.log.log(9, "evaluated()")
			expanded_expr = expr_tree.evaluated()   # FIX: IMPORTANTE NON TOGLIERE QUESTA RIGA O NON FUNZIONA IL GIOCO CON LE VARIABILI OLD

			self.log.log(9, "Expanded to: %s" % (expanded_expr,))

			eval_descr = " %s expanded to %s" % (expr, expanded_expr)
			infos.append(eval_descr)

			exec_context.saveCurrentValue(pcondition.name + "/lastgoodcheck", get_time())

		except ResourceNotFoundException, ex:
			# Tried to access to a bould resource (iface or node) which is not
			# present in the main cache. This check is probably old and
			# was deleted from DB with the bound resource.
			raise ex

		except PollException, e:
			ctime   = get_time()
			result  = None
			infos  += [ str(e) ]

		exec_context.saveExecInfoByName(pcondition.name)

		self.log.debug("Check result %s, result=%s (%s), infos=%s", ctime, result, type(result), infos,)

		#
		# Calculate total execution time
		#
		ts_check_end = time.time()
		
		lastchecktotaltime = ts_check_end - ts_check_start
		
		self.log.debug("Elapsed: %s (%s %s)", lastchecktotaltime,ts_check_end,ts_check_start)

		#
		# Update execution context
		#

		# Update {<name>@}
		if result is not None:
			exec_context.saveCurrentValue(pcondition.name, result)  # {<name>#} at the next check
			exec_context[pcondition.name] = result                  # {<name>@}

		# Update {<name>.checktotaltime@}
		symbol_name = "%s.checktotaltime" % (pcondition.name)
		exec_context.saveCurrentValue(symbol_name, lastchecktotaltime)  # {<name>.checktotaltime#} at the next check
		exec_context[symbol_name] = lastchecktotaltime                  # {<name>@}


		#
		# Update condition fiels
		#
		pcondition.lastdone           = get_time()
		pcondition.lastchecktotaltime = lastchecktotaltime

		new_state = self.calculate_new_state_by_result(pcondition, result, infos)

		self.log.debug("Old state  : %s" % (CSTATE_TO_CODENAME(pcondition.status)))
		self.log.debug("New state  : %s" % (CSTATE_TO_CODENAME(new_state)))

		#
		# Update state if necessary
		#
		if new_state != None:
			self.set_condition_state(pcondition, ctime, new_state, infos)

		#
		# Find updated fields
		#
		self.log.debug("Checking all changed fields...")

		updated_fields = {
			#'mtime' : pcondition.mtime,
			'lastchecktotaltime': lastchecktotaltime,
		}

		for fname in DataGroupChecker.CONDITION_FIELDS_TO_CHECK_FOR_CHANGES:
			new_val = getattr(pcondition, fname)
			old_val = old_field_values[ fname ]
			if new_val != old_val:
				self.log.log(9, "\tfield %s changed %s %s", fname, old_val, new_val)
				updated_fields[ fname ] = new_val

		#if new_state != None:
		if updated_fields:

			update_data = ConditionUpdate(
				ctime, 
				pcondition.uuid, 
				updated_fields,
				infos,
				mtime_sentinel=pcondition.mtime
			)
			self.save_condition_update(pcondition.uuid, update_data)

		return result

	def calculate_new_state_by_result(self, pcondition, result, infos):

		new_state = None

		# Se il risultato della valutazione della condition non e' nullo (uncheckable)
		# cerchiamo di capire se e' riconducibilea ad un True o un False.
		# Se non e' convertibile in un booleano, e' uncheckable
		try:
			if result is not None:
				result = bool(result)
		except Exception, ex:
			result = None

		#
		# UPDATE STATUS
		#

		if result == None:
			if pcondition.uncheckable_fallback == CSTATE_UP:
				self.log.debug("Uncheckable %s %s, forcing verified (up)", pcondition.uuid, pcondition.path)
				result = True

			elif pcondition.uncheckable_fallback == CSTATE_DN:
				self.log.debug("Uncheckable %s %s, forcing not verified (dn)", pcondition.uuid, pcondition.path)
				result = False

		# Verified
		if result == True:

			pcondition.tries = 0

			new_state = CSTATE_UP
			verb      = "UP: %s" % (infos)

		# Unverifield check
		elif result == False:

			oldtries  = pcondition.tries

			if pcondition.status in [CSTATE_UP, CSTATE_FA, CSTATE_TU, CSTATE_UU, CSTATE_DU]:

				pcondition.tries += 1

				if pcondition.tries < pcondition.max_tries:
					new_state = CSTATE_FA
				elif pcondition.tries == pcondition.max_tries:
					new_state = CSTATE_DN
				else:
					# Recovering from alien statusfile or changed maxtries in DB?
					pcondition.tries = pcondition.max_tries
					new_state = CSTATE_DN
			else: # DN, TD, UD
				pcondition.tries = pcondition.max_tries
				new_state = CSTATE_DN

		# Uncheckable
		elif result == None:

			self.log.info("Condition %s %s uncheckable", pcondition.uuid, pcondition.path)

			if pcondition.status in [CSTATE_UP, CSTATE_FA, CSTATE_TU, CSTATE_UU, CSTATE_DU]:
				new_state = CSTATE_UU
			elif pcondition.status in [CSTATE_DN, CSTATE_TD, CSTATE_UD, CSTATE_DD]:
				new_state = CSTATE_UD
			else:
				# state could not be determined
				new_state = CSTATE_UU

		return new_state

	def set_condition_state(self, pcondition, ctime, new_state, infos):

		self.log.debug("Updating condition state")
		self.log.debug("\t%s -> %s", CSTATE_TO_CODENAME(pcondition.status), CSTATE_TO_CODENAME(new_state) )

		if new_state == pcondition.status:
			self.log.debug("Updating penalty--->TODO")

			#verb += self.update_penalty(now)
			pass
		else:
			pcondition.statuslast       = pcondition.status
			pcondition.statuslastchange = ctime

			pcondition.status           = new_state

			if self.log.isEnabledFor(logging.DEBUG):
				self.log.debug("\tstatuslast       = %s", pcondition.statuslast)
				self.log.debug("\tstatuslastchange = %s", pcondition.statuslastchange)
				self.log.debug("\tstatus           = %s", pcondition.status)

			pcondition.verbstatus       = ','.join(infos)

		return

	#--------------------------------------------------------------------#
	# RESUME                                                             #
	#--------------------------------------------------------------------#

	def _resume_if_suspended(self, check_time, pdatagroup):
		"""
		Se il datagroup e' sospeso, questa funziona effettua le operazioni di resume.
		
		NOTA: la funzione restituisce 
				<esito>, <timestamp>
			
			<esito> true se e' stato fatto il resume
			<timestamp> e' l'orario che deve essere usato per effettuare altre modifiche al datagroup.
		"""

		if not pdatagroup.is_suspended():
			return True, check_time

		self.log.info('Datagroup %s suspended. Resume at %s. Now %s', pdatagroup.path, pdatagroup.suspend_resume_time, check_time)
		if pdatagroup.can_be_resumed(check_time):

			# Resume
			self._resume(check_time)

			# Remember resume updates
			self.push_updates(check_time)

			#
			# HACK: se il resume() + check() di un datagroup sono troppo veloci potrebbero risultare eseguite
			#       nello stesso secondo. Questo puo' essere un problema poiche' nei log sono storicizzati
			#       per timestamp e effettuando ordinamenti vari il resume() potrebbe risultare "dopo" il check().
			#       La soluzione piu' brutta, ma piu' "pratica" per evitare problemi e' introdurre una pausa di almeno 1 secondo
			#       tra il resume() e il check().
			#
			time.sleep(DEFAULT_DELAY_AFTER_RESUME)
			check_time = get_time()

			return True, check_time

		# not resumed
		return False, check_time

	def _resume(self, resume_time):

		self.log.info("\tResuming datagroup %s", self.pdatagroup)

		self.lastdone                       = resume_time

		self.pdatagroup.suspended           = False
		self.pdatagroup.suspend_time        = None
		self.pdatagroup.suspend_resume_time = None
		self.pdatagroup.suspend_reason      = ''

		fields = {
			#'mtime'              : self.pdatagroup.mtime,
			'lastdone'           : self.lastdone, 

			'suspended'          : self.pdatagroup.suspended,
			'suspend_time'       : self.pdatagroup.suspend_time,
			'suspend_reason'     : self.pdatagroup.suspend_reason,
			'suspend_resume_time': self.pdatagroup.suspend_resume_time,
		}

		update_data = DataGroupUpdate(resume_time, self.pdatagroup.uuid, fields, mtime_sentinel=self.pdatagroup.mtime)
		self.save_datagroup_update(update_data)

		reason = 'Resumed by poller (at %s)' % (resume_time)

		#
		# RESUME ALL DATASOURCES
		#
		pdatasources = self.get_related_datasources()
		for pdatasource in pdatasources:
			self._resume_datasource(resume_time, pdatasource, reason)

		#
		# RESUME ALL CONDITIONS
		#
		pconditions = self.get_related_conditions()
		for pcondition in pconditions:
			self._resume_condition(resume_time, pcondition, reason)

	def _resume_datasource(self, resume_time, pdatasource, reason):
		infos = [ reason ]

		self.log.debug("\tUpdating datasources %s", pdatasource.path)
		pdatasource.status = DATASOURCE_STATUS_OK
		pdatasource.lastdone = resume_time

		self.log.log(1, "\tCreating updte data for the server")
		# Prepare update data for the server
		fields = {
			'status'  : pdatasource.status,
			'lastdone': pdatasource.lastdone,
		}
		update_data = DataSourceUpdate(resume_time, pdatasource.uuid, pdatasource.name, fields, infos, mtime_sentinel=pdatasource.mtime)
		self.save_datasource_update(pdatasource.uuid, update_data)

	def _resume_condition(self, resume_time, pcondition, reason):

		infos = [ reason ]

		self.log.debug("\tUpdating condition %s", pcondition.path)

		pcondition.statuslastchange = resume_time
		pcondition.lastdone         = pcondition.statuslastchange

		new_status = pcondition.statuslast
		if new_status == CSTATE_IN: # 'IN':
			new_status = CSTATE_UP

		pcondition.statuslast = pcondition.status
		pcondition.status     = new_status   

		#pcondition.update_notable_state(resume_time)

		self.log.log(1, "\tCreating updte data for the server")
		#
		# Find updated fields
		#
		updated_fields = {
			'mtime'           : pcondition.mtime,
			'status'          : pcondition.status,
			'statuslast'      : pcondition.statuslast,
			'statuslastchange': pcondition.statuslastchange,
			'lastdone'        : pcondition.lastdone,
		}
		update_data = ConditionUpdate(
			resume_time, 
			pcondition.uuid, 
			updated_fields,
			infos,
			mtime_sentinel=pcondition.mtime,
		)
		self.save_condition_update(pcondition.uuid, update_data)

	#--------------------------------------------------------------------#
	# DEPENDSON                                                          #
	#--------------------------------------------------------------------#

	def check_datagroup_dependson(self):
		# The depenson chain uses the ".dependson_id" field. Both datagroups and conditions
		# have this field so we can ignore the object type.
		return self.__check_depending_chain(self.pdatagroup)

	def check_condition_dependson(self, pcondition):
		# The depenson chain uses the ".dependson_id" field. Both datagroups and conditions
		# have this field so we can ignore the object type.
		return self.__check_depending_chain(pcondition)

	def __check_depending_chain(self, first_element):

		condition = self.__get_depending_condition(first_element)

		while condition != None:

			can_check, reason = self.__check_condition_for_dependson(condition)
			if not can_check:
				break

			self.log.debug("Depends on %s which has state %s %s. Go on.", condition.path, CSTATE_TO_CODENAME(condition.status), condition.status )

			condition = self.__get_depending_condition(condition)

		# Not verified
		if condition != None:
			return False, reason

		return True, ''

	def __check_condition_for_dependson(self, condition):

			if not self.valid_for_check(condition):
				reason = "Depends on %s which has state %s and is not valid for check" % ( condition.path, condition.status )
				self.log.debug(reason)
				return False, reason

			if condition.status in DEPENDSON_STOP_STATES:
				reason = "Depends on %s which has state %s %s. Stop!" % ( 
					condition.path, 
					CSTATE_TO_CODENAME(condition.status),
					condition.status,
				)
				self.log.debug(reason)
				return False, reason

			return True, ''

	def __get_depending_condition(self, element):

		if element.dependson_id == None:
			return None

		return self.configuration.get_condition(element.dependson_id)

	def update_datagroup_on_dependson(self, pdatagroup, ctime, infos):

		ts_dg_check_start = get_time()

		self.log.debug("Updating datasources state...")

		pdatasources = self.get_related_datasources()
		for pdatasource in pdatasources:
			self.update_datasource_on_dependson(ctime, pdatasource, infos)

		self.log.debug("Updating conditions state...")

		pconditions = self.get_related_conditions()
		for pcondition in pconditions:
			self.update_condition_on_dependson(ctime, pcondition, infos)

		ts_dg_check_end = get_time()

		lastchecktotaltime = ts_dg_check_end - ts_dg_check_start

		self.log.debug("Updating datagroup fields...")

		update_fields = {}

		if lastchecktotaltime != self.pdatagroup.lastchecktotaltime:
			self.pdatagroup.lastchecktotaltime = lastchecktotaltime
			
			update_fields['lastchecktotaltime'] = self.pdatagroup.lastchecktotaltime

		#
		if update_fields:
			update_data = DataGroupUpdate(self.pdatagroup.lastdone, self.pdatagroup.uuid, update_fields, mtime_sentinel=self.pdatagroup.mtime)
			self.save_datagroup_update(update_data)

		return

	def update_datasource_on_dependson(self, ctime, pdatasource, infos):

		if pdatasource.status == DATASOURCE_STATUS_DEPENDSON:
			return

		self.log.debug("\tUpdating datasources %s", pdatasource.path)

		pdatasource.status = DATASOURCE_STATUS_DEPENDSON

		pdatasource.lastchecktotaltime = 0 # not executed

		self.log.debug("\tCreating updte data for the server")

		# Prepare update data for the server
		fields = {
			#'mtime' : pdatasource.mtime,
			'status'            : pdatasource.status,

			'lastchecktotaltime': pdatasource.lastchecktotaltime,
		}
		update_data = DataSourceUpdate(ctime, pdatasource.uuid, pdatasource.name, fields, infos, mtime_sentinel=pdatasource.mtime)
		self.save_datasource_update(pdatasource.uuid, update_data)

	def update_condition_on_dependson(self, ctime, pcondition, infos):

		if   pcondition.status in [CSTATE_UP, CSTATE_FA, CSTATE_TU, CSTATE_UU]: 
			new_status = CSTATE_DU 
		elif pcondition.status in [CSTATE_DN, CSTATE_TD, CSTATE_UD]:            
			new_status = CSTATE_DD 
		else:
			new_status = None

		if not new_status:
			return

		self.log.debug("\tUpdating condition %s", pcondition.path)

		self.set_condition_state(pcondition, ctime, new_status, infos)

		pcondition.lastchecktotaltime = 0 # not executed

		self.log.debug("\tCreating updte data for the server")
		#
		# Find updated fields
		#
		updated_fields = {
			'statuslast'        : pcondition.statuslast,
			'statuslastchange'  : pcondition.statuslastchange,
			'status'            : pcondition.status,
			#'statuslastnotable'      : pcondition.statuslastnotable,
			#'statuslastnotablechange': pcondition.statuslastnotablechange,

			'lastchecktotaltime': pcondition.lastchecktotaltime,
		}

		#if new_state != None:
		if updated_fields:
			update_data = ConditionUpdate(
				ctime, 
				pcondition.uuid, 
				updated_fields,
				infos,
				mtime_sentinel=pcondition.mtime,
			)
			self.save_condition_update(pcondition.uuid, update_data)

	#--------------------------------------------------------------------#

	def valid_for_check(self, condition):

		# FIX: check if datagroup is suspended

		return True




class FakeDataGroupChecker(DataGroupChecker):

	def check(self, check_time, scheduled_time=None):
		"""
		Fake check.
		Generates fake checks with random sleep in time
		"""

		try:
			sleepTime = int(os.environ['SLEEP_TIME'])
		except Exception, e:
			sleepTime = random.randrange(10)

		sghed_group = self.pdatagroup.scheduling_group

		if sghed_group is not None:
			sched_g_name = sghed_group.name
		else:
			sched_g_name = "__NO_SCHED__"

		SANET_HOME = os.environ['_SANET_HOME']
		path = os.path.join('/tmp', 'scheduling_stats-'+sched_g_name+'.log')

		t1 = time.time()

		with open(path, 'a+') as fd:

			data = {
				"type"              : None,
				"group"             : sghed_group.uuid if sghed_group is not None else None,
				"datagroup"         : self.pdatagroup.path,
				"minperiod"         : self.pdatagroup.minperiod,
				"lastdone"          : self.pdatagroup.lastdone,
				"scheduled ts"      : scheduled_time,
				"ts start"          : t1,
				"waiting time"      : sleepTime,
				"ts end"            : None,
				"time elapsed"      : None,
			}

			#
			# LOG START
			#
			data['type']='start'
			fd.write(json.dumps(data) + "\n")
			fd.flush()

			if 1:
				# Fake execution
				self.log.info("*********** START   exection of %s will take %s seconds", self.pdatagroup.path, sleepTime)
				time.sleep(sleepTime)
				self.log.info("***********   END   exection of %s"                     , self.pdatagroup.path)

				t2 = time.time()
				dt = datetime.datetime.fromtimestamp(t2)
				end_str = '%s:%s:%s' % (dt.hour, dt.minute, dt.second)

				# update only lastdone
				self.pdatagroup.lastdone = t2
				update_fields = {
					'lastdone': self.pdatagroup.lastdone
				}
				update_data = DataGroupUpdate(self.pdatagroup.lastdone, self.pdatagroup.uuid, update_fields,
				                              mtime_sentinel=self.pdatagroup.mtime)
				self.save_datagroup_update(update_data)

				self.push_updates(check_time)

			else:
				# Real esecution
				super(FakeDataGroupChecker, self).check(check_time, scheduled_time)
				t2 = time.time()

			#
			# LOG END
			#

			data['type']='end'
			data.update({
				"ts end"            : t2,
				"time elapsed"      : t2 - t1
			})

			fd.write(json.dumps(data) + "\n")
			fd.flush()

