# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>. 

"""

###########################################################################
Expression Evaluation APIs
###########################################################################

.. autoclass:: ExecutionContext

"""

import os
import traceback
import copy
import types

from sanet_common.exprparser.ExprFactory import ParserException, NotEvaluatedError, ExecutionError
from sanet_common.exprparser.ExprFactory import BaseExprFactory, SnmpValueExprTree, DataIdentifierExprTree
from sanet_common.exprparser.Types import SnmpValueType
#from sanet_common.exprparser.utils import parseStringWithFactory
from sanet_common.exprparser.utils import parseString

from sanet_common.sanetdefs.constants import CLUSTER_ROLE_MASTER

from sanet_libsnmp import SnmpWrapperBUGException

from sanet_poller.lib.exceptions import *

#from sanet_poller.lib.snmp.base import snmp_create_sess
from sanet_poller.lib.snmp.base import getvar, gettable, SNMP_DEFAULT_VERSION, SNMP_DEFAULT_COMMUNITY, SNMP_DEFAULT_PORT
from sanet_poller.lib.snmp.base import init_snmp_session, getvar_s, gettable_s

from sanet_poller.lib.utils import crc32



_DONT_USE_SNMP_SESSION = os.environ.get('DONT_USE_SNMP_SESSION', None)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

IFACE_SPECIAL_PARAMS = [
	'$ifindex',
	'$linked_ifindex',
]

STORAGE_SPECIAL_PARAMS = [
	'$stindex'
]

SERVICE_SPECIAL_PARAMS = [
	'$swrunindex'
]

DEVICE_SPECIAL_PARAMS = [
	'$devindex'
]

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

CLUSTER_SPECIAL_PARAMS = [ 
	'$cluster_node', 
	'$cluster_instance', 
	'$cluster_community',
	'$cluster_snmpversion',
	'$cluster_snmpport',
	'$cluster_snmpv3authdata',
]

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class ExecutionContext(dict):
	"""
	This class rappresent the set of symbols (functions and variabiles) used when an expression is evaluated. 
	
	This class is also used by functions for:
	
		* access to the :class:`Logger` object used to log.
		* access to the :class:`Basecheck` object involved in the evaluation.
		* access to the update :class:`PollerConfiguration` object used by the Agent.
		* retrieve the values of symbols used inside and the expression (es: $node).
		* cache all the SNMP values retrieved from the network.
		* save informations about the expression evaluations (errors, ecc...)
	"""
	def __init__(self):

		super(ExecutionContext, self).__init__()

		self.__log           = None
		self.__configuration = None
		self.__cache         = None
		self.__ipc_factory   = None

		self.__snmp_sessions = {}

		self.__old_values     = {}
		self.__current_values = {}

		# Cache for new snmp values and tables.
		self.__snmp_value_cache = {}
		self.__snmp_table_cache = {}

		self.__simple_value_cache = {}

		self.exec_infos = []

		self.exec_infos_by_name = {}

		# Remember 
		self.update_elements_fields_by_id = {}

		# Additional data to send to the server in order to update SanetDB 
		self.additional_update_data = []

	def dispose(self):
		del self.__log
		del self.__configuration
		del self.__cache
		del self.__ipc_factory

		del self.__snmp_sessions

		del self.__old_values
		del self.__current_values
		# Cache for new snmp values and tables.
		del self.__snmp_value_cache
		del self.__snmp_table_cache
		del self.__simple_value_cache

		del self.exec_infos
		del self.exec_infos_by_name

		# Remember 
		del self.update_elements_fields_by_id
		del self.additional_update_data

	#--------------------------------------------------------------#

	def get_or_create_snmp_session(self, host, community_or_authdata, tries, port, version, timeout):
		"""
		Returns new SnmpSession objects used to retrieve SNMP values from network nodes.
		"""

		session_id = "%s:%s:%s" % (host, port, version)

		sess = self.__snmp_sessions.get(session_id, None)

		#print __file__, "get_or_create_snmp_session 1", sess

		if sess is None:
			sess = init_snmp_session(
				host,
				community_or_authdata, 
				tries, 
				port   =port, 
				version=version, 
				timeout=timeout
			)

			#print __file__, "get_or_create_snmp_session CREATED", sess

			self.__snmp_sessions[session_id] = sess

		return sess

	#--------------------------------------------------------------#

	def createNewContext(self):
		"""
		Create a new ExecutionContext object of the same type/class.
		"""
		return self.__class__()

	#--------------------------------------------------------------#

	def setLog(self, log):
		self.__log = log

	def getLog(self):
		return self.__log

	#--------------------------------------------------------------#

	def setConfiguration(self, c):
		self.__configuration = c

	def getConfiguration(self):
		return self.__configuration

	#--------------------------------------------------------------#

	def setCache(self, cache):
		self.__cache = cache

	def getCache(self):
		return self.__cache

	#--------------------------------------------------------------#

	def setIPCFactory(self, ipc_factory):
		self.__ipc_factory = ipc_factory

	def getIPCFactory(self):
		return self.__ipc_factory

	#--------------------------------------------------------------#

	def remember_updated_field(self, element_id, field_name, value):

		e = self.update_elements_fields_by_id.get(element_id, None)
		if e == None:
			e = {}
			self.update_elements_fields_by_id[element_id] = e

		e[field_name] = value

	def get_updated_fields_by_element_id(self):
		return self.update_elements_fields_by_id


	def remember_additional_update_data(self, dtype, data):
		self.additional_update_data.append( (dtype, data) )

	def get_all_additional_update_data(self):
		return self.additional_update_data

	#--------------------------------------------------------------#

	def __contains__(self, key):  # dict.has_key(),  in
		#
		# We are forcing True here in order to force
		# computation when somebody ask for this variables.
		#
		
		# If $ifindex can't be calculated it's not a problem, 
		# an exception will be raised later.
		if key   in IFACE_SPECIAL_PARAMS: 
			return True
		elif key in STORAGE_SPECIAL_PARAMS:
			return True
		elif key in SERVICE_SPECIAL_PARAMS:
			return True
		elif key in DEVICE_SPECIAL_PARAMS:
			return True
		# Same problems about cluster variables
		elif key in CLUSTER_SPECIAL_PARAMS:
			return True
			
		rv = super(ExecutionContext, self).__contains__(key)
		return rv

	def get(self, *params):  # dict.get(key, alt_val)

		key = params[0]

		self.__calculate_special_symbols(key)

		rv = super(ExecutionContext, self).get(*params)

		if isinstance(rv, Exception):
			raise rv

		return rv

	def __getitem__(self, key):

		if key == '$delta':
			# The '$delta' parameter is special. It behaves like
			# and OLD SNMP value. If somebody is asking for it
			# and it's not defined (because we don't have the
			# last 'privatestate' of a target) we must
			# invalidate the expression computation.
			rv = super(ExecutionContext, self).__getitem__(key)
			if rv == None:
				msg = "parameter $delta not defined. No previous state."
				self.__log.debug(msg)
				raise NotEvaluatedError(msg)
		else:
			self.__calculate_special_symbols(key)

			rv = super(ExecutionContext, self).__getitem__(key)

			# If we have got and exception for this key, raise it again
			if isinstance(rv, Exception):
				raise rv
		return rv

	def __calculate_special_symbols(self, key):

		if super(ExecutionContext, self).__contains__(key): 
			return

		# If an exception is raised when calculating these symbols,
		# raise the same exception every time.

		try:
		
			if key in IFACE_SPECIAL_PARAMS:
				self.__calculate_and_save_ifindexes(key)
			elif key in STORAGE_SPECIAL_PARAMS: 
				self.__calculate_and_save_element_index(key, '__storage__', 'stindex')
			elif key in SERVICE_SPECIAL_PARAMS: 
				self.__calculate_and_save_element_index(key, '__service__', 'swrunindex')
			elif key in DEVICE_SPECIAL_PARAMS : 
				self.__calculate_and_save_element_index(key, '__device__' , 'devindex')
			elif key in CLUSTER_SPECIAL_PARAMS:
				# These parameters are valid only when there node & basecheck
				# are configured properly.
				self.__calculate_cluster_parameters(key)

		except ClusterException, e:
			self[key] = e
			# Cluster symbols are alwas calculated together
			self['$cluster_instance'      ] = e
			self['$cluster_node'          ] = e
			self['$cluster_community'     ] = e
			self['$cluster_snmpversion'   ] = e
			self['$cluster_snmpport'      ] = e
			self['$cluster_snmpv3authdata'] = e
			raise e

		except Exception, ex:
			
			# Remember the exception as value (and re-raise it eventually)
			self[key] = ex
			raise ex

	def __calculate_cluster_parameters(self, key):

		if super(ExecutionContext, self).__contains__(key): 
			return

		log           = self.getLog()
		configuration = self.getConfiguration()
		cache         = self.getCache()

		node  = super(ExecutionContext, self).get('__node__'        , None)
		if node == None: raise PollException("%s can not be calculated without node data." % key)

		execcontext = self.copy()

		cluster_node, cluster_instance, changed, infos = calculate_cluster_parameters(log, execcontext, configuration, cache, node)

		for i in infos:
			self.appendReason(i)

		if cluster_node != None:

			if changed:
				t = super(ExecutionContext, self).__getitem__('__check_time__')

				log.debug("Remember cluster changes for %s", node)
				log.debug("\t kind                 : %s", cluster_node.kind)
				log.debug("\t cluster_instance     : %s", cluster_instance)
				log.debug("\t cluster_instance_time: %s", t)
				log.debug("\t cluster_element_id   : %s", cluster_node.uuid)

				self.remember_updated_field(node.uuid, "kind"                 , cluster_node.kind)
				self.remember_updated_field(node.uuid, "cluster_instance"     , cluster_instance)
				self.remember_updated_field(node.uuid, "cluster_instance_time", t)
				self.remember_updated_field(node.uuid, "cluster_element_id"   , cluster_node.uuid)

			log.debug("Adding following symbols")
			log.debug("\t__cluster_node__        = %s", cluster_node)
			log.debug("\t$cluster_instance       = %s", cluster_instance)
			log.debug("\t$cluster_node           = %s", cluster_node.name)
			log.debug("\t$cluster_community      = %s", cluster_node.snmp_community)
			log.debug("\t$cluster_snmpversion    = %s", cluster_node.snmp_version)
			log.debug("\t$cluster_snmpport       = %s", cluster_node.snmp_port)
			log.debug("\t$cluster_snmpv3authdata = %s", cluster_node.snmp_v3_authdata)

			self['__cluster_node__'       ] = cluster_node
			self['$cluster_instance'      ] = cluster_instance
			self['$cluster_node'          ] = cluster_node.name
			self['$cluster_community'     ] = cluster_node.snmp_community
			self['$cluster_snmpversion'   ] = cluster_node.snmp_version
			self['$cluster_snmpport'      ] = cluster_node.snmp_port
			self['$cluster_snmpv3authdata'] = cluster_node.snmp_v3_authdata

	def __calculate_and_save_element_index(self, key, element_symbol, index_symbol):

		if super(ExecutionContext, self).__contains__(key): 
			return

		log           = self.getLog()
		configuration = self.getConfiguration()
		cache         = self.getCache()

		log.debug("Trying to calculate '%s'", key)

		node  = super(ExecutionContext, self).get('__node__'        , None)
		if node  == None: raise PollException("%s can not be calculated without node data." % (key))

		element  = super(ExecutionContext, self).get(element_symbol, None)
		if element  == None: raise PollException("%s can not be calculated without storage data." % (key))

		log.debug("Need to calculate %s for '%s' (node: %s)", key, element.name, node.name)

		local_context = self.copy()

		# <<<<<<<<<<<<<<<<< ifindex no more number <<<<<<<<<<<<<<<<<<<<<<
		#(updated, val) = update_element_index(log, configuration, cache, node, element, index_symbol, int, local_context)
		# >>>>>>>>>>>>>>>>> ifindex no more number >>>>>>>>>>>>>>>>>>>>>>
		(updated, val) = update_element_index(log, configuration, cache, node, element, index_symbol, str, local_context)
		self[key] = val

		if updated:
			log.debug("REMEMBER UPDATED %s.%s = %s", element, index_symbol, val)

			t = super(ExecutionContext, self).__getitem__('__check_time__')

			self.remember_updated_field(element.uuid, index_symbol, val)
			self.remember_updated_field(element.uuid, 'kind', element.kind)  # HACK: ricordiamoci qui il tipo dell'elemento cosi' il server non deve preoccuparsi di capire il tipo

			index_symbol_time = '%s_time' % (index_symbol)
			self.remember_updated_field(element.uuid, index_symbol_time, t)

		log.debug("Calculated %s for '%s' = %s", key, element.name, val)


	def __calculate_and_save_ifindexes(self, key):

		if super(ExecutionContext, self).__contains__(key): 
			return

		log           = self.getLog()
		configuration = self.getConfiguration()
		cache         = self.getCache()

		log.debug("Trying to calculate %s", key)

		if key == '$ifindex':

			node  = super(ExecutionContext, self).get('__node__'        , None)
			if node  == None: raise PollException("%s can not be calculated without node data." % key)

			iface  = super(ExecutionContext, self).get('__iface__'      , None)
			if iface  == None: raise PollException("%s can not be calculated without iface data." % key)

			old_ifindex = iface.ifindex

			log.debug("Need to calculate ifindexs for iface '%s' (node: %s)", iface.name, node.name)

			# xform + instance could involve an expression evaluation. Copy all the
			# current context!
			local_context = self.copy()

			ifindex = update_iface_ifindex(log, configuration, cache, node, iface, local_context)

			self['$ifindex'] = ifindex

			log.debug("Calculated ifindex for iface %s = %s", iface.name, local_context['$ifindex'])

			if old_ifindex != ifindex:

				t = super(ExecutionContext, self).__getitem__('__check_time__')

				self.remember_updated_field(iface.uuid, "kind"        , iface.kind)
				self.remember_updated_field(iface.uuid, "ifindex"     , ifindex)
				self.remember_updated_field(iface.uuid, "ifindex_time", t)

		elif key == '$linked_ifindex':

			iface = super(ExecutionContext, self).get('__linked_iface__', None)
			if iface == None: raise PollException("%s can not be calculated without linked iface data." % key)

			old_ifindex = iface.ifindex

			node  = super(ExecutionContext, self).get('__linked_node__' , None)
			if node  == None: raise PollException("%s can not be calculated without linked node data." % key)

			log.debug("Need to calculate ifindexs for linked iface '%s' (node: %s)", iface.name, node.name)

			# xform + instance could involve an expression evaluation. Copy all the
			# current context!
			# All parameter about the linked iface have the 'linked_' prefix!
			local_context = self.copy()

			local_context['$community'     ] = self['$linked_community'     ]
			local_context['$snmpversion'   ] = self['$linked_snmpversion'   ]
			local_context['$snmpport'      ] = self['$linked_snmpport'      ]
			local_context['$snmpv3authdata'] = self['$linked_snmpv3authdata']

			local_context['$node'          ] = self['$linked_node'         ]
			local_context['$iface'         ] = self['$linked_iface'        ]
			local_context['$instance'      ] = self['$linked_instance'     ]
			local_context['$distinguisher' ] = self['$linked_distinguisher']

			ifindex = update_iface_ifindex(log, configuration, cache, node, iface, local_context)

			self['$linked_ifindex'] = ifindex

			log.debug("Calculated linked_ifindex for linked iface %s = %s", iface.name, local_context['$ifindex'])

			if old_ifindex != ifindex:

				t = super(ExecutionContext, self).__getitem__('__check_time__')

				self.remember_updated_field(iface.uuid, "kind"        , iface.kind)
				self.remember_updated_field(iface.uuid, "ifindex"     , ifindex)
				self.remember_updated_field(iface.uuid, "ifindex_time", t)

	def getCalculatedVariable(self, name):

		if super(ExecutionContext, self).__contains__(name):
			return self[name]

		return None

	#
	# Methods for OLD/CURRENT values (SNMP or others)
	#

	def setOldValues(self, values_dict ):
		"""
		This function should be called berfore the expression evaluation starts.
		It stores inside the execution contexts the OLD expressions' values.
		
		See XXXXXXXXXXXXX section
		"""
		self.__old_values = values_dict

	def setOldValue(self, symbol_name, value):
		self.__old_values[symbol_name] = value

	def getOldValue(self, valid):
		"""
		Returns the OLD value
		"""
		if self.__old_values.has_key(valid):
			v = self.__old_values.get(valid)
			self.__log.debug("using old value for %s=%r" % (valid, v))
			return  v
		else:
			self.__log.debug("Missing OLD value for: %s", valid)

			# This past value is missing. No problem,
			# but we have to invalidate the evaluation result!
			# Just as like as the $delta parameter
			raise NotEvaluatedError("%s missing" % valid)

	def saveCurrentValue(self, val_id, value ):
		"""
		Used to save the CURRENT value of an expression.
		"""
		self.getLog().debug("Saving %s = %r", val_id, value)
		self.__current_values[val_id] = value

	def getCurrentValues(self):
		"""
		Used when the expression evaluation is finished. Returns
		the set of CURRENT values retrieved for the expression evaluations.
		"""
		return self.__current_values

	#
	#
	#
	def cacheSimpleValue(self, name, value):
		self.__simple_value_cache[name] = value

	def getCachedSimpleValue(self, name):
		if self.__simple_value_cache.has_key(name):
			return self.__simple_value_cache[name]
		return None

	#
	# SNMP cache 
	#
	def cacheSnmpValue(self, oid, value):
		"""
		Saves inside the internal SNMP cache the *value* for the given *oid* identifier.
		"""
		self.__snmp_value_cache[oid] = value

	def cacheSnmpTable(self, oid, table):
		"""
		Saves inside the internal SNMP cache the *table* for the given *oid* identifier.
		"""
		self.__snmp_table_cache[oid] = table

	def getCachedSnmpValue(self, oid):
		"""
		Returns the cached value for the given *oid* identifier.
		"""
		if self.__snmp_value_cache.has_key(oid):
			return self.__snmp_value_cache[oid]
		return None

	def getCachedSnmpTable(self, oid):
		"""
		Returns the cached table for the given *oid* identifier.
		"""
		if self.__snmp_table_cache.has_key(oid):
			return self.__snmp_table_cache[oid]
		return None

	#
	#
	#
	def appendReason(self, msg):
		# FIX: remove this method
		self.appendExecutionInfo(msg)

	def getReason(self):
		# FIX: remove this methods
		return self.getExecutionInfos()

	def appendExecutionInfo(self, msg):
		self.exec_infos.append(msg)

	def getExecutionInfos(self):
		return self.exec_infos

	def resetExecInfos(self):
		# FIX: a better way to clean up a list?
		self.exec_infos = []

	def saveExecInfoByName(self, key):
		self.exec_infos_by_name[key] = ','.join(self.exec_infos)

	def getSavedExecInfo(self, key):
		return self.exec_infos_by_name.get(key, None)

	def __str__(self):
		try:
			s = ''
			for k in sorted(self.keys()):
				# IMPORTANT: avoid all context's controls on keys!
				v = super(ExecutionContext, self).__getitem__(k)
				s = s + "%s = %s (%s)" % (k, v,type(v)) + "\n"
		except:
			pass
		return s

	#-----------------------------------------------------------------------------#

	def calculateLazySnmpValue(self, oid, community, host, port):
		# Quando chiediamo al contesto un valore SNMP dobbiamo istanziare
		# un oggetto "Lazy"
		return SanetSnmpValueType(self, oid, community, host, port)

#-----------------------------------------------------------------------------#
# Special function for SNMP expression handling                               #
#-----------------------------------------------------------------------------#

def calculate_implicit_parameters(context, host, community, port):
	"""
	Questa funzione cerca di calcolare i valori effettivamanete da usare da usare
	per calcolare una SnmpExpr in base al contesto e parametri passati in input che 
	possono essere non valorizzati:
	
		1.3.6.1.2.1.1.1.0 [ : <snmp community> ] @ [ <host> [ : <snmp port> ] ]
	
	Mapping:
	
		host              -> host        o $node
		snmp community    -> community   o $community
		snmp port         -> port        o $snmpport
		snmp version      ->               $snmpversion
		snmp v3 authdata  ->               $snmpv3authdata
	"""

	host = host or context.get('$node', None)
	if host == None: raise PollException("Missing $node missing or invalid.")

	# Quando una SnmpExpr utilizza funzioni per il clustering senza specificare alcuni 
	# campi (Es. 1.3.6.1.2.1.1.1@clusterNode()), tutta l'espressione va calcolata
	# "spostandosi" sul clusterNode().
	# Eventuali campi "impliciti" devono essere presi dal clusterNode() e non dal contesto
	# o dal nodo corrente.
	#
	#   1.3.6.1.2.1.1.1 @ clusterNode() ---SI--> 1.3.6.1.2.1.1.1 : $cluster_community @ $cluster_node : $cluster_port  [ $cluster_sv ]
	#                                   ---NO--> 1.3.6.1.2.1.1.1 : $community         @ $node         : $port          [ $sv ]
	#
	# Questi prametri andrebbero presi dal nodo "master" (__cluster_node__) 

	# Se il contesto corrente contiene __cluster_node__ vuol dire che 
	# stiamo valutando una expr di un nodo dependent.
	#
	# Accertiamoci che eventuali campi impliciti vengano presi dal nodo giusto.
	cluster_node = context.get('__cluster_node__', None)
	if cluster_node is not None and cluster_node.name == host:

		sv =              cluster_node.snmp_version
		c  = community or cluster_node.snmp_community
		p  = port      or cluster_node.snmp_port
		a  =              cluster_node.snmp_v3_authdata
	else:
		# Campi impliciti in una situazione normale

		# Bisogna distinguere i seguenti casi:
		# - $snmpversion non e' definito (NONE) -> forziamo SNMPv1
		# - $snmpversion = 0                    -> SNMP disabilitato (caso gesto dalle primitive SNMP del poller)
		# - $snmpversion = 1,2,3                -> SNMP normale
		sv =              context.get('$snmpversion',None) 
		if sv is None:
			sv = SNMP_DEFAULT_VERSION

		c  = community or context.get('$community'     , None) or SNMP_DEFAULT_COMMUNITY
		p  = port      or context.get('$snmpport'      , None) or SNMP_DEFAULT_PORT
		a  =              context.get('$snmpv3authdata', None)

	#print "\t\t-> host=%s, sv=%s, community=%s, port=%s)" % (host, sv, c, p)

	return host, sv, c, p, a


class SanetSnmpValueType(SnmpValueType):

	def __init__(self, *args, **kwargs):
		super(SanetSnmpValueType, self).__init__(*args, **kwargs)

		# Questi sono i campi valorizzati in fase di creazione dell'oggetto
		#   self.context      = ...
		#   self.oid          = ...
		#   self.community    = ...
		#   self.host         = ...
		#   self.port         = ...
		#   self.sv           = ...
		#   self.authdata     = ...

		# I campi che potrebbero non essere valorizzati sono:
		#
		#    - self.community
		#    - self.host
		#    - self.port
		#    - self.sv         (non specificabile in ogni caso)
		#    - self.authdata   (non specificabile in ogni caso)
		#
		# Calcoliamo quali valori impliciti devono essere effetivamente usati
		self.host, self.sv, self.community, self.port, self.authdata = calculate_implicit_parameters(self.context, self.host, self.community, self.port)

	def effective_snmpget(self): #, context, oid, community, host, port):
		"""
		This function is used to retrieve the current (new) value of SNMP expression.
		
		It's a wrapper for the :func:`getvar` but it also does additional operations:
		
		* It looks for cached SNMP values inside the *context*'s SNMP cache.
		* It update the "current" value of every SNMP expression. This value
		  will be used as the "old" value at the next check.
		"""

		context   = self.context
		host      = self.host
		oid       = self.oid
		community = self.community
		port      = self.port
		sv        = self.sv
		authdata  = self.authdata

		tries   = context.get('$shorttries', None)
		timeout = context.get('$timeout'   , None)
		if tries   == None: raise PollException("Parameter $shorttries missing or invalid.")		
		if timeout == None: raise PollException("Parameter $timeout missing or invalid.")
		
		val = None

		# Look in the SNMP cache first
		cache_key = host+oid
		val = context.getCachedSnmpValue(cache_key)
		if val != None:
			return val

		# SNMP GET
		community_or_authdata = authdata or community   # snmp v3

		if _DONT_USE_SNMP_SESSION:
			val = getvar(host, community_or_authdata, tries, oid, port, version=sv, timeout=timeout)
		else:
			sess = self.context.get_or_create_snmp_session(host, community_or_authdata, tries, port, version=sv, timeout=timeout)
			val = getvar_s(sess, oid)


		if val != None:
			# Cache value
			context.cacheSnmpValue(cache_key, val)

			old_id  = "%s%s%s:%s" % (oid, '#', host, port)
			context.saveCurrentValue(old_id, val)

			return val

		# We are trying to calculate and use a NEW snmp value
		# inside a expression evaluation, but something gone wrong.
		# We raise an exception because the expression
		# evaluation witout NEW values is meaningless 
		new_id  = "%s:%s%s%s:%s" % (oid, community, '@', host, port)	

		raise UndefinedSnmpValueException("Current variable %s can't be read or is not present (GET)" % (new_id) )

	def effective_snmpwalk(self): #, context, oid, community, host, port):

		"""
		This function is used to retrieve the current (new) table of SNMP expression.
		
		It's a wrapper for the :func:`getvar` but it also does additional operations:
		
		* It looks for cached SNMP values inside the *context*'s SNMP cache.
		* It update the "current" value of every SNMP expression. This value
		  will be used as the "old" value at the next check.
		"""

		context   = self.context
		host      = self.host
		oid       = self.oid
		community = self.community
		port      = self.port
		sv        = self.sv
		authdata  = self.authdata

		tries   = context.get('$shorttries', None)
		timeout = context.get('$timeout'   , None)
		if tries   == None: raise PollException("Parameter $shorttries missing or invalid.")
		if timeout == None: raise PollException("Parameter $timeout missing or invalid.")
		
		val   = None
		
		# Look in the SNMP cache first
		cache_key = host+oid
		val = context.getCachedSnmpTable(cache_key)
		if val != None:
			return val
		
		# SNMP WALK
		community_or_authdata = authdata or community   # snmp v3



		if _DONT_USE_SNMP_SESSION:
			val = gettable(host, community_or_authdata, tries, oid, int(port), version=sv, timeout=timeout)
		else:
			sess = self.context.get_or_create_snmp_session(host, community_or_authdata, tries, port, version=sv, timeout=timeout)
			val = gettable_s(sess, oid)


		if val != None:
			# cache it for next occurrence in the same expression.
			context.cacheSnmpTable(cache_key, val)
			
			# FIX: Save old tables too?
		
			return val
		
		# We are trying to calculate and use a NEW snmp value
		# inside a expression, but something have gone wrong.
		# We raise an exception because the expression
		# evaluation witout NEW values is meaningless 
		new_id  = "%s:%s%s%s:%s" % (oid, community, '@', host, port)
		raise UndefinedSnmpValueException("Current variable %s can't be read or is not present (WALK)" % (new_id))		

#-----------------------------------------------------------------------------#

class RealSnmpValueExprTree(SnmpValueExprTree):
	"""
	This class implements and extends the :class:`SnmpValueExprTree` base class.
	"""

	def __init__(self, exprOID, eCommunity, charType, eHost, ePort):
		super(RealSnmpValueExprTree, self).__init__(exprOID, eCommunity, charType, eHost, ePort)

	def calculateSnmpValue(self, context, oid, community, ctype, host, port):
		
		# In questo punto di esecuzione i parametri "host", "community", "port" 
		# potrebbero non essere valorizzati:
		#
		#          1.3.6.1.2.1.1.1.0 [ : <community> ] @ [ <host> [ : <port> ] ]
		#

		val = None

		# Current value
		if ctype == '@':
			# LAZY VALUE
			#
			# Non sappiamo ancora se questa espressione SNMP
			# rappresenta una GET o una WALK quindi non restituiamo un
			# valore semplice ma un oggetto SnmpValueType che calcolera'
			# il risultato della GET/WALK solo quando non sara' esplicito
			# il tipo di operazione che si intende fare con questa espressione.
			val = context.calculateLazySnmpValue(oid, community, host, port)
			
		# Last value
		elif ctype == '#':
			
			# I valori precedenti sono salvati nel contesto con chiave.
			# 
			#     <oid>#<host><port>
			# 
			# Se "host" e "port" non sono valorizzati, li prendiamo dal contesto corrente.

			host = host or context.get('$node'    , None)
			port = port or context.get('$snmpport', None) or SNMP_DEFAULT_PORT

			if host == None: raise PollException("Missing $node missing or invalid.")
			if port == None: raise PollException("Missing $snmpport missing or invalid.")

			# I vecchi valori SNMP sono sempre memorizzati nel contesto
			# corrente.
			old_id  = "%s%s%s:%s" % (oid, '#', host, port)
			val = context.getOldValue(old_id)

		return val

#-----------------------------------------------------------------------------#

class RealDataIdentifierExprTree(DataIdentifierExprTree):

	def __init__(self, dataId):
		super(RealDataIdentifierExprTree, self).__init__(dataId)

	def calculateDataValue(self, context):

		if self.dtype == '@':

			if self.id in context:
				return context[self.id]
			else:
				# Questo identificatore manca nel contesto di esecuzione. 
				# E' utile sollevare un errore che indica perche' manca l'identificatore.
				#
				# Ho deciso che per fare questo e' piu' pratico poter salvare le "exec info" in base ad una chiave
				# e poterle recuperare.
				#
				# Vediamo se c'e' una execInfo salvata con lo stesso
				# nome di quell'identificatore. Chi ha creato il contesto di esecuzione
				# deve preoccuparsi di creare l'execInfo se ha i dati per farlo.
				# 
				# Non e' una soluzione pulitissima (troppe assunzioni sul funzionamento specifico di sanet), ma funziona.
				infos = context.getSavedExecInfo(self.id)
				if infos is None:
					infos = "n/a" 

				raise ExecutionError("'%s' is not defined or uncheckable (reason: %s)" % (self.id, infos) )

		elif self.dtype == '#':
			v = context.getOldValue(self.id)
			return v

#-----------------------------------------------------------------------------#

#
# The global expression factory used for expression parsing. 
#
# IMPORTANT: The class is reentrant (and thread-safe!), so we can use a single global instance of it.
#
#
class RealExprFactory(BaseExprFactory):
	"""
	This class extends the default :class:`BaseExprFactory` base class.
	The :class:`BaseExprFactory` can create complete expression trees, but
	for SNMP sub-expression trees this class returns the specific Poller's
	:class:`RealSnmpValueExprTree` class.
	"""
	
	def createSnmpValue(self, exprOID, eCommunity, charType, eHost, ePort):		
		return RealSnmpValueExprTree(exprOID, eCommunity, charType, eHost, ePort)

	def createDataIdentifier(self, dataId):
		return RealDataIdentifierExprTree(dataId)

# Singleton
_exprfactory = RealExprFactory()

#-----------------------------------------------------------------------------#
# UTILITY FUNCTIONS                                                           #
#-----------------------------------------------------------------------------#

def compile_str_expr(log, str_expr):
	"""
	Parses the expression *str_expr* and returns the corresponding :class:`ExprTree` object 
	
	:param  log: logger.Logger object
	
	:param  str_expr: The expression
	:type   str_expr: string
	
	:returns: The parsed :class:`ExprTree` object
		
	:exception: 
	        :class:`PollException` if a parsing error occurred
	"""
	
	try:
		#tree = parseStringWithFactory(str_expr, _exprfactory)
		tree = parseString(str_expr, factory=_exprfactory)
	except ParserException, es:
		raise PollException("Error while parsing expression: %s" % (es))

	return tree


def expreval(log, context, expr_tree):
	"""
	Evaluates :class:`ExprTree` object within the given execution *context*. 
	
	:param  log: logger.Logger object
	
	:param  context: The expression's :class:`ExecutionContext`
	
	:param  expr_tree: The :class:`ExprTree` object.
	
	:returns: On success returns the expression value.
	
	          Returns *None* if the value of any OLD SNMP expression
		  is missing or the expression uses the *$delta* symbol
		  but it can't be calculated.
		
	:exception: 
	        :class:`PollException` If a Poller's internal error occurred.
		
	        :class:`Exception` If any :class:`StandardError` occurred.
	"""
	try:
		value  = expr_tree.evaluate(context)
		
		# FIX: value non deve essere mai NONE (nessuna funzione usata nelle expr deve mai restituire NONE come valore).
		#      
		#      Il valore di ritorno NONE viene restituito da questa funzione (expreval) solo quando mancano
		#      dei valori al check precedente (#) (vedi eccezione NotEvaluatedError)
		#
		#         if value is None:
		#            raise PollException("Expression calculated a null value. This is not allowed.")
		#
		#      Necessario ricontrollare le funzioni del poller e la gestione della valutazione dell'albero
		#      per vedere come sono gestiti i valori NONE.
		#
		#      Ticket aperto: #2386

		return value
	
	# FIX: Convert every possibile (not fatal) exception in a PollException
	#      Poller should never stops if an expression evaluation raise 
	#      an (not fatal) exception.
	
	except NotEvaluatedError, ex:
		# We raise this particular exception only when some past values 
		# are missing!
		# We stop the exception here because we want to save 
		# all the current snmp values for the next evaluation.
		return None
	
	except ExecutionError, ex:
		# Some variables or functions are missing in the context.
		# Very strange! There is probably en error in the expression!
		raise PollException("Execution error: %s " % (ex)) #, traceback.format_exc()))

	except UndefinedSnmpValueException, ex:
		# Somewhere we couldn't read an snmp value.
		raise ex
		
	# HISTORY: Added for ticket #1885.
	except ForceUncheckableException, ex:
		# Somewhere we raised an Exception that force uncheckable state!
		raise ex
		
	except TimeoutException, ex:  #PollException
		raise ex	
	
	except SnmpWrapperBUGException, ex:
		# Special case: the snmp library is bugged and we want to be sure these case will
		#               produce an uncheckable result.
		#
		# HISTORY: Added for ticket #1885.
		#
		raise ForceUncheckableException(str(ex))

	except PollException, ex:
		# An important condition check has failed. (Probably inside
		# a expression function handler).
		raise PollException("Error while evaluating expression: %s " % (ex)) #, traceback.format_exc()) )
	
	except ValueError, ex:
		# Invalid input parameter somewhere
		raise PollException("Invalid parameter somewhere in expression: %s\nTRACE:%s" % (ex, traceback.format_exc()) )
	
	except ZeroDivisionError, ex:
		#raise PollException("Division by Zero: %s " % (ex)) #, traceback.format_exc()) )
		raise PollException("Division by Zero: %s\nTRACE:%s" % (ex, traceback.format_exc()) )
	
	except TypeError, ex:
		#raise PollException("Invalid type somewhere in expression: %s " % (ex)) #, traceback.format_exc()) )
		raise PollException("Invalid type somewhere in expression: %s->%s" % (ex, traceback.format_exc()) )
	
	except BaseException, ex:
		# We want to intercept any kind of exception raised by external libs 
		# that is not a builtin exception.
		if (isinstance(ex, StandardError) or
		    isinstance(ex, StopIteration) or
		    isinstance(ex, Warning) or
		    isinstance(ex, SystemExit) or 
		    isinstance(ex, KeyboardInterrupt) or 
		    isinstance(ex, GeneratorExit)):
			log.error("Generic error: %s -> %s", ex, traceback.format_exc() )
			raise Exception("Generic error: %s -> %s" % (ex, traceback.format_exc()) ) 
		else:
			log.error("Unknown error while evaluating expression: %s -> %s", ex, traceback.format_exc() )
			raise PollException("Unknown error while evaluating expression: %s" % (ex)) #, traceback.format_exc()) )


#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def get_cached_expr_tree(log, cache, expr_str,  expr_id, tag):


	ck = expr_id
	
	# IMPORTANTE: Nella cache viene salvato l'albero sintattico. Se per un qualunque
	#             motivo il caricamento dell'albero non va a buon fine ignoriamo
	#             il problema e ricompiliamo l'espressione.
	#
	#             I possibili motivi sono 3:
	#
	#               a. La cache e' stata compromessa -> ignoriamo!
	#
	#               b. L'albero sintattico salvato non puo' essere ricreato
	#                  in memoria perche' sono cambiate le classi (e quindi gli oggetti)
	#                  che erano state usate 
	#                       -> Problema di aggiornamento software -> tiriamo dritto
	# 
	#               c. Qualcosa non sta funzionando con la cache -
	#                       -> sarebbe da indagare ma ignoriamo 
	#
	# FIX: Trovare un modo per gestire il punto (c)
	
	try:
		cached_expr = cache.get(ck)
	except Exception, e:
		print "ERROR WHILE LOADING EXPR CACHE. Key %s -> %s %s" % (ck, type(e), e)
		cached_expr = None

	if cached_expr is None:
		log.log(9, "NOT hit cached expr: %s" % (expr_id))

	#
	# Verify the cached expression tree is still valid.
	#
	expr_tree_obj = None

	# cached tuple  <timestamp, expr>
	if isinstance(cached_expr, types.TupleType):
		(expr_tag, expr_tree_obj) = cached_expr
		if expr_tag != tag:
			log.log(9, "cached expr expired: tag %s, last tag %s", tag, expr_tag )
			expr_tree_obj = None

	if expr_tree_obj == None:

		#
		# Compile expr tree
		#
		expr_tree_obj = compile_str_expr(log, expr_str)

		#
		# Cache the compiled expression for the future
		#
		# save <timestamp, expr>
		data = (tag, expr_tree_obj)

		#
		# FIX: what if the cache raises an exception here?
		#
		log.log(9, "Saving expr %s in cache: %s", expr_id, ck)
		try:
			cache.set(ck, data)
		except Exception, e:
			print "ERROR WHILE SAVING IN EXPR CACHE. Key %s -> %s %s" % (ck, type(e), e)

	return expr_tree_obj

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def update_element_index(log, configuration, cache, node, element, index_field_name, convert_function, params):
	
	updated = False
	
	index_var_name    = '$' + index_field_name
	index_field_value = getattr(element, index_field_name)

	log.debug("update_element_index(%s:%s)", node.name, element.name)
	log.debug("\txform         = %s" , element.xform)
	log.debug("\tdistinguisher = %s" , element.distinguisher)
	log.debug("\t%s            = %s" , index_field_name, index_field_value)

	if element.xform == "" and (str(index_field_value) == element.distinguisher):
		log.debug("%s eq distinguisher and xform null" % (index_field_name));
		params[index_var_name] = str(index_field_value)
		return (updated, index_field_value)

	# <<<<<<<<<<<<<<<<< ifindex no more number <<<<<<<<<<<<<<<<<<<<<<
	## NB: This is an hack. The $instance variable must be a number!
	##      Sometimes ifaces' configuration is wrong!
	#if not element.xform:
	#	if element.distinguisher is None or not element.distinguisher.isdigit():
	#		log.error("element %s, $distinguisher=%s it not digit!? Skip update stindex", element.name, element.distinguisher);
        #
	#		if getattr(element, index_field_name) is None:
	#			raise PollException("Unable to calculate stindex for %s:%s. Check xform and instance parameters." % (node.name, element.name))
	#			
	#		return element.stindex
	# <<<<<<<<<<<<<<<<< ifindex no more number <<<<<<<<<<<<<<<<<<<<<<

	if element.xform: expr = element.xform
	else:             expr = "$distinguisher"

	try:
		log.debug("Calculating stindex value with: %s", expr)
				
		expr_id  = '%s.xform' % (element.uuid)
		expr_tag = str(crc32(expr))
		
		log.debug("Loading from cache: %s", expr_id)
		xform_expr_tree = get_cached_expr_tree(log, cache, expr,  expr_id, expr_tag)
		
		
		# Create a local context in order to not override parameters
		local_context = ExecutionContext()
		
		local_context.setLog(log)
		local_context.setConfiguration( configuration )
		local_context.setCache(cache)
		local_context.update(params)
		
		new_index_field_value = expreval(log, local_context, xform_expr_tree)
		
	except TimeoutException, ex:
		raise TimeoutException("Cannot determine %s: %s" % (index_field_name, str(ex)) )
	
	# HISTORY: Added for ticket #1885.
	except ForceUncheckableException, ex:
		raise ex

	except PollException, ex:
		raise PollException("Cannot determine %s: %s" % (index_field_name, str(ex)) )
	except Exception, ex:
		raise PollException("Unknown error. Cannot determine %s: %s" % (index_field_name,str(ex)))

	if new_index_field_value == None:
		# Something were missing! Very strange, it should not happen!
		raise PollException("Cannot determine %s value = %s" % (index_field_name, new_index_field_value))
	
	#
	# Convert the "index" value to the disired type
	#
	new_index_field_value = convert_function(new_index_field_value)
	
	params[index_var_name] = new_index_field_value
	
	if index_field_value == new_index_field_value:
		log.debug("%s verified = %s (%s)" % (index_field_name, new_index_field_value, type(new_index_field_value) ))
		pass
	else:
		log.debug("%s: updating %s from %s to %s", element, index_field_name, index_field_value, new_index_field_value)
		#element.stindex = new_index_field_value
		setattr(element, index_field_name, new_index_field_value)
		updated = True
		
	return (updated, getattr(element, index_field_name))


#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def update_iface_ifindex(log, configuration, cache, node, iface, params):
	

	iface.instance = iface.distinguisher

	log.debug("update_iface_ifindex(%s)"            , iface.name)
	log.debug("update_iface_ifindex: xform    = %s" , iface.xform)
	log.debug("update_iface_ifindex: ifindex  = %s" , iface.ifindex)
	log.debug("update_iface_ifindex: instance = %s" , iface.instance)
	
	
	if iface.xform == "" and (str(iface.ifindex) == iface.instance):
		log.debug("ifindex eq instance and xform null");
		params['$ifindex'] = iface.ifindex
		return iface.ifindex
	
	# <<<<<<<<<<<<<<<<< ifindex no more number <<<<<<<<<<<<<<<<<<<<<<
	# NB: This is an hack. The $instance variable must be a number!
	#      Sometimes ifaces' configuration is wrong!
	#if iface.xform == "" and not iface.instance.isdigit():
	#	log.error("iface %s, $instance=%s it not digit!? Skip update ifindex", iface.name, iface.instance);
	#	if iface.ifindex == None:
	#		raise PollException("Unable to calculate ifindex for %s:%s. Check xform and instance parameters." % (iface.node, iface))
	#	return iface.ifindex
	# <<<<<<<<<<<<<<<<< ifindex no more number <<<<<<<<<<<<<<<<<<<<<<
		
	
	if iface.xform: expr = iface.xform
	else:           expr = "$instance"
	
	try:
		log.debug("Calculating ifindex value with: %s", expr)
				
		expr_id  = '%s.xform' % (iface.uuid)
		expr_tag = str(crc32(expr))
		log.debug("Loading from cache: %s", expr_id)
		xform_expr_tree = get_cached_expr_tree(log, cache, expr,  expr_id, expr_tag)
		
		
		# Create a local context in order to not override parameters
		local_context = ExecutionContext()
		local_context.setLog(log)
		local_context.setConfiguration( configuration )
		local_context.setCache(cache)
		local_context.update(params)
		#local_context.update(configuration.get_global_symbols() )

		ifindex_value = expreval(log, local_context, xform_expr_tree)

	except ForceUncheckableException, ex:
		# Special case: the snmp library is bugged and we want to be sure these case will
		#               produce an uncheckable result.
		#
		# HISTORY: Added for ticket #1885.
		#
		raise ex
		
	except TimeoutException, ex:
		raise TimeoutException("Cannot determine ifindex: %s" % str(ex))
		
		
	except PollException, ex:
		raise PollException("Cannot determine ifindex: %s" % str(ex))

	except Exception, ex:
		raise PollException("Unknown error. Cannot determine ifindex: %s" % str(ex))

	if ifindex_value == None:
		# Something were missing! Very strange, it should not happen!
		raise PollException("Cannot determine ifindex: value = %s" % ifindex_value)
	
	# <<<<<<<<<<<<<<<<< ifindex no more number <<<<<<<<<<<<<<<<<<<<<<
	#ifindex_value = int(ifindex_value)								
	# <<<<<<<<<<<<<<<<< ifindex no more number <<<<<<<<<<<<<<<<<<<<<<
	
	params['$ifindex'] = ifindex_value
	
	if iface.ifindex == ifindex_value:
		log.debug("ifindex verified as %s" % ifindex_value)
	else:
		log.debug("iface %s: updating ifindex to %s", iface, ifindex_value)
		# <<<<<<<<<<<<<<<<< ifindex no more number <<<<<<<<<<<<<<<<<<<<<<
		#iface.ifindex = int(ifindex_value)
		# <<<<<<<<<<<<<<<<< ifindex no more number <<<<<<<<<<<<<<<<<<<<<<
		iface.ifindex = str(ifindex_value)
		
	return iface.ifindex

#-----------------------------------------------------------------------------#
#  Clusters                                                                   #
#-----------------------------------------------------------------------------#


def calculate_cluster_parameters(log, execcontext, configuration, cache, node): 

	#
	# Load cluster data
	#
	cluster           = configuration.get_cluster( node.cluster_id )
	
	all_cluster_nodes = configuration.get_cluster_elements( node.cluster_id )
	
	cluster_nodes = [ e for e in all_cluster_nodes if e.cluster_role == CLUSTER_ROLE_MASTER  ]

	if log.isEnabledFor(10): # Debug
		log.debug("Calculating cluster association")
		log.debug("\tnode          = %s (%s)", node.name, node.uuid)
		log.debug("\txform         = %s", node.cluster_xform)
		log.debug("\tdistinguisher = %s", node.cluster_distinguisher)
		log.debug("\tinstance      = %s", node.cluster_instance)
		log.debug("\tclusternodes  = %s", ','.join([n.name for n in cluster_nodes]) )
		log.debug("\tcurr. master  = %s", node.cluster_element_id)

	infos2 = []
	infos2.append( 'masters:[%s]' % (','.join([n.name for n in cluster_nodes]),)  )

	if not cluster_nodes:raise ClusterException("Unable to calculate cluster informations without active cluster nodes: %s nodes" % (len(all_cluster_nodes),))
		
	if node.cluster_distinguisher == None or  node.cluster_distinguisher == '':
		raise ClusterException("Unable to calculate cluster association. Node's distinguisher not set or invalid: %s" % (node.cluster_distinguisher,))

	if node.cluster_xform == None or node.cluster_xform == '':
		raise ClusterException("Unable to calculate cluster association. Node's xform not set or invalid: %sx" % (node.cluster_xform,))

	#
	# Compile/Load cluster_xform expression tree
	#
	log.debug("Loading cluster_xform expr...")

	expr_str = node.cluster_xform
	
	expr_id  = '%s.cluster_xform' % (node.uuid)
	expr_tag = str(crc32(expr_str)) 	
	cluster_xform_expr_tree = get_cached_expr_tree(log, cache, expr_str, expr_id, expr_tag)

	#
	# Calculate cluster node and instance
	#
	infos = [] 
	real_node              = None
	cluster_instance_value = None

	# Try the last known association first
	if node.cluster_element_id != None:
		log.debug("Testing old association with node id %s", node.cluster_element_id)


		# Il nodo a cui eravamo associati potrebbe essere cambiato (cancellato e ricreato)
		# Questo caso e' un po' particolare. Ignoriamo elementi mancanti e riproviamo a calcolare l'associazione.
		try:
			real_node =  configuration.get_element_by_id( node.cluster_element_id )

			# Il nodo a cui eravamo associati esiste, MA appartiene ancora al cluster o la configurazione e' cambiata?
			if real_node not in cluster_nodes:
				log.info("Old association no more valid. %s does not belong to this cluster anymore", real_node.name)
				real_node = None

		except ResourceNotFoundException, e:
			log.info("Old association with node %s need recomputation: node not found: %s", node.cluster_element_id, e)
			real_node = None


		if real_node != None:
			log.debug("Old node is %s", real_node)

			cluster_instance_value, info = verify_association(log, configuration, cache, node, real_node, cluster_xform_expr_tree, execcontext)
			if cluster_instance_value != None:
				log.debug("Old association still active")
				pass
			else:
				# Association not found, but remove this node from the list of nodes to test
				cluster_nodes = [ n for n in  cluster_nodes  if n.uuid != real_node.uuid ]
				
				real_node = None
				log.debug("Last association not verified: %s", info)
				infos.append(info)
		else:
			log.debug("Last cluster association expired, the cluster_node %s is missing",  node.cluster_element_id)
			pass

	if real_node == None:
		log.debug("Recalculating association with all cluster nodes: %s", ','.join([n.name for n in cluster_nodes]))
		
		for real_node in cluster_nodes:
			
			log.debug("Verifying association with node '%s'", real_node.name)
			
			cluster_instance_value, info =  verify_association(log, configuration, cache, node, real_node, cluster_xform_expr_tree, execcontext)
			if cluster_instance_value != None:
				# Association found
				break

			log.debug("Not verified: %s", info)
			infos.append(info)

	# Association changed?
	if real_node != None and cluster_instance_value != None:

		changed = False
		if node.cluster_element_id != real_node.uuid or node.cluster_instance != cluster_instance_value:
			
			changed = True
			
			node.cluster_element_id  = real_node.uuid
			node.cluster_instance    = cluster_instance_value

		return real_node,  cluster_instance_value, changed, infos2
	else:		
		raise ClusterException("Unable to verify cluster association:: %s" % ('\n'.join(infos) ,))



def verify_association(log, configuration, cache, node, real_node, xform_expr_tree, original_exec_context):

	info = ''
	cluster_instance_value = None
		
	try:
		# Don not modify original tree
		xform_expr_tree = copy.deepcopy( xform_expr_tree )

		log.debug("Prepaing context for association")
		
		#
		# Use params from the CONTROLLER
		#
		params = {}
		params.update( real_node.get_expr_symbols() )
		params['__node__'] = real_node

		#
		# ATTENTION: use some params from the DEPENDENT!!
		#
		params['$cluster_distinguisher'] = node.cluster_distinguisher
		params['$shorttries']            = original_exec_context.get('$shorttries')


		context = ExecutionContext()
		context.setLog(log)
		context.setCache(cache)
		context.setConfiguration( configuration )
	
		context.update(params)
		context.update( configuration.get_global_symbols() )

		log.debug("Executing association expression")
		
		cluster_instance_value = expreval(log, context, xform_expr_tree)
		if cluster_instance_value == None:
			raise Exception("Cluster xform returned a null value")
			
		# La cluster instance e' sempre una stringa. Se la expr ha calcolato qualcos'altro
		# forziamo il tipo
		cluster_instance_value = str(cluster_instance_value)

		
		log.debug("Association found!")
		log.debug("\tcluster_node     -> %s", real_node)
		log.debug("\tcluster_instance -> %s (%s)", cluster_instance_value, type(cluster_instance_value))

	except ForceUncheckableException, ex:
		# Special case: the snmp library is bugged and we want to be sure these case will
		#               produce an uncheckable result.
		#
		# HISTORY: Added for ticket #1885.
		#
		raise ex

	except TimeoutException, ex:
		info = '%s not verified (%s)' % (real_node.name, str(ex))

	except PollException, ex:
		info = '%s not verified (%s)' % (real_node.name, str(ex))

	except Exception, ex:
		info = '%s not verified (%s)' % (real_node.name, str(ex))

	return  cluster_instance_value, info

