# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>. 

"""

Moduli dell'agente
==========================

I moduli caricati dell'agente permettono di "estendere" l'agente in due modi:

 * Definire nuove funzioni da utilizzare nelle "expr"
 
 * Definire nuovi controlli da chiamare per verificare in funzionamento interno dell'agente.
 
   Queste funzioni vengono chiamate quando l'agente viene interrogato dall'esterno sul suo
   funzioanmento interno.

Funzioni per le expr
----------------------

Per dichiarare quali funzioni "rendere visibili" nell expr il modulo deve definire il dizionario EXPR_SYMBOLS
che associa ad uno o piu' *nomi* la funzione corrispondente:

::
	
	EXPR_SYMBOLS = {
		<symbol name> : <function>,
		<symbol name> : <function>,
		...
	}

.. _important:: Il dizionario EXPR_SYMBOLS e' obbligatorio per i moduli standard di Sanet. Se non e' definito l'importazione del modulo da errore e l'esecuzione dell'agente viene bloccata. Le plugin esterne (moduli extra) posono non dichiarare questo EXPR_SYMBOLS.

Per dichiarare nuove funzioni da utilizzare nelle expr un modulo deve definire funzioni con la seguente signature:

::
	
	def my_function(context, ...):
		...
		... do something
		...
		return <value>

Il nome della funzione non e' importante (ma per convenzione tutti i nomi rispettano la sintassi "fun_<symbolo expr>").

IL primo parametro della funzione e' obbligatorio e deve essere chiamato *context*. Contiene il riferimento ad un 
oggetto di tipo ExecutionContext(). 


Funzioni di stato
----------------------

Per dichiarare nuovi funzioni di diagnostica dell'agente bisogna definire il dizionario:

::
	
	INTERNAL_STATUS_FUNCTIONS = {
		'<name>': <function>,
	}

Questo dizionario associa ad un nome "logico" una funzione di diagnostica. 

Ogni modulo puo' quindi definire un numero arbitrario di funzioni di diagnostica.

Le funzioni richiamate devono essere definite con la seguente signature:

::
	
	def status_function(agent):
		...
		... do something
		...
		return <status>, <msg>

La funzione deve festiture sempre una tupla di due elementi:

	* <status>  : Booleano. Indica se il controllo e' verificato.
	* <msg>     : Stringa. Informazione descrittiva sull'esisto del controllo.


Esempio struttura di un modulo
------------------------------------

Dato questo modulo:

::
	
	#-----------------------------------------------------------------------------#
	#                                                                             #
	#-----------------------------------------------------------------------------#
	
	def fun_dummy_function(context, a, b, c):
		return a + b + c
	
	
	EXPR_SYMBOLS = {
		'dummy': fun_dummy_function,
	}
	
	#-----------------------------------------------------------------------------#
	#                                                                             #
	#-----------------------------------------------------------------------------#
	
	def _dummy_status(agent):
		
		agent.logger.info("Calculating internal dummy status...")
	
		status = False
		msg    = "DUMMY plugin not working."
	
		return (status, msg)
	
	
	INTERNAL_STATUS_FUNCTIONS = {
		'dummy_status': _dummy_status,
	}

E' possibile eseguire expr che chiamano il symbolo esportato *dummy*:

::
	dummy(1, 1, 1)

"""

import pkgutil
import sys

from sanet_common.pythonutils import load_symbol_from_module

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

EXPR_SYMBOLS_GLOBAL_SYMBOL = 'EXPR_SYMBOLS'

INTERNAL_STATUS_FUNCTIONS_GLOBAL_SYMBOL = 'INTERNAL_STATUS_FUNCTIONS'

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def load_system_expr_functions():
	"""
	Carica dai moduli "standard" di sanet tutti i simboli usabili nelle expr
	e restituisce un dizionario.
	"""
	expr_symbols, agent_status_functions = load_system_expr_and_agentstatus_functions()
	
	return expr_symbols

def load_system_expr_and_agentstatus_functions():
	"""
	Importa tutti i sottomoduli di "sanet_poller.lib.functions" e carica tutti i:
		* simboli delle expr
		* check di disagnostica definiti
	"""

	# Directory dove si trova questo modulo python
	BASE_PATH = __path__[0]
	
	BASE_NAME = 'sanet_poller.lib.functions'
	
	expr_symbols, agent_status_functions = __load_expr_and_agentstatus_functions_from_path( BASE_PATH, BASE_NAME )
	
	return expr_symbols, agent_status_functions

def __load_expr_and_agentstatus_functions_from_path( path, base_name=None ):

	expr_symbols = {}
	agent_status_functions = {}

	for importer, modname, ispkg in pkgutil.iter_modules([path]):

		if base_name:
			module_fullname = '%s.%s' % (base_name, modname)
		else:
			module_fullname = modname
		
		#
		# IL simbolo e' EXPR_SYMBOLS_GLOBAL_SYMBOL. Se non viene importato
		# deve essere sollevata un'eccezione.
		#
		try:
			symbols = load_symbol_from_module( module_fullname, EXPR_SYMBOLS_GLOBAL_SYMBOL)
			
			expr_symbols.update(symbols)

		except Exception, e:
			raise Exception("Error while importing '%s': %s" % (module_fullname, e))


		#
		# Le funzioni di diagnostica in EXPR_SYMBOLS_GLOBAL_SYMBOL sono "facoltative".
		#
		try:
			symbols = load_symbol_from_module( module_fullname, INTERNAL_STATUS_FUNCTIONS_GLOBAL_SYMBOL)

			agent_status_functions.update(symbols)

		except Exception, e:
			pass

	return expr_symbols, agent_status_functions

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def load_extra_expr_functions(extra_pkgs):
	"""
	Carica dai moduli dichiarati nel vettore *extra_pkgs*  e restituisce un dizionario
	con tutti i simboli esportati utilizzabili nelle expr.
	"""
	expr_symbols, status_functions = load_extra_expr_and_agentstatus_functions(extra_pkgs)

	return expr_symbols

def load_extra_expr_and_agentstatus_functions(extra_pkgs):
	"""
	Carica dai moduli dichiarati nel vettore *extra_pkgs*  e restituisce:
		* un dizionario con tutti i simboli esportati utilizzabili nelle expr.
		* un dizionario con tutte le funzioni di diagnostica.
	"""
	expr_symbols     = {}
	status_functions = {}

	for pkg in extra_pkgs:

		#
		# NOTA: Per i moduli extra (non standard) dell'agente i simboli:
		#
		#             - EXPR_SYMBOLS_GLOBAL_SYMBOL
		#             - INTERNAL_STATUS_FUNCTIONS_GLOBAL_SYMBOL
		#
		#       sono "facoltativi".
		#
		try:
			symbols = load_symbol_from_module( pkg, EXPR_SYMBOLS_GLOBAL_SYMBOL)
			expr_symbols.update(symbols)
		except Exception, e:
			#print e
			pass

		try:
			symbols = load_symbol_from_module( pkg, INTERNAL_STATUS_FUNCTIONS_GLOBAL_SYMBOL)
			status_functions.update(symbols)
		except Exception, e:
			#print e
			pass

	return expr_symbols, status_functions
