# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>. 

"""

SNMP Advanced
=============

adjacent
--------

    **Syntax**

      **adjacent** **(** [ *method* **,** *method **,** *...* ] **)**

    **Description**

      This function must be used only for targets of interfaces. It verifies if a *link* exists between the target interface and its linked interface.

      The presence of a *link* is verified by using several methods. Supported methods are: 
      
      * "stp"
      * "cdp"  (Cisco CDP protocol MIBS)
      * "lldp"
      * "edp"  (Extreme EDP protocol MIBS)

      If no methods are specified, the functions tries all the supported methods in this sequence: stp, cdp, lldp.
      
      .. important:: The function remember the last successfull method and tries it first at the next execution.
      
    **Input**

      * variable number of strings.

    **Output**

      Returns a *boolean*: *true* or *false*. Returns *false* at the first invalid method encountered in the input list.

    **Errors**

      Raise an exception and when error occurs while testing the adjacency.

    **Examples**

      +------------------------------------+------------------------------------------------------------------------+---------------------+
      | Expression                         | Description                                                            | Return value        |
      +====================================+========================================================================+=====================+
      | adjacent()                         |                                                                        |                     |
      +------------------------------------+------------------------------------------------------------------------+---------------------+
      | adjacent("stp")                    |                                                                        |                     |
      +------------------------------------+------------------------------------------------------------------------+---------------------+
      | adjacent("stp", "cdp")             |                                                                        |                     |
      +------------------------------------+------------------------------------------------------------------------+---------------------+
      
"""

from sanet_poller.lib.exceptions import PollException
from sanet_poller.lib.exceptions import UndefinedSnmpValueException
from sanet_poller.lib.exceptions import TimeoutException

from sanet_poller.lib.snmp.stp  import areStpAdj
from sanet_poller.lib.snmp.lldp import areLldpAdj
from sanet_poller.lib.snmp.cdp  import areCdpAdj
from sanet_poller.lib.snmp.edp  import areEdpAdj

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_adjacent(context, *methods):
	
	# OLD Syntax: adjacent(port1:community1@host1,port2:community2@host2)
	# NEW Syntax: adjacent() or adjacent( 'stp', ....) 
	# port is always 161 for this function	
	
	log = context.getLog()

	log.debug("areStpAdj(%s)" % (','.join(methods)))

	if len(methods) == 0:
		methods = ['stp', 'cdp', 'lldp']
	
	log.debug("Selected methods: %s", str(methods))
	
	port1 = context['$ifindex']        # This will trigger ifindex calculation or exception
	host1 = context.get('$node'       , None)
	comm1 = context.get('$community'  , None)
	sv1   = context.get('$snmpversion', None)
	
	port2 = context['$linked_ifindex'] # This will trigger ifindex calculation or exception
	host2 = context.get('$linked_node'       , None)
	comm2 = context.get('$linked_community'  , None)
	sv2   = context.get('$linked_snmpversion', None)
	
	if host1    == None: raise PollException('Parameter $host1 missing or invalid')
	if comm1    == None: raise PollException('Parameter $comm1 missing or invalid')
	if port1    == None: raise PollException('Parameter $port1 missing or invalid')
	if sv1      == None: raise PollException('Parameter $sv1 missing or invalid')

	if host2    == None: raise PollException('Parameter $host2 missing or invalid')
	if comm2    == None: raise PollException('Parameter $comm2 missing or invalid')
	if port2    == None: raise PollException('Parameter $port2 missing or invalid')
	if sv2      == None: raise PollException('Parameter $sv2 missing or invalid')

	retval = False
	
	#
	# Load the last successfull method and try it first.
	#
	method_id = "_adj_method/%s:%s/%s:%s" % (host1, port1, host2, port2)
	

	try:
		log.debug("Loading last method")
		last_method = context.getOldValue(method_id)
	except Exception, e:
		log.debug("Last method not found.")
		last_method = None

	#
	# Put methods in order
	#
	log.debug("Last method was: %s (normal order is: %s)", last_method, methods)
	
	if last_method is not None:
		if last_method in methods:
			methods = [ last_method ] + [ m for m in methods if m != last_method ]
	
	#
	# Check 
	#
	log.debug("Checking methods in order: %s", str(methods))
	
	
	successfull_method = None
	
	for method in methods:
		try:
			method = method.lower()
			
			log.debug("Trying method '%s'", method)
			
			if method == 'stp':
				log.debug("Calling areStpAdj(%s, %s, %s, %s, %s, %s, %s, %s)", host1, host2, port1,port2, comm1,comm2, sv1,sv2 )
				(retval, reason) = areStpAdj(log, host1, host2, port1, port2, comm1, comm2, sv1, sv2)

				# Fix: with STP retval can be None. Need to refactor areStpAdj() logic.
				if retval is None:
					retval = False
			elif method == 'cdp':
				log.debug("Calling areCdpAdj(%s, %s, %s, %s, %s, %s, %s, %s)", host1, host2, port1,port2, comm1,comm2, sv1,sv2 )
				(retval, reason) = areCdpAdj(log, host1, host2, port1, port2, comm1, comm2, sv1, sv2)
			elif method == 'lldp':
                                log.debug("Calling areLldpAdj(%s, %s, %s, %s, %s, %s, %s, %s)", host1, host2, port1,port2, comm1,comm2, sv1,sv2 )
                                (retval, reason) = areLldpAdj(log, host1, host2, port1, port2, comm1, comm2, sv1, sv2)

			elif method == 'edp':
				log.debug("Calling areEdpAdj(%s, %s, %s, %s, %s, %s, %s, %s)", host1, host2, port1,port2, comm1,comm2, sv1,sv2 )
				(retval, reason) = areEdpAdj(log, host1, host2, port1, port2, comm1, comm2, sv1, sv2)
			else:
				retval = False
				reason = 'Protocol "%s" not supported' % (method)
	
			context.appendReason('method "%s":%s' % (method, reason) )
			
			if retval == True:
				successfull_method = method
				break
			
		except UndefinedSnmpValueException, e:
			# Just skip to the next method
			reason = 'Method "%s" failed: %s' % (method, str(e))
			log.debug(reason)
			context.appendReason(reason)
			
		except TimeoutException, e:
			# Add informations about which method failed
			reason = 'Timeout error on method "%s": %s' % (method, str(e))
			raise PollException(reason)
		except PollException, e:
			# Add informations about which method failed
			reason = 'Generic error on method "%s": %s' % (method, str(e))
			raise PollException(reason)
	
	#
	# Save the last successfull method 
	#
	if successfull_method:
		context.saveCurrentValue(method_id, successfull_method)

	log.debug("expr become %s", retval);
	
	return retval

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

EXPR_SYMBOLS = {
	'adjacent'                   : fun_adjacent,
}
