# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it> 
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""
Airespace
=========


airespaceAPChannelInterference
------------------------------

    **Syntax**

      calculateAirespaceAccessPointInstance( ap_macaddress ):

    **Description**
      
      Given the Access Point's MAC address *ap_macaddress*, finds the OID instance inside the controller MIB.
      
    **Input**

      * *ap_macaddress*: string containing the hexadecimal MAC Address.

    **Output**

      Returns a string with an OID instance.

    **Errors**

      An exception is raised if an error (es: timeour) occurs,

    **Examples**

      +----------------------------------------------------------+------------------------------------------------------------------------+---------------------+
      | Expression                                               | Description                                                            | Return value        |
      +==========================================================+========================================================================+=====================+
      | calculateAirespaceAccessPointInstance("00AE13CA89")      |                                                                        | *134*               |
      +----------------------------------------------------------+------------------------------------------------------------------------+---------------------+
      

airespaceAPChannelInterference
------------------------------

    **Syntax**

      airespaceAPChannelInterference ( threshold , [ selected_antenna ] )

    **Description**
      
      Read RSSI values inside the Controller MIB for the current AP. All informations are formatted and returned in the
      global verbose status of the expression check.
      
      Only RSSI with value greater than *threshold* are returned. If specified
      
    **Input**

      * *threshold*: Minimum RSSI accepted value.
      * *selected_antenna*: The number of the antenna: 0 for 2.4Gh (default), 1 for 5gh. 

    **Output**

      Returns True if at least one RSSI value satisfies the given threshold/antenna filter.

    **Errors**

      An exception is raised if an error occurs.

    **Examples**

      +-----------------------------------------+------------------------------------------------------------------------+---------------------+
      | Expression                              | Description                                                            | Return value        |
      +=========================================+========================================================================+=====================+
      | airespaceAPChannelInterference(30)      |                                                                        | True                |
      +-----------------------------------------+------------------------------------------------------------------------+---------------------+
      | airespaceAPChannelInterference(30, 1)   |                                                                        |                     |
      +-----------------------------------------+------------------------------------------------------------------------+---------------------+
      

"""
import binascii
import socket
import re

from sanet_poller.lib.snmp.base import getvar, gettable
from sanet_poller.lib.exceptions import PollException, UndefinedSnmpValueException

from sanet_common.sanetdefs.constants import CLUSTER_ROLE_MASTER

#---------------------------------------------------------------------#
#                                                                     #
#---------------------------------------------------------------------#

def __mac_to_dotted_instance(mac_str):
	# 00260BAC5E60  --> 0.38.11.172.94.96
	val = []
	vals = [
		str(int(mac_str[0:2],16)),
		str(int(mac_str[2:4],16)),
		str(int(mac_str[4:6],16)),
		str(int(mac_str[6:8],16)),
		str(int(mac_str[8:10],16)),
		str(int(mac_str[10:12],16))
	]
	return '.'.join(vals)


def __dotted_to_mac_instance(s):
	# "180.164.227.100.112.176" -> 'b4a4e36470b0'
	return ''.join([ hex(int(a))[2:] for a in s.split(".") ])

#---------------------------------------------------------------------#
#                                                                     #
#---------------------------------------------------------------------#

def fun_calculateAirespaceAccessPointInstance(context, ap_macaddress):
	
	log = context.getLog()

	node        = context.get('$node'       , None)
	community   = context.get('$community'  , None)	
	shorttries  = context.get('$shorttries' , None)
	snmpport    = context.get('$snmpport'   , None)
	snmpversion = context.get('$snmpversion', None)

	
	if node        == None: raise PollException('Parameter $node missing.')
	if community   == None: raise PollException('Parameter $community missing.')
	if shorttries  == None: raise PollException('Parameter $shorttries missing.')
	if snmpport    == None: raise PollException('Parameter $snmpport missing.')
	if snmpversion == None: raise PollException('Parameter $snmpversion missing.')

	
	BASE_OID = '1.3.6.1.4.1.14179.2.2.1.1.1'

	log.debug('calculateAirespaceAccessPointInstance(%s)', ap_macaddress)

	instance = __mac_to_dotted_instance(ap_macaddress)

	log.debug("MAC     : %s", ap_macaddress)
	log.debug("Instance: %s", instance)
	
	oid = BASE_OID +'.'+ instance
	
	v = getvar( node, community, shorttries, oid, port=snmpport, version=snmpversion)
	if (v == None):	
		raise UndefinedSnmpValueException("Current variable %s can't be read or is not present (GET) on %s" % (oid, node) )		
		
	log.debug("FOUND   : %s", repr(v))
	
	v = binascii.hexlify(v).upper()
		
	if v == ap_macaddress.upper():
		return instance

	raise Exception("The fetched binary value (%s) doesn't match AP mac address (%s)", v, ap_macaddress)

#---------------------------------------------------------------------#
#                                                                     #
#---------------------------------------------------------------------#

def fun_airespaceAPChannelInterference(context, threshold, selected_antenna=0):


	log         = context.getLog()

	host        = context.get('$cluster_node')
	ap_id       = context.get('$cluster_instance')

        comm        = context.get('$community'  , None)
        snmpport    = context.get('$snmpport'   , None)
        shorttries  = context.get('$shorttries' , None)
        snmpversion = context.get('$snmpversion', None)

	#cluster_nodes = context['$cluster_nodes']

	node              = context.get('__node__')
	configuration     = context.getConfiguration()
	all_cluster_nodes = configuration.get_cluster_elements( node.cluster_id )	
	cluster_nodes = [ e for e in all_cluster_nodes if e.cluster_role == CLUSTER_ROLE_MASTER  ]




	#checkmanager = context.getCheckManager()
	#sanet_nodes  = checkmanager.node_cache
	#nodes_by_name = {}
	#for n in  sanet_nodes.values():
	#	nodes_by_name[ n.name ] = n



        bsnAPIfRxNeighborsEntry = '1.3.6.1.4.1.14179.2.2.17.1'
        bsnAPIfPhyChannelNumber = '1.3.6.1.4.1.14179.2.2.2.1.4'

	log.debug("Getting neighbours entries")

	data_by_ap = {}

	base_oid = bsnAPIfRxNeighborsEntry + '.3.' + ap_id + '.' + str(selected_antenna)

        table = gettable(host, comm, shorttries, base_oid, snmpport, snmpversion)
        if table == None:
                raise UndefinedSnmpValueException("bsnAPIfRxNeighborsEntry (%s) for antenna %s not found!? Verify your AP status!" % (base_oid, selected_antenna))

	log.debug("Read %s entries", len(table))
        for (row_oid, row_val) in table:

		# skip unwanted entries
		if row_val < threshold:
			continue

		ap2_id = row_oid[ 1 + len(base_oid) + 1 : ]

		data_by_ap[ap2_id] = row_val

		log.debug("Found on antenna %s neighbour %s = %s dbm >= %s dbm", selected_antenna, __dotted_to_mac_instance(ap2_id), row_val, threshold)

	if not data_by_ap:
		context.appendReason("No neighbour AP selected")
		log.debug("No neighbour AP selected")
		return False


	log.debug("Getting node antennas' channels")

	channel = getvar(host, comm, shorttries, bsnAPIfPhyChannelNumber + '.' + ap_id + '.' + str(selected_antenna) , snmpport, snmpversion)


	result = ''
	ret = False
		
	for ap2_id, dbm in sorted(data_by_ap.items()):

		log.debug("Reading IP and channel for AP id %s", ap2_id)

		ap2_hostname = ''
		ap2_ip       = ''
		ap2_channel  = None
		ap2_wlc      = None

		#for wlc in cluster_nodes.split(','):
		for wlc in cluster_nodes:

			try:
				log.debug("checking out association with wlc %s", wlc.name )
				#sanet_node = nodes_by_name.get( wlc , None)
				#if sanet_node == None:
				#	log.debug("Unable to find %s in SANET configuration", wlc)
				#	continue

				sanet_node = wlc
				wlc_host = sanet_node.name  # FIX: use right name/address with ipv4,ipv6


				oid = '1.3.6.1.4.1.14179.2.2.1.1.19.' + ap2_id
				ap2_ip   = getvar(wlc_host, sanet_node.snmp_community, shorttries, oid , sanet_node.snmp_port, sanet_node.snmp_version)

				oid = bsnAPIfPhyChannelNumber + '.' + ap2_id + '.' + str(selected_antenna)
				ap2_channel = getvar(wlc_host, sanet_node.snmp_community, shorttries, oid , sanet_node.snmp_port, sanet_node.snmp_version)

				if ap2_channel != None:
					log.debug("\t" + "Found associated WLC %s", wlc.name)
					ap2_wlc = wlc
					break
					
			except Exception, e:
				log.debug("Error while retrieving data from %s: %s", wlc.name, str(e))
				pass

		if ap2_wlc == None:
			log.debug("Unable to find association for neighbor AP %s", ap2_id)
			continue

		if ap2_channel != channel:
			log.debug("antenna %s channel %s != AP2 antenna %s channel %s. Skip",  selected_antenna, channel,  selected_antenna, ap2_channel)
			continue

		ret = True
		
		try:
			log.debug("Reversing neighbor AP %s", ap2_ip)
			names = socket.gethostbyaddr(ap2_ip)
			ap2_hostname = names[0]
		except socket.herror, e:
			log.debug("Unable to find reverse name for %s", ap2_ip)
	
		if ap2_hostname:
			ap2_name_ip = '%s (%s)' % (ap2_hostname, ap2_ip)
		elif ap2_ip:
			ap2_name_ip = '%s' % (ap2_ip)
		else:
			ap2_name_ip = 'N/A'
		
		s = " antenna %s (channel %s) -> AP %s (%s) antenna %s channel %s =  %s dbm (on WLC=%s)" % ( 
			selected_antenna, 
			channel, 
			ap2_name_ip, 
			__dotted_to_mac_instance(ap2_id), 
			selected_antenna, 
			ap2_channel,  
			dbm, 
			ap2_wlc.name
		)

		log.debug(s)

		result += s + '\n'

	context.appendReason(result)

        return ret


#---------------------------------------------------------------------#
#                                                                     #
#---------------------------------------------------------------------#


EXPR_SYMBOLS = {
	'calculateAirespaceAccessPointInstance': fun_calculateAirespaceAccessPointInstance,
	'airespaceAPChannelInterference'       : fun_airespaceAPChannelInterference,
}


