# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>. 

"""
===============
Basic functions
===============

hex
---------------------------------------------

    **Syntax**

      **hex** **(** *byte string* **)**

    **Description**

      Converts the input parameter *bytes* into a UPPERCASE hexadecimal string.

    **Input**

      *bytes*: a sequence of bytes.

    **Output**

      Returns a *string*.

    **Errors**

      If the input parameter can not be converted to a valid string the expression evaluation fails.

    **Examples**

      +------------------------------------+------------------------------------------------------------------------+---------------------+
      | Expression                         | Description                                                            | Return string value |
      +====================================+========================================================================+=====================+
      | hex(1.3.6.1.2.1.2.2.1.6.2@"mason") | Converts into a string the MAC address of the interface with ifindex 2 | *01279AF7A4D*       |
      +------------------------------------+------------------------------------------------------------------------+---------------------+

abs
---------------------------------------------

    **Syntax**

      abs **(** *arg* **)**

    **Description**

      Returns the absolute value of the given argument.

    **Input**

      *args*: number

    **Output**

      Returns a *number*

    **Errors**

      If the input parameter can not be converted to a valid number the expression evaluation fails.

    **Examples**

      +------------+--------+
      | Expression | Value  |
      +============+========+
      | abs(1)     | *1*    |
      +------------+--------+
      | abs(-1)    | *1*    |
      +------------+--------+
      
      
min
---------------------------------------------

    **Syntax**

      min **(** *arg* **)**

    **Description**

      Returns the minimum value.

    **Input**

      *args*: sequence of numeric values.

    **Output**

      Returns a *number*

    **Errors**

      If the input parameter can not be converted to a valid number the expression evaluation fails.

    **Examples**

      +------------+--------+
      | Expression | Value  |
      +============+========+
      | min(1,2,3) | *1*    |
      +------------+--------+
      | min(-1,0)  | *-1*   |
      +------------+--------+      
      
      
max
---------------------------------------------

    **Syntax**

      max **(** *arg* **)**

    **Description**

      Returns the maximum value.

    **Input**

      *args*: sequence of numeric values.

    **Output**

      Returns a *number*

    **Errors**

      If the input parameter can not be converted to a valid number the expression evaluation fails.

    **Examples**

      +------------+--------+
      | Expression | Value  |
      +============+========+
      | max(1,2,3) | *3*    |
      +------------+--------+
      | max(-1,0)  | *0*    |
      +------------+--------+
      
sum
---------------------------------------------

    **Syntax**

      sum **(** *arg* **)**

    **Description**

      Sums all the input arguments. 

    **Input**

      *args*: sequence of numeric values.

    **Output**

      Returns a *number*

    **Errors**

      If the input parameter can not be converted to a valid number the expression evaluation fails.

    **Examples**

      +------------+--------+
      | Expression | Value  |
      +============+========+
      | sum(1,2,3) | *6*    |
      +------------+--------+


ceil
---------------------------------------------

    **Syntax**

      ceil **(** *v* **)**

    **Description**

      Return the ceiling of x as a float, the smallest integer value greater than or equal to x.

    **Input**

      *v*: a float value.

    **Output**

      Returns a *float* value.

    **Errors**

      If the input parameter can not be converted to a valid number the expression evaluation fails.

    **Examples**

      +-----------------------------------------------------+-----------------------------------------------------+
      | Expression                                          | Value                                               |
      +=====================================================+=====================================================+
      | ceil(0.1)                                           | *1.0*                                               |
      +-----------------------------------------------------+-----------------------------------------------------+
      | ceil(1.9)                                           | *2.0*                                               |
      +-----------------------------------------------------+-----------------------------------------------------+
      | ceil(-1.1)                                          | *-1.0*                                              |
      +-----------------------------------------------------+-----------------------------------------------------+

floor
---------------------------------------------

    **Syntax**

      floor **(** *v* **)**

    **Description**

      Return the floor of x as a float, the largest integer value less than or equal to x.

    **Input**

      *v*: a float value.

    **Output**

      Returns a *float* value.

    **Errors**

      If the input parameter can not be converted to a valid number the expression evaluation fails.

    **Examples**

      +-----------------------------------------------------+-----------------------------------------------------+
      | Expression                                          | Value                                               |
      +=====================================================+=====================================================+
      | floor(0.1)                                          | *0.0*                                               |
      +-----------------------------------------------------+-----------------------------------------------------+
      | floor(1.9)                                          | *1.0*                                               |
      +-----------------------------------------------------+-----------------------------------------------------+
      | floor(-1.1)                                         | *-2.0*                                              |
      +-----------------------------------------------------+-----------------------------------------------------+


str
---------------------------------------------

    **Syntax**

      str **(** *arg* **)**

    **Description**

      Tries to converts the input value *arg* to a string. 

    **Input**

      *args*: any value

    **Output**

      Returns a *string*

    **Errors**

      If the input parameter can not be converted to a valid string the expression evaluation fails.

    **Examples**

      +---------------+----------+
      | Expression    | Value    |
      +===============+==========+
      | str(0xFF)     | *255*    |
      +---------------+----------+
      | str(1.0)      | *1.0*    |
      +---------------+----------+
      | str("foo")    | *foo*    |
      +---------------+----------+


float
---------------------------------------------

    **Syntax**

      **float** **(** *arg* **)**

    **Description**

      Tries to converts the input value *arg* to an floating point value. 

    **Input**

      *args*: any value

    **Output**

      Returns a floating point value.

    **Errors**

      If the input parameter can not be converted to a valid floating point value the expression evaluation fails.

    **Examples**

      +----------------+-------------+
      | Expression     | Value       |
      +================+=============+
      | int(0xff)      | *255.0*     |
      +----------------+-------------+
      | int(1.33)      | *1.33*      |
      +----------------+-------------+
      | int("12345")   | *12345.0*   |
      +----------------+-------------+
      | int("foo")     | *ERROR*     |
      +----------------+-------------+

int
---------------------------------------------

    **Syntax**

      **int** **(** *arg* **)**

    **Description**

      Tries to converts the input value *arg* to an integer value. 

    **Input**

      *args*: any value

    **Output**

      Returns a *string*

    **Errors**

      If the input parameter can not be converted to a valid integer value the expression evaluation fails.

    **Examples**

      +----------------+------------+
      | Expression     | Value      |
      +================+============+
      | int(0xFF)      | *255*      |
      +----------------+------------+
      | int(1.33)      | *1*        |
      +----------------+------------+
      | int("12345")   | *12345*    |
      +----------------+------------+
      | int("foo")     | *ERROR*    |
      +----------------+------------+


.. _function-dict:

dict
---------------------------------------------

    **Syntax**

      **dict** **(** *key1* , *value1* [ , *key2* , *value2* ... ] **)**

    **Description**

      Takes a sequence of parameters and created a dictionary object in memory.      . 

    **Input**

      *args*: a sequence of paired *key* / *value* options.

    **Output**

      Returns a *dictionary*

    **Errors**

      Raise an exceptions if the number of paramets is not even.     

    **Examples**

      +---------------------------------------------------------------------------------+--------------------------------------------------------+
      | Expression                                                                      | Value                                                  |
      +=================================================================================+========================================================+
      | dict("key1", 666)                                                               | { "key1: 666 }                                         |
      +---------------------------------------------------------------------------------+--------------------------------------------------------+
      | dict("key1", 666, "foo", "bar")                                                 | { "foo": "bar", "key1":666 }                           |
      +---------------------------------------------------------------------------------+--------------------------------------------------------+

re_escape
---------------------------------------------

    **Syntax**

      **re_escape** **(** *str* **)**

    **Description**

      Returns a string with all non-alphanumerics backslashed.

      This is useful if you want to compose a regular expression from a string that contains regular expression meta-characters (**.**, *****, ... ).

    **Input**

      *str* : string

    **Output**

      Returns a *string*.

    **Examples**

      +----------------------------------------------------------------------------------------------------------+----------------------+------------------------------------------------------------------------------------------+
      |  Expression                                                                                              | Value                | Description                                                                              |
      +==========================================================================================================+======================+==========================================================================================+
      | re_escape("hello")                                                                                       | *hello*              | Nothing to escape                                                                        |
      +----------------------------------------------------------------------------------------------------------+----------------------+------------------------------------------------------------------------------------------+
      | re_escape("192.168.0.1")                                                                                 | *192\\.168\\.0\\.1*     | dots escaped                                                                             |
      +----------------------------------------------------------------------------------------------------------+----------------------+------------------------------------------------------------------------------------------+
      | "192.168.0.1.port1.255.255.255.0"  =~  re_escape("192.168.0.1.") + ".*" + re_escape(".255.255.255.0")    | *true*               | Try to match the pattern:  *ip address* **.** *any character* **.** *ip address*         |
      +----------------------------------------------------------------------------------------------------------+----------------------+------------------------------------------------------------------------------------------+
      | "192.168.0.1.abcdefg.255.255.255.0"  =~  re_escape("192.168.0.1.") + "\\d" + re_escape(".255.255.255.0")  | *false*              | Try to match the pattern:  *ip address* **.** *integer value* **.** *ip address*         |
      +----------------------------------------------------------------------------------------------------------+----------------------+------------------------------------------------------------------------------------------+
  

greplines
---------------------------------------------


    **Syntax**

      **greplines** **(** *multiline text* , *substring* ] **)**

    **Description**

      Searches for the given *substring* inside *multiline text* and returns the first line where substring was found.
     
      If *substring* is not found, the function raises an exception and the expression goes UNCHECKABLE.

    **Input**

      * *multine text*  a (multiline) string 
      * *substring* string
     
    **Output**

      A string

    **Examples**

      +-------------------------------------------+------------------------------------------------------------------------+---------------------+
      | Expression                                | Description                                                            | Return string value |
      +===========================================+========================================================================+=====================+
      | greplines('hello \\\\n sunny word', 'word') |                                                                        |                     |
      +-------------------------------------------+------------------------------------------------------------------------+---------------------+


valueinrange
---------------------------------------------

    **Syntax**

      **valueinrage** **(** *value* **,** *minvalue* **,** *maxvalue* **)**

    **Description**

      Returns *true* if *value* is between *minvalue* and *maxvalue* (included).
      
      Raises an exception otherwise.
      
    **Input**

      * *value* Number
      * *minvalue* Number
      * *maxvalue* Number

    **Output**

      * **boolean** True 

    **Examples**

      +------------------------------------+------------------------------------------------------------------------+---------------------+
      | Expression                         | Description                                                            | Return value        |
      +====================================+========================================================================+=====================+
      | valueinragen(2,1,3)                |                                                                        |                     |
      +------------------------------------+------------------------------------------------------------------------+---------------------+


isDefined()
---------------------------------------------

    **Syntax**

      **isDefined** **(** *varname* **)**

    **Description**

      Returns *true* if a symbol is defined while evaluating the current expression.
      
    **Special symbols**

      Some special symbols are calculated "on-the-fly". The computation could raise an execution error for different reason (network timetout, internal errors, ecc.) . 
      
      The *isDefined()* catches all exception and always return False. 
      
      Some special exception are not handled by the function and lead to an uncheckable result. These are the special cases:
      
      * Exceptions raised by well-known bugs of the SNMP library


    **Input**

      * *varname* : symbol name.

    **Output**

      * **boolean** True if the symbold is defined. False otherwise.

    **Examples**

      +------------------------------------+------------------------------------------------------------------------+---------------------+
      | Expression                         | Description                                                            | Return value        |
      +====================================+========================================================================+=====================+
      | isDefined("ifindex")               | Tells if  $ifindex can be calculated and used inside the espression.   | *true*              |
      +------------------------------------+------------------------------------------------------------------------+---------------------+
      


ifDefined()
---------------------------------------------

    **Syntax**

      **isDefined** **(** *varname* **,** *default_value* **)**

    **Description**

      Check if the symbol *varname* is defined in the current execution context and returns its value.
      
      If the symbol is not defined, the *default_value* is returned.
      
    **Input**

      * *varname* : (string) symbol name.
      * *default_value*: (any type) any value.

    **Output**

      * Return a valid expression value.

    **Examples**

      +------------------------------------+------------------------------------------------------------------------+-----------------------------+
      | Expression                         | Description                                                            | Return value                |
      +====================================+========================================================================+=============================+
      | ifDefined("$ifindex", -1)          | Returns the value of the simbol $ifindex                               | 12                          |
      +------------------------------------+------------------------------------------------------------------------+-----------------------------+
      | ifDefined("$ifindex", -1)          | Returns the value of the simbol $ifindex                               | -1 ($ifindex is not defined |
      +------------------------------------+------------------------------------------------------------------------+-----------------------------+
      
      
      
re_search
---------------------------------------------

    **Syntax**

      **re_search** **(** *string* **,** *regexp* **)**

    **Description**

      Tries to search in *string* the matching sub-string specified by *regexp*.
      If *regexp* uses named groups, the function returns a dictionary with the matching groups.
      If *regexp* uses groups, the function returns a vectory with the matched groups.
      If *regexp* do not use groups, the matching string is returned.

    **Input**

      * *string*  : a string of characters
      * *regexp*  : a regular expression string 

    **Output**

      * A *dictionary* or a *vector* of values.

    **Examples**

      +------------------------------------------------------------------------------------------------+------------------------------------------------------------------------+--------------------------------------+
      | Expression                                                                                     | Description                                                            | Return value                         |
      +================================================================================================+========================================================================+======================================+
      | re_search( "hello world",  "hello" )                                                           |                                                                        | hello                                |
      +------------------------------------------------------------------------------------------------+------------------------------------------------------------------------+--------------------------------------+      
      | re_search( "hello world",  "(lo)" )                                                            |                                                                        | [ 'lo' ]                             |                 
      +------------------------------------------------------------------------------------------------+------------------------------------------------------------------------+--------------------------------------+      
      | re_search( "hello world",  r"(\w+) (\w+)" )                                                    |                                                                        | [ 'hello', 'world' ]                 |
      +------------------------------------------------------------------------------------------------+------------------------------------------------------------------------+--------------------------------------+      
      | re_search( "hello world",  r"(\w+) (\w+)" )[0]                                                 |                                                                        | 'hello'                              |
      +------------------------------------------------------------------------------------------------+------------------------------------------------------------------------+--------------------------------------+      
      | re_search( "John Doe",  r"(?P<name>\w+) (?P<surname>\w+)" )                                    |                                                                        | {'surname': 'Doe', 'name': 'John'}   |
      +------------------------------------------------------------------------------------------------+------------------------------------------------------------------------+--------------------------------------+      
      | re_search( "John Doe",  r"(?P<name>\w+) (?P<surname>\w+)" )['surname']                         |                                                                        | 'Doe'                                |
      +------------------------------------------------------------------------------------------------+------------------------------------------------------------------------+--------------------------------------+      

hexstr2bin
---------------------------------------------

    **Syntax**

      **hexstr2bin** **(** *string* **)**

    **Description**

      Converts the *string* of hexadecimal characters in the corresponding binary rappresentation.

    **Input**

      *string*: a string of hexadecimal ascii characters.

    **Output**

      Returns a binary buffer.

    **Errors**

      Raise an exception when a conversion error is detected.

    **Examples**

      +----------------------------------------------------------------------+--------------------------------------------------+--------------------------------------------------+
      | Expression                                                           | Value                                            | Description                                      |
      +======================================================================+==================================================+==================================================+
      | hexstr2bin('68656c6c6f')                                             | \x68\x65\x6c\x6c\x6f                                            | ascii string for *hello*                         |
      +----------------------------------------------------------------------+--------------------------------------------------+--------------------------------------------------+


ziparray
---------------------------------------------

    **Syntax**

      **ziparray** **(** *array1* **,** *array2* [ **,** *...* ] *)**

    **Description**

      Returns a list of tuples, where the i-th tuple contains the i-th element from each of the arrays.

    **Input**

      Arrays of elements.

    **Output**

      Returns an array.

    **Errors**

      Raise an exception when a conversion error is detected.

    **Examples**

      +----------------------------------------------------------------------+--------------------------------------------------+--------------------------------------------------+
      | Expression                                                           | Value                                            | Description                                      |
      +======================================================================+==================================================+==================================================+
      | ziparray( array(1,2,3), array("a","b","c") )                         | [ (1,'a'), (2,'b'), (3,'c') ]                    | ascii string for *hello*                         |
      +----------------------------------------------------------------------+--------------------------------------------------+--------------------------------------------------+


"""

import time
import os
import re
import array
import binascii
import math

from sanet_common.exprparser.Types import SnmpValueType

from sanet_poller.lib.exceptions import PollException, ForceUncheckableException

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_hex(context, arg):
	"""
	Return a string containing a nicely printable representation of an object.
	
	if **arg** is a :class:`SnmpValueType` object, the SNMP value is
	retrieved and then converted into a string.
	"""
	
	s = ''
	
	if isinstance(arg, SnmpValueType):
		arg = arg.getvalue()
		
	# FIX: do it better (and faster)!
	arg = array.array('B', arg)
	for a in arg:
		s += "%X" % a
		
	return s

#-----------------------------------------------------------------------------#

def fun_str(context, val):
	"""
	Wrapper for the Python's built-in function **str()**
	"""
	
	if isinstance(val, SnmpValueType):
		val = val.getvalue()
			
	return str(val)

#-----------------------------------------------------------------------------#

def fun_repr(context, val):
	"""
	Wrapper for the Python's built-in function **str()**
	"""
	
	if isinstance(val, SnmpValueType):
		val = val.getvalue()
			
	return repr(val)

#-----------------------------------------------------------------------------#

def fun_bool(context, val):
	"""
	Wrapper for the Python's built-in function **bool()**
	"""
	if isinstance(val, SnmpValueType):
		val = val.getvalue()
	
	try:
		return bool(val)
	except Exception, e:
		raise PollException("Unable to convert '%s' (%s) to boolean: %s" % (val, type(val), e))

#-----------------------------------------------------------------------------#

def fun_int(context, arg, base=None):
	"""
	Wrapper for the Python's built-in function **int()**
	"""

	if isinstance(arg, SnmpValueType):
		arg = arg.getvalue()

	try:
		if base is not None:
			return int(arg, base)
		else:
			return int(arg)
	except Exception, e:
		raise PollException("Unable to convert '%s' (%s) to integer value: %s" % (arg, type(arg), e))

#-----------------------------------------------------------------------------#

def fun_float(context, arg):
	"""
	Wrapper for the Python's built-in function **float()**
	"""
	if isinstance(arg, SnmpValueType):
		arg = arg.getvalue()

	try:
		return float(arg)	
	except Exception, e:
		raise PollException("Unable to convert '%s' (%s) to float value: %s" % (arg, type(arg), e))
	
#-----------------------------------------------------------------------------#

def fun_dict(context, arg):
	"""
	Takes a sequence of parameters and created a dictionary object in memory.
	
	Example:
	
	    dict("key1", "value1", "key2", 666) 
	"""
	if isinstance(arg, SnmpValueType):
		arg = arg.getvalue()

	try:
		return float(arg)	
	except Exception, e:
		raise PollException("Unable to convert '%s' (%s) to float value: %s" % (arg, type(arg), e))
	

#-----------------------------------------------------------------------------#

def fun_abs(context, v):
	return abs(v)
	
def fun_max(context, *args):
	return max(*args)
	
def fun_min(context, *args):
	return min(*args)

def fun_sum(context, *args):
	return sum(*args)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_ceil(context, v):

	if isinstance(v, SnmpValueType):
		v = v.getvalue()

	return math.ceil(v)

def fun_floor(context, v):

	if isinstance(v, SnmpValueType):
		v = v.getvalue()

	return math.floor(v)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_greplines(context, text, sub):
	
	if isinstance(text, SnmpValueType):
		text = text.getvalue()

	if isinstance(sub, SnmpValueType):
		sub = sub.getvalue()
	
	for l in text.split("\n"):
		if l.find(sub) > -1:
			return l
			
	raise Exception("'%s' not found" % (sub))

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_re_escape(context, arg):
	return re.escape(arg)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_valueinrange(context, value, minvalue, maxvalue):
	
	if minvalue <= value and value <= maxvalue:
		return value
		
	raise PollException("The value %s not in range (%s,%s)" % (value, minvalue, maxvalue))	

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#


def fun_isDefined(context, varname):
	
	r = False
	
	try:
		varvalue = context[varname]
		r = True
		info = '%s has value %s' % (varname, varvalue)
	
	# Special case
	#         HISTORY: Added for ticket #1885.
	except ForceUncheckableException, e:
		raise PollException('Unable to get value for %s and forcing uncheckable state. Reason: %s' % (varname, str(e)))
	
	except KeyError, e:
		info = 'Unable to get value for %s: not defined %s' % (varname, str(e))		
		
	except PollException, e:
		info = 'Unable to get value for %s: %s' % (varname, str(e))
		
	context.appendReason(info)
	
	return r


def fun_ifDefined(context, varname, else_value):

	try:
		varvalue = context[varname]
		return varvalue

	# Special case
	#         HISTORY: Added for ticket #1885.
	except ForceUncheckableException, e:
		raise PollException('Unable to check value for %s and forcing uncheckable state. Reason: %s' % (varname, str(e)))

	except KeyError, e:
		info = 'Unable to get value for %s: not defined %s' % (varname, str(e))		

	except PollException, e:
		info = 'Unable to get value for %s: %s' % (varname, str(e))

	context.appendReason(info)

	return else_value

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_re_search(context, string, regexp):
	
	m = re.search(regexp, string)
	if not m:
		raise PollException("match not found")
		
	d = m.groupdict()
	if d:
		return d

	v = m.groups()
	if v:
		return v
		
	return m.group()

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_dict(context, *args):
	d = {}
	for i in xrange(len(args)/2):
		k = args[(i*2)]
		v = args[(i*2)+1]
		d[k] = v
	return d

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_hexstr2bin(context, s):
	return binascii.unhexlify(s)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_ziparray(context, *arrays):
	return zip(*arrays)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

EXPR_SYMBOLS = {
	'hex'                   : fun_hex,
	'str'                   : fun_str,
	'repr'                  : fun_repr,
	'bool'                  : fun_bool,
	
	'int'                   : fun_int,
	'number'                : fun_float,
	'float'                 : fun_float,
	
	'abs'                   : fun_abs,
	'max'                   : fun_max,
	'min'                   : fun_min,
	'sum'                   : fun_sum,
	
	'ceil'                  : fun_ceil,
	'floor'                 : fun_floor,

	're_escape'             : fun_re_escape,
	
	'greplines'             : fun_greplines,
	
	'valueinrange'          : fun_valueinrange,
	
	'isDefined'             : fun_isDefined,
	'ifDefined'             : fun_ifDefined,
	
	're_search'             : fun_re_search,

	'dict'                  : fun_dict,

	'hexstr2bin'            : fun_hexstr2bin,

	'ziparray'              : fun_ziparray,
}
