# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it> 
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""
###############
Cisco utils
###############

ciscoNoErrDisabled
=======================================

    **Syntax**
    
        *ciscoNoErrDisabled* **(** [ mib_check_oid ] **)**
    
    **Description**
    
        The function checks the *cErrDisableIfStatusTable* (1.3.6.1.4.1.9.9.548.1.3.1) in the *CISCO-ERR-DISABLE-MIB*, and returns 1 (true) if the table is empty (i.e. there are no errdisabled ports). 
        
        If there are errdisabled ports the function returns 0 (false) and writes in the context a description of the relevant interfaces and their errdisable reason. 
        
        The function verifies that the MIB is supported by fetching the specified OID *mib_check_oid*. The default OID is *cErrDisableRecoveryInterval* (1.3.6.1.4.1.9.9.548.1.1.1).
        
        If the OID *mib_check_oid* can not be fetched, the function raises an exception and the expression evaluation goes *UNCHECKABLE*.
    
    **Input**

       * mib_check_oid: (optional) a string containing a valido OID.

    **Output**

       A booean (true or false)

    **Errors**

       If an error occurs or the OID *mib_check_oid* is not defined (default OID: cErrDisableRecoveryInterval).

    **Examples**

       +-------------------------------------------------------------------------+---------------------+------------------------------------------------------------------------+
       | Expression                                                              | Return value        | Description                                                            |
       +=========================================================================+=====================+========================================================================+
       | ciscoNoErrDisabled()                                                    | 1                   |                                                                        |
       +-------------------------------------------------------------------------+---------------------+------------------------------------------------------------------------+
       | ciscoNoErrDisabled("1.3.6.1.4.1.9.9.548.1.1.3")                         | *UNCHECKABLE*       | The host does not define OID 1.3.6.1.4.1.9.9.548.1.1.3.                |
       +-------------------------------------------------------------------------+---------------------+------------------------------------------------------------------------+

ciscoSensorThresholdsOk
=========================

    **Syntax**

      *ciscoSensorThresholdsOk* **(** [ *ignore_when_notification_disabled* ] [ , *ignore_names_re* ]  [ , ignore_instances ] **)**

    **Description**

      TODO

    **Input**

      * TODO

    **Output**

      TODO

    **Errors**

      TODO

    **Examples**

      +------------------------------------------------------+------------------------------------------------------------------------+---------------------+
      | Expression                                           | Description                                                            | Return string value |
      +======================================================+========================================================================+=====================+
      | iscoSensorThresholdsOk()                             |                                                                        | *01279AF7A4D*       |
      +------------------------------------------------------+------------------------------------------------------------------------+---------------------+
      | iscoSensorThresholdsOk(ignore_instances=1)           |                                                                        | *01279AF7A4D*       |
      +------------------------------------------------------+------------------------------------------------------------------------+---------------------+

"""

from sanet_poller.lib.exceptions import PollException

from sanet_poller.lib.snmp.base import getvar, gettable

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

	
def fun_ciscoSensorThresholdsOk(context, ignore_when_notification_disabled=True, ignore_names_re='', ignore_instances=''):
	# Check CISCO-ENTITY-SENSOR-MIB to see if at leas a sensor is out of its threshold
	# See the description of 1.3.6.1.4.1.9.9.91.1.2.1.1.3 for the algorithm used,
	# e.g. http://tools.cisco.com/Support/SNMP/do/BrowseOID.do?local=en&translate=Translate&objectInput=1.3.6.1.4.1.9.9.91.1.2.1.1.3#oidContent

	# Return true or false, and appends a description of the thresholds violated

	log = context.getLog()
	
	host       = context.get('$node', None)
	community  = context.get('$community', None)
	sv         = context.get('$snmpversion', None)
	port       = context.get('$snmpport', None)
	shorttries = context.get('$shorttries', None)
	
	table_oid = '1.3.6.1.4.1.9.9.91.1.2.1.1'
	sensor_threshold_oid = table_oid + '.' + '4'
	sensor_status_oid = '1.3.6.1.4.1.9.9.91.1.1.1.1.5'
	sensor_value_oid = '1.3.6.1.4.1.9.9.91.1.1.1.1.4'
	sensor_sysuptime_oid = '1.3.6.1.4.1.9.9.91.1.1.1.1.6'
	sensor_relation_oid = table_oid + '.' + '3'
	sensor_notif_oid = table_oid + '.' + '6'
	
	if ignore_names_re:
		ignore_names_re_compiled = re.compile(ignore_names_re)
		
	if ignore_instances:
		ignore_instances_list = ignore_instances.split(' ')	
		
	sensors = 0
	notok = 0
	nonotify = 0
	inthreshold = 0
	violated = 0

	rv = True

	thr = gettable(host, community, shorttries, sensor_threshold_oid, port=port, version=sv) # threshold
	if not thr:
		reason='ciscoSensorThresholdsOk: cannot read threshold values'
		context.appendReason(reason)
		raise PollException(reason)
		
	for oid,sensor_threshold in thr:
		
		if oid.startswith('.'):
			oid = oid[1:]
			
		inst = oid[len(sensor_threshold_oid) + 1:]
		
		if ignore_instances and inst in ignore_instances_list:
			log.debug('Skipping sensor with instance %s' % (inst))
			continue

		sensor_inst = inst.split('.')[0]
		sensor_threshold = int(sensor_threshold)
		
		sensors += 1
		sensor_status = getvar(host, community, shorttries, sensor_status_oid + '.' + sensor_inst, port, sv)
		
				
		descr = ''
		sensor_descr = getvar(host, community, shorttries, '1.3.6.1.2.1.47.1.1.1.1.2' + '.' + sensor_inst, port, sv)
		if sensor_descr is None: sensor_descr = ''

		sensor_name = getvar(host, community, shorttries, '1.3.6.1.2.1.47.1.1.1.1.7' + '.' + sensor_inst, port, sv)
		if sensor_name is None: sensor_name = ''

		if sensor_name == sensor_descr:
			descr = sensor_descr
		else:
			descr = sensor_name + ', ' + sensor_descr

		if ignore_names_re and ignore_names_re_compiled.match(sensor_name):
			log.debug('Skipping sensor by name %s %s' % (inst, sensor_name))
			continue
		
		if sensor_status is None:
			reason='Cannot read status for sensor %s %s' % (inst, sensor_name)
			context.appendReason(reason)
			raise PollException(reason)
			
		sensor_status = int(sensor_status)
		if sensor_status != 1:
			log.debug('sensor %s %s is not ok (%d)' % (inst, sensor_name, sensor_status))
			notok += 1
			continue
			
		# Ignore violations occurred at sysuptime 0, since this appears to be some obscure way to say that it's disabled
		sensor_sysuptime = getvar(host, community, shorttries, sensor_sysuptime_oid + '.' + sensor_inst, port, sv)
		if sensor_sysuptime is None:
			reason='Cannot read sysuptime of measurement for sensor %s %s' % (sensor_inst, sensor_name)
			context.appendReason(reason)
			raise PollException(reason)
			
		sensor_sysuptime = int(sensor_sysuptime)
		if sensor_sysuptime == 0:
			log.debug('sensor %s %s has measurement at sysuptime 0' % (inst, sensor_name))
			notok += 1
			continue
			
		sensor_value = getvar(host, community, shorttries, sensor_value_oid + '.' + sensor_inst, port, sv)
		if sensor_value is None:
			reason='Cannot read value for sensor %s %s' % (inst, sensor_name)
			context.appendReason(reason)
			raise PollException(reason)
			
		sensor_value = int(sensor_value)
		sensor_relation = getvar(host, community, shorttries, sensor_relation_oid + '.' + inst, port, sv)
		if sensor_relation is None:
			reason='Cannot read relation for sensor %s %s' % (inst, sensor_name)
			context.appendReason(reason)
			raise PollException(reason)
			
		sensor_relation = int(sensor_relation)
		
		#log.debug('sensor %s, name %r, value %d, threshold %d, relation %d' % (inst, sensor_name, sensor_value, sensor_threshold, sensor_relation))
		if sensor_relation == 1:
			alarm = sensor_value < sensor_threshold
		elif sensor_relation == 2:
			alarm = sensor_value <= sensor_threshold
		elif sensor_relation == 3:
			alarm = sensor_value > sensor_threshold
		elif sensor_relation == 4:
			alarm = sensor_value >= sensor_threshold
		elif sensor_relation == 5:
			alarm = sensor_value == sensor_threshold
		elif sensor_relation == 6:
			alarm = sensor_value != sensor_threshold
		else:
			context.appendReason('Unknown sensor relation for %s %s: %d' % (inst, sensor_name, sensor_relation))
			raise PollException(reason)
			
		if not alarm:
			inthreshold += 1
			continue
			
		violated += 1
		
		if ignore_when_notification_disabled:
		
			sensor_notif = getvar(host, community, shorttries, sensor_notif_oid + '.' + inst, port, sv)
			if sensor_notif is None:
				reason='Cannot read notification enable for sensor %s %s' % (inst, descr)
				context.appendReason(reason)
				raise PollException(reason)
				
			sensor_notif = (int(sensor_notif) == 1) # 1 true 2 false
			if not sensor_notif:
				log.debug('Ignoring because notification not enabled: sensor %s, descr %s, value %d, threshold %d, relation %d' % (inst, descr, sensor_value, sensor_threshold, sensor_relation))
				nonotify += 1
				continue
		rv = False
		# Get some more info to make reason more readable
		context.appendReason('Threshold violated: instance %s, sensor %s, descr %s, value %d, threshold %d, relation %d' % (inst, sensor_inst, descr, sensor_value, sensor_threshold, sensor_relation))
	context.appendReason('Sensors thresholds checked: %d, sensor not ok: %d, in threshold: %d, threshold violated: %d, notification disabled: %d' % (sensors, notok, inthreshold, violated, nonotify))
	return rv

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

cisco_errdisable_reason_text = {
	1: 'udld',
	2: 'bpduGuard',
	3: 'channelMisconfig',
	4: 'pagpFlap',
	5: 'dtpFlap',
	6: 'linkFlap',
	7: 'l2ptGuard',
	8: 'dot1xSecurityViolation',
	9: 'portSecurityViolation',
	10: 'gbicInvalid',
	11: 'dhcpRateLimit',
	12: 'unicastFlood',
	13: 'vmps',
	14: 'stormControl',
	15: 'inlinePower',
	16: 'arpInspection',
	17: 'portLoopback',
	18: 'packetBuffer',
	19: 'macLimit',
	20: 'linkMonitorFailure',
	21: 'oamRemoteFailure',
	22: 'dot1adIncompEtype',
	23: 'dot1adIncompTunnel',
	24: 'sfpConfigMismatch',
	25: 'communityLimit',
	26: 'invalidPolicy',
	27: 'lsGroup',
	28: 'ekey',
	29: 'portModeFailure',
	30: 'pppoeIaRateLimit',
	31: 'oamRemoteCriticalEvent',
	32: 'oamRemoteDyingGasp',
	33: 'oamRemoteLinkFault',
	34: 'mvrp',
	35: 'tranceiverIncomp',
}


cErrDisableRecoveryInterval = '1.3.6.1.4.1.9.9.548.1.1.1.0'

def fun_ciscoNoErrDisabled(context, mib_check_oid=cErrDisableRecoveryInterval):
	# The function checks the cErrDisableIfStatusTable (1.3.6.1.4.1.9.9.548.1.3.1) in the CISCO-ERR-DISABLE-MIB,
	# and returns 1 (true) if the table i sempty, i.e. there are no errdisabled ports. If there are errdisabled
	# ports the function returns 0 (false) and writes in the context a description of the relevant interfaces
	# and their errdisable reason. 

	log = context.getLog()
	
	host       = context.get('$node', None)
	community  = context.get('$community', None)
	sv         = context.get('$snmpversion', None)
	port       = context.get('$snmpport', None)
	shorttries = context.get('$shorttries', None)

	# Check if the MIB is supported (cErrDisableRecoveryInterval)
	msg = getvar(host, community, shorttries, mib_check_oid, port, sv)
	if msg is None:
		raise PollException("CISCO-ERR-DISABLE-MIB not supported (oid %s not found)." % (mib_check_oid))


	ed_oid = '1.3.6.1.4.1.9.9.548.1.3.1.1.2'
	
	ed = gettable(host, community, shorttries, ed_oid, port=port, version=sv)
	rv = True
	if not ed:
		context.appendReason('Found no errdisabled ports')
		return True
		
	msgs = []
	
	for oid,reason_code in ed:
		
		if oid.startswith('.'):
			oid = oid[1:]
			
		inst = oid[len(ed_oid) + 1:]

		try:
			ifindex,vlan = inst.split('.')
			vlan = int(vlan)
		except ValueError:
			ifindex = instance
			vlan = 0
			
		ifindex = int(ifindex)
		reason_code = int(reason_code)
		
		msg = getvar(host, community, shorttries, '1.3.6.1.2.1.2.2.1.2.%d' % ifindex, port, sv)
		if not msg:
			msg = 'interface without description'
		if vlan:
			msg = msg + ' on VLAN %d' % vlan
		msg = msg + ' ifindex %d' % ifindex
		msg = msg + ' reason %s (%d)' % (cisco_errdisable_reason_text.get(reason_code, 'unknown'), reason_code)
		msgs.append(msg)
		
	context.appendReason(', '.join(msgs))
	
	return False

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

EXPR_SYMBOLS = {
	'ciscoSensorThresholdsOk' : fun_ciscoSensorThresholdsOk,
	'ciscoNoErrDisabled'      : fun_ciscoNoErrDisabled,
}
