# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>. 

"""

External Programs execution
========================================================================

.. _fun_exec:

exec
---------------------------------------------

    **Syntax**

        **exec** **(** *current* **,** *command* **,** [ *param*, ...]  [ **,** *basedir* ] [ **,** *execute_timeout* ] **)**

    **Description**

        The function executes the external command *command* with the given param(s) *param* and returns the content of the standard output produced by the script.

        When the *current* parameter is *true*, the command is executed with the given parameters. 

        If *current* is *false*, the function returns the standard output produced by the command at the previous check for the same datagroup.

        The command must specify the full path (es.: /bin/ls).

    **Input**

        * *current*: boolean. 

        * *command*: string containing the script name.

        * *param*(s): strings containt the script\'s param(s)

        * *basedir*: basedir for the given command (es: basedir=/usr , command=/bin/hello -> /usr/bin/hello)

        * *execute_timeout*: Maximun execution timeout.

        * *exec_on_remote_agent*: The script is executed on the server where the agent related to the datagroup is running on. If *execute_timeout* is not set, it's forced to 60 seconds.

    **Output**

        Returns a *string*.

    **Errors**

    **Examples**

        +-----------------------------------------------------------------------------+----------------------------------+----------------------------------------------------------------+
        | Expression                                                                  |           Return value           | Description                                                    |
        +=============================================================================+==================================+================================================================+
        | exec(True, "mydate.sh")                                                     | *ven nov 26 16:41:30 CET 2010*   | the script *mydate.sh* is: <br /><br />#!/bin/bash<br />date   |
        +-----------------------------------------------------------------------------+----------------------------------+----------------------------------------------------------------+
        | exec(0, "mydate.sh")                                                        | *ven nov 25 16:41:30 CET 2010*   | Returns the output of the last execution                       |
        +-----------------------------------------------------------------------------+----------------------------------+----------------------------------------------------------------+ 
        | exec(0, "mydate.sh", timeout=0)                                             | *exception*                      | Program not returned before timeout                            |
        +-----------------------------------------------------------------------------+----------------------------------+----------------------------------------------------------------+
        | exec(0, "mydate.sh", basedir="/scripts/")                                   | error                            | Program not found in /scripts/ directory                       |
        +-----------------------------------------------------------------------------+----------------------------------+----------------------------------------------------------------+
        | exec(0, "mydate.sh", basedir="/scripts/", exec_on_remote_agent=1)           | error                            | Program not found in /scripts/ directory                       |
        +-----------------------------------------------------------------------------+----------------------------------+----------------------------------------------------------------+

prevExec
---------------------------------------------

    **Syntax**

        **prevExec** **(** *command* **,** [ *param*, ...] **)**

    **Description**

        The function returns the value calculated by *exec()* at the previous datagroup execution.
        
        It's a shortcut for the equivalente version: 
        
        ::
            
            prevExec(...programm..., ...params...)
            exec(1, ...program..., ....params....)

    **Input**

        * *command*: file path of the command
        * *params*: variable number of parameters
     
    **Output**

        Returns a *string*

    **Errors**
    
        Raise an exception if the previous value can is not present.

    **Examples**

        +----------------------------------------------------------------+----------------------------------+----------------------------------------------------------------+
        | Expression                                                     |           Return value           | Description                                                    |
        +================================================================+==================================+================================================================+
        | prevExec("mydate.sh")                                          | *ven nov 26 16:41:30 CET 2010*   | The previews value returned by *mydate.sh*                     |
        +----------------------------------------------------------------+----------------------------------+----------------------------------------------------------------+
 
"""



import os
import sys
import select
import time
import subprocess

from sanet_common.sanetdefs.ipc import AGENT_NETINFOHELPER_QUEUE

from sanet_poller.lib.exceptions import PollException

from sanet_poller.helpers.netinfo import NetInfoAgentHelper
from sanet_poller.lib.functions.sanet_meta import _query_server

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

EXEC_DEFAULT_BASEDIR="/"

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def exec_launch_program(context, log, program_path, args, execute_timeout=None):

	#
	# Launch command
	#
	# FIX: use TIMEOUT !
	process = None

	try: 
		command = [program_path] + args

		log.debug("Launching subprocess")

		retcode = None
		process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=False, close_fds=True)
		
		log.debug("Waiting for output (%s timeout)", execute_timeout)
		rd, wr, er = select.select([process.stdout], 
		                           [], 
		                           [process.stderr],
		                           execute_timeout
		)
		if rd or er:
			log.debug("Incoming output from process...") 

			tstart = time.time()
			process.wait()
			tend   = time.time()

			msg = 'Process terminated after %.3fs. Retcode=%s.' % (tend-tstart, process.returncode)
			log.debug(msg)
			context.appendReason(msg)

		if process.returncode == None:
			raise PollException("Timeout")

		if process.returncode != 0:
			error = process.stderr.read()
			msg = "%s: error return code %s: %s" % (program_path, process.returncode, error)
			raise PollException( msg )

		output = process.stdout.read()

		return output

	except OSError, es:
		raise PollException("exec error %s: %s" % (program_path, es))
	except ValueError, es:
		raise PollException("popen error %s: %s" % (program_path, es))
	finally:
		if process:
			log.debug("Closing process I/O")
			process.stdout.close()
			process.stderr.close()
			process = None

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def exec_launch_program_remote(context, log, program_path, args, execute_timeout=None, request_timeout=None):

	try:
		agent_id = context['__node__'].agent_id
		agent_queue_name = AGENT_NETINFOHELPER_QUEUE(agent_id)

		log.debug("Sending request to agent: %s (%s)", agent_id, agent_queue_name)

		params = {
			'path'           : program_path,
			'args'           : args,
			'execute_timeout': execute_timeout,
		}

		response_data = _query_server(context, 
			NetInfoAgentHelper.REQUEST_NETINFO_EXECUTE, 
			params, 
			timeout      = request_timeout, 
			target_queue = agent_queue_name,
		)

		context.appendReason(response_data['reason'])

		if response_data['exception'] is None:
			return response_data['output']
		else:
			raise Exception(response_data['exception'] )

	except Exception, e:
		raise PollException("Unknown error while executing remote script: %s" % (str(e)))

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def fun_exec(context, getCurrent, program_path, *args, **kwargs):

	log = context.getLog()

	basedir                = kwargs.get('basedir'             , EXEC_DEFAULT_BASEDIR)
	timeout                = kwargs.get('execute_timeout'     , 10)
	exec_on_remoteagent    = kwargs.get('exec_on_remote_agent', False)

	#
	# INPUT
	#
	program_path = str(program_path)
	args = [ str(a) for a in args ]

	log.debug("exec(%s, %s) + args = %s", getCurrent, program_path, args)

	#
	# Already executed?
	#
	val_id = ';'.join([program_path] + args)
	if getCurrent == False:
		#
		# Get the OLD value from the context!
		# 
		# If the OLD value is missing then a NotEvaluatedError will be raised
		# and the evaluation will be invalidated!
		#
		val = context.getOldValue(val_id)
		return val

	#
	# Get from execution cache
	#
	value = context.getCachedSimpleValue(val_id)
	if value != None:
		return value

	program_path = str(os.path.join( basedir, program_path))

	log.debug("Executing command (basedir=%s): %s %s ", basedir, program_path, args)

	if exec_on_remoteagent:

		log.debug("Sending request to remote agent...")

		if timeout is None:
			timeout = 60
		remote_request_timeout = timeout + 10

		output = exec_launch_program_remote(context, log, program_path, args, timeout, request_timeout=remote_request_timeout)
	else:
		log.debug("Local execution...")
		output = exec_launch_program(context, log, program_path, args, timeout)

	#
	# Handle execution ouptut
	#
	context.appendReason(
		"\n========= start output (%s,%s) =========\n" % (program_path, ','.join(args)) 
		+ output + 
		"\n=========  end  output (%s,%s) =========\n" % (program_path, ','.join(args)) 
	)

	#
	# PARSE OUTPUT
	#
	value = output.rstrip("\n")

	# 
	# Cache value in the execution cache
	#
	context.cacheSimpleValue(val_id, value)

	#
	# Save the CURRENT value for the next check.
	#
	# FIX: this is necessary only if we use both OLD and CURRENT
	#      values in the same expression!!
	#
	context.saveCurrentValue(val_id, value)

	return value

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def fun_prevExec(context, program_path, *args, **kwargs):

	log = context.getLog()

	basedir = kwargs.get('basedir', EXEC_DEFAULT_BASEDIR)

	program_path = str(program_path)
	args = [ str(a) for a in args ]

	val_id = ';'.join([program_path] + args)

	val = context.getOldValue(val_id)

	return val

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

EXPR_SYMBOLS = {
	'exec'    : fun_exec,
	'prevExec': fun_prevExec,
}


