# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it> 
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

====================
Filesystem functions
====================

calculateH3CDot11AccessPointInstance
---------------------------------------------

    **Syntax**

      **calculateH3CDot11AccessPointInstance** **(** **)**

    **Description**

      Returns the H3C Access Point (*dependent node*) instance inside the controller node (*master node*) MIB.
      
    **Output**

      String value
      
    **Errors**
    
      Raises an exception if the association can not be calculated.

    **Examples**

      +----------------------------------------+-----------------------+
      | Expression                             | Return string value   |
      +----------------------------------------+-----------------------+
      | calculateH3CDot11AccessPointInstance() | *124.1.58.9*          |
      +----------------------------------------+-----------------------+
  
"""

from sanet_poller.lib.exceptions import PollException, UndefinedSnmpValueException

from sanet_poller.lib.snmp.base import getvar

from sanet_poller.lib.evaluation import SanetSnmpValueType

from sanet_poller.lib.functions.regexps import genericRegexpFunction


#---------------------------------------------------------------------#
#                                                                     #
#---------------------------------------------------------------------#

def fun_calculateH3CDot11AccessPointInstance(context, ap_name):
	
	log = context.getLog()

	node        = context.get('$node'       , None)
	community   = context.get('$community'  , None)	
	shorttries  = context.get('$shorttries' , None)
	snmpport    = context.get('$snmpport'   , None)
	snmpversion = context.get('$snmpversion', None)

	
	if node        == None: raise PollException('Parameter $node missing.')
	if community   == None: raise PollException('Parameter $community missing.')
	if shorttries  == None: raise PollException('Parameter $shorttries missing.')
	if snmpport    == None: raise PollException('Parameter $snmpport missing.')
	if snmpversion == None: raise PollException('Parameter $snmpversion missing.')


	log.debug("Searcing instance for '%s'", ap_name)

	#
	# Find node instance
	#
	OID_INSTANCE_NAME= '1.3.6.1.4.1.2011.10.2.75.2.1.1.1.5'


	oid = OID_INSTANCE_NAME
	match_string =  '^' + ap_name + '$'
	snmpexpr = SanetSnmpValueType(context, OID_INSTANCE_NAME, community, node, snmpport)
	instance = genericRegexpFunction(context, 'VU', snmpexpr, match_string, True, False)
	if instance == None:
		raise UndefinedSnmpValueException("Current variable %s can't be read or is not present (GET) on %s" % (oid, node) )


	log.debug("Found!")
	log.debug("Ap name : %s", ap_name)
	log.debug("Instance: %s", instance)
	

	#
	# Verify association status
	#
	OID_INSTANCE_STATUS = '1.3.6.1.4.1.2011.10.2.75.2.1.1.1.8'
	oid = OID_INSTANCE_STATUS +'.'+ instance	
	v = getvar( node, community, shorttries, oid, port=snmpport, version=snmpversion)
	if (v == None):	
		raise UndefinedSnmpValueException("Current variable %s can't be read or is not present (GET) on %s" % (oid, node) )		
		
	log.debug("Association status : %s", repr(v))
	if v == 1:
		return instance

	raise PollException("Wrong association status (%s)" % (v,) )

#---------------------------------------------------------------------#
#                                                                     #
#---------------------------------------------------------------------#


EXPR_SYMBOLS = {
	'calculateH3CDot11AccessPointInstance': fun_calculateH3CDot11AccessPointInstance, # sanet3
}


