# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it> 
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

Storage utils
=============

hrStorageNoFull
---------------------------------------------

     **Syntax**

      **hrStorageNoFull** **(** *default_percent_threshold* , [ *storage_re* **,** *percent_threshold* [, ... ] ] **)**

     **Description**

      This function takes one mandatory argument, and optional argument pairs. 
      
      When invoked with just the mandatory argument, it will check the occupation of all the filesystems of the current node, as seen in the hrstorage table, having type "Fixed disk" (1.3.6.1.2.1.25.2.1.4) or "Network disk" (.1.3.6.1.2.1.25.2.1.10), against the *default_percent_threshold*. 

      If all filesystems are below thresholds, it will return 0, otherwise it will return 1.

      Optional argument pairs of the form:
      
        *storage_re*, *percent_threshold* 
      
      specify a different threshold for specific filesystems. 
      
      When the hrStorageDescr (1.3.6.1.2.1.25.2.3.1.3) of the filesystem matches the provided *storage_re*, its occupation will be checked against the *percent_threshold* specified next to it, instead of the first argument. 
      
      You can specify **-1** as a threshold to actually disable checking some filesystems.

      If one of the specified filesystems regexp is not found, the function skips it without error. 

     **Returns**
    
      Returns 1 when all selected filesystems satisfy the given thresholds (*default_percent_threshold* for default or *percent_threshold* for the selected filesystems), 0 otherwise.

      If no filesystems are found, or the node doesn't answer to SNMP requests, the function becomes uncheckable. 
      
      If the node has only entries of type different from "Fixed disk" and "Network disk" in the hrStorageTable, the function returns 1.

     **Examples**

      +--------------------------------------------------------------------+-------------------------------------------------------------------------------+
      | Expression                                                         | Return string value                                                           |
      +====================================================================+===============================================================================+
      | hrStorageNoFull(99)                                                | 1                                                                             |
      +--------------------------------------------------------------------+-------------------------------------------------------------------------------+
      | hrStorageNoFull(99, "^/run$", -1)                                  | 1, /run ignored.                                                              |
      +--------------------------------------------------------------------+-------------------------------------------------------------------------------+
      | hrStorageNoFull(99, "^/run/log$", 80, "^/home$", 90)               | 1. Check every partition with 99%, /run/log with 80%, /home with 90%.         |
      +--------------------------------------------------------------------+-------------------------------------------------------------------------------+

"""

import re

from sanet_poller.lib.exceptions import PollException

from sanet_poller.lib.snmp.base import getvar, gettable

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#


def fun_hrStorageNoFull(context, defperc, *args):
	
	if len(args) % 2:
		raise PollException('hrStorageNoFull invoked with odd number of arguments')

	if not (0<=defperc and defperc<=100):
		raise PollException("Invalid default threshold (value=%s)" % defperc)

		
	log = context.getLog()
	host       = context.get('$node', None)
	community  = context.get('$community', None)
	sv         = context.get('$snmpversion', None)
	port       = context.get('$snmpport', None)
	shorttries = context.get('$shorttries', None)

	hrStorage = '1.3.6.1.2.1.25.2.3.1'
	hrStorageType  = hrStorage + '.2'
	hrStorageDescr = hrStorage + '.3'
	hrStorageSize  = hrStorage + '.5'
	hrStorageUsed  = hrStorage + '.6'

	hrStorageFixedDisk   = '1.3.6.1.2.1.25.2.1.4'
	hrStorageNetworkDisk = '1.3.6.1.2.1.25.2.1.10'
	
	log.debug("hrStorageNoFull(%s, %s)", defperc, args)

	dt = gettable(host, community, shorttries, hrStorageDescr, port=port, version=sv)
	if not dt:
		reason = 'hrStorageNoFull: cannot find any filesystem in the hrStorageTable'
		context.appendReason(reason)
		raise PollException(reason)
		
		
	ret = True
	
	for oid, descr in dt:

		perc = float(defperc)
		
		log.debug("%s %s", oid, descr)
		
		if oid.startswith('.'):
			oid = oid[1:]
			
		inst = oid[len(hrStorageDescr) + 1:]

		typ  = getvar(host, community, shorttries, hrStorageType + '.' + inst , port, sv)
		if typ and typ.startswith('.'):
			typ = typ[1:]
			
		if not typ or (typ not in [ hrStorageFixedDisk , hrStorageNetworkDisk ]):
			continue
			
		i = 0
		while i < len(args):
			regexp, thisperc = args[i:i+2]
 			i += 2
			if re.compile(regexp).match(descr):
				perc = float(thisperc)
				break
		
		if perc == -1:
			reason = '%-5s storage %s ignored.' % ('-', descr)
			context.appendReason(reason)
			continue
		
		size = getvar(host, community, shorttries, hrStorageSize + '.' + inst, port, sv)
		if size is None:
			reason = 'Cannot read size for storage %s' % descr
			context.appendReason(reason)
			raise PollException(reason)
			
		if float(size) == 0:
			continue

		used = getvar(host, community, shorttries, hrStorageUsed + '.' + inst, port, sv)
		if used is None:
			reason = 'Cannot read used for storage %s' % descr
			context.appendReason(reason)
			raise PollException(reason)

		full = float(used) / float(size) > (perc/100.0)
		if full:
			ret = False

		reason = '%-5s storage %s has %.f used, %.f size, %.02f%% occupation against %.02f%% threshold.' % (
			'OK' if not full else 'FAIL',
			descr, 
			float(used), 
			float(size), 
			float(used) / float(size) * 100.0, 
			perc
		)
		context.appendReason(reason)

	return (1 if ret else 0)


EXPR_SYMBOLS = {
	'hrStorageNoFull'         : fun_hrStorageNoFull,
}

