"""

===============
Interface utils
===============

isEtherFd
---------------------------------------------

    **Syntax**

      **isEtherFd** **(** *ifindex* **)**

    **Description**

      The function tries to verify if the interface with the given *ifindex* is in ethernet Full-duplex mode. 

      The function looks inside the *execution context* for the following informations:
      * $node: the hostname of the remote host
      * $community: the SNMP community to use
      * $snmpversion: SNMP version to use
      * $snmpport: SNMP port to use
      * $shorttries: number of SNMP tries

    **Input**

      *ifindex*: integer value.

    **Output**

      Returns a *boolean*

    **Errors**

      The function raises an exception if it can not retrieve enough informations about the interface.

    **Examples**

      +-------------------------------------+--------------+--------------------------------------------------------------------------------------+
      | Expression                          | Return value | Description                                                                          |
      +=====================================+==============+======================================================================================+
      | isEtherFd( 1 )                      | *true*       |                                                                                      |
      +-------------------------------------+--------------+--------------------------------------------------------------------------------------+
      | isEtherFd( $myIfindex )             | *true*       |                                                                                      |
      +-------------------------------------+--------------+--------------------------------------------------------------------------------------+
      | isEtherFd( 1.3.6.1.2.1.2.2.1.1.1@ ) | *false*      | Retrieve the ifindex stored in the mib of the current node (1.3.6.1.2.1.2.2.1.1.1@)  |
      +-------------------------------------+--------------+--------------------------------------------------------------------------------------+
 

getHCInOctets
---------------------------------------------

     **Syntax**

       *getHCInOctets* **(** [ *host* ] [ **,** *ifindex*] [ **,** *community* ] [ **,** *shorttries* ] **)**

     **Description**

       Returns the value of the *ifHCInOctets* for the given *host*/*iface*.

     **Input**

       * *host*: the node.
       * *ifindex*: (optional) the ifindex of the current interface.
       * *comm*: (optional) the SNMP community to use.
       * *shorttries*: (option) the max number of SNMP packets to send before abort.

     **Output**

       Returns a number

     **Errors**

       The function raises an exception if an error occurs.

     **Examples**

       +-------------------------------------+--------------+--------------------------------------------------------------------------------------+
       | Expression                          | Return value | Description                                                                          |
       +=====================================+==============+======================================================================================+
       | getHCInOctets($node)                | 123131       |                                                                                      |
       +-------------------------------------+--------------+--------------------------------------------------------------------------------------+
 


getHCOutOctets
---------------------------------------------

     **Syntax**
     
       *getHCOutOctets* **(** [ *host* ] [ **,** *ifindex*] [ **,** *community* ] [ **,** *shorttries* ] **)**
 
     **Description**

       Returns the value of the *ifHCOutOctets* for the given *host*/*iface*.

     **Input**

       * *host*: the node.
       * *ifindex*: (optional) the ifindex of the current interface.
       * *comm*: (optional) the SNMP community to use.
       * *shorttries*: (option) the max number of SNMP packets to send before abort.

     **Output**

       Returns a *boolean*

     **Errors**

       The function raises an exception if it can not retrieve enough informations about the interface.

     **Examples**
    
       +-------------------------------------+--------------+--------------------------------------------------------------------------------------+
       | Expression                          | Return value | Description                                                                          |
       +=====================================+==============+======================================================================================+
       | getHCOutOctets($node)               | 5678         |                                                                                      |
       +-------------------------------------+--------------+--------------------------------------------------------------------------------------+

ifSpeed
---------------------------------------------

     **Syntax**
     
       **ifSpeed** **(** ifindex [ **,** default ]  [ **,** ignore_snmp_errors ] **)**
       
     **Description**

       Tries to fetch the interface speed via SNMP for the given *ifindex* interface. Returns the number of bits/second.
       
       The functions search for the ifHighSpeed (1.3.6.1.2.1.31.1.1.1.15) or the ifSpeed (1.3.6.1.2.1.2.2.1.5) OIDS.
       
       if *ignore_snmp_errors* is set to *true* (or 1), any error while accessing the *OIDS* is ignored. If *default* is specified and *ignore_snmp_errors* is enabled, *default* is returned.

     **Input**

       * *ifindex*           : ifindex of the selected interface.
       * *default*           : (optional) Default return value if an error occurs and *ignore_snmp_errors* is enabled.
       * *ignore_snmp_errors*: (optional) the SNMP community to use.

     **Output**

       Returns a numeric value (bit/s).

     **Errors**

       The function raises an exception when an error occurs and *ignore_snmp_errors* is not enabled.

     **Examples**
    
       +-------------------------------------+--------------+--------------------------------------------------------------------------------------+
       | Expression                          | Return value | Description                                                                          |
       +=====================================+==============+======================================================================================+
       | ifSpeed($ifindex)                   | 1000000      |                                                                                      |
       +-------------------------------------+--------------+--------------------------------------------------------------------------------------+


"""

from sanet_common.exprparser.Types import SnmpValueType

from sanet_poller.lib.exceptions import PollException

from sanet_poller.lib.snmp.base import getvar

from sanet_poller.lib.evaluation import SanetSnmpValueType

# FIX: BAD, use a common code instead
from .regexps import fun_byRegexpUnique

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def _getHCvar(context, oid, host=None, ifindex=None, comm=None, shorttries=None):

	log = context.getLog()

	host       = host       if host       is not None else context.get('$node'      , None)
	ifindex    = ifindex    if ifindex    is not None else context.get('$ifindex'   , None)
	comm       = comm       if comm       is not None else context.get('$community' , None)
	shorttries = shorttries if shorttries is not None else context.get('$shorttries', None)

	port       = context.get('$snmpport', None)

	# Force right SNMP version for HC vars
	# and handler IPv correctly
	sv         = context.get('$snmpversion', 2)
	if sv == 1:
		sv = 2

	# Handle right SNMP version with IPv6
	ipv = context.get('$ipversion', 4)
	if ipv == 6:
		if sv < 62:
			sv = 62
			log.debug("Forcing SNMP with IPv6. Changed version to %s", sv)

	if host       == None: raise PollException('Parameter $host missing or invalid.')
	if ifindex    == None: raise PollException('Parameter $ifindex missing or invalid.')
	if comm       == None: raise PollException('Parameter $comm missing or invalid.')
	if port       == None: raise PollException('Parameter $port missing or invalid.')
	if shorttries == None: raise PollException('Parameter $shorttries missing or invalid.')

	# Be sure the ifindex is a valid interger value.
	ifindex = int(ifindex)

	log.debug("getHCvar(%s,host=%s,ifindex=%s,community=%s,tries=%s,snmpport=%s,snmpversion=%s)", oid, host, ifindex, comm, shorttries, port,sv)

	result = getvar(host, comm, shorttries, '%s.%s' % (oid, ifindex), port, sv)
	if result == None:
		raise PollException("HC %s.%s not found" % (oid, ifindex))

	return result

#-----------------------------------------------------------------------------#

def fun_getHCInOctets(context, host=None, ifindex=None, community=None, shorttries=None):

	return _getHCvar(context, '1.3.6.1.2.1.31.1.1.1.6', host=host, ifindex=ifindex, comm=community, shorttries=shorttries)

#-----------------------------------------------------------------------------#

def fun_getHCOutOctets(context, host=None, ifindex=None, community=None, shorttries=None):

	return _getHCvar(context, '1.3.6.1.2.1.31.1.1.1.10', host=host, ifindex=ifindex, comm=community, shorttries=shorttries)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_isEtherFd(context, ifindex):

	reason=''

	log = context.getLog()

	log.debug("isEtherFd(%s)", ifindex)

	host       = context.get('$node', None)
	community  = context.get('$community', None)
	sv         = context.get('$snmpversion', None)
	port       = context.get('$snmpport', None)
	shorttries = context.get('$shorttries', None)

	if ifindex    == None: raise PollException('Parameter $ifindex missing or invalid')
	if host       == None: raise PollException('Parameter $node missing or invalid')
	if community  == None: raise PollException('Parameter $community missing or invalid')
	if sv         == None: raise PollException('Parameter $snmpversion missing or invalid')
	if port       == None: raise PollException('Parameter $snmpport missing or invalid')
	if shorttries == None: raise PollException('Parameter $shorttries missing or invalid')

	# Be sure the ifindex values is a valid interger value.
	ifindex = int(ifindex)

	# Return (value, reason)

	# Try Ether-like MIB
	ans = getvar(host, community, shorttries, '1.3.6.1.2.1.10.7.2.1.19.%s' % ifindex , port, sv)
	if ans:
		reason = '1.3.6.1.2.1.10.7.2.1.19.%s = %r' % (ifindex, ans)
		
		context.appendReason(reason)
		
		if int(ans) == 3:
			return True	
		elif int(ans) == 2:
			return False    
		else:
			# Il caso Unknown (1) lo gestiamo come mib non funzionante
			context.appendReason('Unknown (1) state')
			pass

	reason = "Ether-like MIB unusable, "

	# Try MAU MIB
	ans = getvar(host, community, shorttries, '1.3.6.1.2.1.26.2.1.1.3.%s.1' % ifindex , port, sv)
	if ans:
		reason +=  '1.3.6.1.2.1.26.2.1.1.3.%s.1 = %r' % (ifindex, ans)

		context.appendReason(reason)

		if   ans == '1.3.6.1.2.1.26.4.1': #  dot3MauTypeAUIot3MauTypeAUI
			# Ignore this type. Mib not working
			pass			
		elif ans == '1.3.6.1.2.1.26.4.11': # dot3MauType10BaseTFD
			return True 
		elif ans == '1.3.6.1.2.1.26.4.13': # dot3MauType10BaseFLFD
			return True 
		elif ans == '1.3.6.1.2.1.26.4.16': # dot3MauType100BaseTXFD
			return True 
		elif ans == '1.3.6.1.2.1.26.4.18': # dot3MauType100BaseFXFD
			return True 
		elif ans == '1.3.6.1.2.1.26.4.20': # dot3MauType100BaseT2FD
			return True 
		elif ans == '1.3.6.1.2.1.26.4.22': # dot3MauType1000BaseXFD
			return True 
		elif ans == '1.3.6.1.2.1.26.4.24': # dot3MauType1000BaseLXFD
			return True 
		elif ans == '1.3.6.1.2.1.26.4.26': # dot3MauType1000BaseSXFD
			return True 
		elif ans == '1.3.6.1.2.1.26.4.28': # dot3MauType1000BaseCXFD
			return True 
			
		elif ans == '1.3.6.11.3.6.1.2.1.10.7.2.1.19.2.1.26.4.30': # dot3MauType1000BaseTFD
			return True 
			
		elif ans == '1.3.6.1.4.1.11.2.3.7.8.11.8': # HPstack
			return True 
		elif ans == '1.3.6.1.4.1.11.2.3.7.8.11.12': # hpicfMauType10GigBaseCX4
			return True 
		elif ans == '1.3.6.1.4.1.11.2.3.7.8.11.13': # hpicfMauType10GigBaseSR
			return True
		elif ans == '1.3.6.1.4.1.11.2.3.7.8.11.14': # hpicfMauType10GigBaseER
			return True
		elif ans == '1.3.6.1.4.1.11.2.3.7.8.11.15': # hpicfMauType10GigBaseLR
			return True
		elif ans == '1.3.6.1.4.1.11.2.3.7.8.11.2': # hpicfMauType1000BaseSXFD
			return True
		elif ans == '1.3.6.1.4.1.11.2.3.7.8.11.4': # hpicfMauType1000BaseLXFD
			return True
		elif ans == '1.3.6.1.4.1.11.2.3.7.8.11.6': # hpicfMauType1000BaseTFD
			return True
		elif ans == '1.3.6.1.4.1.11.2.3.7.8.11.8': # hpicfMauType1000BaseStkFD
			return True
		elif ans == '1.3.6.1.4.1.11.2.3.7.8.11.9': # hpicfMauType1000BaseLHFD
			return True
			
		elif ans == '1.3.6.1.4.1.1916.3.1.1.4': # extremeMauType1000BaseSXFD
			return True
		elif ans == '1.3.6.1.4.1.1916.3.1.1.5': # extremeMauType1000BaseLXFD
			return True	
		elif ans == '1.3.6.1.4.1.1916.3.1.1.6': # extremeMauType1000BaseCXFD
			return True
		elif ans == '1.3.6.1.4.1.1916.3.1.1.8': # extremeMauType1000BaseWDMFD
			return True
		elif ans == '1.3.6.1.4.1.1916.3.1.1.10': # extremeMauType1000BaseLX70FD
			return True
		elif ans == '1.3.6.1.4.1.1916.3.1.1.12': # extremeMauType1000BaseZXFD
			return True
		elif ans == '1.3.6.1.4.1.1916.3.1.1.14': # extremeMauType1000BaseLX100FD
			return True
			
		elif ans == '0.0':
			# Ignore this type. Mib not working
			pass						
			

		else:
			# hpicfMauType1000BaseStkHD     1.3.6.1.4.1.11.2.3.7.8.11.7
			return False
	
	reason += "MAU MIB unusable, ";

	# Try CISCO STACK MIB
	try:
		snmptable = SanetSnmpValueType(context, '1.3.6.1.4.1.9.5.1.4.1.1.11', community, host, port)
		inst      = fun_byRegexpUnique(context, snmptable, '^%s$' % ifindex );
		
		ans = getvar(host, community, shorttries, '1.3.6.1.4.1.9.5.1.4.1.1.10.%s' % inst, port, sv)
		if ans:
			reason += '1.3.6.1.4.1.9.5.1.4.1.1.10.%s = %r' % (inst,ans)
			
			context.appendReason(reason)
	
			if ans == 2:
				return True
			elif ans == 1:
				return False
			
			reason += "CISCO STACK MIB unusable: 1.3.6.1.4.1.9.5.1.4.1.1.10.%s = %r, " % (inst,ans)
		else:
			reason += "CISCO STACK MIB unusable, ";
			
	except PollException, ex:
		reason += "CISCO STACK MIB unusable: cannot find %s, " % (ifindex);

	# Try CTRON MIB
	# Only unit 1
	ans = getvar(host, community, shorttries, '1.3.6.1.4.1.52.3.9.1.10.7.2.1.1.4.1.%s' % ifindex, port, sv )
	if ans:
		reason +=  '1.3.6.1.4.1.52.3.9.1.10.7.2.1.1.4.1.%s = %r' % (ifindex, ans)
		
		context.appendReason(reason)
		
		if ans == 2:    # 10full
			return True
		elif ans == 4:  # 100full
			return True
		elif ans == 6:  # 1000full
			return True
		else:
			return False

	# Try CISCO c2900 mib
	# 1.3.6.1.4.1.9.9.87.1.4.1.1.25 is IfIndex
	# 1.3.6.1.4.1.9.9.87.1.4.1.1.32 duplex (1 full 2 half)
	try:
		snmptable = SanetSnmpValueType(context, '1.3.6.1.4.1.9.9.87.1.4.1.1.25', community, host, port)
		inst      = fun_byRegexpUnique(context, snmptable, '^%s$' % ifindex )
		
		ans = getvar(host, community, shorttries, '1.3.6.1.4.1.9.9.87.1.4.1.1.32.%s' % inst, port, sv)
		if ans:
			reason = '1.3.6.1.4.1.9.9.87.1.4.1.1.32.%s = %r' % (inst,ans)
			
			context.appendReason(reason)
			
			if ans == 1:
				return True
			elif ans == 2:
				return False
			
			reason += "CISCO c2900 MIB unusable: 1.3.6.1.4.1.9.9.87.1.4.1.1.32.%s = %r" % (inst,ans)
		else:
			reason += "CISCO c2900 MIB unusable, "
	
	except PollException, ex:
		reason += "CISCO c2900 MIB unusable: cannot find %s, " % ifindex

	reason += "CTRON MIB unusable";
	raise PollException(reason)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_ifSpeed(context, ifindex, default=None, ignore_snmp_errors=False):

	reason=''

	log = context.getLog()

	log.debug("ifSpeed(%s)", ifindex)

	host       = context.get('$node', None)
	community  = context.get('$community', None)
	sv         = context.get('$snmpversion', None)
	port       = context.get('$snmpport', None)
	shorttries = context.get('$shorttries', None)

	if ifindex    == None: raise PollException('Parameter $ifindex missing or invalid')
	if host       == None: raise PollException('Parameter $node missing or invalid')
	if community  == None: raise PollException('Parameter $community missing or invalid')
	if sv         == None: raise PollException('Parameter $snmpversion missing or invalid')
	if port       == None: raise PollException('Parameter $snmpport missing or invalid')
	if shorttries == None: raise PollException('Parameter $shorttries missing or invalid')


	try:
		# Return (value, reason)
		oid = '1.3.6.1.2.1.31.1.1.1.15.%s' % (ifindex)
		log.debug("Getting %s", oid)

		val = getvar(host, community, shorttries, oid, port, sv)
		if val is not None:
			# NOTE: An estimate of the interface's current bandwidth in units of 1,000,000 bits per second.
			val = val * 1000000
			return val

		#
		# FIX: forse bisogna invertire queste due FETCH e verificare che 1.3.6.1.2.1.2.2.1.5 sia == 4294967295. (tappo)
		#

		oid   = '1.3.6.1.2.1.2.2.1.5.%s'     % (ifindex)
		log.debug("Getting %s", oid)

		val = getvar(host, community, shorttries, oid, port, sv)
		if val is not None:
			return val
			
	except Exception, e:
		msg = "Unable to calculate interface speed: %s:%s" % (type(e), str(e))
		log.warning(msg)
		if not ignore_snmp_errors:
			raise PollException(msg)

	if default is not None:
		return default
	
	raise PollException("Unable to calculate interface speed")



EXPR_SYMBOLS = {
	'getHCInOctets'         : fun_getHCInOctets,  # tested
	'getHCOutOctets'        : fun_getHCOutOctets, # tested
	'isEtherFd'             : fun_isEtherFd,      # tested

	'ifSpeed'               : fun_ifSpeed,
}
