
"""
********************************************************
JMX Protocol
********************************************************

jmx
=======================

	**IMPORTANT**: this function requires the *jmx-utils* package installed where machine the agent using this function is running on.

	**Syntax**

		**jmx** **(** *bean_type* , *bean_name* , [ *options* ] **)**

	**Description**

		Retrieves the value of the given java bean: *bean_type*, *bean_name*

		Valid optinal arguments are:

		+--------------------+------------------------------------------------------------------------------------------------------------------------+
		| Option             | Description                                                                                                            |
		+====================+========================================================================================================================+
		| getlast            | When 1 the function returns the value fetched in the last datagroup execution.                                         |
		+--------------------+------------------------------------------------------------------------------------------------------------------------+
		| host               | JMX server. Default is taken from execution context $node.                                                             |
		+--------------------+------------------------------------------------------------------------------------------------------------------------+
		| port               | JMX port. Default is taken from execution context $port                                                                |
		+--------------------+------------------------------------------------------------------------------------------------------------------------+
		| username           | JMX username.                                                                                                          |
		+--------------------+------------------------------------------------------------------------------------------------------------------------+
		| password           | JMX password                                                                                                           |
		+--------------------+------------------------------------------------------------------------------------------------------------------------+
		| java_bin           | JVM shell command. Default is: /usr/bin/java.                                                                          |
		+--------------------+------------------------------------------------------------------------------------------------------------------------+
		| jmxcmd_bin         | JMX Client program installed with  the *jmx-utils* package. Default is /usr/local/jmxcmd/bin/jmxcmd.jar,               |
		+--------------------+------------------------------------------------------------------------------------------------------------------------+
		| timeout            | Timeout in seconds.                                                                                                    |
		+--------------------+------------------------------------------------------------------------------------------------------------------------+

	**Input**

		1. bean_stype: string
		2. bean_name: string

	**Outpu**

		Returns the bean's value. 

	**Errors**

		Raise an exception if an error occours.

	**Examples**

		+-------------------------------------------------------------+-------------------------------------+-------------------------------------------------------+
		| Expression                                                  | Return value                        |  Description                                          |
		+=============================================================+=====================================+=======================================================+
		| jmx('java.lang:type=Runtime', 'Uptime')                     | 123141151                           |                                                       |
		+-------------------------------------------------------------+-------------------------------------+-------------------------------------------------------+
		| jmx('java.lang:type=Memory', 'HeapMemoryUsage.committed')   | 561231                              |                                                       |
		+-------------------------------------------------------------+-------------------------------------+-------------------------------------------------------+
      
"""

import os
import subprocess
import select
import time
import json

from sanet_common.sanetdefs.constants import *

from sanet_poller.lib.exceptions import PollException

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

JAVA_BIN_DEFAULT_PATH    = '/usr/bin/java'

JMXCMD_BIN_DEFAULT_PATH  = '/usr/local/jmxcmd/bin/jmxcmd.jar'

JMX_TIMEOUT = 10000 # 10 seconds

def fun_jmx(context, bean_type, bean_name, getlast=0, host=None, port=None, username=None, password=None, java_bin=None, jmxcmd_bin=None, timeout=None):

	log = context.getLog()

	if host       == None: host  = context.get('$node', None)
	if port       == None: port  = context.get('$port', 1090)

	if host == None:
		raise PollException("node not specified")


	attribute_match = None

	if bean_name.find('.') > -1:
		bean_name, attribute_match = bean_name.split('.',1)

	log = context.getLog()

	val_id = ';'.join([host +'.jmx.'+bean_name] )
	if getlast:
		val = context.getOldValue(val_id)
		return val

	
	json_value = launch_jmxcmd(context, log, bean_type, bean_name, host=host, port=port, username=username, password=password, java_bin=java_bin, jmxcmd_bin=jmxcmd_bin, timeout=timeout)

	try:
		object_value = json.loads(json_value)
	except Exception, e:
		raise PollException("Malformed data from JMX call: %s" % (e))
	
	if isinstance(object_value, types.DictType):  # simple value
		value = object_value['v']

	elif isinstance(object_value, types.ListType): # Multiple values
		
		r = []
		for o in object_value:
			log.debug("\t%s = %s" % (o['k'], o['v']))
			if attribute_match:
				if o['k'].find(attribute_match) > -1:
					log.debug("\t\tmatched")
					r.append( (o['k'],o['v']) )
			else: 
				r.append( (o['k'],o['v']) )
		
		if len(r) == 1:
			value = r[0][1]
		else:
			value = r
	else:
		raise PollException("Invalid object returned from JMX utility")
		
	
	# 
	# Cache value in the execution cache
	#
	context.cacheSimpleValue(val_id, value)

	#
	# Save the CURRENT value for the next check.
	#
	# FIX: this is necessary only if we use both OLD and CURRENT
	#      values in the same expression!!
	#
	context.saveCurrentValue(val_id, value)	
	
	return value
	
def launch_jmxcmd(context, log, bean_type, bean_name, host=None, port=None, username=None, password=None, java_bin=None, jmxcmd_bin=None, timeout=None):

	if jmxcmd_bin == None: jmxcmd_bin = os.environ.get('SANET3_JMXCMD_BIN', JMXCMD_BIN_DEFAULT_PATH)
	if java_bin   == None: java_bin   = os.environ.get('JAVA_BIN'         , JAVA_BIN_DEFAULT_PATH) 
	if timeout    == None: timeout    = os.environ.get('JMX_TIMEOUT'      , JMX_TIMEOUT)

	process = None

	try: 
		auth_params = "-:-"
		if username:
			auth_params = "%s:%s" % (username, password)

		remote_host = "%s:%s" % (host, port)

		command = [java_bin, "-jar", jmxcmd_bin, auth_params, remote_host, bean_type, bean_name]

		log.debug("Launching subprocess: %s", ' '.join(command))

		log.debug("\n");
		
		env = os.environ.copy()
		if timeout is not None:
			log.debug("Using timeout: %s", timeout)
			env['TIMEOUT'] = str(int(timeout))
		

		retcode = None
		process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=False, close_fds=True, env=env)
		
		log.debug("Waiting for output (%s timeout)", timeout)
		rd, wr, er = select.select([process.stdout], [], [process.stderr], float(timeout) / 1000)
		if rd or er:
			log.debug("Incoming output from process...") 

			tstart = time.time()
			process.wait()
			tend   = time.time()

			msg = 'Process terminated after %.3fs. Retcode=%s.' % (tend-tstart, process.returncode)
			log.debug(msg)
			context.appendReason(msg)

		if process.returncode == None:
			raise PollException("Timeout")

		error_msgs = ''
		log.debug("Error messages? %s", len(er))
		if er:
			error_msgs = process.stderr.read()

		if error_msgs:
			log.debug("Error messages: %s", error_msgs)
			context.appendReason('Error: ' + error_msgs)

		if process.returncode != 0:
			#error = process.stderr.read()

			msg = "error return code %s: %s" % (process.returncode, error_msgs)
			raise PollException( msg )

		output = process.stdout.read()
		
		log.debug("OUTPUT: len=%s content=%s", len(output), output)

		return output

	except OSError, es:
		raise PollException("exec error %s: %s" % (command, es))
	except ValueError, es:
		raise PollException("popen error %s: %s" % (command, es))
	finally:
		if process:
			log.debug("Closing process I/O")
			process.stdout.close()
			process.stderr.close()
			process = None



#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

EXPR_SYMBOLS = {
	'jmx'       : fun_jmx,
}
