# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>. 

"""
===============
MS SQL function
===============

mssql_check
---------------------------------------------

    **Syntax**

      **mssql_check** **(** *mode, username [, password ][, host ][, port ][, db_name][, ...]* **)**

    **Description**

      Performs a check specified by **mode** argument and returns its value. If not specified, **host** argument is taken from the checked node.

    **Input**

      *mode*: kind of control to perform on MS SQL Server.

      *username*: username used to execute queries on MS SQL Server.

      *password*: password associated to *username* for authentication.

      *host*: host name or ip address of MS SQL Server instance. Default is $node.

      *port*: port to connecto to on *host* machine. Default is 1433.
      
      *db_name*: name of a specific database to perform the control on (meaningful only for some controls). If not specified all databases are considered.

      *threshold*: a threshold value to use on specific mode checks, mandatory for them.
      
      **Modes**
            * Server controls
            
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | Modes            | Description                                                                                                                          | Return value | Threshold | DB_name  |
            +==================+======================================================================================================================================+==============+===========+==========+
            | connections      | Number of open connections to MS SQL server.                                                                                         |    int       |     No    |    No    |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | cpu              | Percentage of CPU utilization by MS SQL Server.                                                                                      |  percentage  |     No    |    No    |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | bufferhitratio   | Percentage of pages found in the buffer cache without having to read from disk. The ratio is the total number of cache hits          |  percentage  |     No    |    No    |
            |                  | divided by the total number of cache lookups over the last few thousand page accesses.                                               |              |           |          |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | pagelooks        | Page lookups per second.                                                                                                             | float(delta) |     No    |    No    |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | freepages        | Total number of free pages (cumulative).                                                                                             |    int       |     No    |    No    |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | totalpages       | Total number of pages (cumulative).                                                                                                  |    int       |     No    |    No    |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | targetpages      | Ideal number of pages in buffer pool (cumulative).                                                                                   |    int       |     No    |    No    |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | databasepages    | Number of pages in the buffer pool with database content (cumulative).                                                               |    int       |     No    |    No    |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | stolenpages      | Number of pages used for miscellaneous server purposes (including procedures cache).                                                 |    int       |     No    |    No    |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | lazywrites       | Number of buffers written per second by the buffer manager's lazy writer. The lazy writer is a system process that flushes out       | float(delta) |     No    |    No    |
            |                  | batches of dirty, aged buffers (buffers that contain changes that must be written back to disk before the buffer can be reused       |              |           |          |
            |                  | for a different page) and makes them available to user processes. The lazy writer eliminates the need to perform frequent            |              |           |          |
            |                  | checkpoints in order to create available buffers.                                                                                    |              |           |          |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | readahead        | Number of pages read per second in anticipation of use.                                                                              | float(delta) |     No    |    No    |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | pagereads        | Number of physical database page reads that are issued per second. This statistic displays the total number of physical page reads   | float(delta) |     No    |    No    |
            |                  | across all databases.                                                                                                                |              |           |          |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | checkpoints      | Number of pages flushed to disk per second by a checkpoint or other operation that require all dirty pages to be flushed.            | float(delta) |     No    |    No    |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | pagewrites       | Number of physical database page writes issued per second.                                                                           | float(delta) |     No    |    No    |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | lockrequests     | Number of new locks and lock conversions per second requested from the lock manager.                                                 | float(delta) |     No    |    No    |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | locktimeouts     | Number of lock requests per second that timed out, including requests for NOWAIT locks.                                              | float(delta) |     No    |    No    |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | deadlocks        | Number of lock requests per second that resulted in a deadlock.                                                                      | float(delta) |     No    |    No    |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | lockwaits        | Number of lock requests per second that required the caller to wait.                                                                 | float(delta) |     No    |    No    |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | lockwait         | Total wait time (in milliseconds) for locks in the last second.                                                                      |     int      |     No    |    No    |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | averagewait      | Average amount of wait time (in milliseconds) for each lock request that resulted in a wait.                                         |     int      |     No    |    No    |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | pagesplits       | Number of page splits per second that occur as the result of overflowing index pages.                                                | float(delta) |     No    |    No    |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | cachehit         | Ratio between cache hits and lookups.                                                                                                |     int      |     No    |    No    |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | sqlcompilations  | Number of compilations per second. Indicates the number of times the compile code path is entered. Includes compiles caused          | float(delta) |     No    |    No    |
            |                  | by statement-level recompilations in SQL Server. After SQL Server user activity is stable, this value reaches a steady state.        |              |           |          |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | batchreq         | Number of Transact-SQL command batches received per second. This statistic is affected by all constraints (such as I/O, number       | float(delta) |     No    |    No    |
            |                  | of users, cache size, complexity of requests, and so on). High batch requests mean good throughput.                                  |              |           |          |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | fullscans        | Number of unrestricted full scans per second. These can be either base-table or full-index scans.                                    | float(delta) |     No    |    No    |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | pagelife         | Number of seconds a page will stay in the buffer pool without references.                                                            |    float     |     No    |    No    |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            | time2connect     | Time to connecto to the database.                                                                                                    |    float     |     No    |    No    |
            +------------------+--------------------------------------------------------------------------------------------------------------------------------------+--------------+-----------+----------+
            
            
            * Database controls

            On these controls, the function checks if everydatabase (or the one specified by *db_name* parameter) has the property to check below the value given by *threshold** parameter.

            +------------------+---------------------------------------------------------------------------------------------------------------+-----------+----------+
            | Modes            | Description                                                                                                   | Threshold | DB_name  |
            +==================+===============================================================================================================+===========+==========+
            | transpec         | Number of transactions started for the database per second. Returns 1 if all trans/sec are below the given    |    Yes    |   Yes    |
            |                  | threshold, 0 + list of databases with excess of trans/sec in execution info otherwise.                        |           |          |
            +------------------+---------------------------------------------------------------------------------------------------------------+-----------+----------+
            | db-free          | Free storage percentage per database file groups. Returns 1 if all percenteges are below the given threshold, |    Yes    |   Yes    |
            |                  | 0 + list of saturated file groups in execution info otherwise.                                                |           |          |
            +------------------+---------------------------------------------------------------------------------------------------------------+-----------+----------+
            | databaseonline   | Online databases. Returns 1 if all databases are online, 0 + list of databases with state != ONLINE in        |     No    |   Yes    |
            |                  | execution info otherwise.                                                                                     |           |          |
            +------------------+---------------------------------------------------------------------------------------------------------------+-----------+----------+
            | time2connect     | Time to connecto to the database.                                                                             |     No    |   Yes    |
            +------------------+---------------------------------------------------------------------------------------------------------------+-----------+----------+
            | connections      | Number of open connections to MS SQL server.                                                                  |     No    |   Yes    |
            +------------------+---------------------------------------------------------------------------------------------------------------+-----------+----------+

    **Output**

      Output changes depending on the specified mode.

    **Errors**

      * If an invalide mode is passed to mssql_check, a PollException is raised.
      
      * If, on checks that requires it, threshold value is not specified a PollException is raised.
      
      * Some checks need a former value to compute their result. If it's the first time this kind of check is performed, and so a former value is not present, a PollException is rised in order to allow the computation to continue flawlessly. The second time the check is performed, everything should be fine.

    **Examples**

      +--------------------------------------------------------------------------------------------------+----------------------------------------------+-----------------------+
      | Expression                                                                                       | Description                                  | Return string value   |
      +==================================================================================================+==============================================+=======================+
      | mssql_check(username="sanet", password="sanet", mode="time2connect")                             | Returns time spent to connect to the server. | 0.0164229869843       |
      +--------------------------------------------------------------------------------------------------+----------------------------------------------+-----------------------+
      | mssql_check(username="sanet", password="sanet", mode="db-free", threshold=30)                    | Check if all the databases on                | 1 if they have        |
      |                                                                                                  | MS SQL Server have free space over 30%.      |                       |
      +--------------------------------------------------------------------------------------------------+----------------------------------------------+-----------------------+
      | mssql_check(username="sanet", password="sanet", mode="db-free", threshold=30, instance="tempdb") | Check if database tempdb on MS SQL Server    | 1 if they have        |
      |                                                                                                  | has free space over 30%.                     |                       |
      +--------------------------------------------------------------------------------------------------+----------------------------------------------+-----------------------+
      | mssql_check(username="sanet", password="sanet", mode="databaseonline")                           | Check if all the databases on                | 1 if they are online  |
      |                                                                                                  | MS SQL Server are online.                    |                       |
      +--------------------------------------------------------------------------------------------------+----------------------------------------------+-----------------------+
      | mssql_check(username="sanet", password="sanet", mode="databaseonline", instance="tempdb")        | Check if database tempdb on MS SQL Server    | 1 if they are online  |
      |                                                                                                  | is online.                                   |                       |
      +--------------------------------------------------------------------------------------------------+----------------------------------------------+-----------------------+

"""

import traceback

from sanet_poller.lib.exceptions import PollException

__error = None

try:
	from sanet_poller.externals.mssql.check_mssql_server import _mssql_check_server2008
	from sanet_poller.externals.mssql.check_mssql_server2005 import _mssql_check_server2005
except ImportError, e:
	__error = e

from sanet_common.formatutils import calc_num_prefix

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def db_free_output(context, result, threshold):
	
	threshold = int(threshold)
	
	res = 1
	
	for db, values in result.items():
		
		if type(values) == str:
			res = 0
			context.appendExecutionInfo( values )
			continue
		
		# Data files
		for group, gData in values['ROWS'].items():
			tmp = calc_num_prefix( gData['max_size'], 1024 )
			max_size = (tmp[0], tmp[3])
			
			tmp = calc_num_prefix( gData['used'], 1024 )
			used = (tmp[0], tmp[3])
			free = (100 - 100 * float(gData['used'])/float(gData['max_size']))
			
			if free < threshold:
				res = 0
				context.appendExecutionInfo(db + ' DATA file(s) in group "%s":\n' % group)
				context.appendExecutionInfo('\tmax size: %.2f %sB\n' % max_size)
				context.appendExecutionInfo('\tused: %.2f %sB\n' % used)
				context.appendExecutionInfo('\tfree: %.2f %%\n' % free)
		
		# Log files
		for group, gData in values['LOG'].items():
			tmp = calc_num_prefix( gData['max_size'], 1024 )
			max_size = (tmp[0], tmp[3])
			
			tmp = calc_num_prefix( gData['used'], 1024 )
			used = (tmp[0], tmp[3])
			free = (100 - 100 * float(gData['used'])/float(gData['max_size']))
			
			if free < threshold:
				res = 0
				context.appendExecutionInfo(db + ' LOG file(s) in group "%s":\n' % group)
				context.appendExecutionInfo('\tmax size: %.2f %sB\n' % max_size)
				context.appendExecutionInfo('\tused: %.2f %sB\n' % used)
				context.appendExecutionInfo('\tfree: %.2f %%\n' % free)
	
	return res


def fun_mssql_check2008(context, mode, username, password=None, host=None, port=1433, db_name=None, *args, **kwargs):
	return mssql_check(context, mode, username, password=password, host=host, port=port, db_name=db_name, mssql_version=2008, *args, **kwargs)

def fun_mssql_check2005(context, mode, username, password=None, host=None, port=1433, db_name=None, *args, **kwargs):
	return mssql_check(context, mode, username, password=password, host=host, port=port, db_name=db_name, mssql_version=2005, *args, **kwargs)


def mssql_check(context, mode, username, password=None, host=None, port=1433, db_name=None, mssql_version=2008, *args, **kwargs):

	if __error is not None:
		raise PollException(str(__error))
	
	hostname = host or context['$node']
	database = 'master'
	
	if db_name:
		data = context.getCachedSimpleValue(db_name) or {}
	else:
		data = context.getCachedSimpleValue(database) or {}
	
	dcache = context.getCache()
	logger = context.getLog()
	
	if not mode in data.keys():
	
		try:
			if mssql_version == 2008:
				result = _mssql_check_server2008(dcache=dcache, hostname=hostname, user=username, password=password, port=int(port), mode=mode, database=database, instance=db_name)
			elif mssql_version == 2005:
				result = _mssql_check_server2005(dcache=dcache, hostname=hostname, user=username, password=password, port=int(port), mode=mode, database=database, instance=db_name)

			data[mode] = result
			
		except KeyError:
			raise PollException('Unknown check: %s.' % mode)
		
		else:
			if db_name:
				context.cacheSimpleValue(db_name, data)
			else:
				context.cacheSimpleValue(database, data)
		
	else:
		result = data[mode]
	
	
	if mode == 'databaseonline':
		res = 0
		v_message = []
		
		for elem in result:
			if elem['state'] >= res and elem['state'] != 0:
				res = elem['state']
				v_message.append( elem['verbose'] )
		
		v_message = '\n'.join(v_message)
		
		if res:
			context.appendExecutionInfo(v_message)
			return 0
		else:
			return 1
	
	
	elif mode == 'db-free':
		
		if not 'threshold' in kwargs:
			raise PollException('Please insert a threshold value for this check. Expected a percentage value.')
		
		threshold = kwargs[ 'threshold' ]
		
		return db_free_output(context, result, threshold)
	
	
	elif mode == 'transpec':
		
		if not 'threshold' in kwargs:
			raise PollException('Please insert a threshold value for this check. Expected an integer value.')
		
		threshold = kwargs[ 'threshold' ]
		
		res = 1
		
		for db, tps in result.items():
			
			if type(tps) == str:
				res = 0
				context.appendExecutionInfo( tps )
				continue
			
			if tps > threshold:
				res = 0
				context.appendExecutionInfo(db + ' Transactions/sec: %s' % tps)
		
		return res
	
	else:
		return result


EXPR_SYMBOLS = {
	'mssql_check': fun_mssql_check2008,
	'mssql_check_2008': fun_mssql_check2008,
	'mssql_check_2005': fun_mssql_check2005
}
