# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

Network utils
==============


resolve
---------------------------------------------

    **Syntax**
         
         *resolve* **(** *host* [ **,** *ipversion* ] [**,** *stable_order* ] **)**
         
    **Description**
         
         Resolves the address for the give *host*. The DNS record A or AAAA depends of the *ipversion*.
         
         If there are multiple addresses of *host*, the first address is returned. If the parameter *stable_order* is *true* (default *false*), all the 
         resolved IPs are sorted by network order and the first address (smallest) is always returned).
         
    **Input**
   
         * *host* : an IP address
         * *ipversion*: IP version: 4 (default) or 6.
         * *stable_order*: (optional) *true* or *false*. Default *false*.
    
    **Output**
         
         A string
         
    **Errors**
         
         An exception is raised if an error occurs.
         
    **Examples**
   
         +-------------------------------------+-----------------+--------------------------------------------------------------------------------------+
         | Expression                          | Return value    | Description                                                                          |
         +=====================================+=================+======================================================================================+
         | resolve("www.google.com")           | 173.194.113.230 |                                                                                      |
         +-------------------------------------+-----------------+--------------------------------------------------------------------------------------+

resolveall
---------------------------------------------

    **Syntax**
         
         *resolveall* **(** *host* [ **,** *ipversion* ] [**,** *stable_order* ] **)**
         
    **Description**
         
         Resolves all the address for the give *host*. The DNS record A or AAAA depends of the *ipversion*.
         
         The function returns an array with all resolved IP addresses.
         
         If the parameter *stable_order* is *true* (default), the returned array is sorted by network order criteria.
         
    **Input**
         
         * *host* : an IP address
         * *ipversion*: IP version: 4 (default) or 6.
         * *stable_order*: (optional) *true* or *false*. Default *true*.
         
    **Output**
         
         A vector of strings.
         
    **Errors**
         
         An exception is raised if an error occurs.
         
    **Examples**
         
         +--------------------------------------------------+--------------------------------------------------------------------------------------+--------------------------------------------------------------------------------------+
         | Expression                                       | Return value                                                                         | Description                                                                          |
         +==================================================+======================================================================================+======================================================================================+
         | resolveall("yahoo.it", 4, stable_order=1)        | ['74.6.50.24', '77.238.184.24', '98.137.236.24', '106.10.212.24', '212.82.102.24']   |                                                                                      |
         +--------------------------------------------------+----------------------------------------------------------------------+------------------------------------------------------------------------------------------------------+
         | resolveall("yahoo.it", 4, stable_order=0)        | ['106.10.212.24', '212.82.102.24', '74.6.50.24', '77.238.184.24', '98.137.236.24']   |                                                                                      |
         +--------------------------------------------------+--------------------------------------------------------------------------------------+--------------------------------------------------------------------------------------+
         | resolveall("yahoo.it", 4, stable_order=1)[0]     | '74.6.50.24'                                                                         |                                                                                      |
         +--------------------------------------------------+--------------------------------------------------------------------------------------+--------------------------------------------------------------------------------------+


buff2IP
---------------------------------------------

    **Syntax**
    
         *buff2IP* **(** *string* **)**
    
    **Description**
    
         Convert the buffer of bytes *string* storing a IPv4 address to it's dotted-string rappresentation (es: 192.168.0.24).
    
    **Input**
    
         * *string* : a 4 byte buffer string
    
    **Output**
    
         A string
    
    **Errors**
    
         An exception is raised if the input string does not store an IPv4 address. 
    
    **Examples**
    
         +-------------------------------------+-------------------+--------------------------------------------------------------------------------------+
         | Expression                          | Return value      | Description                                                                          |
         +=====================================+===================+======================================================================================+
         | buff2IP( 1.2.3.6.1.2.3@ )           | "173.194.113.230" | OID's values is a 4-byte buffer storing the IP address *173.194.113.230*.            |
         +-------------------------------------+-------------------+--------------------------------------------------------------------------------------+

"""
import socket

from sanet_common.netutils import resolve_ip_addr

from sanet_poller.lib.exceptions import PollException

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_resolve(context, host, ipversion=4, stable_order=False):

        reason = None
        addrs  = None

        try:
                addrs = resolve_ip_addr(host, ipversion, stable_order=stable_order)
        except socket.timeout, t:
                reason = "reachability NOK: socket timeout on %s: %s" % (host,t)
        except socket.error, e:
                reason = "reachability NOK: socket error on %s: %s" % (host, e)
        except Exception, e:
                reason = "reachability NOK: error on %s: %s" % (host, e)

        if not addrs:
                reason = "No addresses for host %s version %d" % (host, ipversion)

        if reason:
                raise PollException(reason)

        context.appendReason("Hostname %s resolved with: %s" % (host, addrs))

        return addrs[0]

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_resolveall(context, host, ipversion=4, stable_order=True):

        reason = None
        addrs  = None

        try:
                addrs = resolve_ip_addr(host, ipversion, stable_order=stable_order)
                
                print addrs
        except socket.timeout, t:
                reason = "reachability NOK: socket timeout on %s: %s" % (host,t)
        except socket.error, e:
                reason = "reachability NOK: socket error on %s: %s" % (host, e)
        except Exception, e:
                reason = "reachability NOK: error on %s: %s" % (host, e)

        if not addrs:
                reason = "No addresses for host %s version %d (%s)" % (host, ipversion, reason)

        if reason:
                raise PollException(reason)

        context.appendReason("Hostname %s resolved with: %s" % (host, addrs))

        return addrs

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_buff2IP(context, buf):

	return socket.inet_ntop(socket.AF_INET, buf)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

EXPR_SYMBOLS = {
	'resolve': fun_resolve,
	
	'resolveall': fun_resolveall,

	'buff2IP': fun_buff2IP,
}
