"""


NTP
========================================================================

isNtpOk
---------------------------------------------

    **Syntax**

      **isNtpOk** **(** *hostname* **)**

    **Description**

      Checks if the host *hostname* is a valid NTP server.
      The functions performs a NTP request and checks the NTP response. It verifies if:
      * The server's clock is syncronized
      * THe server is among the secondary NTP servers (server *stratum* between 1 and 15).

    **Input**

      *hostname*: the string containing the hostname of the remote server.

    **Output**

      Returns a *true* or *false*.

    **Errors**

      If the hostname does not respond before a 3 seconds timeout, the function returns *false*.

    **Examples**

	+-----------------------+--------------+
	| Expression            | Return value |
	+-----------------------+--------------+
	| isNtpOk("mason")      | *true*       |
	+-----------------------+--------------+
	| isNtpOk("google.com") | *false*      |
	+-----------------------+--------------+

NTPTimeStamp
---------------------------------------------

    **Syntax**

      **NTPTimeStamp** **(** *buffer* **)**

    **Description**

      Converts *buffer* (a 64-bit NTP timestamps [RFC-1305]) to Unix Timestamp UTC (seconds since 1970-01-01).

    **Input**

      *buffer*: the 64-bit string containing a NTP timestamps

    **Output**

      Returns a floating point value.

    **Errors**

      Raise an exception if the input buffer can not be converted to valid timestamp.

    **Examples**

	+-----------------------------------------------------------------------------------+----------------------------------------------+-----------------------------------------------------------------------------------+
	| Expression                                                                        | Return value                                 | Description                                                                       |
	+===================================================================================+==============================================+===================================================================================+
	| NTPTimeStamp( hexstr2bin("0000000000000000" ) )                                   | -2208988800                                  | NTP uses an epoch of 1900-01-01 00:00:00.                                         |
	+-----------------------------------------------------------------------------------+----------------------------------------------+-----------------------------------------------------------------------------------+
	| NTPTimeStamp( hexstr2bin("0000000000000000" ) ) + 2208988800                      | 0                                            | Thu Jan  1 00:00:00 CET 1970 (NTP epoch + seconds to 1970 (2208988800)            |
	+-----------------------------------------------------------------------------------+----------------------------------------------+-----------------------------------------------------------------------------------+
	| NTPTimeStamp( hexstr2bin("C232B916FAE071DB" ) )                                   | 1049115286.98                                | Mon Mar 31 14:54:46 CEST 2003                                                     |
	+-----------------------------------------------------------------------------------+----------------------------------------------+-----------------------------------------------------------------------------------+

"""

import struct
import socket
import select

from sanet_common.exprparser.Types import SnmpValueType

from sanet_common.netutils import resolve_ipv4

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#


def fun_isNtpOk(context, host):	
	
	log = context.getLog()
	
	try:
		recv_data = None
		
		# Don't block while resolving names
		addrs = resolve_ipv4(host, False, 10)
		if addrs == None or len(addrs) < 1:
			# Fix: Possible without exceptions?
			reason = "NOK: unable to resolve host name %s" % (host)
			log.debug(reason)
			context.appendReason(reason)
			return False
		
		ip = addrs[0]
		s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, 0)		
		s.connect( (ip, 123) )
	
		# Prepare NTP packet for probe
		data = struct.pack("B47s", 0x1b, ('\x00' * 47))
		
		# Send probes!
		for i in xrange(3):
			#FIX: do it better? Exceptions from send() or select()?
			s.send(data)
			rd, wt, er = select.select([s], [], [s], 3.0)
			if rd:
				try:
					recv_data, peer_addr = s.recvfrom(4096)
				except socket.error, er:
					log.error("%s", er)
					continue
				
				break
		s.close()
		
		if recv_data == None:
			reason = "No NTP data received"
			log.debug(reason)
			context.appendReason(reason)
			return False
		
		#
		# Read response header's fields
		#
		# http://www.ietf.org/rfc/rfc5905.txt
		#
		#   0                   1                   2                   3
		#   0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
		#  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
		#  |LI | VN  |Mode |    Stratum     |     Poll      |  Precision   |
		#  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
		#  |                       ... 11 word   ...                       |
		#  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
		#  |                       ... optionals ...                       |
		#
		#  
		#     LI Leap Indicator (leap): 2-bit integer warning of an impending leap
   		#  second to be inserted or deleted in the last minute of the current
   		#  month.
		#
		#          | 0     | no warning                             |
		#          | 1     | last minute of the day has 61 seconds  |
		#          | 2     | last minute of the day has 59 seconds  |
		#          | 3     | unknown (clock unsynchronized)         |
		#           
		#     Stratum (stratum): 8-bit integer representing the stratum:
		#
		#          | 0      | unspecified or invalid                              |
		#          | 1      | primary server (e.g., equipped with a GPS receiver) |
		#          | 2-15   | secondary server (via NTP)                          |
		#          | 16     | unsynchronized                                      |
		#          | 17-255 | reserved                                            |		
		t = struct.unpack('!4B11I', recv_data)[0]
		stratum = struct.unpack('!4B11I', recv_data)[1]
		lih = (t & (2**7)) > 0
		lil = (t & (2**6)) > 0
		li = 2*lih + lil
		# (b1, stratum, dummy, precision) = struct.unpack("BBBb", recv_data[0:4])
		
		if stratum == 0:
			stratum = 16
		
		reason = "li=%d, stratum=%d" % (li, stratum)
		log.debug(reason)
		context.appendReason(reason)
		
		# Not unsynchronized and among secondary servers
		if (li != 3 and (stratum > 0) and (stratum < 15)):
			return True
		else:
			return False
		
	except socket.timeout, t:
		# resolve_ipv4()?
		reason = "NOK: socket timeout on %s: %s" % (host, t)
	except socket.error, e:
		# resolve_ipv4()?
		reason = "NOK: socket error on %s: %s" % (host, e)
	except Exception, e:
		# resolve_ipv4()?
		reason = "network error on %s: %s" % (host, e)
		
	log.debug(reason)
	context.appendReason(reason)
	
	return False
	

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

#
# One is the offset between the two epochs. Unix uses an epoch located at 1/1/1970-00:00h (UTC) 
# and NTP uses 1/1/1900-00:00h. This leads to an offset equivalent to 70 years 
# in seconds (there are 17 bisiest years between the two dates so the offset is
#
# (70*365 + 17)*86400 = 2208988800
#
NTP_OFFSET = 2208988800

def fun_NTPTimeStamp(context, data):

	#log = context.getLog()

	if isinstance(data, SnmpValueType):
		data = data.getvalue()

	# 
	# NTPTimeStamp ::= OCTET STRING(SIZE(8))
	# -- TEXTUAL-CONVENTION
	# --  DspHint
	# --    4d.4d
	# --  Status
	# --    mandatory
	# --  Descr
	# --    NTP timestamps are represented as a 64-bit
	# --    unsigned fixed-point number, in seconds relative to
	# --    00:00 on 1 January 1900.  The integer part is in the
	# --    first 32 bits and the fraction part is in the last
	# --    32 bits.
	# --  Refer
	# --    D.L. Mills, 'Network Time Protocol (Version 3)',
	# --    RFC-1305, March 1992, Section 3.1

	integer_part,frac_part = struct.unpack("!LL", data)

	#log.debug("Integer part: %s", integer_part)
	#log.debug("Frac    part: %s", frac_part   )

	seconds  = float(integer_part - NTP_OFFSET)

	#
	# NTP uses 1/2^32 sec as its unit of fractional time.
	#
	seconds += frac_part * (1.0 / (2**32))

	#log.debug("Seconds: %s", seconds)

	return seconds

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

EXPR_SYMBOLS = {
	'isNtpOk'               : fun_isNtpOk, #tested
	'NTPTimeStamp'          : fun_NTPTimeStamp,
}

