"""

Oracle
==============

oracle_raw_query
------------------------

    **Syntax**
    
      *oracle_raw_query* **(** *dsn* **,** *username* **,**  *password* **,** *query* **)**

    **Description**

      Connects to the given *dsn* oracle instance and executes the given *query* with the give *username*/*password*.

    **Input**

      * *dsn*: oracle instance identifier
      * *username*: string
      * *password*: string
      * *query*: a valid SQL command

    **Output**

      Returns the first column of the first row of the returned record set.

    **Errors**
    
      An exception is raised when an error occurs.

    **Examples**

      +-----------------------------------------------------------------------------------------+--------------------------+------------------------------------------------------+
      | Expression                                                                              | Return string value      | Description                                          |
      +=========================================================================================+==========================+======================================================+
      | oracle_raw_query("dsn1", "user", "password", "SELECT count(*) FROM table1")             | *9*                      |                                                      |
      +-----------------------------------------------------------------------------------------+--------------------------+------------------------------------------------------+

oracle_dsn
------------------------

TODO

"""
import time

from sanet_poller.lib.exceptions import PollException

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

__cx_oracle_import_info = None
try:
	import cx_Oracle
except ImportError, e:
	__cx_oracle_import_info = e

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

#import threading
#_GLOBAL_LOCK = threading.Semaphore()

def __oracle_raw_query(context, dsn, username, password, query):

	dsn = str(dsn)
	username = str(username)
	password = str(password)
	query = str(query)

	log = context.getLog()

	try:
		log.debug('oraclequery: connetting...')
		#print dsn
		#print username
		#print password
		#print query 

		#log.info("ORACLE LOCK: %s", [ dsn, username, password, query ] )
		#_GLOBAL_LOCK.acquire()

		oracle_connection = cx_Oracle.connect(user=username, password=password, dsn=dsn, threaded=True)

		log.debug('oraclequery: db version: %s', oracle_connection.version)

		log.debug('oraclequery: opening cursor')
		oracle_cursor = oracle_connection.cursor()

		log.debug('oraclequery: executing query')

		log.debug('---')
		log.debug(query)
		log.debug('---')

		oracle_cursor.execute(query)
		ret = oracle_cursor.fetchall()
		if ret:
			log.debug('oraclequery: query result = %s', ret[0][0])
			retval = ret[0][0]

			log.debug('oraclequery: RESULT: %s', retval)

			return retval
		else:
			raise Exception("The query returned 0 records")

		log.debug('oraclequery: closing connection...')

		oracle_cursor.close()
		oracle_connection.close()

		log.debug('oraclequery: done.')

	except Exception, e:
		msg = str(e)

		log.debug(msg)

		raise Exception("Exception while executing query (type %s): %s" % ( type(e), msg) )
	finally:
		#log.info("ORACLE UNLOCK: %s", [  dsn, username, password, query ] )
		#_GLOBAL_LOCK.release()
		pass


def __oracle_raw_query2(context, dsn, username, password, query, return_type='value'):

	log = context.getLog()

	cache_id = '.'.join([dsn, username, password, query])

	check_data = context.getCachedSimpleValue(cache_id)
	if check_data == None:

		ts_start = time.time()
		try:
			retval = __oracle_raw_query(context, dsn, username, password, query)
		except Exception, e:
			log.warning("Query returned exception: %s %s. Query: %s" % (type(e), str(e), query))
			retval = e
		ts_end   = time.time()

		check_data = {
			'value'    : retval,
			'exectime' : ts_end - ts_start
		}

		# 
		# Cache value in the execution cache
		#
		context.cacheSimpleValue(cache_id, check_data)
	else:
		log.debug("Using cached value for query: %s", query)

	if return_type == 'value':

		v = check_data['value']
		if isinstance(v, Exception):
			raise v
		return v

	elif return_type == 'exectime':
		return check_data['exectime']

	else:
		raise PollException("Invalid parameter return_type='%s'" % (return_type))

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def __oracle_dsn(context, hostname, port, sid):
	#
	# Funzione creata per il ticket #1788, ma mai usata
	#
	return cx_Oracle.makedsn(hostname, str(port), sid)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#


if __cx_oracle_import_info != None:

	def __oracle_raw_query(*args,**kwargs):
		raise PollException("python module cx_Oracle not imported or not installed: %s" % (__cx_oracle_import_info) )

	def __oracle_raw_query2(*args,**kwargs):
		raise PollException("python module cx_Oracle not imported or not installed: %s" % (__cx_oracle_import_info) )

	def __oracle_dsn(*args, **kwargs):
		raise PollException("python module cx_Oracle not imported or not installed: %s" % (__cx_oracle_import_info) )

#
# Nomi funzioni visibili richiamabili dal poller
#
EXPR_SYMBOLS = {
	'oracle_raw_query'  : __oracle_raw_query,
	'oracle_raw_query2' : __oracle_raw_query2,
	'oracle_dsn'        : __oracle_dsn,
}

