"""
==================
Postgres functions
==================

psql_query
---------------------------------------------

    **Syntax**

        **psql_query** **(** *dbname* **,** *query* [ **,** *host* ] [**,** *port* ] [**,** *user*] [**,** *password*]  [**,** *timeout*] **)**     

    **Description**

        Connect to the database dbname and executes the given query and returns the first row of the recordset. 

    **Input**

        * dbname   - the database name
        * host     - database host address (defaults to UNIX socket if not provided)
        * port     - connection port number (defaults to 5432 if not provided)
        * user     - user name used to authenticate
        * password - password used to authenticate

    **Output**

        A tuple of values (the first row of the queryset) 

    **Errors**

        An exception is raised on connect timout or when the query returns an error.

    **Examples**

        +----------------------------------------------------------------------------+------------------------------------------------------------------------+---------------------+
        | Expression                                                                 | Description                                                            | Return string value |
        +============================================================================+========================================================================+=====================+
        | psql_query("mydb", "SELECT * FROM table1")                                 | Connect to localhost and returns the first row of the queryset.        |                     |
        | psql_query("mydb", "SELECT * FROM table1")[0]                              | Returns the first column value of the first row of the queryset.       |                     |
        | psql_query("mydb", "SELECT * FROM table1", user="foo", password="bar")[0]  | Returns the first column value of the first row of the queryset.       |                     |
        +----------------------------------------------------------------------------+------------------------------------------------------------------------+---------------------+
"""

import psycopg2

from sanet_poller.lib.exceptions import PollException

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

PSQL_DEFAULT_TIMEOUT = 3

def fun_psql_query(context, dbname, query, host=None, port=None, user=None, password=None, timeout=None):
	
	if timeout is None:
		timeout = PSQL_DEFAULT_TIMEOUT

	connection_options = {
		'database'       : dbname,
		'connect_timeout': timeout,
	}
	
	if host     is not None: connection_options['host'    ] = host
	if port     is not None: connection_options['port'    ] = port
	if user     is not None: connection_options['user'    ] = user
	if password is not None: connection_options['password'] = password
	if timeout  is not None: connection_options['connect_timeout'] = timeout

	try:
		conn = psycopg2.connect(**connection_options)
	except Exception, e:
		raise PollException("Unable to connect to postgresql server: %s" % (e))
	
	cur = conn.cursor()
	
	cur.execute(query)
	
	rows = cur.fetchall()
	if rows:
		return rows[0]
	else:
		raise PollException("Query returned 0 records")

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

EXPR_SYMBOLS = {
	'psql_query' : fun_psql_query, 
}

