# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it> 
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

isReachable
---------------------------------------------

    **Syntax**

      **isReachable** **(** *host* [ **,** *probesize* [ **,** *shorttries* [ **,** *keep_rtt_measure* [ **,** *timeout* [ **,** *do_frag* [ **,** *bind_address*]]]]]] **)**

    **Description**

      Tests if a remote host identified by *host* responds to ICMP ECHO requests. 

      The parameter *host* can be a string containing a valid IPv4 address or a hostname string. If an hostname is given, it is resolved and the reachability test is performed for every resolved IP until the remote host responds or the test fails.

      If the parameters *probesize*, *shorttires* and *timeout* are not specified then the function looks for the corresponding variables *$probesize*, *$shorttries* and *$timeout* inside the current *execution context*. If they are missing, the test raises an exceptions and the expression evaluation fails.

      If the *keep_rtt_measure* parameter's value is *true*, the max, min and average RTT values calculated during the test are stored inside a RRD file with name: *host*.rrd. If the file does not exists, it will be created on the fly. (TODO: link to the file format)

      If the *do_frag* parameter is *false*, the IP packets are sent with the DF flag set and they are not fragmented. Default value is *True*.

      if *bind_address* is specified, outgoing IP packets will use *bind_address* as the source address.

    **Input**

      * *hostname*        : string. 
      * *probesize*       : integer number. 
      * *shorttries*      : interger number
      * *keep_rtt_measure*: boolean value. **DEPRECATED: not used. for compatibility only**
      * *timeout*         : integer number. (default: 3)
      * *do_frag*         : boolean value. (default: true)
      * *bind_address*    : string

    **Output**

      Returns a *true* or *false*.

    **Errors**

      If the parameter *host* is not valid or can not be translated into a valid IP(s) the function returns *false*.

      If the value for 

    **Examples**

	 +----------------------------------------------------------+----------------------------------------------------------------------------------------------------+
	 | Expression                                               | Description                                                                                        |
	 +----------------------------------------------------------+----------------------------------------------------------------------------------------------------+
	 | isReachable("mason")                                     | Pings host "mason".                                                                                |
	 +----------------------------------------------------------+----------------------------------------------------------------------------------------------------+
	 | isReachable($hostname)                                   |                                                                                                    |
	 +----------------------------------------------------------+----------------------------------------------------------------------------------------------------+
	 | isReachable("mason", 1472)                               |                                                                                                    |
	 +----------------------------------------------------------+----------------------------------------------------------------------------------------------------+
	 | isReachable("mason", $packetlen)                         |                                                                                                    |
	 +----------------------------------------------------------+----------------------------------------------------------------------------------------------------+
	 | isReachable("mason", 1472, 3)                            |                                                                                                    |
	 +----------------------------------------------------------+----------------------------------------------------------------------------------------------------+
	 | isReachable("mason", 1472, $tries)                       |                                                                                                    |
	 +----------------------------------------------------------+----------------------------------------------------------------------------------------------------+
	 | isReachable("mason", 1472, 3, 1)                         |                                                                                                    |
	 +----------------------------------------------------------+----------------------------------------------------------------------------------------------------+
	 | isReachable("mason", 1472, 3, False)                     | Disabled keep_rtt_measure.                                                                         |
	 +----------------------------------------------------------+----------------------------------------------------------------------------------------------------+
	 | isReachable("mason", 1472, 3, False, 3)                  | Disabled keep_rtt_measure, 3 seconds of timeout for every ICMP packet.                             |
	 +----------------------------------------------------------+----------------------------------------------------------------------------------------------------+
	 | isReachable("mason", 1472, 3, False, $timeout)           |                                                                                                    |
	 +----------------------------------------------------------+----------------------------------------------------------------------------------------------------+
	 | isReachable("mason", 1472, 3, False, $timeout, False)    | Sends 3 ICMP packets. The ICMP packets' payload is 1472 bytes and the IP packet is not fragmented. |
	 +----------------------------------------------------------+----------------------------------------------------------------------------------------------------+
	 | isReachable("mason", 1472, bind_address="127.0.0.1")     | Tries to ping host 'mason' with source address 127.0.0.1. It will fail.                            |
	 +----------------------------------------------------------+----------------------------------------------------------------------------------------------------+
	 | isReachable("localhost", 1472, bind_address="127.0.0.1") | Tries to ping localhost with source address 127.0.0.1. Valid address.                              |
	 +----------------------------------------------------------+----------------------------------------------------------------------------------------------------+
 

isReachable6
---------------------------------------------

     **Syntax**

      **isReachable6** **(** *host* [ **,** *probesize* [ **,** *shorttries* [ **,** *keep_rtt_measure* [ **,** *timeout* [ **,** *bind_address* ]]]]]]]] **)**

     **Description**

      This function perform the same reachability test of the **isReachable()** function, but uses ICMPv6 instead of ICMPv4.

      See the documentation of **isReachable()** for more details.

icmpPmtu
---------------------------------------------

    **Syntax**

      **icmpPmtu** **(** *host* [ **,** *probesize* [ **,** *shorttries* [ **,** *timeout*]]] **)**

    **Description**

     Checks if the PMTU to a remote host is enough for an IPv4 ICMP ECHO requests with paylod *probesize*. 

     The host is pinged with DF bit set, and if it answers the value is *true*. If it doesn't answer (after *shorttries* attempts with *timeout*), host is pinged with DF bit clear, and if it answers the values is *false*. If it doesn't answer even with DF clear, the value is *true*. In other words:

     * If the host answers to pings with the specified payload, but the PMTU doesn't allow them through, the value is *false*
     * If the host is alive but doesn't anser to pings with the specified paylod, whatever DF bit we use, the value is *true* (i.e., it is not a PMTU problem)
     * If the host is dead, the value is *true* (again, it is not a PMTU problem)

     If the host has multiple IPv4 addresses, each of them is tried with DF bit set, and at the first answer the value is *true*. Then each of them is tried with DF clear, and at the first answer the value is *false*. If there is no answer at all, the value is *true*.


packetLoss
---------------------------------------------

    **Syntax**

      **packetLoss** **(** *host* [ **,** *probesize* [ **,** *shorttries* [ **,** *keep_rtt_measure* [ **,** *timeout* [ **,** *do_frag*]]]]] **)**

    **Description**

     Calculates the percentage of packet loss for the given *host*.

     The functions accepts the same parameters (with the same semantics) used for *isReachable()*.

    **Input**

      *hostname* : string
      *probesize*: integer number
      *shorttries*: interger number
      *keep_rtt_measure*: boolean value
      *timeout*: integer number
      *do_frag*: boolean value

    **Output**

      Returns packet loss percentage between 0 and 1.

    **Errors**

      If the parameter *host* is not valid or can not be translated into a valid IP(s) the function returns *false*.

      If the value for 

    **Examples**

	+-------------------------------------+---------------------------+
	| Expression                          | Description               |
	+-------------------------------------+---------------------------+
	| packetLoss("mason")                 | 0.33~                     |
	+-------------------------------------+---------------------------+
	| isReachable("www.google.com", 1472) | 0.0                       |
	+-------------------------------------+---------------------------+
 


"""
import time
import sys, array, types, re
import subprocess, os
import string
import random
from datetime import datetime

# socket & HTTP protocol
import socket, select
import urllib2

import ipaddr
	
# DNS 
from sanet_common.netutils import resolve_ip
from sanet_common.netutils import Ping

from sanet_poller.lib.exceptions import PollException

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

# Default values in seconds
REACHABLE_DEFAULT_TIMEOUT = 3
REACHABLE_MIN_TIMEOUT = 1
REACHABLE_MAX_TIMEOUT = 5

def hostIsReachable(log, 
                    ipversion, 
                    host, 
                    probesize=None, 
                    shorttries=None, 
                    #keep_rtt_measure=None, 
                    timeout=None, 
                    #sanet_node=None, 
                    #sanet_target=None, 
                    do_frag=True, 
                    bind_address=None
):
	
	#if keep_rtt_measure == None: 
	#	keep_rtt_measure = 1
		
	if timeout == None: 
		timeout = REACHABLE_DEFAULT_TIMEOUT
	
	if timeout < REACHABLE_MIN_TIMEOUT: timeout = REACHABLE_MIN_TIMEOUT
	if timeout > REACHABLE_MAX_TIMEOUT: timeout = REACHABLE_MAX_TIMEOUT
	
	#log.debug('isReachable(host=%s, probesize=%s, shorttries=%s, keep_rtt=%s, timeout=%s, do_frag=%s, bind_address=%s)', host, probesize, shorttries, keep_rtt_measure, timeout, do_frag, bind_address)
	log.debug('isReachable(host=%s, probesize=%s, shorttries=%s, timeout=%s, do_frag=%s, bind_address=%s)', host, probesize, shorttries,timeout, do_frag, bind_address)

	(result, loss, details) = check_reachability(log, 
	                                             ipversion,
	                                             host,
	                                             host,
	                                             int(probesize),
	                                             #keep_rtt_measure,
	                                             int(shorttries),
	                                             timeout,
	                                             do_frag,
	                                             #sanet_node,
	                                             #sanet_target,
	                                             bind_address=bind_address
	)
	
	return (result, loss, details)

def icmpPmtu(log, ipversion, host, probesize=None, shorttries=None, timeout=REACHABLE_DEFAULT_TIMEOUT, sanet_node=None, sanet_target=None):
	if timeout < REACHABLE_MIN_TIMEOUT: timeout = REACHABLE_MIN_TIMEOUT
	if timeout > REACHABLE_MAX_TIMEOUT: timeout = REACHABLE_MAX_TIMEOUT
	dfdetails=''
	for i in range(0,shorttries):
		(result, loss, details) = check_reachability(log, 
			ipversion,
			host,
			host,
			int(probesize),
			0,
			1,
			timeout,
			0,
			sanet_node,
			sanet_target,
		)
		if result:
			return(result, 'Answer with DF: ' + details)
	dfdetails = details
	for i in range(0,shorttries):
		(result, loss, details) = check_reachability(log, 
			ipversion,
			host,
			host,
			int(probesize),
			0,
			1,
			timeout,
			1,
			sanet_node,
			sanet_target,
		)
		if result:
			return(False, 'No answer with DF set, but answer with DF clear: ' + dfdetails + ', ' + details)
	return (True, 'No answer, with or without DF set: ' + dfdetails + ', ' + details)
	
	
def check_reachability(log, 
	ipversion, 
	name, 
	host, 
	probesize, 
	#keep_rtt_measure, 
	shorttries, 
	timeout, 
	do_frag, 
	#sanet_node, 
	#sanet_target, 
	bind_address=None
):
	#
	# FIX: when host == name ?
	#
	verbose = True
	delay = 0
	preload = 0
	
	#
	# Get host addresses
	#
	addrs = None
	reason = None
	try:
		host = str(host)
		#if settings.POLLER_ASYNC_RESOLVE_HOSTNAME:
		#	addrs = resolve_ip(host, ipversion, False, 10)		
		#else:
		#addrs = resolve_ip_addr(host, ipversion)
		print host
		addrs = resolve_ip(host, ipversion, False, 10)		
		
	except socket.timeout, t:
		reason = "reachability NOK: socket timeout on %s: %s" % (host,t)
		log.debug(reason)
	except socket.error, e:
		reason = "reachability NOK: socket error on %s: %s" % (host, e)
		log.debug(reason)
	except Exception, e:
		reason = "reachability NOK: error on %s: %s" % (host, e)
		log.debug(reason)
		
	if not addrs:
		reason = "No addresses for host %s version %d" % (host, ipversion)

	if reason:
		return (False, 1.0, "Cannot resolve %s: %s" % (host, reason))

	fisher_yates_shuffle(addrs)
	
	log.debug("%s has %s addresses: %s", host, len(addrs), addrs)
	
	# Calculate round trip times
	for addr in addrs:
		(rttavg, rttmin, rttmax, loss_avg, infos) = do_rtt_measuring(log, ipversion, addr, shorttries, delay, preload, probesize, timeout, do_frag, bind_address=bind_address)
		if loss_avg < 1: # pinged!
			break
		
	if infos:
		infos_text = ' Info:' + ','.join(infos)
	else:
		infos_text = ''
		
	details = "host %s, addr %s, IPv%s, latency: %.3f ms avg, %.3f ms min, %.3f ms max. Loss: %.2f %%.%s" % (host, addr, ipversion, rttavg*1000, rttmin*1000, rttmax*1000, loss_avg*100, infos_text)
	
	if loss_avg == 0:
		log.debug( "target %s pinged: %s", name, details )
	elif loss_avg == 1:	# 100% of packet lost
		details = "host %s unreachable, addr %s: %s" % (host, string.join(addrs,','), details)
		log.debug( "target %s: %s", name, details )
	elif loss_avg > 0:	# some of packet lost
		log.debug( "target %s: %s", name, details )

	return ( True if loss_avg<1 else False , loss_avg, details)

def do_rtt_measuring(log, ipversion, target, num_probes, delay, preload, size, timeout, do_frag, bind_address=None): 
	i = 0
	dmin = 3600
	dmax = 0
	davg = 0
	num_lost = 0
	
	# For every probe we get  an  rtt and a verbose info about errors or timeouts
	
	if ipversion == 4:
		(recv_rrts, infos) = do_rtt_measuring_ipv4(log, target, num_probes, delay, preload, size, timeout, do_frag, bind_address=bind_address)
	elif ipversion == 6:
		(recv_rrts, infos) = do_rtt_measuring_ipv6(log, target, num_probes, delay, preload, size, timeout, do_frag, bind_address=bind_address)
	else:
		raise PollException("do_rtt_measuring: Invalid IP version %s" % ipversion)
	
	for i in xrange(len(recv_rrts)):
		rtt = recv_rrts[i]
		if rtt != None:
			log.debug("%s: got ping answer #%s in %s s", target, i, rtt )
			dmin = min(rtt, dmin)
			dmax = max(rtt, dmax)
			davg += rtt
		else:
			log.debug("%s: lost ping answer #%s: %s", target, i, infos[i] )
			num_lost += 1
		i += 1

	if num_lost == num_probes:
		davg = timeout;
		dmin = timeout;
		dmax = timeout;
	else:
		davg = float(davg) / float((num_probes - num_lost))
		
	# remove unnecessary infos
	infos = [i for i in infos if i ]

	return (davg, dmin, dmax, float(num_lost) / float(num_probes), infos)


def do_rtt_measuring_ipv4(log, target, num_probes, delay, preload, size, timeout, do_frag, bind_address=None): 
	
	probe_rtts = []
	probe_infos = []
	
	log_callback=(lambda x: log.debug("ping %s: %s" % (target, x)) )
	log_callback = None
	
	if preload > 0:
		log.debug("%s: Sending IPv4 preload packets", target )
		p = Ping(target, size=size, timeout=timeout, count=preload, callback=log_callback, do_frag=do_frag, bind_address=bind_address)
		rrts = p.ping()
		p.close()

	log.debug("%s: Sending IPv4 pings.", target)
	
	try:
		p = Ping(target, size=size, timeout=timeout, count=num_probes, callback=log_callback, do_frag=do_frag, bind_address=bind_address)
		(probe_rtts, probe_infos) = p.ping()
		p.close()
	except socket.error, e:
		for i in xrange(num_probes):
			probe_rtts.append(None)
			probe_infos.append(str(e))
	
	return (probe_rtts, probe_infos)
	

def do_rtt_measuring_ipv6(log, target, num_probes, delay, preload, size, timeout, do_frag, bind_address=None): 
	
	# do-frag is currently ignored
	probe_rtts = []
	
	ipversion = 6
	
	if preload > 0:
		log.debug("%s: Sending IPv6 preload packets", target )
		for i in xrange(preload):
			exec_ping(log, 6, target, size, timeout, bind_address=bind_address);	
	
	log.debug("%s: Sending IPv6 pings.", target)
	for i in xrange(num_probes):
		t0 = datetime.now()
		pinged = exec_ping(log, ipversion, target, size, timeout, bind_address=bind_address)
		delta = datetime.now() - t0
		if pinged:
			probe_rtts.append(delta.seconds + (delta.microseconds / 1000000.0))
		else:
			probe_rtts.append(None)
	
	probe_infos = ['' for n in xrange(len(probe_rtts))]
	
	return (probe_rtts, probe_infos)

def exec_ping(log, ipversion, target, size, timeout, bind_address=None):
	
	if ipversion == 4:  
		ping='ping'
	elif ipversion == 6: 
		ping='ping6'
	else:
		raise ValueException("wrong version of ping = %s" % ipversion)
			
	# Examples:
	#
	# root@localhost# ping -c 1 -s 1472 -A -W 3 -n 1.0.0.1
	# PING 1.0.0.1 (1.0.0.1) 1472(1500) bytes of data.
	# 1480 bytes from 1.0.0.1: icmp_seq=1 ttl=64 time=1.35 ms
	# ...
	#
	# root@localhost:~# ping6 -c 3 -s 1452 -A -l 4 -n -W 3 2001:760:2E00:AFFE::1
	# PING 2001:760:2E00:AFFE::1(2001:760:2e00:affe::1) 1423 data bytes
	# 1460 bytes from 2001:760:2e00:affe::1: icmp_seq=1 ttl=64 time=2.00 ms
	# ...
 
	command = "%s -c 1 -s %s -A -W %s -n" % (ping, size, timeout)

	if bind_address != None:
		command += " -I '%s'" % (bind_address)

	command += " '%s'" % (target)
	
	
	log.debug("exec_ping: %s", command)
	
	process = subprocess.Popen(command, stdout=subprocess.PIPE, shell=True)
	os.waitpid(process.pid, 0)
	output = process.stdout.read().strip()
	
	if re.search(" bytes from ", output):
		return True
	else:
		return False
	
def fisher_yates_shuffle(array):
	return random.shuffle(array)







def fun_isReachable(context, host, probesize=None, shorttries=None, keep_rtt_measure=None, timeout=None, do_frag=True, bind_address=None):	
	"""
	TODO 
	
	:param  context:  Current execution context.
	:type context:  :class:`pollerlib.ExecutionContext`
	
	:returns: True or False
	:rtype: boolean
	
	:exception: 
	        :class:`PollException` on internal error
	"""
	
	if probesize  == None: probesize  = context.get('$probesize' , None)  # 0 len payload allowed
	if shorttries == None: shorttries = context.get('$shorttries', None)
	if timeout    == None: timeout    = context.get('$timeout'   , None)

	if probesize  == None: raise PollException('Parameter $probesize missing.')
	if shorttries == None: raise PollException('Parameter $shorttries missing.')
	
	if shorttries == 0: raise PollException('Parameter $shorttries=%s invalid.' % shorttries)

	sanet_node   = context.get('__node__' , None)
	sanet_target = context.get('__check__', None)
	
	(val,loss,info) = hostIsReachable(context.getLog(), 
	                                  4, 
	                                  host, 
	                                  probesize=probesize, 
	                                  shorttries=shorttries, 
	                                  #keep_rtt_measure=keep_rtt_measure, 
	                                  timeout=timeout, 
	                                  #sanet_node=sanet_node, 
	                                  #sanet_target=sanet_target, 
	                                  do_frag=do_frag, 
	                                  bind_address=bind_address
	                                 )
	context.appendReason(info)
	
	return val

def fun_isReachable6(context, host, probesize=None, shorttries=None, keep_rtt_measure=None, timeout=None, bind_address=None):
	"""
	TODO 
	
	:param  context:  Current execution context.
	:type context:  :class:`pollerlib.ExecutionContext`
	
	:returns: True or False
	:rtype: boolean
	
	:exception: 
	        :class:`PollException` on internal error
	"""
	if probesize  == None: probesize  = context.get('$probesize' , None)  # 0 len payload allowed
	if shorttries == None: shorttries = context.get('$shorttries', None)
	if timeout    == None: timeout    = context.get('$timeout'   , None)

	if probesize  == None: raise PollException('Parameter $probesize missing.')
	if shorttries == None: raise PollException('Parameter $shorttries missing.')
	
	if shorttries == 0: raise PollException('Parameter $shorttries=%s invalid.' % shorttries)
	
	sanet_node   = context.get('__node__' , None)
	sanet_target = context.get('__check__', None)

	(val,loss,info) = hostIsReachable(context.getLog(), 
	                                  6, 
	                                  host, 
	                                  probesize=probesize, 
	                                  shorttries=shorttries, 
	                                  #keep_rtt_measure=keep_rtt_measure, 
	                                  timeout=timeout, 
	                                  #sanet_node=sanet_node, 
	                                  #sanet_target=sanet_target, 
	                                  bind_address=bind_address
	                                 )
	
	context.appendReason(info)
	
	return val

def fun_icmpPmtu(context, host, probesize=None, shorttries=None, timeout=None):
	if probesize  == None: probesize  = context.get('$probesize' , None)  # 0 len payload allowed
	if shorttries == None: shorttries = context.get('$shorttries', None)
	if timeout    == None: timeout    = context.get('$timeout'   , None)

	if probesize  == None: raise PollException('Parameter $probesize missing.')
	if shorttries == None: raise PollException('Parameter $shorttries missing.')
	
	if shorttries == 0: raise PollException('Parameter $shorttries=%s invalid.' % shorttries)
	
	(val,info) = icmpPmtu(context.getLog(), 4, host, probesize, shorttries, timeout)
	
	context.appendReason(info)
	
	return val


def fun_packetLoss(context, host, probesize=None, shorttries=None, keep_rtt_measure=None, timeout=None):	
	""" DO NOT USE THIS!!! TESTING """
	
	if probesize  == None:
		probesize  = context.get('$probesize', None)
	if shorttries == None:
		shorttries =  context.get('$shorttries', None)

	if probesize  == None: raise PollException('Parameter $probesize missing or invalid: %s' % (probesize))
	if shorttries == None: raise PollException('Parameter $shorttries missing or invalid: %s' % (shorttries))

	(val,loss,info) = hostIsReachable(context.getLog(), 4, host, probesize, shorttries, False, timeout)

	context.appendReason(info)

	return loss





EXPR_SYMBOLS = {
	'isReachable'           : fun_isReachable,
	'isReachable6'          : fun_isReachable6,
	'icmpPmtu'              : fun_icmpPmtu,
}
