# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it> 
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

==========
RegExpr
==========

byBinaryValue
---------------------------------------------

    **Syntax**

       **byBinaryValue** **(** *snmpexpr* **,** *str hex* **)**

    **Description**

       Scans the MIB identified by the SNMP expression *snmpexpr*. The *binary* value of every entry in the mib is compared with the hexadecimal string stored in the parameter *str hex* and the first matching sub-OID is returned.

    **Input**

       *snmpexpr* : A SNMP expression indentifing a valid MIB.
       *str hex* : A string containing a valid hexadecimal string.

    **Output**

       Returns a *string* OID.

    **Errors**

      The function raises an exception when the matching sub-OID is not found.

    **Examples**

	 +------------------------------------------------------+---------------+------------------------------------------------------------------------------------------+
	 | Expression                                           | Return value  | Description                                                                              |
	 +======================================================+===============+==========================================================================================+
	 |  byBinaryValue(1.3.6.1.2.1.2.2.1.6@, "fefd00000000") | *2*           | It returns the ifPhysAddress  sub-oid of the interface with MAC address "fefd00000000".  |
	 +------------------------------------------------------+---------------+------------------------------------------------------------------------------------------+
 

byRegexpUnique
---------------------------------------------

    **Syntax**

      **byRegexpUnique** **(** *snmpexpr* **,**  *str_regexpr* **)**

    **Description**

      Scans the MIB identified by the SNMP expression *snmpexpr*. The value of every entry in the mib is compared with the regular expression stored in the parameter *str_regexpr* and the first matching sub-OID is returned.

    **Input**

      *snmpexpr* : A SNMP expression indentifing a valid MIB.
      *snmpexpr* : A string containing a valid regular expression.

    **Output**

      Returns a *string* OID.

    **Errors**

      The function raises an exception when the matching sub-OID is not found.

    **Examples**

	 +---------------------------------------------+--------------+-----------------------------------------------------------------------------------------------------------------------+
	 | Expression                                  | Return value | Description                                                                                                           |
	 +=============================================+==============+=======================================================================================================================+
	 | byRegexpUnique(1.3.6.1.2.1.2.2.1.@, "eth0") | *2.4*        | It returns the ifDescr sub-oid for the interface matching the string "eth0" ( 1.3.6.1.2.1.2.2.1.**2.4** = *eth0* ).   |
	 +---------------------------------------------+--------------+-----------------------------------------------------------------------------------------------------------------------+
 


byMultiRegexpUnique
---------------------------------------------

    **Syntax**

      **byMultiRegexpUnique** **(** snmparg **,** [ *suboid1* **,** *regexpr1* ] **,** [ *suboid2* **,** *regexpr2* ] **,** ... **)**

    **Description**

      Scans the MIB identified by the SNMP expression *snmparg* and verifies all the sub-oids <suboid1>, <suboid2>, ect.  match with the corresponding regular expressions  <regexpr1>,  <regexpr2>, etc. The instance of the matching table entry is returned. 

      The second element of each tuple rappresent the matching expression and must be a string of this form: 

      *value type* **/** *regular expression*

      The *value type* must be *b* (binary) or *s* (string).

    **Input**

      * *snmpexpr* : A SNMP expression indentifing a valid MIB.
      * sequence of *suboid1*, *regexpr1* : A string containing a valid regular expression.

    **Output**

      Returns a *string* OID.

    **Errors**

      The function raises an exception when the matching sub-OID is not found.

    **Examples**

	+----------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------------------------------------------+
	| Expression                                                                                                     | Return value | Description                                                                                                      |
        +================================================================================================================+==============+==================================================================================================================+
	| byMultiRegexpUnique(1.3.6.1.2.1.25.4.2.1@ , "2" , "s/^apache.exe$", "4" , "s/^/bin/apache$" , "5", "s/^foo$" ) | *4232*       | Searches for a process 'apache.exe' in the hrprocess MIB with path '/bin/apache' and called with parameter 'foo' |
	+----------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------------------------------------------+
	| byMultiRegexpUnique(1.3.6.1.2.1.25.4.2.1@ , "2" , "s/^" + re_escape("c:\\apache.exe") + "$")                    | *123*        | Searches for a process 'c:\\apache.exe'.                                                                          |
	+----------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------------------------------------------+
 

 
byMultiRegexpCount
---------------------------------------------

    **Syntax**

      **byMultiRegexpCount** **(** snmparg **,** [ *suboid1* **,** *regexpr1* ] **,** [ *suboid2* **,** *regexpr2* ] **,** ... **)**

    **Description**

      Scans the MIB identified by the SNMP expression *snmparg* and counts all the sub-oids <suboid1>, <suboid2>, ect.  match with the corresponding regular expressions  <regexpr1>,  <regexpr2>, etc. 

      The second element of each tuple rappresent the matching expression and must be a string of this form: 

       *value type* **/** *regular expression*

      The *value type* must be *b* (binary) or *s* (string).

    **Input**

      * *snmpexpr* : A SNMP expression indentifing a valid MIB.
      * sequence of *suboid1*, *regexpr1* : A string containing a valid regular expression.

    **Output**

      Returns a *string* OID.

    **Errors**

      The function raises an exception if an error occours.

    **Examples**

      +----------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------------------------------------------+
      | Expression                                                                                                     | Return value | Description                                                                                                      |
      +================================================================================================================+==============+==================================================================================================================+
      | byMultiRegexpCount(1.3.6.1.2.1.25.4.2.1@ , "2" , "s/^apache.exe$", "4" , "s/^/bin/apache$" , "5", "s/^foo$" )  | 42           |                                                                                                                  |
      +----------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------------------------------------------+

existsRegexp
---------------------------------------------

    **Syntax**

      **existsRegexp** **(** *snmpexpr* **,** *str_regexpr* **)**

    **Description**

      Scans the MIB identified by the SNMP expression *snmpexpr*. The value of every entry in the mib is compared with the regular expression stored in the parameter *str_regexpr*. The function returns *true* if there is a least one matching entry in the MIB.

    **Input**

      * *snmpexpr* : A SNMP expression indentifing a valid MIB.
      * *str_regexpr* : A valid regular expression.

    **Output**

      Returns a *string* OID.

    **Errors**

    **Examples**

	 +-------------------------------------------+--------------+------------------------------------------------+
	 | Expression                                | Return value | Description                                    |
	 +===========================================+==============+================================================+
	 | existsRegexp( 1.3.6.1.2.1.2.2.1@, "eth0") | *1*          |  Checks if there is an interface called "eth0" |
	 +-------------------------------------------+--------------+------------------------------------------------+
 

existsRegexpUnique
---------------------------------------------

    **Syntax**

      **existsRegexpUnique** **(** *snmpexpr* **,** *str_regexpr* **)**

    **Description**

      Scans the MIB identified by the SNMP expression *snmpexpr*. The value of every entry in the MIB is compared with the regular expression stored in the parameter *str_regexpr*. The function returns *true* if there is *only one* matching entry in the MIB.

    **Input**

      * *snmpexpr* : A SNMP expression indentifing a valid MIB.
      * *str_regexpr* : A valid regular expression.

    **Output**

      Returns a *string* OID.

    **Errors**

    **Examples**

	 +-------------------------------------------------+--------------+-------------------------------------------------+
	 | Expression                                      | Return value | Description                                     |
	 +=================================================+==============+=================================================+
	 | existsRegexpUnique( 1.3.6.1.2.1.2.2.1@, "eth0") | *1*          | There is only one "eth0" interface              |
	 +-------------------------------------------------+--------------+-------------------------------------------------+
	 | existsRegexpUnique( 1.3.6.1.2.1.2.2.1@, "^eth") | *0*          | There are too many interfaces with prefix "eth" |
	 +-------------------------------------------------+--------------+-------------------------------------------------+
 

byBinaryIP
---------------------------------------------

    **Syntax**

     **byBinaryIP** **(** *snmpexpr* **,** *str_ip* **)**

    **Description**

      Scans the MIB identified by the SNMP expression *snmpexpr*. The string value of every entry is compared with the hexadecimal representation of the IPv4 address *str_ip*.

    **Input**

      * *snmpexpr* : A SNMP expression indentifing a valid MIB.
      * *str_ip* : A valid regular expression.

    **Output**

      Returns a *string* OID.

    **Errors**

    **Examples**

	 +----------------------------------------------------+---------------------------------------------------------------------------------+
	 | Expression                                         | Description                                                                     |
	 +====================================================+=================================================================================+
	 | existsRegexp( 1.3.6.1.2.1.2.2.1@, "255.255.255.0") | Checks if there is an entry in the MIB matching the string *FFFFFF00*.          |
	 +----------------------------------------------------+---------------------------------------------------------------------------------+
 

"""

import re
import socket
import struct

from sanet_common.exprparser.Types import SnmpValueType

from sanet_poller.lib.exceptions import PollException, UndefinedSnmpValueException
from sanet_poller.lib.snmp.base import getvar, gettable
from sanet_poller.lib.utils import *

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_byRegexpUnique(context, snmpexpr, str_regexpr):
	"""
	TODO 
	
	:param  context:  Current execution context.
	:type   context:  :class:`pollerlib.ExecutionContext`
	
	:returns: TODO
	:rtype: number
	
	:exception: 
	        :class:`UndefinedSnmpValueException` Unable to calculate the SNMP instance value.
	"""
	instance = genericRegexpFunction(context, 'VU', snmpexpr, str_regexpr, True, False)
	if instance != None:
		return instance
	else:
		# We raise an expcetion because this functions must return a valid OID instance!
		raise UndefinedSnmpValueException("cannot determine instance for byRegexpUnique(%s,%s)" % (str(snmpexpr), str_regexpr))

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_byBinaryValue(context, snmpexpr, str_hex):
	"""
	TODO 
	
	:param  context:  Current execution context.
	:type   context:  :class:`pollerlib.ExecutionContext`
	
	:returns: TODO
	:rtype: boolean
	
	:exception: 
	        :class:`UndefinedSnmpValueException` Unable to calculate the SNMP instance value.
	"""
	
	# Convert the SNMP variable from binary to its hexadecimal rappresentation
	# and matches with str_hex
	instance = genericRegexpFunction(context, 'VH', snmpexpr, '^%s$' % str_hex, False, True)
	if instance != None:
		return instance
	else:
		# We raise an expcetion because this functions must return a valid OID instance!
		raise UndefinedSnmpValueException("cannot determine instance for byBinaryValue(%s,%s)" % (str(snmpexpr), str_hex))

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_existsRegexpUnique(context, snmpexpr, str_regexpr):
	"""
	TODO 
	
	:param  context:  Current execution context.
	:type   context:  :class:`pollerlib.ExecutionContext`
	
	:returns: TODO
	:rtype: boolean
	
	:exception: 
	"""
	rv = genericRegexpFunction(context, 'EU', snmpexpr, str_regexpr, True, False)
	if rv != None:
		return 1
	else:
		return 0

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_existsRegexp(context, snmpexpr, str_regexpr):	
	"""
	TODO 
	
	:param  context:  Current execution context.
	:type   context:  :class:`pollerlib.ExecutionContext`
	
	:returns: TODO
	:rtype: boolean
	
	:exception: 
	"""
	rv = genericRegexpFunction(context, 'EN', snmpexpr, str_regexpr, False, False)
	if rv != None:
		return 1
	else:
		return 0

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_byBinaryIP(context, snmpexpr, str_ip):
	"""
	TODO 
	
	:param  context:  Current execution context.
	:type   context:  :class:`pollerlib.ExecutionContext`
	
	:returns: TODO
	:rtype: boolean
	
	:exception: 
	        :class:`PollException` Invalid IP address *str_ip*
	        :class:`UndefinedSnmpValueException` Unable to calculate the instance for the input IP address.
	"""
	
	try:
		h = "%02X%02X%02X%02X" % struct.unpack("4B", socket.inet_aton(str_ip))
	except socket.error, er:
		raise PollException('Invalid IP address %s' % str_ip)
	
	instance = genericRegexpFunction(context, 'VH', snmpexpr, '^%s$' % h, False, True)
	if instance != None:
		return instance
	else:
		# We raise an expcetion because this functions must return a valid instance!
		raise UndefinedSnmpValueException("cannot determine instance for byRegexpUnique(%s,%s)" % (str(snmpexpr), str_ip))

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def genericRegexpFunction(context, what, snmparg, str_re, mustBeUnique, hexlifyValue):
	"""
	TODO 
	
	:param  context:  Current execution context.
	:type   context:  :class:`pollerlib.ExecutionContext`
	
	:returns: TODO
	:rtype: boolean
	
	:exception: 
	        :class:`PollException` Invalid type for *snmparg* or any other internal error.
		:class:`UndefinedSnmpValueException` Unable to retrieve some SNMP values.
	"""
	
	if not isinstance(snmparg, SnmpValueType):
		# INTERNAL ERROR! Function called with wrong parameter expressions!
		#raise TypeError("byRegexpUnique: snmparg type is invalid (%s)" % type(snmparg))
		raise PollException("genericRegexpFunction: snmp arg type is invalid (%s). Malformed expression?" % type(snmparg))
	
	log    = context.getLog()
	dcache = context.getCache()
	
	instance = None
	
	baseoid   = snmparg.oid
	host      = snmparg.host
	port      = snmparg.port
	community = snmparg.community

	if not context.has_key('$snmpversion'): raise PollException("Parameter $snmpversion missing or invalid")
	if not context.has_key('$shorttries') : raise PollException("Parameter $shorttries missing or invalid")
	
	path      = context.get('$path', '<none>')
	sv        = context['$snmpversion']
	tries     = context['$shorttries' ]
	
	if __debug__: log.debug("genericRegexpUnique, target %s, host %s, comm %s, port %s tries %s, baseoid %s, regexpt %s", path, host, community, port, tries, baseoid, str_re)
	
	# 
	# Call to this function is expensive. Calculate function's return value once 
	# and cache it inside the execution context.
	#
	retval_context_key = "regexp/%s/%s/%s/%s/RE.%s.%s" % (community,host,str(port),baseoid,what,str_re)
	if context.has_key(retval_context_key):
		instance = context[retval_context_key]
		if __debug__: log.debug("    -> using cached return value (%s) = %s" % (retval_context_key, instance))	
		return instance
			
	# If we are going to match "str_re" with a binary value converted to string,
	# we should perform a case insensitive comparison.
	if hexlifyValue:
		compiled_re = re.compile(str_re, re.IGNORECASE)
	else:
		compiled_re = re.compile(str_re)

	#
	# Search 'instance' in the cache
	#
	ck = 'snmp_cache' + '/' + community + '/' + host + '/' + str(port) + '/' + baseoid + '/RE.' + what + '.' + str(crc32(str_re))

	if __debug__: log.debug("genericRegexpFunction: ck = %s", ck)

	cached_re = dcache.get(ck + '/re')
	
	if (cached_re != None and cached_re.pattern == compiled_re.pattern):
		saved_instance = dcache.get(ck + '/in')
		if saved_instance != None:
			if __debug__: log.debug("using saved instance %s -> snmpget( %s )", saved_instance, (baseoid+'.'+saved_instance))
			
			# ExternalErrorException (PollException) raised here
			val = getvar(host, community, tries, baseoid+'.'+saved_instance, port, version=sv)
			if val:
				if hexlifyValue:
					val = binascii.hexlify(val)
					
				m = compiled_re.match(str(val))
				if m:
					instance = saved_instance
				else:
					if __debug__: log.debug("Cache obsolete: %s.%s:%s@%s:%s = %s (=~ %s)", baseoid,saved_instance,community,host,port,instance, str_re)
					pass
			else:
				# Strange! We didn't get a value with the saved instance! Possible?
				# Network is slow?
				if __debug__: log.debug("Undefined value value for: %s.%s:%s@%s:%s", baseoid,saved_instance,community,host,port)
				pass
							
	elif cached_re:
		if __debug__: log.debug("Cache RE collision: %s (%s) %s (%s)", cached_re.pattern, crc32(cached_re.pattern), compiled_re.pattern, crc32(compiled_re.pattern) )
		pass
	
	if instance == None:
		#
		# Not found in cache, search 'instance' inside the SNMP table
		#
		if __debug__: log.debug("instance not found in cache: searching with SNMP." )
		
		# If the value for 'snmparg' can't be read
		# a UndefinedSnmpValueException is raised here!
		snmp_table = snmparg[:]
		
		nfound = 0
		for (oid, val) in snmp_table:
			if __debug__: log.debug("got %s -> %r (%s) to match up with %s", oid, val, type(val), str_re )
			
			#if what == 'VH':
			if hexlifyValue:				
				val = binascii.hexlify(val)
			
			m = compiled_re.match(str(val))
			if m: 
				nfound += 1

				# We searched in the table    1.3.6.1.2.1.2.2.1.2  for the value 2.
				# We found the tuple         .1.3.6.1.2.1.2.2.1.2.X = 2
				# We return the instance id ----------------------^
				instance = oid[1 + len(baseoid) + 1:]
				value    = val
				if __debug__: log.debug("found instance %s: %s %r", instance, oid, value);

				# If we don't need to check the uniqueness
				# of the matching express then we can stop on 
				# at the first match
				if not mustBeUnique:
					break
					
		if nfound == 1:
			
			if __debug__: log.debug("Cache miss: %s.%s:%s@%s:%s = %r (=~ %s)", baseoid,instance,community,host,port,value, str_re)
			
			# FIX: What should we do if an error occurs here? Just skip?
			try:
				if __debug__: log.debug("SAVING CACHE")
				dcache.set(ck + '/re', compiled_re)
				dcache.set(ck + '/in', instance)
			except Exception, e:
				print e
				pass
				
			
		
	#
	# Cache function's return value for this execution
	#
	context[retval_context_key] = instance
		
	return instance



#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#


def fun_byMultiRegexpUnique(context, snmparg, *params, **kwargs):

        return __byMultiRegexp(context, "unique", snmparg, *params, **kwargs)


def fun_byMultiRegexpCount(context, snmparg, *params, **kwargs):
        matched = __byMultiRegexp(context, "count", snmparg, *params, **kwargs)
        return len(matched)



def __byMultiRegexp(context, action, snmparg, *params, **kwargs):

	def __compare_val(valtype, valcmp, val):
		if valtype == "b":
			v = binascii.hexlify(val)
		else:
			v = str(val)
		return (v, re.match(valcmp, v) != None)


	log = context.getLog()
	dcache = context.getCache()

	log.debug("byMultiRegexpUnique()")
	
	if len(params) % 2 != 0:
		raise PollException("Invalid number of filter parameters")

	if not isinstance(snmparg, SnmpValueType):
		raise PollException("snmp arg type is invalid (%s). Malformed expression?" % type(snmparg))

	instance = None

	baseoid   = snmparg.oid
	host      = snmparg.host
	port      = snmparg.port
	community = snmparg.community
	
	check_host = context.get('$node', 'none')

	#if not context.has_key('$path')       : raise PollException("Parameter $path missing or invalid")
	if not context.has_key('$snmpversion'): raise PollException("Parameter $snmpversion missing or invalid")
	if not context.has_key('$shorttries') : raise PollException("Parameter $shorttries missing or invalid")

	#path      = context['$path']	
	sv        = context['$snmpversion']
	tries     = context['$shorttries' ]

	filter_conditions = []
	for i in xrange(len(params)/2):
		sub_oid = str(params[i*2])
		valtype, valcmp = str(params[(i*2)+1]).split('/',1)
		filter_conditions.append( [ sub_oid, valtype, valcmp] )
		log.debug("condition -> %s = %s : %s", sub_oid, valtype, valcmp)
		log.debug("condition -> %s = %s : %s", type(sub_oid), type(valtype), type(valcmp))

	#
	# CACHE KEYS
	#

	# Cache key inside current execution context
	cache_key = "multiregexp/%s/%s/%s/%s/TUPLES/" % (community,host,str(port),baseoid)
	cache_key += '/'.join( [ "%s=%s:%s" % (c[0],c[1],c[2]) for c in filter_conditions ] )

	# Global cache key
	global_cache_key = 'snmp_cache' + '/' + cache_key + '/' + check_host + "/in"


	#
	# Optimized search for unique
	#
	if action == 'unique':

		#
		# 
		#
		if context.has_key(cache_key):
			instance = context[cache_key]
			log.debug("    -> using cached return value (%s) = %s" % (cache_key, instance))	
			return instance

		#
		# Search 'instance' in the cache
		#
		log.debug("Checking in GLOBAL CACHE (%s)", global_cache_key)
		instance = dcache.get(global_cache_key)

		log.debug("\tinstance was: %s", instance)

		if instance != None:

			# Verify all filters with cached instance
			match_all = True
			for (suboid, valtype, valcmp) in filter_conditions:

				o = baseoid + '.' + str(suboid) + '.' + str(instance)
				v = getvar(host, community, tries, o, port, version=sv)

				log.debug("\tget %s -> %s", o, v)

				if v == None:
					# WARNING: with some agents a GETVAR() on  hrSWRunParameters
					#          returns None (not defined on the agent) and
					#          GETTABLE() returns empty string. Just fix it.
					if (baseoid + '.' + str(suboid))  == '1.3.6.1.2.1.25.4.2.1.5':
						v = ''

				if v == None:
					match_all = False
					break
		
				compared_value, result = __compare_val(valtype, valcmp, v)

				if not result:
					match_all = False
					break

			if match_all:
				log.debug("\tOLD INSTANCE MATCHED")
				return instance

	#
	# WALK
	#
	first_condition = filter_conditions.pop(0)
	first_suboid  = first_condition[0]
	first_valtype = first_condition[1]
	first_valcmp  = first_condition[2]

	base_oid = baseoid + '.' + str(first_suboid)

	log.debug("Getting first table: %s", base_oid)
	base_table = gettable(host, community, tries, base_oid, port=port, version=sv)
	if not base_table:
		reason = "Table not found."
		context.appendReason(reason)
		raise PollException(reason)

	matched = []

	log.debug("Filtering entries")
	for oid, v in base_table:

		msg = "%s -> %s" % (oid, baseoid)

		instance = oid[ 1 + len(base_oid) + 1: ]
		msg += "{ %s = %s }" % (first_suboid, v)

		match_all = False

		compared_value, result = __compare_val(first_condition[1], first_condition[2], v)
		if result:
			match_all = True

			for (suboid, valtype, valcmp) in filter_conditions:

				o = baseoid + '.' + str(suboid) + '.' + str(instance)
				v = getvar(host, community, tries, o, port, version=sv)

				if v == None:
					# WARNING: with some agents a GETVAR() on  hrSWRunParameters
					#          returns None (not defined on the agent) and
					#          GETTABLE() returns empty string. Just fix it.
					if (baseoid + '.' + str(suboid))  == '1.3.6.1.2.1.25.4.2.1.5':
						v = ''

				if v == None:
					msg += "{ %s=%s }" % (suboid, v)
					match_all = False
					break

				compared_value, result = __compare_val(valtype, valcmp, v)

				if v == None:
					msg += "{ %s=%s } (None) }" % (suboid,compared_value)
				else:
					msg += "{ %s=%s }" % (suboid,compared_value)

				if not result:
					match_all = False
					break

		log.debug(msg)

		if match_all:

			log.debug("found instance")

			if action == 'unique':
				log.debug("Saving in execution cache.")
				context[cache_key] = instance
				context.appendReason(msg)

				log.debug("Saving in global cache: %s = %s", global_cache_key, instance)
				dcache.set(global_cache_key, instance)

				return instance
			else:
				matched.append(instance)

	if action == 'count':
		return matched
	else:
		raise PollException("Instance not found")

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

EXPR_SYMBOLS = {
	'byBinaryIP'            : fun_byBinaryIP,
	'byBinaryValue'         : fun_byBinaryValue,
	'byRegexpUnique'        : fun_byRegexpUnique,
	'existsRegexp'          : fun_existsRegexp,
	'existsRegexpUnique'    : fun_existsRegexpUnique,

	'byMultiRegexpUnique'   : fun_byMultiRegexpUnique,
	'byMultiRegexpCount'    : fun_byMultiRegexpCount,
}
