# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it> 
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""
***************
ICMP Functions
***************

Ping & RTT
==========

.. _poller_function_getICMPStat:

getICMPStat
-----------

    **Syntax**

      getICMPStat ( info_type , host [, probesize] [,  shorttries] [, timeout ] [, do_frag] [, bind_address] [, ipversion ])

    **Description**

      Performs a reachability test using the ICMP protocol and calculates the RTT Min, RTT Max, RTT Average and Packet Loss percentage.
      
      Returns the value type specified by *info_type* string parameter
      
      .. attention:: Parameters *ipversion* and *probesize* should be configured properly.

    **Input**

      * *info_type*    : (string) *rttavg*, *rttmin*, *rttmax*, *loss*.
      * *host*         : (string) The hostname or IP address.
      * *probesize*    : (number) The ICMP packet size. Default: 1472. 
      * *shorttries*   : (number) Number of ICMP echo pachets. Default: 5.
      * *timeout*      : (number) Timeout for every echo packet. Default: 3.
      * *do_frag*      : (boolean) Set to *true* (es. 1) to enable IP fragmentation, *false* (es: 0) to disable fragmentantion. Default is: *true*.
      * *bind_address* : (string) IP source address. Default: not set.
      * *ipversion*    : (number) IP version: 4 for IPv4, 6 for IPv6. Default: 4.
      
      .. warning:: *probesize* default values is always 1472, even if *ipversion* is 6.
      
    **Output**

      Returns an floating point value

    **Errors**

      An exception is raised when an error occurs.

    **Examples**

      +-------------------------------------+------------------------------------------------------------------------+---------------------+
      | Expression                          | Description                                                            | Return string value |
      +-------------------------------------+------------------------------------------------------------------------+---------------------+
      | getICMPStat ('rttmin', "mason")     |                                                                        |                     |
      +-------------------------------------+------------------------------------------------------------------------+---------------------+
      | getICMPStat ('rttmin', $node, 1472) |                                                                        |                     |
      +-------------------------------------+------------------------------------------------------------------------+---------------------+

getRttInfo
-----------

    Alias for :ref:`poller_function_getICMPStat`.

"""

import time
import sys, array, types, re
import subprocess, os
import string
import random

from sanet_poller.lib.exceptions import PollException

from sanet_poller.lib.icmp import SanetIcmpPinger, ResolveException

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def fun_getRttInfo(context, info_type, host, probesize=None,  shorttries=None, timeout=None, do_frag=False, bind_address=None, ipversion=None):	

	log = context.getLog()

	log.debug("getRttInfo(ipv=%s, host=%s, tries=%s, probe=%s, timeout=%s, frag=%s, bind=%s)", ipversion, host, shorttries, probesize, timeout, do_frag, bind_address)

	if probesize  == None: probesize  = context.get('$probesize' , 1472)
	if shorttries == None: shorttries = context.get('$shorttries', 5)
	if timeout    == None: timeout    = context.get('$timeout'   , 3)
	if ipversion  == None: ipversion  = context.get('$ipversion' , 4)

	probesize  = int(probesize)
	shorttries = int(shorttries)
	timeout    = float(timeout)
	ipversion  = float(ipversion)

	ex = None

	cache_key = ':'.join(str(e) for e in [host, probesize,  shorttries, timeout, do_frag, bind_address, ipversion])
	data = context.getCachedSimpleValue(cache_key)
	if data == None:

		#
		# POLICY
		#
		log.debug("Using normal ping")
		p = SanetIcmpPinger()

		#
		# CHECK
		#
		ex = None
		try:
			rttavg, rttmin, rttmax, loss_avg, infos = p.measure_rtt(log, ipversion, host, shorttries, probesize, timeout, do_frag, bind_address=bind_address)
		except ResolveException, ex: # non si risolve = non si raggiunge
			rttavg, rttmin, rttmax, loss_avg, infos = None, None, None, 1   , [str(ex)]
			ex = None
		except PollException, ex:
			rttavg, rttmin, rttmax, loss_avg, infos = None, None, None, None, [str(ex)]

			log.debug("Error while measuring rtt: %s", str(ex) )

		#
		# DATA
		#
		data = (rttavg, rttmin, rttmax, loss_avg, infos, ex)

		log.debug("avg=%s, min=%s, max=%s, loss=%s, infos=%s ex=%s" % (rttavg, rttmin, rttmax, loss_avg, infos, ex) )

		context.cacheSimpleValue(cache_key, data)
	else:
		rttavg, rttmin, rttmax, loss_avg, infos, ex = data

	if infos:
		context.appendReason(','.join(infos))

	if ex != None:
		log.debug("Got exception %s:%s. Raise it", type(ex), str(ex))
		raise ex

	if  info_type == "rttavg":
		if rttavg is None: raise PollException("Packet loss=%s. Unable to calculate %s" % (loss_avg, info_type) )
		return rttavg
	elif info_type == "rttmin":
		if rttmin is None: raise PollException("Packet loss=%s. Unable to calculate %s" % (loss_avg, info_type) )
		return rttmin
	elif info_type == "rttmax":
		if rttmax is None: raise PollException("Packet loss=%s. Unable to calculate %s" % (loss_avg, info_type) )
		return rttmax
	elif info_type == "loss":
		return loss_avg

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

EXPR_SYMBOLS = {
	'getRttInfo'            : fun_getRttInfo,
	'getICMPStat'           : fun_getRttInfo,
}





