# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>. 

"""
================================
Monitoraggio condition
================================

isUp
---------

    **Syntax**
    
      **isUp** **(** *identity string* [ **,** uncheckable_fallback **)** 

    **Description**

      The parameter *identity string* can be:
      
         * A *condition path*.
         * A *classification*. The condition is searched inside the node related to the datagroup executing the expression. If more than one conditions if found, it's choosed randomly.

      Returns *true* if the condition *path* is UP. 
      Returns *false* if the condition is FAILING or DOWN.
      
      If uncheckable_fallback is not specifided, raises an exception if the condition status is UNCHECKABLE or SUSPENDED
      
      When uncheckable_fallback is specified and it is 'UP', if condition's status is UNCHECKABLE or SUSPENDED, return True.
      When uncheckable_fallback is specified and it is 'DN', if condition's status is UNCHECKABLE or SUSPENDED, return False.
      
      If the condition's status can not be calculated, the functions raises and exception.

    **Input**

      1. *path*: the full path or classification of the condition
      2. *uncheckable_fallback*: (string). Valid values are 'UP', 'DN'

    **Outpu**

      Returns a boolean.

    **Errors**

      Raise an exceptions and become uncheckable when the given condition is not found or when uncheckable_fallback is not specified and the condition's status is not UP,DOWN,FA.

    **Examples**
      
      +---------------------------------------------------------------------------+----------------------------------------------+
      | Expression                                                                | Return value                                 |
      +===========================================================================+==============================================+
      | isUp("localhost;;dg;condition_up")                                        | *true*                                       |
      +---------------------------------------------------------------------------+----------------------------------------------+
      | isUp("localhost;;dg;condition_down")                                      | *false*                                      |
      +---------------------------------------------------------------------------+----------------------------------------------+
      | isUp("localhost;;xxx;xxx")                                                | *exception*                                  |
      +---------------------------------------------------------------------------+----------------------------------------------+
      | isUp("localhost;;xxx;xxx", uncheckable_fallback="UP")                     | *true*                                       |
      +---------------------------------------------------------------------------+----------------------------------------------+
      | isUp("localhost;;xxx;xxx", uncheckable_fallback="DN")                     | *false*                                      |
      +---------------------------------------------------------------------------+----------------------------------------------+
      | isUp("system.interface.status")                                           | *false*                                      |
      +---------------------------------------------------------------------------+----------------------------------------------+

UncheckableIfNotUp
-----------------------------------

    **Syntax**
    
      **UncheckableIfNotUp** **(** *path* [ **,** ... ] **)** 

    **Description**

      It turns uncheckable if any of the given transition is not "UP".

    **Input**

      1. *paths*: path string

    **Outpu**

      Returns *true* if all conditions are UP. Uncheckable otherwise.

    **Errors**

      If a condition does not exists, the expression goes uncheckable.


    **Examples**

      +----------------------------------------------------------+----------------------------------------------+
      | Expression                                               | Return value                                 |
      +==========================================================+==============================================+
      | UncheckableIfNotUp("foo;;test", "bar;;test")             | *true*                                       |
      +----------------------------------------------------------+----------------------------------------------+

====================
Monitoraggio dei Tag
====================

tagStatus
---------

    **Syntax**
    
      **tagStatus** **(** *tagpath* **)** 

    **Description**

      Return the status of the given tag.

    **Input**

      1. *tagpath*: the fullpath of the given tag (es: "geo:/Global vision/tag1/tag2")

    **Outpu**

      Returns the status code of the given tag. (es. "UP")

    **Errors**

      The expression goes uncheckable if the tag is not found.

    **Examples**

      +---------------------------------------------------------+----------------------------------------------+
      | Expression                                              | Description                                  |
      +=========================================================+==============================================+
      | tagStatus("geo:/Global vision/tag1/tag2")               | "UP"                                         |
      +---------------------------------------------------------+----------------------------------------------+

=======================
Monitoraggio Datasource
=======================

measureAvgInInterval
--------------------

    **Syntax**
    
      **measureAvgInInterval** **(** *path* **,** *interval* **)** 

    **Description**

      Calculate the average value of the given datasource *path* in the last *interval* (in seconds).

    **Input**

      1. *path*: datasource path
      2. *interval*: time interval in seconds.

    **Outpu**

      Numeric value

    **Errors**

      The expression goes down if the datasourse does not exits.

    **Examples**

      +----------------------------------------------------------------------------------+----------------------------------------------+
      | Expression                                                                       | Description                                  |
      +==================================================================================+==============================================+
      | measureAvgInInterval("localhost;;icmp-reachability:lossperc", 600)               |                                              |
      +----------------------------------------------------------------------------------+----------------------------------------------+

pktLossAvgInInterval
--------------------

    **Syntax**
    
      **pktLossAvgInInterval** **(** *interval* [, *discard_number* ] [, *discard_threshold* ] [, *node_name* ] [, *ds_name* ] **)**

    **Description**

      Return the average packet loss of the node related to the expression where this function is called.
      The value is calculate in the last *interval* seconds.
  

    **Input**

      1. *interval*: time range in seconds.
      2. *discard_number*: Max number of NaN entries ignored when calculating the average.
      3. *discard_threshold*: Values under this threshold are ignored.
      4. *node_name*: Node name (optional)
      5. *ds_name*: packetloss datasource name. Default is *lossperc*.

    **Outpu**

      Numeric value

    **Errors**

      It goes uncheckable if the function can not find useful data to fetch.

    **Examples**

      +----------------------------------------------------------------------------------+----------------------------------------------+
      | Expression                                                                       | Description                                  |
      +==================================================================================+==============================================+
      | pktLossAvgInInterval(600)                                                        | 0.004                                        |
      +----------------------------------------------------------------------------------+----------------------------------------------+

pktLossAvgAndVarInInterval
-----------------------------

    **Syntax**
    
      **pktLossAvgAndVarInInterval** **(** *interval* [, *discard_number* ] [, *discard_threshold* ] [, *node_name* ] [, *ds_name* ] **)**

    **Description**

      The function behaves in the same way of pktLossAvgAndVarInInterval() but returns a tuple with two values:
      
      ::
          
          average , variance  

    **Input**

      See pktLossAvgAndVarInInterval()

    **Outpu**

      A tuple with two floating point values.

    **Errors**

      See pktLossAvgAndVarInInterval()

    **Examples**

      +----------------------------------------------------------------------------------+----------------------------------------------+
      | Expression                                                                       | Description                                  |
      +==================================================================================+==============================================+
      | pktLossAvgAndVarInInterval(600)[0]                                               | 0.004                                        |
      +----------------------------------------------------------------------------------+----------------------------------------------+
      | pktLossAvgAndVarInInterval(600)[1]                                               | 0.34512                                      |
      +----------------------------------------------------------------------------------+----------------------------------------------+




      
      
getConditionStatusTotalTime
---------------------------

    **Syntax**
    
      **getConditionStatusTotalTime** **(** [ *path* ]  [, *time_range*] [, *down_states*] [, *perc*] **)**

    **Description**

      Returns the total down time of a particular condition with the given *path*
      
      If *path* is not specified, the primary condition of the monitored element is selected.
      
      The default timerange is the last 24H. You can specify the differente time range in seconds.
      
      If *perc* is *true* or a positive interge, the percentage value is returned instead of seconds.

    **Input**

      1. *path*      : (optional) the path of the selected condition.
      2. *time_range*: (optional) Time range in seconds.
      3. *down_state*: (optional) Sommand-separated-value string of the down states (es: "DN,FA,UD").
      4. *perc*      : (optional) Boolean

    **Outpu**

      Numeric value

    **Errors**

      It goes uncheckable if the function can not find useful data to fetch.

    **Examples**

      +----------------------------------------------------------------------------------+----------------------------------------------+
      | Expression                                                                       | Description                                  |
      +==================================================================================+==============================================+
      | getConditionStatusTotalTime()                                                    | 123123                                       |
      +----------------------------------------------------------------------------------+----------------------------------------------+
      | getConditionStatusTotalTime(time_range=60*60*24*20, down_states="FA,DN")         | 123123                                       |
      +----------------------------------------------------------------------------------+----------------------------------------------+





DataSourceAvgInInterval
--------------------------

    **Syntax**
    
      **DataSourceAvgInInterval** **(** classification [ **,** interval ]  [ **,** discard_high_threshold ] [ **,** discard_high_max_number ] [ **,** null_max_number ] **)**

    **Description**

      Calculate the average value of the datasource, with the given **classification** in the last *interval* seconds. 
      
      .. important:: The datasource is selected among the datasources of the **current monitored element** at execution time. If not found, it's selected among the *current parent node*.
      
      The maximum allowed value for *interval* is MAX_INTERVAL (one hour). An exception is raised if *interval* is above MAX_INTERVAL.
      
      If *null_max_number* is specified, the function raise an exception when the number of null values in the datasource series is above *null_max_number*.
      
      If *discard_high_threshold* is specified, datasource values above this threshold will be ignored for the average computation. 
      If *discard_high_max_number* is specified, the number of ignored values will be at most *discard_high_threshold*.
      
    **Input**

      1. *classification*: datasource classification.
      2. *interval*      : time interval in seconds.
      3. *discard_high_threshold*: number
      4. *discard_high_max_number*: number
      5. *null_max_number*: number

    **Outpu**

      Numeric value

    **Errors**

      The function raise an exception if an error occurs.

    **Examples**

      +----------------------------------------------------------------------------------------------------------------------+----------------------------------------------+
      | Expression                                                                                                           | Description                                  |
      +======================================================================================================================+==============================================+
      | DataSourceAvgInInterval('network.interface.traffic.in')                                                              |                                              |
      +----------------------------------------------------------------------------------------------------------------------+----------------------------------------------+
      | DataSourceAvgInInterval('network.interface.traffic.in', interval=180)                                                |                                              |
      +----------------------------------------------------------------------------------------------------------------------+----------------------------------------------+
      | DataSourceAvgInInterval('network.interface.traffic.in', interval=180, null_max_number=20)                            |                                              |
      +----------------------------------------------------------------------------------------------------------------------+----------------------------------------------+

DataSourceAvgAndVarInInterval
--------------------------------

    **Syntax**
    
      **DataSourceAvgAndVarInInterval** **(** classification [ **,** interval ]  [ **,** discard_high_threshold ] [ **,** discard_high_max_number ] [ **,** null_max_number ] **)**

    **Description**

      The function behaves in the same way of DataSourceAvgInInterval() but returns a tuple with:
      
      ::
          
          average , variance
      
    **Input**

      See DataSourceAvgAndVarInInterval().

    **Outpu**

      A tuple with two floating point values.

    **Errors**

      See DataSourceAvgAndVarInInterval().

    **Examples**

      +----------------------------------------------------------------------------------------------------------------------+----------------------------------------------+
      | Expression                                                                                                           | Description                                  |
      +======================================================================================================================+==============================================+
      | DataSourceAvgAndVarInInterval('network.interface.traffic.in')[0]                                                     | average value                                |
      +----------------------------------------------------------------------------------------------------------------------+----------------------------------------------+
      | DataSourceAvgAndVarInInterval('network.interface.traffic.in')[1]                                                     | variance value                               |
      +----------------------------------------------------------------------------------------------------------------------+----------------------------------------------+
"""

import json
import time
import socket

import math

from sanet_common.uuids import create_global_uuid

from sanet_common.sanetdefs.constants import *

from sanet_common.sanetdefs.ipc import SANETD_STATE_QUEUE
from sanet_common.sanetdefs.ipc import AGENT_TEMP_REPLY_QUEUE

from sanet_common.ipc import encode_request_with_id
from sanet_common.ipc import decode_response
from sanet_common.ipc import RESPONSE_ERROR


from sanet_poller.lib.exceptions import PollException, RemoteQueryException

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def _query_server(context, cmd, params, timeout=None, target_queue=SANETD_STATE_QUEUE):

	timeout = timeout or 5

	logger = context.getLog()

	try:
		ipc_factory = context.getIPCFactory()
		# OPEN
		logger.log(9, "Opening IPC connection")
		connection = ipc_factory.create_connection()
		connection.open()

		# BIND
		replay_id = '-'.join([create_global_uuid()])	
		replay_address = AGENT_TEMP_REPLY_QUEUE( replay_id )
		connection.bind(replay_address)
		logger.log(9, "Bound to %s", replay_address)

		# ENCODE
		logger.log(9, "Encoding request")
		mid = str(int(time.time()))
		request = encode_request_with_id(cmd, mid, params)

		# SEND
		logger.log(9, "Sending...")
		connection.send_msg(target_queue, request, timeout=timeout)

		# RECEIVE
		(addr, msg) = connection.receive_msg(ack_msg=True, timeout=timeout)
		if (msg == None):
			logger.info("chiedo %s %s", addr, msg)
			raise Exception("Timeout while waiting for datasource data from server.")

		# DECODE
		response = decode_response(msg)
		if response['id'] != mid:
			raise RemoteQueryException("Waiting for response with id %s. Received %s" % (mid, response['id']))

		if response['status'] == RESPONSE_ERROR:
			raise RemoteQueryException("Reponse Error: %s" % ( response['content'] ) )

		# HANDLE RESPONSE
		return response['content']

	except RemoteQueryException, e:
		raise e

	except Exception, e:
		raise RemoteQueryException("Unknown error: %s:%s" % (type(e), str(e)))


#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def _get_datasource_data(context, ts_start, ts_end, path=None, node_name=None, ds_name=None, subelement_kind=None, subelement_name=None, ds_classification=None, timeout=None):

	configuration = context.getConfiguration()

	tenant_id = configuration.get_tenant_id()

	log = context.getLog()

	try:
		params = {
			'tenant_id'      : tenant_id,
			'ts_start'       : ts_start,
			'ts_end'         : ts_end,

			'node_name'      : node_name,
			'subelement_kind': subelement_kind,
			'subelement_name': subelement_name,
                                         
			'path'           : path,
			'ds_name'        : ds_name,

			'classification' : ds_classification,
		}

		log.debug("Fetching data for datasource: %s", params)

		response = _query_server(context, 'get_datasource_data', params, timeout=timeout)

		return response['data']

	except RemoteQueryException, ex:
		raise PollException("Unable to get data for ds '%s': %s" % (path, ex))


	return data


def _datasource_avg(context, interval, discard_high_number=0, discard_high_threshold=0.0,  path=None, node_name=None, ds_name=None):

	log = context.getLog()

	ts_end   = int(time.time())
	ts_start = int(ts_end - interval)

	data = _get_datasource_data(context, ts_start, ts_end, path=path, node_name=node_name, ds_name=ds_name)

	log.debug("Num entries read from rrd: %s", len(data))

	skipped        = 0
	total          = 0
	count          = 0
	discarded_high = 0
	
	variance_entries = []

	for e in data:   
		# e = [ <timestamp>, <ds0>, <ds1>, ... <dsN>]
		#loss = e[1+(ds_index-1)]
		loss  = e
		log.log(9, "Entry: %s", e)

		if loss == None:
			skipped += 1
			continue

		if (discarded_high < discard_high_number) and (loss >= discard_high_threshold):
			discarded_high += 1
			log.debug("Discarded #%d, %s >= %s" % (discarded_high, loss, discard_high_threshold))
			continue

		total += loss
		count += 1

		variance_entries.append(loss)

	log.debug("Null values: %s", skipped)
	if count == 0:
		raise Exception("Unable to calculate loss average. No valid entries fetched. (tot entries=%s, tot NAN=%s, discarded=%s)" % (len(data), skipped, discarded_high,))

	#return total / count
	average = total / count
	
	variance = 0
	for v in variance_entries:
		variance += math.pow( average - v, 2)
	variance = variance / len(variance_entries)
	
	return average, variance


#-----------------------------------------------------------------------------#
# CONDITIONS                                                                  #
#-----------------------------------------------------------------------------#

from sanet_poller.lib.exceptions import ConditionNotFoundException

def fun_isUp(context, path_or_classification, uncheckable_fallback=None ):

	if uncheckable_fallback is not None:
		if uncheckable_fallback not in ['UP','DN']:
			raise PollException("Invalid uncheckable_fallback state: %r" % (uncheckable_fallback))

	log = context.getLog()

	node    = context['__node__']
	element = context['__element__']

	condition = None
	source    = None

	#
	# Look for condition status 
	#
	log.debug("Searching condition %s in local configuration...", path_or_classification)

	local_configuration = context.getConfiguration()

	# PATH?
	try:
		condition = local_configuration.get_condition_by_path(path_or_classification)
	except ConditionNotFoundException, e:
		log.debug("Condition not found by path '%s'", path_or_classification);
		condition = None

	if condition is None:
		# CLASSIFICATION?
		try:
			condition = local_configuration.get_condition_by_classification(path_or_classification, node.uuid, element.uuid)
		except ConditionNotFoundException, e:
			log.debug("Condition not found by classification '%s'", path_or_classification);
			condition = None

	if condition is not None:
		log.debug("Condition %s found in local configuration!", path_or_classification)

		source = 'local'
		status = condition.status
	else:
		log.debug("Condition %s not found in local configuration. Getting from sanetd server.", path_or_classification)

		try:
			search_params = { 
				'tenant_id'   : node.tenant_id, 
				'path'        : path_or_classification,

				'node_name'   : node.name, 
				'element_name': element.name,
			}

			response = _query_server(context, 'get_condition_status', search_params )

			source = 'remote'
			status = response['data']

		except RemoteQueryException, ex:
			#
			# FIX: notify this problem. #2942
			#
			raise PollException("Unable to get status for condition \"%s\": %s" % (path_or_classification, ex))

	#
	# Found
	#
	msg = "Status for %s (%s) = %s" % (path_or_classification, source, CSTATE_TO_CODENAME(status))
	context.appendExecutionInfo(msg)

	log.debug(msg)

	if status == CSTATE_UP:
		return True
	elif status in [ CSTATE_FA, CSTATE_DN ]:
		return False
	else:
		if uncheckable_fallback is not None:
			if uncheckable_fallback == 'UP':
				return True
			else:
				return False

		raise PollException("Status for %s is uncheckable (%s)" % (path_or_classification, CSTATE_TO_CODENAME(status)))

#-----------------------------------------------------------------------------#

def fun_UncheckableIfNotUp(context, *condition_paths ):

	log = context.getLog()

	node = context['__node__']

	local_configuration = context.getConfiguration()

	states = []
	for path in condition_paths:
		condition = None
		is_local = False

		if condition is None:
			log.debug("Condition %s not found in local configuration", path)

			try:
				response = _query_server(context, 'get_condition_status', { 'tenant_id': node.tenant_id, 'path': path })
				status = response['data']
			except RemoteQueryException, ex:
				raise PollException("Unable to get status for condition \"%s\": %s" % (path, ex))

		states.append( (path, status, is_local) )

	log.debug("Loaded %s states", len(states))

	up_count = 0
	for path, status, is_local in states:

		if status == CSTATE_UP:
			up_count += 1

		info = "Status for %s (%s) = %s" % (path, 'local' if is_local else 'remote', CSTATE_TO_CODENAME(status))
		context.appendReason( info )

		log.debug(info)

	if up_count < len(states): 
		raise PollException("Target forced uncheckable by targets status" )

        return True

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_getConditionStatusTotalTime(context, path=None, time_range=None, down_states=None, perc=False ):

	log = context.getLog()

	node = context['__node__']

	log.debug("getNodeUpTime(%s, %s)", node.name, node.uuid)

	try:
		
		log.debug("Sending request...")
		response = _query_server(context, 'get_node_up_time', {
			'tenant_id': node.tenant_id,
			'node_id'  : node.uuid, 
			'time_range' : time_range,
			'path'     : path,
		})
		totals = response['states_totals']

		log.debug("Received stats: " + str(totals))

		selected_states = []
		if down_states is None:
			selected_states = CSTATE_MACRO_DOWN
		else:
			#if uncheckable_as_down: selected_states += [CSTATE_UU, CSTATE_DU]
			#if failing_as_down    : selected_states += [CSTATE_FA]
			for code in down_states.split(','):
				selected_states += [ CSTATE_FROM_CODENAME(code) ]

		total_down = sum([ totals.get(CSTATE_TO_CODENAME(s), 0)  for s in selected_states  ])

		if perc:
			total = sum( totals.values() )
			if total != 0:
				total_down = float(total_down) / total
			else:
				total_down = 0

		return total_down

	except RemoteQueryException, ex:
		raise PollException("Error on remote server: %s" % (ex))

#-----------------------------------------------------------------------------#
# TAGS                                                                        #
#-----------------------------------------------------------------------------#

def fun_tagStatus(context, tagid):

	log = context.getLog()

	tenant_id = None
	element = context.get('__element__', None)
	if element is not None:
		tenant_id = element.tenant_id

	log.debug("asking status for tag %s (tenant=%s) to sanet server....", tagid, tenant_id)

	try:
		response = _query_server(context, 'get_tag_status', {
			'tenant_id': tenant_id,
			'id'       : tagid, 
		})

		status =  response['data']

		return CSTATE_TO_CODENAME(status)

	except RemoteQueryException, ex:
		raise PollException("Unable to get status for tag \"%s\": %s" % (tagid, ex))


#------------------------------------------------------------------------------#
# DATASOURCES                                                                  #
#------------------------------------------------------------------------------#

def fun_measureAvgInInterval(context, path, interval):

	average, variance = _datasource_avg(context, interval, path=path)

	return average

#------------------------------------------------------------------------------#

def fun_pktLossAvgInInterval(context, interval, discard_number=0, discard_threshold=0.0, node_name=None, ds_name=None):

	average, variance = fun_pktLossAvgAndVarInInterval(context, interval, discard_number=discard_number, discard_threshold=discard_threshold, node_name=node_name, ds_name=ds_name)

	return average

def fun_pktLossAvgAndVarInInterval(context, interval, discard_number=0, discard_threshold=0.0, node_name=None, ds_name=None):

	log = context.getLog()

	#
	# Check input parameters
	#
	if (discard_number != 0 and discard_threshold == 0.0) or (discard_number == 0 and discard_threshold != 0.0):
		raise Exception("second and third arg to pktLossAvgInInterval must be both zero or both nonzero")

	node_name = node_name or context['__node__'].name
	ds_name   = ds_name   or 'lossperc'

	#
	# Check expr cache
	#
	val_id = '_'.join([ str(o).lower() for o in [interval, discard_number, discard_threshold, node_name, ds_name] ])

	log.debug("Expr cache key: %s", val_id)
	ret_val = context.getCachedSimpleValue(val_id)
	log.debug("Cached value: type=%s val=%r", type(ret_val), ret_val)

	if ret_val != None:
		return ret_val

	#
	# Calculate
	#
	try:
		ret_val = _datasource_avg(context, interval, discard_high_number=discard_number, discard_high_threshold=discard_threshold, node_name=node_name, ds_name=ds_name)

	except PollException, e:
		raise e

	except Exception, e:
		raise PollException("Unknown error: %s:%s" % (type(e), str(e)))

	# 
	# Cache value in the execution cache
	#
	log.debug("Saving to cache: %s %r", type(ret_val), ret_val)
	context.cacheSimpleValue(val_id, ret_val)

	average, variance = ret_val

	return average, variance

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

MAX_INTERVAL = 3600

def _round_to_minute(ts):
	return int(ts) - (int(ts) % 60)

def fun_DataSourceAvgInInterval(context, classification, interval=MAX_INTERVAL,  discard_high_threshold=None, discard_high_max_number=None, null_max_number=None):

	average, variance = fun_DataSourceAvgAndVarInInterval(
		context, 
		classification          = classification, 
		interval                = interval,  
		discard_high_threshold  = discard_high_threshold, 
		discard_high_max_number = discard_high_max_number, 
		null_max_number         = null_max_number
	)

	return average

def fun_DataSourceAvgAndVarInInterval(context, classification, interval=MAX_INTERVAL,  discard_high_threshold=None, discard_high_max_number=None, null_max_number=None):

	log = context.getLog()

	log.debug("elementDataSourceAvgInInterval()")

	#
	# Time interval
	#
	if (interval > MAX_INTERVAL):
		raise Exception("Interval %s too big" % (interval))

	ts_end   = int(time.time())
	ts_start = int(ts_end - interval)

	#
	# Select monitored element
	#
	element = context['__element__']
	node    = context['__node__']

	if node != element:
		subelement_name = element.name
		subelement_kind = element.kind
	else:
		subelement_name = None
		subelement_kind = None

	#
	# Fetch data
	#

	# NOTE: la funzione DataSourceAvgInInterval() puo' comparire piu' volte all'interno della stessa expr. Facciamo caching a livello di singola esecuzione.
	#       Siccome nella chiave compare anche ts_start e ts_end e' possibile che due chiamate siano disallineate di un secondo (o piu') vanificando la cache.
	#       Per essere sicuri, arrotondiamo ts_start e ts_end al minuto.
	val_id = '_'.join([ str(o).lower() for o in [_round_to_minute(ts_start), _round_to_minute(ts_end), node.name, subelement_kind, subelement_name, classification ] ])

	log.debug("Expr cache key: %s", val_id)
	ret_val = context.getCachedSimpleValue(val_id)
	log.debug("Cached value: type=%s val=%r", type(ret_val), ret_val)

	if ret_val != None:
		return ret_val

	try:
		log.debug("    node           = %s", node.name)
		log.debug("    subname        = %s", subelement_name)
		log.debug("    subkind        = %s", subelement_kind)
		log.debug("    classification = %s", classification)
		log.debug("    interval       = %s", interval                )
		log.debug("    threshold      = %s", discard_high_threshold  )
		log.debug("    max_number     = %s", discard_high_max_number )
		log.debug("    null max       = %s", null_max_number         )
		log.debug("    node           = %s", node                    )

		data = _get_datasource_data(context, ts_start, ts_end, path=None, node_name=node.name, ds_name=None, subelement_kind=subelement_kind, subelement_name=subelement_name, ds_classification=classification)

		log.debug("Num entries read from rrd: %s", len(data))
		log.log(1, " %s",  data)

		num_none      = 0
		num_discarded = 0
		valid_values  = 0
		total         = 0

		variance_entries = []

		for value in data:

			if value is None:
				if null_max_number is not None and num_none >= null_max_number:
					raise Exception("Too many null values.")
				num_none += 1
				continue

			if discard_high_threshold is not None:
				if value >= discard_high_threshold and (discard_high_max_number is None or num_discarded < discard_high_max_number):
					num_discarded += 1
					continue

			log.log(1, "\t%s", value)

			total        += value
			valid_values += 1

			variance_entries.append(value)

		log.debug("Final entries to process: %s", valid_values)

		if valid_values == 0:
			raise PollException("No data to process (discarded=%s, null=%s)" % (num_discarded, num_none))

		average = total / valid_values

		variance = 0
		for v in variance_entries:
			variance += math.pow( average - v, 2)

		ret_val = average, variance

	except PollException, e:
		raise e

	except Exception, e:
		raise PollException("Unknown error: %s:%s" % (type(e), str(e)))

	# 
	# Cache value in the execution cache
	#
	log.debug("Saving to cache: %s %r", type(ret_val), ret_val)
	context.cacheSimpleValue(val_id, ret_val)

	return ret_val

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

EXPR_SYMBOLS = {
	'isUp'                          : fun_isUp,

	'UncheckableIfNotUp'            : fun_UncheckableIfNotUp,

	'tagStatus'                     : fun_tagStatus,

	'measureAvgInInterval'          : fun_measureAvgInInterval,

	'pktLossAvgInInterval'          : fun_pktLossAvgInInterval,
	'pktLossAvgAndVarInInterval'    : fun_pktLossAvgAndVarInInterval,

	'getConditionStatusTotalTime'   : fun_getConditionStatusTotalTime, 

	'DataSourceAvgInInterval'       : fun_DataSourceAvgInInterval,
	'DataSourceAvgAndVarInInterval'	: fun_DataSourceAvgAndVarInInterval,
}

