# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it> 
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""
================================
SNMP Utils
================================

str2oid
-------

    **Syntax**::
    
      str2oid( *string* ) 

    **Description**

      Convert a char string to the corrisponding OID string.

    **Input**

      * *string*: 

    **Outpu**

      Return a valid OID string.

    **Errors**

      Raises and exception if the input string can not be converted to a valid OID string.

    **Examples**

      +---------------------------+----------------------------------------------+----------------------------------------------+
      | Expression                | Return value                                 | Description                                  |
      +---------------------------+----------------------------------------------+----------------------------------------------+
      | str2oid("foo bar")        | 102.111.111.32.98.97.114                     |                                              |
      +---------------------------+----------------------------------------------+----------------------------------------------+

tryget
------

    **Syntax**::
    
      tryget **(** *snmp expr*  **,** *default value* **)**

    **Description**

      Execute the SNMP expression and returns its value. If the value can not be retrieved, the *default value* is returned instead.

    **Input**

      1. *snmp expr*: a valid Sanet SNMP expression
      2. *default value*: any value

    **Outpu**

      Any value

    **Errors**

      None

    **Examples**

      +-----------------------------------------------------+----------------------------------------------+------------------------------------------------------------------------+
      | Expression                                          | Return value                                 | Description                                                            |
      +-----------------------------------------------------+----------------------------------------------+------------------------------------------------------------------------+
      | tryget( 1.3.6.1.2.1.1.1.0@, "foo" )                 | * a value * or *foo*                         |                                                                        |
      +-----------------------------------------------------+----------------------------------------------+------------------------------------------------------------------------+

testget
-------

    **Syntax**::
    
      testget **(** *snmpexpr* **)**

    **Description**

      Tries to calculate the value for the given SNMP expression.

    **Input**

      1. *snmpexpr*: a SNMP expression.

    **Outpu**

      TODO

    **Errors**

      TODO

    **Examples**

      +-----------------------------------------------------+----------------------------------------------+------------------------------------------------------------------------+
      | Expression                                          | Return value                                 | Description                                                            |
      +=====================================================+==============================================+========================================================================+
      |  testget( 1.3.6.1.2.1.1.1.0@ )                      | *true*                                       |                                                                        |
      +-----------------------------------------------------+----------------------------------------------+------------------------------------------------------------------------+
      |  testget( 9.8.7.6.2.1.1.1.0@ )                      | *false*                                      |                                                                        |
      +-----------------------------------------------------+----------------------------------------------+------------------------------------------------------------------------+


sumall
---------------------------------------------

    **Syntax**

      *sumall** **(** *oid* **,** *community* **,** *hostname* [ **,** *hostname* ] **)**

    **Description**

      Retrieves the value for the given *oid* from every *hostanme* with the given *community* and sums all the values together.

      If some hosts do not export the required MIB or an error occurs, the host is skipped.

    **Input**

      *oid*: String containin the OID.
      *community*: SNMP community
      *hostname*: node hostname

    **Output**

      Returns a *number*

    **Errors**

      The function raises an exception if the SNMP MIB identified by *snmp expr* can not be found.

    **Examples**

	+-----------------------------------------------------+-----------------+
	| Expression                                          | value           |
	+=====================================================+=================+
	| sumall("1.3.6.1.2.1.1.3.0", "public", "foo","bar")  | *1531782725*    |
	+-----------------------------------------------------+-----------------+
"""



import urllib2
import socket
import re

from sanet_poller.lib.snmp.base import getvar

from sanet_poller.lib.exceptions import PollException, UndefinedSnmpValueException, TimeoutException

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_str2oid(context, s):

	return '.'.join( [ str(ord(c)) for c in s ] )

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_sumall(context, oid, community, *hostnames):
	
	log = context.getLog()
	
	shorttries  = context.get('$shorttries' , None)
	snmpport    = context.get('$snmpport'   , None)
	snmpversion = context.get('$snmpversion', None)
	
	if shorttries  == None: raise PollException('Parameter $shorttries missing.')
	if snmpport    == None: raise PollException('Parameter $snmpport missing.')
	if snmpversion == None: raise PollException('Parameter $snmpversion missing.')
	
	log.debug("sumall(%s,%s,...)", oid, community)
	log.debug("oid          = %s", oid)	
	log.debug("community    = %s", community)	
	log.debug("$shorttries  = %s", shorttries)	
	log.debug("$snmpport    = %s", snmpport)	
	log.debug("$snmpversion = %s", snmpversion)	
	sum = 0
	
	for h in hostnames:
		try:
			v = getvar( h, community, shorttries, oid, port=snmpport, version=snmpversion)
			if (v == None):	
				raise UndefinedSnmpValueException("Current variable %s can't be read or is not present (GET) on %s" % (oid, h) )		
			log.debug("Read %s %s=%s", h, oid,v)
			sum += v		
		except TimeoutException, e:
			raise TimeoutException("%s: %s" % (h, e))
			
	return sum

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_tryget(context, snmparg, default):
	
	log = context.getLog()
	
	baseoid   = snmparg.oid
	host      = snmparg.host
	port      = snmparg.port
	community = snmparg.community
	
	try:
		log.debug("checking if mib is supported for: %s:%s@%s:%s", baseoid, community, host, port)
		
		val = snmparg.getvalue()  
		
		log.debug("%s", val)
			
		return val
	
	except UndefinedSnmpValueException, er:
		log.debug("mib not supported for %s:%s@%s:%s -> %s", baseoid, community, host, port, er)
		return default


def fun_testget(context, snmparg):
	
	log = context.getLog()
	
	baseoid   = snmparg.oid
	host      = snmparg.host
	port      = snmparg.port
	community = snmparg.community
	
	try:
		log.debug("checking if mib is supported for: %s:%s@%s:%s", baseoid, community, host, port)
		
		val = snmparg.getvalue()  
		
		log.debug("%s", val)
			
		return True
		
	except UndefinedSnmpValueException, er:
		msg = "mib not supported for %s:%s@%s:%s -> %s" % (baseoid, community, host, port, er)
		log.debug(msg)

		context.appendReason(msg)
		return True

	except TimeoutException, er:
		msg = "Timeout on get %s:%s@%s:%s" % ( baseoid, community, host, port)

		log.debug(msg)
		context.appendReason(msg)

		return False

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#
	

EXPR_SYMBOLS = {

	'str2oid': fun_str2oid,

	'tryget' : fun_tryget, #
	'testget' : fun_testget, #
	
	'sumall': fun_sumall,
	
}




