"""

SNMP Walk utils
===============

walkAverage
---------------------------------------------

    **Syntax**

      *walkAverage* **(** *snmp expr* **)**

    **Description**

      Calculates the the arithmetic average of the elements of the MIB specified by *snmp expr*.

      Not numreic MIB's elements will be ignored

    **Input**

      * *snmp expression*: A SNMP expression 

    **Output**

      Returns a *integer*

    **Errors**

      The function raises an exception if the SNMP MIB identified by *snmp expr* can not be found.

    **Examples**

	 +----------------------------------------+--------------------+
	 | Expression                             | Return Value       |
	 +========================================+====================+
	 | walkAverage( 1.3.6.1.2.1.2.2.1@"bar" ) | * 1935680570.0*    |
	 +----------------------------------------+--------------------+


walkCount
---------------------------------------------

    **Syntax**

      *walkCount* **(** *snmp expr* **)**

    **Description**

      Returns the number of elements of the MIB specified by *snmp expr*.

    **Input**

      * *snmp expression*: A SNMP expression 

    **Output**

      Returns a *integer*

    **Errors**

      The function raises an exception if the SNMP MIB identified by *snmp expr* can not be found.

    **Examples**

	+--------------------------------------+--------------+
	| Expression                           | Return Value |
	+======================================+==============+
	| walkCount( 1.3.6.1.2.1.2.2.1@"bar" ) | *44.0*       |
	+--------------------------------------+--------------+


walkCountMatch
---------------------------------------------

    **Syntax**

      *walkCountMatch* **(** *snmpexpr* **,** *str_regexpr_oid* **,** *str_regexpr_val* **)**

    **Description**

      Scans the MIB identified by the SNMP expression *snmpexpr*. The MIB sub-oid and the value is compared with the regular expressions *str_regexpr_oid* and *str_regexpr_val*. The number of matching entries is returned.

    **Input**

      * *snmpexpr*:  SNMP Expression identifying the MIB to scan.
      * *str_regexpr_prefix*: Regular expression string matching the sub-OIDs in the MIB
      * *str_regexpr_val*: Regular expression string matching the OIDs' value

    **Output**

      Returns the number of matching entries

    **Errors**

      The function raises an exception if the SNMP MIB identified by *snmp expr* can not be found.

    **Examples**

	+-----------------------------------------------+--------------+
	| Expression                                    | Return Value |
	+===============================================+==============+
	| walkCountMatch( 1.2.3.4.5@ , "^6.7.8", "^5$") | *44.0*       |
	+-----------------------------------------------+--------------+


walkMax
---------------------------------------------

    **Syntax**

      *walkMax* **(** *snmp expr* **)**

    **Description**

      Returns the maximum **numeric** value found inside the specified MIB.

      Not numreic MIB's elements will be ignored

    **Input**

      * *snmp expression*: A SNMP expression 

    **Output**

      Returns a *integer*

    **Errors**

      The function raises an exception if the SNMP MIB identified by *snmp expr* can not be found.

    **Examples**

	+------------------------------------+----------------------+
	| Expression                         | Return Value         |
	+====================================+======================+
	| walkMax( 1.3.6.1.2.1.2.2.1@"bar" ) | * 1935680570.0*      |
	+------------------------------------+----------------------+


walkMaxPos
---------------------------------------------

    **Syntax**

      *walkMaxPos* **(** *snmp expr* **)**

    **Description**

      Returns the maximum **positive* **numeric** value found inside the specified MIB.

      Not numreic MIB's elements will be ignored

    **Input**

      * *snmp expression*: A SNMP expression 

    **Output**

      Returns a *integer*

    **Errors**

      The function raises an exception if the SNMP MIB identified by *snmp expr* can not be found.

    **Examples**

	+----------------------------------------+--------------------+
	| Expression                             | Return Value       |
	+========================================+====================+
	| walkMaxPos( 1.3.6.1.2.1.2.2.1@"bar" )  | * 1935680570.0*    |
	+----------------------------------------+--------------------+


walkMin
---------------------------------------------

     **Syntax**

      *walkMin* **(** *snmp expr* **)**

    **Description**

      Returns the minimum **numeric** value found inside the specified MIB.

      Not numreic MIB's elements will be ignored

    **Input**

      * *snmp expression*: A SNMP expression 

    **Output**

      Returns a *integer*

    **Errors**

      The function raises an exception if the SNMP MIB identified by *snmp expr* can not be found.

    **Examples**

	+------------------------------------+-------------------+
	| Expression                         | Return Value      |
	+====================================+===================+
	| walkMin( 1.3.6.1.2.1.2.2.1@"bar" ) | * 1935680570.0*   |
	+------------------------------------+-------------------+


walkMinPos
---------------------------------------------
     **Syntax**

      *walkMinPos* **(** *snmp expr* **)**

    **Description**

      Returns the maximum **numeric** **positive** value found inside the specified MIB.

      Not numreic MIB's elements will be ignored.

    **Input**

      * *snmp expression*: A SNMP expression 

    **Output**

      Returns a *integer*

    **Errors**

      The function raises an exception if the SNMP MIB identified by *snmp expr* can not be found.

    **Examples**

	+---------------------------------------+-------------------+
	| Expression                            | Return Value      |
	+=======================================+===================+
	| walkMaxPos( 1.3.6.1.2.1.2.2.1@"bar" ) | * 1935680570.0*   |
	+---------------------------------------+-------------------+


walkSum
---------------------------------------------

    **Syntax**

      *walkSum* **(** *snmp expr* **)**

    **Description**

      Sums together all the elements of the MIB specified by *snmp expr*.

      MIB's element not suitable for arithmetic addition are ignored.
	
    **Input**

      * *snmp expression*: A SNMP expression 

    **Output**

      Returns a *integer*

    **Errors**

      The function raises an exception if the SNMP MIB identified by *snmp expr* can not be found.

    **Examples**

	+---------------------------------------------+-------------------+
	| Expression                                  | Return Value      |
	+=============================================+===================+
	| walkSum( 1.3.6.1.2.1.2.2.1:"public"@"foo" ) | *5294132713.0*    |
	+---------------------------------------------+-------------------+

walkDiff
---------------------------------------------

    **Syntax**

      *walkDiff* **(** *snmp expr* [ **,** *threshold* [ **,** *verbose_info* ] ] **)**

    **Description**

      Performs a SNMP Walk on the given *snmp expr* and compare the returned table content with the content of the same table fetched at the previous check.

      Returns *true* if the content is not changed.

      If *threshold* is specified, the table entries with a value difference <= given *threshold* are considered unchanged.

      If *verbose_info* is set to 1, the function adds verbose informations to the execution text.

    **Input**

      * *snmp expression*: A SNMP expression.
      * *threshold*: Value threshold.

    **Output**

      Returns a *boolean*

    **Errors**

      The function raises an exception if the given SNMP MIB does not exists.

    **Examples**

	+---------------------------------------------+-------------------+
	| Expression                                  | Return Value      |
	+=============================================+===================+
	| walkDiff( 1.3.6.1.2.1.2.2@)                 | *True*            |
	+---------------------------------------------+-------------------+
	| walkDiff( 1.3.6.1.2.1.2.2@, 10)             | *True*            |
	+---------------------------------------------+-------------------+


walkFilter
---------------------------------------------

    **Syntax**

      *walkFilter* **(** *snmp expr* **,** *selected suboid*   **,**  *suboid 1* **,** *val 1*  **,**  *suboid 2* **,** *val 2* *...* **)**

    **Description**

      Scans the MIB identified by the SNMP expression *snmparg* and returns all the values of the *selected suboid* if the table entry matches <suboid1>, <suboid2>, ect.  and the corresponding values specified by the regular expressions  <regexpr1>,  <regexpr2>, etc. 

      Each matching expression (es: regexpr1) must be a string of this form: 

        *value type* **/** *regular expression*
     
      The *value type* must be *b* (binary) or *s* (string).

      Example:
      
        s/foo
        i/1034

    **Input**

      * *snmp expr*: Base oid
      * *selected suboid*: oid string
      * *suboid n*  : suboid string
      * *val n* :  string with 

    **Output**

      Returns a vector of values

    **Errors**

      The function raises an exception when an error occours.

    **Examples**

	+---------------------------------------------------------------------+-------------------+---------------------------------------------------------------------------+
	| Expression                                                          | Return Value      | Description                                                               |
	+=====================================================================+===================+===========================================================================+
	| sum(walkFilter(1.3.6.1.2.1.2@, "2.1.4", "2.1.7", "i/1"))            | 31436             | Sums all MTU (...2.1.4) of active interfaces (...2.1.7)                   |
	+---------------------------------------------------------------------+-------------------+---------------------------------------------------------------------------+


"""

import json

import re

from sanet_common.exprparser.Types import SnmpValueType
from sanet_common.exprparser.ExprFactory import NotEvaluatedError

from sanet_common.sanetdefs.constants import ELEMENT_KIND_IFACE

from sanet_poller.lib.exceptions import PollException
from sanet_poller.lib.snmp.base import getvar, gettable

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_walkCount(context, snmpexpr):
	
	# If the value for 'snmpexpr' can't be read
	# a UndefinedSnmpValueException is raised here!
	count = len(snmpexpr[:])
	
	return count

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_walkSum(context, snmpexpr):

	log = context.getLog()
	
	sum = 0
	
	# If the value for 'snmpexpr' can't be read
	# a UndefinedSnmpValueException is raised here!
	table = snmpexpr[:]		
	
	print "OK"
		
	if len(table) == 0:
		raise PollException("walkSum called on empty table")
		
	for (oid, a) in table:
		try:
			v = float(a)        #FIX right conversion?????
			sum = sum + v
		except ValueError, er:
			# FIX: TODO?
			log.error("walkSum: error %s", er)	
			pass
		
	context.appendReason("%s elements" % len(table))
	
	return sum

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_walkMin(context, snmpexpr):
	
	log = context.getLog()
	
	valmin = None
	
	# If the SNMP values for 'snmpexpr' can't be read
	# a UndefinedSnmpValueException is raised here!
	table = snmpexpr[:]
	
	if len(table) == 0:
		raise PollException("walkMin called on empty table")
		
	for (oid,a) in table:
		try:
			v = float(a)        #FIX right conversion?????
			if valmin == None or valmin > v:
				valmin = v
		except ValueError, er:
			log.error("walkSum: error %s", er)	
			# FIX: TODO?
	
	context.appendReason("%s elements" % len(table))
	
	return valmin

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_walkMax(context, snmpexpr):
	
	log = context.getLog()
	
	valmax = None
		
	# If the value for 'snmpexpr' can't be read
	# a UndefinedSnmpValueException is raised here!
	table = snmpexpr[:]
	
	if len(table) == 0:
		raise PollException("walkMax called on empty table")
	
	for (oid,a) in table:
		try:
			v = float(a)        #FIX right conversion?????
			if valmax == None or valmax < v:
				valmax = v
		except ValueError, er:
			# FIX: TODO?
			log.error("walkSum: error %s", er)	
			
	context.appendReason("%s elements" % len(table))
	
	return valmax

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_walkAverage(context, snmpexpr):
	
	log = context.getLog()
	
	sum = 0
	
	# If the value for 'snmpexpr' can't be read
	# a UndefinedSnmpValueException is raised here!
	table = snmpexpr[:]
	
	if len(table) == 0:
		raise PollException("walkAverage called on empty table")
	
	for (oid,a) in table:
		try:
			v = float(a)        #FIX right conversion?????
			sum = sum + v
		except ValueError, er:
			# FIX: TODO?
			log.error("walkSum: error %s", er)	
	
	context.appendReason("%s elements" % len(table))
	
	return (float(sum) / float(len(table)))

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_walkMinPos(context, snmpexpr):
		
	log = context.getLog()
	
	minp = None
	
	# If the value for 'snmpexpr' can't be read
	# a UndefinedSnmpValueException is raised here!
	table = snmpexpr[:]
	
	if len(table) == 0:
		raise PollException("walkMinPos called on empty table")
	
	for (oid,a) in table:
		try:
			v = float(a)        #FIX right conversion?????
			if v > 0:
				if minp == None or minp > v:
					minp = v
		except ValueError, er:
			# FIX: TODO?
			log.error("walkMin: error %s", er)	
				
	context.appendReason("%s elements" % len(table))
	
	return minp

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_walkMaxPos(context, snmpexpr):
	
	log = context.getLog()
	
	maxp = None
	
	# If the value for 'snmpexpr' can't be read
	# a UndefinedSnmpValueException is raised here!
	table = snmpexpr[:]
	
	if len(table) == 0:
		raise PollException("walkMaxPos called on empty table")
	
	for (oid,a) in table:
		try:
			v = float(a)        #FIX right conversion?????
			if v > 0:
				if maxp == None or maxp < v:
					maxp = v
		except ValueError, er:
			# FIX: TODO?
			log.error("walkMaxPos: error %s", er)	
				
	context.appendReason("%s elements" % len(table))
	
	return maxp

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_walkCountMatch(context, snmpexpr, str_regexpr_prefix, str_regexpr_val):
	
	log = context.getLog()
	
	# Get OID TABLE
	table = snmpexpr[:]
	
	if len(table) == 0:
		return 0
	
	common_oid = '.'+snmpexpr.oid
	
	prefix_re = re.compile(str_regexpr_prefix)
	val_re    = re.compile(str_regexpr_val)
		
	count = 0	
	
	for (oid, a) in table:
		
		# remove the common oid prefix .1.2.3.4.x.x.x.x.... = y
		#                              ^^^^^^^^^
		prefix = oid[ len(common_oid)+1: ]
		
		# Match OID prefix
		m1 = prefix_re.match( prefix )
		if m1:
			# Match OID value
			m2 = val_re.match( str(a) )
			if m2:
				log.debug('%s = %r  matched', oid, a)
				count += 1	
		
	return count

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def _walk_diff(context, snmparg, threshold=None, verbose_info=1, time_delta=None, filter_function=None):
	

	log    = context.getLog()
	dcache = context.getCache()

	baseoid   = snmparg.oid
	host      = snmparg.host
	port      = snmparg.port
	community = snmparg.community

	ck = '/'.join(['snmp_cache', 'walkdiff', host, community, str(port), baseoid])

	#
	# Load old data from cache
	#
	old_table = None
	try:
		old_table = dcache.get(ck)
		if old_table is not None:
			old_table = json.loads(old_table)
	except Exception, e:
		old_table = None
		context.appendExecutionInfo("Error while loading Old MIB: %s:%s" % (type(e), e))

	#
	# Get current data
	#
	raw_table = snmparg[:]  # If the value for 'snmparg' can't be read UndefinedSnmpValueException is raised here!
	current_table = []

	for oid, v in raw_table:
		current_table.append( (oid[len(baseoid) + 1 : ] , v) )

	# update cache with new data
	dcache.set(ck, json.dumps(current_table))

	#
	# Compare
	#
	ret = True

	if old_table is None:
		raise NotEvaluatedError("Unable to compare current MIB (%s) with old values. MIb not saved yet. " % (baseoid))

	valid, only_current, only_old = merge_diff(current_table, old_table)

	#
	# Calculate differences
	#
	ret = True

	reasons = []

	for new_entry, old_entry in valid:
		
		oid     = new_entry[0]
		
		if filter_function:
			if not filter_function(oid):
				continue
		
		new_val = new_entry[1]
		old_val = old_entry[1]
		
		if threshold is not None:
			try:
				d = new_val - old_val
				if time_delta is not None:
					d = float(d) / time_delta

				if (d > threshold):
					reasons.append( (oid, "diff   : old %s -> new %s" % (repr(old_val), repr(new_val)) ) )
					ret = False
				elif verbose_info:
					reasons.append( (oid, "ok     : old %s = new %s" %  (repr(old_val), repr(new_val)) ) )
			except Exception, e:
				reasons.append( (oid, "unable to compare values with threshold: %s, %s" % (repr(old_val), repr(new_val)) ) )
				ret = False
		else:
			if (new_val != old_val):
				reasons.append( (oid, "changed: old %s <> new %s" % (repr(old_val), repr(new_val)) ) )
				ret = False
			elif verbose_info:
				reasons.append( (oid, "ok     : old %s = new %s" %  (repr(old_val), repr(new_val)) ) )

	if only_current:
		for oid, val in only_current:
			
			if filter_function:
				if not filter_function(oid):
					continue
			
			reasons.append( (oid, "new entry" ) )

		# Ignore new interfaces
		ret = True

	if only_old:
		for oid, val in only_old:
			
			if filter_function:
				if not filter_function(oid):
					continue
			
			reasons.append( (oid, "not found in current data" ) )

		# Ignore vanished interfaces
		ret = True

	return ret, reasons
	

def merge_diff(t1, t2):
	valid      = []
	only_t1    = []
	only_t2    = []
	len1 = len(t1)
	len2 = len(t2)
	i1, i2 = 0,0
	while (i1 < len1) and (i2 < len2):
		oid1 = t1[i1][0]
		oid2 = t2[i2][0]
		if oid1 == oid2:
			valid.append( (t1[i1], t2[i2]) )
			i1 += 1
			i2 += 1
		elif oid1 < oid2:  # missing from t2
			only_t1.append(t1[i1])
			i1 += 1
		elif oid1 > oid2:  # missing from t1
			only_t2.append(t2[i2])
			i2 += 1
	if i1 < len1:
		only_t1 += t1[i1:]
	if i2 < len2:
		only_t2 += t2[i2:]
	return valid, only_t1, only_t2

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_walkDiff(context, snmparg, threshold=None, verbose_info=1, time_delta=None):
	
	def __filter_f(oid):
		return True

	ret, reasons = _walk_diff(context, snmparg=snmparg, threshold=threshold, verbose_info=verbose_info, time_delta=time_delta, filter_function=__filter_f)
	
	context.appendReason( '\n'.join([ "%s - %s" % (r[0], r[1]) for r in reasons ]))
	
	return ret

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_walkDiffIface(context, snmparg, threshold=None, verbose_info=1, time_delta=None, policy=None):
	
	POLICY_ALL         = "all"
	POLICY_BACKBONE    = "backbone"
	POLICY_NOTBACKBONE = "not-backbone"
	
	log = context.getLog()
	current_config = context.getConfiguration()

	element = context.get('__element__', None)

	#
	# Load sanet interfaces by ifindex 
	#
	sanet_iface_by_ifindex = {}
	
	if element:
		ifaces = current_config.get_element_by_parent_id(element.uuid, ELEMENT_KIND_IFACE)
		log.debug("Node: %s", element)
		for i in ifaces:
			log.debug("\tiface %s, ifindex=%s, backbone=%s", i, i.ifindex, i.backbone)
			sanet_iface_by_ifindex[ i.ifindex ] = i

	#
	# DIFF 
	#

	# Check interface backbone flag
	def __filter_f(oid):
		if policy == POLICY_ALL: return True
		#ifindex = int(oid[1:])
		ifindex = oid[1:]
		iface = sanet_iface_by_ifindex.get(ifindex, None)
		if policy == POLICY_BACKBONE:
			if iface is None: 
				return False
			return iface.backbone
		else:
			if iface is None: 
				return True
			return not iface.backbone

	ret, reasons = _walk_diff(context, snmparg=snmparg, threshold=threshold, verbose_info=verbose_info, time_delta=time_delta, filter_function=__filter_f)
	
	#
	# Load ifName from host
	#
	ifnames_by_ifindex = {}

	host       = snmparg.host
	port       = snmparg.port
	community  = snmparg.community
	sv         = element.snmp_version if element else context.get('$snmp_version', 1)
	shorttries = context.get('$shorttries', 3)

	oid = '1.3.6.1.2.1.2.2.1.2'
	table = gettable(host, community, shorttries, oid, port=port, version=sv) # threshold
	if table:
		for suboid, name in table:
			log.debug("%s = %s", suboid, name)
			ifindex = suboid[ 1+len(oid)+1:]
			#ifnames_by_ifindex[ int(ifindex) ] = name
			ifnames_by_ifindex[ ifindex      ] = name

	#
	# Produce output
	#
	result = []
	for r in reasons:
		#ifindex = int(r[0][1:])
		ifindex = r[0][1:]
		
		sanet_name = '*n/a*'
		iface = sanet_iface_by_ifindex.get(ifindex, None)
		if iface is not None: sanet_name = iface.name

		ifname = ifnames_by_ifindex.get(ifindex, ifindex)

		result.append( "%s: %s - Sanet: %s , ifName: %s." % (r[0], r[1], sanet_name, ifname) )
	
	text = '\n'.join(result)
	
	context.appendReason( text )
	
	return ret

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_walkFilter(context, snmparg, selected_suboid, *params, **kwargs):

	log = context.getLog()

	log.debug("byMultiRegexpUnique()")
	
	if len(params) % 2 != 0:
		raise PollException("Invalid number of filter parameters")

	if not isinstance(snmparg, SnmpValueType):
		raise PollException("snmp arg type is invalid (%s). Malformed expression?" % type(snmparg))
	
	instance = None
	
	baseoid   = snmparg.oid
	host      = snmparg.host
	port      = snmparg.port
	community = snmparg.community

	if not context.has_key('$snmpversion'): raise PollException("Parameter $snmpversion missing or invalid")
	if not context.has_key('$shorttries') : raise PollException("Parameter $shorttries missing or invalid")

	sv        = context['$snmpversion']
	tries     = context['$shorttries' ]

	filter_conditions = []
	for i in xrange(len(params)/2):
		sub_oid = str(params[i*2])
		valtype, valcmp = str(params[(i*2)+1]).split('/',1)
		filter_conditions.append( [ sub_oid, valtype, valcmp] )
		log.debug("condition -> %s = %s : %s", sub_oid, valtype, valcmp)
		log.debug("condition -> %s = %s : %s", type(sub_oid), type(valtype), type(valcmp))

	#
	#
	#
	first_condition = filter_conditions.pop(0)
	first_suboid  = first_condition[0]
	first_valtype = first_condition[1]
	first_valcmp  = first_condition[2]

	base_oid = baseoid + '.' + str(first_suboid)

	log.debug("Getting first table: %s", base_oid)
	base_table = gettable(host, community, tries, base_oid, port=port, version=sv)
	if not base_table:
		reason = "Table not found."
		context.appendReason(reason)
		raise PollException(reason)
	
	def __compare_val(valtype, valcmp, val):	
	
		if valtype == "b":	
			v = binascii.hexlify(val)
		else:
			v = str(val)
		return (v, re.match(valcmp, v) != None)

	values = []

	log.debug("Filtering entries")
	for oid, v in base_table:

		msg = "%s -> %s" % (oid, baseoid)
		
		instance = oid[ 1 + len(base_oid) + 1: ]
		msg += "{ %s = %s }" % (first_suboid, v)
		
		match_all = False

		compared_value, result = __compare_val(first_condition[1], first_condition[2], v)
		if result:
			match_all = True

			for (suboid, valtype, valcmp) in filter_conditions:

				o = baseoid + '.' + str(suboid) + '.' + str(instance)
				v = getvar(host, community, tries, o, port, version=sv)

				if v == None:
                                	# WARNING: with some agents a GETVAR() on  hrSWRunParameters
                                        #          returns None (not defined on the agent) and
                                        #          GETTABLE() returns empty string. Just fix it.
                                        if (baseoid + '.' + str(suboid))  == '1.3.6.1.2.1.25.4.2.1.5':
                                                v = ''

				if v == None:
					msg += "{ %s=%s }" % (suboid, v)
					match_all = False
					break
		
				compared_value, result = __compare_val(valtype, valcmp, v)
	                        if v == None:
                                        msg += "{ %s=%s } (None) }" % (suboid, compared_value)
                                else:
                                        msg += "{ %s=%s }" % (suboid,compared_value)
				if not result:
					match_all = False
					break

					
			if not match_all:
				continue
			
			log.debug(msg)

			o = baseoid + '.' + selected_suboid + '.' + instance
			v = getvar(host, community, tries, o, port, version=sv)
		
			values.append(v)
			
			
			
	log.info("values = %s", values)
	
	return values

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

EXPR_SYMBOLS = {
	'walkAverage'           : fun_walkAverage,
	'walkCount'             : fun_walkCount,
	'walkMax'               : fun_walkMax,
	'walkMaxPos'            : fun_walkMaxPos,
	'walkMin'               : fun_walkMin,
	'walkMinPos'            : fun_walkMinPos,
	'walkSum'               : fun_walkSum,
	'walkCountMatch'        : fun_walkCountMatch,
	
	'walkDiff'              : fun_walkDiff,

	'walkDiffIface'         : fun_walkDiffIface,

	'walkFilter'            : fun_walkFilter,
}
