
"""


============
TCP
============

isTcpOpen
---------------------------------------------

    **Syntax**

      **isTcpOpen** **(** *host* **,** *port* **)**

    **Description**

      Verifies if a TCP connection can be established with the remote peer *host*:*port*.

      The parameter *host* can be a valid IPv4 addresso or a hostname. If a hostname is given, it is resolved and the test tries the first IP returned.

      The functions returns *false* the address resolution fails of if the TCP connection can not be established before a 10 seconds timeout.

    **Input**

      * *host*: hostname or IP address.
      * *port*: port number

    **Output**

      Returns a *boolean*

    **Errors**

    **Examples**

	 +------------------------------------------+--------------+---------------+
	 | Expression                               | Return value | Description   |
	 +==========================================+==============+===============+
	 | isTcpOpen("www.google.com", 80)          | *true*       |               |
	 +------------------------------------------+--------------+---------------+
	 | isTcpOpen("thisisnotavalidhostname", 80) | *false*      |               |
	 +------------------------------------------+--------------+---------------+
 
tcpReceive
---------------------------------------------

    **Syntax**

      **tcpReceive** **(** node **,** tcp_port [ **,** timeout] [, num_lines ] **)**

    **Description**

      Connects to *node* and download the first *num_lines* of text data from the listening application.

    **Input**

      * *node*: hostname
      * *tcp_port*: integer
      * *timeout*: integer

    **Output**

      Returns a *string*.

    **Errors**

    **Examples**

	 +------------------------------------------+--------------+---------------+
	 | Expression                               | Return value | Description   |
	 +==========================================+==============+===============+
	 | tcpReceive("www.google.com", 80)         | "<html>...." |               |
	 +------------------------------------------+--------------+---------------+ 
"""

import time
import socket

from sanet_common.netutils import resolve_ipv4
from sanet_common.netutils import resolve_ip

from sanet_poller.lib.exceptions import *

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_isTcpOpen(context, host, port, timeout=3):
	"""
	Verifies if a TCP connection can be established with the remote peer **host**:**port**.
	
	:param  context:  Current execution context.
	:type   context:  :class:`pollerlib.ExecutionContext`
	
	:param  host:  Host name
	:type   host:  string
	
	:param  port:  TCP port
	:type   port:  number
	
	:returns: True if the remote peer is reachable, False otherwise or the given *host* name can't be resolved.
	:rtype: boolean
	
	:exception: 
	        :class:`PollException` when *host* string parameter is an empty.
	"""

	# parameter from database is float. The socket library doesn't like it!
	port = int(port)

	host = host or ''
	if host == '':
		raise PollException("isTcpOpen invalid argument: host=%s, port=%s" % (host, port))
	
	log = context.getLog()

	try:
		log.debug("Resolving %s", host)
		# Dont' block while resolving host name
		addrs = resolve_ipv4(host, False, 10)
		if addrs == None or len(addrs) < 1:
			# Fix: Possible without exceptions?
			reason = "isTcpOpen(%s,%s) NOK: unable to resolve host name" % (host,port)
			log.debug(reason)
			context.appendReason(reason)
			return False
		
		ip = addrs[0]
		log.debug("isTcpOpen(%s,%s): ip = %s", host,port,ip)
		
		sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
		sock.settimeout( timeout ) # seconds		
		
		t1 = time.time()
		sock.connect( (ip,port) )
		t2 = time.time()
		(peerhost, peerport) = sock.getpeername()
		
		reason = "Connected in %.5f s to %s:%s (address %s:%s)" % ((t2-t1), host, port, peerhost, peerport)
		log.debug( reason )
		context.appendReason(reason)
		
		return True
	
	except socket.timeout, t:
		reason = "isTcpOpen(%s,%s) NOK: socket timeout %s" % (host,port, t)
	except socket.error, e:
		reason = "isTcpOpen(%s,%s) NOK: socket error %s" % (host,port, e)
	except Exception, e:
		# resolve_ipv4 raised an exception (not socket)?
		reason = "isTcpOpen(%s,%s) NOK: error %s" % (host,port, e)

	log.debug(reason)
	context.appendReason(reason)
	
	return False
	
	#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def fun_tcpReceive(context, node, tcp_port, timeout=None, num_lines=0):
	
	#
	# FIX: handle IPv6
	#      handle 
	#
	
	#BUFFER_LEN      = 1024
	#if num_lines > 0:
	#	BUFFER_LEN = 1
	
	DEFAULT_TIMEOUT = 30 
	
	log = context.getLog()
	
	timeout = timeout or context.get('$timeout', DEFAULT_TIMEOUT)
	
	try:
		log.debug("connecting to %s (timeout=%s)...", node, timeout)
		s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		
		s.settimeout(timeout)
		
		s.connect( (node, tcp_port) )
		
		log.debug("reading data...")
		
		lines_count = 0
		totdata = ''
		while True:
			data = s.recv(1)
			if len(data) == 0:
				break
			totdata += data

			if data == '\n':
				lines_count += 1
				if lines_count == num_lines:
					break
			
		log.debug("Read %s bytes", len(totdata))
		
		return totdata
	
	except socket.error, er:
		raise PollException( "Socket error while connecting to %s: %s" % (node, er) )
		
	except Exception, er:
		raise PollException( "Unknown error with node '%s': %s" % (node, er) )

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

EXPR_SYMBOLS = {
	'isTcpOpen'    : fun_isTcpOpen,
	'tcpReceive'   : fun_tcpReceive,
}	
