
"""
********************************************************
Test functions
********************************************************

iftrue
=======================

    **Syntax**
    
      **iftrue** **(** *test*, *val1*, *val2* **)**

    **Description**

      Returns *val1* if the boolean value of *test* is *true*. Returns *val2* otherwise.

    **Input**

      1. *test*: a value
      2. *val1*: a value
      3. *val2*: a value

    **Outpu**

      Any value

    **Errors**

      None

    **Examples**

      +----------------------------------------------------+-------------------------------------+-------------------------------------------------------+
      | Expression                                         | Return value                        |  Description                                          |
      +====================================================+=====================================+=======================================================+
      | iftrue(1, "yes", "no")                             | "yes"                               |                                                       |
      +----------------------------------------------------+-------------------------------------+-------------------------------------------------------+
      | iftrue(0, "yes", "no")                             | "no"                                |                                                       |
      +----------------------------------------------------+-------------------------------------+-------------------------------------------------------+

onlyif
=======================

    **Syntax**
    
      **onlyif** **(** *test_val*, *val* [ , *uncheck_msg* ] **)**

    **Description**

      Returns *val1* if the boolean value of *test_val* is *true*. Raise an exception otherwise.

    **Input**

      1. *test_val*: value (boolean)
      2. *val*: a value
      3. *uncheck_msg*: error message. (optional)

    **Outpu**

      The value *val*.

    **Errors**

      Raise an exception (with optional message *uncheck_msg*).

    **Examples**

      +----------------------------------------------------+-------------------------------------+-------------------------------------------------------+
      | Expression                                         | Return value                        |  Description                                          |
      +====================================================+=====================================+=======================================================+
      | onlyif(true, 1)                                    | 1                                   |                                                       |
      +----------------------------------------------------+-------------------------------------+-------------------------------------------------------+
      | onlyif(false, 2)                                   | *exception*                         |                                                       |
      +----------------------------------------------------+-------------------------------------+-------------------------------------------------------+
      | onlyif(false, 2, "error message")                  | *exception: error message*          |                                                       |
      +----------------------------------------------------+-------------------------------------+-------------------------------------------------------+



test_symbol
=======================

    **Syntax**
    
      **test_symbol** **(** *name* **)**

    **Description**

      Tests if the symbol with name *name* exists or can be calculated (es. $ifindex, $cluster_instance, ecc.) in the current execution context.

    **Input**

      1. *name*: string

    **Outpu**

      *true* or *false*

    **Errors**

      None

    **Examples**

      +----------------------------------------------------+-------------------------------------+-------------------------------------------------------+
      | Expression                                         | Return value                        |  Description                                          |
      +====================================================+=====================================+=======================================================+
      | test_symbol("$ifindex")                            | *true*                              |                                                       |
      +----------------------------------------------------+-------------------------------------+-------------------------------------------------------+
      | test_symbol("$_xyz_  ")                            | *false*                             |                                                       |
      +----------------------------------------------------+-------------------------------------+-------------------------------------------------------+



test_state
=======================

    **Syntax**
    
      **test_state** **(** *statename* **)**

    **Description**

      Returns *true* if *statename* is "UP". Returns *false* if *statename* is "DN" or "FA* or "IN". Raise an exception otherwise.

    **Input**

      1. *name*: a valid status code.

    **Output**

      A boolean value.

    **Errors**

      None

    **Examples**

      +----------------------------------------------------+-------------------------------------+-------------------------------------------------------+
      | Expression                                         | Return value                        |  Description                                          |
      +====================================================+=====================================+=======================================================+
      | test_state("UP")                                   | *true*                              |                                                       |
      +----------------------------------------------------+-------------------------------------+-------------------------------------------------------+
      | test_state("FA")                                   | *false*                             |                                                       |
      +----------------------------------------------------+-------------------------------------+-------------------------------------------------------+
      | test_state("UN")                                   | *exception*                         |                                                       |
      +----------------------------------------------------+-------------------------------------+-------------------------------------------------------+

"""

from sanet_common.sanetdefs.constants import *
from sanet_poller.lib.exceptions import PollException

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_test_symbol(context, name):
	# Usiamo dei blocchi try except a causa della natura implementativa dell'oggetto context:
	# alcune chiavi come ifindex vengono calcolate al volo, quindi verrebbero sempre trovate
	# anche se il loro calcolo genererebbe una PollException

	log = context.getLog()
	
	try:
		c = context[name]
		return True
	except KeyError, e:
		#log.debug("save_value: symbol '%s' not defined", name)
		pass
	except PollException, e:
		#log.debug("save_value: error for '%s': %s:%s", name, type(e),str(e))
		pass

	return False

#-----------------------------------------------------------------------------#

def fun_test_state(context, statename):

	statename = statename.upper()

	context.appendExecutionInfo("state = %s" % (statename))

	s = CSTATE_FROM_CODENAME(statename)
	
	context.appendExecutionInfo("state = %s" % (s))

	#if s in [CSTATE_UD, CSTATE_UU, CSTATE_DU, CSTATE_DD]:
	#	raise PollException("Uncheckable test")

	if s == CSTATE_UP:
		return True

	if s in [CSTATE_FA, CSTATE_DN, CSTATE_IN]:
		return False

	if s in [CSTATE_UD, CSTATE_UU, CSTATE_DU, CSTATE_DD]:
		raise PollException("Uncheckable test (%s)" % (statename))
	
	raise PollException("Invalid state (%s)" % (statename))

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_iftrue(context, test_val, val1, val2 ):

	if test_val:
		return val1
	else:
		return val2

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_onlyif(context, test_val, val, uncheck_msg=None):
	
	if test_val:
		return val
		
	if uncheck_msg is None:
		uncheck_msg = "Test value '%s' = %s" % (test_val, val)
	
	raise PollException(uncheck_msg)
	
#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

EXPR_SYMBOLS = {
	'iftrue'     : fun_iftrue,
	'onlyif'     : fun_onlyif,

	'test_symbol': fun_test_symbol,
	'test_state' : fun_test_state,	
}
