# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>. 

"""

==============
Time functions
==============

time
---------------------------------------------

    **Syntax**

      **time** **(** **)**

    **Description**

      Return the time as a floating point number expressed in seconds since the epoch (1970-01-01 00:00:00), in UTC.

    **Output**

      Returns a floating point value.

    **Examples**

      +--------------+---------------------+
      | Expression   | Return string value |
      +==============+=====================+
      | time()       | 1291971078.92       |
      +--------------+---------------------+

UTCTZOffset
---------------------------------------------

    **Syntax**

      **UTCOffset** **(** [ *timezone_name* ] **)**

    **Description**

      Returns the offset in seconds of the given *timezone_name*. 
      
      .. warning:: the returned value counts the Daylight Saving Time (DST) if it's in effect at the moment of the call.
      
      If *timezone_name* is not specified, the local timezone is used.

    **Output**

      Offset in seconds

    **Errors**

      Raise an exception if the input timezone is not valid.

    **Examples**

      +------------------------------------+---------------------+---------------------------------------------------------------------------------------------------+
      | Expression                         | Value               | Description                                                                                       |
      +====================================+=====================+===================================================================================================+
      | UTCOffset()                        | 3600                | Offset of local timezone (Europe/Rome) when *Daylight Saving Time (DST)* is not active. (+01:00)  |
      +------------------------------------+---------------------+---------------------------------------------------------------------------------------------------+
      | UTCOffset()                        | 7600                | Offset of local timezone (Europe/Rome) when *Daylight Saving Time (DST)* is not active. (+02:00)  |
      +------------------------------------+---------------------+---------------------------------------------------------------------------------------------------+
      | UTCOffset('America/Chicago')       | -18000              | Offset of local timezone (Europe/Rome) when *Daylight Saving Time (DST)* is not active. (+02:00)  |
      +------------------------------------+---------------------+---------------------------------------------------------------------------------------------------+



create_dateandtime
---------------------------------------------

    **Syntax**

      **create_dateandtime** **(** year **,** month **,** day **,** hour **,** minutes **,** seconds  **,** deci_seconds  [ **,** utc_dir  **,** utc_hours  **,** utc_minutes ] **)**

    **Description**

      Create a dateandtime byte vector with the given time informations.
      
      When called without parameters, the function returns a dateandtime vector for the current *local* time (with timezone informations included, if available)
      
      If *utc_dir* is '+' or '-', the byte vector will contain timezone informations (11 bytes length). Default value for *utc_hours* and *utc_minutes* is 0 (00:00).
      
      if *utc_dir* is not specified the byte vector will code local time and without timezone informations (8 bytes length).
      
      The DateAndTime specification:
      
          ::
              
              > ---------------------------------------------------------------------------
              >    DateAndTime ::= TEXTUAL-CONVENTION
              >    DISPLAY-HINT "2d-1d-1d,1d:1d:1d.1d,1a1d:1d"
              >    STATUS       current
              >    DESCRIPTION
              >            "A date-time specification.
              >
              >            field  octets  contents                  range
              >            -----  ------  --------                  -----
              >              1      1-2   year*                     0..65536
              >              2       3    month                     1..12
              >              3       4    day                       1..31
              >              4       5    hour                      0..23
              >              5       6    minutes                   0..59
              >              6       7    seconds                   0..60
              >                           (use 60 for leap-second)
              >              7       8    deci-seconds              0..9
              >              8       9    direction from UTC        '+' / '-'
              >              9      10    hours from UTC*           0..13
              >             10      11    minutes from UTC          0..59
              >
              >            * Notes:
              >            - the value of year is in network-byte order
              >            - daylight saving time in New Zealand is +13
              >
              >            For example, Tuesday May 26, 1992 at 1:30:15 PM EDT would be
              >            displayed as:
              >
              >                             1992-5-26,13:30:15.0,-4:0
              >
              >
              >            Note that if only local time is known, then timezone
              >            information (fields 8-10) is not present."
              >    SYNTAX       OCTET STRING (SIZE (8 | 11))
              > ---------------------------------------------------------------------------

    **Output**

      A byte vector (8 byte or 11 byte length)

    **Examples**
      
      +------------------------------------+--------------------------------------------------------+--------------------------------------------------------------------------------------------------+
      | Expression                         | Value                                                  | Description                                                                                      |
      +====================================+========================================================+==================================================================================================+
      | create_dateandtime()               | x07 xe0 x04 x1d x0a x02 x02 x00 + x02 x00              | 2016/4/29 10:02 +02:00                                                                           |
      +------------------------------------+--------------------------------------------------------+--------------------------------------------------------------------------------------------------+
      | create_dateandtime(1970,1,1,1,1)   | x07 xb2 x01 x01 x01 x01 x00 x00                        | 1970/1/1 01:01  (no timezone info)                                                               |
      +------------------------------------+--------------------------------------------------------+--------------------------------------------------------------------------------------------------+


dateandtime2tuple
---------------------------------------------

    **Syntax**

      **dateandtime2tuple** **(** [ *dateandtime value* ] **)**

    **Description**

      Convert the given *dateandtime value* byte vector to the corrisponding time tuple.
      
      This is the "inverse" of *create_dateandtime()*.

    **Output**

      A tuple of values.
      
    **Errors**
    
      Raise an exception if the input values is not valid dateandtime byte vector.

    **Examples**

      +------------------------------------------------------------------------------------+-----------------------------------------------------------------+--------------------------------------------------------------------------------------------------+
      | Expression                                                                         | Value                                                           | Description                                                                                      |
      +====================================================================================+=================================================================+==================================================================================================+
      | dateandtime2tuple(create_dateandtime(1970,1,1,1,0,0,utc_dir="+", utc_hours=1))     | (1970, 1, 1, 1, 0, 0, 0, '+', 1, 0)                             |                                                                                                  |
      +------------------------------------------------------------------------------------+-----------------------------------------------------------------+--------------------------------------------------------------------------------------------------+
      | dateandtime2tuple(create_dateandtime())                                            | (2016, 4, 28, 18, 30, 13, 0, '+', 2, 0)                         |                                                                                                  |
      +------------------------------------------------------------------------------------+-----------------------------------------------------------------+--------------------------------------------------------------------------------------------------+

dateandtime2timestamp
---------------------------------------------

    **Syntax**

      **dateandtime2timestamp** **(**  *dateandtime value* [ , *default_tz* ] **)**

    **Description**

      Convert the given *dateandtime value* byte vector to the corrisponding timestamp UTC.
      
      If the dateandtime *value* is a local time tuple (8 byte, without timezone informations), the *default_tz* parameter
      is used for the computation, otherwise the local timezone is used.
      
      THe *default_tz* string parameter can specify the timezone with two different formats:
      
      1) Timezone code (es. "Europe/Rome")
      2) An offset in the format  ("+"|"-")(<HH>:<MM>). (Es: "+01:00")
      
      When *default_tz* contains an empty string (""), the local timezone is used.
      
      When *dateandtime value* do not contains timezone infos, the function tries to adjust the timestamp with Daylight Saving TIme (DTS) settings for the used timezone (*default_tz* or local timezone) active when the function is called.
      
      **NOTE**: When *default_tz* uses the *offset format* (2) the function do not handle DST.

    **Input**
      
      * *dateandtime value* : a DateAndTime value (8 o 11 byte length vector)
      * *default_tz*: (optional) a string

    **Output**

      A UTC timestamp.
      
    **Errors**
    
      Raise an exception if the input values is not valid dateandtime byte vector or the timezone is not

    **Examples**

      +----------------------------------------------------------------------------------------------------------------+-----------------------------------------------------------------+--------------------------------------------------------------------------------------------------+
      | Expression                                                                                                     | Value                                                           | Description                                                                                      |
      +================================================================================================================+=================================================================+==================================================================================================+
      | dateandtime2timestamp( 1.3.6.1.2.1.25.1.2@ )                                                                   | 1461915587                                                      | Fri Apr 29 07:39:47 UTC 2016 (UTC) / Fri Apr 29 09:39:47 CEST 2016 (localtime)                   |
      +----------------------------------------------------------------------------------------------------------------+-----------------------------------------------------------------+--------------------------------------------------------------------------------------------------+
      | dateandtime2timestamp(  create_dateandtime() )                                                                 | 1461915587                                                      | Fri Apr 29 07:39:47 UTC 2016 (UTC) / Fri Apr 29 09:39:47 CEST 2016 (localtime)                   |
      +----------------------------------------------------------------------------------------------------------------+-----------------------------------------------------------------+--------------------------------------------------------------------------------------------------+
      | dateandtime2timestamp(  create_dateandtime(2016,4,29, 09,30) )                                                 | 1461915000                                                      | Fri Apr 29 07:30:00 UTC 2016 (UTC) / Fri Apr 29 09:30:00 CEST 2016 (localtime)                   |
      +----------------------------------------------------------------------------------------------------------------+-----------------------------------------------------------------+--------------------------------------------------------------------------------------------------+
      | dateandtime2timestamp(  create_dateandtime(2016,4,29, 09,30), "+02:00" )                                       | 1461915000                                                      | Fri Apr 29 07:30:00 UTC 2016 (UTC) / Fri Apr 29 09:30:00 CEST 2016 (localtime)                   |
      +----------------------------------------------------------------------------------------------------------------+-----------------------------------------------------------------+--------------------------------------------------------------------------------------------------+
      | dateandtime2timestamp(  create_dateandtime(2016,4,29, 09,30), "Europe/Rome" )                                  | 1461915000                                                      | Fri Apr 29 07:30:00 UTC 2016 (UTC) / Fri Apr 29 09:30:00 CEST 2016 (localtime)                   |
      +----------------------------------------------------------------------------------------------------------------+-----------------------------------------------------------------+--------------------------------------------------------------------------------------------------+
      | dateandtime2timestamp(  create_dateandtime(2016,4,29, 09,30, utc_dir="+") )                                    | 1461922200                                                      | Fri Apr 29 09:30:00 UTC 2016 (UTC) / Fri Apr 29 11:30:00 CEST 2016 (localtime)                   |
      +----------------------------------------------------------------------------------------------------------------+-----------------------------------------------------------------+--------------------------------------------------------------------------------------------------+
      | dateandtime2timestamp(  create_dateandtime(2016,4,29, 09,30, utc_dir="+"), "America/Chicago" )                 | 1461922200                                                      | Fri Apr 29 09:30:00 UTC 2016 (UTC) / Fri Apr 29 11:30:00 CEST 2016 (localtime)                   |
      +----------------------------------------------------------------------------------------------------------------+-----------------------------------------------------------------+--------------------------------------------------------------------------------------------------+
      | int(dateandtime2timestamp(  create_dateandtime() )) - int(time())                                              | 0                                                               | ok, same current time for time() and create_dateandtime()                                        |
      +----------------------------------------------------------------------------------------------------------------+-----------------------------------------------------------------+--------------------------------------------------------------------------------------------------+

"""

import time
import struct

import datetime

import calendar
import pytz

from sanet_common.exprparser.Types import SnmpValueType

from sanet_poller.lib.exceptions import PollException

#-----------------------------------------------------------------------------#
# FUNZIONI DI COMODO                                                          #
#-----------------------------------------------------------------------------#

def _localTzname():
	"""
	Questa funzione serve per calcolare il nome "genrico" ("Etc/GMT<offset>") della timezone locale a
	dove viene eseguita.
	"""
	if time.daylight:
		offsetHour = time.altzone / 3600
	else:
		offsetHour = time.timezone / 3600
	return 'Etc/GMT%+d' % (offsetHour)

#-----------------------------------------------------------------------------#

def _localoffset(current_date=None):
	"""
	Questa funzioen calcola l'offset in secondi dall'UTC per la timezone locale.
	L'offset tiene conto del Daylight Saving Time (DST) attivo per l'oggetto datetime() specificato in input (current_date) o  datetime.datetime.now().
	"""

	if current_date is None:
		current_date = datetime.datetime.now()
		
	timezone_name = _localTzname()
	
	pst = pytz.timezone(timezone_name)
	
	offset = pst.utcoffset(current_date).total_seconds()
	
	return offset

#-----------------------------------------------------------------------------#

def __decode_dateandtime(val):
	"""
	Prende in input il vettore di byte *val* che codifica un valore DateAndTime a 8 o 11 byte e restituisce
	una tupla:
	
		(year, month, day, hour, minutes, seconds, deci_seconds, None, None, None)                  8 byte
	o
		(year, month, day, hour, minutes, seconds, deci_seconds, "+"|"-", utc_hours, utc_minutes)  11 byte
	"""

	utc_dir     = None
	utc_hours   = None
	utc_minutes = None

	if len(val) == 8: # no timezone
		(year, month, day, hour, minutes, seconds, deci_seconds) = struct.unpack('!hBBBBBB', val)

	elif len(val) == 11:
		(year, month, day, hour, minutes, seconds, deci_seconds, utc_dir, utc_hours, utc_minutes) = struct.unpack('!hBBBBBBBBB', val)
		utc_dir = chr(utc_dir)
	else:
		raise ValueError('Invalid ADVA DateAndTime string: %r (len=%s)' % (val, len(val)))
	
	return (year, month, day, hour, minutes, seconds, deci_seconds, utc_dir, utc_hours, utc_minutes)

#-----------------------------------------------------------------------------#
# FUNZIONI ESPORTATE                                                          #
#-----------------------------------------------------------------------------#

def fun_time(context):
	"""Current GMT time with decimals in a float"""
	return time.time()

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_UTCOffset(context, timezone_name=None):
	"""
	Restituisce il numero di secondi di offset dall'UTC della timezone passata in input.
	Se la timezone non e' specificata, viene restituito l'offset locale.
	"""
	
	current_date = datetime.datetime.now()
	
	if timezone_name is None:
		timezone_name = _localTzname()
	
	
	pst = pytz.timezone(timezone_name)
	
	offset = pst.utcoffset(current_date).total_seconds()
	
	context.getLog().debug("UTCTZOffset(%s) = %s", timezone_name, offset)
	
	return offset

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_create_dateandtime(context, year=None, month=None, day=None, hour=None, minutes=None, seconds=None, deci_seconds=None, utc_dir=None, utc_hours=None, utc_minutes=None):
	"""
	Questa funzione crea un vettore di byte contenente la rappresentazione binaria di un valore
	DateAndTime come da specifiche SNMPV2-TC.
	
	Il vettore in output puo' essere lungo:
	
	* 8 byte  quando la data indicata non contiene informazioni sulla timezone (la data e' quindi in localtime)
	* 11 byte quando la data indicata contiene informazioni sulla timezone.
	
	Se nessuno dei parametri viene specificato, genera un valore dateandtime corrispondente all'ora attuale con il timezone locale.
	
	> ---------------------------------------------------------------------------
	>    DateAndTime ::= TEXTUAL-CONVENTION
	>    DISPLAY-HINT "2d-1d-1d,1d:1d:1d.1d,1a1d:1d"
	>    STATUS       current
	>    DESCRIPTION
	>            "A date-time specification.
	>
	>            field  octets  contents                  range
	>            -----  ------  --------                  -----
	>              1      1-2   year*                     0..65536
	>              2       3    month                     1..12
	>              3       4    day                       1..31
	>              4       5    hour                      0..23
	>              5       6    minutes                   0..59
	>              6       7    seconds                   0..60
	>                           (use 60 for leap-second)
	>              7       8    deci-seconds              0..9
	>              8       9    direction from UTC        '+' / '-'
	>              9      10    hours from UTC*           0..13
	>             10      11    minutes from UTC          0..59
	>
	>            * Notes:
	>            - the value of year is in network-byte order
	>            - daylight saving time in New Zealand is +13
	>
	>            For example, Tuesday May 26, 1992 at 1:30:15 PM EDT would be
	>            displayed as:
	>
	>                             1992-5-26,13:30:15.0,-4:0
	>
	>
	>            Note that if only local time is known, then timezone
	>            information (fields 8-10) is not present."
	>    SYNTAX       OCTET STRING (SIZE (8 | 11))	
	> ---------------------------------------------------------------------------
	"""
	
	
	# Se non viene passato in input nessun parametro, dobbiamo restiture un dateandtime riferito all'ora attule con informazione della timezone in uso.
	if year==None and month==None and day==None and hour==None and minutes==None and seconds==None and deci_seconds==None and utc_dir==None and utc_hours==None and utc_minutes==None:
	
	
		(year, month, day, hour, minutes, seconds, wday, yday, isdst) = time.localtime()
		
		deci_seconds = 0
	
		if time.daylight:
			offsetHour = time.altzone 
		else:
			offsetHour = time.timezone
		
		utc_dir     = '-' if offsetHour > 0 else '+'
		offsetHour = abs(offsetHour)
		utc_minutes = offsetHour % 3600
		utc_hours   = (offsetHour - utc_minutes) / 3600
	else:
		year          = year         if year         is not None else 1970
		month         = month        if month        is not None else 1
		day           = day          if day          is not None else 1
		hour          = hour         if hour         is not None else 0
		minutes       = minutes      if minutes      is not None else 0
		seconds       = seconds      if seconds      is not None else 0
		deci_seconds  = deci_seconds if deci_seconds is not None else 0
		
	
	print year, month, day, hour, minutes, seconds, deci_seconds, utc_dir, utc_hours, utc_minutes
	
	if utc_dir is None:
		print year, month, day, hour, minutes, seconds, deci_seconds
		return struct.pack('!hBBBBBB', year, month, day, hour, minutes, seconds, deci_seconds)
	else:
		utc_hours     = utc_hours    if utc_hours    is not None else 0
		utc_minutes   = utc_minutes  if utc_minutes  is not None else 0

		print year, month, day, hour, minutes, seconds, deci_seconds, ord(utc_dir) if utc_dir is not None else "<NO DIR>", int(utc_hours), int(utc_minutes)
		return struct.pack('!hBBBBBBBBB', year, month, day, hour, minutes, seconds, deci_seconds, ord(utc_dir), int(utc_hours), int(utc_minutes))


#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_dateandtime2tuple(context, val):
	"""
	Converte il valore dateandtime "val" nella corrispondente tupla templorale con 8 o 11 elementi.
	"""

	if isinstance(val, SnmpValueType):
		val = val.getvalue()

	return __decode_dateandtime(val)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def _UTCoffsetspec2seconds(s):
	"""
	Questa funzione converte in secondi la stringa in input che rappresenta una timezone.
	
	I formati possibili della timezone sono due:
	
	1) Codice Timezone (es. Europe/Rome)
	2) Offset   (+|-)(HH:MM)
	
	Se viene specificato il codice della timezone (1) il numero di secondi restituito tiene
	conto anche del Daylight Saving Time (DTS) in essere al momento della chiamata per quella timezone.
	"""
	try:
		pst = pytz.timezone(s)
		current_date = datetime.datetime.now()
		offset = pst.utcoffset(current_date).total_seconds()
		return offset
	except Exception, e:
		pass
	
	try:
		utc_dir = s[0]
		utc_hours, utc_minutes = s[1:].split(':')
		return ((int(utc_hours) * 60 * 60) + (int(utc_minutes) * 60)) * (1 if utc_dir == '+' else -1)
	except TypeError, e:
		pass
	except ValueError, e:
		pass

	raise PollException("Invalid timezone string: %s (%s)" % (s, type(s)) )


def fun_dateandtime2timestamp(context, val, default_tz=None):

	if isinstance(val, SnmpValueType):
		val = val.getvalue()

	log = context.getLog()

	log.debug("calling dateandtime2timestamp( default_tz=%s )", default_tz)

	# *val* dovrebbe essere un vettore di byte.
	val = str(val)
	if len(val) == 0:
		raise PollException("Input dateandtime value (%r) is invalid." % (val))

	#log.debug("Converting date: %s", repr(val))

	# Decodifichiamo la tupla.
	t_tuple =  __decode_dateandtime(val)

	log.debug("Calculated time tuple from dateandtime value: %s", t_tuple)

	t_struct_time = ( 
		t_tuple[0],  # tm_year	(for example, 1993)                                          
		t_tuple[1],  # tm_mon	range [1, 12]                                         
		t_tuple[2],  # tm_mday	range [1, 31]                                         
		t_tuple[3],  # tm_hour	range [0, 23]                                         
		t_tuple[4],  # tm_min	range [0, 59]                                         
		t_tuple[5],  # tm_sec	range [0, 61]; see (2) in strftime() description                                         
		0,           # tm_wday	range [0, 6], Monday is 0                                         
		0,           # tm_yday	range [1, 366]                                         
		0            # tm_isdst	0, 1 or -1; see below                                         
	)
	
	# Convertiamo la tupla temporale in secondi UTC e non il localtime
	# perche' tanto gli aggiustamenti dell'offset della timezone li effettuiamo dopo.
	t = calendar.timegm( t_struct_time )
	
	log.debug("Calculated timegm timestamp: %s", t)
	
	# dateandtime with timezoneinfo
	if t_tuple[7] is not None:  
	
		utc_offset = ((t_tuple[8] * 60 * 60) + (t_tuple[9] * 60)) * (1 if t_tuple[7] == '+' else -1)
		
	# dateandtime is in localtime
	else: 
		log.debug("datetimetuple is local time only")
		
		if default_tz is not None and default_tz != "":
			log.debug("Forced timezone string: %s", default_tz)
			utc_offset = _UTCoffsetspec2seconds(default_tz)
		else:
			log.debug("No timezone specified. Calculating local offset")
			utc_offset = _localoffset()
	
	log.debug("UTC Timezone offset: %s", utc_offset )
	
	t -= utc_offset
	
	log.debug("Calculated final timestamp : %s", t)
	
	log.debug("Final:")
	log.debug("\tOriginal dateandtime is: %s", t_tuple)
	log.debug("\tReturned timestamp     : %s (UTC %s)", t, time.asctime(time.gmtime(t))   )
	log.debug("\tCurrent UTC TS is      : %s", time.time())

	return t

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#


EXPR_SYMBOLS = {

	'time'                  : fun_time,

	'UTCOffset'             : fun_UTCOffset,
	
	'create_dateandtime'    : fun_create_dateandtime,

	'dateandtime2timestamp' : fun_dateandtime2timestamp,
	
	'dateandtime2tuple'     : fun_dateandtime2tuple,
}
