# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it> 
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

===
UDP
===

isUdpOpen
---------------------------------------------

    **Syntax**

      **isUdpOpen** **(** *host* **,** *port* [ , *options* ] **)**

    **Description**

      Verifies if a UDP port is open on the remote *host*.

    **Input**

      * *host*: hostname or IP address.
      * *port*: port number
      * *timeout* : **Optional**. set the maximum response timeout. (Default: 3 seconds)
      * *open_on_timeout* : **Optional**, forces to return *True* on timeout. (Default: True)

    **Output**

      Returns a *boolean*

    **Errors**

      The function returns an uncheckable value when any unmanaged network error occurs.

    **Examples**

	 +---------------------------------------------------------+--------------+------------------------------------------------------------------------------------------------------------------------------------------+
	 | Expression                                              | Return value | Description                                                                                                                              |
	 +=========================================================+==============+==========================================================================================================================================+
	 | isUdpOpen("mydns", 53, timeout=10)                      | *True*       | Received 12 bytes.,udpOpen($node, 53, timeout=3) expands to True                                                                         |
	 +---------------------------------------------------------+--------------+------------------------------------------------------------------------------------------------------------------------------------------+
	 | isUdpOpen("mydns", 666,)                                | *False*      | Connection refused. UDP port 666 closed,udpOpen($node, 666, timeout=3) expands to False                                                  |
	 +---------------------------------------------------------+--------------+------------------------------------------------------------------------------------------------------------------------------------------+
	 | isUdpOpen("ns2.google.com", 666, open_on_timeout=True)  | *True*       | Connection timeout. Port 666 filtered?,udpOpen("ns2.google.com", 666, timeout=3, open_on_timeout=True) expands to True                   |
	 +---------------------------------------------------------+--------------+------------------------------------------------------------------------------------------------------------------------------------------+
 
"""
import socket

from sanet_poller.lib.exceptions import PollException

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_isUdpOpen(context, node, port, timeout=3, open_on_timeout=True):

	log = context.getLog()

	s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
	
	s.settimeout(timeout)
	
	s.connect( (node, port) )  # UDP sockets receive ICMP errors only when connected
	
	log.debug("Sending UDP probe with timeout %ss" % timeout)
	s.sendto( 'sanetudptest' , (node, port) )
	
	try:
		log.debug("Waiting for response...")
		d = s.recv(1024)
		context.appendReason("Received %s bytes." % (len(d)))

	except socket.timeout, e:
		context.appendReason("Connection timeout. Port %s filtered?" % (port))
		if open_on_timeout:
			return True
		else:
			return False
	except socket.error, e:
		if e.errno == 111:  # Connection refused
			context.appendReason("Connection refused. UDP port %s closed" % (port))
			return False
		else:
			raise PollException("Error while testing UDP port: %s" % (e))
	except Exception, e:
		raise PollException("Error while testing UDP port %s: %s (%s)" % (port, e, type(e)))

	return True

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

EXPR_SYMBOLS = {
	'isUdpOpen'   : fun_isUdpOpen,
}

