"""

.. _saveval_getlastval:

********************************************************
Execution values 
********************************************************

.. _saveval:

saveval
=======================

    **Syntax**
    
      **saveval** **(** *name*, *value* **)**

    **Description**

      Save the value *value* with the name *name*. Returns *value*. The value of *name* will be available
      the next time the expression is evaluated. See :ref:`getlastval`.

    **Input**

      1. *name*: string
      2. *value*: any value

    **Outpu**

      Returns the value of *value*.

    **Errors**

      None

    **Examples**

      +----------------------------------------------------+-------------------------------------+-------------------------------------------------------+
      | Expression                                         | Return value                        |  Description                                          |
      +====================================================+=====================================+=======================================================+
      | saveval("x", 1)                                    | 1                                   | Stores 1 in the internal state for the next execution |
      +----------------------------------------------------+-------------------------------------+-------------------------------------------------------+
      | saveval("x", walkCount(1.3.6.1.2.1.1.1.@)          | The value of the walkCount()        | Packet loss average in the last 180 seconds.          |
      +----------------------------------------------------+-------------------------------------+-------------------------------------------------------+
      | saveval("x", ....) - getlastval("x") < $delta      | *true* or *false*                   | Compare the current value of "x" with the last.       |
      +----------------------------------------------------+-------------------------------------+-------------------------------------------------------+

.. _getlastval:

getlastval
=======================

    **Syntax**
    
      **getlastval** **(** *name* **)**

    **Description**

      Returns the value of *name* saved at the last execution time by the function *saveval()* (See: :ref':`saveval`.)

    **Input**

      1. *name*: string

    **Outpu**

      Any value

    **Errors**

      Raise exception in there is no value for the given variable name.

    **Examples**

      +----------------------------------------------------+-------------------------------------+-------------------------------------------------------+
      | Expression                                         | Return value                        |  Description                                          |
      +====================================================+=====================================+=======================================================+
      | getlastval("x")                                    | *a value*                           |  The the last value of "x".                           |
      +----------------------------------------------------+-------------------------------------+-------------------------------------------------------+
      | saveval("x", ....) - getlastval("x") < $delta      | *true* or *false*                   |  Compare the current value of "x" with the last.      |
      +----------------------------------------------------+-------------------------------------+-------------------------------------------------------+
     
"""

from sanet_common.exprparser.Types import SnmpValueType

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_saveval(context, val_name, value):

	log = context.getLog()

	#
	# Snmp expressions are "lazy". Remember to retrieve the value
	# from the network before saving it.
	#
	if isinstance(value, SnmpValueType):
		value = value.getvalue()

	log.debug("Saving value %s = %s", val_name, value)
	
	context.saveCurrentValue(val_name, value)

	return value

def fun_getlastval(context, val_name):
		
	log = context.getLog()
	
	log.debug("Getting old value for '%s'...", val_name)

	value = context.getOldValue(val_name)

	return value

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

EXPR_SYMBOLS = {

	'saveval'   : fun_saveval,
	'getlastval': fun_getlastval,
}	
