"""
********************************************************
VMware tools
********************************************************

vmware_esx
=======================


	**Syntax**

		**vmware_esx** **(** *username, password, esx_server, select, subselect [, sslport] [, vmware_script_path] [, timeout] [, whitelist] [, is_datacenter]* **)**

	**Description**

		The function returns the selected information about the ESX server.
		
		+--------------------+----------------------------------------------------------------------------------------------------------------------------------+
		| Argument           | Description                                                                                                                      |
		+====================+==================================================================================================================================+
		| username           | VMware Username                                                                                                                  |
		+--------------------+----------------------------------------------------------------------------------------------------------------------------------+
		| password           | VMware Password                                                                                                                  |
		+--------------------+----------------------------------------------------------------------------------------------------------------------------------+
		| esx_server         | ESX hostname                                                                                                                     |
		+--------------------+----------------------------------------------------------------------------------------------------------------------------------+
		| select             | Could be "cpu" (CPU information), "mem" (memory information), "runtime" (runtime information) or "storage" (storage information) |
		+--------------------+----------------------------------------------------------------------------------------------------------------------------------+
		| subselect          | See table below                                                                                                                  |
		+--------------------+----------------------------------------------------------------------------------------------------------------------------------+
		| sslport            | If a SSL port different from 443 is used                                                                                         |
		+--------------------+----------------------------------------------------------------------------------------------------------------------------------+
		| vmware_script_path | If a VMware script path different from "*/usr/local/sanet-check-vmware/check_vmware_esx.pl*" is used                             |
		+--------------------+----------------------------------------------------------------------------------------------------------------------------------+
		| timeout            | If a timeout different from 10 seconds is setted (in milliseconds)                                                               |
		+--------------------+----------------------------------------------------------------------------------------------------------------------------------+
		| whitelist          | whitelist as regexp                                                                                                              |
		+--------------------+----------------------------------------------------------------------------------------------------------------------------------+
		| is_datacenter      | Flag (1 o 0) if the esx_server is a normal server o  datacenter                                                                  |
		+--------------------+----------------------------------------------------------------------------------------------------------------------------------+
		
		Valid *subselect* arguments are:
		
			1. cpu
				- ready (ms): time that the virtual machine was ready, but could not get scheduled to run on the physical CPU. 
				- wait (ms): CPU time spent in wait state. The wait total includes time spent the CPU idle, CPU swap wait, and CPU I/O wait states.
				- usage (%): actively used CPU of the host, as a percentage of the total available CPU.
				
			2. mem
				- usage (%): average mem usage in percentage.
				- consumed (byte): amount of machine memory used on the host.
				- swapused (byte): amount of memory that is used by swap.
				- overhead (byte): additional mem used by VM Server.
				- memctl (byte): the sum of all vmmemctl values for all powered-on virtual machines, plus vSphere services on the host.
				
			3. runtime
				- con: shows connection state (connected, disconnected, inaccessible, invali, orphaned).
				- status: overall object status. The state can be gray (unknown), green (ok), red (definitely has a problem), yellow (might have a problem).
				- health: checks cpu/storage/memory/sensor status.
				- issues: lists all configuration issues for the host. Output could be "All X health checks are GREEN" or could show some problem.
				- storagehealth: local storage status check.
				- listhost: list all host and their status.
				
			4. storage
				- lun: list SCSI logical units with their status.

		
	**Input**

		1. username: string
		2. password: string
		3. esx_server: string
		4. select: string (cpu, mem, runtime)
		5. subselect: string (see above)
		6. sslport: string or integer (default: 443)
		7. vmware_script_path: string (default: "*/usr/local/sanet-check-vmware/check_vmware_esx.pl*")
		8. timeout: integer (default: 10 seconds) in milliseconds
		9. whitelist: string (default: None), available only for commands:
			- select=runtime, subselect=health
			- select=runtime, subselect=issues
			- select=runtime, subselect=storagehealth
			- select=storage, subselect=lun
		10. is_datacenter: boolean (default: False)

		If whitelist is used for not supported commands the output will be unchanged.
		
		
	**Output**

		Returns the requested information. 

	**Errors**

		Raise an exception if an error occours.

	**Examples**

		+-----------------------------------------------------------------------+--------------------------------------+-------------------------------------------------------+
		| Expression                                                            | Return value                         | Description                                           |
		+=======================================================================+======================================+=======================================================+
		| vmware_esx("<username>", "<password>", "virtuoso", "cpu", "ready")    | 183.00                               | time that the virtual machine was ready in ms         |
		+-----------------------------------------------------------------------+--------------------------------------+-------------------------------------------------------+
		| vmware_esx("<username>", "<password>", "virtuoso", "mem", "consumed") | 6467753082.88                        | amount of machine memory used on the host in byte     |
		+-----------------------------------------------------------------------+--------------------------------------+-------------------------------------------------------+
		| vmware_esx("<username>", "<password>", "virtuoso", "runtime", "con")  | OK: connection state=connected       | connection state                                      |
		+-----------------------------------------------------------------------+--------------------------------------+-------------------------------------------------------+
		
		
vmware_vm
=======================


	**Syntax**

		**vmware_vm** **(** *username, password, esx_server, vm_name, sel, subselect [, sslport, vmware_script_path, timeout]* **)**

	**Description**

		The function returns the selected information about the virtual machine.
		
		+--------------------+------------------------------------------------------------------------------------------------------------------------+
		| Argument           | Description                                                                                                            |
		+====================+========================================================================================================================+
		| username           | VMware Username                                                                                                        |
		+--------------------+------------------------------------------------------------------------------------------------------------------------+
		| password           | VMware Password                                                                                                        |
		+--------------------+------------------------------------------------------------------------------------------------------------------------+
		| esx_server         | ESX hostname                                                                                                           |
		+--------------------+------------------------------------------------------------------------------------------------------------------------+
		| vm_name            | Virtual machine name                                                                                                   |
		+--------------------+------------------------------------------------------------------------------------------------------------------------+
		| select             | Could be "cpu" (CPU information), "mem" (memory information) or "runtime" (runtime information)                        |
		+--------------------+------------------------------------------------------------------------------------------------------------------------+
		| subselect          | See the table below                                                                                                    |
		+--------------------+------------------------------------------------------------------------------------------------------------------------+
		| sslport            | If a SSL port different from 443 is used                                                                               |
		+--------------------+------------------------------------------------------------------------------------------------------------------------+
		| vmware_script_path | If a VMware script path different from "*/usr/local/sanet-check-vmware/check_vmware_esx.pl*" is used                   |
		+--------------------+------------------------------------------------------------------------------------------------------------------------+
		| timeout            | If a timeout different from 10 seconds is setted (in milliseconds)                                                     |
		+--------------------+------------------------------------------------------------------------------------------------------------------------+
		
		
		Valid *subselect* arguments are:
		
			1. cpu
				- ready (ms): time that the virtual machine was ready, but could not get scheduled to run on the physical CPU. 
				
				- wait (ms): CPU time spent in wait state. The wait total includes time spent the CPU idle, CPU swap wait, and CPU I/O wait states.
				
				- usage (%): actively used CPU of the host, as a percentage of the total available CPU.
				
			2. mem
				- usage (%): average mem usage in percentage.
				
				- consumed (byte): amount of machine memory used on the host.
				
				- memctl (byte): the sum of all vmmemctl values for all powered-on virtual machines, plus vSphere services on the host.
				
			3. runtime
				- con: shows connection state (connected, disconnected, inaccessible, invali, orphaned).
				
				- powerstate: virtual machine power state (poweredOn, poweredOff, suspended).
				
				- status: overall object status. The state can be gray (unknown), green (ok), red (definitely has a problem), yellow (might have a problem).
				
				- consoleconnections: console connections to VM.
				
				- gueststate: guest OS status. Can be: running, shuttingdown, resetting, standby, notrunning, unknown.
				
				- tools: Vmware tools status. Can be: 
				    * VMware tools are starting.
				    * VMware tools are not running.
				    * VMware tools are running.
				    * VMware tools are installed, but the installed version is known to have a grave bug and should be immediately upgraded.
				    * VMware tools are installed, but the version is not current.
				    * VMware tools were never been installed.
				    * VMware tools are installed, supported, and newer than the versionavailable on the host.
				    * No information about VMware tools available. 
				  New since vSphere API 5.0:
				    * VMware tools are installed, and the version is current.
				    * VMware tools are installed, supported, but a newer version is available.
				    * VMware tools are installed, and the version is known to be too new to work correctly with this virtual machine.
				    * VMware tools are installed, but the version is too old.
				    * VMware tools are installed, but it is not managed by VMWare.
				
				- issues: all issues for the host. Can be: X config issues.
		
	**Input**

		1. username: string
		2. password: string
		3. esx_server: string
		4. vm_name: string
		5. select: string (cpu, mem, runtime)
		6. subselect: string (see above)
		7. sslport: string or integer (default: 443)
		8. vmware_script_path: string (default: "*/usr/local/sanet-check-vmware/check_vmware_esx.pl*")
		9. timeout: integer (default: 10 seconds) in milliseconds
		
		
	**Output**

		Returns the requested information. 

	**Errors**

		Raise an exception if an error occours.

	**Examples**

		+--------------------------------------------------------------------------------------+-------------------------------------+-------------------------------------------------------+
		| Expression                                                                           | Return value                        | Description                                           |
		+======================================================================================+=====================================+=======================================================+
		| vmware_vm("<username>", "<password>", "virtuoso", "loggit", "cpu", "usage")          | 0.48                                | actively used CPU as a percentage                     |
		+--------------------------------------------------------------------------------------+-------------------------------------+-------------------------------------------------------+
		| vmware_vm("<username>", "<password>", "virtuoso", "loggit", "mem", "memctl")         | 0.0                                 | sum of all vmmemctl values in byte                    |
		+--------------------------------------------------------------------------------------+-------------------------------------+-------------------------------------------------------+
		| vmware_vm("<username>", "<password>", "virtuoso", "loggit", "runtime", "powerstate") | OK: Power state: poweredOn          | power state                                           |
		+--------------------------------------------------------------------------------------+-------------------------------------+-------------------------------------------------------+
      
"""



import os, re
import subprocess
import select
import time

from sanet_poller.lib.exceptions import PollException

VMWARE_TIMEOUT = 10000 #10 seconds

VMWARE_DEFAULT_SCRIPT_PATH = "/usr/local/sanet-check-vmware/check_vmware_esx.pl"


# ESEMPI #

#vmware_esx(username, password,   esx_server, select, subselect, sslport=None, vmware_script_path=None, timeout=None)
#vmware_esx("devz",   <password>, "virtuoso", "cpu",  "ready")
#vmware_esx("devz",   <password>, "virtuoso", "cpu",  "ready",   sslport=111, vmware_script_path="/usr/local/vattelappesca")
#vmware_esx("devz",   <password>, $node,      "cpu",  "ready",   sslport=111)

#vmware_vm(username, <password>, esx_server,     vm_name, select, subselect, sslport)
#vmware_vm(username, <password>, "server_1",     $node,   select, subselect, sslport)
#vmware_vm(username, <password>, $cluster_node , $node,   select, subselect, sslport)


###### HOST COMMAND LIST ########

# il valore nei dizionari e' utile per il parsing e il caching

HOST_MEM_COMMAND = {
	'usage'   : "mem_usage"      ,
	'consumed': "consumed_memory",
	'swapused': "mem_swap"       ,
	'overhead': "mem_overhead"   ,
	'memctl'  : "mem_memctl"     ,
}

HOST_CPU_COMMAND = {
	'wait' : 'cpu_wait' ,
	'ready': 'cpu_ready',
	'usage': 'cpu_usage',
}

HOST_RUNTIME_COMMAND = {
	'con'          : 'connection state',
	'status'       : 'overallstatus'   ,
	'health'       : 'health'          ,
	'issues'       : 'issues'          ,
	'storagehealth': 'storagehealth'   ,
	'listhost'     : 'listhost',
}

HOST_STORAGE_COMMAND = {
	'lun': 'lun',
}

HOST_COMMAND_LIST = {
	'mem'    : HOST_MEM_COMMAND,
	'cpu'    : HOST_CPU_COMMAND,
	'runtime': HOST_RUNTIME_COMMAND,
	'storage': HOST_STORAGE_COMMAND,
}

########## VIRTUAL MACHINE COMMAND LIST ########

VM_MEM_COMMAND = {
	'usage'   : "mem_usage"      ,
	'consumed': "consumed_memory",
	'overhead': "mem_overhead"   ,
	'active'  : "mem_active"     ,
	'memctl'  : "memctl"         ,
}

VM_CPU_COMMAND = {
	'wait' :'cpu_wait' ,
	'ready':'cpu_ready',
	'usage':'cpu_usage',
}

VM_RUNTIME_COMMAND = {
	'con'               : 'Connection state'   ,
	'powerstate'        : 'Power state'        ,
	'status'            : 'Overall status'     ,
	'consoleconnections': 'Console connections',
	'gueststate'        : 'Guest state'        ,
	'tools'             : 'Tools state'        ,
	'issues'            : 'issues'             ,
	'listhost'          : 'List hosts',
}

VM_COMMAND_LIST = {
	'mem'    : VM_MEM_COMMAND,
	'cpu'    : VM_CPU_COMMAND,
	'runtime': VM_RUNTIME_COMMAND,
}


# per ottimizzare i tempi (siccome lo script in perl e' lento...) ove possibile non viene eseguito il comando
# specificando il "subselect", ma solo il "select". In questo modo con una sola chiamata si ottengono piu'
# informazioni e si salvano in cache. Alcuni comandi pero' se eseguiti solo con "select" o anche con "subselect"
# forniscono risultati differenti. Per questi comandi (e solo per questi) viene eseguito il comando comprensivo di "subselect".

NOT_CACHED_COMMANDS = [
	'storage',
	'runtime',
]

# per alcuni comandi e' possibile specificare una whitelist:
# tuple formate da (<select>,<subselect>)

WHITELIST_SUPPORTED_COMMAND = [
	('runtime','health'       ),
	('runtime','issues'       ),
	('runtime','storagehealth'),
	('storage','lun'          ),
]


# controlla se i parametri passati sono supportati

def check_command(sel, subselect, command_list):
	
	if sel not in command_list:
		raise PollException("Malformed command: %s not in supported commands" % (sel))
		
	if subselect not in command_list[sel]:
		raise PollException("Malformed command: select=%s subselect=%s not in supported commands" % (sel, subselect))



# inserisce i valori nella cache

def cache_value(context, command_list, sel, output, vm = 0):
	
	for command in command_list:
		if sel == command:
			for c in command_list[command]:
				tocache = parse_value(command_list[command][c], output, command, c, vm)
				context.cacheSimpleValue(command_list[command][c], tocache)


# effettua il parsing dell'output dello scripit per adattarlo a come lo vogliamo mostrarlo

def parse_value(param, output, sel, subselect, vm = 0):
	
	try:
		
		if sel == "mem" or sel == "cpu":
			str_start = "'" + param + "'="
			str_end = ";"
			value = (output.split(str_start))[1].split(str_end)[0]

			if subselect == "usage": # e' in percentuale
				try:
					float(value[:-1])
					value = value[:-1]
				except ValueError, e: # potrebbe fornire un output del tipo 'Not available'
					pass

			else:
				try:
					float(value[:-2])
					value = value[:-2]
					
					if sel == "mem": # megabyte, li convertiamo in byte
						value = float(value) * 1024 * 1024

				except ValueError, e: # potrebbe fornire un output del tipo 'Not available'
					pass
				
				
		else: # sel == runtime or sel == storage per gli host // sel == runtime per le VM
			if subselect == 'health':
				regex = re.compile("All (.*) health checks are GREEN")
				match = regex.search(output)
			
				if match == None:
					value = output
				else:
					value = "All " + match.group(1) + " health checks are GREEN"
			
			elif (vm == 0 and subselect == 'issues') or (subselect == 'storagehealth') or (sel == 'storage'): # get the first line
				value = output.split('\n', 1)[0]
			
			else: # forniamo l'output cosi' com'e'
				value = output
			
		
		return value

	# Problemi generici con output non riconosciuto. Non tutti i casi li abbiamo gestiti
	except IndexError, e:
		raise PollException("Output format not recognized. Output: %s" % (output) )

	# La stringa potrebbe essere formattata male o contenere caratteri invalidi, ecc. Gestiamo genericamente qualunque errore di parsing.
	except Exception, e:
		raise PollException("Unknown error while parsing data: %s %s. Data: %s" % (type(e), str(e), output))


# esegue lo script in perl

def exec_command(context, command, timeout, log):
	
	env = os.environ.copy()
	
	log.log(1, "Executing: %s", ' '.join(command))
	process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=False, close_fds=True, env=env)

	rd, wr, er = select.select([process.stdout], [], [process.stderr], float(timeout) / 1000)
	
	if rd or er:
		log.debug("Incoming output from process...") 

		tstart = time.time()
		process.wait()
		tend   = time.time()

		msg = 'Process terminated after %.3fs. Retcode=%s.' % (tend-tstart, process.returncode)
		log.debug(msg)
	
	if process.returncode == None:
		raise PollException("Timeout")
		
	error_msgs = ''
	log.debug("Error messages? %s", len(er))
	if er:
		error_msgs = process.stderr.read()
		
	if error_msgs:
		raise PollException("Error message from check_vmware_esx.pl: %s" % (error_msgs))

	output = process.stdout.read()
	if output == None or output == "":
		raise PollException(process.stderr.read())
	
	log.debug("OUTPUT: len=%s content=%s", len(output), output)


	return output



# funzione per monitorare gli host

def fun_vmware_esx(context, username, password, esx_server, sel, subselect, sslport=443, vmware_script_path=VMWARE_DEFAULT_SCRIPT_PATH, timeout=VMWARE_TIMEOUT, whitelist=None, is_datacenter=False):
	
	log = context.getLog()
	
	param = None
	value = None
	
	check_command(sel, subselect, HOST_COMMAND_LIST)
	
	#bisogna fare questo controllo perche' lo script non e' omogeneo nel formato di output.
	param = HOST_COMMAND_LIST[sel][subselect]
	
	option_H_or_D = '-H' if not is_datacenter else '-D'
	
	if sel in NOT_CACHED_COMMANDS:
		
		if whitelist != None and (sel,subselect) in WHITELIST_SUPPORTED_COMMAND:
			command = [vmware_script_path, "-W", whitelist, option_H_or_D, esx_server, "-u", username, "-p", password, "--sslport", str(sslport), "-S", sel, "-s", subselect]
				
		else:
			command = [vmware_script_path, option_H_or_D, esx_server, "-u", username, "-p", password, "--sslport", str(sslport), "-S", sel, "-s", subselect]
		
		output = exec_command(context, command, timeout, log)
		
		value = parse_value(param, output, sel, subselect)
		
	else:
		cached_value = context.getCachedSimpleValue(param)
		
		if cached_value == None: #non e' nella cache
			
			command = [vmware_script_path, option_H_or_D, esx_server, "-u", username, "-p", password, "--sslport", str(sslport), "-S", sel]
			output = exec_command(context, command, timeout, log)
			
			value = parse_value(param, output, sel, subselect)
			
			#se non e' nella cache devo inserirglielo
			cache_value(context, HOST_COMMAND_LIST, sel, output)
			
		else: #e' nella cache
			value = cached_value
	
	return value



# funzione per monitorare le macchine virtuali

def fun_vmware_vm(context, username, password, esx_server, vm_name, sel, subselect, sslport=443, vmware_script_path=VMWARE_DEFAULT_SCRIPT_PATH, timeout=VMWARE_TIMEOUT):
	
	log = context.getLog()
	
	param = None
	value = None
	
	check_command(sel, subselect, VM_COMMAND_LIST)
	
	#bisogna fare questo controllo perche' lo script non e' omogeneo nel formato di output.
	param = VM_COMMAND_LIST[sel][subselect]


	if sel in NOT_CACHED_COMMANDS:
		
		command = [vmware_script_path, '-H' , esx_server, "-N", vm_name, "-u", username, "-p", password, "--sslport", str(sslport), "-S", sel, "-s", subselect]
		output = exec_command(context, command, timeout, log)
		
		value = parse_value(param, output, sel, subselect)
	
	else:
		
		cached_value = context.getCachedSimpleValue(param)
		
		if cached_value == None: #non e' nella cache
			
			command = [vmware_script_path, '-H', esx_server, "-N", vm_name, "-u", username, "-p", password, "--sslport", str(sslport), "-S", sel]
			output = exec_command(context, command, timeout, log)
			
			value = parse_value(param, output, sel, subselect, vm = 1)
			
			#se non e' nella cache devo inserirglielo
			cache_value(context, VM_COMMAND_LIST, sel, output, vm = 1)
			
		else: #e' nella cache
			value = cached_value
	

	return value
	



EXPR_SYMBOLS = {
	'vmware_esx': fun_vmware_esx,
	'vmware_vm' : fun_vmware_vm,
}
