# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>. 

"""


HTTP
========================================================================

urlContentDoesntMatch
---------------------------------------------

    **Syntax**

        **urlContentDoesntMatch** **(** *url* **,** *regexpr* [ **,** *case_insensitive_flag* [**,** *optional params* ] ] **)**

    **Description**

        Verifies if content of the given *url*  **does not match** the regular expression *regexpr_str*.
        
        See the function *urlContentMatches()* for more details.


.. _sanet-poller.lib.functions.urlContentMatches:

urlContentMatches
---------------------------------------------

    **Syntax**
      
        **urlContentMatches** **(** *url* **,** *regexpr* [ **,** *case_insensitive_flag* [**,** *optional params* ] ] **)**
      
    **Description**
        
        Verifies if content of the given *url*  matches the regular expression *regexpr_str*.
        
        The parameter *case_insensitive_flag* is used to switch to *case insensitive* mode.
        
        The parameter *regexpr_str* can be enclosed between *"/"* characters (eg. *"/foo/"*). In this case the case insensitive test can be enabled by appending a *"i"* at the end of the expression. (eg: *"/foo/i"*).
        The *"i"* character modifier has precedence on the *case_insensitive* parameter.
        
        The *timeout* parameter is an *optional* parameter and is used to set the maximum connection timeout in seconds (default: 20 seconds).

    **Optional parameters**
        
        This is the list of optional parameters:
        
            +----------------+-------------------------------------------------------------------------+
            | Variable       | Description                                                             |
            +================+=========================================================================+
            | include_data   | Include http data in the verbose status (valid only for conditions)     |
            +----------------+-------------------------------------------------------------------------+
            | http_headers   | String containing HTTP Headers (use "\\n" for multi headers)             |
            +----------------+-------------------------------------------------------------------------+
            | use_tls1       | true/false to force TLSv1 instead of SSL2/3                             |
            +----------------+-------------------------------------------------------------------------+
            | postdata       | String containing data for POST request                                 |
            +----------------+-------------------------------------------------------------------------+
            | ciphers        | OpensSSL Ciphers String                                                 |
            +----------------+-------------------------------------------------------------------------+
            | no_fix_ssl3    | (default False) Disable all special Sanet3 fixes for SSL3 and TLS.      |
            +----------------+-------------------------------------------------------------------------+
            | ignore_errors  | Specify how to handle errors. See :ref:`Ignore errors <ignore-errors>`. |
            +----------------+-------------------------------------------------------------------------+

        .. important:: You should not change no_fix_ssl3 defeault if you python version is < 2.7.9.
      
    **Use HTTP Basic/NTML Authentication**
        
        In order to use a HTTP basic/NTML authentication, the following *optional parameters* must be defined:
        
            +-----------------+-------------------------------------------------------------------+
            | Variable        | Description                                                       |
            +=================+===================================================================+
            | basic_user      | username                                                          |
            +-----------------+-------------------------------------------------------------------+
            | basic_password  | password                                                          |
            +-----------------+-------------------------------------------------------------------+
            | basic_ntlm      | true/false. Switch to NTML authentication. (Default False)        |
            +-----------------+-------------------------------------------------------------------+
            | basic_challenge | true/false. Send authentication info challenge (Default: True)    |
            +-----------------+-------------------------------------------------------------------+
        
        This parameters can be passed to the function as context variables too ( $basic_user, $basic_password ).
    
    **Use a proxy**
        
        In order to use a HTTP proxy, the following *optional parameters* must be defined:
        
            +----------------+-------------------------------------------------------------------+
            | Variable       | Description                                                       |
            +================+===================================================================+
            | proxy_protocol | "http" (default) or "https"                                       |
            +----------------+-------------------------------------------------------------------+
            | proxy_server   | Proxy's hostname                                                  |
            +----------------+-------------------------------------------------------------------+
            | proxy_port     | Proxy's port                                                      |
            +----------------+-------------------------------------------------------------------+
        
        If you need proxy authentication you must define the following parameters:
        
            +----------------+-------------------------------------------------------------------+
            | Variable       | Description                                                       |
            +================+===================================================================+
            | proxy_realm    | Proxy's realm (Can be \*)                                         |
            +----------------+-------------------------------------------------------------------+
            | proxy_user     | Username                                                          |
            +----------------+-------------------------------------------------------------------+
            | proxy_password | Password                                                          |
            +----------------+-------------------------------------------------------------------+
            | proxy_ntlm     | NTLM support flag. Set to 1 to enable NTLM authentication         |
            +----------------+-------------------------------------------------------------------+
        
        This parameters can be passed to the function as context variables too ( $basic_user, $basic_password ).

    **Use local Proxy and HTTP Basic Authentication on remote server**
    
        You can use Proxy support and basic authentication at the same time. Remember that proxy authentication is performed first.
    
    **Use local Proxy with NTLM Authentication**
    
        **IMPORTANT**: this feature is enabled only if NTLM python modules are installed.
        
        If *proxy_ntlm* is set to 1, NTLM Proxy Authentication is used instead of HTTP Proxy Authentication.
        
        Parameter *proxy_realm* is used used as the DOMAIN name. The real user used for the authentication is:
        
            ::
                
                <domain>\<user>
        
        Es:
        
            ::
                
                proxy_user = "foo"     
                proxy_realm = "DOMAIN"
    
    **POST requests**
    
        The HTTP request will be a POST instead of a GET when the *postdata* parameter is provided. 
        
        The value can be:
        
             a) a string  encoded in the standard **application/x-www-form-urlencoded**.

                 ::
                     
                     postdata="var1=value1&var2=value2%20123"
                 
                Will send this data:
                
                 ::
                     
                     var1: value1
                     var2: value2 123
             
             b) a dictionary objects <key, value> created with the function :ref:`function-dict`.
         
                 ::
                     
                     postdata=dict("var1","value1", "var2","value2", "var3","value3")
                 
                Will send this data:
                
                 ::
                     
                     var1: value1
                     var2: value2
                     var3: value3
                     
                  
        
    .. _ignore-errors:

    **Ignore errors**
        
        **All errors are ignored by default**.
        
        The parameter *ignore_errors* can change this behaviour. It is a comma-separated sequence of *tokens*:
        
            ::
                
                token ::= "NONE" | "ALL" | ["+","-"] <type>
                type  ::= "TIMEOUT" | "HTTP" | "SSL" | "OTHERS"
        
        The *type* string can be:
        
            * TIMEOUT: network connection's timeout.
            * HTTP: HTTP/S protocol errors.
            * OTHERS: Any other error not handled by other types.
        
        .. warning:: HTTP errors may include proxy errors.
        
        The "+" or "-" flag is optional. 
        
            * "+": the specified error type is ignored. **THIS IS THE DEFAULT**.
            * "-": the specified error type is NOT ignored and will raise an exception. 
        
        You can specify an arbitrary sequence of *tokens*. It will be evaluted from left to right. 
        
            ::
                
                Sequence             Behaviour
                ---------            ----------
                
                NONE,ALL             ALL wins. Ignore all errors
                NONE,+TIMEOUT        DO NOT IGNORE all kind of errors, BUT IGNORE TIMEOUT.
                ALL,-SSL             ignore all errors, BUT raise SSL errors
                ALL,-SSL,-HTTP       ignore all errors, BUT raise SSL and HTTP errors
                
                ALL,-SSL,NONE        NONE wins. DO NOT IGNORE ANY ERROR
                
                ALL,+SSL             Equivalent to ALL
                NONE,-SSL            Equivalent to NONE
        
    **Output**
        
        Returns a *boolean*
        
    **Errors**
        
        An exception is raised when error is detected.
        
    **Examples**
        
        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+
        | Expression                                                                                                            | Return Value | Description                                                                  |
        +=======================================================================================================================+==============+==============================================================================+
        | urlContentMatches("http://www.google.com", "google")                                                                  | *true*       |                                                                              |
        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+
        | urlContentMatches("http://www.google.com", "GOOGLE")                                                                  | *false*      |                                                                              |
        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+
        | urlContentMatches("http://www.google.com", "GOOGLE", "i")                                                             | *true*       |                                                                              |
        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+
        | urlContentMatches("http://www.google.com", "/GOOGLE/i")                                                               | *true*       |                                                                              |
        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+
        | urlContentMatches("http://www.google.com", "/GOOGLE/i", basic_user="foo", basic_password="12345abc")                  | *true*       |  (with HTTP Basic Authentication)                                            |
        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+
        | urlContentMatches("http://www.google.com", "/GOOGLE/i", include_data=True)                                            | *true*       |  (with data in the verbstatus)                                               |
        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+
        | urlContentMatches("http://www.google.com", "/GOOGLE/i", http_headers="User-Agent: dummy")                             | *true*       |  (with "dummy" as user agent                                                 |
        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+
        | urlContentMatches("http://www.google.com", "/GOOGLE/i", postdata="ciccio")                                            | *ERROR*      |  (postdata is not a  x-www-form-urlencoded                                   |
        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+
        | urlContentMatches("http://www.google.com/login", "/Exit/i", postdata=dict("username", "user", "password", "12345"))   | *true*       |                                                                              |
        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+
        | urlContentMatches("http://www.google.com", "GOOGLE", ciphers="DEFAULT:!DH")                                           | *true*       |  (Disabled Diffie-Helfman cipher)                                            |
        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+
        | urlContentMatches("https://www.google.com/login", "/Exit/i", ciphers="LOW")                                           | *ERROR*      |  (ssl error)                                                                 |
        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+

        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+
        | Expression                                                                                                            | Return Value | Description                                                                  |
        +=======================================================================================================================+==============+==============================================================================+
        | urlContentMatches("https://www.google.com/login", "/google/i", ignore_errors="+TIMEOUT")                              | True         | without any error                                                            |
        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+
        | urlContentMatches("https://www.google.com/login", "/google/i", ignore_errors="+TIMEOUT")                              | False        | with a timeout                                                               |
        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+
        | urlContentMatches("https://www.google.com/login", "/google/i", ignore_errors="-TIMEOUT")                              | *exception*  | with a timeout                                                               |
        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+
        | urlContentMatches("https://www.google.com/login", "/google/i", ignore_errors="+TIMEOUT")                              | False        | with a SSL error (because "+" is the defautl behaviour)                      |
        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+
        | urlContentMatches("https://www.google.com/login", "/google/i", ignore_errors="+TIMEOUT,+SSL")                         | False        | with a SSL error (because "+" is the defautl behaviour)                      |
        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+
        | urlContentMatches("https://www.google.com/login", "/google/i", ignore_errors="+TIMEOUT,-SSL")                         | *exception*  | with a SSL error                                                             |
        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+
        | urlContentMatches("https://www.google.com/login", "/google/i", ignore_errors="NONE,+TIMEOUT")                         | True         | without any error                                                            |
        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+
        | urlContentMatches("https://www.google.com/login", "/google/i", ignore_errors="NONE,+TIMEOUT")                         | False        | with a timeout                                                               |
        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+
        | urlContentMatches("https://www.google.com/login", "/google/i", ignore_errors="NONE,+TIMEOUT")                         | *exception*  | with any error but timeout                                                   |
        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+
        
    **Troubleshoot**
        
        If you get the following execution error:
        
            ::
                
                error: <urlopen error [Errno 8] _ssl.c:504: EOF occurred in violation of protocol
        
        You need to specify TLS protocol:
        
            ::
                
                urlContentMatches("http://www.google.com", "google", use_tls1=1)
        


httpRequest
---------------------------------------------

    **Syntax**
        
        **httpRequest** **(** *url* [**,** *optional params* ] **)**
        
    **Description**
        
        Sends an HTTP request (GET or POST) and returns the HTTP response body.
        
        By default, the function performs GET request. If optional parameter *postdata* is specified, the function will perform a POST request.
        
        See the function :ref:`sanet-poller.lib.functions.urlContentMatches` for more details aboud *options params*.
        
    **Output**
        
        Returns a *string* containing the response body.d
        
    **Errors**
        
        An exception is raised when error is detected. If an error occurrs and it is ignored, the function will return an emtpy string ('').
        
    **Examples**
        
        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+
        | Expression                                                                                                            | Return Value | Description                                                                  |
        +=======================================================================================================================+==============+==============================================================================+
        | httpRequest("http://www.google.com")                                                                                  | *true*       | Performs a GET.                                                              |
        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+
        | httpRequest("http://www.google.com", postdata="ciccio")                                                               | *ERROR*      | Performs a POST (postdata is not a  x-www-form-urlencoded).                  |
        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+
        | httpRequest("http://www.google.com/login", postdata=dict("username", "user", "password", "12345"))                    | *true*       | Performs a POST (postdata is converted to a x-www-form-urlencoded string)    |
        +-----------------------------------------------------------------------------------------------------------------------+--------------+------------------------------------------------------------------------------+

"""

import sys
import threading
import types
import re
import array

# socket & HTTP protocol
import socket, select
import urllib2
from urllib2 import Request
from cookielib import CookieJar
import base64

from django.utils import encoding

import urllib
import ssl

import ipaddr

from sanet_poller.lib.exceptions import PollException

# NTLM Support (optional)
__ntlm_supported = False
try:

	from sanet_poller.lib.ntlmutils import LabsAbstractNtlmAuthHandler

	from sanet_poller.lib.ntlmutils import LabsHTTPNtlmAuthHandler

	__ntlm_supported = True

	#print "NTML supported"

	class SquidProxyNtlmAuthHandler(LabsAbstractNtlmAuthHandler, urllib2.BaseHandler):

		auth_header = 'Proxy-authorization'

		def __init__(self, *args, **argv):
			LabsAbstractNtlmAuthHandler.__init__(self, *args, **argv)

		def http_error_407(self, req, fp, code, msg, headers):

			result = self.http_error_authentication_required('proxy-authenticate', req, fp, headers)
			if result.code == 407:
				raise Exception("Authentication error %s: %s" % (result.code, result.msg))

			return result

except Exception, e:
	#print 'Error while loading NTLM support'
	pass





# BUG With SSLv3
# http://bugs.python.org/issue11220

python_version = sys.version_info
if python_version[0] == 2 and ((python_version[1] == 6 and python_version[1] >= 6) or (python_version[1] > 6)):
	_need_ssl3_fix = 1

	# custom HTTPS opener, banner's oracle 10g server supports SSLv3 only
	import httplib, ssl, urllib2, socket

	# Workaround for Debian8 SSLv3 missing support (Ticket LABS #1494)
	_sslv3_supported = True
	try:
		x = ssl.PROTOCOL_SSLv3
	except AttributeError, e:
		_sslv3_supported = False

	class HTTPSConnectionV3(httplib.HTTPSConnection):
		def __init__(self, *args, **kwargs):
			
			httplib.HTTPSConnection.__init__(self, *args, **kwargs)
			
			#
			# we can not pass "ciphers" in "kwargs" options. Leave None here and 
			# call self.set_ciphers().
			#
			self._ciphers = None
			
		def set_ciphers(self, ciphers):
			self._ciphers = ciphers

		def connect(self):
			
			sock = socket.create_connection((self.host, self.port), self.timeout)
			if self._tunnel_host:
				self.sock = sock
				self._tunnel()
				
			options = {}
			if self._ciphers:
				options['ciphers'] = self._ciphers
			
			try:
				self.sock = ssl.wrap_socket(sock, self.key_file, self.cert_file, ssl_version=ssl.PROTOCOL_SSLv3, **options)
			except ssl.SSLError, e:
				print("Trying SSLv3.")
				self.sock = ssl.wrap_socket(sock, self.key_file, self.cert_file, ssl_version=ssl.PROTOCOL_SSLv23, **options)

			except AttributeError, e:
				# Workaround for Debian8 SSLv3 missing support (Ticket LABS #1494)
				if _sslv3_supported == False:
					print ("SSLv3 not supported in this python version. Forcing v23")
					self.sock = ssl.wrap_socket(sock, self.key_file, self.cert_file, ssl_version=ssl.PROTOCOL_SSLv23, **options)
				else:
					# Unknown AttributeError
					raise e

	class HTTPSConnectionTLSv1(httplib.HTTPSConnection):
		def __init__(self, *args, **kwargs):
			httplib.HTTPSConnection.__init__(self, *args, **kwargs)

			#
			# FIX: handler ciphers here too?
			#

		def connect(self):
			sock = socket.create_connection((self.host, self.port), self.timeout)
			if self._tunnel_host:
				self.sock = sock
				self._tunnel()

			#print ("Trying TLSv1")
			
			#
			# FIX: handler ciphers here too?
			#
			self.sock = ssl.wrap_socket(sock, self.key_file, self.cert_file, ssl_version=ssl.PROTOCOL_TLSv1)

	class HTTPSHandlerV3(urllib2.HTTPSHandler):
		
		def __init__(self, *args, **kwargs):
			#
			# IMPORTANT: urllib2.HTTPSHandler is a "old" python object. No super().
			#
			urllib2.HTTPSHandler.__init__(self, *args, **kwargs)

			#
			# we can not pass "ciphers" in "kwargs" options. Leave None here and 
			# call self.set_ciphers().
			#
			self._ciphers = None

		def set_ciphers(self, ciphers):
			self._ciphers = ciphers

		def https_open(self, req):
			
			#
			# The only way to pass more options is to wrap the costructor class
			# and set all the options here.
			#
			#    CALLER -->  HTTPSHandlerV3().do_open() -> _connection_class_wrapper -> HTTPSConnectionV3() -> ...
			# 
			def _connection_class_wrapper(*args, **kwargs):
				h = HTTPSConnectionV3(*args, **kwargs)
				h.set_ciphers(self._ciphers)
				return h
				
			return self.do_open(_connection_class_wrapper, req)
			
			#return self.do_open(HTTPSConnectionV3, req)

	class HTTPSHandlerTLSv1(urllib2.HTTPSHandler):
		def https_open(self, req):
			return self.do_open(HTTPSConnectionTLSv1, req)

else:
	_need_ssl3_fix = 0


#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class _url_opener(threading.Thread):
	def __init__(self, client, url, log):
		super(_url_opener, self).__init__()
		self.log       = log
		self.client    = client
		self.url       = url
		self.exception = None
		self.result    = None
		self.setName(self.url)
	def run ( self ):
		try:
			self.log.debug("_url_opener(%s) run", self.getName())
			self.result = self.client.open(self.url)
			self.log.debug("_url_opener(%s) got : %r", self.getName(), self.result)
		except Exception, es:
			self.exception = es
			self.log.debug("_url_opener(%s): got exception = %s\n%s", self.url, str(es), traceback.format_exc())

def parse_ignore_errors(default, ignore_errors_val):

	ignored_errors = {
		'TIMEOUT' : default,
		'HTTP'    : default,
		'OTHERS'  : default,
	}

	for val in ignore_errors_val.split(','):
		val = str(val)
		if val == "ALL":
			for k in ignored_errors:
				ignored_errors[k] = True
		elif val == "NONE":
			for k in ignored_errors:
				ignored_errors[k] = False
		else:
			flag = True

			if val[0] in ['+','-']:
				if val[0] == '-': 
					flag = False
				val2 = val[1:]
			else:
				val2 = val

			if val2 not in ignored_errors.keys() and val not in ['ALL','NONE']:
				raise Exception("Invalid parameter value. Invalid error code in %s: %s" % (repr(ignore_errors_val), val) )

			ignored_errors[val2] = flag

	return ignored_errors

def _http_request(context, *args, **kwargs):
	if kwargs.get('use_curl', False):
		return _http_request_curl(context, *args, **kwargs)
	else:
		return _http_request_urllib(context, *args, **kwargs)


def _http_request_urllib(context, url, 
                        proxy_server=None, 
                        proxy_port=None, 
                        proxy_realm=None, 
                        proxy_user=None, 
                        proxy_password=None, 
                        proxy_ntlm=None, 
                        basic_user=None, 
                        basic_password=None, 
                        timeout=None, 
                        include_data=False,
                        proxy_protocol=None,
                        http_headers=None,
                        use_tls1=False,
                        postdata=None,
                        basic_ntlm=None,
                        basic_challenge=True,
                        ciphers=None,
                        no_fix_ssl3=False,
                        ignore_errors=None,
                        **options
):

	log = context.getLog()

	# Bound parameters (inside context) have less priority then positional arguments
	proxy_server    = proxy_server   if proxy_server   != None else context.get('$proxy_server'  , None)
	proxy_port      = proxy_port     if proxy_port     != None else context.get('$proxy_port'    , None)
	proxy_realm     = proxy_realm    if proxy_realm    != None else context.get('$proxy_realm'   , None)
	proxy_user      = proxy_user     if proxy_user     != None else context.get('$proxy_user'    , None)
	proxy_password  = proxy_password if proxy_password != None else context.get('$proxy_password', None)
	proxy_ntlm      = proxy_ntlm     if proxy_ntlm     != None else context.get('$proxy_ntlm'    , False)
	proxy_protocol  = proxy_protocol if proxy_protocol != None else context.get('$proxy_protocol', 'http')

	basic_user      = basic_user     if basic_user     != None else context.get('$basic_user'    , None)
	basic_password  = basic_password if basic_password != None else context.get('$basic_password', None)
	basic_ntlm      = basic_ntlm     if basic_ntlm     != None else context.get('$basic_ntlm'    , False)

	http_timeout    = timeout        if timeout        != None else 20

	#
	# Handle 
	#
	if ignore_errors == None:
		ignore_errors = "ALL"

	# FIX: different default for urlContentMatches() and urlContentDoesntMatch()
	#if invert:
	#	def_ignore = True
	#else:
	#	def_ignore = False
	def_ignore = True
	ignored_errors = parse_ignore_errors( def_ignore, ignore_errors)

	log.debug("IGNORED ERRORS: %s", ignored_errors)

	#
	# PROXY?
	#
	if proxy_server:

		if not proxy_port     : raise PollException('Parameter $proxy_port missing or invalid.')
		#if proxy_realm == None: raise PollException('Parameter $proxy_realm missing.') # this parameter can be ''
		if proxy_realm:
			if not proxy_user     : raise PollException('Parameter $proxy_user missing or invalid.')
			if not proxy_password : raise PollException('Parameter $proxy_password missing or invalid.')

		proxy_ntlm = bool(proxy_ntlm)
		proxy_port = int(proxy_port)

		log.debug("Using proxy")

		if proxy_ntlm:
			if not __ntlm_supported:
				raise PollException("NTLM python modules not loaded correctly!")
			else:
				log.debug("NTLM Proxy authentication enabled")
	else:
		log.debug("Not using proxy")

	if basic_user:
		if not basic_password : raise PollException('Parameter $basic_password missing or invalid.')
	else:
		log.debug("NO BASIC AUTH")

	data = None
	
	try:
		log.debug("_http_request( url='%s')", url)

		#
		# Create HTTP CLIENT
		#

		client = None

		handlers = []

		#
		# Use PROXY?
		#
		if proxy_server:

			log.debug("Using proxy('%s', '%s', '%s', '%s', '%s')", proxy_server, proxy_port, proxy_realm, proxy_user, proxy_password)

			if proxy_ntlm:
				url_to_proxy = '%s://%s:%s' % (
					proxy_protocol,
					proxy_server,
					proxy_port
				)
			else:
				if proxy_realm:
					url_to_proxy = '%s://%s:%s@%s:%s' % (
						proxy_protocol,
						proxy_user,
						proxy_password,
						proxy_server,
						proxy_port
					)
				else:
					url_to_proxy = '%s://%s:%s' % (
						proxy_protocol,
						proxy_server,
						proxy_port
					)

			log.debug("Url to proxy: %s", url_to_proxy)

			proxies = {}
			proxies[proxy_protocol] = url_to_proxy
			proxy_handler = urllib2.ProxyHandler(proxies)

			if not proxy_ntlm:
				if proxy_realm:
					proxy_auth_handler = urllib2.ProxyBasicAuthHandler()
					proxy_auth_handler.add_password(
						proxy_realm,
						proxy_server,
						proxy_user,
						proxy_password
					)
			else:
				# Use realm as domain
				#       <domain>\<user>
				proxy_user2 = r'%s\%s' % (proxy_realm, proxy_user)

				log.debug("NTLM User = %s", proxy_user2)

				proxy_ntlm_pass  = urllib2.HTTPPasswordMgrWithDefaultRealm()
				proxy_ntlm_pass.add_password(None, url, proxy_user2, proxy_password)

				proxy_auth_handler = SquidProxyNtlmAuthHandler(proxy_ntlm_pass)

			handlers.append(proxy_handler)

			if proxy_realm:
				handlers.append(proxy_auth_handler)
		else:
			# Ignore environment http_proxy settings
			proxy_handler = urllib2.ProxyHandler({})
			handlers.append(proxy_handler)


		if _need_ssl3_fix and not no_fix_ssl3:
			if url.find("https") == 0:
				if use_tls1:
					log.info("Using TLSv1 FIX")
					handlers.append( HTTPSHandlerTLSv1() )
					#
					# FIX: handler ciphers here too?
					#
				else:
					log.info("Using SSLv3 FIX")

					h = HTTPSHandlerV3()
					h.set_ciphers(ciphers)

					handlers.append(h)

		#
		# Use HTTP BASIC AUTH ?
		#
		if basic_user:
			log.debug("Using basic authentication: user=%s, password=%s", basic_user, basic_password)
			passman = urllib2.HTTPPasswordMgrWithDefaultRealm()
			passman.add_password(None, url, basic_user, basic_password)

			if not basic_ntlm:
				authhandler = urllib2.HTTPBasicAuthHandler(passman)
			else:
				authhandler = LabsHTTPNtlmAuthHandler( passman )

			handlers.append(authhandler)

		# Use cookies in order to handle requests/redirects that need a session
		handlers.append( urllib2.HTTPCookieProcessor( CookieJar()  ) )

		#
		# Performing HTTP REQUEST
		#
		client = urllib2.build_opener(*handlers)

		if postdata is not None:
			log.debug("Post data: %s", postdata)
			if isinstance(postdata, types.DictType):
				postdata = urllib.urlencode(postdata)

		log.debug("Post data: %s", postdata)

		if postdata is not None:
			url_request = Request(url, postdata)
		else:
			url_request = Request(url)

		#
		# HACK: (ticket #2550) Alcuni servizi web non gestiscono l'auth challenge come previsto e quindi bisogna
		#                      forzare ANCHE negli header della richiesta iniziale l'header "Authorization".
		#                      REF: https://ecommerce.shopify.com/c/shopify-apps/t/python-authentication-27662
		#
		if basic_user:
			if not basic_challenge:
				base64string = base64.encodestring('%s:%s' % (basic_user, basic_password)).replace('\n', '')
				url_request.add_header("Authorization", "Basic %s" % base64string)   

		if http_headers is not None:
			headers = http_headers.split('\n')
			for h in headers:
				header_key, header_value = h.split(':', 1)
				header_value = header_value.strip()
				log.debug("HTTP Header: %s = %s", header_key, header_value)
				url_request.add_header(header_key, header_value)

		# FIX: this is necessary because urlib2 does not handle timeouts
		#      with python 2.5
		python_version = sys.version_info
		if python_version[0] == 2 and python_version[1] >= 6:
			log.debug("open with timeout")
			f = client.open(url_request, timeout=http_timeout)
		else:
			log.debug("open with thread timeout")
			t = _url_opener(client, url_request, log)
			t.start()
			log.debug("Waiting for response")
			t.join(http_timeout)
			if t.isAlive():
				raise socket.timeout("Timeout expired (%ss)" % (timeout))

			if t.exception:
				#
				# FIX: should rise urllib2.URLError(reason="time out")
				#
				raise t.exception

			f = t.result

		log.debug("HTTP Response: %s %s", f.code, f.msg)

		contenttype = f.info().getheaders("Content-Type")
		log.debug("HTTP Response Content-type: %s", contenttype )
		if contenttype:
			contenttype = contenttype[0]
		else:
			contenttype = 'text'   # fallback when response has not content-type

		log.debug("got content: %s", contenttype)

		data = f.read()

		data = encoding.smart_str(data, encoding='ascii', errors='replace')

		nbytes = len(data)

		log.info("got %d bytes", nbytes)

		# Search in 'text/*' content only!
		if contenttype[0:4] == 'text' or \
		   contenttype      == 'application/x-ns-proxy-autoconfig':  # wpad data
			data = ''.join(data)
		else:
			context.appendReason("Content-Type != text/*")

		f.close()

		log.info("Got %d bytes", len(data))

		log.log(9, "Response Data: %s", data)

	except urllib2.URLError, er:

		log.error("URLError: %s:%s", type(er), str(er)) 

		#
		# FIX: better way to detect timeouts?
		#
		if str(er).find('timed out') > -1: 
			msg = "urlContentMatches(%s) network timeout error: %s" % (url,  str(er))

			if ignored_errors['TIMEOUT']:
				context.appendReason(msg)
				log.info(msg)
			else:
				raise PollException(msg) 
		else:
			#msg = "urlContentMatches(%s,%s) generic URL error: %s" % (url, regexpr_str, str(er))
			msg = "urlContentMatches(%s) generic URL error: %s" % (url,  str(er))

			if ignored_errors['HTTP']:

				# Try to read response body
				try:
        				data = er.read()
					log.info("ERROR body: %s", data)
				except Exception, e:
					log.error("Unable to read HTTP error body")

				context.appendReason(msg)
				log.info(msg)
			else:
				raise PollException(msg) 

	except urllib2.HTTPError, er:
		msg = "urlContentMatches(%s) HTTP error: %s" % (url,  str(er))

		if ignored_errors['HTTP']:
	
			# Try to read response body
			try:
       				data = er.read()
			except Exception, e:
				log.error("Unable to read HTTP error body")

			context.appendReason(msg)
			log.info(msg)
		else:
			raise PollException(msg) 

	except Exception, er:
		msg = "urlContentMatches(%s) uknown error: %s:%s" % (url,  type(er), str(er))

		if ignored_errors['OTHERS']:
			context.appendReason(msg)
			log.info(msg)
		else:
			raise PollException(msg) 

	return data


import pycurl
import StringIO


def _http_request_curl(context, url, 
                        proxy_server=None, 
                        proxy_port=None, 
                        proxy_realm=None, 
                        proxy_user=None, 
                        proxy_password=None, 
                        proxy_ntlm=None, 
                        basic_user=None, 
                        basic_password=None, 
                        timeout=None, 
                        include_data=False,
                        proxy_protocol=None,
                        http_headers=None,
                        use_tls1=False,
                        postdata=None,
                        basic_ntlm=None,
                        basic_challenge=True,
                        ciphers=None,
                        no_fix_ssl3=False,
                        ignore_errors=None,
                        **options
):
	log = context.getLog()

	# Bound parameters (inside context) have less priority then positional arguments
	proxy_server    = proxy_server   if proxy_server   != None else context.get('$proxy_server'  , None)
	proxy_port      = proxy_port     if proxy_port     != None else context.get('$proxy_port'    , None)
	proxy_realm     = proxy_realm    if proxy_realm    != None else context.get('$proxy_realm'   , None)
	proxy_user      = proxy_user     if proxy_user     != None else context.get('$proxy_user'    , None)
	proxy_password  = proxy_password if proxy_password != None else context.get('$proxy_password', None)
	proxy_ntlm      = proxy_ntlm     if proxy_ntlm     != None else context.get('$proxy_ntlm'    , False)
	proxy_protocol  = proxy_protocol if proxy_protocol != None else context.get('$proxy_protocol', 'http')

	basic_user      = basic_user     if basic_user     != None else context.get('$basic_user'    , None)
	basic_password  = basic_password if basic_password != None else context.get('$basic_password', None)
	basic_ntlm      = basic_ntlm     if basic_ntlm     != None else context.get('$basic_ntlm'    , False)

	http_timeout    = timeout        if timeout        != None else 20


	full_http_headers = []



	if ignore_errors == None:
		ignore_errors = "ALL"

	def_ignore = True
	ignored_errors = parse_ignore_errors( def_ignore, ignore_errors)

	data = None

	try:

		response = StringIO.StringIO()

		c = pycurl.Curl()
		
		c.setopt(c.VERBOSE, 1)
		c.setopt(pycurl.NOSIGNAL, 1) # IMPORTANT: required for thread-safety!!!!

		c.setopt(pycurl.SSL_VERIFYPEER, False)
		c.setopt(pycurl.SSL_VERIFYHOST, False)

		c.setopt(pycurl.FOLLOWLOCATION, True)

		c.setopt(pycurl.TIMEOUT       , http_timeout)
		c.setopt(pycurl.CONNECTTIMEOUT, http_timeout)
		
		
		c.setopt(c.URL, url)
		c.setopt(c.WRITEFUNCTION, response.write)


		if proxy_server:
			c.setopt(pycurl.PROXYTYPE, pycurl.PROXYTYPE_HTTP)

			proxy_url = "%s://%s" % (proxy_protocol, proxy_server)

			c.setopt(pycurl.PROXY    , proxy_url)
			if proxy_port:
				c.setopt(pycurl.PROXYPORT, proxy_port  )

			if proxy_ntlm:
				print "NTML"
				c.setopt(pycurl.PROXYAUTH, pycurl.HTTPAUTH_NTLM) 

			c.setopt(pycurl.PROXYAUTH, pycurl.HTTPAUTH_NTLM) 

			if proxy_user and proxy_port:
				c.setopt(pycurl.PROXYUSERPWD, "%s:%s" % (proxy_user, proxy_password))


		print "*" * 80, basic_user, basic_password
		if basic_user and basic_password:
			if basic_ntlm:
				print "*" * 80, 1
				c.setopt(pycurl.HTTPAUTH, pycurl.HTTPAUTH_NTLM)
				c.setopt(pycurl.USERNAME, basic_user)
				c.setopt(pycurl.USERPWD , basic_password)
			else:
				print "*" * 80, 2
				c.setopt(pycurl.HTTPAUTH, pycurl.HTTPAUTH_BASIC)
				c.setopt(pycurl.USERPWD, basic_user + ':' + basic_password)
				#c.setopt(pycurl.USERNAME, basic_user)
				#c.setopt(pycurl.USERPWD , basic_password)


		if basic_user:
			if not basic_challenge:
				print "*" * 80, 3
				base64string = base64.encodestring('%s:%s' % (basic_user, basic_password)).replace('\n', '')
				full_http_headers.append( ("Authorization", "Basic %s" % base64string)   )
	 
	 
		#
		# Headers
		#
		if full_http_headers:

			full_http_headers += http_headers

			curl.setopt(pycurl.HTTPHEADER, http_headers)
			
		if postdata:
			c.setopt(pycurl.POSTFIELDS, data)

		if postdata is not None:
			c.setopt(pycurl.POST, 1)
			
			log.debug("Post data: %s", postdata)
			if isinstance(postdata, types.DictType):
				postdata = urllib.urlencode(postdata)

			c.setopt(pycurl.POSTFIELDS, postdata)

		c.perform()
		c.close()
		data = response.getvalue()
		response.close()


	except pycurl.error, er:

		log.error("URLError: %s:%s", type(er), str(er)) 

		#
		# FIX: better way to detect timeouts?
		#
		if str(er).find('timed out') > -1: 
			msg = "urlContentMatches(%s) network timeout error: %s" % (url,  str(er))

			if ignored_errors['TIMEOUT']:
				context.appendReason(msg)
				log.info(msg)
			else:
				raise PollException(msg) 
		else:
			#msg = "urlContentMatches(%s,%s) generic URL error: %s" % (url, regexpr_str, str(er))
			msg = "urlContentMatches(%s) generic URL error: %s" % (url,  str(er))

			if ignored_errors['HTTP']:

				# Try to read response body
				try:
        				data = er.read()
					log.info("ERROR body: %s", data)
				except Exception, e:
					log.error("Unable to read HTTP error body")

				context.appendReason(msg)
				log.info(msg)
			else:
				raise PollException(msg) 

	except Exception, er:
		msg = "urlContentMatches(%s) uknown error: %s:%s" % (url,  type(er), str(er))

		if ignored_errors['OTHERS']:
			context.appendReason(msg)
			log.info(msg)
		else:
			raise PollException(msg) 




	return data



#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_urlContentMatches(context, url, regexpr_str, case_insensitive=' ',
                          proxy_server=None,
                          proxy_port=None,
                          proxy_realm=None,
                          proxy_user=None,
                          proxy_password=None,
                          proxy_ntlm=None,
                          basic_user=None,
                          basic_password=None,
                          timeout=None,
                          include_data=False,
                          proxy_protocol=None,
                          http_headers=None,
                          use_tls1=False,
                          postdata=None,
                          basic_ntlm=None,
                          basic_challenge=True,
                          ciphers=None,
                          no_fix_ssl3=False,
                          ignore_errors=None,
                          **options
):
	"""
	Verifies if content of the given *url*  matches the regular expression *regexpr_str*.
	
	The parameter *regexpr_str* can be enclosed between '/' characters (eg. '/foo/'). In this
	case the case insensitive test can be enabled by appending a 'i'
	at the end of the expression. (eg: '/foo/i').
	The 'i' character modifier has precedence on the *case_insensitive* parameter.
	
	:param  url:  The content URL 
	:type   url:  :class:`pollerlib.ExecutionContext`
	
	:param  regexpr_str:  regualr expression
	:type   regexpr_str:  string
	
	:param  case_insensitive:  When 'i' enables case-insensitive test
	:type   case_insensitive:  char
	
	:returns: True the content matches, False otherwise.
	:rtype: boolean
	"""

	if regexpr_str[0] == '/':
		regexpr_str = regexpr_str[1:]
	if regexpr_str[-1] == '/':
		regexpr_str = regexpr_str[:-1]
	if regexpr_str[-2:] == '/i':
		regexpr_str = regexpr_str[:-2]
		case_insensitive='i'

	return __urlContentMatches(context, url, regexpr_str, case_insensitive, False, 
	                           proxy_server    = proxy_server, 
	                           proxy_port      = proxy_port, 
	                           proxy_realm     = proxy_realm, 
	                           proxy_user      = proxy_user, 
	                           proxy_password  = proxy_password, 
	                           proxy_ntlm      = proxy_ntlm, 
	                           basic_user      = basic_user, 
	                           basic_password  = basic_password, 
	                           timeout         = timeout, 
	                           include_data    = include_data,
	                           proxy_protocol  = proxy_protocol,
	                           http_headers    = http_headers,
	                           use_tls1        = use_tls1,
	                           postdata        = postdata,
	                           basic_ntlm      = basic_ntlm,
	                           basic_challenge = basic_challenge,
	                           ciphers         = ciphers,
	                           no_fix_ssl3     = no_fix_ssl3,
	                           ignore_errors   = ignore_errors,
	                           **options
	                          )

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_urlContentDoesntMatch(context, url, regexpr_str, case_insensitive=' ', 
                              proxy_server=None,
                              proxy_port=None,
                              proxy_realm=None,
                              proxy_user=None,
                              proxy_password=None,
                              proxy_ntlm=None,
                              basic_user=None,
                              basic_password=None,
                              timeout=None,
                              include_data=False,
                              proxy_protocol=None,
                              http_headers=None,
                              use_tls1=False,
                              postdata=None,
                              basic_ntlm=None,
                              basic_challenge=True,
                              ciphers=None,
                              no_fix_ssl3=False,
                              ignore_errors=None,
                              **options
):
	"""
	Verifies if content of the given *url*  DOES NOT match the regular expression *regexpr_str*.
	
	See :func:`fun_urlContentMatches`.
	"""
	
	if regexpr_str[0] == '/':
		regexpr_str = regexpr_str[1:]
	if regexpr_str[-1] == '/':
		regexpr_str = regexpr_str[:-1]
	if regexpr_str[-2:] == '/i':
		regexpr_str = regexpr_str[:-2]
		case_insensitive='i'

	return __urlContentMatches(context, url, regexpr_str, case_insensitive, True, 
	                           proxy_server    = proxy_server, 
	                           proxy_port      = proxy_port, 
	                           proxy_realm     = proxy_realm, 
	                           proxy_user      = proxy_user, 
	                           proxy_password  = proxy_password, 
	                           proxy_ntlm      = proxy_ntlm, 
	                           basic_user      = basic_user, 
	                           basic_password  = basic_password, 
	                           timeout         = timeout, 
	                           include_data    = include_data,
	                           proxy_protocol  = proxy_protocol,
	                           http_headers    = http_headers,
	                           use_tls1        = use_tls1,
	                           postdata        = postdata,
	                           basic_ntlm      = basic_ntlm,
	                           basic_challenge = basic_challenge,
	                           ciphers         = ciphers,
	                           no_fix_ssl3     = no_fix_ssl3,
	                           ignore_errors   = ignore_errors,
	                           **options
	                          )

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def __urlContentMatches(context, url, regexpr_str, case_insensitive, invert,
                        proxy_server=None, 
                        proxy_port=None, 
                        proxy_realm=None, 
                        proxy_user=None, 
                        proxy_password=None, 
                        proxy_ntlm=None, 
                        basic_user=None, 
                        basic_password=None, 
                        timeout=None, 
                        include_data=False,
                        proxy_protocol=None,
                        http_headers=None,
                        use_tls1=False,
                        postdata=None,
                        basic_ntlm=None,
                        basic_challenge=True,
                        ciphers=None,
                        no_fix_ssl3=False,
                        ignore_errors=None,
                        **options
):
	log = context.getLog()

	found = False

	data = _http_request(context, url, 
	                           proxy_server    = proxy_server, 
	                           proxy_port      = proxy_port, 
	                           proxy_realm     = proxy_realm, 
	                           proxy_user      = proxy_user, 
	                           proxy_password  = proxy_password, 
	                           proxy_ntlm      = proxy_ntlm, 
	                           basic_user      = basic_user, 
	                           basic_password  = basic_password, 
	                           timeout         = timeout, 
	                           include_data    = include_data,
	                           proxy_protocol  = proxy_protocol,
	                           http_headers    = http_headers,
	                           use_tls1        = use_tls1,
	                           postdata        = postdata,
	                           basic_ntlm      = basic_ntlm,
	                           basic_challenge = basic_challenge,
	                           ciphers         = ciphers,
	                           no_fix_ssl3     = no_fix_ssl3,
	                           ignore_errors   = ignore_errors,
	                           **options
	                          )
	
	#
	# Verify response body
	#
	if data is not None:        

		log.debug("Checking match width data (len %s)", len(data))

		if include_data:
			context.appendReason(data)

		re_flags = re.DOTALL | re.MULTILINE

		if case_insensitive == 'i':
			r = re.compile(regexpr_str, re.IGNORECASE | re_flags)
			m = r.search(data)
		else:
			r = re.compile(regexpr_str, re_flags)
			m = r.search(data)

		if m != None:
			found = True

		#m = None
		#data = None

	if found:
		if not invert:
			log.debug("OK, not inverting, content matching");
			return True
		else:
			log.debug("OK, inverting, content not matching");
			return False
	else:
		if not invert:
			log.debug("NOK, not inverting, content not matching");
			return False
		else:
			log.debug("NOK, inverting, content matching");
			return True

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_httpRequest(context, url, 
                          proxy_server=None,
                          proxy_port=None,
                          proxy_realm=None,
                          proxy_user=None,
                          proxy_password=None,
                          proxy_ntlm=None,
                          basic_user=None,
                          basic_password=None,
                          timeout=None,
                          include_data=False,
                          proxy_protocol=None,
                          http_headers=None,
                          use_tls1=False,
                          postdata=None,
                          basic_ntlm=None,
                          basic_challenge=True,
                          ciphers=None,
                          no_fix_ssl3=False,
                          ignore_errors=None,
                          **options
):

	data = _http_request(context, url, 
	                           proxy_server    = proxy_server, 
	                           proxy_port      = proxy_port, 
	                           proxy_realm     = proxy_realm, 
	                           proxy_user      = proxy_user, 
	                           proxy_password  = proxy_password, 
	                           proxy_ntlm      = proxy_ntlm, 
	                           basic_user      = basic_user, 
	                           basic_password  = basic_password, 
	                           timeout         = timeout, 
	                           include_data    = include_data,
	                           proxy_protocol  = proxy_protocol,
	                           http_headers    = http_headers,
	                           use_tls1        = use_tls1,
	                           postdata        = postdata,
	                           basic_ntlm      = basic_ntlm,
	                           basic_challenge = basic_challenge,
	                           ciphers         = ciphers,
	                           no_fix_ssl3     = no_fix_ssl3,
	                           ignore_errors   = ignore_errors
	                           ,**options
	                          )

	# If data is none, the request failed and errors were ignored. Return empty string.
	if data is None:
		return ''

	return data

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

EXPR_SYMBOLS = {
	'urlContentDoesntMatch' : fun_urlContentDoesntMatch,
	'urlContentMatches'     : fun_urlContentMatches,
	
	'httpRequest'           : fun_httpRequest,
}
