# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it> 
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

.. _poller_fun_execWebScenario:

execWebScenario
----------------

	**Syntax**

		**execWebScenario** **(** *name* [, *ssl_version* ] **)**

	**Description**

		Retrieves from Sanet configuration the configuration of the given *webscenario* and tries to execute it.
		
		Returns a dictionary containing the following keys:
		
			* *status*    : 1 Ok, 2 Error, 0 Unknown
			* *totaltime* : Total execution time in seconds
			
	**SSL Version**
	
		These are the valid values for the *ssl_version* parameter:
		
			* "SSLv2"
			* "SSLv3"
			* "TLSv1"

	**Macro expansion**
	
		Macros string (es. {{*name**}} ) used inside steps' URL or POST VARS will be expanded also with $*name* parameters taken from the execution context (es. datagroup parameters, $node, ecc.)

	**Input**

		* *name*: the name of the web scenario
		* *ssl_version*: (optinal) The SSL version to use.

	**Output**

		Returns a *dictionary*

	**Errors**

		Raise an exception if an unknown error occurs

	**Examples**

		+------------------------------------------------------------------------------------+-------------------------------------+-------------------------------------------------------+
		| Expression                                                                         | Return value                        |  Description                                          |
		+====================================================================================+=====================================+=======================================================+
		| execWebScenario("testwebapp")["status"]                                            | 1                                   |                                                       |
		+------------------------------------------------------------------------------------+-------------------------------------+-------------------------------------------------------+
		| execWebScenario("testwebapp")["totaltime"]                                         | 0.141                               |                                                       |
		+------------------------------------------------------------------------------------+-------------------------------------+-------------------------------------------------------+
		| execWebScenario("testwebapp", ssl_version="SSLv3")["totaltime"]                    | 0.141                               | Force SSLv3                                           |
		+------------------------------------------------------------------------------------+-------------------------------------+-------------------------------------------------------+
     
"""

import traceback
import pprint

from sanet_poller.lib.exceptions import PollException

from sanet_poller.lib.webtests.api import execute_scenario, WEBSCENARIO_STATUS_UP, WEBSCENARIO_STATUS_DOWN, STEP_STATUS_OK

from sanet_poller.lib.functions.sanet_meta import RemoteQueryException, _query_server

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fun_execWebScenario(context, ws_name, ssl_version=None):

	log = context.getLog()
	configuration = context.getConfiguration()
	

	val_id = 'webscenario_data:%s' % (ws_name)
	
	result = context.getCachedSimpleValue(val_id)
	if result != None:
		return result
	
	

	log.debug("Launcing execWebTest(%s)", ws_name)

	#
	# Load web scenario from Sanet server
	#
	log.debug("Loading web scenario from server...")
	
	web_scenario_data = None
	
	try:
		log.debug("Loading web scenario from sanet server")

		response = _query_server(context, 'get_webscenario_data', {
			'tenant_id': configuration.get_tenant_id(),
			'ws_name'  : ws_name,
		})

		web_scenario_data = response['ws_data']

		log.debug("Received web scenario.")
		log.log(1, "web scenario data: %s", pprint.pformat(web_scenario_data))

	except RemoteQueryException, ex:
		raise PollException("Unable to get data for webscenario '%s': %s" % (ws_name, ex))
	
	#
	# Execute scenario
	#
	try:
		additional_macros = {}
		for k, v in context.copy().items():
			if k.startswith('$'):
				additional_macros[k[1:]]=v
		
		result = execute_scenario(log, web_scenario_data, additional_macros=additional_macros, ssl_version=ssl_version)

		log.debug("Web scenario status = %s", result['status'])

		# Cache value in the execution cache
		context.cacheSimpleValue(val_id, result)
		
		# Save data to send to the server
		context.remember_additional_update_data( 'webscenario', result )

		i = 1
		reason_data = []
		for step in result['steps']:
			if step['exec_code'] == STEP_STATUS_OK:
				reason_data.append("%s: status=OK     exec-time=%s http-code=%s  time=%s matched=%s info=%s" % (i, step['exec_time'], step['http_code'], step['totaltime'], step['matched'], step['info']))
			else:
				reason_data.append("%s: status=ERROR  exec-time=%s info=%s"                                  % (i, step['exec_time'], step['info']))
			i += 1
		reason_data.append("TOTAL TIME: %s" % (result['totaltime']))
		
		context.appendReason( '\n'.join(reason_data) )

		
		if result['status'] == WEBSCENARIO_STATUS_UP:
			r = {
				'status'   : result['status'],
				'totaltime': result['totaltime'],
			}
			
		elif result['status'] == WEBSCENARIO_STATUS_DOWN:
			
			
			r = {
				'status'   : result['status'],
				'totaltime': result['totaltime'],
			}
		else:
			PollException("Unknown error")

		return r
		
	except Exception, e:
		print traceback.format_exc()
		raise PollException("Error while executing scenario: %s" % (e))

	
#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

EXPR_SYMBOLS = {
	'execWebScenario': fun_execWebScenario,
}	
