# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it> 
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import time
import sys
import re
import subprocess
import os
import string
import random

from datetime import datetime

# socket & HTTP protocol
import socket, select

import ipaddr

# DNS 
from sanet_common.netutils import resolve_ip

from sanet_common.netutils import Ping

from sanet_poller.lib.exceptions import PollException 

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class ResolveException(PollException):
	pass

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class SanetIcmpPinger(object):
	"""
	"""
	CRE_PING6_OUTPUT = re.compile('.* bytes from .* time=([0-9.]+) ms.*', flags=re.DOTALL)


	def __init__(self):
		pass

	def measure_rtt(self, log, ipversion, host, num_probes, probesize, timeout, do_frag, bind_address=None):

		addrs = self.resolve_host(log, ipversion, host)
	
		# Calculate round trip times
		for addr in addrs:
			(rttavg, rttmin, rttmax, loss_avg, infos) = self.do_rtt_measuring(log, ipversion, addr, num_probes, probesize, timeout, do_frag, bind_address=bind_address)
			if loss_avg < 1: # pinged!
				return (rttavg, rttmin, rttmax, loss_avg, infos)
			
		return (None, None, None, 1, ['ICMP Timeout on all addresses'])

	def resolve_host(self, log, ipversion, host):
		
		log.debug("Resolving ipv=%s, host=%s", ipversion, host)
		
		addrs = None
		reason = None
		try:
			host = str(host)
			addrs = resolve_ip(host, ipversion, False, 10)
		except socket.timeout, e:
			reason = "reachability NOK: socket timeout on %s: %s" % (host,e)
		except socket.error, e:
			reason = "reachability NOK: socket error on %s: %s" % (host, e)
		except Exception, e:
			reason = "reachability NOK: error on %s: %s" % (host, e)

		if not addrs:
			reason = "No addresses for host %s version %d" % (host, ipversion)
		
		# Per qualunque ragione, se non posso risolvere il nome, sollevo una eccezione.
		if reason:
			raise ResolveException(reason)
		
		random.shuffle(addrs)
		
		return addrs

	def do_rtt_measuring(self, log, ipversion, target, num_probes, size, timeout, do_frag, bind_address=None): 

		i = 0
		dmin = 3600
		dmax = 0
		davg = 0
		num_lost = 0
		
		# For every probe we get  an  rtt and a verbose info about errors or timeouts
		
		if ipversion == 4:
			(recv_rrts, infos) = self.do_rtt_measuring_ipv4(log, target, num_probes, size, timeout, do_frag, bind_address=bind_address)
		elif ipversion == 6:
			(recv_rrts, infos) = self.do_rtt_measuring_ipv6(log, target, num_probes, size, timeout, do_frag, bind_address=bind_address)
		else:
			raise PollException("do_rtt_measuring: Invalid IP version %s" % ipversion)

		for i in xrange(len(recv_rrts)):
			rtt = recv_rrts[i]
			if rtt != None:
				log.debug("%s: got ping answer #%s in %s s", target, i, rtt )
				dmin = min(rtt, dmin)
				dmax = max(rtt, dmax)
				davg += rtt
			else:
				log.debug("%s: lost ping answer #%s: %s", target, i, infos[i] )
				num_lost += 1
			i += 1

		if num_lost == num_probes:
			davg = timeout;
			dmin = timeout;
			dmax = timeout;
		else:
			davg = float(davg) / float((num_probes - num_lost))
			
		# remove unnecessary infos
		infos = [i for i in infos if i ]

		return (davg, dmin, dmax, float(num_lost) / float(num_probes), infos)

	def do_rtt_measuring_ipv4(self, log, target, num_probes, size, timeout, do_frag, bind_address=None): 

		probe_rtts = []
		probe_infos = []
		
		log_callback=(lambda x: log.debug("ping %s: %s" % (target, x)) )
		log_callback = None
		

		log.debug("%s: Sending IPv4 pings.", target)
		
		try:
			p = Ping(target, size=size, timeout=timeout, count=num_probes, callback=log_callback, do_frag=do_frag, bind_address=bind_address)
			(probe_rtts, probe_infos) = p.ping()
			p.close()
		except socket.error, e:
			for i in xrange(num_probes):
				probe_rtts.append(None)
				probe_infos.append(str(e))
		
		return (probe_rtts, probe_infos)

	def do_rtt_measuring_ipv6(self, log, target, num_probes, size, timeout, do_frag, bind_address=None): 

		# do-frag is currently ignored
		probe_rtts  = []
		probe_infos = []

		ipversion = 6

		log.debug("%s: Sending IPv6 pings.", target)
		for i in xrange(num_probes):

			#t0 = datetime.now()
			#pinged = self.exec_ping(log, ipversion, target, size, timeout, bind_address=bind_address)
			#delta = datetime.now() - t0

			pinged, delta, info = self.exec_ping(log, ipversion, target, size, timeout, bind_address=bind_address)

			if pinged:
				#probe_rtts.append(delta.seconds + (delta.microseconds / 1000000.0))
				probe_rtts.append(delta)
			else:
				probe_rtts.append(None)
				
			probe_infos.append(info)

		#probe_infos = ['' for n in xrange(len(probe_rtts))]

		return (probe_rtts, probe_infos)

	def exec_ping(self, log, ipversion, target, size, timeout, bind_address=None):

		if ipversion == 4:  
			ping='ping'
		elif ipversion == 6: 
			ping='ping6'
		else:
			raise ValueException("wrong version of ping = %s" % ipversion)

		# Examples:
		#
		# root@localhost# ping -c 1 -s 1472 -A -W 3 -n 1.0.0.1
		# PING 1.0.0.1 (1.0.0.1) 1472(1500) bytes of data.
		# 1480 bytes from 1.0.0.1: icmp_seq=1 ttl=64 time=1.35 ms
		# ...
		#
		# root@localhost:~# ping6 -c 3 -s 1452 -A -l 4 -n -W 3 2001:760:2E00:AFFE::1
		# PING 2001:760:2E00:AFFE::1(2001:760:2e00:affe::1) 1423 data bytes
		# 1460 bytes from 2001:760:2e00:affe::1: icmp_seq=1 ttl=64 time=2.00 ms
		# ...
	 
		command = "%s -c 1 -s %s -A -W %s -n" % (ping, size, timeout)

		if bind_address != None:
			command += " -I '%s'" % (bind_address)

		command += " '%s'" % (target)
		
		
		log.debug("exec_ping: %s", command)

		process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
		os.waitpid(process.pid, 0)
		output = process.stdout.read().strip()
		stderr = process.stderr.read().strip()

		#if re.search(" bytes from ", output):
		#	return True
		#else:
		#	return False

		m = SanetIcmpPinger.CRE_PING6_OUTPUT.match(output)
		if m:
			ms = float(m.group(1)) / 1000.0
			return True , ms, ""
		else:
			return False, 0 , "ping6 output not parsed ('%s')." % (output+stderr)
