# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>. 

"""

.. _snmp-v3-options-string:

Opzioni SNMPv3 
===========================

La stringa di configurazione delle opzioni SNMPv3 deve rispettare la seguente sintassi:

::
	
	<option> : <value> [ , <option> : <value> ]*


.. important:: La stringa e' CASE SENSITIVE


La *option* puo' essere una delle seguenti:

	+-------------+-----------------------------+----------------------------+--------------+
	| Opzione     | Descrizione                 | Valori ammessi             | Default      |
	+=============+=============================+============================+==============+
	| context     | Context Name                | *stringa*                  | ''           |
	+-------------+-----------------------------+----------------------------+--------------+
	| username    | Securety Name               | *stringa*                  | 'initial'    |
	+-------------+-----------------------------+----------------------------+--------------+
	| sec_level   | Security Level              | noauth, auth, crypt        | noauth       |
	+-------------+-----------------------------+----------------------------+--------------+
	| auth_proto  | Authentication protocol     | md5, sha                   | md5          |
	+-------------+-----------------------------+----------------------------+--------------+
	| auth_pass   | Authentication passphrase   | *stringa*                  |              |
	+-------------+-----------------------------+----------------------------+--------------+
	| crypt_proto | Privacy protocol            | *stringa*                  |              |
	+-------------+-----------------------------+----------------------------+--------------+
	| crypt_pass  | Privacy passphrase          | *stringa*                  |              |
	+-------------+-----------------------------+----------------------------+--------------+

Le opzioni *auth_proto* e *auth_pass* sono richieste quando *sec_level* e' valorizzato a *auth* or *crypt*.

Le opzioni *crypt_proto* e *crypt_pass* sono richieste quando *sec_level*  e' valorizzato a  *crypt*.

Esempi:

	::
		
		context:mycontext
		
		sec_level:auth,username:foo,auth_proto:md5,auth_pass:mypassword
		
		sec_level:noauth,context:mycontext

"""

import types
import re

from sanet_libsnmp import *

from sanet_poller.lib.exceptions import *

#FIX: spostare e farla dinamica in base alla configurazione?
POLLER_DEFAULT_SNMP_TIMEOUT = 3

#-----------------------------------------------------------------------------#
# SNMPv3 option string constants                                              #
#-----------------------------------------------------------------------------#

snmpv3_valid_options = {
	#POLLER PARAM    SNMP WRAPPER PARAM
	 'username'    : 'SecName'
	,'sec_level'   : 'SecLevel'
	,'auth_proto'  : 'AuthProto'
	,'auth_pass'   : 'AuthPass'
	,'context'     : 'Context'

	# FIX: The use of the SNMPv3 Priv protocol is not stable (coredump)
	,'crypt_proto' : 'PrivProto'
	,'crypt_pass'  : 'PrivPass'
}	

snmpv3_sec_levels = {
	'noauth': 'noAuthNoPriv',
	'auth'  : 'authNoPriv',
	
	# FIX: The use of the SNMPv3 Priv protocol is not stable (coredump)
	'crypt' : 'authPriv'
}

snmpv3_protos = {
	'md5': 'MD5',
	'sha': 'SHA',
	
	# FIX: The use of the SNMPv3 Priv protocol is not stable (coredump)
	'des': 'DES', 
	'aes': 'AES',
}

#-----------------------------------------------------------------------------#
# Function for handling SNMPv3 parameters inside config string                #
#-----------------------------------------------------------------------------#

KEY_VAL_REGEXP_MATCH = r'((?:\\.|[^\:,]+)*)\:("(?:\\.|[^"\\]+)*"|(?:\\.|[^,"\\]+)*)'

KEY_VAL_REGEXPR = re.compile(KEY_VAL_REGEXP_MATCH)

def parse_snmpv3_auth_options(string_data):
	
	if not string_data:
		return {}
	
	parameters = KEY_VAL_REGEXPR.findall(string_data)
	if parameters:
		for opt_name, opt_value in parameters:

			if opt_name not in snmpv3_valid_options:
				raise Exception("Invalid SNMPv3 option: '%s'" % (opt_name))
				
			elif opt_name in ['auth_proto','crypt_proto']:
				if opt_value not in snmpv3_protos.keys():
					raise Exception("Invalid SNMPv3 protocol: '%s'" % (opt_value))

			elif opt_name == 'sec_level':
				if opt_value not in snmpv3_sec_levels.keys():
					raise Exception("Invalid SNMPv3 security level: '%s'" % (opt_value))
		
		return parameters
	else:
		raise ValueError("Invalid SNMPv3 options string: '%s'" % (string_data))
	
#-----------------------------------------------------------------------------#

def _get_snmpv3_auth_options_from_string(config_string):
	"""
	"""
	
	options = {}
	
	try:
		parameters = parse_snmpv3_auth_options(config_string)
		
		for (opt_name, opt_val) in parameters:
			
			opt_name = opt_name.strip().lower()
			opt_val  = opt_val.strip().encode('ascii') #.lower()
		
			real_opt_name = snmpv3_valid_options.get(opt_name, None)
			if real_opt_name == None:
				raise PollException("Invalid SNMPv3 option '%s=%s'!" % (opt_name, opt_val) )

			if   opt_name == 'sec_level'  : real_opt_val = snmpv3_sec_levels[opt_val]
			elif opt_name == 'auth_proto' : real_opt_val = snmpv3_protos[opt_val]
			elif opt_name == 'auth_pass'  : real_opt_val = opt_val.replace(r"\,",',')
			elif opt_name == 'context'    : real_opt_val = opt_val

			# FIX: The use of the SNMPv3 Priv protocol is not stable (coredump)
			elif opt_name == 'crypt_proto': real_opt_val = snmpv3_protos[opt_val]
			elif opt_name == 'crypt_pass' : real_opt_val = opt_val.replace(r"\,",',')

			else:
				#
				# FIX: should raise an exception here?
				#
				real_opt_val = opt_val
			
			options[ real_opt_name ] = real_opt_val
			
		return options
		
	except ValueError, e:
		raise PollException("Error while parsing SNMPv3 option string.")
		

#-----------------------------------------------------------------------------#
# SNMP LOW-LEVEL FUNCTIONS. Everybody MUST use these two.                     #
#-----------------------------------------------------------------------------#

def getvar( host, comm_or_auth, tries, oid, port=SNMP_DEFAULT_PORT, version=SNMP_DEFAULT_VERSION, timeout=POLLER_DEFAULT_SNMP_TIMEOUT):
	"""
	Low level function used to perform SNMP GET requests.
	
	:param  host: remove SNMP agent's host name.
	:type   host: string
	
	:param  comm_or_auth: SNMP community or SNMPv3 authentication string.
	:type   comm_or_auth: string
	
	:param  tries: Retries before failure.
	:type   tries: number
	
	:param  oid: dotted-decimal numeric OID (es. '1.2.3')
	:type   oid: string
	
	:param  port: remote SNMP agent's port number (Default: :const:`SNMP_DEFAULT_PORT`).
	:type   port: number
	
	:param  version: SNMP protocol version (Default: :const:`SNMP_DEFAULT_VERSION`).
	:type   version: (number)
	
	:returns: On success returns the SNMP value for the given OID. Returns *None* if the
	          given value can't be retrieved.
	
	:exception: 
	        :class:`ExternalErrorException` when a low level error is returned by the SNMP library.
	"""

	tries   = int(tries)
	port    = int(port)
	version = int(version)

	if version == 3:
		other_snmp_options = _get_snmpv3_auth_options_from_string(comm_or_auth)
		community = None
	else:
		other_snmp_options = {}
		community = comm_or_auth
	
	try:
		#print "DEBUG......getvar(%s:%s, %s:%s, tries=%s, version=%s, other_options=%s)"% (host,port, oid,comm, tries, version, other_snmp_options) 
		val = snmp_get(oid, host, port=port, community=community, timeout=timeout, retries=tries-1, Version=version, **other_snmp_options)
		
		# quick hack for having OIDS w/out leading dot
		if isinstance(val, types.StringTypes):
			if len(val) > 0:
				if val[0] == '.':
					val = val[1:]
		return val
	
	except SnmpWrapperBUGException, e:
		# Special case: the snmp library is bugged and we want to be sure these case will
		#               produce an uncheckable result.
		#
		# HISTORY: Added for ticket #1885.
		#
		raise ForceUncheckableException("Forced uncheckable value on SNMP GET: %s" % (str(e)))
	
	except SnmpTimeout, e:
		# If we get a timeout on a single request, then we will probably get 
		# a timeout on any other request. This is fatal problem.
		raise TimeoutException("Timeout on SNMP Get (%s) on %s:%s comm=%s v=%s" % (oid, host, port, community, version))
	
	except SnmpWrapperException, e:
		raise ExternalErrorException( "Low-level error getvar(oid=%s  host=%s:%s comm=%s v=%s) reason: %s" % (oid, host, port, community, version, e))

#-----------------------------------------------------------------------------#

def gettable(host, comm_or_auth, tries, oid, port=SNMP_DEFAULT_PORT, version=SNMP_DEFAULT_VERSION, timeout=POLLER_DEFAULT_SNMP_TIMEOUT):
	"""
	Low level function used to perform SNMP WALK requests. The returned tables
	is a list of tuples of two elements: the OID string and a value.
	
	:param  host: remove SNMP agent's host name.
	:type   host: string
	
	:param  comm_or_auth: SNMP community or SNMPv3 authentication string.
	:type   comm_or_auth: string
	
	:param  tries: Retries before failure.
	:type   tries: number
	
	:param  oid: dotted-decimal numeric OID (es. '1.2.3')
	:type   oid: string
	
	:param  port: remote SNMP agent's port number (Default: :const:`SNMP_DEFAULT_PORT`).
	:type   port: number
	
	:param  version: SNMP protocol version (Default: :const:`SNMP_DEFAULT_VERSION`).
	:type   version: (number)
	
	:returns: On success returns the SNMP table (list of tuples <oid,value>) for the given OID. Returns *None* if the
	          given value can't be retrieved.
	
	:exception: 
	        :class:`ExternalErrorException` when a low level error is returned by the SNMP library.
	"""
	
	tries   = int(tries)
	port    = int(port)
	version = int(version)
	
	if version == 3:
		other_snmp_options = _get_snmpv3_auth_options_from_string(comm_or_auth)
		community = None
	else:
		community = comm_or_auth
		other_snmp_options = {}
	
	try:
		#print "DEBUG......gettable(%s:%s, %s:%s,tries=%s,version=%s)" % (host,port, oid,comm, tries,version) 
		
		#if settings.POLLER_FIX_WALK_BUG:
		#	print "using custom walk"
		#	val = snmp_smartwalk(oid, host, int(port), comm, POLLER_DEFAULT_SNMP_TIMEOUT, tries-1, Version=version, **other_snmp_options)
		#else:
		val = snmp_walk(oid, host, port=int(port), community=community, timeout=timeout, retries=tries-1, Version=version, **other_snmp_options)
		
		return val
		
	except SnmpWrapperBUGException, e:
		# Special case: the snmp library is bugged and we want to be sure these case will
		#               produce an uncheckable result.
		#
		# HISTORY: Added for ticket #1885.
		#
		raise ForceUncheckableException("Forced uncheckable value on SNMP WALK: %s" % (str(e)))
		
	except SnmpTimeout, e:
		# If we get a timeout on a single request, then we will probably get 
		# a timeout on any other request. This is fatal problem.
		#raise TimeoutException("Timeout on SNMP Walk (%s)" % oid)
		raise TimeoutException("Timeout on SNMP Get (%s) on %s:%s comm=%s v=%s" % (oid, host, port, community, version))
		
	except SnmpWrapperException, e:
		raise ExternalErrorException( "Low-level error gettable(oid=%s host=%s:%s comm=%s v=%s) reason: %s" % (oid, host, port, community, version, e))

#-----------------------------------------------------------------------------#
# Sessions                                                                    #
#-----------------------------------------------------------------------------#

def init_snmp_session(host, 
                      comm_or_auth, 
                      tries, 
                      port=SNMP_DEFAULT_PORT, 
                      version=SNMP_DEFAULT_VERSION, 
                      timeout=POLLER_DEFAULT_SNMP_TIMEOUT
):
	tries   = int(tries)
	port    = int(port)
	version = int(version)

	if version == 3:
		other_snmp_options = _get_snmpv3_auth_options_from_string(comm_or_auth)
		community = None
	else:
		other_snmp_options = {}
		community = comm_or_auth

	sess = snmp_create_sess(
		host, 
		port      = port, 
		community = community, 
		timeout   = timeout,
		retries   = tries,
		Version   = version,
		**other_snmp_options
	)

	return sess


def getvar_s( sess, oid ):

	try:
		val = snmp_get_s(sess, oid)
		
		# quick hack for having OIDS w/out leading dot
		if isinstance(val, types.StringTypes):
			if len(val) > 0:
				if val[0] == '.':
					val = val[1:]
		return val

	except SnmpWrapperBUGException, e:
		# Special case: the snmp library is bugged and we want to be sure these case will
		#               produce an uncheckable result.
		#
		# HISTORY: Added for ticket #1885.
		#
		raise ForceUncheckableException("Forced uncheckable value on SNMP GET: %s" % (str(e)))

	except SnmpTimeout, e:
		# If we get a timeout on a single request, then we will probably get 
		# a timeout on any other request. This is fatal problem.
		raise TimeoutException("Timeout on SNMP Get (%s) on %s comm=%s v=%s" % (oid, sess.DestHost, sess.Community, sess.Version))

	except SnmpWrapperException, e:
		raise ExternalErrorException( "Low-level error getvar_s(oid=%s  host=%s comm=%s v=%s) reason: %s" % (oid, sess.DestHost, sess.Community, sess.Version, e))


def gettable_s(sess, oid):

	try:
		val = snmp_walk_s(sess, oid)

		return val

	except SnmpWrapperBUGException, e:
		# Special case: the snmp library is bugged and we want to be sure these case will
		#               produce an uncheckable result.
		#
		# HISTORY: Added for ticket #1885.
		#
		raise ForceUncheckableException("Forced uncheckable value on SNMP WALK: %s" % (str(e)))

	except SnmpTimeout, e:
		# If we get a timeout on a single request, then we will probably get 
		# a timeout on any other request. This is fatal problem.
		#raise TimeoutException("Timeout on SNMP Walk (%s)" % oid)
		raise TimeoutException("Timeout on SNMP Walk (%s) on %s comm=%s v=%s" % (oid, sess.DestHost, sess.Community, sess.Version))

	except SnmpWrapperException, e:
		raise ExternalErrorException( "Low-level error gettable_s(oid=%s  host=%s comm=%s v=%s) reason: %s" % (oid, sess.DestHost, sess.Community, sess.Version, e))
