# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import struct
import re

from sanet_poller.lib.exceptions import *

from sanet_poller.lib.snmp.base import getvar, gettable

from sanet_poller.lib.utils import macbin2blob

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#


def areStpAdj(log, host1, host2, port1, port2, comm1, comm2, sv1=1, sv2=1):
	reason = None
	
	log.debug( "areStpAdj(%s,%s,%s,%s,%s,%s,%s,%s)", host1,host2, port1,port2, comm1,comm2, sv1,sv2)
	
	#
	# Check if we need to rewrite the community parameter (for Cisco's devices)
	#
	comm1 = ciscoExtendCommunity(log, host1, comm1, port1, sv1)
	comm2 = ciscoExtendCommunity(log, host2, comm2, port2, sv2)
	
	#
	p1 = ifIndex2stp(log, host1, port1, comm1, sv1)
	#if p1 == None:
		#return (None,  "cannot find %s in bridge mib for %s" % (port1, host1) )
	port1 = p1;
	
	#
	p2 = ifIndex2stp(log, host2, port2, comm2, sv2)
	#if p2 == None:
		#return (None, "cannot find %s in bridge mib for %s" % (port2, host2) )
	port2 = p2
	
	#
	# Check if Host1 is the designated bridge for host2
	#
	(v1, r) = __is1DesigFor2(log, host1, host2, port1, port2, comm1,  comm2,  sv1,  sv2)
	if v1 == False or v1 == None:
		reason = "%s is not DB for %s because (%s)"        % (host1, host2, r)
	elif v1 == True:
		return (True, r)
	
	#
	# Check if Host2 is the designated bridge for host1
	#
	(v2, r) = __is1DesigFor2(log, host2, host1, port2, port1, comm2, comm1, sv2, sv1)
	if v2 == False or v2 == None:
		reason="%s, %s is not DB for %s because (%s)"        % (reason, host2, host1, r)
	elif v2 == True:
		reason="%s, %s is DB for %s because (%s)"            % (reason, host2, host1, r)
	
	#
	# End
	#
	#if (v1 == None) or (v2 == None):
	#	return (None, reason)
	
	return (v1 or v2, reason)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def __is1DesigFor2(log, host1, host2, port1, port2, comm1, comm2, sv1, sv2):
	"""
	   HOST1                 HOST2
	    ----+               +----
	        |               |
	   port1|== ----      ==|
	        |       \       |
	        |==      ---- ==|port2
	        |               |
	    ----+               +----
	"""

	# This bridge MAC address
	myIdOid          = '1.3.6.1.2.1.17.1.1.0'   
	      
	# The Bridge Identifier of the bridge which this port considers to be 
	# the Designated Bridge for this ports segment.
	#
	#    bridge priority (2 bytes)  +  port's MAC address (6 bytes)]
	#
	desigBaseOid     = '1.3.6.1.2.1.17.2.15.1.8'      
	
	# The Port Identifier of the port on the Design3473DDated Bridge for this 
	# ports segment." (2 bytes) (network order)
	#    priority (4 bit) + port (12 bit)
	desigPortBaseOid = '1.3.6.1.2.1.17.2.15.1.9'      
	
	# Priority of a port (2 bytes) (network order)
	# "The value of the priority field which is contained in the first (in 
	# network byte order) octet of the (2 octet long) Port ID. The other
	# octet of the Port ID is given by the value of dot1dStpPort.
	portPrioBaseOid  = '1.3.6.1.2.1.17.2.15.1.2'      

	log.debug ("Entering is1DesigFor2(%s, %s, %s, %s, %s, %s,%s,%s)", host1, host2, port1, port2, comm1, comm2,sv1,sv2 )
	
	# Get host1's MAC address (just for output info)
	var = getvar(host1, comm1, 3, myIdOid, 161, sv1)
	if var == None:
		#raise UndefinedSnmpValueException("cannot read %s from %s: %s" % (myIdOid, host1, lastsnmperror) )
		raise UndefinedSnmpValueException("cannot read %s from %s" % (myIdOid, host1) )
	mac_addr = macbin2blob(var)
	
	
	# Get Designated bridge ID for host1 on port 1
	db1 = getvar(host1, comm1, 3, "%s.%s" % (desigBaseOid,port1), 161, sv1)
	if db1 == None:
		#raise UndefinedSnmpValueException("cannot read %s.%s from %s: %s" % (desigBaseOid,port1,host1,lastsnmperror) )
		raise UndefinedSnmpValueException("cannot read %s.%s from %s" % (desigBaseOid,port1,host1) )
	
	# Extract MAC address
	(priority, db1) = struct.unpack('2s6s', db1)
	db1 = macbin2blob(db1)
	if db1 == '000000000000':
		return (None, "no DB on %s from %s: STP disabled?" % (port1, host1))
	
	
	# Get Designated bridge ID for host2 on port2
	db2 = getvar(host2, comm2, 3, "%s.%s" % (desigBaseOid,port2), 161, sv2)
	if db2 == None:
		#raise UndefinedSnmpValueException("cannot read %s.%s from %s: $lastsnmperror" % (desigBaseOid,port2,host2,lastsnmperror) )
		raise UndefinedSnmpValueException("cannot read %s.%s from %s" % (desigBaseOid,port2,host2) )
	
	# Extract MAC address
	(priority, db2) = struct.unpack('2s6s', db2)
	db2 = macbin2blob(db2)
	if db2 == '000000000000':
		return (None, "no DB on %s from %s: STP disabled?" % (port2,host2))
	
	#
	# The designated bridge is different?
	#
	log.debug( "id1=%s, db1=%s, db2=%s", mac_addr, db1, db2 )
	
	if db1 != db2:
		return (None, "%s != %s" % (db1,db2))
	
	#
	# Ok, now control Port Identifiers.
	#
	#    NOTE: HP 2524 have inverted bytes
	#          Cisco do not use priority
	
	#
	# Get ports priority (for verbose output)
	#
	pp1 = getvar(host1, comm1, 3, "%s.%s" % (portPrioBaseOid, port1), 161, sv1)
	if pp1 == None:
		#raise UndefinedSnmpValueException("cannot read %s.%s from %s: %s" % (portPrioBaseOid,port1,host1,lastsnmperror))
		raise UndefinedSnmpValueException("cannot read %s.%s from %s" % (portPrioBaseOid,port1,host1))

	pp2 = getvar(host2, comm2, 3, "%s.%s" % (portPrioBaseOid, port2), 161, sv2)
	if pp2 == None:
		#raise UndefinedSnmpValueException(None, "cannot read %s.%s from %s: %s" % (portPrioBaseOid,port2,host2,lastsnmperror))
		raise UndefinedSnmpValueException(None, "cannot read %s.%s from %s" % (portPrioBaseOid,port2,host2))

	#
	#
	dp2 = getvar(host2, comm2, 3, "%s.%s" % (desigPortBaseOid, port2), 161, sv2)
	if dp2 == None:
		#raise UndefinedSnmpValueException("cannot read %s.%s from %s: %s" % (desigPortBaseOid,port2,host2,lastsnmperror))
		raise UndefinedSnmpValueException("cannot read %s.%s from %s" % (desigPortBaseOid,port2,host2))
	
	(dp2,) = struct.unpack('!H', dp2)    # convert 2 bytes (network order!) in short int 	
	need_swap = needsDPswap(host2, comm2)
	if need_swap:
		dp2 = ((dp2<<8) & 0xFF00) | ((dp2>>8) & 0x00FF)  # swap bytes
		
	log.debug( "host %s port %s dp: %s (first=%s, second=%s) (swapped: %s), host %s port %s pp=%s", host2,port2,hex(dp2),(dp2>>8),(dp2&0x0F),need_swap,host1,port1,pp1 )
			
	# Remove priority's bits
	dp2 = 0x0FFF & dp2
	
	#
	#	
	dp1 = getvar(host1, comm1, 3, "%s.%s" % (desigPortBaseOid, port1), 161, sv1)
	if dp1 == None:
		#raise UndefinedSnmpValueException("cannot read %s.%s from %s: %s" % (desigPortBaseOid,port1,host1,lastsnmperror))
		raise UndefinedSnmpValueException("cannot read %s.%s from %s" % (desigPortBaseOid,port1,host1))
	
	(dp1,) = struct.unpack('!H', dp1)   # convert 2 bytes (network order!) in short int 
	need_swap = needsDPswap(host1, comm1);
	if need_swap:
		dp1 = ((dp1<<8) & 0xFF00) | ((dp1>>8) & 0x00FF)
		
	log.debug( "host %s port %s dp: %s (first=%s, second=%s) (swapped: %s), host %s port %s pp=%s", host1,port1,hex(dp1),(dp1>>8),(dp1&0x0F),need_swap,host1,port1,pp1 )
	
	# remove priority's bits
	dp1 = 0x0FFF & dp1
	
	#
	# Final check (port1 is a string!)
	#
	log.debug("dp1=%s, dp2=%s", dp1, dp2)
	
	if dp1 != int(port1):
		return (False, "%s ne %s" % (dp1,port1))
	
	if dp2 != int(port1):
		return (False, "%s ne %s" % (dp2,port1))
	
	return (True, "id1=%s, db1=%s, db2=%s, dp1=%s, dp2=%s" % (mac_addr,db1,db2,dp1,dp2))

def needsDPswap(host,comm,sv=1):
   
	sysOid   = '1.3.6.1.2.1.1.2.0'
	sysDescr = '1.3.6.1.2.1.1.1.0'
	
	s = getvar(host, comm, 3, sysOid, 161, sv)
	if s == None:
		raise UndefinedSnmpValueException("cannot determine if %s needs DP byte swapping" % (host))

	#print "needsDPswap(%s,%s,%s) = %s " % (host,comm,sv,s)

	if s == '1.3.6.1.4.1.11.2.3.7.11.9':  
		# HP 4000/8000, all versions
		return True;
	
	if s == '1.3.6.1.4.1.11.2.3.7.11.7':  
		# Documented as HP 8000, never seen
		return True;
	
	if ((s == '1.3.6.1.4.1.11.2.3.7.11.18') or (s == '1.3.6.1.4.1.11.2.3.7.11.19')):  
		# HP 2512 or HP 2524: 
		s = getvar(host, comm, 3, sysDescr, 161, sv)
		if s == None:
			raise UndefinedSnmpValueException("cannot determine if %s needs DP byte swapping" % (host))
		
		# Example: "HP J4813A ProCurve Switch 2524, revision F.05.22, ROM F.01.01  (/sw/code/build/info(s02))"
		#if ($s !~ /revision F\.([0-9]+)\.([0-9]+),/):
		#	return undef;
		m = re.match(r".+revision F\.([0-9]+)\.([0-9]+),", s)
		if not m:
			raise PollException("cannot determine if %s needs DP byte swapping" % (host))
		
		(major,minor) = m.groups()
		
		if int(major) > 5  : return False
		if int(minor) >= 34: return False
		
		return True
	
	return False

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#
	
def ciscoExtendCommunity(log, host, comm, port, sv):
	# Per leggere la VLAN di una porta access:
	# .1.3.6.1.4.1.9.9.68.1.2.2.1.2.porta = vlan
	
	# Per leggere la VLAN nativa di una porta trunk
	# .1.3.6.1.4.1.9.9.46.1.6.1.1.5.porta = nvlan
	
	# Bitmask di VLAN su una porta:
	# .1.3.6.1.4.1.9.9.46.1.6.1.1.4.porta = bitmask	
	
	if not sv:
		sv = 1

	cisco = isCisco(host, comm, sv)
	if cisco == False:
		log.debug("ciscoExtendCommunity(): host %s not Cisco", host)
		return comm
	
	# Access:
	v = getvar(host, comm, 3, '1.3.6.1.4.1.9.9.68.1.2.2.1.2.' + str(port), 161, sv)
	if v != None:
		log.debug("ciscoExtendCommunity(): port %s of %s is access, vlan %s", port, host, v)
		return str(comm + "@" + str(v))
		
	# Trunk:
	#v = getvar(host, comm, 3, '1.3.6.1.4.1.9.9.46.1.6.1.1.5.' + str(port), 161, sv)
	#if v != None:
	#	log.debug("ciscoExtendCommunity(): port %s of %s is trunk, native vlan %s" % (port, host, v))
	#	return str(comm + "@" + str(v))
	
	# Try these vlan bitmasks:
	#
	for (submib, vlan_range_start) in [ (4, 0), (17,1024), (18, 2048), (19,3072) ]:
		bitmask = getvar(host, comm, 3, '1.3.6.1.4.1.9.9.46.1.6.1.1.'+str(submib)+'.'+str(port), 161, sv)
		if bitmask:
			vlan = scan_vlan_bit_mask(log, bitmask, vlan_range_start)
			if vlan != None:
				log.debug("ciscoExtendCommunity(): port %s of %s is trunk, first member vlan is %s", port, host, vlan)
				return str(comm + "@" + str(vlan))



				
	# E qui cosa facciamo? Altro?
	return comm


def scan_vlan_bit_mask(log, bytes, vlan_range_start):	
	# Search for the first bit '1' in the bitmask. That is the member vlan
	n = 0
	for byte in bytes:
		(byte,) = struct.unpack('B',byte)
		log.debug("n=%s, byte is %s", n+vlan_range_start, hex(byte))
		for j in xrange(8):
			bit = (byte & (0x80 >> j))
			if bit != 0:
				return n+vlan_range_start
			else:
				n += 1
	return None

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#


def isCisco(host, comm, sv=1):
	if not sv:
		sv = 1
	
	# The vendor's authoritative identification of the
	# network management subsystem contained in the
	# entity. This value is allocated within the SMI
	# enterprises subtree (1.3.6.1.4.1) and provides an
	# easy and unambiguous means for determining `what
	# kind of box' is being managed. For example, if
	# vendor `Flintstones, Inc.' was assigned the
	# subtree 1.3.6.1.4.1.4242, it could assign the
	# identifier 1.3.6.1.4.1.4242.1.1 to its `Fred
	# Router'.
	sysOid = '1.3.6.1.2.1.1.2.0'
	
	s = getvar(host, comm, 3, sysOid, 161, sv)
	if s == None:
		# NOTE: really strange. Isn't this OID supported by everyone?!
		return False
	
	if s[0:14] ==  "1.3.6.1.4.1.9.":
		return True
	else:
		return False

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def ifIndex2stp(log, host, ifIndex, comm, sv):
	
  	# IfIndex of a port
	# The value of the instance of the ifIndex object,
	# defined in MIB-II, for the interface corresponding
	# to this port." 
	portIfIndexBaseOid = '1.3.6.1.2.1.17.1.4.1.2'
	
	if not sv:
		sv = 1
	
	log.debug("ifIndex2stp(%s, %s, %s, %s)", host, ifIndex, comm, sv)

	s = getvar(host, comm, 3, portIfIndexBaseOid + '.' + str(ifIndex), 161, sv)
	if s != None:
		if s == ifIndex:
			log.debug("ifIndex2stp(): %s %s.%s = %s", host, portIfIndexBaseOid, ifIndex, s)
			return s

	if ifIndex > 100 and ifIndex < 300:
		# 3com SSII 1100 e 3300 ?
		t = ifIndex % 100 + 32 * int(ifIndex/100) - 32
		s = getvar(host, comm, 3, portIfIndexBaseOid + '.' + str(t), 161, sv)
		if s != None:
			if s == ifIndex:
				log.debug( "ifIndex2stp(): %s %s.%s = %s", host, portIfIndexBaseOid, t, s )
				return t

	t = gettable(host, comm, 3, portIfIndexBaseOid, 161, sv)
	if t == None:
		raise UndefinedSnmpValueException("can't get snmp table %s in bridge mib for %s" % (portIfIndexBaseOid, host))
	
	for (k,v) in t:
		# We searched the table '1.2.3' and we got a list of:
		#   [('.1.2.3.xxx', vvvv ), (....)...]
		# Let's extract the 'xxx' part.
		if v == int(ifIndex):
			log.debug("%s. %s = %s  --> %s", host,k,v, k[len(portIfIndexBaseOid)+2:] )
			return ( k[len(portIfIndexBaseOid)+2:] ) 

	raise UndefinedSnmpValueException("cannot find %s in bridge mib for %s" % (ifIndex, host))

