# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import sys
import traceback

import time
from time import time as get_time

import threading

import Queue

import heapq

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def get_current_thread():
	return threading.currentThread()

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class PollSimpleThread(threading.Thread):
	"""
	Helper class.
	"""
	def __init__(self, handler, *args):

		# Do not block main program on exit
		super(PollSimpleThread,self).__init__()

		# Don't block main program on exit
		# FIX: hack for old API version.
		if (hasattr(self, "daemon")):
			self.daemon = True

		if hasattr(self, "setDaemon"):
			self.setDaemon(True)

		self.handler = handler
		self.args    = args

	def run(self):
		self.handler(*self.args)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class PollWorkRequest(object):
	"""
	This object stores all the informations about a single job
	performed by the thread pool.
	"""
	def __init__(self, function, args=None):
		"""
		:param function: The function to execute for this request.
		:type function: function
		
		:param args: Function's arguments.
		:type args: list of objects
		"""

		self.function = function
		self.args = args or []

		self.create_time = get_time()

	def execute(self, executing_thread):
		execution_delay = get_time() - self.create_time

		result = self.function(executing_thread,  execution_delay, *self.args)

	def __str__(self):
		return "%s" % (','.join([str(a) for a in self.args ]))

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

MAX_STATS_HISTORY = 10

class PollWorkerThread(threading.Thread):

	def __init__(self, id, logger, requests_queue, poll_timeout=5, pre_run_callback=None):

		super(PollWorkerThread,self).__init__()

		self.id              = id

		self.__STOP          = False

		self._logger         = logger
		self._requests_queue = requests_queue
		self._poll_timeout   = poll_timeout

		# Info about current job
		self.current_work        = None    # set by this object
		self.current_work_start  = None    # set by this object
		self.current_work_status = None    # set by the current_work (request) handler

		self.count   = 0
		self.errors  = 0

		# Do not block main program on exit
		# FIX: hack for old API version.
		if (hasattr(self, "daemon")):
			self.daemon = True
		else:
			self.setDaemon(True)

		#self.__dismissed = threading.Event()
		self.time_stats = [ 0 for i in xrange(MAX_STATS_HISTORY) ]
		self.time_pos   = 0

		self.pre_run_callback = pre_run_callback

	def stop(self):
		self._logger.info("[worker thread %s] stop received.", self.id)

		self.__STOP = True

	def getLogger(self):
		return self._logger

	def run(self):

		if self.pre_run_callback != None:
			thread_obj = self
			self.pre_run_callback(thread_obj)

		queue   = self._requests_queue
		timeout = self._poll_timeout

		while 1:
			request = None

			if self.__STOP:
				break

			# get next work request. If we don't get a new request from the
			# queue after self._poll_timout seconds, we jump to the start of
			# the while loop again, to give the thread a chance to exit.
			try:
				request = queue.get(True, timeout)
			except Queue.Empty, er:
				continue
			except KeyboardInterrupt, er:
				break
			except SystemExit, er:
				break

			if request != None:

				start = get_time()

				self.current_work_status = None    # \
				self.current_work_start  = start   #  Attention: these variables should be set atomically. possibible race condition with get_current_work()
				self.current_work        = request # /

				try:
					result = request.execute( self )
				except Exception, ex:
					self.errors += 1
					self._logger.error("[thread %s]: caught error %s -> %s" % (self.id, ex, traceback.format_exc()))

				elapsed = get_time() - start

				self.time_stats[ self.time_pos % MAX_STATS_HISTORY ] = elapsed
				self.time_pos += 1

				self.count += 1

				self.current_work        = None  # \
				self.current_work_status = None  #  Attention: these variables should be set atomically. possibible race condition with get_current_work()
				self.current_work_start  = None  # /

		self._logger.info("[worker thread %s] exit", self.id)

	def get_stats(self):

		if self.time_pos > 0:
			avg = (sum(self.time_stats) /  float(min(self.time_pos, MAX_STATS_HISTORY)))
		else:
			avg = 0.0

		mins_v = [ v for v in self.time_stats if v != 0 ]
		if mins_v:
			min_v = min(mins_v)
		else:
			min_v = 0

		return ( min_v, max(self.time_stats), avg )

	def get_current_work(self):

		# ATTENTION: thise self.current_* have to be set atomically by run(), but this is not true. At least, be sure all variables are not None
		status   = self.current_work_status
		ts_start = self.current_work_start
		work     = self.current_work

		if (work is not None) and (ts_start is not None) and (status is not None):
			return (status, work, get_time() - ts_start)
		else:
			return (None, None, 0)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def float_format(number, left_pad=7, precision=2):
	"""
	Format a number in a way that will align decimal points.
	"""
	outer = '{0:>%i}.{1:<%i}' % (left_pad, precision)
	inner = '{:.%if}' % (precision,)
	return outer.format(*(inner.format(number).split('.')))

class PollThreadPool(object):

	def __init__(self, logger, worker_class=PollWorkerThread, queue_size=-1, resq_size=0, poll_timeout=5):

		self.start_time = get_time()

		self.__STOP = False

		self.logger = logger

		self.worker_class = worker_class

		# single queue with infinite size
		self.logger.debug("Creating thread queue...")
		self._requests_queue = Queue.Queue(queue_size)

		self.workers = []

	def stop(self):

		self.logger.info("Stopping thread pool")
		self.__STOP = True

		for w in self.workers:
			w.stop()

	def createWorkers(self, num_workers, poll_timeout=5, pre_run_callback=None):

		for i in xrange(num_workers):
			self.addWorker(poll_timeout, pre_run_callback)

	def addWorker(self, poll_timeout=5,  pre_run_callback=None):

		wid = str(len(self.workers) + 1)

		w = self.worker_class(
			wid,
			self.logger, 
			self._requests_queue, 
			poll_timeout          = poll_timeout,  
			pre_run_callback      = pre_run_callback,
		)

		self.workers.append(w)
		w.start()

	def __putRequest(self, request, block=True, timeout=0):
		"""
		Put work request into work queue and save its id for later.
		"""
		block = False
		timeout = 0

		self._requests_queue.put(request, block, timeout)

	def execJob(self, function, args_list=None):

		if self.__STOP:
			return

		self.__putRequest( PollWorkRequest(function, args_list) )

	def findRequest(self, function, fun_cmp_params):

		for request in self._requests_queue.queue:
			if request.function == function:
				if fun_cmp_params(request.args) == True:
					return request.create_time
		return None

	def get_info_text(self, all_threads_info=True):

		infos = self.get_info(all_threads_info=all_threads_info)

		t_info = []

		format_str = '  '.join(    ('%10s'    , '%7s'    ,'%5s'  , '%4s' , '%8s'  , '%8s'  , '%8s'  , '%8s'  , '%s'  ) )
		header     = format_str  % ('thread'  , 'working', 'done', 'errs', 'min'  , 'max'  , 'avg'  , 'elap' , 'job' )
		t_info.append(header)

		for line in infos['threads']:
			t_line = format_str % line
			t_info.append(t_line)

		t_info.append('    Uptime: %sm %ss'           % (infos['uptime']/60, infos['uptime']%60) )
		t_info.append('    Jobs / s: %.3f'            % (infos['rate']         ) )
		t_info.append('    Jobs tot: %s'              % (infos['processed']    ) )
		t_info.append('    Request queue size: %s (max:%s)' % (infos['wait_queue'], infos['wait_queue_max']   ) )

		return '\n'.join(t_info)

	def get_info(self, all_threads_info=True):

		infos = {
			'critical_problems': {}
		}

		#
		# Threads
		#
		MIN = 0xffffffff

		run_tot   = 0
		count_tot = 0
		errs_tot  = 0
		min_tot   = MIN
		max_tot   = 0
		avg_tot   = 0
		elap_tot  = 0

		t_info = []
		for t in self.workers:
			name = t.getName()[7:]
			(status, work, elapsed) = t.get_current_work()

			working   = '*'  if work is not None else ''
			work_info = '%s: %s' % (status, work) if work is not None else ''

			count   = t.count
			errors  = t.errors

			mine, maxe, avge = t.get_stats()

			if (all_threads_info==True) or (working and not all_threads_info):
				line = ( 
					name,
					working, 
					count, 
					errors, 
					float_format(mine, 4, 3) if mine != 0 else '-',
					float_format(maxe, 4, 3) if maxe != 0 else '-',
					float_format(avge, 4, 3) if avge != 0 else '-',
					float_format(elapsed, 4, 3) if elapsed != 0 else '-',
					work_info,
				)
				t_info.append(line)

			run_tot   += 1 if work is not None else 0
			count_tot += count
			errs_tot  += errors

			if mine > 0: min_tot = min(min_tot, mine)

			max_tot  = max(max_tot, maxe)
			avg_tot  = max(avg_tot, avge)
			elap_tot = max(elap_tot, elapsed)

		min_tot = min_tot if min_tot != MIN else -1

		line = (
			"TOT: %s" % len(self.workers), 
			run_tot, 
			count_tot, 
			errs_tot, 
			float_format(min_tot , 4, 3) if min_tot  != 0 else '-',
			float_format(max_tot , 4, 3) if max_tot  != 0 else '-',
			float_format(avg_tot , 4, 3) if avg_tot  != 0 else '-',
			float_format(elap_tot, 4, 3) if elap_tot != 0 else '-',
			''
		)
		t_info.append(line)

		infos['threads'] = t_info



		#
		# Other 
		#
		uptime = int(get_time() - self.start_time)

		infos['uptime']         = int(get_time() - self.start_time)

		infos['processed']      = count_tot
		
		infos['process'] = {
			'min': min_tot,
			'max': max_tot,
			'avg': avg_tot,
		}

		if uptime > 0:
			infos['rate']           = float(count_tot) / float(uptime)
		else:
			infos['rate']           = 0

		infos['wait_queue']     = self._requests_queue.qsize()
		infos['wait_queue_max'] = self._requests_queue.maxsize

		return infos

	def get_queue_size(self):
		return self._requests_queue.qsize()

	def get_workder(self, id):
		for thread in self.workers:
			if thread.id == id:
				return thread
		return None

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class TimeQueue(object):
	"""
	This class mantains a time queue of events. Every events is rappresented
	by:
	
	* event's start time
	* event's data (Tipically a tuple <handler,args>)
	"""
	
	def __init__(self):
		# Priority Queue (heap queue).
		self.__priority_queue = []

	def size(self):
		"""Returns the number of events saved in the time queue."""
		return len(self.__priority_queue)

	def flush(self):
		del self.__priority_queue[:]

	def popNextEvent(self):
		"""
		Extracts the next event from the time queue.
		
		:exception: 
			:class:`IndexError`  if the time queue is empty
		"""
		# FIX: there is anything faster?
		return heapq.heappop(self.__priority_queue)

	def addEvent(self, starttime, event_data):
		"""
		Insert a new event inside the time queue.
		
		:param starttime: Event start time.
		:type starttime: number
		
		:param data: Event's data.
		:type data: object
		"""

		# Use time as priority
		# FIX: there is anything faster?
		heapq.heappush( self.__priority_queue, (starttime, event_data) )

	def getNextTimeout(self):
		"""
		Returns the start time of the first element of the time queue.
		Returns *None* if the time queue is empty.
		"""
		if len(self.__priority_queue) > 0:
			(starttime, event_data) = self.__priority_queue[0]
			return starttime
		else:
			return None

	def findEvent(self, fun_cmp):
		for (starttime, event_data) in self.__priority_queue:
			if fun_cmp(event_data):
				return starttime
		return -1

	def peekEvents(self, num):
		events = self.__priority_queue[0:num]
		if len(events) > 0:
			events.sort()
			return events
		else:
			return []

	def moveEvent(self, fun_find, new_starttime):
		"""
		Move event described by fun_find inside scheduled heap at ts new_starttime.

		:param fun_find: lambda function which describes an event
		:param new_starttime: timestamp to move given event
		:return: True if found and moved, False otherwise
		"""
		for i in xrange(len(self.__priority_queue)):
			(starttime, event) = self.__priority_queue[i]
			if fun_find(event):

				# FIX: rimuoviamo un elemento e ricreiamo lo HEAP
				#      E' costosissimo. Fortuna che e' una operazione che facciamo "poco" spesso.
				del self.__priority_queue[i]
				heapq.heapify(self.__priority_queue)

				self.addEvent( new_starttime, event )
				return True

		return False

	def removeEvent(self, fun_find):
		"""
		Remove event described by fun_find from scheduled heap.

		:param fun_find: lambda function which describes an event
		:return: event if found and removed, None otherwise
		"""

		for i in xrange(len(self.__priority_queue)):
			(starttime, event) = self.__priority_queue[i]
			if fun_find(event):
				# FIX: rimuoviamo un elemento e ricreiamo lo HEAP
				#      E' costosissimo. Fortuna che e' una operazione che facciamo "poco" spesso.
				del self.__priority_queue[i]
				heapq.heapify(self.__priority_queue)

				return event

		return None

#-----------------------------------------------------------------------------#
# PollOrganizer                                                               #
#-----------------------------------------------------------------------------#

class PollOrganizer(threading.Thread):
	"""
	This class implements a thread which aim is to check wether a datagroup can be executed
	according to its scheduling group or not.
	"""

	def __init__(self, logger, handler, polltimeout=5):
		super(PollOrganizer, self).__init__()

		self.organizerHandler = handler

		self.logger = logger

		self.inputQueue = Queue.Queue()
		self.timeout = polltimeout

		self.__STOP = False

		# Ritardo dell'ultimo job tra il tempo di inserimento nella coda e tempo di 
		# "gestione" da parte dell'Organizer
		self._last_run_delay = 0
		self._start_ts = 0
		self._run_rate = 0
		self._run_count = 0

	def get_info(self):

		return {
			'queue_size'       : self.inputQueue.qsize(),
			'run_delay'        : self._last_run_delay,
			'run_rate'         : float(self._run_count) / float(time.time() - self._start_ts),
			'critical_problems': { },
		}

	def stop(self):
		self.__STOP = True

	def newJob(self, handler, params):
		ts = time.time()
		self.inputQueue.put( (ts, handler, params) )

	def run(self):
		queue            = self.inputQueue
		timeout          = self.timeout
		organizerHandler = self.organizerHandler

		self._start_ts = time.time()
	
		while True:
			handler = None
			params = None

			# Keeping the same thread execution structure
			if self.__STOP:
				break

			# get next work request. If we don't get a new request from the
			# queue after self._poll_timout seconds, we jump to the start of
			# the while loop again, to give the thread a chance to exit.
			try:
				ts, handler, params = queue.get(True, timeout)

				self._run_count += 1
				self._last_run_delay = time.time() - ts

				organizerHandler(handler, params)
			except Queue.Empty, er:
				continue
			except KeyboardInterrupt, er:
				break
			except SystemExit, er:
				break
			except Exception, er:
				self.logger.error('[organizer] caught error %s:%s -> %s' % (type(er), str(er), traceback.format_exc()))

#-----------------------------------------------------------------------------#

class PollScheduler(threading.Thread):
	"""
	This class implement a simple event scheduler. An event is simply a function that will be called after a specified timeout.
	
	The scheduler simply performs timeout/scheduling computation. Nothing else.
	
	Each scheduled event will be submitted as a job to a ThreadPool object.
	
	**WARNING: The implemented scheduling algorithm is neither hard nor soft real-time.**
	"""
	def __init__(self, schedule_callback):
		super(PollScheduler, self).__init__()

		self.schedule_callback = schedule_callback

		# Don't block main program on exit
		# FIX: hack for old API version.
		if (hasattr(self, "daemon")):
			self.daemon = True
		else:
			self.setDaemon(True)

		self.__STOP = False

		self.__queue       = TimeQueue()
		self.__queueIsFull = threading.Condition()

		self.__my_time     = get_time()

	def scheduleNewEvent(self, timeout, handler, handler_params):
		self.__queueIsFull.acquire()
		self.__queue.addEvent(timeout, (handler, handler_params) )
		self.__queueIsFull.notify()
		self.__queueIsFull.release()
		return True

	def whenIsEvent(self, handler, fun_match_data):

		# Every scheduled event is a tuple (handler, handler_params)
		fun_cmp = lambda x:  True if (x[0]==handler and fun_match_data(x[1])) else False
		event_time = self.__queue.findEvent(fun_cmp)

		return event_time

	def alterEventScheduleTime(self, handler, fun_match_data, new_time):

		fun_find = lambda x:  True if (x[0]==handler and fun_match_data(x[1])) else False

		# Lock della coda & modifica dell'evento (se c'e')
		# Se non abbiamo toccato niente, non ha senso segnalare modifiche alla coda.
		self.__queueIsFull.acquire()
		modified = self.__queue.moveEvent(fun_find, new_time)
		if modified:
			self.__queueIsFull.notify()
		self.__queueIsFull.release()

		return modified

	def removeEvent(self, handler, fun_match_data):

		fun_find = lambda x: True if (x[0] == handler and fun_match_data(x[1])) else False

		# Lock della coda & modifica dell'evento (se c'e')
		# Se non abbiamo toccato niente, non ha senso segnalare modifiche alla coda.
		self.__queueIsFull.acquire()
		event = self.__queue.removeEvent(fun_find)
		if event:
			self.__queueIsFull.notify()
		self.__queueIsFull.release()

		return event

	def getScheduledEvents(self, num):

		events = self.__queue.peekEvents(num)

		els = []
		for e in events:
			starttime                 = e[0]
			(handler, handler_params) = e[1]
			els.append(  (starttime, handler, handler_params) )

		return els

	def stop(self):
		self.__STOP = True

	def run(self):
		self.schedule()

	def flush(self):
		self.__queueIsFull.acquire()
		self.__queue.flush()
		self.__queueIsFull.notify()
		self.__queueIsFull.release()

	def schedule(self):
		schedule_callback = self.schedule_callback

		queue = self.__queue
		isFull = self.__queueIsFull

		isFull.acquire()      

		while 1:
			if self.__STOP:
				break

			# Is there an event to schedule?
			nexttimeout = queue.getNextTimeout()

			if nexttimeout == None:
				self.__my_time = get_time()

				# Wait for event with higher priority or stop signal
				# Scheduler queue is empty, wait for events or stop signal
				isFull.wait()
			else:
				now = get_time()
				waittime = nexttimeout - now

				handler, handler_params = None, None

				self.__my_time = now

				# if the scheduler is late then 'waittime' can be <= 0,
				# but we don't mind! Wait only for future! 
				if waittime > 0.005:  # speed up a bit!

					start = get_time()
					isFull.wait(waittime)                        # UNLOCK, WAIT, LOCK
					now = get_time()

					elapsed = now - start

					# Timeout elapsed?
					if elapsed >= waittime:
						starttime, (handler, handler_params) = queue.popNextEvent()
					else:
						# new events with higher priority or stop signal?
						pass
				else:
					starttime, (handler, handler_params) = queue.popNextEvent()

				# No new events, only a timeout!
				if handler != None:
					isFull.release()                             # UNLOCK

					schedule_callback(handler, handler_params)

					isFull.acquire()                             # LOCK

		# End loop
		isFull.release()                                                     # UNLOCK

	def get_schedule_size(self):
		return self.__queue.size()

	def get_text_info(self):
		return "N. scheduled = %s (last op time:%s)" % (self.__queue.size(), time.ctime(self.__my_time))


class AgentThread(threading.Thread):

	def __init__(self, name, run_callback, callback_params_dict=None):

		# Do not block main program on exit
		super(AgentThread, self).__init__()

		# Don't block main program on exit
		# FIX: hack for old API version.
		if (hasattr(self, "daemon")):
			self.daemon = True
		else:
			self.setDaemon(True)

		self.name = name
		self.run_callback = run_callback
		self.callback_params_dict = callback_params_dict

	def run(self):
		if self.callback_params_dict is not None:
			self.run_callback(**self.callback_params_dict)
		else:
			self.run_callback()
