# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import re, os
import StringIO
import time
import urllib
import tempfile
import logging
import traceback

from HTMLParser import HTMLParser

try:
	import pycurl
except ImportError, e:
	pass

from sanet_poller.lib.exceptions import PollException

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

WEBSCENARIO_STATUS_UNKNOWN = 0
WEBSCENARIO_STATUS_UP      = 1
WEBSCENARIO_STATUS_DOWN    = 2

STEP_STATUS_UNKNOWN  = 0
STEP_STATUS_OK       = 1
STEP_STATUS_ERROR    = 2
STEP_STATUS_SKIPPED  = 3

HOST_IP_RE = "(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])"
HOST_NAME_RE = "(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])"
PROXY_CONFIG_STRING_RE = "^(https?)://((?P<user>[^@]+)@)?(?P<host>" + HOST_IP_RE + "|" + HOST_NAME_RE + ")(:(?P<port>\w+))?$"

SAVE_STEP_LOG_NO            = 0 
SAVE_STEP_LOG_ONLY_ERRORS   = 1
SAVE_STEP_LOG_ALL_RESPONSES = 2

# Nome del "gruppo" riconosciuto da usare per estrarre variabili dalla pagine usando espressioni regolari coi gruppi.
VAR_TO_FETCH_GROUP_NAME="VAR_TO_FETCH"

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class BaseWebException(Exception):
	pass

class TimeOutException(BaseWebException):
	pass

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def get_proxy_config(config_str):
	
	if not config_str:
		return None
	
	m = re.match(PROXY_CONFIG_STRING_RE, config_str)
	if m:
		return m.groupdict()
	
	raise BaseWebException("Invalid proxy configuration")


def extract_used_vars(s):
	used_vars = re.findall("(?:{{([^}]+)}})", s)
	return used_vars

def replace_var(s, var_name, var_value):	
	return s.replace("{{%s}}" % (var_name), var_value)


#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

SSL_VERSION_TO_CURL = {
	'SSLv2': pycurl.SSLVERSION_SSLv2,
	'SSLv3': pycurl.SSLVERSION_SSLv3,
	'TLSv1': pycurl.SSLVERSION_TLSv1,
}

def execute_scenario(logger, ws, follow_location=1, additional_macros=None, ssl_version=None):
	
	if additional_macros is None:
		additional_macros = {}
	

	no_error = True
	step_results = []
	status = WEBSCENARIO_STATUS_UP
	

	logger.debug("Executing scenario: %s", ws['name'])

	proxy_config = get_proxy_config(ws['proxy'])
	logger.debug("Using proxy: %s", proxy_config)

	ts_start = time.time()
	
	logger.debug("Creating curl session")
	curl_session = pycurl.Curl()
	
	#
	# SSL VERSION
	#
	ssl_version = ssl_version or ws['ssl_version']
	if ssl_version:
		ssl_version = SSL_VERSION_TO_CURL[ ssl_version ]
		curl_session.setopt(pycurl.SSLVERSION, ssl_version)

	logger.debug("Using SSL: %s", ssl_version)
	
	current_request_vars = {}
	for step in ws['steps']:

		# Se abbiamo preso errore da uno step precedente ignoriamo tutti quelli successivi
		if no_error:
			try:
				r = exec_single_step(logger, 
				                     curl_session, 
				                     step, 
				                     ws['agent'], 
				                     request_vars      = current_request_vars, 
				                     follow_location   = follow_location, 
				                     proxy_config      = proxy_config, 
				                     additional_macros = additional_macros, 
				                     save_mode         = ws['save_step_log'],
				)
				if r['matched']:
					r['exec_code'] = STEP_STATUS_OK
					r['info'     ] = 'success'
				else:
					r['exec_code'] = STEP_STATUS_ERROR
					r['info'     ] = 'not matched'
					no_error = False
					
					status = WEBSCENARIO_STATUS_DOWN

				step_results.append(r)

				# Le variabili estratte dallo step vengono usate per lo step successivo
				current_request_vars = r['vars']

			except BaseWebException, e:
				step_results.append({
					'uuid'     : step['uuid'],
					'exec_code': STEP_STATUS_ERROR, 
					'exec_time': time.time(),
					'info'     : str(e)
				}) 
				no_error = False
				
				status = WEBSCENARIO_STATUS_DOWN
			except Exception, e:
				msg = "Error on step %s: %s:%s" % (step, type(e), str(e))
				logger.debug(msg)
				raise PollException(msg)
				
		else:
			step_results.append({
				'uuid'     : step['uuid'],
				'exec_code': STEP_STATUS_SKIPPED, 
				'exec_time': time.time(),
				'info'     : 'skipped' 
			}) 
			continue


	ts_end = time.time()

	#
	execution_data = {
		'uuid'     : ws['uuid'],
		'name'     : ws['name'],

		'status'   : status,
		'exec_time': ts_start,
		'totaltime': ts_end - ts_start,

		'steps'    : step_results,
	}

	return execution_data

def exec_single_step(logger, curl_session, step, agent, request_vars = None, follow_location=1, proxy_config=None, additional_macros=None, save_mode=SAVE_STEP_LOG_NO):
	
	if additional_macros is None:
		additional_macros = {}
		
	logger.info("start step %s", step['order'])

	buffer_header = StringIO.StringIO()
	buffer_page   = StringIO.StringIO()

	curl_session.setopt(pycurl.WRITEFUNCTION , buffer_page.write  )
	curl_session.setopt(pycurl.HEADERFUNCTION, buffer_header.write)

	if proxy_config: 
		curl_session.setopt(pycurl.PROXY    , proxy_config['host'])

		if proxy_config.has_key('port'):
			curl_session.setopt(pycurl.PROXYPORT, int(proxy_config['port']))

		if proxy_config.has_key('user'):
			#
			# FIX: 'user' contiene sia username che password. E se la password contiene ':' che succede? Ricontrollare documentazione CURL
			#
			#credentials = '%s:%s' % (proxy_config['username'], proxy_config['password'])
			credentials = proxy_config['user']
			print credentials
			curl_session.setopt(pycurl.PROXYUSERPWD, credentials)
	else:
		curl_session.setopt(pycurl.PROXY, '')

	#
	# L'url puo' essere espanso che le variabili prese al giro precedente
	#
	url = str(step['url'])
	logger.debug("\tURL: %s", url)
	
	used_vars = extract_used_vars(url)
	if used_vars:
		logger.debug("Using vars: %s", used_vars)
		for used_var in used_vars:
			if used_var in request_vars:
				value = request_vars[used_var]
				url = url.replace("{{%s}}" % (used_var), value)
			elif used_var in additional_macros:
				value = additional_macros[used_var]
				url = url.replace("{{%s}}" % (used_var), value)
				
		logger.debug("\t\tEXPANDED: %s", str(url))
	
	url = url.encode('ascii')
	curl_session.setopt(pycurl.URL, url)
	
	curl_session.setopt(pycurl.FOLLOWLOCATION, follow_location)
	curl_session.setopt(pycurl.COOKIEFILE, "")
	#curl_session.setopt(pycurl.COOKIEJAR , "")
	
	#curl_session.setopt(pycurl.VERBOSE, True)
	logger.debug("\tdisable ssl verify peer")
	curl_session.setopt(pycurl.SSL_VERIFYPEER, 0)
	curl_session.setopt(pycurl.SSL_VERIFYHOST, 0)

	
	logger.info("\ttimeout: %s", step['timeout'])
	curl_session.setopt(pycurl.TIMEOUT, step['timeout']) #timeout in secondi
	
	if agent != None:
		logger.debug("\tuser agent: %s", str(agent))
		curl_session.setopt(pycurl.USERAGENT, str(agent))


	if step['headers'] != None:
		logger.debug("\theader: %s", step['headers'])
		
		headers = step['headers']
		used_vars = extract_used_vars(headers)
		if used_vars:
			logger.debug("Using vars: %s", used_vars)
			for used_var in used_vars:
				if used_var in request_vars:
					value = request_vars[used_var]
					headers = headers.replace("{{%s}}" % (used_var), value)
				elif used_var in additional_macros:
					value = additional_macros[used_var]
					headers = headers.replace("{{%s}}" % (used_var), value)
		
		logger.debug("\theader: %s", headers)
		
		curl_session.setopt(pycurl.HTTPHEADER, str(headers).split('\n'))	
	

	logger.info("\tmethod: %s", step['method'])

	if step['method'] == 'post':

		logger.debug("\tPOST VARS");

		post_vars     = step['post']
		raw_post_data = step['raw_post_data']
		
		if raw_post_data:
			value = raw_post_data
			
			logger.debug("\t\t\t raw_post_data = %s", value);
			
			used_vars = extract_used_vars(value)
			if used_vars:
				for used_var in used_vars:
					if used_var in request_vars:
						v = request_vars[used_var]
						value = value.replace("{{%s}}" % (used_var), v)
					elif used_var in additional_macros:
						v = additional_macros[used_var]
						value = value.replace("{{%s}}" % (used_var), v)
			
			logger.debug("\t\t\t EXPANDED = %r", value);

			post_data = value
			
			post_data = post_data.encode('ascii')
			
			curl_session.setopt(pycurl.POST, 1)
			curl_session.setopt(pycurl.POSTFIELDSIZE, len(post_data))
			
			import cStringIO
			fin = cStringIO.StringIO(post_data)

			def _ioctl(*args):
				#print __file__, "_ioctl", args
				if cmd == curl.IOCMD_RESTARTREAD:
					fin.seek(0)
			
			def _read(*args):
				try:
					#print __file__, "_read", args
					data = fin.read(*args)
					return data
				except Exception, e:
					print __file__,  "_read", "ERROR", str(e)
					raise e
			
			curl_session.setopt(pycurl.IOCTLFUNCTION, _ioctl)
			curl_session.setopt(pycurl.READFUNCTION, _read)
		
		else:
			d = {}
			for name, value, uuid in post_vars:
				logger.debug("\t\t%s = %s", name, value);
				
				used_vars = extract_used_vars(value)
				if used_vars:
					for used_var in used_vars:
						if used_var in request_vars:
							v = request_vars[used_var]
							value = value.replace("{{%s}}" % (used_var), v)
						elif used_var in additional_macros:
							v = additional_macros[used_var]
							value = value.replace("{{%s}}" % (used_var), v)
					logger.debug("\t\t\t EXPANDED = %s", value);
				
				d[name] = value
			
			d.update(request_vars)
			
			"""
			VECCHIO CODICE
			
			post_data = urllib.urlencode(d)
			
			curl_session.setopt(pycurl.POST, 1)
			curl_session.setopt(pycurl.POSTFIELDS, post_data)
			"""
			
			post_data = '&'.join(['%s=%s' % (name, urllib.quote(value)) for name, value in d.items() ])

			logger.debug("\t\t\t EXPANDED = %r", post_data);

			import cStringIO
			fin = cStringIO.StringIO(post_data)

			def _ioctl(*args):
				#print __file__, "_ioctl", args
				if args[0] == pycurl.IOCMD_RESTARTREAD:
					fin.seek(0)
			
			def _read(*args):
				try:
					#print __file__, "_read", args
					data = fin.read(*args)
					return data
				except Exception, e:
					print __file__,  "_read", "ERROR", str(e)
					raise e
			
			curl_session.setopt(pycurl.IOCTLFUNCTION, _ioctl)
			curl_session.setopt(pycurl.READFUNCTION, _read)


			curl_session.setopt(pycurl.POST, 1)
			curl_session.setopt(pycurl.POSTFIELDSIZE, len(post_data))

			
	else:
		curl_session.setopt(pycurl.POST, 0) 


	try:
		logger.debug("\tPerforming...")
		curl_session.perform()

		logger.debug("\t\tdone")
		
		http_code = curl_session.getinfo(pycurl.HTTP_CODE)

		logger.debug("\tHTTP code: %s", http_code)
		
		header_content = buffer_header.getvalue()
		page_content   = buffer_page.getvalue()
		
		logger.log(1, header_content)
		logger.log(1, page_content)

		buffer_header.close()
		buffer_page.close()

		#
		# FIX: spostare il processing dei dati della pagina in un altro metodo
		#
		
		matched = True
		
		if step['required_status_code']:
			if step['required_status_code'] != http_code:
				logger.debug("\tHTTP code %s != required %s", http_code, step['required_status_code'] )
				matched = False
		
		if matched:
			if step['required_string']:
				try:
					if not re.search(str(step['required_string']), page_content):
						matched = False
				except Exception, e:
					matched = False
				
				logger.debug("\tmatching with string '%s'? %s", step['required_string'], matched)
		
		fetched_vars = {}
		for name, regexp, uuid in step['vars']:
			try:
				logger.debug("\tSearching variable '%s' regexpr: %s", str(name), str(regexp))
				m = re.search(str(regexp), page_content)
				
				if m != None:
					
					# Match con uso del gruppo o match semplice?
					foundvalue = m.groupdict().get(VAR_TO_FETCH_GROUP_NAME, None)
					if foundvalue is not None:
						logger.debug("\t\tfound %s match value %s", VAR_TO_FETCH_GROUP_NAME, str(foundvalue))
					else:
						foundvalue = m.groups()[0]
						logger.debug("\t\tfound match value %s", str(foundvalue))

					logger.debug("\t\tfound value %s", str(foundvalue))
					
					contains_escaped_chars = re.search(r'&\w+;', foundvalue)
					
					if contains_escaped_chars:
						
						# Replace escaped chars founded
						
						parser = HTMLParser()
						foundvalue_escaped = parser.unescape(foundvalue)
						logger.debug("\t\tfound value %s", str(foundvalue_escaped))
						foundvalue = foundvalue_escaped

					fetched_vars[name] = foundvalue
				else:
					logger.warning("\t\tStep variabile %s NOT FOUND.", name)
			except Exception, e:
				logger.debug(str(e))
		
		logger.debug("\tsave mode = %s", str(save_mode))
		
		
		if save_mode == SAVE_STEP_LOG_NO or (matched and save_mode == SAVE_STEP_LOG_ONLY_ERRORS):
			header_content = None
			page_content   = None
		
		#
		# Pycurl non effettua encoding e decoding della pagina, ma noi dobbiamo
		# avere la stringa page_content in unicode.
		# Tentiamo di capire il charset cercando nell'header e tentiamo una
		# conversione. Se non ci riusciamo forziamo utf-8 ignorando eventuali
		# caratteri strani.
		#
		if page_content is not None:

			try:
				encoding = re.findall('Content-Type: .*; charset=(.*)',header_content)[-1]
				page_content = page_content.decode(encoding)

			except Exception, e:
				page_content = page_content.decode('utf-8', errors='replace')
		
		return_data = {
			'uuid'       : step['uuid'],
			'exec_time'  : time.time(),
		
			'http_code'  : http_code,
			'totaltime'  : curl_session.getinfo(curl_session.TOTAL_TIME),
			
			'header'     : header_content,
			'response'   : page_content,
			
			'matched'    : matched,
			
			'vars'       : fetched_vars
		}
		
		return return_data
	
	except pycurl.error, e:
		if e[0] == pycurl.E_OPERATION_TIMEOUTED:
			logger.error("timeout")
			raise TimeOutException("timeout")
		else:
			logger.error("Error on step: %s:%s -> %s", type(e), str(e), traceback.format_exc())
			raise BaseWebException("%s:%s" % (type(e), str(e)) )

