# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import signal
import getopt
import logging
import logging.handlers
import json

import threading

from sanet_common.pythonutils import load_symbol

from sanet_poller.lib.functions import load_extra_expr_and_agentstatus_functions

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def __hack_keyboard_interrupt():

	from sanet_common.osutils.proctitle import get_proc_title, set_proc_title

	print "*** handling Ctrl+C ***"

	original_proc_title = get_proc_title()

	# Fork. The parent process waits for a Ctrl+C. The child process 
	# performs the business logic.
	childpid = os.fork()
	if childpid <> 0:
		# MAIN PROCESS
		try:
			set_proc_title( original_proc_title + ' (keyboard supervisor)' )

			# wait for child termination
			os.wait()
		except KeyboardInterrupt:
			# I put the capital B in KeyBoardInterrupt so I can
			# tell when the Watcher gets the SIGINT
			#log.debug('POLLER: KeyBoardInterrupt! killing child process.')
			print "Ctrl+C"
			try:
				os.killpg(os.getpgrp(), signal.SIGKILL)
			except OSError: 
				pass
		sys.exit()
	else:
		# CHILD PROCESS: ignore SIGINT
		signal.signal(signal.SIGINT, signal.SIG_IGN)

		set_proc_title( original_proc_title + ' (checker)' )

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

LOG_CRITICAL = 50
LOG_ERROR    = 40
LOG_WARNING  = 30
LOG_INFO     = 20
LOG_DEBUG    = 10

#: Constants for mapping log level names to numeric values.
log_levels = {
	'CRITICAL': LOG_CRITICAL,
	'ERROR'   : LOG_ERROR,
	'WARNING' : LOG_WARNING,
	'INFO'    : LOG_INFO,
	'DEBUG'   : LOG_DEBUG
}

logger = None

DEFAULT_LOG_FILE = "/var/log/sanet/poller_log"

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def init_logger(agent_id, log_file, log_level, logtostdout=False):	

	global logger

	identifier = '%s' % (agent_id)

	format_str = '%(asctime)s %(levelname)s agent/' + identifier + ' > %(message)s'

	logger = logging.getLogger('poller-%s' % (agent_id))
	logger.setLevel( log_level )

	hdlr = logging.handlers.WatchedFileHandler(log_file, "a")
	hdlr.setFormatter( logging.Formatter(format_str) )
	hdlr.setLevel( log_level )
	logger.addHandler(hdlr)

	if logtostdout:
		hdlr = logging.StreamHandler()
		hdlr.setFormatter( logging.Formatter(format_str) )
		hdlr.setLevel( log_level )
		logger.addHandler(hdlr)	

	return logger

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def init_agent_cache_manager(handler_module_name, init_params):

	handler_class = load_symbol( handler_module_name )
	handler       = handler_class(**init_params)

	return handler

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

DEFAULT_SANET_CACHE_MODULE = "sanet_common.cache.rediscache.RedisCacheManager"

def launch_agent(
	agent,

	address,
	auth,

	tenantid,
	agentid,

	threads_number,

	logfile,
	loglevel,
	logtostdout,

	cache_module,
	cache_module_params,

	extra_pkgs
):
	# Set thread stack size to 1Mb. FIX: make it configurable.y
	threading.stack_size( 1024 * 1024 )

	logger = init_logger(agentid, logfile, loglevel, logtostdout=logtostdout )

	logger.info("*****************************************************************")
	logger.info("************************* new start *****************************")
	logger.info("*****************************************************************")

	logger.info("Stack size: %s", threading.stack_size() )

	logger.info("Loading cache manager %s...", cache_module)
	logger.info("\t%s", cache_module_params)

	cache_manager = init_agent_cache_manager(cache_module, cache_module_params)

	#
	# Loading modules
	#
	if extra_pkgs:
		logger.info("Loading extra symbols %s...", extra_pkgs)
		extra_pkgs = extra_pkgs.split(':')

		extra_symbols, extra_agentstatus_functions = load_extra_expr_and_agentstatus_functions(extra_pkgs)
	else:
		extra_symbols, extra_agentstatus_functions = None, None

	#
	# Setup agent
	#
	logger.debug("Creating Agent")
	agent.setup(tenantid, agentid, logger, cache_manager , num_working_threads=threads_number, extra_symbols=extra_symbols, extra_agentstatus_functions=extra_agentstatus_functions)

	agent.set_broker( address )
	agent.set_auth( auth )

	#
	# Start
	#
	logger.debug("Starting Agent")
	agent.start()

	return

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def poller_main(agent, argv):

	opt_tenantid       = None
	opt_agentid        = None
	opt_broker         = None
	opt_loglevel       = LOG_INFO
	opt_logfile        = DEFAULT_LOG_FILE

	opt_threads_number = None 

	opt_extra_pkgs     = None

	opt_cache_module   = DEFAULT_SANET_CACHE_MODULE

	opt_cache_module_params = {}

	opt_verbose        = False
	opt_keyboard_int   = False

	try:
		opts, args = getopt.getopt(argv, "hB:t:i:L:l:T:X:c:C:vk")

		for o, a in opts:

			if   o in ("-h"):
				show_help()
				sys.exit(0)

			elif o in ("-B"): 
				opt_broker = a

			elif o in ("-t"): 
				opt_tenantid = a
			elif o in ("-i"): 
				opt_agentid = a

			elif o in ("-L"):
				if log_levels.has_key(a):
					opt_loglevel = log_levels[a]
				else:
					opt_loglevel = int(a)
			elif o in ("-l"):
				opt_logfile = a
			elif o in ("-T"):
				n = int(a)
				if n != 0:
					opt_threads_number = n
			elif o in ("-X"):
				opt_extra_pkgs = a

			elif o in ("-c"):
				opt_cache_module = a
			elif o in ("-C"):
				opt_cache_module_params = json.loads(a)

			elif o in ("-v"): 
				opt_verbose = True
			elif o in ("-k"):
				opt_keyboard_int = True

	except getopt.error, msg:
		print "Error while parsing options: %s" % msg
		show_help()
		sys.exit()
	except ValueError, ex:
		print "Invalid value for input parameter: %s" % ex
		show_help()
		sys.exit()

	if opt_agentid == None:
		print "agent id not specified"
		sys.exit(-1)

	data = opt_broker.split('@')
	if len(data) > 1:
		opt_auth, opt_address = data
		opt_auth = opt_auth.split('/')
	else:
		opt_address = data[0]
		opt_auth = None

	opt_address = opt_address.split(':')
	if len(opt_address) > 1:
		opt_address[1] = int(opt_address[1])

	if opt_keyboard_int:
		__hack_keyboard_interrupt()

	launch_agent(
		agent, 

		opt_address,
		opt_auth,

		opt_tenantid,
		opt_agentid,

		opt_threads_number,

		opt_logfile,
		opt_loglevel,
		opt_verbose,
		opt_cache_module,
		opt_cache_module_params,
		opt_extra_pkgs
	)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def show_help():

	print "Usage: <options>" 
	print ""
	print "Options:"
	print "\t", "-B <broker address>   ", "Broker address <user>:<password@<host>"
	print "\t", "-t <tenant name>      ", "Tenant name"
	print "\t", "-i <agent name>       ", "Agent name"
	print "\t", "-h                    ", "Show help and exit"
	print "\t", "-L <log level>        ", "Log level"
	print "\t", "-l <log file>         ", "Log file name"
	print "\t", "-T <number of threads>", ""
	print "\t", "-X <extra packages>   ", ""
	print "\t", "-c <cache module>     ", ""
	print "\t", "-C <cache options>    ", ""
	print "\t", "-k                    ", "Stop agent with Ctrl+C"
	print "\t", "-v                    ", "Show log on standard output"

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

if __name__ == "__main__":

	import os
	import signal

	from sanet_poller.agent               import BaseAgent
	from sanet_poller.agent_new_scheduler import BaseAgentAdvancedScheduling

	SANET_NO_SCHEDULINGGROUPS = os.environ.get('SANET_NO_SCHEDULINGGROUPS', False)
	if SANET_NO_SCHEDULINGGROUPS:
		Sanet3Agent = BaseAgent
	else:
		Sanet3Agent = BaseAgentAdvancedScheduling

	agent = Sanet3Agent()
	
	#from sanet_poller.agent_new_buffered  import BufferedAgent
	#agent = BufferedAgent()

	poller_main(agent, sys.argv[1:])

