# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""
REMOTE AGENT Launcher
============================================================================================================

Il server per lanciare i singoli agenti remoti defini nel file di configurazione.


Protocollo di lancio agenti
----------------------------

1) Il launcher legge il file di configurazione globale
2) Per ogni *sezione agente* definita nel file di configurazione esegure un agente remoto passandogli in input i parametri di esecuzione (pidfile e configfile)
3) Ogni agente funziona da supervisor di se stesso. Legge il proprio file di configurazione ed esegue tutti i sotto processi e li riavvia se necessario.

::
    
                                         launcher.py -----------> <globlal config file>
                                           |    |
                                           |    |
                              *exec*       |    |
                                 +---------+    +---------+
                                 |                        |
                                 v                        v
       <conf file> <---remotepoller/main.py      remotepoller/main.py---> <conf file>
                                 |     
                                 | *exec*    
                      +----------+----------+---------------+
                      |          |          |               |
                      v          v          v               v
                   remote remote-snmptrap remote-checker  remote-discovery


Variabile d'ambiente SANET_REMOTE_HOME
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Il launcher cerca gli script e gli altri file dell'agente remoto nella directory dov'e' posizionato il file launch.py.

La directory di default puo' essere cambiata impostando la variabile d'ambiente SANET_REMOTE_HOME:

::
	
	export SANET_REMOTE_HOME=/usr/local/sanetremotepoller


"""

import os
import os.path
import sys
import getopt
import subprocess

#==============================================================================#

SANET_REMOTE_INSTALL_DIR=None

SANET_REMOTE_INSTALL_DIR = os.environ.get("SANET_REMOTE_HOME", SANET_REMOTE_INSTALL_DIR)

if not SANET_REMOTE_INSTALL_DIR:
	SANET_REMOTE_INSTALL_DIR = os.path.abspath(os.path.join(os.path.dirname( os.path.abspath(__file__) ) ) )

sys.path.append( SANET_REMOTE_INSTALL_DIR )

#-----------------------------------------------------------------------------#
# CONSTANTS                                                                   #
#-----------------------------------------------------------------------------#

GLOBAL_CONFIG_FILE_NAME = "global.ini"

DEFAULT_CONFIG_DIR      = '/etc/sanet3-remote'

VAR_DIR                 = '/var/run'

REMOTE_AGENT_BIN        = os.path.join(SANET_REMOTE_INSTALL_DIR, 'remotepoller/main.py')

#-----------------------------------------------------------------------------#
# GLOBALS                                                                     #
#-----------------------------------------------------------------------------#

verbose = False

def verbose_print(*args):
	if verbose: print ' '.join(args)

#-----------------------------------------------------------------------------#
# MAIN                                                                        #
#-----------------------------------------------------------------------------#

def main():

	global verbose

	config_file = None

	try:
		opts, args = getopt.getopt(sys.argv[1:], "hc:v")
		for o, a in opts:

			if   o in ("-h"):
				print_help()
				sys.exit()

			elif o in ("-c"):
				config_file = a

			elif o in ("-v"):
				verbose = True

	except getopt.error, msg:
		sys.stderr.write("Error while parsing options: %s\n" % (msg,))
		print_help()
		sys.exit()
	except ValueError, e:
		sys.stderr.write("Invalid value for input parameter: %s\n" % (e))
		print_help()
		sys.exit()

	if not args:
		print_help()
		sys.exit(-1)
	
	#
	# Read Global config file
	#
	if not config_file:
		config_file = get_config_file_path()
		verbose_print("Using global configuration file:", config_file)


	configuration = read_configuration_file(config_file)

	
	#
	# Process arguments
	#
	cmd = args[0]

	agent_id = None
	if len(args) > 1:
		agent_id = args[1]


	if cmd == 'list':
		show_list(configuration)

	elif cmd == 'test':
		start_all(configuration, show_only=True)

	elif cmd == 'start':

		if not agent_id:
			start_all(configuration)
		else:
			if agent_id not in configuration.keys():
				raise Exception("Invalid agent id: %s" % (agent_id))

			start_agent(configuration, agent_id)

	elif cmd == 'stop':

		if not agent_id:
			stop_all(configuration)
		else:
			if agent_id not in configuration.keys():
				raise Exception("Invalid agent id: %s" % (agent_id))

			stop_agent(configuration, agent_id)

	else:
		print_help()

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def print_help():
	print "Sanet remote agent launcher:"
	print "%s [-v] [-c <config file> ] list               List configured agents." % (sys.argv[0])
	print "%s [-v] [-c <config file> ] test               Show execution commands for each agent." % (sys.argv[0])
	print "%s [-v] [-c <config file> ] start [agent_id]   Start agents or single agent <agent_id>." % (sys.argv[0])
	print "%s [-v] [-c <config file> ] stop  [agent_id]   Stop agents or single agent <agent_id>." % (sys.argv[0])


#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def get_config_file_path():

	if os.environ.get("SANET_REMOTE_HOME"):
		config_file = os.path.join( os.environ.get("SANET_REMOTE_HOME"), "etc", GLOBAL_CONFIG_FILE_NAME )
	else:
		config_file = os.path.join( DEFAULT_CONFIG_DIR                        , GLOBAL_CONFIG_FILE_NAME )

	return config_file


def read_configuration_file(file_path):

	import ConfigParser
	
	if not os.path.isabs(file_path):
		file_path = os.path.abspath(file_path)



	try:
		if verbose:
			verbose_print("Loading global settings from", file_path)

		configfile = ConfigParser.ConfigParser()
		configfile.readfp(open(file_path, 'r'))

		configuration = {}
		for s in configfile.sections():

			agent_config = dict( configfile.items(s) )

			if agent_config.has_key('config_file'):
				config_file = agent_config['config_file']
			else:
				config_file = os.path.join('agents', s)

			# Se il file di configurazione e' relativo
			# lo posizioniamo all'interno della directory  etc/agents
			if not os.path.isabs(config_file):
				config_dir  = os.path.dirname(file_path)
				config_file = os.path.join(config_dir, config_file)
				agent_config['config_file'] = config_file

			if not agent_config.has_key('pid_file'):
				file_name = 'sanet_remote_%s.pid' % (s)
				agent_config['pid_file'] = os.path.join(VAR_DIR, file_name)

			configuration[s] = agent_config

		return configuration

	except IOError, e:
		raise Exception("Error on configuration file '%s': %s" % (file_path, str(e),))
	except ConfigParser.Error, e:
		raise Exception("Error while reading configuration file '%s': %s" % (file_path, str(e),))

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def show_list(configuration):

	print "Agents:"
	for agent_id, config in sorted(configuration.items()):
		print "\t", agent_id, "(conf: %s, pidfile: %s)" % (config['config_file'], config['pid_file'])

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def start_all(configuration, show_only=False):

	verbose_print("Starting all defined agents...")

	for agent_id in sorted(configuration.keys()):
		start_agent(configuration, agent_id, show_only=show_only)

	return

#-----------------------------------------------------------------------------#

def start_agent(configuration, agent_id, show_only=False):

	if not show_only:
		verbose_print("Starting", agent_id)

	config_file = configuration[agent_id]['config_file'] 
	pid_file    = configuration[agent_id]['pid_file'] 

	args = [
		'python',
		REMOTE_AGENT_BIN, 
		'-n', agent_id,
		'-c', config_file, 
		'-p', pid_file, 
		'start'
	]

	if not show_only:

		verbose_print("Executing command", ' '.join(args))
		subprocess.call(args)
	else:
		print ' '.join(args)

	return
#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def stop_all(configuration):
	"""
	In base al file di configurazione <configuration> cerca di eseguire lo stop di tutti gli agenti definiti
	"""

	verbose_print("Stopping all defined agents...")

	for agent_id in sorted(configuration.keys()):
		stop_agent(configuration, agent_id)

	return

#-----------------------------------------------------------------------------#

def stop_agent(configuration, agent_id):
	"""
	In base alla configurazione globale <configuration> esegue l'<agente_id> con il comando di stop.
	"""

	verbose_print("Stopping", agent_id)

	pid_file    = configuration[agent_id]['pid_file'] 
	config_file = configuration[agent_id]['config_file']

	args = [
		'python',
		REMOTE_AGENT_BIN, 
		'-n', agent_id,
		'-c', config_file, 
		'-p', pid_file, 
		'stop'
	]

	verbose_print("Cmd:", ' '.join(args))

	subprocess.call(args)

	return

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

if __name__ == "__main__":

	main()




