# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import time
from time import time as get_time
import traceback

import logging
import logging.handlers

from StringIO import StringIO

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

from sanet_common.sanetdefs.ipc import SANETD_CLT_QUEUE, AGENT_CHECKER_QUEUE

from sanet_common.ipc import AbstractIPCFactory

from sanet_poller.agent import BaseAgent
from sanet_poller.agent import RemoteServerException

from remotepoller.lib.ipc import *
from remotepoller.lib.checker import RemoteDataGroupChecker

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class RemoteCheckerIPCFactory(AbstractIPCFactory):

	def __init__(self, logger, agent):
		super(RemoteCheckerIPCFactory, self).__init__(logger)
		self.agent = agent

	def create_connection(self):
		return self.agent.create_ipc_connection()

#-----------------------------------------------------------------------------#

class RemoteCheckerAgent(object):

	def __init__(self, logger, settings):
	
		self._stop = False

		self.settings = settings

		self.logger    = logger

		self.tenant_id = None
		self.agent_id  = None

		self.load_config_timeout     = 3
		self.load_config_retry_delay = 5

		self.restart_delay = 5  # pausa in secondi tra un turno di discovery e l'altro

	def create_ipc_connection(self):

		# Rispetto alla versione base, per creare una connessione usiamo
		# i parametri letti dal file di configurazione

		self.broker_address = (self.settings['server_host']          , self.settings.get('server_port'    , None) )
		self.broker_auth    = (self.settings.get('server_user', None), self.settings.get('server_password', None) )

		connect_options = {}

		self.logger.debug("Broker address: %s", self.broker_address )

		host = self.broker_address[0]
		if len(self.broker_address) > 1: 
			try:
				if self.broker_address[1]:
					connect_options['port']     = int(self.broker_address[1])
			except Exception, e:
				pass

		if self.broker_auth:
			connect_options['username'] = self.broker_auth[0]
			if len(self.broker_auth) > 1:
				connect_options['password'] = self.broker_auth[1]

		self.logger.debug("Opening connection with server %s", host )
		self.logger.debug("Options:")
		for k,v in connect_options.items():
			self.logger.debug("\t%s = %s", k, v)

		c = IPCConnection( host, **connect_options)

		self.logger.debug("Connection created!")

		return c

	def launch(self):

		self.logger.debug("Init")

		self.load_settings_from_server()

		self.handle_commands()

		return

	def is_stopping(self):
		return self._stop

	def load_settings_from_server(self):

		tenant_name = self.settings['tenant']
		agent_name  = self.settings['agent']

		#
		# FIX: fissare un numero massimo di tentativi per cercare di scaricare la configurazione
		#      dal server.
		#
		while 1:
			try:
				self.logger.info("Loading settings from server...")

				# Create
				connection = self.create_ipc_connection()
				connection.open()

				# Bind
				replyid = 'temp-agentctl-%s' % (os.getpid())
				connection.bind(replyid)

				# Send/Receive
				config_from_server = self._send_request_and_receive(connection, 'get_agent_config', tenant_name, agent_name, timeout=self.load_config_timeout)

				# Close
				connection.close()

				self._update_settings_from_config(config_from_server)

				break

			except RemoteServerException, er:
				self.logger.critical("Server error. Unable to load settings from server: %s", er)
			except Exception, er:
				self.logger.critical("Unknow error while loading settings from server: %s", er)

			# Something gone wrong. Wait and retry
			time.sleep(self.load_config_retry_delay)

		return True

	def _send_request_and_receive(self, connection, rtype, *params, **kwargs):

		DEFAULT_TIMEOUT = 60
		timeout = kwargs.get('timeout', DEFAULT_TIMEOUT)

		mid = str(get_time())
		msg = encode_request_with_id(rtype, mid, *params)
		connection.send_msg(SANETD_CLT_QUEUE, msg, timeout=timeout)

		(addr, msg) = connection.receive_msg(ack_msg=True, timeout=timeout)

		if (msg == None):
			self.logger.info("chiedo %s %s", addr, msg)
			raise Exception("Timeout while loading configuration for %s" % rtype)

		response = decode_response(msg)

		if response['id'] != mid:
			raise RemoteServerException("Waiting for response with id %s. Received %s" % (mid, response['id']))

		if response['status'] == RESPONSE_ERROR:
			raise RemoteServerException("Unable to get elements from server: %s" % ( response['content'] ) )

		return response['content']

	def _update_settings_from_config(self, config_from_server):

		self.logger.info("Config from server: %s", config_from_server)		

		# Set all global settings
		self.tenant_id            = config_from_server['tenantid']
		self.agent_id             = config_from_server['agentid']

	def handle_commands(self):

		self.logger.info("Start...")

		while True:

			if self.is_stopping():
				break

			try:
				self.logger.info("Creating connection")

				self.connection = self.create_ipc_connection()
				self.connection.open()

				self.connection.bind( AGENT_CHECKER_QUEUE(self.agent_id), persistent=False)

				while not self.is_stopping():
					self.receive_command()

				self.connection.ack_pending()
				self.connection.close()

			except IPCConnectionError, e:
				self.logger.error("Connection lost")
			finally:
				try:
					if self.connection:
						self.connection.close()
				except Exception, e:
					pass

		self.logger.info("Exit")


	def receive_command(self):

		try:
			self.logger.debug("[HELPER] Receiving...")

			(msg_addr, msg) = self.connection.receive_msg(timeout=1, ack_msg=False)
			if msg is None:
				return

			self.logger.debug("[ctl] Received %s:%s", msg_addr, msg)

			self.single_request_handler(msg_addr, msg)

		except IPCNotFound, e:
			self.logger.error("[ctl] Error on connection: %s", e)
			pass
		except IPCException, e:
			self.logger.error("[ctl] Error on connection: %s", e)
			raise e	
		except Exception, e:
			self.logger.warning("[ctl] Error while processing message: %s", e)

	def single_request_handler(self, msg_addr, msg):

		try:
			request = decode_request(msg)

			request_type   = request['type']
			request_id     = request['id']
			request_params = request['params']

			if   request_type == 'exec_datagroup': response_data = self.exec_datagroup(request_params)
			elif request_type == 'stop_checker'  : response_data = self.stop_checker(request_params)
			else:
				raise Exception("Invalid request: %s" % (request['type']))

			response = encode_response_with_id( RESPONSE_SUCCESS, request_id, response_data )
		except Exception, ex:
			response = encode_response_with_id( RESPONSE_ERROR  , request_id, "Error while processing request: %s:%s" % (str(ex), traceback.format_exc()) )

		try:
			self.connection.send_msg( msg_addr, response, blocking=False)
		except IPCException, ex:
			self.logger.error("[HELPER] Error while processing request for %s: %s", msg_addr, str(ex))

	def exec_datagroup(self, params):
		params = params[0]

		datagroup_id        = params['datagroup_id']
		log_level           = params['loglevel']

		self.logger.info("Executing datagroup %s", datagroup_id)

		#
		# START
		#

		self.logger.debug("Init ipc factory")
		ipc_factory = RemoteCheckerIPCFactory(self.logger, self)

		connection = self.create_ipc_connection()
		connection.open()

		self.logger.debug("Init datagroup check")
		remotechecker = RemoteDataGroupChecker()


		#
		# Init cache
		#
		self.logger.debug("Init cache manager")
		try:
			cache_handler =            self.settings['cache_module']
			cache_params  = json.loads(self.settings['cache_config'])
		except Exception, e:
			self.logger.debug("Error in cache manager configuration: %s. Using default manager.", str(e))
			cache_handler = "sanet_common.cache.rediscache.RedisCacheManager"
			cache_params  = {}
			
		self.logger.debug("Cache manager settings: %s %s",cache_handler, cache_params)

		remotechecker.set_cache_params(cache_handler, cache_params)

		#
		# LOG
		#
		stream = StringIO()
		exec_logger = remotechecker.init_log(log_level, stream=stream)

		self.logger.debug("executing")
		
		#
		# EXECUTE
		#
		try:
			updates_data = remotechecker.execute(exec_logger, connection, self.tenant_id, datagroup_id, ipc_factory)
		except Exception, e:
			self.logger.error("ERROR WHILE EXECUTING CHECK: %s", str(e))
			updates_data = None

		self.logger.debug("finish")
		log = stream.getvalue()

		remotechecker.dispose()

		stream.close()

		self.logger.debug(">>>>>>>>>>>>>>>>>>>>> DATAGROUP EXECUTION LOG START <<<<<<<<<<<<<<<<<<<<<<")
		for l in log.split('\n'):
			self.logger.debug('LOG >>>> ' + l)
		self.logger.debug(">>>>>>>>>>>>>>>>>>>>> DATAGROUP EXECUTION LOG END  <<<<<<<<<<<<<<<<<<<<<<<")
		#
		# END
		#

		data = {
			'updates_data': BaseAgent.encode_update_events(updates_data),
			'log'         : log,
		}
		
		return data

	def stop_checker(self, params):
		params = params[0]

		self.logger.info("Received stop command %s")

		self._stop = True

		data = {}

		return data

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def init_logger(agent_id, log_file, log_level, logtostdout=False):	

	global logger

	identifier = '%s' % (agent_id)

	format_str = '%(asctime)s %(levelname)s checker/' + identifier + ' > %(message)s'

	logger = logging.getLogger('poller-%s' % (agent_id))
	logger.setLevel( log_level )


	hdlr = logging.handlers.WatchedFileHandler(log_file, "a")
	hdlr.setFormatter( logging.Formatter(format_str) )
	hdlr.setLevel( log_level )
	logger.addHandler(hdlr)

	if logtostdout:
		hdlr = logging.StreamHandler()
		hdlr.setFormatter( logging.Formatter(format_str) )
		hdlr.setLevel( log_level )
		logger.addHandler(hdlr)	

	return logger

def launch_remote_checker_from_settings(settings, logtostdout=True):

	tenantid            = settings['tenant'],
	agentid             = settings['agent'],

	logfile             = settings['log_file']
	loglevel            = 10 #settings['log_level']

	#
	# Init log
	#
	logger = init_logger(agentid, logfile, loglevel, logtostdout=logtostdout )

	logger.info("*****************************************************************")
	logger.info("************************* new start *****************************")
	logger.info("*****************************************************************")

	logger.debug('Using settings:')
	for k, v in sorted(settings.items()):
		logger.debug('    %s=%s', k, v)

	#
	# Launch agent
	#
	checkeragent = RemoteCheckerAgent(logger, settings)
	checkeragent.launch()

	# Never returns
	return
