
# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

####################################################################################
MAIN
####################################################################################

TODO

"""

import os
import sys
import time
from time import time as get_time

import logging

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

from sanet_common.sanetdefs.ipc import AUTOCONF_CTL_QUEUE

from sanet_poller.agent import RemoteServerException
from sanet_poller.discovery import BaseDiscoveryAgent

from remotepoller.lib.ipc import *

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class RemoteDiscoveryAgent(BaseDiscoveryAgent):

	"""
	Versione "remota" dell'agente di discovery.

	Rispetto alla versione base questa versione si connetta al server
	autoconfd e scarica alcuni parametri di configurazione prima di iniziare.
	"""

	def __init__(self, logger, settings):
		super(RemoteDiscoveryAgent, self).__init__(logger)

		self.settings = settings

		self.load_config_timeout     = 3
		self.load_config_retry_delay = 5

	def init(self):
		super(RemoteDiscoveryAgent, self).init()

		# La connessione col server e' gia' stata creata. 
		# Scarichiamo i parametri di configurazione
		self._load_settings_from_server()

	def create_ipc_connection(self):
		# Rispetto alla versione base, per creare una connessione usiamo
		# i parametri letti dal file di configurazione
		self.broker_address = (self.settings['server_host']          , self.settings.get('server_port'    , None) )
		self.broker_auth    = (self.settings.get('server_user', None), self.settings.get('server_password', None) )

		connect_options = {}

		self.logger.debug("Broker address: %s", self.broker_address )

		host = self.broker_address[0]
		if len(self.broker_address) > 1: 
			try:
				if self.broker_address[1]:
					connect_options['port']     = int(self.broker_address[1])
			except Exception, e:
				pass

		if self.broker_auth:
			connect_options['username'] = self.broker_auth[0]
			if len(self.broker_auth) > 1:
				connect_options['password'] = self.broker_auth[1]

		self.logger.debug("Opening connection with server %s", host )
		self.logger.debug("Options:")
		for k,v in connect_options.items():
			self.logger.debug("\t%s = %s", k, v)

		c = IPCConnection( host, **connect_options)

		self.logger.debug("Connection created!")

		return c

	#--------------------------------------------------------------#
	# Metodi privati                                               #
	#--------------------------------------------------------------#

	def _load_settings_from_server(self):
		
		tenant_name = self.settings['tenant']
		agent_name  = self.settings['agent']

		#
		# FIX: fissare un numero massimo di tentativi per cercare di scaricare la configurazione
		#      dal server.
		#
		while 1:
			try:
				self.logger.info("Loading settings from server...")

				connection = self.connection

				replyid = 'temp-agentctl-%s' % (os.getpid())
				connection.bind(replyid)

				config_from_server = self._send_request_and_receive(connection, 'get_agent_config', tenant_name, agent_name, timeout=self.load_config_timeout)

				self._update_settings_from_config(config_from_server)

				break
			except RemoteServerException, er:
				self.logger.critical("Server error. Unable to load settings from server: %s", er)
			except Exception, er:
				self.logger.critical("Unknow error while loading settings from server: %s", er)

			# Something gone wrong. Wait and retry
			time.sleep(self.load_config_retry_delay)

		return True

	def _send_request_and_receive(self, connection, rtype, *params, **kwargs):

		DEFAULT_TIMEOUT = 60
		timeout = kwargs.get('timeout', DEFAULT_TIMEOUT)

		mid = str(get_time())
		msg = encode_request(rtype, *params)
		connection.send_msg(AUTOCONF_CTL_QUEUE, msg, timeout=timeout)

		(addr, msg) = connection.receive_msg(ack_msg=True, timeout=timeout)

		if (msg == None):
			self.logger.info("chiedo %s %s", addr, msg)
			raise Exception("Timeout while loading configuration for %s" % rtype)

		response = decode_response(msg)

		if response['status'] == RESPONSE_ERROR:
			raise RemoteServerException("Unable to get elements from server: %s" % ( response['content'] ) )

		return response['content']

	def _update_settings_from_config(self, config_from_server):

		self.logger.info("Config from server: %s", config_from_server)

		# Set all global settings
		self.tenant_id            = config_from_server['tenantid']
		self.agent_id             = config_from_server['agentid']

		self.source               = self.agent_id

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def init_logger(agent_id, log_file, log_level, logtostdout=False):	

	global logger

	identifier = '%s' % (agent_id)

	format_str = '%(asctime)s %(levelname)s discovery/' + identifier + ' > %(message)s'

	logger = logging.getLogger('poller-%s' % (agent_id))
	logger.setLevel( log_level )

	hdlr = logging.handlers.WatchedFileHandler(log_file, "a")
        hdlr.setFormatter( logging.Formatter(format_str) )
	hdlr.setLevel( log_level )
	logger.addHandler(hdlr)

	if logtostdout:
		hdlr = logging.StreamHandler()
		hdlr.setFormatter( logging.Formatter(format_str) )
		hdlr.setLevel( log_level )
		logger.addHandler(hdlr)	

	return logger


def launch_remote_discovery_from_settings(settings, logtostdout=True):

	tenantid            = settings['tenant'],
	agentid             = settings['agent'],
	
	logfile             = settings['discovery_log_file']
	loglevel            = settings['discovery_log_level']

	#
	# Init log
	#
	logger = init_logger(agentid, logfile, loglevel, logtostdout=logtostdout )

	logger.info("*****************************************************************")
	logger.info("************************* new start *****************************")
	logger.info("*****************************************************************")

	logger.debug('Using settings:')
	for k, v in sorted(settings.items()):
		logger.debug('    %s=%s', k, v)

	#
	# Launch agent
	#
	discoveryagent = RemoteDiscoveryAgent(logger, settings)
	discoveryagent.launch()

	# Never returns
	return



