# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import time
from time import time as get_time
import traceback

import logging
import logging.handlers

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

from sanet_common.sanetdefs.ipc import SANETD_CLT_QUEUE

from sanet_common.uuids import create_global_uuid

from sanet_poller.agent import RemoteServerException

from remotepoller.lib.ipc import *

from sanet_poller.helpers.netinfo import NetInfoAgentHelper

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class RemoteNetInfoAgent(NetInfoAgentHelper):
	
	"""
	Questa classe implementa solo la logica:
		- per collegarsi al server centrale e scaricare la configurazione
		- far partire il vero "helper" NetInfoAgentHelper
	"""

	def __init__(self, logger, settings):

		self.settings = settings

		self.logger    = logger

		self.tenantid = None
		self.agentid  = None

		self.load_config_timeout     = 3
		self.load_config_retry_delay = 5

	def create_ipc_connection(self):

		# Rispetto alla versione base, per creare una connessione usiamo
		# i parametri letti dal file di configurazione

		self.broker_address = (self.settings['server_host']          , self.settings.get('server_port'    , None) )
		self.broker_auth    = (self.settings.get('server_user', None), self.settings.get('server_password', None) )

		connect_options = {}

		self.logger.debug("Broker address: %s", self.broker_address )

		host = self.broker_address[0]
		if len(self.broker_address) > 1: 
			try:
				if self.broker_address[1]:
					connect_options['port']     = int(self.broker_address[1])
			except Exception, e:
				pass

		if self.broker_auth:
			connect_options['username'] = self.broker_auth[0]
			if len(self.broker_auth) > 1:
				connect_options['password'] = self.broker_auth[1]

		self.logger.debug("Opening connection with server %s", host )
		self.logger.debug("Options:")
		for k,v in connect_options.items():
			self.logger.debug("\t%s = %s", k, v)

		c = IPCConnection( host, **connect_options)

		self.logger.debug("Connection created!")

		return c

	def launch(self):

		self.logger.debug("Init")

		self.load_settings_from_server()

		self.handle_commands()

		return


	def load_settings_from_server(self):

		tenant_name = self.settings['tenant']
		agent_name  = self.settings['agent']

		#
		# FIX: fissare un numero massimo di tentativi per cercare di scaricare la configurazione
		#      dal server.
		#
		while 1:
			try:
				self.logger.info("Loading settings from server...")

				# Create
				connection = self.create_ipc_connection()
				connection.open()

				# Bind
				replyid = 'temp-agentctl-%s-%s' % (agent_name, create_global_uuid())
				connection.bind(replyid)

				# Send/Receive
				config_from_server = self._send_request_and_receive(connection, 'get_agent_config', tenant_name, agent_name, timeout=self.load_config_timeout)

				# Close
				connection.close()

				self._update_settings_from_config(config_from_server)

				break

			except RemoteServerException, er:
				self.logger.critical("Server error. Unable to load settings from server: %s", er)
			except Exception, er:
				self.logger.critical("Unknow error while loading settings from server: %s", er)

			# Something gone wrong. Wait and retry
			time.sleep(self.load_config_retry_delay)

		return True

	def _send_request_and_receive(self, connection, rtype, *params, **kwargs):

		DEFAULT_TIMEOUT = 60
		timeout = kwargs.get('timeout', DEFAULT_TIMEOUT)

		mid = str(get_time())
		msg = encode_request_with_id(rtype, mid, *params)
		connection.send_msg(SANETD_CLT_QUEUE, msg, timeout=timeout)

		(addr, msg) = connection.receive_msg(ack_msg=True, timeout=timeout)

		if (msg == None):
			self.logger.info("chiedo %s %s", addr, msg)
			raise Exception("Timeout while loading configuration for %s" % rtype)

		response = decode_response(msg)

		if response['id'] != mid:
			raise RemoteServerException("Waiting for response with id %s. Received %s" % (mid, response['id']))

		if response['status'] == RESPONSE_ERROR:
			raise RemoteServerException("Unable to get elements from server: %s" % ( response['content'] ) )

		return response['content']

	def _update_settings_from_config(self, config_from_server):

		self.logger.info("Config from server: %s", config_from_server)

		# Set all global settings
		self.tenantid            = config_from_server['tenantid']
		self.agentid             = config_from_server['agentid']

	#-----------------------------------------------------------------------------#
	#                                                                             #
	#-----------------------------------------------------------------------------#

	def handle_commands(self):

		self.logger.info("Start receiving commands")

		# Execute NetInfoAgentHelper.handleHelperCommands()
		self.handleHelperCommands()

		# Never returns
		return

	def is_stopping(self):
		#
		# Implement NetInfoAgentHelper.is_stopping()
		#
		return False
	
	def create_connection(self):
		#
		# Implement NetInfoAgentHelper.create_connection()
		#
		return self.create_ipc_connection()


#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def init_logger(agent_id, log_file, log_level, logtostdout=False):	

	global logger

	identifier = '%s' % (agent_id)

	format_str = '%(asctime)s %(levelname)s checker/' + identifier + ' > %(message)s'

	logger = logging.getLogger('poller-%s' % (agent_id))
	logger.setLevel( log_level )


	hdlr = logging.handlers.WatchedFileHandler(log_file, "a")
	hdlr.setFormatter( logging.Formatter(format_str) )
	hdlr.setLevel( log_level )
	logger.addHandler(hdlr)

	if logtostdout:
		hdlr = logging.StreamHandler()
		hdlr.setFormatter( logging.Formatter(format_str) )
		hdlr.setLevel( log_level )
		logger.addHandler(hdlr)	

	return logger


def launch_remote_netinfo_from_settings(settings, logtostdout=True):

	tenantid            = settings['tenant'],
	agentid             = settings['agent'],

	# Netinfo settings
	if not settings.get('netinfo_log_file' , None): settings['netinfo_log_file' ] = "/var/log/sanet_remote_%s_netinfo.log" % (settings['agent'])
	if not settings.get('netinfo_log_level', None): settings['netinfo_log_level'] = 20
	settings['netinfo_log_level'] = int(settings['netinfo_log_level'])

	logfile  = settings['netinfo_log_file']
	loglevel = settings['netinfo_log_level']

	#
	# Init log
	#
	logger = init_logger(agentid, logfile, loglevel, logtostdout=logtostdout )

	logger.info("*****************************************************************")
	logger.info("************************* new start *****************************")
	logger.info("*****************************************************************")

	logger.debug('Using settings:')
	for k, v in sorted(settings.items()):
		logger.debug('    %s=%s', k, v)

	#
	# Launch
	#
	checkeragent = RemoteNetInfoAgent(logger, settings)
	checkeragent.launch()

	# Never returns
	return
