# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""
========================================================================================
SNMPTRAPPER AGENT
========================================================================================


::
    
                                   REMOTE SERVER                                          SANET SERVER
                                   
                +---------------------------------------------------+                 +----------------------+
                | SNMPTRAPPER                                       |                 | TRAPD                |
                |                                                   |                 |                      |
                |               (data + address )                   |                 |                      |
                |    UDP   ---------QUEUE---------->  MAIN  --------------IPC QUEUE--------> * process *     |
                |   Thread                           Thread         |                 |                      |
                |     |                                             |                 |                      |
                |     |                          RemoteTrapperAgent |                 |                      |
     SNMP       |     v                                             |                 |                      |
     TRAP -------->* UDP SOCKET                                     |                 |                      |
                |                                                   |                 |                      |
                +---------------------------------------------------+                 +----------------------+

"""

import os
import sys
import time
from time import time as get_time
import traceback
import json
import base64

import logging
import logging.handlers

import threading
import Queue

from sanet_common.sanetdefs.ipc import SANETD_CLT_QUEUE, SNMPTRAP_DATA_QUEUE, AGENT_TEMP_REPLY_QUEUE

from sanet_poller.agent import RemoteServerException

from remotepoller.lib.ipc import *

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class TrapReceiverThread(threading.Thread):

	def __init__(self, server, logger, queue):

		super(TrapReceiverThread, self).__init__()

		# Don't block main program on exit
		# FIX: hack for old API version.
		if (hasattr(self, "daemon")):
			self.daemon = True

		if hasattr(self, "setDaemon"):
			self.setDaemon(True)

		self.server = server
		self.logger = logger
		self.queue  = queue

		self.listen_socket = None
		self.restart_wait = 3

	def run(self):

		self.logger.debug("[UDP trap receiver] starting...")

		while 1:

			listen_address = ( self.server.settings['snmptrap_listen_address'], self.server.settings['snmptrap_listen_port'] )

			try:
				self.logger.debug("[UDP trap receiver] Opening listen socket")

				self.listen_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
				self.listen_socket.bind( listen_address )

				while 1:

					self.logger.info('[UDP trap receiver] Waiting for trap on UDP %s...', listen_address)

					(data, address_tuple) = self.listen_socket.recvfrom(4096)

					data = {
						'ts'     : time.time(),
						'address': address_tuple,
						'data'   : base64.b64encode(data),
					}

					self.logger.debug("[UDP trap receiver] Received: %s", data['data'])

					self.queue.put_nowait(data)

			except Exception, e:
				self.logger.error("[UDP trap receiver] error: %s", e)

			if self.listen_socket:
				try:
					self.listen_socket.close()
				except Exception, e:
					pass

			self.logger.error("[UDP trap receiver] Wait %s seconds before restart...", self.restart_wait)
			time.sleep(self.restart_wait)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#


class RemoteTrapperAgent(object):

	def __init__(self, logger, settings):

		self._stop = False

		self.settings = settings

		self.logger    = logger

		self.tenant_id = None
		self.agent_id  = None

		self.load_config_timeout     = 3
		self.load_config_retry_delay = 5

		self.restart_delay = 5  

	def launch(self):

		self.logger.info("Init")

		self.load_settings_from_server()

		self.handle_traps()

		return

	def load_settings_from_server(self):

		tenant_name = self.settings['tenant']
		agent_name  = self.settings['agent']

		#
		# FIX: fissare un numero massimo di tentativi per cercare di scaricare la configurazione
		#      dal server.
		#
		while 1:
			try:
				self.logger.info("Loading settings from server...")

				# Create
				connection = self.create_ipc_connection()
				connection.open()

				# Bind
				#replyid = 'temp-agentctl-%s' % (os.getpid())
				replyid = AGENT_TEMP_REPLY_QUEUE( 'snmp-trapper-%s' % (os.getpid()))
				connection.bind(replyid)

				# Send/Receive
				config_from_server = self._send_request_and_receive(connection, 'get_agent_config', tenant_name, agent_name, timeout=self.load_config_timeout)

				# Close
				connection.close()

				self._update_settings_from_config(config_from_server)

				break

			except RemoteServerException, er:
				self.logger.critical("Server error. Unable to load settings from server: %s", er)
			except Exception, er:
				self.logger.critical("Unknow error while loading settings from server: %s", er)

			# Something gone wrong. Wait and retry
			self.logger.critical("Wait %s seconds before restart", self.load_config_retry_delay)
			time.sleep(self.load_config_retry_delay)

		return True

	def _send_request_and_receive(self, connection, rtype, *params, **kwargs):

		DEFAULT_TIMEOUT = 60
		timeout = kwargs.get('timeout', DEFAULT_TIMEOUT)

		mid = str(get_time())
		msg = encode_request_with_id(rtype, mid, *params)
		connection.send_msg(SANETD_CLT_QUEUE, msg, timeout=timeout)

		(addr, msg) = connection.receive_msg(ack_msg=True, timeout=timeout)

		if (msg == None):
			self.logger.info("chiedo %s %s", addr, msg)
			raise Exception("Timeout while loading configuration for %s" % rtype)

		response = decode_response(msg)

		if response['id'] != mid:
			raise RemoteServerException("Waiting for response with id %s. Received %s" % (mid, response['id']))

		if response['status'] == RESPONSE_ERROR:
			raise RemoteServerException("Unable to get elements from server: %s" % ( response['content'] ) )

		return response['content']

	def _update_settings_from_config(self, config_from_server):

		self.logger.info("Config from server: %s", config_from_server)

		# Set all global settings
		self.tenant_id            = config_from_server['tenantid']
		self.agent_id             = config_from_server['agentid']

	def is_stopping(self):
		return self._stop

	def handle_traps(self):

		self.logger.info("Starting trap receiver thread")

		traps_queue = Queue.Queue(-1)

		self.logger.debug("Launching UDP listen thread")
		trapd_thread = TrapReceiverThread(self, logger, traps_queue)
		trapd_thread.start()


		self.logger.debug("Starting main loop")
		while True:

			if self.is_stopping():
				break

			try:
				self.logger.info("Creating connection with server")
				self.connection = self.create_ipc_connection()
				self.connection.open()

				while 1:

					try:
						self.logger.debug("Waiting for new traps")
						trap_data = traps_queue.get(True, 1) # block waiting

						self.logger.debug("Trap data received")
						json_data = json.dumps(trap_data)

						self.logger.info("Sending trap data to server")
						self.connection.send_msg( SNMPTRAP_DATA_QUEUE, json_data, blocking=False)

					except Queue.Empty, er:
						self.logger.debug("No traps")
						continue

				self.connection.ack_pending()
				self.connection.close()

			except IPCConnectionError, e:
				self.logger.error("Connection lost")
			finally:
				try:
					if self.connection:
						self.connection.close()
				except Exception, e:
					pass

		self.logger.info("Exit")


	def create_ipc_connection(self):

		# Rispetto alla versione base, per creare una connessione usiamo
		# i parametri letti dal file di configurazione

		self.broker_address = (self.settings['server_host']          , self.settings.get('server_port'    , None) )
		self.broker_auth    = (self.settings.get('server_user', None), self.settings.get('server_password', None) )

		connect_options = {}

		self.logger.debug("Broker address: %s", self.broker_address )

		host = self.broker_address[0]
		if len(self.broker_address) > 1: 
			try:
				if self.broker_address[1]:
					connect_options['port']     = int(self.broker_address[1])
			except Exception, e:
				pass

		if self.broker_auth:
			connect_options['username'] = self.broker_auth[0]
			if len(self.broker_auth) > 1:
				connect_options['password'] = self.broker_auth[1]

		self.logger.debug("Opening connection with server %s", host )
		self.logger.debug("Options:")
		for k,v in connect_options.items():
			self.logger.debug("\t%s = %s", k, v)

		c = IPCConnection( host, **connect_options)
				
		self.logger.debug("Connection created!")

		return c

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def init_logger(agent_id, log_file, log_level, logtostdout=False):	

	global logger

	identifier = '%s' % (agent_id)

	format_str = '%(asctime)s %(levelname)s trapper/' + identifier + ' > %(message)s'

	logger = logging.getLogger('poller-%s' % (agent_id))
	logger.setLevel( log_level )

	hdlr = logging.handlers.WatchedFileHandler(log_file, "a")
	hdlr.setFormatter( logging.Formatter(format_str) )
	hdlr.setLevel( log_level )
	logger.addHandler(hdlr)

	if logtostdout:
		hdlr = logging.StreamHandler()
		hdlr.setFormatter( logging.Formatter(format_str) )
		hdlr.setLevel( log_level )
		logger.addHandler(hdlr)	

	return logger

def launch_remote_trapper_from_settings(settings, logtostdout=True):

	tenantid = settings['tenant'],
	agentid  = settings['agent'],
                 
	logfile  = settings['log_file']
	loglevel = settings['snmptrap_log_level']

	#
	# Init log
	#
	logger = init_logger(agentid, logfile, loglevel, logtostdout=logtostdout )

	logger.info("*****************************************************************")
	logger.info("************************* new start *****************************")
	logger.info("*****************************************************************")

	logger.debug('Using settings:')
	for k, v in sorted(settings.items()):
		logger.debug('    %s=%s', k, v)

	#
	# Launch agent
	#
	trapperagent = RemoteTrapperAgent(logger, settings)
	trapperagent.launch()

	# Never reach this
	return

