# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import sys
import time
import string
import re
import os
from operator import attrgetter
import traceback

import chardet

import readline

from cli.clilib.lex import UnescapeString
from cli.clilib.lex import LexException, lex_split, lex_join

from sanet_common.termutils import term_bold, term_color, TERM_COLOR_GREEN, TERM_COLOR_YELLOW

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class FatalCliException(Exception):
	pass

class BaseCliException(Exception):
	pass

class CliParsingException(BaseCliException):
	pass

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def JOIN_LIST(wlist):
	return ' '.join(wlist)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#


def CLI_KEYWORD(s):
	return UnescapeString(term_bold(s))

def CLI_COMMENT(s):
	return UnescapeString(term_color(TERM_COLOR_GREEN, s))

def CLI_WARNING(s):
	return UnescapeString(term_color(TERM_COLOR_YELLOW, s))

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class HelpingString(unicode):

	help = ''

	def __new__(cls, word, help=''):
		self = super(HelpingString, cls).__new__(cls, word)
		self.help = help
		return self

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class Token(object):

	token_repr   = '<TOKEN>'

	token_regexp = None     # Stringa con espressione regolare per fare match puramente sintattico.

	expand_help = False

	def __init__(self, *args, **kwargs):

		self.callback      = None

		self.callback_help = None

		self.subtree       = []
		self.help          = ''
		self.pipe          = 'DEFAULT'

		for k,v in kwargs.items():
			if not hasattr(self, k):
				raise Exception("Invalid argument %s" % (k))
			getattr(self,k) # We throw exception if setting an unknown attribute
			setattr(self,k, v)

		if self.callback is not None:
			
	 		if self.pipe == 'DEFAULT':
				self.pipe = default_pipe

			#if isinstance(self.pipe, T_Pipe) and not self.pipe in self.get_subtree():
			#	self.get_subtree().append(self.pipe)

	def __repr__(self):
		return "<%s>" % (self.__class__.__name__)

	def matches_syntax(self, cli, word):
		"""
		Questo metodo verifica puramente sintatticamente con la stringa *word*.
		
		La verifica viene effettuata se e' stato specificato una "token_regexp" in fase di inizializzazione dell'ogetto.
		"""
		if self.token_regexp is not None:
			return True if re.match(self.token_regexp, word) else False

		return True

	def matches(self, cli, word):
		"""
		Questo metodo verifica che il token faccia match "semanticamente" con la stringa *word*.
		"""
		return False

	def canonical(self, cli, word):
		"""
		Restituisce la versione "canonica" della stringa *word*
		"""
		return HelpingString(word, self.help)

	def canonical_syntax(self, cli, word):
		"""
		Restituisce la versione "canonica" della stringa *word* in caso di controllo puraente sintattico.
		"""
		#
		# FIX: Questo metodo e' un hack. Invece di passare a canonical() un parametro in piu' (e riscrivere tutti i
		#      metodi canonical() in giro per sanet, ne abbiamo definito uno nuovo che fa la stessa cosa,
		#      ma che deve essere chiamato solo durnate il parsing sintattico.
		return word

	def completion(self, cli, text):
		"""
		Restituisce un elenco di stringhe riconosciute dal token che potrebbero completare la stringa *text*.
		"""
		return []

	def get_subtree(self):
		subtree = [ t for t in self.subtree ]

		if isinstance(self.pipe, T_Pipe) and not self.pipe in subtree:
			subtree.append(self.pipe)

		return subtree

	ordering = 0

	def __cmp__(self, other):
		assert isinstance(other, Token)
		if self.__class__ == other.__class__:
			return self.cmp_sameclass(other)
		else:
			return cmp(self.ordering, other.ordering)

	def cmp_sameclass(self, other):
		return cmp(self, other)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def SORTED_KEYWORDS( keywords ):  
	return sorted(keywords, key=attrgetter('token_repr'))


class T_Keyword(Token):

	def __init__(self, kw, **kwargs):
		super(T_Keyword, self).__init__(**kwargs)
		self.kw         = kw
		self.token_repr = kw

	def __repr__(self):
		return "<%s:%s>" % (self.__class__.__name__, self.kw)

	#def __str__(self):
	#	return "<T_Keyword:%s>"%(self.kw,)

	def __unicode__(self):
		return self.kw

	def matches_syntax(self, cli, word):
		#
		# La verfica sintattica di una keword richiede che "word" faccia match
		# esattamente con la keyword (self.kw)
		#
		# HACK: per permettere che di default venga fatto l'autocompletamento 
		#       e' necessario che la verifica sintattica di un token sia valida
		#       anche parzialmente. Quindi usiamo startswith().
		return self.kw.startswith(word)

	def matches(self, cli, word):
		#
		# Per le keyword in generale la verifica "semantica" e' identica a quella sintattica 
		#
		return self.kw.startswith(word)

	def canonical_syntax(self, cli, word):
		return self.kw

	def canonical(self, cli, word):
		return HelpingString(self.kw, self.help)

	def completion(self, cli, text):
		if self.kw.startswith(text):
			return [HelpingString(self.kw, self.help)]
		else:
			return []

	def cmp_sameclass(self, other):
		return cmp(self.kw, other.kw)

	def change_kw(self, newkw):
		self.kw = newkw
		self.token_repr = newkw

	ordering = 1

BLANK_ENTRY = T_Keyword('', token_repr='<**CLI-EMPTY-TO**>', help='', pipe=None)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class T_Pipe(T_Keyword):

	@staticmethod
	def real_callback(c):
		callback = lambda a: None
		i = len(c.matched_tokens) - 1
		while i >= 0:
			if isinstance(c.matched_tokens[i], T_Pipe):
				callback=c.matched_tokens[i-1].callback
				break
			i = i - 1
		# Uh?
		c.prepipe_matched_words = c.matched_words
		c.prepipe_matched_tokens = c.matched_tokens
		c.prepipe_canonical_words = c.canonical_words
		c.matched_words=c.matched_words[:i]
		c.matched_tokens=c.matched_tokens[:i]
		c.canonical_words=c.canonical_words[:i]
		c.debug("Using callback from %d %s with %r" % (i-1, str(c.matched_tokens[i-1]), c.canonical_words))
		return callback(c)

	ordering = 1000

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class T_Number(Token):
	def __init__(self, *args, **kwargs):
		super(T_Number, self).__init__(*args, **kwargs)
		self.token_repr = '<NUM>'

	def matches(self, cli, word):
		for c in word:
			if not c in string.digits:
				return False
		return True

	ordering = 2

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class T_Float(Token):

	def __init__(self, *args, **kwargs):
		super(T_Float, self).__init__(*args, **kwargs)
		self.token_repr = '<float>'

	def matches(self, cli, word):
		try:
			float(word)
			return True
		except ValueError, e:
			return False

	ordering = 2

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class T_PositiveInteger(Token):

	def __init__(self, *args, **kwargs):
		super(T_PositiveInteger, self).__init__(*args, **kwargs)
		self.token_repr = '<num>'

	def matches_syntax(self, cli, word):
		try:
			if int(word) >= 0:
				return True
			else:
				return False
		except ValueError, e:
			return False

	def matches(self, cli, word):
		try:
			if int(word) >= 0:
				return True
			else:
				return False
		except ValueError, e:
			return False

	ordering = 2

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class T_Line(Token):

	def __init__(self, *args, **kwargs):
		super(T_Line, self).__init__(*args, **kwargs)
		self.token_repr = '<LINE>'

	def matches(self, cli, word):
		return True

	ordering = 3

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class T_Anyword(Token):

	def __init__(self, *args, **kwargs):
		super(T_Anyword, self).__init__(*args, **kwargs)
		if "token_repr" not in kwargs:
			self.token_repr = '<WORD>'

	def matches(self, cli, word):
		return True

	ordering = 4

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class T_ValidWord(Token):

	def __init__(self, valid_re, *args, **kwargs):

		self.token_repr = '<WORD>'
		self.valid_re = valid_re

		super(T_ValidWord, self).__init__(*args, **kwargs)
		
	def matches_syntax(self, cli, word):
		"""
		Semplice controllo sintattico identico.
		"""
		return self.valid_re.match(word)

	def matches(self, cli, word):
		return self.valid_re.match(word)

	ordering = 4

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class T_Alternatives(T_Anyword):

	expand_help=True

	def __init__(self, alternatives, *args, **kwargs):
		super(T_Alternatives, self).__init__(*args, **kwargs)

		self.alternatives = sorted(alternatives)

		if len(self.token_repr) > 8:
			self.token_repr = '<ALT>'
		else:
			self.token_repr = '|'.join([ a[0] for a in self.alternatives] ) #map(lambda x:x[0], self.alternatives))

	def matches_syntax(self, cli, word):
		# Sintatticamente non abbiamo le regexp delle "alternative", ma ci basta che 
		# una delle alternative faccia match (come sotto-stringa)
		for i in map(lambda x:x[0], self.alternatives):
			if word is not None and i.startswith(word):
				return True
		return False
		
	def matches(self, cli, word):
		# Se anche solo una delle stringhe tra le alternative matcha (anche una sottostringa di queste) va bene.
		for i in map(lambda x:x[0], self.alternatives):
			if word is not None and i.startswith(word):
				return True
		return False

	def canonical(self, cli, word):
		rv = []
		for i in self.alternatives:
			if word is not None and i[0].startswith(word):
				rv.append(i[0])
		
		if len(rv) > 0:
			return rv[0]

		return None

		if len(rv) == 1:
			return rv[0]
		elif len(rv) >1:
			for i in rv:
				if i == word:
					return i
		return super(T_Alternatives,self).canonical(cli, word)

	def completion(self, cli, text):
		try:
			rv = []
			for i in self.alternatives:
				if text is not None and i[0].startswith(text):
					rv.append(HelpingString(i[0], i[1]))
			return rv
		except Exception, e:
			return [str(e), '']

	def cmp_sameclass(self, other):
		return 0

	ordering = 5

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class T_Boolean(Token):

	vals = ['true', 'false']

	def __init__(self, *args, **kwargs):
		super(T_Boolean, self).__init__(*args, **kwargs)
		self.token_repr = '<BOOL>'

	def matches(self, cli, word):
		for i in self.vals:
			if i.startswith(word):
				return True
		return False

	def canonical(self, cli, word):
		rv = []
		for v in self.vals:
			if v.startswith(word):
				rv.append(v)
		if len(rv) == 1:
			return rv[0]
		elif len(rv) > 1:
			for i in rv:
				if i == word:
					return i
		return super(T_Boolean, self).canonical(cli, word)

	def completion(self, cli, text):
		try:
			rv = []
			for i in self.vals:
				if i.startswith(text):
					rv.append(HelpingString(i, i))
			return rv
		except Exception, e:
			return [str(e), '']

	ordering = 2

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def include_cbk(c):
	
	INCLUDE_KEYWORDS = [ 'include', 'match', 'grep' ]

	try:
		compiled = re.compile(c.canonical_words[-1])
	except Exception, e:
		c.outln("Error: invalid regular expression %s (%s: %s)" % (c.canonical_words[-1], e.__class__.__name__, e))
		return

	#if c.canonical_words[-2] == 'include' or c.canonical_words[-2] == 'match':
	if c.canonical_words[-2] in INCLUDE_KEYWORDS:
		c.match_re = compiled
		T_Pipe.real_callback(c)
		c.match_re = None
	elif c.canonical_words[-2] == 'exclude':
		c.exclude_re = compiled
		T_Pipe.real_callback(c)
		c.exclude_re = None
	else:
		c.outln("Unimplemented")

def begin_cbk(c):
	try:
		compiled = re.compile(c.canonical_words[-1])
	except Exception, e:
		c.outln("Error: invalid regular expression %s (%s: %s)" % (c.canonical_words[-1], e.__class__.__name__, e))
	c.begin_re      = compiled
	c.begin_matched = False
	T_Pipe.real_callback(c)
	c.begin_re      = None
	c.begin_matched = False
	
def end_cbk(c):
	try:
		compiled = re.compile(c.canonical_words[-1])
	except Exception, e:
		c.outln("Error: invalid regular expression %s (%s: %s)" % (c.canonical_words[-1], e.__class__.__name__, e))
	c.end_re      = compiled
	c.end_matched = False
	T_Pipe.real_callback(c)
	c.end_re      = None
	c.end_matched = False
	
def count_cbk(c):
	c.count_lines   = True
	c.lines_counted = 0
	T_Pipe.real_callback(c)
	c.out('%d lines of output\n' % c.lines_counted)
	c.count_lines   = False
	c.lines_counted = 0

def last_cbk(c):
	n = 40
	if isinstance(c.matched_tokens[-1], T_Number):
		n=int(c.canonical_words[-1])
	c.last    = n
	c.lastbuf = []
	T_Pipe.real_callback(c)
	c.last = 0
	c.out('\n'.join(c.lastbuf) + '\n')
	c.lastbuf = []

def debug_cbk(c):
	olda                = c.debugging
	oldi                = c.debug_interpreter
	c.debugging         = True
	c.debug_interpreter = True
	T_Pipe.real_callback(c)
	c.debugging         = olda
	c.debug_interpreter = oldi

def elapsed_cbk(c):
	start   = time.time()
	T_Pipe.real_callback(c)
	elapsed = time.time() - start
	c.out('\nElapsed: %f s\n' % elapsed)


default_pipe = T_Pipe('|', help='Pipe output through filter',
	subtree=[
		T_Keyword('begin', help='Display only output after a line matching a regex',
			subtree=[
				T_Anyword(help='Regexp to match',
					callback=begin_cbk,
					pipe=None,
				),
			],
		),
		T_Keyword('count', help='Display only the number of lines in output',
			callback=count_cbk,
			pipe=None,
		),
		T_Keyword('display', help='Display additional information',
			subtree=[
				T_Keyword('debug', help='Display debugging information when available',
					callback=debug_cbk,
					pipe=None,
				),
				T_Keyword('elapsed', help='Display time elapsed in the operation',
					callback=elapsed_cbk,
					pipe=None,
				),
			]
		),
		T_Keyword('end', help='Display only output before and excluding a line matching a regex',
			subtree=[
				T_Anyword(help='Regexp to match',
					callback=end_cbk,
					pipe=None,
				),
			],
		),
		T_Keyword('exclude', help='Display only output not matching a regex',
			subtree=[
				T_Anyword(help='Regexp not to match',
					callback=include_cbk,
					pipe=None,
				),
			],
		),
		T_Keyword('last', help='Display only a number of lines at the end of output',
			callback=last_cbk,
			pipe=None,
			subtree=[
				T_Number(help='Number of lines, default 40',
					callback=last_cbk,
					pipe=None,
				),
			],
		),
		T_Alternatives([
			('include', 'Display only output matching a regex'),
			('match'  , 'Display only output matching a regex'),
			('grep'   , 'Display only output matching a regex'),
		], help='Display only output matching a regex',
			subtree=[
				T_Anyword(help='Regexp to match',
					callback=include_cbk,
					pipe=None,
				),
			],
		),
		
		
		#T_Keyword('include', help='Display only output matching a regex',
		#	subtree=[
		#		T_Anyword(help='Regexp to match',
		#			callback=include_cbk,
		#			pipe=None,
		#		),
		#	],
		#),
		#T_Keyword('match', help='Display only output matching a regex (same as include)',
		#	subtree=[
		#		T_Anyword(help='Regexp to match',
		#			callback=include_cbk,
		#			pipe=None,
		#		),
		#	],
		#),
		
		T_Keyword('tail', help='Display only a number of lines at the end of output (same as last)',
			callback=last_cbk,
			pipe=None,
			subtree=[
				T_Number(help='Number of lines, default 40',
					callback=last_cbk,
					pipe=None,
				),
			],
		),
	],
)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class CliEnv(dict):

	def __init__(self, d):
		super(CliEnv, self).__init__()

		self.update(d)

		self.level = 0

	def __str__(self):
		try:
			s = ''
			for k in sorted(self.keys()):
				# IMPORTANT: avoid all context's controls on keys!
				v = super(CliEnv, self).__getitem__(k)
				s = s + "%s = %s (%s)" % (k, v,type(v)) + "\n"
		except:
			pass
		return s

	def clone(self):
		new_dict = dict(self)

		new_env = self.__class__(new_dict)

		new_env.level = self.level + 1

		return new_env

	def get_prompt(self):
		return ""

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class Cli(object):

	DEFAULT_EXIT_WORDS = ['exit', 'quit', 'logout']

	def __init__(self, tree, env, *args, **kwargs):

		self.is_input_tty = sys.stdin.isatty()

		self._debug_completer_counter = 0

		self.tree   = tree
		self.env    = env

		self.indent_spaces = ' ' * 4
		self.commentchars = '#!'
		
		self.exitwords = [] 

		self.use_rawinput = True

		self.debug_interpreter = False
		self.debugging = False

		self.exception_cleanup = self.reset_parser

		self.current_words   = []
		self.unmatched_words = []
		self.matched_words   = []

		self.matched_tokens = []
		#self.matched_token = None

		self.canonical_words = []

		self.match_re      = None
		self.exclude_re    = None
		self.begin_re      = None
		self.begin_matched = False
		self.end_re        = None
		self.end_matched   = False
		self.count_lines   = False
		self.lines_counted = 0
		self.last          = 0
		self.lastbuf       = []

		self.precmd  = None
		self.postcmd = None

		#self.context = None

		self.set_exit_words(self.DEFAULT_EXIT_WORDS)

		for k,v in kwargs.items():
			getattr(self,k) # So we throw exception if setting an unknown attribute
			setattr(self,k, v)

		#self._isatty = True if sys.stdout.isatty() else False


	#----------------------------------------------------------------------#

	def clean(self):

		self.matched_words   = []
		self.canonical_words = []

		self.matched_tokens = []
		#self.matched_token  = None

		self.unmatched_words = self.current_words

		self.match_re      = None
		self.exclude_re    = None
		self.begin_re      = None
		self.begin_matched = False
		self.end_re        = None
		self.end_matched   = False
		self.count_lines   = False
		self.lines_counted = 0
		self.last          = 0
		self.lastbuf       = []

	#----------------------------------------------------------------------#

	def set_exit_words(self, words):
		self.exitwords = words

	def get_exit_words(self):
		return self.exitwords

	#----------------------------------------------------------------------#

	def CLI_KEYWORD(self, s):
		if self.env.colors_enabled():
			return CLI_KEYWORD(s)
		return s

	def CLI_COMMENT(self, s):
		if self.env.colors_enabled():
			return CLI_COMMENT(s)
		return s

	def CLI_WARNING(self, s):
		if self.env.colors_enabled():
			return CLI_WARNING(s)
		return s

	#----------------------------------------------------------------------#

	def out(self,s):
		sys.stdout.write(s.encode('utf-8'))

	def outln(self,s):
		self.out(s + '\n')

	def comment(self, s):
		sys.stderr.write(s.encode('utf-8') + '\n')

	def echo(self, s):
		sys.stderr.write(s.encode('utf-8') + '\n')

	def info(self, s):
		sys.stderr.write(s.encode('utf-8') + '\n')

	def warning(self, s):
		sys.stderr.write("WARNING: " + s.encode('utf-8') + '\n')

	def error(self, s):
		sys.stderr.write("ERROR:" + s.encode('utf-8') + '\n')

	def debug(self, s):

		if self.debug_interpreter:

			if isinstance(s, unicode):
				s = s.encode('utf-8')

			s = '\n'.join([ '***DEBUG***> %s' % (s) for s in s.split('\n')])

			f = open("./ioslike_debug.log", "a+")
			f.write(s+'\n')
			f.close()

			sys.stderr.write( s + '\n')

	#----------------------------------------------------------------------#

	def prnt(self,s):
		
		if self.match_re is not None:
			ar=[]
			for l in s.split('\n'):
				if self.match_re.search(l):
					ar.append(l)
			if len(ar) > 0:
				self.out('\n'.join(ar) + '\n')

		elif self.exclude_re is not None:
			ar=[]
			for l in s.split('\n'):
				if not self.exclude_re.search(l):
					ar.append(l)
			if len(ar) > 0:
				self.out('\n'.join(ar) + '\n')

		elif self.begin_re is not None:
			ar=[]
			for l in s.split('\n'):
				if self.begin_matched:
					ar.append(l)
				elif self.begin_re.search(l):
					ar.append(l)
					self.begin_matched=True
			if len(ar) > 0:
				self.out('\n'.join(ar) + '\n')

		elif self.end_re is not None:
			ar=[]
			for l in s.split('\n'):
				if self.end_matched:
					continue
				elif self.end_re.search(l):
					self.end_matched=True
				else:
					ar.append(l)
			if len(ar) > 0:
				self.out('\n'.join(ar) + '\n')
		elif self.count_lines:
			for l in s.split('\n'):
				self.lines_counted = self.lines_counted + 1
				#self.out(l+'\n')
		elif self.last:
			for l in s.split('\n'):
				self.lastbuf.append(l)
				if len(self.lastbuf) > self.last:
					self.lastbuf = self.lastbuf[1:]

		else:
			self.out(s+'\n')

	def piqj(self, n, *vals):
		"""(p)rint (i)ndented (q)uoted (j)oined
		"""
		vals2 = []
		for v in vals:
			if isinstance(v, UnescapeString):
				v = v
			else:
				if isinstance(v, unicode):
					v = v.encode('UTF-8')
			vals2.append(v)
		vals = vals2
		
		
		self.prnt(self.indent_spaces * n +  lex_join(vals))

	def pij(self, n, *vals):
		vals2 = []
		for v in vals:
			if isinstance(v, unicode):
				v = v.encode('UTF-8')
			vals2.append(v)
		vals = vals2
		
		self.prnt(self.indent_spaces * n +  ' '.join(vals))

	#----------------------------------------------------------------------#

	def reset_parser(self):

		self.config_completer()

		self.current_words = []

		self.clean()

	def config_completer(self):

		d = readline.get_completer_delims()
		d = d.replace('-','')
		d = d.replace(':','')
		d = d.replace(';','')
		d = d.replace('|','')
		d = d.replace('/','')
		readline.set_completer_delims(d)

		readline.set_completer(self.completer_handler)

		# Complete with tabs
		readline.parse_and_bind("tab: complete")

		# Completition inside a token do not insert text inside it
		#readline.parse_and_bind("set skip-completed-text on")


	def completer_handler(self, text, state):

		try:
			self.debug("----------- COMPLETER START %s %s -------------" % (repr(text), state))

			self._debug_completer_counter += 1
			self.debug("COUNTER: %s" % (self._debug_completer_counter))
			
			self.debug("begin: %s , end: %s" % ( readline.get_begidx(), readline.get_endidx()) )
			self.debug("completer called with: %r %d" % (text, state))

			# Parse what we have so far
			line = readline.get_line_buffer()
			self.debug("completer get_line_buffer: %r" % (line))

			ret  = self.completions(line)

			self.debug('completions returned %s, len %d, state %d' % (ret, len(ret), state))

			# If one only, return with space appended and quoted
			if len(ret) == 1 and state == 0:
				self.debug('\ndoing qjoin: %r' % (ret))
				retval = ret[0]
				self.debug('\ndone qjoin: %r' % (retval,))

			# If still in range, return nth element
			elif state < len(ret):
				retval =  ret[state]
				self.debug('done: %r\n' % (retval,))
			else:
				self.debug('done x: None')
				retval = None

			need_mangle = True

	 		#XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX 
	 		# Poiche' il token inserito dall'utente potrebbe essere
	 		# diverso da quello che ha fatto match (es. ci sono
	 		# degli escape) dobbiamo fare degli aggiustamenti
	 		# sulla stringa realmente restituita a readline
	 		# per effettuare il completamento
	 		
	 		if text != '': # stiamo cercando di completare qualcosa (no tab a vuoto)
				if retval and retval != "|": # se abbiamo trovato un possibile candidato bisogna fare gli aggiustamenti del caso
				
					tokens = self.split(line, get_raw_tokens=True)
					if tokens:
						self.debug("\ntokens: %r" % ( tokens ))

						s_tokens = [ t for t in tokens ]
						
						quoted     = s_tokens[-1].code == "QUOTED"
						last_token = s_tokens[-1].s
						
						self.debug('\nlast token: %r' % (last_token,))
						
						self.debug('\nretval1: %r' % (retval,))
						
						retval = retval + ''
						retval.encode('UTF-8')
						retval = lex_join([retval,])

						need_mangle = False
						
						if quoted:
							retval = retval[1:]
							if retval[-1] == '"':
								retval = retval[:-1]
						

						if text != last_token or quoted:
							num_chars_to_remove = len(last_token)
							retval = retval[ num_chars_to_remove: ]
							
						self.debug('\nretval2: %r' % (retval,))

						self.debug('\ntext   : %r' % (text,))
						if text != last_token or quoted:
							retval = text + retval

						self.debug('\nretval3: %r' % (retval,))

			self.debug("Complete end...")

			self.debug('done selection of retval: %r' % (retval,))

			self.debug("----------- COMPLETER END 1 %s %s %r-------------" % (repr(text), state, retval))

			if need_mangle:
				if retval:
					retval = retval + ''
					if retval != '|':
						retval.encode('UTF-8')
						retval = lex_join([retval,])

			if (len(ret) > 0):
				if retval:
					retval += ' '

			# If more than one option and at last of range, display help
			if (len(ret) > 1) and ((state + 1) == len(ret)):
				self.debug('doing help for: %s' % (ret,))
				self.display_help(ret)

				self.out(self.env.get_prompt() + ' ' + readline.get_line_buffer())
				readline.redisplay()

			if self.debug_interpreter:
				self.debug('completer returning: %r' % (retval,))
				self.debug("")
				readline.redisplay()

			self.debug("----------- COMPLETER END 2 %s %s %r-------------" % (repr(text), state, retval))

			return retval

		except Exception, e:
			self.debug('completer got exception: %s: %s' % (e, traceback.format_exc()))
			# Exceptions here are trapped by readline

	def completions(self, line, forhelp=False):

		try:
			self.current_words = self.split(line)
		except:
			return []

		# Completion of last word, when no trailing space and all words matched
		if len(line) and line[-1] != ' ':
			self.match(skiplast=True)
			#text = line.split(' ')[-1]
			text = self.split(line)[-1]
		else:
			self.match()
			if len(self.unmatched_words) == 1:
				text = self.unmatched_words[0]
			else:
				text = ''

		if len(self.unmatched_words) > 1:
			self.debug("completions returning because many unmatched words: %r" % (self.unmatched_words))
			return []

		tree = self.tree

		if len(self.matched_tokens) > 0:
			
			# If no subtree??
			if self.matched_tokens[-1].get_subtree():
				tree = self.matched_tokens[-1].get_subtree()
			else:
				self.debug("No more subtree to match")
				return []

		if self.debug_interpreter:
			self.debug("completions matched words: %r" % (self.matched_words))
			self.debug("completions unmatched words: %r" % (self.unmatched_words))
			self.debug("completions completing text: %r" % (text))
			td = ' '.join(map(lambda e: repr(e), tree))
			self.debug("completions searching subtree: %s" % td)

		# See what our subtree can return
		cando=[]

		for alt in tree:
			if forhelp and not alt.expand_help:
				#self.debug("TOKEN")
				morealts = [HelpingString(alt.token_repr, alt.help)]
			else:
				#self.debug("CALLING")
				morealts = alt.completion(self, text)

			if self.debug_interpreter:
				self.debug("completions looking completion of %r in alternative %r (%r), got %r" % (text, alt, alt, morealts))

			cando = cando + morealts

		if self.debug_interpreter:
			self.debug("completions found %r" % cando)

		# Filter what matches with the text provided, not really useful
		ret = []
		for alt in cando:
			if alt.startswith(text):
				ret.append(alt)

		if self.debug_interpreter:
			self.debug("completions filtered %r" % ret)

		if forhelp and len(self.matched_tokens) and (self.matched_tokens[-1].callback):
			if (self.matched_tokens[-1].callback):
				if self.matched_tokens[-1].callback_help:
					ret.append(HelpingString('<CR>', self.matched_tokens[-1].callback_help))
				else:
					ret.append(HelpingString('<CR>', ''))

		return ret


	#def split(self, line, allow_invalid_tokens=True, get_raw_tokens=False):
	def split(self, line, get_raw_tokens=False):

		#self.debug("Splitting line: %s %s %s" % (line, repr(line), type(line)) )

		if len(line) == 0:
			return []

		ret = []

		try:
			str_tokens = lex_split(line)

			#if not allow_invalid_tokens:
			#	for t in str_tokens:
			#		self.debug("%s %s" % ( t, t.is_wellformed() ))
			#		if not t.is_wellformed():
			#			self.out("Invalid input starting at: %s\n" % (t.pos))
			#			self.out("%s\n" % (line))
			#			self.out("%s^\n" % ( " " * t.pos ))
			#			return []
			
			if get_raw_tokens:
				ret = str_tokens
			else:
				ret = [ t.s for t in str_tokens ]

		except LexException, e:
			self.on_lex_error(e)

		#self.debug("Split returned:")
		#for s in ret:
		#	self.debug("\t%s (%s)" % (s, s))
		return ret

	def on_lex_error(self, e):
		col = e.args[1]
		#self.out("%s^\n" % ( " " * (len(self.env.get_prompt()) + col) ) )
		#self.out("Invalid input starting at column: %s\n" % (col))
		self.error("Invalid input starting at column: %s\n" % (col))


	def display_help(self, ret):

		if not ret:
			return

		self.out('\n')
		for i in ret:
			if hasattr(i, 'help'):
				self.out(' %-18s %s\n' % (i,i.help))
			else:
				self.out(' %-18s\n' % (i))
		self.out('\n')

	def load_history(self, fpath):
		try:
			readline.read_history_file(fpath)
		except IOError, e:
			pass

	def save_history(self, fpath):
		try:
			readline.write_history_file(fpath)
		except IOError, e:
			pass

	def get_history_items(self):

		try:
			# FIX: very inefficient. See readline doc for a better solution.
			r = []
			for i in xrange(1, readline.get_current_history_length()+1):
				s = readline.get_history_item(i)
				if s == None:
					continue
				if s[0] in self.commentchars:
					continue
				r.append(s)
			return r
		except Exception, e:
			#print e
			return []

	def read_next_input_line(self, show_prompt):
		return raw_input(self.env.get_prompt()) if show_prompt else raw_input()


	def cmdloop(self):#, argv=None, quit=False):


		loop = True

		while loop:

			try:

				loop = self.read_input_parse_perform()

			# a) Ctlr+D 
			# b) EOF on stdin
			except EOFError, e:
				self.on_eof()
				#print ""  # andiamo a capo per evitare brutti disallineamenti.
				break

			# Ctrl+C
			except KeyboardInterrupt:

				sys.stderr.write('\nInterrupted!\n')
				sys.stderr.flush()

				self.exception_cleanup()

			#except BaseException, e:
			except Exception, e:
				self.on_generic_exception(e)

	def on_eof(self):
		print ""
		return

	def read_input_parse_perform(self):
		"""
		Questo metodo gestisce:

			* la lettura dell'input
			* il parsing e l'esecuzione del comando associato.
			* il match
			* il perform del comando
		
		OUTPUT
			Restituisce False se la riga letta in input corrisponde ad un comando di exit (EXIT WORDS).
			Restituisce True altriment.
		
		ECCEZIONI
		
			TODO
		"""

		self.reset_parser()

		line = self.read_input_line()

		if line == '':
			return True

		if self.env.get('ECHO',False):
			self.echo( line )

		# solo caratteri di spaziatura?
		if len(line.lstrip()) == 0:
			return True

		# Stringa con un commento?
		if line.lstrip()[0] in self.commentchars:
			return True

		# Comando incompleto. Mostriamo l'help.
		if line.rstrip().endswith('?'):
			line = line[:-1]
			self.display_help(self.completions(line, forhelp=True))
			return True

		#self.debug("Splitting words...")
		self.current_words = self.split(line) #, allow_invalid_tokens=False)

		#self.debug("Current words: ")
		#for i in xrange(len(self.current_words)):
		#	s = self.current_words[i]
		#	self.debug("\t%s) %s , %r , %s, unicode?=%s" % (i, s, s, type(s), isinstance(s, unicode))) 

		self.clean()

		if len(self.current_words) == 1:
			if self.current_words[0] in self.get_exit_words():

				#if self.env.get('ECHO',False):
				#	self.outln( line_pre )

				#break
				return False

		self.match()

		self.perform()

		return True

	def read_input_line(self):
		"""
		Questo metodo restituisce la prossima riga dall'input.
		
		La funzione cerca capire l'encoding utilizzato per codificare la stringa in input e di convertire la stringa in caratteri UNICODE.
		
		In caso di EOF questa funzione deve sollevare EOF.
		"""
		line = self.read_next_input_line(show_prompt=self.is_input_tty)

		if line:
			encoding = chardet.detect(line)['encoding']

			self.debug("INPUT %s %s %s %s" % (encoding, line, repr(line), type(line)))
			line = line.decode(encoding)

			#
			# FIX: use this code?
			#
			#if sys.stdin.encoding:
			#	line = line.decode(sys.stdin.encoding)
			#else:
			#	encoding = chardet.detect(line)['encoding']
			#
			#	self.debug("INPUT %s %s: %s %s" % (encoding, type(line), line, repr(line)))
			#	line = line.decode(encoding)

			self.debug(" ->   %s %s %s" % (line, repr(line), type(line)))

		return line

	def on_generic_exception(self, e):
		#if not (self.debugging or self.debug_interpreter):
		#	
		#	sys.stderr.write(u'\nError: %s: %s\n' % (e.__class__.__name__, e))
		#	sys.stderr.flush()
		#	
		#	self.exception_cleanup()
		#else:
		#	print "Raising exception because debugging"
		#	raise e
		#sys.exit(-1)

		if self.debug_interpreter:
			print "Raising exception because debugging..."
			print traceback.format_exc()
			raise e
		
		self.error("Generic error: %s:%s -> %s" % (type(e), str(e), traceback.format_exc()) )
		self.exception_cleanup()

	def match(self, skiplast=False):
		"""
		Try to match current_words to our tree and shift words from unmatched to matched.
		"""

		self.debug("---MATCH---")
		self.debug("\tcurrent   words = %r" % (self.current_words) )
		self.debug("\tunmatched words = %r" % (self.unmatched_words) )
		self.debug("\tmatched   words = %r" % (self.matched_words) )
		self.debug("\tcanonical words = %r" % (self.canonical_words) )

		self.unmatched_words = []

		if len(self.current_words) == 0:
			return

		if skiplast:
			matchable_words = self.current_words[:-1]
		else:
			matchable_words = self.current_words

		self.debug("\tmachable words = %r" % (matchable_words) )

		subtree = self.tree

		self.clean()
		self.debug("Entering macth method, self.self.matched_words = %r" % (self.matched_words))


		for word in matchable_words:
			
			matched = []
			matched_canonical = []
			
			chosen  = None

			for i in xrange(len(subtree)):
				
				candidate = subtree[i]

				self.debug("Trying candidate %s: %r" % (i, candidate,) )

				if isinstance(candidate, T_Line):
					# Always match and consume all line
					self.debug("Line match: matched_words %r, unmatched_words %r" % (self.matched_words, self.unmatched_words))
					
					n = len(self.matched_words)
					word = ' '.join(self.unmatched_words)
					self.matched_words.append(word)
					self.canonical_words.append(word)
					self.unmatched_words = []
					
					self.matched_tokens.append(candidate)
					
					self.debug("Line match after %d words, word %r, matched_words %r, " % (n, word, self.matched_words))
					return

				#if candidate.matches(self, word):
				if self.test_token_match(candidate, word):

					#canonical_word = candidate.canonical(self, word)
					canonical_word = self.get_token_canonical(candidate, word)
					self.debug("\tMatched! Canonical word: %r %s" % (canonical_word, type(canonical_word)) )
					
					matched.append(candidate)
					matched_canonical.append(canonical_word)
					
					if canonical_word == word:
						chosen = len(matched) - 1
			
			if len(matched) == 0:
				break
				
			if len(matched) == 1:
				chosen = 0 # It is the index, not the token
				
			self.debug("matched           = %s" % (matched))
			self.debug("matched_canonical = %s" % (matched_canonical))
			self.debug("chosen            = %s" % (chosen))
			
			if chosen == None:
				break
				
			self.matched_words.append(word)

			#c = matched[chosen].canonical(self, word)
			c = matched_canonical[chosen]

			self.canonical_words.append(c + '') # Add '' to make it of expected type
			self.unmatched_words = self.unmatched_words[1:]
			
			self.matched_tokens.append(matched[chosen])
			
			# Pass to the next token
			subtree = matched[chosen].get_subtree()

		return

	def test_token_match(self, candidate, word):
		"""
		Metodo che verifica il match tra il token candidato "candiate" e la stringa "word".
		
		Restituisce true se il token fa match, falso altrimenti.
		"""
		return candidate.matches_syntax(self, word) and candidate.matches(self, word)

	def get_token_canonical(self, candidate, word):
		"""
		Restituisce la stringa canconica per il token *candidate* della stringa *word* 
		"""
		return candidate.canonical(self, word)



	def perform(self):

		self.debug("---PERFORM---")

		if len(self.unmatched_words) > 0:
			if len(self.matched_tokens):
				msg = "Unrecognized command after %s: %s" % (JOIN_LIST(self.canonical_words), JOIN_LIST(self.unmatched_words),)
			else:
				msg = "Unrecognized command: %s" % (JOIN_LIST(self.unmatched_words))
			self.on_unmatched_words(msg)
			return

		if len(self.matched_tokens) <= 0:
			return

		token = self.matched_tokens[-1]

		if token.callback == None:
			self.on_incomplete_command()
			return

		self.debug("Doing %r, pipe is %s, subtree is %r" % (token, str(token.pipe), token.get_subtree()))
			
		#if self.precmd: 
		#	self.debug("Calling precmd")
		#	self.precmd(self)
		
		self.do_pre_cmd()
		
		self.debug("Calling cmd")
		try:

			if self.env.get('ECHO',False):
				self.echo( ' '.join([ repr(w) for w in self.canonical_words ]) )

			#s = "PERFORM:" + ' '.join(self.canonical_words + ['->', token.callback.__name__, token.callback.__code__.co_filename] )
			#print s
			#self.debug( s )

			self.call_token_callback(token)

			self.after_token_callback()

		except Exception, e:
			self.debug("---PERFORM--- CALL CMD EXCEPTION ---- \n %s" % (traceback.format_exc()))
			self.on_perform_exception(e)


	def do_pre_cmd(self):
		return

	def do_post_cmd(self):
		if self.postcmd: 
			self.debug("Calling precmd")
			self.postcmd(self)

		return

	def on_unmatched_words(self, msg):
		"""
		Metodo chiamato quando alcune sotto parti della stringa letta in input non hanno fatto match con i token.
		
		La stringa *msg* contiene un messaggio di dettaglio sul problema di match.
		"""
		self.error(msg)

	def on_incomplete_command(self):
		self.error("Incomplete command: %r" % (' '.join(self.current_words)))

	#------------------------------------------------------------------------------#

	def call_token_callback(self, token):
		"""
		Esegue la callback del tocket "token"
		"""
		token.callback(self)

	def after_token_callback(self):
		"""
		Metodo eseguito dopo che e' stata eseguita la callback di un tocket.
		"""
		return

	def on_perform_exception(self, ex):
		"""
		Metodo chiamato quando la callback di un token solleva una qualunque eccezione.
		"""
		self.error(traceback.format_exc())

	def enter_subcli(self, cmds, env):
		"""
		Questo metodo deve essere utilizzato dalle callback dei tocket quando vogliono lanciare sotto-cli.

		Invoca l'esecuzione di una sotto-cli con comandi "cmds" e environemnt "env".
		
		Il metodo ritorna quando la sottocli termina la sua esecuzione (exit commands) o viene
		sollevata una eccezione non intercettata dal loop dalla sotto-cli.
		
		Le eccezioni non intercettate dalla sotto cli vengono restituite.
		"""

		new_cli = self.get_class()(cmds, env, debug_interpreter=env['DEBUG'])
		new_cli.cmdloop()

	def get_class(self):
		"""
		Quando in un punto della Cli devo creare una nuova sotto-CLI
		devo istanziare un nuovo oggetto di questa stessa classe.

		Con questo metodo tutto il codice della CLI non deve fare un
		import ad un modulo particolare.
		"""
		return self.__class__

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

if __name__ == '__main__':

	debug_interpreter = True
	if len(sys.argv) > 1 and sys.argv[1] == 'no-debug':
		debug_interpreter = False

	def cbk(c):
		c.prnt("cbk_called with %d tokens: %r (%r)" % (len(c.matched_tokens), ' '.join(c.matched_words), ' '.join(c.canonical_words)))
		#c.prnt("cbk_called with %s tokens: %r (%r)" % (c.matched_token, ' '.join(c.matched_words), ' '.join(c.canonical_words)))

	def compatible_cbk(wl):
		print "compatible_cbk called with: %r" % ' '.join(wl)

	def show_version(c):
		c.prnt("You are using the ioslike python library demo CLI.")

	def configure(c):
		Cli([], prompt='configure# ').cmdloop()

	test_prompt = []

	test_prompt.append(T_Keyword('show',
		help='Show running system information',
		subtree=[
			T_Keyword('version', help='Site information and software version', callback=show_version,
				pipe='DEFAULT',
			), 
			T_Keyword('vermouth', help='Sweet alcoholic beverage information', callback=cbk), 
			T_Keyword('configuration', help='Show configuration', callback=cbk,
				pipe='DEFAULT',
			),
			T_Alternatives(
				[
					['mon', 'Monday'],
					['tue', 'Tuesday'],
					['wed', 'Wednsday'],
					['thu', 'Thursday'],
					['fri', 'Friday'],
					['sat', 'Saturday'],
					['sun', 'Sunday'],
				],
				token_repr='<DOW>',
				help='Day of week', callback=cbk,
			),
		],
	))


	test_prompt.append(T_Keyword('shower', help='Have shower',
		pipe='DEFAULT',
		callback=cbk, 
	))

	test_prompt.append(T_Keyword('refresh', help='Refresh dynamic tables', callback=cbk, subtree=[
		T_Keyword('this', help='Refresh this', callback=cbk),
		T_Keyword('thing', help='Refresh many diverse things', callback=cbk),
		T_Keyword('that', help='Refresh that', callback=cbk),
	]))

	test_prompt.append(T_Keyword('configure',
		help='Configure system',
		subtree=[
			T_Keyword('terminal', help='Configure system from the terminal this', callback=configure),
			T_Keyword('web interface', help='Configure system from the terminal this', callback=configure),
		],
	))

	test_prompt.append(T_Keyword(u'Citt\u0155', callback=cbk,
		help='Stinga Unicode',
	))

	test_prompt.append(T_Keyword('test', subtree=[
		T_Keyword("number"       , subtree=[ T_Number(callback=cbk),                                ], ),
		T_Keyword("float"        , subtree=[ T_Float(callback=cbk),                                 ], ),
		T_Keyword("boolean"      , subtree=[ T_Boolean(callback=cbk),                               ], ),
		T_Keyword("integer"      , subtree=[ T_PositiveInteger(callback=cbk),                       ], ),
		T_Keyword("any"          , subtree=[ T_Anyword(callback=cbk),                               ], ),
		T_Keyword("valid"        , subtree=[ T_ValidWord(re.compile("abc"), callback=cbk),                      ], ),
		T_Keyword("alternatives" , subtree=[ T_Alternatives([("a","help A"), ("b", "help B")], callback=cbk), ], ),
		T_Keyword("line"         , subtree=[ T_Line(callback=cbk), ], ),
	]))



	env = CliEnv({'PROMPT':'cli-test# '})

	cli = Cli(test_prompt, env, debug_interpreter=debug_interpreter)
	cli.cmdloop()

