# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import sys

import codecs

import csv

from pyparsing import ParseException
from pyparsing import White
from pyparsing import StringEnd
from pyparsing import Literal
from pyparsing import Regex
#from pyparsing import OneOrMore
from pyparsing import Optional
from pyparsing import Or
from pyparsing import ZeroOrMore
#from pyparsing import Suppress
from pyparsing import FollowedBy
#from pyparsing import SkipTo

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

_codec = codecs.lookup('string_escape')
# The following is needed for compatibility with python 2.4 and 2.5
try:
	_cenc  = _codec.encode
	#  = _codec.decode
	_echar = None
except AttributeError:
	# 2.4
	_cenc  = _codec[0]
	#_cdec  = _codec[1]
	_echar = '\\'

#------------------------------------------------------------------------------#
# UNESCAPE                                                                     #
#------------------------------------------------------------------------------#

#class StringEscapeException(BaseException):
#	pass

class StringEscapeException(Exception):
	pass

def string_unescape(s, escape_spaces=True):

	#if isinstance(s, unicode):
	#	s = s.encode('utf-8')
		
	#if isinstance(s, unicode):
	#	s = s.encode('ascii')

	escaped=False
	
	isascii=False
	try:
		s.encode('ascii')
		isascii=True
	except Exception, e:
		pass
		
	NORMAL    = 1
	ESCAPE    = 2
	HEXESCAPE = 3
	
	state = NORMAL
	
	#s2=b''
	s2 = []
	
	binary = False
	
	i=0
	while i < len(s):
		
		if state == NORMAL:
			c = s[i]
			if isascii: c = c.encode('ascii')
			i += 1

			if c=="\\": 
				state = ESCAPE
				continue
			#s2 = s2 + c
			s2.append(c)
		elif state == ESCAPE:
			c = s[i]
			i += 1

			if   c=='\"': c = "\""
			elif c=='\'': c = "'"
			elif c=='\\': c = "\\"
			elif c=='x':
				binary = True
				state = HEXESCAPE
				continue
			elif escape_spaces:
				if   c == 'b': c = "\b"
				elif c == 't': c = "\t"
				elif c == 'n': c = "\n"
				elif c == 'f': c = "\f"
				elif c == 'r': c = "\r"
				elif c == ' ': c = " "
			elif not escape_spaces:
				if c in "btnfr ":
					raise StringEscapeException("Escaped spaces not allowed (pos: %s)" % (i))
			#s2 = s2 + c
			if isascii: c = c.encode('ascii')
			s2.append(c)
			state = NORMAL
		elif state == HEXESCAPE:
			x = s[i:i+2]
			i += 2

			if not isascii:
				raise Exception("Unable to unescape unicode strings")

			c = chr(int(x,16))
			s2.append(c)
			state = NORMAL
			
			escaped = True

	s2 = b''.join(s2)
	if escaped:
		s2 = s2.decode('UTF-8')

	return s2


def string_escape(s):
	
	if s.__class__ == unicode:
		s = s.encode('utf-8')
		
	# FIX: this trick have some drawbacks
	# 1) escapes single quotes (')
	# 2) Escapes double quotes (") with ("") and not with (\")
	# 3) Do not escapes "|" if there are spaces inside the string
	s = _cenc(s)[0] 
	
	
	# FIX(3)
	if s.find(' ') == -1:
		s = s.replace ("|", r"\|")

	# FIX (2)
	s = s.replace("\\'", "'")
	
	# FIX (2)
	s = s.replace("\\'", "'")
	
	# FIX (1)
	#s = s.replace('""', '\\"')
	s2 = ''
	for c in s:
		if c == '"':
			s2 += "\\"
		s2 += c

	return s2


#------------------------------------------------------------------------------#
# ELEMNTI DEL LESSICO                                                          #
#------------------------------------------------------------------------------#

class Tok(object):

	def __init__(self, s, pos):
		self.code = 'TOK'
		self.pos      = pos
		self.s        = s
		
		self.original = s

	def __repr__(self):
		
		s = self.s
		if isinstance(self.s, unicode):
			sv = s.encode('utf-8')
		else:
			sv = s
			
		o = self.original 
		#if isinstance(o, unicode):
		#	o = o.encode('utf-8')
			
		return "{%s,%s,%s(%s,%s),%r(%s,%s)}" % (
		        self.code,
			self.pos,
			sv,type(s),repr(s), 
			o,type(o),repr(o),
		)
		
	def is_wellformed(self):
		return True

class TokPipe(Tok):
	def __init__(self, s, pos):
		super(TokPipe, self).__init__(s, pos)
		self.code = 'PIPE'

class TokUnquoted(Tok):
	def __init__(self, s, pos):
		super(TokUnquoted, self).__init__(s, pos)
		self.s = string_unescape(s, escape_spaces=False)

class TokQuoted(Tok):
	def __init__(self, s, pos, closed_quotes=True):
		super(TokQuoted, self).__init__(s, pos)
		self.code = 'QUOTED'

		self.s = string_unescape(s, escape_spaces=True)

		self.original = '"'+s
		if closed_quotes:
			self.original += '"'

		self.closed = closed_quotes

	def is_wellformed(self):
		return self.closed


__DEBUG_PARSING = False

LITERAL = Literal
SPACE   = White().suppress()
EOL     = StringEnd().suppress()

#
# PIPE   
#
# |
#
def _pipe_action(original_string, string_pos, matched_tokens):
	s = matched_tokens[0]
	if __DEBUG_PARSING: print "PIPE(%s)" % (s)
	return TokPipe(s, string_pos)

TOKEN_PIPE = LITERAL("|").setParseAction(_pipe_action).setDebug(__DEBUG_PARSING)

# 
# UNQUOTED
#
# Qualunque sequenza di caratteri che:
# 1) non inizia per \
# 2) non contenga 
# 
#    <spazi>      \n \r \t \a
#    |
#    "
#
# I caratteri |," sono ammessi solo se preceduti da escape ( \| , \" )
# 


def _unquoted_action(original_string, string_pos, matched_tokens):
	s = matched_tokens[0]
	return TokUnquoted(s, string_pos)

re_unquoted     = r'([^"\|\\\s]+(?:\\[^ntra ][^"\|\\\s]*)*)'
TOKEN_UNQUOTED = Regex(re_unquoted).setParseAction(_unquoted_action).setDebug(__DEBUG_PARSING)

#
# QUOTED
#
# Sequenza di caratteri che
# 1) Inizia per "
# 2) Contiene una stringa che non contenga " o \  non escapate
# 3) Finisce con
#      "
#      EOF     per gestire le stringhe non complete
#
# "the quick\ brown\nfox jumps over the lazy dog\|"
#
#
def _quoted_action(original_string, string_pos, matched_tokens):

	if len(matched_tokens) == 3:  #  " foo "
		s = matched_tokens[1]
		return TokQuoted(s, string_pos, closed_quotes=True)

	elif len(matched_tokens) == 2: # " foo
		s = matched_tokens[1]
		return TokQuoted(s, string_pos, closed_quotes=False)
	else:
		ParseException("Unknown parsing error at pos: %s" % (string_pos))

re_quoted  = r'([^"\\]*(?:\\.[^"\\]*)*)'
TOKEN_QUOTED = ( \
                       LITERAL('"') + \
                       Optional(Regex(re_quoted)) + \
                       ( \
                        LITERAL('"') + FollowedBy( EOL | SPACE) \
                        | \
                        EOL
                       ) \
               ).setParseAction(_quoted_action).setDebug(__DEBUG_PARSING)
                

TOKENS        = TOKEN_PIPE | TOKEN_QUOTED | TOKEN_UNQUOTED

GRAMMAR = TOKENS + ZeroOrMore(SPACE + TOKENS) + EOL.suppress()


#class LexException(BaseException):
#	pass
class LexException(Exception):
	pass

def lex_split(s):
	
	try:
		tokens = GRAMMAR.parseString( s )
		return [ t for t in tokens ] 
	except ParseException, err:
		raise LexException("Invalid string starting position: %s" % (err.col), err.col)
	except Exception, e:
		print __file__, "lex_split()", type(e), e
		raise e


#------------------------------------------------------------------------------#
# JOIN                                                                         #
#------------------------------------------------------------------------------#

class UnescapeString(unicode):
	"""
	Questa classe "stringa" viene usata per identificare le stringhe
	che non devono essere soggette ad operazione di escape.
	
	Vedi funzione lex_join().
	"""
	def __new__(cls, word):
		self = super(UnescapeString, cls).__new__(cls, word)
		return self

def lex_join(strings):
	"""
	Concatena le stringhe specificate dal vettorey "string" in un unica stringa
	effettando l'escapring delle singole stringhe se necessario.
	
	Le stringhe in input di tipo UnescapeString non vengono escapate ma semplicemente
	concatenate al risultato finale.
	"""
	
	u8 = []
	
	for s in strings:
		
		# Questo tipo di stringhe non sono soggette ad escape.
		if isinstance(s, UnescapeString):
			u8.append(s)
			continue
		
		escaped = string_escape(s)
		
		if len(s) == 0:   # stringa vuota
			escaped = '"' + escaped + '"'
			
		elif len(escaped) != len(s):  #stringa con caratteri escaped
			escaped = '"' + escaped + '"'
			
		elif s.find(' ') > -1:  # stringa con spazi 
			escaped = '"' + escaped + '"'

		elif s[-1] == "?":  # stringa che finisce con '?' (es. show node nam?)
			escaped = '"' + escaped + '"'
			
		u8.append(escaped)

	return ' '.join(u8)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#



if __name__ == "__main__":
	
	def split_chars(s):
		l = []
		for c in s:
			o = ord(c)
			if 32 <= o and o <= 127:
				l.append(c)
			else:
				l.append(hex(o))
		return l

	def print_repr(s):
		print "========================================================"
		print split_chars(s)
		print "___",s,"___"

	s1 = raw_input()
	
	print "*************************** INPUT ****************************"
	print s1
	print split_chars(s1)


	print "*************************** TEST ****************************"
	
	
	"""
	#s2 = string_unescape(s1)
	s2 = string_unescape(s1)

	#s3 = ASCII_TO_UNICODE(s2)
	s3 = s2
	
	#s4 = UNICODE_TO_ASCII(s3)
	s4 = s3 
	s5 = string_escape(s4)
	
	
	print_repr(s1)
	print_repr(s2)
	print_repr(s3)
	print_repr(s4)
	print_repr(s5)
	"""
	
	print "*************************** PARSING ****************************"
	
	tokens = lex_split(s1)

	for t in tokens:
		print "%s" % (t)
	
	s_tokens = [ t.s for t in tokens if not isinstance(t, TokPipe) ]

	print "*************************** JOINING ****************************"


	sf = lex_join(s_tokens)

	print sf
	
	print "***************************** REVERSE TEST *****************************"

	se = lex_join( [t.s for t in lex_split(sf)] )

	print "ORIGIN: ", s1
	print "FINAL : ", se
	
	assert (se == sf)


