# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import time
import traceback
import getpass
import logging

from django.conf import settings
from django.utils.translation import ugettext as _

from sanet_common.pythonutils import load_symbol_from_module
from sanet_common.sanetdefs.constants import SOURCE_LIBRARY, SOURCE_LOCAL
from sanet_common.termutils import term_color, TERM_COLOR_RED, TERM_COLOR_YELLOW, TERM_COLOR_CYAN

from conf import get_configuration
DEBUG_CONFIG = get_configuration().DEBUG_CONFIG or os.environ.get('DEBUG_CONFIG',False)


from django.contrib.auth.models import User

from core.exceptions import ApplicationException

from core.models import get_user_default_tenant
from core.api.base import db_set_current_tenant_schema
from core.api.base import tenant_get_primary, tenant_get


from core.capabilities import CAPABILITIES_IDS

from cli.clilib.ioslike import Cli as BaseCli
from cli.clilib.ioslike import CliEnv as BaseCliEnv, SORTED_KEYWORDS
from cli.clilib.ioslike import FatalCliException

from cli.prompts import cli_get_registered_menus
from cli.prompts import cli_register_menu

from cli.prompts.common import common_commands
from cli.prompts.common import exec_commands
from cli.prompts.common import set_tenant_commands

from cli.prompts.show   import show_commands
from cli.prompts.show.library import show_commands_library

from cli.prompts.config import cmd_config_prompt
from cli.prompts.config import cmd_config_library_prompt
from cli.prompts.config import config_prompt, cmd_config_prompt, config_prompt_command
from cli.prompts.config.tenants import _conf_tenant

from sanet_common.termutils import term_color, TERM_COLOR_RED, TERM_COLOR_YELLOW, TERM_COLOR_CYAN


#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

ERROR_COLOR   = TERM_COLOR_RED

WARNING_COLOR = TERM_COLOR_YELLOW

INFO_COLOR    = TERM_COLOR_CYAN

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

_CLI_LOGGER = logging.getLogger('cli.logging')

class SanetCliEnv(BaseCliEnv):
	"""
	La Cli di Sanet (SanetDefaultCli) e' piu' sofisticata della CLI di base ed
	ha bisogno di un environment piu' sofisticato.
	
	Questa classe server per gestire l'environment usato da SanetDefaultCli.
	"""

	def clone(self):
		return super(SanetCliEnv, self).clone()

	def get_prompt(self):

		username = None

		# IMPORTANT: This function checks the environment variables:
		#
		#    LOGNAME
		#    USER
		#    LNAME
		#    USERNAME
		#  
		# If none are set, the login name from the password database is returned 
		# on systems which support the pwd module, otherwise, an exception is raised.

		user = self.get('CURRENT_USER', None)
		if user:
			username = user.username

		if username is None:
			username = getpass.getuser()

		s = 'cli:'

		if self['SOURCE'] == SOURCE_LIBRARY:
			s += '%s@library:' % (username)

			s = s + self.get('PROMPT', '')

			s = s + '#'
		else:
			if self.get('MODE', None) != 'configuration':
				s += '%s@%s' % (username, self['CURRENT_TENANT'])
				s = s + '>'
			else:
				if self.get('CONFIG_TENANT'):
					s += '%s@' % (username)
				else:
					s += '%s@global-config:' % (username)
				s = s + ''

				s = s + self.get('PROMPT', '')

				s = s + '#'

		if isinstance(s, unicode):
			s = s.encode('utf-8')

		return s

	def colors_enabled(self):
		return self.get('COLORS', False) # == 'yes'

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class SanetDefaultCli(BaseCli):
	"""
	Classe con funzioni specifiche per Sanet3.
	"""

	def get_class(self):
		"""
		Quando in un punto della Cli devo creare una nuova sotto-CLI
		devo istanziare un nuovo oggetto di questa stessa classe.

		Con questo metodo tutto il codice della CLI non deve fare un
		import ad un modulo particolare.
		"""
		return self.__class__

	def echo(self, s):
		
		s = "ECHO: " + s.encode('utf-8')
		
		if self.env.colors_enabled():
			s = term_color(INFO_COLOR, s)
		super(SanetDefaultCli, self).echo(s)

	def info(self, s):

		s = "INFO: " + s.encode('utf-8')

		if self.env.colors_enabled():
			s = term_color(INFO_COLOR, s)
		super(SanetDefaultCli, self).info(s)

	def error(self, s):
		s = "ERROR: " + s.encode('utf-8') + '\n'

		if self.env.colors_enabled():
			s = term_color(ERROR_COLOR, s)

		sys.stderr.write(s)

	def warning(self, s):
		s = "WARNING: " + s.encode('utf-8') + '\n'

		if self.env.colors_enabled():
			s = term_color(WARNING_COLOR, s)

		sys.stderr.write(s)

	def call_token_callback(self, token):
		"""
		Nella cli di Sanet3 l'esecuzione della callback di un token puo' sollevare diverti tipi di eccezioni.
		
			* FatalCliException    -> eccezione non gestita

			* Altre eccezioni
			   
			     Se la Cli viene invocata con -X (EXIT_ON_ERROR)
 			        -> trasformate in FatalCliException.
 			     Altrimenti
 			        -> messaggio di errore
			     
		"""

		try:
			super(SanetDefaultCli, self).call_token_callback(token)

		except ApplicationException, e:
			
			msg = str(e)

			# La cli e' stata invocata con l'opzione -X quindi qualunque errore applicativo genera una eccezione fatale.
			if self.env['EXIT_ON_ERROR']:
				raise FatalCliException(msg)
			
			self.error(msg )

		except Exception, e:

			# La cli e' stata invocata con l'opzione -X quindi alla minima eccezione lanciare.
			if isinstance(e, FatalCliException):
				raise e

			msg =  "%s" % str(e)

			if self.env['EXIT_ON_ERROR']:
				raise FatalCliException(msg)

			self.error(msg)

	def on_unmatched_words(self, msg):

		# La cli e' stata invocata con l'opzione -X quindi anche un semplice errore di match si trasforma in una eccezione fatale.
		if self.env['EXIT_ON_ERROR']:
			raise FatalCliException(msg)

		super(SanetDefaultCli, self).on_unmatched_words(msg)

	def on_incomplete_command(self):

		# La cli e' stata invocata con l'opzione -X quindi anche un semplice comando incompleto si trasforma in una eccezione fatale.
		if self.env['EXIT_ON_ERROR']:
			raise FatalCliException("Exception incomplete command!")

		super(SanetDefaultCli, self).on_incomplete_command()

	def on_generic_exception(self, ex):

		# La cli e' stata invocata con l'opzione -X quindi qualunque eccezione si trasforma in una eccezione fatale.

		if isinstance(ex, FatalCliException):
			raise ex

		if self.env['EXIT_ON_ERROR']:
			# trasformiamo l'eccezione in una eccezione fatale
			raise FatalCliException("Exception detected: %s:%s" % (type(ex), str(ex)))

		super(SanetDefaultCli, self).on_generic_exception(ex)

	def on_perform_exception(self, ex):

		# La cli e' stata invocata con l'opzione -X quindi alla minima eccezione lanciare.

		if not isinstance(ex, FatalCliException):
			raise ex

		if self.env['EXIT_ON_ERROR']:
			# trasformiamo l'eccezione in una eccezione fatale
			raise FatalCliException("Error while performing commands: %s" % (str(ex)))

		super(SanetDefaultCli, self).on_perform_exception(ex)

		return

	#
	# Metodi nuovi della CLI di Sanet3
	#

	#def set_current_tenant(self):
	#	# Credo che non sia usato da nessuna parte
	#	# se e' stato scritto per un eventuale uso futuro bisogna correggere cli.env con self.env
	#
	#	tenant_name = cli.env['CONFIG_TENANT']
	#	tenant = tenant_get(tenant_name)
	#	db_set_current_tenant_schema(tenant_name)
	#	return tenant

	def get_config_element(self):
		"""
		Quando dobbiamo modificare/aggiornare elementi del monitoraggio
		la cli deve rileggere sempre l'oggetto dal DB cosi' da lavorare
		su un oggetto sembre aggiornato.
		Cosi' facendo si evitano problemi di ConcurrentUpdate che
		in realta' non esistono.
		Questo metodo legge dall'environmente della CLI l'oggetto
		corrente e restituisce sempre un oggetto appena letto dal DB
		"""
		old_element = self.env['CONFIG_ELEMENT_DATA']
		model_class = old_element.__class__
		model_id    = old_element.uuid

		db_set_current_tenant_schema(old_element.tenant.name)

		element     = model_class.objects.get(uuid=model_id)

		return element

	def set_config_element(self, e):
		self.env['CONFIG_ELEMENT_DATA'] = e
		return


#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def create_history_file():

	fpath = os.sep.join(["%s" % os.environ['HOME'], ".sanet-cli-history"])

	if not os.path.exists(fpath):
		f = open(fpath, "w+")
		f.write('')
		f.close()

	return fpath

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def register_apps_cli_hooks():

	for django_app_name in settings.INSTALLED_APPS:
		try:
			cli_module = '%s.cli.prompts' % (django_app_name)
			module_loader = load_symbol_from_module( cli_module, 'register_cli_hooks')
			module_loader()
			if DEBUG_CONFIG: print "Registered app hooks %s" % (django_app_name)
		except SyntaxError, e:
			raise Exception("Error in module %s: %s" % ( django_app_name, e) )

		except Exception, e:
			if DEBUG_CONFIG: print "ERROR:", type(e), e

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def launch_sanet_cli(debug=False, library=False, config=False, tenant_name=None, echo_mode=False, colors_mode=False, cli_class=None, user=None, exit_on_error=False):

	"""
	Funzione principale per lanciare la CLI di SANET.
	"""

	#
	# Calculate user for the cli
	#
	if library and user:
		if not user.is_superuser:
			raise ApplicationException(_("User '%s' can not access to Library") % (user.username))

	# Nella configurazione globale (non dentro un tenant) non possono entrare utenti normali.
	if config and not tenant_name and user and not user.is_superuser:
			raise ApplicationException(_("User '%s' can not access global configuration") % (user.username))

	#
	# Prepare history
	#

	enter_global_config = True

	if library and not config:
		raise ApplicationException(_("You can not enter in the library without config mode on."))

	history_file_path = create_history_file()

	#
	# Calcoliamo il tenant di default
	#
	default_tenant = None
	if library == False:
		if tenant_name == None:
			if user is not None:
				default_tenant = get_user_default_tenant(user)
				if default_tenant is None:
					if user.is_superuser:
						default_tenant = tenant_get_primary()
				if default_tenant is None:
					raise ApplicationException(_("User '%s' not allowed to access to any tenant.") % (user.username))
			else:
				default_tenant = tenant_get_primary()
		else:
			enter_global_config = False
			default_tenant = tenant_get(tenant_name)
			if default_tenant == None:
				raise ApplicationException(_("Tenant '%s' not found") % (tenant_name,))
			if user:
				if not user.get_profile().can_view_tenant(default_tenant):
					raise ApplicationException(_("User '%s' not allowed to access to tenant '%s'") % (user.username, default_tenant))

	source = SOURCE_LIBRARY if library == True else SOURCE_LOCAL

	#
	# Prepariamo l'environment iniziale per la CLI
	#
	env = {
		'ECHO'                     : echo_mode, 
		'DEBUG'                    : debug,

		'EXIT_ON_ERROR'            : exit_on_error,

		'PROMPT'                   : '',
		'COLS'                     : None,

		'COLORS'                   : colors_mode and sys.stdout.isatty(), 

		'SOURCE'                   : source,

		'CURRENT_TENANT'           : default_tenant.name if default_tenant != None else None,

		# if authenticated CLI is disabled, the user is not selected (NONE) and the current user has all CAPABILITIES
		'CURRENT_USER'             : user if user is not None else None,
		'CURRENT_USER_CAPABILITIES': user.get_capabilities(default_tenant) if user is not None else CAPABILITIES_IDS,
	}
	env = SanetCliEnv(env)

	#
	# Carichiamo gli hook principali e quelli delle applicazioni
	# abilitate in INSTALLED_APPS
	#
	register_apps_cli_hooks()

	# La Cli da usare e' quella di default di Sanet.
	# FUTURE: sarebbe utile avere altri tipi di CLI per poter leggere da standard input
	#         comandi/file CLI ed elaborare i dati in maniera diversa.
	cli_class = cli_class or SanetDefaultCli

	#
	# Lanciamo la cli con il menu' iniziale in base alla modalita' selezionata
	# FIX: refactoriy di questo codice per renderlo piu' pulito.
	#
	if config:

		cli = cli_class( [] , env, debug_interpreter=env['DEBUG'] )

		cli.load_history(history_file_path)

		if sys.stdout.isatty():
			if user:
				cli.info( _("Using user '%s' as current user") % (user.username) )

		if library == True:

			if sys.stdout.isatty():
				cli.info( _("Entering LIBRARY") )
				cli.info( _('Type "help" for help'))

			cli_register_menu('__show_commands__'  , 'library'              , show_commands_library)
			cli_register_menu('__common_commands__', 'base_common_commands' , common_commands)

			cmd_config_library_prompt(cli)
		else:

			if sys.stdout.isatty():
				cli.info( _("Entering global configuration") )
				cli.info( _('Type "help" for help'))

			cli_register_menu('__show_commands__'  , 'base_show_commands'   , show_commands)
			cli_register_menu('__common_commands__', 'base_common_commands' , common_commands)
			cli_register_menu('__common_commands__', 'other_common_commands', exec_commands)
			cli_register_menu('common:set'         , 'tenant'               , set_tenant_commands)

			if enter_global_config:
				cmd_config_prompt(cli)
			else:
				_conf_tenant(cli, default_tenant.name)
	else:

		cli_register_menu('__show_commands__'  , 'base_show_commands'   , show_commands)
		cli_register_menu('__common_commands__', 'base_common_commands' , common_commands)
		cli_register_menu('__common_commands__', 'other_common_commands', exec_commands)
		cli_register_menu('common:set'         , 'tenant'               , set_tenant_commands)

		sanet_commands = SORTED_KEYWORDS(
			cli_get_registered_menus('__show_commands__')
			+
			cli_get_registered_menus('__common_commands__')
			+
			cli_get_registered_menus('__apps_commands__')
			+
			config_prompt_command
		)

		cli = cli_class( sanet_commands , env, debug_interpreter=env['DEBUG'] )

		cli.load_history(history_file_path)

		if sys.stdout.isatty():

			cli.info( _("Sanet v%s CLI") % (settings.SANET_VERSION_FULL_VERSION) )

			if user:
				cli.info( _("Using user '%s' as current user") % (user.username) )

			cli.info( _("Using tenant '%s' as current tenant") % (default_tenant.name) )
			cli.info( _('Type "help" for help'))

		cli.cmdloop()

	#
	# Exit: save history
	#
	cli.save_history(history_file_path)
	
