# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

Questo modulo contiene le classi necessarie per effettuare il parsing di un file di configurazione
della CLI.

	ParsingOnlyCli

"""
import sys


from cli.clilib.lex import lex_join
from cli.clilib.ioslike import CliParsingException

from cli.clilib.ioslike import Cli as BaseCli
from cli.clilib.ioslike import CliEnv as BaseCliEnv
#from cli.clilib.ioslike import JOIN_LIST

from cli.main import register_apps_cli_hooks

from cli.prompts.config.tenants import _conf_tenant

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class ParsingOnlyCli(BaseCli):

	"""
	Versione speciale della CLI per effettuare solo il parsing sintattico.
	
	Questa classe sovrascrive alcuni metodi specifici della CLI per 
	produrre una versione in memoria dell'albero dei comandi letti 
	da file.
	
	Anche la gestioni degli errori e' stata leggermente modificata per
	cambiare i messaggi di errore.
	
	Differenze principali:
	
		* Non mostra prompt.
		* Legge la configurazione da uno stream di caratteri
		* Non effettua il completamento dei comandi.
		* Restituisce messaggi di errore con informazioni sulla riga dell'input che ha generato il problema.
	"""

	def __init__(self, tree, env, *args, **kwargs):
		super(ParsingOnlyCli, self).__init__(tree, env, *args, **kwargs)
		
		self._input_stream = None

		self.current_line_number = 0

		# Struttura dati contenente l'albero dei comandi parsati
		self.parsed_commands = None

	def set_input_stream(self, input_stream):
		self._input_stream = input_stream

	def config_completer(self):
		# NIENTE COMPLETAMENTO
		return

	def completer_handler(self, text, state):
		raise Exception("method not implemented")

	def completions(self, line, forhelp=False):
		raise Exception("method not implemented")

	def display_help(self, ret):
		raise Exception("method not implemented")

	def load_history(self, fpath):
		raise Exception("method not implemented")

	def save_history(self, fpath):
		raise Exception("method not implemented")

	def get_history_items(self):
		raise Exception("method not implemented")

	def display_help(self, ret):
		return

	def info(self, s):
		super(SanetDefaultCli, self).info(s)

	def error(self, s):
		super(ParsingOnlyCli, self).error( "line %s: %s" % (self.current_line_number, s) )
		sys.stderr.write(s)

	def warning(self, s):
		sys.stderr.write(s)

	def read_next_input_line(self, show_prompt):
		"""
		Questa versione della CLI legge da un stream senza mostrare prompt o completamento.
		
		Per segnalare gli errori tiene traccia internamente del numero di riga a cui siamo arrivati.
		"""
		line = self._input_stream.readline()

		if not line:
			raise EOFError("END OF FILE")

		self.current_line_number += 1

		return line

	def on_eof(self):
		# Di default viene stampato un andata a capo per l'utente, ma questa non e' una shell userfriendly
		pass
	
	def on_lex_error(self, e):
		pos = e.args[1]
		raise CliParsingException("line %s: Invalid input starting at column: %s\n" % (self.current_line_number, pos))


	def test_token_match(self, candidate, word):
		"""
		Il controllo di match normale sarebbe
		
		candidate.matches_syntax(self, word) and matches_syntax.matches(self, word)
		^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^     ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
		          sintattico                  +            semantico
		"""
		return candidate.matches_syntax(self, word)

	def get_token_canonical(self, candidate, word):
		"""
		Questa versione della CLI non deve "triggerare" il calcolo della stringa canonica normale
		perche' altrimenti alcuni token vanno a tocciare sul DB.
		"""
		return candidate.canonical_syntax(self, word)

	def on_unmatched_words(self, msg):
		raise CliParsingException("line %s: %s" % (self.current_line_number, msg))

	def on_incomplete_command(self):
			msg = "Incomplete command: %r" % (' '.join(self.current_words))
			
			raise CliParsingException("line %s: %s" % (self.current_line_number, msg))

	def call_token_callback(self, token):

		if self.parsed_commands is None:
			self.parsed_commands = []

		line_tokens = list(self.canonical_words)

		line_tokens = [ lex_join([t]) for t in line_tokens ]

		parsed_command = { 
			'command'    : line_tokens, #list(self.canonical_words), 
			'subcommands': None,
		}

		self.parsed_commands.append( parsed_command )

		super(ParsingOnlyCli, self).call_token_callback(token)

		return

	def on_perform_exception(self, ex):
		raise ex

	def on_generic_exception(self, ex):

		# Se l'eccezione e' di tipo CliParsinException allora e' stata generata
		# da codice "nostro".
		if isinstance(ex, CliParsingException):
			raise ex
			
		# Se e' di un'altro tipo, probabilmente un comando non gestisce il parsing. Limitiamoci a dire a quale riga e' avvenuto il problema.
		raise CliParsingException("Error while parsing at line: %s" % (self.current_line_number))

	def enter_subcli(self, cmds, env):
		
		current_parsed_commands = None

		if self.parsed_commands is None:
			self.parsed_commands = []
		else:
			current_parsed_commands = self.parsed_commands

			if len(current_parsed_commands) > 0:
				if current_parsed_commands[-1]['subcommands'] is None:
					current_parsed_commands[-1]['subcommands'] = []

				self.parsed_commands = current_parsed_commands[-1]['subcommands']

		#
		# Enter sub-section (recursion)
		#
		old_cmds, old_env = self.tree, self.env

		self.tree   = cmds
		self.env    = env

		self.cmdloop()

		#
		# Exit sub-section
		#
		self.tree, self.env = old_cmds, old_env	

		if current_parsed_commands is not None:
			self.parsed_commands = current_parsed_commands

		return



	#def set_current_tenant(self):
	#	raise Exception("Not implemented")

	def get_config_element(self):
		raise Exception("Not implemented")

	def set_config_element(self, e):
		raise Exception("Not implemented")

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def launch_sanet_cli_for_parsing_only(input_stream, debug_mode=False):
	"""
	Funzione di comodo per lanciare la CLI di parsing da dentro il tenant 
	di default.
	"""

	#
	# Prepariamo l'environment iniziale per la CLI per fare solo parsing
	#
	env = {
		'PARSE_ONLY'               : True,
		'DEBUG'                    : debug_mode,
	}
	env = BaseCliEnv(env)

	env.input_stream = input_stream

	#
	# Carichiamo gli hook principali e quelli delle applicazioni
	# abilitate in INSTALLED_APPS
	#
	register_apps_cli_hooks()

	# La Cli da usare e' quella di default di Sanet.
	# FUTURE: sarebbe utile avere altri tipi di CLI per poter leggere da standard input
	#         comandi/file CLI ed elaborare i dati in maniera diversa.
	cli = ParsingOnlyCli( [] , env, debug_interpreter=env['DEBUG'] )

	cli.set_input_stream(input_stream)

	_conf_tenant(cli, None)

	return cli.parsed_commands


