# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import time
import sys
import csv

from sanet_common.timeutils import timestamp_to_isostr

from django.utils.translation import ugettext as _
from django.core.management import call_command

from conf import get_configuration

from core.exceptions import ApplicationException

from core.api.base import tenant_get

from core.models import User


from cli.clilib.ioslike import T_Pipe, T_Keyword
from cli.clilib.ioslike import T_Keyword, T_Number, T_Line, T_Anyword, T_Alternatives
from cli.clilib.ioslike import SORTED_KEYWORDS

from cli.tokens import T_Tenantcompleter, T_KeywordWithHook, T_Agentcompleter

from cli.tokens import T_DataGroupCompleter
from cli.tokens import T_TimeGraphCompleter
from cli.tokens import T_Usercompleter

from cli.textformat import *

from cli.prompts.defs import *

from cli.prompts.defs import _cli_verbose_mode_to_loglevel


#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

csv.register_dialect('cli', delimiter=',')

def csv_output(data):

	spamwriter = csv.writer(sys.stdout, dialect='cli')

	for r in data:
		spamwriter.writerow(r)

	return


def output_table(cli, data):

	cols          = cli.env['COLS']
	output_format = cli.env.get('OUTPUT_FORMAT', "normal")
	output_opts   = cli.env.get('OUTPUT_FORMAT_OPTS', "")

	if output_format == "normal":
		cli.prnt(text_table(data, cols))

	elif output_format == "csv":

		if output_opts:
			csv.register_dialect('cli', delimiter=output_opts)

		csv_output(data[1:])

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def cmd_show_env(cli):

	cols = cli.env['COLS']

	r = [[
		[_('NAME') , 20],
		[_('VALUE'), 60],
	]]

	env = dict(cli.env.items())

	env['SANET_DIR'] = get_configuration().SANET_INSTALL_DIR

	for k,v in sorted(env.items()):
		r.append([
			k, repr(v)
		])

	cli.prnt(text_table(r, cols))

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def cmd_show_date(cli):
	cli.out("%s\n" % ( timestamp_to_isostr(time.time())))

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def cmd_set_term_width(cli):

	# set terminal width <val>|auto
	val = cli.canonical_words[3]

	if val == "auto":
		val = None
	else:
		val = int(val)

	cli.env['COLS'] = val

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def cmd_set_term_format(cli):
	# set terminal format <format> <params>

	if len(cli.canonical_words) > 4:
		params = cli.canonical_words[4]
	else:
		params = ""

	val = cli.canonical_words[3]
	cli.env['OUTPUT_FORMAT'] = val
	cli.env['OUTPUT_FORMAT_OPTS'] = params

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def cmd_set_term_exitwords(cli):

	# set terminal exitwords <w>
	words = cli.canonical_words[3]

	words = words.strip()

	words = words.split(' ')
	for w in words:
		if not w:
			cli.error( _("Invalid word '%s'") % (w) )
			return

	cli.info( _("Setting exit words: %(exit_words)s\n") % {'exit_words': (', '.join(words)) })

	cli.set_exit_words(words)

def cmd_set_term_color(cli):

	# set terminal colors yes/no
	mode = cli.canonical_words[3]

	cli.env['COLORS'] = True if mode == 'yes' else False

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def cmd_use_tenant(cli):

	env = cli.env

	tenant_name = cli.canonical_words[2]

	if tenant_name == env['CURRENT_TENANT']:
		return

	if env.get('CONFIG_TENANT',None):
		cli.error( _("Can not switch to tenant '%s' because current config tenant is '%s'\n") % (tenant_name, env['CONFIG_TENANT'] ) )
		return

	t = tenant_get(tenant_name)

	env['CURRENT_TENANT'] = tenant_name

	cli.out( _("Switched to tenant: %(name)s\n") % { 'name': t.name } )

	#
	# Recalculate capabilities for the current tenant
	#
	user = env['CURRENT_USER']
	if user is not None:
		env['CURRENT_USER_CAPABILITIES'] = user.get_capabilities(t)



#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def __r_help_entry(cli, e, cmdstr):

	rows = []

	# Remove the all the pipe commands (every entry after the Pipe )
	selected_entries = []
	for sub_e in SORTED_KEYWORDS(e.get_subtree()):
		if isinstance(sub_e, T_Pipe): continue
		selected_entries.append(sub_e)

	cmdstr += e.token_repr

	if (e.callback):
		rows += [ [cmdstr, e.help ] ]

	for sub_e in selected_entries:
		rows += __r_help_entry(cli, sub_e, cmdstr + ' ')

	return rows



def cmd_show_help(cli):

	rows = []

	for e in SORTED_KEYWORDS(cli.tree):
		rows += __r_help_entry(cli, e, '')

	# Calculate column width
	maxlen = 0
	for row in rows:
		maxlen = max(maxlen, len(row[0]))

	format_str = "  %%-%ss %%s" % (maxlen)
	for row in rows:
		cli.prnt( format_str % (row[0], row[1]))

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def cmd_show_history(cli):

	for i in cli.get_history_items():
		try:
			cli.prnt(i)
		except Exception, e:
			cli.prnt(repr(i))

def cmd_echo(cli):
	cli.prnt(' '.join(cli.canonical_words[1:]))

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def cmd_execute_datagroup(cli):

	tenant_name = cli.env['CURRENT_TENANT']
	db_set_current_tenant_schema(tenant_name)

	# 0       1         2    3        4        5
	# execute datagroup this [verbose <mode>]
	# execute datagroup path <path>   [verbose <mode>]
	# execute datagroup path <uuid>   [verbose <mode>]
	
	if len(cli.canonical_words) >= 3 and cli.canonical_words[2] == 'this':
		
		verbose_args = cli.canonical_words[3:]

		element          = cli.env['CURRENT_ELEMENT_DATA']
		dg_instance_name = cli.env.get('CONFIG_DATAGROUP_INSTANCE', None)
		
		if dg_instance_name is None:
			cli.warning("You are not inside a datagroup configuration section")
			return
		
		dg = DataGroup.objects.get(tenant__name=tenant_name, element=element, name=dg_instance_name)
		
		uuid_or_path = dg.uuid
	else:
		uuid_or_path = cli.canonical_words[3]
		
		verbose_args = cli.canonical_words[4:]

	loglevel = _cli_verbose_mode_to_loglevel(verbose_args)

	# NOTE: Instead of command line options you must use the "dest" name specified
	# in the "make_option" object:
	#
	#    "-l 10" -> loglevel=10
	#
	call_command('exec_datagroup', uuid_or_path, loglevel=loglevel, tenant=tenant_name, interactive=False, verbosity=True, show_traceback=True)

	return

def cmd_execute_timegraph_by_path(cli):
	
	# 0       1         2    3      4      
	# execute timegraph path <path> [ <seconds> ]


	path   = cli.canonical_words[3]

	if len(cli.canonical_words) > 4:
		kwargs = { 'seconds': cli.canonical_words[4] }
	else:
		kwargs = {} 

	args = path.rsplit(';', 1)

	call_command('exec_timegraph', *args, interactive=False, verbosity=True, show_traceback=True, **kwargs)

	return

def cmd_execute_timegraph_by_uuid(cli):
	
	# 0       1         2    3        4
	# execute timegraph path <uuid> [ <seconds> ]
	
	args = cli.canonical_words[3:4]
	if len(cli.canonical_words) > 4:
		kwargs = { 'seconds': cli.canonical_words[4] }
	else:
		kwargs = {}
	
	call_command('exec_timegraph', *args, interactive=False, verbosity=True, show_traceback=True, **kwargs)
	
	return



from core.api.base import db_set_current_tenant_schema
from core.models import DataGroup

def cmd_check_datagroup(cli):

	current_tenant = cli.env.get('CURRENT_TENANT', None)

	# check datagroup path <path> now

	path = cli.canonical_words[3]
	
	db_set_current_tenant_schema(current_tenant)
	
	dg = DataGroup.objects.get(path=path)

	args = [
		#'--tenant' , dg.tenant.name,
		#'--agent'  , dg.agent.name,
		'check_now',
		path,
	]
	
	call_command('agentinfo', *args, tenant=dg.tenant.name, agent=dg.agent.name, interactive=False, verbosity=True, show_traceback=True)


def cmd_agent_update_config(cli):

	# signal agent <agent> <cmd>

	current_tenant = cli.env.get('CURRENT_TENANT', None)
	db_set_current_tenant_schema(current_tenant)
	
	agent_name = cli.canonical_words[2]

	args = [
		'update_config',
	]

	call_command('agentinfo', 
		*args, 
		tenant=current_tenant, 
		agent=agent_name, 
		interactive=False, verbosity=True, show_traceback=True
	)

exec_commands = [

	T_KeywordWithHook('execute', help=_('Evaluate datagroups/timegraphs locally'), pipe=None, 

		hook_name='tenant:execute',

		subtree=[
		T_Keyword('datagroup', help=_('Check a datagroup'), pipe=None, subtree=[
			T_Keyword('path', help=_('Specify by path'), pipe=None, subtree=[
				T_DataGroupCompleter('id', help=_('Execute datagroup by path'), pipe=None, callback=cmd_execute_datagroup, subtree=[
					T_Keyword('verbose', help=_('Execute datagroup by path verbosely'), pipe=None, callback=cmd_execute_datagroup, subtree=[
						T_Alternatives(VERBOSE_OPTIONS_LABEL , help=_('Execute datagroup by path verbosely'), pipe=None, callback=cmd_execute_datagroup)
					])
				]),
			]),
			T_Keyword('id', help=_('Specify by id'), pipe=None, subtree=[
				T_Anyword('id', help=_('Execute datagroup by id'), callback=cmd_execute_datagroup, subtree=[
					T_Keyword('verbose', pipe=None, callback=cmd_execute_datagroup, subtree=[
						T_Alternatives(VERBOSE_OPTIONS_LABEL , pipe=None, callback=cmd_execute_datagroup)
					])
				]),
			]),
			T_Keyword('this', help=_('Specify by path'), pipe=None, callback=cmd_execute_datagroup, subtree=[
					T_Keyword('verbose', pipe=None, callback=cmd_execute_datagroup, subtree=[
						T_Alternatives(VERBOSE_OPTIONS_LABEL , pipe=None, callback=cmd_execute_datagroup)
					])
				]
			)
		]),

		T_Keyword('timegraph', help=_('Show timegraph data'), pipe=None, subtree=[
			T_Keyword('path', help=_('Show timegraph data by datagroup path and timegraph name'), pipe=None, subtree=[
				T_TimeGraphCompleter('path', help=_('Show timegraph data by datagroup path and timegraph name'), pipe=None, callback=cmd_execute_timegraph_by_path, subtree=[
					T_Number(help=_('Show timegraph data by datagroup path and timegraph name and period in seconds'), pipe=None, callback=cmd_execute_timegraph_by_path)
				]),

				#T_DataGroupCompleter('id', help=_('Show timegraph data by datagroup path and timegraph name'), pipe=None, subtree=[
				#	T_Anyword('name', help=_('Show timegraph data by datagroup path and timegraph name'), pipe=None, callback=cmd_execute_timegraph_by_path, subtree=[
				#		T_Number(help=_('Show timegraph data by datagroup path and timegraph name and period in seconds'), pipe=None, callback=cmd_execute_timegraph_by_path)
				#	]),
				#])
			]),
			T_Keyword('id', help=_('Show timegraph data by id'), pipe=None, subtree=[
				T_Anyword('id', help=_('Show timegraph data by id'), pipe=None, callback=cmd_execute_timegraph_by_uuid, subtree=[
					T_Number(help=_('Show timegraph data by id and period in seconds'), pipe=None, callback=cmd_execute_timegraph_by_uuid)
				]),
			]),
		]),
	]),
	
	T_KeywordWithHook('check', help=_('Force execution of datagroup by its agent'), pipe=None, subtree=[
		T_Keyword('datagroup', help=_('Force execution of datagroup by its agent'), pipe=None, subtree=[
			T_Keyword('path', help=_('Force execution of datagroup by its agent'), pipe=None, subtree=[
				T_DataGroupCompleter('id', help=_('Force execution of datagroup by its agent'), pipe=None, subtree=[
					T_Keyword('now', help=_('Force execution of datagroup by its agent'), callback=cmd_check_datagroup)
				])
			])
		])
	]),

	T_KeywordWithHook('signal', help=_('Force execution of datagroup by its agent'), pipe=None, subtree=[
		T_Keyword('agent', help=_('Force execution of datagroup by its agent'), pipe=None, subtree=[
			T_Agentcompleter('agent', help=_('Force execution of datagroup by its agent'), pipe=None, subtree=[
				T_Keyword('update-config', help=_('Force execution of datagroup by its agent'), callback=cmd_agent_update_config)
			])
		])
	])
]

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

set_tenant_commands = [
	T_Keyword('tenant', help=_('Set current tenant'), subtree=[
		T_Tenantcompleter(help=_('Tenant name'), callback=cmd_use_tenant, pipe=None),
	]),
]

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def cmd_set_errors(cli):

	# set errors default 
	# set errors ignore

	if cli.canonical_words[2] == 'ignore':
		cli.env['DEF_EXIT_ON_ERROR'] = cli.env['EXIT_ON_ERROR']
		cli.env['EXIT_ON_ERROR'] = False

		cli.warning("Errors disabled (exit on error was: %s)" % (cli.env['DEF_EXIT_ON_ERROR']))

	elif cli.canonical_words[2] == 'default':
		if cli.env.has_key('DEF_EXIT_ON_ERROR'):
			cli.env['EXIT_ON_ERROR'] = cli.env['DEF_EXIT_ON_ERROR']
			cli.warning("Errors reset to default (exit on error=%s)" % (cli.env['EXIT_ON_ERROR']))

			del cli.env['DEF_EXIT_ON_ERROR']

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def cmd_sleep(cli):

	if len(cli.canonical_words) > 1:
		seconds = int(cli.canonical_words[1])
	else:
		seconds = 1

	time.sleep(seconds)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def cms_set_current_user(cli):
	
	# set user <username>

	current_user     = cli.env['CURRENT_USER']
	_real_user       = cli.env.get('_REAL_USER', None)
	
	
	username = cli.canonical_words[2]
	user = User.objects.get(username=username)


	if current_user is None:
		raise ApplicationException("CLI does not support user authentication. Command not working.")


	if _real_user is None:

		if not current_user.is_superuser:
			raise ApplicationException("Current user is not admin.")

		cli.env['_REAL_USER'] = current_user
	else:
		
		if not _real_user.is_superuser:
			raise ApplicationException("Only admin users can switch identity.")

	cli.env['CURRENT_USER'] = user

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

common_commands = [

	T_Keyword('env' , help=_('Print shell environment settings (for debug)'), callback=cmd_show_env ),

	T_Keyword('date', help=_('Print current time'), callback=cmd_show_date ),

	T_Keyword('echo', help=_('Print current time'), subtree=[
		T_Line(help="line", callback=cmd_echo)
	]),

	T_Keyword('sleep', help=_('Wait'), subtree=[
		T_Number(help=_('Wait'), callback=cmd_sleep)
	]),


	T_KeywordWithHook('set', help=_('Change shell working environemnt'), 

		hook_name='common:set',

		subtree=[
			T_Keyword('terminal', help=_('Set terminal properties for current session'), subtree=[
				T_Keyword('width', help=_('Set terminal witdh in columns'), subtree=[
					T_Keyword('auto', help=_('Autodetect terminal width'), callback=cmd_set_term_width, pipe=None),
					T_Number(help=_('Number of columns, 0 for unlimited columns'), callback=cmd_set_term_width, pipe=None),
				]),
				T_Keyword('exitwords', help=_('Words recognized as exit comands, for this level of the CLI'), subtree=[
					T_Line(help=_('Space separated words, default: exit logout quit'), callback=cmd_set_term_exitwords, pipe=None,)
				]),
				T_Keyword('format', help=_('Set output format'), subtree=[
					T_Alternatives([('normal',''),('csv','')], help=_('Set output format'), callback=cmd_set_term_format, pipe=None, subtree=[
						T_Line(help='',callback=cmd_set_term_format),
					]),
				]),
				T_Keyword('colors', help=_('Color'), subtree=[
					T_Alternatives([('yes',''),('no','')], help=_('Set output format'), callback=cmd_set_term_color, pipe=None),
				]),
			]),
			T_Keyword('errors', help=_('Error handling for the current section'), subtree=[
				T_Keyword('default'  , help=(''), callback=cmd_set_errors),
				T_Keyword('ignore'   , help=(''), callback=cmd_set_errors),
			]),
			
			T_Keyword('user', help=_('Set current user'), subtree=[
				T_Usercompleter(help=_('Set current user'), callback=cms_set_current_user, pipe=None)
			]),
		]
	),

	T_Keyword('history', help=_('View CLI history'),
		callback=cmd_show_history,
	),

	T_Keyword('help', help=_('Print help'),
		callback=cmd_show_help,
	),

]

