# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import datetime
import re
import traceback

import pytz

from django.db import transaction
from django.utils.timezone import now

from django.db import transaction

from core.db_utils import transaction_commit_on_success

from core.models import PriorityLabel

from core.api.base import tenant_get, tenant_update, tenant_create, tenant_remove, tenant_get_primary, tenant_change_primary_by_name
from core.api.base import node_remove
from core.api.base import link_modify


DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S'


from cli.clilib.ioslike import *
from cli.textformat import *

from cli.tokens import *


from cli.prompts.defs import PROMPT_SEP
from cli.prompts.defs import LINK_ALTERNATIVES, LINK_CLI_TO_TYPE

from cli.prompts import cli_get_registered_menus

from cli.prompts.config.schedulinggroup import cmd_schedulinggroup_config_prompt

from cli.prompts.config.agent import cmd_agent_config_prompt
from cli.prompts.config.agent import cmd_agent_remove
from cli.prompts.config.agent import cmd_agent_rename

from cli.prompts.config.cluster import cmd_cluster_rename
from cli.prompts.config.cluster import cmd_cluster_config_prompt
from cli.prompts.config.cluster import cmd_cluster_remove

from cli.prompts.config.nodes import cmd_node_config_prompt

from cli.prompts.config.templates.datagroups import cmd_datagroup_template_config_prompt
from cli.prompts.config.templates.datagroups import cmd_datagroup_template_move_to_library

from cli.prompts.config.templates.elements import cmd_node_template_config_prompt
from cli.prompts.config.templates.elements import cmd_node_template_remove
from cli.prompts.config.templates.elements import cmd_interface_template_config_prompt
from cli.prompts.config.templates.elements import cmd_interface_template_remove
from cli.prompts.config.templates.elements import cmd_storage_template_config_prompt
from cli.prompts.config.templates.elements import cmd_storage_template_remove
from cli.prompts.config.templates.elements import cmd_service_template_config_prompt
from cli.prompts.config.templates.elements import cmd_service_template_remove
from cli.prompts.config.templates.elements import cmd_device_template_config_prompt
from cli.prompts.config.templates.elements import cmd_device_template_remove

#------------------------------------------------------------------------------#
# TENANTS                                                                      #
#------------------------------------------------------------------------------#

@transaction_commit_on_success
def cmd_tenant_remove(cli):
	
	from django import db
	
	db.reset_queries()
	
	# no tenant <name> <name>
	try:
		name  = cli.canonical_words[2]
		name2 = cli.canonical_words[3]
		
		if name != name2:
			raise Exception("Name mismatch")

		cli.warning("*****************************************")
		cli.warning("EVERY DATA WILL BE LOST")
		cli.warning("*****************************************")

		answer =  raw_input("Are you really sure [y/N]") 
		if answer != 'y':
			return

		cli.info("Deleting tenant...")

		tenant_remove(name)
		cli.info("Tenant deleted.")
		
		if name == cli.env['CURRENT_TENANT']:

			t = tenant_get_primary()

			cli.info("Switching to default tenant '%s'" % (t.name))
			
			t.db_set_current_schema()
			
			cli.env['CURRENT_TENANT'] = t.name

			cli.info("done.")

	
	except Exception, e:
		cli.error(str(e))

	return

def cmd_tenand_set_primary(cli):

	tenant_name = cli.env['CONFIG_TENANT']

	if tenant_change_primary_by_name(tenant_name):
		cli.info( "Primary changed" )
	

def cmd_tenand_set_attribute(cli):

	param = cli.canonical_words[0]
	value = cli.canonical_words[1]

	tenant_name = cli.env['CONFIG_TENANT']	
	tenant = tenant_select_all(name_filter_str=tenant_name)[0]

	if param == 'long-name': 
		tenant.long_name = value

		updated_fields = ['long_name']

	elif param == 'timezone': 

		try:
			tz = pytz.timezone(value)
		except pytz.exceptions.UnknownTimeZoneError, e:
			raise e

		tenant.timezone = value

		updated_fields = ['timezone']

	else:
		raise Exception("Invalid parameter '%s'" % (param))
			
	tenant_update(tenant, updated_fields)


#------------------------------------------------------------------------------#
# NODES                                                                        #
#------------------------------------------------------------------------------#


def cmd_node_remove(cli):

	if cli.env.get('PARSE_ONLY', False):
		return
		
	# no node <n> <n>

	tenant_name  = cli.env['CONFIG_TENANT']

	name  = cli.canonical_words[2]
	name2 = cli.canonical_words[3]
	
	if name != name2:
		cli.error("Node name mismatch (%s <> %s)" % (name, name2))
		return

	node_remove(name, tenant_name=tenant_name)

#------------------------------------------------------------------------------#
# LINKS                                                                        #
#------------------------------------------------------------------------------#

def cmd_link_modify(cli):

	if cli.env.get('PARSE_ONLY', False):
		return

	# [u'link', u'nodo1', u'iface1', u'nodo2', u'iface2', u'prova', 'virtual', u'pippero']
	
	tenant_name = cli.env['CONFIG_TENANT']

	n1_name = cli.canonical_words[1]
	i1_name = cli.canonical_words[2]
	n2_name = cli.canonical_words[3]
	i2_name = cli.canonical_words[4]
	
	lname = None
	if len(cli.canonical_words) >= 6:
		lname = cli.canonical_words[5]
		
	ltype = None
	if len(cli.canonical_words) >= 7:
		ltype = cli.canonical_words[6]
		ltype = LINK_CLI_TO_TYPE[ltype]

	ldescr = None
	if len(cli.canonical_words) >= 8:
		ldescr = cli.canonical_words[7]
	
	link_modify(n1_name, i1_name, n2_name, i2_name,  lname=lname, ltype=ltype, ldescr=ldescr, tenant_name=tenant_name)


#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

@transaction_commit_on_success
def cmd_set_prioritylevel(cli):
	
	tenant_name = cli.env['CONFIG_TENANT']

	# level <n> <lable> <descr>
	# level <label> <min> <max> <label> 
	label       = cli.canonical_words[1]
	min_level   = int(cli.canonical_words[2])
	max_level   = int(cli.canonical_words[3])
	description = cli.canonical_words[4]

	tenant = tenant_get(tenant_name)
	
	db_set_current_tenant_schema(tenant.name)


	if PriorityLabel.can_add_new() == False:
		cli.error("Too many priority levels (max number %s)" % (PriorityLabel))
		return

	created = False
	try:
		l = PriorityLabel.objects.get(tenant=tenant, label=label)
	except Exception, e:
		l = PriorityLabel(tenant=tenant, label=label)
		created = True
		
		
	l.min_level = min_level
	l.max_level = max_level
	l.description = description
	
	l.validate()
	
	if created:
		l.save()
	else:
		l.concurrent_save(fields_name=['min_level','max_level','description'])
	
	return
	
	
@transaction_commit_on_success
def cmd_del_prioritylevel(cli):
	
	tenant_name = cli.env['CONFIG_TENANT']

	# no level <lable> <label>
	label = cli.canonical_words[2]
	label2 = cli.canonical_words[3]
	
	if label != label2:
		cli.error("Name mismatch: %s %s" % (label, label2))
		return

	tenant = tenant_get(tenant_name)
	
	db_set_current_tenant_schema(tenant.name)

	l = PriorityLabel.objects.get(tenant=tenant, label=label)
	l.delete()

#"""
#@transaction_commit_on_success
#def cmd_change_prioritylevel(cli):
#	
#	print "SET"
#	
#	tenant_name = cli.env['CONFIG_TENANT']
#
#	# change <n> <n>
#	old_level       = int(cli.canonical_words[1])
#	new_level       = int(cli.canonical_words[2])
#
#	tenant = tenant_get(tenant_name)
#	
#	db_set_current_tenant_schema(tenant.name)
#	
#	try:
#		l = PriorityLevel.objects.get(tenant=tenant, level=old_level)
#		l.level = new_level
#		l.save()
#
#	except PriorityLevel.DoesNotExist, e:
#		print "not found"
#"""
#
#"""
#mt_config_priority_levels = [
#	T_Keyword('priority-level', help='', pipe=None, subtree=[
#		T_Anyword       (help='label', callback=cmd_set_prioritylevel , pipe=None, subtree=[
#			T_Number(help='level', pipe=None, subtree=[
#				T_Number(help='level', pipe=None, subtree=[
#					T_Anyword       (help='description', callback=cmd_set_prioritylevel , pipe=None, )
#				]),
#			]),
#		]),
#	]),
#]
#"""
#
#"""
#T_Keyword('add', help='', pipe=None, subtree=[
#	T_Anyword       (help='label', callback=cmd_add_prioritylevel , pipe=None, subtree=[
#		T_Anyword       (help='description', callback=cmd_add_prioritylevel , pipe=None,
#		)
#	]),
#]),
#T_Keyword('change', help='', pipe=None, subtree=[
#	T_Number(help='old level', pipe=None, subtree=[
#		T_Number(help='new level', callback=cmd_change_prioritylevel , pipe=None )
#	]),
#]),
#"""
#"""
#def cmd_prioritylevels_config_prompt(cli):
#	
#	tenant_name = cli.env['CONFIG_TENANT']
#
#
#	env = cli.env.clone()
#	env['PROMPT'] = 'config-%s#' % ( PROMPT_SEP.join([ tenant_name, 'levels' ]) )
#	
#	cmds = mt_config_priority_levels  + SORTED_KEYWORDS(cli_get_registered_menus('__common_commands__') + cli_get_registered_menus('__show_commands__'))
#
#	cli = cli.get_class()(cmds, env, debug_interpreter=env['DEBUG'])
#	cli.cmdloop()	
#"""	

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def cmd_refresh_datagroup_template(cli):
	# refresh datagroup-template <name>

	tenant_name       = cli.env['CONFIG_TENANT']
	tenant  = tenant_get(tenant_name)		
	if tenant:
		db_set_current_tenant_schema(tenant.name)
	
	dgt_name = cli.canonical_words[2]


	dgt = datagroup_template_get(dgt_name, tenant=tenant)
	
	from core.api.templates import refresh_datagroup_template
	
	refresh_datagroup_template(dgt)

def cmd_refresh_node(cli):
	# refresh datagroup-template <name>

	tenant_name       = cli.env['CONFIG_TENANT']
	tenant  = tenant_get(tenant_name)		
	if tenant:
		db_set_current_tenant_schema(tenant.name)
	
	# refresh node <name>
	node_name = cli.canonical_words[2]

	node = MonitoredNode.objects.get(name=node_name)

	from core.models import ElementTemplate
	from core.api.templates import refresh_element
	from core.api.templates import refresh_element_template
	
	refresh_element(node)
	
	#XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
	
	#for et in ElementTemplate.objects.filter(tenant=tenant, melement=node):
	#	
	#	print "---------------------", et.telement.name
	#	
	#	refresh_element_template(tenant, et.telement, node)
	
	#XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
	
	
def cmd_refresh_datagroup(cli):
	
	# refresh datagroup <path>	
	tenant_name       = cli.env['CONFIG_TENANT']
	tenant  = tenant_get(tenant_name)		
	if tenant:
		db_set_current_tenant_schema(tenant.name)
		
	path = cli.canonical_words[2]
		
	dg = DataGroup.objects.get(tenant=tenant, path=path)
	
	from core.api.templates import refresh_datagroup
	
	refresh_datagroup(dg)
	
	
def cmd_refresh_all_nodes(cli):

	from core.api.templates import refresh_element

	# refresh datagroup <path>	
	tenant_name       = cli.env['CONFIG_TENANT']
	tenant  = tenant_get(tenant_name)		
	if tenant:
		db_set_current_tenant_schema(tenant.name)
		
	for e in MonitoredElement.objects.all():
		cli.info( "Refreshing %s" % ( e.name ) )
		refresh_element(e)


#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def cmd_rename_node(cli):

	# refresh datagroup-template <name>

	tenant_name = cli.env['CONFIG_TENANT']
	tenant = tenant_get(tenant_name)
	
	# rename node <name>
	node_name = cli.canonical_words[2]
	
	name  = cli.canonical_words[2]
	name2 = cli.canonical_words[3]
	
	from core.api.elements import element_rename
	
	element = MonitoredElement.objects.get(tenant=tenant, kind=ELEMENT_KIND_NODE, parent=None, name=name)
	
	element_rename(element, name2)
	
def cmd_rename_link(cli):

	tenant_name = cli.env['CONFIG_TENANT']
	tenant = tenant_get(tenant_name)
	
	# rename link <name> <name>
	node_name = cli.canonical_words[2]
	
	name  = cli.canonical_words[2]
	name2 = cli.canonical_words[3]
	
	from core.api.base import link_rename
	
	link_rename(tenant, name, name2)


def cmd_remove_link(cli):

	tenant_name = cli.env['CONFIG_TENANT']
	tenant = tenant_get(tenant_name)
	
	# no link <name> <name>
	node_name = cli.canonical_words[2]
	
	name  = cli.canonical_words[2]
	name2 = cli.canonical_words[3]
	
	from core.api.base import link_rename
	
	link_delete(tenant, name)	

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

from core.api.templates import datagroup_template_rename_in_tenant

def cmd_rename_datagroup_template(cli):

	tenant_name = cli.env['CONFIG_TENANT']
	tenant = tenant_get(tenant_name)
	

	# rename datagroup-template <n> <n>
	old_name = cli.canonical_words[2]
	new_name = cli.canonical_words[3]

	fix_logs = cli.canonical_words[4] == 'fix-logs'

	if fix_logs:
		cli.warning("Update existing logs could take a very long time.")
		cli.warning("Are you sure to continue [y/N]")
		response = cli.read_next_input_line(show_prompt=False)
		if response.strip().lower() != 'y':
			cli.info("Cancelled")
			return
	else:
		cli.warning("Renaming a template without fixing existing logs could cause problems with future data elaborations.")
		cli.warning("Are you sure to continue [y/N]")
		response = cli.read_next_input_line(show_prompt=False)
		if response.strip().lower() != 'y':
			cli.info("Cancelled")
			return

	cli.info( "Renaming datagroups in element templates" )
	
	def _feedback(msg):
		cli.info(msg)

	#_rename_datagroup_template(tenant, old_name, new_name, feedback_callback=_feedback)

	datagroup_template_rename_in_tenant(tenant, old_name, new_name, fix_logs=fix_logs, feedback_callback=_feedback)

	return



def cmd_rename_node_template(cli):
	# rename node-template <name> <new name>
	return _cmd_rename_element_template(cli, ELEMENT_KIND_NODE)

def cmd_rename_interface_template(cli):
	# rename node-template <name> <new name>
	return _cmd_rename_element_template(cli, ELEMENT_KIND_IFACE)

def cmd_rename_storage_template(cli):
	# rename node-template <name> <new name>
	return _cmd_rename_element_template(cli, ELEMENT_KIND_STORAGE)

def cmd_rename_service_template(cli):
	# rename node-template <name> <new name>
	return _cmd_rename_element_template(cli, ELEMENT_KIND_SERVICE)

def cmd_rename_device_template(cli):
	# rename node-template <name> <new name>
	return _cmd_rename_element_template(cli, ELEMENT_KIND_DEVICE)


@transaction_commit_on_success
def _cmd_rename_element_template(cli, kind):

	tenant_name = cli.env['CONFIG_TENANT']
	tenant = tenant_get(tenant_name)
	
	db_set_current_tenant_schema(tenant_name)
	
	name     = cli.canonical_words[2]
	new_name = cli.canonical_words[3]
	
	if name == new_name:
		return

	#print name, new_name
	# FIX:  move to core.api.template
	
	mtime = time.time()
	
	
	telement = TElement.objects.get(tenant=tenant, name=name, kind=kind)
	
	telement.name = new_name
	
	element_template_path = telement.name
	
	
	#
	# Aggiorniamo il name di tutti i sub-element template
	#
	
	for sub_telement in telement.get_sub_elements().all():
		
		parent_template_name, sub_element_name = sub_telement.name.split(':')
	
		new_name = '%s:%s' % (telement.name, sub_element_name)
		
		sub_telement.name = new_name
		sub_telement.save()
		
	#
	# Update datagroups
	#

	mes = ElementTemplate.objects.filter(tenant=tenant, telement=telement).select_related('melement').all()
	
	for e in mes:
		
		element = e.melement
		
		#print element

		for dg in element.datagroups.all():
			# FIX: trasformare in un metodo del modello
			path = dg.path

			if path.find(name) < 0:  # FIX: dobbiamo modificare solo i datagroup appartenenti all'elemente template. Filtriamo cosi'. Fa schifo, ma funziona.
				#print "skipped", path
				continue

			path = path.split(DATAGROUP_PATH_SEP)
			path[PATH_PART_TEMPLATE_POS] = element_template_path
			path = DATAGROUP_PATH_SEP.join(path)

			#print dg.path, path

			dg.path = path

			dg.concurrent_save(['path'], new_sentinel_value=mtime)

		for ds in element.monitored_datasources.all():
			# FIX: trasformare in un metodo del modello
			path = ds.path

			if path.find(name) < 0:  # FIX: dobbiamo modificare solo i datagroup appartenenti all'elemente template. Filtriamo cosi'. Fa schifo, ma funziona.
				#print "skipped", path
				continue


			path = path.split(DATAGROUP_PATH_SEP)
			path[PATH_PART_TEMPLATE_POS] = element_template_path
			path = DATAGROUP_PATH_SEP.join(path)

			#print ds.path, path

			ds.path = path

			ds.concurrent_save(['path'], new_sentinel_value=mtime)

		for c in element.monitored_conditions.all():
			# FIX: trasformare in un metodo del modello
			path = c.path		

			if path.find(name) < 0: # FIX: dobbiamo modificare solo i datagroup appartenenti all'elemente template. Filtriamo cosi'. Fa schifo, ma funziona.
				#print "skipped", path
				continue

			path = path.split(DATAGROUP_PATH_SEP)
			path[PATH_PART_TEMPLATE_POS] = element_template_path
			path = DATAGROUP_PATH_SEP.join(path)

			#print c.path, path
			c.path = path

			c.concurrent_save(['path'], new_sentinel_value=mtime)

		# Questo dovrebbe assicurarci che l'aggiornamento dell'elemento e 
		# tutti i sotto-elementi sia cmq atomico.
		# Es:  utente A aggiorna node1  e utente B aggiorna solo node1:eth0
		element.concurrent_save(['name'], new_sentinel_value=mtime)
		

	# Save now the main element (2phase commit)	Commit
	telement.concurrent_save(['name'], new_sentinel_value=mtime)



#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def parse_suspend_upto(buf):
	# Formats: "YYYY-MM-DD HH:MM:SS", "forever", number of seconds
	if buf.lower() == "forever":
		return 0
	try:
		return int(time.time()) + int(buf)
	except:
		raise ValueError('Expected number of seconds, or date and time in "YYYY-MM-DD HH:MM:SS" format (with quotes), or "forever", got:' + buf)


def parse_suspend_options(options):
	
	DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S'
	
	# for <n> <reason>
	if options[0] == 'for':
		seconds = int(options[1])
		t       = time.time() + seconds
		
		options  = options[2:]
		
	# forever <reason>
	elif options[0] == 'forever':
		t       = 0
		
		options  = options[1:]

	# until <date> <reason>
	elif options[0] == 'until':
		date    = options[1]
		try:
			d = datetime.datetime.strptime(date, DATETIME_FORMAT)
		except ValueError:
			raise ValueError('Expected date and time in "YYYY-MM-DD HH:MM:SS" format (with double quotes), got: ' + date)

		t = int(time.mktime(d.timetuple()))

		options  = options[2:]
		
	if options:
		reason = options[0]
	else:
		reason = 'Suspended by CLI'

	return t, reason

from core.api.base import *
from core.api.elements import element_suspend, element_resume
from core.api.elements import datagroup_suspend, datagroup_resume

def cmd_suspend(cli):

	tenant_name = cli.env['CONFIG_TENANT']
	tenant = tenant_get(tenant_name)
	
	# suspend <type> <name>         for <num>     [ <reason> ]
	# suspend <type> <node> [<sub>] forever       [ <reason> ]
	# suspend <type> <node> [<sub>] until <date>  [ <reason> ]
	
	rtype = cli.canonical_words[1]
	
	if rtype == 'datagroup':
		t, reason = parse_suspend_options(cli.canonical_words[3:])
		
		path = cli.canonical_words[2]
		datagroup = DataGroup.objects.get(tenant=tenant, path=path)
		
		datagroup_suspend(datagroup, t, reason)

	elif rtype == 'node':
		t, reason = parse_suspend_options(cli.canonical_words[3:])
		
		node_name = cli.canonical_words[2]
		element = node_get(node_name, tenant_name=tenant_name)
		
		element_suspend(element, t, reason)
	else:
		t, reason = parse_suspend_options(cli.canonical_words[4:])

		node_name  = cli.canonical_words[2]
		name = cli.canonical_words[3]

		if rtype == 'interface':
			element = iface_get(node_name, name, tenant_name=tenant_name)
		elif rtype == 'storage':
			element = storage_get(node_name, name, tenant_name=tenant_name)
		elif rtype == 'device':
			element = device_get(node_name, name, tenant_name=tenant_name)
		elif rtype == 'service':
			element = service_get(node_name, name, tenant_name=tenant_name)

		element_suspend(element, t, reason)

def cmd_resume(cli):
	
	def get_reason(words):
		
		reason = None
		
		if words:
			reason = ''.join(words)
		
		if reason:
			return reason
			
		return 'Resumed by CLI (%s)' % (time.ctime())
			
	tenant_name = cli.env['CONFIG_TENANT']
	tenant = tenant_get(tenant_name)
	
	# resume <type> <name>       [ <reason> ]
	# resume <type> <node>       [ <reason> ]
	# resume <type> <node> <sub> [ <reason> ]
	
	rtype = cli.canonical_words[1]

	#reason = 'Resumed by CLI (%s)' % (time.ctime())
	
	if rtype == 'datagroup':
		
		path = cli.canonical_words[2]
		
		reason =  get_reason(cli.canonical_words[3:])
		
		datagroup = DataGroup.objects.get(tenant=tenant, path=path)
		
		datagroup_resume(datagroup, reason)

	elif rtype == 'node':
		
		node_name = cli.canonical_words[2]
		
		reason =  get_reason(cli.canonical_words[3:])
		
		element = node_get(node_name, tenant_name=tenant_name)				
		
		element_resume(element, reason)
	else:
		node_name  = cli.canonical_words[2]
		name = cli.canonical_words[3]

		reason =  get_reason(cli.canonical_words[4:])

		if rtype == 'interface':
			element = iface_get(node_name, name, tenant_name=tenant_name)
		elif rtype == 'storage':
			element = storage_get(node_name, name, tenant_name=tenant_name)
		elif rtype == 'device':
			element = device_get(node_name, name, tenant_name=tenant_name)
		elif rtype == 'service':
			element = service_get(node_name, name, tenant_name=tenant_name)

		element_resume(element, reason)

_suspend = [ 
	
	T_Keyword('for', subtree=[
		T_Number(help='Number of seconds, or  format (with quotes), or "forever"', callback=cmd_suspend, subtree=[
			T_Line(help='Reason for suspension', callback=cmd_suspend, pipe=None)
		])
	]),
	T_Keyword('forever', callback=cmd_suspend, subtree=[
		T_Line(help='Reason for suspension', callback=cmd_suspend, pipe=None)
	]),
	T_Keyword('until', help="date and time in YYYY-MM-DD HH:MM:SS", subtree=[
		T_Anyword(help='', callback=cmd_suspend, subtree=[
			T_Line(help='Reason for suspension', callback=cmd_suspend, pipe=None)
		])
	]),
]
	
mt_suspend = T_Keyword('suspend', help='Suspend targets and measures', subtree=[
	T_Keyword('node', help='', subtree=[
		T_MonitoredNodecompleter(help='Node name', subtree=_suspend),
	]),
	T_Keyword('interface', help='', subtree=[
		T_MonitoredNodecompleter(help='Node name', subtree=[
			T_InlineIfaceCompleter(help='Iface name', subtree=_suspend),
		]),
	]),
	T_Keyword('storage', help='', subtree=[
		T_MonitoredNodecompleter(help='Node name', subtree=[
			T_InlineStorageCompleter(help='Storage name', subtree=_suspend),
		]),
	]),
	T_Keyword('device', help='', subtree=[
		T_MonitoredNodecompleter(help='Node name', subtree=[
			T_InlineDeviceCompleter(help='Service name', subtree=_suspend),
		]),
	]),
	T_Keyword('service', help='', subtree=[
		T_MonitoredNodecompleter(help='Node name', subtree=[
			T_InlineServiceCompleter(help='Device name', subtree=_suspend),
		]),
	]),
	
	T_Keyword('datagroup', help='', subtree=[
		T_DataGroupCompleter(help='Datagroup',  subtree=_suspend),
	])
])

mt_resume = T_Keyword('resume', help='Suspend targets and measures', subtree=[
	T_Keyword('node', help='', subtree=[
		T_MonitoredNodecompleter(help='Node', callback=cmd_resume, subtree=[
			T_Line(help='Reason for suspension', callback=cmd_suspend, pipe=None)
		]),
	]),
	T_Keyword('interface', help='', subtree=[
		T_MonitoredNodecompleter(help='Node', subtree=[
			T_InlineIfaceCompleter(help='Iface', callback=cmd_resume, subtree=[
				T_Line(help='Reason for suspension', callback=cmd_suspend, pipe=None)
			]),
		]),
	]),
	T_Keyword('storage', help='', subtree=[
		T_MonitoredNodecompleter(help='Node name', subtree=[
			T_InlineStorageCompleter(help='Storage name', callback=cmd_resume, subtree=[
				T_Line(help='Reason for suspension', callback=cmd_resume, pipe=None)
			]),
		]),
	]),
	T_Keyword('service', help='', subtree=[
		T_MonitoredNodecompleter(help='Node', subtree=[
			T_InlineServiceCompleter(help='Service name', callback=cmd_resume, subtree=[
				T_Line(help='Reason for suspension', callback=cmd_resume, pipe=None)
			]),
		]),
	]),
	T_Keyword('device', help='', subtree=[
		T_MonitoredNodecompleter(help='Node name', subtree=[
			T_InlineDeviceCompleter(help='Device name', callback=cmd_resume, subtree=[
				T_Line(help='Reason for suspension', callback=cmd_resume, pipe=None)
			]),
		]),
	]),
	
	T_Keyword('datagroup', help='', subtree=[
		T_DataGroupCompleter(help='Datagroup path', callback=cmd_resume, subtree=[
			T_Line(help='Reason for suspension', callback=cmd_resume, pipe=None)
		]),
	])
])

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

from django.core.management import call_command

def cmd_restart_agent(cli):

	# restart agent <name>

	tenant_name = cli.env['CONFIG_TENANT']
	tenant = tenant_get(tenant_name)

	agent_name = cli.canonical_words[2]

	call_command('agentinfo', 'stop', agent=agent_name, loglevel=10, tenant=tenant_name, interactive=False, verbosity=True, show_traceback=True)


mt_restart = T_Keyword('restart', help='Restart subsystem', subtree=[
	T_Keyword('agent', help='Send restart signal to agent', subtree=[
		T_Agentcompleter(help='Existing agent name', callback=cmd_restart_agent)
	])
])


#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def cmd_set_minimum_priority_level(cli):

	tenant_name = cli.env['CONFIG_TENANT']
	tenant = tenant_get(tenant_name)
	db_set_current_tenant_schema(tenant.name)
	
	if cli.canonical_words[0] != 'no':

		selected_label = cli.canonical_words[1]
		selected_level = PriorityLabel.objects.get(tenant=tenant, label=selected_label)

		current_minimum_priority  = tenant.get_setting(TenantSetting.SETTING_MINIMUM_LEVEL)

		if selected_level.min_level == current_minimum_priority:
			return

		current_critical_priority = tenant.get_setting(TenantSetting.SETTING_CRITICAL_LEVEL)

		if selected_level.min_level > current_critical_priority:
			cli.error( "Minimum level (%s) can not be greater than critical level (%s)" % (selected_level.min_level, current_critical_priority) )
			return

		TenantSetting.save_setting(tenant, TenantSetting.SETTING_MINIMUM_LEVEL, selected_level.min_level)
	else:
		TenantSetting.save_setting(tenant, TenantSetting.SETTING_MINIMUM_LEVEL, PRIORITY_NONE)


def cmd_set_critical_priority_level(cli):

	tenant_name = cli.env['CONFIG_TENANT']
	tenant = tenant_get(tenant_name)
	db_set_current_tenant_schema(tenant.name)

	if cli.canonical_words[0] != 'no':

		selected_label = cli.canonical_words[1]
		selected_level = PriorityLabel.objects.get(tenant=tenant, label=selected_label)

		current_critical_priority  = tenant.get_setting(TenantSetting.SETTING_CRITICAL_LEVEL)

		if selected_level.min_level == current_critical_priority:
			return

		current_minimum_priority = tenant.get_setting(TenantSetting.SETTING_MINIMUM_LEVEL)

		if selected_level.min_level < current_minimum_priority:
			cli.error( "Critical level (%s) can not be lower than minimum level (%s)" % (selected_level.min_level, current_minimum_priority) )
			return

		TenantSetting.save_setting(tenant, TenantSetting.SETTING_CRITICAL_LEVEL, selected_level.min_level)
	else:
		TenantSetting.save_setting(tenant, TenantSetting.SETTING_CRITICAL_LEVEL, PRIORITY_MAX + 1)


#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

mt_config_tenant = SORTED_KEYWORDS([

	T_Keyword('long-name', help='Set tenant long name', subtree=[
		T_Line(help='', callback=cmd_tenand_set_attribute, pipe=None),
	]),

	T_Keyword('primary', help='Set tenant as default', callback=cmd_tenand_set_primary, pipe=None ),

	T_Keyword('timezone', help='Set default timezone for this tenant', subtree=[
		T_Anyword(help='', callback=cmd_tenand_set_attribute, pipe=None),
	]),

	#T_Keyword('group', help='Specify group authorized to access to the tenant', subtree=[
	#	T_Groupcompleter(help='', callback=cmd_tenant_add_group, pipe=None),
	#	T_Anyword       (help='', callback=cmd_tenant_add_group, pipe=None),
	#]),

	T_Keyword('agent', help='Create or modify an monitoring agent', subtree=[
		T_Agentcompleter(help='Existing agent name', callback=cmd_agent_config_prompt, pipe=None,),
		T_ValidWord     (VALID_AGENT_NAME, help='New agent name', callback=cmd_agent_config_prompt, pipe=None,),
	]),

	T_Keyword('scheduling-group', help='XXXXXXXXXX', subtree=[
		T_SchedulingGroupcompleter(help='XXXXXXXXXX', callback=cmd_schedulinggroup_config_prompt, pipe=None,),
		T_ValidWord(VALID_ELEMENT_NAME, help='NXXXXXXXXXXXX', callback=cmd_schedulinggroup_config_prompt, pipe=None )
	]),


	T_Keyword('cluster', help='Create or modify a monitoring cluster', subtree=[
		T_Clustercompleter(help='Existing cluster', callback=cmd_cluster_config_prompt, pipe=None,),
		T_ValidWord(VALID_CLUSTER_NAME, help='New cluster name', callback=cmd_cluster_config_prompt, pipe=None,)
	]),

	T_Keyword('node', help='Configure a node', subtree=[
		T_MonitoredNodecompleter(help='Name of an existing node', callback=cmd_node_config_prompt, pipe=None,),
		T_ValidWord(VALID_NODE_NAME, help='Name of a new node that will be created', callback=cmd_node_config_prompt, pipe=None,)
	]),

	T_Keyword('link', help='Link two interfaces', subtree=[ 

		T_MonitoredNodecompleter(help='Node 1', subtree=[

			#T_MonitoredIfacecompleter(help='Iface 1', subtree=[
			T_InlineIfaceCompleter(help='Iface 1', subtree=[

				T_MonitoredNodecompleter(help='Node 2', subtree=[

					#T_MonitoredIfacecompleter(help='Iface 2', callback=cmd_link_modify, subtree=[
					T_InlineIfaceCompleter(help='Create/Modify a link between interfaces', callback=cmd_link_modify, subtree=[

						T_ValidWord(VALID_LINK_NAME, token_repr='<LINK NAME>', help='Create/Modify a link with the given name', callback=cmd_link_modify, subtree=[
							T_Alternatives(LINK_ALTERNATIVES, help='Create/Modify a link with the given name and type', callback=cmd_link_modify, subtree=[
								T_Line(help='Create/Modify a link between interfaces', callback=cmd_link_modify,)
							]),
						]),

					]),
				],),
			],),
		],),
	]),

	T_Keyword('priority-level', help='Define a new priority level label', pipe=None, subtree=[
		T_ValidWord(VALID_PRIORITY_LEVEL, help='Label', pipe=None, subtree=[
			T_Number(help='Minimum level number', pipe=None, subtree=[
				T_Number(help='Maximum level number', pipe=None, subtree=[
					T_Line (help='Description', callback=cmd_set_prioritylevel , pipe=None, )
				]),
			]),
		]),
	]),

	T_Keyword('datagroup-template', help='', subtree=[
		T_DataGroupTemplate(help='', callback=cmd_datagroup_template_config_prompt, pipe=None,),
		T_LibraryDataGroupTemplate(help='', callback=cmd_datagroup_template_config_prompt, pipe=None,),
		T_ValidWord(VALID_DATAGROUP_NAME, help='new datagroup template', callback=cmd_datagroup_template_config_prompt, pipe=None),
	]),
	
	T_Keyword('move', help='', subtree=[
		T_Keyword('datagroup-template', help='', subtree=[
			T_DataGroupTemplate(help='', pipe=None, subtree=[
				T_Keyword('to-library', help='', callback=cmd_datagroup_template_move_to_library, ),
			]),
		]),
	]),
	
	
	T_Keyword('node-template', help='', subtree=[
		T_NodeTemplate   (help='', callback=cmd_node_template_config_prompt     , pipe=None,),
		T_ValidWord        (VALID_TEMPLATE_NAME, help='', callback=cmd_node_template_config_prompt     , pipe=None),
	]),
	T_Keyword('interface-template', help='', subtree=[
		T_IfaceTemplate  (help='', callback=cmd_interface_template_config_prompt, pipe=None,),
		T_ValidWord        (VALID_TEMPLATE_NAME, help='', callback=cmd_interface_template_config_prompt, pipe=None),
	]),
	T_Keyword('storage-template', help='', subtree=[
		T_StorageTemplate(help='', callback=cmd_storage_template_config_prompt  , pipe=None),
		T_ValidWord        (VALID_TEMPLATE_NAME, help='', callback=cmd_storage_template_config_prompt  , pipe=None),
	]),
	T_Keyword('service-template', help='', subtree=[
		T_ServiceTemplate(help='', callback=cmd_service_template_config_prompt  , pipe=None),
		T_ValidWord        (VALID_TEMPLATE_NAME, help='', callback=cmd_service_template_config_prompt  , pipe=None),
	]),
	T_Keyword('device-template', help='', subtree=[
		T_DeviceTemplate (help='', callback=cmd_device_template_config_prompt   , pipe=None),
		T_ValidWord        (VALID_TEMPLATE_NAME, help='', callback=cmd_device_template_config_prompt   , pipe=None),
	]),
	
	T_Keyword('refresh', help='' , subtree=[
		T_Keyword('datagroup', help='', subtree=[
			T_DataGroupCompleter(help='', callback=cmd_refresh_datagroup, pipe=None)
		]),
		T_Keyword('datagroup-template', help='', subtree=[
			T_DataGroupTemplate(help='', callback=cmd_refresh_datagroup_template, pipe=None,),
		]),	
		T_Keyword('node', help='Refresh a node', subtree=[
			T_MonitoredNodecompleter(help='Name of an existing node', callback=cmd_refresh_node, pipe=None,),
		]),
		T_Keyword('all', help='Refresh all configured nodes', callback=cmd_refresh_all_nodes, pipe=None,),
	]),
	
	T_KeywordWithHook('rename', help='' ,
	
		hook_name='configuration:tenant:rename'
		, 
		subtree=[

		
		
		T_Keyword('agent', help='', subtree=[
			T_Agentcompleter(help='Existing agent name', subtree=[
				T_ValidWord     (VALID_AGENT_NAME, help='New agent name', callback=cmd_agent_rename, pipe=None,)
			]),
		]),
	
		T_Keyword('cluster', help='', subtree=[
			T_Clustercompleter(help='Existing cluster',  subtree=[ #
				T_Anyword (help='New cluster name', callback=cmd_cluster_rename, pipe=None,)
			]),
		]),
		
		T_Keyword('node', help='', subtree=[
			T_MonitoredNodecompleter(help='', subtree=[ #
				T_ValidWord (VALID_NODE_NAME, help='Name', callback=cmd_rename_node, pipe=None,),
			]),
		]),

		T_Keyword('link', help='Rename link', subtree=[
			T_LinkCompleter(help='', token_repr='<LINK NAME>', subtree=[ #
				T_ValidWord (VALID_LINK_NAME, token_repr='<LINK NAME>', help='Name', callback=cmd_rename_link, pipe=None,),
			]),
		]),

		T_Keyword('datagroup-template', help='', subtree=[
			T_DataGroupTemplate(help='', pipe=None, subtree=[
				T_ValidWord(VALID_DATAGROUP_NAME, help='', subtree=[
					T_Keyword('fix-logs'   , help='', callback=cmd_rename_datagroup_template, pipe=None),
					T_Keyword('no-fix-logs', help='', callback=cmd_rename_datagroup_template, pipe=None),
				]) 
			])
		]),

		T_Keyword('node-template', help='', subtree=[
			T_NodeTemplate(help='', pipe=None, subtree=[
				T_ValidWord(VALID_TEMPLATE_NAME, help='', callback=cmd_rename_node_template, pipe=None),
			])
		]),
		T_Keyword('interface-template', help='', subtree=[
			T_IfaceTemplate(help='', pipe=None, subtree=[
				T_ValidWord(VALID_TEMPLATE_NAME, help='', callback=cmd_rename_interface_template, pipe=None),
			])
		]),
		T_Keyword('storage-template', help='', subtree=[
			T_StorageTemplate(help='', pipe=None, subtree=[
				T_ValidWord(VALID_TEMPLATE_NAME, help='', callback=cmd_rename_storage_template, pipe=None),
			])
		]),
		T_Keyword('service-template', help='', subtree=[
			T_ServiceTemplate(help='', pipe=None, subtree=[
				T_ValidWord(VALID_TEMPLATE_NAME, help='', callback=cmd_rename_service_template, pipe=None),
			])
		]),
		T_Keyword('device-template', help='', subtree=[
			T_DeviceTemplate(help='', pipe=None, subtree=[
				T_ValidWord(VALID_TEMPLATE_NAME, help='', callback=cmd_rename_device_template, pipe=None),
			])
		]),



	])
	,
	mt_suspend
	,
	mt_resume
	,
	mt_restart
	,
	T_KeywordWithHook('no', help='Remove or disable configuration',
	
		hook_name='configuration:tenant:no'
		,
		subtree=[
	
		#T_Keyword('group', help='Unauthorize a group', subtree=[
		#	T_Groupcompleter(help='Existing group', subtree=[
		#	T_Groupcompleter(help='Confirm group', callback=cmd_tenant_remove_group,),
		#	]),
		#]),

		T_Keyword('agent', help='Remove a monitoring agent', subtree=[
			T_Agentcompleter(subtree=[
			T_Agentcompleter(help='Existing agent name', callback=cmd_agent_remove,),
			]),
		]),

		T_Keyword('scheduling-group', help='Remove a scheduling group', subtree=[
			T_SchedulingGroupcompleter(subtree=[
				T_SchedulingGroupcompleter(help='Existing agent name', callback=cmd_schedulinggroup_config_prompt, subtree=[
					T_Keyword('force', help='', callback=cmd_schedulinggroup_config_prompt)
				]),
			]),
		]),
		
		T_Keyword('cluster', help='Remove a cluster', subtree=[
			T_Clustercompleter(subtree=[
			T_Clustercompleter(help='Existing cluster name', callback=cmd_cluster_remove,),
			]),
		]),
		T_Keyword('node', help='Remove a node', subtree=[
			T_MonitoredNodecompleter(subtree=[
			T_MonitoredNodecompleter(help='Node name', callback=cmd_node_remove,),
			]),
		]),
		T_Keyword('link', help='Remove a link', subtree=[
			T_LinkCompleter(subtree=[
			T_LinkCompleter(help='link name', callback=cmd_remove_link,)
			])
		]),
		T_Keyword('datagroup-template', help='Delete a datagroup temaplte', subtree=[
			T_DataGroupTemplate(help='datagroup template name', pipe=None, subtree=[
				T_DataGroupTemplate(help='confirm datagroup tempalte name', callback=cmd_datagroup_template_config_prompt, pipe=None, subtree=[
					T_Keyword('force', help='', callback=cmd_datagroup_template_config_prompt)				
				])
			]),
		]),
		T_Keyword('priority-level', help='Delete a priority level', pipe=None, subtree=[
			T_Anyword(help='Label', pipe=None, subtree=[
				T_Anyword(help='Label', pipe=None, callback=cmd_del_prioritylevel)
			])
		]),
			
		
		T_Keyword('node-template'     , help='', subtree=[
			T_NodeTemplate(help='', pipe=None, subtree=[
			T_NodeTemplate(help='', callback=cmd_node_template_remove, pipe=None,)
			]),
		]),
		T_Keyword('interface-template', help='', subtree=[
			T_IfaceTemplate(help='', pipe=None, subtree=[
			T_IfaceTemplate(help='', callback=cmd_interface_template_remove, pipe=None,)
			]),
		]),
		T_Keyword('storage-template'  , help='', subtree=[
			T_StorageTemplate(help='', pipe=None, subtree=[
			T_StorageTemplate(help='', callback=cmd_storage_template_remove, pipe=None,)
			]),
		]),
		T_Keyword('service-template'  , help='', subtree=[
			T_ServiceTemplate(help='', pipe=None, subtree=[
			T_ServiceTemplate(help='', callback=cmd_service_template_remove, pipe=None,)
			]),
		]),
		T_Keyword('device-template'   , help='', subtree=[
			T_DeviceTemplate(help='', pipe=None, subtree=[
			T_DeviceTemplate(help='', callback=cmd_device_template_remove, pipe=None,)
			]),
		]),
	]
	),

])

def cmd_tenant_config_prompt_direct(cli):
	# conf tenant <name>
	tenant_name = cli.canonical_words[2]
	
	_conf_tenant(cli, tenant_name)

def cmd_tenant_config_default_prompt(cli):

	tenant_name = cli.env['CURRENT_TENANT']

	if tenant_name == None:
		cli.warning("Default tenant not selected\n")
		return

	_conf_tenant(cli, tenant_name)
	

def cmd_tenant_create_and_config_prompt(cli):

	# tenant <name> 
	name = cli.canonical_words[1]
	
	_create_and_config(cli, name, create_new=True)


def cmd_tenant_config_prompt(cli):

	# tenant <name>
	name = cli.canonical_words[1]

	_create_and_config(cli, name)


def _create_and_config(cli, name, create_new=False):
	
	created = False
	try:
		tenant = tenant_get(name)
		if tenant == None:
			if create_new:
				
				cli.info("Creating tenant %s" % (name ))
	
				tenant = tenant_create(name)
				created = True
				
			else:
				cli.error("Tenant not found")

	except Exception, e:
		cli.error("%s %s %s" % (type(e), str(e), traceback.format_exc()))

		tenant_remove(name, full=False)
		return

	_conf_tenant(cli, tenant.name)
	

def _conf_tenant(cli, tenant_name):
	
	if not cli.env.get('PARSE_ONLY', False):
	
		tenant = tenant_get(tenant_name)

		env = cli.env.clone()

		cli.info( "Entering configuration tenant: %s" % (tenant.name) )

		#env['PROMPT'] = 'config-%s#' % ( PROMPT_SEP.join([ tenant.name ]) )
		
		env['PROMPT'] = '%s' % ( PROMPT_SEP.join([ tenant.name ]) )
		
		env['MODE'] = 'configuration'

		env['CONFIG_TENANT'] = tenant.name

		# With this setting "show" command will work on the same tenant.
		env['CURRENT_TENANT'] = tenant.name

		cli.info( "Current tenant is now: %s" % (tenant.name) )

	else:
		env = cli.env.clone()

	cmds = SORTED_KEYWORDS(mt_config_tenant + cli_get_registered_menus('configuration:tenant') \
	       ) \
	       + \
	       [BLANK_ENTRY] \
	       + \
	       SORTED_KEYWORDS(cli_get_registered_menus('__common_commands__') \
	       + \
	       cli_get_registered_menus('__show_commands__') \
	)

	#newcli = cli.get_class()(cmds, env, debug_interpreter=env['DEBUG'])
	#newcli.cmdloop()
	
	cli.enter_subcli(cmds, env)
	

	if not cli.env.get('PARSE_ONLY', False):
		cli.info( "Restoring current tenant: %s" % (cli.env['CURRENT_TENANT']) )






