# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import os
import sanet_common
import sanet_poller

from django.core import management
from django.utils.translation import ugettext as _


import core
from core.api.base import tenant_select_all
from core.api.base import db_set_current_tenant_schema
from core.api.base import link_select_all

from core.api.cache import init_cache_manager_from_conf

from server.sanetd.sanetdlogic import calculate_late_checks

from cli.clilib.ioslike import T_Keyword, T_Number, T_Line

from cli.decorators import cli_admin_command

from cli.tokens import T_KeywordWithHook

from cli.tokens import *  
from cli.tokens import T_Tenantcompleter
from cli.tokens import T_SchedulingGroupcompleter

from cli.prompts.show.users    import cmd_show_users, cmd_show_groups, cmd_show_users_permissions, cmd_show_groups_permissions

from cli.prompts.show.tenants  import *
from cli.prompts.show.agents   import *
from cli.prompts.show.clusters import *
from cli.prompts.show.nodes    import *
from cli.prompts.show.ifaces   import *
from cli.prompts.show.storages import *
from cli.prompts.show.services import *
from cli.prompts.show.devices  import *

from cli.prompts.show.datagroups import *

from cli.prompts.show.datagroup_templates import cmd_show_datagroup_templates

from cli.prompts.show.configuration import *

from cli.prompts.show.templates import *
from cli.prompts.show.templates import cmd_show_element_templates

from cli.prompts.show.logs import *

from cli.prompts.show.prioritylevels import cmd_show_priority_levels

from cli.prompts import cli_get_registered_hooks

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def cmd_show_info(cli):

	# show info [full]

	user = cli.env['CURRENT_USER']

	cols = cli.env['COLS']

	env = cli.env

	show_full = len(cli.canonical_words) > 2

	cli.prnt(   ""   )
	cli.prnt( _("SANET 3 SYSTEM INFORMATIONS") )
	cli.prnt(   "==========================="  )
	cli.prnt(   ""   )

	if user and user.is_superuser:
		r = [[
			[_('USERS' ), 12],
			[_('GROUPS'), 5],
		]]
		r.append([ 
			User.objects.count(), 
			Group.objects.count() 
		])
		cli.prnt(text_table(r, cols))

	r = [[
		[_('TENANT'     ), 12],
		[_('AGENTS'     ), 5],
		[_('NODES'      ), 5],
		[_('IFACES'     ), 5],
		[_('STORAGES'   ), 5],
		[_('SERVICES'   ), 5],
		[_('DEVICES'    ), 5],
		[_('LINKS'      ), 5],
		[_('CLUSTER'    ), 5],
		[_('DATAGROUPS' ), 5],
		[_('DATASOURCES'), 5],
		[_('CONDITIONS' ), 5],
		[_('TIMEGRAPHS' ), 5],
	]]

	if user:
		tenants = user.get_profile().get_visible_tenants().all()
	else:
		tenants = tenant_select_all().all()

	for t in tenants:

		db_set_current_tenant_schema(t.name)

		cursor = connection.cursor()
		cursor.execute("SELECT kind, count(*) as count FROM monitoredelement GROUP BY kind")
		rows = cursor.fetchall()

		kinds = {
			ELEMENT_KIND_NODE   : 0,
			ELEMENT_KIND_IFACE  : 0,
			ELEMENT_KIND_STORAGE: 0,
			ELEMENT_KIND_SERVICE: 0,
			ELEMENT_KIND_DEVICE : 0,
		}
		kinds.update( dict( rows ) )

		r.append([
			t.name,
			int(Agent.objects.count()),
			kinds[ELEMENT_KIND_NODE],
			kinds[ELEMENT_KIND_IFACE],
			kinds[ELEMENT_KIND_STORAGE],
			kinds[ELEMENT_KIND_SERVICE],
			kinds[ELEMENT_KIND_DEVICE],
			int(Link.objects.count()),
			Cluster.objects.count(),
			int(DataGroup.objects.count()),
			int(DataSource.objects.count()),
			int(Condition.objects.count()),
			int(TimeGraph.objects.count()),
		])

	cli.prnt(text_table(r, cols))

	hooks = cli_get_registered_hooks('show:info')
	for h in hooks:
		h(cli)

	if show_full:
		cli.prnt( _("Environment") )
		cli.prnt( "-----------" )
		cli.prnt( "" )
		cli.prnt( "\tINSTALL DIR: %s" % (settings.BASE_DIR) )
		cli.prnt( "\tSANET_HOME : %s" % (os.environ.get('SANET_HOME','-not set-') ) )
		cli.prnt( "" )

	if show_full:
		cli.prnt( _("Installation info") )
		cli.prnt( "-----------------" )
		cli.prnt( "" )
		cli.prnt( "\tVersion        : %s" % (settings.SANET_VERSION_FULL_VERSION))
		cli.prnt( "\tServer  version: %s" % (core.__version__ ) )
		cli.prnt( "\tCommon lib ver.: %s" % (sanet_common.__version__  ) )
		cli.prnt( "\tAgent version  : %s" % (sanet_poller.__version__ ) )
		cli.prnt( "" )

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def cmd_show_late(cli):

	# 0    1    2    3  4
	# show info late by (datagroup|agent|node)

	cols        = cli.env['COLS']

	tenant_name = cli.env['CURRENT_TENANT']
	tenant = tenant_get(tenant_name)

	db_set_current_tenant_schema(tenant.name)

	args = cli.canonical_words[4:]
	mode = 'datagroup'
	if len(args) > 0:
		mode = args[0]

	cache_manager = init_cache_manager_from_conf()

	late_datagroups = calculate_late_checks(time.time(), cache_manager) 

	late_by_uuid = dict([ (uuid, lastdone) for uuid, path, lastdone in late_datagroups ])

	datagroups = DataGroup.objects.filter(uuid__in=late_by_uuid.keys()).select_related('element', 'element__parent', 'agent')

	last = ''

	r = []
	if mode == 'datagroup':
		
		datagroups = datagroups.extra(select={'path2':"replace(datagroup.path,';','1')"}, order_by=['path2'])

		r.append([
			[_('DATAGROUP'), 30],
			[_('ELEMENT'  ), 10],
			[_('AGENT'    ), 10],
			[_('LASTDONE' ), 10],
		])

		for dg in datagroups :
			path = dg.path if dg.path != last else ''
			last = dg.path

			r.append([
				path,
				dg.element.path,
				dg.agent.name,
				timestamp_to_isostr( late_by_uuid[dg.uuid] ),
			])

	elif mode == 'agent':

		datagroups = datagroups.extra(select={'path2':"replace(datagroup.path,';','1')"}, order_by=['agent','path2'])

		r.append([
			[_('AGENT'     ), 30],
			[_('DATAGROUP' ), 10],
			[_('ELEMENT'   ), 10],
			[_('LASTDONE'  ), 10],
		])

		for dg in datagroups :
			name = dg.agent.name if dg.agent.name != last else ''
			last = dg.agent.name
			r.append([
				name,
				dg.path,
				dg.element.path,
				timestamp_to_isostr( late_by_uuid[dg.uuid] ),
			])

	elif mode == 'element':

		datagroups = datagroups.extra(select={'path2':"replace(datagroup.path,';','1')"}, order_by=['agent','path2'])

		r.append([
			[_('ELEMENT'   ), 30],
			[_('DATAGROUP' ), 10],
			[_('AGENT'     ), 10],
			[_('LASTDONE'  ), 10],
		])

		for dg in datagroups :
			path = dg.element.path if dg.element.path != last else ''
			last = path
			
			r.append([
				path,
				dg.path,
				dg.agent.name,
				timestamp_to_isostr( late_by_uuid[dg.uuid] ),
			])
		
	cli.prnt(text_table(r, cols))
		
#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def cmd_show_links(cli):
	
	user        = cli.env['CURRENT_USER']

	cols        = cli.env['COLS']
	tenant_name = cli.env['CURRENT_TENANT']

	links = link_select_all(tenant_name=tenant_name, user=user)
	
	r = [[
		[_('NODE1'), 30],
		[_('IFACE1'), 10],
		[_('NODE2'), 30],
		[_('IFACE2'), 10],
		[_('NAME'), 30],
		[_('LEVEL'), 10],
		[_('DESCR'), 10],
	]]
	for l in links:
		r.append([
			l.node1.name, 
			l.iface1.name,
			l.node2.name,
			l.iface2.name,
			l.name,
			LINK_TYPE_TO_CLI[l.ltype],
			l.description,
		])
	cli.prnt(text_table(r, cols))

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def show_ip(cli):

	tenant_name = cli.env['CURRENT_TENANT']
	tenant = tenant_get(tenant_name)

	db_set_current_tenant_schema(tenant.name)

	# show ip nodes a b c ...
	# show ip allnodes

	from sanet_common.netutils import resolve_ip

	details = []
	nodes   = {}
	names   = []

	if cli.canonical_words[2] == 'allnodes':
		for n in MonitoredNode.objects.all():
			names.append(n.name)
	else:
		names = cli.canonical_words[3].split(' ')

	for name in names:
		v4 = []
		try:
			v4 = resolve_ip(name, 4)
		except Exception, e:
			cli.warning( _("Problems while resolving '%(name)s': %(error)s") % {'name':name, 'error':e} )

		details.append([name, ' '.join(v4)])
		for i4 in v4:
			if not nodes.has_key(i4):
				nodes[i4]=[name]
			else:
				nodes[i4].append(name)
	if details:
		cli.prnt(text_details(details))

	for i4 in nodes.keys():
		if len(nodes[i4]) > 1:
			cli.prnt( _("\nOVERLAP: %(ip)s %(names)s\n") % {'ip':i4, 'names': ' '.join(nodes[i4]) }  )

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#


@cli_admin_command
def cmd_broker_info_queues(cli):
	management.call_command('broker_info', 'queues', verbosity=0, interactive=False)

@cli_admin_command
def cmd_cache_info(cli):
	management.call_command('cache_info', verbosity=0, interactive=False)

@cli_admin_command
def cmd_cache_keys(cli):

	# show system cache keys [* [sorted]]
	if len(cli.canonical_words) >= 5:
		pattern = cli.canonical_words[4]
	else:
		pattern = '*'
		
	sorted = ''
	if len(cli.canonical_words) >= 6:
		sorted = cli.canonical_words[5]
		
	management.call_command('cache_info', 'keys', pattern, sorted, verbosity=0, interactive=False)

@cli_admin_command
def cmd_db_usage(cli):
	management.call_command('dbusage', verbosity=0, interactive=False)


@cli_admin_command
def cmd_sanetd_info(cli):
	# show system sanetd [args]
	
	args = cli.canonical_words[3:]
	if args:
		args = args[0].split('\n')
	else:
		args = ['info']
	
	
	management.call_command('sanetdinfo', *args, verbosity=0, interactive=False)

@cli_admin_command
def cmd_queries(cli):
	
	for q in connection.queries:
		print q['sql']

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def cmd_show_lastchanged(cli):

	tenant_name = cli.env['CURRENT_TENANT']
	tenant = tenant_get(tenant_name)
	
	db_set_current_tenant_schema(tenant.name)
	
	words = cli.canonical_words[2:]
	# show lastchanged
	# show lastchanged seconds <n>
	
	seconds = 60
	if len(words) > 0:
		seconds = int(words[1])
		
	mtime = time.time() - seconds
	
	cli.prnt( _("Elements changed after: %(time)s") % {'time': timestamp_to_isostr(mtime) } )
	
	for c in Cluster.objects.filter(mtime__gt=mtime).order_by('mtime').all():
		cli.prnt( "%s %s %s" % (timestamp_to_isostr(c.mtime), "cluster", c) )
	
	for n in MonitoredElement.objects.filter(mtime__gt=mtime).order_by('kind', 'mtime').all():
		cli.prnt( "%s %s %s" % (timestamp_to_isostr(n.mtime), "node", n) )
	
	for dg in DataGroup.objects.filter(mtime__gt=mtime).order_by('mtime').all():
		cli.prnt( "%s %s %s" % (timestamp_to_isostr(dg.mtime), "datagroup", dg.path) )
	
	for e  in DataGroupParameter.objects.filter(mtime__gt=mtime).order_by('mtime').all(): cli.prnt("%s %s %s" % (timestamp_to_isostr(e.mtime) , "param"      , ("\t" + str(e))) )
	for e  in DataSource.objects.filter(mtime__gt=mtime).order_by('mtime').all()        : cli.prnt("%s %s %s" % (timestamp_to_isostr(e.mtime) , "datasource" , ("\t" + str(e.path))) )
	for e  in Condition.objects.filter(mtime__gt=mtime).order_by('mtime').all()         : cli.prnt("%s %s %s" % (timestamp_to_isostr(e.mtime) , "condition"  , ("\t" + str(e.path))) )
			
	for e in TimeGraph.objects.filter(mtime__gt=mtime).order_by('mtime').all(): 
		cli.prnt( "%s %s %s" % (timestamp_to_isostr(e.mtime), "timegraph", ("\t" + (str(e)))) )
			
	for tgs in TimeGraphSerie.objects.filter(mtime__gt=mtime).order_by('mtime').all():
		cli.prnt( "%s %s %s" % (timestamp_to_isostr(tgs.mtime), "serie", ("\t" + (str(tgs)))) )
	
	return

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

from core.models import SchedulingGroup

def cmd_show_scheduling_group(cli):
	
	# show scheduling-group [all]

	user           = cli.env['CURRENT_USER']

	current_tenant = cli.env.get('CURRENT_TENANT', None)
	current_user   = cli.env.get('CURRENT_USER'  , None)
	cols           = cli.env['COLS']

	tenant = tenant_get(current_tenant)

	tenant.db_set_current_schema()

	r = [[
		[_('UUID'        ), -4],
		[_('NAME'        ), 10],
		[_('DESCRIPTION' ), 20],
		[_('MIN GAP'     ), 10],
		[_('MAX PARALLEL'), 10],
		[_('MAX DEADLINE'), 10],
	]]
	
	show_all = len(cli.canonical_words) > 2 and cli.canonical_words[2] == 'all'

	sgs = SchedulingGroup.objects.filter(tenant=tenant)
	
	if not show_all:
		sgs = sgs.filter(is_real=True)
	
	sgs = sgs.order_by('name')

	for sg in sgs:

		r.append([
			sg.uuid,

			sg.name,

			sg.description,

			sg.min_gap      if sg.min_gap      is not None else '',
			sg.max_parallel if sg.max_parallel is not None else '',
			sg.max_deadline if sg.max_deadline is not None else '',
		])

	output_table(cli, r)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

show_commands = [

	T_KeywordWithHook('show'  , help=_('Show running system information'), 
	
		hook_name='show',
		
		subtree=SORTED_KEYWORDS([	

			#T_Keyword('lastchanged'  , help=_('Show last configured entities'), callback=cmd_show_lastchanged, subtree=[
			#	T_Keyword('seconds'  , help=_('Show last configured entities'), pipe=None, subtree=[
			#		T_Anyword(help=_('Show last configured entities'), callback=cmd_show_lastchanged)
			#	])
			#]),

			T_Keyword('info'  , help=_('Show global informations'), callback=cmd_show_info, subtree=[
				T_Keyword('full'  , help='Show global informations (full data)', callback=cmd_show_info, ),
				
				#T_Keyword('late'  , help='Show late checks', callback=cmd_show_late, subtree=[
				#	T_Keyword('by', subtree=[
				#		T_Keyword('datagroup', help='Show late check by datagroup', callback=cmd_show_late),
				#		T_Keyword('element'  , help='Show late check by element'  , callback=cmd_show_late),
				#		T_Keyword('agent'    , help='Show late check by agent'    , callback=cmd_show_late),
				#	])
				#])
			]),

			T_Keyword('system' , help=_('Show internal informations'), subtree=[
				T_Keyword('cache', help='', subtree=[
					T_Keyword('info' , help='', callback=cmd_cache_info, ),
					T_Keyword('keys' , help='', callback=cmd_cache_keys, subtree=[
						T_Anyword(help='', callback=cmd_cache_keys, subtree=[
							T_Keyword('sorted' , help='', callback=cmd_cache_keys)
						]),
					]),
				]),
				T_Keyword('queues', help=_('Show internal queue system informations'), callback=cmd_broker_info_queues, ),

				T_Keyword('db', help=_('Show internal database informations'), subtree=[
					T_Keyword('usage'  , help='', callback=cmd_db_usage, ),
					T_Keyword('queries', help='', callback=cmd_queries,),
				]),

				T_Keyword('sanetd', help=_('Show central sanetd server informations'), callback=cmd_sanetd_info, subtree=[
					T_Line(help='', callback=cmd_sanetd_info, ),
				]),

			]),


			T_Keyword('user' , help=_('Show defined Users'), callback=cmd_show_users, subtree=[
				T_Usercompleter(help=_('username'), callback=cmd_show_users, subtree=[
					T_Keyword('permissions' , help=_('Show permissions/capabilities'), callback=cmd_show_users_permissions)
				]),
			]),

			T_Keyword('group' , help=_('Show defined Groups'), callback=cmd_show_groups, subtree=[
				T_Groupcompleter(help=_('username'), callback=cmd_show_groups, subtree=[
					T_Keyword('permissions' , help=_('Show permissions/capabilities'), callback=cmd_show_groups_permissions)
				]),
			]),
			
			T_Keyword('tenant', help=_('Show defined Tenants'), callback=cmd_show_tenants, subtree=[
				T_Tenantcompleter(help=_('tenant name'), callback=cmd_show_tenant_details, ),
			]),

			T_Keyword('priority-levels', help=_('Show priority levels range'), callback=cmd_show_priority_levels, ),
			
			
			T_Keyword('agent' , help=_('Show define agents'), callback=cmd_show_agents , subtree=[
				T_Agentcompleter(help=_('Show selected agent'), pipe=None, callback=cmd_show_agent, subtree=[
					T_Keyword('info'    , help=_('Show agent running internal data')  , pipe=None, callback=cmd_show_agent_info,),
					T_Keyword('schedule', help=_('Show agent scheduling informations'), pipe=None, callback=cmd_show_agent_info, subtree=[
						T_Anyword(help=_('Show the next <n> scheduled datagroups'), pipe=None, callback=cmd_show_agent_info,)
					]),
				]),
			]),

			T_Keyword('scheduling-group' , help=_('Show clusters'), callback=cmd_show_scheduling_group, subtree=[
				T_Keyword('all', help='', callback=cmd_show_scheduling_group)
			]),


			T_Keyword('cluster' , help=_('Show clusters'), callback=cmd_show_clusters, subtree=[
				T_Clustercompleter(help=_('Show selected cluster'), callback=cmd_show_clusters, ),
			]),
			
			T_KeywordWithHook('node' , help=_('Show nodes'), callback=cmd_show_nodes, 

				hook_name='show:node',

				subtree=[
			
				T_Keyword('name', help=_('Show node informations'), subtree=[
					T_MonitoredNodecompleter(help=_('Show selected node'), callback=cmd_show_nodes, 
						hook_name='show:node:name',
						subtree=[
							T_Keyword('run-state', help="", callback=cmd_show_nodes_runstate)
						]
					)
				]),
				T_Keyword('agent', help=_('Show nodes by agent'), subtree=[
					T_Agentcompleter(help=_('Show nodes assigned to the specified agent'), callback=cmd_show_nodes),
				]),
				T_Keyword('template', help=_('Show nodes by template'), subtree=[
					T_NodeTemplate(help=_('Show node by template'), callback=cmd_show_nodes,)
				]),

				T_Keyword('cluster', help=_('Show nodes by cluster'), subtree=[
					T_Clustercompleter(help=_('Show node by cluster name'), callback=cmd_show_nodes,)
				]),
				
				]
			),

			T_Keyword('interface' , help=_('Show interfaces'), callback=cmd_show_ifaces, subtree=[
				T_Keyword('node' , help=_('Show interfaces for the given node'), subtree=[
					T_MonitoredNodecompleter(help=_('Show interfaces for the given node'), callback=cmd_show_ifaces, subtree=[
						T_InlineIfaceCompleter(help=_('Show the selected interface for the given node'), callback=cmd_show_ifaces ),
					]),
				]),
				T_Keyword('template', help=_('Show interface by template'), subtree=[
					T_IfaceTemplate(help=_('Show interface by template'), callback=cmd_show_ifaces,)
				]),
			]),
			T_Keyword('storage' , help=_('Show storages'), callback=cmd_show_storages, subtree=[
				T_Keyword('node' , help=_('Show storages for the given node'), subtree=[
					T_MonitoredNodecompleter(help=_('Show storages for the given node'), callback=cmd_show_storages, subtree=[
						T_InlineStorageCompleter(help=_('Show storage informations for the given node'), callback=cmd_show_storages ),
					]),
				]),
				T_Keyword('template', help=_('Show storage by template'), subtree=[
					T_StorageTemplate(help=_('Show storage by template'), callback=cmd_show_storages,)
				]),
			]),
			T_Keyword('service' , help=_('Show services'), callback=cmd_show_services, subtree=[
				T_Keyword('node' , help=_('Show interfaces for the given node'), subtree=[
					T_MonitoredNodecompleter(help=_('Show interfaces for the given node'), callback=cmd_show_services, subtree=[
						T_InlineServiceCompleter(help=_('Show interface informations'), callback=cmd_show_services ),
					]),
				]),
				T_Keyword('template', help=_('Show service by template'), subtree=[
					T_ServiceTemplate(help=_('Show service by template'), callback=cmd_show_services,)
				]),
			]),
			T_Keyword('device' , help=_('Show devices'), callback=cmd_show_devices, subtree=[
				T_Keyword('node' , help=_('Show devices for the given node'), subtree=[
					T_MonitoredNodecompleter(help=_('Show devices for the give node'), callback=cmd_show_devices, subtree=[
						T_InlineDeviceCompleter(help=_('Show device informations'), callback=cmd_show_devices ),
					]),
				]),
				T_Keyword('template', help=_('Show device by template'), subtree=[
					T_DeviceTemplate(help=_('Show device by template'), callback=cmd_show_devices,)
				]),
			]),

			T_Keyword('link' , help=_('Show links'), callback=cmd_show_links, ),

			T_Keyword('templates'  , help=_('Show configuration templates')  , callback=cmd_show_element_templates, subtree=[

				T_Keyword('all'      , help=_('Show node templates')     , callback=cmd_show_all_templates),
			
				T_Keyword('node'     , help=_('Show node templates')     , callback=cmd_show_node_templates, subtree=[
					T_NodeTemplate(help=_('Show node template informations'), callback=cmd_show_node_templates,)
				]),
				T_Keyword('interface', help=_('Show interface templates'), callback=cmd_show_interface_templates, subtree=[
					T_IfaceTemplate(help=_('Show interface template informations'), callback=cmd_show_interface_templates,)
				]),
				T_Keyword('storage'  , help=_('Show storage templates')  , callback=cmd_show_storage_templates, subtree=[
					T_StorageTemplate(help=_('Show storage template informations'), callback=cmd_show_storage_templates,)
				]),
				T_Keyword('service'  , help=_('Show service templates')  , callback=cmd_show_service_templates, subtree=[
					T_ServiceTemplate(help=_('Show service template informations'), callback=cmd_show_service_templates,)
				]),
				T_Keyword('device'   , help=_('Show device templates')   , callback=cmd_show_device_templates, subtree=[
					T_DeviceTemplate(help=_('Show device template informations'), callback=cmd_show_device_templates,)
				]),
				T_Keyword('datagroup', help=_('Show datagroup templates'), callback=cmd_show_datagroup_templates, subtree=[
					T_DataGroupTemplate(help=_('Show datagroup template informations'), callback=cmd_show_datagroup_templates, )
				]),

				T_Keyword('used', help='', subtree=[
					T_Keyword('datagroup', help='', callback=cmd_show_datagroup_templates, subtree=[
						T_DataGroupTemplate(help='', callback=cmd_show_datagroup_templates, )
					]),
				]),

			]),

			#T_Keyword('datagroup' , help=_('Show configured datagroups'), callback=cmd_show_datagroup , subtree=[
			T_KeywordWithHook('datagroup' , help=_('Show configured datagroups'), callback=cmd_show_datagroup , 

				hook_name='show:datagroup',

				subtree=[

				T_Keyword('path'     , help=_('Show datagrop'), subtree=[
					T_DataGroupCompleter(all_path=True, help=_('Show datagroup'), callback=show_datagroup_details, subtree=[
						T_Keyword('run-state'     , help=_('Show datagrop running state'), callback=cmd_show_datagroup_details_runstate)
					]), 
				]),

				T_Keyword('node'     , help=_('Show datagroup for the give node'), subtree=[ 
					T_MonitoredNodecompleter(help="Show node's datagroups",  callback=cmd_show_datagroup_by_node) ]),
				T_Keyword('interface', help="Show interface's datagroups", subtree=[ 
					T_MonitoredNodecompleter(help="Show interface's datagroups", subtree=[ 
						T_InlineIfaceCompleter  (help="Show interface's datagroups", callback=cmd_show_datagroup_by_iface) ]) ]),
				T_Keyword('storage'  , help="Show storage's datagroups", subtree=[ 
					T_MonitoredNodecompleter(help="Show storage's datagroups", subtree=[ 
						T_InlineStorageCompleter(help="Show storage's datagroups", callback=cmd_show_datagroup_by_storage)	]) ]),
				T_Keyword('service'  , help="Show service's datagroups", subtree=[ 
					T_MonitoredNodecompleter(help="Show 's datagroups", subtree=[ 
						T_InlineServiceCompleter(help="Show service's datagroups", callback=cmd_show_datagroup_by_service) ]) ]),
				T_Keyword('device'   , help="Show device's datagroups", subtree=[ 
					T_MonitoredNodecompleter(help="Show device's datagroups", subtree=[ 
						T_InlineDeviceCompleter (help="Show device's datagroups", callback=cmd_show_datagroup_by_device) ]) ]),

				T_Keyword('template', help=_('Show datagroup by template'), subtree=[
					T_DataGroupTemplate(help='', callback=cmd_show_datagroup_by_template, subtree=[
						T_Keyword('and-subtemplates', help=_('Show datagroup by template'), callback=cmd_show_datagroup_by_template,)
					])
				]),

				T_Keyword('scheduling-group', help=_('Show datagroup by scheduling group'), subtree=[
					T_SchedulingGroupcompleter(help='', callback=cmd_show_datagroup_by_scheduling_group)
				])
			]),

			#T_Keyword('datasource', help=_('Show configured datasources'), callback=cmd_show_datasource, subtree=[
			T_KeywordWithHook('datasource' , help=_('Show configured datasources'), callback=cmd_show_datasource, 

				hook_name='show:datasource',
			
				subtree=[

				T_Keyword('id'     , help="Show configured datasource", subtree=[
					T_Anyword(help='', callback=cmd_show_datasource_details, subtree=[
						T_Keyword('data', help='', callback=cmd_show_datasource_details_data, subtree=[
							T_Number(help='', callback=cmd_show_datasource_details_data)
						]),	
						T_Keyword('run-state'    , help=_('Show datasource running state'), callback=cmd_show_datasource_details_runstate),
						T_Keyword('storeinfo'    , help=_('Show datasource storage info '), callback=cmd_show_datasource_details_storage),
					])
				]),
				T_Keyword('path'     , help="Show configured datasource", subtree=[
					T_DataSourceCompleter(all_path=True, help="Show configured datasource", callback=cmd_show_datasource_details, subtree=[
						T_Keyword('data', help='', callback=cmd_show_datasource_details_data, subtree=[
							T_Number(help='', callback=cmd_show_datasource_details_data)
						]),	
						T_Keyword('run-state'    , help=_('Show datasource running state'), callback=cmd_show_datasource_details_runstate),
						T_Keyword('storeinfo'    , help=_('Show datasource storage info '), callback=cmd_show_datasource_details_storage),
					]),
				]),
				T_Keyword('node'     , help="Show node's datasources", subtree=[
					T_MonitoredNodecompleter(help=_("Show node's datasources"), callback=cmd_show_datasource_by_node)]),
					
				T_Keyword('interface', help=_("Show interface's datasources"), subtree=[
					T_MonitoredNodecompleter(help=_("Show interface's datasources"), subtree=[
						T_InlineIfaceCompleter  (help=_("Show interface's datasources"), callback=cmd_show_datasource_by_iface)])]),
						
				T_Keyword('storage'  , help=_("Show storage's datasources"), subtree=[
					T_MonitoredNodecompleter(help=_("Show storage's datasources"), subtree=[
						T_InlineStorageCompleter  (help=_("Show storage's datasources"), callback=cmd_show_datasource_by_storage)])]),
						
				T_Keyword('service'  , help=_("Show service's datasources"), subtree=[
					T_MonitoredNodecompleter(help=_("Show service's datasources"), subtree=[
						T_InlineServiceCompleter  (help=_("Show service's datasources"), callback=cmd_show_datasource_by_service)])]),
						
				T_Keyword('device'   , help=_("Show device's datasources" ), subtree=[
					T_MonitoredNodecompleter(help=_("Show device's datasources" ), subtree=[
						T_InlineDeviceCompleter   (help=_("Show device's datasources" ), callback=cmd_show_datasource_by_device)])]),


				T_Keyword('template', help=_('Show datasources by datagroup template'), subtree=[
					T_DataGroupTemplate(help='', callback=cmd_show_datasource_by_template, subtree=[
						T_Keyword('and-subtemplates', help=_('Show datasources by datagroup template'), callback=cmd_show_datasource_by_template,)
					])
				]),

				T_Keyword('state'     , help="Show datasources by state", subtree=[
				
					T_Alternatives([ 
							('ok'         , 'Successfull'),
							('uncheckable', 'Uncheckable'),
							('strangeval' , 'Strange values (outside min/max or invalid)'),
							('cascade'    , 'Ignored datasource because of cascade option'),
							('error'      , 'Error'),
					], callback=cmd_show_datasource_by_state),
					#T_Keyword('sv'     , help="Show datasources by state", callback=cmd_show_datasource_by_state) 
				])

			]),
			
			#T_Keyword('condition' , help=_('Show conditions'), callback=cmd_show_condition , subtree=[
			T_KeywordWithHook('condition' , help=_('Show conditions'), callback=cmd_show_condition , 

				hook_name='show:condition',

				subtree=[

				T_Keyword('status'     , help='', subtree=[
					T_Alternatives([ 
							('down'       , 'Show down conditions'),
							('up'         , 'Show up conditions'),
							('suspended'  , 'Show suspended conditions'),
							('uncheckable', 'Show uncheckable conditions'),
						],
						'status'     , help=_('Show condition with the given status'), callback=cmd_show_condition_by_status
					),
				]),

				T_Keyword('priority'    , help='', subtree=[
					T_Keyword('undefined'     , help='', callback=cmd_show_condition_by_value ),
					
					T_Keyword('value'     , help='', subtree=[
						T_ValidPriorityValue(help='', callback=cmd_show_condition_by_value),
					]),
					T_Keyword('minimun'     , help='', subtree=[
						T_Number(help='', callback=cmd_show_condition_by_min_priority),
					]),
					T_Keyword('level', help=_('Delete a priority level'), pipe=None, subtree=[
						T_PriorityLevelCompleter(help=_('Label'), pipe=None, callback=cmd_show_condition_by_priority_level)
					]),
				]),
			
				T_Keyword('path'     , help='', subtree=[
					T_ConditionCompleter    (all_path=True, help="Show condition", callback=cmd_show_condition_details, subtree=[
						T_Keyword('run-state'     , help=_('Show condition running state'), callback=cmd_show_condition_details_runstate)
					]),
				]),
					
				T_Keyword('node'     , help='', subtree=[
					T_MonitoredNodecompleter(help="Show node's conditions", callback=cmd_show_condition_by_node)]),
					
				T_Keyword('interface', help="Show interface's conditions", subtree=[
					T_MonitoredNodecompleter(help="Show interface's conditions", subtree=[
						T_InlineIfaceCompleter    (help="Show interface's conditions", callback=cmd_show_condition_by_iface)])]),
						
				T_Keyword('storage'  , help="Show storage's conditions", subtree=[
					T_MonitoredNodecompleter(help="Show storage's conditions", subtree=[
						T_InlineStorageCompleter  (help="Show storage's conditions", callback=cmd_show_condition_by_storage)])]),
						
				T_Keyword('service'  , help="Show service's conditions", subtree=[
					T_MonitoredNodecompleter(help="Show service's conditions", subtree=[
						T_InlineServiceCompleter  (help="Show service's conditions", callback=cmd_show_condition_by_service)])]),
						
				T_Keyword('device'   , help="Show device's conditions", subtree=[
					T_MonitoredNodecompleter(help="Show device's conditions", subtree=[
						T_InlineDeviceCompleter   (help="Show device's conditions", callback=cmd_show_condition_by_device)])]),
						
				T_Keyword('template', help=_('Show conditions by datagroup template'), subtree=[
					T_DataGroupTemplate(help='', callback=cmd_show_conditions_by_template, subtree=[
						T_Keyword('and-subtemplates', help=_('Show conditions by datagroup template'), callback=cmd_show_conditions_by_template,)
					])
				]),
						
			]),

			#T_Keyword('timegraph' , help=_('Show timegraphs'), callback=cmd_show_timegraph  , subtree=[
			T_KeywordWithHook('timegraph' , help=_('Show timegraphs'), callback=cmd_show_timegraph  , 

				hook_name='show:timegraph',

				subtree=[


				T_Keyword('path'     , help=_('Show timegraph'), subtree=[
					T_TimeGraphCompleter    (all_path=True, help=_('Show timegraph'), callback=cmd_show_timegraph_details), ]),
			
			
				T_Keyword('node'     , help='', subtree=[
					T_MonitoredNodecompleter(help="Show node's timegraphs", callback=cmd_show_timegraph_by_node)]),
					
				T_Keyword('interface', help="Show interface's timegraphs", subtree=[
					T_MonitoredNodecompleter(help="Show interface's timegraphs", subtree=[
						T_InlineIfaceCompleter    (help="Show interface's timegraphs", callback=cmd_show_timegraph_by_iface)])]),
						
				T_Keyword('storage'  , help="Show storage's timegraphs", subtree=[
					T_MonitoredNodecompleter(help="Show storage's timegraphs", subtree=[
						T_InlineStorageCompleter  (help="Show storage's timegraphs", callback=cmd_show_timegraph_by_storage)])]),
						
				T_Keyword('service'  , help="Show service's timegraphs", subtree=[
					T_MonitoredNodecompleter(help="Show service's timegraphs", subtree=[
						T_InlineServiceCompleter  (help="Show service's timegraphs", callback=cmd_show_timegraph_by_service)])]),
						
				T_Keyword('device'   , help="Show device's timegraphs", subtree=[
					T_MonitoredNodecompleter(help="Show device's timegraphs", subtree=[
						T_InlineDeviceCompleter   (help="Show device's timegraphs", callback=cmd_show_timegraph_by_device)])])
			]),
			T_Keyword('timegraphserie', help=_('Show timegraph series'), callback=cmd_show_timegraphserie, subtree=[

				T_Keyword('datagroup'  , help="", subtree=[
					T_DataGroupCompleter    (all_path=True, help="", callback=cmd_show_timegraphserie_by_datagroup ) ]), 

				T_Keyword('node'     , help='', subtree=[
					T_MonitoredNodecompleter(help=_('Show timegraph series'), callback=cmd_show_timegraphserie_by_node)]),
				T_Keyword('interface', help='', subtree=[
					T_MonitoredNodecompleter(help='', subtree=[
						T_InlineIfaceCompleter    (help=_('Show timegraph series'), callback=cmd_show_timegraphserie_by_iface)])]),
				T_Keyword('storage'  , help='', subtree=[
					T_MonitoredNodecompleter(help='', subtree=[
						T_InlineStorageCompleter  (help=_('Show timegraph series'), callback=cmd_show_timegraphserie_by_storage)])]),
				T_Keyword('service'  , help='', subtree=[
					T_MonitoredNodecompleter(help='', subtree=[
						T_InlineServiceCompleter  (help=_('Show timegraph series'), callback=cmd_show_timegraphserie_by_service)])]),
				T_Keyword('device'   , help='', subtree=[
					T_MonitoredNodecompleter(help='', subtree=[
						T_InlineDeviceCompleter   (help=_('Show timegraph series'), callback=cmd_show_timegraphserie_by_device)])]),

			]),
			
			T_KeywordWithHook('log' , help="Show log", callback=cmd_show_logs, 
			
				hook_name='show:log',
			
				subtree=[
		
				T_Keyword('id', help="Show log entry by ID", subtree=[
					T_Anyword(help="Show log entry with the given ID", callback=cmd_show_log_by_id)
				]),
		
				T_Keyword('last', help="Show last <n> logs", subtree=[
					T_Number(help="Show last <n> logs", callback=cmd_show_logs)
				]),

		
				T_Keyword('node'     , help="Show node's log", subtree=[ 
					T_MonitoredNodecompleter(help="Show node's log", callback=cmd_show_node_logs, subtree=[
						T_Keyword('last', help="Show last <n> logs of the given node", subtree=[
							T_Number(help="Show last <n> logs of the given node", callback=cmd_show_node_logs)
						]),
					]) 
				]),
				T_Keyword('interface', help="Show interface's log", subtree=[ 
					T_MonitoredNodecompleter(help="Show interface's log", subtree=[ 
						T_InlineIfaceCompleter  (help="Show interface's log", callback=cmd_show_interface_logs, subtree=[
							T_Keyword('last', help="Show last <n> logs of the given interface", subtree=[
								T_Number(help="Show last <n> logs of the given interface", callback=cmd_show_interface_logs)
							]),
						]) 
					]) 
				]),
				T_Keyword('storage'  , help="Show storage's log", subtree=[ 
					T_MonitoredNodecompleter(help="Show storage's log", subtree=[ 
						T_InlineStorageCompleter(help="Show storage's log", callback=cmd_show_storage_logs, subtree=[
							T_Keyword('last', help="Show last <n> logs of the given storage", subtree=[
								T_Number(help="Show last <n> logs of the given storage", callback=cmd_show_storage_logs)
							]),
						])
					]) 
				]),
				T_Keyword('service'  , help="Show service's log", subtree=[ 
					T_MonitoredNodecompleter(help="Show service's log", subtree=[ 
					T_InlineServiceCompleter(help="Show service's log", callback=cmd_show_service_logs, subtree=[
							T_Keyword('last', help="Show last <n> logs of the given service", subtree=[
								T_Number(help="Show last <n> logs of the given service", callback=cmd_show_service_logs)
							]),
						]) 
					]) 
				]),
				T_Keyword('device'   , help="Show device's log", subtree=[ 
					T_MonitoredNodecompleter(help="Show device's log", subtree=[ 
						T_InlineDeviceCompleter (help="Show device's log", callback=cmd_show_device_logs, subtree=[
							T_Keyword('last', help="Show last <n> logs of the given device", subtree=[
								T_Number(help="Show last <n> logs of the given device", callback=cmd_show_device_logs)
							]),
						]) 
					]) 
				]),
				T_Keyword('datagroup'     , help="Show datagrop", subtree=[
					T_DataGroupCompleter(help="Show datagroup", callback=cmd_show_datagroup_logs, subtree=[
							T_Keyword('last', help="Show last <n> logs of the given datagroup", subtree=[
								T_Number(help="Show last <n> logs of the given datagroup", callback=cmd_show_datagroup_logs)
							]),
						]), 
				]),
				T_Keyword('condition'     , help="Show condition", subtree=[
					T_ConditionCompleter(help="Show condition", callback=cmd_show_condition_logs, subtree=[
							T_Keyword('last', help="Show last <n> logs of the given datagroup", subtree=[
								T_Number(help="Show last <n> logs of the given datagroup", callback=cmd_show_condition_logs)
							]),
					]), 
				]),
				
				T_Keyword('status'     , help='', subtree=[
					T_Alternatives([ 
							('down'       , 'Show down conditions'),
							('up'         , 'Show up conditions'),
							('suspended'  , 'Show suspended conditions'),
							('uncheckable', 'Show uncheckable conditions'),
						],
						'status'     , help=_('Show log with the given status'), callback=cmd_show_status_logs,  subtree=[
							T_Keyword('last', help="Show last <n> logs of the given datagroup", subtree=[
								T_Number(help="Show last <n> logs of the given datagroup", callback=cmd_show_status_logs)
							]),
						]
					),
				]),				
			]),

			T_KeywordWithHook('ip', help=_('IPv4 and IPv6 addresses via DNS'),
				subtree=[
					T_Keyword('allnodes', help=_('Resolved IP addresses for all the node names'),
						callback=show_ip,
					),
					T_Keyword('nodes', help=_('Resolved IP addresses for specific nodes or names'),
						subtree=[
							T_Line(help=_('Name to resolve, or quoted space-separated list of names'),
								callback=show_ip,
							),
						],
					),
				]
				,
				hook_name='show:ip'
			)
 		] 
		+ 
		show_conf_command
		)
	)
]

