#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.
#

"""

#############################################################################
conf: Gestione della configurazione Django
#############################################################################

Introduzione
------------

Il modulo *conf* contiene tutti i moduli (file) python per gestire le costanti/variabili base del sistema.

Alcune di questi file sono modificabili solo in fase di sviluppo, altri sono modificabili in fase di deploy (installazione) del sistema.

Schema degli import dei file di configurazione:

::
    
                      MODIFICABILI SOLO DURANTE                           |         MODIFICABILI
                             LO SVILUPPO                                  |      IN FASE DI DEPLOY
                                                                          |
    sanet_version.py ------------------------+                            |
                                             |                            |
                                             +---> default_settings.py ---|---> settings.py 
                                             |                            |        |
    install_dir.py   ---> default_dirs.py ---+                            |        |
                                                                          |        |
                                                                          |        +---> debug_settings.py
                                                                          |        |
                                                                          |        +---> wsgi_settings.py
                                                                          |
                                                                          |

Moduli base
---------------

.. automodule:: conf.sanet_version

.. automodule:: conf.install_dir

.. automodule:: conf.default_dirs

.. automodule:: conf.default_settings

settings
------------------------------------------------------------------------

Questo file e' il file di configurazione che **deve** essere modificato
quando si vuole cambiare una parametrizzazione di Sanet.

Il suo contento base dovrebbe essere:

::
	
	from default_settings import *
	
	#
	# Add custom settings below
	#
	


Moduli configurazione aggiuntivi
-----------------------------------

.. automodule:: conf.debug_settings

.. automodule:: conf.wsgi_settings

.. automodule:: conf.uwsgi_settings



Gestione base delle URL per l'interfaccia WEB
-------------------------------------------------

.. automodule:: conf.urls

"""
import os
import os.path
import sys

import pwd

from django.conf import settings as GLOBAL_DJANGO_SETTINGS

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def get_configuration():
	"""
	Questo metodo restituisce un oggetto per accedere a tutte le costanti di configurazione
	del sistema.
	
	Non sappiamo se la gestione della configurazione cambiera' col tempo
	quindi non chiamiamo direttamente le funzioni di django.
	"""

	#
	# IDEE: invece che gestire una configurazione "statica" basata sui file necessari a django (settings.py, ecc.)
	#       sarebbe meglio mettere molta della configurazione nel database e fare in modo che
	#       questo metodo restituisca un oggetto wrapper che legge dal DB ma si comporta
	#       esattamente come django.conf.settings.
	#

	return GLOBAL_DJANGO_SETTINGS

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def get_broker_connection_string():
	"""
	Metodo che restituisce la *connection string* per collegarsi al BROKER IPC di Sanet.
	
	Questo metodo si base sulle costanti di sistema BROKER_AUTH, BROKER_ADDRESS.
	"""
	return '%s/%s@%s' % (settings.BROKER_AUTH[0], settings.BROKER_AUTH[1], settings.BROKER_ADDRESS[0])

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def check_base_dirs(logger):

	def _check_dir(logger, path):
		if os.path.exists(path):
			return
		logger.info("Creating %s", path)
		os.makedirs(path)
		return

	_check_dir(logger, get_configuration().LOG_DIR)
	_check_dir(logger, get_configuration().PID_DIR)
	_check_dir(logger, get_configuration().DATA_DIR)

	set_dir_permissions(logger, get_configuration().DATA_DIR)

	return

def set_dir_permissions(logger, dirname):

	user = get_configuration().WEBSERVER_USER

	if not user:
		return

	try:
		userinfo = pwd.getpwnam(user)

		uid = userinfo[2]
		gid = userinfo[3]

		os.chown( dirname, uid, gid )

		if logger: logger.info( "Assign %s to %s" % ( dirname, user ) )

	except Exception, e:
		#if DEBUG: 
		if logger: logger.error("Unable to assign %s to %s. Reason: %s:%s" % (dirname, user, type(e), str(e)))

