# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""
========================================================================
:mod:`conf.urls`
========================================================================

Introduzione
--------------

Questo modulo rappresenta il *punto di ingresso* per la gestione del URL in Sanet.

Sanet prevede che le URL gestite debbano avere la seguente struttura:

	::
		
		http://<hostname>/  {{ BASE_URL_PREFIX }}  <.........> ? <parametri in GET>
		^^^^^^^^^^^^^^^^^^  ^^^^^^^^^^^^^^^^^^^^^  ^^^^^^^^^^^
		     (1)                   (2)                 (3)

La parte (1) viene gestita dal webserver (e probabilmente da modulo wsgi) che viene utilizzato
per gestire le richiestse HTTP a basso livello.

La parte (2) della URL e' la parte *riconosciuta* da Sanet. La costante *{{BASE_URL_PREFIX}}* e' configurabile da file di configurazione (vedi :py:mod:`conf.default_settings` e :py:mod:`settings`).

.. important:: {{ <BASE_URL_PREFIX> }} deve sempre contenere come ultimo carattere "/".

La parte (3) e' il resto della URL che viene riconosciuta e gestita dalla diverse sub-app di Sanet.


Gestione dei file statici e file media
-----------------------------------------

I file web *statici* (javascript, immagini e file CSS standard) vengono gestiti dal codice dell URL
di Sanet in maniera che la gestione dei file statici sia uguale sia quando Sanet e' eseguito
in modalita *DEBUG* sia quando viene eseguito normalmente.



Tutti i file statici presenti nella directory:

	::
		
		{{ SANET_INSTALL_DIR }} / {{STATIC_WEB_ROOT }} / <path>
		
		{{ SANET_INSTALL_DIR }} / {{MEDIA_WEB_ROOT }} / <path>

Sono accessibili dalla URL:

	::
		
		http://<hostname>/  {{ BASE_URL_PREFIX }} static/ <path>
		
		http://<hostname>/  {{ BASE_URL_PREFIX }} media/ <path>

.. important:: Questa gestione delle URL dei file statici ha come effetto collaterale che anche 
               i file *statici* vengono gestiti dai diversi *middleware* di django.
               
               In particolare, passando anche attraverso il *SessionMiddleware* di Django,
               l'accesso ai file statici produce accessi al DB non necessari.

.. warning:: questa gestione **e' contraria alle linee guida di Django** per la gestione dei file statici.
             
             E' stato deciso di gestire i file statici in questo modo perche' si voleva:
             
                  * ridurre al minimo le configurazioni dei webserver necessarie per gestire i file statici.
                  * avere piu' liberta' nella gestione delle URL (compresi i file statici) delle sub-app.


Calcolare la URL di un file statico
---------------------------------------

URL calcolata nel codice:

	::
		
		reverse( 'static.file.serve', kwargs={'path': '<relative path> ' )


URL calcolata nei template:

	::
		
		{% url 'static.file.serve' '<relative path> ' %}



Gestione dinamica delle URL gestite dalls sub-app
------------------------------------------------------------------------

Per semplificare il modo con cui nuove sub-app possono *estenedere* l'interfaccia
web di Sanet, le URL riconosciute come valide vengono generate:

	* In maniera "statica" per le sub-app principali
	* In maniera *dinamica* per ogni sub-app aggiunta tra le INSTALLED_APPS.

Per ogni *sub-app* che non fa parte delle {{SANET_BASE_APPS}}, se la *sub-app* conviene un file *urls.py*,
tutte le URL della *sub-app* verranno rese accessibili a partire da:

	::
		
		http://<hostname>/{{ BASE_URL_PREFIX }}{{ sub-app name }}/ .....



.. _sanet-urls-sub-app-statics:

Gestione dei file statici delle sub-app
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Le *sub-app* possono utilizzare file statici contenuti nella directory della *sub-app* stessa.

Sanet gestisce automaticamente queste URL nel seguente modo:

	::
		
		{{SANET_INSTALL_DIR}}/<subapp>/static/<relative path>  -> http://<hostname>/{{ BASE_URL_PREFIX }} <subapp>/static/<relative path> 


Calcolare le URL
++++++++++++++++++

Da codice:

	::
		
		reverse( 'static.file.<subapp>', kwargs={'path': '<relative path> ' )

Da template:

	::
		
		{% url 'static.file.<subapp>' '<relative path> ' %}

"""

import os

from django.conf import settings

from django.conf.urls import include, url

from django.conf.urls.static import static
import django.views.static

# Uncomment the next two lines to enable the admin:
from django.contrib import admin
admin.autodiscover()


# This is the list of (python) packages that contain a 'locale' directory 
# for javascript code
# 	localejs/locale/<language>/LC_MESSAGES/djangojs.po
#js_info_dict = {
#    'domain'  : 'djangojs',
#    'packages': ('localejs',),
#}


BASE_URL_PREFIX        = settings.BASE_URL_PREFIX
STATIC_ROOT            = settings.STATIC_ROOT
STATIC_WEB_ROOT        = settings.STATIC_WEB_ROOT
MEDIA_WEB_ROOT         = settings.MEDIA_WEB_ROOT
BASE_WEB_URL_PREFIX    = settings.BASE_WEB_URL_PREFIX

#--------------------------------------------------------------------------------#
#                                                                                #
#--------------------------------------------------------------------------------#

handler403 = 'webui.views.errors.custom_400_bad_request'
handler403 = 'webui.views.errors.custom_403_permission_denied'
handler404 = 'webui.views.errors.custom_404_page_not_found'
handler500 = 'webui.views.errors.custom_500_server_error'

#--------------------------------------------------------------------------------#
#                                                                                #
#--------------------------------------------------------------------------------#

#
# Dinamically add sub-app urls (if the subup defines the module)
#
# Give the <app name>, automatically load <app name>/urls.py and create :
#
#   1) A global entry point for subapp urls:
#
#      <BASE_URL_PREFIX/ <app name> / <subapp urls>
#
#   2) Static url 
#
#          <BASE_URL_PREFIX/ <app name> / static / <relative path> 
#
#      for subapp static files inside the directory:
#
#           <subapp>/static/ <relative path> 
#
#      Subapp code must access to static files using:
#
#              reverse( 'static.file.<subapp>', kwargs={'path': '<relative path> ' )
#
#              {% url 'static.file.<subapp>' '<relative path> ' %}
#

def __module_exits( module_path ):
	try:
		mod = __import__( module_path , {}, {}, [])
		return True
	except ImportError, e:
		#print __file__, module_path, str(e)
		return False

sub_app_urls = []
for app_name in settings.INSTALLED_APPS:
	# Le app base di sanet sono gia' "hardcoded" 
	if not (app_name in settings.SANET_BASE_APPS):
		urls_module = '%s.urls' % (app_name)
		if __module_exits(urls_module):
			
			#
			# FIX: Se la sub-app e' nel PYTHONPATH ma non nella directory di Sanet questa gestione NON FUNZIONA!
			#      Trovare il modo di calcolare il path della sub-app in maniera che funzioni sempre!
			#
			app_static_root      = os.path.join( settings.BASE_DIR , app_name, 'static' )
			app_static_view_name = 'static.file.%s' % (app_name)
			
			sub_app_urls.append( url(r'^%s%s/'                     % (BASE_URL_PREFIX, app_name), include(urls_module)) )

			# Changed for django 1.11
			sub_app_urls.append( url(r'^%s%s/static/(?P<path>.*)$' % (BASE_URL_PREFIX, app_name), django.views.static.serve  , {'document_root': app_static_root },  name=app_static_view_name ), )

#--------------------------------------------------------------------------------#
#                                                                                #
#--------------------------------------------------------------------------------#

#from core.views import index as core_index
import core.views

BASE_URLS = [
	# http://<hostname>/<BASE_URL_PREFIX>
	url(r'^%s$'                    % BASE_URL_PREFIX       , core.views.index, name='core.views.index'),

	# http://<hostname>/<base_url_prefix>/web/...
	url(r'^%s/'                    % BASE_WEB_URL_PREFIX   , include('webui.urls') ),        
]

BASE_URLS += [
	# http://<hostname>/<BASE_URL_PREFIX>/static/...  ->  <STATIC_ROOT>/...
	url(r'^%sstatic/(?P<path>.*)$' % BASE_URL_PREFIX  , django.views.static.serve  , {'document_root': STATIC_WEB_ROOT }, name='static.file.serve'),
	
	# http://<hostname>/<BASE_URL_PREFIX>/media/...   ->  <MEDIA_WEB_ROOT>/...
	url(r'^%smedia/(?P<path>.*)$'  % BASE_URL_PREFIX  , django.views.static.serve  , {'document_root': MEDIA_WEB_ROOT  }, name='media.file.serve' ),
]

all_urls = BASE_URLS + sub_app_urls

#--------------------------------------------------------------------------------#
#                                                                                #
#--------------------------------------------------------------------------------#

# REMOVED OFR DJANGO 1.11
#urlpatterns = patterns('',
#
#	# javascript locales used by rest javascript code
#	#url(r'^%s/jsi18n/$'% URL_PREFIX, 'django.views.i18n.javascript_catalog', js_info_dict),
#
#	# Uncomment the admin/doc line below to enable admin documentation:
#	# url(r'^admin/doc/', include('django.contrib.admindocs.urls')),    
#	# Uncomment the next line to enable the admin:
#	#url(r'^%sadmin/'               % BASE_URL_PREFIX       , include(admin.site.urls)),
#
#	*all_urls
#)  

urlpatterns = all_urls

