# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import os
import time
import sys
import pwd
import logging

from django.db import connection
#from django.db import transaction
from django.db import models
from django.db import IntegrityError
from django.db.models import Q

#from django.utils.datastructures import SortedDict  # removed since Django 1.9
from collections import OrderedDict as SortedDict

from django.conf import settings
from django.core.management.color import no_style as UNFORMATTED
from core.compatibility import reverse
from django.contrib.auth.models import User, Group

from sanet_common.sanetdefs.constants import *

from conf import get_configuration

DEBUG = get_configuration().DEBUG_CONFIG or os.environ.get('DEBUG_CONFIG',False)

from core.db_utils import transaction_commit_on_success


from core.exceptions import DatabaseException, ApplicationException
from core.exceptions import InvalidConfigurationException

from core.models import sanet_tenant_create
from core.models import sanet_tenant_create_vardir

from core.models import ELEMENT_KIND_NODE, ELEMENT_KIND_IFACE, ELEMENT_KIND_STORAGE, ELEMENT_KIND_SERVICE, ELEMENT_KIND_DEVICE, ELEMENT_KIND_STR
from core.models import LINK_TYPE_L2
from core.models import DBSetting
from core.models import TenantDependentModel
from core.models import Tenant
from core.models import TenantGroup
from core.models import TenantUserSettings
from core.models import Agent
from core.models import PriorityLabel
from core.models import Cluster
from core.models import MonitoredElement
from core.models import MonitoredNode
from core.models import MonitoredIface
from core.models import MonitoredStorage
from core.models import MonitoredService
from core.models import MonitoredDevice
from core.models import Link, LinkElement
from core.models import DataGroup,  Condition, DataSource, TimeGraph


from core.api.datasource import init_datasource_store_from_conf

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def db_version():
	return DBSetting.get_version()

def set_db_version(v):
	DBSetting.set_version(v)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def tenant_name_to_schema(tenant_name):
	return Tenant.tenant_name_to_schema(tenant_name)

def db_set_current_tenant_schema(tenant_name, include_public_schema=True, db_cursor=None):
	Tenant.db_set_current_tenant_schema(tenant_name, include_public_schema=include_public_schema, db_cursor=db_cursor)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def db_set_public_schema(db_cursor=None):

	if db_cursor == None:
		db_cursor = connection.cursor()
		
	db_cursor.execute('SET search_path TO "$user",public;' )

def get_current_schema():
	
	c = connection.cursor()
	c.execute('show search_path;')
	r = c.fetchall()
	return r[0][0]  # [ ( <searchpath> , ) , ]

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#


def create_tenant_environment(tenant):
	
	create_tenant_schema(tenant.name)

	tenant_create_datadir(tenant.uuid, tenant.name)

	datasource_store = init_datasource_store_from_conf()
	datasource_store.init_tenant(tenant)

	create_tenant_defaults(tenant)


def create_tenant_schema(tenant_name):
	
	schema_name = tenant_name_to_schema(tenant_name)

	# Create tenant schema
	cursor = connection.cursor()
	
	cursor.execute("SELECT count(*) FROM pg_namespace WHERE nspname='%s'" % (schema_name))
	data = cursor.fetchall()
	if data[0][0] > 0:
		# this should never happen
		raise DatabaseException("DB Schema for tenant '%s' already exists." % (tenant_name))
		
	cursor.execute('CREATE SCHEMA %s;' % (schema_name) )
	
	tenant_syncdb(tenant_name, cursor)
	

def create_tenant_defaults(tenant):

	agent_create_r(tenant, DEFAULT_AGENT_NAME, True)
	
	#
	# More to do?
	#
	check_priority_levels(tenant)
	
	check_common_icons(tenant)

	#check_tags(tenant)  #FIX: non mi piace qui. Dovremmo definire un segnale e registrare tutto quello che va fatto quando viene creato un nuovo tenantinit_tenant
	
	#if 'entables' in get_configuration().INSTALLED_APPS:
	#	from entables.tablesstore.djangostore import init_entables_chains, set_default_tenant_rules, fix_names
	#	init_entables_chains(tenant)
	#	set_default_tenant_rules(tenant)
	#	fix_names(tenant)

	sanet_tenant_create.send(sender=tenant.__class__, instance=tenant)


	return

def check_primary_agent(tenant):
	
	tenant.db_set_current_schema()

	# No primary agent? Should never happen!
	if Agent.objects.filter(is_primary=True).count() == 0:
		a, created = Agent.objects.get_or_create(tenant=tenant, name=DEFAULT_AGENT_NAME)
		a.is_primary = True
		a.save()


def check_priority_levels(tenant):

	db_set_current_tenant_schema(tenant.name)

	if PriorityLabel.objects.count() > 0:
		return

	#print "Setting default priority level"
	
	LEVELS = [
		(  1, 30, 'low'     , "Priorita' bassa"   ),
		( 31, 60, 'medium'  , "Priorita' media"   ),
		( 61,100, 'critical', "Priorita' critica" ),
	]

	for minl, maxl, label, description in LEVELS:
		pl = PriorityLabel(tenant=tenant, min_level=minl, max_level=maxl, label=label, description=description)
		pl.save()

	return

def check_common_icons(tenant):

	#print "Checking standard icons..."

	def symlink_exists(alias):
		
		# FIX: CHECK IF LINK EXISTS
		try:
			os.lstat(alias)
			return True
		except OSError, e:
			if e.errno == 2:
				return False
			else:
				raise

	tenant_icons_base_dir = tenant_get_icons_dir(tenant.uuid)

	icons_dir = os.path.join( get_configuration().SANET_INSTALL_DIR, "share", "icons")

	for file_name in os.listdir(icons_dir):
		if file_name.startswith('.'): continue

		icon_file_path = os.path.join(icons_dir, file_name)

		icon_dest_file_path = os.path.join(tenant_icons_base_dir, file_name)


		if not symlink_exists(icon_dest_file_path):
			try:
				os.symlink(icon_file_path, icon_dest_file_path)
				#print "Created link", icon_file_path, "->", icon_dest_file_path
			except Exception, e:
				print >> sys.stderr, "Unable to create link for %s: %s" % (icon_dest_file_path, e)




def sync_all_tenants(connection, logger=None):
	
	logger = logger or logging.getLogger('sanet.global')

	for t in Tenant.objects.all():
		tenant_syncdb(t.name, connection, logger=logger)

	db_cursor = connection.cursor()
	db_cursor.execute("SET search_path TO public")
	
	return


def tenant_syncdb(tenant_name, cursor, logger=None):

	logger = logger or logging.getLogger('sanet.global')

	#
	from django.db import DEFAULT_DB_ALIAS, connections
	connection = connections[DEFAULT_DB_ALIAS]

	#from django.db.migrations.executor import MigrationExecutor
	#
	#c = SanetMigrateCommand()
	#c.verbosity = 1
	#
        ## Get the database we're operating from
        #
        ## Hook for backends needing any database preparation
        #connection.prepare_database()
	#
        ## Hook for backends needing any database preparation
        ##connection.prepare_database()
        ## Work out which apps have migrations and which do not
        #executor = MigrationExecutor(connection, c.migration_progress_callback)
        #
        ## Raise an error if any migrations are applied before their dependencies.
        #executor.loader.check_consistent_history(connection)
        #
	#print "Syncing...."
	#c.tenant_name = tenant_name
	#c.sync_apps(connection, executor.loader.unmigrated_apps)
	#
	#print "OK"
	#
	#return
	
	
	
	cursor = connection.cursor()
	cursor.execute("set search_path to %s;" % (tenant_name));
	
	
	# Move to tenant schema
	#db_set_current_tenant_schema(tenant_name, include_public_schema=False)#, db_cursor=cursor)

	# Get informations
	tables = connection.introspection.table_names()
	seen_models = connection.introspection.installed_models(tables)
	created_models = set()
	pending_references = {}
	
	# Get all tenant dependent models for each app
	all_models = []

	#for app in models.get_apps(): 
	from django.apps import apps
	for app in apps.get_app_configs(): 
	
		logger.debug("Checking app: %s", app )
	
		#app_name = app.__name__.split('.')[-2]
		app_name = app.name
		#app_name = app
		
		#tenant_models = [m for m in models.get_models(app, include_auto_created=True) if (issubclass(m, TenantDependentModel)) ]
		tenant_models = [m for m in app.get_models(include_auto_created=True) if (issubclass(m, TenantDependentModel)) ]

		if settings.DEBUG: 
			for m in tenant_models:
				logger.log(9, "\tfound model: %s", m)

		all_models.append( (app_name, tenant_models) )

	def __model_installed(model):
		opts = model._meta
		converter = connection.introspection.table_name_converter
		return not ((converter(opts.db_table) in tables) or (opts.auto_created and converter(opts.auto_created._meta.db_table) in tables))

	manifest = SortedDict(
		(app_name, filter(__model_installed, model_list)) for app_name, model_list in all_models
	)

	with connection.schema_editor() as schema_editor:

		# Creating models
		for app_name, model_list in manifest.items():

			for model in model_list:

				"""
				sql, references = connection.creation.sql_create_model(model, UNFORMATTED(), seen_models)

				seen_models.add(model)
				created_models.add(model)

				for refto, refs in references.items():
					pending_references.setdefault(refto, []).extend(refs)
					if refto in seen_models:
						sql.extend(connection.creation.sql_for_pending_references(refto, UNFORMATTED(), pending_references))

				sql.extend(connection.creation.sql_for_pending_references(model, UNFORMATTED(), pending_references))

				for statement in sql:
					cursor.execute(statement)
				"""

				logger.debug("\tCreating model: %s", m)
				
				data = schema_editor.create_model(model)

				#if settings.DEBUG: print __file__, data

				tables.append(connection.introspection.table_name_converter(model._meta.db_table))

		cursor = connection.cursor()
		cursor.execute("set search_path to %s, public;" % (tenant_name));


	#transaction_commit_unless_managed()

	# Install SQL indices for all newly created models
	for app_name, model_list in manifest.items():
		for model in model_list:
			if model in created_models:
				index_sql = connection.creation.sql_indexes_for_model(model, UNFORMATTED())
				if index_sql:
					for sql in index_sql:
						#try:
							cursor.execute(sql)
						#except Exception, e:
						#	transaction_rollback_unless_managed(using=db)
						#else:
						#	transaction_commit_unless_managed(using=db)
						
	# Move to tenant schema
	db_set_public_schema()


def symlink_exists(alias):
	
	# FIX: CHECK IF LINK EXISTS
	try:
		os.lstat(alias)
		return True
	except OSError, e:
		if e.errno == 2:
			return False
		else:
			raise

def check_indexes_on_tables(tenant, logger=None):

	cursor = connection.cursor()

	#print "Init special indexes for tenant", tenant.name

	db_set_current_tenant_schema(tenant.name, db_cursor=cursor)

	INDEXES = [
		('monitoredelement',
		 'monitoredelement_tenant_id_kind_name_parent_id_null_key',
		 "CREATE UNIQUE INDEX monitoredelement_tenant_id_kind_name_parent_id_null_key ON monitoredelement (tenant_id, kind, name, COALESCE(parent_id,''));" )
	]

	for table, index, create_query in INDEXES:

		if logger: logger.info( "\t" + "Checking %s %s", table, index)

		query = "select count(*) from pg_indexes where schemaname = '%s' and tablename = '%s' and indexname='%s'" % (tenant.name.lower(), table, index)
		cursor.execute(query)
		val = cursor.fetchall()[0][0] # [ (1,) , ]
		if val == 0:
			cursor.execute(create_query)
		else:
			if logger: logger.info( "\t" + "\t" + "Already exists")

	cursor.close()





#
# FIX: queste funzioni dovrebbero stare in un modulo a parte e non qui per evitare
#      problemi di import circolari. Fare refactory il prima possibile
#
def tenant_get_data_dir(tenant_id, dir_name=None):
	args = [get_configuration().DATA_DIR, tenant_id]
	if dir_name:
		args.append(dir_name)
	return os.sep.join( args )

def tenant_get_media_dir(tenant_id):
	return tenant_get_data_dir(tenant_id, get_configuration().TENANT_MEDIA_DIRNAME)

def tenant_get_icons_dir(tenant_id):
	base = tenant_get_data_dir(tenant_id, get_configuration().TENANT_MEDIA_DIRNAME)
	return os.path.join(base, 'icons')

def tenant_get_media_file_path(tenant_id, file_path):
	return os.path.join( tenant_get_media_dir(tenant_id) , file_path )

def tenant_get_media_url_path(tenant_id, file_path):	
	return '/'.join( [tenant_id, get_configuration().TENANT_MEDIA_DIRNAME, file_path] )

def tenant_get_media_url(tenant_id, media_file_path):
	media_sub_url = tenant_get_media_url_path(tenant_id, media_file_path)
	return reverse('media.file.serve', kwargs={'path':media_sub_url})
	
def tenant_get_icon_url(tenant_id, icon_file_name):
	return tenant_get_media_url(tenant_id, 'icons/' + icon_file_name)

# FIX: move this in an generic module
def static_file_url(file_path):
	return reverse('static.file.serve', kwargs={'path':file_path})


from conf import set_dir_permissions

def makedirs_for_webuser(path):

	try:
		os.makedirs(path)
		#os.mkdir(base_path)
	except OSError, e: 
		if e.errno == 17: # file exits:
			pass
		else:
			raise e
	
	set_dir_permissions(None, path)
	
	return

def tenant_create_datadir(tenant_id, tenant_name):
	
	RRDS_DIRNAME = 'rrds'
	
	# /<id>/rrd
	tenant_base_data_dir = tenant_get_data_dir(tenant_id)
	
	if not os.path.exists(tenant_base_data_dir):
		if DEBUG: print "Creating tenant dir", tenant_base_data_dir
		makedirs_for_webuser(tenant_base_data_dir)


	# /<id>/media
	tenant_media_dir = tenant_get_media_dir(tenant_id)
	if not os.path.exists(tenant_media_dir):
		if DEBUG: print "Creating %s" % ( tenant_media_dir )
		makedirs_for_webuser(tenant_media_dir)


	MEDIA_SUBDIRS = [
		'icons',
	]
	for dir_name in MEDIA_SUBDIRS:
		dir_path = os.path.join( tenant_media_dir, dir_name)
		if DEBUG: print "Checking", dir_path
		if not os.path.exists(dir_path):
			if DEBUG: print "Creating %s" % ( dir_path )
			makedirs_for_webuser(dir_path)

	#
	# Signal to other subapp
	#
	sanet_tenant_create_vardir.send(sender=Tenant, tenant_id=tenant_id)


	#
	# Creating symbolic links
	#

	# /<name>/ -> <id>
	alias_base_data_dir  = os.sep.join( [get_configuration().DATA_DIR, tenant_name] )
	if DEBUG: print "Creating alias", alias_base_data_dir

	if symlink_exists(alias_base_data_dir):
		if DEBUG: print "THe link alread exists"
		linked_dir = os.path.realpath(alias_base_data_dir)

		if linked_dir != tenant_base_data_dir:
			if DEBUG: print "Wrong link: delete old link"
			os.unlink(alias_base_data_dir)

	#
	# Create: {{VAR_LIB}}/<tenant_name>/ -> {{VAR_LIB}}/<tenant_id>/
	#
	if not symlink_exists(alias_base_data_dir):
		if DEBUG: print "creating link "
		os.symlink(tenant_base_data_dir, alias_base_data_dir)

	return 

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def tenant_get(tenant_name):
	try:
		return Tenant.objects.get(name=tenant_name)
	except Tenant.DoesNotExist, e:
		return None
		
def tenant_get_by_id(tenant_uuid):
	try:
		return Tenant.objects.get(uuid=tenant_uuid)
	except Tenant.DoesNotExist, e:
		return None



def tenant_create(tenant_name, long_name='', timezone=DEFAULT_TIMEZONE , primary=False):
	
	timezone = timezone if timezone != None else DEFAULT_TIMEZONE
	primary  = primary  if primary  != None else False
	
	#
	# FIX: check if primary is True and raise exception if neccessary
	#
	tenant = _tenant_create(tenant_name, long_name, timezone, primary)

	check_indexes_on_tables(tenant)
	
	return tenant

@transaction_commit_on_success
def _tenant_create(tenant_name, long_name, timezone, primary):
	
	new_tenant = Tenant(
		name       = tenant_name,
		long_name  = long_name,
		is_primary = primary,
		timezone   = timezone,
	)
	new_tenant.save()
	
	
	create_tenant_environment(new_tenant)

	return new_tenant

@transaction_commit_on_success
def tenant_update(tenant, fields_name=None):
	# More to do?
	#tenant.save()
	tenant.concurrent_save(fields_name=fields_name)


@transaction_commit_on_success
def tenant_remove(tenant_name, full=True):
	
	t = Tenant.objects.get(name=tenant_name)
	
	if t.is_primary:
		raise Exception("Can not remove primary tenant")

	if full:
		schema_name = tenant_name_to_schema(tenant_name)
		new_name = '%s_deleted_%s' % (schema_name, int(time.time()))

		db_cursor = connection.cursor()	
		db_cursor.execute('DROP SCHEMA %s CASCADE;' % (schema_name) )

	db_set_current_tenant_schema('public', include_public_schema=False)
	
	t.delete()

@transaction_commit_on_success
def tenant_rename(tenant_name, new_name):
	raise Exception("tenant_rename not implemented yet")



@transaction_commit_on_success
def tenant_addgroup(tenant_name, group_name):
	
	t = Tenant.objects.get(name=tenant_name)
	g = Group.objects.get(name=group_name)
	
	tg, created = TenantGroup.objects.get_or_create(tenant=t, group=g)
	if created:
		tg.save()

		
def tenant_getgroups(tenant_name):
	gs = []
	t = Tenant.objects.get(name=tenant_name)
	for tg in TenantGroup.objects.filter(tenant=t):
		gs.append(tg.group)
	return gs
	


@transaction_commit_on_success
def tenant_removegroup(tenant_name, group_name):
	
	t = Tenant.objects.get(name=tenant_name)
	g = Group.objects.get(name=group_name)
	
	tg = TenantGroup.objects.get(tenant=t, group=g)
	tg.delete()

	# Cancelliamo le configurazioni a livello utente sul tenant
	# che ormai non e' piu' visibile dagli utenti del gruppo.
	for u in g.user_set.all():
		TenantUserSettings.objects.filter(user=u, tenant=t).delete()

@transaction_commit_on_success
def tenant_toggle_group(tenant, group):

	
	tg, created = TenantGroup.objects.get_or_create(tenant=tenant, group=group)
	
	if created:
		return True
	else:
		
		tg.delete()
		
		# Cancelliamo le configurazioni a livello utente sul tenant
		# che ormai non e' piu' visibile dagli utenti del gruppo.
		for u in group.user_set.all():
			TenantUserSettings.objects.filter(user=u, tenant=tenant).delete()
		
		return False

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def tenant_change_primary_by_name(tenant_name):
	t = Tenant.objects.get(name=tenant_name)
	_tenant_change_primary(t)

def tenant_change_primary(t):
	_tenant_change_primary(t)

@transaction_commit_on_success
def _tenant_change_primary(t):

	try:
		p = Tenant.objects.get(is_primary=True)
	except Tenant.DoesNotExist, e:
		p = None
	
	if p == t:
		return False

	if p:
		p.is_primary = False

		p.concurrent_save(['is_primary'])

	t.is_primary = True

	t.concurrent_save(['is_primary'])

	return True


def tenant_get_primary():
	try:
		return Tenant.objects.get(is_primary=True)
	except Tenant.DoesNotExist, e:
		return None

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def tenant_select_all(user_name=None, name_filter_str=None):

	if user_name != None:
		groups  = User.objects.get(username=user_name)
		tenants = Tenant.objects.filter(groups=groups)
	else:
		tenants = Tenant.objects

	if name_filter_str:
		pattern = '%s' % (name_filter_str)
		tenants = tenants.filter(name__startswith=pattern)

	return tenants.all()

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

@transaction_commit_on_success
def user_create(username, fields=None):

	u = User(username=username)
	
	if fields:
		for f_name, fval in fields.items():
			
			if f_name == 'password':
				u.set_password(fval)
			else:
				setattr(u, f_name, fval)
	u.save()

	return u

@transaction_commit_on_success
def user_delete(username):

	u = User.objects.get(username=username)
	u.delete()

@transaction_commit_on_success
def user_rename(name, new_name):

	user = User.objects.get(username=name)
	profile = user.get_profile()

	user.username = new_name
	user.save()
	
	profile.concurrent_save()
	

def user_find(username):
	u = User.objects.get(username=username)
	return u

@transaction_commit_on_success
def user_update(user):

	profile = user.get_profile()
	user.save()
	profile.concurrent_save()
	
@transaction_commit_on_success
def user_set_password(user, password, encripted=True):

	profile = user.get_profile()

	if encripted:
		user.set_password(password)  # FIX: Che succede se questa operazione non rispetta il contesto transazionale in caso di crash?	
	else:
		user.password = password
		
	user.save()
	
	profile.concurrent_save()

def user_set_enabled(user, enabled):


	if user.is_active == enabled:
		return

	profile = user.get_profile()

	user.is_active = enabled
	user.save()

	profile.concurrent_save()

def _user_add_to_group(user, group):
	# FIX: more to do?
	user.groups.add(group)

def _user_remove_from_group(user, group):
	# FIX: more to do
	user.groups.remove(group)

def _user_add_or_remove_to_group(user, group):

	if group in user.groups.all():
		_user_remove_from_group(user, group)
		return False
	else:
		_user_add_to_group     (user, group)
		return True

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def group_get(group_name):
	try:
		return Group.objects.get(name=group_name)
	except Group.DoesNotExist, e:
		return None

def group_get_data(group_name):
	try:
		g = Group.objects.get(name=group_name)
		return g.sanetgroupprofile_set.all()[0]
	except Group.DoesNotExist, e:
		return None

@transaction_commit_on_success
def group_create(groupname):

	g = Group(name=groupname)
	g.save()

	return g

@transaction_commit_on_success
def group_delete(groupname):

	if groupname == get_configuration().DEFAULT_ADMIN_GROUP:
		raise InvalidConfigurationException("Unable to delate system group '%s'" % (get_configuration().DEFAULT_ADMIN_GROUP,))

	g = Group.objects.get(name=groupname)
	g.delete()

@transaction_commit_on_success
def group_adduser(groupname, username):

	g = Group.objects.get(name=groupname)
	u = User.objects.get(username=username)

	_user_add_to_group(u, g)

@transaction_commit_on_success
def group_removeuser(groupname, username):

	g = Group.objects.get(name=groupname)
	u = User.objects.get(username=username)

	_user_remove_from_group(u, g)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def agent_select_all(tenant_name):

	if tenant_name:
		db_set_current_tenant_schema(tenant_name)
		t = tenant_get(tenant_name)

	clusters = Agent.objects.filter(tenant=t)
		
	return clusters.all()

def agent_get(name, tenant_name=None):

	if tenant_name:
		t = tenant_get(tenant_name)
		db_set_current_tenant_schema(tenant_name)

	try:
		a = Agent.objects.get(name=name)
		return a
	except Agent.DoesNotExist, e:
		return None
		
def agent_get_by_id(aid, tenant):

	db_set_current_tenant_schema(tenant.name)

	try:
		a = Agent.objects.get(tenant=tenant, uuid=aid)
		return a
	except Agent.DoesNotExist, e:
		return None


def agent_get_primary(tenant):

	db_set_current_tenant_schema(tenant.name)

	try:
		a = Agent.objects.get(tenant=tenant, is_primary=True)
		return a
	except Agent.DoesNotExist, e:
		return None

@transaction_commit_on_success
def agent_create(agent_name, primary_if_first=False, tenant_name=None):

	if tenant_name:
		db_set_current_tenant_schema(tenant_name)
		t = tenant_get(tenant_name)

	primary = False
	if primary_if_first == True:	
		if Agent.objects.filter(is_primary=True).count() == 0:
			primary = True

	a = agent_create_r(t, agent_name, primary)

	#
	# FIX: race condition. Se aggiorniamo anche l'mtime del tenant "t"
	#      dovremmo riuscire a rendere "atomica questa operazione (2-phase commit)
	#
	#t.concurrent_save(['mtime'])

	return a

	
def agent_create_r(tenant, agent_name, primary):

	db_set_current_tenant_schema(tenant.name)

	a, created = Agent.objects.get_or_create(tenant=tenant, name=agent_name, is_primary=primary)
	a.save()

	return a



	
@transaction_commit_on_success
def agent_update(agent, *fields_name):

	# FIX: altro da fare?
	agent.concurrent_save(fields_name=fields_name)


@transaction_commit_on_success
def agent_set_enabled(agent, enabled):

	oldmtime    = agent.mtime

	#FIX: WHAT TO DO?	
	agent.is_enabled = enabled
	
	agent.concurrent_save(['is_enabled'])
	
@transaction_commit_on_success
def agent_set_remote(agent, remote):
	
	agent.is_remote = remote
	
	agent.concurrent_save(['is_remote'])

@transaction_commit_on_success
def agent_set_primary(agent):

	mtime = time.time()

	# Toggle old primary
	agents = Agent.objects.filter(is_primary=True).exclude(uuid=agent.uuid).all()
	if len(agents) >= 0:
		
		for a in agents:
			a.is_primary = False
			a.concurrent_save(fields_name=['is_primary', 'mtime'], new_sentinel_value=mtime)
		

	agent.is_primary = True
	agent.concurrent_save(fields_name=['is_primary'], new_sentinel_value=mtime)

	# FIX: something changed. what to do?
	
	return

@transaction_commit_on_success
def agent_rename(tenant, name, new_name):

	db_set_current_tenant_schema(tenant.name)
	
	a = Agent.objects.get(tenant=tenant, name=name)

	a.name  = new_name
	
	a.concurrent_save(['name']) 
	
	return

def agent_remove(agent_name, tenant_name=None):
	
	if tenant_name:
		db_set_current_tenant_schema(tenant_name)
		t = tenant_get(tenant_name)
		
	a = Agent.objects.get(tenant=t, name=agent_name)

	return agent_remove_safe(a, t)
	
@transaction_commit_on_success
def agent_remove_safe(a, tenant):

	if a.is_primary:
		raise Exception("Unable to remove primary agent")
	
	p = Agent.objects.get(tenant=tenant, is_primary=True)
	
	if a.monitoredelement_set.count() > 0:
		raise Exception("Agent busy")
	
	#
	#FIX: fix linked models
	#   - monitored nodes?
	#
	
	a.delete()

def agent_select_all(tenant_name, name_filter_str=None):

	db_set_current_tenant_schema(tenant_name)
	t = tenant_get(tenant_name)

	agents = Agent.objects
	if name_filter_str:
		pattern = '%s' % (name_filter_str)
		agents = agents.filter(name__startswith=pattern)

	a = agents.all()
	return a

def agent_select_all_locals():

	local_agents = []

	for t in Tenant.objects.all():

		db_set_current_tenant_schema(t.name)

		for a in Agent.objects.filter(tenant=t, is_enabled=True, is_remote=False):
			local_agents.append(a)

	return local_agents

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def cluster_get(cluster_name, tenant_name=None):

	if tenant_name:
		db_set_current_tenant_schema(tenant_name)
		t = tenant_get(tenant_name)

	try:
		c = Cluster.objects.get(name=cluster_name)
		return c
	except Cluster.DoesNotExist, e:
		return None

@transaction_commit_on_success
def cluster_save(cluster, *fields_name):

	cluster.concurrent_save(fields_name=fields_name)

	return

@transaction_commit_on_success
def cluster_rename(tenant, name, new_name, action_user):

	db_set_current_tenant_schema(tenant.name)

	if action_user:
		#action_user.assert_capability(self.tenant, 'core.template.element.acl.edit')
		self.assert_permission_readwrite(action_user)

	c = Cluster.objects.get(name=name)

	c.name = new_name

	c.concurrent_save(fields_name=['name'])

@transaction_commit_on_success
def cluster_remove(cluster_name, tenant_name=None):

	if tenant_name:
		db_set_current_tenant_schema(tenant_name)
		t = tenant_get(tenant_name)

	cluster = Cluster.objects.get(name=cluster_name)
	
	#
	# FIX: now we update handle only node clusters
	#
	
	#
	# FIX: use UPDATE and not SAVE!!
	#
	elements = MonitoredElement.objects.filter(kind=ELEMENT_KIND_NODE, cluster=cluster)
	for e in elements:

		_node_set_cluster(e, None)
	
	#
	# FIX: anything else to do?
	#
	cluster.delete()
	
	return

@transaction_commit_on_success
def cluster_create(cluster_name, tenant_name=None, owner=None):

	if tenant_name:
		db_set_current_tenant_schema(tenant_name)
		t = tenant_get(tenant_name)

	c = Cluster(name=cluster_name)
	c.tenant = t
	c.owner = owner
	c.save()

	return c


def cluster_select_all(tenant_name, name_filter_str=None):

	if tenant_name:
		db_set_current_tenant_schema(tenant_name)
		t = tenant_get(tenant_name)

	clusters = Cluster.objects
	if name_filter_str:
		pattern = '%s' % (name_filter_str)
		clusters = clusters.filter(name__startswith=pattern)
		
		
	return clusters.all()

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def element_get_all(tenant_name, uuid=None):

	db_set_current_tenant_schema(tenant_name)

	el_data = MonitoredElement.objects.get(uuid = uuid)
	el = MonitoredElement.objects.filter(uuid = uuid).values('uuid','name')[0]

	if uuid:
		all_els =  MonitoredElement.objects.filter(tenant__name=tenant_name, parent__uuid=uuid)
	else:
		all_els =  MonitoredElement.objects.filter(tenant__name=tenant_name, parent__isnull=True)

	# FIX: els e' gia' un dizionario. Se devi cambiare solo il kind puoi farlo senza ricreare un nuovo vettore di dati?
	# FIX: sarebbe meglio impostare 'kind' fuori dalle api di base.

	els = {}
	els['elements'] = []
	for a_el in all_els:
		els['elements'].append(element_get_info(a_el.uuid))

	if el_data.kind != 1:
		els['parent'] = element_get_info(el_data.parent.uuid)

	els['mode'] = el
	els['additional'] = {}
	els['additional']['datagroups'] = DataGroup.objects.filter(element__uuid = uuid).values('uuid','treedirentry','classification', 'suspended', 'suspend_time', 'suspend_resume_time', 'suspend_reason', 'mtime', 'element_id', 'element__name', 'title', 'name')
	els['additional']['conditions'] = Condition.objects.filter(element__uuid = uuid).values('uuid','statuschange_action','expr', 'classification', 'uncheckable_fallback', 'msg_upsubj', 'primary', 'mtime', 'statuslastchange', 'statuslast', 'element_id', 'element__name', 'datagroup_id', 'datagroup__name', 'title', 'name', 'status')
	els['additional']['datasources'] = DataSource.objects.filter(element__uuid = uuid).values('uuid','treedirentry', 'gauge','classification', 'cascade', 'expr', 'order', 'mtime', 'title', 'status', 'element_id', 'element__name', 'datagroup_id', 'datagroup__name', 'title', 'name', 'status')
	els['additional']['timegraphs'] = TimeGraph.objects.filter(element__uuid = uuid).values('uuid','primary', 'stacked')

	return els

def element_get(tenant_name, kind, name):

	db_set_current_tenant_schema(tenant_name)

	return MonitoredElement.objects.get(kind=kind, name=name, tenant__name=tenant_name)

def element_get_info(uuid, child=False):
	ret = {}

	me = MonitoredElement.objects.get(uuid = uuid)
	ret['rt'] = me.kind
	
	if me.kind == 1:
		ret['values'] = MonitoredNode.objects.filter(uuid=me.uuid).values()[0]
	
	elif me.kind == 2:
		ret['values'] = MonitoredIface.objects.filter(uuid=me.uuid).values()[0]
		
	elif me.kind == 3:
		ret['values'] = MonitoredStorage.objects.filter(uuid=me.uuid).values()[0]
	
	elif me.kind == 4:
		ret['values'] = MonitoredService.objects.filter(uuid=me.uuid).values()[0]

	
	elif me.kind == 5:
		ret['values'] = MonitoredDevice.objects.filter(uuid=me.uuid).values()[0]
		
	ret['values']['kind'] = ELEMENT_KIND_STR[ret['values']['kind']]
	return ret

def get_struct(uuid, tenant_name):

	db_set_current_tenant_schema(tenant_name)
	ret = {}
	ret['data'] = element_get_info(uuid)
	if ret['data']['values']['kind'] != ELEMENT_KIND_STR[1]:
		ret['parent'] = element_get_info(ret['data']['values']['parent_id'])
	ret['content'] = element_get_all(tenant_name, uuid)
	return ret
   
#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

@transaction_commit_on_success
def node_create(node_name, tenant_name=None, owner=None):


	if tenant_name:
		db_set_current_tenant_schema(tenant_name)
		t = tenant_get(tenant_name)

	if owner and t:
		owner.assert_capability(t, 'core.element.create')

	n = MonitoredNode(name=node_name)
	n.owner  = owner
	n.tenant = t

	n.timezone = t.timezone

	try:
		a = Agent.objects.get(is_primary=True)
	except Agent.DoesNotExist, e:
		raise ApplicationException("Impossible to create node. No primary agent defined.")

	n.agent = a

	#
	# FIX: set default agent
	# FIX: set default timezone
	#
	n.save()

	return n

@transaction_commit_on_success
def node_set_agent(node, agent, user=None):

	if user:
		user.assert_capability(node.tenant, 'core.element.conf')

	mtime = time.time() 

	e_uuids = []

	e_uuids.append(node.uuid)

	#
	# FIX: se le interfacce di un link stanno su due "agent" diversi ci sono dei
	#      problemi perche' un agente carica solo le interfacce a lui associate
	#      e quindi controlli di adiacenza vanno in errore grave.
	#
	#      Bloccare questo aggiornamento se interfacce linkate tra loro stanno su 
	#      tenant diversi.
	#

	for i in MonitoredIface.objects.filter(parent=node).all():
		
		i.agent = agent
		i.concurrent_save(['agent'], new_sentinel_value=mtime)
		
		e_uuids.append(i.uuid)
		
		#for link, li in i.get_linked_ifaces():
		#	if li.agent != agent:
		#		raise Exception("Can't set agent. Interface %s is linked with %s which belongs to agent %s" % (i.path, li.path, li.agent.name))
				
	for s in MonitoredStorage.objects.filter(parent=node).all():
		s.agent = agent
		s.concurrent_save(['agent'], new_sentinel_value=mtime)

		e_uuids.append(s.uuid)

	for s in MonitoredService.objects.filter(parent=node).all():
		s.agent = agent
		s.concurrent_save(['agent'], new_sentinel_value=mtime)

		e_uuids.append(s.uuid)

	for d in MonitoredDevice.objects.filter(parent=node).all():
		d.agent = agent
		d.concurrent_save(['agent'], new_sentinel_value=mtime)

		e_uuids.append(d.uuid)

	#
	# Modifichiamo tutti di datagroup delle entita' coinvolte
	#
	for dg in DataGroup.objects.filter(element__uuid__in=e_uuids):
		dg.agent = agent
		dg.concurrent_save(['agent'], new_sentinel_value=mtime)

	for ds in DataSource.objects.filter(element__uuid__in=e_uuids):
		ds.agent = agent
		ds.concurrent_save(['agent'], new_sentinel_value=mtime)

	for c in Condition.objects.filter(element__uuid__in=e_uuids):
		c.agent = agent
		c.concurrent_save(['agent'], new_sentinel_value=mtime)

	#
	# IMPORTANTE: il test di aggiornameento concorrente lo facciamo dopo aver modificato
	#             tutte le entita' coinvolte cosi' se c'e' stato un aggiornamento 
	#             concorrente lo rileviamo in un solo aggiornamento, quello del nodo
	#             che ha triggerato tutto.
	#

	node.agent = agent
	node.concurrent_save(['agent'], new_sentinel_value=mtime)

	return

@transaction_commit_on_success
def node_set_cluster(node, cluster, user=None):

	if user:
		user.assert_capability(node.tenant, 'core.element.conf')

	return _node_set_cluster(node, cluster)

def _node_set_cluster(node, cluster):

	if node.cluster == cluster:
		return

	prev = node.cluster

	node.cluster = cluster

	if cluster == None:
		node.cluster_role          = None
		node.cluster_distinguisher = None
		node.cluster_xform         = None
		node.cluster_instance      = None
	else:
		if prev != node.cluster:
			node.cluster_role = CLUSTER_ROLE_DEPENDENT

	node.concurrent_save(['cluster', 'cluster_role', 'cluster_distinguisher', 'cluster_xform', 'cluster_instance'])

	#
	# FIX: change all related objects with UPDATE and not SAVE
	# - datagroups
	# - sub elements (interfaces, storages, ecc...)
	#
	#for e in MonitoredElement.objects.filter(parent=node).all():
	#	e.cluster = cluster
	#	# FIX: update config time
	#	e.save()


@transaction_commit_on_success
def node_update(node, *fields_name, **options):
	"""
	Il parametro options puo' contenere le seguenti chiavi:
		'mtime'
	"""

	# FIX: signal modification

	node.validate_fields()

	node.concurrent_save(fields_name=fields_name, new_sentinel_value=options.get('mtime',None))


@transaction_commit_on_success
def node_remove(node_name, tenant_name=None):

	if tenant_name:
		db_set_current_tenant_schema(tenant_name)
		t = tenant_get(tenant_name)

	n = MonitoredElement.objects.get(name=node_name, kind=ELEMENT_KIND_NODE)
	n = n.get_specific_element()
	
	#
	# Remove involved links
	#
	ifaces = n.get_ifaces()
	for i in ifaces:
		i = i.get_specific_element()
		_remove_links_on_iface(i)

	#
	# FIX: adjust depends on before delete!!!
	#
	n.delete()

	return

@transaction_commit_on_success
def node_remove_sub_element(node, kind, element_name, user=None):

	db_set_current_tenant_schema(node.tenant.name)

	if user:
		user.assert_capability(node.tenant, 'core.element.conf')

	#
	# FIX: adjust depends on before delete!!!
	#

	element = MonitoredElement.objects.get(parent=node, kind=kind, name=element_name)
	element.delete()

	#
	# FIX: update node mtime for transactional update? uhm... maybe not necessary
	#

	return


def node_get(node_name, tenant_name=None):

	if tenant_name:
		db_set_current_tenant_schema(tenant_name)

	try:
		n = MonitoredNode.objects.get(name=node_name)
		return n
	except MonitoredNode.DoesNotExist, e:
		return None

def node_get_by_uuid(uuid, tenant=None):

	if tenant:
		db_set_current_tenant_schema(tenant.name)

	try:
		n = MonitoredNode.objects.get(uuid=uuid)
		return n
	except MonitoredNode.DoesNotExist, e:
		return None


def node_select_all(tenant, name_filter_str=None, agent=None):

	db_set_current_tenant_schema(tenant.name)

	nodes = MonitoredNode.objects.filter(tenant=tenant)

	if agent:
		nodes = nodes.filter(agent=agent)

	if name_filter_str:
		pattern = '%s' % (name_filter_str)
		nodes = nodes.filter(name__startswith=pattern)

	return nodes

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def iface_get(node_name, iface_name, tenant_name=None):

	if tenant_name:
		db_set_current_tenant_schema(tenant_name)

	try:
		node = MonitoredNode.objects.get(name=node_name)

		iface = MonitoredIface.objects.get(name=iface_name, parent=node)
		
		return iface

	except MonitoredIface.DoesNotExist, e:
		return None
		
def iface_get_by_id(uuid, tenant=None):
	if tenant:
		db_set_current_tenant_schema(tenant.name)
	try:
		return MonitoredIface.objects.get(tenant=tenant, uuid=uuid)
	except MonitoredIface.DoesNotExist, e:
		return None

@transaction_commit_on_success
def iface_create(node_name, iface_name, tenant_name=None):

	if tenant_name:
		db_set_current_tenant_schema(tenant_name)
		t = tenant_get(tenant_name)

	n = MonitoredNode.objects.get(name=node_name)

	if n.tenant != t:
		raise Exception("Unable to create interface on tenant %s. Node tenant is %s" % (t, n.tenant))

	i = MonitoredIface(name=iface_name)

	i.kind = ELEMENT_KIND_IFACE
	i.tenant = t
	i.parent = n

	i.agent = n.agent

	i.owner = n.owner

	# 
	# Other to do?
	#

	i.save()

	return i

@transaction_commit_on_success
def iface_update(iface, *fields_name, **options):

	iface.validate_fields()

	iface.concurrent_save(fields_name=fields_name, new_sentinel_value=options.get('mtime',None))
	
	return

def iface_select_all(tenant, node=None, name_filter_str=None, agent=None):

	db_set_current_tenant_schema(tenant.name)

	ifaces = MonitoredIface.objects.filter(tenant=tenant)

	if node != None:
		ifaces = ifaces.filter(parent=node)

	if agent:
		ifaces = ifaces.filter(agent=agent)

	if name_filter_str:
		pattern = '%s' % (name_filter_str)
		ifaces = ifaces.filter(name__startswith=pattern)

	return ifaces.all()

@transaction_commit_on_success
def iface_delete(tenant, node_name, iface_name, user=None):

	if user:
		user.assert_capability(tenant, 'core.element.conf')

	db_set_current_tenant_schema(tenant.name)

	node = MonitoredNode.objects.get(name=node_name)

	iface = MonitoredIface.objects.get(name=iface_name, parent=node)

	_remove_links_on_iface(iface)

	iface.delete()
	
	return

def _remove_links_on_iface(iface):
	
	for linked_iface in iface.links.all():
		link = linked_iface.link
		link.delete()

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

@transaction_commit_on_success
def storage_create(node, storage_name):

	# FIX: move to API
	storage = MonitoredStorage(
		tenant = node.tenant, 
		name   = storage_name, 
		parent = node,
		kind   = ELEMENT_KIND_STORAGE,
		agent  = node.agent,
		
		owner = node.owner,
	)
	storage.save()

	return storage

@transaction_commit_on_success
def storage_update(storage, *fields_name, **options):
	
	storage.validate_fields()
	
	storage.concurrent_save(fields_name=fields_name, new_sentinel_value=options.get('mtime',None))

def storage_get(node_name, storage_name, tenant_name=None):

	if tenant_name:
		db_set_current_tenant_schema(tenant_name)

	try:
		# FIX: do it with one query node__name
		node = MonitoredNode.objects.get(name=node_name)
		storage = MonitoredStorage.objects.get(name=storage_name, parent=node)		
		return storage
	except MonitoredStorage.DoesNotExist, e:
		return None

def storage_get_by_id(uuid):
	try:
		storage = MonitoredStorage.objects.get(uuid=uuid)
		return storage
	except MonitoredStorage.DoesNotExist, e:
		return None

def storage_select_all(node=None, name_filter_str=None, tenant_name=None):

	db_set_current_tenant_schema(tenant_name)
	
	elements = MonitoredStorage.objects
	
	if node != None:
		elements = elements.filter(parent=node)
	
	if name_filter_str:
		pattern = '%s' % (name_filter_str)
		ielements = elements.filter(name__startswith=pattern)

	return elements.all()

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

@transaction_commit_on_success
def service_update(service, *fields_name, **options):	
	
	service.validate_fields()
	
	service.concurrent_save(fields_name=fields_name, new_sentinel_value=options.get('mtime',None))

#------------------------------------------------------------------------------#

@transaction_commit_on_success
def service_create(node, service_name):

	# FIX: move to API
	service = MonitoredService(
		tenant = node.tenant, 
		name   = service_name, 
		parent = node,
		kind   = ELEMENT_KIND_SERVICE,
		agent  = node.agent,
		
		owner  = node.owner,
	)
	service.save()

	return service
	
#------------------------------------------------------------------------------#

def service_get(node_name, service_name, tenant_name=None):

	if tenant_name:
		db_set_current_tenant_schema(tenant_name)

	try:
		# FIX: do it with one query node__name
		node = MonitoredNode.objects.get(name=node_name)
		service = MonitoredService.objects.get(name=service_name, parent=node)
		return service
	except MonitoredService.DoesNotExist, e:
		return None

#------------------------------------------------------------------------------#

def service_get_by_id(uuid):
	try:
		service = MonitoredService.objects.get(uuid=uuid)
		return service
	except MonitoredService.DoesNotExist, e:
		return None

#------------------------------------------------------------------------------#

def service_select_all(node=None, name_filter_str=None, tenant_name=None):

	db_set_current_tenant_schema(tenant_name)
	
	elements = MonitoredService.objects
	
	if node != None:
		elements = elements.filter(parent=node)
	
	if name_filter_str:
		pattern = '%s' % (name_filter_str)
		elements = elements.filter(name__startswith=pattern)

	return elements.all()
	
#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

@transaction_commit_on_success
def device_create(node, device_name):

	# FIX: move to API
	device = MonitoredDevice(
		tenant = node.tenant, 
		name   = device_name, 
		parent = node,
		kind   = ELEMENT_KIND_DEVICE,
		agent  = node.agent,
		
		owner  = node.owner,
	)
	device.save()

	return device

#------------------------------------------------------------------------------#

@transaction_commit_on_success
def device_update(device, *fields_name, **options):		
	
	device.validate_fields()
	
	device.concurrent_save(fields_name=fields_name, new_sentinel_value=options.get('mtime',None))
	
	return

#------------------------------------------------------------------------------#

def device_select_all(node=None, name_filter_str=None, tenant_name=None):

	db_set_current_tenant_schema(tenant_name)
	
	elements = MonitoredDevice.objects
	
	if node != None:
		elements = elements.filter(parent=node)
	
	if name_filter_str:
		pattern = '%s' % (name_filter_str)
		elements = elements.filter(name__startswith=pattern)

	return elements.all()

#------------------------------------------------------------------------------#

def device_get(node_name, name, tenant_name=None):

	if tenant_name:
		db_set_current_tenant_schema(tenant_name)

	try:
		# FIX: do it with one query node__name
		node = MonitoredNode.objects.get(name=node_name)
		device = MonitoredDevice.objects.get(name=name, parent=node)
		return device
	except MonitoredDevice.DoesNotExist, e:
		return None

#------------------------------------------------------------------------------#

def device_get_by_id(uuid):
	try:
		device = MonitoredDevice.objects.get(uuid=uuid)
		return device
	except MonitoredDevice.DoesNotExist, e:
		return None

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

@transaction_commit_on_success
def link_rename(tenant, name, new_name):
	
	db_set_current_tenant_schema(tenant.name)
	
	# FIX: validate new_name
	
	l = Link.objects.get(tenant=tenant, name=name)
	
	l.name = new_name
	l.mtime = time.time()
	l.save()

@transaction_commit_on_success
def link_delete(tenant, name):
	
	db_set_current_tenant_schema(tenant.name)
	
	# FIX: validate new_name
	
	l = Link.objects.get(tenant=tenant, name=name)
	
	# This will delete linked elements too
	l.delete()
	

@transaction_commit_on_success
def link_modify(n1, i1, n2, i2, lname=None, ltype=None, ldescr=None, tenant_name=None):

	if tenant_name==None:
		raise Exception("INVALID TENTANT")


	db_set_current_tenant_schema(tenant_name)
	tenant = tenant_get(tenant_name)

	node1 = node_get(n1)
	if node1 == None:
		raise Exception('Node %s not found' % node1)

	iface1 = iface_get(n1, i1)
	if iface1 == None:
		raise Exception('Interface %s not found' % iface1)

	node2 = node_get(n2)
	if node2 == None:
		raise Exception('Node %s not found' % node2)

	iface2 = iface_get(n2, i2)
	if iface2 == None:
		raise Exception('Interface %s not found' % iface2)

	if ltype == None:
		ltype = LINK_TYPE_L2


	#
	# Troviamo i link gia' esistenti che coinvolgono almeno una delle due interfacce
	#
	link_counter = {}
	for le in LinkElement.objects.filter(iface__in=[iface1,iface2]):
		link_counter.setdefault( le.link.uuid , [le.link, 0] )
		if le.iface in [iface1, iface2]:
			link_counter[ le.link.uuid ][1] += 1


	#
	# CERCHIAMO DI CAPIRE SE STIAMO MODIFICANDO UN LINK ESISTENTE O SE NE DOBBIAMO CREARE UNO NUOVO
	#
	link = None
	
	default_lname  = Link.compose_link_name(ltype, node1, iface1, node2, iface2)
	default_lname2 = Link.compose_link_name(ltype, node2, iface2, node1, iface1)

	# I link di tipo L2 sono particolari
	if ltype == LINK_TYPE_L2:

		for existing_link, counter in link_counter.values():

			# Il link coinvolge una o entrambe le interfacce, ma non e' del tipo che ci interessa. Ignoriamo.
			if existing_link.ltype != LINK_TYPE_L2: continue

			#print "Existing link:", existing_link.name, counter

			# Esiste gia' un link di tipo L2 che coinvolge una sola delle interfacce. Non possiamo creare altri link L2 quindi.
			if counter != 2:
				raise Exception("One or both interfaces already belong to a different link of type L2 (link name: %s)" % (le.link.name))

			# Abbiamo trovato il link. Vediamo se il nome da dargli e' specificato e sed dobbiamo usare quello di default.
			link = existing_link

			if lname is None:
				# HACK: In mancanza di un nome "custom", viene impostato un default
				#       Per risolere un bug il nome di default potrebbe essere in due versioni, forziamo alla versione corretta!
				if link.name == default_lname or link.name == default_lname2:
					lname = Link.default_link_name(ltype, node1, iface1, node2, iface2)
				else:
					# custom name, do not touch!
					lname = link.name

			break
	else:
		for existing_link, counter in link_counter.values():

			# Il link non coinvolge entrambe le interfacce oppure non e' del tipo giusto -> Ignoriamo
			if existing_link.ltype != ltype or counter != 2: continue

			# Se l'utente ha specificato un nome dobbiamo, il link da modificare deve avere quel nome.
			if lname is not None:
				if existing_link.name == lname:
					link = existing_link
					break
			else:
				if existing_link.name == default_lname or existing_link.name  == default_lname2:
					link = existing_link
					lname = Link.default_link_name(ltype, node1, iface1, node2, iface2)
				else:
					# custom name, do not touch!
					pass

	#print "-----", link, lname, ldescr

	try:

		if link is None:

			if lname is None:
				lname = Link.default_link_name(ltype, node1, iface1, node2, iface2)
			
			if ldescr is None:
				ldescr = ''

			link = Link(tenant=tenant, name=lname, ltype=ltype, description=ldescr)
			link.save()
		
			le1 = LinkElement(tenant=tenant, link=link, iface=iface1)
			le1.save()
		
			le2 = LinkElement(tenant=tenant, link=link, iface=iface2)
			le2.save()
		else:
			if lname is None:
				lname = Link.default_link_name(ltype, node1, iface1, node2, iface2)

			link.name = lname
			link.ltype = ltype

			if ldescr is not None:
				link.description = ldescr

			link.concurrent_save(fields_name=['name','ltype','description'])
	except IntegrityError, e:
		raise InvalidConfigurationException("Duplicated link name '%s'" % (link.name))

	return


def link_select_all(tenant_name=None, node=None, user=None):
	"""
	Estrare i dati dei link.
	
	Se tenant_name non e' specificato, utilizza il tenant primario.
	
	Se "node" e' specificato, mostra i link che coinvolgono quel nodo, altrimenti
	se "user" e' specificato, mostra i link che coinvolgono una qualunque interfaccia
	visibile dall'utente.
	
	In qualunque altro caso, mostra tutti i link del tenant scelto.
	"""

	#
	# Select tenant
	#
	if tenant_name != None:
		tenant = Tenant.objects.get(name=tenant_name)
	else:
		tenant = Tenant.objects.get(is_primary=True)
	tenant.db_set_current_schema()

	#
	# Find links
	#
	if node is not None:
		links = [ le.link for le in LinkElement.objects.filter(Q(iface__parent=node)).select_related('link').all() ]
	else:
		if user:
			links = tenant.get_links_by_user(user)
		else:
			links = Link.objects.filter(tenant=tenant).all()

	#
	# Load related nodes/ifaces objects
	#
	r = []
	for l in links:
		ifaces = list(LinkElement.objects.filter(link=l).select_related('iface', 'iface__parent').all())

		iface1 = ifaces[0].iface
		iface2 = ifaces[1].iface
		
		node1 = iface1.parent
		node2 = iface2.parent
		
		if node1.name > node2.name:
			node1, iface1, node2, iface2 = node2, iface2, node1, iface1
			

		l.node1  = node1
		l.iface1 = iface1
		l.node2  = node2
		l.iface2 = iface2

		r.append(l)

	def _stable_sort(l1, l2):
		if l1.node1.name  < l2.node1.name : return -1
		if l1.node1.name  > l2.node1.name : return  1
		if l1.iface1.name < l2.iface1.name: return -1
		if l1.iface1.name > l2.iface1.name: return  1
		if l1.node2.name  < l2.node2.name : return -1
		if l1.node2.name  > l2.node2.name : return  1
		if l1.iface2.name < l2.iface2.name: return -1
		if l1.iface2.name > l2.iface2.name: return  1
		if l1.name        < l2.name       : return -1
		if l1.name        > l2.name       : return  1
		return 0

	r = sorted(r, cmp=_stable_sort )

	return r

def link_export_all_links(tenant):

	db_set_current_tenant_schema(tenant.name)
	
	r = {
		'links': [],
		'linked_ifaces': []
	}
	for ld in Link.objects.filter(tenant=tenant).values('uuid', 'ltype', 'name'):
		r['links'].append(ld)

	for led in LinkElement.objects.filter(tenant=tenant).values('link_id','iface_id'):
		r['linked_ifaces'].append(led)

	return r

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def icon_select_all(tenant_name, name_filter_str=None):

	t = tenant_get(tenant_name)
	
	icon_dir_path = os.path.join( tenant_get_media_dir(t.uuid) , 'icons' )	
	
	icons = os.listdir(icon_dir_path)
	
	# Remove extension
	icons = [ i.rsplit(".",-1)[0] for i in icons ]
	
	if name_filter_str:
		icons = [ i for i in icons if i.startswith(name_filter_str) ]

	return sorted(icons)

