# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""
DataSource Store
======================================================================

Introduzione
------------------------------------------------------------------------

La storicizzazione dei datasource (dati storici) e' gestita attraverso un oggetto che implementa la classe AbstractDataSourceStore.

Sanet carica dinamicamente il modulo pyhon che implementa la classe in base alla configurazione in settings.py chiamando le funzioni init_datasource_store_from_conf().


Configurazione di Sanet
------------------------------------------------------------------------

I parametri in default_settings.py sono due:


::
	
	DATASOURCE_STORAGE_HANDLER = '<modulo python>.<classe>'
	
	DATASOURCE_STORAGE_PARAMS  = { 
		
		.... dizionario con i parametri di configurazione del modulo... 
		
	}


API
------------------------------------------------------------------------

Eccezioni
~~~~~~~~~~~~~~

Queste sono le eccezioni sollevate dai metodi della classe AbstractDataSourceStore:

::
	
	DataSourceStoreBaseException
	   |
	   +---- DataSourceStoreException
	   |
	   +---- DataSourceStoreInitException
	   |
	   `---- DataSourceStoreUpdateException


Classi
~~~~~~~~~~~~~~

.. autoclass:: core.api.datasource.AbstractDataSourceStore
	:members:

.. autoclass:: core.api.datasource.NullDataSourceStore
	:members:

Funzioni
~~~~~~~~~~~~~~

.. autofunction:: core.api.datasource.init_datasource_store

.. autofunction:: core.api.datasource.init_datasource_store_from_conf

"""


from sanet_common.pythonutils import load_symbol

from conf import get_configuration

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class DataSourceStoreBaseException(Exception):
	"""
	Base exception related to DataSourceStore
	"""
	pass

class DataSourceStoreException(DataSourceStoreBaseException):
	"""
	Generic exception
	"""
	pass

class DataSourceStoreInitException(DataSourceStoreBaseException):
	"""
	Raised on initialization error
	"""
	pass

class DataSourceStoreUpdateException(DataSourceStoreException):
	"""
	Raised on update
	"""
	pass

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class AbstractDataSourceStore(object):
	"""
	Abstract class for every DataSource store module.
	"""

	def init_store(self, datasource):
		"""
		Call this methods when the datasource is created inside sanet database
		"""
		raise NotImplementedError()
		
	def check_sanet_conf(self):
		"""
		Call sanet configuration for errors of problems.
		"""
		return

	def init_tenant(self, tenant):
		"""
		Init internal structures for the tenant
		"""
		raise NotImplementedError()

	def exists(self, tenant_id, datasource_id):
		raise NotImplementedError()

	def save_value(self, tenant_id, datasource_id, timestamp, value):
		"""
		Update datasource storage
		"""
		raise NotImplementedError()

	def fetch_values(self, tenant_id, datasource_id, ts_start=None, ts_end=None, resolution=None, min_value=None, max_value=None):
		"""
		Fetch values
		"""
		raise NotImplementedError()

	@staticmethod
	def validate_storage_spec_options(options_string):
		"""
		Returns true if "options_string" is a valid configuration string, Raises an exception otherwise.
		"""
		raise Exception("NOT IMPLEMENTED")

	def check_and_update_storage_spec(self, tenant_id, datasource_id, minperiod, gauge_flag, storage_spec_string):
		"""
		Verifies and updates (if necessary) if the current storage settings for datasource <tenant_id>:<datasource_id> match
		with the given options string <storage_spec_string>.
		"""
		raise Exception("NOT IMPLEMENTED")
	
	def stats(self, timeout=None):
		"""
		Verify working status of the datasource store
		"""
		raise Exception("NOT IMPLEMENTED")


#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class NullDataSourceStore(AbstractDataSourceStore):
	"""
	This special store does nothing. 
	
	For test/debug only.
	"""

	def __init__(seflf, *params, **kwargs):
		pass

	def init_store(self, datasource):
		return True

	def check_sanet_conf(self):
		return True

	def init_tenant(self, tenant):
		return True

	def exists(self, tenant_id, datasource_id):
		return True

	def save_value(self, tenant_id, datasource_id, timestamp, value):
		return True

	def fetch_values(self, tenant_id, datasource_id, ts_start=None, ts_end=None, resolution=None, min_value=None, max_value=None):
		msg = "No data with this Datasource Store"
		print __name__, msg
		raise DataSourceStoreException( msg )

	@staticmethod
	def validate_storage_spec_options(options_string):
		return True

	def check_and_update_storage_spec(self, tenant_id, datasource_id, minperiod, gauge_flag, storage_spec_string):
		return True

	def stats(self, timeout=None):
		return { "info": "NullDataSourceStore is working"}

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def init_datasource_store(handler_module_name, init_params):
	"""
	Create a DataSourceStore instance from module *handler_module_name* with initial parameters *init_params*.
	
	:param oid:  handler_module_name 
	:type oid: string
	
	:param host: init_params
	:type host: dictionary
	"""

	handler_class = load_symbol( handler_module_name )
	handler       = handler_class(**init_params)

	return handler

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def init_datasource_store_from_conf():
	"""
	TODO
	"""

	config = get_configuration()

	return init_datasource_store(config.DATASOURCE_STORAGE_HANDLER, config.DATASOURCE_STORAGE_PARAMS)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def check_datasource_store_conf():
	"""
	TODO
	"""

	dss = init_datasource_store_from_conf()

	dss.check_sanet_conf()


