# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

RRDTOOL DataSource Store module
=======================================================================

Per storicizzare i dati dei datasource tramite RRDTOOL e' necessario usare il modulo:

	::
		
		DATASOURCE_STORAGE_HANDLER = 'core.api.datasource.rrd.RRDTOOLDataSourceStore'


Configurazione modulo
-------------------------------------------

I parametri di configurazione del modulo sono:

	+-------------------------------+------------------------------------------------------------------------------------------------------------+
	| Parametro                     | Descrizione                                                                                                |
	+===============================+============================================================================================================+
	| storage_base_dir              | Directory dove salvare i file RRD                                                                          |
	+-------------------------------+------------------------------------------------------------------------------------------------------------+
	| heartbeat_factor              | L'heartbeat dei file RRD viene alcolato moltiplicando il *minperiod* con questo valore.                    |
	+-------------------------------+------------------------------------------------------------------------------------------------------------+
	| rrdcached_flush_before_read   | Se impostato a *true*, qualuneuq operazione di lettura dati causa un flush preventivo del demone RRDCACHED |
	+-------------------------------+------------------------------------------------------------------------------------------------------------+

Esempio:

	::
		
		DATASOURCE_STORAGE_PARAMS  = { 
			# Base directory where the storage will put RRD files ( $DATA_DIR / <tenant_id> /rrds/ <rrd file id>.rrd )
			"storage_base_dir"           : DATA_DIR, 
			
			# RRD file's heartbead = DATASOURCE.min_period * heartbeat_factor
			#heartbeat_factor: 60,
			
			# Send FLUSH command to RRDCACHED before read fetching data from RRDs
			# Be sure RRDCACHED_ADDRESS is in your environment!!
			#"rrdcached_flush_before_read": False
		}


Configurazione avanzata dei datasource
-------------------------------------------

Il parametro *storage-spec* puo' essere utilizzato per specificare parametri specifici di RRDTOOL non gestibili direttamente da Sanet. 

.. danger:: quando viene modificato il parametro *storage-spec* (in un template o in una configurazione specifica), deve essere effettuata una operazione di aggiornamento dei file RRD. NON e' predicibile quanto tempo sia necessaio a RRDTOOL per aggiornare tutti i file RRD potenzialmente coinvolti da questa configurazione.

.. danger:: La modifica della configurazione di Sanet e l'aggioranmento dei file RRD non sono operazioni *atomiche* e/o *transazionali*. In caso di errori di aggiornamento dei file RRD, Sanet non e' in grado di garantire che alcuni file siano stati gia' modificati con una configurazione inconsistente rispetto a quanto specificato nel database di configurazione di Sanet.

La configurazione avviene impostato una stringa con la seguente sintassi:

	::
		
		    <option>=<value> <option>=<value> ...

I parametri che e' possibile specificare sono i seguenti:

	+------+----------+---------+--------------------------------------------------------------------------------------------------------------------------------------------------------------------+
	| Nome | Tipo     | Default | Descrizione                                                                                                                                                        |
	+======+==========+=========+====================================================================================================================================================================+
	|  DST | stringa  |         | I valori ammessi sono: COUNTER, DERIVE. Questo parametro viene considerato solo se il datasource e' impostato di tipo COUNTER (*gauge* = *false*).                 | 
	+------+----------+---------+--------------------------------------------------------------------------------------------------------------------------------------------------------------------+
	|  min | numero   |         | Valore minimo valido per la memorizzazione. Se il valore da salvare e' inferiore, viene storicizzato il valore nullo (NaN).                                        | 
	+------+----------+---------+--------------------------------------------------------------------------------------------------------------------------------------------------------------------+
	|  max | numero   |         | Valore massimo valido per la memorizzazione. Se il valore da salvare e' superiore, viene storicizzato il valore nullo (NaN).                                       | 
	+------+----------+---------+--------------------------------------------------------------------------------------------------------------------------------------------------------------------+

Esempio:

	::
		
		DST=DERIVE min=0 max=5000000000



"""

import os
import os.path
import time
import math
import traceback

import rrdtool

from django.conf import settings

from core.api.base import makedirs_for_webuser

from core.api.datasource import DataSourceStoreException
from core.api.datasource import DataSourceStoreUpdateException

from core.api.datasource import AbstractDataSourceStore

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

try:
	_RRD_VERSION = tuple([int(n) for n in rrdtool.__version__.split('.')])
except Exception, e:
	_RRD_VERSION = (0,0,0)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

DEFAULT_STEP = 60                    # 60 seconds

DEFAULT_HEARTBEAT_FACTOR = 60        # 60 times the step (minperiod)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

import sys
import socket
import select
import ctypes
from fcntl import ioctl
from termios import FIONREAD

def get_sock_available_bytes(sock):
	store_buffer = ctypes.c_int()
	ioctl(sock, FIONREAD, store_buffer )
	return store_buffer.value


def rrdcached_stats(server_address, timeout=3):

	server_address = server_address.replace("unix:", "")

	sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
	#
	# CONNECT
	#
	sock.connect(server_address)

	#
	# SEND
	#
	sock.sendall(b"STATS\n")

	#
	# RECEIVE
	#

	# Close write channel. Wait for all "read" data + EOF
	sock.shutdown(socket.SHUT_WR)

	data = ''
	while 1:
		r, w, e = select.select([sock],[],[sock], timeout)
		if sock in r:
			buff_size = get_sock_available_bytes(sock)
			buff = sock.recv(buff_size)
			if len(buff) > 0: # EOF?
				data += buff
			else:
				break

	return data

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class BaseRRDTOOLDataSourceStore(AbstractDataSourceStore):

	GLOBAL_RRD_DIRNAME = 'rrds'

	RRD_UNKNOWN_VALUE = 'U'

	def __init__(self, storage_base_dir, heartbeat_factor=DEFAULT_HEARTBEAT_FACTOR, rrdcached_flush_before_read=False, RRDCACHED_ADDRESS=None):

		super(BaseRRDTOOLDataSourceStore, self).__init__()

		self.storage_base_dir            = storage_base_dir 
		self.heartbeat_factor            = heartbeat_factor
		self.rrdcached_flush_before_read = rrdcached_flush_before_read

		self._rrdcached_address         = RRDCACHED_ADDRESS or os.environ.get('RRDCACHED_ADDRESS', None)

		return

	#-----------------------------------------------------------------------------#

	def init_tenant(self, tenant):
		return self._init_tenant(tenant)

	def exists(self, tenant_id, datasource_id):
		return self._exists(tenant_id, datasource_id)

	def save_value(self, tenant_id, datasource_id, timestamp, value):
		return self._save_value(tenant_id, datasource_id, timestamp, value)

	def fetch_values(self, tenant_id, datasource_id, ts_start=None, ts_end=None, resolution=None, min_value=None, max_value=None):
		return self._fetch_values(tenant_id, datasource_id, ts_start=ts_start, ts_end=ts_end, resolution=resolution, min_value=min_value, max_value=max_value)

	def check_sanet_conf(self):
		return self._check_sanet_conf()

	@staticmethod
	def validate_storage_spec_options(options_string):
		raise Exception("NOT IMPLEMENTED")

	def check_and_update_storage_spec(self, tenant_id, datasource_id, minperiod, gauge_flag, storage_spec_string):
		raise Exception("NOT IMPLEMENTED")

	#-----------------------------------------------------------------------------#
	# Metodi specifici per la gestione tramite RRDTOOL
	#-----------------------------------------------------------------------------#

	def _init_tenant(self, tenant):
		rrds_dir  = self._get_tenant_dir(tenant.uuid)
		if not os.path.exists(rrds_dir):
			makedirs_for_webuser(rrds_dir)

	def _exists(self, tenant_id, datasource_id):
		dirname, filename = self._get_datasource_rrd_filename(tenant_id, datasource_id)
		return os.path.exists(filename)

	def _save_value(self, tenant_id, datasource_id, timestamp, value):

		dirname, filename = self._get_datasource_rrd_filename(tenant_id, datasource_id)

		rrd_update_str = "%s:%s" % (int(timestamp), str(value) if value != None else BaseRRDTOOLDataSourceStore.RRD_UNKNOWN_VALUE)

		try:
			if self._rrdcached_address:
				#
				# FIX: when _rrdcached_address is not a unix socket absolute path are not allowed.
				#
				#           filename = filename[len(self.storage_base_dir)+1:]
				#
				#args = ['--skip-past-updates', '--daemon', self._rrdcached_address, filename, rrd_update_str]
				args = ['--daemon', self._rrdcached_address, filename, rrd_update_str]
			else:
				#args = ['--skip-past-updates', filename, rrd_update_str]
				args = [filename, rrd_update_str]

			rrdtool.update(*args)

		except rrdtool.error, er:
			raise DataSourceStoreUpdateException("Unable to update %s with %s: %s" % (datasource_id, value, str(er)))

		except Exception, er:
			raise Exception("Unknown error while updating RRDTOOL FILE: %s:%s -> %s" % (type(er), str(er), traceback.format_exc()))

	def _fetch_values(self, tenant_id, datasource_id, ts_start=None, ts_end=None, resolution=None, min_value=None, max_value=None):

		dirname, file_name = self._get_datasource_rrd_filename(tenant_id, datasource_id)

		try:
			#
			# FLUSH BEFORE READ
			#
			if self._rrdcached_address and self.rrdcached_flush_before_read:
				rrdtool.flushcached('--daemon', self._rrdcached_address, file_name)
			else:
				pass

			#
			# SET TIME RANGE
			#
			scalex_orig = [ts_start, ts_end, resolution]

			now = int(time.time())

			if ts_start == None:
				ts_start = now - 60*60*24*7

			if ts_end == None:
				ts_end = now

			ts_start = int(ts_start)
			ts_end   = int(ts_end)

			#
			# READ DATA from FILE (or DAEMON)
			#
			rrd_params = []

			if resolution == None:
				rrd_params += [ 'AVERAGE',                        '-s', str(ts_start), '-e', str(ts_end) ]
			else:
				rrd_params += [ 'AVERAGE', '-r', str(resolution), '-s', str(ts_start), '-e', str(ts_end) ]

			if self._rrdcached_address:
				rrd_params += [ '--daemon', self._rrdcached_address ] + rrd_params

			data = rrdtool.fetch(file_name, *rrd_params)

			# [
			#   (<ts start>, <ts_end>, <step> ),
			#   (<dsname>, ),
			#   ( (<val1,), (val2,) , ... ),
			# ]
			
			
			# FIX: too much precision makes JSONized floats too long.
			#      This is not a good solution in some cases.
			#def _round(v, n):
			#	e = (10**n)
			#	return float(int(v * e)/e)
				
			dataseries = []
			for v in data[2]:
				v = v[0]

				# FIX: disabled.
				#if v is not None:
				#	v = _round(v,6)

				if min_value is not None: 
					if v < min_value: v = None
				if max_value is not None: 
					if v > max_value: v = None
				dataseries.append(v)

			#
			# Remove None values from RRD output
			#
			e_ts_start = data[0][0]
			e_ts_end   = data[0][1]
			rrd_resolution = data[0][2]

			# Alle volte rrd restituisce dei valori nel futuro che sono nulli. Togliamoli
			n = 0
			while e_ts_end > ts_end:
				del dataseries[-1]
				e_ts_end -= rrd_resolution
				n += 1

			scalex = [e_ts_start, e_ts_end, rrd_resolution ]

			#
			# Adjust resolution
			#
			if resolution is not None:
				if resolution != rrd_resolution:
					scalex, dataseries = adjust_dataserie_resolution(scalex, dataseries, resolution)

			r = {
				"scalex"   : scalex,
				"scaley"   : None,
				"dataseries": dataseries,
				"info"     : {
					'removed'           : n,
					'len'               : len(dataseries),
					'resolution'        : resolution,
					'params'            : scalex_orig,
					'altered_resolution': rrd_resolution,
				}
			}

			return r

		except rrdtool.error, er:
			raise DataSourceStoreException("Can't fetch data for datasource %s: %s" % (file_name, er))

	def _check_sanet_conf(self):

		import sys
		from sanet_common.termutils import term_red

		from_settings = settings.DATASOURCE_STORAGE_PARAMS.get('RRDCACHED_ADDRESS', None)
		from_env      = os.environ.get('RRDCACHED_ADDRESS', None)

		if from_settings != from_env:
			print term_red("""*** The RRDCACHED_ADDRESS configuration is invalid ***\nSANET setttings: %s\nEnvironment    : %s\nCheck your settings.py file and your environment!\n""" % (from_settings, from_env))
			sys.exit(-1)

	def _get_tenant_dir(self, tenant_id):

		dir_path = os.path.join(
			self.storage_base_dir,
			tenant_id,
			BaseRRDTOOLDataSourceStore.GLOBAL_RRD_DIRNAME,
		)

		return dir_path

	def _get_datasource_rrd_filename(self, tenant_id, datasource_key):

		dir_path = self._get_tenant_dir(tenant_id)

		file_name = datasource_key + '.rrd' # FIX: convert special characters

		# FIX: use os.path
		file_path = os.path.join(dir_path, file_name)

		return (str(dir_path), str(file_path))

	def init_store(self, datasource):

		dirname, filename = self._get_datasource_rrd_filename(datasource.tenant.uuid, datasource.uuid)

		if os.path.exists(filename):
			return False

		try:
			makedirs_for_webuser(dirname)
		except Exception, e:
			if e.errno != 17:
				raise DataSourceStoreException("Unable to create storage dir (%s): %s:%s" % (dirname, type(e), str(e)))

		self.create_rrd_file(filename, datasource)

		return True

	def create_rrd_file(self, file_name, datasource, start_ts=0):

		step = datasource.datagroup.minperiod
		if step == 0:
			step = DEFAULT_STEP

		dtype     = 'GAUGE' if datasource.gauge else 'COUNTER'
		heartbeat = step * self.heartbeat_factor

		maxv = BaseRRDTOOLDataSourceStore.RRD_UNKNOWN_VALUE
		minv = BaseRRDTOOLDataSourceStore.RRD_UNKNOWN_VALUE

		ds_def = "DS:data:%s:%s:%s:%s" % (dtype, heartbeat, minv, maxv)

		ONE_SECOND = 1
		ONE_MINUTE = ONE_SECOND * 60
		ONE_HOUR   = ONE_MINUTE * 60
		ONE_DAY    = ONE_HOUR   * 24
		ONE_YEAR   = ONE_DAY    * 365

		max_period = ONE_YEAR

		series = [
			#
			# When we use only 1 consolidated data point for a RRA
			# all consolidation functions are the same.
			#
			#'RRA:LAST:0.5:1:%s'   %  ( max_period / step ),
			#'RRA:MAX:0.5:1:%s'    %  ( max_period / step ),
			#'RRA:MIN:0.5:1:%s'    %  ( max_period / step ),
			'RRA:AVERAGE:0.5:1:%s'    %  ( max_period / step ),
			#'RRA:AVERAGE:0.5:5:%s' % ( max_period/ (step * (ONE_DAY/step)))
		]

		try:

			rrd_params = []

			rrd_params += [
				'--step', str(step), 
				ds_def, 
			] + series

			if start_ts:
				rrd_params += ['--start', str(start_ts)]

			rrdtool.create(str(file_name), *rrd_params)

		except rrdtool.error, er:
			raise DataSourceStoreException("Can't create RRD file %s: %s" % (file_name, er))

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class RRDTOOLDataSourceStoreNormal(BaseRRDTOOLDataSourceStore):

	@staticmethod
	def validate_storage_spec_options(options_string):
		# Ignore storage_spec options
		return True

	def check_and_update_storage_spec(self, tenant_id, datasource_id, minperiod, gauge_flag, storage_spec_string):
		# Ignore storage_spec options
		return True

	def stats(self, timeout=None):
		return {}

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

import pprint

class RRDTOOLDataSourceStore(BaseRRDTOOLDataSourceStore):

	@staticmethod
	def validate_storage_spec_options(options_string):
		r = RRDTOOLDataSourceStore._parse_storage_spec_options(options_string)
		return True

	def check_and_update_storage_spec(self, tenant_id, datasource_id, minperiod, gauge_flag, storage_spec_string):
		return self._check_and_update_storage_spec(tenant_id, datasource_id, minperiod, gauge_flag, storage_spec_string)

	#-----------------------------------------------------------------------------#
	#                                                                             #
	#-----------------------------------------------------------------------------#

	def _check_and_update_storage_spec(self, tenant_id, datasource_id, minperiod, gauge_flag, storage_spec_string):

		# If the RRD file does not exists, return without changes
		dirname, filename = self._get_datasource_rrd_filename(tenant_id, datasource_id)
		if not os.path.exists(filename):
			#print "\t", "NO EXIST"
			return False

		#
		# Parse NEW settings
		#
		new_settings = RRDTOOLDataSourceStore._parse_storage_spec_options(storage_spec_string)
		if new_settings is None:
			new_settings = {}

		# Set default for missing options
		if not new_settings.has_key('DST'): new_settings['DST'] = 'GAUGE' if gauge_flag else 'COUNTER'
		if not new_settings.has_key('min'): new_settings['min'] = None 
		if not new_settings.has_key('max'): new_settings['max'] = None 

		# heatbeat based on default step or the datasource (datagroup) minperiod
		step = minperiod
		if step == 0:
			step = DEFAULT_STEP

		new_settings['heartbeat'] = step * DEFAULT_HEARTBEAT_FACTOR

		#
		# Read current file settings
		#
		try:
			file_data = rrdtool.info(filename)
		except rrdtool.error, er:
			raise DataSourceStoreException("Unable to read RRD file infos %s: %s" % (filename, er))
		except Exception, er:
			raise DataSourceStoreException("Unable to read RRD file infos %s: %s:%s" % (filename, type(e), str(e)))

		current_settings = {
			'DST'      : file_data['ds[data].type'             ],
			'heartbeat': file_data['ds[data].minimal_heartbeat'],
			'min'      : file_data['ds[data].min'              ],
			'max'      : file_data['ds[data].max'              ],
		}

		#
		# Update ?
		#
		if current_settings == new_settings:
			#print "\t", "NOT CHANGED"
			return False

		try:
			minv = new_settings['min'] if new_settings['min'] is not None else BaseRRDTOOLDataSourceStore.RRD_UNKNOWN_VALUE
			maxv = new_settings['max'] if new_settings['max'] is not None else BaseRRDTOOLDataSourceStore.RRD_UNKNOWN_VALUE

			rrd_params = [
				'--data-source-type' , 'data:%s' % (new_settings['DST'      ]),
				'--heartbeat'        , 'data:%s' % (new_settings['heartbeat']),
				'--minimum'          , 'data:%s' % (minv),
				'--maximum'          , 'data:%s' % (maxv),
			]

			rrdtool.tune(str(filename), *rrd_params)

			return True

		except rrdtool.error, er:
			raise DataSourceStoreException("Unable to update RRD configuration for file %s: %s" % (filename, er))
		except Exception, er:
			raise DataSourceStoreException("Unable to update RRD configuration for file %s: %s:%s" % (filename, type(er), str(er)))

	#-----------------------------------------------------------------------------#

	#
	#
	#    <option>=<value> <option>=<value> ...
	#
	#
	#
	#   DST=[COUNTER | DERIVE | DCOUNTER | DDERIVE]
	#   min=<value>
	#   max=<value>
	#

	VALID_STORAGE_SPEC_OPTIONS = {
		'DST': str,
		'min': int,
		'max': int,
	}

	@staticmethod
	def _parse_storage_spec_options(options_string):

		if not options_string:
			return None

		try:
			r = {}

			parts = options_string.split(' ')
			for part in parts:
				part = part.strip()
				name, value = part.split('=')

				r[name] = RRDTOOLDataSourceStore.VALID_STORAGE_SPEC_OPTIONS[ name ](value)

			return r

		except Exception, e:
			raise ValueError("Invalid syntax or format or parameter: %s" % (options_string))

	def create_rrd_file(self, file_name, datasource, start_ts=0):

		step = datasource.datagroup.minperiod
		if step == 0:
			step = 60

		#
		# DEFAULT CONFIGURATION
		#
		dtype     = 'GAUGE' if datasource.gauge else 'COUNTER'

		heartbeat = step * self.heartbeat_factor

		maxv = RRDTOOLDataSourceStore.RRD_UNKNOWN_VALUE
		minv = RRDTOOLDataSourceStore.RRD_UNKNOWN_VALUE

		#
		# SPECIFIC STORAGE OPTIONS
		#
		if dtype == 'COUNTER':
			storage_spec_options = RRDTOOLDataSourceStore._parse_storage_spec_options(datasource.storage_spec)
			if storage_spec_options is not None:
				dtype = storage_spec_options.get('DST', dtype)
				maxv  = storage_spec_options.get('max', maxv)
				minv  = storage_spec_options.get('min', minv)

		ds_def = "DS:data:%s:%s:%s:%s" % (dtype, heartbeat, minv, maxv)

		ONE_SECOND = 1
		ONE_MINUTE = ONE_SECOND * 60
		ONE_HOUR   = ONE_MINUTE * 60
		ONE_DAY    = ONE_HOUR   * 24
		ONE_YEAR   = ONE_DAY    * 365

		max_period = ONE_YEAR

		series = [
			#
			# When we use only 1 consolidated data point for a RRA
			# all consolidation functions are the same.
			#
			#'RRA:LAST:0.5:1:%s'   %  ( max_period / step ),
			#'RRA:MAX:0.5:1:%s'    %  ( max_period / step ),
			#'RRA:MIN:0.5:1:%s'    %  ( max_period / step ),
			'RRA:AVERAGE:0.5:1:%s'    %  ( max_period / step ),
			#'RRA:AVERAGE:0.5:5:%s' % ( max_period/ (step * (ONE_DAY/step)))
		]

		try:
			rrd_params = []


			if start_ts:
				rrd_params += ['--start', str(start_ts)]

			rrd_params += [
				'--step', str(step), 
				ds_def, 
			] + series

			# IMPORTANT: rrdcreate doesn't support --daemon before version 1.6
			if self._rrdcached_address and _RRD_VERSION >= (1,6,0):
				rrd_params += [
					'--daemon', self._rrdcached_address
				]

			rrdtool.create(str(file_name), *rrd_params)

		except rrdtool.error, er:
			raise DataSourceStoreException("Can't create RRD file %s: %s" % (file_name, er))


	def stats(self, timeout=None):

		if self._rrdcached_address:

			data = rrdcached_stats(self._rrdcached_address, timeout=timeout)
		else:
			data = {}

		return data


def adjust_dataserie_resolution(scalex, dataseries, resolution):
	"""
	A partire dalla serie di dati "dataseries" caratterizzata dalle specifiche 
	temporali in scalex ( <ts_start, ts_end, resolution>  )
	la fuznione calcola una nuova serie dati in base alla nuova risoluzione
	specificata.
	"""

	old_ts_start   = scalex[0]
	old_ts_end     = scalex[1]
	old_resolution = scalex[2]

	new_ts_start = old_ts_start
	new_ts_end   = old_ts_end  

	if resolution > old_resolution:
		
		#
		# Se la nuova risoluzione e' piu' alta della precedente, dobbiamo
		# prendere N valori della vecchia serie e mostrarne solo uno.
		#

		# secondi   0   1   2   3   4   5   6   7   8   9   10  11  12
		#           |   |   |   |   |   |   |   |   |   |   |   |   |   
		#
		# DATA      |   2   |   3   |   4   |  -5   |  -11  |  13   |
		#
		# NEW1      |     3     |     4     |    -11    |     13    |
		#
		# NEW2      |       3       |       -5      |       13      |

		new_dataseries = []

		last_value    = 0
		value_to_show = last_value
		last_dt       = 0

		i = 0
		ts            = new_ts_start
		next_data_ts  = new_ts_start + old_resolution
		next_new_ts   = new_ts_start + resolution

		for i, val in enumerate(dataseries):

			if val is not None:
				if value_to_show is None:
					value_to_show = val
				else:
					# Invece di prendere il MAX tra i valori
					# prendiamo quello che in assoluto rappresenta il picco piu' "significativo"
					# Questo dovrebbe permettere di mostrare degli aggregati interessanti
					# anche in presenza di valori negativi
					if (abs(val) >= abs(value_to_show)):
						value_to_show = val

			last_value = val

			ts += old_resolution

			if ts >= next_new_ts:

				new_dataseries.append(value_to_show)

				value_to_show = val
				value_to_show = None

				next_new_ts += resolution

	else:
		# secondi   0   1   2   3   4   5   6   7   8   9   10  11  12  13  14  15  16  17  18  19  20  21  22  23  24
		#           |   |   |   |   |   |   |   |   |   |   |   |   |   |   |   |   |   |   |   |   |   |   |   |   |   
		#
		# DATA      |       2       |       3       |       4       |       -5      |      11       |      13       |
		#
		# NEW1      |     2     |     3     |     3     |     4     |    -5     |    -5     |     11    |     13    |
		#
		# NEW2      |   2   |   2   |   -5  |  -5   |    4  |   4   |   -5  |  -5   |  11   |  11   |  13   |  13   |
                       
		
		new_dataseries = []

		current_ts      = old_ts_start
		current_rrd_pos = 0

		while current_ts < new_ts_end:

			if current_rrd_pos < len(dataseries):
				new_dataseries.append( dataseries[current_rrd_pos] )
			else:
				new_dataseries.append( dataseries[current_rrd_pos-1] )
			current_ts += resolution

			if (current_ts >= (old_ts_start + (current_rrd_pos+1) * old_resolution)) and current_rrd_pos < (len(dataseries)-1):
				current_rrd_pos += 1

	new_scalex = [new_ts_start, new_ts_end, resolution ]

	return new_scalex, new_dataseries












#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

if __name__ == "__main__":

	import os
	import sys
	import time
	import random


	base_path = os.environ["SANET_HOME"]
	if not base_path in sys.path:
		sys.path.insert(0, base_path)
	os.environ.setdefault("DJANGO_SETTINGS_MODULE", "conf.settings")

	from conf import get_configuration
	from core.models import *

	from core.api.datasource.rrd import RRDDataSourceStore

	t = Tenant(name='t1')
	dg = DataGroup(minperiod=1)

	ds = DataSource(path='prova;prova',
		tenant = t,
		datagroup=dg,
		absolute_value = True, 
		min_val        = None,
		max_val        = None,
		storage_spec   = None
	)

	store = RRDDataSourceStore(get_configuration().DATA_DIR, heartbeat_factor=2)

	if len(sys.argv) > 1:
		data = store.fetch_last(ds, int(sys.argv[1]))
	else:
		random.seed(os.getpid())

		ts_start = int(time.time())
		for i in xrange(1000):
			t = int(time.time())
			v = random.randint(0, 10000)

			store.save_value(ds, t, v)
