# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import os

from django.db import transaction
from django.db.models import Q

from conf import get_configuration
DEBUG = get_configuration().DEBUG_CONFIG or os.environ.get('DEBUG_CONFIG',False)

from core.db_utils import transaction_commit_on_success

from core.models import DataGroup, DataGroupParameter
from core.models import CDataGroup
from core.models import CDataGroupParameterValue
from core.models import CDataGroupFieldValue
from core.models import TimeGraph
from core.models import TField
from core.models import CTimeGraphFieldValue
from core.models import CTimeGraphSerieFieldValue
from core.models import CDataSourceFieldValue
from core.models import CConditionFieldValue
from core.models import TimeGraphSerie

from core.api.base import db_set_current_tenant_schema
from core.api.templates import *

from core.api.templates import refresh_element_template

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

from sanet_common.sanetdefs.constants import DATAGROUP_PATH_SEP

from django.db import IntegrityError

from core.exceptions import InvalidConfigurationException
from core.models import MonitoredElement

@transaction_commit_on_success
def element_rename(element, new_name, user=None):

	#
	# FIX: eventuali collisioni sui nomi sono verificate dal DB e dai vincoli di integrita',
	#      ma effettuiamo comunque il controllo nella speranza di sollevare il DB
	#      da inutili perdite di tempo.
	#

	db_set_current_tenant_schema(element.tenant.name)

	if user:
		user.assert_capability(element.tenant, 'core.element.conf_monitoring')

	q = MonitoredElement.objects.filter(tenant=element.tenant, kind=element.kind, name=new_name)
	if element.parent:
		q = q.filter(parent=element.parent)
	if q.count() > 0:
		raise InvalidConfigurationException("Element '%s' already exists!" % (new_name), [])

	# Rename
	try:
		return element_rename_raw(element, new_name)
	except IntegrityError, e:
		raise InvalidConfigurationException("Element '%s' already exists!" % (new_name), [str(e)])
	
	
def element_rename_raw(element, new_name):

	mtime = time.time()
	
	element_old_path = element.calculate_path()
	
	# Rename del nome
	element.name = new_name

	# Prima aggiorniamo i figli in questo modo evitiamo strane race condition
	# quando entra in gioco l'aggiornamento concorrente nei vari livelli
	for subelement in element.monitoredelement_set.filter(parent=element):
		_update_element_paths(element, subelement, mtime)
	
	_update_element_paths(element.parent, element, mtime)
	
	return

def _update_element_paths(parent, element, mtime):

	ppath = parent.calculate_path() if parent is not None else None

	element_new_path = element.calculate_path_with_parent(ppath)
	
	for dg in element.datagroups.all():
		
		# FIX: trasformare in un metodo del modello
		path = dg.path
		path = path.split(DATAGROUP_PATH_SEP)
		path[0] = element_new_path
		path = DATAGROUP_PATH_SEP.join(path)
		
		dg.path = path

		dg.concurrent_save(['path'], new_sentinel_value=mtime)

	for ds in element.monitored_datasources.all():

		# FIX: trasformare in un metodo del modello
		path = ds.path
		path = path.split(DATAGROUP_PATH_SEP)
		path[0] = element_new_path
		path = DATAGROUP_PATH_SEP.join(path)
		
		ds.path = path

		ds.concurrent_save(['path'], new_sentinel_value=mtime)

	for c in element.monitored_conditions.all():
		
		# FIX: trasformare in un metodo del modello
		path = c.path		
		path = path.split(DATAGROUP_PATH_SEP)
		path[0] = element_new_path
		path = DATAGROUP_PATH_SEP.join(path)
		
		c.path = path

		c.concurrent_save(['path'], new_sentinel_value=mtime)

	# Questo dovrebbe assicurarci che l'aggiornamento dell'elemento e 
	# tutti i sotto-elementi sia cmq atomico.
	# Es:  utente A aggiorna node1  e utente B aggiorna solo node1:eth0
	element.concurrent_save(['name'], new_sentinel_value=mtime)
	

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

from sanet_common.sanetdefs.constants import *

@transaction_commit_on_success
def element_suspend(monitored_element, resume_time, reason, user=None):

	db_set_current_tenant_schema(monitored_element.tenant.name)

	if user:
		user.assert_capability(monitored_element.tenant, 'core.element.conf_monitoring')

	suspend_time = int(time.time())

	for dg in monitored_element.get_all_datagroups():
		if not dg.suspended:
			dg.suspend(suspend_time=suspend_time, resume_time=resume_time, reason=reason)

@transaction_commit_on_success
def datagroup_suspend(datagroup, resume_time=0, reason=None, user=None):

	db_set_current_tenant_schema(datagroup.tenant.name)

	if user:
		user.assert_capability(datagroup.tenant, 'core.element.conf_monitoring')

	suspend_time = int(time.time())

	datagroup.suspend(suspend_time=suspend_time, resume_time=resume_time, reason=reason)

#------------------------------------------------------------------------------#

@transaction_commit_on_success
def element_resume(monitored_element, reason, user=None):

	db_set_current_tenant_schema(monitored_element.tenant.name)

	if user:
		user.assert_capability(tenant, 'core.element.conf_monitoring')

	resume_time = int(time.time())

	for dg in monitored_element.get_all_datagroups():
		#_datagroup_resume(dg, reason=reason)
		if dg.suspended:
			dg.resume(resume_time=resume_time, reason=reason)

@transaction_commit_on_success
def datagroup_resume(datagroup, reason, user=None):

	db_set_current_tenant_schema(datagroup.tenant.name)

	if user:
		user.assert_capability(datagroup.tenant, 'core.element.conf_monitoring')

	resume_time = time.time()

	datagroup.resume(resume_time=resume_time, reason=reason)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

@transaction_commit_on_success
def element_set_template(element, new_telement, user=None):
	"""
	Associa all'elemento "element" il l'element-template  "new_telement".
	"""

	return _element_set_template(element, new_telement, user=user)

def _element_set_template(element, new_telement, user=None):

	from core.models import DataGroup, DataGroupParameter, Condition, DataSource
	from core.models import TEDataGroup

	tenant = element.tenant
	db_set_current_tenant_schema(tenant.name)

	if user:
		user.assert_capability(tenant, 'core.element.conf_monitoring')

	## Vediamo se l'elemento aveva gia' un template associato
	#
	try:
		e_t = ElementTemplate.objects.get(tenant=tenant, melement=element, telement=new_telement)
		
		if e_t.telementappliedtemplate is not None:
			raise InvalidConfigurationException("Tring to set template '%s' for element '%s', but this reference already exists because element was generated by template '%s'" % (
				new_telement.name,
				element.path,
				element.telement_generator.name,
			))

		if DEBUG: print "FOUND", tenant, element, new_telement

		return

	except ElementTemplate.DoesNotExist, e:
		e_t = None

	try:
		applyorder = ElementTemplate.objects.filter(tenant=tenant, melement=element).latest('applyorder').applyorder + 1
	except ElementTemplate.DoesNotExist, e:
		applyorder = 0

	# Create association
	e_t = ElementTemplate(tenant=tenant, melement=element, telement=new_telement, applyorder=applyorder)
	e_t.save()

	telement = new_telement

	refresh_element_template(tenant, telement, element)

	return

#------------------------------------------------------------------------------#

@transaction_commit_on_success
def element_unset_template(element, element_template_name=None, user=None):

	return _element_unset_template(element, element_template_name, user=user)

def _element_unset_template(element, element_template_name=None, user=None): #if element_template_name is None unset all templates associated to the element (DA CHIEDERE)

	tenant = element.tenant

	db_set_current_tenant_schema(tenant.name)

	if user:
		user.assert_capability(tenant, 'core.element.conf_monitoring')

	try:
		if element_template_name:
			
			el_t = ElementTemplate.objects.get(tenant=tenant, melement=element, telement__name=element_template_name) #retrive the association from the element and the requested template
			element_template_list = [el_t]
			
		else:
			element_template_list = ElementTemplate.objects.filter(tenant=tenant, melement=element)
			
	except ElementTemplate.DoesNotExist, e:
		if DEBUG: print "%s not found for element %s" % (element_template_name, element)
		element_template_list = None

	if element_template_list == None:
		return
	
	for e_t in element_template_list:
		#
		# Questa associazione e' stata creata perche' l'elemento e' stato generato da un sub-element template
		# a cui sono associati tanti element-template di configurazione.
		#
		if e_t.telementappliedtemplate is not None:
			raise InvalidConfigurationException("Can not remove reference to template '%s'. This reference belongs to sub-element template '%s'" % (e_t.telement.name, e_t.telementappliedtemplate.telement.name))
		
		
		telement = e_t.telement
		
		print '*** Sto rimuovendo il template %s ***' % (str(e_t.telement.name))
		#
		# Remove all sub-elements created from from this template
		#
		if DEBUG: print "Removing sub-elements"
		
		for sub_element_template in telement.get_sub_elements():
			if DEBUG: print "Removing sub-element", sub_element_template.name
			for subelement in element.get_sub_elements().filter(telement_generator=sub_element_template).all():
				
				# Removing datagroup linked to 
				_element_unset_template_clean_pending_datagroups(element.tenant, subelement, sub_element_template)
				
				subelement.delete()
		
		#
		# Remove datagroup
		#
		
		
		if DEBUG: print "Removing association with old template", e_t
		e_t.delete()

		_element_unset_template_clean_pending_datagroups(tenant, element, telement)
	
	#FIX: refresh delle sole dipendenze
	if len(element_template_list) == 1:
		refresh_element(element)
	
	return


def _element_unset_template_clean_pending_datagroups(tenant, element, telement):

	# Rimuoviamo tutti i datagroup / configurazioni a livello di elemento che 
	# dipendono dall'element-template
	if DEBUG: print "Removing configured datagroups"
	cdgs = CDataGroup.objects.filter(tenant=tenant, element=element, tdatagroup__isnull=True, tedatagroup__telement=telement).all()
	for cdg in cdgs:
		if DEBUG: print "\t", cdg

		# Cancelliamo tutti i datagroup effettivi
		DataGroup.objects.filter (tenant=tenant, element=element, cdatagroup=cdg).delete()
		
		# Cancelliamo tutte le configurazioni
		cdg.delete()

	if DEBUG: print "Removing template-datagroup"
	for tedg in TEDataGroup.objects.filter(tenant=tenant, telement=telement).all():
		if DEBUG: print "\t", tedg
		for cdg in CDataGroup.objects.filter(tenant=tenant, tedatagroup=tedg, element=element):
			if DEBUG: print "\t", "\t", "Removing cdg ", cdg
			cdg.delete()

	return

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

#@transaction_commit_on_success
def element_datagroup_create_from_element_template(element, datagroup_template, instance_name, tenant, mtime):

	return _element_datagroup_create(element, None, datagroup_template, instance_name, tenant, mtime=mtime)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def _element_datagroup_create(element, tdatagroup, tedatagroup, instance_name, tenant, mtime=None, user=None):

	mtime = time.time() if mtime is None else mtime

	db_set_current_tenant_schema(tenant.name)

	if user:
		user.assert_capability(tenant, 'core.element.conf_monitoring')

	cdg, created = CDataGroup.objects.get_or_create(
		tenant=tenant,
		element=element,
		tdatagroup=tdatagroup, 
		tedatagroup=tedatagroup, 
		name=instance_name,
	)
	
	if created:
		cdg.mtime = mtime
		cdg.save()

	#
	# Create Datagroup from a Configured Datagroup
	#
	try:
		d = DataGroup.objects.get(
			tenant     = tenant,
			element    = element,
			name       = instance_name,
			cdatagroup = cdg,
		)
	except DataGroup.DoesNotExist, e:
		d = DataGroup(
			tenant     = tenant,
			element    = element,

			name       = instance_name,
			cdatagroup = cdg,

			agent      = element.agent,

			owner      = element.owner,
		)

	d.update_from_config(mtime=mtime)
	#d.save()

	return cdg

#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#                                                                              #
#                                                                              #
# ADDITIONAL                                                                   #
#                                                                              #
#                                                                              #
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#

def element_additional_datagroup_get(element, instance_name, tenant):

	db_set_current_tenant_schema(tenant.name)

	created = False
	try:
		cdg = CDataGroup.objects.get(tenant=tenant, element=element, tedatagroup__isnull=True, tdatagroup__isnull=False, name=instance_name, )
		return cdg
	except CDataGroup.DoesNotExist, e:
		return None

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

@transaction_commit_on_success
def element_datagroup_create_additional(element, datagroup_template, instance_name, tenant, user=None):

	return _element_datagroup_create(element, datagroup_template, None, instance_name, tenant, user=user)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

@transaction_commit_on_success
def element_datagroup_remove_additional(element, datagroup_instance_name, user=None):

	mtime = time.time()

	tenant = element.tenant

	db_set_current_tenant_schema(tenant.name)

	if user:
		user.assert_capability(tenant, 'core.element.conf_monitoring')

	try:
		cdg = CDataGroup.objects.get(tenant=tenant, element=element, name=datagroup_instance_name,  tedatagroup=None, tdatagroup__isnull=False)
	except CDataGroup.DoesNotExist, cdg:
		if DEBUG: print __file__, "element_datagroup_remove_additional()", "additional datagroup not defined"
		cdg = None

	if cdg is None:
		return

	#
	# FIX: dobbiamo ricordarci di aggiornare tutti i datagroup influenzati dall'eliminazione di questo elemento
	#

	#dg = cdg.get_datagroup()
	#datagroups_to_update = [ dg.uuid for dg in dg.get_depending_datagroups() ]

	#
	# Delete
	#

	#
	# >>>> CONCURRENT DELETE 
	#
	cdg.delete() # IMPORTANTE: non e' concorrente

	#for dg in DataGroup.objects.filter(tenant=tenant, uuid__in=datagroups_to_update).all():
	#	dg.update_from_config(mtime=mtime)

	# IMPORTANTE: aggiorniamo l'elemento per gestire l'aggiornamento concorrente
	#             visto che cdg.delete() non e' concorrente
	element.concurrent_save(['mtime'], new_sentinel_value=mtime)
	# <<<< CONCURRENT DELETE

	return

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

@transaction_commit_on_success
def element_additional_datagroup_set_conditions_priority(element, datagroup_template_name, datagroup_instance_name, priority, user=None):
	
	tenant = element.tenant

	db_set_current_tenant_schema(tenant.name)

	if user:
		user.assert_capability(tenant, 'core.element.conf_monitoring')

	mtime = time.time()

	cdg  = element_additional_datagroup_get(element, datagroup_instance_name, tenant)

	tdg  = datagroup_template_get(datagroup_template_name, tenant=tenant)

	tf   = TField.objects.get(fclass=TField.FIELD_CONDITION, name='priority')

	dg = DataGroup.objects.get(tenant=tenant, element=element, name=datagroup_instance_name, cdatagroup=cdg)
	
	for c in dg.conditions.all():
		
		tc = cdg.find_tcondition(c.name)

		try:
			ccfv = CConditionFieldValue.objects.get(cdatagroup=cdg, tcondition=tc, tfield=tf, tenant=tenant)
		except CConditionFieldValue.DoesNotExist, e:
			ccfv = CConditionFieldValue(cdatagroup=cdg, tcondition=tc, tfield=tf, tenant=tenant, mtime=mtime)
			ccfv.save()

		ccfv.value = str(priority)
		ccfv.concurrent_save(['value'], new_sentinel_value=mtime)

		c.update_single_field_and_save(ccfv, mtime=mtime)



#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#


@transaction_commit_on_success
def element_additional_datagroup_field_set(element, datagroup_template_name, datagroup_instance_name, field_name, field_str_value, tenant=None):

	# FIX: do it more efficently

	db_set_current_tenant_schema(tenant.name)

	mtime = time.time()

	cdg  = element_additional_datagroup_get(element, datagroup_instance_name, tenant)

	tdg  = datagroup_template_get(datagroup_template_name, tenant=tenant)


	#field_name  = CLI_FIELD_TO_DATAGROUP_FIELD(field_name)
	tf   = TField.objects.get(fclass=TField.FIELD_DATAGROUP, name=field_name)

	real_value = tf.validate_value(field_str_value)

	try:
		cdgfv = CDataGroupFieldValue.objects.get(cdatagroup=cdg, tfield=tf, tenant=tenant)
	except CDataGroupFieldValue.DoesNotExist, e:
		cdgfv = CDataGroupFieldValue(cdatagroup=cdg, tfield=tf, tenant=tenant)
		cdgfv.save()

	cdgfv.value = field_str_value
	
	#cdgfv.mtime = time.time()
	#cdgfv.save()
	cdgfv.concurrent_save(['value'], new_sentinel_value=mtime)
	
	
	# Update 
	dg = DataGroup.objects.get(tenant=tenant, element=element, cdatagroup=cdg)

	dg.update_single_field_and_update(cdgfv, mtime=mtime)

	#
	# Suspend the datagroup if any of the suspended field is set.
	#
	if tf.name in  ['suspended']: #, 'suspend_time', 'suspend_reason','suspend_resume_time']:

		# Suspending?
		if (real_value):

			# Keep the current value for the following fields

			# Keep current settings, if any
			suspend_time   = dg.suspend_time
			resume_time    = dg.suspend_resume_time
			suspend_reason = dg.suspend_reason

			dg.suspend(
				resume_time  = resume_time, 
				suspend_time = suspend_time, 
				reason       = suspend_reason, 
				force=True
			)
		else:
			# suspended false
			resume_time = int(time.time())
			reason      = "No reason. (Resumed at %s)" % (resume_time)

			dg.resume(resume_time=int(time.time()), reason=reason, force=True)


	return

@transaction_commit_on_success
def element_additional_datagroup_field_unset(element, datagroup_template_name, datagroup_instance_name, field_name, tenant=None):

	db_set_current_tenant_schema(tenant.name)
	
	mtime = time.time()
	
	cdg  = element_additional_datagroup_get(element, datagroup_instance_name, tenant)

	tdg  = datagroup_template_get(datagroup_template_name, tenant=tenant)

	#field_name  = CLI_FIELD_TO_DATAGROUP_FIELD(field_name)
	tf   = TField.objects.get(fclass=TField.FIELD_DATAGROUP, name=field_name)

	try:
		cdgfv = CDataGroupFieldValue.objects.get(cdatagroup=cdg, tfield=tf, tenant=tenant)
	except CDataGroupFieldValue.DoesNotExist, e:
		if DEBUG: print "Field %s.%s.%s.%s not configured" % (element.name, datagroup_template_name, datagroup_instance_name, field_name)
		return

	dg = DataGroup.objects.get(tenant=tenant, element=element, cdatagroup=cdg)	


	#
	# 2-phase commmit
	#
	cdgfv.delete()
	
	fv = cdg.get_configured_field_or_default(tf.name)
	
	# Update 
	dg.update_single_field_and_update(fv, mtime=mtime)

	return


#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#
	
@transaction_commit_on_success
def element_additional_datagroup_parameter_set(element, datagroup_template_name, datagroup_instance_name, param_name, param_vtype, param_value, tenant=None, user=None):

	db_set_current_tenant_schema(element.tenant.name)

	if user:
		user.assert_capability(element.tenant, 'core.element.conf_monitoring')

	mtime = time.time()

	tdg = datagroup_template_get(datagroup_template_name, tenant=tenant)

	cdg  = element_additional_datagroup_get(element, datagroup_instance_name, tenant)

	dg = DataGroup.objects.get(tenant=tenant, element=element, cdatagroup=cdg)
	

	# FIX:check type mismatch with parameters defined  in datagroup-template

	# Check if the parameter already exists (in the anchestors too) with the same type
	configured_params = cdg.get_configured_params_dict(filter_names=[param_name,])

	param = configured_params.get( param_name, None )
	if param:
		if param.vtype != param_vtype:
			raise ValueError("Can not change the param type (from type %s to %s)" % ( param.vtype, param_vtype))

	try:
		cdgpv = CDataGroupParameterValue.objects.get(cdatagroup=cdg, name=param_name, tenant=tenant)
	except CDataGroupParameterValue.DoesNotExist, e:
		cdgpv = CDataGroupParameterValue(cdatagroup=cdg, name=param_name, tenant=tenant)
		cdgpv.save()

	# 2-phase commit
		
	# FIX: validate
	cdgpv.value = param_value
	cdgpv.vtype = param_vtype
	#cdgpv.mtime = time.time()
	#cdgpv.save()
	cdgpv.concurrent_save(['value','vtype'], new_sentinel_value=mtime)

	# Update 

	#dgp, created = DataGroupParameter.objects.get_or_create(tenant=tenant, datagroup=dg, name=param_name)
	#v = convert_to_param_type(dgp.vtype, param_value)
	#dgp.value = param_value
	#dgp.mtime = time.time()
	# FIX: usa update
	#dgp.save()
	dg.update_single_parameter_and_save(cdgpv, mtime=mtime)
	
	
	return

#------------------------------------------------------------------------------#

@transaction_commit_on_success
def element_additional_datagroup_parameter_unset(element, datagroup_template_name, datagroup_instance_name, param_name, tenant=None, user=None):

	mtime = time.time()

	db_set_current_tenant_schema(tenant.name)

	if user:
		user.assert_capability(element.tenant, 'core.element.conf_monitoring')


	tdg = datagroup_template_get(datagroup_template_name, tenant=tenant)

	cdg  = element_additional_datagroup_get(element, datagroup_instance_name, tenant)

	dg = DataGroup.objects.get(tenant=tenant, element=element, cdatagroup=cdg)

	cdgpv = CDataGroupParameterValue.objects.get(cdatagroup=cdg, name=param_name, tenant=tenant)

	dgp = DataGroupParameter.objects.get(tenant=tenant, datagroup=dg, name=param_name)

	# 2-phase commit


	cdgpv.delete()


	#
	# FIX: SPOSTARE IN UN METODO DEL  CDataGroup
	#
	# Update the real datagroup

	#dg.mtime = mt
	#dg.save() # FIX: update only mtime

	pv = cdg.get_configured_param(param_name)
	if pv:
		dgp.value = pv.value
		#dgp.mtime = mt
		#dgp.save()
		
		#time.sleep(2)
		dgp.concurrent_save(['value'], new_sentinel_value=mtime)
	else:
		dgp.delete()

	dg.concurrent_save(['mtime'], new_sentinel_value=mtime)

	return

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

@transaction_commit_on_success
def element_additional_datasource_field_set(element, datagroup_instance_name, datasource_name, field_name, field_str_value, tenant, user=None):

	db_set_current_tenant_schema(tenant.name)

	if user:
		user.assert_capability(tenant, 'core.element.conf_monitoring')

	mtime = time.time()

	tf  = TField.objects.get(fclass=TField.FIELD_DATASOURCE, name=field_name)
	tf.validate_value(field_str_value)
	
	cdg = element_additional_datagroup_get(element, datagroup_instance_name, tenant)
	
	tds = cdg.tdatagroup.get_last_defined_datasource(datasource_name)

	dg = DataGroup.objects.get(tenant=tenant, element=element, name=datagroup_instance_name, cdatagroup=cdg)
	ds = DataSource.objects.get(tenant=tenant, datagroup=dg, name=datasource_name)


	# Assicuriamoci di non lasciare record "spuri" associati a tdatasource meno "vicine" nella gerarchia dei datagroup template. Tutti i "fieldvalue" con lo stesso nome ma di tdatasource "antenati" vanno tolti.
	for old_cdsfv in CDataSourceFieldValue.objects.filter(cdatagroup=cdg, tfield=tf, tenant=tenant).filter(tdatasource__in=tds.get_ancestors()).all():
		old_cdsfv.delete()

	try:
		cdsfv = CDataSourceFieldValue.objects.get(cdatagroup=cdg, tdatasource=tds, tfield=tf, tenant=tenant)
	except CDataSourceFieldValue.DoesNotExist, e:
		cdsfv = CDataSourceFieldValue(cdatagroup=cdg, tdatasource=tds, tfield=tf, tenant=tenant)
		cdsfv.save()
		
	#
	# 2-phase commit
	#
	
	# FIX: validate
	real_value = tf.validate_value(field_str_value)
	
	cdsfv.value = field_str_value
	
	#cdsfv.tenant = tenant
	#cdsfv.mtime = time.time()
	#cdsfv.save()
	
	cdsfv.concurrent_save(['value'], new_sentinel_value=mtime)
	

	# Update the working tables
	ds.update_single_field_and_save(cdsfv, mtime=mtime)
	
	return	

@transaction_commit_on_success
def element_additional_datasource_field_unset(element, datagroup_instance_name, datasource_name, field_name, tenant, user=None):

	db_set_current_tenant_schema(tenant.name)

	if user:
		user.assert_capability(tenant, 'core.element.conf_monitoring')

	mtime = time.time()

	tf  = TField.objects.get(fclass=TField.FIELD_DATASOURCE, name=field_name)
	
	cdg = element_additional_datagroup_get(element, datagroup_instance_name, tenant)
	
	tds = cdg.tdatagroup.get_last_defined_datasource(datasource_name)

	dg = DataGroup.objects.get(tenant=tenant, element=element, name=datagroup_instance_name, cdatagroup=cdg)
	ds = DataSource.objects.get(tenant=tenant, datagroup=dg, name=datasource_name)

	try:
		cdsfv = CDataSourceFieldValue.objects.get(cdatagroup=cdg, tdatasource=tds, tfield=tf, tenant=tenant)
	except CDataSourceFieldValue.DoesNotExist, e:
		if DEBUG: print "Field %s.%s.%s.%s not configured" % (element.name, datagroup_instance_name, datasource_name, field_name)
		return

	# 2-phase commit
	
	cdsfv.delete()

	#
	# Aggiornare il valore in base ai template
	#

	fv = cdg.get_configured_datasource_field_or_default(datasource_name, field_name)

	# Update the working tables
	ds.update_single_field_and_save(fv, mtime=mtime)

	return


#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

@transaction_commit_on_success
def element_additional_condition_field_set(element, datagroup_instance_name, condition_name, field_name, field_str_value, tenant, user=None):

	db_set_current_tenant_schema(tenant.name)

	if user:
		user.assert_capability(tenant, 'core.element.conf_monitoring')

	mtime = time.time()

	tf  = TField.objects.get(fclass=TField.FIELD_CONDITION, name=field_name)

	tf.validate_value(field_str_value)


	cdg = element_additional_datagroup_get(element, datagroup_instance_name, tenant)

	dg = DataGroup.objects.get(tenant=tenant, element=element, name=datagroup_instance_name, cdatagroup=cdg)
	c  = Condition.objects.get(tenant=tenant, datagroup=dg, name=condition_name)

	tc = cdg.find_tcondition(condition_name)

	# Assicuriamoci di non lasciare record "spuri" associati a tcondition meno "vicine" nella gerarchia dei datagroup template. Tutti i "fieldvalue" con lo stesso nome ma di tcondition diverse vanno tolti.
	for old_ccfv  in CConditionFieldValue.objects.filter(cdatagroup=cdg, tfield=tf, tenant=tenant).filter(tcondition__in=tc.get_ancestors()).all():
		old_ccfv.delete()

	try:
		ccfv = CConditionFieldValue.objects.get(cdatagroup=cdg, tcondition=tc, tfield=tf, tenant=tenant)
	except CConditionFieldValue.DoesNotExist, e:
		ccfv = CConditionFieldValue(cdatagroup=cdg, tcondition=tc, tfield=tf, tenant=tenant, mtime=mtime)
		ccfv.save()

	real_value = tf.validate_value(field_str_value)

	# FIX: validate
	ccfv.value = field_str_value
	#ccfv.tenant = tenant
	#ccfv.save()
	ccfv.concurrent_save(['value'], new_sentinel_value=mtime)

	# Update the working tables
	changed = c.update_single_field_and_save(ccfv, mtime=mtime)

	if changed:
		if DEBUG: print "CHANGED:", field_name
		if field_name == 'primary':
			for dg2 in DataGroup.objects.filter(tenant=tenant).filter(Q(dependson=c)|Q(element=dg.element)|Q(element__in=dg.element.get_all_sub_elements())).exclude(uuid=dg.uuid):
				if DEBUG: print "Updating", dg2
				dg2.update_from_config(mtime=mtime)

	return

@transaction_commit_on_success
def element_additional_condition_field_unset(element, datagroup_instance_name, condition_name, field_name, tenant, user=None):

	db_set_current_tenant_schema(tenant.name)

	if user:
		user.assert_capability(tenant, 'core.element.conf_monitoring')

	mtime = time.time()

	tf  = TField.objects.get(fclass=TField.FIELD_CONDITION, name=field_name)

	cdg = element_additional_datagroup_get(element, datagroup_instance_name, tenant)

	tc = cdg.find_tcondition(condition_name)

	dg = DataGroup.objects.get(tenant=tenant, element=element, name=datagroup_instance_name, cdatagroup=cdg)
	c  = Condition.objects.get(tenant=tenant, datagroup=dg, name=condition_name)

	try:
		ccfv = CConditionFieldValue.objects.get(cdatagroup=cdg, tcondition__name=condition_name, tfield=tf, tenant=tenant)
	except CConditionFieldValue.DoesNotExist, e:
		if DEBUG: print "Field %s.%s.%s.%s not configured" % (element.name, datagroup_instance_name, condition_name, field_name)
		return

	ccfv.delete()

	#
	# Update the working tables
	#
	#fv = cdg.get_configured_condition_field(condition_name, tf.name)
	fv = cdg.get_configured_condition_field_or_default(condition_name, tf.name)

	changed = c.update_single_field_and_save(fv, mtime=mtime)

	if changed:
		if field_name == 'primary':
			for dg in DataGroup.objects.filter(tenant=tenant, dependson=c):
				dg.update_from_config(mtime=mtime)

	return

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

@transaction_commit_on_success
def element_additional_timegraph_field_set(element, datagroup_instance_name, timegraph_name, field_name, field_str_value, tenant, user=None):

	db_set_current_tenant_schema(tenant.name)

	if user:
		user.assert_capability(tenant, 'core.element.conf_monitoring')

	mtime = time.time()

	tf  = TField.objects.get(fclass=TField.FIELD_TGRAPH, name=field_name)
	
	tf.validate_value(field_str_value)

	cdg = element_additional_datagroup_get(element, datagroup_instance_name, tenant)
	
	ttg = cdg.tdatagroup.get_last_defined_timegraph(timegraph_name)

	dg = DataGroup.objects.get(tenant=tenant, element=element, name=datagroup_instance_name, cdatagroup=cdg)
	tg = TimeGraph.objects.get(tenant=tenant, datagroup=dg, name=timegraph_name)


	try:
		ctgfv = CTimeGraphFieldValue.objects.get(tenant=tenant, cdatagroup=cdg, ttimegraph=ttg, tfield=tf)
	except CTimeGraphFieldValue.DoesNotExist, e:
		ctgfv = CTimeGraphFieldValue(tenant=tenant, cdatagroup=cdg, ttimegraph=ttg, tfield=tf, mtime=mtime)
		ctgfv.save()
	
	# FIX: validate
	
	ctgfv.value = field_str_value
	#ctgfv.save()
	ctgfv.concurrent_save(['value'], new_sentinel_value=mtime)
	

	# Update the working tables
	tg.update_single_field_and_save(ctgfv, mtime=mtime)
	
	return

@transaction_commit_on_success
def element_additional_timegraph_field_unset(element, datagroup_instance_name, timegraph_name, field_name, tenant, user=None):

	mtime = time.time()

	db_set_current_tenant_schema(tenant.name)

	if user:
		user.assert_capability(tenant, 'core.element.conf_monitoring')

	tf  = TField.objects.get(fclass=TField.FIELD_TGRAPH, name=field_name)
	
	cdg = element_additional_datagroup_get(element, datagroup_instance_name, tenant)
	
	ttg = cdg.tdatagroup.get_last_defined_timegraph(timegraph_name)

	try:
		ctgfv = CTimeGraphFieldValue.objects.get(tenant=tenant, cdatagroup=cdg, ttimegraph=ttg, tfield=tf)
	except CTimeGraphFieldValue.DoesNotExist, e:
		if DEBUG: print "Field %s.%s.%s.%s not configured" % (element.name, datagroup_instance_name, timegraph_name, field_name)
		return

	# Update the working tables
	dg = DataGroup.objects.get(tenant=tenant, element=element, name=datagroup_instance_name, cdatagroup=cdg)
	tg = TimeGraph.objects.get(tenant=tenant, datagroup=dg, name=timegraph_name)

	ctgfv.delete()

	# FIX: IMPLEMENTARE fv = cdg.get_configured_timegraph_field_or_default(timegraph_name, tf.name)
	#                   tg.update_single_field_and_save(fv, mtime=mtime)

	dg.update_from_config(mtime=mtime) #FIX: update only selected field

	return

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

@transaction_commit_on_success		
def element_additional_timegraph_serie_field_set(element, datagroup_instance_name, timegraph_name, serie_name, field_name, field_str_value, tenant, user=None):

	db_set_current_tenant_schema(tenant.name)

	mtime = time.time()

	if user:
		user.assert_capability(tenant, 'core.element.conf_monitoring')

	tf   = TField.objects.get(fclass=TField.FIELD_TSERIE, name=field_name)

	tf.validate_value(field_str_value)

	cdg = element_additional_datagroup_get(element, datagroup_instance_name, tenant)
	ttgs = cdg.tdatagroup.get_last_defined_timegraphserie(timegraph_name, serie_name)

	dg = DataGroup.objects.get(tenant=tenant, element=element, cdatagroup=cdg)#name=datagroup_instance_name)
	tg = TimeGraph.objects.get(tenant=tenant, datagroup=dg, name=timegraph_name)
	ts = TimeGraphSerie.objects.get(tenant=tenant, timegraph=tg, name=serie_name)

	try:
		ctgsfv = CTimeGraphSerieFieldValue.objects.get(tenant=tenant, cdatagroup=cdg, ttimegraphserie=ttgs, tfield=tf)
	except CTimeGraphSerieFieldValue.DoesNotExist, e:
		ctgsfv = CTimeGraphSerieFieldValue(tenant=tenant, cdatagroup=cdg, ttimegraphserie=ttgs, tfield=tf, mtime=mtime)
		ctgsfv.save()

	# FIX: validate

	ctgsfv.value = field_str_value
	#ctgsfv.save()
	ctgsfv.concurrent_save(['value'], new_sentinel_value=mtime)

	# Update the working tables
	#dg = DataGroup.objects.get(tenant=tenant, element=element, name=datagroup_instance_name)

	ts.update_single_field_and_save(ctgsfv, mtime=mtime)

	return

@transaction_commit_on_success		
def element_additional_timegraph_serie_field_unset(element, datagroup_instance_name, timegraph_name, serie_name, field_name, tenant, user=None):

	mtime = time.time()

	db_set_current_tenant_schema(tenant.name)

	if user:
		user.assert_capability(tenant, 'core.element.conf_monitoring')

	tf   = TField.objects.get(fclass=TField.FIELD_TSERIE, name=field_name)
			
	cdg = element_additional_datagroup_get(element, datagroup_instance_name, tenant)
	ttgs = cdg.tdatagroup.get_last_defined_timegraphserie(timegraph_name, serie_name)

	dg = DataGroup.objects.get(tenant=tenant, element=element, cdatagroup=cdg)#name=datagroup_instance_name)
	tg = TimeGraph.objects.get(tenant=tenant, datagroup=dg, name=timegraph_name)
	ts = TimeGraphSerie.objects.get(tenant=tenant, timegraph=tg, name=serie_name)

	try:
		ctgsfv = CTimeGraphSerieFieldValue.objects.get(tenant=tenant, cdatagroup=cdg, ttimegraphserie=ttgs, tfield=tf)
	except CTimeGraphSerieFieldValue.DoesNotExist, e:
		if DEBUG: print "Field %s.%s.%s.%s.%s not configured" % (element.name, datagroup_instance_name, timegraph_name, serie_name, field_name)
		return

	ctgsfv.delete()

	# FIX: IMPLEMENTARE fv = cdg.get_configured_timegraphserie_field_or_default(timegraph_name, tf.name)
	#                   ts.update_single_field_and_save(fv, mtime=mtime)
	
	#ts.update_single_field_and_save(ctgsfv)
	dg.update_from_config(mtime=mtime) # FIX: update only selected field

	return
