# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

Stato dei processi di sanet
=============================

Lo stato dei diversi processi di sanet viene salvato e letto tramite le api:

	* sanet_subsystem_get_current_state( )
	* sanet_subsystem_set_current_state( )

Lo stato e' rappresentato dalla struttura:

::
	
	{
		'checktime': <timestamp>,
		'running'  : <boolean>,
		'info'     : <text>,
	}

"""

import os
import os.path
import time
import types
import json

import rrdtool
import subprocess

from PIL import Image

from conf import get_configuration

from core.api.base import tenant_get_data_dir
from core.api.base import makedirs_for_webuser

from core.api.cache import init_cache_manager_from_conf

from core.api.datasource import check_datasource_store_conf

from django.db import connection

from sanet_common.termutils import term_red


#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

def check_system_settings():

	#
	# Check internal DB parameters
	#
	_check_db_server_run_settings()

	#
	# Check datasource store settings
	#
	check_datasource_store_conf()

	#
	# Fix: other checks?
	#

	return

def _check_db_server_run_settings():

	conf = get_configuration()

	c = connection.cursor() 

	c.execute("SHOW max_connections")
	max_connections = int(c.fetchone()[0])

	si = len(conf.SANETD_INSTANCES)
	wt = conf.SANETD_WORKING_THREADS_NUM

	if (wt * si) >= max_connections:

		msg = """*** Postgres configuration not valid. Please check 'max_connections'.\nCurrent value: %s.\nShould be at least: %s ( SANETD_INSTANCES (%s) * SANETD_WORKING_THREADS_NUM (%s))""" % (max_connections, si * wt, si, wt)

		print term_red( msg )

		sys.exit(-1)



#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

def sanet_subsystem_get_current_state(name):
	"""
	
	"""
	try:
		cache_manager = init_cache_manager_from_conf()

		current_state_json = cache_manager.get(name)
		if current_state_json is None:
			return

		current_state = json.loads(current_state_json)

		return current_state
	except Exception, e:
		return None

def sanet_subsystem_set_current_state(name, checktime, running, info):
	"""
	
	"""
	try:
		cache_manager = init_cache_manager_from_conf()

		current_state = {
			'checktime': checktime,
			'running'  : running,
			'info'     : info,
		}
		current_state_json = json.dumps(current_state)

		cache_manager.set(name, current_state_json)
	except Exception, e:
		pass

	return

#
# Metodi di comodo
#
def sanetd_get_current_state(instance_id=None):

	# FIX: old. remove 
	if not instance_id:
		return sanet_subsystem_get_current_state('sanetd_state')

	return sanet_subsystem_get_current_state('sanetd-%s_state' % (instance_id))

def sanetd_set_current_state(checktime, running, info, instance_id=None):

	# FIX: old. remove 
	if not instance_id:
		return sanet_subsystem_set_current_state('sanetd_state'                   , checktime, running, info)
                                                                             
	return sanet_subsystem_set_current_state('sanetd-%s_state' % (instance_id), checktime, running, info)

def entables_get_current_state():
	return sanet_subsystem_get_current_state('entables_state')

def entables_set_current_state(checktime, running, info):
	return sanet_subsystem_set_current_state('entables_state', checktime, running, info)

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

# Le statistiche numeriche vengono salvate in file RRD nelle directory:
# 
#     {{DATA_DIR}} / _global_stats
#
#		sanetd.rrd
#
#     {{DATA_DIR}} / <tenantid> / stats / 
#
#		<agentid>.rrd
#

GLOBAL_STATS_DIR = "_global_stats"

TENANT_STATS_DIR = 'stats'


#
# Parametri di memroizzazione RRD
#
MAX_PERIOD = 60 * 60 * 24 * 30 * 6
STEP       = 60   
HEARTBIT   = STEP * 2

SERIES = [
	'RRA:AVERAGE:0.5:1:%s' % ( MAX_PERIOD / STEP )
]


def _get_rrd_stats_file_path(tenant_id, name):

	file_name = '%s_stats.rrd' % (name)

	tenant_base_dir = tenant_get_data_dir(tenant_id)

	stats_dir = os.path.join(tenant_base_dir, TENANT_STATS_DIR)

	if not os.path.exists(stats_dir):
		#os.makedirs(stats_dir)
		makedirs_for_webuser(stats_dir)

	file_path = os.path.join(stats_dir, file_name)

	return file_path

def _create_rrd_stats_file(tenant_id, name, ds_defs, file_name=None):

	if file_name is None:
		file_name = agent_get_rrd_stats_file_path(tenant_id, name)

	#
	# RRD GLOBAL 
	#
	rrd_params = [
		'--start', str(0),
		'--step' , str(STEP), 
	]

	#
	# DSs
	#
	ds_params = []
	for ds_name in ds_defs: 
		ds_params.append( "DS:%s:GAUGE:%s:U:U" % (ds_name, HEARTBIT) )

	rrd_params = rrd_params + ds_params

	#
	# SERIESs
	#
	rrd_params = rrd_params + SERIES

	try:
		rrdtool.create(str(file_name), *rrd_params)
	except rrdtool.error, er:
		raise Exception("Can't create RRD file %s: %s" % (file_name, er))


def _update_rrd_stats_file(tenant_id, name, ds_defs, stats_dict, interval=None):

	timestamp = int(time.time())

	file_name = _get_rrd_stats_file_path(tenant_id, name)

	file_name = str(file_name)

	if not os.path.exists(file_name):
		_create_rrd_stats_file(tenant_id, name, ds_defs, file_name)

	try:
		if interval is not None:

			interval = int(interval)

			valid_heartbeat = interval * 3

			fileinfo = rrdtool.info(file_name)

			for ds_name in ds_defs:
				heartbeat = fileinfo['ds[%s].minimal_heartbeat' % (ds_name) ] 
				if heartbeat != valid_heartbeat:
					print __file__, "Changing heartbeat for '%s'" % (ds_name)
					rrdtool.tune(file_name, "--heartbeat", "%s:%s" % (ds_name, valid_heartbeat))
	except Exception, e:
		print "ERRORE: ", e
		pass

	if 1:
		values = []
		for ds_name in ds_defs:
			v = stats_dict.get(ds_name, 'U')
			if v is None:
				v = 'U'
			values.append(str(v))

		rrd_update_str = str(timestamp) + ":" + ':'.join(values)

		try:
			rrdtool.update(file_name, rrd_update_str)
		except rrdtool.error, er:
			raise Exception("Unable to update %s: %s" % (file_name, str(er)))

#-------------------------------------------------------------------------------#
# Statistiche sanetd                                                            #
#-------------------------------------------------------------------------------#

SANETD_STAT_DS_ELEMENTS = [
	'queuesize',
	'maxqueuesize',
]

SANETD_COLORS = {
	'queuesize'       : 'FF1300',  # rosso
	'maxqueuesize'    : '0000FC',  # blue
	
	'processqueuesize': 'FF00FF',  # viola
}


def sanetd_get_rrd_stats_file_path(stats_type='queues'):

	file_name = 'sanetd_%s.rrd' % (stats_type)

	stats_dir = os.path.join(get_configuration().DATA_DIR, GLOBAL_STATS_DIR)

	if not os.path.exists(stats_dir):
		makedirs_for_webuser(stats_dir)

	file_path = os.path.join(stats_dir, file_name)

	return file_path

def sanetd_create_rrd_stats_file(stats_type='queues'):

	file_name = sanetd_get_rrd_stats_file_path(stats_type)

	ds_defs   = SANETD_STAT_DS_ELEMENTS

	#
	# RRD GLOBAL 
	#
	rrd_params = [
		'--start', str(0),
		'--step', str(STEP), 
	]

	#
	# DSs
	#
	ds_params = []
	for ds_name in ds_defs: 
		ds_params.append( "DS:%s:GAUGE:%s:U:U" % (ds_name, HEARTBIT) )

	rrd_params = rrd_params + ds_params

	#
	# SERIESs
	#
	rrd_params = rrd_params + SERIES

	try:
		rrdtool.create(str(file_name), *rrd_params)
	except rrdtool.error, er:
		raise Exception("Can't create RRD file %s: %s" % (file_name, er))

def sanetd_update_rrd_stats_file(stats_dict, interval, stats_type='queues'):

	ds_defs   = SANETD_STAT_DS_ELEMENTS


	timestamp = int(time.time())

	file_name = sanetd_get_rrd_stats_file_path(stats_type)

	file_name = str(file_name)

	if not os.path.exists(file_name):
		sanetd_create_rrd_stats_file()

	try:
		if interval is not None:

			interval = int(interval)

			valid_heartbeat = interval * 3

			fileinfo = rrdtool.info(file_name)

			for ds_name in ds_defs:
				heartbeat = fileinfo['ds[%s].minimal_heartbeat' % (ds_name) ] 
				if heartbeat != valid_heartbeat:
					rrdtool.tune(file_name, "--heartbeat", "%s:%s" % (ds_name, valid_heartbeat))
	except Exception, e:
		print "ERRORE: ", e
		pass

	if 1:
		values = []
		for ds_name in ds_defs:
			v = stats_dict.get(ds_name, 'U')
			if v is None:
				v = 'U'
			values.append(str(v))

		rrd_update_str = str(timestamp) + ":" + ':'.join(values)

		try:
			rrdtool.update(file_name, rrd_update_str)
		except rrdtool.error, er:
			raise Exception("Unable to update %s: %s" % (file_name, str(er)))

def sanetd_create_rrd_graph_file_elements(outdir, output_image_filename, ds_list, colors, ltype='LINE', stacked=False, title='', ts_start=None, ts_end=None, stats_type='queues'):

	if not os.path.exists(outdir):
		makedirs_for_webuser(outdir)

	if ts_end is None:
		ts_end = int(time.time())

	if ts_start is None:
		ts_start = ts_end - 60*60*8

	NAN_COLOR = 'ffffaa'

	outfile = os.path.join(outdir, output_image_filename)
	if os.path.exists(outfile):
		os.remove(outfile)

	file_name = sanetd_get_rrd_stats_file_path(stats_type)

	print __file__, "IMAGE", file_name

	defs  = []
	lines = []
	ticks = []
	prints = []
	for ds_name in ds_list: #STAT_DS_ELEMENTS:
		d = 'DEF:%s="%s":%s:AVERAGE' % ( ds_name, file_name, ds_name )
		defs.append(d)

		d = "CDEF:%s_nan=%s,UN"      % ( ds_name, ds_name)
		defs.append(d)

		l = 'TICK:%s_nan#%s:1:"no data\l"' % ( ds_name, NAN_COLOR)
		if not ticks: ticks.append(l)

		color = colors[ds_name]	
		if stacked:
			l = '%s:%s#%s:"%s\l":STACK' % (ltype, ds_name, color, ds_name)
		else:
			l = '%s:%s#%s:"%s\l"' % (ltype, ds_name, color, ds_name)

		lines.append(l)

		for up, low in [('MAX','max'),('AVERAGE','avg'),('LAST','current')]:
			legend_attr = ds_name
			entry_kw = {
				'display_var' : ds_name,
				'up'          : up, 
				'low'         : low,
				legend_attr   : ds_name,
				'idem'        : '%s', 
				'2lf'         : '%.2lf',
				'shortlegend' : '',
			}
			if up == "LAST":
				entry_kw['shortlegend'] = "%s\l" % entry_kw['shortlegend']

			TEMPLATE = 'GPRINT:%(display_var)s:%(up)s:"%(low)s %('+legend_attr+')s %(2lf)s %(idem)s%(shortlegend)s"'
			
			prints.append(TEMPLATE % entry_kw)

	dimensions = [ 
		"-t", "\"%s\"" % (title),
		"-h", "50",
		"-w", "600",
	]

	now = time.time()

	args = dimensions + defs + ticks + lines + prints
	args = ["rrdtool", "graph", outfile] + args + [ "-s" , str(ts_start), "-e", str(ts_end) ]

	r = subprocess.call(" ".join(args), shell=True)

	im = Image.open(outfile)

	img_width, img_height = im.size

	return outfile, img_width, img_height

#-------------------------------------------------------------------------------#
# Statistiche agenti                                                            #
#-------------------------------------------------------------------------------#

AGENT_STAT_DS_ELEMENTS = [
	'nodes',
	'ifaces',
	'storages',
	'services',
	'devices',
]

"""
Descrizione 			Nome 		Nome
-----------			----		-----
numero di nodi 			nodes 		DS:nodes:GAUGE:120:U:U
noumeo di interfacce 		ifaces 		DS:ifaces:GAUGE:120:U:U
numero di dischi 		storages 	DS:storages:GAUGE:120:U:U
numero di service 		services 	DS:services:GAUGE:120:U:U
numero di device 		devices 	DS:devices:GAUGE:120:U:U
numero di datagroup 		datagroups 	DS:datagroups:GAUGE:120:U:U
numero di datasource 		datasources 	DS:datasources:GAUGE:120:U:U
nuemro di condition 		conditions 	DS:conditions:GAUGE:120:U:U
numero di datagroup al secondo 	dgrate 		DS:dgrate:GAUGE:120:U:U
lunghezza wai./ting queue 	wqueuelen 	DS:wqueuelen:GAUGE:120:U:U
lunghezza sending queue 	squeuelen 	DS:squeuelen:GAUGE:120:U:U
memoria (virtuale) occupata 	vmemory 	DS:vmemory:GAUGE:120:U:U

RRA SERIES
RRA:AVERAGE:0.5:1:%s' % ( MAX_PERIOD / STEP )

Comando:

rrdtool create     agentstats.%s.rrd          --start 0 --step STEP    DS.... RRA......

Test:

rrdtool create     agentstats.%s.rrd          --start 0 --step 60  DS:nodes:GAUGE:120:U:U  DS:ifaces:GAUGE:120:U:U  DS:storages:GAUGE:120:U:U  DS:services:GA

"""


AGENT_STAT_DS_ELEMENTS = [
	'nodes',
	'ifaces',
	'storages',
	'services',
	'devices',
]

AGENT_STAT_DS_MONITOR = [
	'datagroups',
	'datasources',
	'conditions',
]

AGENT_STAT_DS_RATES = [
	'dgrate',
]

AGENT_STAT_DS_QUEUES = [
	'wqueuelen',
	'squeuelen',
]

AGENT_STATS_DS_MEMORY = [
	'vmemory',
	'rmemory',
]

AGENT_STAT_DS = AGENT_STAT_DS_ELEMENTS + AGENT_STAT_DS_MONITOR + AGENT_STAT_DS_RATES + AGENT_STAT_DS_QUEUES + AGENT_STATS_DS_MEMORY


AGENT_COLORS = {
	'nodes'      : 'FF1300',  # rosso
	'ifaces'     : 'CE5BC0',  # viola
	'storages'   : '0000FC',  # blue
	'services'   : '29FF00',  # verde
	'devices'    : 'DED00F',  # giallo

	'datagroups' : 'B24B1C',  # marrone
	'datasources': 'FF00FF',  # grigioblue
	'conditions' : 'FF007F',  # fuxia

	'dgrate'     : 'ff0000',  # green

	'wqueuelen'  : 'ff0000',  # red
	'squeuelen'  : '0000ff',  # blue

	'vmemory'    : '00ff00',  # green
	'rmemory'    : '0000ff',  # green
}

def agent_get_rrd_stats_file_path(agent):
	return _get_rrd_stats_file_path(agent.tenant.uuid, agent.uuid)

def agent_create_rrd_stats_file(agent, file_name=None):	
	return _create_rrd_stats_file(agent.tenant.uuid, agent.uuid, AGENT_STAT_DS, file_name=file_name)

def agent_update_rrd_stats_file(agent, stats_dict, interval):
	return _update_rrd_stats_file(agent.tenant.uuid, agent.uuid, AGENT_STAT_DS, stats_dict, interval=interval)

def agent_create_rrd_graph_file_elements(agent, outdir, name, ds_list, colors, ltype='LINE', stacked=False, title='', ts_start=None, ts_end=None):

	if ts_end is None:
		ts_end = int(time.time())

	if ts_start is None:
		ts_start = ts_end - 60*60*8

	NAN_COLOR = 'ffffaa'

	outfile = os.path.join(outdir, name)

	file_name = agent_get_rrd_stats_file_path(agent)
	

	defs  = []
	lines = []
	ticks = []
	prints = []
	for ds_name in ds_list: #STAT_DS_ELEMENTS:
		d = 'DEF:%s="%s":%s:AVERAGE' % ( ds_name, file_name, ds_name )
		defs.append(d)

		d = "CDEF:%s_nan=%s,UN"      % ( ds_name, ds_name)
		defs.append(d)

		l = 'TICK:%s_nan#%s:1:"no data\l"' % ( ds_name, NAN_COLOR)
		if not ticks: ticks.append(l)

		color = colors[ds_name]	
		if stacked:
			l = '%s:%s#%s:"%s\l":STACK' % (ltype, ds_name, color, ds_name)
		else:
			l = '%s:%s#%s:"%s\l"' % (ltype, ds_name, color, ds_name)

		lines.append(l)

		for up, low in [('MAX','max'),('AVERAGE','avg'),('LAST','current')]:
			legend_attr = ds_name
			entry_kw = {
				'display_var' : ds_name,
				'up'          : up, 
				'low'         : low,
				legend_attr   : ds_name,
				'idem'        : '%s', 
				'2lf'         : '%.2lf',
				'shortlegend' : '', #ds_name,
			}
			if up == "LAST":
				entry_kw['shortlegend'] = "%s\l" % entry_kw['shortlegend']

			TEMPLATE = 'GPRINT:%(display_var)s:%(up)s:"%(low)s %('+legend_attr+')s %(2lf)s %(idem)s%(shortlegend)s"'
			prints.append(TEMPLATE % entry_kw)

	dimensions = [ 
		"-t", "\"%s\"" % (title),
		"-h", "50",
		"-w", "600",
	]

	now = time.time()

	args = dimensions + defs + ticks + lines + prints
	args = ["rrdtool", "graph", outfile] + args + [ "-s" , str(ts_start), "-e", str(ts_end) ]

	r = subprocess.call(" ".join(args), shell=True)

	im = Image.open(outfile)

	img_width, img_height = im.size

	return outfile, img_width, img_height


























#-------------------------------------------------------------------------------#
#-------------------------------------------------------------------------------#
# STATITICS V2                                                                  #
#-------------------------------------------------------------------------------#
#-------------------------------------------------------------------------------#

def _get_rrd_stats_file_path_v2(stat_id, tenant_id=None):

	if tenant_id is None:
		stats_dir = os.path.join(get_configuration().DATA_DIR, GLOBAL_STATS_DIR)
	else:
		tenant_base_dir = tenant_get_data_dir(tenant_id)
		stats_dir = os.path.join(tenant_base_dir, TENANT_STATS_DIR)

	if not os.path.exists(stats_dir):
		makedirs_for_webuser(stats_dir)

	file_name = "%s.rrd" % (stat_id)

	file_path = os.path.join(stats_dir, file_name)

	return file_path


def _create_rrd_stats_file_v2(file_name):

	#
	# RRD GLOBAL 
	#
	rrd_params = [
		'--start', str(0),
		'--step' , str(STEP), 
	]

	#
	# DSs
	#
	ds_params = []
	ds_params.append( "DS:gauge:GAUGE:%s:U:U"  % (HEARTBIT) )
	ds_params.append( "DS:rate:COUNTER:%s:U:U" % (HEARTBIT) )

	rrd_params = rrd_params + ds_params

	#
	# SERIESs
	#
	rrd_params = rrd_params + SERIES

	try:
		rrdtool.create(str(file_name), *rrd_params)
	except rrdtool.error, er:
		raise Exception("Can't create RRD file %s: %s" % (file_name, er))


def sysstats_update_stat_file(stat_id, value, tenant_id=None, interval=None):

	file_name = _get_rrd_stats_file_path_v2(stat_id, tenant_id=tenant_id)

	file_name = str(file_name)

	if not os.path.exists(file_name):
		_create_rrd_stats_file_v2(file_name)

	try:
		if interval is not None:
	
			interval = int(interval)
	
			valid_heartbeat = interval * 3
	
			fileinfo = rrdtool.info(file_name)
	
			for ds_name in ['gauge', 'rate']: #ds_defs:

				heartbeat = fileinfo['ds[%s].minimal_heartbeat' % (ds_name) ] 
				if heartbeat != valid_heartbeat:
					print __file__, "Changing heartbeat for '%s'" % (ds_name)
					rrdtool.tune(file_name, "--heartbeat", "%s:%s" % (ds_name, valid_heartbeat))
	except Exception, e:
		print "ERRORE: ", e
		pass

	timestamp = int(time.time())

	if value is not None:
		gauge = value
		rate  = int(value)
	else:
		gauge = "U"
		rate  = "U"

	rrd_update_str = "%s:%s:%s" % (str(timestamp), gauge, rate)

	try:
		rrdtool.update(file_name, rrd_update_str)
	except rrdtool.error, er:
		raise Exception("Unable to update %s: %s" % (file_name, str(er)))
	except Exception, e:
		raise Exception("Unable to update %s: %s" % (file_name, str(er)))

def sysstats_create_rrd_graph_file(outdir, output_image_filename, selected_stats, colors, ltype='LINE', stacked=False, title='', ts_start=None, ts_end=None, tenant_id=None, selected_ds_type='gauge', legend=None):

	legend = legend or {}

	if not os.path.exists(outdir):
		makedirs_for_webuser(outdir)

	if ts_end is None:
		ts_end = int(time.time())

	if ts_start is None:
		ts_start = ts_end - 60*60*8

	NAN_COLOR = 'ffffaa'

	outfile = os.path.join(outdir, output_image_filename)
	if os.path.exists(outfile):
		os.remove(outfile)

	defs  = []
	lines = []
	ticks = []
	prints = []


	for selected_stat in selected_stats:

		stat_id = selected_stat

		file_name = _get_rrd_stats_file_path_v2(stat_id, tenant_id=tenant_id)
		
		if 1:
			fake_ds_name = stat_id.rsplit('.', 1)[-1]

			legend_name = legend.get(stat_id, None) or fake_ds_name

			d = 'DEF:%s="%s":%s:AVERAGE' % ( fake_ds_name, file_name, selected_ds_type )
			defs.append(d)

			d = "CDEF:%s_nan=%s,UN"      % ( fake_ds_name, fake_ds_name)
			defs.append(d)

			l = 'TICK:%s_nan#%s:1.0:' % ( fake_ds_name, NAN_COLOR)
			if not ticks: 
				ticks.append(l)

			# LINE or AREA
			color = colors.get(fake_ds_name, '0000ff')
			if stacked:
				l = '%s:%s#%s:"%s":STACK' % (ltype, fake_ds_name, color, legend_name)
			else:
				l = '%s:%s#%s:"%s"'       % (ltype, fake_ds_name, color, legend_name)

			lines.append(l)

			l = ''
			l += 'COMMENT:"%-20s" ' % (legend_name)
			l += ' ' .join([
				'GPRINT:%s:MAX:"Max %%4.2lf %%s\\t"'     % (fake_ds_name,) ,
				'GPRINT:%s:AVERAGE:"Avg %%4.2lf %%s\\t"'     % (fake_ds_name,) ,
				'GPRINT:%s:LAST:"Last %%4.2lf %%s\\t\l"' % (fake_ds_name,) ,
			])

			prints.append(l)

	dimensions = [ 
		"-t", '"%s"' % (title),
		#"-D",
		"-h", "40",
		"-w", "650",
		"--border", "1",
		#"--font", "DEFAULT:8:Arial",
		"--disable-rrdtool-tag",
		"--color", "BACK#ffffff",
	]
	

	now = time.time()

	CR = [ 'COMMENT:"\l"' ]

	args = dimensions + defs + ticks + lines + CR + prints
	args = ["rrdtool", "graph", outfile] + args + [ "-s" , str(ts_start), "-e", str(ts_end) ]

	#import pprint
	#pprint.pprint(args)

	r = subprocess.call(" ".join(args), shell=True)

	im = Image.open(outfile)

	img_width, img_height = im.size

	return outfile, img_width, img_height

#-------------------------------------------------------------------------------#


def sanetd_update_stats(logger, sanetd_instance, stats_dict, interval=None):

	for key, value in stats_dict.items():

		if not (isinstance(value, types.IntType) or isinstance(value, types.LongType) or isinstance(value, types.FloatType) or isinstance(value, types.NoneType)):
			continue

		stat_id = "sanetd.%s.%s" % (sanetd_instance, key)

		try:
			sysstats_update_stat_file(stat_id, value, interval=interval) 
			logger.info("Updated %s: %s", stat_id, value)
		except Exception, e:
			logger.error("Unable to update sanetd stats '%s': %s %s", stat_id, type(e), str(e))


def agent_update_stats(logger, agent, stats_dict, interval=None):

	for key, value in stats_dict.items():

		if not (isinstance(value, types.IntType) or isinstance(value, types.LongType) or isinstance(value, types.FloatType) or isinstance(value, types.NoneType)):
			continue

		stat_id = 'agent.%s.%s' % (agent.name, key)

		try:
			sysstats_update_stat_file(stat_id, value, tenant_id=agent.tenant_id, interval=interval)
			logger.info("Updated agent stat %s: %s", stat_id, value)
		except Exception, e:
			logger.error("Unable to update agent stats '%s': %s %s", stat_id, type(e), str(e))

#-------------------------------------------------------------------------------#
