"""

Espressioni per le serie calcolate
==================================

All'interno dei timegraph le serie "calcolate" sono delle serie calcolate on-demand utilizzando i valori delle altre serie
definite nello stesso timegraph.

Ipotizziamo di avere un datagroup cosi' definito ::

	datagroup prova

		datasource ds1
		exit

		datasource ds2
		exit

		timegraph grafico

			serie serie1
				var-value ds1
				legend "Prima serie"
			exit

			serie serie2
				var-value ds2
				legend "Seconda serie"
			exit

			serie calcolata
				computed-value  serie1 + serie2
				legend "Serie calcolata"
			exit

		exit
	
	exit
	
In questo datagroup e' definito un timegraph chiamato "grafico" che grafica 3 serie: serie1, serie2 e calcolata.

I valori delle prime due  (serie1 e serie2) vengono estratti da dati reali, piu' precisamente dai due datasource
ds1 e ds2 definiti all'interno del datagroup.

I valori della serie "calcolata" vengono calcolati prendendo sommando i valori corrispondenti delle altre
due serie gia' definite.

Esempio: ::

	            t0    t1    t2    t3    t4

	serie1       1     2     3     4     5
	
	serie2      25    -3    45    34    23

	calcolata   26    -1    48    38    25


Operatori ammessi
-----------------

L'espressione deve essere un'espressione matematica valida. Esempio: ::

	1 + 2 / 3 * abs(-3)

Gli operatiri matematici disponibili sono:

	+-----------+------------------------------------------------------------------------------------------+
	| Operatore | Descrizione                                                                              |
	+===========+==========================================================================================+
	| a + b     |                                                                                          |
	+-----------+------------------------------------------------------------------------------------------+
	| a - b     |                                                                                          |
	+-----------+------------------------------------------------------------------------------------------+
	| a * b     |                                                                                          |
	+-----------+------------------------------------------------------------------------------------------+
	| a / b     |                                                                                          |
	+-----------+------------------------------------------------------------------------------------------+
	| a % b     |                                                                                          |
	+-----------+------------------------------------------------------------------------------------------+
	| a ** b    |                                                                                          |
	+-----------+------------------------------------------------------------------------------------------+

All'interno dell'espressione possono essere usate le seguenti funzioni:

	+-------------------+------------------------------------------------------------------------------------------+
	| funzione          | Descrizione                                                                              |
	+===================+==========================================================================================+
	| abs(x)            |                                                                                          |
	+-------------------+------------------------------------------------------------------------------------------+
	| divmod(x)         |                                                                                          |
	+-------------------+------------------------------------------------------------------------------------------+
	| int(x)            |                                                                                          |
	+-------------------+------------------------------------------------------------------------------------------+
	| pow(x,e)          |                                                                                          |
	+-------------------+------------------------------------------------------------------------------------------+
	| float(x)          |                                                                                          |
	+-------------------+------------------------------------------------------------------------------------------+
	| long(x)           |                                                                                          |
	+-------------------+------------------------------------------------------------------------------------------+
	| max(x)            |                                                                                          |
	+-------------------+------------------------------------------------------------------------------------------+
	| min(x)            |                                                                                          |
	+-------------------+------------------------------------------------------------------------------------------+
	| log(n,e)          |                                                                                          |
	+-------------------+------------------------------------------------------------------------------------------+
	| sqrt(x)           |                                                                                          |
	+-------------------+------------------------------------------------------------------------------------------+
	| sin(x)            |                                                                                          |
	+-------------------+------------------------------------------------------------------------------------------+
	| cos(x)            |                                                                                          |
	+-------------------+------------------------------------------------------------------------------------------+
	| radians(x)        |                                                                                          |
	+-------------------+------------------------------------------------------------------------------------------+

Funzioni speciali:

	+-------------------+------------------------------------------------------------------------------------------+
	| funzione          | Descrizione                                                                              |
	+===================+==========================================================================================+
	| iftrue(c,t,e)     | If "c" is "true" returns "t" else returns "e"                                            |
	+-------------------+------------------------------------------------------------------------------------------+
	| ifnull(x,d)       |                                                                                          |
	+-------------------+------------------------------------------------------------------------------------------+

Variabili disponibili
---------------------

All'interno delle espressioni possono essere sempre usate come variabili i nomi delle serie che definiscono
il parametro "var-value".

Sono disponibili i seguenti altri simboli/costanti:

	+------------------+------------------------------------------------------------------------------------------+
	| funzione         | Descrizione                                                                              |
	+==================+==========================================================================================+
	| NONE o None      | Costante nulla                                                                           |
	+------------------+------------------------------------------------------------------------------------------+
	| __ts_start__     | timestamp di inizio serie (x0)                                                           |
	+------------------+------------------------------------------------------------------------------------------+
	| __ts_end__       | timestamp di fine serie (xn)                                                             |
	+------------------+------------------------------------------------------------------------------------------+
	| __step__         | delta temporale tra i valori della serie                                                 |
	+------------------+------------------------------------------------------------------------------------------+
	|  __t__           | timestamp corrente all'interno della serie                                               |
	+------------------+------------------------------------------------------------------------------------------+
	| __i__            | indice dell'elemento calcolato nella serie                                               |
	+------------------+------------------------------------------------------------------------------------------+
	
Esempio: 

::
	
	datagroup test
		...
		datasource ds1                 <--+
		exit                              | referenzia "ds1"
		...                               |
		timegraph grafico1                |
			serie v                   |           <--+
				var-value ds1  <--+              | utilizza "v"
			exit                                     |
			serie x                                  |
				computed-value v * __i__ * 3  <--+
			exit
		exit
	exit

Valori nulli nelle serie
------------------------

Se all'interno di una espressione viene usata una serie che contiene i valori nulli, il singolo valori della
serie viene considerato equivalente alla costante "NONE".

Il valore "NONE" e' un valore speciale e quando viene usato all'interno di una espressione aritmetica provoca
come risultato sempre "NONE".

L'unico modo per evitare questo comportamento e' usare la funzione isnull().


Esempio: ::

	datagroup prova
		...
		datasource ds1
			...
		exit
		...
		timegraph grafico1
			serie x
				var-value ds1
			exit
			serie y1
				computed-value "x * 3"
			exit

			serie y2
				computed-value "ifnull(x, 0) * 3"
			exit
		exit
		...
	exit
	
Ordine di valutazione delle serie
---------------------------------

Quando vengono calcolate le serie l'ordine di valutazione e' il seguente:
- estrazione dei valori per tutte le serie "reali" (con parametro 'var-value')
- per ogni serie con parametro "computed-value" calcolo valori in base alle serie "reali"

Questo significa che non e' possibile utilizzare all'interno di una espressione di una serie
calcolate i valori di un'altra serie calcolata.

"""

import re
import math
import time

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#



class SafeNoneType(object):

	def __init__(self):
		pass

	def __repr__(self):                      return None
	def __str__(self):                       return str(None)
	def __int__(self):                       return None
	def __float__(self):                     return None
	def __bool__(self):                      return None
	def __complex__(self):                   return None

	def __eq__(self, other):
		return False
	def __ne__(self, other):                 
		return False
	def __lt__(self, other):                 
		return False
	def __gt__(self, other):                 
		return False
	def __le__(self, other):                 
		return False
	def __ge__(self, other):                 
		return False
	def __nonzero__(self):
		return True

	def __neg__(self):                       return None
	def __pos__(self):                       return None
	def __add__(self, other):                return None
	def __radd__(self, other):               return None
	def __sub__(self, other):                return None
	def __rsub__(self, other):               return None
	def __mul__(self, other):                return None
	def __rmul__(self, other):               return None
	def __div__(self, other):                return None
	def __rdiv__(self, other):               return None
	def __truediv__(self, other):            return None
	def __rtruediv__(self, other):           return None
	def __floordiv__(self, other):           return None
	def __rfloordiv__(other):                return None
	def __mod__(self, other):                return None
	def __rmod__(self, other):               return None
	def __pow__(self, othr, modulo=None):    return None
	def __rpow__(self, other, modulo=None):  return None
	
	def __lshift__(self, other):             return None
	def __rlshift__(self, other):            return None
	def __rshift__(self, other):             return None
	def __rrshift__(self, other):            return None
	def __and__(self, other):                return None
	def __rand__(self, other):               return None
	def __or__(self, other):                 return None
	def __ror__(self, other):                return None
	def __xor__(self, other):                return None
	def __rxor__(self, other):               return None
	def __invert__(self):                    return None
	def __abs__(self):                       return None
	def __round__(self, n=None):             return None
	def __divmod__(self, other):             return None

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def TIMEGRAPH_EXPR_SERIE_SYMBOL(serie_name):
	
	return serie_name

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def _fun_ifnull(v, d):
	if v is None or isinstance(v, SafeNoneType):
		return d
	return v

def _fun_iftrue(i, t, e):
	if i:
		return t
	else:
		return e

def create_timegraphserie_expr_contex(symbols=None):


	context = {
	}

	for symbol in dir(__builtins__):
		context[ symbol ] = None

	context['abs'   ] = abs
	context['divmod'] = divmod
	context['int'   ] = int
	context['pow'   ] = pow
	context['float' ] = float
	context['long'  ] = long
	context['max'   ] = max
	context['min'   ] = min
	
	context['log'   ] = math.log
	context['sqrt'  ] = math.sqrt
	
	context['sin'   ] = math.sin
	context['cos'   ] = math.cos
	context['radians'] = math.radians
	
	context['iftrue'] = _fun_iftrue
	context['ifnull'] = _fun_ifnull
	
	context['NONE'  ] = SafeNoneType()
	context['None'  ] = SafeNoneType()

	if symbols is not None:
		context.update(symbols)

	return context
	

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def compile_timegraphserie_expr(expr_str):

	#return expr_str
	return compile( expr_str, '<string>', 'eval')
	
def validate_timegraphserie_expr(expr_str):
	system_symbols = create_timegraphserie_expr_contex()
	try:
		eval(expr_str, {}, system_symbols)
	except SyntaxError, e:
		raise ValueError("Invalid expression: %s" % (e))
	except Exception, e:
		pass
		#print e
	
	return True

def evaluate_timegraphserie_expr(exec_context, expr_str):
	
	return eval(expr_str, {}, exec_context)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#


def calculate_timegraph_series_data(datasource_store, tg, ts_start, ts_end, g_type=None, include_errors=False, cache_manager=None):
	
		if cache_manager is None:
			cache_manager    = init_cache_manager_from_conf()

		# set start and end time for all rrd data
		now = int(time.time())
		if ts_start == None:
			ts_start = now - 60*60*24*7
		else:
			# timestamp must be integer value
			ts_start = int(ts_start)

		if ts_end == 'now':
			ts_end = now

		tenant = tg.tenant

		dg = tg.datagroup

		real_series = tg.get_real_series().all()

		# Troviamo i datasource referenziati	
		datasources_by_name = dict([ (ds.name, ds) for ds in dg.datasources.filter(name__in=[ s.variable_value for s in real_series ]) ])

		#
		# CARICAMENTO DATI DELLE SERIE
		#
		max_val = 0
		min_val = 0
		max_num_values = 0

		step = dg.minperiod
		
		# il fetch dei dati potrebbe restituire dei dati con dei range temporali "diversi" da quelli di partenza
		effective_ts_start = ts_start
		effective_ts_end   = ts_end
		effective_step     = step

		#print "calculate_timegraph_series_data", effective_ts_start, effective_ts_end, effective_step
		
		
		# Cerchiamo di ottenere i dati delle serie 
		for s in real_series:
			try:
				ds = datasources_by_name[s.variable_value] 
				ds_data = datasource_store.fetch_values(tenant.uuid, ds.uuid, ts_start=ts_start, ts_end=ts_end, resolution=step, min_value=ds.min_val, max_value=ds.max_val)

				s.ds_data = ds_data
				
				max_val            = max(max_val, max(ds_data['dataseries']) )
				min_val            = max(min_val, min(ds_data['dataseries']) )

				#print ds, ds_data['scalex'], len(ds_data['dataseries'])
				
				effective_ts_start = min(effective_ts_start, ds_data['scalex'][0] )
				effective_ts_end   = max(effective_ts_end  , ds_data['scalex'][1] )
				#effective_step     = min(effective_step    , ds_data['scalex'][2] )
				effective_step     = ds_data['scalex'][2]
			except Exception, e:
				pass

		#print "effective"
		#print "\t", effective_ts_start
		#print "\t", effective_ts_end
		#print "\t", effective_step


		# Se per alcune serie non abbiamo i valori, produciamo dei dati nulli compatibili con
		# le serie che invece hanno dati reali.
		if effective_step == 0:
			effective_step = 60
		max_num_values = int((effective_ts_end - effective_ts_start) / effective_step)
		NULL_SERIE = [ None for n in xrange( max_num_values ) ]
		fake_ds_data = {
			'scalex'    : [effective_ts_start,effective_ts_end, effective_step],
			'dataseries': NULL_SERIE
		}

		for s in real_series:
			if not hasattr(s, 'ds_data'):
				s.ds_data = fake_ds_data


		#
		# CALCOLIAMO I VALORI DELLE SERIE "COMPUTED"
		#
		
		# Definiamo le variabili comuni per il calcolo
		c = {
			"__ts_start__" : effective_ts_start,
			"__ts_end__"   : effective_ts_end,
			"__step__"     : effective_step,
			"__max_val__"  : max_val,
			"__min_val__"  : min_val,
		}
		expr_context = create_timegraphserie_expr_contex(c)
		
		computed_series = tg.get_computed_series().all()
		for cs in computed_series:
			
			compiled_expr = compile_timegraphserie_expr(cs.computed_value_expr)

			cs_values = []
			t = float(effective_ts_start)
			
			errors = []
			
			for i in xrange(max_num_values):

				try:
					for real_s in real_series:
						values = real_s.ds_data['dataseries']
						v = values[i]
						if v is None:
							v = SafeNoneType()
						
						#serie_symbol_name = TIMEGRAPH_EXPR_SERIE_SYMBOL(real_s.name)
						
						#expr_context[ serie_symbol_name ] = v
						expr_context[ real_s.name ] = v
						
					expr_context[ "__i__" ] = i
					expr_context[ "__t__" ] = t
						
					computed_val = evaluate_timegraphserie_expr(expr_context, compiled_expr)

				except Exception, e:
					computed_val = None

					if include_errors:
						errors.append(str(e))

				cs_values.append(computed_val)
				
				t += effective_step

			ds_data = {
				'scalex'    : [ effective_ts_start, effective_ts_end, effective_step ],
				'dataseries': cs_values,
			}
			
			if include_errors:
				ds_data['errors'] = errors 

			cs.ds_data = ds_data

		series_by_name = dict([ (s.name, s) for s in list(real_series) + list(computed_series) ])
		
		return datasources_by_name, series_by_name




