 
# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

#
# Da TimeGraph a rrdgraph
# =======================
#
# Per poter renderizzare i timegrapgh usando RRDTOOL e' necessario "descrivere"
# in formato rrdgraph le diverse serie del timegraph.
# 
# Questa operazione e' banale per le serie "normali", mentre e' un problema per 
# le serie "computed value" 
# 
#   * I timegraph esprimono le serie "computed" usando espressioni in "notazione infissa"
#   * RRDTOOL descrive le espressioni usando "notazione postfissa" o (notazione polacca inversa)
#   * Gli operatori e funzioni matematiche utilizzabili nelle serie "computed" non hanno
#     un corrispondente "diretto" con gli operatori/funzioni usabili nelle espressioni RRDTOOL.
# 
# Per gestire questa cosa e' necessario "convertire" le espressioni delle serie "computed"
# da formato "timegraph" a formato RRDTOOL.
#
# Poiche' le espressioni "computed" usano sintassi python possiamo utilizzare
# il modulo parser di python per:
#
#    1) PARSING    : Parsare l'espressione "computed"
#
#    2) CONVERSIONE: Manipolare l'albero dell'espressione (expr tree) per rendere 
#       l'espressione piu' compatibile con gli operatori usati da RRDTOOL
#
#    3) ESPORTAZIONE: esportare l'albero espressione  in notazione postfissa
# 

import sys
import ast


#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def _print(l, tree):
	s = '\t' * l + str(tree) + '\n'
	return s

class BaseTree(object):

	def __init__(self, type, value, params):
		self.type   = type
		self.value  = value
		self.params = params

	def __repr__(self):
		return self.type

	def __str__(self):
		return '' #self.type

	def visit_tree(self, l=0, visit_fun=_print):

		s = visit_fun(l, self)
		
		if self.params:
			for a in self.params:
				s += a.visit_tree(l=l+1, visit_fun=visit_fun)
				
		return s

	def visit_tree_rpn(self, l=0, mapping=None):

		s = ''
		if self.params:
			for a in self.params:
				s += a.visit_tree_rpn(l=l+1, mapping=mapping)

		#s += repr(self)

		return s


class NumTree(BaseTree):

	def __init__(self, type, value, params):
		super(NumTree, self).__init__(type, value, params)
		
	def __str__(self):
		return repr(self.value)

	def visit_tree(self, l=0, visit_fun=_print):
		return visit_fun(l, self)

	def visit_tree_rpn(self, l=0, mapping=None):
		return str(self.value)


class NameTree(BaseTree):

	def __init__(self, type, value, params):
		super(NameTree, self).__init__(type, value, params)

	def __str__(self):
		#return "<id: %s>" % (self.value)
		return self.value

	def visit_tree(self, l=0, visit_fun=_print):
		return visit_fun(l, self)

	def visit_tree_rpn(self, l=0, mapping=None):
		
		if mapping:
			return mapping.get(self.value, self.value)
			
		return self.value

class StrTree(BaseTree):

	def __init__(self, type, value, params):
		super(StrTree, self).__init__(type, value, params)

	def __str__(self):
		return repr(self.value)

	def visit_tree(self, l=0, visit_fun=_print):
		return visit_fun(l, self)

	def visit_tree_rpn(self, l=0, mapping=None):
		return self.value
		
class UnaryOpTree(BaseTree):

	OPERATORS = {
		'Not' : 'not',
	}	

	def __init__(self, type, value, params):
		super(UnaryOpTree, self).__init__(type, value, params)

	def __str__(self):
		return UnaryOpTree.OPERATORS.get( self.params[0].type , self.params[0].type )

	def visit_tree(self, l=0, visit_fun=_print):
		
		s = ''
		
		s += visit_fun(l, self)
		s += ' '
		s += self.params[1].visit_tree(l=l+1, visit_fun=visit_fun)
		
		return s

	def visit_tree_rpn(self, l=0, mapping=None):
		s = ''
		s += self.params[1].visit_tree_rpn(l=l+1, mapping=mapping)
		s += ','
		s += UnaryOpTree.OPERATORS.get( self.params[0].type , self.params[0].type )
		return s

class BinOpTree(BaseTree):
	
	OPERATORS = {
		'Add' : '+',
		'Sub' : '-',
		'Mult': '*',
		'Div' : '/',
	}

	def __init__(self, type, value, params):
		super(BinOpTree, self).__init__(type, value, params)

	def __str__(self):
		return "%s" % (BinOpTree.OPERATORS.get( self.params[1].type , self.params[1].type ))

	def visit_tree(self, l=0, visit_fun=_print):
		s = ''
		s += '('
		s += self.params[0].visit_tree(l=l+1, visit_fun=visit_fun)
		s += ' '
		s += visit_fun(l, self)
		s += ' '
		s += self.params[2].visit_tree(l=l+1, visit_fun=visit_fun)
		s += ')'
		
		return s

	def visit_tree_rpn(self, l=0, mapping=None):
		s = ''
		s += self.params[0].visit_tree_rpn(l=l+1, mapping=mapping)
		s += ','
		s += self.params[2].visit_tree_rpn(l=l+1, mapping=mapping)
		s += ','
		s += BinOpTree.OPERATORS.get( self.params[1].type , self.params[1].type )
		return s

class BoolOpTree(BaseTree):

	OPERATORS = {
		'And': 'and',
		'Or' : 'or',
	}

	def __init__(self, type, value, params):
		super(BoolOpTree, self).__init__(type, value, params)

	def __str__(self):
		return " %s " % (BoolOpTree.OPERATORS.get( self.params[0].type , self.params[0].type ))

	def visit_tree(self, l=0, visit_fun=_print):

		s = ''

		s += '('
		
		o = visit_fun(l, self)

		#values = []
		#if self.params:
		#	for a in self.params[1:]:
		#		v = a.visit_tree(l=l+1, visit_fun=visit_fun)
		#		values.append(v)
		#		
		#s += o.join(values)
		
		s += o
				
		s += ')'
		
		return s

	def visit_tree_rpn(self, l=0, mapping=None):
		s = ''

		if self.params:
			for a in self.params[1:]:
				s += a.visit_tree_rpn(l=l+1, mapping=mapping)
				s += ','
		
		s += BoolOpTree.OPERATORS.get( self.params[0].type , self.params[0].type )
		return s

class CompareTree(BaseTree):

	OPERATORS = {
		'Gt' : '>',
		'Lt' : '<',
		'LtE': '<=',
		'GtE': '>=',
	}

	def __init__(self, type, value, params):
		super(CompareTree, self).__init__(type, value, params)

	def __str__(self):
		return "%s" % (CompareTree.OPERATORS.get( self.params[1].type , self.params[1].type ))

	def visit_tree(self, l=0, visit_fun=_print):
		
		s = ''

		s += '('
		s += self.params[0].visit_tree(l=l+1, visit_fun=visit_fun)
		s += ' '
		s += visit_fun(l, self)
		s += ' '
		s += self.params[2].visit_tree(l=l+1, visit_fun=visit_fun)
		s += ')'
		
		return s

	def visit_tree_rpn(self, l=0, mapping=None):
		s = ''
		s += self.params[0].visit_tree_rpn(l=l+1, mapping=mapping)
		s += ','
		s += self.params[2].visit_tree_rpn(l=l+1, mapping=mapping)
		s += ','
		
		v = CompareTree.OPERATORS.get( self.params[1].type , self.params[1].type )
		if mapping:
			v = mapping.get(v,v)
		s += v
		
		return s



class CallTree(BaseTree):

	def __init__(self, type, value, params):
		super(CallTree, self).__init__(type, value, params)

	def __str__(self):
		return self.params[0].value

	def visit_tree(self, l=0, visit_fun=_print):

		s = ''

		s += visit_fun(l, self)

		s += '('
		if self.params:
			values = []
			for a in self.params[1:]:
				v = a.visit_tree(l=l+1, visit_fun=visit_fun)
				values.append(v)
				
			s += ','.join(values)
		s += ')'
				
		return s

	def visit_tree_rpn(self, l=0, mapping=None):
		
		s = ''
		
		if self.params:
			for a in self.params[1:]:
				s += a.visit_tree_rpn(l=l+1, mapping=mapping)
				s += ','

		s += self.params[0].visit_tree_rpn(l=l+1, mapping=mapping)
		
		return s

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class PythonExprParser(ast.NodeVisitor):
	
	TREE_TYPES = {
		'Call'   : CallTree,
		'BinOp'  : BinOpTree,
		'BoolOp' : BoolOpTree,
		'UnaryOp': UnaryOpTree,
		'Compare': CompareTree,
		
		'Num'    : NumTree,
		'Str'    : StrTree,
		'Name'   : NameTree,
	}	

	def __init__(self, *args, **kwargs):
		
		super(PythonExprParser, self).__init__(*args, **kwargs)
		
		self.level = 0
		self.visit_stack = [[]]
		self.expr_tree = None

	def generic_visit(self, node):
		self._produce_epxr_tree(node)


	def _produce_epxr_tree(self, node):

		value    = None
		subtrees = None
		
		ast_node_type = type(node).__name__

		if  ast_node_type == 'Name':

			value = node.id

		elif  ast_node_type == 'Num':

			value = node.n

		elif  ast_node_type == 'Str':
			value = node.s
		else:
			self.visit_stack.append([])
			self.level += 1
			
			ast.NodeVisitor.generic_visit(self, node)
			
			self.level -= 1

			subtrees = self.visit_stack.pop()
			
		klass = PythonExprParser.TREE_TYPES.get( ast_node_type , BaseTree )

		self.expr_tree = klass(ast_node_type, value, subtrees)

		self.visit_stack[-1].append(self.expr_tree)


	def parse_string(self, s):
		
		ast_tree = ast.parse(s)
		
		self.visit( ast_tree )
		
		return self.expr_tree

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def _produce_rrd_compatible_expr_string(l, tree):
	
	if tree.type == 'BoolOp':
		
		if tree.params[0].type == 'Or':
			#  a or b  ->  iftrue( iftrue(a, 1, 0) + iftrue(b, 1, 0) , 1, 0)

			r = []
			for p in tree.params[1:]:
				s = 'iftrue(%s,1,0)' % (p.visit_tree(l=l+1, visit_fun=_produce_rrd_compatible_expr_string))
				r.append( s )
				
			s = 'iftrue(%s,1,0)' % ('+'.join(r))
			

		elif tree.params[0].type == 'And':
			#  a and b  ->  iftrue( a, iftrue(b, 1, 0), 0)
		
			s = ''
			
			def _concat_if(params):

				s1 = ''
				s1 += 'iftrue('
				
				a = params[0]
				
				s1 += a.visit_tree(l=l+1, visit_fun=_produce_rrd_compatible_expr_string)
				s1 += ','
				
				if params[1:]:
					s1 += _concat_if(params[1:])
				else:
					s1 += '1'
				
				s1 += ','
				s1 += '0'
				s1 += ')'
				
				return s1

			s += _concat_if(tree.params[1:])
		
		return s
	else:
		s = str(tree)
		return s
	
#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#


def convert_expr_to_rpn(s):
	

	
	
	print "============================================"

	print "Expr: ", s

	parser = PythonExprParser()

	print "Parsing expression..."
	expr_tree = parser.parse_string(s)
	
	print "Converting/chaning expression to RRDTOOL COMPATIBLE expr..."
	
	s2 = expr_tree.visit_tree(visit_fun=_produce_rrd_compatible_expr_string)
	print "Expr: ", s2
	
	print "Parsing (again)..."
	expr_tree = parser.parse_string(s2)

	print "Converting notation to RPN..."
	mapping = {
		'iftrue': 'IF',
		'<'     : 'LT',
		'>'     : 'GT',
		'<='    : 'LE',
		'>='    : 'GE',
	}
	
	rpn_expr_string = expr_tree.visit_tree_rpn(mapping=mapping)

	print "RPN expression: ", rpn_expr_string
	
	print "============================================"

	return rpn_expr_string

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

if __name__ == "__main__":
	
	import sys
	
	print convert_expr_to_rpn(sys.argv[1])
