# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

Schema di come la classe AbstractLocalPollerConfiguration viene implementata/estesa                        
per essere usata durante l'esecuzione/check da parte di:                                                   
                                                                                                           
    * exec_datagroup  con datagroup locali                                                                    
    * exec_datagroup  con datagroup gestiti da agenti remoti                                                  
    * exec di datagroup su un agente locale                                                                   
                                                                                                           
                                                                                                           
                                DATABASE                                                             
                                   ^                                                                 
                                   |      .........................                                                          
                                          v                       .
    PollerConfiguration    LocalConfiguration        LocalToRemoteConfiguration       +------- RemoteConfiguration      
             ^                     ^                              |                   |                 ^               
             |                     |                              |                   |                 |               
             |                     |                              |                   |                 |               
           AGENT                   |                           SANETD <---------------+           REMOTE AGENT          
                                   |                                                                    ^               
                                   |                                                                    |               
                                   +-------------------------+    +-------------------------------------+               
                                                             |    |                                        
                                                         exec_datagroup                                    
                                                                                                            
"""

from sanet_poller.lib.exceptions import *

from sanet_poller.lib.classes   import PNode, PIface, PStorage, PDevice, PService

from sanet_poller.lib.classes import AbstractLocalPollerConfiguration

from core.models import *

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

class LocalConfiguration(AbstractLocalPollerConfiguration):
	"""
	Questo oggetto e' una versione "fast" di:
	
		sanet_poller.lib.classes.AbstractLocalPollerConfiguration 
	
	Server per accedere direttamente al DB di sanet caricando "on demand" i 
	nodi/interfacce/ecc. solo quando vengono chiesti.
	
	Puo' essere usata solo localmente e andrebbe usata SOLO per operazioni
	di test / debug.
	"""
	
	def __init__(self, logger, tenant_id, cache_manager):
		super(LocalConfiguration, self).__init__(logger, tenant_id, cache_manager)

		self._tenant = Tenant.objects.get(uuid=self.tenant_id)

	def _set_tenant_schema(self):
		self.logger.debug("SETTING SCHEMA: %s", self._tenant.name)
		self._tenant.db_set_current_schema()

	def _get_cluster(self, cid):
		self._set_tenant_schema()
		try:
			c = Cluster.objects.get(uuid=cid)
			return c.get_dict()
		except Cluster.DoesNotExist, e:
			raise ClusterException("cluster %s not found. (not loaded?)" % (cid))

	def _get_cluster_elements_ids(self, cid):
		self._set_tenant_schema()
		try:
			c = Cluster.objects.get(uuid=cid)
			elements = MonitoredElement.objects.filter(cluster=c).values_list("uuid")		
			data = [row[0] for row in elements]
			return data
		except Cluster.DoesNotExist, e:
			raise ClusterException("cluster %s not found. (not loaded?)" % (cid))

	def _get_datagroup(self, dgid):
		self._set_tenant_schema()
		try:
			dg = DataGroup.objects.get(uuid=dgid)
			dg.load_fields_from_cache(self.cache_manager)
			self._cache[dgid] = dg
			return dg.get_dict_with_ds_and_c()
		except DataGroup.DoesNotExist, e:
			raise DataGroupNotFoundException("DataGroup %s not found (%s:%s)" % (dgid, type(e), e))

	def _get_datagroup_params(self, dgid):
		self._set_tenant_schema()
		try:
			dg = DataGroup.objects.get(uuid=dgid)
			raw_params = DataGroupParameter.objects.filter(datagroup=dg)
			return [ p.get_dict() for p in raw_params ]
		except DataGroup.DoesNotExist, e:
			raise DataGroupNotFoundException("DataSource %s not found" % (dsid))

	def _get_datasource(self, dsid):
		self._set_tenant_schema()
		try:
			ds = DataSource.objects.get(uuid=dsid)
			ds.load_fields_from_cache(self.cache_manager)
			return ds.get_dict()
		except DataSource.DoesNotExist, e:
			raise DataSourceNotFoundException("DataSource %s not found" % (dsid))

	def _get_condition(self, cid):
		self._set_tenant_schema()
		try:
			c = Condition.objects.get(uuid=cid)
			c.load_fields_from_cache(self.cache_manager)
			return c.get_dict()
		except Condition.DoesNotExist, e:
			raise ConditionNotFoundException("Condition %s not found" % (cid))

	def _get_condition_by_path(self, path):
		self._set_tenant_schema()
		try:
			c = Condition.objects.get(path=path)
			c.load_fields_from_cache(self.cache_manager)
			return c.get_dict()
		except Condition.DoesNotExist, e:
			raise ConditionNotFoundException("Condition %s not found (%s)" % (path, str(e)))

	def _get_condition_by_classification(self, classification, node_id, subelement_id):
		self._set_tenant_schema()
		try:

			c = Condition.objects.filter(classification=classification).filter(Q(element__uuid=node_id)|Q(element__uuid=subelement_id))
			c = c.order_by('-element__kind')
			c = c.all()[:1]
			if len(c) == 0:
				raise ConditionNotFoundException("Condition %s not found" % (classification))
			c = c[0]

			c.load_fields_from_cache(self.cache_manager)
			return c.get_dict()

		except Condition.DoesNotExist, e:
			raise ConditionNotFoundException("Condition with classification '%s' not found (%s)" % (classification, str(e)))

	def _get_element_by_id(self, e_id):
		self._set_tenant_schema()
		try:
			me = MonitoredElement.objects.get(uuid=e_id)
			e = me.get_specific_element()
			e.load_fields_from_cache(self.cache_manager)
			return e.get_dict()
		except MonitoredElement.DoesNotExist, e:
			raise ResourceNotFoundException("element:%s" % (e_id))

	def _get_element_by_parent_id(self, p_id, kind):
		r = []
		for me in MonitoredElement.objects.filter(parent__uuid=p_id, kind=kind):
			e = me.get_specific_element()
			e.load_fields_from_cache(self.cache_manager)
			r.append(e.get_dict())
		return r

	def _get_linked_iface_by_id(self, iface_id, link_type = None):
		self._set_tenant_schema()
		try:
			i = MonitoredIface.objects.get(uuid=iface_id)
		except Exception, e:
			raise ResourceNotFoundException("Interface %s not found" % (iface_id))

		link_elements = LinkElement.objects.filter(iface=i)
		if link_type != None:
			link_elements = link_elements.filter(link__ltype=link_type)
		link_elements = link_elements.all()

		if link_elements: 
			le1 = link_elements[0]
			link = le1.link
			try:
				le2 = LinkElement.objects.filter(link=link).exclude(iface=i)
				le2 = le2[0]
				return le2.iface.get_dict() 
			except Exception, e:
				raise ResourceNotFoundException("Linked iface %s for %s not found" % (iid, iface_id))
		return None

	#----------------------------------------------------------------------#


	#
	# Useful methods for local evaluation
	#
	def get_node_by_name(self, name):

		node = MonitoredNode.objects.get(name=name)

		node.load_fields_from_cache(self.cache_manager)

		pn = PNode()
		pn.load_from_dict( node.get_dict() )

		return pn

	def get_subelement_by_name(self, node_name, kind, name):

		node = MonitoredNode.objects.get(name=node_name)
		me  = MonitoredElement.objects.get(kind=kind, name=name, parent=node)

		e = me.get_specific_element()
		e.load_fields_from_cache(self.cache_manager)

		if   e.kind == ELEMENT_KIND_NODE   : pe = PNode()
		elif e.kind == ELEMENT_KIND_IFACE  : pe = PIface()
		elif e.kind == ELEMENT_KIND_STORAGE: pe = PStorage()
		elif e.kind == ELEMENT_KIND_SERVICE: pe = PService()
		elif e.kind == ELEMENT_KIND_DEVICE : pe = PDevice()
		else: raise Exception("INVALID KIND: %s" % (e.kind))

		pe.load_from_dict( e.get_dict() )
		
		return pe

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

class LocalToRemoteConfiguration(LocalConfiguration):
	"""
	Questa classe viene utilizzata per caricare le informazioni di configurazione
	da inviare agli agenti remoti per fare check on-demand.
	
	Rispetto alla LocalToRemoteConfiguration questa classe restituisce i dati
	nel formato "dizionario" (dict) in modo che siano serializzabili in formato
	JSON e poter essere inviati in risposta all'agente remoto.
	"""

	def get_cluster(self, cid):
		# Return raw data (dict) directly
		return self._get_cluster(cid)

	def get_cluster_elements(self, cid):
		
		# Return raw data (dict) directly
		elements_ids = self.get_cluster_elements_ids(cid)
		r = []
		for eid in elements_ids:
			element = self.get_element_by_id(eid)
			r.append(element)
		return r

	def get_cluster_elements_ids(self, cid):
		return self._get_cluster_elements_ids(cid)
		
	#	self._set_tenant_schema()
	#	try:
	#		c = Cluster.objects.get(uuid=cid)
	#		elements = MonitoredElement.objects.filter(cluster=c).values_list("uuid")		
	#		data = [row[0] for row in elements]
	#		return data
	#	except Cluster.DoesNotExist, e:
	#		raise ClusterException("cluster %s not found. (not loaded?)" % (cid))




	def get_datagroup(self, dgid):
		# Return raw data (dict) directly
		return self._get_datagroup(dgid)

	def get_datagroup_params(self, dgid):
		# Return raw data (dict) directly
		return self._get_datagroup_params(dgid)

	def get_datasource(self, dsid):
		# Return raw data (dict) directly
		return self._get_datasource(dsid)

	def get_condition(self, cid):
		# Return raw data (dict) directly
		return self._get_condition(cid)

	def get_condition_by_path(self, path):
		# Return raw data (dict) directly
		return self._get_condition_by_path(path)

	def get_condition_by_classification(self, classification, node_id, subelement_id):
		# Return raw data (dict) directly
		return self._get_condition_by_classification(path, node_id, subelement_id)

	def get_element_by_id(self, e_id):
		# Return raw data (dict) directly
		return self._get_element_by_id(e_id)

	def get_element_by_parent_id(self, p_id, kind):
		# Return raw data (dict) directly
		return self._get_element_by_parent_id(p_id, kind)

	def get_linked_iface_by_id(self, iface_id, link_type = None):
		# Return raw data (dict) directly
		return self._get_linked_iface_by_id(iface_id, link_type=link_type)




