# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

check_schema
========================================================================

Syntax: 

	::
		
		sanet-mange check_schema

Updates Sanet DataBase schema.

.. warning:: This is an internal Sanet command. It is not recommended to use this command.

"""

import os
import sys
import time
import logging
import getpass
import pytz
from optparse import make_option

import django
from django.conf import settings
from django.db import connection
from django.db import transaction
from django.utils.translation import ugettext as _
from django.contrib.auth.models import Group

from sanet_common.sanetdefs.constants import *

from conf import check_base_dirs

from core.db_utils import transaction_commit_on_success
#from core.db_utils import transaction_set_dirty
from core.db_utils import transaction_commit

from core.models import DBSetting
from core.models import sanet_post_update_schema

from core.api.base import tenant_select_all
from core.api.base import db_version, set_db_version
from core.api.base import tenant_select_all
from core.api.base import check_indexes_on_tables

from core.management.commands import SanetBaseCommand

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

def _signal_schema_update(schema_version):

	#
	# TODO: need to handle return informations after updates?
	#
	sanet_post_update_schema.send(None, schema_version=schema_version)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

def check_and_update_schema(logger):

	_version = tuple([int(n) for n in django.get_version().split('.')])

	version =  db_version()	

	if _version > (1,4):
		
		# Last version before django 1.11 
		#assert version == 45
		
		return


	logger.info( "Checking schema (version: %s)" % (db_version()) )

	if version <  1: update_schema_0_1()
	if version <  2: update_schema_1_2()
	if version <  3: update_schema_2_3()
	if version <  4: update_schema_3_4()
	if version <  5: update_schema_4_5()
	if version <  6: update_schema_5_6()
	if version <  7: update_schema_6_7()
	if version <  8: update_schema_7_8()
	if version <  9: update_schema_8_9()
	if version < 10: update_schema_9_10()
	if version < 11: update_schema_10_11()
	if version < 12: update_schema_11_12()
	if version < 13: update_schema_12_13()
	if version < 14: update_schema_13_14()
	if version < 15: update_schema_14_15()
	if version < 16: update_schema_15_16()
	if version < 17: update_schema_16_17()
	if version < 18: update_schema_17_18()
	if version < 19: update_schema_18_19()
	if version < 20: update_schema_19_20()
	if version < 21: update_schema_20_21()
	if version < 22: update_schema_21_22()
	if version < 23: update_schema_22_23()
	if version < 24: update_schema_23_24()
	if version < 25: update_schema_24_25()
	if version < 26: update_schema_25_26()
	if version < 27: update_schema_26_27()

	if version < 28: update_schema_27_28() # multi-template + sub-element template
	
	if version < 29: update_schema_28_29() # log webscenario
	
	if version < 30: update_schema_29_30() # fix multi-template update

	if version < 31: update_schema_30_31() # *index -> string

	if version < 32: update_schema_31_32() # tagged_parent

	if version < 33: update_schema_32_33() # Config profilation (CLI + WEB)

	if version < 34: update_schema_33_34() # WebScenario: Raw Data in POST (#2864)
	if version < 35: update_schema_34_35() # Scheduling Groups
	if version < 36: update_schema_35_36() # owner in Cluster
	if version < 37: update_schema_36_37() # custom group instance
	if version < 38: update_schema_37_38() # fix custom group instance unique index

	if version < 39: update_schema_38_39() # map background
	if version < 40: update_schema_39_40() # ifspeed (for heatmap)

	if version < 41: update_schema_40_41() # ctime for monitoredelements
	
	if version < 42: update_schema_41_42() # report params VARCHAR -> TEXT
	
	if version < 43: update_schema_42_43() # remove logicmap.group_perm

	if version < 44: update_schema_43_44() # event.condition/element_uuid

	if version < 45: update_schema_44_45() # Last schema version before porting to django 1.11

	#if version < 46: update_schema_45_46() # Last schema version before porting to django 1.11


	return

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

def table_exists(name):
	c=connection.cursor()
	cmd="select count(*)>0 from pg_tables where tablename=%s and tableowner=%s"
	c.execute(cmd, [name, settings.DATABASES['default']['USER']])
	rv=c.fetchall()[0][0]
	return rv

def column_exists(table_name, column_name, schema_name=None):

	ret = False

	cursor = connection.cursor()

	if schema_name is None: schema_name = 'public';

	schema_name = schema_name.lower()

	cmd = 'set search_path TO  %s;' % (schema_name)
	cursor.execute( cmd )

	cmd = """SELECT EXISTS (SELECT 1 FROM information_schema.columns WHERE table_schema='%s' AND table_name='%s' AND column_name='%s');""" % (schema_name, table_name, column_name)

	cursor.execute( cmd )
	exists = cursor.fetchone()[0]

	return exists

def index_exists(schema, table, indexname):
	
	c = connection.cursor()
	
	cmd = '''SELECT count(*) FROM pg_indexes WHERE schemaname = '%s' and tablename = '%s' and indexname = '%s';''' % (schema, table, indexname)
	
	c.execute( cmd )
	data = c.fetchall()
	if data[0][0] >= 1:
		return True
	else:
		return False

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_0_1():
	print "fake schema update"
	set_db_version(1)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_1_2():

	print "Updating schema version from 1 to 2"
	
	db_cursor = connection.cursor()

	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:
		print "Updating schema %s" % (name)
		
	
	db_cursor = connection.cursor()
	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:		
		db_cursor.execute("SET search_path TO %s;" % (name))
		db_cursor.execute("ALTER TABLE condition      ADD COLUMN deleted BOOLEAN NOT NULL DEFAULT false;")
		db_cursor.execute("ALTER TABLE datasource     ADD COLUMN deleted BOOLEAN NOT NULL DEFAULT false;")
		db_cursor.execute("ALTER TABLE timegraph      ADD COLUMN deleted BOOLEAN NOT NULL DEFAULT false;")
		db_cursor.execute("ALTER TABLE timegraphserie ADD COLUMN deleted BOOLEAN NOT NULL DEFAULT false;")
  
	db_cursor.execute("SET search_path TO public;")
	set_db_version(2)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_2_3():

	print "Updating schema version from 2 to 3"
	
	db_cursor = connection.cursor()

	db_cursor = connection.cursor()
	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:		
		print "Updating schema tenant %s" % (name)
		db_cursor.execute("SET search_path TO %s;" % (name))
		q = "ALTER TABLE timegraph      ADD COLUMN classification TEXT NOT NULL DEFAULT '';"
		print q
		db_cursor.execute(q)
  
	db_cursor.execute("SET search_path TO public;")
	set_db_version(3)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_3_4():

	print "Updating schema version from 3 to 4"
	
	db_cursor = connection.cursor()
	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:		
		print "Updating schema %s" % (name)
		db_cursor.execute("SET search_path TO %s;" % (name))
		db_cursor.execute("ALTER TABLE tagtree        ADD COLUMN uniques  BOOLEAN NOT NULL DEFAULT false;")
  
	db_cursor.execute("SET search_path TO public;")
	set_db_version(4)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_4_5():

	print "Updating schema version from 4 to 5"
	
	db_cursor = connection.cursor()
	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:		
		print "Updating schema %s" % (name)
		db_cursor.execute("SET search_path TO %s;" % (name))
		
		sql = "ALTER TABLE  logicmapconnection DROP  CONSTRAINT IF EXISTS logicmapconnection_tenant_id_logicmap_id_entity1_id_entity2_key;"
		print "\t", sql
		db_cursor.execute(sql)
		
		sql = "ALTER TABLE  logicmapconnection ADD COLUMN icon TEXT DEFAULT NULL;"
		print "\t", sql
		db_cursor.execute(sql)
  
	db_cursor.execute("SET search_path TO public;")
	set_db_version(5)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_5_6():

	print "Updating schema version from 5 to 6"
	
	db_cursor = connection.cursor()
	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:		
		print "Updating schema %s" % (name)
		db_cursor.execute("SET search_path TO %s;" % (name))
		
		sql = "ALTER TABLE datagroup ALTER COLUMN cdatagroup_id SET NOT NULL;"
		print "\t", sql
		db_cursor.execute(sql)
  
	db_cursor.execute("SET search_path TO public;")
	set_db_version(6)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_6_7():

	print "Updating schema version from 6 to 7"
    
	db_cursor = connection.cursor()
	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:		
		print "Updating schema %s" % (name)
		db_cursor.execute("SET search_path TO %s;" % (name))
		
		sql = 'ALTER TABLE tags_elementtag ADD COLUMN "is_explosive" boolean NOT NULL default false;'
		print "\t", sql
		db_cursor.execute(sql)

		sql = "ALTER TABLE tags_elementtag ADD COLUMN explosive_group integer;"
		print "\t", sql
		db_cursor.execute(sql)
    

	db_cursor.execute("SET search_path TO public;")
	set_db_version(7)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_7_8():

	print "Updating schema version from 6 to 7"
    
	db_cursor = connection.cursor()
	
	sql = 'ALTER TABLE "tenant_user" ADD COLUMN "is_default" boolean NOT NULL default false;';
	print "\t", sql
	db_cursor.execute(sql)

	set_db_version(8)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_8_9():

	print "Updating schema version from 8 to 9"
    
	db_cursor = connection.cursor()
	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:		
		print "Updating schema %s" % (name)
		db_cursor.execute("SET search_path TO %s;" % (name))
		
		sql = 'ALTER TABLE "agent" ADD COLUMN "config_update_minperiod" smallint CHECK ("config_update_minperiod" >= 0) NOT NULL DEFAULT 180;';
		print "\t", sql
		db_cursor.execute(sql)

	db_cursor.execute("SET search_path TO public;")
	set_db_version(9)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#


"""

update dbsetting set value = 9;

SET search_path to ....;
ALTER TABLE prioritylabel drop column mtime;
ALTER TABLE "c_ctimegraphfieldvalue" drop column mtime;
ALTER TABLE "c_ctimegraphseriefieldvalue" drop column mtime;
ALTER TABLE "c_telement" drop column mtime;

SET search_path to ....;
ALTER TABLE prioritylabel drop column mtime;
ALTER TABLE "c_ctimegraphfieldvalue" drop column mtime;
ALTER TABLE "c_ctimegraphseriefieldvalue" drop column mtime;
ALTER TABLE "c_telement" drop column mtime;

SET search_path to ....;
ALTER TABLE prioritylabel drop column mtime;
ALTER TABLE "sanetgroupprofile" drop column mtime;
ALTER TABLE "tenant_group" drop column mtime;
ALTER TABLE "tenant_user" drop column mtime;
ALTER TABLE "c_ctimegraphfieldvalue" drop column mtime;
ALTER TABLE "c_ctimegraphseriefieldvalue" drop column mtime;
ALTER TABLE "c_telement" drop column mtime;
"""

@transaction_commit_on_success
def update_schema_9_10():
	
	print "Updating schema version from 9 to 10"

		
	db_cursor = connection.cursor()
	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:
		print "Updating prioritylabel table for tenant:", name
		db_cursor.execute("SET search_path TO %s;" % (name))

		TABLES = [ 
			"prioritylabel",
			"c_ctimegraphfieldvalue",
			"c_ctimegraphseriefieldvalue",
		]

		for table_name in TABLES:
			sql = 'ALTER TABLE "%s" ADD COLUMN "mtime" INTEGER NOT NULL DEFAULT 0;' % (table_name)
			print "\t", sql
			db_cursor.execute( sql )


	db_cursor.execute("SET search_path TO public;")
	
	TABLES = [ 
		"sanetgroupprofile", 
		"tenant_group", 
		"tenant_user",
		"c_telement",
	]

	for table_name in TABLES:

		sql = 'ALTER TABLE "%s" ADD COLUMN "mtime" INTEGER NOT NULL DEFAULT 0;' % (table_name)
		print "\t", sql
		db_cursor.execute( sql )

	set_db_version(10)

	return



#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_10_11():
	
	print "Updating schema version from 10 to 11"
	
	_change_mtime_field()#skip_tables=['prioritylabel'])
	
	set_db_version(11)

def _change_mtime_field(): #skip_tables):
	
	from django.db import models
	from core.models import TenantDependentModel, MTimeModel

	#
	# Change mtime type for all models
	#
	db_cursor = connection.cursor()
	
	# Get informations
	tables = connection.introspection.table_names()
	seen_models = connection.introspection.installed_models(tables)

	def __model_installed(model):
		opts = model._meta
		converter = connection.introspection.table_name_converter
		return ((converter(opts.db_table) in tables) or (opts.auto_created and converter(opts.auto_created._meta.db_table) in tables))
	
	# Get all tenant dependent models for each app
	all_models = []
	tenant_models = []
	
	for app in models.get_apps(): 
		app_name = app.__name__.split('.')[-2]
		
		for m in models.get_models(app, include_auto_created=True):
			
			#if m._meta.db_table in skip_tables:
			#	continue

			if 'mtime' not in [ f.name for f in m._meta.local_fields]:
				continue

			if not (issubclass(m, MTimeModel)):
				continue
				
			all_models.append(m)	
			
			if (issubclass(m, TenantDependentModel)):
				tenant_models.append(m)
				
	all_models    = filter(__model_installed, all_models)
	tenant_models = filter(__model_installed, tenant_models)
	
	db_cursor.execute("SET search_path TO public;")
	for m in all_models:
		q = 'ALTER TABLE "%s" ALTER COLUMN mtime TYPE INTEGER;' % (m._meta.db_table)
		print "\t", q
		db_cursor.execute(q)
		
	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:
		db_cursor.execute("SET search_path TO %s;" % (name))
		
		for m in tenant_models:

			if 'mtime' not in [ f.name for f in m._meta.local_fields]:
				continue

			q = 'ALTER TABLE "%s" ALTER COLUMN mtime TYPE INTEGER;' % (m._meta.db_table)
			print "\t", q
			db_cursor.execute(q)
		
		db_cursor.execute("SET search_path TO public;")
			


#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_11_12():

	print "Updating schema version from 11 to 12"

	db_cursor = connection.cursor()

	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:

		db_cursor.execute("SET search_path TO %s;" % (name))
		
		_change_minperiod_type(db_cursor)

		_add_statuslastnotablechange(db_cursor)

	db_cursor.execute("SET search_path TO public;")
	set_db_version(12)
	
def _change_minperiod_type(db_cursor):

	
	query = 'ALTER TABLE datagroup ALTER COLUMN minperiod TYPE INTEGER;'
	print query
	db_cursor.execute(query)

def _add_statuslastnotablechange(db_cursor):
	
	query = 'ALTER TABLE condition ADD COLUMN "statuslastnotablechange" integer CHECK ("statuslastnotablechange" >= 0) NOT NULL DEFAULT 0;'
	print query
	db_cursor.execute(query)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_12_13():

	print "Updating schema version from 12 to 13"

	db_cursor = connection.cursor()

	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:
		print "SCHEMA: ", name
		db_cursor.execute("SET search_path TO %s;" % (name))
#		_fix_tags_tables(db_cursor)

	db_cursor.execute("SET search_path TO public;")
	set_db_version(13)


def _fix_tags_tables(db_cursor):
	
	query = "ALTER TABLE  tagtree DROP COLUMN cached;"
	print query
	db_cursor.execute(query)
	
	query = "ALTER TABLE  tags_elementtag DROP COLUMN path;"
	print query
	db_cursor.execute(query)

	query = "ALTER TABLE  tags_elementtag DROP COLUMN state;"
	print query
	db_cursor.execute(query)

	query = "ALTER TABLE  tagentry DROP COLUMN state;"
	print query
	db_cursor.execute(query)

	query = "ALTER TABLE tags_elementtag ADD COLUMN directly_tagged BOOLEAN NOT NULL DEFAULT true;"
	print query
	db_cursor.execute(query)


#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_13_14():

	print "Updating schema version from 13 to 14"

	db_cursor = connection.cursor()

	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:
		print "SCHEMA: ", name
		db_cursor.execute("SET search_path TO %s;" % (name))

		query = 'ALTER TABLE logicmap       ADD COLUMN "tagref_id"     varchar(32) REFERENCES "tagentry"        ("uuid") DEFERRABLE INITIALLY DEFERRED DEFAULT NULL;'
		db_cursor.execute(query) 

		query = 'ALTER TABLE logicmapentity ADD COLUMN "tagref_id"     varchar(32) REFERENCES "tagentry"        ("uuid") DEFERRABLE INITIALLY DEFERRED DEFAULT NULL;'
		db_cursor.execute(query) 
		query = 'ALTER TABLE logicmapentity ADD COLUMN "elementtag_id" varchar(32) REFERENCES "tags_elementtag" ("uuid") DEFERRABLE INITIALLY DEFERRED DEFAULT NULL;'
		db_cursor.execute(query) 
		
		query = 'CREATE UNIQUE INDEX "logicmapentity_tenant_id_logicmap_id_parent_id_tagref_id_key"    ON logicmapentity (tenant_id, logicmap_id, parent_id, tagref_id);'
		db_cursor.execute(query) 
		query = 'CREATE UNIQUE INDEX "logicmapentity_tenant_id_logicmap_id_parent_id_elementag_id_key" ON logicmapentity (tenant_id, logicmap_id, parent_id, elementtag_id);'
		db_cursor.execute(query) 
		

	db_cursor.execute("SET search_path TO public;")
	set_db_version(14)


#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#


@transaction_commit_on_success
def update_schema_14_15():

	from tags.models import TagTree, Tag, ElementTag

	print "Updating schema version from 14 to 15"

	db_cursor = connection.cursor()

	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:
		print "SCHEMA: ", name
		db_cursor.execute("SET search_path TO %s;" % (name))

		query = 'ALTER TABLE tags_elementtag ADD COLUMN "tree_id" varchar(32);'
		print query
		db_cursor.execute(query)

		query = 'ALTER TABLE tags_elementtag ADD CONSTRAINT "tags_elementtag_tree_id_fkey" FOREIGN KEY (tree_id) REFERENCES tagtree(uuid) DEFERRABLE INITIALLY DEFERRED'
		print query
		db_cursor.execute(query)
		
		query = 'CREATE INDEX "tags_elementtag_tree_id" ON tags_elementtag (tree_id);';
		print query
		db_cursor.execute(query)

		query = 'ALTER TABLE tags_elementtag ADD COLUMN  "unique_sentinel" varchar(32) NOT NULL DEFAULT md5(random()::text);'
		print query
		db_cursor.execute(query)

		#for tree in TagTree.objects.all():
		#	for et in ElementTag.objects.select_related('tag', 'tag__tree').all():
		#		et.tree = et.tag.tree
		#		et.save()		
		#for tree in TagTree.objects.all():
		#	for tag in tree.tags.all():
		#		print tag.path
		#		count = ElementTag.objects.filter(tag=tag).update(tree=tag.tree)
		#		print "\t", count				
				

	db_cursor.execute("SET search_path TO public;")
	set_db_version(15)

@transaction_commit_on_success
def update_schema_15_16():
	
	#from tags.models import *

	print "Updating schema version from 15 to 16"

	db_cursor = connection.cursor()

	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:
		print "SCHEMA: ", name
		db_cursor.execute("SET search_path TO %s;" % (name))

		query = 'ALTER TABLE tags_elementtag ALTER COLUMN "tree_id" SET NOT NULL;'
		print query
		db_cursor.execute(query)
		
		
		query = 'CREATE UNIQUE INDEX "tags_elementtag_tenant_id_tree_id_elementtype_elementid_unique_sentinel_key" ON tags_elementtag (tenant_id, tree_id, elementtype, elementid, unique_sentinel);';
		print query
		db_cursor.execute(query)

	db_cursor.execute("SET search_path TO public;")
	set_db_version(16)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_16_17():
	
	#from tags.models import *

	print "Updating schema version from 16 to 17"

	db_cursor = connection.cursor()

	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:
		print "SCHEMA: ", name
		db_cursor.execute("SET search_path TO %s;" % (name))

		query = 'ALTER TABLE tagtree ADD COLUMN "nostate" BOOLEAN NOT NULL DEFAULT false;'
		print query
		db_cursor.execute(query)

	db_cursor.execute("SET search_path TO public;")
	set_db_version(17)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_17_18():
	
	print "Updating schema version from 17 to 18"

	db_cursor = connection.cursor()

	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:
		print "SCHEMA: ", name
		db_cursor.execute("SET search_path TO %s, public;" % (name))
		
		
		query = 'ALTER TABLE logicmap ADD COLUMN    owner_id   integer  REFERENCES auth_user ("id") DEFERRABLE INITIALLY DEFERRED;'
		print query
		db_cursor.execute(query)

		query = 'ALTER TABLE logicmap ADD COLUMN    group_perm smallint CHECK ("group_perm" >= 0) NOT NULL DEFAULT 0;'
		print query
		db_cursor.execute(query)

	db_cursor.execute("SET search_path TO public;")
	set_db_version(18)


	from webui.capabilities import CAPABILITIES
	from webui.models import Capability

	for t in tenant_select_all():
		for g in Group.objects.all():
			for capability_id in CAPABILITIES.keys():
				c, created = Capability.objects.get_or_create(tenant=t, group=g, code=capability_id)	
				if created:
					print c
			

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_18_19():
	
	print "Updating schema version from 17 to 18"

	db_cursor = connection.cursor()

	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:
		print "SCHEMA: ", name
		db_cursor.execute("SET search_path TO %s, public;" % (name))
		
		query = 'ALTER TABLE iface ADD COLUMN "backbone" BOOLEAN NOT NULL DEFAULT false';
		print query
		db_cursor.execute(query)
	
	db_cursor.execute("SET search_path TO public;")
	set_db_version(19)


#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_19_20():
	
	print "Updating schema version from 19 to 20"

	db_cursor = connection.cursor()

	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:

		print "SCHEMA: ", name
		db_cursor.execute("SET search_path TO %s, public;" % (name))
		
		query = 'ALTER TABLE datagroup ADD COLUMN "passive_match" text NOT NULL DEFAULT \'\';'
		print query
		db_cursor.execute(query)
		
		query = 'CREATE INDEX "datagroup_passive_match_key" ON datagroup (passive_match);';
		print query
		db_cursor.execute(query)
	
	db_cursor.execute("SET search_path TO public;")
	set_db_version(20)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_20_21():
	
	print "Updating schema version from 20 to 21"

	db_cursor = connection.cursor()

	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:
		
		print "SCHEMA: ", name
		db_cursor.execute("SET search_path TO %s, public;" % (name))
		
		query = 'ALTER TABLE node ADD COLUMN "effective_ip" text DEFAULT NULL;'
		print query
		db_cursor.execute(query)
	
	db_cursor.execute("SET search_path TO public;")
	set_db_version(21)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_21_22():
	
	print "Updating schema version from 21 to 22"

	db_cursor = connection.cursor()

	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:
		
		print "SCHEMA: ", name
		db_cursor.execute("SET search_path TO %s, public;" % (name))
		
		query = 'ALTER TABLE monitoredelement ADD COLUMN "notify_address" text DEFAULT NULL;'
		print query
		db_cursor.execute(query)
	
	db_cursor.execute("SET search_path TO public;")
	set_db_version(22)


#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_22_23():
	
	print "Updating schema version from 22 to 23"

	db_cursor = connection.cursor()

	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:
		
		print "SCHEMA: ", name
		db_cursor.execute("SET search_path TO %s, public;" % (name))
		
		query = 'ALTER TABLE tagtree ADD COLUMN nomap boolean NOT NULL DEFAULT false;'
		print query
		db_cursor.execute(query)
	
	db_cursor.execute("SET search_path TO public;")
	set_db_version(23)
	
#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_23_24():
	
	print "Updating schema version from 23 to 24"

	db_cursor = connection.cursor()

	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:
		
		print "SCHEMA: ", name
		db_cursor.execute("SET search_path TO %s, public;" % (name))
		
		query = 'ALTER TABLE geo_elementpos ADD COLUMN zoom double precision NOT NULL default 6;'
		print query
		db_cursor.execute(query)
	
		query = 'ALTER TABLE geo_tagpos ADD COLUMN zoom double precision NOT NULL default 6;'
		print query
		db_cursor.execute(query)
	
	db_cursor.execute("SET search_path TO public;")
	set_db_version(24)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_24_25():
	
	print "Updating schema version from 24 to 25"

	db_cursor = connection.cursor()

	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:
		
		print "SCHEMA: ", name
		db_cursor.execute("SET search_path TO %s, public;" % (name))
		
		query = 'ALTER TABLE logicmap ADD COLUMN zoom double precision NOT NULL DEFAULT 1'
		print query
		db_cursor.execute(query)
	
	db_cursor.execute("SET search_path TO public;")
	set_db_version(25)


#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_25_26():
	
	print "Updating schema version from 25 to 26"
	
	if 'webtests' in settings.INSTALLED_APPS:
		
		db_cursor = connection.cursor()
	
		schemas = ['public',] + [ t.name for t in tenant_select_all() ]
		for name in schemas:
			
			print "SCHEMA: ", name
			db_cursor.execute("SET search_path TO %s, public;" % (name))
			
			query = 'ALTER TABLE webtests_webscenario ADD COLUMN store_interval integer NOT NULL DEFAULT 60;'
			print query
			db_cursor.execute(query)
		
		db_cursor.execute("SET search_path TO public;")
		
	else:
		
		print 'Webscenarios disabled. Skipped updating schema versione from 25 to 26.'
		
	set_db_version(26)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_26_27():

	print "Updating schema version from 26 to 27"

	db_cursor = connection.cursor()

	db_cursor.execute("SET search_path TO public;")

	print "Updating reporttype table"

	query =  'ALTER TABLE reporttype ADD COLUMN version integer CHECK ("version" >= 0) NOT NULL DEFAULT 1'

	print query
	db_cursor.execute(query)

	db_cursor.execute("SET search_path TO public;")

	set_db_version(27)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_27_28():

	print "Updating schema version from 27 to 28"

	db_cursor = connection.cursor()

	schemas = ['public',]
	for name in schemas:

		queries = [
			"ALTER TABLE c_telement ADD COLUMN parent_id     integer DEFAULT NULL;",
			"ALTER TABLE c_telement ADD COLUMN distinguisher text    DEFAULT NULL;",
			"ALTER TABLE c_telement ADD COLUMN xform         text    DEFAULT NULL;",
			"ALTER TABLE c_telement ADD COLUMN instance      text    DEFAULT NULL;",
		]
		for q in queries:
			print q
			db_cursor.execute(q)

		query = """ALTER TABLE "c_telement" ADD CONSTRAINT "parent_id_refs_id" FOREIGN KEY ("parent_id") REFERENCES "c_telement" ("id") DEFERRABLE INITIALLY DEFERRED;"""
		print query
		db_cursor.execute(query)


	schemas = [ t.name for t in tenant_select_all() ]
	for name in schemas:

		print "SCHEMA: ", name
		db_cursor.execute("SET search_path TO %s, public" % (name))

		query = """ALTER TABLE monitoredelement ADD COLUMN telement_generator_id integer REFERENCES "c_telement" ("id") DEFERRABLE INITIALLY DEFERRED"""
		print query
		db_cursor.execute(query)

		query = """ALTER TABLE c_element_template ADD COLUMN applyorder integer NOT NULL DEFAULT 0;"""
		print query
		db_cursor.execute(query)

		query = """ALTER TABLE c_element_template ADD COLUMN telementappliedtemplate_id integer REFERENCES c_telementappliedtemplate ("id") DEFERRABLE INITIALLY DEFERRED"""
		print query
		db_cursor.execute(query)


	db_cursor.execute("SET search_path TO public;")

	set_db_version(28)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_28_29():

	print "Updating schema version from 28 to 29"
	
	if 'webtests' in settings.INSTALLED_APPS and table_exists('webtests_webscenario') and table_exists('webtests_steplog'):
		
		db_cursor = connection.cursor()
		
		schemas = ['public',] + [ t.name for t in tenant_select_all() ]
		for name in schemas:
			
			print "SCHEMA: ", name
			db_cursor.execute("SET search_path TO %s, public;" % (name))

			query =  'ALTER TABLE webtests_webscenario ADD COLUMN ssl_version text;'
			print query
			db_cursor.execute(query)
			
			query =  'ALTER TABLE webtests_webscenario ADD COLUMN save_step_log integer NOT NULL DEFAULT 0;'
			print query
			db_cursor.execute(query)

		db_cursor.execute("SET search_path TO public;")
	
	else:
		
		print 'Webscenarios disabled. Skipped updating schema versione from 28 to 29.'
		
	set_db_version(29)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_29_30():
	
	print "Updating schema version from 29 to 30: fix"
	
	
	if not column_exists('monitoredelement', 'telement_generator_id', 'public'):
	
		db_cursor = connection.cursor()
		
		db_cursor.execute("SET search_path TO public")
		
		query = """ALTER TABLE monitoredelement ADD COLUMN telement_generator_id integer REFERENCES "c_telement" ("id") DEFERRABLE INITIALLY DEFERRED"""
		print query
		db_cursor.execute(query)
	
	set_db_version(30)


#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_30_31():
	
	print "Updating schema version from 29 to 30"
	
	db_cursor = connection.cursor()
	
	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:
		
		print "SCHEMA: ", name
		db_cursor.execute("SET search_path TO %s, public;" % (name))
		
		ENTITIES = [
			('iface'            , 'ifindex'   ),
			('storage'          , 'stindex'   ),
			('service'          , 'swrunindex'),

			('discoverediface'  , 'ifindex'   ),
			('discoveredstorage', 'stindex'   ),
			('discoveredservice', 'swrunindex'),
			('discovereddevice' , 'devindex'  ),
		]
		
		QUERIES = [
			'alter table %(table)s add column    %(field)s_2 text;',
			'update      %(table)s set           %(field)s_2 = cast( %(field)s as text) ;',
			'alter table %(table)s drop column   %(field)s;',
			'alter table %(table)s rename column %(field)s_2 to %(field)s;',
		]
		
		for table, field in ENTITIES:
			
			for q in QUERIES:
				q = q % {'table': table, 'field': field}
				print q
				db_cursor.execute(q)
	
	db_cursor.execute("SET search_path TO public;")
	set_db_version(31)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_31_32():

	print "Updating schema version from 31 to 32"

	db_cursor = connection.cursor()

	schemas = ['public',] + [ t.name for t in tenant_select_all() ]
	for name in schemas:
		
		print "SCHEMA: ", name
		db_cursor.execute("SET search_path TO %s;" % (name))

		if not column_exists('tags_elementtag', 'tagged_parent', schema_name=name):
			query = 'ALTER TABLE tags_elementtag ADD COLUMN "tagged_parent" VARCHAR(32);'
			print query
			db_cursor.execute(query)

		if not column_exists('tagtree', 'stime', schema_name=name):
			query = 'ALTER TABLE tagtree ADD COLUMN "stime" double precision NOT NULL DEFAULT 0;'
			print query
			db_cursor.execute(query)

	db_cursor.execute("SET search_path TO public;")
	set_db_version(32)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_32_33():
	print "Updating schema version from 32 to 33"

	db_cursor = connection.cursor()


	schemas = ['public',]
	for name in schemas:

		print "SCHEMA: ", name

		for table_name in [ 'c_telement', 'c_tdatagroup']:

			if not column_exists(table_name, 'owner_id', schema_name=name):

				db_cursor.execute("SET search_path TO %s;" % (name))

				query = """ ALTER TABLE "%s"  ADD COLUMN "owner_id" integer REFERENCES "auth_user" ("id") DEFAULT NULL;""" % (table_name)
				print query
				db_cursor.execute(query)

				query = """ CREATE INDEX "%(table_name)s_owner_id"  ON "%(table_name)s" ("owner_id"); """ % { "table_name": table_name}
				print query
				db_cursor.execute(query)

	schemas = [ 'public' ] + [ t.name for t in tenant_select_all() ]
	for name in schemas:

		print "SCHEMA: ", name

		for table_name in [ 'monitoredelement', 'datagroup', 'datasource', 'condition', 'timegraph', 'timegraphserie' ]:

			if not column_exists(table_name, 'owner_id', schema_name=name):

				db_cursor.execute("SET search_path TO %s,public;" % (name))

				query = """ ALTER TABLE "%s"  ADD COLUMN "owner_id" integer REFERENCES "auth_user" ("id") DEFAULT NULL;""" % (table_name)
				print query
				db_cursor.execute(query)

				query = """ CREATE INDEX "%(table_name)s_owner_id"  ON "%(table_name)s" ("owner_id"); """ % { "table_name": table_name}
				print query
				db_cursor.execute(query)

	db_cursor.execute("SET search_path TO public;")
	set_db_version(33)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_33_34():
	print "Updating schema version from 33 to 34"

	db_cursor = connection.cursor()

	schemas = [ 'public' ] + [ t.name for t in tenant_select_all() ]
	for name in schemas:

		print "SCHEMA: ", name
		db_cursor.execute("SET search_path TO %s,public;" % (name))

		query = '''ALTER TABLE webtests_step            ADD COLUMN "raw_post_data"      text NOT NULL DEFAULT '';'''
		print "\t", query
		db_cursor.execute(query)

	set_db_version(34)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_34_35():
	print "Updating schema version from 34 to 35"
	
	db_cursor = connection.cursor()
	
	schemas = [ 'public' ] + [ t.name for t in tenant_select_all() ]
	for name in schemas:
		
		print "SCHEMA: ", name
		db_cursor.execute("SET search_path TO %s,public;" % (name))
		
		query = 'ALTER TABLE agent            ADD COLUMN "scheduling_min_gap"      smallint CHECK ("scheduling_min_gap" >= 0)      DEFAULT NULL'
		print "\t", query
		db_cursor.execute(query)
		
		query = 'ALTER TABLE agent            ADD COLUMN "scheduling_max_parallel" smallint CHECK ("scheduling_max_parallel" >= 0) DEFAULT NULL'
		print "\t", query
		db_cursor.execute(query)
		
		query = 'ALTER TABLE agent            ADD COLUMN "scheduling_max_deadline" smallint CHECK ("scheduling_max_deadline" >= 0) DEFAULT NULL'
		print "\t", query
		db_cursor.execute(query)
		
		query = 'ALTER TABLE cluster          ADD COLUMN "owner_id"                integer REFERENCES "auth_user" ("id") DEFERRABLE INITIALLY DEFERRED DEFAULT NULL'
		print "\t", query
		db_cursor.execute(query)
		
		query = "ALTER TABLE cluster          ADD COLUMN scheduling_group_id       varchar(32) REFERENCES scheduling_group (uuid) ON DELETE SET NULL"
		print "\t", query
		db_cursor.execute(query)
		
		query = "ALTER TABLE monitoredelement ADD COLUMN scheduling_group_id       varchar(32) REFERENCES scheduling_group (uuid) ON DELETE SET NULL"
		print "\t", query
		db_cursor.execute(query)
		
		query = "ALTER TABLE datagroup        ADD COLUMN scheduling_group_id       varchar(32) REFERENCES scheduling_group (uuid) ON DELETE SET NULL"
		print "\t", query
		db_cursor.execute(query)
		
	set_db_version(35)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_35_36():
	print "Updating schema version from 35 to 36"
	
	db_cursor = connection.cursor()
	
	schemas = [ 'public' ] + [ t.name for t in tenant_select_all() ]
	for name in schemas:

		print "SCHEMA: ", name
		db_cursor.execute("SET search_path TO %s,public;" % (name))
		for table_name in [ 'cluster', ]:

			if not column_exists(table_name, 'owner_id', schema_name=name):

				query = """ ALTER TABLE "%s"  ADD COLUMN "owner_id" integer REFERENCES "auth_user" ("id") DEFAULT NULL;""" % (table_name)
				print query
				db_cursor.execute(query)

				query = """ CREATE INDEX "%(table_name)s_owner_id"  ON "%(table_name)s" ("owner_id"); """ % { "table_name": table_name}
				print query
				db_cursor.execute(query)

	db_cursor.execute("SET search_path TO public;")
	set_db_version(36)


#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_36_37():
	print "Updating schema version from 36 to 37"

	db_cursor = connection.cursor()

	schemas = [ 'public' ] + [ t.name for t in tenant_select_all() ]
	for name in schemas:

		print "SCHEMA: ", name
		db_cursor.execute("SET search_path TO %s,public;" % (name))

		query = "ALTER TABLE customfield_customfieldvalue ADD COLUMN custom_instance_field_value_id varchar(32) REFERENCES customfield_custominstancefieldvalue (uuid) ON DELETE CASCADE DEFAULT NULL"
		print query
		db_cursor.execute(query)

		query = "ALTER TABLE customfield_customfieldvalue ADD COLUMN element_to_custom_group_instance_id varchar (32) REFERENCES customfield_elementcustomgroupinstance (uuid) ON DELETE CASCADE DEFAULT NULL"
		print query
		db_cursor.execute(query)

		query = "ALTER TABLE customfield_customfieldvalue ADD COLUMN personalized BOOLEAN DEFAULT FALSE"
		print query
		db_cursor.execute(query)

	set_db_version(37)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#



@transaction_commit_on_success
def update_schema_37_38():
	print "Updating schema version from 36 to 37"

	db_cursor = connection.cursor()

	schemas = [ 'public' ] + [ t.name for t in tenant_select_all() ]
	for name in schemas:

		print "SCHEMA: ", name
		db_cursor.execute("SET search_path TO %s,public;" % (name.lower()))

		print "Dropping old index"
		db_cursor.execute('''ALTER TABLE customfield_customgroupinstance DROP CONSTRAINT IF EXISTS customfield_customgroupinstance_name_key;''')

		if not index_exists(name.lower(), 'customfield_customgroupinstance', 'customfield_customgroupinstan_tenant_id_custom_group_id_nam_key'):
			print "Creating new index"
			db_cursor.execute('''CREATE UNIQUE INDEX customfield_customgroupinstan_tenant_id_custom_group_id_nam_key ON %s.customfield_customgroupinstance (tenant_id, custom_group_id, name);''' % (name))

	set_db_version(38)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_38_39():

	print "Updating schema version from 38 to 39"

	db_cursor = connection.cursor()

	schemas = [ 'public' ] + [ t.name for t in tenant_select_all() ]
	for name in schemas:

		print "SCHEMA: ", name
		db_cursor.execute("SET search_path TO %s,public;" % (name))

		query = "ALTER TABLE logicmap ADD COLUMN background_img   text;"
		db_cursor.execute(query)

		query = "ALTER TABLE logicmap ADD COLUMN background_color text;"
		db_cursor.execute(query)

	set_db_version(39)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_39_40():

	print "Updating schema version from 39 to 40"

	db_cursor = connection.cursor()

	schemas = [ 'public' ] + [ t.name for t in tenant_select_all() ]
	for name in schemas:

		print "SCHEMA: ", name
		db_cursor.execute("SET search_path TO %s,public;" % (name))

		query = '''ALTER TABLE iface ADD COLUMN  "speed" bigint;'''
		db_cursor.execute(query)

	set_db_version(40)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_40_41():

	print "Updating schema version from 40 to 41"

	db_cursor = connection.cursor()

	schemas = [ 'public' ] + [ t.name for t in tenant_select_all() ]
	for name in schemas:

		db_cursor.execute("SET search_path TO %s,public;" % (name))
		
		db_cursor.execute('ALTER TABLE monitoredelement ADD COLUMN  "ctime" integer NOT NULL DEFAULT 0;')

	set_db_version(41)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_41_42():

	print "Updating schema version from 41 to 42"

	db_cursor = connection.cursor()

	schemas = [ 'public' ] + [ t.name for t in tenant_select_all() ]
	for name in schemas:

		db_cursor.execute("SET search_path TO %s,public;" % (name))

		db_cursor.execute('ALTER TABLE "reportparam"         ALTER COLUMN value TYPE TEXT;')
		db_cursor.execute('ALTER TABLE "periodicreportparam" ALTER COLUMN value TYPE TEXT;')

	set_db_version(42)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_42_43():

	print "Updating schema version from 42 to 43"

	db_cursor = connection.cursor()

	schemas = [ 'public' ] + [ t.name for t in tenant_select_all() ]
	for name in schemas:

		db_cursor.execute("SET search_path TO %s,public;" % (name))

		db_cursor.execute('ALTER TABLE logicmap DROP COLUMN group_perm ;')

	set_db_version(43)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_43_44():

	print "Updating schema version from 43 to 44"

	db_cursor = connection.cursor()

	schemas = [ 'public' ] + [ t.name for t in tenant_select_all() ]
	for name in schemas:

		db_cursor.execute("SET search_path TO %s,public;" % (name))

		db_cursor.execute('ALTER TABLE event ADD COLUMN "element_uuid"   varchar(32);')
		db_cursor.execute('ALTER TABLE event ADD COLUMN "condition_uuid" varchar(32);')

	set_db_version(44)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def update_schema_44_45():

	# Schema version 45 is the last schema before django 1.11
	_signal_schema_update(44)

	set_db_version(45)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

#@transaction_commit_on_success
#def update_schema_45_46():
#
#	# Schema version 45 is the last schema before django 1.11
#	_signal_schema_update(45)
#
#	print "Updating schema version from 45 to 46"
#
#	db_cursor = connection.cursor()
#
#	schemas = [ t.name for t in tenant_select_all() ]
#	for name in schemas:
#
#		db_cursor.execute("SET search_path TO %s" % (name))
#
#		db_cursor.execute('ALTER TABLE "condition" ADD COLUMN "cascade" boolean  DEFAULT false NOT NULL')
#		db_cursor.execute('ALTER TABLE "condition" ADD COLUMN "order"   smallint DEFAULT 0     NOT NULL CHECK ("order" >= 0)')
#
#	set_db_version(46)


#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

def _check_every_tenant(logger):
	for t in tenant_select_all():
		_t_check_single_tenant(logger, t)

@transaction_commit_on_success
def _t_check_single_tenant(logger, tenant):
	check_indexes_on_tables(tenant, logger=logger)

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#


class Command(SanetBaseCommand):

	help = _("Check DB schema version and perform updates")

	sanet_internal = True

	#option_list = SanetBaseCommand.option_list + (
	#	make_option('--adminpass',
	#		action='store',
	#		dest='adminpass',
	#		type='string',
	#		default=None,
	#		help='Default admin pass'),
	#	make_option('--defaulttenant-name',
	#		action='store',
	#		type='string',
	#		dest='defaulttenantname',
	#		default=None,
	#		help='Default Tenant Name'),
	#	make_option('--defaulttenant-longname',
	#		action='store',
	#		type='string',
	#		dest='defaulttenantlongname',
	#		default=None,
	#		help='Default Tenant Long Name'),
	#	make_option('--defaulttenant-timezone',
	#		action='store',
	#		type='string',
	#		dest='defaulttenanttimezone',
	#		default=None,
	#		help='Default Tenant Timezone'),
	#    
	#)


	def add_arguments(self, parser):
		super(Command, self).add_arguments(parser)

		parser.add_argument('--adminpass'             , action='store', type=str, dest='adminpass'            , default=None, help='Default admin pass'),
		parser.add_argument('--defaulttenant-name'    , action='store', type=str, dest='defaulttenantname'    , default=None, help='Default Tenant Name'),
		parser.add_argument('--defaulttenant-longname', action='store', type=str, dest='defaulttenantlongname', default=None, help='Default Tenant Long Name'),
		parser.add_argument('--defaulttenant-timezone', action='store', type=str, dest='defaulttenanttimezone', default=None, help='Default Tenant Timezone'),


	def handle(self, *args, **options):

		logger = logging.getLogger('sanet.global')

		check_base_dirs(logger)

		#
		# Sync missing tables
		#
		from django.core.management import call_command
		#call_command('syncdb', interactive=False, verbosity=True, show_traceback=True)


		#for django_app_name in settings.INSTALLED_APPS:
		#	try:
		#		print django_app_name
		#		module_fullname = '%s.migrations' % (django_app_name)
		#		mod = __import__(module_fullname, {}, {}, [])
		#		has_migrations = True
		#	except ImportError, e:
		#		print type(e), str(e)
		#		has_migrations = False
		#	django_app_name = django_app_name.rsplit('.')[-1]
		#	if has_migrations:
		#		print django_app_name
		#		call_command('migrate', django_app_name, 'initial', interactive=False, verbosity=True, show_traceback=True)
		#return
		

		logger.debug("#")
		logger.debug("# Migrate")
		logger.debug("#")

		call_command('migrate', interactive=False, verbosity=True, show_traceback=True, fake_initial=1)

		logger.debug("#")
		logger.debug("# Loading data")
		logger.debug("#")
		#call_command('loaddata', 'dbsetting', app='core', interactive=False, verbosity=True, show_traceback=True)
		
		db_version, created = DBSetting.objects.get_or_create(name='version')
		if created:
			db_version.set_value(0)
			db_version.save()
		
		

		logger.debug("#")
		logger.debug("# Syncing DB")
		logger.debug("#")
		from django.core.management import call_command
		call_command('syncdbtenants', interactive=False, verbosity=True, show_traceback=True)

		#
		# Close pending connections
		#
		connection.close()

		#
		# Check consistency
		#
		check_and_update_schema(logger)

		_check_every_tenant(logger)

		return




