# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.
"""

clean_rrds
========================================================================

Syntax: 

	::
		
		sanet-mange clean_rrds [options]

Removes orphans RRD files.

A RRD file is orphan when :

	* it's related to a deleted datasource.
	* it's related to a datasource with the *save-value* flag set to *False*.

Options:
	* --print-only : Does not perform any action.
	* --remove-no-save-value : Removes files linked to datasources with flag "save-value" set to *False*.

"""

import os
import sys
import getpass

from django.utils.translation import ugettext as _

from conf import get_configuration

from core.models import DataSource

from core.api.datasource import init_datasource_store
from core.api.datasource.rrd import BaseRRDTOOLDataSourceStore

from core.management.commands import TenantBasedCommand

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

class Command(TenantBasedCommand):

	help = "Print old/unused datasources files (RRD files)"

	#option_list = TenantBasedCommand.option_list + (
	#	make_option('-l', '--log-level'            , action='store'     , type='string', dest='log_level'            , default=20   , help='Log Level'),
	#	make_option(      '--print-only'           , action='store_true'               , dest='print_only'           , default=False, help='Do not perform any action.'),
	#	make_option(      '--remove-no-save-value' , action='store_true'               , dest='remove_no_save_value' , default=False, help='Remove RRDs related to datasources with "save-value" set to False.'),
	#)
	
	def add_arguments(self, parser):
		super(Command, self).add_arguments(parser)

		parser.add_argument('-l', '--log-level'            , action='store'     , dest='log_level'            , default=20   , help='Log Level'),
		parser.add_argument(      '--print-only'           , action='store_true', dest='print_only'           , default=False, help='Do not perform any action.'),
		parser.add_argument(      '--remove-no-save-value' , action='store_true', dest='remove_no_save_value' , default=False, help='Remove RRDs related to datasources with "save-value" set to False.'),


	def handle(self, *args, **options):

		tenant = self.get_tenant(options)
		tenant.db_set_current_schema()

		logger = self.init_stdout_log(int(options['log_level']))

		logger.info("Cleaning old rrds for tenant %s" % (tenant))

		conf = get_configuration()
		datasource_store = init_datasource_store(conf.DATASOURCE_STORAGE_HANDLER, conf.DATASOURCE_STORAGE_PARAMS)

		if not isinstance(datasource_store, BaseRRDTOOLDataSourceStore):
			logger.info("Datasource store backend does not store any RRD: %s", conf.DATASOURCE_STORAGE_HANDLER)
			return

		#
		# Select valid rrds
		#
		datasources = DataSource.objects

		if options['remove_no_save_value']:
			datasources = datasources.exclude(save_value=False)

		valid_ids = set([r[0] for r in datasources.all().values_list('uuid')])

		#
		# Scan dir
		#

		base_path = datasource_store._get_tenant_dir(tenant.uuid)

		for fname in os.listdir(base_path):

			name = fname.replace('.rrd','')

			if name not in valid_ids:

				fpath = os.path.join(base_path, fname)

				try:
					if not options['print_only']:
						os.remove(fpath)

					logger.info( "\t" + "\t" + "removed %s", fpath)
				except OSError, e:
					logger.info( "\t" + "\t" + "error   %s -> %s", fpath, str(e))
