# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

exec_expr
=============================================

Launch: 

	::
		
		cd $SANET_INSTALL_DIR
		python manage.py exec_expr [options] <node>[:<subentity>] [ <expression> ]

The program will evaluate a generic *expression* using informations about the specified *node* / *subentity* (interface, storage, device, service).

If *node* (or the *subentity*) are present inside SANET's configuration, the programm will load all default *node*/*subentity* variables .

Options:

	::
		
		--stdin           Additional parameters (and the expression) will be read line by line from standard input. The parameter expression can be read from standard input too. See later in this section.
		
		--tenant <name>   The expression will be evaluated with data taken from the selected tenant instead of from the primary tenant. 


Examples
--------

Generic expression on a generic host not defined inside SANET: 

	::
		
		# python manage.py exec_expr localhost '1+1'

Output:

	::
		
		2013-07-25 18:53:46,638 DEBUG MainThread > Loading tenant environment
		2013-07-25 18:53:46,654 DEBUG MainThread > Using default tenant: <Tenant:10a279a6ce5c4438a98f9c33c0b77cde:site>
		...
		...
		...
		2013-07-25 18:53:46,790 DEBUG MainThread > evaluated()
		VALUE: 2
		INFOS:


Retrieve SysName from node mason defined inside Sanet: 

	::
		
		# python manage.py exec_expr mason '1.3.6.1.2.1.1.1.0@'

Output:

	::
		
		2013-07-25 18:55:31,989 DEBUG MainThread > Loading tenant environment
		...
		...
		...
		2013-07-25 18:55:32,201 DEBUG MainThread > evaluated()
		VALUE: Linux mason 2.4.34.5 #12 Tue Jun 26 14:09:29 CEST 2007 i686
		INFOS:


Specify additional variables and expression from standard input
---------------------------------------------------------------


Additional variables can be specified line by line throw standard input which must have the following syntax: 

	::
		
		$variable_name [ "=" <type> ] "=" <value>
	

Where:

	* $*variable_name* : Must be a valid expression variable (es: $foo)
	* *type* : (optional) Must be a valid type-string: num, string. Default is: string.
	* *value* : The symbols' value.
	


Example: 

::
	
	# python manage.py exec_expr --stdin mason '10 * $myvariable'
	$myvariable=num=5
	
	*Ctrl+D*
	

Output:

	::
		
		2013-07-25 18:57:53,205 DEBUG MainThread > Loading tenant environment
		2013-07-25 18:57:53,220 DEBUG MainThread > Using default tenant: <Tenant:10a279a6ce5c4438a98f9c33c0b77cde:site>
		...
		...
		...
		2013-07-25 18:58:03,262 DEBUG MainThread > evaluated()
		VALUE: 50.0
		INFOS:
		

Specify the expression from standard input
---------------------------------------------------------------

If you want to specify the expression from standard input, you must append a blank line after parameters' lines and enter the expression string.

Esample: 

	::
		
		# python manage.py exec_expr --stdin mason '10 * $myvariable'
		$myvariable1=hello
		$myvariable2=world
		$myvariable3=it's
		$myvariable4=me
		
		$myvariable + 1
		*Ctrl+D*


Output:

	::
		
		2013-07-25 18:57:53,205 DEBUG MainThread > Loading tenant environment
		2013-07-25 18:57:53,220 DEBUG MainThread > Using default tenant: <Tenant:10a279a6ce5c4438a98f9c33c0b77cde:site>
		...
		...
		...
		2013-07-25 18:58:03,262 DEBUG MainThread > evaluated()
		VALUE: 6
		INFOS:

"""


import os
import sys
import time
import logging
import traceback

from optparse import make_option

from sanet_poller.lib.functions import load_system_expr_functions

from sanet_poller.lib.evaluation import ExecutionContext
from sanet_poller.lib.evaluation import compile_str_expr
from sanet_poller.lib.evaluation import expreval

from sanet_poller.lib.functions import load_extra_expr_functions

#from django.core.management.base import BaseCommand


from conf import get_configuration

from core.management.commands import TenantBasedCommand

from core.models import *

from core.api.base import tenant_get_primary, tenant_get
from core.api.base import db_set_current_tenant_schema

from core.api.cache import init_cache_manager_from_conf

from core.localpolling import LocalConfiguration

from server.sanetd.sanetdlogic import ServerIPCFactory

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

from sanet_poller.utils.shell import load_symbols_from_stdin

import StringIO


class Command(TenantBasedCommand):
	
	help = "Execute a generic expression."

	#option_list = TenantBasedCommand.option_list + (
	#	make_option('--stdin'       , action='store_true', dest='use_stdin'  , default=False, help=''),
	#	make_option('--no-defaults' , action='store_true', dest='no_defaults', default=False, help=''),
	#	make_option('-l'            , action='store'     , dest='debug_level', default=20, help=''),
	#)		
	
	def add_arguments(self, parser):

		super(Command, self).add_arguments(parser)

		parser.add_argument('--stdin'       , action='store_true', dest='use_stdin'  , default=False, help=''),
		parser.add_argument('--no-defaults' , action='store_true', dest='no_defaults', default=False, help=''),
		parser.add_argument('-l'            , action='store'     , dest='debug_level', default=20, help=''),

		#
		# Positional
		#
		parser.add_argument('node_name', nargs='?', default=None)
		parser.add_argument('expr_str' , nargs='?', default=None)
		
	#def handle(self, node_name, expr_str=None, *args, **options):
	def handle(self, *args, **options):
		
		node_name = options['node_name']
		expr_str  = options['expr_str' ]

		self.logger = None

		try:
			#tenant_name = options['tenant']
			#if tenant_name == None:
			#	tenant  = tenant_get_primary()
			#	if tenant == None:
			#		raise Exception("No primary tenant found. You must specify a tenant name")
			#else:
			#	tenant = tenant_get(tenant_name)
			#db_set_current_tenant_schema(tenant.name)
			
			tenant = self.get_tenant(options)


			#
			# Handler parameters
			#
			symbols = {}
			if options['use_stdin']:
				expr, symbols = load_symbols_from_stdin()
				if expr:
					
					if expr_str is not None:
						raise Exception("Can not specify an expression from command line and from standard input at the same time")
					
					expr_str = expr


			#
			# Init Log
			#
			output_stream = StringIO.StringIO()

			logger, handler = init_log(output_stream, int(options['debug_level']))

			self.logger   = logger
			self.log_hdlr = handler 
			
			#
			# Test 
			#
			value = None
			infos = []
			
			try:
				value, infos = text_expression(logger, tenant, node_name, expr_str, symbols, options['no_defaults'])
			except Exception, e:
				print type(e), str(e), traceback.format_exc()

			#
			# Print output
			#
			print output_stream.getvalue()
			
			output_stream.close()

			print "VALUE:", value
			print "INFOS:"
			for i in infos:
				print i

		finally:
			self.remove_logger()

	def remove_logger(self):
		
		if self.logger:
			self.logger.removeHandler(self.log_hdlr)



def text_expression(logger, tenant, node_name, expr_str, symbols, no_defaults=False):

			config = get_configuration()
			
			logger.debug( "Loading cache manager...")
			cache_manager = init_cache_manager_from_conf()


			logger.debug( "Creating ipc factory...")
			ipc_factory = ServerIPCFactory(logger)
			
			logger.debug( "Loading system symbols")
			system_symbols = {}
			system_symbols.update( load_system_expr_functions() )

			logger.info("Loading extra symbols from %s", config.LOCAL_POLLER_EXTRA_MODULES)
			extra_symbols = load_extra_expr_functions(config.LOCAL_POLLER_EXTRA_MODULES)
			
			expr_symbols = {}
			expr_symbols.update( {
				'$shorttries': 5,
				'$timeout'   : 1,
			})


			
			global_symbols = {}
			global_symbols.update(symbols)
			global_symbols.update(system_symbols)
			global_symbols.update(extra_symbols)

			#
			# Prepare Configuration
			#
			c = LocalConfiguration(logger, tenant.uuid, cache_manager)
			c.set_global_symbols(global_symbols)

			#
			# Get element parameters
			#
			node = None
			element = None
				
			sub_element_name = None
			try:
				node_name, sub_element_name = node_name.split(':')
			except ValueError, e:
				pass

			try:
				node = c.get_node_by_name(node_name)

				expr_symbols['__element__'] = node
				expr_symbols['__node__'   ] = node
				
				node.calculate_effective_ip()

				expr_symbols.update( node.get_expr_symbols() )
				
				if sub_element_name is not None:

					for kind in [ELEMENT_KIND_IFACE, ELEMENT_KIND_STORAGE, ELEMENT_KIND_SERVICE, ELEMENT_KIND_DEVICE]:
						try:
							element = c.get_subelement_by_name(node_name, kind, sub_element_name)
							logger.info("Found: %s", element)

							expr_symbols['__element__'] = element

							ref_key = {
								ELEMENT_KIND_IFACE   : "__iface__",
								ELEMENT_KIND_STORAGE : "__storage__",
								ELEMENT_KIND_SERVICE : "__service__",
								ELEMENT_KIND_DEVICE  : "__device__",
							}
					

							
							expr_symbols.update( node.get_expr_symbols() )

							expr_symbols[ref_key[kind]] = element
							
							if kind == ELEMENT_KIND_IFACE:
								expr_symbols['__iface__'] = element
								expr_symbols.update( element.get_expr_symbols() )
								
								linked_iface = c.get_linked_iface_by_id( element.uuid, link_type=LINK_TYPE_L2 )
								if linked_iface != None:
									linked_node = get_element_node( c, linked_iface )
									
									expr_symbols['__linked_iface__'] = linked_iface
									expr_symbols['__linked_node__' ] = linked_node
									
									expr_symbols.update( linked_iface.get_expr_symbols(prefix='linked_') )
									expr_symbols.update( linked_node.get_expr_symbols(prefix='linked_') )

							break
							
						except Exception, e:
							pass
				
				if element:
					if node.uuid != element.uuid:
						expr_symbols.update( node.get_expr_symbols() )

			except Exception, e:
				logger.warning("Node %s not found: %s", node_name, e)

				node_data = {
					'$node'       : node_name,
				}

				if not no_defaults:
					node_data.update({
						'$community'  : 'public',
						'$snmpversion': 1,
						'$snmpport'   : 161,
					})

				expr_symbols.update(node_data)


			#
			# SubElement
			#
			#if sub_element:
			#	logger.info("Loading sub element: %s", sub_element)
			#	try:
			#		sub = c.get_subelement_by_name(node_name, ELEMENT_KIND_IFACE, sub_element)
			#		logger.info("Found: %s", sub)
			#		expr_symbols['__iface__'] = sub
			#		expr_symbols.update( sub.get_expr_symbols() )
			#		
			#		linked_iface = c.get_linked_iface_by_id( sub.uuid, link_type=LINK_TYPE_L2 )
			#		if linked_iface != None:
			#			linked_node = get_element_node( c, linked_iface )
			#			
			#			expr_symbols['__linked_iface__'] = linked_iface
			#			expr_symbols['__linked_node__' ] = linked_node
			#			
			#			expr_symbols.update( linked_iface.get_expr_symbols(prefix='linked_') )
			#			expr_symbols.update( linked_node.get_expr_symbols(prefix='linked_') )
			#		
			#		
			#	except Exception, e:
			#		print e
			#		pass
				
			#expr_symbols.update( stdin_symbols )

			expr_symbols.update( c.get_global_symbols() )

			logger.debug("Context:")
			if expr_symbols:
				max_len=str(max([len(s) for s in expr_symbols.keys()]))
				for s,v in sorted(expr_symbols.items()):
					if isinstance(v, types.FunctionType):
						logger.debug("\t%-"+max_len+"s = function %s"      , s, v.__module__)
					else:
						logger.debug("\t%-"+max_len+"s = %s %-30s", s, type(v).__name__, repr(v)  )

			#
			# Execution Context
			#
			exec_context = ExecutionContext()
			exec_context.setLog( logger )
			exec_context.setConfiguration( c )
			exec_context.setCache(cache_manager)
			exec_context.setIPCFactory( ipc_factory )

			exec_context.update( expr_symbols )
			
			exec_context['__check_time__'] = time.time()

			#
			# Evaluate expr
			#
			expr_tree = compile_str_expr(logger, expr_str)


			try:
				value = expreval(logger, exec_context, expr_tree)
				
				logger.debug("evaluated()")
				expanded_expr = expr_tree.evaluated() # FIX: IMPORTANTE NON TOGLIERE QUESTA RIGA O NON FUNZIONA IL GIOCO CON LE VARIABILI OLD

				infos = exec_context.getExecutionInfos()

				eval_descr = "expression %s expanded to %s" % (expr_str, expanded_expr)
				infos.append( eval_descr)
				
				additional_data = exec_context.get_all_additional_update_data()
				if additional_data:
					logger.debug("Additional data")
					for id, data in additional_data:
						logger.debug("\t%s = %s", id, (str(data)[0:30]+"...") if len(str(data)) > 30 else str(data))
			except Exception, e:
				logger.debug("Execution exception detected! %s:%s", type(e), str(e))
				
				value = None
				infos = [ "ERROR: %s:%s\nTRACE : %s" % (type(e), str(e), traceback.format_exc()) ]

			return value, infos

def init_log(stream, level=20):

	log_level = level # logging.DEBUG

	format_str = '%(asctime)s %(levelname)s %(threadName)s > %(message)s'
	
	logger = logging.getLogger('test')
	logger.setLevel( log_level )
	
	hdlr = logging.StreamHandler(stream)
	hdlr.setFormatter( logging.Formatter(format_str) )
	hdlr.setLevel( log_level )
	logger.addHandler(hdlr)	
	
	return logger, hdlr

def get_element_node(configuration, element):
	"""
	Restituisce l'elemento Nodo che contiene l'elemento "element".
	"""
	e = element
	while e.parent_id != None:
		e = configuration.get_element_by_id(e.parent_id)
	return e



