# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

exec_timegraph
=============================================

Launch: 

	::
		
		sanet-manage exec_timegraph [opzioni] <datagroup> <timegraph name>
		
		sanet-manage exec_timegraph [opzioni] <timegraphid>

Testare un timegraph

Le opzioni disponibili sono: 

	::
		
		--tenant <nome>   Esegue il datagroup del tenant specificato. Default: Tenant primario
		
		--seconds <n>     Imposta quanto indietro nel tempo rispetto all'istante di esecuzione utilizzare per calcolare i dati delle serie.
		
		--export          Le serie del timegraph vengono esportate in formato CSV
		
		--graph           Show graph with ascii.

Esempi
--------    

Alcuni esempi:

	::
		
		sanet-manege exec_timegraph 8f4e95cd2de54a888cfd92eb974ae83b --seconds=3000 --export
		
		sanet-manage exec_timegraph 'localhost;;icmp-reachability'  rttgraph  --seconds=3000 --export

"""

import os
import sys
import time
import pprint
from optparse import make_option
import math

# Global Sanet libraries
from sanet_common.ipc import *
from sanet_common.timeutils import timestamp_to_isostr

from sanet_common.termutils import *  # all colors

# Local libraries (Sanet+Django related)
#from django.core.management.base import BaseCommand

from conf import get_configuration

from core.management.commands import TenantBasedCommand
from core.models import *
from core.api.datasource import init_datasource_store_from_conf

from webui.views import calculate_timegraph_series


_COLORS = [
	TERM_COLOR_GREEN          ,
	TERM_COLOR_BLUE           ,
	TERM_COLOR_CYAN           ,
	TERM_COLOR_YELLOW         ,
	TERM_COLOR_MAGENTA        ,
	TERM_COLOR_LIGHT_BLUE     ,
	TERM_COLOR_LIGHT_MAGENTA  ,
	TERM_COLOR_LIGHT_GRAY     ,
	TERM_COLOR_LIGHT_CYAN     ,
	TERM_COLOR_WHITE          ,
	TERM_COLOR_RED            ,
	TERM_COLOR_DARK_GRAY      ,
	TERM_COLOR_LIGHT_RED      ,
	TERM_COLOR_LIGHT_GREEN    ,
	TERM_COLOR_LIGHT_YELLOW   ,
]

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

class Command(TenantBasedCommand):

	help = "Processe and produce timegraph's series"

	#option_list = TenantBasedCommand.option_list + (
	#	make_option('--seconds', action='store'     , dest='seconds'       , default=60*5, help=''),
	#	make_option('--export' , action='store_true', dest='export'        , default=False, help=''),
	#	make_option('--graph'  , action='store_true', dest='graph'         , default=False, help=''),
	#)

	def add_arguments(self, parser):
		super(Command, self).add_arguments(parser)

		parser.add_argument('--seconds' , action='store'     , dest='seconds'       , default=60*5, help=''),
		parser.add_argument('--export'  , action='store_true', dest='export'        , default=False, help=''),
		parser.add_argument('--graph'   , action='store_true', dest='graph'         , default=False, help=''),
		
		parser.add_argument('parameters', action='store'     , nargs='*', help=''),
		
	def handle(self, *args, **options):

		datasource_store = init_datasource_store_from_conf()
		cache_manager    = init_cache_manager_from_conf()

		conf = get_configuration()

		tenant = self.get_tenant(options)

		seconds = int(options['seconds'])

		parameters = options['parameters']

		timegraph = None
		if len(parameters) == 2:
			try:
				datagroup = DataGroup.objects.get(tenant=tenant      , path=parameters[0])
				timegraph = TimeGraph.objects.get(datagroup=datagroup, name=parameters[1])
			except TimeGraph.DoesNotExist, e:
				print "Not found timegraph with uuid: %s %s" % (parameters[0], parameters[1])
				return
		else:
			try:
				timegraph = TimeGraph.objects.get(uuid=parameters[0])
			except TimeGraph.DoesNotExist, e:
				print "Not found timegraph with uuid: %s" % (parameters[0])
				return
		
		ts_end   = int(time.time())
		ts_start = ts_end - seconds
		
		series = calculate_timegraph_series(datasource_store, timegraph, ts_start, ts_end, include_errors=True, cache_manager=cache_manager)
		
		if options['export']:
			
			tot = len(series[0]['values'])
			
			ts_start = series[0]['timescale']['from']
			step     = series[0]['timescale']['step']
			
			print "timestamp", ", ".join( [ s['title'] for s in series ] )
			
			#print "["
			for i in xrange(tot):
				#print "[", 
				print ts_start,
				for s in series:
					print "," , s['values'][i],
				#print "],"
				print ""
				
				ts_start += step
			#print "]"

		elif options['graph']:

			from operator import itemgetter
			series = sorted(series, key=itemgetter('format', 'order'))


			num_series = len(series)
			tot        = len(series[0]['values'])
			ts_start   = series[0]['timescale']['from']
			step       = series[0]['timescale']['step']


			min_val = 0
			max_val = 0.0000000001

			for i in xrange(num_series):
				serie = series[i]
				s_min = 0
				s_max = 0.0000000001
				for v in serie['values']:
					if v is not None:
						min_val = min(min_val, v)
						max_val = max(max_val, v)

						s_min = min(s_min, v)
						s_max = max(s_max, v)

				if serie['format'] != 'tick':
					print term_color(_COLORS[i], serie['label'] ), s_min, s_max
				else:
					print term_color(_COLORS[i], serie['label'] )


			term_size

			max_range        = max_val
			MAX_TERM_COLUMNS = term_size()[0] - 60
			factor           = float(MAX_TERM_COLUMNS) / max_range


			GRAPH_PATTERN = "%19s %30.10f |%-s"

			#      1234567890123456789 123456789012345678901234567890 |
			BORDER = "                                                   +" + '-' * MAX_TERM_COLUMNS + '+'
			
			print BORDER
			
			ts = ts_start
			for x in xrange(tot):

				buf    = ''
				buflen = 0
				
				maxv = 0
				

				for s in xrange(num_series):

					serie = series[s]
					v = serie['values'][x]
					if v is None: 
						continue

					if serie['format'] == 'tick':
						c = '#'
						if v != 0:
							n = MAX_TERM_COLUMNS

							buf = c * n
							buflen += n

					else:
						maxv = max(maxv,v)

						n    = int(v * factor)

						l    = n - buflen 
						
						c = 'x' #styles[s]
						line = c * l
						
						buflen += len(line) 
						
						buf += term_color(_COLORS[s], line)


				print GRAPH_PATTERN % (timestamp_to_isostr(ts), maxv, buf, ) + ' ' * (MAX_TERM_COLUMNS - buflen) + '|'

				ts += step

			print BORDER

		else:
			for s in series:
				
				print s['label']
				ts_start = s['timescale']['from']
				step     = s['timescale']['step']
				
				ts = ts_start
				for v in s['values']:
					print "\t %s [%s] = %s" % (timestamp_to_isostr(ts), ts, v)
					ts += step
				
				if s.has_key('errors'):
					ts = ts_start
					for e in s['errors']:
						print "\t %s [%s] = %s" % (timestamp_to_isostr(ts), ts, v)
						ts += step



