# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""
push_event
========================================================================

Push generic event and process it with the matching with passive datagroup

TODO 

"""

import os
import sys
import time
import re

import pprint
import logging
from optparse import make_option

from sanet_common.ipc import *

from sanet_poller.agent import BaseAgent

from core.models import MonitoredNode, Event

from core.api.base import db_set_current_tenant_schema

from core.api.push import guess_node_from_ip

from server.sanetd.sanetdlogic import create_ipc_connection_from_config

from core.management.commands import TenantBasedCommand

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

class UnknownEventException(Exception):
	pass
	
class UnManagedEventException(Exception):
	pass

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

class Command(TenantBasedCommand):

	help = "Push generic event and process it with the matching with passive datagroup"

	def add_arguments(self, parser):
		super(Command, self).add_arguments(parser)

		parser.add_argument('-L'             , action='store'     , dest='loglevel'     , default=20   , help='Set log level (<20 DEBUG, <30 INFO, <40 WARNING, <50 ERROR.'),

		parser.add_argument('--stdin'        , action='store_true', dest='use_stdin'    , default=False, help='Load event data from standard input in JSON format'),
		parser.add_argument('--use-env'      , action='store_true', dest='use_env'      , default=False, help='Load event data from environment symbols PUSH_EVENT_<name>'),

		parser.add_argument('--element'      , action='store'     , dest='element'      , default=None , help='Specify the Sanet monitored element\'s path.'),
		parser.add_argument('--sourceip'     , action='store'     , dest='sourceip'     , default=None , help='Specify the source ip of the event generator.'),

		parser.add_argument('--passive-match', action='store'     , dest='passive_match', default=None , help='Force a passive match string.'),

		#
		# Positional arguments
		#
		parser.add_argument('parameters'     , action='store'     , nargs='*'           , help='')

	def handle(self, *args, **options):

		#
		# Set the right tenant
		#
		tenant = self.get_tenant(options)

		parameters = options['parameters']
		#
		# Init Log
		#
		l = int(options['loglevel'])
		self.logger = None
		logger = self.init_log(l)

		#
		# Read raw alarm from options or standard in
		#

		if len(parameters) > 0:
			logger.debug("Reading alarm from command line argument")
			raw_event_string = parameters[0]
			print raw_event_string
		elif options['use_stdin']:
			logger.debug("Reading alarm from standard input")
			raw_event_string = sys.stdin.read()
		elif options['use_env']:
			logger.debug("Reading alarm from environment")

			raw_event = {}
			for k, v in os.environ.items():
				if k.startswith('PUSH_EVENT_'):
					name = k[ len('PUSH_EVENT_') : ]
					raw_event[ name ] = v

			raw_event_string = json.dumps(raw_event)
		else:
			raise Exception("You must specify alarm\'s data")

		logger.debug("Raw event data: %s", raw_event_string)
		event_data = json.loads(raw_event_string)

		#
		# Add minimum fields
		#
		ts           = int(time.time())
		source_ip    = options['sourceip']
		element_path = options['element']

		logger.debug("ts           = %s", ts          )
		logger.debug("source ip    = %s", source_ip   )
		logger.debug("element path = %s", element_path)

		if not (element_path or source_ip):
			raise Exception("YOu must specify a source_ip or an element path")

		if element_path:
			sanet_element      = MonitoredNode.objects.get(tenant=tenant, name=element_path)
			sanet_element_name = sanet_element.path
			sanet_element_uuid = sanet_element.uuid
		else:
			sanet_element_name = source_ip
			sanet_element_uuid = None

			sanet_element      = MonitoredNode.get_from_sourceip(source_ip)

			if sanet_element is not None:
				sanet_element_name = sanet_element.name
			else:
				sanet_element, sanet_element_name = guess_node_from_ip(logger, source_ip)

			if sanet_element is not None:
				sanet_element_uuid = sanet_element.uuid

		try:
			#
			# Salviamo l'evento sul DB per iniziare
			#
			logger.debug("Saving custom event on log...")
			Event.log_custom_event(tenant, int(ts), source_ip, sanet_element_name, sanet_element_uuid, raw_event_string)
			logger.debug("Saved")
			
			if sanet_element is None:
				raise UnknownEventException("Unable to find a monitored element related to the event")
				
			time.sleep(0.5)

			#
			# Controllo passivo
			#

			logger.debug("Processing event with passive datagroup...")
			
			passive_match_value = options['passive_match'] or event_data.get('passive_match',None)
			if not passive_match_value:
				raise UnManagedEventException("Passive match string not provided for the alarm")

			ipc_connection = create_ipc_connection_from_config(logger)

			process_event_with_datagroup(logger, ipc_connection, ts, source_ip, sanet_element, passive_match_value, event_data)

			logger.debug("Finish")
			
		except UnknownEventException, e:
			print e
			
		except UnManagedEventException, e:
			print e

		self.remove_logger()
		
		return


	def init_log(self, log_level):

		format_str = '%(asctime)s %(levelname)s %(threadName)s > %(message)s'
		logger = logging.getLogger('test')
		logger.setLevel( log_level )

		self.logger = logger

		hdlr = logging.StreamHandler()
		hdlr.setFormatter( logging.Formatter(format_str) )
		hdlr.setLevel( log_level )
		logger.addHandler(hdlr)

		self.log_hdlr = hdlr

		return logger

	def remove_logger(self):

		if self.logger:
			self.logger.removeHandler(self.log_hdlr)




def process_event_with_datagroup(logger, ipc_connection, receive_time, source_ip, sanet_node, passive_match_value, event_data, check_time=None):

	from sanet_poller.lib.classes   import DataGroupChecker
	from sanet_poller.lib.functions import load_system_expr_functions
	from sanet_poller.lib.functions import load_extra_expr_functions
	from sanet_poller.lib.communication import AgentUpdateEvent
	
	from django.conf import settings
	
	from core.models import DataGroup
	from core.api.cache import init_cache_manager_from_conf
	from core.api.datasource import init_datasource_store_from_conf
	from core.localpolling import LocalConfiguration
	
	from server.sanetd.sanetdlogic import ServerIPCFactory
	
	if check_time is None:
		check_time = time.time()
	
	logger.info("Processing event '%s' on node %s", passive_match_value, sanet_node)
	
	datagroups_to_process = []
	
	#
	# Calcoliamo quale datagroup deve processare l'evento
	#
	
	# 1. match esatto sul passive_match
	base_query = DataGroup.objects.filter(element=sanet_node, minperiod=0)
	
	for dg in base_query.all():
		logger.debug("%s -> %s", dg.path, dg.passive_match)
	
	datagroups_to_process += list( base_query.filter(passive_match=passive_match_value).all() )
	
	# 2. Datagroup che fanno match usando passive_match come espressione regolare
	for dg in base_query.filter(passive_match__startswith=DataGroup.PASSIVE_MATCH_REGEX_PREFIX).all():
		regexp = dg.passive_match.replace(DataGroup.PASSIVE_MATCH_REGEX_PREFIX, '')
		if re.match(regexp, passive_match_value):
			datagroups_to_process.append(dg)
	
	logger.info("Found %s matching eventid", len(datagroups_to_process) )
	
	if len(datagroups_to_process) == 0:
		raise UnManagedEventException("No datagroup found for passive match '%s'" % (passive_match_value))
	
	#
	# PROCESS
	#
	for dg in datagroups_to_process:
		
		logger.info("Processing %s", dg.path)
		
		tenant = dg.tenant
		db_set_current_tenant_schema(tenant.name)
		datagroup_id = dg.uuid
		
		cache_manager = init_cache_manager_from_conf()
		ipc_factory   = ServerIPCFactory(logger)
		datasource_store = init_datasource_store_from_conf()
		
		#
		# Simboli disponiboli
		#
		symbols = {}
		
		# Simboli PUSH
		logger.debug("Loading event data")
		
		for k, v in event_data.items():
			logger.debug( "\t%s = %s", k, v)
			symbols['$'+k] = v
		
		logger.debug( "Loading system symbols")
		system_symbols = load_system_expr_functions()
		
		logger.debug("Loading extra symbols from %s", settings.LOCAL_POLLER_EXTRA_MODULES)
		extra_symbols = load_extra_expr_functions(settings.LOCAL_POLLER_EXTRA_MODULES)
		
		symbols.update(system_symbols)
		symbols.update(extra_symbols)
		
		# Creiamo l'oggetto che ha il compito fornire i vari oggetti del monitoraggio
		# (nodi,interfacce,ecc...) che potranno servire durante la valutazione
		# del datagroup.
		c = LocalConfiguration(logger, tenant.uuid, cache_manager)
		
		c.set_global_symbols(symbols) 
		
		# Otteniamo l'oggetto PDataGroup associato al datagroup selezioanto
		pdg = c.get_datagroup(datagroup_id)
		
		dgc  = DataGroupChecker(c, logger, cache_manager, ipc_factory, pdg)
		 
		logger.debug( "########################################################" )
		logger.debug( "###################### CHECKING ########################" )
		logger.debug( "########################################################" )
		
		logger.info( "Start checking...")
		
		# CHECK
		t1 = time.time()
		dgc.check( check_time )
		t2 = time.time()
		
		logger.debug( "########################################################" )
		logger.debug( "####################### SENDING ########################" )
		logger.debug( "########################################################" )
		
		agent = None
		
		updates_data = dgc.get_datagroup_all_updates()
		
		#
		# I dati dell'evento pushed vengono inseriti nel messaggio dati
		# inviato al server centrale.
		#
		updates_data[0]['pushed_data'] = event_data
		
		#
		# INVIO
		#
		logger.debug( "Creating update event" )
		fake_event = AgentUpdateEvent(check_time, tenant.uuid, agent, updates_data)
		
		logger.debug("Sending update data: %s", pprint.pformat(fake_event))
		
		logger.info("Sending update data to sanetd broker" )
		
		events = [fake_event, ]
		
		connection = ipc_connection
		connection.open()
		
		#events = json.dumps(events)
		events = BaseAgent.encode_update_events(events)
		
		connection.send_msg( 'sanetd-events', events, blocking=False)
		
		connection.close()
		logger.info("Sent. Exit." )
	
	return

