# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

sanetdata
====================================================================================

Lancio: 

	::
		
		sanet-manage sanetdata


Questo comando effettua un checkup dell'installazione effettuando le seguenti operazioni automatiche:
	* inizializzazione di directory/file di sistema
	* aggiornamento del database da una release all'altra
	* fix di dati corrotti
 
.. danger:: questo comando deve essere SEMPRE eseguito quando sanet NON e' in esecuzione.
 
.. note:: Questo comando puo' essere lanciato un numero illimitato di volte.

.. attention:: Questo comando cancella le sessioni WEB attualmente memorizzate dal sistema. Una volta lanciato, tutte le sessioni WEB che richiedono autenticazione (login) devono essere ripristinate.

"""

import time
import os
import sys
import logging
import getpass
import pytz
from optparse import make_option

from sanet_common.termutils import term_green

from sanet_common.sanetdefs.constants import *

from django.db import connection
from django.db import transaction
from django.core.management import call_command
from django.core.management.base import BaseCommand
from django.contrib.auth.models import User

from conf import get_configuration
from conf import check_base_dirs

from core.db_utils import transaction_commit_on_success, transaction_commit

from core.management.commands import SanetBaseCommand
from core.models import sanet_init_subapp
from core.models import sanet_check_tenant
from core.models import sanet_cleanup_tenant
from core.models import sanet_post_check
from core.models import CDataGroup, DataGroup
from core.models import TField
from core.models import Link
from core.models import MonitoredElement
from core.models import SANETUserProfile
from core.models import Capability

from core.api.sysstats import check_system_settings


from core.api.base import tenant_select_all
from core.api.base import db_set_current_tenant_schema
from core.api.base import group_get, group_create
from core.api.base import check_common_icons
from core.api.base import check_primary_agent
from core.api.base import check_priority_levels	
from core.api.base import tenant_get_primary, tenant_select_all
from core.api.base import tenant_create, tenant_select_all
from core.api.base import tenant_create_datadir
from core.api.base import group_get, group_adduser

from core.api.elements import element_rename_raw

#from core.api.datasource import check_datasource_store_conf
from core.api.datasource import init_datasource_store_from_conf
from core.api.datasource import init_datasource_store

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

def check_default_groups():
	logger = logging.getLogger('sanet.global')

	logger.info( "Checking group '%s'...", get_configuration().DEFAULT_ADMIN_GROUP,)
	g = group_get(get_configuration().DEFAULT_ADMIN_GROUP)

	if g == None:
		logger.info(  "Creating group '%s'", get_configuration().DEFAULT_ADMIN_GROUP,)
		g = group_create(get_configuration().DEFAULT_ADMIN_GROUP)


def check_default_users(options):

	logger = logging.getLogger('sanet.global')

	#
	# FIX: use core.api
	#

	logger.info( "Checking admins..." )
	admins = User.objects.filter(is_superuser=True)
	if len(admins) == 0:

		logger.info( "No administrators defined..." )

		logger.info( "Creating default admin '%s'", get_configuration().DEFAULT_ADMIN_USERNAME )

		admin = User(
			username     = get_configuration().DEFAULT_ADMIN_USERNAME,
			is_superuser = True,
			is_active    = True,
		)

		if options['adminpass'] is None:	
			password  = getpass.getpass("Enter '%s' password: " % (admin.username))
			password2 = getpass.getpass("Confirm password: ")

			if not password:
				raise Exception("Invalid password")

			if password != password2:
				raise Exception("Password not confirmed")
		else:
			 password = options['adminpass']
			 logger.info( "using default pass '%s'" % (password) )

		admin.set_password(password)

		admin.save()

	active_admins = User.objects.filter(is_superuser=True, is_active=True).count()
	if active_admins == 0:
		logger.info( "WARNING: no active admin users" )


def check_default_auth():

	logger = logging.getLogger('sanet.global')
	
	logger.info( "Checking default authorizations")
	
	g = group_get(get_configuration().DEFAULT_ADMIN_GROUP)
	if g == None:
		raise Exception("Unable to find group '%s'" % (g.name))

	# FIX: user core.api !!
	admins = User.objects.filter(is_superuser=True)

	for a in admins:
		logger.info( "Added user '%s' to '%s'", a.username, g.name )
		group_adduser(g, a)

def check_missing_profiles():
	
	logger = logging.getLogger('sanet.global')
	
	logger.info( "Checking user profile data")
	
	for u in User.objects.all():
		try:
			u.get_profile()
		except SANETUserProfile.DoesNotExist, e:
			logger.debug("\t"+"Creating data for %s" % (u.username))
			up = SANETUserProfile(user=user)
			up.save()

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

def check_every_tenant():

	for t in tenant_select_all():

		check_single_tenant(t)

		#if 'entables' in get_configuration().INSTALLED_APPS:
		#	from entables.tablesstore.djangostore import init_entables_chains, set_default_tenant_rules, fix_names
		#	init_entables_chains(t)
		#	set_default_tenant_rules(t)
		#	fix_names(t)

		sanet_check_tenant.send(sender=t.__class__, instance=t)

	return

@transaction_commit_on_success
def check_single_tenant(tenant):

	logger = logging.getLogger('sanet.global')

	logger.info( "Checking configuration and setting for tenant '%s'", tenant.name )

	tenant_create_datadir(tenant.uuid, tenant.name)
	
	check_datasource_store(tenant)
	
	check_primary_agent(tenant)

	check_common_icons(tenant)

	check_priority_levels(tenant)

	fix_capabilities(tenant)

	check_invalid_names(tenant)
	
	check_invalid_dg_names(tenant)
	
	check_invalid_dg_suspended(tenant)
	
	return

#----------------------------------------------------------------------#

def check_datasource_store(tenant):
	
	logger = logging.getLogger('sanet.global')
	
	logger.info( "Init datasource module" )
	datasource_store = init_datasource_store_from_conf()  # FIX: should init with tenant?
	
	datasource_store.init_tenant(tenant)
	
	return

#----------------------------------------------------------------------#

def fix_capabilities(tenant):
	
	logger = logging.getLogger('sanet.global')
	
	logger.info("Checking capabilities")
	
	tenant.db_set_current_schema()
	
	for old, new in [ ('webscenario.element.edit', 'webscenario.element.create' ), ('webscenario.read', 'webscenario.access') ]:
		for c in Capability.objects.filter(tenant=tenant, code=old):
			c.code = new
			c.save()
	return

#----------------------------------------------------------------------#

def check_invalid_names(tenant):
	
	logger = logging.getLogger('sanet.global')

	logger.info( "Fixing invalid names" )

	for m in MonitoredElement.objects.filter(tenant=tenant).filter(kind__in=[ELEMENT_KIND_STORAGE, ELEMENT_KIND_SERVICE, ELEMENT_KIND_DEVICE]):

		if re.match(VALID_ELEMENT_NAME,  m.name):
			continue
		
		base_new_name = _fix_element_name(m.kind, m.name)
		
		if not re.match(VALID_ELEMENT_NAME,  base_new_name):
			raise Exception("Element %s has an invalid name ('%s'), but cant figure out how to fix it. Tried with '%s'" % (m.path, m.name, base_new_name))

		new_name = base_new_name
		if MonitoredElement.objects.filter(tenant=tenant).filter(kind=m.kind, parent=m.parent, name=base_new_name).count() > 0:

			counter  = 2
			new_name = "%s-%s" % (base_new_name, counter)
			while MonitoredElement.objects.filter(tenant=tenant).filter(kind=m.kind, parent=m.parent, name=new_name).count() > 0:
				counter += 1
				new_name = "%s-%s" % (base_new_name, counter)

		element_rename_raw(m, new_name)

	return

def _fix_element_name(kind, name):
	
	if kind == ELEMENT_KIND_STORAGE:
		if name == '/':
			name = 'root'

	name = name.strip(";").strip(":").strip(" ").strip("/")

	name = name.replace(":", "")
	name = name.replace(";", "")
	name = name.replace(" ", ".")
	name = name.replace("/", ".")

	return name

#----------------------------------------------------------------------#

def check_invalid_dg_names(tenant):
	"""
	A causa di un bug nel rename ci sono in giro delle incosistenze che potrebbero
	saltare fuori. 
	"""
	
	logger = logging.getLogger('sanet.global')
	
	logger.info( "Fixing datagroup names with configured names" )

	tenant.db_set_current_schema()
	
	cdg_name_by_id = dict([ (e[0],e[1]) for e in CDataGroup.objects.filter(tenant=tenant).values_list('id', 'name').all() ])
	
	for dg in DataGroup.objects.filter(tenant=tenant).all():

		name = cdg_name_by_id[ dg.cdatagroup_id ]
		
		if dg.name != name:
			
			logger.debug("Fixed datagroup %s. Wrong name '%s' (configured name %s)" % (dg.path, dg.name, name))
			dg.name = name
			dg.save()

	return

def check_invalid_dg_suspended(tenant):

	#
	# This routine fixes problems with inconsistent suspended datagroups
	#

	logger = logging.getLogger('sanet.global')

	logger.info( "Fixing suspended datagroups" )

	tenant.db_set_current_schema()

	#
	# Find datagroups flagged suspended with wrong conditions
	#
	c = connection.cursor()
	c.execute("""
SELECT uuid 
  FROM datagroup 
 WHERE datagroup.path IN (SELECT dg.path 
                            FROM datagroup dg  INNER JOIN condition c ON (dg.uuid = c.datagroup_id) 
                           WHERE dg.suspended = true AND c.status != 70        
                           GROUP by dg.path 
                          HAVING count(*) > 0
                         )
""")
	uuids = [r[0] for r in c.fetchall()]

	dgs_to_fix =  DataGroup.objects.filter(uuid__in=uuids).order_by('path').all()

	if dgs_to_fix:
		logger.info( "Removing wrong parameters")
		resume_time = time.time()
		for dg in dgs_to_fix:
			logger.info( "Fixing datagroup: %s", dg.path)
			dg.remove_configured_suspend_fields()
			dg.suspended = False
			dg.save()

		logger.info( "Wait for suspension")
		time.sleep(1)

		logger.info( "Suspending datagroups")
		suspend_time = time.time()
		for dg in dgs_to_fix:
			logger.info( "Suspending: %s", dg.path)
			dg.suspend(suspend_time=suspend_time, reason="Suspended by sanetdata")

	return

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

def cleanup_every_tenant():

	for t in tenant_select_all():
		
		connection.close()  # just to be sure we are working with a clean connection from now on
		
		cleanup_single_tenant(t)


def cleanup_single_tenant(tenant):
	
	logger = logging.getLogger('sanet.global')
	
	logger.info( "Cleaning tenant '%s'", tenant.name )
	
	_t_cleanup_single_tenant(tenant)

	# Cancelliamo i file RRD spusi.
	# IMPORTANTE: di default non cancelliamo i file RRD con "save-value=False" perche' e' pericoloso. Per questo motivo
	#             al comando "clean_rrds" non passiamo il parametro "remove-no-save-value".
	call_command('clean_rrds'    , tenant=tenant.name, log_level=logger.getEffectiveLevel(), interactive=False, verbosity=True, show_traceback=True)

	call_command('fix_dead_users', tenant=tenant.name, interactive=False, verbosity=True, show_traceback=True)

	#
	# Trigger del segnal di clean per tutte le applicazioni.
	#
	sanet_cleanup_tenant.send(sender=tenant.__class__, instance=tenant)

	return

@transaction_commit_on_success
def _t_cleanup_single_tenant(tenant):

	db_cursor = connection.cursor()
	db_set_current_tenant_schema(tenant.name, db_cursor=db_cursor)

	remove_broken_links(db_cursor)

	remove_dead_datagroups(db_cursor)

	fix_wrong_priority(db_cursor)

	return

def fix_wrong_priority(db_cursor):
	
	logger = logging.getLogger('sanet.global')
	
	logger.info( "Fixing wrong priority" )

	db_cursor.execute("UPDATE monitoredelement SET priority = 0 where priority < 0 or priority > 100")
	db_cursor.execute("UPDATE condition        SET priority = 0 where priority < 0 or priority > 100")

	return

def remove_broken_links(db_cursor):
	
	logger = logging.getLogger('sanet.global')
	
	logger.info( "Remove broken links" )

	for l in Link.objects.all():

		if l.ifaces.count() < 2:
			logger.info( "Deleting link %s", l )
			l.delete()

	return

def remove_dead_datagroups(db_cursor):
	
	logger = logging.getLogger('sanet.global')
	
	logger.info("Deleting zombie datagroups")

	find_datagroups_id_query = """
                select id
                  from (select dg.uuid  as id,
                               dg.path  as path, 
                               cdg.id   as cdg_id, 
                               tdg.uuid as tdg_id, 
                               tedg.id  as tedg_id
                          from datagroup dg left join c_cdatagroup   cdg on (dg.cdatagroup_id   = cdg.id)
                                            left join c_tdatagroup   tdg on (cdg.tdatagroup_id  = tdg.uuid)
                                            left join c_tedatagroup tedg on (cdg.tedatagroup_id = tedg.id)
                       ) t
                 where tdg_id is null and tedg_id is null

	"""

	#--
	#-- Trova e cancella tutti i datagroup in un tenant che non sono sono piu' associati a nessun template in configurazione
	#--

	queries = [
		"""delete from datasource          where datagroup_id in (""" + find_datagroups_id_query + """);""",
		"""delete from condition           where datagroup_id in (""" + find_datagroups_id_query + """);""",
		"""delete from timegraphserie      where timegraph_id in (select uuid from timegraph where datagroup_id in (""" + find_datagroups_id_query + """) );""",
		"""delete from timegraph           where datagroup_id in (""" + find_datagroups_id_query + """);""",
		"""delete from datagroup_parameter where datagroup_id in (""" + find_datagroups_id_query + """);""",
		"""delete from datagroup           where uuid         in (""" + find_datagroups_id_query + """);""",
	]

	count = 0
	for query in queries:
		db_cursor.execute(query)
		count += db_cursor.rowcount

	logger.info( "\tDeleted %s entities.", count)

	return


#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

def check_default_tenant(options):

	t = tenant_get_primary()
	if t == None:

		tenants = tenant_select_all()
		if len(tenants) > 1:
			raise Exception("Multiple tenants but not a primary one! Something is wrong")

		create_primary_tenant(options)

def create_primary_tenant(options):

	logger = logging.getLogger('sanet.global')

	logger.info( "Creating default tenant" )

	# Ask name

	if options['defaulttenantname'] is None:
		name      = raw_input("Name [%s]:"       % (DEFAULT_PRIMARY_TENANT_NAME) )
		if not name:
			name = DEFAULT_PRIMARY_TENANT_NAME
	else:
		name = options['defaulttenantname']

	# Ask long name
	if options['defaulttenantlongname'] is None:
		long_name = raw_input("Long name [%s]:"       % (DEFAULT_PRIMARY_TENANT_LONG_NAME) )
	else:
		long_name = options['defaulttenantlongname']

	# Ask timezone
	while 1:
		if options['defaulttenanttimezone'] is None:
			tz = raw_input("Timezone [%s]" % (DEFAULT_TIMEZONE))
		else:
			tz = options['defaulttenanttimezone']

		if not tz:
			tz = DEFAULT_TIMEZONE
		
		try:
			pytz.timezone(tz)
			break
		except pytz.exceptions.UnknownTimeZoneError, e:
			logger.error( e )

		options['defaulttenanttimezone'] = None

	t = tenant_create(name, long_name=long_name, primary=True, timezone=tz)

	#from core.api.base import check_indexes_on_tables
	#check_indexes_on_tables(t)

	return

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@transaction_commit_on_success
def check_internal_data():

	check_tfields()

	return


def check_tfields():

	TFIELDS = [

		# DATAGROUP
		(1, 'str'       , 'title'),
		(1, 'str'       , 'dependson'),
		(1, 'int'       , 'minperiod'),
		(1, 'int'       , 'shorttries'),
		(1, 'float'     , 'timeout'),
		(1, 'bool'      , 'deleted'),
		(1, 'bool'      , 'suspended'),
		(1, 'str'       , 'suspend_reason'),
		(1, 'timestamp' , 'suspend_time'),
		(1, 'timestamp' , 'suspend_resume_time'),
		(1, 'str'       , 'classification'),
		(1, 'str'       , 'passive_match'),
		(1, 'str'       , 'scheduling_group'),

		# DATASOURCE
		(2, 'str'  , 'title'),
		(2, 'str'  , 'expr'),
		(2, 'int'  , 'order'),
		(2, 'bool' , 'cascade'),
		(2, 'bool' , 'save_value'),
		(2, 'bool' , 'gauge'),
		(2, 'float', 'min_val'),
		(2, 'float', 'max_val'),
		(2, 'str'  , 'storage_spec'),
		(2, 'str'  , 'classification'),
		(2, 'bool' , 'deleted'),

		# CONDITION
		(3, 'str'  , 'title'),
		(3, 'bool' , 'primary'),
		(3, 'str'  , 'dependson'),

		(3, 'int'  , 'order'),
		(3, 'bool' , 'cascade'),

		(3, 'str'  , 'expr'),
		(3, 'int'  , 'max_tries'),
		(3, 'int'  , 'uncheckable_fallback'),
		(3, 'str'  , 'statuschange_action'),
		(3, 'float', 'priority'),
		(3, 'str'  , 'classification'),

		(3, 'str'  , 'email'),       
		(3, 'str'  , 'upemail'),     
		(3, 'str'  , 'msg_downsubj'),
		(3, 'str'  , 'msg_downbody'),
		(3, 'str'  , 'msg_upsubj'),  
		(3, 'str'  , 'msg_upbody'),  
		(3, 'bool' , 'deleted'),

		# TIMEGRAPH
		(4, 'str' , 'title'),
		(4, 'bool', 'primary'),
		(4, 'bool', 'stacked'),
		(4, 'str' , 'ylegend'),
		(4, 'bool', 'deleted'),
		(4, 'str' , 'classification'),

		# TIMEGRAPHSERIE
		(5, 'str' , 'color'),
		(5, 'int' , 'render_order'),
		(5, 'str' , 'variable_value'),
		(5, 'str' , 'computed_value_expr'),
		(5, 'str' , 'legend'),
		(5, 'str' , 'style'),
		(5, 'bool', 'deleted'),
	]

	valid = set()

	for fclass, vtype, name in TFIELDS:
		f, created = TField.objects.get_or_create(fclass=fclass, vtype=vtype, name=name)
		if created:
			print "Added missing field %s:%s:%s" % (fclass, vtype, name)
		f.save()

		valid.add( "%s_%s_%s" % (fclass, vtype, name) )

	for tf in TField.objects.all():
		k = "%s_%s_%s" % (tf.fclass, tf.vtype, tf.name)
		if k not in valid:
			print "Deleted old tfield '%s'" % (k)
			tf.delete()

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

class Command(SanetBaseCommand):

	help = "Perform maintenance operations on Sanet installation."

	#option_list = SanetBaseCommand.option_list + (
	#	make_option('--adminpass'             , action='store', type='string', dest='adminpass'            , default=None, help='Default admin pass'),
	#	make_option('--defaulttenant-name'    , action='store', type='string', dest='defaulttenantname'    , default=None, help='Default Tenant Name'),
	#	make_option('--defaulttenant-longname', action='store', type='string', dest='defaulttenantlongname', default=None, help='Default Tenant Long Name'),
	#	make_option('--defaulttenant-timezone', action='store', type='string', dest='defaulttenanttimezone', default=None, help='Default Tenant Timezone'),
	#	make_option('-l', '--log-level'       , action='store', type='string', dest='log_level'            , default=None, help='Log Level'),
	#)

	def add_arguments(self, parser):

		super(Command, self).add_arguments(parser)

		parser.add_argument('--adminpass'             , action='store', type=str, dest='adminpass'            , default=None, help='Default admin pass'),
		parser.add_argument('--defaulttenant-name'    , action='store', type=str, dest='defaulttenantname'    , default=None, help='Default Tenant Name'),
		parser.add_argument('--defaulttenant-longname', action='store', type=str, dest='defaulttenantlongname', default=None, help='Default Tenant Long Name'),
		parser.add_argument('--defaulttenant-timezone', action='store', type=str, dest='defaulttenanttimezone', default=None, help='Default Tenant Timezone'),
		parser.add_argument('-l', '--log-level'       , action='store', type=str, dest='log_level'            , default=None, help='Log Level'),

	def handle(self, *args, **options):
		
		logger = logging.getLogger('sanet.global')
		
		check_base_dirs(logger)
		
		if options['log_level']:
			l = int(options['log_level'])
			logger.setLevel( l )
			for handler in logger.handlers:
				handler.setLevel( l )
		
		#
		# BEFORE EVERYTHING 
		#
		# Verify system is configured properly for Sanet.
		#
		#
		check_system_settings()

		# Be sure nothing is pending with the DB
		connection.close()


		#
		# STEP 1: update db schema if necessary
		#

		#print term_green("STEP 1")
		call_command('check_schema', interactive=False, verbosity=True, show_traceback=True)

		# Be sure nothing is pending with the DB
		connection.close()

		#
		# STEP 2: Check data consistency
		#
		#print term_green("STEP 2")
		check_internal_data()

		check_default_groups()
		check_default_users(options)
		check_default_auth()

		check_missing_profiles()

		sanet_init_subapp.send(None)

		check_default_tenant(options)

		check_every_tenant()

		cleanup_every_tenant()

		#
		# STEP 3: notify sub apps/plugins 
		#
		#print term_green("STEP 3")
		sanet_post_check.send(None)

		#
		# FINISH
		#
		msg = "Sanet update completed successfully."
		logger.info(msg)
		
		print term_green(msg)

		return

