# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import time
import datetime
import logging

from sanet_common.timeutils import datetime_to_timestamp

from django.db import models
#from django.contrib.contenttypes import generic     # removed in 1.11
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.auth.models import User
from django.dispatch import receiver

from core.models import sanet_init_subapp
from core.models import sanet_check_tenant
from core.models import sanet_cleanup_tenant
from core.models import sanet_post_check
from core.models import CapabilityException
from core.models import Tenant, MonitoredElement, Condition, UUIDKeyModel, TenantDependentModel

from tags.models import Tag
from tags.models import TAG_FILTER

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class UserNote(UUIDKeyModel, TenantDependentModel):

	# uuid         =
	# tenant       = 

	content_type   = models.ForeignKey(ContentType, related_name="%(class)s_usernotes")
	object_pk      = models.TextField()
	content_object = GenericForeignKey(ct_field="content_type", fk_field="object_pk")	

	user           = models.ForeignKey(User, blank=True, null=True, related_name="%(class)s_notes")
	user_name      = models.CharField(max_length=30, blank=True)

	submit_date    = models.DateTimeField(null=False, default=None)

	text           = models.TextField(null=False, blank=False)

	@property
	def submit_date_utcts(self):
		return datetime_to_timestamp(self.submit_date)

	@property
	def isodate(self):
		return self.submit_date.strftime("%Y-%m-%d %H:%M")

	@staticmethod
	def can_edit_element_notes(tenant, user, element):
		r = True
		if isinstance(element, Tenant):
			r = r and "notes.note.tenant.edit"  in user.get_capabilities(tenant)
		elif isinstance(element, MonitoredElement):
			r = r and 'notes.note.element.edit' in user.get_capabilities(tenant)
			r = r and element.can_write(user) 
		elif isinstance(element, Tag):
			r = r and 'notes.note.tag.edit'     in user.get_capabilities(tenant)
			r = r and element.can_write(user) 
		return r

	@staticmethod
	def assert_edit_element_notes(tenant, user, element):

		if not UserNote.can_edit_element_notes(tenant, user, element):
			raise CapabilityException("User '%s' can not edit notes on element '%s'" % (user.username, element))

		return True

	class Meta:
		db_table = "note"
		unique_together = (
		)

#------------------------------------------------------------------------------#
#                                                                              #
# REGISTER NEW METHODS                                                         #
#                                                                              #
#                                                                              #
# Usiamo il metodo "monkey patch" per estendere le classi del core di sanet    #
# con nuove funzionalita'.                                                     #
#                                                                              #
# Per usare i metodi "aggiunti" bisogna importare anche questo modulo.         #
#                                                                              #
#------------------------------------------------------------------------------#

def __get_monitoredelement_notes(self):

	resource_id    = getattr(self, 'uuid')

	return UserNote.objects.filter(object_pk=resource_id)

MonitoredElement.add_to_class('notes', __get_monitoredelement_notes)

def __get_monitoredelement_allnotes(self):

	ids    = [ getattr(self, 'uuid') , ]

	ids.extend( [ v[0] for v in self.get_all_sub_elements().values_list('uuid') ] )

	return UserNote.objects.filter(object_pk__in=ids)

MonitoredElement.add_to_class('allnotes', __get_monitoredelement_allnotes)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def __get_tag_notes(self):
	return UserNote.objects.filter(object_pk=self.uuid)

Tag.add_to_class('notes', __get_tag_notes)

#------------------------------------------------------------------------------#

def __get_tenant_notes(self):
	return UserNote.objects.all()

Tenant.add_to_class('notes', __get_tenant_notes)

#------------------------------------------------------------------------------#

def __get_condition_notes(self):
	return UserNote.objects.filter(object_pk=self.uuid)

Condition.add_to_class('notes', __get_condition_notes)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def __get_tenant_notes_by_user(self, user):

	elements = UserNote.objects.all()

	if user is None or user.is_superuser:
		return elements

	elements = TAG_FILTER(self, user, elements, 'note', 'object_pk', user_field='user_id')

	return elements

Tenant.add_to_class('notes_by_user', __get_tenant_notes_by_user)

#------------------------------------------------------------------------------#

def __get_node_notes(self, user):

	uuids = [ self.uuid ]

	#uuids += [ sube.uuid for sube in self.get_sub_elements_by_user(user).all() ]
	uuids += [ sube[0] for sube in self.get_sub_elements_by_user(user).all().values_list('uuid') ]

	elements = UserNote.objects.filter(object_pk__in=uuids)

	return elements

MonitoredElement.add_to_class('notes_by_user', __get_node_notes)

#------------------------------------------------------------------------------#

def __get_condition_notes(self, user):

	uuids = [ self.uuid ]

	#uuids += [ sube.uuid for sube in self.get_sub_elements_by_user(user).all() ]

	elements = UserNote.objects.filter(object_pk__in=uuids)

	return elements

Condition.add_to_class('notes_by_user', __get_condition_notes)

#------------------------------------------------------------------------------#
# SANET SIGNALS                                                                #
#------------------------------------------------------------------------------#

@receiver(sanet_init_subapp)
def _sanet_init_subapp_notes(sender, **kwargs):

	logger = logging.getLogger('sanet.global')

	logger.info(  "[notes] Init" )

@receiver(sanet_check_tenant)
def _sanet_check_tenant_notes(sender, **kwargs):

	logger = logging.getLogger('sanet.global')

	t = kwargs['instance']

	logger.info(  "[notes] Check tenant %s" % (t)  )

@receiver(sanet_cleanup_tenant)
def _sanet_cleanup_tenant_notes(sender, **kwargs):

	logger = logging.getLogger('sanet.global')

	t = kwargs['instance']

	logger.info(  "[notes] Cleanup tenant %s" % (t) )

@receiver(sanet_post_check)
def _sanet_post_check_notes(sender, **kwargs):

	logger = logging.getLogger('sanet.global')

	logger.info(  "[notes] Post check" )

#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
# WEBUI SIGNALS                                                                #
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#

from webui.models import webui_load_page_extension

@receiver(webui_load_page_extension)
def _webui_load_page_extension_dummy(sender, **kwargs):

	page = kwargs['page']
	
	from notes.web import NotesPageExtension

	return NotesPageExtension(page)

