# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import traceback

from core.compatibility import reverse

from core.models import *
from core.models import Tenant

from webui.decorators import ajax_tenant_dependent

from core.models import CapabilityException

from webui.api.response import HttpJsonSuccessResponse, HttpJsonErrorResponse

from notes.models import UserNote
from notes.api.base import usernotes_add
from notes.api.base import usernotes_delete

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

ENTITY_TYPE_TO_CLASS = {
	'tenant'   : Tenant,
	'node'     : MonitoredNode,
	'interface': MonitoredIface,
	'storage'  : MonitoredStorage,
	'service'  : MonitoredService,
	'device'   : MonitoredDevice,
	'condition': Condition,
}

CLASS_TO_ENTITY_TYPE = dict([(v,k) for k,v in ENTITY_TYPE_TO_CLASS.items()])

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

@ajax_tenant_dependent
def get_notes(request, entity_type, entity_id):

	user   = request.user
	tenant = request.tenant

	model = ENTITY_TYPE_TO_CLASS[ entity_type ]
	resource = model.objects.get(uuid=entity_id)

	data = _get_notes(tenant, user, resource )

	return HttpJsonSuccessResponse(data=data)

def _get_notes(tenant, user, resource ):

	main_entity_id   = resource.uuid
	main_entity_type = CLASS_TO_ENTITY_TYPE[resource.__class__]

	notes = []
	for n in resource.notes_by_user(user).order_by('-submit_date'):

		o = n.content_object
		
		#
		# Dati della nota
		#
		if o is not None:
			entity_id   = o.uuid
			entity_type = CLASS_TO_ENTITY_TYPE[o.__class__]
			elem        = o.name
			url_elem    = reverse('resource_page', kwargs={'rt':entity_type, 'uuid': entity_id})
		else:
			entity_id = ''
			entity_type = None
			elem        = None
			url_elem    = ''

		try:
			if n.user:
				username = n.user.username
			elif n.user_name:
				username = n.user_name
			else:
				username = ''
		except User.DoesNotExist, e:
			username = ''

		#
		# Permessi
		#
		
		# L'utente puo' "cancellare" la singola nota solo se ha i permessi di scrittura sulla risorsa associata
		# Se la nota e' associata al Tenant, non ha senso controllare qui i permeessi qui perche'
		# questo punto del codice e' raggiungibile solo se l'utente ha accesso al tenant.
		if isinstance(user, Tenant):
			can_write = o.can_write(user)
		else:
			can_write = True

		notes.append({
			'id'      : n.uuid,
			'date'    : n.submit_date_utcts,
			'user'    : username, 
			'elem'    : elem,
			'url_elem': url_elem,
			'text'    : n.text,
			
			'can_edit': can_write,
		})


	#
	# Permessi di modifica delle note sulla risorsa
	#
	can_edit = UserNote.can_edit_element_notes(tenant, user, resource)

	#
	# Dati finali
	#
	data = {
		'meta': {
			'element_id': resource.uuid,
			
			'can_edit' : can_edit,

			'urls': {
				#
				# FIX: 
				#
				'refresh'   : reverse('notes.views.get_notes'  , kwargs={'entity_type':main_entity_type, 'entity_id': main_entity_id}) if can_edit else None,
				'ops_add'   : reverse('notes.views.add_note'   , kwargs={'entity_type':main_entity_type, 'entity_id': main_entity_id}) if can_edit else None,
				'ops_remove': reverse('notes.views.remove_note')                                                                       if can_edit else None, 
			},
			
			'visible'   : 0,      # FIX: DISMESSO?
		},
		'content': notes
	}

	return data

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

@ajax_tenant_dependent
def add_note(request, entity_type, entity_id):

	tenant = request.tenant
	user   = request.user

	text   = request.GET_OR_POST()['text']

	try:
		model = ENTITY_TYPE_TO_CLASS[ entity_type ]
		resource = model.objects.get(uuid=entity_id)

		note = usernotes_add(resource, user, text, action_user=user)

		n = note
		entity_id   = resource.uuid
		entity_type = CLASS_TO_ENTITY_TYPE[resource.__class__]
		data = {
			'id'      : n.uuid,
			'date'    :  int(n.submit_date.strftime('%s')), 
			'user'    : n.user.username or n.username or user_name or '', 
			'elem'    : resource.name,
			'url_elem': reverse('resource_page', kwargs={'rt':entity_type, 'uuid': entity_id}),
			'text'    : n.text,
		}

		return HttpJsonSuccessResponse('Added note %s' % (note.uuid), data=data)

	except CapabilityException, e:
		return HttpJsonErrorResponse(str(e))

	except Exception, e:
		return HttpJsonErrorResponse('Error while  adding note: %s' % (str(e)))

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#


@ajax_tenant_dependent
def remove_note(request):

	user    = request.user
	tenant  = request.tenant

	note_id = request.GET_OR_POST()['noteid']

	try:
		note = UserNote.objects.get(tenant=tenant, uuid=note_id)

		usernotes_delete(note, user=user)

		data = {
			'uuid': note_id
		}

		return HttpJsonSuccessResponse('Note removed', data=data)

	except CapabilityException, e:
		return HttpJsonErrorResponse(str(e))

	except Exception, e:
		return HttpJsonErrorResponse('Error while deleting note %s: %s' % (note_id, str(e)))
    
