# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.


import time
import datetime
import json
import pprint

from django.db.models import Q
from django.http import HttpResponse, HttpResponseNotFound

from core.models import DataSource
from core.models import DataGroup

from core.api.cache import init_cache_manager_from_conf
from core.api.datasource import DataSourceStoreException
from core.api.datasource import init_datasource_store_from_conf
from core.api.utils import get_ds_avg_in_timespec
from core.api.utils import get_ds_max_in_timespec
from core.api.utils import get_ds_min_in_timespec

from tags.models import Tag, TagTree

from webui.api.types import WEB_ELEMENT_TYPE_TO_MODEL

from rest_api.decorators import rest_api_enabled
from rest_api.http_responses import HttpResponseUnprocessableEntity
from rest_api.http_responses import HttpResponseNotImplemented

#------------------------------------------------------------------------------#
# COSTANTI                                                                     #
#------------------------------------------------------------------------------#

GLOBAL_PARAMS = [ 
	'sanet3auth',
	'username', 
	'tenantid', 
]


# FIX:
# I tipi di risorse disponibili dalle API REST dovrebbero essere
# gli stessi dell'interfaccia WEB (WEB_ELEMENT_TYPE_TO_MODEL)
# ma ci sono tipi di risorse che l'interfaccia web non gestisce.
#
REST_RESOURCES_MODELS = {}

REST_RESOURCES_MODELS.update(WEB_ELEMENT_TYPE_TO_MODEL)

REST_RESOURCES_MODELS.update({
	'tagtree': TagTree,
})

#------------------------------------------------------------------------------#
# Funzioni di comodo                                                           #
#------------------------------------------------------------------------------#

def extract_get_params(request):
	"""
	Estrae i parametri in GET della request che non sono parametri globali
	dell'interfaccia REST.
	"""
	return dict([ (name,v) for name, v in request.GET.items() if name not in GLOBAL_PARAMS])

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

@rest_api_enabled
def resource_list(request, resource_type):

	api_ts_start = time.time()

	user   = request.user
	tenant = request.tenant

	tenant.db_set_current_schema()

	cache_manager = init_cache_manager_from_conf()

	#
	# All GET params are filters (excluded global params)
	#
	filter_params = extract_get_params(request)

	try:
		#
		# "rest_api/nodes" -> "node"
		#
		resource_type = resource_type[:-1]

		#
		# Load Model
		#
		elem_object = REST_RESOURCES_MODELS[resource_type]

	except (AttributeError, KeyError), e:

		if resource_type in REST_RESOURCES_MODELS:
			return HttpResponseNotImplemented('API for resource ' + resource_type + ' currently not implemented.')
		else:
			return HttpResponseUnprocessableEntity('Unknown resource name')

	#
	# Load elements
	#
	elements = elem_object.get_visible_by_user(tenant, user)

	#
	# Apply filter
	#
	if filter_params:
		elements = elements.filter(**filter_params)

	result = []

	for elem in elements:

		#
		# Load dictionary
		#
		if hasattr(elem, 'load_fields_from_cache'):
			elem.load_fields_from_cache(cache_manager=cache_manager)

		elem_dict = elem.get_dict()

		#
		# Common field useful for REST API
		#
		elem_dict.update({
			'path': elem.path
		})

		result.append( elem_dict )


	response = {}

	response['elapsed_time'] = time.time() - api_ts_start
	response['operation'] = 'fetch_all_' + resource_type

	response['result'] = result

	return HttpResponse(json.dumps(response))

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

@rest_api_enabled
def get_node_info(request, uuid_or_path):

	api_ts_start = time.time()

	user = request.user
	tenant = request.tenant

	tenant.db_set_current_schema()

	#
	# LOAD
	#
	try:
		elem_object = WEB_ELEMENT_TYPE_TO_MODEL['node']
		elem = elem_object.get_visible_by_user(tenant, user).get(Q(uuid=uuid_or_path)|Q(name=uuid_or_path))
	except elem_object.DoesNotExist, e:
		return HttpResponseNotFound('Node not found')

	#
	# PROCESS
	#
	cache_manager = init_cache_manager_from_conf()

	elem.load_fields_from_cache(cache_manager=cache_manager)
	elem_dict = elem.get_dict()
	
	elem_dict.update({'path': elem.calculate_path()})

	#
	# Resposne
	#
	response = {}

	response['elapsed_time'] = time.time() - api_ts_start
	response['operation'] = 'get_node_info'

	response['result'] = elem_dict


	return HttpResponse(json.dumps(response))


@rest_api_enabled
def get_primary_condition_status_change(request, uuid_or_path):

	api_ts_start = time.time()

	user = request.user
	tenant = request.tenant

	tenant.db_set_current_schema()

	ts_start = int(request.GET.get('ts_start', api_ts_start - 60*60*8))
	ts_end   = int(request.GET.get('ts_end'  , api_ts_start))

	#
	# LOAD
	#
	try:
		elem_object = WEB_ELEMENT_TYPE_TO_MODEL['node']
		elem = elem_object.get_visible_by_user(tenant, user).get(Q(uuid=uuid_or_path)|Q(name=uuid_or_path))
	except elem_object.DoesNotExist, e:
		return HttpResponseNotFound('Node not found')

	#
	# PROCESS
	#
	condition = elem.get_primary_condition()
	if condition is None:
		return HttpResponseNotFound('Node ' + elem.name + ' has no primary condition.')

	result = condition_stat_changes(condition, ts_start, ts_end)

	#
	# Resposne
	#
	response = {}

	response['elapsed_time'] = time.time() - api_ts_start
	response['operation'] = 'get_primary_condition_status_change'

	response['result'] = result

	response['condition_path'] = condition.calculate_path()
	response['element_name'] = elem.name
	response['element_uuid'] = elem.uuid
	response['element_path'] = elem.calculate_path()

	return HttpResponse(json.dumps(response))


def condition_stat_changes(condition, ts_start, ts_end):

	changes = condition.logchanges_sequence(ts_start, ts_end)

	tot_by_status = []

	for s, e, tot, status in changes:
		tot_by_status.append((s, e, status))

	return tot_by_status

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

@rest_api_enabled
def get_interface_info(request, uuid_or_path):

	api_ts_start = time.time()

	user   = request.user
	tenant = request.tenant
	
	tenant.db_set_current_schema()

	operation = 'get_interface_info'

	#
	# LOAD
	#

	try:
		elem_object = WEB_ELEMENT_TYPE_TO_MODEL['interface']

		if uuid_or_path.find(':') > -1:
			node_name, iface_name = uuid_or_path.split(':')

			elem = elem_object.get_visible_by_user(tenant, user).get(name=iface_name, parent__name=node_name)
		else:
			elem = elem_object.get_visible_by_user(tenant, user).get(uuid=uuid_or_path)

	except elem_object.DoesNotExist, e:
		return HttpResponseNotFound('Node or interface not found.')

	#
	# PROCESS
	#
	cache_manager = init_cache_manager_from_conf()

	elem.load_fields_from_cache(cache_manager=cache_manager)
	elem_dict = elem.get_dict()

	elem_dict.update({'path': elem.calculate_path()})

	#
	# Resposne
	#
	response = {}

	response['elapsed_time'] = time.time() - api_ts_start
	response['operation']    = operation

	response['result']       = elem_dict

	response['element_info'] = elem_dict   # perche' questa chiave? dovremmo gia' usare "result"

	return HttpResponse(json.dumps(response))


@rest_api_enabled
def get_traffic(request, uuid_or_path):

	api_ts_start = time.time()

	user   = request.user
	tenant = request.tenant

	tenant.db_set_current_schema()

	operation = 'get_traffic'

	ts_start = int(request.GET.get('ts_start', api_ts_start-60*60*8))
	ts_end = int(request.GET.get('ts_end', api_ts_start))

	#
	# LOAD
	#
	
	try:
		elem_object = WEB_ELEMENT_TYPE_TO_MODEL['interface']

		if uuid_or_path.find(':') > -1:
			node_name, iface_name = uuid_or_path.split(':')

			elem = elem_object.get_visible_by_user(tenant, user).get(name=iface_name, parent__name=node_name)
		else:
			elem = elem_object.get_visible_by_user(tenant, user).get(uuid=uuid_or_path)

	except elem_object.DoesNotExist, e:
		return HttpResponseNotFound('Node or interface not found.')

	#
	# PROCESS
	#

	try:
		result = interface_traffic_data(elem, ts_start, ts_end, tenant)
	except Exception, e:
		return HttpResponseNotFound(e)


	#
	# Response
	#
	response = {}

	response['elapsed_time'] = time.time() - api_ts_start
	response['operation'] = operation

	response['result'] = result

	response['element_name'] = elem.name
	response['element_path'] = elem.path
	response['element_uuid'] = elem.uuid

	return HttpResponse(json.dumps(response))



def interface_traffic_data(interface, ts_start, ts_end, tenant):

	try:
		datagroup = DataGroup.objects.get(name='iface-traffic', element=interface)

	except DataGroup.DoesNotExist, e:
		raise Exception('There is no check for traffic data on interface ' + interface.path)

	try:
		ds_in = DataSource.objects.get(element=interface, name='in', datagroup=datagroup)
		ds_out = DataSource.objects.get(element=interface, name='out', datagroup=datagroup)

	except DataSource.DoesNotExist, e:
		raise Exception("Can't find correct datasources. There might be a missconfiguration.")

	in_dataseries = datasource_data_series(ds_in, ts_start, ts_end, tenant)
	out_dataseries = datasource_data_series(ds_out, ts_start, ts_end, tenant)

	in_tot = 0
	in_max = 0
	for value in in_dataseries:
		if value is None:
			value = 0

		elif value > in_max:
			in_max = value

		in_tot += value

	in_avg = in_tot/len(in_dataseries)

	out_tot = 0
	out_max = 0
	for value in out_dataseries:
		if value is None:
			value = 0

		elif value > out_max:
			out_max = value

		out_tot += value

	out_avg = out_tot/len(out_dataseries)

	return {
		'in_max': in_max,
		'in_avg': in_avg,

		'out_max': out_max,
		'out_avg': out_avg
	}

def datasource_data_series(datasource, ts_start, ts_end, tenant):

	datasource_store = init_datasource_store_from_conf()

	try:
		data = datasource_store.fetch_values(
			tenant.uuid,
			datasource.uuid,
			ts_start=ts_start,
			ts_end=ts_end,
			resolution=60,
			min_value=datasource.min_val,
			max_value=datasource.max_val
		)
	except DataSourceStoreException, e:
		raise e

	return data['dataseries']


#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#


@rest_api_enabled
def datasource_info(request, uuid_or_path, operation='get_avg_value'):
	return _datasource_info(request, uuid_or_path, operation=operation)

@rest_api_enabled
def get_avg_value(request, uuid_or_path):
	return _datasource_info(request, uuid_or_path, 'get_avg_value')

@rest_api_enabled
def get_max_value(request, uuid_or_path):
	return _datasource_info(request, uuid_or_path, 'get_max_value')

@rest_api_enabled
def get_min_value(request, uuid_or_path):
	return _datasource_info(request, uuid_or_path, 'get_min_value')

def _datasource_info(request, uuid_or_path, operation='get_avg_value'):

	api_ts_start = time.time()

	user      = request.user
	tenant    = request.tenant

	tenant.db_set_current_schema()

	ts_start = int(request.GET.get('ts_start', api_ts_start - 60*60*8))
	ts_end   = int(request.GET.get('ts_end'  , api_ts_start))

	timespec = request.GET.get('timespec', None)

	#
	# LOAD
	#

	try:
		ds = DataSource.objects.get(Q(path=uuid_or_path)|Q(uuid=uuid_or_path))
	except DataSource.DoesNotExist, e:
		return HttpResponseNotFound("Datasource not found. Please check provided uuid: %s" % (uuid_or_path,))

	if not ds.datagroup.assert_permission_read(user):
		return HttpResponseNotFound("Datasource not found. Please check provided uuid: %s" % (uuid_or_path,))

	#
	# PROCESS
	#

	if operation == 'get_avg_value':
		result = get_ds_avg_in_timespec(ds, ts_start, ts_end, timespec)

	elif operation == 'get_max_value':
		result = get_ds_max_in_timespec(ds, ts_start, ts_end, timespec)

	elif operation == 'get_min_value':
		result = get_ds_min_in_timespec(ds, ts_start, ts_end, timespec)

	else:
		return HttpResponseNotImplemented('Request ' + operation + ' is unknown or not currently implemented.')

	#
	# Response
	#

	response = {}

	response['elapsed_time'] = time.time() - api_ts_start
	response['operation'] = operation

	response['result'] = result

	response['datasource_name'] = ds.name
	response['datasource_path'] = ds.path
	response['datasource_uuid'] = ds.uuid

	response['element_name'] = ds.element.name
	response['element_uuid'] = ds.element.uuid

	return HttpResponse(json.dumps(response))

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

@rest_api_enabled
def tag_datasources_get_last_value(request, uuid_or_path):
	"""
	Ultimo valore dei datasource contenuti in un tag
	------------------------------------------------------------------------

	**DESCRIZIONE**

		Elabora tutti i datasource del tag (e sottotag) e restituisce l'ultimo valore registrato.
		
			::
				
				http://<hostname>/sanet/rest_api/tag/datasources/get_last_value?<params>
		
		Esempio:
		
			::
				
				http://<hostname>/sanet/rest_api/tag/geo:/Geografico/datasources/get_last_value?sanet3auth=12345&sum_series=by-name&tag=geo:/Geografico
				http://<hostname>/sanet/rest_api/tag/datasources/get_last_value?sanet3auth=12345&sum_series=by-name&tag=geo:/Geografico

	**INPUT**

		Questi sono i parametri in Get:
		
			* tag (stringa)
			
				UUID o PATH del tag da elaborare.
			
			* sum_by (stringa)
			
				Indica se sommare i valori dei datasource elabrorati e in che modo.
				
				I valori ammessi sono:
				
					"by-name"     Somma i valori dei datasource raggruppando per nome del datasource
					"no"          Restituisce informazioni sull'ultimo valore per tutti i datasource elaborati.
				
				Facoltativo: Si

	**OUTPUT**

		Restituisce un messaggio nel formato:
			
			::
				
				{
					'elapsed_time': <tempo in secondi>
					,
					'errors'      : <numero di anomalie in fase di elaborazione>
					,
					'result'      : <dati della richiesta>
				}
		
		Il contentuto/formato del campo *data* dipende dal valore specificato dai parametri in input.
		
		Quando *sum_series* valore *"no"*, il campo *data* contiene un elenco di tuple:
				
			::
				
				<parent_id, parent_name, element_id, element_name, ds_uuid, ds_name, value>
			
		Quando *sum_series* valore *"by-name"*,il campo *data* contiene un elenco di tuple:
			
			::
				
				< datasource name , total >
	"""

	datasource_store = init_datasource_store_from_conf()

	api_ts_start = time.time()
	errors       = 0

	user      = request.user
	tenant    = request.tenant

	tenant.db_set_current_schema()

	sum_by = request.GET.get('sum_by', None)

	if sum_by:
		if sum_by not in ['name']:
			return HttpResponseNotImplemented("Invalid option value '%s'." % (sum_by))

	ts_end   = time.time()
	ts_start = ts_end - 60*60 * 48;

	#
	# Find Dag
	#
	tag = Tag.objects.get(Q(path=uuid_or_path)|Q(uuid=uuid_or_path))

	#
	# Load all datasources
	#
	datasources_to_process = tag.get_all_datasources().filter(save_value=True).all()

	datasources_to_process = datasources_to_process.select_related('element', 'element__parent')


	#
	# Extract data
	#

	datasources = []
	for ds in datasources_to_process:

		v = None

		try:
			ret = datasource_store.fetch_values(
				tenant.uuid,
				ds.uuid,
				ts_start   = ts_start,
				ts_end     = ts_end,
				resolution = 60,
				min_value  = ds.min_val,
				max_value  = ds.max_val
			)

			data = ret['dataseries']

			data.reverse()

			for v in data:
				if v is not None: 
					break

		except DataSourceStoreException, e:
			errors += 1

		if ds.element.parent_id:
			parent_id   = ds.element.parent_id
			parent_name = ds.element.parent.name
		else:
			parent_id   = None
			parent_name = None


		datasources.append((parent_id, parent_name, ds.element.uuid, ds.element.name, ds.uuid, ds.name, ds.path, v))

	if sum_by == 'name':

		by_name = {}
		for parent_id, parent_name, element_id, element_name, ds_uuid, ds_name, ds_path, v in datasources:

			v = v if v is not None else 0

			tot = by_name.get(ds_name, by_name.setdefault(ds_name, 0))
			tot += v
			by_name[ds_name] = tot

		data = sorted(by_name.items())
	else:
		data = datasources

	#
	# RESPONSE
	#
	response = {
		'elapsed_time': time.time() - api_ts_start,
		'operation'   : 'get_last_value',

		'result'      : data,

		'errors'      : "%s datasources not loaded" % (errors),
	}

	return HttpResponse(json.dumps(response))





@rest_api_enabled
def tag_datasources_get_last_year_values(request, uuid_or_path):
	"""
	Valori mensili dell'anno precedente dei datasource contenuti in un tag
	------------------------------------------------------------------------
	
	**DESCRIZIONE**
	
		Elabora tutti i datasource del tag (e sottotag) e restituisce il valore registrato all'inizio di ogni mese dell'anno.
		
			::
				
				http://<hostname>/sanet/rest_api/tag/<uuid or path>/datasources/get_last_year_values?<params>
		
		Esempio:
		
			::
				
				http://<hostname>/sanet/rest_api/tag/geo:/Geografico/datasources/get_last_year_values?sanet3auth=12345
				http://<hostname>/sanet/rest_api/tag/422ce2179631474eb34a27cff467102a/datasources/get_last_year_values?sanet3auth=12345
	
	**INPUT**
	
		Questi sono i parametri all'interno della URL:
		
			* <uuid or path> (stringa)
			
				UUID o PATH del tag da elaborare.
		
		Questi sono i parametri in GET:
		
			* valid_time_window (intero)
			
				In presenza di valori *nulli* registrati, numero di secondi di tolleranza entro il quale cercare il prima volore valido.
				Se non vengono trovati valori non *nulli* all'interno del periodo di tolleranza, il valore registrato per quel meso sara' None.
	
	**OUTPUT**
	
		Restituisce un messaggio nel formato:
		
			::
				
				{
					'elapsed_time': <tempo in secondi>
					,
					'errors'      : <numero di anomalie in fase di elaborazione>
					,
					'result'      : <dati della richiesta>
				}
		
		Quando *sum_series* valore *"no"*, il campo *data* contiene un elenco di tuple:
		
			::
				
				<parent_id, parent_name, element_id, element_name, ds_uuid, ds_name, date, ts, value>
	
	**Esempi**
	
		::
		        
		        {'elapsed_time': 0.014796972274780273,
		         'errors': '0 datasources not loaded',
		         'operation': 'get_last_year_values',
		         'result': [('79b977c52f6e4c84aa496ba942ba70d8', 'localhost', 'ec6e4cb188b64ebcbce204d14380de4d', 'eth0', 'ccd306b551a94e9ca67673d7a96f89fb', 'in', 'localhost:eth0;;iface-traffic;in', '2016-01-01', 1451602800.0, None),
		                    ('79b977c52f6e4c84aa496ba942ba70d8', 'localhost', 'ec6e4cb188b64ebcbce204d14380de4d', 'eth0', 'ccd306b551a94e9ca67673d7a96f89fb', 'in', 'localhost:eth0;;iface-traffic;in', '2016-02-01', 1454281200.0, None),
		                    ('79b977c52f6e4c84aa496ba942ba70d8', 'localhost', 'ec6e4cb188b64ebcbce204d14380de4d', 'eth0', 'ccd306b551a94e9ca67673d7a96f89fb', 'in', 'localhost:eth0;;iface-traffic;in', '2016-03-01', 1456786800.0, None),
		                    ('79b977c52f6e4c84aa496ba942ba70d8', 'localhost', 'ec6e4cb188b64ebcbce204d14380de4d', 'eth0', 'ccd306b551a94e9ca67673d7a96f89fb', 'in', 'localhost:eth0;;iface-traffic;in', '2016-04-01', 1459461600.0, None),
		                    ('79b977c52f6e4c84aa496ba942ba70d8', 'localhost', 'ec6e4cb188b64ebcbce204d14380de4d', 'eth0', 'ccd306b551a94e9ca67673d7a96f89fb', 'in', 'localhost:eth0;;iface-traffic;in', '2016-05-01', 1462053600.0, None),
		                    ('79b977c52f6e4c84aa496ba942ba70d8', 'localhost', 'ec6e4cb188b64ebcbce204d14380de4d', 'eth0', 'ccd306b551a94e9ca67673d7a96f89fb', 'in', 'localhost:eth0;;iface-traffic;in', '2016-06-01', 1464732000.0, None),
		                    ('79b977c52f6e4c84aa496ba942ba70d8', 'localhost', 'ec6e4cb188b64ebcbce204d14380de4d', 'eth0', 'ccd306b551a94e9ca67673d7a96f89fb', 'in', 'localhost:eth0;;iface-traffic;in', '2016-07-01', 1467324000.0, None),
		                    ('79b977c52f6e4c84aa496ba942ba70d8', 'localhost', 'ec6e4cb188b64ebcbce204d14380de4d', 'eth0', 'ccd306b551a94e9ca67673d7a96f89fb', 'in', 'localhost:eth0;;iface-traffic;in', '2016-08-01', 1470002400.0, None),
		                    ('79b977c52f6e4c84aa496ba942ba70d8', 'localhost', 'ec6e4cb188b64ebcbce204d14380de4d', 'eth0', 'ccd306b551a94e9ca67673d7a96f89fb', 'in', 'localhost:eth0;;iface-traffic;in', '2016-09-01', 1472680800.0, None),
		                    ('79b977c52f6e4c84aa496ba942ba70d8', 'localhost', 'ec6e4cb188b64ebcbce204d14380de4d', 'eth0', 'ccd306b551a94e9ca67673d7a96f89fb', 'in', 'localhost:eth0;;iface-traffic;in', '2016-10-01', 1475272800.0, None),
		                    ('79b977c52f6e4c84aa496ba942ba70d8', 'localhost', 'ec6e4cb188b64ebcbce204d14380de4d', 'eth0', 'ccd306b551a94e9ca67673d7a96f89fb', 'in', 'localhost:eth0;;iface-traffic;in', '2016-11-01', 1477954800.0, None),
		                    ('79b977c52f6e4c84aa496ba942ba70d8', 'localhost', 'ec6e4cb188b64ebcbce204d14380de4d', 'eth0', 'ccd306b551a94e9ca67673d7a96f89fb', 'in', 'localhost:eth0;;iface-traffic;in', '2016-12-01', 1480546800.0, None),
		                    ('79b977c52f6e4c84aa496ba942ba70d8', 'localhost', 'ec6e4cb188b64ebcbce204d14380de4d', 'eth0', 'a66f9268b838471d98d5b579ad4f9c0c', 'out', 'localhost:eth0;;iface-traffic;out', '2016-01-01', 1451602800.0, None),
		                    ('79b977c52f6e4c84aa496ba942ba70d8', 'localhost', 'ec6e4cb188b64ebcbce204d14380de4d', 'eth0', 'a66f9268b838471d98d5b579ad4f9c0c', 'out', 'localhost:eth0;;iface-traffic;out', '2016-02-01', 1454281200.0, None),
		                    ('79b977c52f6e4c84aa496ba942ba70d8', 'localhost', 'ec6e4cb188b64ebcbce204d14380de4d', 'eth0', 'a66f9268b838471d98d5b579ad4f9c0c', 'out', 'localhost:eth0;;iface-traffic;out', '2016-03-01', 1456786800.0, None),
		                    ('79b977c52f6e4c84aa496ba942ba70d8', 'localhost', 'ec6e4cb188b64ebcbce204d14380de4d', 'eth0', 'a66f9268b838471d98d5b579ad4f9c0c', 'out', 'localhost:eth0;;iface-traffic;out', '2016-04-01', 1459461600.0, None),
		                    ('79b977c52f6e4c84aa496ba942ba70d8', 'localhost', 'ec6e4cb188b64ebcbce204d14380de4d', 'eth0', 'a66f9268b838471d98d5b579ad4f9c0c', 'out', 'localhost:eth0;;iface-traffic;out', '2016-05-01', 1462053600.0, None),
		                    ('79b977c52f6e4c84aa496ba942ba70d8', 'localhost', 'ec6e4cb188b64ebcbce204d14380de4d', 'eth0', 'a66f9268b838471d98d5b579ad4f9c0c', 'out', 'localhost:eth0;;iface-traffic;out', '2016-06-01', 1464732000.0, None),
		                    ('79b977c52f6e4c84aa496ba942ba70d8', 'localhost', 'ec6e4cb188b64ebcbce204d14380de4d', 'eth0', 'a66f9268b838471d98d5b579ad4f9c0c', 'out', 'localhost:eth0;;iface-traffic;out', '2016-07-01', 1467324000.0, None),
		                    ('79b977c52f6e4c84aa496ba942ba70d8', 'localhost', 'ec6e4cb188b64ebcbce204d14380de4d', 'eth0', 'a66f9268b838471d98d5b579ad4f9c0c', 'out', 'localhost:eth0;;iface-traffic;out', '2016-08-01', 1470002400.0, None),
		                    ('79b977c52f6e4c84aa496ba942ba70d8', 'localhost', 'ec6e4cb188b64ebcbce204d14380de4d', 'eth0', 'a66f9268b838471d98d5b579ad4f9c0c', 'out', 'localhost:eth0;;iface-traffic;out', '2016-09-01', 1472680800.0, None),
		                    ('79b977c52f6e4c84aa496ba942ba70d8', 'localhost', 'ec6e4cb188b64ebcbce204d14380de4d', 'eth0', 'a66f9268b838471d98d5b579ad4f9c0c', 'out', 'localhost:eth0;;iface-traffic;out', '2016-10-01', 1475272800.0, None),
		                    ('79b977c52f6e4c84aa496ba942ba70d8', 'localhost', 'ec6e4cb188b64ebcbce204d14380de4d', 'eth0', 'a66f9268b838471d98d5b579ad4f9c0c', 'out', 'localhost:eth0;;iface-traffic;out', '2016-11-01', 1477954800.0, None),
		                    ('79b977c52f6e4c84aa496ba942ba70d8', 'localhost', 'ec6e4cb188b64ebcbce204d14380de4d', 'eth0', 'a66f9268b838471d98d5b579ad4f9c0c', 'out', 'localhost:eth0;;iface-traffic;out', '2016-12-01', 1480546800.0, None)]}
	"""

	datasource_store = init_datasource_store_from_conf()


	api_ts_start = time.time()
	errors       = 0

	user      = request.user
	tenant    = request.tenant

	tenant.db_set_current_schema()

	#
	# Parametri richiesta
	#
	valid_time_window = int(request.GET.get('valid_time_window', 60*60*8))

	#
	# Find Tag
	#
	tag = Tag.objects.get(Q(path=uuid_or_path)|Q(uuid=uuid_or_path))

	#
	# Load all datasources
	#
	datasources_to_process = tag.get_all_datasources().filter(save_value=True).all()

	datasources_to_process = datasources_to_process.select_related('element', 'element__parent', 'datagroup')

	#
	# Extract data
	#

	year = datetime.datetime.now().year - 1
	#
	# FOR ALL MONTHS
	#

	datasources = []

	for ds in datasources_to_process:

		for month in xrange(1, 12+1):

			date_start   = datetime.date( year, month    , 1)

			ts_start = time.mktime(date_start.timetuple())
			ts_end   = ts_start + valid_time_window

			ts_start_effective = ts_start

			v = None

			try:
				ret = datasource_store.fetch_values(
					tenant.uuid,
					ds.uuid,
					ts_start   = ts_start,
					ts_end     = ts_end,
					resolution = ds.datagroup.minperiod,
					min_value  = ds.min_val,
					max_value  = ds.max_val
				)

				data = ret['dataseries']

				# Search for the first not null value
				step = ret['scalex'][2]
				while data and ts_start_effective < ts_end:
					v = data.pop(0)
					if v is not None:
						break

					ts_start_effective += step

				if not data:
					ts_start_effective = ts_start

			except DataSourceStoreException, e:
				errors += 1

			if ds.element.parent_id:
				parent_id   = ds.element.parent_id
				parent_name = ds.element.parent.name
			else:
				parent_id   = None
				parent_name = None

			datasources.append(
				(
				parent_id                      .encode('ascii'), 
				parent_name                    .encode('ascii'), 
				ds.element.uuid                .encode('ascii'), 
				ds.element.name                .encode('ascii'), 
				ds.uuid                        .encode('ascii'), 
				ds.name                        .encode('ascii'), 
				ds.path                        .encode('ascii'), 
				date_start.strftime("%Y-%m-%d").encode('ascii'), 
				ts_start,
				v
				)
			)

	data = datasources

	#
	# RESPONSE
	#
	response = {
		'elapsed_time': time.time() - api_ts_start,
		'operation'   : 'get_last_year_values',

		'result'      : data,

		'errors'      : "%s datasources not loaded" % (errors),
	}
	
	#pprint.pprint(response, width=300)

	return HttpResponse(json.dumps(response))






