# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import time
import datetime
import json
import pprint

from django.http import HttpResponse

from sanet_common.sanetdefs.constants import CSTATE_DN

from core.models import Event, TenantSetting, PriorityLabel
from core.api.cache import init_cache_manager_from_conf

from tags.models import Tag
from tags.models import ElementTag, TagMetaData

from rest_api.decorators import rest_api_enabled

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#
"""    
Questa view espone tutti gli allarmi in corso.

Il formato di output e' una lista di dizionari con i dati associati all'allarme con questi campi:

    {
        "uuid"      : "7eb499dd9fb54f0a87afd351d0c4ce56"                                 *identificativo univoco dell'allarme*
        "ts"        : 1511001368,                                                        *timestamp UTC**di generazione dell'allarme*
        "element": {
             "uuid"       : "e173f52135b845a5bf1792e33ad508ac"                            *identificativo univoco dell'elemento associato all'allarme*
             "path"       : "dm10lx03:rootfs",                                            *identificativo univoco mnemonico dell'elemento associato all'allarme*
             "name"       : "rootfs",                                                     *nome dell'elemento*
             "description": "Root FS",                                                    *descrizione**dell'elemento*
             "type"       : 3,                                                            *tipo dell'elemento (1=nodo,2=interfaccia,3=storage,4=service,5=device)*
             "parent_id"  : "a12937d61c874890914182b6d25ebbb8",                           *identificativo dell'elemento padre (l'uuid del nodo)*
             "parent_name": "dm10lx03",                                                   *identificativo univoco dell'elemento padre (il nome del nodo)*
        },
        "condition": {
            "uuid"            : "050db78a671e48358ca266a92e9c9682"                       *identificativo univoco della condition che ha generato l'allarme*
            "path"            : "dm10lx03:rootfs;storage;storage;check-storageperc",     *identificativo mnemonico*
            "name"            : "check-storageperc",                                     *nome della **condition*
            "description"     : "Check occupazione filesystem root",                     *descrizione**della condition*
            "classification"  : "system.storage.usage",                                  *classificazione della **condition*
            "priority"        : 60,                                                      *importanza/priorita' associata alla **condition*
            "priority_level"  : "medium",                                                *classe di importanza/priorita' associata alla **condition*
            "statuslastchange": 1511001368,                                              *timestamp ***UTC**della *variazione di stato precedente della **condition*
        },
                   #"datasources": {                                                                 *opt. serie dati utilizzate per generare l'allarme*
                   #    "available": {                                                               *nome della serie*
                   #        "id"   : "09298c8603974662bf067f451bca0670",                             *identificativo univoco della serie*
                   #        "value": True                                                            *valore della serie al momento dell'allarme*
                   #    },
                   #    "installed": {
                   #        "id"   : "d3d0f7a84891485b8558c8fdb87e3a20",
                   #        "value": 2199020105728
                   #    },
                   #    "used": {
                   #        "id"   : "460abc40cfac456e8026ed7e13588771",
                   #        "value": 2155118809088
                   #    }
                   #},
        "result": {                                                                      *risultati del check*
            "status"    : 10,                                                            *stato della verifica della condition*
            "laststatus": 20,                                                            *stato precedente*
        },
        "info": {                                                                        *informazioni dell'elemento*
            "ip4s"       : ['10.0.25.201'],
        },
        "tags": [                                                                        *opt. **classificazioni/tag assegnati alla condizione o all'elemento associati all'allarme*
            "dmh:/DMH/Sistemi Critici/3.Labs"
        ],
        "cluster": {                                                                     ***opt. *cluster al quale l'elemento e' associato*
            "name"       : ['VMWare-Produzione'],
        },
        "meta": {                                                                        ****opt. *informazioni addizionali dell'elemento*
            "referenti" : [                                                              *referenti associati all'elemento*
                'Sergio Rizzi <sergio.rizzi@labs.it>', 
                'Tiziano Fogli <tiziano.fogli@labs.it>'
            ]
        },
    }
"""

@rest_api_enabled
def alarm_current(request):

	cache_manager = init_cache_manager_from_conf()

	valid_states = [ CSTATE_DN, ]

	tenant = request.tenant
	tenant.db_set_current_schema()

	user   = request.user

	#
	# Parameters
	#
	priority_level_name = request.REQUEST.get('priority_level', None)

	#
	# Select conditions
	#
	selected_conditions = tenant.get_all_conditions_by_user(user)

	selected_conditions = selected_conditions.select_related('element', 'element__parent')

	selected_conditions = selected_conditions.filter(status__in=valid_states)


	# Filter on priority
	if priority_level_name is not None:
		pl = PriorityLabel.objects.get(label=priority_level_name)
		
		selected_conditions = selected_conditions.filter(priority__range=(pl.min_level, pl.max_level))
	else:
		minimum_priority  = tenant.get_setting(TenantSetting.SETTING_MINIMUM_LEVEL)
		if minimum_priority:
			selected_conditions = selected_conditions.exclude(priority__lt=minimum_priority)

	#
	# Process
	#
	result = []

	for condition in selected_conditions:

		element = condition.element

		#event = Event.objects.filter(condition_path=condition.path)
		events = condition.get_logs()
		
		events = events.filter(status=condition.status, statuslast=condition.statuslast)[0:1]
		
		if len(events) == 0:
			continue

		event = events[0]

		alarm_data = {}
		
		alarm_data['element'] = {
			"uuid"       : element.uuid, 
			"path"       : element.path, 
			"name"       : element.name, 
			"description": element.description,
			"type"       : element.kind, 
			"parent_id"  : element.parent_id,
			"parent_name": element.parent.name if element.parent_id is not None else None, 
		}

		#
		# condition
		#
		alarm_data["condition"] = {
			"uuid"            : condition.uuid,              #"050db78a671e48358ca266a92e9c9682"                       *identificativo univoco della condition che ha generato l'allarme*
			"path"            : condition.path,              #"dm10lx03:rootfs;storage;storage;check-storageperc",     *identificativo mnemonico*
			"name"            : condition.name,              #"check-storageperc",                                     *nome della **condition*
			"description"     : condition.expanded_title(cache_manager=cache_manager),       #"Check occupazione filesystem root",                     *descrizione**della condition*
			"classification"  : condition.classification,    #"system.storage.usage",                                  *classificazione della **condition*
			"priority"        : condition.priority,          #60,                                                      *importanza/priorita' associata alla **condition*
			"priority_level"  : condition.priority_label,    #"medium",                                                *classe di importanza/priorita' associata alla **condition*
			"statuslastchange": condition.statuslastchange,  #1511001368,                                              *timestamp ***UTC**della *variazione di stato precedente della **condition*
		}

		#
		# datasources
		#
		alarm_data["datasources"] = { }

		#
		# Result
		#
		alarm_data["result"] = {
		     "status"    : event.status,    
		     "laststatus": event.statuslast,
		}

		#
		# ip4s
		#
		ips = element.get_node().get_ipv4_addresses() or []
		alarm_data["info"] = {
		     "ip4s"       : ips,
		}

		#
		# Tags
		#
		alarm_data["tags"] = extract_tags(element.kind, element.uuid, condition.uuid)

		#
		# Cluster
		#
		if element.cluster:
			cluster_name = element.cluster.name if element.cluster else None

			alarm_data["cluster"] = {
			     "name"       : [ cluster_name ],
			}

		#
		# Metas
		#
		metas = extract_meta(condition.element_id, condition.uuid) or None

		alarm_data["meta"] = metas

		result.append(alarm_data)

	return HttpResponse(json.dumps(result))

from django.db.models import Q
from tags.models import ELEMENT_KIND_TO_TAG_ELEMENTTYPE
from tags.models import ElementTag

def extract_tags(element_kind, element_id, condition_id):

	e_elementtype = ELEMENT_KIND_TO_TAG_ELEMENTTYPE.get( element_kind , None)
	if e_elementtype == None:
		return []

	c_uuid        = condition_id
	c_elementtype = 'condition'

	tags_name_by_tree = {}
	paths = set()
	
	f  = Q(elementid=element_id, elementtype=e_elementtype)
	f |= Q(elementid=c_uuid    , elementtype=c_elementtype)
	
	for et in ElementTag.objects.filter(f).select_related('tag').all():
		paths.add(et.tag.path)

	return list(paths)

def extract_meta(element_id, condition_id):

	ids = [ element_id, condition_id ]

	tagged = ElementTag.objects.filter(elementid__in=ids)

	tagged = tagged.all()

	tags = [et.tag for et in tagged]

	metas = TagMetaData.objects.filter(tag__in=tags)
	metas = metas.all()

	by_name = {}
	for md in metas:
		l = by_name.get( md.name, by_name.setdefault( md.name, [] ))
		l.append(md.value)

	vals = by_name

	return vals
