# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET3
# SANET3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

#
# Supervisor for sanetd
#

import sys
import os
import getopt
import time
import errno 
import signal
import json

#-----------------------------------------------------------------------------#

import logging
logging._srcfile     = None # dont' collect information about where calls were made from.
logging.logThreads   = 0    # dont' collect threading information.
logging.logProcesses = 0    # dont' collect process information.
import logging.handlers

#-----------------------------------------------------------------------------#
# SHARED LIBRARIES                                                            #
#-----------------------------------------------------------------------------#

from sanet_common.osutils.daemonize import daemonize
from sanet_common.osutils.proctitle import get_proc_title, set_proc_title
from sanet_common.osutils.osutils   import convert_exit_status_to_verb_reason

#-----------------------------------------------------------------------------#
# IMPORT CONFIGURATION FROM DANGO                                             #
#-----------------------------------------------------------------------------#

SANET_INSTALL_DIR=None

from os.path import join, dirname, abspath, exists
if not SANET_INSTALL_DIR:
	SANET_INSTALL_DIR = abspath(os.path.join(os.path.dirname(abspath(__file__)),"..",".." ))

sys.path.append( SANET_INSTALL_DIR )
        
#-----------------------------------------------------------------------------#
# DJANGO SETUP
#-----------------------------------------------------------------------------#

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "conf.settings")

# Added for django 1.11
import django
django.setup()  

#-----------------------------------------------------------------------------#
# LOCAL IMPORTS                                                               #
#-----------------------------------------------------------------------------#

from django.db import connection
from conf import get_configuration

#-----------------------------------------------------------------------------#
# Program Options                                                             #
#-----------------------------------------------------------------------------#

PROGRAM_NAME = 'sanet_localpollers_supervisor'

SUPERVISED_POLLER_SCRIPT = 'server/poller/supervised_poller.py'


#-----------------------------------------------------------------------------#
# Global files                                                                #
#-----------------------------------------------------------------------------#

poller_supervisor_log_filename = PROGRAM_NAME
poller_supervisor_log_file = None
poller_supervisor_pid_file = None

# The main logger object.
log = None

#-----------------------------------------------------------------------------#
# Program Options                                                             #
#-----------------------------------------------------------------------------#

# -h
opt_help = False

# -L log level
opt_loglevel = logging.INFO

log_levels = {
	'DEBUG': logging.DEBUG,
	'INFO' : logging.INFO,
}


#-----------------------------------------------------------------------------#
# MAIN PROGRAM                                                                #
#-----------------------------------------------------------------------------#

def poller_supervisor_main(argv=None):

	global  opt_help, opt_loglevel

	if argv is None:
		argv = sys.argv

	try:
		opts, args = getopt.getopt(sys.argv[1:], "hnL:")
		for o, a in opts:
			if   o in ("-h"):
				opt_help = True
			elif o in ("-L"):
				if log_levels.has_key(a):
					opt_loglevel = log_levels[a]
				else:
					opt_loglevel = int(a)

	except getopt.error, msg:
		print "Error while parsing options: %s" % msg
		print_help()
		sys.exit()
	except ValueError, ex:
		print "Invalid value for input parameter: %s" % ex
		print_help()
		sys.exit()

	if opt_help:
		print_help()
		sys.exit()

	if len(args) != 1:
		print "Wrong number of arguments: %s" % (args)
		print_help()
		sys.exit(-1)

	load_conf()

	cmd = args[0]

	if   cmd == 'start':
		start_poller_supervisor()

	elif cmd == 'stop':
		stop_poller_supervisor()
		
	elif cmd == 'test':
		show_poller_launch_commands()
	else:
		print "Invalid command: %s" % (cmds[0])
		print_help()
		sys.exit(-1)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def print_help():
	print "%s [options] <cmd>" % (sys.argv[0])
	print "Options:"
	print "\t", "-h    ", "Show help"
	print "\t", "-L <n>", "Set log level"
	print "Commands: "
	print "\t", "start : start local polling"
	print "\t", "stop  : stop polling"
	print "\t", "test  : show start command line for each local agent"

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def load_conf():

	global poller_supervisor_pid_file, \
	       poller_supervisor_log_file 

	poller_supervisor_log_file = os.sep.join([get_configuration().LOG_DIR, poller_supervisor_log_filename + '.log'])

	poller_supervisor_pid_file = os.sep.join([get_configuration().PID_DIR, PROGRAM_NAME + '.pid'])

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def start_poller_supervisor():
	
	try:
		#connection.close() # BAD BAD BAD
		
		# Check for a pidfile to see if the daemon is already running
		try:
			pid = read_pid_file(poller_supervisor_pid_file)
		except IOError:
			pid = None
		if pid:
			if is_supervisor_process(pid):
				sys.stderr.write("Already running. (PID %s)\n" % pid)
				sys.exit(-1)
			else:
				if is_process_running(pid):
					sys.stderr.write("Unknown process with pid %s!\n" % pid)
					sys.exit(-1)
				else:
					# if the process is dead, go on!
					sys.stderr.write("PID file not empty, but process %s not running!\n" % pid)

		daemonize()

		# Do after daemonize!
		write_pid_on_file(poller_supervisor_pid_file)

		init_log()
		
		init_children_jobs()

		# Poller should not lock database. Close connection!
		connection.close() 
		
		respawn_children_until_stop()
		
		remove_pid_file(poller_supervisor_pid_file)

	except Exception, e:
		sys.stdout.write(str(e))
		__log( os.getpid(), "Error: %s" % (e) )
		
	return 0	

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def get_process_cmd_line(agent):

	__log(os.getpid(), " subprocess for %s running" % (agent))
	
	program = os.path.join(get_configuration().BASE_DIR, SUPERVISED_POLLER_SCRIPT)

	ipc_broker = get_configuration().BROKER_ADDRESS
	ipc_auth   = get_configuration().BROKER_AUTH
	
	broker = ':'.join(ipc_broker)
	if ipc_auth is not None and ipc_auth[0]:
		broker = '%s/%s@%s' % (ipc_auth[0], ipc_auth[1], broker)
	
	cache_module = get_configuration().CACHE_HANDLER
	cache_params = get_configuration().CACHE_HANDLER_PARAMS	
	
	log_file = os.path.join(get_configuration().LOG_DIR, 'poller.log')
	
	extra_modules = get_configuration().LOCAL_POLLER_EXTRA_MODULES 
	extra_modules = ':'.join(extra_modules)

	options = [
		get_configuration().BASE_DIR,
		
		'-B', broker,
		
		#'-t', str(agent.tenant.uuid),
		#'-i', str(agent.uuid),

		'-t', str(agent.tenant.name),
		'-i', str(agent.name),
		
		#'-n', str(agent.name),
		
		'-l', str(log_file),
		'-T', str(agent.threads),
		'-c', cache_module,
		'-C', json.dumps(cache_params),
	]
	
	if extra_modules:
		options += [
			'-X', extra_modules,
		]

	args = [
		  "python"
		, "python"
		, str(program)
	] + options

	return args

def start_poller_subprocess(params=None):

	try:

		agent = params['agent']

		args = get_process_cmd_line(agent)

		__log(os.getpid(), "%s execlp..." % (str(' '.join(args))))

                os.execlp( *args )
                
        except Exception, e:
		__log(os.getpid(), "Error while launching execd: %s" % (e))
                pass	
		

def show_poller_launch_commands():

	from pipes import quote
	from core.api.base import agent_select_all_locals

	agents = agent_select_all_locals()

	for a in agents:
		if not a.is_enabled:
			continue
			
		args = get_process_cmd_line(a)
		print ' '.join([quote(a) for a in args ])

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def stop_poller_supervisor():
	
	try:
		pid = read_pid_file(poller_supervisor_pid_file)
	except IOError:
		pid = None

	if not pid:
		sys.stderr.write("pidfile %s does not exist. Process not running?\n" % (poller_supervisor_pid_file) )
		sys.exit(1)
	else:
		if not is_supervisor_process(pid):
			sys.stderr.write("Unknown process with pid %s!\n" % pid)
			sys.exit(1)

	#
	# Send command to commit and stop
	#
	
	# Try killing the daemon process       
	try:
		sys.stdout.write("Sending term signal to poller (pid=%s)" % (pid))
		sys.stdout.flush()

		# Send signals until the process dies
		while 1:
			os.kill(pid, signal.SIGTERM)

			sys.stdout.write(".")
			sys.stdout.flush()

			time.sleep(1)

	except OSError, err:
		# did we kill the process or what else?
		if err.errno == errno.ESRCH:
			print ""
			print "Process terminated.",
		else:
			print "Error while sending signal to supervisor", err
			sys.exit(1)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#


def init_log():
	"""
	global log, opt_loglevel

	log = logging.getLogger(PROGRAM_NAME)
	log.setLevel( opt_loglevel )

	hdlr = logging.FileHandler(poller_supervisor_log_file, "a")
	hdlr.setFormatter( logging.Formatter('%(asctime)s %(levelname)s> %(message)s') )
	hdlr.setLevel( opt_loglevel )
	log.addHandler(hdlr)
	"""
	pass

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

# FIX: move to default_settings.py?
WAITTIME_AFTER_KILL = 1   #seconds


class SubProcessInfo(object):

	def __init__(self, name, action, pidfile_name, params={}):

		super(SubProcessInfo, self).__init__()

		self.name        = name
		self.action      = action
		self.params      = params
		self.pid         = 0
		self.pidfile     = os.sep.join( [get_configuration().PID_DIR, pidfile_name] )
		self.term_sent   = 0   # time the term signal was sent
		self.graceperiod = 10  # seconds. Interval between the (soft) term signal and the kill signal
		
	def __str__(self):
		return '<%s:%s>' % (self.name, self.pid if self.pid != None else '-')
		
	def run_action(self):
		self.action(self.params)



STOP_SUPERVISOR = False

subprocesses = []


def init_children_jobs():

	global subprocesses

	from core.api.base import agent_select_all_locals

	__log(os.getpid(), "Preparing subprocesses...")

	agents = agent_select_all_locals()

	for a in agents:

		__log(os.getpid(), "Found local agent: %s. Active = %s" % (a, a.is_enabled))
		
		if not a.is_enabled:
			continue
		

		process_name          = "%s-%s" % (a.name, a.uuid)
		process_pid_filename = process_name + '.pid'

		job = SubProcessInfo(name=process_name, 
		                     action=start_poller_subprocess, 
		                     pidfile_name=process_pid_filename, 
		                     params={
		                         'agent': a
		                     }
		)  
		subprocesses.append(job)

	for s in load_extra_subprocesses():
		subprocesses.append(s)

	for p in subprocesses:
		__log(os.getpid(), "Loaded job: %s" % (p.name))

	return

def load_extra_subprocesses():

	from sanet_common.pythonutils import load_symbol

	__log(os.getpid(), "Loading extra subprocesses...")

	jobs = []

	conf = get_configuration()

	for conf_line in conf.LOCAL_POLLER_EXTRA_PROCESSES:
		
		if len(conf_line) == 3:
			name, proc_handler_name, params = conf_line
		else:
			name, proc_handler_name = conf_line
			params = {}
			
		__log(os.getpid(), "Loading handler %s:%s" % (proc_handler_name, params))

		try:
			handler = load_symbol( proc_handler_name )
		except Exception, ex:
			print "Unable to load handler %s:%s" % (proc_handler_name, ex)
			__log(os.getpid(), "Unable to load handler %s:%s" % (proc_handler_name, ex))
			continue

		job = SubProcessInfo(
			name=name, 
			action=handler, 
			pidfile_name=name + '.pid', 
			params=params
		)  

		jobs.append(job)

	return jobs


def supervisor_on_sigterm(signum, frame):
	global STOP_SUPERVISOR
	STOP_SUPERVISOR = True


def respawn_children_until_stop():

	global STOP_SUPERVISOR, subprocesses

	not_running_by_name = dict( [ (p.name, p) for p in subprocesses ] )
	running_by_pid      = {}

	try:

		STOP_SUPERVISOR = False

		supervisor_pid = os.getpid()

		signal.signal(signal.SIGTERM, supervisor_on_sigterm)

		original_proc_title = get_proc_title()

		set_proc_title( original_proc_title + ' (supervisor)' )

		__log(supervisor_pid, "Starting main loop")
		
		while True:
			if not STOP_SUPERVISOR:

				for child in not_running_by_name.values():

					pid = os.fork()
					if pid:
						# father
						child.pid = pid

						del not_running_by_name[ child.name ] 
						running_by_pid[ pid ] = child

						#__log(supervisor_pid, "spawned %s PID: %s" % (child.name, child.pid))
						__log(supervisor_pid, "spawned %s" % (child))
					else:
						# son
						signal.signal(signal.SIGTERM, signal.SIG_DFL )

						my_pid = os.getpid()
						__log(my_pid, "started child %s" % (child))
						write_pid_on_file(child.pidfile)

						set_proc_title( original_proc_title + ' (%s)' % (child.name,))

						try:
							child.run_action() # This should not return
						except Exception, e:
							__log(my_pid, "EXCEPTION child %s: %s" % (child, e))
						exit() # This should not happen
			else:
				for child in running_by_pid.values():

					if not child.term_sent:
						__log(supervisor_pid, "sending TERM to %s " % (child))
						os.kill(child.pid, signal.SIGTERM)
						child.term_sent = time.time()

					elif time.time() - child.term_sent < child.graceperiod:
						__log(supervisor_pid, "re-sending TERM to %s" % (child))
						os.kill(child.pid, signal.SIGTERM)

					else:
						__log(supervisor_pid, "sending KILL to %s" % (child))
						os.kill(child.pid, signal.SIGKILL)

					break # Go through a wait before killing next one
					
				if len(running_by_pid) == 0:
					__log(supervisor_pid, "no more children, exiting")
					return
				else:
					time.sleep(WAITTIME_AFTER_KILL)
					
			try:
				(pid, status) = os.wait()
			except OSError, e:
				__log(supervisor_pid, "Wait interrupted with error %d: %s" % (e.errno, os.strerror(e.errno)))
				__log(supervisor_pid, "Stop signal recieved: %s" % ( STOP_SUPERVISOR ), )
			else:
				child = running_by_pid[ pid ]

				#__log(supervisor_pid, "Child %s exited" % (child))

				try:
					reason = convert_exit_status_to_verb_reason(status)
				except Exception, e:
					reason = "Unreadable reason: %s" % (str(e),)
					
				__log(supervisor_pid, "Child %s exited with status %d: %s" % (child, status, reason))
				
				if os.path.exists(child.pidfile):
					os.remove(child.pidfile)
					
				child.pid = 0

				del running_by_pid[pid]
				not_running_by_name[child.name] = child


	except Exception, e:
		__log("%s", e)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def __log(pid, status):

	global poller_supervisor_log_file
	
	try:
		filename = poller_supervisor_log_file
		f = open(filename,'a')
		f.write("%s [poller supervisor %s] %s\n" % (time.time(), pid, status) )
		f.close()
	except:
		pass


#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def write_pid_on_file(pid_file):
	pid = str(os.getpid())
	f = open(pid_file, 'w+')
	f.write("%s\n" % (pid))
	f.close()
	
def remove_pid_file(pid_file):
	try:
		if os.path.exists(pid_file):
			os.remove(pid_file)
	except Excetion, ex:
		print "Unable to remove pid file", ex
		pass

def read_pid_file(name):
	f = file(name,'r')
	try:
		pid = int(f.read().strip())
	except ValueError, e:
		pid = None
	f.close()
	return pid

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def is_supervisor_process(pid):
	
	try:
		f = file('/proc/%s/cmdline' % pid,'r')
		cmd_argv = f.read().split('\x00')
		f.close()
	except IOError, er:
		sys.stderr.write("Process %s does not exist.\n" % pid)
		return False

	# Without setproctitle
	
	if cmd_argv[0] == 'python' and cmd_argv[1].find('server/poller/main.py') != -1:
		return True

	# With setproctitle
	if cmd_argv[0].find('python') != -1 and cmd_argv[0].find('server/poller/main.py') != -1:
		return True
	
	return False

def is_process_running(pid):
	
	process_filename = '/proc/%s/cmdline' % pid
	
	if os.path.exists(process_filename):
		return True
	else:
		return False

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

if __name__ == "__main__":
	poller_supervisor_main()

#BDCA


