# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

"""

import threading
import Queue

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class SanetdSimpleMailSender(object):
	
	def __init__(self, logger, mailhandler, poll_timeout=5):
		
		super(SanetdSimpleMailSender, self).__init__()

		self._errors       = -1
		self._low_errors   = -1
		self._delivered    = -1
		
		self._sending      = False
		
		self._logger       = logger
		self._handler      = mailhandler
		
		self._logger.debug( "[%s] init", self.name)
		
	@property
	def name(self):
		return self.__class__.__name__

	def start(self):
		self._logger.log(19, "[%s] start", self.name)
		self._errors       = 0
		self._low_errors   = 0
		self._delivered    = 0
		
	def num_pending_messages(self):
		if self._sending: 
			return 1
		else: 
			return 0

	def get_text_info(self):
		return "queue=%s, delivered=%s, errors=%s, low errors=%s" % (self.num_pending_messages(), self._delivered, self._errors, self._low_errors, )
	
	def push_email(self, frm, to, subject, body, extra_headers):
			
		self._sending = True
		
		self._logger.debug("[%s] push mail to %s: %s", self.name, to, subject)
		
		result = self._handler.send(
			  frm
			, to
			, subject
			, body
			, extra_headers
		)

		self._logger.info("[%s] mail delivery: %s", self.name, result)
		
		# Fix: This object is shared. These counters are not thread safe (?)
		if result == 0:
			self._delivered  += 1
		elif result == -1:
			self._errors     += 1
		elif result == -1:
			self._low_errors += 1
			
		self._sending = False
			
			
	def stop(self):
		self._logger.debug("[%s] stop", self.name)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class SanetdThreadedMailSender(threading.Thread):
	"""
	This is a multithreaded version. This object creates a thread which
	waits for emails.
	"""
	
	def __init__(self, logger, handler, poll_timeout=5):

		super(SanetdThreadedMailSender, self).__init__()

		# Do not block main program on exit
		# FIX: hack for old API version.
		if (hasattr(self, "daemon")):
			self.daemon = True
		else:
			self.setDaemon(True)

		self._poll_timeout = poll_timeout
		self._mail_queue   = Queue.Queue(-1)
		self._errors       = 0
		self._low_errors   = 0
		self._delivered    = 0

		self._stop_thread  = False

		self._logger       = logger
		self._handler      = handler

		self._logger.debug("[%s] init", self.name)
		
	@property
	def name(self):
		return self.__class__.__name__

	def start(self):
		self._logger.debug("[%s] start", self.name)
		super(SanetdThreadedMailSender, self).start()

	def get_text_info(self):
		return "queue=%s, delivered=%s, errors=%s, low errors=%s" % (self._mail_queue.qsize(), self._delivered, self._errors, self._low_errors)
	
	def push_email(self, frm, to, subj, body, extra_headers):
			
		self._logger.debug("[poller thread mailer] pushing new mail for %s", email)	

		mail_data = (frm, to, subj, body, extra_headers)
		
		block = False
		timeout = 0
		self._mail_queue.put(mail_data, block, timeout)

	def run(self):
		
		self._logger.debug("[poller thread mailer] started.")
		
		queue   = self._mail_queue
		timeout = self._poll_timeout
		
		while True:
			try:
				mail_data = queue.get(True, timeout)
			except Queue.Empty, er:
				continue
			except KeyboardInterrupt, er:
				break
			except SystemExit, er:
				break
			
			if mail_data != None:
				try:
					self.__real_send( mail_data )
				except Exception, ex:
					self._logger.error("[poller threaded mailer] exception: %s", ex)
					self._low_errors += 1
					
			else:
				if self._stop_thread == True:
					self._logger.debug("[poller threaded mailer] exiting...")
					return

	def __real_send(self, mail_data ):

		#
		# FIX: group several emails inside only one message?
		# 		
		
		(frm, to, subj, body, extra_headers) = mail_data
		
		self._logger.debug("[poller thread mailer] sending mail to %s: %s", to, subj)
		
		#result = notify(
		#	  self._logger
		result = self._handler.send(
			  frm
			, to
			, subj
			, body
			, extra_headers
		)
		
		if result == 0:
			self._delivered  += 1
		elif result == -1:
			self._errors     += 1
		elif result == -1:
			self._low_errors += 1


	def stop(self):
		
		self._stop_thread = True
		self._logger.debug("[%s] stopping....", self.name)
		
		block = False
		timeout = 0
		self._mail_queue.put(None, block, timeout)

		self._logger.debug("[%s] waiting for thread stop", self.name)
		self.join()
		
		self._logger.debug("[%s] stopped", self.name)



if __name__ == "__main__":
	
	
	import os
	import logging
	

	logger = logging.getLogger("prova")

	log_format = '%(asctime)s %(levelname)s> %(message)s'
	logger.setLevel(10)
	hdlr = logging.StreamHandler()
	hdlr.setFormatter( logging.Formatter(log_format) )
	hdlr.setLevel( 10 )
	logger.addHandler(hdlr)


	os.environ.setdefault("DJANGO_SETTINGS_MODULE", "conf.settings")	

	#from sanet_common.mailing.sendmail import SendMailShellHandler as handler
	#h = handler(logger, { "command": "/usr/local/bin/sendmail -f noc@labs.it -t" } )

	from sanet_common.mailing.djangosmtp import DjangoMailHandler as handler
	
	h = handler(logger, { 
		"HOST"               : "localhost",
		"PORT"               : 25,
		"DEFAULT_DOMAIN_NAME": "localdomain",
		"ENVELOPE_SENDER"    : "noc@localhost",
	})



	#sender = SanetdSimpleMailSender(logger, h)
	sender = SanetdThreadedMailSender(logger, h)
	sender.start()
	sender.push_email("recipient@localhost", "subject", "body", {})
	sender.stop()
	
