# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""


"""

import os
import time
import types
import traceback
from socket import gethostname
import json
import pprint
import re

from django.conf import settings
#from django import db
#from django.db import transaction
from django.db import connection
from django.db.models import Q

from sanet_common.uuids import create_global_uuid
from sanet_common.sanetdefs.constants import *
from sanet_common.ipc import IPCConnection
from sanet_common.ipc import AbstractIPCFactory
from sanet_common.formatutils import number_to_str
from sanet_common.binary import datestr
from sanet_common.osutils.daemonize import launch_daemon

from core.exceptions import *

from core.db_utils import transaction_commit_on_success, transaction_set_dirty

# Core signals
from core.models import sanetd_process_changed_conditions
from core.models import sanetd_alarm_msg_create_template_symbols
from core.models import sanetd_process_event_additional_updates

# Core models
from core.models import Agent
from core.models import SchedulingGroup
from core.models import Cluster
from core.models import MonitoredElement, MonitoredNode, MonitoredIface, MonitoredStorage, MonitoredService, MonitoredDevice, Link
from core.models import ELEMENT_MODEL_BY_KIND
from core.models import DataGroup, DataGroupParameter, DataSource, Condition
from core.models import Event, Alarm

from core.models import render_msg_template
from core.api.base import db_set_current_tenant_schema
from core.api.base import tenant_get
from core.api.base import *

from core.api.datasource import DataSourceStoreUpdateException
from core.api.cache import init_cache_manager_from_conf


# TAGS
from tags.models import ELEMENT_KIND_TO_TAG_ELEMENTTYPE
from tags.models import ElementTag
from tags.server.tagsd.api import signal_condition_changed




#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class ServerIPCFactory(AbstractIPCFactory):

	def __init__(self, logger):
		super(ServerIPCFactory, self).__init__(logger)

	def create_connection(self):
		return create_ipc_connection_from_config(self.logger)

#-----------------------------------------------------------------------------#

def create_ipc_connection_from_config(logger):
	return create_ipc_connection(logger, settings.BROKER_ADDRESS, settings.BROKER_AUTH)

def create_ipc_connection(logger, broker_address, broker_auth):

	connect_options = {}
	host = broker_address[0]
	if len(broker_address) > 1:
		connect_options['port'] = broker_address[1]

	if broker_auth is not None and broker_auth[0]:
		connect_options['username'] = broker_auth[0]
		connect_options['password'] = broker_auth[1]

	logger.debug("Opening connection with server %s" % (host) )
	logger.log(9, "Options:")

	if logger.isEnabledFor(9):
		for k,v in connect_options.items():
			logger.log(9, "\t%s = %s", k, v)

	c = IPCConnection( host, **connect_options)

	return c


#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def create_raw_query_params(table_name, rid, fields, sentinel_mtime=None, pk_column_name='uuid'):

	query_string = 'UPDATE %s SET ' % (table_name)
	query_values = []

	qfields = []
	for f_name, f_value in fields.items(): 
		qfields.append( '%s=%%s' % (f_name) )
		query_values.append(f_value)
	query_string += ', '.join( qfields )

	query_string += ' WHERE %s=%%s' % (pk_column_name)
	query_values.append( rid )

	if sentinel_mtime:
		# BUG: Siamo costretti a convertire in INTEGER il campo mtime
		#      per un problema di perdita di precisione. Da indagare.
		query_string += ' AND mtime::integer=%s'
		query_values.append( int(sentinel_mtime) )

	return query_string, query_values

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def CREATE_ALARM_MSG_DATA(element, condition, event, alarm, flow_info_entries, datasources=None, pushed_data=None, saved_variables=None):

	parent = element.parent

	msg = {
		'ts'        : alarm.mtime, 

		'tenantid'  : element.tenant.uuid,
		'tenantname': element.tenant.name,

		'uuid': alarm.uuid,

		'event': {
			'uuid': event.uuid,
		},

		'element'    : {
			'uuid'       : element.uuid,
			'name'       : element.name,
			'kind'       : element.kind,
			'type'       : element.kind,
			'path'       : element.path,

			'parent_id'  : parent.uuid if parent else None,
			'parent_name': parent.name if parent else None,
			
			'nodedescr'  : parent.description if parent else element.description,
		},

		'condition'  : {
			'uuid'            : condition.uuid,
			'name'            : condition.name,
			'path'            : condition.path,
			'priority'        : condition.priority,
			'priority_level'  : PriorityLabel.get_label(condition.priority),
			'expr'            : condition.expr,
			'maxtries'        : condition.max_tries,
			'statuslastchange': condition.statuslastchange,
			'classification'  : condition.classification,
		},

		'params'   : condition.datagroup.get_safe_params_dict(),

		'datasources': datasources,

		'saved_variables': saved_variables,

		'result'     : {
			'status'     : condition.status,
			'laststatus' : condition.statuslast,
			'verbstatus' : condition.verbstatus,
		},

		'flow-information' : flow_info_entries,

		'mail': None,

		# Questo campo memorizza tutti i dati inviati con eventi push
		'pushed_data': pushed_data,
	}

	return msg

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def load_agent_config(tenant_name, agent_name):

	tenant = tenant_get(tenant_name)
	if tenant == None:
		raise Exception("Invalid tenant: %s" % (tenant_name))

	db_set_current_tenant_schema(tenant.name)

	try:
		agent = Agent.objects.get(tenant=tenant, name=agent_name)
		
		if not agent.is_enabled:
			raise Exception("Agent %s not enabled" % (agent_name))

	except Agent.DoesNotExist, e:
		raise Exception("Invalid agent: %s" % (agent_name))

	#
	# Base settings
	#
	ret = {
		'tenantid': tenant.uuid,
		'agentid' : agent.uuid,

		'threads' : agent.threads,

		'config_update_minperiod': agent.config_update_minperiod,
	} 

	#
	# Advanced settings for scheduling groups
	#
	ret.update({
		'scheduling_min_gap'     : agent.scheduling_min_gap,
		'scheduling_max_parallel': agent.scheduling_max_parallel,
		'scheduling_max_deadline': agent.scheduling_max_deadline,
	})

	# FIX: disabled, buffering not working yet.
	# Advanced settings for remote agents
	#
	#ret.update( {
	#	'sendbuffer_max_data_size': settings.AGENT_REMOTE_SEND_BUFFER_MAX_SIZE ,
	#})

	return ret

def load_scheduling_group_config(tenantid, cache_manager):

	tenant = tenant_get_by_id(tenantid)
	db_set_current_tenant_schema(tenant.name)

	ret = list(SchedulingGroup.load_for_agent(tenant))

	return ret


def load_cluster_config(tenantid, cache_manager):

	tenant = tenant_get_by_id(tenantid)
	db_set_current_tenant_schema(tenant.name)

	ret = list(Cluster.load_for_agent(tenant))

	return ret


def load_nodes_config(tenantid, agentid, cache_manager):

	tenant = tenant_get_by_id(tenantid)
	db_set_current_tenant_schema(tenant.name)

	agent    = agent_get_by_id(agentid, tenant)

	ret = list(MonitoredNode.load_for_agent(tenant, agent, cache_manager))

	return ret

def load_ifaces_config(tenantid, agentid, cache_manager):

	tenant   = tenant_get_by_id(tenantid)
	db_set_current_tenant_schema(tenant.name)

	agent    = agent_get_by_id(agentid, tenant)

	ret = list(MonitoredIface.load_for_agent(tenant, agent, cache_manager))

	return ret
	
def load_storages_config(tenantid, agentid, cache_manager):

	tenant = tenant_get_by_id(tenantid)
	db_set_current_tenant_schema(tenant.name)

	agent  = agent_get_by_id(agentid, tenant)

	ret = list(MonitoredStorage.load_for_agent(tenant, agent, cache_manager))

	return ret

def load_devices_config(tenantid, agentid, cache_manager):

	tenant = tenant_get_by_id(tenantid)
	db_set_current_tenant_schema(tenant.name)

	agent  = agent_get_by_id(agentid, tenant)

	ret = list(MonitoredDevice.load_for_agent(tenant, agent, cache_manager))

	return ret

def load_services_config(tenantid, agentid, cache_manager):

	tenant = tenant_get_by_id(tenantid)
	db_set_current_tenant_schema(tenant.name)

	agent  = agent_get_by_id(agentid, tenant)

	ret = list(MonitoredService.load_for_agent(tenant, agent, cache_manager))

	return ret

def load_links_config(tenantid, cache_manager):

	tenant = tenant_get_by_id(tenantid)
	db_set_current_tenant_schema(tenant.name)

	# NOTE: do not filter by agent because linked interfaces could belong to different agents
	ret = link_export_all_links(tenant)

	return ret

def load_datagroup_params_config(tenantid, agentid, cache_manager):
	
	tenant   = tenant_get_by_id(tenantid)
	db_set_current_tenant_schema(tenant.name)

	agent    = agent_get_by_id(agentid, tenant)

	ret = list(DataGroupParameter.load_for_agent(tenant, agent))

	return ret


def load_datagroups_config(tenantid, agentid, cache_manager):

	tenant   = tenant_get_by_id(tenantid)
	db_set_current_tenant_schema(tenant.name)

	agent    = agent_get_by_id(agentid, tenant)

	datagroups = list(DataGroup.load_for_agent(tenant, agent, cache_manager))

	return datagroups

def load_datasources_config(tenantid, agentid, cache_manager):

	tenant   = tenant_get_by_id(tenantid)
	db_set_current_tenant_schema(tenant.name)

	agent    = agent_get_by_id(agentid, tenant)

	datagroups = list(DataSource.load_for_agent(tenant, agent, cache_manager))

	return datagroups

def load_conditions_config(tenantid, agentid, cache_manager):

	tenant   = tenant_get_by_id(tenantid)
	db_set_current_tenant_schema(tenant.name)

	agent    = agent_get_by_id(agentid, tenant)

	ret = list(Condition.load_for_agent(tenant, agent, cache_manager))

	return ret


#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def process_update_event(logger, cache_manager, datasource_store, tenant, agentid, event):
	"""
	Questa e' la funzione principale che serve per "processare" i dati di aggiornamento di un datagroup 
	prodotti dalla valutazione di un datagroup.
	"""

	# 
	# Aggiornamento transazionale di tutto il datagroup
	#
	alarms, conditions_changed = _process_update_event_t(logger, cache_manager, datasource_store, tenant, agentid, event)

	#
	# Notifica alle altre parti del server (processi) quali sono le condition
	# che hanno subito delle variazioni.
	#
	if conditions_changed:
		try:
			#db.reset_queries()

			logger.debug("Signaling processed conditions...")

			sanetd_process_changed_conditions.send(None, 
			                                       log                = logger, 
			                                       tenant             = tenant, 
			                                       agentid            = agentid, 
			                                       changed_conditions = conditions_changed
			                                      )

			#logger.debug("db queries: %s", len(connection.queries))
			#for q in connection.queries:
			#	logger.debug("%s", q['sql'])
		except Exception, e:
			logger.error("An error occurred when signaling changed conditions: %s:%s -> %s", type(e), str(e), traceback.format_exc())

	return alarms


def _SET_TENANT(tenant, context):
	if not context['settenant']:
		db_set_current_tenant_schema(tenant.name)
		context['settenant'] = True

@transaction_commit_on_success
def _process_update_event_t(logger, cache_manager, datasource_store, tenant, agentid, event):

	alarms             = []
	conditions_changed = []

	context = {
		'settenant': False
	}

	for update_event_chunk in event['data']:

		#
		# Update elements from the event
		#
		# Se un elemento non esiste questo metodo solleva un'eccezione e l'intera operazione viene
		# annullata.
		_update_elements(logger, cache_manager, tenant, agentid, update_event_chunk, context)

		logger.debug("element db queries: %s", len(connection.queries))

		#
		# Update additional object from the event
		#
		_update_additional(logger, cache_manager, tenant, agentid, update_event_chunk, context)

		logger.debug("additional db queries: %s", len(connection.queries))

		#
		# Update Datagroup from the event
		#
		new_alarms_to_send, conditions_changed = update_datagroup(logger, cache_manager, datasource_store, tenant, agentid, update_event_chunk , context)

		logger.debug("datagroup db queries: %s", len(connection.queries))

		alarms += new_alarms_to_send

	return alarms, conditions_changed

def _update_elements(logger, cache_manager, tenant, agentid, update_event_chunk, context):

	#_SET_TENANT(tenant, context)

	elements = update_event_chunk.get('elements', None)
	if elements == None:
		return

	for element_id, fields in elements.items():
		if fields:
			_update_element(logger, cache_manager, tenant, agentid, update_event_chunk, element_id, fields, context)


def _update_element(logger, cache_manager, tenant, agentid, update_event_chunk, element_id, fields, context):

	#_SET_TENANT(tenant, context)

	#
	# Update fields
	#

	kind       = fields.get('kind', None)
	if kind:

		model      = ELEMENT_MODEL_BY_KIND[ kind ]

		c = None

		#
		# CACHE
		#
		cache_fields    = dict( [ (f,v) for f,v in fields.items() if f    in model.cached_fields] )
		if cache_fields:
			try:
				logger.log(19, "Updating element %s %s on cache: %s", model, element_id, cache_fields)
				model.raw_save_fields_in_cache( cache_manager, element_id, cache_fields )
			except Exception, e:
				logger.warning("Unable to save values in cache: %s", e)

		#
		# DB FIELS
		#
		common_fields   = dict([ (f,v) for f,v in fields.items() if f     in MonitoredElement.CLUSTER_FIELDS  ])
		if common_fields:

			_SET_TENANT(tenant, context)
			
			table_name = 'monitoredelement'
			logger.debug("\t\tUpdating monitoredelement: %s", common_fields)

			query_string, query_values = create_raw_query_params(table_name, element_id, common_fields, )

			if c is None:
				c = connection.cursor()

			c.execute(query_string, query_values)
			if c.rowcount == 1:
				logger.info("Updated %s %s: %s",table_name, element_id, common_fields)
				transaction_set_dirty()
		
		element_fields  = dict([ (f,v) for f,v in fields.items() if f not in model.cached_fields + MonitoredElement.CLUSTER_FIELDS + ['kind',] ])
		if element_fields:
			_SET_TENANT(tenant, context)

			table_name = model._meta.db_table
			pk_column  = model._meta.pk.get_attname()

			logger.debug("\t\tUpdating %s: %s", table_name, element_fields)

			query_string, query_values = create_raw_query_params(table_name, element_id, element_fields, pk_column_name=pk_column)

			if c is None:
				c = connection.cursor()

			c = connection.cursor()
			c.execute(query_string, query_values)
			if c.rowcount == 1:
				logger.info("Updated %s %s: %s",table_name, element_id, element_fields)
				transaction_set_dirty()
	else:
		#
		# OLD METHOD. REMOVE AS SOON AS POSSIBLE
		#


		#
		# CACHE FIELDS
		#
		cache_fields    = dict( [ (f,v) for f,v in fields.items() if f    in model.cached_fields] )
		if cache_fields:
			try:
				logger.log(19, "Updating element %s %s on cache: %s", model, element_id, cache_fields)
				model.raw_save_fields_in_cache( cache_manager, element_id, cache_fields )
			except Exception, e:
				logger.warning("Unable to save values in cache: %s", e)

		#
		# DB FIELS
		#

		_SET_TENANT(tenant, context)

		# Find the element
		e = MonitoredElement.objects.get(uuid=element_id)
		e = e.get_specific_element()

		model = e.__class__

		db_fields = dict( [ (f,v) for f,v in fields.items() if f  not in model.cached_fields ] )

		logger.debug("\tUpdating element %s with: %s", e, fields)

		# FIX: update only if configuration time is the same?
		for f_name, value in db_fields.items():
			setattr(e, f_name, value)
		e.save()



def _update_additional(logger, cache_manager, tenant, agentid, update_event_chunk, context):

	additional_updates = update_event_chunk.get('additional_updates', None)
	if additional_updates == None:
		logger.debug("No additional updates")
		return

	#
	# Trigger creation of symbols from other installed apps
	#
	

	logger.debug("Triggerin signal for additional_updates...")

	ret = sanetd_process_event_additional_updates.send(
		sender             = None, 
		log                = logger, 
		tenant             = tenant, 
		agentid            = agentid,
		additional_updates = additional_updates
	)
	
	logger.debug("Additional updates processed by %s subapps", len(ret))

	return



def update_datagroup(logger, cache_manager, datasourcestore, tenant, agentid, update_event_chunk, context  ):

	alarms_to_send = []

	try:
		datagroup_id = update_event_chunk['datagroup_id']
		element_id   = update_event_chunk['element_id']

		updates = update_event_chunk['datagroup_updates']

		if updates:
			fields         = updates['fields']
			sentinel_mtime = updates['smtime']

			cache_fields  = dict( [ (f,v) for f,v in fields.items() if f     in DataGroup.cached_fields] )
			db_fields     = dict( [ (f,v) for f,v in fields.items() if f not in DataGroup.cached_fields] )

			if db_fields:
				logger.debug("Updating datagroup fields: %s", db_fields)

				_SET_TENANT(tenant, context)

				query_string, query_values = create_raw_query_params('datagroup', datagroup_id, db_fields, sentinel_mtime=sentinel_mtime)

				# Exec query
				c = connection.cursor()
				c.execute(query_string, query_values)
				if c.rowcount:
					# IMPORTANT: this is necessary for correct transaction handling!		
					transaction_set_dirty()

					logger.debug("datagroup %s updated!", datagroup_id)
				else:
					raise ConcurrentUpdateException("%s deleted or changed before commit update" % (datagroup_id))

				# Nel caso di datagroup sospesi devo rimuoveri i parametri di sospensione
				if 'suspended' in db_fields:
					suspended = db_fields['suspended']
					if suspended == False:

						logger.debug( "Updating datagroup suspension configured fields")
						_SET_TENANT(tenant, context)
						dg = DataGroup.objects.get(tenant=tenant, uuid=datagroup_id)
						dg.remove_configured_suspend_fields()
						transaction_set_dirty()

			#
			# Update Cache fields
			#
			try:
				if cache_fields:
					logger.log(19, "Updating datagroup %s on cache: %s", datagroup_id, cache_fields)
					DataGroup.raw_save_fields_in_cache( cache_manager, datagroup_id, cache_fields )
			except Exception, e:
				logger.warning("Unable to save values in cache: %s", e)

		update_datasources(logger, cache_manager, datasourcestore, tenant, agentid, datagroup_id, element_id, update_event_chunk, context)

		alarms_to_send, conditions_changed = update_conditions(logger, cache_manager, tenant, agentid, datagroup_id, element_id, update_event_chunk, context)

		return alarms_to_send, conditions_changed

	except DataGroup.DoesNotExist, e:
		logger.error("\tTrying to update datagroup %s but does not exists" % (dg_id))

	return alarms_to_send


def update_datasources(logger, cache_manager, datasourcestore, tenant, agentid, datagroup_id, element_id, update_event_chunk, context ):

	tenant_id     = tenant.uuid

	dg_data = update_event_chunk
	if not dg_data.has_key('datasources_updates'):
		return

	# logger.warning(dg_data['datasources_updates'])
	for datasource_id , ds_updates in dg_data['datasources_updates'].items():

		# Per ragioni di perormance carichiamo il datasource dal database SOLO se
		# realmente necessario (dobbiamo aggiornare il db o dobbiamo creare lo store RRD)
		datasource = None

		try:
			ts             = ds_updates['time']
			sentinel_mtime = ds_updates.get('smtime', None)
			fields         = ds_updates['fields']
			verbstatus     = ','.join( ds_updates['infos'] )

			data_to_cache = dict( [ (f,v) for f,v in fields.items() if f     in DataSource.cached_fields] )  #FIX: test di performance indicano che questo e' leggermente piu' lento che fare il ciclo for nel modo classico
			data_to_db    = dict( [ (f,v) for f,v in fields.items() if f not in DataSource.cached_fields] )  #FIX: test di performance indicano che questo e' leggermente piu' lento che fare il ciclo for nel modo classico

			#
			# Update DB
			#
			if data_to_db:
				logger.info("Updating datasource %s on DB: %s", datasource_id, data_to_db)

				if datasource is None:
					_SET_TENANT(tenant, context)
					datasource = DataSource.objects.get(tenant=tenant, uuid=datasource_id)

				#
				# QUery
				#
				query_string, query_values = create_raw_query_params('datasource', datasource.uuid, data_to_db, sentinel_mtime=sentinel_mtime)

				# Exec query
				c = connection.cursor()
				c.execute(query_string, query_values)
				if c.rowcount:

					# IMPORTANT: this is necessary for correct transaction handling!
					transaction_set_dirty()

					logger.debug("%s: updated!", datasource.path)
				else:
					raise ConcurrentUpdateException("%s deleted before commit update" % (datasource.uuid))

			#
			# Update data serie in the storage
			#
			try:  # Questo try...except e' piu' veloce di un IF con get() dal dizionario in caso di valore mancante
				value = fields['lastvalue']
			except Exception, ex:
				#continue
				value = None

			if ds_updates.get('nosave', 0):
				logger.debug("Not saving datasource %s", datasource_id)
			elif value is not None:
				if isinstance(value, types.FloatType) or isinstance(value, types.IntType) or isinstance(value, types.LongType):

					# Create storage if it does not exist
					exists = datasourcestore.exists(tenant_id, datasource_id)

					if not exists:
						if datasource is None:
							_SET_TENANT(tenant, context)
							datasource = DataSource.objects.get(tenant=tenant, uuid=datasource_id)

						logger.debug("INIT datasource (%s:%s) path %s)", tenant_id, datasource_id, datasource.path)
						datasourcestore.init_store(datasource)

					#logger.log(19, "Updating datasource %s on store with value: %r (%s)", datasource_id, value, type(value))
					datasourcestore.save_value(tenant_id, datasource_id, ts, value)
						
				else:
					logger.warning("Trying to save datasource %s on store with invalid type: %s = %r", datasource_id, type(value), value)

			#
			# Update Cache fields
			#
			try:
				data_to_cache['verbstatus'] = verbstatus
				
				#if data_to_cache['lastvalue'] == None:
				#	logger.warning("Updating datasource %s on cache: %s", datasource_id, data_to_cache['verbstatus'])
				logger.debug("Updating datasource %s on cache: %s", datasource_id, data_to_cache)
				DataSource.raw_save_fields_in_cache( cache_manager, datasource_id, data_to_cache )
			except Exception, e:
				logger.warning("Unable to save values in cache: %s", e)

		except DataSource.DoesNotExist, e:
			logger.warning("Datasource %s does not exist. skip" % (datasource_id,) )
			continue

		except DataSourceStoreUpdateException, e:
			logger.warning("Error while updating datasource (%s:%s:%s): %s:%s", tenant_id, agentid, datasource_id, e, traceback.format_exc())
			continue

	return


def update_conditions(logger, cache_manager, tenant, agentid, datagroup_id, element_id, update_event_chunk, context):

	alarms_to_send     = []

	conditions_changed = []

	dg_data = update_event_chunk
	if not dg_data.has_key('conditions_updates'):
		return alarms_to_send


	for c_id, update_data in dg_data['conditions_updates'].items():

		condition = None

		old_status     = None
		ts             = update_data['time']
		sentinel_mtime = update_data.get('smtime', None)
		verbstatus     = ','.join(update_data['infos'])

		data_to_cache = dict([ (f,v) for f,v in update_data['fields'].items() if f     in Condition.cached_fields ])
		data_to_db    = dict([ (f,v) for f,v in update_data['fields'].items() if f not in Condition.cached_fields ])

		#
		# Update DB
		#
		if data_to_db:

			logger.debug("Updating condition %s on db: %s", c_id, data_to_db)

			_SET_TENANT(tenant, context)

			try:
				condition = Condition.objects.get(uuid=update_data['uuid'])
			except Condition.DoesNotExist, e:
				logger.warning("Condition %s does not exist. Skip" % (update_data['uuid']))
				continue

			condition.load_fields_from_cache(cache_manager)

			old_status = condition.status

			#
			# Update conditions fields because we are going to use this object
			#
			for field_name, value in data_to_db.items():
				setattr(condition, field_name, value)

			if condition.update_notable_state(ts):
				data_to_db['statuslastnotablechange'] = ts

			#
			# UPDATE QUERY
			#

			#
			# Poiche' Server e agente sono separati puo' esitere un problema di inconsistenza tra lo stato sul server (DB)
			# e lo stato della condition dell'agente.
			#
			# Regole AUREE: 
			#  - Vale sempre lo stato sul DB (storico, log, ecc.)
			#  - se la condition e' stata modificata da configurazione (mtime cambiato), gli aggiornamenti dell'agente 
			#    potrebbero non essere validi.
			#
			# Per evitare di salvare transazioni "incosistenti" l'agente comunica al server lo stato della condition attuale
			# e lo stato precedente. 
			# Se in fase di aggiornamento, lo stato della condition nel DB non corrisponde allo stato che l'agente considera 
			# "precedente" l'aggiornamento viene perso.
			#
			# AGENT      CHECK                       -                 -     +                     +
			#            STATE                  UP   FA                DN    UP   FA!!!!           UP                    
			#                   L---------------u----*--------L--------*-----*----u------------L---*-----------u--->
			#                    .             .      \       .         \     \  .              .   \         .  
			#                     .           .        \       .         \     \.                .   \FA>UP  .            Legend
			#                      .         .    UP>FA \       .         \    .\                 .   \     .                ... config messages
			#                       .       .            \       .         \  .  \                 .   \   .                 \   update messages
			#                        .     .              \       .         \.    \                 .   \ .      
			#                         .   .                \       .        .\     \                 .   \       
			#                          . .                  \       .      .  \     \                 . . \      
			#                    -------.--------------------#-------.----.----#-----#-----------------.---\------>
			# SERVER   CURRENT   UP                          FA                DN    UP                     
			#             LAST   *                           UP                FA    DN                     WRONG
			#

			query_string, query_values = create_raw_query_params('condition', condition.uuid, data_to_db, sentinel_mtime=sentinel_mtime)

			# verifica stato precedente
			if data_to_db.has_key('statuslast'):
				query_string += " AND status = %s"
				query_values.append( data_to_db['statuslast'] )

			logger.debug("%s -> %s", query_string, query_values)

			# Exec query
			c = connection.cursor()
			db_set_current_tenant_schema(tenant.name, db_cursor=c)
			c.execute(query_string, query_values)
			if c.rowcount == 1:
				logger.debug("-----> DB UPDATED")

				# IMPORTANT: this is necessary for correct transaction handling!
				transaction_set_dirty()
			else:
				raise ConcurrentUpdateException("Condition %s changed or deleted before commit update" % (condition.uuid))

		# 
		# Update cache fields
		#
		try:
			data_to_cache['verbstatus'] = verbstatus

			if condition is not None:
				condition.verbstatus = verbstatus

			logger.debug("Updating condition %s on cache: %s", c_id, data_to_cache)

			Condition.raw_save_fields_in_cache(cache_manager, c_id, data_to_cache)
		except Exception, ex:
			logger.error("Unable to save cached fields for %s: %s", c_id, str(ex))

		#
		# New event/alarms?
		#
		if condition and old_status != condition.status:

			conditions_changed.append(condition)

			# IMPORTANTE: il codice seguente potrebbe alterare i cached field dell'oggetto.
			#             Assicuriamoci che l'oggetto abbia TUTTI i cached fields aggiornati
			#             da qui in avanti.
			condition.load_fields_from_cache(cache_manager)


			#
			# Load element data
			#
			_SET_TENANT(tenant, context)

			element = MonitoredElement.objects.get(uuid=element_id, tenant=tenant)

			#
			# Exec external script
			#
			logger.debug("Launching external script (%s, %s)", CSTATE_TO_CODENAME(old_status), CSTATE_TO_CODENAME(condition.status))
			exec_script_on_status_change(logger, tenant, agentid, condition)

			#
			# Store log event
			#
			alarm_to_send = create_eventlog(logger, cache_manager, tenant, agentid, element, ts, condition, raw_data=dg_data)

			# FIX
			if alarm_to_send:
				alarms_to_send.append(alarm_to_send)

	return alarms_to_send, conditions_changed

#def signal_changed_conditions(logger, cache_manager, tenant, condition):
#
#	try:
#		# La condition indicata ha cambiato di stato quindi bisogna
#		# aggiornare lo stato di tutto quello che dipende dalle condition:
#		#
#		#  - aggiornare lo stato dei tag      -> notificare a tags
#		#  - aggiornare lo stato delle mappe  -> aggiornato tutte le volte che viene visualizzata una mappa
#		#
#
#		signal_condition_changed(tenant.uuid, condition.uuid)
#
#	except Exception, e:
#		logger.error("Error while invalidating cache: %s:%s" % (type(e), str(e)))
#
#	return

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def create_eventlog(logger, cache_manager, tenant, agentid, element, ts, condition, raw_data=None):

	logger.debug("create_eventlog %s", condition)

	mtime = int(time.time())

	new_event = Event(
		tenant           = tenant,
		mtime            = mtime,

		kind             = EVENT_CONDITION,

		timestamp        = ts,
		msg              = condition.verbstatus,

		element_path     = element.path,
		condition_path   = condition.path,

		element_uuid     = element.uuid,
		condition_uuid   = condition.uuid,

		status           = condition.status, 
		statuslast       = condition.statuslast,
		statuslastchange = condition.statuslastchange,

		priority         = condition.priority,
	)

	new_event.save()

	#
	# 
	#
	verbstatus_msgbody = ''
	
	alarm_msg_data = None

	#
	# Produce new alarm? 
	#
	if is_alarm_transition(condition.statuslast, condition.status):

		#
		# New alarm
		#
		flow_informations_step_0 = [
		]

		if agentid != None:
			flow_informations_step_0.append({
				'agent_id': agentid,
			})

		new_alarm = Alarm(
			mtime      = new_event.mtime,
			tenant     = tenant,
			
			event      = new_event,
			ack_needed = is_alarm_ack_needed( condition.statuslast, condition.status ),
			ack        = None,
		)

		new_alarm.set_flow_informations( flow_informations_step_0 )
		new_alarm.save()

		#
		# Prepare datastruct for the notify manager
		#

		datasources = get_ds_info_from_raw_data(raw_data)

		saved_variables = None
		if raw_data is not None:
			saved_variables = raw_data.get('saved_variables', None)

		# Per gli eventi "pushed" il campo "pushed_data" contiene i dati pushed associati all'aggiornamento del datagroup
		pushed_data = raw_data.get('pushed_data', None) if raw_data is not None else None

		#
		# Creazione dell'allarme
		#
		alarm_msg_data = CREATE_ALARM_MSG_DATA(element, condition, new_event, new_alarm, flow_informations_step_0, datasources=datasources, pushed_data=pushed_data, saved_variables=saved_variables)

		logger.debug("Alarm data: %s", pprint.pformat(alarm_msg_data))

		# mail
		mail_data = create_email(logger, cache_manager, tenant, agentid, element, ts, condition, raw_data=raw_data)

		alarm_msg_data['mail'] = mail_data

		logger.log(9, "Final alarm data: %s", alarm_msg_data)

		#
		#
		#
		body = mail_data.get('body', None)
		if body:
			verbstatus_msgbody = body

	#
	# Update verbose status message
	#

	# Fallback
	if not verbstatus_msgbody:
		verbstatus_msgbody = condition.verbstatus

	# Update
	condition.verbstatus_msgbody = verbstatus_msgbody
	condition.save_fields_in_cache(cache_manager)

	return alarm_msg_data

def get_ds_info_from_raw_data(raw_data):

	all_ds_values = {}
	if raw_data:
		for ds_id, ds_data in raw_data.get('datasources_updates', {}).items():
			try:
				# Se e' presente "rate" questo datasource e' un COUNTER
				if ds_data['fields'].has_key('lastrate'):
					v = ds_data['fields']['lastrate']
				else:
					v = ds_data['fields']['lastvalue']

				#ds_name_for_template = ds_data['name'].replace('-','_') # Fix name
				ds_name_for_template = ds_data['name']

				printable_value = fix_object_for_jsondump(v)
				all_ds_values[ ds_name_for_template ] = {
					'id'   : ds_id,
					'value': printable_value
				}
			except KeyError, e:
				pass

	return all_ds_values

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def exec_script_on_status_change(logger, tenant, agentid, condition):

	if not condition.statuschange_action:
		return 

	try:
		logger.info("Launching script %s", condition.statuschange_action )

		#
		# Calculating environment
		#
		env = dict([(k,v) for k,v in os.environ.items()])

		env.update( calculate_external_script_environment_from_condition(condition) )

		action_id = create_global_uuid()
		env['SANET_ACTION_UUID'] = action_id

		#
		# Prepare execution arguments
		#
		import shlex 
		params = shlex.split( condition.statuschange_action )
		script_path   = params[0]
		script_params = params[0:]
		logger.info("Path  : %s", script_path  )
		logger.info("Params: %s", script_params )
		
		#
		# Locate script path
		#
		if settings.SANETD_EXEC_DIR:
			script_path = os.path.join(settings.SANETD_EXEC_DIR, script_path)

		try:
			logger.info("Excuting action %s : %s %r", action_id, script_path, script_params)

			#
			# Launch exteranl script in a de-attached daemon process.
			#
			# ATTENTION: do not user "subprocess" library here because external script will
			#            be children of sanetdserver process. If sanetdserver does not call
			#            wait() on terminated scripts, a lot of <defunct> processes will
			#            be hang and consuming kernel PIDs.
			#
			# All errors will be printed on syslog.
			#
			#
			launch_daemon(script_path, script_params, env=env, log_syslog=True)

			logger.info("Launched action %s", action_id)

		except OSError,er:
			logger.error("Unable to exec status action %s - %s: %s:%s", action_id, script_path, type(er), str(er))

	except Exception, e:
		logger.error("Error while launching script on status change for condition %s: %s:%s -> %s", condition.path, type(e), str(e), traceback.format_exc())

def calculate_external_script_environment_from_condition(condition):

	tenant  = condition.tenant
	element = condition.element
	node    = element.get_node()

	ret = {}
	ret['SANET_TENANT_NAME'         ] = tenant.name
	ret['SANET_TENANT_UUID'         ] = tenant.uuid

	ret['SANET_NODE_UUID'           ] = node.uuid
	ret['SANET_NODE_NAME'           ] = node.name

	ret['SANET_ELEMENT_UUID'        ] = element.uuid
	ret['SANET_ELEMENT_NAME'        ] = element.name

	ret['SANET_CONDITION_UUID'      ] = condition.uuid
	ret['SANET_CONDITION_PATH'      ] = condition.path

	ret['SANET_CONDITION_LASTCHANGE'] = str(int(condition.statuslastchange))

	ret['SANET_CONDITION_OLD_STATE' ] = condition.get_statelast()
	ret['SANET_CONDITION_NEW_STATE' ] = condition.get_state()

	return ret

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def create_email(logger, cache_manager, tenant, agentid, element, ts, condition, raw_data=None):

	logger.debug("create_email()")

	logger.debug("%s -> %s", CSTATE_TO_CODENAME(condition.statuslast), CSTATE_TO_CODENAME(condition.status))

	if not is_alarm_transition(condition.statuslast, condition.status):
		logger.debug("Transizione di stato non importante")
		return None

	if is_down_transition(condition.statuslast, condition.status):
		logger.debug("-> MAIL DOWN")

		email    = condition.email
		msg_subj = condition.msg_downsubj
		msg_body = condition.msg_downbody

	elif is_failing_transition(condition.statuslast, condition.status):
		logger.debug("FAILING")
		return None

	elif is_up_transition(condition.statuslast, condition.status):
		logger.debug("-> MAIL UP")

		email    = condition.upemail or condition.email
		msg_subj = condition.msg_upsubj
		msg_body = condition.msg_upbody

	else:
		log.error("Transizione di stato non gestita da invio email")
		return None

	if not email:
		logger.warning("Condition %s: email address is null."  , condition.path)
		email = ''

	# >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
	# IMPORTANT: Sanet2 behaviour removed by ticket #722
	#if email == "none":
	#	email = ''
	#	logger.warning("Condition %s: email address is 'none'", condition.path)
	# <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

	mail_data = {
		'to'     : email,
		'subject': msginterp(logger, cache_manager, condition, msg_subj       , raw_data=raw_data),
		'body'   : msginterp(logger, cache_manager, condition, msg_body, True , raw_data=raw_data),
		'headers': { 
			"X-SANET-Target-Id"     : condition.path,
			"X-SANET-Target-Status" : CSTATE_TO_CODENAME(condition.status),
			"X-SANET-Priority"      : condition.priority
		}
	}

	return mail_data

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def fix_object_for_jsondump(o):
    # Questa funzione serve per convertire l'oggetto "o" in una struttura dati
    # che sia:
    #   - JSON-compatibile
    #   - senza valori binary "spuri" (vettori di byte)
    # 
    # La funzione json.dumps() non e' in grado di fare il dump di stringhe che in 
    # realta' sono vettori di byte puri (no UNICODE, no ASCII).
    # 
    # Per riuscire a fare il dump in questi casi, "massaggiamo" l'oggetto (dizionario)
    # originale e cambiamo le stringhe (array) di byte nella loro rappresentazione (repr()).
    # 
    # Fa veramente schifo questa soluzione, ma e' l'unica che mi e' venuta in mente 
    # per gestire questa situazione.

    def _fix_object_for_jsondump(o):
        if isinstance(o, types.TupleType):
            return tuple([ _fix_object_for_jsondump(a) for a in o ])
        elif isinstance(o, types.ListType ):
            return [ _fix_object_for_jsondump(a) for a in o ]
        elif isinstance(o, types.DictType ):
            K = dict([ (k, _fix_object_for_jsondump(v)) for k,v in o.items() ] )
            return K
        elif isinstance(o, types.StringType) or isinstance(o, types.UnicodeType):
            try:
                encoder = json.JSONEncoder()
                encoder.encode(o)
                return o
            except Exception, e:
                #print __file__, "_fix_object_for_jsondump", o, e
                return ("%r" % (o))[1:-1]
        else:
            return o

    return _fix_object_for_jsondump(o)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class SnmpVariablesExpander(object):
	"""
	Classe di comodo per espandere in maniera "smart" le OID
	nei messaggi.
	
		::
			
			{{ exec_snmp.1.3.6.1.2.1.1.1.0 }}
			{{ exec_snmp.1.3.6.1.2.1 }}
			
			{% for oid , v in exec_snmp.items %}
				{{ oid }} = {{ v }}
			{% endfor %}
	
	Test:
		::
			
			from server.sanetd.sanetdlogic import SnmpVariablesExpander
			
			oid = raw_input()
			
			oids = { 
				'1.3.6.1.2.1.1.1.0':"ancona", 
				'1.3.6.1.2.1.1.2.0':"bologna",
				'1.3.6.1.3.1.1.1.0':"como", 
				'1.3.6.1.3.1.1.2.0':"domodossola",
				'1.3.6.1.3.1.2.1.0':"empoli",
				'1.3.6.1.3.1.2.2.0':"firenze",
				'2.1.1.1.1.1.1.1.0':"zzzz",
			}
			
			v = SnmpVariablesExpander( oids )
			
			print v
			print v[1]
			print v[1][3][6][1][2][1][1][2]
			print v[1][3][6][1][2][1][1][2][0]
			
			print "-------------"
			if oid:
				for n in oid.split('.'):
					v = v[int(n)]
			print type(v), v
	"""

	def __init__(self, variables, index=0, valid_oids=None):

		# Il dizionario "variables" potrebbe contenere delle OID con valori binary (non stampabili)
		# e quindi inutili se mostrati nel corpo di una mail.
		#
		# Qui "correggiamo" i dati in modo da avere una struttura dati che "espanda" dati sempre 
		# e comunque in maniera "decente".
		self.variables = fix_object_for_jsondump(variables)

		self.index      = index
		self.valid_oids = valid_oids if (valid_oids is not None) else sorted(self.variables.keys())

		return

	def __getitem__(self, key):

		if key == 'items':
			return sorted(self.variables.items())
		else:
			try:
				index = self.index

				new_valid_oids = []
				for oid in self.valid_oids:
						parts = oid.split('.')
						if int(key) == int(parts[index]):
							new_valid_oids.append(oid)

				if len(new_valid_oids) == 0:
					raise IndexError()

				elif len(new_valid_oids) == 1:
					if index >= len(new_valid_oids[0].split('.')) - 1:
						return self.variables[new_valid_oids[0]]

				return SnmpVariablesExpander(self.variables, index=index+1, valid_oids=new_valid_oids)

			except IndexError, e:
				raise e

	def __len__(self):               
		return len(self.valid_oids)

	def __iter__(self):                      
		var_list = [ (oid, self.variables[oid] ) for oid in sorted(self.valid_oids) ]
		return iter(var_list)

	def __str__(self):

		var_list = []
		for oid in self.valid_oids:
			v = self.variables[oid]
			#oid = '.'.join(oid.split('.')[self.index:])
			var_list.append((oid, v))

		if len(var_list) == 1:
			if var_list[0][0] == '':
				return var_list[0][1]

		return str(var_list)



def msginterp(log, cache_manager, condition, msg, is_body=False, raw_data=None):

	def timestamp_to_str(t):
		return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(t))

	log.debug("Substituting parameters")

	# Avoid Unicode characters
	msg = msg.encode('ascii','replace')

	t_now = time.time() 

	datagroup = condition.datagroup

	# QUESTION: How can we output a plain '\n' ?
	node         = None
	iface        = None
	linked_iface = None
	linked_node  = None

	element = condition.element.get_specific_element()
	node = element.get_node()
	node.load_fields_from_cache(cache_manager=cache_manager)

	if element.kind == ELEMENT_KIND_IFACE:
		iface = element

	#---------------------------------------------------------------
	# Simboli $.......
	#---------------------------------------------------------------

	element_symbols = get_expr_symbols_from_element(element, cache_manager)
	params_symbols  = get_expr_symbols_from_condition(condition)

	all_symbols = {}
	all_symbols.update(element_symbols)
	all_symbols.update(params_symbols)

	for param, value in all_symbols.items():
		if param[0] == '$':
			log.log(1, "Replacing %s = %s", param, str(value))
			msg = msg.replace(param, str( number_to_str(value)  ))

	if raw_data is not None:
		for key, value in raw_data.get('current_values',{}).items():
			msg = msg.replace(key, str( number_to_str(value)  ))

	#---------------------------------------------------------------
	# Wildcard %
	#---------------------------------------------------------------
	if node != None:
		msg = msg.replace('%H', node.name)

	if iface != None:

		if msg.find('%IDsnmp') >= 0:
			try:
				d = iface.description_snmp
				
				msg = msg.replace('%IDsnmp', (d if d else '<no iface snmp description>') )
			except Exception, e:
				msg = msg.replace('%IDsnmp', "***%s:%s***" % (type(e), str(e)) )

		if msg.find('%ID') >= 0:
			try:
				d = iface.description
				msg = msg.replace('%ID', (d if d != None else '<no iface snmp description>') )
			except Exception, e:
				msg = msg.replace('%ID', "***%s:%s***" % (type(e), str(e)) )

		msg = msg.replace('%I', iface.name)

	# Indirizzi IPV4
	if msg.find('%A') >= 0: 
		msg = msg.replace('%A', node.ipv4_addresses_spaced )

	# Descrizione del nodo
	try:
		msg = msg.replace('%D', node.description)
	except Exception, e:
		msg = msg.replace('%D', "***%s:%s***" % (type(e), str(e)) )

	msg = msg.replace('%E', str(condition.expr))

	msg = msg.replace('%H', node.name)

	if iface != None: 
		msg = msg.replace('%I', iface.name)

	if iface:
		linked_iface = iface.get_linked_iface()
		if linked_iface:
			linked_node = linked_iface.get_node()

	if linked_node:
		msg = msg.replace('%L', '%s:%s' % (linked_node.name, linked_iface.name))

	msg = msg.replace('%M', str(number_to_str(condition.max_tries)))

	if msg.find('%N') >= 0:  
		msg = msg.replace('%N', condition.tenant.visible_name)

	msg = msg.replace('%m', str(condition.verbstatus))

	msg = msg.replace('%n', str(number_to_str(condition.tries)) )

	msg = msg.replace('%p', '%')

	msg = msg.replace('%P', str(number_to_str(condition.priority)))

	msg = msg.replace('%T', condition.path)

	msg = msg.replace('%t', str(number_to_str(t_now)) )

	msg = msg.replace('%s', timestamp_to_str(t_now) )


	# Show the fetched snmp value
	if raw_data is not None:
		__fetched_snmp_variables = raw_data.get('saved_variables',{}).get('snmp_variables',[])
		
		if len(__fetched_snmp_variables) == 1:
			v = repr(__fetched_snmp_variables.values()[0])
			# Some SNMP variables are not printable types (binary arrays, eccc...)
			msg = msg.replace('%V', v )
		else:
			msg = msg.replace('%V', '**too many OIDs***' )


	d = datestr()
	msg = msg.replace('%y', d)

	# URL to the resource's page
	condition.alarm_url = get_condition_url(condition)

	if msg.find('%U') != -1:   #expand user notes on that controll?
		#url = get_condition_url(condition)
		#msg = msg.replace('%U', url)
		msg = msg.replace('%U', condition.alarm_url )

	#---------------------------------------------------------------
	# NUOVE WILDCARD {{ }} (Template di django)
	#---------------------------------------------------------------

	msg = re.sub('{{\s*}}','', msg)

	try:
		#
		# Wildcard di default
		#
		template_symbols = {
			'__logger__'  : log,

			'alarm'       : {
			                   'ts'    : str(number_to_str(t_now)),
			                   'ts_iso': timestamp_to_str(t_now),
			                   'ts_ISO': datestr(),
			                },
			'tenant'      : element.tenant,
			'element'     : element,
			'node'        : node,

			#'ds'          : all_ds_values,

			'condition'   : condition,

			#'params'      : params_d,
		}

		if element.kind == ELEMENT_KIND_IFACE:
			template_symbols.update({
				'iface'       : element,
				'linked_node' : linked_node,
				'linked_iface': linked_iface,
			})

		if element.kind == ELEMENT_KIND_STORAGE:
			template_symbols.update({
				'storage'       : element,
			})

		if element.kind == ELEMENT_KIND_SERVICE:
			template_symbols.update({
				'service'       : element,
			})

		if element.kind == ELEMENT_KIND_DEVICE:
			template_symbols.update({
				'device'       : element,
			})

		#
		# GESTIONE DEI NOMI DEI DATASOURCE NEI TEMPLATE
		#
		#    Un datasource puo' contenere un '-' nel nome, ma questo non piace alla libreria di django.#
		#
		#        {{ ds.foo-bar.value }}
		#
		#    Per aggirare questo problema, i nomi usati nei template vengono "massaggiati" per
		#    contenre '_' al posto di '-'.
		#
		#        {{ ds.foo-bar.value }}  -> {{ ds.foo_bar.value }}
		#    
		#
		all_ds_values = {}
		for ds_name, info in get_ds_info_from_raw_data(raw_data).items():

			ds_name_for_template = ds_name.replace('-','_') # Fix name

			all_ds_values[ ds_name_for_template ] = info

		template_symbols['ds'] = all_ds_values

		#
		# Parametri del datagroup
		#
		#    {{ params.<name> }}
		#
		params_d = dict([(p.name, str(p.get_value())) for p in datagroup.datagroupparameter_set.all() ])

		template_symbols['params'] = params_d

		#
		# Variabili salvate nelle expr e OID SNMP lette dalla rete
		#
		template_symbols_add_saved_variables(template_symbols, raw_data)

		#
		# Tutti i dati di eventi "pushed" saranno accessibili tramite macro:
		#
		#	{{ pushed_data.<key>[.<sub-key> ... ] }}
		#
		# Esempio:
		#
		#	{{ pushed_data.var1 }}
		#
		if raw_data:
			if raw_data.has_key('pushed_data'):
				template_symbols['pushed_data'] = raw_data['pushed_data']

		#
		# Trigger creation of symbols from other installed apps
		#
		try:

			# Try to not creae a cache_manager too many times while rendering wildcards.
			# Wildcard Handlers should use get_cache_manager().
			node.set_cache_manager(cache_manager)
			element.set_cache_manager(cache_manager)

			sanetd_alarm_msg_create_template_symbols.send(sender           = None, 
			                                              log              = log, 
			                                              tenant           = condition.tenant, 
			                                              node             = node, 
			                                              element          = element, 
			                                              condition        = condition, 
			                                              template_symbols = template_symbols,
			                                             )
		except Exception, e:
			log.error("Error while loading wildcards from subapps: %s:%s -> %s", type(e), str(e), traceback.format_exc())

		#
		# RENDER
		#
		msg = render_msg_template(msg, template_symbols)

	except Exception, e:
		# Se ci sono degli errori non espandiamo niente
		log.error("Error while expanding wildcards {{}} in message: %s. \n%s %s - %s", msg, type(e), str(e), traceback.format_exc())
		pass

	return msg


def template_symbols_add_saved_variables(template_symbols, raw_data):

	#
	# Espandiamo tutte le variabili calcolate dal datagroup
	#
	#    {{ exec_vars.<name> }}
	#
	if raw_data is not None:
		exec_vars = raw_data.get('saved_variables',{}).get('free_variables',{})
	else:
		exec_vars = {}

	template_symbols['exec_vars'] = exec_vars

	#
	# Espandiamo tutte le OID SNMP raccolte dal datagroup (walk escluse!)
	#
	#    {{ exec_snmp.<oid> }}
	#
	if raw_data is not None:
		snmp_variables = raw_data.get('saved_variables',{}).get('snmp_variables',{})
	else:
		snmp_variables = {}

	exec_snmp = SnmpVariablesExpander(snmp_variables)

	template_symbols['exec_snmp'] = exec_snmp

	return


def get_condition_url(condition):

	try:
		# Necessary in order to add  global_page_url() method to models.
		import webui.models

		url = condition.element.get_node().global_page_url

		base = settings.ALARM_URL_BASE % { 'hostname': gethostname() }

		url = base + url + '#techincal'
	except Exception, e:
		print __file__, "UNABLE TO CALCULATE CONDITION URL" % (str(e))
		url = '**error**'

	return url

def get_expr_symbols_from_element(element, cache_manager):

	symbols = {}

	#
	# Element
	#
	element = element.get_specific_element()
	element.load_fields_from_cache(cache_manager=cache_manager)

	symbols.update( get_element_check_symbols(element) )

	#
	# Node
	#
	node = element.get_node()
	if node != element:
		node.load_fields_from_cache(cache_manager=cache_manager)
		symbols.update( get_element_check_symbols(node) )

	#
	# Linked
	#
	if element.kind == ELEMENT_KIND_IFACE:
		linked_iface = element.get_linked_iface()
		if linked_iface is not None:
			linked_iface.load_fields_from_cache(cache_manager=cache_manager)

			linked_node  = linked_iface.get_node()
			linked_node.load_fields_from_cache(cache_manager=cache_manager)

			symbols.update( get_element_check_symbols(linked_iface, prefix='linked_'))
			symbols.update( get_element_check_symbols(linked_node , prefix='linked_'))

	return symbols

def get_expr_symbols_from_condition(condition):

	r = {}

	dg = condition.datagroup
	for p in dg.datagroupparameter_set.all():
		name = '$%s' % (p.name)
		try:
			v    = p.get_value()
		except Exception, e:
			v    = ''
		r[name] = v

	r['$path'] = condition.path

	return r

from sanet_poller.lib.classes import PNode, PIface, PStorage, PService, PDevice

def get_element_check_symbols(element, prefix=''):

	if   element.kind == ELEMENT_KIND_NODE   : 
		c = PNode()
		cached = {}
	elif element.kind == ELEMENT_KIND_IFACE  : 
		c = PIface()
		cached = {
			'$%sifindex' % (prefix): element.ifindex
		}
	elif element.kind == ELEMENT_KIND_STORAGE: 
		c = PStorage()
		cached = {
			'$%sstindex' % (prefix): element.stindex
		}
	elif element.kind == ELEMENT_KIND_SERVICE: 
		c = PService()
		cached = {
			'$%sswrunindex' % (prefix): element.swrunindex
		}
	elif element.kind == ELEMENT_KIND_DEVICE : 
		c = PDevice()
		cached = {
			'$%sdevindex' % (prefix): element.devindex
		}
	else:
		return {}

	c.load_from_dict( element.get_dict() )

	d = c.get_expr_symbols(prefix=prefix)
	d.update(cached)

	return d

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def load_additional_alarm_informations(log, alarm_dict):
	"""
	Questo metodo aggiunge al json dell'allarme informazioni aggiuntive
	che potrebbero essere utili per filtrare l'allarme
	"""
	try:
		log.log(9, "Loading informations for alarm: %s" % (pprint.pformat(alarm_dict)))

		_load_base_info(log, alarm_dict)

		# Tags
		_load_tags_info(log, alarm_dict)

	except Exception, e:
		log.error("Error while loading additional informations about alarm: %s" % (e))
		log.error("ERROR: %s",alarm_dict)

	return

def _load_base_info(log, alarm_dict):

	log.info("Loading base info")

	if alarm_dict['element']['parent_id'] != None:
		node_id = alarm_dict['element']['parent_id']
		sub_element_id = alarm_dict['element']['uuid']
	else:
		node_id = alarm_dict['element']['uuid']
		sub_element_id = None

	try:
		n = MonitoredNode.objects.get(uuid=node_id)
	except MonitoredNode.DoesNotExist, e:
		raise Exception("Monitored node %s not found" % (node_id))

	#
	# Node info
	#
	alarm_dict['element']['nodedescr'] = n.description

	#
	# SNMP Sys*
	#
	alarm_dict['info'] = {}
	alarm_dict['info']['sysdescr'] = n.get_sysdescr()
	alarm_dict['info']['sysname' ] = n.get_sysname()
	alarm_dict['info']['syslocation'] = n.get_sysname()

	# ATTENZIONE: se il nodo e' monitorato da un agente remoto, questa risoluzione viene chiesta all'agente remoto
	ips = n.ipv4_addresses
	if ips:
		alarm_dict['info']['ip4s'] = ips

	#
	# Linked elements
	#
	alarm_dict['linked'] = {}
	alarm_dict['linked']['node']  = None  
	alarm_dict['linked']['iface'] = None  

	if sub_element_id is not None:
		try:
			sub_e = MonitoredElement.objects.get(uuid=sub_element_id)
			sub_e = sub_e.get_specific_element()

			if sub_e.kind == ELEMENT_KIND_IFACE:

				linked_iface = sub_e.get_linked_iface()
				if linked_iface is not None:
					linked_node  = linked_iface.get_node()

					alarm_dict['linked']['node']   = linked_node.name
					alarm_dict['linked']['iface']  = linked_iface.name

		except MonitoredNode.DoesNotExist, e:
			raise Exception("Monitored node %s not found" % (node_id))

	#
	# Refs
	#
	c = Condition.objects.get(uuid=alarm_dict['condition']['uuid'])
	alarm_dict['url'] = get_condition_url(c)

	return

def _load_tags_info(log, alarm_dict):



	log.info("Loading tags info")

	alarm_dict['tags'] = []

	# nota: icket #505 - ci servono i nomi dei tag di un particolare tree
	alarm_dict['tags_name'] = {}

	try:
		e_uuid        = alarm_dict['element']['uuid']

		e_elementtype = ELEMENT_KIND_TO_TAG_ELEMENTTYPE.get( alarm_dict['element']['type'] , None)
		if e_elementtype == None:
			log.error("load_tags_info(): no element type")
			return

		c_uuid        = alarm_dict['condition']['uuid']
		c_elementtype = 'condition'

		log.debug("Loading tags for element %s:%s" % (e_uuid, e_elementtype))

		# TODO / FIX: I tag da caricare nell'allarme sono solo i tag direttamente associati all'elemento
		#             o anche altri tag "ereditati"?
		#             Esempio: gli allarmi su un'interfaccia ereditano i tag del nodo?

		# nota: ticket #505. ci servono i nomi dei tag di un particolare tree
		tags_name_by_tree = {}
		paths = set()
		for et in ElementTag.objects.filter(Q(elementid=e_uuid, elementtype=e_elementtype)|Q(elementid=c_uuid, elementtype=c_elementtype)).select_related('tag', 'tree').all():

			log.debug("\ttag %s" % (et.tag.path))
			paths.add(et.tag.path)

			tag  = et.tag
			tree = et.tree

			# nota: Ticket #505
			#tags_name = set([t.name for t in tag.get_tags_path()])
			tags_name = set([tag.name])
			names = tags_name_by_tree.get(tree.name, set())
			names = names | tags_name
			tags_name_by_tree[ tree.name ] = names

		alarm_dict['tags'] = list(paths)

		# nota: Ticket #505
		alarm_dict['tags_name'] = dict([ (tree_name, list(tags_name))   for tree_name, tags_name in tags_name_by_tree.items()  ])
		#log.debug("Loaded tags_name: %s", alarm_dict['tags_name'])

	except Exception, e:
		log.error("Error while loading tag informatiosn: %s" % (str(e)))

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

"""

             +--------------------------------------------------------------+
             | sanetd                                                       |
             |                                                              |
    QUEUE -------> *processing* ---> [ alarms_queue ] ---> AlarmProcessor   |
             |                                                              |
             +--------------------------------------------------------------+
"""


class AbstractAlarmProcessor(object):
	"""
	Classe base dell'entita' che processa gli allarmi prodotti da Sanet per essere "inviati".
	"""
	
	def __init__(self, logger):
		self.logger = logger
	
	def process_alarm(self, server, alarm_data):
		raise Exception("NotImplemented")


class MailAlarmProcessor(AbstractAlarmProcessor):
	"""
	Processor di base che invia gli allarmi tramite email.
	"""

	def process_alarm(self, server, alarm_data):

		logger      = self.logger

		mail_sender = server.mail_sender

		try:
			logger.error("[event] Tyring backup delivery")

			#
			# Estrae dall'allarme i campi della mail
			#
			mail_data = alarm_data.get('mail',None)
			if mail_data:

				frm      = 'sanetd.daemon'
				x_header = mail_data['headers'] 
				to       = mail_data['to']
				subject  = mail_data['subject']
				body     = mail_data['body']

				# >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
				# IMPORTANT: Sanet2 behaviour removed by ticket #722
				#
				#if not to:
				#	logger.error("[event] Null recipient. Email discarded: %s", alarm_data)
				#	return
				# 
				#if to == "none":
				#	logger.error("[event] Recipient 'none'. Email discarded: %s", alarm_data)
				#	return
				# <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

				mail_sender.push_email( frm, to, subject, body, x_header )

				return True
			else:
				logger.error("[event] Untable to delivery email for alarm: %s. No delivery informations", alarm_data)

		except Exception, e:
			logger.error("[event] Unable to use backup delivery: %s:%s -> %s", type(e), e, traceback.format_exc())

#-----------------------------------------------------------------------------#

from sanet_common.pythonutils import load_symbol

ALARM_PROCESSOR_FALLBACK_MODULE = 'server.sanetd.sanetdlogic.MailAlarmProcessor'

def load_alarm_handlers(logger):

	logger.info("Loading alarm handlers")

	config_entries = settings.SANET_ALARM_PROCESSORS

	if not config_entries:
		logger.error("No alarm processors configured. Forcing processor '%s'.", ALARM_PROCESSOR_FALLBACK_MODULE)
		config_entries = [ ALARM_PROCESSOR_FALLBACK_MODULE ]

	handlers = []

	# Load all configured alarm processors classes
	for config_entry in config_entries:

		try:
			#
			# TODO: <config_entry> should be:
			#       a.  a simple module name string
			#       b.  a tuple ( <module name string>, <config options dict> )
			#
			handler_module_name = config_entry

			logger.info("Loading %s", handler_module_name)

			handler_class = load_symbol( handler_module_name )
			handler       = handler_class(logger)

			handlers.append(handler)

		except Exception, e:
			logger.fatal("Unable to load alarm processor '%s': %s %s -> %s", handler_module_name, type(e), str(e), traceback.format_exc())

	return handlers

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def calculate_late_checks(now, cache_manager, mf=3):

	late_datagroups = []

	query = """
SELECT DISTINCT dg.uuid, 
       dg.path, 
       dg.minperiod
  FROM datagroup dg INNER JOIN condition c ON (c.datagroup_id = dg.uuid) 
 WHERE dg.minperiod > 0
   AND  c.status not in (%s, %s, %s, %s, %s, %s, %s)
   FOR READ ONLY
;
"""

	params = [CSTATE_UU, CSTATE_UD, CSTATE_TU, CSTATE_TD, CSTATE_IN, CSTATE_DD, CSTATE_DU]

	cursor = connection.cursor()
	cursor.execute(query, params)
	ret = cursor.fetchall()

	#
	# Find lastdone
	# 
	for (dg_uuid, dg_path, dg_minperiod) in ret:

		key = DataGroup.cache_key_from_uuid(dg_uuid)

		data = DataGroup.raw_load_fields_from_cache(cache_manager, key, ['lastdone'])

		lastdone = data.get('lastdone', None)
		if lastdone is None:
			lastdone = 0
		else:
			lastdone = int(lastdone)

		#
		# Check delay
		#
		elapsed = now - lastdone
		if (elapsed > (mf * dg_minperiod)):
			late_datagroups .append( (dg_uuid, dg_path, lastdone) )

	return late_datagroups 

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def force_datagroup_status(datagroup, logger, new_status, change_time=None, selected_condition_names=None, reason=None, cache_manager=None):

	if new_status is CSTATE_IN:
		raise ApplicationException("Can not force CSTATE_IN status.")

	alarms_to_send = []

	self = datagroup

	if cache_manager is None:
		cache_manager = init_cache_manager_from_conf()

	if change_time is None:
		change_time = time.time()

	if reason is None:
		reason = 'Datagroup to status %s at %s' % (CSTATE_TO_CODENAME(new_status), change_time)
	#
	# Update datagroup
	#
	self.lastdone = change_time
	self.save_fields_in_cache(cache_manager)

	#
	# Update conditions
	#
	conditions = self.conditions
	if selected_condition_names is not None:
		conditions = conditions.filter(name__in=selected_condition_names)

	conditions = conditions.all()
	for c in conditions:
		logger.info("Condition %s", c.path)

		last               = c.status

		if last == new_status:
			logger.info("not changed")
			continue

		c.lastdone         = change_time
		c.status           = new_status
		c.statuslast       = last
		c.statuslastchange = change_time
		c.verbstatus       = reason

		c.update_notable_state(change_time)

		c.mtime            = change_time

		#
		# Update cache
		#
		c.save_fields_in_cache(cache_manager)

		#
		# Update DB
		#
		c.save()

		logger.info("\tupdated")

		alarm_to_send = create_eventlog(logger, cache_manager, c.tenant, None, c.element, change_time, c)
		if alarm_to_send:
			alarms_to_send.append(alarm_to_send)

	return alarms_to_send

