# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET3
# SANET3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

Watchdog
============================

Il processo di watchdog (wdog) comunica con tutti gli altri processi e salva l'esito dell'interrogazione
nella cache:

::
	
	sanetd     ------+
	                 |
	entables   ------+
	                 |               WRITE                      READ
	                 +------- wdog ---------> CURRENT STATUS <-------- WEB
	                 |
	agents     ------+
	                 |
	tagsd      ------+

Processi vitali e processi opzionali
-------------------------------------

Wdog controlla lo stato di diversi processi, ma invia notifiche di errore solo quando i processi
"vitali" di sanet non sono in esecuzione.

Processi vitali:

	* Sanetd
	* Enatables
	* Agenti
	
Processi opzionali:

	* Tagsd

"""

import os
import sys
import time
import logging
import json
import traceback

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def load_wdog_settings():
	#
	# FIX: i parametri di configurazione del watchdog dovrebbero stare nel DB
	#      e non nella configurazione "statica" (django) di SANET.
	#      In questo modo possono essere modificabili via web.
	#
	#      Spostare parte della configurazione sul database e reimplementare
	#      questa funzione. Vedi ticket #620
	#
	return get_configuration().SANETD_WATCHDOG_CONF

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def sanetdwdog_main(logger, conf, wdog_settings):

	try:
		watchdog_interval      = wdog_settings['interval']
		watchdog_to            = wdog_settings['to']
		watchdog_from          = wdog_settings['from']
		watchdog_subj          = wdog_settings['subject']
		watchdog_body          = wdog_settings['body']
		wdog_send_notification_on_errors = wdog_settings['send_notification_on_errors']

		round_num = 0

		mail_sender = init_mail_sender_from_conf(logger)

		logger.info("*" * 80)
		logger.info("Starting")
		logger.info("*" * 80)
		while True:

			# Do not leave open connections. Save db usage.
			connection.close()

			try:
				if round_num != 0:
					logger.info("Pause until the next round  (%s seconds)...", watchdog_interval)
					time.sleep(watchdog_interval)

				logger.info("Started round: %s", round_num)

				#
				# FIX: check sanetd internal state
				#
				
				report_text = []

				s1, t1 = check_sanetd(logger, conf, wdog_settings)
				report_text += t1
				if not s1:
					logger.critical("Sanetd not working")

				s2, t2 = check_every_tenant(logger, watchdog_interval, wdog_settings)
				report_text += t2
				if not s2:
					logger.critical("Some agents are not working")

				#
				# FIX: MOVE THIS CODE TO "entables" subapp + SIGNALS 
				#
				s3, t3 = check_entables(logger, conf)
				report_text += t3
				if not s3:
					logger.critical("Entables is not working")

				#
				# FIX: MOVE THIS CODE TO "TAGS" subapp + SIGNALS
				#
				# Tagsd non e' un processo fondamentale quindi il watchdog non si allarma
				s4, t4 = check_tagsd(logger, conf)
				report_text += t4
				if not s4:
					logger.warning("Tagsd is not working")

				# Datasource Store
				s5, t5 = check_datasource_store(logger, conf, wdog_settings)
				report_text += t5
				if not s5:
					logger.warning("RRDCACHED is not working")

				# Do not leave open connections. Save db usage.
				connection.close()

				if not(s1 and s2 and s3 and s5):
					logger.critical("Something is not working: sanetd=%s agents=%s entables=%s storage=%s", s1, s2, s3, s5)
					logger.critical("Details:\n" + '\n'.join(report_text))

					# Skip email notification when something is not working
					if not wdog_send_notification_on_errors:
						logger.info("Notification on errors disabled.")
						continue

				logger.info("Watchdog time. Mail address: %s", watchdog_to)
				if watchdog_to:

					logger.info("Sending mail to: %s", watchdog_to)

					extra_header = {
						'X-Sanet-Watchdog-time': str(time.time()),
						'X-Sanet-Watchdog-round': round_num,
					}

					SEPARATOR = "*" * 80

					mail_body = watchdog_body + '\n' + SEPARATOR + '\n' + '\n'.join(report_text) + '\n' + SEPARATOR

					mail_sender.send(watchdog_from, watchdog_to, watchdog_subj, mail_body, extra_header)
				else:
					logger.warning("Watchdog recipient not set. Notification skipped")

			except Exception, e:
				logger.error("error: %s: %s", e, traceback.format_exc())
			finally:
				try:
					connection.close()
				except Exception, e:
					logger.error("error before restart round: %s: %s", e, traceback.format_exc())

				round_num += 1

	except Exception, e:
		logger.error("Error on round %s: %s: %s", round_num, e, traceback.format_exc())

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def check_datasource_store(logger, conf, wdog_settings):

	timeout = GET_OR_NONE(wdog_settings, 'timeouts', 'datasource_store') or GET_OR_NONE(wdog_settings, 'timeouts', 'default') or 3

	report_text = []

	logger.info("Checking datasource store (timeout=%s)", timeout)

	try:
		status     = True
		status_msg = ''

		datasouce_store = init_datasource_store_from_conf()

		data = datasouce_store.stats(timeout=timeout)

		status     = True
		status_msg = "Datasource storage is working"

		logger.info(status_msg)

		report_text += [ status_msg ]

	except Exception, e:
		status     = False
		status_msg = "Error while fetching Datasource storage status: (type %s) %s" % (type(e), str(e))

		logger.error(status_msg)

		report_text += [ status_msg ]

	sanet_subsystem_set_current_state("datasourcestore", time.time(), status, status_msg)

	return status, report_text




def check_sanetd(logger, conf, wdog_settings):

	timeout = GET_OR_NONE(wdog_settings, 'timeouts', 'sanetd') or GET_OR_NONE(wdog_settings, 'timeouts', 'default') or 3

	report_text = []

	logger.info("Checking sanetd (timeout=%s)", timeout)

	broker = '%s/%s@%s' % (conf.BROKER_AUTH[0], conf.BROKER_AUTH[1], conf.BROKER_ADDRESS[0])


	if 0:
		try:
			status     = True
			status_msg = ''

			#
			# STATS VERSION 1
			#
			json_data = sanetd_ctl(broker, 'info_json', [], timeout=timeout)

			data = json.loads(json_data)

			stats_dict = {}
			
			try:
				stats_dict.update({
					'queuesize'   : data['queue']['size'],
					'maxqueuesize': wdog_settings['sanetd_event_queue_max_size'],
				})
				
				sanetd_update_rrd_stats_file(stats_dict, wdog_settings['interval'])
				
			except Exception, e:
				logger.error("Unable to update sanetd stats: %s %s", type(e), str(e))

			try:
				stats_dict.update({
					'processqueuesize': data['processqueue']['size'],
				})
				sanetd_update_rrd_stats_file(stats_dict, wdog_settings['interval'], stats_type='queues') 
			except Exception, e:
				logger.error("Unable to update sanetd stats: %s %s", type(e), str(e))



			#
			# Checks 
			#

			# Process problems
			critical_problems = data['critical_problems']
			if critical_problems:
				status      = False
				status_msg += "Process has internal problems: %s" % (' '.join(critical_problems))

				report_text += [ status_msg ]

			# Limits
			if wdog_settings.get('sanetd_event_queue_max_size', None) is not None:
				if data['queue']['size'] >= wdog_settings['sanetd_event_queue_max_size']:
					status      = False
					status_msg += "Sanet Event queue too big: %s (max %s)" % (data['queue']['size'], wdog_settings['sanetd_event_queue_max_size'])

					report_text += [ status_msg ]

			if status:
				status_msg = "Sanetd server is running." + status_msg
				logger.info(status_msg)
			else:
				status_msg = "Sanetd server not running properly." + status_msg
				logger.error(status_msg)

			#
			# Save verbose status
			#
			sanetd_set_current_state(time.time(), True, status_msg)


		except Exception, e:

			status     = False
			status_msg = "Error while fetching sanetd status: (type %s) %s" % (type(e), str(e))

			sanetd_set_current_state(time.time(), status, status_msg)

			logger.error(status_msg)

			report_text += [ status_msg ]

	else:
		#
		# STATS VERSION 2
		#

		status     = True

		for instance_id in get_configuration().SANETD_INSTANCES:

			status_msg = ''

			try:
				#
				# GET STATS
				#
				json_data = sanetd_ctl(broker, 'info_json', [], timeout=timeout, instance_id=instance_id)

				data = json.loads(json_data)

				#
				# UPDATE STATS
				#
				
				
				stats_dict = {
					'queueread'        : data['queue']['read'],
					'queuesize'        : data['queue']['size'],
					'maxqueuesize'     : wdog_settings['sanetd_event_queue_max_size'],

					'processqueuesize' : data['processqueue']['size'],

					'processqueuerate' : data['processqueue']['rate'],
					'processqueuetot'  : data['processqueue']['processed'],

					'process.min'      : data['process']['min'] if data['process']['min'] >=0 else None ,
					'process.max'      : data['process']['max'] if data['process']['max'] >=0 else None ,
					'process.avg'      : data['process']['avg'] if data['process']['avg'] >=0 else None ,
				}

				#import pprint
				#pprint.pprint(stats_dict)

				sanetd_update_stats(logger, instance_id, stats_dict, interval=wdog_settings['interval'])

				#
				# ANALYZE
				#
				
				# Process problems
				critical_problems = data['critical_problems']
				if critical_problems:
					status      = False
					status_msg += "Process has internal problems: %s" % (' '.join(critical_problems))

					report_text += [ status_msg ]

				# Limits
				if wdog_settings.get('sanetd_event_queue_max_size', None) is not None:
					if data['queue']['size'] >= wdog_settings['sanetd_event_queue_max_size']:
						status      = False
						status_msg += "Sanet Event queue too big: %s (max %s)" % (data['queue']['size'], wdog_settings['sanetd_event_queue_max_size'])

						report_text += [ status_msg ]

				if status:
					status_msg = "Sanetd server %s is running. %s" % (instance_id, status_msg)
					logger.info(status_msg)
				else:
					status_msg = "Sanetd server %s not running properly. %s"  % (instance_id, status_msg)
					logger.error(status_msg)

				#
				# Save verbose status
				#
				sanetd_set_current_state(time.time(), True, status_msg, instance_id=instance_id)

			except Exception, e:

				status     = False
				status_msg = "Error while fetching sanetd %s status: (type %s) %s" % (instance_id, type(e), str(e))

				sanetd_set_current_state(time.time(), status, status_msg, instance_id=instance_id)

				logger.error(status_msg)

				report_text += [ status_msg ]

	return status, report_text

#-----------------------------------------------------------------------------#

def check_entables(logger, conf):

	timeout = GET_OR_NONE(wdog_settings, 'timeouts', 'entables') or GET_OR_NONE(wdog_settings, 'timeouts', 'default') or 3

	logger.info("Checking entables (timeout=%s)", timeout)

	report_text = []

	broker = '%s/%s@%s' % (conf.BROKER_AUTH[0], conf.BROKER_AUTH[1], conf.BROKER_ADDRESS[0])
	try:
		text = entables_ctl(broker, 'info', [], timeout=timeout)

		entables_set_current_state(time.time(), True, text)

		report_text += [ "Entables is running." ]

		status = True

	except Exception, e:
		logger.error("Error while fetching entables status: %s", e)
		entables_set_current_state(time.time(), False, str(e))

		report_text += [ "Entables is not running: %s" % (str(e)) ]

		status = False

	return status, report_text

#-----------------------------------------------------------------------------#

def check_tagsd(logger, conf):
	
	timeout = GET_OR_NONE(wdog_settings, 'timeouts', 'tagsd') or GET_OR_NONE(wdog_settings, 'timeouts', 'default') or 3

	logger.info("Checking tagsd (timeout=%s)", timeout)

	report_text = []

	try:
		status = get_tagsd_server_status(timeout=timeout)

		if status is None:
			raise Exception("Unable to receive Tagsd status.")

		sanet_subsystem_set_current_state("tagsd", time.time(), True, status['info'])

		report_text += [ "Tagsd is running." ]

		status = True

	except Exception, e:
		logger.error("Error while fetching entables status: %s", e)

		sanet_subsystem_set_current_state("tagsd", time.time(), False, str(e))

		report_text += [ "Tags is not running: %s" % (str(e)) ]

		status = False

	return status, report_text

#-----------------------------------------------------------------------------#

def check_every_tenant(logger, interval, wdog_settings):

	status = True

	logger.info("Checking tenants...")

	report_text = []

	for tenant in Tenant.objects.all():

		db_set_current_tenant_schema(tenant.name)

		s, text = check_tenant_status(logger, tenant, interval, wdog_settings)

		status = status and s
		report_text += text

	return status, report_text

#-----------------------------------------------------------------------------#

def check_tenant_status(logger, tenant, interval, wdog_settings):

	logger.info("Checking tenant %s....", tenant)

	status = True
	report_text = []

	agents = Agent.objects.filter(tenant=tenant, is_enabled=True).all()

	for agent in agents:

		agent_status, agent_text = check_agent(logger, agent, interval, wdog_settings)

		status = status and agent_status
		report_text += agent_text

	return status, report_text

#-----------------------------------------------------------------------------#

def GET_OR_NONE(d, *keys):
	r = d
	for k in keys:
		if not r.has_key(k):
			return None
		r = r.get(k,{})
	return r

def check_agent(logger, agent, interval, wdog_settings):

	timeout = GET_OR_NONE(wdog_settings, 'timeouts', 'tenants', agent.tenant.name, 'agents', agent.name) or GET_OR_NONE(wdog_settings, 'timeouts', 'default') or 3

	report_text = []

	now = time.time()

	logger.info("Checking agent: %s (timeout=%s)", agent, timeout)

	try:
		cmd = 'infojson'
		broker_address = get_broker_connection_string()

		#
		# GET STATS
		#
		json_str       = agent_ctl(broker_address, agent.uuid, cmd, [], timeout=timeout)

		data = json.loads(json_str)

		if data is None:
			# Should never happen
			raise Exception("BUG: Agent returned no info status")

		#
		# UPDATE
		#

		# STATS VERSION 1
		try:
			agent_update_rrd_stats_file(agent, data, interval)
		except Exception, e:
			logger.error("Unable to update agent %s stats: %s %s", agent.name, type(e), str(e))

		# STATS VERSION 2
		try:
			agent_update_stats(logger, agent, data, interval=interval)
		except Exception, e:
			logger.error("Unable to update agent %s stats: %s %s", agent.name, type(e), str(e))

		#
		# ANALYZE
		#

		status = True
		msg    = []

		#
		# Agent is running normaly or it detected some problems inside itself?
		#
		if data['critical_problems']:

			status = False

			for name, info in sorted(data['critical_problems'].items()):
				text = '%s: %s' % (name, info)
				msg.append( text )

		#
		# Wdog's Limits exeeded?
		#
		agent_settings = wdog_settings['agents'].get( agent.name, None )
		if agent_settings:

			# Wait queue size?
			wait_queue_max_size = agent_settings.get('wait_queue_max_size', None)
			if wait_queue_max_size is not None:
				if data['wqueuelen'] >= wait_queue_max_size:
					status = False
					text = 'Internal queue too big. Current size %s. Should be less than %s.' % (data['wqueuelen'], wait_queue_max_size)

					msg.append( text )

					logger.warning("%s", msg)

			# Datagroup rate?
			min_check_per_second = agent_settings.get('min_check_per_second', None)
			if min_check_per_second is not None:
				if data['dgrate'] <= min_check_per_second:
					status = False
					text = 'Check rate too small: %s/s. Minimum should be %s/s)' % (data['dgrate'], min_check_per_second)

					msg.append( text )

		if status:
			info_text = 'Agent %s:%s is running' % (agent.tenant.name, agent.name)
			logger.info("%s", info_text)
		else:
			info_text = 'Agent %s:%s has internal problems. %s' % (agent.tenant.name, agent.name, ';'.join(msg))
			logger.error("%s", info_text)

		agent.set_current_state(now, status, '\n'.join(msg))

		report_text += [ info_text ]

	except Exception, e:
		logger.error("error while checking agent %s: %s", agent, e)
		agent.set_current_state(now, False, "%s" % (str(e)))

		report_text += [ 'Agent %s:%s is not running: %s' % (agent.tenant.name, agent.name, str(e)) ]
		status = False

	return status, report_text

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def init_logger(log_leve, log_file):

	log_format = '%(asctime)s %(levelname)s wdog> %(message)s'

	logger = logging.getLogger('wdog')
	logger.setLevel( log_level )

	hdlr = logging.FileHandler(log_file, "a")
	hdlr.setFormatter( logging.Formatter(log_format) )
	hdlr.setLevel( log_level )
	logger.addHandler(hdlr)

	hdlr = logging.StreamHandler()
	hdlr.setFormatter( logging.Formatter(log_format) )
	hdlr.setLevel( log_level )
	logger.addHandler(hdlr)

	return logger

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

if __name__ == "__main__":

	# Be sure django will find modules with the right path
	sanet_install_dir = sys.argv[1]
	if not sanet_install_dir in sys.path:
		sys.path.insert(0, sanet_install_dir)
	os.environ.setdefault("DJANGO_SETTINGS_MODULE", "conf.settings")

	# Changed for django 1.11
	import django
	django.setup()

	#
	# Now we are inside django
	#

	from django.db import connection

	from conf import get_configuration
	from conf import  get_broker_connection_string

	from core.models import Tenant, Agent
	from core.api.base import db_set_current_tenant_schema 
	from core.api.mail import init_mail_sender_from_conf
	from core.api.datasource import init_datasource_store_from_conf

	from core.api.sysstats import sanetd_set_current_state

	from core.api.sysstats import entables_set_current_state

	from core.api.sysstats import sanet_subsystem_set_current_state

	# stats V1
	from core.api.sysstats import sanetd_update_rrd_stats_file
	from core.api.sysstats import agent_update_rrd_stats_file

	# stats V2
	from core.api.sysstats import agent_update_stats
	from core.api.sysstats import sanetd_update_stats

	#
	# Funzioni per ottenere lo stato dei processi
	#
	from server.sanetd.sanetdctl import sanetd_ctl

	from sanet_poller.agentctl import agent_ctl

	from entables.management.commands.entables_ctl import entables_ctl

	from tags.server.tagsd.api import get_tagsd_server_status

	#
	# INIT
	#
	static_conf   = get_configuration()
	wdog_settings = load_wdog_settings()

	# Set log 
	log_level = wdog_settings['log_level']
	log_file  = os.sep.join( [static_conf.LOG_DIR, 'wdog_log'])

	logger = init_logger(log_level, log_file)

	# Just a warning message
	if sanet_install_dir != static_conf.SANET_INSTALL_DIR:
		msg = "Install dir %s is different from configuration install dir %s" % (sanet_install_dir, static_conf.SANET_INSTALL_DIR)
		sys.stderr.write(msg)
		logger.error(msg)

	#
	# LAUNCH
	#
	sanetdwdog_main(logger, static_conf, wdog_settings)

