# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""
Statusd Server
====================================================================================

Introduzione
---------------------------------------------------------------------------

Statusd e' il processo di Sanet che viene utilizzato dagli agenti per ottenere on-demand,
dal server centrale, dati che non riguardano configurazione per effettuare monitoraggio.

Schema thread interni e strutture dati
---------------------------------------------------------------------------

.. image:: ../_static/stated.png
	:width: 800

Gestione delle richieste
---------------------------------------------------------------------------

Le richieste inviate al demone tramite la CODA sono tuple del tipo:

::
	
	( "<name>", <params dict> )

I parametri sono:

	* *name* : stringa con il nome della richiesta
	* *params dict*: dizionario (key,value) con i parametri specifici per quella richiesta.

I dati restituiti dopo aver processato la richiesta dipendono dalla richiesta, ma **e' fortmente
consigliato** utilizzare un dizionario per codificare i dati nella risposta, in modo
che possa essere successivamente esteso.

Le *richieste* supportate dal Statusd possono essere:

	* Richieste *standard*: la gestione di queste richieste e' prevista di base dal server.
	* Richieste *non standard* la cui gestione viene delegata alle sub-app. Vedi :ref:`statusd-register-new-request-handlers`.
	
Richieste standard
---------------------------------------------------------------------------

TODO





.. _statusd-register-new-request-handlers:

Estendere il demone da una sub-app
---------------------------------------------------------------------------

Il demone puo' essere esteso (senza modifiche al codice) da sub-app per gestire
nuovi tipi di richieste

Si rimanda alla sezione :ref:`dummy-plugin-extend-statusd-register-new-requests`

..
			XXXXXXXXXXXXXXXXXXXXXXX OLD SCHEMA
			::
			   
				 IPC      |
				LAYER     |                  SERVER
					  |  
			   STATED_QUEUE --|------------>  ListenThread
					  |                    |         
					  |             [request queue]  
					  |                    |         
					  |                    +------------------+         
					  |                    |                  |
					  |                    v                  v          
					  |              worker thread      worker thread    
					  |                    |                  |
					  |                    |<-----------------+
					  |                    v
					  |             [response queue]
					  |                    |
					  |                    v
					  |              ResponseThread
					  |                    |
				      * <-|--------------------+
					  |
			XXXXXXXXXXXXXXXXXXXXXXX OLD SCHEMA
"""

import os
import sys
import time
import logging

import threading
import Queue

import socket

from sanet_common.threadutils import AbstractDaemonThread

from sanet_common.sanetdefs.ipc import SANETD_STATE_QUEUE

from sanet_common.ipc import RESPONSE_SUCCESS, RESPONSE_ERROR
from sanet_common.ipc import decode_request, encode_response_with_id


# These imports depend on django
from django.db import connection
from django.db import connection

from conf import get_configuration
from core.models import *

from conf import get_configuration

from core.models import *

from core.api.base import *
from core.api.datasource import init_datasource_store_from_conf

from server.sanetd.sanetdlogic import create_ipc_connection_from_config

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class AbstractThread(AbstractDaemonThread):
	"""
	Classe base di tutti i thread di questo server.
	"""
	def __init__(self, server):
		
		super(AbstractThread, self).__init__()

		self.server = server

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class WorkerThread(AbstractThread):
	"""
	Questi thread vengono istanziati dal ListenThread per processare in parallelo
	piu' richieste (ThreadPool).
	"""

	def __init__(self, id, server, listen_thread):

		super(WorkerThread, self).__init__(server)

		self.id = id
		self.listen_thread = listen_thread

	def run(self):

		logger = self.server.logger

		logger.info("Started worker: %s", self.id)
		while 1:
			try:
				try:
					data = self.listen_thread.request_queue.get(True, 1) 
				except Queue.Empty, er:
					continue

				logger.log(1, "Worker(%s): processing request: %s", self.id, data)

				request_id, reply_address, command, params = data

				self.listen_thread.handle_request( request_id, reply_address, command, params )

			except Exception, e:
				logger.error("Worker(%s): %s:%s", self.id, type(e), str(e))

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class ListenThread(AbstractThread):

	def __init__(self, server):

		super(ListenThread, self).__init__(server)

		self.threads       = None
		self.request_queue = None

	def run(self):

		logger = self.server.logger

		self.request_queue = Queue.Queue(-1)

		self.threads = [ WorkerThread(x, self.server, self) for x in xrange(5) ]
		for t in self.threads:
			t.start()

		logger.debug("Running IPC listener...")

		while 1:
			if not self.server.isRunning():
				break

			ipc_connection = None
			try:
				self.server.init_connection()

				ipc_connection = self.server.get_connection()

				self.ipc_connection = ipc_connection

				logger.debug("Bount to %s", SANETD_STATE_QUEUE)

				while self.server.isRunning():

					# N.B: x django: free db connections
					connection.close()

					# Read request
					logger.debug("Waiting...")
					(replay_addr, msg) = ipc_connection.receive_msg(timeout=1, ack_msg=True)
					if msg == None:
						continue

					logger.info("Decoding request")
					try:
						request = decode_request(msg)
					except Exception, e:
						logger.error("[ctl] Invalid request format: %s", e)
						continue

					request_id = request['id']

					# Handle request
					command    = request['type']
					params     = request['params'][0]    

					self.process_request(request_id, replay_addr, command, params)

			except Exception, e:
				logger.error("Unknown error in listen thread: %s:%s" % (type(e), str(e)))
				time.sleep(1)
			finally:
				if ipc_connection:
					ipc_connection.close()

		logger.info("Stopping")

	def process_request(self, request_id, reply_address, command, params):
		self.request_queue.put_nowait( (request_id, reply_address, command, params) )

	def signal_server_to_stop(self):
		self.server.stop_server()

	def handle_request(self, request_id, replay_addr, command, params):
		try:
			response = self._handle_request(command, params)
			response = encode_response_with_id( RESPONSE_SUCCESS, request_id, response )
		except Exception, e:
			response = "%s:%s" % (type(e), str(e))

			self.server.logger.error(response)

			response = encode_response_with_id( RESPONSE_ERROR, request_id, response )

		self.server.queue_response(replay_addr, response)

	def _handle_request(self, command, params):

		response = None

		connection.close()

		logger = self.server.logger

		try:
			if  command == 'get_server_status':

				#
				# RESPONSE
				#
				response = { 
					'data': { 
						'status': 'OK', 
						'info'  : 'Server is running normally.' 
					} 
				}

			elif command == 'get_condition_status':

				#
				# PARAMS
				#
				tenant_id = params['tenant_id']
				rid       = params['path']
				node_name = params.get('node_name', None)
				
				#
				# GET STATE
				#
				c   = _get_condition(tenant_id, rid, node_name)
				
				#
				# RESPONSE
				#
				response = { 
					'data': c.status 
				}

			elif command == 'get_datasource_data':

				logger.debug('get_datasource_data( %s ) ' % (params))

				#
				# Parameters
				#
				tenant_id       = params['tenant_id']
				ts_end          = params['ts_end']
				ts_start        = params['ts_start']

				path            = params.get('path'           , None)
				ds_name         = params.get('ds_name'        , None)

				node_name       = params.get('node_name'      , None)
				subelement_kind = params.get('subelement_kind', None)
				subelement_name = params.get('subelement_name', None)

				classification  = params.get('classification' , None)

				# move to tenant
				tenant = tenant_get_by_id(tenant_id)
				db_set_current_tenant_schema(tenant.name)

				search_params = { }

				#
				# Search
				#
				try:
					if path:
						search_params['path'] = path 
					else:
						if classification:
							search_params['classification'] = classification 

						if ds_name:
							search_params['name'] = ds_name

					ds = DataSource.objects.filter(**search_params)

					if subelement_name and subelement_kind:
						#search_params.update({ 
						#	'element__parent__name': node_name ,
						#	'element__kind'        : subelement_kind,
						#	'element__name'        : subelement_name,
						#})

						#ds = DataSource.objects.get(**search_params)

						f_node_sub_element = { 
							'element__parent__name': node_name ,
							'element__kind'        : subelement_kind,
							'element__name'        : subelement_name,
						}
						f_node = {
							'element__name': node_name ,
							'element__kind': ELEMENT_KIND_NODE,
						}

						ds = ds.filter(Q(**f_node_sub_element)|Q(**f_node))
						ds = ds.order_by('-element__kind')
					else:
						#search_params.update({ 
						#	'element__name'          : node_name,
						#	'element__parent__isnull': True,
						#})

						ds = ds.filter(Q(element__name=node_name, element__kind=ELEMENT_KIND_NODE)|Q(element__parent__name=node_name))
						ds = ds.order_by('-kind')

					#
					# Fetch
					#
					ds = ds.all()[:1]
					if ds:
						ds = ds[0]
					else:
						raise Exception("Not found")

				except DataSource.DoesNotExist, e:
					raise Exception("Datasource not found! Filters: %s" % (search_params))

				#
				# fetch data
				#
				ds_store = init_datasource_store_from_conf()
				fetched_data = ds_store.fetch_values(tenant_id, ds.uuid, ts_start=ts_start, ts_end=ts_end)

				serie_values = fetched_data['dataseries']

				response = {
					'data': serie_values 
				}

			elif command == 'get_node_up_time':

				tenant_id = params['tenant_id']
				node_id   = params['node_id']
				time_range  = params['time_range']
				path      = params['path']

				ts_end = time.time()
				if time_range is None:
					time_range = 60 * 60 * 24
				ts_start = ts_end - time_range

				# move to tenant
				tenant = tenant_get_by_id(tenant_id)
				db_set_current_tenant_schema(tenant.name)

				node = MonitoredNode.objects.get(uuid=node_id)

				if path is None:
					c = node.monitored_conditions.get(primary=True)
				else:
					c = Condition.objects.get(tenant=tenant, path=path)

				logger.debug("Using condition: %s", c.path)

				changes = c.logchanges_sequence(ts_start, ts_end)

				totals_by_state = {}
				for ts_start,ts_end,duration,status in changes:
					#self.logger.debug("%s %s %s %s", ts_start, ts_end, duration, status)
					code = CSTATE_TO_CODENAME(status)
					v = totals_by_state.get(code, 0)
					v += duration
					totals_by_state[code] = v

				response = { 'states_totals': totals_by_state }

			elif command  == 'restart':
				self.signal_server_to_stop()
				response = {
					'info': "Command received. Stopping..." 
				}
			else:
				#
				# Trigger handlers from installed apps
				#
				from core.models import statusd_handle_request

				logger.info("Signal request to sub-apps")

				signal_ret_data = statusd_handle_request.send(
					sender  = None,
					server  = self.server, 
					command = command,
					params  = params,
				)

				logger.info("Request processed by %s subapps", len(signal_ret_data))

				#
				# Any valid response?
				#
				response = None

				for signal_handler, ret_data in signal_ret_data:
					if ret_data is not None:
						logger.debug("Request processed: Return data: %s", ret_data)
						response = ret_data
						break

				if response is None:
					raise Exception("Invalid request '%s'" % (command))

		finally:
			connection.close()

		return response

	def send_response(self, reply_address, response):
		self.server.responses_queue.put_nowait( (reply_address, response) )

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class ResponseThread(AbstractThread):
	"""
	Questo thread legge dalla coda  server.reaponses_queue  le risposte da inviare
	via IPC a chi aveva mandato le richieste.
	"""

	def __init__(self, server):
		super(ResponseThread, self).__init__(server)

	def run(self):

		self.server.logger.info("ResponseThread: Response thread started...")

		while 1:
			try:
				try:
					response_to_send = self.server.responses_queue.get(True, self.server.ipc_timeout)
				except Queue.Empty, er:
					response_to_send = None

				if not response_to_send:
					continue

				self.server.logger.log(1, "ResponseThread: sending response: %s", response_to_send)

				reply_address, response = response_to_send

				self.server.send_response(reply_address, response)

			except Exception, e:
				self.server.logger.error("ResponseTread: error %s:%s", type(e), str(e))

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def _get_condition(tenant_id, rid, node_name):

	tenant = Tenant.objects.get(uuid=tenant_id)
	tenant.db_set_current_schema()


	c = None
	try:
		c = Condition.active_objects.get(uuid=rid)
	except Condition.DoesNotExist, e:
		print "NOT FOUND1"
		pass

	if not c:
		try:
			c = Condition.objects.get(path=rid)
		except Condition.DoesNotExist, e:
			print "NOT FOUND2"
			pass

	# Classification
	if not c:
		qs = Condition.objects.filter(classification=rid)

		if node_name:
			qs = qs.filter(Q(element__name=node_name, element__kind=ELEMENT_KIND_NODE)|Q(element__parent__name=node_name))
			qs = qs.order_by('-element__kind')
		
		print qs.count()
		
		qs = qs[:1]
		if qs:
			c = qs[0]
			print "Found:", c.path
		else:
			print "NOT FOUND3"
			pass


	if c == None:
		raise Exception("Condition with id '%s' not found" % (rid))

	return c

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#


def init_logger(log_level, log_file):

	logger = logging.getLogger('statusd')
	logger.setLevel( log_level )

	hdlr = logging.FileHandler(log_file, "a")
	hdlr.setFormatter( logging.Formatter('%(asctime)s %(levelname)s discovery > %(message)s') )
	hdlr.setLevel( log_level )
	logger.addHandler(hdlr)

	hdlr = logging.StreamHandler()
	hdlr.setFormatter( logging.Formatter('%(asctime)s %(levelname)s discovery > %(message)s') )
	hdlr.setLevel( log_level )
	logger.addHandler(hdlr)

	return logger


#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#


class StatusDServer(object):

	def __init__(self, logger):
		self.logger = logger
		self._loop = 1

		self.ipc_timeout    = 1
		self.ipc_connection = None

		self.responses_queue = None

	def init_connection(self):
		self.ipc_connection = create_ipc_connection_from_config(self.logger)
		self.ipc_connection.bind(SANETD_STATE_QUEUE)
		self.ipc_connection.open()

	def get_connection(self):
		return self.ipc_connection

	def queue_response(self, reply_address, response):
		self.responses_queue.put_nowait( (reply_address, response) )

	def send_response(self, reply_address, response):

		# Send response
		self.logger.info("Sending response...")

		self.ipc_connection.send_msg( reply_address, response, blocking=False)

	def start(self):
		self.logger.info("Starting server...")

		self.responses_queue = Queue.Queue(-1)

		self.spawn_sub_threads()

	def stop_server(self):
		self.logger.info("Server stop signaled.")
		self._loop = 0

	def isRunning(self):
		return (self._loop == 1)

	def load_configuration(self):
		conf = get_configuration()

	def spawn_sub_threads(self):

		threads = []

		self.logger.info("Init IPC listener")
		t = ListenThread(self)
		threads.append(t)

		t = ResponseThread(self)
		threads.append(t)

		self.logger.info("Launching")
		for t in threads:
			t.start()

		# N.B: x django: free db connections
		connection.close()

		self.logger.info("Waiting for threads")
		for t in threads:
			while t.isAlive():
				t.join(1)

		self.logger.info("SERVER STOP")

