# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

# TEST WITH 
#    snmptrap -v 2c -c public localhost '' 1.2.3.4.5   1.3.6.1.2.1.1.6.0     s "just here"
"""
Trapd Server
=====================================================================================

Il server Trapd riceve le trap in due modi:

	* Stando in ascolto su un socket UDP
	* Stando in ascolto sulla coda  SNMPTRAP_DATA_QUEUE. Su questa cosa arrivano le trap inoltrate da agenti remoti.

TODO  immagine schema thread

"""
import os
import sys
import re
import logging
import traceback

import time
import socket

import pickle
import base64

import threading

from sanet_common.ipc import *
from sanet_common.sanetdefs.ipc import SNMPTRAP_DATA_QUEUE

from sanet_common.netutils import ber
from sanet_common.netutils.snmptrap import decode as snmptrap_decode

from sanet_common.timeutils import timestamp_to_isostr

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class BaseTrapReceiverThread(threading.Thread):

	def __init__(self, logger, server):

		super(BaseTrapReceiverThread, self).__init__()

		# Don't block main program on exit
		# FIX: hack for old API version.
		if (hasattr(self, "daemon")):
			self.daemon = True

		if hasattr(self, "setDaemon"):
			self.setDaemon(True)

		self.logger = logger

		self.server = server

#-----------------------------------------------------------------------------#

class UDPTrapReceiverThread(BaseTrapReceiverThread):

	def __init__(self, logger, server, listen_port):
		super(UDPTrapReceiverThread, self).__init__(logger, server)
		self.listen_port = listen_port

	def run(self):

		listen_socket = None

		while 1:
			try:
				self.logger.info("[UDP trap] Opening UDP listen socket")

				listen_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
				listen_socket.bind( self.listen_port )

				while 1:

					self.logger.info( '[UDP trap] Waiting for trap')
					(data, address_tuple) = listen_socket.recvfrom(4096)

					receive_time = time.time()

					self.server.process_trap_data(receive_time, address_tuple, data)
			except Exception, e:
				self.logger.error("[UDP trap] %s", e)

			try:
				if listen_socket:
					listen_socket.close()
			except Exception, e:
				self.log.error("[UDP trap] %s", e)

			time.sleep(2)
			self.logger.info("[UDP trap] Restarting...")


#-----------------------------------------------------------------------------#

class QueueTrapReceiverThread(BaseTrapReceiverThread):
	
	def __init__(self, logger, server):
		super(QueueTrapReceiverThread, self).__init__(logger, server)
	
	def run(self):
		
		self.load_configuration()
		
		self.logger.info("[IPC trap] Start...")
		
		while True:
			
			try:
				self.logger.info("[IPC trap] Creating connection...")
				self.queue_events = self.create_ipc_connection()
				self.queue_events.open()
				
				self.queue_events.bind(SNMPTRAP_DATA_QUEUE, persistent=True)
				
				self.logger.log(9,"[IPC trap] Reading event from %s", SNMPTRAP_DATA_QUEUE)
				
				while 1:
					(addr, msg) = self.queue_events.receive_msg(timeout=1, ack_msg=False)
					self.logger.debug("received = %s", msg)
					
					if msg is None:
						continue
					
					data = json.loads(msg)
					receive_time  = data['ts']
					address_tuple = data['address']
					data          = base64.b64decode(data['data'])
					
					#
					# FIX: se l'orario di ricezione della trap originale e' troppo "sfasato" rispetto
					#      all'orario di arrivo dei dati in questo punto forse sarebbe il caso di
					#      droppare la TRAP.
					#
					
					self.server.process_trap_data(receive_time, address_tuple, data)
				
				self.logger.log(9,"[IPC trap] stop reading event from %s", self.SANETD_EVENTS_QUEUE)
			
			except IPCConnectionError, e:
				self.logger.error("[IPC trap] Problem on connection. Trying reconnecting")
			except Exception, e:
				self.logger.error("[IPC trap] Unkonwn: %s:%s -> %s" % (type(e), e, traceback.format_exc()))
			finally:
				pass
			
			try:
				self.logger.warning("[IPC trap] Closing event queue")
				if self.queue_events:
					self.queue_events.close()
			except Exception, e:
				self.logger.warning("[IPC trap] error while closing queue_events: %s:%s -> %s", e, traceback.format_exc())
				pass
			
			self.logger.warning("[IPC trap] Restart...")
			
			time.sleep(2)
			
		self.logger.debug("[IPC trap] Exit")

	def create_ipc_connection(self):
		return create_ipc_connection(self.logger, self.queue_broker_address, self.queue_broker_auth)		

	def load_configuration(self):
		
		conf = get_configuration()
		
		# Set log 
		self.log_level = conf.SANETD_LOG_LEVEL
		self.log_file  = os.sep.join([conf.LOG_DIR, 'sanetd.log'])
		self.queue_broker_address = conf.BROKER_ADDRESS
		self.queue_broker_auth    = conf.BROKER_AUTH		

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

from sanet_poller.lib.exceptions import UndefinedSnmpValueException
from sanet_poller.lib.evaluation import ExecutionContext
from sanet_poller.lib.evaluation import SanetSnmpValueType

from sanet_poller.lib.classes import DataGroupChecker

class TrapSanetSnmpValueType(SanetSnmpValueType):
	"""
	Quando si usano le trap i valori SNMP calcolati dalle expr non sono normalmente letti dalla rete
	ma sono presi nel contesto di esecuzione.
	"""
	def __init__(self, *args, **kwargs):
		super(TrapSanetSnmpValueType, self).__init__(*args, **kwargs)
	
	def effective_snmpget(self):
		
		context   = self.context
		host      = self.host
		oid       = self.oid
		community = self.community
		port      = self.port
		sv        = self.sv
		
		#
		# FIX: non so se sarebbe utile poter prendere dalla rete OID di altri nodi e dal 
		#      contesto locale le OID della trap per il nodo corrente.
		#      Per ora stiamo sul semplice e limitiamoci ai dati della TRAP.
		#
		try:
			v = self.context.trap_varbindlist_by_oid[self.oid]
			
			old_id  = "%s%s%s:%s" % (oid, '#', host, port)
			context.saveCurrentValue(old_id, v)
			
			return v
		except KeyError, e:
			raise UndefinedSnmpValueException("OID %s not found in TRAP varbindlist" % (oid) )
	
	def effective_snmpwalk(self):
		#
		# FIX: ha senso poter fare delle walk sui dati della trap?
		#
		raise Exception("SNMP Walk not supported on TRAP varbindlist data")

class SnmpTrapExecutionContext(ExecutionContext):
	"""
	Il constesto di esecuzione delle TRAP e' diverso da quello normale. C'e' una diversa
	gestione dei valori SNMP (vedi TrapSanetSnmpValueType)
	"""
	def __init__(self, trap_varbindlist):
		super(SnmpTrapExecutionContext, self).__init__()
		
		self.trap_varbindlist_by_oid = dict(trap_varbindlist)
	
	def calculateLazySnmpValue(self, oid, community, host, port):
		return TrapSanetSnmpValueType(self, oid, community, host, port)

class SnmpTrapDataGroupChecker(DataGroupChecker):
	"""
	La classe per la valutazione di un datagroup deve gestire in maniera particolari i dati
	della trap
	"""
	def __init__(self, *args, **kwargs):
		super(SnmpTrapDataGroupChecker, self).__init__(*args, **kwargs)
		self.varbindlist = []
	
	def setSnmpVarBindList(self, trap_varbindlist):
		self.trap_varbindlist = trap_varbindlist
	
	def createNewExecutionContext(self):
		return SnmpTrapExecutionContext(self.trap_varbindlist)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class TrapDServer(object):
	
	def __init__(self, logger, mail_sender, listen_port, restart_pause):
		
		self._loop = 0
		
		self.logger = logger
		self.mail_sender = mail_sender
		
		self.listen_port   = listen_port
		self.restart_pause = restart_pause
	
	def load_configuration(self):
		conf = get_configuration()
		self.queue_broker_address = conf.BROKER_ADDRESS
		self.queue_broker_auth    = conf.BROKER_AUTH
	
	def create_ipc_connection(self):
		return create_ipc_connection(self.logger, self.queue_broker_address, self.queue_broker_auth)
	
	def start(self):
		
		self._loop = 1
		
		self.logger.info("Starting trapd server...")
		
		self.logger.info("Loading configuration")
		self.load_configuration()
		
		udp_thread = UDPTrapReceiverThread(self.logger, self, self.listen_port)
		udp_thread.start()
		
		ipc_thread = QueueTrapReceiverThread(self.logger, self)
		ipc_thread.start()
		
		while self._loop:
				self.logger.debug("idle")
				time.sleep(10)
	
	@staticmethod
	def trapdata_to_msg(trap_data):
		msg = "Version: %s\nCommunity: %s\nVarbind list:\n%s" % (
			trap_data['version'], 
			trap_data['community'],
			'\n'.join( [ "%s = %s" % (oid,repr(val)) for oid,val in trap_data['varbindlist'] ] ),
		)
		return msg
	
	def process_trap_data(self, receive_time, address_tuple, data):
		
		connection.close()
		
		tenant = Tenant.objects.get(is_primary=True)
		db_set_current_tenant_schema(tenant.name)
		
		#
		# DECODIFICHIAMO LA TRAP
		#
		self.logger.info( 'Trap from %s port %d' % (address_tuple[0], address_tuple[1])) 
		
		ber_data  = ber.decode(data)
		trap_data = snmptrap_decode(ber_data)
		
		if self.logger.isEnabledFor(logging.DEBUG):
			
			self.logger.debug("RAW  Data: %s", repr(data))
			self.logger.debug("BER  Data: %s", repr(ber_data))
			self.logger.debug("TRAP Data: %s", repr(trap_data))
		
		# Get trap oid
		OID = '1.3.6.1.6.3.1.1.4.1.0'
		trapoid = None
		for oid,v in trap_data['varbindlist']:
			if oid.startswith(OID):
				trapoid = v
				break
		if trapoid == None:
			trapoid = '1.3'
		
		self.logger.info("Trap OID: %s", trapoid)
		
		saved_data = pickle.dumps( ber_data )
		saved_data = base64.b64encode( saved_data )
		
		#
		# Try to find the name used in sanet
		#
		source_ip = address_tuple[0]
		
		# Find sanet node
		sanet_node = None
		sanet_node_name = None
		
		self.logger.info("Remote host has ip = %s", source_ip)
		
		#sanet_node = TrapDServer.get_sanet_node_from_sourceip(logger, source_ip)
		sanet_node = MonitoredNode.get_from_sourceip(source_ip)
		
		if sanet_node is not None:
			self.logger.info("Found sanet node with forced ip %s = %s", source_ip, sanet_node.name)
			sanet_node_name = sanet_node.name
		else:
			sanet_node, sanet_node_name = guess_node_from_ip(self.logger, source_ip)
		
		logger.info("Source ip          = %s", source_ip)
		logger.info("Selected node name = %s", sanet_node_name)
		
		#
		# SALVIAMO LA TRAP NEL LOG
		#
		l = self.new_snmptrap_log(
			self.logger,
			tenant, 
			receive_time,
			source_ip, 
			trapoid, 
			trap_data,
			node_name = sanet_node_name
		)
		l.save()
		
		#
		# PROCESSIAMO LA TRAP CON DATAGROUP
		#	
		if sanet_node is not None:
			try:
				ipc_connection = self.create_ipc_connection()
				self.process_trap_with_datagroup(logger, ipc_connection, receive_time, source_ip, sanet_node, trapoid, trap_data)
			except Exception, e:
				logger.critical("Unable to update datagroup with trap: %s %s", type(e), e)
		else:
			self.signal_unknown_trap(logger, 
				receive_time,
				source_ip, 
				trapoid, 
				trap_data
			)
	
	def new_snmptrap_log(self, logger, tenant, timestamp, source_ip, trapoid, trap_data, site_name=None, node_name=None):
		
		#
		# Salviamo l'evento sul DB per iniziare
		#
		e = Event()
		
		e.tenant = tenant
		
		if node_name == None:
			node_name = sourceip
		
		e.kind         = EVENT_SNMPTRAP
		e.timestamp    = timestamp
		e.sourceip     = source_ip
		e.element_path = node_name
		e.trapoid      = trapoid 
		e.varbindlist  = trap_data['varbindlist']
		e.msg          = TrapDServer.trapdata_to_msg(trap_data)
		e.save()
		
		logger.info("Saved trap on DB")
		
		return e
	
	def process_trap_with_datagroup(self, logger, ipc_connection, receive_time, source_ip, sanet_node, trapoid, trap_data):
		
		logger.info("Processing trap %s on node %s", trapoid, sanet_node)
		
		datagroups_to_process = []
		
		#
		# Troviamo tutti i datagroup da processare
		#
		
		# 1. Datagroup che fanno match esatto con sul campo "passive_match"
		base_query = DataGroup.objects.filter(element=sanet_node, minperiod=0)
		
		datagroups_to_process += list(base_query.filter(passive_match=trapoid).all())
		
		# 2. Datagroup che fanno match usando passive_match come espressione regolare
		for dg in base_query.filter(passive_match__startswith=DataGroup.PASSIVE_MATCH_REGEX_PREFIX).all():
			logger.debug("\tmatching %s passive_match=%s with %s", dg.path, dg.passive_match, trapoid)
			regexp = dg.passive_match.replace(DataGroup.PASSIVE_MATCH_REGEX_PREFIX, '')
			if re.match(regexp, trapoid):
				datagroups_to_process.append(dg)
		
		logger.info("Found %s matching trapoid", len(datagroups_to_process) )
		
		if len(datagroups_to_process) == 0:
			
			self.signal_unmanaged_trap(logger, 
				receive_time,
				source_ip, 
				sanet_node,
				trapoid, 
				trap_data
			)
			return
		
		#
		# PROCESS
		#
		varbindlist = trap_data['varbindlist']
		
		for dg in datagroups_to_process:
			logger.info("Processing %s", dg.path)
			
			tenant = dg.tenant
			db_set_current_tenant_schema(tenant.name)
			datagroup_id = dg.uuid
			
			import pprint
			import json
			from sanet_poller.lib.classes   import DataGroupChecker
			from sanet_poller.lib.functions import load_system_expr_functions
			from sanet_poller.lib.functions import load_extra_expr_functions
			from sanet_poller.lib.communication import AgentUpdateEvent
			
			from core.api.cache import init_cache_manager_from_conf
			from core.api.datasource import init_datasource_store_from_conf
			from core.localpolling import LocalConfiguration
			
			from server.sanetd.sanetdlogic import ServerIPCFactory
			
			cache_manager = init_cache_manager_from_conf()
			ipc_factory   = ServerIPCFactory(logger)
			datasource_store = init_datasource_store_from_conf()
			
			#
			# Simboli disponiboli
			#
			symbols = {}
			
			logger.debug( "Loading system symbols")
			system_symbols = load_system_expr_functions()
			
			logger.debug("Loading extra symbols from %s", conf.LOCAL_POLLER_EXTRA_MODULES)
			extra_symbols = load_extra_expr_functions(conf.LOCAL_POLLER_EXTRA_MODULES)
			
			symbols.update(system_symbols)
			symbols.update(extra_symbols)
			
			# Creiamo l'oggetto che ha il compito fornire i vari oggetti del monitoraggio
			# (nodi,interfacce,ecc...) che potranno servire durante la valutazione
			# del datagroup.
			c = LocalConfiguration(logger, tenant.uuid, cache_manager)
			
			c.set_global_symbols(symbols) 
			
			# Otteniamo l'oggetto PDataGroup associato al datagroup selezioanto
			pdg = c.get_datagroup(datagroup_id)
			
			# Istanziamo la classe che contiene tutta la logica per la valutazione di un 
			# datagroup.
			dgc = SnmpTrapDataGroupChecker(c, logger, cache_manager, ipc_factory, pdg )#, system_symbols)
			
			dgc.setSnmpVarBindList(varbindlist)
			
			logger.debug( "########################################################" )
			logger.debug( "###################### CHECKING ########################" )
			logger.debug( "########################################################" )
			
			logger.info( "Start checking...")
			
			# CHECK
			t1 = time.time()
			dgc.check( time.time() )
			t2 = time.time()
			
			logger.debug( "########################################################" )
			logger.debug( "####################### SENDING ########################" )
			logger.debug( "########################################################" )
			
			agent = None
			
			updates_data = dgc.get_datagroup_all_updates()
			
			logger.debug( "Creating update event" )
			fake_event = AgentUpdateEvent(time.time(), tenant.uuid, agent, updates_data)
			
			logger.debug("Sending update data: %s", pprint.pformat(fake_event))
			
			logger.info("Sending update data to sanetd broker" )
			
			events = [fake_event, ]
			
			connection = ipc_connection
			connection.open()
			
			events = json.dumps(events)
			connection.send_msg( 'sanetd-events', events, blocking=False)
			
			connection.close()
			logger.info("Sent. Exit." )
		
		return
	
	def signal_unknown_trap(self, logger, receive_time, source_ip, trapoid, trap_data):
		
		mail_params = settings.SNMPTRAP_MAIL_CONF['unknown']
		
		self._signal_strange_trap(logger, mail_params, receive_time, source_ip, trapoid, trap_data)

	def signal_unmanaged_trap(self, logger, receive_time, source_ip, sanet_node, trapoid, trap_data):
		
		mail_params = {}
		mail_params.update(settings.SNMPTRAP_MAIL_CONF['unmanaged'])
		
		# Se il nodo specifica un indirizzo per le segnalazioni, usiamo quello invece del default.
		if sanet_node.notify_address:
			logger.info("Sending 'unmanaaged trap' infos to %s for node %s (instead of %s)", sanet_node.notify_address, sanet_node.name, mail_params['to'])
			mail_params['to'] = sanet_node.notify_address
		
		additional_data = {
			'sanet_node': sanet_node.name
		}
		
		self._signal_strange_trap(logger, mail_params, receive_time, source_ip, trapoid, trap_data, additional_data=additional_data)
	
	def _signal_strange_trap(self, logger, mail_params, receive_time, source_ip, trapoid, trap_data, additional_data={}):
		
		def render_template_string(string, context):
			r = string+""
			for k,v in context.items():
				r = r.replace("{{"+k+"}}", str(v))
			return r
		
		try:
			d = {
				'receive_time': timestamp_to_isostr(receive_time),
				'source_ip'   : source_ip,
				'trap_oid'    : trapoid,
				'trap_msg'    : TrapDServer.trapdata_to_msg(trap_data),
			}
			d.update(additional_data)
			
			mail_from = settings.SNMPTRAP_MAIL_CONF['from']
			mail_to   = render_template_string(mail_params['to']     , d)
			mail_subj = render_template_string(mail_params['subject'], d)
			mail_body = render_template_string(mail_params['body']   , d)
			
			self.mail_sender.send(mail_from, mail_to, mail_subj, mail_body)
			
		except Exception, e:
			logger.error("Unable to notify SNMP trap: %s %s\n%s", type(e), e, traceback.format_exc())

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def init_logger(log_leve, log_file):
	
	logger = logging.getLogger('trapd')
	logger.setLevel( log_level )
	
	hdlr = logging.FileHandler(log_file, "a")
	hdlr.setFormatter( logging.Formatter('%(asctime)s %(levelname)s trapd > %(message)s') )
	hdlr.setLevel( log_level )
	logger.addHandler(hdlr)
	
	hdlr = logging.StreamHandler()
	hdlr.setFormatter( logging.Formatter('%(asctime)s %(levelname)s trapd > %(message)s') )
	hdlr.setLevel( log_level )
	logger.addHandler(hdlr)
	
	return logger


if __name__ == "__main__":

	#
	# Be sure django will find modules with the right path
	#
	sanet_install_dir = sys.argv[1]
	if not sanet_install_dir in sys.path:
		sys.path.insert(0, sanet_install_dir)
	os.environ.setdefault("DJANGO_SETTINGS_MODULE", "conf.settings")
	
	# Added for django 1.11
	import django
	django.setup()
	
	# These imports depend on django
	from django.db import connection
	from django.conf import settings
	from conf import get_configuration
	
	from core.models import Event, Tenant, EVENT_SNMPTRAP, MonitoredNode, DataGroup
	
	from core.api.base import db_set_current_tenant_schema
	from core.api.mail import init_mail_sender_from_conf
	from core.api.push import guess_node_from_ip
	
	from server.sanetd.sanetdlogic import create_ipc_connection
	
	#
	# Start
	#
	conf = get_configuration()
	
	# Set log 
	log_level = get_configuration().SNMPTRAP_LOG_LEVEL
	log_file  = os.sep.join( [conf.LOG_DIR, 'trapd_log'])
	
	logger = init_logger(log_level, log_file)
	
	if sanet_install_dir != conf.SANET_INSTALL_DIR:
		msg = "Install dir %s is different from configuration install dir %s" % (sanet_install_dir, conf.SANET_INSTALL_DIR)
		print msg
		logger.critical(msg)
	
	#
	# Enabled?
	#
	if not get_configuration().SNMPTRAP_LISTEN:

		logger.warning("Trapd not enabled. Configured SNMPTRAP_LISTEN parameter in settings!")

		# Block process
		import select
		select.select([],[],[], None)


	#
	# Start
	#
	mail_sender = init_mail_sender_from_conf(logger)
	
	s = TrapDServer(logger, mail_sender, conf.SNMPTRAP_LISTEN, conf.SNMPTRAP_ON_ERROR_RESTART)
	s.start()

	# never returns



