/**
 * Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
 * Created: 5/22/13, 2:45 PM
 * Author: Antonio Vaccarino
 */

/* TIMEGRAPH
format =
{

	flags: {
		timeline:   bool //display timeline or not, default true
		controls:    bool // display conrtol buttons or not, default true
		locked:     bool // if the user can make selections and change range, default false
		dataview:   bool // if the user can view/download the data from the graph, default false,
		legend:    bool // display the legend for the colours, independent from other controls
	},
	title: string, // can be null or missing
	url: string, // if missing the map goes static, with both uuid and url it's dynamic
	series: [
		{
			label:      string, // mandatory
			timescale: {        // all mandatory
				from:    number, // Python timestamp
				to:    number, // Python timestamp
				step:   number // Python timestamp
			},
			values:     [ number ], //mandatory
			format:     string(line | area | bars | tick | bartick), //defaults to bars|bartick
			//tick:       boolean, // default false, determines if the series is normalised to tick
			scaleauth:  boolean, // default true, ignored if tick is true, if the series contributes to scaling when applicable (min/max == null)
			threshold:  number, // markers only?? could be used for callout TOO, default is 0
			threshside: boolean, //defaults to true (greater than) if missing
			color:      hexstring, // defaults to lwp standard colors
			unit:       string // can be missing/null
		}
	],
	presel: {
		from:   number, // Python timestamp,
		to: number, // Python timestamp
	}
	scale: {            // mandatory as structure
		min:    number,     // mandatory but can be null
		max:    number,     // mandatory but can be null
		log:    boolean,     //if scale is logaritmic, handled via Flotr2, defaults to false
		unit:   string      // can be  missing/null
	},
	zorder:     [ number ] // could be implied by series (i.e. missing)

};*/

timegraph = {};

timegraph.const = {};

// ratio of movement when using the step tool compared to width of sel or graph
timegraph.const.shiftratio = 0.25;
timegraph.const.minzoomsteps = 5;
timegraph.const.zoomratio = 2;

// minimum number of pixels between two x scale ticks
timegraph.const.xtickspace = 160;

timegraph.const.reqpx_vis = 112;
timegraph.const.reqpx_legend = 36;
timegraph.const.reqpx_timeline = 64;

timegraph.const.formats = ['line', 'area', 'bar', 'tick', 'bartick', 'linetick'];
timegraph.const.numeric = ['line', 'area', 'bar'];
timegraph.const.boolean = ['tick', 'bartick', 'barshadow', 'linetick'];

// minute in millisec
timegraph.const.time1min = 60000;

timegraph.const.intervals = {
	'1h':       timegraph.const.time1min*60,
	'8h':       timegraph.const.time1min*60*8,
	'24h':      timegraph.const.time1min*60*24,
	'3d':       timegraph.const.time1min*60*24*3,
	'7d':       timegraph.const.time1min*60*24*7,
	'1m':       timegraph.const.time1min*60*24*31,
	'3m':       timegraph.const.time1min*60*24*93,
	'6m':       timegraph.const.time1min*60*24*186,
	'1y':       timegraph.const.time1min*60*24*366
};

timegraph.const.interval_list = [
	'1y', '6m', '3m', '1m', '7d', '3d', '24h', '8h', '1h'
];


timegraph.getdiv_legend = function()
{
	return $(this).find("#"+lwp.run.call(this, 'getid_legend'));
};

timegraph.getdiv_timeline = function ()
{
	return $(this).find("#"+lwp.run.call(this, 'getid_timeline'));
};

timegraph.getdiv_toolbar = function ()
{
	return $(this).find("#"+lwp.run.call(this, 'getid_toolbar'));
};

timegraph.getdiv_vis = function ()
{
	return $(this).find("#"+lwp.run.call(this, 'getid_vis'));
};

timegraph.getid_timeline = function ()
{
	return $(this).attr('id')+"_timeline";
};

timegraph.getid_toolbar = function ()
{
	 return $(this).attr('id')+"_toolbar";
};

timegraph.getid_vis = function ()
{
	 return $(this).attr('id')+"_vis";
};

timegraph.getid_legend = function ()
{
	return $(this).attr('id')+"_legend";
};


timegraph.numTickFormatter = function (value)
{
//    console.log("Formatting request");
//    console.log("value ",value," context ",this);
	return lwp.renderNumeric(value, this.tickUnit);
};


timegraph.validate = function validate(data)
{
	// validating structural root level fields

//	console.error("Checking for widget ",this);
//	console.error("Timegraph validating data ",data);


	if (!$.isArray(data['series']))
	{
		console.error("ERROR: series list ",data['series']," not in array format");
		return false;
	}

	// validating structural fields per series

	for (var s = 0; s < data['series'].length; s++)
	{
		var series = data['series'][s];

		if (timegraph.const.formats.indexOf(series['format'])==-1)
		{
			console.error("ERROR: invalid graph format declaration "+series['format']);
			return false;
		}

		if (typeof series['label'] != 'string')
		{
			console.error("ERROR: missing data label");
			return false;
		}

		var timescale = series['timescale'];
		if (!timescale || typeof timescale['from'] != 'number' || typeof timescale['to'] != 'number' || typeof timescale['step'] != 'number')
		{
			console.error("ERROR: bad format for timescale");
			return false;
		}

		if (!$.isArray(series['values']))
		{
			console.error("ERROR: series values not in array form ");
			return false;
		}

		// validating datapoints: all must be numeric or null
		for (var i = 0; i < series['values']; i++)
		{
			var v = series['values'][i];
			if (typeof v != "number" && v != null)
			{
				console.error("ERROR: non-numeric or null datapoint n."+i+" of series "+s);
				return false;
			}
		}
	}
	return true;
};

timegraph.aux = {};
timegraph.aux.log10 = function (value)
{
	return Math.log(value)/Math.LN10;
};

timegraph.scaleapprox = function (min, max, do_min, do_max)
{

	if (min == null && max == null)
	{
		return [0, 1];
	}

	if (min == max)
	{
		if (min == 0)
		{
			return [0, 1];
		}
		else if (min < 0)
		{
			return [min, -min];
		}
		else
		{
			return [0, min];
		}
	}


//    console.log("approximating scale from min ",min," max ",max);
	var ladder = [1,2,5,10];




	var upward = max > 0;
	var nmin = upward ? min : Math.abs(max);
	var nmax = upward ? max : Math.abs(min);

	// forcing nmin while we search for a more complex handling
	nmin = 0;

	var li = Math.round(timegraph.aux.log10(nmax));
	var v = nmax * Math.pow(10, -li);
	if (v<0)
	{
		li--;
		v = nmax * Math.pow(10, -li);
	}

	var t = 0;
	for (var i = 0; i < ladder.length; i++)
	{
		if (v <= ladder[i])
		{
			t = ladder[i];
			break;
		}
	}

//    console.log("Approximating max "+v+" to ",t," * ",Math.pow(10,li));
	nmax = t * Math.pow(10, li);

	if (upward)
	{
		return [nmin, nmax];
	}
	else
	{
		return [-nmin, -nmax];
	}
};

timegraph.emerge = function ()
{

	var conf = {};
	var data = clone($(this).data('wdata'));


//	console.log("Emerging WDATA ",data);

	conf['dynamic'] = typeof data['url'] == "string" && $.trim(data['url']).length > 0;
	if (conf['dynamic'])
	{
		conf['url'] = data['url'];
	}

	if (typeof data['title'] == 'string')
	{
		conf['title'] = data['title'];
	}
	else
	{
		conf['title'] = null;
	}

	// completing the zorder array
	// shows the index of the series in the order they should be drawn into
	var zorder = data.hasOwnProperty('zorder') ? data['zorder'] : [];
//    console.log("TGE: checking series", data['series']);
	for (var s = 0; s < data['series'].length; s++)
	{
		if (zorder.indexOf(s)==-1)
		{
			zorder.push(s);
		}
	}

	// re-assembling series for zorder in one place only
	var series = [];
	for (var i = zorder.length-1; i > -1; i--)
	{
		var z = zorder[i];
		series.push(data['series'][z]);
	}

	// setting scaling data
	var scale  = {
		min:    null,
		max:    null,
		log:    false
	};
	if (data.hasOwnProperty('scale'))
	{
		scale = data['scale'];
	}

	var scaleext_min = scale['min'] == null;
	var scaleext_max = scale['max'] == null;
	if (scaleext_min)
	{
		var extended_min = Infinity;
	}
	if (scaleext_max)
	{
		var extended_max = -Infinity;
	}

	var gfrom = Infinity;
	var gto = -Infinity;

	// writing all to flotr2 formats
//    console.log("TGE: walking series for scaling", data['series']);
	for (var s = 0; s < series.length; s++)
	{
		// if the series contributes to general scaling
		var cseries = series[s];
		// data graphed  according to Flotr2
		var gF = {};
		var scaleauth = cseries['scaleauth'];
		var timescale = cseries['timescale'];
		timescale['from'] = lwp.toJavascriptTimeStamp(timescale['from']);
		timescale['to'] = lwp.toJavascriptTimeStamp(timescale['to']);
		timescale['step'] = lwp.toJavascriptTimeStamp(timescale['step']);

		gfrom = Math.min(gfrom, timescale['from']);
		gto = Math.max(gto, timescale['to']);

		var bools = timegraph.const.boolean.indexOf(cseries.format)!=-1;
		if (bools)
		{
			gF['yaxis'] = 2;
			var isgt = cseries['threshgt'] !== false;
			var thresh = cseries['threshold'];
//            console.log("Tickable series, must be over ("+isgt+") "+thresh);
		}

//        console.log("Data aggregation for series");

		var caggr = timegraph.getSeriesAggregate.call(this, cseries);

		var graphed = [];
		for (var t = 0; t < cseries['values'].length; t++)
		{
			var moment = timescale['from']+timescale['step']*t;
			var value = cseries['values'][t];
			if (bools)
			{
				// writing data set to boolean according to threa
				var passed = isgt ? value!= null && value > thresh : value!= null && value < thresh;
				value = passed ? 1 : null;
			}
			else if (scaleauth)
			{
				// updating min and max on scale in case it's requested
				if (scaleext_max)
				{
					extended_max = Math.max(extended_max, value);
				}
				if (scaleext_min)
				{
					extended_min = Math.min(extended_min, value);
				}
			}
			graphed.push([moment, value]);
		}
		gF['data'] = graphed;
		gF['label'] =  cseries['label'];

		var gformat = cseries['format'];

		if (gformat == 'tick')
		{
			gformat = 'bartick';
		}
		else if (gformat == 'linetick')
		{
			gformat = 'tick';
		}

		if (gformat == 'bar')
		{
			gF['lines'] = { show: false };
			gF['bars'] = {
				show:           true,
				lineWidth:      1,
				barWidth:       timescale['step'],
				fill:           true,
				fillOpacity:    0.6
			};
		}
		else if (gformat == 'line')
		{
			gF['lines'] = {
				show: true,
				fill: false,
				lineWidth: 1
			}
		}
		else if (gformat == 'area')
		{
			gF['lines'] = {
				fill:           true,
				fillopacity:    0.6,
				lineWidth:      1
			};
		}
		else if (gformat == 'tick')
		{
			gF['lines'] = { show: false };
			gF['bars'] = {
				show:           true,
				lineWidth:      1,
				barWidth:       0,
				fill:           false
			};
		}
		else if (gformat == 'bartick')
		{
			gF['lines'] = { show: false };
			gF['bars'] = {
					show: true,
					lineWidth: 1,
					barWidth: timescale['step'],
					fill: true,
					fillOpacity: 1
			};

		}

		cseries['graphed'] = gF;
		cseries['aggregate'] = caggr;

	}
//    console.log("TGE: finished scaling walk");

//    console.error("TGE: setting scale");
	if (scaleext_max)
	{
		if (extended_max != -Infinity)
		{
			scale['max'] = extended_max;
		}
	}
	if (scaleext_min)
	{
		if (extended_min != Infinity)
		{
			scale['min'] = extended_min;
		}
	}

//    console.error("existing scale: ", scale);
	var newscale = timegraph.scaleapprox(scale['min'], scale['max'], scaleext_min, scaleext_max);
	scale['min'] = newscale[0];
	scale['max'] = newscale[1];
//    console.error("TGE: approx scale: ",scale);

	conf['scale'] = scale;
	conf['series'] = series;

//    console.log("TGE: Setting UI flags");
	conf['ui'] = {};
	try
	{
		conf['ui']['timeline'] = data['flags']['timeline'] !== false;
		conf['ui']['controls'] = data['flags']['controls'] !== false;
		conf['ui']['locked'] = data['flags']['locked'] === true;
		conf['ui']['dataview'] = data['flags']['dataview'] === true;
		conf['ui']['legend'] = data['flags']['legend'] !== false;
	}
	catch (err_undef)
	{
		conf['ui'] = {
			timeline:   true,
			controls:   true,
			locked:     false,
			dataview:   false,
			legend: true
		}
	}

//    console.log("TGE: setting presel if applicable");
	if (data.hasOwnProperty('presel') &&
		typeof data['presel']['from'] == 'number' &&
		typeof data['presel']['to'] == 'number')
	{
		conf['ui']['presel'] = {
			from:   lwp.toJavascriptTimeStamp(data['presel']['from']),
			to:   lwp.toJavascriptTimeStamp(data['presel']['to'])
		};
	}
	else
	{
		conf['ui']['presel'] = null;
	}

//    console.log("TGE: passing all to wlive");
	lwp.toWlive.call(this, conf, 'wgraphed');

	// data can now be passed directly to Flotr2


//    console.log("TGE: setting flotr options");
	// fixed rendering options for Flotr, parametrized added later
	var graphopts = {
		legend: { show:  false},
		shadowSize: 0,
		y2axis: { // boolean display axis
			min:    0,
			max:    1,
			noTicks: 0
		},
		grid: {
			backgroundColor: "#ffffff"
		},
		xaxis: {
			mode:   'time',
			timeMode: 'local',

			tickFormatter: lwp.dateTickFormatter,
			noTicks: Math.ceil($(this).width()/timegraph.const.xtickspace)
		}
	};

	if (conf['ui']['locked'] == false)
	{
		graphopts['selection'] = { mode: 'x', color: "#000000" };
	}

//    console.log("TGE: setting colors and other GRAPHOPTS");
	var valS = [];
	var colors = [];
	var fallbackcolors = lwp.getColors();
	var fallbackidx = 0;
	for (var s = 0; s < conf['series'].length; s++)
	{
		valS.push(conf['series'][s]['graphed']);
		if (!conf['series'][s].hasOwnProperty('color') || !lwp.isColorString(conf['series'][s]['color']))
		{
			conf['series'][s]['color'] =  fallbackcolors[fallbackidx];
			fallbackidx = (fallbackidx + 1) % fallbackcolors.length;
		}
		colors.push(conf['series'][s]['color']);
	}

	graphopts['colors'] = colors;
	graphopts['yaxis'] = {
		tickFormatter: timegraph.numTickFormatter,
		tickUnit: conf['scale']['unit']

	};
	if (conf['scale']['log'])
	{
		graphopts['yaxis']['scaling'] = 'logarithmic';
		graphopts['yaxis']['base'] = 10;
	}
	else
	{
		graphopts['yaxis']['scaling'] = 'linear';
	}
	var scalemin = conf['scale']['min'];
	if (scalemin != null)
	{

		graphopts['yaxis']['min'] = scalemin;
	}
	var scalemax = data['scale']['max'];
	if (conf['scale']['max'] != null)
	{

		graphopts['yaxis']['max'] = scalemax;
	}

	lwp.toWlive.call(this, valS, 'wseries');
	lwp.toWlive.call(this, graphopts, 'wgopts');

//    console.log("TGE: GRAPHED DATA ",conf);
};

timegraph.ctrl = {};

timegraph.ctrl.move_fwd = function (ev)
{
	var widget = ev.data['context'];
	timegraph.ctrl.move_step.call(widget, 1);
};

timegraph.ctrl.move_back = function (ev)
{
	var widget = ev.data['context'];
	timegraph.ctrl.move_step.call(widget, -1);
};

timegraph.getMaxStep = function ()
{
	var series = lwp.fromWlive.call(this)['wseries'];
	var steps = [];
	for (var i = 0; i < series.length; i++)
	{
		// defining the step and calculating from there, no walk
		steps.push (series[i]['data'][1][0]-series[i]['data'][0][0])
	}
	return Math.max.apply(null, steps);

};

timegraph.getMinStep = function ()
{
	var graphdata = lwp.fromWlive.call(this)['wgraphed'];
	//    console.log("Series info: ",graphdata);
	var minspan = Infinity;
	for (var i = 0; i < graphdata['series'].length; i++)
	{
		var cstep = graphdata['series'][i]['timescale']['step'];
		if (cstep < minspan) { minspan = cstep; }
	}

	return minspan < Infinity ? minspan : null;
};

timegraph.getClosestStampInSeries = function (timestamp, series)
{
	// gets the closest in-series timestamp to a candidate timestamp
//    console.log("Data series structure: ",series);
//	console.error("Timestamp ",timestamp," in series ",series);

	var base;
	var step;
	var scaletype = "";
	if (!series.hasOwnProperty('timescale'))
	{
		scaletype = "scaled";
		// we are checking a WSERIES series
		base = series['data'][0][0];
		if (series['data'].length == 1)
		{
			return base;
		}
		step = series['data'][1][0] - base;
	}
	else
	{
		scaletype = "graphed";
		// we are checking a WGRAPHED series
		base = series['timescale']['from'];
		step = series['timescale']['step'];
	}
	var dist = timestamp - base;
	var steps = Math.round (dist / step);
//	console.error("Timestamp ",timestamp," in series of type "+scaletype+" nearest to idx "+(base+steps));
	return (base+step*steps);
};

timegraph.getClosestStamp = function (timestamp, sid)
{

	// gets the closest timestamp to a candidate timestamp by checking on either ALL series or to an indexed series

	var series = !sid ? lwp.fromWlive.call(this)['wseries'] : [lwp.fromWlive.call(this)['wseries'][sid]];


	var stamps = [];
	for (var i = 0; i < series.length; i++)
	{
		try
		{
			stamps.push(timegraph.getClosestStampInSeries.call(this, timestamp, series[i]));
		}
		catch (ex)
		{
			// skipping empty series
		}

	}
//	console.error("Timestamps series ",stamps);
	return stamps.length > 0 ? Math.min.apply(null, stamps) : timestamp;

};

timegraph.getClosestIndex = function (timestamp, sid, early)
{
	// if early, the index is likely to be BEFORE the stamp, otherwise it is AFTER
	early = early !== false;
	var series = !sid ? lwp.fromWlive.call(this)['wseries'] : [lwp.fromWlive.call(this)['wseries'][sid]];

	var steps = [];
	for (var i = 0; i < series.length; i++)
	{
		// defining the step and calculating from there, no walk
		var base = series[i]['data'][0][0];
		var step = series[i]['data'][1][0] - base;
		var dist = timestamp - base;
		var nsteps = Math.floor(dist / step);
		steps.push(nsteps);
	}

	if (early)
	{
		return Math.min.apply(null, steps);
	}
	else
	{
		return Math.max.apply(null, steps);
	}

};

timegraph.getClosestEarlyStamp = function (timestamp, sid)
{

	var series = !sid ? lwp.fromWlive.call(this)['wseries'] : [lwp.fromWlive.call(this)['wseries'][sid]];

	var stamps = [];
	for (var i = 0; i < series.length; i++)
	{
		// defining the step and calculating from there, no walk
		var base = series[i]['data'][0][0];
		var step = series[i]['data'][1][0] - base;
		var dist = timestamp - base;
		var steps = Math.floor(dist / step);
		stamps.push(base+step*steps);
	}
	//    console.log("Timestamps series ",stamps);
	return Math.min.apply(null, stamps);

};

timegraph.getMostRecentStamp = function ()
{
	var series = lwp.fromWlive.call(this)['wseries'];
//    console.log("SERIES INFO: ",series);

	var latest = [];
	for (var i = 0; i < series.length; i++)
	{
		latest.push(series[i]['data'][series[i]['data'].length-1][0]);
	}

	return Math.min.apply(null, latest);
};

timegraph.ctrl.softBounce = function (coretime, sel_width)
{
	// moves the current selection to the timelimits of the graph keeping its size
	var ext_graph = timegraph.ctrl.getGraphExtent.call(this);

	// map is static, going to graph end
	if (sel_width > 0)
	{
		var selfrom = coretime - sel_width / 2;
		var selto = coretime + sel_width / 2;

		if (selfrom < ext_graph[0])
		{
			selfrom = ext_graph[0];
			selto = selfrom + sel_width;
		}

		if (selto > ext_graph[1])
		{
			selto = ext_graph[1];
			var latest = timegraph.getMostRecentStamp.call(this);
			selfrom = timegraph.getClosestEarlyStamp.call(this, latest-sel_width-timegraph.getMaxStep.call(this)*2);
		}

		var evt = {
			x1: selfrom,
			x2: selto
		};
		this.data('wrender_timeline').selection.setSelection(evt, false);
	}

};

timegraph.ctrl.rebase = function (coretime, graph_width, sel_width, sel_core)
{
	var url = lwp.fromWlive.call(this, 'wgraphed')['url'];
	if (!sel_core)
	{
		sel_core = null;
		sel_width = 0;
	}

//    console.log("URL for rebase: ",url);
	if (!url)
	{
		timegraph.ctrl.softBounce.call(this, coretime, sel_width);
		return;
	}

	var now = new Date().getTime();
	var graphto = Math.min(coretime + graph_width/2, now);
	var graphfrom = graphto - graph_width;

	var ts_from = lwp.fromJavascriptTimeStamp(graphfrom);
	var ts_to = lwp.fromJavascriptTimeStamp(graphto);

	var req = {
		requester: 'timegraph',
		from: ts_from,
		to: ts_to
	};



//    console.log("Making ajax rebase request for times");
//    console.log("From ",new Date(graphfrom)," (",graphfrom,")");
//    console.log("To ",new Date(graphto)," (",graphto,")");

	var context = {
		widget: this,
		coretime: coretime,
		sel_width: sel_width,
		sel_core: sel_core
	};

	$.ajax ({
		url: url,
		context: context,
		data: req
		}).done(
			function (data)
			{
				var widget = this['widget'];
				var sel_width = this['sel_width'];
				var coretime = this['coretime'];

				var graphdata = typeof  data == 'string' ? JSON.parse(data) : data;

				graphdata['flags'] = lwp.fromWlive.call(widget, 'wgraphed')['ui'];

//                console.log("new graph data ");

				var latestkey = lwp.fromWlive.call(widget, "latestkey");

				lwp.dropWlive.call(widget);
				lwp.winit.call(widget);
				lwp.populate.call(widget, graphdata);
				lwp.toWlive.call(widget, latestkey, "latestkey");
				lwp.render.call(widget);

				if (sel_width > 0)
				{
					var selcoretime = sel_core != null ? sel_core : coretime;
					var presel = {
						x1: selcoretime - sel_width/2,
						x2: selcoretime + sel_width/2
					};

					var div_vis = timegraph.getdiv_vis.call(widget);
					var div_timeline = timegraph.getdiv_timeline.call(widget);

					// autozooming to preexisting selection
					if (div_timeline.length > 0)
					{
						$(div_timeline).prop('selectedIndex', 0);
						widget.data('wrender_timeline').selection.setSelection(presel, true);
						Flotr.EventAdapter.fire(div_timeline.get(0), 'flotr:select', presel);
					}
					else
					{
						timegraph.ctrl.zoomSelection.call(widget, div_vis, presel['x1'], presel['x2']);
					}
				}
			}
		).fail(
			function ()
			{
				timegraph.ctrl.softBounce.call(this['widget'], this['coretime'], this['sel_width']);
			}
		);

};

timegraph.ctrl.tolatest = function (ev)
{
	var widget = ev.data['context'];
	var tskey = ev.data['timespan'];
	var timespan = timegraph.const.intervals[tskey];



//	console.log("Setting latest key for ",widget," to "+tskey);
	lwp.toWlive.call(widget, tskey, "latestkey");
//	$(widget).toWlive("latestkey", tskey);

//	console.log("FULL WLIVE is: ",lwp.fromWlive.call(widget));

	var coretime = new Date().getTime();
	var to = coretime;
	var from = coretime - timespan;

	timegraph.ctrl.rebase.call(widget, coretime, (to-from));
};

timegraph.ctrl.zoom_step = function (ev)
{
	var widget = ev.data['context'];
	// true: in, false: out
	var dir = ev.data['dir'];

	var graph_ext = timegraph.ctrl.getGraphExtent.call(widget);
	var timespan = dir ? (graph_ext[1]-graph_ext[0]) / timegraph.const.zoomratio : (graph_ext[1]-graph_ext[0]) * timegraph.const.zoomratio;
//    var graph_width = graph_ext[1]-graph_ext[0];
	var sel_ext = timegraph.ctrl.getSelExtent.call(widget);
	var sel_width = sel_ext[1]-sel_ext[0];
	var sel_core = (sel_ext[1] + sel_ext[0])/2;
	var graph_core = (graph_ext[1] + graph_ext[0])/2;
	var coretime = sel_width > 0 ? sel_core : graph_core;
	var to = coretime + timespan/2;
	var from = coretime - timespan/2;
	if (sel_width > 0)
	{
		if (sel_ext[0] < from && sel_ext[1] > to)
		{
			sel_width = 0;
		}
		else if ((sel_ext[1] >= from && sel_ext[1] <= to) ||
				(sel_ext[0] >= from && sel_ext[0] <= to))
		{
			sel_ext[0] = Math.max(sel_ext[0], from);
			sel_ext[1] = Math.min(sel_ext[1], to);
			sel_core = (sel_ext[1]+sel_ext[0])/2;
			sel_width = sel_ext[1]-sel_ext[0];
		}
		else
		{
			sel_width = 0;
		}
	}

	timegraph.ctrl.rebase.call(widget, coretime, (to-from), sel_width, sel_core);

};



timegraph.ctrl.changescope = function (ev)
{
	var widget = ev.data['context'];
	var tskey = ev.data['timespan'];
	var timespan = timegraph.const.intervals[tskey];

//    console.log("Requested scope change to width of "+tskey+" -> "+timespan);

	var graph_ext = timegraph.ctrl.getGraphExtent.call(widget);
//    var graph_width = graph_ext[1]-graph_ext[0];
	var sel_ext = timegraph.ctrl.getSelExtent.call(widget);
	var sel_width = sel_ext[1]-sel_ext[0];
	var sel_core = (sel_ext[1] + sel_ext[0])/2;
	var graph_core = (graph_ext[1] + graph_ext[0])/2;
	var coretime = sel_width > 0 ? sel_core : graph_core;
	var to = coretime + timespan/2;
	var from = coretime - timespan/2;
	if (sel_width > 0)
	{
		if (sel_ext[0] < from && sel_ext[1] > to)
		{
			sel_width = 0;
		}
		else if ((sel_ext[1] >= from && sel_ext[1] <= to) ||
				(sel_ext[0] >= from && sel_ext[0] <= to))
		{
			sel_ext[0] = Math.max(sel_ext[0], from);
			sel_ext[1] = Math.min(sel_ext[1], to);
			sel_core = (sel_ext[1]+sel_ext[0])/2;
			sel_width = sel_ext[1]-sel_ext[0];
		}
		else
		{
			sel_width = 0;
		}
	}
	timegraph.ctrl.rebase.call(widget, coretime, (to-from), sel_width, sel_core);
};

timegraph.ctrl.move_now = function (ev)
{
	var widget = ev.data['context'];
	var graph_ext = timegraph.ctrl.getGraphExtent.call(widget);
	var graph_width = graph_ext[1]-graph_ext[0];
	var to = new Date().getTime();
	var from = to - graph_width;
	var coretime = (to + from) / 2;
	var sel_ext = timegraph.ctrl.getSelExtent.call(widget);
	var sel_width = sel_ext[1]-sel_ext[0];
	var sel_core = null;
	if (sel_width > 0)
	{
		if (sel_ext[0] < from && sel_ext[1] > to)
		{
			sel_width = 0;
		}
		else if ((sel_ext[1] >= from && sel_ext[1] <= to) ||
				(sel_ext[0] >= from && sel_ext[0] <= to))
		{
			sel_ext[0] = Math.max(sel_ext[0], from);
			sel_ext[1] = Math.min(sel_ext[1], to);
			sel_core = (sel_ext[1]+sel_ext[0])/2;
			sel_width = sel_ext[1]-sel_ext[0];
		}
		else
		{
			sel_width = 0;
		}
	}
	timegraph.ctrl.rebase.call(widget, coretime, graph_width, sel_width, sel_core);
};

timegraph.ctrl.move_step = function (dir)
{

	// dir should be a number, otherwise it's boolean true/false 1/-1 defaulting to true
	// determines direction and (if needed) multiplicator for shift
	dir = typeof dir != "number" ? dir !== false ? 1 : -1 : dir;

	var ext_graph = timegraph.ctrl.getGraphExtent.call(this);
	var graph_width = ext_graph[1] - ext_graph[0];
	var ext_sel = timegraph.ctrl.getSelExtent.call(this);
	var sel_width = ext_sel[1]-ext_sel[0];
	var coretime = timegraph.ctrl.getSelCenter.call(this);
//    console.log("Centering on ", coretime);

	var has_sel = ext_sel[0] != ext_sel[1];

//    var now = lwp.fromJavascriptTimeStamp(new Date().getTime());

	var coreshift = has_sel ? sel_width : graph_width;
	coreshift = coreshift * timegraph.const.shiftratio * dir;

	var newcore = coretime + coreshift;
//    console.log("new core is ",newcore);
	var newsel = null;
	if (has_sel)
	{
		newsel = [newcore - sel_width/2, newcore + sel_width/2];
	}

	if ((has_sel && (newsel[0] < ext_graph[0] || newsel[1] > ext_graph[1])) || !has_sel)
	{
//        console.log("NEED REBASE");
		timegraph.ctrl.rebase.call(this, newcore, graph_width, sel_width);
	}
	else
	{
		//    console.log("back to ",newsel);
		var evt = {
			x1: newsel[0],
			x2: newsel[1]
		};
		this.data('wrender_timeline').selection.setSelection(evt, false);
	}

};

timegraph.ctrl.getGraphExtent = function ()
{
	// returns the two timestamps of the graph
	var xaxis;
	try
	{
		xaxis = $(this).data('wrender_timeline')['axes']['x'];
	}
	catch (err)
	{
//        console.log("Missing timeline, getting limits from vis");
		xaxis = $(this).data('wrender_vis')['axes']['x'];
	}

	return [xaxis['min'], xaxis['max']];
};

timegraph.ctrl.getSelExtent = function ()
{
	// returns the width of the selection
	var sel_area;
	try
	{
		sel_area = $(this).data('wrender_timeline').selection.getArea();
	}
	catch (err)
	{
		var pre_area = $(this).data('wrender_vis')['axes']['x'];
		sel_area = { 'x1': pre_area['min'], 'x2': pre_area['max']};
	}
	return [sel_area['x1'], sel_area['x2']];
};

timegraph.ctrl.hasTracedSel = function ()
{
	// returns if the user has selected a time range INSIDE the main timespan
	var sel_area = timegraph.ctrl.getSelExtent.call(this);
	return sel_area[0] != sel_area[1];
};

timegraph.ctrl.getSelCenter = function ()
{
	// returns the center of the current selection OR of the graph itself
	var sel_area = timegraph.ctrl.getSelExtent.call(this);
	var xaxis = timegraph.ctrl.getGraphExtent.call(this);
	//    console.log("GENERAL SEL INFO: ",sel_area);
	//    console.log("Current SELECTION", sel_area.getArea());
	//    console.log("Current GRAPH", xaxis['min']," ",xaxis['max']);
	var center = sel_area[0] != sel_area[1] ? (sel_area[0]+sel_area[1]) / 2 : (xaxis[0] + xaxis[1]) / 2;
	return center;
};

timegraph.ctrl.zoomSelection = function (dest, x1, x2)
{
//	console.log("optical zoom set");
	// zooms the selection on dest widget
	var data = lwp.fromWlive.call(this, 'wseries');
	var opts = lwp.fromWlive.call(this, 'wgopts');

	var minspan = timegraph.getMinStep.call(this) * timegraph.const.minzoomsteps;
	var diff = minspan - (x2 - x1);
	if (diff > 0)
	{
		var padding = diff / 2;
		x2 = x2 + padding;
		x1 = x1 - padding;
	}

//    console.log("Graph opts ",opts);

	var visF = dest.get(0);
	//    Flotr.draw(visF, data, opts);
	var zoomopts = clone(opts);
	zoomopts['xaxis'] = $.extend({'min': x1}, opts['xaxis']);
	zoomopts['xaxis'] = $.extend({'max': x2}, zoomopts['xaxis']);

	lwp.toWlive.call(this, data, 'lastZoomReq');
//    Flotr.draw(visF, data, zoomopts);
	timegraph.render_vis.call(this, zoomopts);
};

timegraph.getOrdForTime = function (timestamp, series)
{
	// determines the ordinal point in the data series where a certain timestamp applies
	var sstamp = timegraph.getClosestStampInSeries.call(this, timestamp, series);
	var timescale = series['timescale'];
	return (sstamp - timescale['from']) / timescale['step'];
};

timegraph.getSeriesAggregate = function (series, tsfrom, tsto)
{

//    console.log("Getting aggregation data for: ",series);

	var bools = timegraph.const.boolean.indexOf(series.format)!=-1;

//    console.log("Boolean series? "+bools);

	var dataset = series['values'];
	if (typeof tsfrom == 'number' && typeof tsto == 'number')
	{
		var ordfrom = timegraph.getOrdForTime.call(this, tsfrom, series);
		var ordto = timegraph.getOrdForTime.call(this, tsto, series);
//        console.log("series: ",series, ordfrom, ordto);
		dataset = series['values'].slice(ordfrom, ordto+1);
	}

	var isgt = series['threshgt'] !== false;
	var thresh = series['threshold'];

	var caggr = {};
	var linqer = Enumerable.From(dataset);
//    console.log("Aggregation  object ",linqer);
	var lvalues = linqer.Except([null]);
	var lmiss = linqer.Count("$==null");
	var lmisslen = lmiss * series['timescale']['step'];
	var lmisses = linqer.GroupBy("$==null");
//    console.log("MISSES SEQ: ",lmisses);
	if (!bools)
	{
		try
		{
			caggr['max'] = lvalues.Max();
			caggr['min'] = lvalues.Min();
			caggr['avg'] = lvalues.Average();
		}
		catch (err)
		{
			caggr['max'] = "N/A";
			caggr['min'] = "N/A";
			caggr['avg'] = "N/A";
		}

		caggr['miss'] = lmisslen;

//        console.log("Numeric series unpolled for ",Math.round(lmisslen/60000)," mins, aggr ");
	}
	else
	{
		var comparestring = isgt ? "$>"+thresh : "$<"+thresh;
		var lvalid = linqer.Count(comparestring);
		caggr['valid'] = lvalid * series['timescale']['step'];
		if (typeof caggr['valid'] == 'undefined')
		{
			caggr['valid'] = "N/A";
		}
//        console.log("Boolean series true for ",Math.round(lmisslen/60000)," mins, aggr ");
	}

//    console.log("Aggregated: ",caggr);
	return caggr;


};

timegraph.render_vis = function (coreopts)
{
	var data = lwp.fromWlive.call(this, 'wseries');
	var visF = timegraph.getdiv_vis.call(this).get(0);
	var title = lwp.fromWlive.call(this, 'wgraphed')['title'];
	var dview = lwp.fromWlive.call(this, 'wgraphed')['ui']['dataview'];


	var visopts = {
		crosshair:
		{
			mode:   'xy',
			color:  '#666666'
		}
	};

	// moved to separate div
//	if (title != null)
//	{
//		visopts['title'] = title;
//	}

	if (dview === true)
	{
		var csvopts = {
			spreadsheet: {
				show: true,
				tickFormatter: function(e) {
					return lwp.renderTimeString(e);
				}
			}
		};
		$.extend(visopts, csvopts);
	}

	$.extend(visopts, coreopts);


	var xaxis = visopts['xaxis'];
	var datasection;
	var wtype = $(this).data('wtype');
//	console.log("selecting on type "+wtype);
	if (wtype !== "bandgraph" && xaxis.hasOwnProperty('min') && xaxis.hasOwnProperty('max'))
	{
//		console.log("Xaxis data in vopts : ",visopts);

		var timefrom = visopts['xaxis']['min'];
		var timeto = visopts['xaxis']['max'];

//		console.log("Extracting "+timefrom+" to "+timeto);
		datasection = timegraph.extractDataSection.call(this, data, timefrom, timeto);
//		console.log("Datasection ",datasection);
		var autoscale = {
			autoscale: true,
			autoscaleMargin: 0.1
		};
		$.extend(visopts['yaxis'], autoscale);
		delete visopts['yaxis']['min'];
		delete visopts['yaxis']['max'];
	}
	else
	{
		datasection = data;
	}

//	console.log("rendering with opts ",visopts);
//	console.log("with data ",data);
	$(this).data('wrender_vis',Flotr.draw(visF, datasection, visopts));
};

timegraph.extractDataSection = function (data, timefrom, timeto)
{
	var idxfrom = timegraph.getClosestIndex.call(this, timefrom, null, true);
	var idxto = timegraph.getClosestIndex.call(this, timeto, null, false);

//	console.log("Getting data from "+idxfrom+" to "+idxto);
	var redata = [];
	for (var s = 0; s < data.length; s++)
	{
		var series = clone(data[s]);
		var sdata = data[s].data.slice(idxfrom, idxto+2);
		series.data = sdata;
		redata.push(series);
	}

	return redata;

};

timegraph.destroy = function()
{
	$(this).data('wrender_vis').destroy();
};

timegraph.autoScaleSelection = function (data, timefrom, timeto)
{
	var idxfrom = timegraph.getClosestIndex.call(this, timefrom, null, true);
	var idxto = timegraph.getClosestIndex.call(this, timeto, null, false);

//	console.log("autoscaling selection by going through "+idxfrom+"  to "+idxto);

	var max = -Infinity;
	var min = Infinity;

	for (var s = idxfrom; s <= idxto; s++)
	{
		var series = data[s];
		if (series.hasOwnProperty("bars"))
		{
			continue;
		}

		var els = series.data;
		for (var i = 0; i <  els.length; i++)
		{
			var value = els[i][1];
			if (value > max)
			{
				max = value;
			}
			if (value < min)
			{
				min = value;
			}
		}
	}

	return {
		min: min,
		max: max
	}
};

timegraph.getreqpx_vis = function ()
{
	return timegraph.const.reqpx_vis;
};

timegraph.getreqpx_timeline = function ()
{
	return timegraph.const.reqpx_timeline;
};

timegraph.wrender = function ()
{
	var vis = lwp.toDisplay.call(this);
	vis.empty();


//	console.warn("Starting wrender");

	var data = lwp.fromWlive.call(this, 'wseries');
	var opts = lwp.fromWlive.call(this, 'wgopts');
	var conf = lwp.fromWlive.call(this, 'wgraphed');
	//    console.log("Rendering with data ",data," opts ",opts, " and conf ", conf);
	var tot_height = 0;
	
	if (conf == null)
		return;

	if (conf["title"] != null)
	{
		var div_title = $("<div/>");
		div_title.addClass("jswlib_tg_title");
		div_title.append(conf["title"]);
		vis.append(div_title);
		tot_height += div_title.height();
	}


	var div_vis = $("<div/>");
	div_vis.addClass("jswlib_timegraph_vis");
	div_vis.attr('id', timegraph.getid_vis.call(this));
	var height_vis_min = lwp.run.call(this, 'getreqpx_vis');
	vis.append(div_vis);
	tot_height += div_vis.height();

//	console.warn("Need height ",height_vis_min);

	var ctrl_height = 0;
	if (conf['ui']['controls'] === true)
	{
		var div_ctrl = timegraph.drawControls.call(this);
		vis.append(div_ctrl);
		ctrl_height += div_ctrl.height();
	}

	// checking if we have space for legend
	var height_needed = timegraph.const.reqpx_legend;
	var has_space = vis.height()-ctrl_height-height_needed > height_vis_min;
//    console.log("Enough space for legend from "+vis.height()+"? "+has_space);

//	if (has_space && (conf['ui']['legend'] !== false))
	if (conf['ui']['legend'] !== false)
	{
		timegraph.drawLegend.call(this);
		var div_legend = timegraph.getdiv_legend.call(this);
//		console.warn("Legend has height "+div_legend.height());
//		var leg_height = Math.min(div_legend.height(), timegraph.const.reqpx_legend);
//		div_legend.height(leg_height);
		ctrl_height += div_legend.height();
	}

	var timefrom = Infinity;
	var timeto = -Infinity;
	var timeseries = conf['series'];
	for (var i = 0; i < timeseries.length; i++)
	{
		timefrom = Math.min(timefrom, timeseries[i]['timescale']['from']);
		timeto = Math.max(timeto, timeseries[i]['timescale']['to']);
	}


	height_needed = lwp.run.call(this, 'getreqpx_timeline');
	has_space = vis.height()-ctrl_height-height_needed > height_vis_min;

//    console.log("Enough space for timeline? "+has_space);

//	if (has_space && conf['ui']['timeline'] === true)
	if (conf['ui']['timeline'] === true)
	{
		var div_timeline = $("<div/>");
		div_timeline.addClass("jswlib_timegraph_timeline");
		div_timeline.attr('id', timegraph.getid_timeline.call(this));
		div_timeline.height(height_needed);
		//TODO: set specific options for timeline
		vis.append(div_timeline);
		var tlopts = clone(opts);
		if (this.data('wtype') == 'timegraph')
		{
			tlopts['yaxis']['noTicks'] = 2;
		}

		$.extend (tlopts['xaxis'], { min: timefrom, max: timeto});
//        console.log("Drawing timeline with tlopts ",tlopts," and data ",data);
		$(this).data('wrender_timeline',Flotr.draw(div_timeline.get(0), data, tlopts));
		ctrl_height += div_timeline.height();

		// binding selection event on timeline
		var evt_launcher = div_timeline.get(0);
		Flotr.EventAdapter.observe (evt_launcher, 'flotr:select', function(evt) {
//            console.log("Selecting optical zoom on timeline");
//            console.log("Selection info ",evt);
			var selfrom = evt.x1;
			var selto = evt.x2;
			// this launches optical zoom on graph_vis
			var widget = $(evt_launcher).parents('.jswlib_widget');
			div_vis = timegraph.getdiv_vis.call(widget);
//            console.log("launching optical zoom on div_vis ", div_vis);
			timegraph.ctrl.zoomSelection.call(widget, div_vis, selfrom, selto);
		});

		Flotr.EventAdapter.observe (evt_launcher, 'flotr:click', function(evt)
		{
//            console.log("Cancelling zoom on timeline");
			var widget = $(evt_launcher).parents('.jswlib_widget');
//            timegraph.ctrl.resetZoom.call(widget);
			timegraph.wrender.call(widget);
		});
	}

//	console.warn("Rendered support elements, passing to vis");

//	console.log("Working on base height ",vis.height()," with controls taking ",ctrl_height);

	var vis_height = Math.max(vis.height()-(tot_height+ctrl_height), height_vis_min);
//	var vis_height = height_vis_min;
	div_vis.height(vis_height);
	timegraph.render_vis.call(this, opts);
	var visF = div_vis.get(0);
	div_vis.bind('mouseleave', {context: this}, function(evt)
	{
		var widget = evt.data['context'];
		widget.find('.jswlib_callout').remove();

	});

	tot_height += vis_height + ctrl_height;

	Flotr.EventAdapter.observe (visF, 'flotr:click', function(evt)
	{
//        console.log("Cancelling zoom on timeline");
		var widget = $(visF).parents('.jswlib_widget');
//            timegraph.ctrl.resetZoom.call(widget);
		timegraph.wrender.call(widget);
	});

	Flotr.EventAdapter.observe (visF, 'flotr:mousemove', timegraph.drawContextInfo);
	Flotr.EventAdapter.observe (visF, 'flotr:select', function(evt) {
//        console.log("Selecting optical zoom on graph");
//        console.log("Selection info ",evt);

		// this launches optical zoom on graph_vis and overlay on timeline
		var widget = $(visF).parents('.jswlib_widget');

//        console.log("Widget range as ",timegraph.ctrl.getGraphExtent.call(widget));

		var maxrange = timegraph.ctrl.getGraphExtent.call(widget);

//        console.log("Selection out back? ", evt.x1 < maxrange[0]);
//        console.log("Selection out front? ", evt.x2 > maxrange[1]);
		evt.x1 = Math.max(evt.x1, maxrange[0]);
		evt.x2 = Math.min(evt.x2, maxrange[1]);

		div_vis = timegraph.getdiv_vis.call(widget);
		div_timeline = timegraph.getdiv_timeline.call(widget);

		if (div_timeline.length > 0)
		{
//            console.log("zooming through timeline event");
			$(div_timeline).prop('selectedIndex', 0);
			widget.data('wrender_timeline').selection.setSelection(evt, true);
			Flotr.EventAdapter.fire(div_timeline.get(0), 'flotr:select', evt);
		}
		else
		{
		//            console.log("autozoom");
//            var selfrom = Math.max(evt.x1, maxrange[0]);
//            var selto = Math.min(evt.x2, maxrange[1]);
			timegraph.ctrl.zoomSelection.call(widget, div_vis, evt.x1, evt.x2);
		}
	});

	if (conf['ui']['presel']!= null)
	{

		var zoomctx = {
			x1: conf['ui']['presel']['from'],
			x2: conf['ui']['presel']['to']
		};
        
        div_timeline = timegraph.getdiv_timeline.call(this);
        
        if (div_timeline.length > 0)
		{
            $(div_timeline).prop('selectedIndex', 0);
            $(this).data('wrender_timeline').selection.setSelection(zoomctx , true);
            Flotr.EventAdapter.fire(div_timeline.get(0), 'flotr:select', zoomctx );
        }
		conf['ui']['presel'] = null;
	}

	var div_legend = timegraph.getdiv_legend.call(this);
	if (div_legend.length > 0)
	{
		vis.append(div_legend);
	}

	vis.height(tot_height);
	$(this).height(tot_height);

};

timegraph.drawContextInfo = function (evt)
{


	if (evt.originalEvent.type != 'mousemove')
	{
//        console.log("BAD INTERCEPT!");
		$('.jswlib_callout').remove();
		return true;
	}

//    console.log("Data on highlight: ",evt,"\nIn context: ",this);




	//console.log("Received FLOTR event on MULTIGRAPH");
		//console.log(evt);
	var widget = lwp.getWidgetSelf.call($(evt.currentTarget));

	// bangraph draws slightly different
//	console.warn("WIDGET TYPE ON DIV ",$(widget),": "+$(widget).data('wtype'));
	var isband = $(widget).data('wtype') == "bandgraph";


	if (widget.find(".flotr-datagrid-container").length > 0)
	{
//             console.log("has dataview open");
		$('.jswlib_callout').remove();
		return true;
	}


	var graph = $(evt.currentTarget)[0].graph;

//        console.log("Master widget: ",widget," DATA: ",lwp.fromWlive.call(widget));
//        console.log("Extra data: ",widget.data());
//        console.log("Graph reference ",graph);



	var cx = graph.lastMousePos.x;
	if (cx >= graph.axes.x.min && cx <= graph.axes.x.max)
	{
		var i;
		var graphdata = lwp.fromWlive.call(widget, 'wgraphed');
		var series = graphdata['series'];

		var wdata = widget.data('wdata');
		var rawseries = wdata['series'];
		var rez = {};
		var zorder = wdata['zorder'];
		var zlen = $.isArray(zorder) ? zorder.length : 0;
		var unordered = 0;
		var n_series = series.length;
		for (i = 0; i < n_series; i++)
		{
			var zforced, hasZ;
			if (zlen > 0)
			{
				zforced = zorder.indexOf(i);
				hasZ = zforced != -1;
			}
			else
			{
				hasZ = false;
			}

			var realz = hasZ ? zforced : zlen+unordered;
			var rearranged = n_series - (realz+1);
			if (!hasZ)
			{
				unordered++;
			}
			rez [rearranged] = i;
		}
//			console.log("RESORTED: ",rez);




		var calloutstring = "";
		var pubDate = new Date();
		pubDate.setTime(cx);
		var realdate = timegraph.getClosestStamp.call(widget, pubDate.getTime());

//		console.log("calculating callout data from ",graphdata," in "+realdate);
//			console.log("With wlive ",lwp.fromWlive.call(widget));
//			console.log("And wdata ",$(widget).data('wdata'));

		if (!isband)
		{
			calloutstring += "<b>"+lwp.renderTimeString(realdate)+"</b><br>";

			var pointdata = {};
			for (i = 0; i < series.length; i++)
			{
//	            console.log("Current series data ",series[i]);
				var sstamp = timegraph.getClosestStamp.call(widget, pubDate.getTime(), i);
				var timescale = series[i]['timescale'];
				var ord = (sstamp - timescale['from']) / timescale['step'];
	//                console.log("Seeking ord ",ord);
				var sdata = null;
				var oob = false;
				try
				{
					var rawidx = rez[i];
					sdata = rawseries[rawidx]['values'][ord];
	//					console.log(sdata);
	//					sdata = series[i]['graphed']['data'][ord][1];
				}
				catch (err)
				{
					oob = true;
					break;
	//                    console.log("seeking out of bounds!");
				}

	//                console.log("Local stamp ",sstamp);
	//                console.log('POINT IS ',ord,": ",series[i]['graphed']['data'][ord]);

					pointdata[series[i]['label']] = sdata != null ? lwp.renderNumeric(sdata, series[i]['unit']) : "N/A";

			}
			calloutstring += lwp.rendertextcell(pointdata);

		}
		else
		{
			// bandgraph handles only current active band
			calloutstring += "<b>"+lwp.renderTimeString(realdate)+"</b>: ";
			for (i = 0; i < series.length; i++)
			{
//				console.log("Current series "+series[i].label+" data ",series[i]);
				var sstamp = timegraph.getClosestStamp.call(widget, pubDate.getTime(), i);
				var timescale = series[i]['timescale'];
				var ord = (sstamp - timescale['from']) / timescale['step'];
				var sdata = null;
				var oob = false;
				try
				{
					sdata = series[i]['graphed']['values'][ord];
					if (sdata == 1)
					{
						calloutstring += series[i].label;
					}
//					console.warn("BAND "+series[i].label+" has "+sdata);
	//					console.log(sdata);
	//					sdata = series[i]['graphed']['data'][ord][1];
				}
				catch (err)
				{
					oob = true;
					break;
	//                    console.log("seeking out of bounds!");
				}
			}
		}



//        console.warn("LOGGING FOR CALLOUT: ",pointdata);
		if (!oob)
		{
			lwp.spawncallout(evt, calloutstring);
		}
		else
		{
			$('.jswlib_callout').remove();
		}

//            console.log(calloutstring);


		}


};

timegraph.drawLegend = function ()
{
	var vis = lwp.toDisplay.call(this);
	var div_legend = $("<div/>");
	div_legend.addClass("jswlib_timegraph_legend");
	div_legend.attr('id', timegraph.getid_legend.call(this));

	var blocks = [];
	var series = lwp.fromWlive.call(this, 'wgraphed')['series'];
	for (var i = 0; i < series.length; i++)
	{
		var label = series[i]['label'];
		var color = series[i]['color'];
		var aggr = series[i]['aggregate'];

		var lblock = $("<div/>");
		lblock.addClass('jswlib_legendblock');
		lblock.css({
			'max-height': '1em',
			'line-height': '1em',
			'vertical-align': 'middle'
		});

		var lmark = $("<span/>");
		lmark.addClass('jswlib_legendblock_marker');
		lmark.css('background-color', color);
		lmark.append("&nbsp;");

		var llabel = $("<span/>");
		llabel.addClass('jswlib_legendblock_label');
		llabel.append(label);

		lblock.append(lmark).append(llabel);

		lblock.bind('hover', {context: this, series: series[i]}, timegraph.drawSeriesInfo);
		lblock.bind('mouseleave', function ()
		{
			$('.jswlib_callout').remove();
		});

		div_legend.append(lblock);

//		console.log("added "+label+" (",lblock,") to legend");
	}

	// setting height
	vis.append(div_legend);
};

timegraph.formatTimeSpan = function (candidate)
{
	// since it's for graphing polling data we suppose a down of DAYS will be far enough
	var spans_inc = [ timegraph.const.time1min, timegraph.const.intervals['1h'], timegraph.const.intervals['24h']];

	var days = Math.floor (candidate / spans_inc[2]);
	var hours = Math.floor ((candidate % spans_inc[2])/spans_inc[1]);
	var mins = Math.floor ((candidate % spans_inc[1])/spans_inc[0]);

	var timestr = "";
	if (days > 0)
	{
		timestr += days+"d ";
	}
	if (hours > 0)
	{
		timestr += hours+"h ";
	}
	if (mins > 0)
	{
		timestr += mins+"m ";
	}

	return days+hours+mins != 0 ? timestr : null;

};

timegraph.drawSeriesInfo = function (ev)
{

	var widget = ev.data['context'];
	var series = ev.data['series'];
	var notbools = series['tick'] !== true;
	//    console.log("Drawing detailed callout for series ",series);
	// TODO: add selection data for compare
	var selection = timegraph.ctrl.getSelExtent.call(widget);
	var hassel = selection[0] != selection[1];
	if (hassel)
	{
		var selaggregate = timegraph.getSeriesAggregate.call(widget, series, selection[0], selection[1]);
//        console.log("Selection aggregate data:  ",selaggregate);
	}

	var toCalloutData = function (aggregate, series)
	{
		var datadict = {};
		if (series['tick'] !== true)
		{
			datadict['average'] = lwp.renderNumeric(aggregate['avg'], series['unit']);
			datadict['max'] = lwp.renderNumeric(aggregate['max'], series['unit']);
			datadict['min'] = lwp.renderNumeric(aggregate['min'], series['unit']);
			datadict['missing'] = aggregate['miss'] != 0 ? timegraph.formatTimeSpan(aggregate['miss']) : 0;
		}
		else
		{
			//        console.log("Boolean series has aggregate: ",series['aggregate']);
			datadict['time'] = aggregate['valid'] != 0 ? timegraph.formatTimeSpan(aggregate['valid']) : 0;
		}
		return datadict;
	};

	var info_graph = toCalloutData(series['aggregate'], series);
	var info_sel = hassel ? toCalloutData(selaggregate, series) : null;

	var infotable = $("<table/>");

	var tableheaders = hassel ? ['Series', 'Selection'] : null;
	var tabledata = hassel ? [info_graph, info_sel] : [info_graph];

	if (tableheaders)
	{
		var hcontainer = $("<tr/>");
		for (var i = 0; i < tableheaders.length; i++)
		{
			hcontainer.append ("<th>"+tableheaders[i]+"</th>");
		}
		infotable.append(hcontainer);
	}

	var dcontainer = $("<tr/>");
	for (var i = 0; i < tabledata.length; i++)
	{
		var datacell = $("<td/>");
		datacell.append(lwp.rendertextcell(tabledata[i]));
		dcontainer.append(datacell);
	}
	infotable.append(dcontainer);

	lwp.spawncallout(ev, $("<div/>").append(infotable).html());

};


timegraph.drawControls = function ()
{

	var div_ctrl = $("<div/>");
	div_ctrl.addClass("jswlib_toolbar");
	div_ctrl.attr('id', timegraph.getid_toolbar.call(this));
	div_ctrl.css('text-align', 'center');
	div_ctrl.css('width', "100%");

	var btn_move_fwd = $("<span/>");
	btn_move_fwd.css('margin-right', '1em');
	btn_move_fwd.addClass("jswlib_btncontrol");
	btn_move_fwd.addClass("jswlib_flag_notextsel");
	btn_move_fwd.append('<i class="icon-forward"></i>');
	btn_move_fwd.attr('title', 'Forward');
	btn_move_fwd.click({context: this}, timegraph.ctrl.move_fwd);

	var btn_move_back = $("<span/>");
	btn_move_back.css('margin-left', '1em');
	btn_move_back.addClass("jswlib_btncontrol");
	btn_move_back.addClass("jswlib_flag_notextsel");
	btn_move_back.append('<i class="icon-backward"></i>');
	btn_move_back.attr('title', 'Backward');
	btn_move_back.click({context: this}, timegraph.ctrl.move_back);

	var intervals = timegraph.const.intervals;
	var tskey = lwp.fromWlive.call($(this), "latestkey");

//	console.log("Current latest key on ",this," is "+tskey);
//	console.log("FULL WLIVE is: ",lwp.fromWlive.call(this));

	for (var i = 0; i < timegraph.const.interval_list.length; i++)
	{
		var ckey = timegraph.const.interval_list[i];
		var btn_preset = $("<span/>");
		btn_preset.addClass("jswlib_btncontrol");
		btn_preset.addClass("jswlib_flag_notextsel");
		btn_preset.append(ckey);
		if (ckey === tskey)
		{
			btn_preset.addClass("jswlib_timegraph_lastperiod");
		}
		btn_preset.attr('title', ckey);
//        btn_preset.click({context: this, timespan: ckey}, timegraph.ctrl.changescope);
		btn_preset.click({context: this, timespan: ckey}, timegraph.ctrl.tolatest);
		div_ctrl.append(btn_preset);
	}

	var btn_now = $("<span/>");
	btn_now.css('margin-left', '1em');
	btn_now.addClass("jswlib_btncontrol");
	btn_now.addClass("jswlib_flag_notextsel");
	btn_now.append('<i class="icon-fast-forward"></i>');
	btn_now.click({context: this}, timegraph.ctrl.move_now);

	div_ctrl.append(btn_move_back);
	div_ctrl.append(btn_move_fwd);

	var isdynamic = lwp.fromWlive.call(this, 'wgraphed')['dynamic'];
	if (isdynamic)
	{
		var btn_zoom_out = $("<span/>");
		btn_zoom_out.css('margin-left', '1em');
		btn_zoom_out.addClass("jswlib_btncontrol");
		btn_zoom_out.addClass("jswlib_flag_notextsel");
		btn_zoom_out.append('<i class="icon-minus"></i>');
		btn_zoom_out.attr('title', 'Zoom out');
		btn_zoom_out.click({context: this, dir: false}, timegraph.ctrl.zoom_step);

		var btn_zoom_in = $("<span/>");
		btn_zoom_in.css('margin-right', '1em');
		btn_zoom_in.addClass("jswlib_btncontrol");
		btn_zoom_in.addClass("jswlib_flag_notextsel");
		btn_zoom_in.append('<i class="icon-plus"></i>');
		btn_zoom_in.attr('title', 'Zoom in');
		btn_zoom_in.click({context: this, dir: true}, timegraph.ctrl.zoom_step);

		div_ctrl.append(btn_zoom_out);
		div_ctrl.append(btn_zoom_in);
	}

	div_ctrl.append(btn_now);
	return div_ctrl;
};




/* BANDGRAPH
format =
{

	flags: {
		timeline:   bool //display timeline or not, default true
		controls:    bool // display conrtol buttons or not, default true
		locked:     bool // if the user can make selections and change range, default false
	},
	title: string
	url: string, // if missing the map goes static, with both uuid and url it's dynamic
	sequence: [
		{
			from: number, // Python timestamp
			to: number, // Python timestamp
			value: number, // index of state, can be null/undefined
			details: object/string // textcell item with details, can be null/undefined
		}
	],
	states: [
		{
		label:    string, // name of the state
		color:    hexstring
		}
	],
	presel: {
		from:   number, // Python timestamp,
		to: number, // Python timestamp
	}

};*/


bandgraph = clone(timegraph);

bandgraph.const.reqpx_timeline = 56;
bandgraph.const.reqpx_vis = 56;

bandgraph.getreqpx_vis = function ()
{
	return bandgraph.const.reqpx_vis;
};

bandgraph.getreqpx_timeline = function ()
{
	return bandgraph.const.reqpx_timeline;
};

bandgraph.validate = function (data)
{
	// validating structural root level fields

//	console.warn("Bandgraph validating data ",data);

	if (!$.isArray(data['sequence']))
	{
		console.error("ERROR: sequence list ",data['sequence']," not in array format");
		return false;
	}

	var states = null;
	for (var i = 0; i < data['sequence']; i++)
	{
		if (!$.isPlainObject(data['sequence'][i]))
		{
			console.error("ERROR: Non-object in position "+i+" of data sequence");
			return false;
		}
		else if (typeof data['sequence'][i]['to'] != 'number' || typeof data['sequence'][i]['from'] != 'number')
		{
			console.error("ERROR: Element "+i+" in data sequence missing/invalid timespan parameters");
			return false;
		}

		var cstate = data['sequence'][i]['state'];

		if (['undefined', 'number'].indexOf(typeof (cstate))==-1 && cstate != null)
		{
			console.error('ERROR: Non valid state index for element '+i+" in data sequence");
			return false;
		}

		if (typeof cstate == 'number')
		{
			states = Math.max(states, cstate);
		}
	}

	if (!$.isArray(data['states']))
	{
		console.error("ERROR: Missing state definitions for sequence");
		return false;
	}

	if (data['states'].length < states+1)
	{
		console.error("ERROR: Unreferenced state indexes in data sequence");
		return false;
	}

	return true;
};


bandgraph.extractStep = function (spans)
{
	// get the greatest common divider (step) for a number of timespans
//    console.log("Extracting step from spans ",spans);
	if (spans.length == 1)
	{
		return spans[0];
	}

	var step = spans.pop();
	for (var i = 0; i < spans.length; i++)
	{
		var paragon = spans[i];
		while (paragon != 0)
		{
			var rest = step % paragon;
			step = paragon;
			paragon = rest;
		}
//        step = gcd (step, paragon);
	}

	return step;
};


bandgraph.emerge = function ()
{

	var conf = {};
	var data = $(this).data('wdata');

	conf['dynamic'] = typeof data['url'] == "string" && $.trim(data['url']).length > 0;
	if (conf['dynamic'])
	{
		conf['url'] = data['url'];
	}

		if (typeof data['title'] == 'string')
	{
		conf['title'] = data['title'];
	}
	else
	{
		conf['title'] = null;
	}


//    console.log("Emerging WDATA ",data);

//    var spans = [];
	var gfrom = Infinity;
	var gto = -Infinity;
	for (var i = 0; i < data['sequence'].length; i++)
	{
		var span = data['sequence'][i];
//        console.log("Current span: ",span);
		var tfrom = lwp.toJavascriptTimeStamp(span['from']);
		var tto = lwp.toJavascriptTimeStamp(span['to']);
		gfrom = Math.min(gfrom, tfrom);
		gto = Math.max(gto, tto);
//        var tlen = (tto - tfrom);
//        spans.push(tlen);
	}
//    spans.push(gto-gfrom);
//    var step = bandgraph.extractStep(spans);
	var step = bandgraph.const.time1min;
//    console.log("STEP IS ",step);
//    console.log("TOT STEPS: ",(gto-gfrom)/step);
//    console.log("BANDING FROM ",gfrom," to ",gto);
	var sequence = [];
	var vals = [];
	for (var s = 0; s < data['states'].length; s++)
	{
		sequence[s] = [];
		vals[s] = [];
	}
	//    console.log("Sequence pre: ",sequence);
	for (var i = 0; i < data['sequence'].length; i++)
	{
		var span = data['sequence'][i];
		var tfrom = lwp.toJavascriptTimeStamp(span['from']);
		var tto = lwp.toJavascriptTimeStamp(span['to']);
		var tlen = tto - tfrom;
		var polls = tlen / step;
		var cval = span['value'];
		for (var p = 0; p < polls; p++)
		{
			for (var s = 0; s < data['states'].length; s++)
			{
				var cseqval = s == cval ? 1 : null;
				sequence[s].push([tfrom + step*p, cseqval]);
				vals[s].push(cseqval);
			}
		}
	}

	conf['ui'] = {};
	try
	{
		conf['ui']['timeline'] = data['flags']['timeline'] !== false;
		conf['ui']['controls'] = data['flags']['controls'] !== false;
		conf['ui']['locked'] = data['flags']['locked'] === true;
		conf['ui']['dataview'] = data['flags']['dataview'] === true;
	}
	catch (err_undef)
	{
		conf['ui'] = {
			timeline:   true,
			controls:   true,
			locked:     false,
			dataview:   false
		}
	}

	if (data.hasOwnProperty('presel') &&
		typeof data['presel']['from'] == 'number' &&
		typeof data['presel']['to'] == 'number')
	{
		conf['ui']['presel'] = {
			from:   lwp.toJavascriptTimeStamp(data['presel']['from']),
			to:   lwp.toJavascriptTimeStamp(data['presel']['to'])
		};
	}
	else
	{
		conf['ui']['presel'] = null;
	}

//    console.log("Sequence data: ",sequence);

	var wgraphed = [];
	var wseries = [];
	var fallbackcoloridx = 0;
	var fallbackcolors = lwp.getColors();
	var colors = [];

	for (var i = 0; i < data['states'].length; i++)
	{
		//label: data['states'][i]['label'],
		var cseries = {};
		cseries['label'] = data['states'][i]['label'] ? data['states'][i]['label'] : "";
		if (lwp.isColorString(data['states'][i]['color']))
		{
			cseries['color'] = data['states'][i]['color'];
		}
		else
		{
			cseries['color'] = fallbackcolors[fallbackcoloridx];
			fallbackcoloridx = (fallbackcoloridx+1)%fallbackcolors.length;
		}
		colors.push(cseries['color']);

		cseries['format'] = 'barshadow';
		cseries['tick'] = true;
		cseries['timescale'] = {
			from:   gfrom,
			to:     gto,
			step:   step
		};

		cseries['graphed'] = {};
		cseries['graphed']['data'] = sequence[i];
		cseries['graphed']['values'] = vals[i];

		cseries['graphed']['lines'] = { show: false };
		cseries['graphed']['bars'] = {
			show:           true,
			fill:           true,
			lineWidth:      0,
			barWidth:       cseries['timescale']['step'],
			fillOpacity:    1,
			shadowSize:     0
		};

		var caggr = bandgraph.getSeriesAggregate.call(this, cseries);
		cseries['aggregate'] = caggr;

		wgraphed.push(cseries);
		wseries.push(cseries['graphed']);
	}



//    console.log("RANGE graphed in time: ",lwp.dateTickFormatter(gfrom)," - ",lwp.dateTickFormatter(gto));

	var graphopts = {
		legend: { show:  false},
		shadowSize: 0,
		yaxis: { // boolean display axis
			min:    0,
			max:    1,
			noTicks: 0,
			ticks: null
		},
		grid: {
			backgroundColor: "#ffffff"
		},
		xaxis: {
			mode:   'time',
			timeMode: 'local',
			tickFormatter: lwp.dateTickFormatter,
			noTicks: Math.ceil($(this).width()/timegraph.const.xtickspace)
		}
	};

	if (conf['ui']['locked'] == false)
	{
		graphopts['selection'] = { mode: 'x', color: "#000000" };
	}

	conf['series'] = wgraphed;

	graphopts['colors'] = colors;

	lwp.toWlive.call(this, conf, 'wgraphed');
	lwp.toWlive.call(this, wseries, 'wseries');
	lwp.toWlive.call(this, graphopts, 'wgopts');

};


jswlib['timegraph'] = timegraph;
jswlib['bandgraph'] = bandgraph;
buildWidgetsTree();
