# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

API di alto livello
========================================================================

.. contents:: Contenuti

Introduzione
------------------------------------------------------------------------

Questo modulo contiene le funzioni "ad alto livello" per gestire:

	* creazione di tagtree
	* creazione di tag
	* tagging / untagging
	* aggiornamento stati tag (tagsd)

Gli altri moduli di sanet dovrebbero utilizzare la api di questo file
se non vogliono preoccuparsi di gestire manualmente la creazione di 
oggetti nel DB e la gestione degli stati.

"""

from core.exceptions import ApplicationException

from core.db_utils import transaction_commit_on_success, transaction_set_dirty

from core.api.base import db_set_current_tenant_schema
from core.decorators import preserve_db_schema_on_exception

from tags.models import *

from tags.models import _tag_element
from tags.models import _tag_datagroup
from tags.models import _tag_condition

from tags.models import _untag_element
from tags.models import _untag_datagroup
from tags.models import _untag_condition
from tags.models import _tagtree_recalculate_tagged_elements

from tags.server.tagsd.api import signal_state_refresh

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def tagtree_get(tenant, name, create=False, owner=None):
	"""
	
	"""
	tree, created = _tagtree_get_t(tenant, name, create=create, owner=owner)

	if created:
		signal_state_refresh(tenant.uuid, [tree.uuid], force=True)

	return tree

@preserve_db_schema_on_exception
@transaction_commit_on_success
def _tagtree_get_t(tenant, name, create=False, owner=None):

	return _tagtree_get(tenant, name, create=create, owner=owner)

def _tagtree_get(tenant, name, create=False, owner=None):

	created = False

	if create:

		try:
			tree = TagTree.objects.get(tenant=tenant, name=name)
			created = False
		except TagTree.DoesNotExist, e:
			tree = None

		if tree is None:

			if owner:
				owner.assert_capability(tenant, 'tags.tagtree.create')

			tree = TagTree(tenant=tenant, name=name)
			tree.save()
			created = True

		if created and (owner is not None):
			tree.owner = owner
			tree.save()
	else:
		tree = TagTree.objects.get(tenant=tenant, name=name)

	return tree, created

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def tagtree_delete(tenant, name, user=None):

	tenant.db_set_current_schema()

	tree_id = _tagtree_delete_t(tenant, name, user=user)

	signal_state_refresh(tenant.uuid, [tree_id], force=True)

	return

@preserve_db_schema_on_exception
@transaction_commit_on_success
def _tagtree_delete_t(tenant, name, user=None):

	ret = _tagtree_delete(tenant, name, user=user)

	transaction_set_dirty()

	return ret


def _tagtree_delete(tenant, name, user=None):

	tree =  TagTree.objects.get(tenant=tenant, name=name)
	cursor = connection.cursor()

	if user:
		user.assert_capability(tenant, 'tags.tagtree.delete')
		tree.assert_permission_readwrite(user)

	# Salviamo l'UUID che viene rimosso dopo la delete
	uuid = tree.uuid

	#
	# Send signal to other sub-apps to perform delete operations on objects releated to
	# tagtree.
	#
	sanet_tag_tagtree_pre_delete.send(None, instance=tree, cursor=cursor)


	#
	# Delete permissions
	#

	# Template per le eliminazioni successive
	DELETE_BASE_QUERY = """
DELETE FROM %(table)s TG 
 WHERE EXISTS (
               SELECT 1 
                 FROM tagentry T 
                WHERE TG.%(column_name)s = T.uuid 
                  AND T.tree_id = '%(tree)s' 
              )
"""

	#
	# Elimino i tagGroupAuth
	#
	query = DELETE_BASE_QUERY % {
		'table'      : 'tagentry_group',
		'column_name': 'root_direntry_id',
		'tree'       : uuid,
	}
	cursor.execute(query)

	#
	# Elimino i tagentryInsertRule
	#
	query = DELETE_BASE_QUERY % {
		'table'      : 'tagentryinsertrule',
		'column_name': 'tagentry_id',
		'tree'       : uuid
	}
	cursor.execute(query)

	#
	# Elimino i tagentryMetadata
	#
	query = DELETE_BASE_QUERY % {
		'table'      : 'tagentrymetadata',
		'column_name': 'tag_id',
		'tree'       : uuid
	}
	cursor.execute(query)

	#
	# Delete tags/tagtree
	#
	
	# delete all tagged elements
	cursor.execute( "DELETE FROM tags_elementtag WHERE tree_id='%s';" % (tree.uuid) )

	# delete all tags
	cursor.execute( "DELETE FROM tagentry        WHERE tree_id='%s';" % (tree.uuid) )

	# delete the tree
	tree.delete()

	# TODO:
	#
	# La cancellazione dei modelli tramite django e' troppo lenta con tree molto grandi
	# e la gestione dei segnali nel "post_delete" degli ElementTag 
	# rallenta tutto.
	#
	#    cursor = connection.cursor()
	#    cursor.execute('DELETE FROM tags_elementtag where tenant_id=%s and tree_id=%s', [tree.tenant_id, tree.uuid])
	#    cursor.execute('DELETE FROM tagentry        where tenant_id=%s and tree_id=%s', [tree.tenant_id, tree.uuid])
	#
	#  Il problema di questo sistema e' che non vengono triggerati i segnali
	#  delle mapper per invalidre cache e simili.

	#tree.delete()

	return uuid

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def tagtree_set_nostate(tree, nostate):

	db_set_current_tenant_schema(tree.tenant.name)

	r = _tagtree_set_nostate_t(tree, nostate)

	signal_state_refresh(tree.tenant.uuid, [tree.uuid])

	return

@preserve_db_schema_on_exception
@transaction_commit_on_success
def _tagtree_set_nostate_t(tree, nostate):

	if tree.nostate == nostate:
		return False

	tree.nostate = nostate

	if tree.nostate == False:
		tree.nomap = False

	#
	# Altro da fare?
	#

	tree.save()

	return True

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def tag_create(tagtree, name, parent=None, user=None):


	if user:
		user.assert_capability(tagtree.tenant, 'tags.tag.create')
		tagtree.assert_permission_readwrite(user)

	tag, created = _tag_create_t(tagtree, name, parent=parent)

	if created:
		signal_state_refresh(tagtree.tenant.uuid, [tag.tree.uuid], force=True)

	return tag

@preserve_db_schema_on_exception
@transaction_commit_on_success
def _tag_create_t(tagtree, name, parent=None):
	return _tag_create(tagtree, name, parent=parent)

def _tag_create(tagtree, name, parent=None):

	if parent is None:
		if tagtree.tags.exclude(name=name).filter(parent=None).count() > 0:
			raise ApplicationException("Can not create multiple root tags.")

	tag, created = Tag.objects.get_or_create(tenant=tagtree.tenant, tree=tagtree, name=name, parent=parent)

	if created:
		tag.update_path()
		tag.save()

	return tag, created

#-----------------------------------------------------------------------------#

def tag_create_root(tenant, name, owner=None):

	tag, created = _tag_create_root_t(tenant, name, owner=owner)

	if created:
		signal_state_refresh(tenant.uuid, [tag.tree.uuid], force=True)

	return tag

@preserve_db_schema_on_exception
@transaction_commit_on_success
def _tag_create_root_t(tenant, name, owner=None):

	tagtree, created = _tagtree_get(tenant, name, create=True, owner=owner)

	tag, created =  _tag_create(tagtree, name)

	return tag, created

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def tag_remove(tag, autodelete_tree=False, user=None):

	is_root   = tag.parent is None
	tenant_id = tag.tenant_id
	tree_id   = tag.tree_id

	if user:
		user.assert_capability(tag.tenant, 'tags.tag.delete')
		tag.tree.assert_permission_readwrite(user)

	_tag_remove_t(tag, autodelete_tree=autodelete_tree)

	# FIX: refresh only 
	signal_state_refresh(tenant_id, [tree_id], force=True)

@preserve_db_schema_on_exception
@transaction_commit_on_success
def _tag_remove_t(tag, autodelete_tree=False):

	if tag.parent is None:

		if autodelete_tree == True:
			tag.tree.delete()
		else:
			tag.delete()

	else:
		tag.delete()

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

@preserve_db_schema_on_exception
@transaction_commit_on_success
def tag_rename(tag, new_name, user=None):

	if user:
		user.assert_capability(tag.tenant, 'tags.tag.edit')
		tag.tree.assert_permission_readwrite(user)

	old_path = tag.path

	subtags = tag.sub_tags

	tag.name = new_name
	tag.update_path()
	tag.save()

	for subtag in subtags.all():

		trailing_path = subtag.path[ len(old_path)+1: ]

		new_path = TAG_PATH_DIVIDER.join([ tag.path, trailing_path ])

		subtag.path = new_path
		subtag.save()

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#


def tagtree_recalculate_tagged_elements(tree):

	tagged_count = tagtree_recalculate_tagged_elements_t(tree)

	# IMPORTANTE: Se qualcosa nel tagging e' cambiato inviamo una notifica finale di refresh
	#             Qualche notifica e' forse gia' stata fatta partire, ma una notifica
	#             finale dovrebbe garantire che lo stato di tagsd venga aggiornato.
	if tagged_count > 0:
		signal_state_refresh(tree.tenant_id, [tree.uuid], force=True)

	return tagged_count

#@transaction_commit_on_success	
def tagtree_recalculate_tagged_elements_t(tree):

	#
	# IMPORTANTE: l'operazione di refresh di un tagtree richia di inviare troppe
	#             notifiche al tagsd.
	#
	#             Per evitare troppe notifiche, inviamo notifiche di aggiornamento
	#             del tree interessato solo un tot alla volta.
	#
	c = {
		'signal_count': 0
	}
	def signal_tagged_element(context, tree, tag, elementtag, node):
		print node
		context['signal_count'] += 1
		if context['signal_count'] > 30:
			print ""
			signal_state_refresh(tag.tenant_id, [tag.tree_id], force=True)
			context['signal_count'] = 0
	
	tagged_count = _tagtree_recalculate_tagged_elements(tree, tagged_callback=signal_tagged_element, tagged_callback_context=c)

	return tagged_count

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def tag_element(tag, element, user=None, force_retagging=False):

	if user and not user.is_superuser:
		user.assert_capability(tag.tenant, 'tags.tag.tag_element')
		tag.tree.assert_permission_readwrite(user)

		# Quando un tagtree ha permessi READ-WRITE associati ad almeno un gruppo, 
		# l'aggiunta di elementi nei tag (tagging) e' possibile  solo da parte 
		# dell'OWNER dell'elemento o degli UTENTI dei gruppi che hanno gia' 
		# permessi READ-WRITE sull'elemento (attraverso altri tagtree).  
		# Questo vincolo impedisce ad un utente U di auto-assegnarsi permessi 
		# READ-WRITE  "taggando" un elemento E un tagtree TT1 a cui avrebbe 
		# accesso  in sola LETTURA in un tagtree TT2.

		if user is not tag.tree.owner:
			
			groups_with_rw = [ ttgp.group for ttgp in TagTreeGroupPermission.objects.filter(dirtree=tag.tree, auth_type=DIRENTRY_AUTH_READWRITE) ]
			
			if groups_with_rw:
				can_tag = False
				for g in user.groups.all():
					if g in groups_with_rw:
						can_tag = True
				if not can_tag:
					raise ApplicationException("User '%s' can not add element '%s' to tagtree '%s' with READ-WRITE because he has only READ permission on that element" % (user.username, element, tag.tree.name))


	tagged = _tag_element_t(tag, element, force_retagging)

	if tagged:
		signal_state_refresh(tag.tenant_id, [tag.tree_id], force=True)

	return tagged

@preserve_db_schema_on_exception
@transaction_commit_on_success
def _tag_element_t(tag, element, force_retagging):
	return _tag_element(tag, element, force_retagging=force_retagging)

#-----------------------------------------------------------------------------#

@preserve_db_schema_on_exception
@transaction_commit_on_success
def tag_element_indirect(tag, element, user=None):

	if user:
		user.assert_capability(tag.tenant, 'tags.tag.tag_element')
		tag.tree.assert_permission_readwrite(user)

	return _tag_element(tag, element, directly_tagged=False)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def untag_element(tag, element, user=None):

	if user:
		user.assert_capability(tag.tenant, 'tags.tag.tag_element')
		tag.tree.assert_permission_readwrite(user)

	untagged = _untag_element_t(tag, element)

	if untagged:
		signal_state_refresh(tag.tenant_id, [tag.tree_id], force=True)

	return untagged

@preserve_db_schema_on_exception
@transaction_commit_on_success
def _untag_element_t(tag, element):
	return _untag_element(tag, element)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def tag_datagroup(tag, datagroup):

	tagged = _tag_datagroup_t(tag, datagroup)

	if tagged:
		signal_state_refresh(tag.tenant_id, [tag.tree_id], force=True)

	return tagged

@preserve_db_schema_on_exception
@transaction_commit_on_success
def _tag_datagroup_t(tag, datagroup):
	return _tag_datagroup(tag, datagroup)

#-----------------------------------------------------------------------------#

@preserve_db_schema_on_exception
@transaction_commit_on_success
def tag_datagroup_indirect(tag, datagroup, user=None):

	if user:
		user.assert_capability(tag.tenant, 'tags.tag.tag_element')
		tag.tree.assert_permission_readwrite(user)
		
	return _tag_datagroup(tag, datagroup, directly_tagged=False)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def untag_datagroup(tag, datagroup):
	untagged = _untag_datagroup_t(tag, datagroup)

	if untagged:
		signal_state_refresh(tag.tenant_id, [tag.tree_id], force=True)

	return untagged

@preserve_db_schema_on_exception
@transaction_commit_on_success
def _untag_datagroup_t(tag, datagroup):
	return _untag_datagroup(tag, datagroup)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def tag_condition(tag, condition, user=None):

	if user:
		user.assert_capability(tag.tenant, 'tags.tag.tag_element')
		tag.tree.assert_permission_readwrite(user)

	tagged = _tag_condition_t(tag, condition)

	if tagged:
		signal_state_refresh(tag.tenant_id, [tag.tree_id], force=True)

	return tagged

@preserve_db_schema_on_exception
@transaction_commit_on_success
def _tag_condition_t(tag, condition):
	return _tag_condition(tag, condition)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def untag_condition(tag, condition, user):

	if user:
		user.assert_capability(tag.tenant, 'tags.tag.tag_element')
		tag.tree.assert_permission_readwrite(user)

	untagged = _untag_condition_t(tag, condition)

	if untagged:
		signal_state_refresh(tag.tenant_id, [tag.tree_id], force=True)

	return untagged

@preserve_db_schema_on_exception
@transaction_commit_on_success
def _untag_condition_t(tag, condition):
	return _untag_condition(tag, condition)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def elementtag_set_explosive(et, explosive, explosive_group):

	r = _elementtag_set_explosive_t(et, explosive, explosive_group)

	if r:
		signal_state_refresh(et.tenant_id, [et.tree_id], force=True)

	return r

@preserve_db_schema_on_exception
@transaction_commit_on_success
def _elementtag_set_explosive_t(et, explosive, explosive_group):

	changed = False

	if et.is_explosive != explosive:
		et.is_explosive = explosive
		changed = True

	if et.explosive_group != explosive_group:
		et.explosive_group = explosive_group
		changed = True

	if changed:
		et.save()

	return changed

