# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.
"""

Strutture dati per il calcolo degli stati
=============================================================================

Classi 
----------------------------------------------------------------------

Per calcolare gli stati di un Tag/TagTree viene utilizzata una struttura ad albero di elementi *TreeNode*.

La classe *TreeNode* viene estesa da diverse altre 4 classi per gestire tutti i possibili elementi di un TagTree:

	* Tag ->  TagTreeNode
	* Elementi -> ElementTreeNode
	* Datagroup -> DataGroupTreeNode
	* Condition  -> ConditionTreeNode
	
Esempio:

::
    
                    tree1
                      |                
                    Global                  ->                  TagTreeNode
                    /   \                                          /     \
                   /     \                                        /       \
           localhost     phink0             ->      ElementTreeNode        ElementTreeNode
               /             \                                  /           \
      icmp-reach            ntp-status      ->   DataGroupTreeNode          DataGroupTreeNode
             /                 \                              /               \
        reach                   check       ->   ConditionTreeNode            ConditionTreeNode
        

Creazione dell'albero
------------------------------

La creazione dell'albero puo' essere fatta in qualunque modo (vedi modulo state2.py)


Calcolo degli stati per ogni elemento dell'albero
--------------------------------------------------

TODO

"""

import os
import time
import json

from sanet_common.sanetdefs.constants import *

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

_DEBUG = int(os.environ.get('DEBUG', 0) or 0)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

"""
Queste classi servono per costruire la struttura ad albero necessaria
per calcolare lo stato di un albero di tag.

"""

def GROUP_COUNTERS():
	return {
		'tot': 0,
		'dn' : 0,
	}

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def _default_visit(l, e):
	print e, '    ' * l, e.__class__.__name__

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class TreeNode(object):

	"""
	Per calcolare lo stato di un TagTree utilizziamo una struttura ad albero.
	
	Ogni elemento dell'albero e' un oggetto di tipo TreeNode.
	
	"""

	# ATTENZIONE:  L'utilizzo di __slots__ serve per ottimizzare il consumo di memoria
	#              I campi definiti qui devono essere gli unici inizializzati nell' __init__!!
	__slots__ = [
		 'id'

		,'r_id'

		,'children'
		,'conditions'

		,'explosive'
		,'group'

		,'calc_status'
		,'visible_status'
		,'exploded'
		,'alarms'
		,'critical_alarms'

		,'ts'
	]

	def __init__(self, id, r_id=None, ts=None):

		# UUID univoco dell'elemento dell'albero 
		#
		#   * TAG           ->    Tag.UUID 
		#   * ElementTag    ->    ElementTag.UUID
		#
		self.id              = id              

		# UUID dell'elemento "effettivo" dell'albero:
		#
		#   * TAG           -> UUID del TAG 
		#   * ElementTag    -> ElementTag.elementid (UUID del MonitoredElement, DataGroup, Condition)
		#
		self.r_id            = r_id            # reale id dell'oggetto puntato da questo element (tag.uuid, elementtag.uuid)

		self.children        = []
		self.conditions      = []

		self.explosive       = False
		self.group           = None

		# Campi valorizzati dopo il calcolo dello stato
		self.ts              = ts    # Momento di calcolo di stato del nodo
		self.alarms          = None  # contatore allarmi presenti nell'albero
		self.critical_alarms = None  # contatore allarmi critici presenti nell'albero
		self.calc_status     = None  # stato calcolato e utilizzato per il calcolo degli stati durante la visita nell'albero
		self.visible_status  = None  # stato finale che deve essere mostrato (puo' essere diverso da "calc_status" in diverse situazioni)
		self.exploded        = None  # Flag che indica che il nodo e' "esploso" (contiene elementi "esplosivi" DOWN)

	def __repr__(self):
		return "id=%s r_id=%s status=%-4s finalstatus=%s exp=%s alarms=%s critical=%s"       % (self.id, self.r_id, CSTATE_TO_CODENAME(self.calc_status), CSTATE_TO_CODENAME(self.visible_status), self.exploded, self.alarms, self.critical_alarms)

	def state_info(self):
		return "status=%s finalstatus=%s exp=%s alarms=%s critical=%s"       % (CSTATE_TO_CODENAME(self.calc_status) if self.calc_status else '  ', CSTATE_TO_CODENAME(self.visible_status), self.exploded, self.alarms, self.critical_alarms)

	def get_state_dict(self):
		return self._get_state_dict()

	def _get_state_dict(self):
		return {
			'_ts'           : self.ts,
			#'status'        : self.calc_status, 
			'visible_status': self.visible_status,
			'exploded'      : self.exploded, 
			'alarms'        : self.alarms,
			'critical'      : self.critical_alarms,
		}

	def set_no_state(self):
		"""
		Visita ricorsivamente tutto l'albero e impostat stato Uncheckable (UU) per tutti i nodi.
		"""
		self.__set_no_state()
		for treenode in self.children:   treenode.set_no_state()
		for treenode in self.conditions: treenode.set_no_state()

	def __set_no_state(self):
		self.ts              = None
		self.calc_status     = CSTATE_UU
		self.visible_status  = CSTATE_UU
		self.exploded        = False
		self.alarms          = 0
		self.critical_alarms = 0

	def visit(self, visit_fun=_default_visit, level=0):
		"""
		Funzione di comodo per "visita" ricorsiva dell'albero.
		
		La funzione "visit_fun" deve essere una funzione con la seguente signature:
		
			def  visit_fun( level, treenode ):
				...
		
		I parametri:
		
			<level>   : livello di profondita' dell'albero (parte da 0)
			<treenode>: elemento TreeNode dell'albero da visitare.
		
		"""

		if visit_fun is not None:
			visit_fun(level, self)

		for c in self.children:
			c.visit(visit_fun=visit_fun, level=level+1)

		for c in self.conditions:
			c.visit(visit_fun=visit_fun, level=level+1)

		return

	def calculate_state(self, conditions_data_dict, critical_priority=PRIORITY_NONE, minimum_priority=PRIORITY_NONE, ts=None):
		"""
		Il parametro "conditions_data_dict" e' un dizionario di tuple che contiene solo i seguent dati:
		
			<uuid> , <status> , <priority> , <primary>
		
		NOTA: Per il calcolo degli stati non servono ulteriori informazioni sulle condition. 
		"""

		ts = ts or int(time.time())

		self._calculate_state(conditions_data_dict, critical_priority=critical_priority, minimum_priority=minimum_priority, ts=ts)

		return self.visible_status

	def _calculate_state(self, conditions_data_dict, critical_priority=PRIORITY_NONE, minimum_priority=PRIORITY_NONE, level=0, ts=None):
		
		raise Exception("ABSTRACT METHOD NOT IMPLEMENTED")


#-----------------------------------------------------------------------------#

class TagTreeNode(TreeNode):

	def _calculate_state(self, conditions_data_dict, critical_priority=PRIORITY_NONE, minimum_priority=PRIORITY_NONE, level=0, ts=None):

		if _DEBUG: print '\t' * level, "%s" % (self.__class__.__name__,)

		global_status = None 
		exploded      = False

		alarms_set          = set()
		critical_alarms_set = set()

		explosive_groups = {}

		tot              = 0
		tot_skipped      = 0
		tot_primary      = 0
		tot_primary_down = 0
		tot_up           = 0
		tot_fa           = 0
		tot_down         = 0
		tot_uncheckable  = 0
		tot_in           = 0
		
		has_conditions = 0

		for nt in self.children:

			explosive, group_id = nt.explosive, nt.group

			calculated_state = nt._calculate_state(conditions_data_dict, critical_priority=critical_priority, minimum_priority=minimum_priority, level=level+1, ts=ts)
			if calculated_state is None:
				continue

			fake_primary, status, num_conditions, exploded, alarms_set2, critical_alarms_set2 = calculated_state
			
			has_conditions += 1 if num_conditions > 0 else 0

			alarms_set          = alarms_set.union(alarms_set2)
			critical_alarms_set = critical_alarms_set.union(critical_alarms_set2)

			if group_id is not None:
				counters = explosive_groups.get(group_id, explosive_groups.setdefaul(group_id, GROUP_COUNTERS() ))
				counters['tot'] += 1
				if status == CSTATE_DN:
					counters['dn'] += 1

			tot += 1
			tot_primary += 1
			tot_primary_down += 1 if (status == CSTATE_DN) else 0
			tot_down         += 1 if (status == CSTATE_DN) else 0
			tot_uncheckable  += 1 if (status in CSTATES_UNCHECKABLE or status is None) else 0
			tot_up           += 1 if (status == CSTATE_UP) else 0
			tot_fa           += 1 if (status == CSTATE_FA) else 0
			tot_in           += 1 if (status == CSTATE_IN) else 0

		if _DEBUG: print '\t' * (level+1), "COUNT 1", "tot=%s, pr=%s/%s, dn=%s, un=%s, up=%s" % ( tot       , tot_primary     , tot_primary_down, tot_down        , tot_uncheckable , tot_up          )

		for nt in self.conditions:

			explosive, group_id = nt.explosive, nt.group

			calculated_state = nt._calculate_state(conditions_data_dict, critical_priority=critical_priority, minimum_priority=minimum_priority, level=level+1, ts=ts)
			if calculated_state is None:
				continue

			primary, status, num_conditions, exploded, alarms_set2, critical_alarms_set2 = calculated_state

			has_conditions += 1 if num_conditions > 0 else 0

			alarms_set          = alarms_set.union(alarms_set2)
			critical_alarms_set = critical_alarms_set.union(critical_alarms_set2)

			if group_id is not None:
				counters = explosive_groups.get(group_id, explosive_groups.setdefaul(group_id, GROUP_COUNTERS() ))
				counters['tot'] += 1
				if status == CSTATE_DN:
					counters['dn'] += 1

			tot += 1
			tot_primary      += 1 if primary else 0

			tot_primary_down += 1 if (primary and (status == CSTATE_DN)) else 0
			tot_down         += 1 if (status == CSTATE_DN) else 0
			tot_uncheckable  += 1 if (status in CSTATES_UNCHECKABLE or status is None) else 0
			tot_up           += 1 if (status == CSTATE_UP) else 0
			tot_fa           += 1 if (status == CSTATE_FA) else 0
			tot_in           += 1 if (status == CSTATE_IN) else 0

		real_tot = tot - tot_uncheckable

		if _DEBUG: print '\t' * (level+1), "COUNT 2", "tot=%s, real_tot=%s, pr=%s/%s, dn=%s, un=%s, up=%s" % ( tot, real_tot, tot_primary, tot_primary_down, tot_down, tot_uncheckable , tot_up)

		#if 0:
		#	print self
		#	print len(self.children)
		#	print len(self.conditions)
		#	print "real_tot          ", real_tot
		#	print "tot               ", tot              
		#	print "tot_skipped       ", tot_skipped      
		#	print "tot_primary       ", tot_primary      
		#	print "tot_primary_down  ", tot_primary_down 
		#	print "tot_up            ", tot_up           
		#	print "tot_fa            ", tot_fa           
		#	print "tot_down          ", tot_down         
		#	print "tot_uncheckable   ", tot_uncheckable  
		#	print "tot_in            ", tot_in           

		global_status = CSTATE_UU

		if tot == 0 and (len(self.conditions) + len(self.children)) > 0 :
			global_status = CSTATE_UP

		if real_tot == 0:
			global_status = CSTATE_UU
		elif real_tot == tot_up:
			global_status = CSTATE_UP
		elif tot_down > 0:
			if tot_primary > 0 and tot_primary == tot_primary_down:
				global_status = CSTATE_DN
			elif tot_down >= real_tot:
				global_status = CSTATE_DN
			else:
				global_status = CSTATE_FA
		elif tot_fa > 0:
			global_status = CSTATE_FA
		elif tot_in > 0:
			if real_tot == tot_in: 
				global_status = CSTATE_IN
			else:
				global_status = CSTATE_UP
		else:
			global_status = CSTATE_UU

		# Ci sono dei gruppi esplosi?
		for group_id, counters in explosive_groups.items():
			if counters['tot'] == counters['dn']:
				exploded = True
				global_status = CSTATE_DN

		self.ts              = ts
		self.alarms          = len(alarms_set)
		self.critical_alarms = len(critical_alarms_set)
		self.exploded        = exploded

		self.calc_status          = global_status

		self.visible_status    = self.calc_status if self.critical_alarms == 0 else CSTATE_DN

		if _DEBUG: print '\t' * (level+1), 'FINAL status=%s, finalstatus=%s, exploded=%s, alarms=%s, critical=%s' % ( self.calc_status, self.visible_status, self.exploded, self.alarms, self.critical_alarms,  )

		return True, self.calc_status, has_conditions, self.exploded, alarms_set, critical_alarms_set

#-----------------------------------------------------------------------------#

class ElementTreeNode(TreeNode):

	def _calculate_state(self, conditions_data_dict, critical_priority=PRIORITY_NONE, minimum_priority=PRIORITY_NONE, level=0, ts=None):

		if _DEBUG: print '\t' * level, "%s" % (self.__class__.__name__,)

		global_status = None 
		exploded      = False

		alarms_set          = set()
		critical_alarms_set = set()

		explosive_groups = {}

		tot_valid = 0

		is_down        = False
		up             = 0
		fa             = 0
		down           = 0
		uncheckable    = 0
		uncheckable_up = 0
		uncheckable_dn = 0
		suspended      = 0

		has_primary = False

		has_conditions = 0

		if len(self.children) == 0 and len(self.conditions) == 0:
			self.ts              = ts
			self.alarms          = 0
			self.critical_alarms = 0
			self.visible_status    = CSTATE_DEFAULT
			self.calc_status          = None
			self.exploded        = False

			return has_primary, self.calc_status, 0, self.exploded, alarms_set, critical_alarms_set

		for nt in self.children:

			explosive, group_id = nt.explosive, nt.group

			calculated_state = nt._calculate_state(conditions_data_dict, critical_priority=critical_priority, minimum_priority=minimum_priority, level=level+1, ts=ts)
			if calculated_state is None:
				continue

			primary, status, num_conditions,  exploded, alarms_set2, critical_alarms_set2 = calculated_state
			
			has_conditions += 1 if num_conditions > 0 else 0
			
			alarms_set          = alarms_set.union(alarms_set2)
			critical_alarms_set = critical_alarms_set.union(critical_alarms_set2)

			if group_id is not None:
				counters = explosive_groups.get(group_id, explosive_groups.setdefaul(group_id, GROUP_COUNTERS() ))
				counters['tot'] += 1
				if status == CSTATE_DN:
					counters['dn'] += 1

			if status is not None:
				tot_valid += 1

				has_primary = has_primary or primary

			#
			# STATUS
			#
			if isinstance(nt, DataGroupTreeNode) and primary:
				if status == CSTATE_DN:
					is_down = True

			if   status == CSTATE_UP             : up          += 1
			elif status == CSTATE_FA             : fa          += 1
			elif status == CSTATE_DN             : down        += 1
			elif status == CSTATE_IN             : suspended   += 1
			elif status in [CSTATE_UU, CSTATE_DU, CSTATE_TU]:
				uncheckable    += 1
				uncheckable_up += 1
			elif status in [CSTATE_UD, CSTATE_DD, CSTATE_TD]:
				uncheckable    += 1
				uncheckable_dn += 1


		if _DEBUG: print '\t' *  (level+1), self, is_down        , up             , fa             , down           , uncheckable    , uncheckable_up , uncheckable_dn , suspended      

		for nt in self.conditions:

			explosive, group_id = nt.explosive, nt.group

			calculated_state = nt._calculate_state(conditions_data_dict, critical_priority=critical_priority, minimum_priority=minimum_priority, level=level+1, ts=ts)
			if calculated_state is None:
				continue

			primary, status, num_conditions, exploded, alarms_set2, critical_alarms_set2 = calculated_state

			has_conditions += 1 if num_conditions > 0 else 0
			
			alarms_set          = alarms_set.union(alarms_set2)
			critical_alarms_set = critical_alarms_set.union(critical_alarms_set2)

			if group_id is not None:
				counters = explosive_groups.get(group_id, explosive_groups.setdefaul(group_id, GROUP_COUNTERS() ))
				counters['tot'] += 1
				if status == CSTATE_DN:
					counters['dn'] += 1

			if status is not None:
				tot_valid += 1

				has_primary = has_primary or primary

			#
			# STATUS
			#
			if primary:
				if status == CSTATE_DN:
					is_down = True

			if   status == CSTATE_UP             : up          += 1
			elif status == CSTATE_FA             : fa          += 1
			elif status == CSTATE_DN             : down        += 1
			elif status == CSTATE_IN             : suspended   += 1
			elif status in [CSTATE_UU, CSTATE_DU, CSTATE_TU]:
				uncheckable    += 1
				uncheckable_up += 1
			elif status in [CSTATE_UD, CSTATE_DD, CSTATE_TD]:
				uncheckable    += 1
				uncheckable_dn += 1

		# Ci sono dei gruppi esplosi?
		for group_id, counters in explosive_groups.items():
			if counters['tot'] == counters['dn']:
				exploded = True

		global_status = CSTATE_UU

		if _DEBUG: print '\t' *  (level+1), self, is_down        , up             , fa             , down           , uncheckable    , uncheckable_up , uncheckable_dn , suspended      

		while 1:

			if is_down:
				global_status = CSTATE_DN
				break

			if tot_valid == 0:
				# nessuno dei sotto-elementi ha stati significativi,
				# ma ci sono condition (quindi stiamo monitorando) -> Facciamo vedere verde.
				if has_conditions:
					global_status = CSTATE_UP
				else:
					# Se non ci sono condition nei sotto-elementi non stiamo monitorando niente, quindi in realta' non 
					# sappiamo dire nulla sullo stato.
					global_status = CSTATE_UU
				break

			if down:
				if up > 0:
					global_status = CSTATE_FA
					break
				elif fa > 0:
					global_status = CSTATE_FA
					break
				else:
					#print "YYYYYYYYYYY"
					global_status = CSTATE_DN
					break

			if fa:
				global_status = CSTATE_FA
				break

			if up:
				global_status = CSTATE_UP
				break

			#if tot_valid > 0 and suspended == tot_valid:
			if suspended == tot_valid:

				if suspended == has_conditions:
					global_status = CSTATE_IN
				else:
					global_status = CSTATE_UP
				break

			# Se tutte le condition sono uncheckable l'elemento e' uncheckable
			if uncheckable_dn:
				global_status = CSTATE_UD
				break

			if uncheckable_up:
				global_status = CSTATE_UU
				break

			break

		#
		# Aggiustamenti finali
		#
		self.ts              = ts
		self.alarms          = len(alarms_set)
		self.critical_alarms = len(critical_alarms_set)
		self.exploded        = exploded
		
		self.visible_status    = CSTATE_DN if self.critical_alarms != 0 or exploded else global_status

		self.calc_status          = global_status if tot_valid > 0 else None

		return True, self.calc_status, has_conditions, self.exploded, alarms_set, critical_alarms_set

#-----------------------------------------------------------------------------#

class DataGroupTreeNode(TreeNode):

	def _calculate_state(self, conditions_data_dict, critical_priority=PRIORITY_NONE, minimum_priority=PRIORITY_NONE, level=0, ts=None):

		#if _DEBUG: print '\t' * level, "%s" % (self.__class__.__name__,)

		global_status = None 
		exploded      = False

		alarms_set          = set()
		critical_alarms_set = set()

		explosive_groups = {}


		tot_valid = 0

		is_down        = False
		up             = 0
		fa             = 0
		down           = 0
		uncheckable    = 0
		uncheckable_up = 0
		uncheckable_dn = 0
		suspended      = 0

		has_primary = False

		for nt in self.conditions:

			explosive, group_id = nt.explosive, nt.group

			calculated_state = nt._calculate_state(conditions_data_dict, critical_priority=critical_priority, minimum_priority=minimum_priority, level=level+1, ts=ts)
			if calculated_state is None:
				continue

			primary, status, num_conditions, exploded, alarms_set2, critical_alarms_set2 = calculated_state

			alarms_set          = alarms_set.union(alarms_set2)
			critical_alarms_set = critical_alarms_set.union(critical_alarms_set2)

			if group_id is not None:
				counters = explosive_groups.get(group_id, explosive_groups.setdefaul(group_id, GROUP_COUNTERS() ))
				counters['tot'] += 1
				if status == CSTATE_DN:
					counters['dn'] += 1

			if status is not None:
				tot_valid += 1

				has_primary = has_primary or primary

			#
			# STATUS
			#
			if primary:
				if status == CSTATE_DN:
					is_down = True

			if   status == CSTATE_UP             : up          += 1
			elif status == CSTATE_FA             : fa          += 1
			elif status == CSTATE_DN             : down        += 1
			elif status == CSTATE_IN             : suspended   += 1
			elif status in [CSTATE_UU, CSTATE_DU, CSTATE_TU]:
				uncheckable    += 1
				uncheckable_up += 1
			elif status in [CSTATE_UD, CSTATE_DD, CSTATE_TD]:
				uncheckable    += 1
				uncheckable_dn += 1

		# Ci sono dei gruppi esplosi?
		for group_id, counters in explosive_groups.items():
			if counters['tot'] == counters['dn']:
				exploded = True


		if _DEBUG: print '\t' *  (level+1), self, is_down        , up             , fa             , down           , uncheckable    , uncheckable_up , uncheckable_dn , suspended      

		#
		# CALCOLO STATO
		#

		global_status = CSTATE_UU

		while 1:

			if is_down:
				global_status = CSTATE_DN


			if tot_valid == 0:
				global_status = CSTATE_UP

			if down:
				if up > 0:
					global_status = CSTATE_FA
				elif fa > 0:
					global_status = CSTATE_FA
				else:
					#print "YYYYYYYYYYY"
					global_status = CSTATE_DN

			if fa:
				global_status = CSTATE_FA

			if up:
				global_status = CSTATE_UP

			if tot_valid > 0 and suspended == tot_valid:
				global_status = CSTATE_IN

			# Se tutte le condition sono uncheckable l'elemento e' uncheckable
			if uncheckable_up:
				global_status = CSTATE_UU

			if uncheckable_dn:
				global_status = CSTATE_UD

			break


		#
		# Aggiustamenti finali
		#

		self.alarms          = len(alarms_set)
		self.critical_alarms = len(critical_alarms_set)
		self.exploded        = exploded

		self.visible_status    = CSTATE_DN if self.critical_alarms != 0 or exploded else global_status
		
		self.calc_status          = global_status if tot_valid > 0 else None

		#if _DEBUG: print '\t' * (level+1), 'FINAL status=%s, finalstatus=%s, exploded=%s, alarms=%s, critical=%s' % ( self.calc_status, self.visible_status, self.exploded, self.alarms, self.critical_alarms,  )

		return has_primary, self.calc_status, len(self.conditions), self.exploded, alarms_set, critical_alarms_set


#-----------------------------------------------------------------------------#

class ConditionTreeNode(TreeNode):
	"""
	Condition taggata nell'albero
	"""

	def _calculate_state(self, conditions_data_dict,  critical_priority=PRIORITY_NONE, minimum_priority=PRIORITY_NONE, level=0, ts=None):

		if _DEBUG: print '\t' * level, "%s" % (self.__class__.__name__)

		# Questo nodo dell'albero e' relativo ad una condition, ma se l'UUID non compare in "conditions_data_dict"
		# allora:
		#	* La condition e' DELETED
		#	* C'e' un bug e questo nodo non si sa a cosa e' associato
		#
		# In entrambi i casi questo nodo viene "IGNORATO"
		#
		data = conditions_data_dict.get(self.r_id, None)
		if not data:
			return None

		#
		# Calcolo dello stato
		#
		uuid, status, priority, primary = data[0], data[1], data[2], data[3]

		alarms_set          = set()
		critical_alarms_set = set()

		if priority >= minimum_priority:

			exploded = False

			if status is CSTATE_DN:

				if priority >= minimum_priority:
					alarms_set.add(uuid)

				if priority >= critical_priority:
					critical_alarms_set.add( uuid )

			self.exploded        = exploded
			self.alarms          = len(alarms_set)
			self.critical_alarms = len(critical_alarms_set)
			self.visible_status    = status

			self.calc_status          = status   # Stato da passare al padre per il calcolo degli stati
		else:
			self.exploded        = False
			self.alarms          = 0
			self.critical_alarms = 0
			self.visible_status    = status

			self.calc_status          = None     # Stato da passare al padre per il calcolo degli stati

		self.ts = ts

		if _DEBUG: print '\t' * (level+1), 'FINAL status=%s, exploded=%s, alarms=%s, critical=%s, priority=%s' % ( self.calc_status, self.exploded, self.alarms, self.critical_alarms, priority )

		return primary, self.calc_status, 1, self.exploded, alarms_set, critical_alarms_set

