# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.
"""

API per la creazione degli alberi degli stati
=====================================================================================

Per creare gli alberi degli stati bisogna utilizzare le funzioni per caricare tutti i dati di un tree:

::
	
	load_tree_tagging_data()

Questa funzione restiuiscono tutte in output 3 dizionari:

::
	
	tags_by_parent
	tagged_by_tag 
	tagged_by_parent


Esistono altre due funzioni "specifiche", ma non dovrebbero essere usate se non per debugging/testing:

::
	load_tag_tagging_data()
	load_tags_tagging_data()



Questi dizionari vengono passati alla funzione:

::
	
	create_tag_statetree_structure()

Questa funzione crea l'albero degli stati e restituisce il puntatore all'oggetto TreeNode che rappresenta il tag root dell'albero.


Esempio di creazione dell'albero:

::

	tree = Tree.objects.get(...)

	root_tag = tree.root
	
	tags_by_parent, tagged_by_tag, tagged_by_parent = load_tag_tagging_data(root_tag)
	
	statetree_dict = {}
	
	treenode_root_tag = create_tag_statetree_structure(statetree_dict, tags_by_parent, tagged_by_tag, tagged_by_parent, root_tag)
	
	treenode_root_tag.calculate_state(.......)

"""

import os

from django.db import connection

from core.models import MonitoredElement, Condition, DataGroup

from tags.models import ElementTag, Tag

from tags.api.state import *

#------------------------------------------------------------------------------------#
#                                                                                    #
#------------------------------------------------------------------------------------#

_DEBUG = int(os.environ.get('DEBUG', 0) or 0)

#------------------------------------------------------------------------------------#
#                                                                                    #
#------------------------------------------------------------------------------------#

def load_conditions_for_tagging(condition_ids=None):
	"""
	Questa funzione carica i dati delle condition necessari per effettuare il calcolo degli stati.
	
	Restituisce un dizionario con il mapping:
	
		UUID conditio ->  ( <uuid> , <status>, <priority> , <primary>, <path> )
	
	Questo dizionario viene utilizzato dal metodo TreeNode.calculate_state().
	"""

	# Solo le condition "effettive" (non deleted) devono essere caricate
	selected_conditions = Condition.active_objects

	if condition_ids:
		selected_conditions = selected_conditions.filter(uuid__in=condition_ids)

	# Caricamento dal DB
	selected_conditions = selected_conditions.all().values_list('uuid', 'status', 'priority', 'primary', 'path')

	# Dizionario
	return dict([ (e[0], e) for e in selected_conditions ])



#------------------------------------------------------------------------------------#
#                                                                                    #
#------------------------------------------------------------------------------------#

def load_tree_tagging_data(tree):
	"""
	Per il "tree", restituisce una tupla di tre dizionari:
	
		tags_by_parent    : elengo Tag indicizzati per UUID del Tag parent
		
		tagged_by_parent  : elenco degli ElementTag  per UUID del parent nella gerarichia del calmpo "tagged_parent"
		
		tagged_by_tag     : elenco degli ElementTag  per UUID del tag in cui c'e' tagging DIRETTO
	"""
	tag = tree.root
	return load_tag_tagging_data(tag)


def load_tag_tagging_data(tag):
	"""
	Per il "tag", restituisce una tupla di tre dizionari:
	
		tags_by_parent    : elengo Tag indicizzati per UUID del Tag parent
		
		tagged_by_parent  : elenco degli ElementTag  per UUID del parent nella gerarichia del calmpo "tagged_parent"
		
		tagged_by_tag     : elenco degli ElementTag  per UUID del tag in cui c'e' tagging DIRETTO
	"""
	return load_tags_tagging_data([tag])


def load_tags_tagging_data(root_tags):
	"""
	Per tutti i tag "root_tags", restituisce una tupla di tre dizionari:
	
		tags_by_parent    : elengo Tag indicizzati per UUID del Tag parent
		
		tagged_by_parent  : elenco degli ElementTag  per UUID del parent nella gerarichia del calmpo "tagged_parent"
		
		tagged_by_tag     : elenco degli ElementTag  per UUID del tag in cui c'e' tagging DIRETTO
	"""

	tags_by_parent   = {}
	tagged_by_parent = {}
	tagged_by_tag    = {}
	
	t1 = time.time()

	tags =[]
	
	for tag in root_tags:
		
		if tag is None:
			continue
		
		# Carichiamo i TAG 
		tags += [tag] 
		for t in tag.sub_tags.all():
			tags.append(t)

	for t in tags:
		l = tags_by_parent.get(t.parent_id, tags_by_parent.setdefault( t.parent_id, []) )
		l.append(t)

	# Carichiamo gli elementi taggati
	tagged = ElementTag.objects.filter(tag__in=tags).all().values()
	for t in tagged:

		l = tagged_by_parent.get(t['tagged_parent'], tagged_by_parent.setdefault( t['tagged_parent'], [] ))
		l.append(t)

		# Per i soli elementi taggati direttamente indicizziamo per tag
		if t['directly_tagged'] == True:  
			l = tagged_by_tag.get(t['tag_id'], tagged_by_tag.setdefault( t['tag_id'], [] ))
			l.append(t)

	t2 = time.time()

	if _DEBUG:
		print __file__, "load_tag_tagging_data(%s)" % (tag.path), t2 - t1

	return tags_by_parent, tagged_by_tag, tagged_by_parent

#------------------------------------------------------------------------------------#
#                                                                                    #
#------------------------------------------------------------------------------------#

def create_tag_statetree_structure(statetree_dict, tags_by_parent, tagged_by_tag, tagged_by_parent, tag,  tree_parent=None,    l=1,     debug=False , ts=None):
	"""
	Crea un albero di TreeNode del tag "tag" in base alle informazioni di tagging dei tre prametri:
	
		tags_by_parent
		tagged_by_tag 
		tagged_by_parent
	
	Il parametro in input "statetree_dict" deve essere un dizionario e viene popolato con i riferimenti per UUID a tutti gli elementi dell'albero.
	

	"""
	
	ts = ts or int(time.time())

	if debug: print '    ' * l, 'T', tag.name

	#
	# Creiamo il nodo del tag
	#
	statetree_node = TagTreeNode(tag.uuid, tag.uuid, ts=ts)

	statetree_dict[tag.uuid] = statetree_node

	if tree_parent is not None:
		tree_parent.children.append(statetree_node)

	#
	# Creiamo i nodi per tutti i sub tag  (ricorsivamente)
	#
	if debug: print '    ' * (l+1), '--subtags'

	sub_tags = tags_by_parent.get(tag.uuid, [])
	for subtag in sub_tags:
		# Ricorsione
		create_tag_statetree_structure(statetree_dict, tags_by_parent, tagged_by_tag, tagged_by_parent, subtag, tree_parent=statetree_node,   l=l+1, debug=debug  )

	#
	# Creiamo i nodi dell'albero per tutti gli elementi taggati direttamente (elementi di monitoraggio o condition)
	#
	if debug: print '    ' * (l+1), '--subelements'

	for tagged in tagged_by_tag.get(tag.uuid, []):

		if tagged['directly_tagged'] == False:
			continue

		_create_tagged(statetree_dict,  tagged_by_parent, tagged,   tree_parent=statetree_node,  l=l+1, debug=debug)

	return statetree_node


def _create_tagged(states_dict, tagged_by_parent, tagged, tree_parent=None, l=1, debug=False):
	
	if debug: print '    '  * l, tagged['elementtype'], tagged['uuid'], tagged['elementid'], tagged['directly_tagged']
	
	if   tagged['elementtype'] == "tag"       : klass = TagTreeNode
	elif tagged['elementtype'] == "datagroup" : klass = DataGroupTreeNode
	elif tagged['elementtype'] == "condition" : klass = ConditionTreeNode
	else                                      : klass = ElementTreeNode
	
	if tagged['elementtype'] == "condition":
		tree_node = klass(tagged['uuid'], tagged['elementid'])
		tree_parent.conditions.append(tree_node)
	else:
		tree_node = klass(tagged['uuid'], tagged['elementid'])
		tree_parent.children.append(tree_node)

	subs = tagged_by_parent.get(tagged['uuid'], [])

	for sub in subs:

		if sub['directly_tagged'] == True:
			continue

		_create_tagged(states_dict, tagged_by_parent, sub, tree_parent=tree_node, l=l+1, debug=debug)

	states_dict[tagged['uuid']] = tree_node
	
	return tree_node





#------------------------------------------------------------------------------------#
#------------------------------------------------------------------------------------#
#------------------------------------------------------------------------------------#
# METODI PER EFFETTUARE IL FIX dei riferimenti ElementTag.tagged_parent              #
#------------------------------------------------------------------------------------#
#------------------------------------------------------------------------------------#
#------------------------------------------------------------------------------------#

FORCE_UPDATE = False

def _load_data_for_tagged_parent_fix():

	t1 = time.time()

	conditions = Condition.objects.all().values_list('uuid', 'datagroup_id')

	conditions_by_parent = {}
	for uuid, parent_id in conditions:
		l = conditions_by_parent.get(parent_id, conditions_by_parent.setdefault(parent_id, []))
		l.append(uuid)

	t2 = time.time()

	datagroups = DataGroup.objects.all().values_list('uuid', 'element_id')

	datagroups_by_parent = {}
	for uuid, parent_id in datagroups:
		l = datagroups_by_parent.get(parent_id, datagroups_by_parent.setdefault(parent_id, []))
		l.append(uuid)

	t3 = time.time()

	elements = MonitoredElement.objects.all().values_list('uuid', 'parent_id')

	elements_by_parent = {}
	for uuid, parent_id in elements:
		l = elements_by_parent.get(parent_id, elements_by_parent.setdefault(parent_id, []))
		l.append(uuid)

	t4 = time.time()

	if 1:
		print "LOAD condition  %10s %s" % ( len(conditions) , t2 - t1 )
		print "LOAD datagroup  %10s %s" % ( len(datagroups) , t3 - t2 )
		print "LOAD elements   %10s %s" % ( len(elements)   , t4 - t3 )

		#print "conditions      ",           #, "\t", SIZE(conditions)
		#print "datagroups      ",           #, "\t", SIZE(datagroups)
		#print "elements        ",           #, "\t", SIZE(elements)
		#print "conditions index", len(conditions_by_parent)#, "\t", SIZE(conditions_by_parent)
		#print "datagroups index", len(datagroups_by_parent)#, "\t", SIZE(datagroups_by_parent)
		#print "elements   index", len(elements_by_parent  )#, "\t", SIZE(elements_by_parent)

	return elements_by_parent, datagroups_by_parent, conditions_by_parent


def _update_tree_tagged_parent(elements_by_parent, datagroups_by_parent, conditions_by_parent, tree, l=1, force=False):
	
	print "TREE:", tree.name
	
	if tree.root is None:
		return 0
	
	
	tagged_by_tag_uuid = {}
	tagged = ElementTag.objects.filter(tree=tree).all().values()
	count = 0
	for et in tagged:
		elements = tagged_by_tag_uuid.get(et['tag_id'], tagged_by_tag_uuid.setdefault( et['tag_id'], [] ) )
		elements.append(et)
		count += 1
		
	print "\t", count
		
	cursor = connection.cursor()

	updated = _update_tag_tagged_parent(cursor, tagged_by_tag_uuid, elements_by_parent, datagroups_by_parent, conditions_by_parent, tree.root, l=l, force=force)
	
	cursor.close()
	
	return updated

def _update_tag_tagged_parent(cursor, tagged_by_tag_uuid, elements_by_parent, datagroups_by_parent, conditions_by_parent, tag      , l=1, force=False):

	updated = 0
	

	print "\t" * l, "TAG", tag.path

	for subtag in tag.child_tags.all():
		updated += _update_tag_tagged_parent(cursor, tagged_by_tag_uuid, elements_by_parent, datagroups_by_parent, conditions_by_parent, subtag, l=l+1, force=force)


	cursor.execute("COMMIT;")

	#
	# Troviamo tutti gli elementi taggati in questo tag e aggiorniamo i puntatori in base
	#
	et_by_uuid = {}
	
	#elements = tag.tagged.all().values() 
	elements = tagged_by_tag_uuid.get(tag.uuid,[])
	
	for et in elements:
		et_by_uuid[et['elementid']] = et

	for et in et_by_uuid.values():
		
		# Dobbiamo andare in ricorsione a partire dagli elementi taggati direttamente.
		if et['directly_tagged'] == False: 
			continue
		
		updated += _update_element_tagged_parent(cursor, elements_by_parent, datagroups_by_parent, conditions_by_parent, et, None, et_by_uuid, l=l+1, force=force)


	print "\t" * (l+1), updated


	return updated



def _update_element_tagged_parent(cursor, elements_by_parent, datagroups_by_parent, conditions_by_parent, elementtag, parent_elementtag, et_by_uuid, l=1, force=False):

	updated = 0

	element_id = elementtag['elementid']

	if _DEBUG: print "\t" * l, elementtag['elementtype'], element_id

	if parent_elementtag:

		if (elementtag['directly_tagged'] == False):
			parent_id = parent_elementtag['uuid']
		else:
			parent_id = None

		if elementtag['tagged_parent'] != parent_id or force:

			cursor.execute("UPDATE tags_elementtag SET tagged_parent='%s' WHERE uuid='%s'" % (parent_id, elementtag['uuid']))
			updated = cursor.rowcount
			
			if _DEBUG: print "\t" * l, "UPDATE", updated



	#if _DEBUG: print ""

	#
	# Datagroups
	#
	datagroup_ids = datagroups_by_parent.get(element_id, [])
	for dg_id in datagroup_ids:
		updated += _update_datatagroup_tagged_parent(cursor, elements_by_parent, datagroups_by_parent, conditions_by_parent, dg_id, elementtag, et_by_uuid, l=l+1, force=force)

	#
	# Sotto elementi
	#
	subelements_ids = elements_by_parent.get(element_id,[])
	for sub_element_id in subelements_ids:

		sub_elementtag = et_by_uuid.get(sub_element_id, None)
		if not sub_elementtag: continue

		updated += _update_element_tagged_parent(cursor, elements_by_parent, datagroups_by_parent, conditions_by_parent, sub_elementtag, elementtag, et_by_uuid, l=l+1, force=force)

	return updated

def _update_datatagroup_tagged_parent(cursor, elements_by_parent, datagroups_by_parent, conditions_by_parent, datagroup_id, element_elementtag, et_by_uuid, l=1,  force=False):

	updated = 0

	datagroup_elementtag = et_by_uuid[datagroup_id]

	if _DEBUG: print "\t" * l, "DG", datagroup_elementtag['elementtype'], datagroup_id

	parent_id = element_elementtag['uuid'] if (datagroup_elementtag['directly_tagged'] == False) else None
	if datagroup_elementtag['tagged_parent'] != parent_id or force:

		#updated += ElementTag.objects.filter(uuid=datagroup_elementtag.uuid).update(tagged_parent=parent_id)

		cursor.execute("UPDATE tags_elementtag SET tagged_parent='%s' WHERE uuid='%s'" % (parent_id, datagroup_elementtag['uuid']))

		if _DEBUG: print "\t" * l, cursor.rowcount

		updated += cursor.rowcount

	conditions_ids = conditions_by_parent.get(datagroup_id, [])

	for c_id in conditions_ids:
		updated += _update_condition_tagged_parent(cursor, elements_by_parent, datagroups_by_parent, conditions_by_parent, c_id, datagroup_elementtag, et_by_uuid, l=l+1, force=force)

	return updated

def _update_condition_tagged_parent(cursor, elements_by_parent, datagroups_by_parent, conditions_by_parent, condition_id, datagroup_elementtag, et_by_uuid, l=1,  force=False):

	# La condition che sto guardando e' figlia del datagroup, ma non e' taggata nel tag corrente.
	#
	#    Succede quando un tree e' unique=True e la condition e' stata taggata direttamente in un tag differente da quello del nodo.
	# 
	# quindi non la trovero dentro et_by_uuid.
	#

	et = et_by_uuid.get(condition_id, None)
	if et is None:
		return 0

	updated = 0

	if _DEBUG: print "\t" * l, "C", et['elementtype'], condition_id
	
	p = datagroup_elementtag['uuid'] if (et['directly_tagged'] == False) else None
	if et['tagged_parent'] != p or force:
		#et.tagged_parent = p
		#et.save()
		#updated += ElementTag.objects.filter(uuid=et.uuid).update(tagged_parent=p)
		
		cursor.execute("UPDATE tags_elementtag SET tagged_parent='%s' WHERE uuid='%s'" % (p, et['uuid']))
		updated = cursor.rowcount

		if _DEBUG: print "\t" * l, cursor.rowcount

	#if _DEBUG: print ""

	return updated
