# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

Base Dati e API a Basso livello
========================================================================

.. contents:: Contenuti

Introduzione
----------------------------------------------------------------

TODO

Modello ER
----------------------------------------------------------------

TODO


Operazioni che triggerano il refresh degli stati dei tag
----------------------------------------------------------------

::
	
	CODED	CASE	Type		OPERATION	DETAILS			TO DO
	-----	-----	-----		---------	-------			------
	
	x	1	TagTree		CREATE					load tree data + refresh tree state
	
		2	TagTree		RENAME					-
	
	x	3	TagTree		DELETE					free tree data
	
	
	x	4	TagTree		SET 		nostate = true		refresh tree state
			
	x	5	TagTree		SET 		nostate = false		refresh tree state
	
	
	x	6	Tag		CREATE					load tree data + refresh tree state
	
	x	7	Tag		DELETE					load tree data + refresh tree state
	
	
	x	8	element		TAG 					load tree data + refresh tree state
	
	x	9	element		UNTAG					load tree data + refresh tree state
	
	x	10	element		SET		explosive_group = ...	load tree data + refresh tree state
	
	
	x	11	prioritylevel   CREATE					refresh tree state
			
	x	12	prioritylevel   DELETE 					refresh tree state

"""

import time
import logging
import json
import re
import traceback
import logging

from sanet_common.uuids import MAX_UUID_LEN, create_global_uuid
from sanet_common.sanetdefs.constants import *
from sanet_common.sanetdefs.constants import ELEMENT_KIND_NODE, ELEMENT_KIND_IFACE, ELEMENT_KIND_STORAGE, ELEMENT_KIND_SERVICE, ELEMENT_KIND_DEVICE

#from django.db import transaction
from django.db import models
from django.db import connection
from django.db.models import Q
from django.db.models import Max
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver
from django.dispatch import Signal
from django.contrib.auth.models import User, Group

from core.db_utils import transaction_commit_on_success

from core.models import sanet_init_subapp
from core.models import sanet_check_tenant
from core.models import sanet_cleanup_tenant
from core.models import sanet_post_check
from core.models import sanet_model_update

from core.models import sanetd_process_changed_conditions

from core.models import sanetd_alarm_msg_create_template_symbols

from core.models import statusd_handle_request

from core.models import Tenant
from core.models import * # FIX: remove this
from core.statecolor import get_tag_state_color

from tags.server.tagsd.api import signal_state_refresh

from tags.server.tagsd.api import signal_conditions_changed

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

DEBUG = os.environ.get('DEBUG_SANET_TAGS', False)

#-----------------------------------------------------------------------------#
# CONSTANTS                                                                   #
#-----------------------------------------------------------------------------#

TAG_PATH_DIVIDER = '/'

TAG_ELEMENTTYPE_NODE      = 'node'
TAG_ELEMENTTYPE_IFACE     = 'interface'
TAG_ELEMENTTYPE_STORAGE   = 'storage'
TAG_ELEMENTTYPE_SERVICE   = 'service'
TAG_ELEMENTTYPE_DEVICE    = 'device'
TAG_ELEMENTTYPE_DATAGROUP = 'datagroup'
TAG_ELEMENTTYPE_CONDITION = 'condition'
TAG_ELEMENTTYPE_TIMEGRAPH = 'timegraph'

TAG_ELEMENTTYPE_TO_MODEL = {
	'node'     : MonitoredNode,
	'interface': MonitoredIface,
	'storage'  : MonitoredStorage,
	'service'  : MonitoredService,
	'device'   : MonitoredDevice,
	'datagroup': DataGroup,
	'condition': Condition,
	'timegraph': TimeGraph,
}

#-----------------------------------------------------------------------------#

MODEL_TO_TAG_ELEMENTTYPE = dict([ (v,k) for k,v in TAG_ELEMENTTYPE_TO_MODEL.items()])

ELEMENT_KIND_TO_TAG_ELEMENTTYPE = {
	ELEMENT_KIND_NODE    : TAG_ELEMENTTYPE_NODE,
	ELEMENT_KIND_IFACE   : TAG_ELEMENTTYPE_IFACE,
	ELEMENT_KIND_STORAGE : TAG_ELEMENTTYPE_STORAGE,
	ELEMENT_KIND_SERVICE : TAG_ELEMENTTYPE_SERVICE,
	ELEMENT_KIND_DEVICE  : TAG_ELEMENTTYPE_DEVICE,
}

TAG_ELEMENTTYPE_TO_ELEMENT_KIND = {
	TAG_ELEMENTTYPE_NODE    : ELEMENT_KIND_NODE    ,
	TAG_ELEMENTTYPE_IFACE   : ELEMENT_KIND_IFACE   ,
	TAG_ELEMENTTYPE_STORAGE : ELEMENT_KIND_STORAGE ,
	TAG_ELEMENTTYPE_SERVICE : ELEMENT_KIND_SERVICE ,
	TAG_ELEMENTTYPE_DEVICE  : ELEMENT_KIND_DEVICE  ,
}

#-----------------------------------------------------------------------------#
# SIGNALS                                                                     #
#-----------------------------------------------------------------------------#

"""
Segnale lanciato al termine di una operazione di tagging.
Poiche' il tagging coinvolge l'elemento taggato e tutto quello che contiene,
inviamo un unico segnale al termine del tagging.
"""
sanet_tag_post_tagging_changed = Signal(providing_args=["instance"])

"""
Segnale lanciato prima di della cancellazione effettiva di un tagtree.
Le sub-app che referenziano TagTree o TagEntry devono provvedere alla cancellazione
di tutte le entita' collegate.

   * instance: Tagtree
   * cursor  : connection.cursor 
   
"""
sanet_tag_tagtree_pre_delete = Signal(providing_args=['instance', 'cursor'])

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class TagTree(UUIDKeyModel, TenantDependentModel, MTimeModel): 
	"""
	TODO
	"""
	# uuid   = 
	# tenant =
	# mtime  = 

	name     = models.CharField(max_length=254, unique=True, null=False, blank=False, db_index=True)
	descr    = models.TextField(blank=True, null=False, default='', db_index=False)

	owner    = models.ForeignKey(User, null=True, default=None, db_index=True, related_name="owned_trees")

	allnodes = models.BooleanField(null=False, default=False)

	uniques  = models.BooleanField(null=False, default=False)

	nostate  = models.BooleanField(null=False, default=False)

	nomap    = models.BooleanField(null=False, default=False)

	# Questo campo viene aggiornato quando il demone tagsd ricalcola lo stato di un tagtree
	stime    = models.FloatField(null=False, default=time.time, db_index=False)

	def __str__(self):
		return "<Tagree:%s>" % (self.name)

	def get_dict(self):
		
		return {
			'uuid'           : self.uuid,
			'name'           : self.name,
		}

	def update_state_time(self, ts=None):
		self.stime = ts or time.time()
		self.concurrent_save(['stime'])

	def get_state_time(self):
		return self.stime

	#@property
	#def path_entry(self):
	#	return self.name

	@property
	def path(self):
		return "%s:" % (self.name)

	@property
	def root(self):
		try:
			return self.tags.get(parent__isnull=True)
		except Tag.DoesNotExist, e:
			return None

	@transaction_commit_on_success
	def rename(self, new_name, user=None):

		if user:
			user.assert_capability(self.tenant, 'tags.tagtree.edit')
			self.assert_permission_readwrite(user)

		path_prefix     = self.path
		path_prefix_len = len(path_prefix)

		# FIX: check if new_name is a valid name
		self.name = new_name
		self.mtime = time.time()
		self.save()  # FIX: update only the path

		new_prefix = self.path

		for tag in self.tags.select_related('parent'):
			tag.path = TAG_PATH_DIVIDER.join([ new_prefix ,  tag.path[ path_prefix_len : ] ])
			tag.save() # FIX: update only the path

	@transaction_commit_on_success
	def set_uniques(self, uniques):

		if self.uniques == uniques:
			return

		# L'albero deve gestire solo tag univoci quindi vediamo
		if self.uniques == False:
			# Trasformiamo gli elementtag  in univoci forzando il valore del campo "unique_sentinel"
			# Se un elemento e' taggato in due punti dell'albero questa operazione dovrebbe generare
			# un errore 
			#updated_count = ElementTag.objects.filter(tag__tree=self).update(unique_sentinel=ElementTag.TAGGED_UNIQUE)
			updated_count = ElementTag.objects.filter(tree=self).update(unique_sentinel=ElementTag.TAGGED_UNIQUE)

		self.uniques = uniques
		self.save()

	def get_groups(self):
		return [ ttgp.group for ttgp in TagTreeGroupPermission.objects.filter(dirtree=self) ]

	def get_group_permissions(self):
		return TagTreeGroupPermission.objects.filter(dirtree=self)

	def can_edit(self, user):
		if user.is_superuser:
			return True
		if user == self.owner:
			return True
		return TagTreeGroupPermission.objects.filter(dirtree=self).filter(group__in=user.groups.all(),  auth_type=DIRENTRY_AUTH_READWRITE).count() > 0

	def assert_permission_read(self, user):
		if user.is_superuser:
			return True
		if user == self.owner:
			return True
		return TagTreeGroupPermission.objects.filter(dirtree=self).filter(group__in=user.groups.all(),  auth_type=DIRENTRY_AUTH_READONLY).count() > 0

	def assert_permission_readwrite(self, user):
		if not self.can_edit(user):
			raise ApplicationException("User '%s' has no permission on tree '%s'." % (user.username, self.name))
		return True

	@staticmethod
	def get_visible_by_user(tenant, user):
		return TagTree.get_visible_trees_by_user(tenant=tenant,user=user)
		
	@staticmethod
	def get_visible_trees_by_user(tenant, user, auth_type=None):

		trees = TagTree.objects

		#if exclude_tagtrees:
		#	trees.exclude(uuid__in=[t.uuid for t in exclude_tagtrees])

		if user is None or user.is_superuser:
			trees = trees.all()
			return trees

		#
		# Dobbiamo creare una query per selezionare tutti i tagtree associati ai gruppi a cui appartiene l'utente
		#
		visible_trees_sub_query = """
					select distinct t.uuid
					  from tagtree t inner join tagtree_group     t_g ON (t.uuid       = t_g.dirtree_id)
							 inner join auth_user_groups  u_g ON (t_g.group_id = u_g.group_id  )
					 where u_g.user_id = %s
					""" 

		# Se abbiamo anche specificato il tipo di accesso al TagTree (read-only o read-write)
		if auth_type:
			visible_trees_sub_query += ' and t_g.auth_type = %s' % (auth_type)

		visible_trees_sub_query = visible_trees_sub_query % (user.id)
		
		#
		# Prendiamo tutti i tree che soddisfano il filtro sulle autorizzazioni o che appartengono all'utente.
		#
		trees = trees.extra(
			where=[
				"("  # Non togliere queste parentesi tonde altrimenti il queriset non e' ulteriormente filtrabile
				+ 
				"(tagtree.uuid IN (" + visible_trees_sub_query + "))" + " OR " + '(tagtree.owner_id = %s)' % (user.id) 
				+ 
				")"  # Non togliere queste parentesi tonde altrimenti il queriset non e' ulteriormente filtrabile
			]
		)

		return trees


	def set_owner(self, user, action_user=None):
		
		if action_user:
			if not action_user.is_superuser:
				raise ApplicationException("Only superuser can change set owner.")

		self.owner = user
		self.save()

	def unset_owner(self, action_user=None):

		if action_user:
			if not action_user.is_superuser:
				raise ApplicationException("Only superuser can change set owner.")

		self.owner = None
		self.save()

	class Meta:
		unique_together = (
			("tenant", "owner", "name"),
		)
		db_table = "tagtree"


class TagTreeGroupPermission(TenantDependentModel, models.Model):
	"""
	TODO
	"""

	dirtree   = models.ForeignKey(TagTree)
	group     = models.ForeignKey(Group)

	auth_type = models.PositiveSmallIntegerField(null=False, default=DIRENTRY_AUTH_READWRITE) 

	def str_auth_type(self):
		if self.auth_type == DIRENTRY_AUTH_READWRITE:
			return "RW"
		else:
			return  "RO"

	class Meta:
		db_table = "tagtree_group"

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class MetaTag(UUIDKeyModel, TenantDependentModel, MTimeModel): 
	"""
	TODO
	"""

	class Meta:
		abstract = True

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class Tag(MetaTag):
	"""
	TODO
	"""

	# uuid   = 
	# tenant =
	# mtime  = 

	tree        = models.ForeignKey(TagTree, null=False, db_index=True, related_name="tags")

	name        = models.TextField(null=True, blank=True, default=None, db_index=True)
	description = models.TextField(null=True, blank=True, default=None)                         # FIX: schifo, usare not null with default ''

	parent      = models.ForeignKey('self', null=True, default=None, db_index=True, related_name="child_tags")

	icon        = models.TextField(null=True, blank=True, default=None)

	path        = models.TextField(null=False, blank=False, db_index=True)


	def __str__(self):
		return '<TAG: %s>' % (self.name)
		
		
	def get_dict(self):
		
		return {
			'uuid'           : self.uuid,
			'name'           : self.name,
			'mtime'          : self.mtime, 
			'path'           : self.path,
			'icon'           : self.icon,
			'tree_id'        : self.tree_id,
			'parent_id'      : self.parent_id,
		}

	@staticmethod
	def get_visible_by_user(tenant, user, auth_type=None):
		
		trees = TagTree.get_visible_trees_by_user(tenant, user, auth_type=auth_type)
		
		return Tag.objects.filter(tree__in=[t.uuid for t in trees])

	@property
	def descr(self):
		return self.description or ''

	@property
	def sub_path(self):
		return self.name

	def update_path(self):
		if self.parent != None:
			path = self.parent.path
		else:
			path = self.tree.path
		self.path = TAG_PATH_DIVIDER.join([path, self.sub_path])

	def get_tag_element_counter(self):
		
		query = """
			 select count(elementid), elementtype
	
			  from tags_elementtag tagelement inner join tagentry tag on (tagelement.tag_id = tag.uuid)
			  
			  where (
			         tag.path =    '""" + self.path.replace("'","''") + """'       
			         or 
			         tag.path like '""" + self.path.replace("'","''") + """/%%'    
			        )
			  group by elementtype;
		"""
		
		cursor = connection.cursor()

		cursor.execute(query)
		rows = cursor.fetchall()
			
		return rows
		
	@property
	def sub_tags(self):
		prefix = self.path + TAG_PATH_DIVIDER  # NOTA: aggiungedo il TAG_PATH_DIVIDER in fondo, escludo automaticamente il tag "self"
		return Tag.objects.filter(path__startswith=prefix)

	def ancestors_tags_list(self):
		r = []
		p = self.parent
		while p:
			r.append(p)
			p = p.parent
		r.reverse()
		return r

	def get_tags_path(self):
		return self.ancestors_tags_list() + [ self ]

	def get_directly_tagged_elements(self):

		ids = [ e[0] for e in self.tagged.filter(directly_tagged=True).values_list('elementid') ]

		return MonitoredElement.objects.filter(uuid__in=ids)

	def get_elementtags_with_elements(self, only_directly_tagged=True, path_names=False):

		from django.db import connection
		from django import db

		db.reset_queries()

		tagged =  self.tagged
		if only_directly_tagged:
			tagged = tagged.filter(directly_tagged=True)
		tagged = tagged.all()

		element_ids = [ e.elementid for e in tagged ]

		if path_names:
			elements  = [ e for e in MonitoredElement.objects.filter(uuid__in=element_ids).select_related('parent').all() ]
		else:
			elements  = [ e for e in MonitoredElement.objects.filter(uuid__in=element_ids).all() ]

		elements += [ e for e in DataGroup.active_objects.filter(uuid__in=element_ids).all() ]
		elements += [ e for e in Condition.active_objects.filter(uuid__in=element_ids).all() ]

		#
		# FIX: noi dobbiamo far vedere solo i datagroup e condition NON DELETED.
		#      Bisogna trovare un modo piu' efficiente per estrarre subito solo i tagged NON DELETED dal DB e non
		#      filtrarli dopo.
		#
		valid = set( [ e.uuid for e in elements ] )
		tagged = [ t for t in tagged if t.elementid in valid ]

		return tagged, elements

	def get_all_nodes(self, recursive=True):

		filter_condition = """
			tag.path = '""" + self.path.replace("'","''") + """'       -- tag itself
		"""

		if recursive:
			filter_condition += """
				or 
				tag.path like '""" + self.path.replace("'","''") + """/%%'    -- subtags
			"""

		query = """
			 select me.*
			   from monitoredelement me inner join tags_elementtag te  on (me.uuid = te.elementid)
			                            inner join tagentry        tag on (te.tag_id = tag.uuid)
			  where (
			         @@filter_condition@@
			        )
			    and 
			        te.elementtype = 'node'
			  order by me.name
		"""

		query = query.replace("@@filter_condition@@", filter_condition)

		return [ n for n in MonitoredElement.objects.raw(query) ]

	def get_all_storages(self):
		query = """
			 select me.*
			   from monitoredelement me inner join tags_elementtag te  on (me.uuid = te.elementid)
			                            inner join tagentry        tag on (te.tag_id = tag.uuid)
			  where (
			         tag.path =    '""" + self.path.replace("'","''") + """'       -- tag itself
			         or 
			         tag.path like '""" + self.path.replace("'","''") + """/%%'    -- subtags
			        )
			    and te.elementtype = 'storage'
			  order by me.name
		"""
		return [ n for n in MonitoredElement.objects.raw(query) ]

	def get_all_conditions_by_user(self, user):
		# NOTA: se chiamiamo questo metodo probabimente lo stiamo facendo per un utente che abbiamo gia' verificato
		# quindi fare un'altra volta filtraggi e' inutile. Sarebbe meglio farli di nuovo.
		return self.get_all_conditions()

	def get_all_timegraphs_by_user(self, user):
		# NOTA: se chiamiamo questo metodo probabimente lo stiamo facendo per un utente che abbiamo gia' verificato
		# quindi fare un'altra volta filtraggi e' inutile. Sarebbe meglio farli di nuovo.
		return self.get_all_timegraphs()

	def get_all_datagroups(self):

		datagroups = DataGroup.active_objects

		datagroups = datagroups.extra(
			tables = ['tags_elementtag', 'tagentry'],
			where  = [
				"((tagentry.path             = %s) or (tagentry.path like %s))",
				"tags_elementtag.tag_id      = tagentry.uuid",
				"tags_elementtag.elementtype = 'datagroup'",
				"datagroup.uuid              = tags_elementtag.elementid",
				"datagroup.deleted           = false",
			],
			params=[
				self.path,
				self.path+'/%%'
			]
		).distinct()

		return datagroups


	def get_all_conditions(self):

		conditions = Condition.active_objects

		conditions = conditions.extra(
			tables = ['tags_elementtag', 'tagentry'],
			where  = [
				"((tagentry.path             = %s) or (tagentry.path like %s))",
				"tags_elementtag.tag_id      = tagentry.uuid",
				"tags_elementtag.elementtype = 'condition'",
				"condition.uuid              = tags_elementtag.elementid",
				"condition.deleted           = false",
			],
			params=[
				self.path,
				self.path+'/%'
			]
		).distinct()

		return conditions

	def get_all_datasources(self):

		dss = DataSource.active_objects

		dss = dss.extra(
			where  = [
			          """exists (
			                SELECT 1
			                  FROM tags_elementtag et inner join tagentry  tag on (et.tag_id = tag.uuid)
			                                          inner join datagroup d   on (d.uuid    = et.elementid and et.elementtype = 'datagroup')
			                 WHERE 
			                       ((tag.path             = %s) or (tag.path like %s))
			                   AND d.deleted = false
			                   AND d.uuid = datasource.datagroup_id
			          )"""
			],
			params=[
				self.path,
				self.path+'/%'
			]
		)

		return dss

	def get_all_timegraphs(self):

		query = """
		        SELECT d.uuid
		          FROM tags_elementtag et inner join tagentry  tag on (et.tag_id = tag.uuid)
		                                  inner join datagroup d   on (d.uuid    = et.elementid and et.elementtype = 'datagroup')
		         WHERE 
		               ((tag.path             = %s) or (tag.path like %s))
		           AND d.deleted = false
		"""

		params=[
			self.path,
			self.path+'/%'
		]

		cursor = connection.cursor()
		cursor.execute(query, params)
		
		ids = [r[0] for r in cursor.fetchall()]

		return TimeGraph.active_objects.filter(datagroup__uuid__in=ids).order_by('datagroup__path', 'name')

	def get_logs(self):

		events = Event.objects

		events = events.extra(
			tables = ['condition', 'tags_elementtag', 'tagentry'],
			where  = [
				"tagentry.path               like '%s'" % (self.path + "%%"),
				"tags_elementtag.tag_id      = tagentry.uuid ",
				"tags_elementtag.elementtype = 'condition' ",
				"condition.uuid              = tags_elementtag.elementid ",
				"condition.deleted           = false ",
				"((event.condition_path = condition.path) or (event.condition_uuid = condition.uuid))",
			]
		)

		return events

	def get_logs_by_user(self, user=None):
		"""
		This method is just for code compatibility with other user-dependent code.
		If we access to a Tag, we are sure "user" can view all it's logs.
		"""
		return self.get_logs()

	def get_path(self):
		return self.path

	def get_icon_url(self):
		icon_name = self.get_icon_name_or_default()
		icon_url = reverse('webui.views.icons.get_colored_icon3', kwargs = {'icon': icon_name, 'tenantid': self.tenant_id})
		return icon_url

	def get_icon_url_with_state(self, status_code):
		return self.get_icon_url() + get_tag_state_color(status_code)

	def get_icon_name_or_default(self):
		if self.icon: return self.icon
		return 'tag'

	def get_all_elements(self, get_queryset = False):
		sub_tree_prefix = self.path + TAG_PATH_DIVIDER
		return ElementTag.objects.filter( Q(tag__path=self.path) | Q(tag__path__startswith=sub_tree_prefix) )

	@property
	def subnodes(self):
		return self.get_directly_tagged_elements().filter(kind=ELEMENT_KIND_NODE)

	def can_edit(self, user):
		groups = user.groups.all()
		tree = self.tree

		auth_type = TagTreeGroupPermission.objects.filter(group__in = groups, dirtree = tree).aggregate(Max('auth_type'))

		return auth_type

	class Meta:

		db_table = "tagentry"

		unique_together = (
			("tenant", "tree", "parent", "name"),
		)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class TagMetaData(UUIDKeyModel, TenantDependentModel, MTimeModel):
	"""
	Meta informazioni associate ad un tag. Usato per codificare dati
	da poter utilizzare nella messaggistica.
	"""

	tag         = models.ForeignKey(Tag, related_name="metadata", db_index=True)

	name        = models.TextField(null=False, blank=True, db_index=True)
	value       = models.TextField(null=False, blank=True, default="")                       

	def value_lines(self):
		return self.value.split("\n")

	class Meta:

		db_table = "tagentrymetadata"

		unique_together = (
			("tenant", "tag", "name"),
		)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class ElementTag(MetaTag):
	"""
	TODO
	"""

	TAGGED_UNIQUE = 'UNIQUE'

	tag             = models.ForeignKey(Tag, related_name="tagged", db_index=True)

	elementtype     = models.TextField(null=False, db_index=True)
	elementid       = models.TextField(null=False, db_index=True)

	#
	# Campi per gestione esplosivita'
	#

	# Si/No: indica se l'elemento taggato all'interno del tag deve essere "esplosivo", ovvero
	#        il suo stato "down"/rosso forza a catena lo stato "down" del tag in cui e' contenuto.
	is_explosive    = models.BooleanField(null=False, default=False, db_index=True, )

	# All'interno dello stesso e' possibile "raggruppare diversi elementi in gruppi" e
	# scatenare l'effetto "esplosivo" solo se tutti gli elementi del gruppo sono
	# "down".
	explosive_group = models.IntegerField(null=True, default=None, db_index=True, )

	#
	# Campi di comodo
	#

	# Flag per sapere se questo elemento e stato taggato "direttamente"
	# o se e' stato taggato in seguito al tag ricorsivo di un altro elemento.
	# In questo modo siamo in grado di selezionare immediatamente gli elementi
	# interni di un tag escludendo sotto-elementi taggati.
	directly_tagged = models.BooleanField(null=False, default=True, db_index=True, )

	# Campi di comodo per 
	# Nei tagtree con flag unique=True  un elemento non puo' essere taggato in due tag
	# dello stesso tree. 
	# Per fare in modo che questo non succeda mai, usiamo un campo di comodo con un indice
	# univoco:
	#
	#     <tenant, tree, elementid, unique_sentinel>:
	#
	# Se deve essere univoco questo campo viene valorizzato con il valore TAGGED_UNIQUE,
	# in questo modo due taggedelement in due tag diversi creano errore per 
	# chiave duplicata.
	# 
	# Se puo' essere duplicato viene valorizzato con un UUID casuale.	
	tree            = models.ForeignKey(TagTree, null=False, db_index=True, related_name="taggedelements") # NOTA: il campo tree e' "di fatto" duplicato perche' e' valorizzato con lo stesso valore di self.tag.tree
	unique_sentinel = models.CharField(null=False, max_length=MAX_UUID_LEN, default=create_global_uuid)

	# Questo e' un campo di comodo per poter ricostruire velocemente la gerarchia 
	# degli elementtag rispettando il tagging diretto o indiretto.
	# Se un elementag e' taggato indirettamente allora deve essere "figlio" di un altro elementag taggato "direttamente"
	# e questo campo contiene l'UUID dell'ElementTag "padre".
	tagged_parent   = models.CharField(null=True, max_length=MAX_UUID_LEN, default=None)

	def __str__(self):
		return "<ElementTag:uuid=%s, parent=%s, etype=%20s, eid=%s, explosive=%s, group=%s >" % (self.uuid, self.tagged_parent, self.elementtype, self.elementid, self.is_explosive, self.explosive_group)

	@property
	def path(self):
		return self.tag.path + TAG_PATH_DIVIDER

	@property
	def element(self):
		return TAG_ELEMENTTYPE_TO_MODEL[self.elementtype].objects.get(uuid=self.elementid)

	class Meta:
		unique_together = [
			("tenant", "tree", "elementid", "unique_sentinel"),
		]

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class TagEntryInsertRule(TenantDependentModel):
	"""
	TODO
	"""

	tagentry     = models.ForeignKey(Tag, null=False, db_index=True, related_name="tagrules")

	# Espressione regolare per fare match sul nome del nodo quando viene creato
	rule         = models.TextField(null=False, blank=False, db_index=True) 

	def match(self, element):
		if re.search(self.rule, element.name):
			return True
		else:
			return False

	class Meta:
		unique_together = (
			("tenant", "tagentry", "rule"),
		)
		db_table = "tagentryinsertrule"

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

class TagGroupAuth(TenantDependentModel):
	"""
	TODO
	"""

	#tenant        = 
	group          = models.ForeignKey(Group, null=False)
	root_direntry  = models.ForeignKey(Tag, null=False)

	auth_type      = models.PositiveSmallIntegerField(null=False, default=DIRENTRY_AUTH_READWRITE) 

	class Meta:
		db_table = "tagentry_group"

#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#

def _QUERY_VISIBLE_ELEMENTS_BY_USER(user, trees, elementtypes=None, table_name='monitoredelement', key_field='uuid', subquery_key_field=None):

	# This join is slower.
	#
	#filter_query = """
	#         SELECT distinct et.elementid as uuid
	#           FROM auth_user u INNER JOIN  auth_user_groups ug ON (u.id          = ug.user_id)
	#                    INNER JOIN  auth_group       g  ON (ug.group_id   = g.id)
	#                    INNER JOIN  tagtree_group    tg ON (g.id          = tg.group_id)
	#                    INNER JOIN  tagtree          tt ON (tg.dirtree_id = tt.uuid)
	#                    INNER JOIN  tagentry         te ON (tt.uuid       = te.tree_id)
	#                    INNER JOIN  tags_elementtag  et ON (te.uuid       = et.tag_id)
	#          WHERE u.id = %s
	#""" % (user.id)

	#visible_trees = TagTree.get_visible_trees_by_user(user, excluded_tagtrees=excluded_tagtrees)

	if subquery_key_field:
		key_field = subquery_key_field

	if not trees:
		# L'utente non puo' vedere nulla quindi restituiamo una query di filtro che ritorna sempre un queryset vuoto
		filter_query = """
		  SELECT et.elementid as uuid
			FROM tags_elementtag as et
		   WHERE false
		"""
	else:
		filter_query = """
		    SELECT 1
		      FROM tags_elementtag  et
		     WHERE  
		           %(table_name)s.%(key_field)s = et.elementid
		       AND 
		           et.tree_id in (%(in_clause)s)
		""" % { 
			'table_name': table_name,
			'key_field' : key_field,
			'in_clause' : ','.join(["'%s'" % (tree.uuid) for tree in trees]) 
		}

		if elementtypes is not None:
			filter_query += """
			    AND 
			        et.elementtype in (%s)
			""" % ( ','.join(["'%s'" % (e) for e in elementtypes]) )

	return filter_query



def TAG_FILTER(tenant, user, queryset, table_name, key_field='uuid', user_field='owner_id', elementtypes=None, excluded_tagtrees=None):
	"""
	Questa funzione prende in input un "queryset" django e aggiunge le necessarie condizioni/subquery
	per filtrare gli elementi visibili in base all'utente / tagging.

		* queryset : il queryset django
		* table_name: e' il nome della tabella principale del queryset.
		* key_field : e' il nome del campo della chiave primaria nella tabella "table_name"
		* user_field: e' il nome del campo che contiene la foreign key all'utente
		* elementtypes: tipi di elementi taggati che si vuole selezionare. Default tutti
		* excluded_tagtrees: elenco uuid dei tagtree da escludere nel calcolo della visibilita'.
	"""

	# Gli utenti admin di default vedono tutto, a meno che non siano stati specificati dei tagtree da escludere
	if user.is_superuser:
		if not excluded_tagtrees:
			return queryset

	# Utenti normali o admin con dei filtri

	sql_where_filters = []

	#
	# Visible elements
	#
	visible_trees = TagTree.get_visible_trees_by_user(tenant, user)

	visible_filter_query = _QUERY_VISIBLE_ELEMENTS_BY_USER(user, visible_trees , elementtypes=elementtypes, table_name=table_name)

	sql_where_filters.append( """
		(
		 exists ( %(subquery)s )
		 OR
		 %(table_name)s.%(user_field)s = %(user_id)s
		)
	""" % {
		"table_name": table_name,
		"key_field" : key_field,
		"user_field": user_field,

		"user_id"   : user.id if user else '-1',

		"subquery"  : visible_filter_query ,
	})

	#
	# Excluded elements
	#
	if excluded_tagtrees:
		exclude_filter_query = _QUERY_VISIBLE_ELEMENTS_BY_USER(user, excluded_tagtrees, elementtypes=elementtypes, table_name=table_name)
		sql_where_filters.append( """ NOT exists ( %s ) """ % ( exclude_filter_query) )

	# Final queryset
	queryset = queryset.extra(where = sql_where_filters )

	return queryset


# >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
# >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
# FUNZIONI TEMPORANEE FINCHE' NON AGGIUSTIAMO IL TAGGING DI TIMEGRAPH E TIMEGRAPHSERIE
# >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
# >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

def TAG_FILTER_BY_DATAGROUP(tenant, user, queryset, table_name, key_field='uuid', user_field='owner_id', excluded_tagtrees=None, subquery_key_field=None):

	# Gli utenti admin di default vedono tutto, a meno che non siano stati specificati dei tagtree da escludere
	if user.is_superuser:
		if not excluded_tagtrees:
			return queryset

	# Utenti normali o admin con dei filtri

	sql_where_filters = []

	#
	# Visible elements
	#
	visible_trees = TagTree.get_visible_trees_by_user(tenant, user)
	visible_filter_query = _QUERY_VISIBLE_ELEMENTS_BY_USER(user, visible_trees , elementtypes=['datagroup'], table_name=table_name, key_field=key_field, subquery_key_field=subquery_key_field)
	sql_where_filters.append( """ 
		(
		  exists ( %(subquery)s )
		 OR 
		 %(table_name)s.%(user_field)s = %(user_id)s
		) 
	""" % { 
		"table_name": table_name, 
		"key_field" : key_field, 
		"user_field": user_field,

		"user_id"   : user.id if user else '-1',
		
		"subquery"  : visible_filter_query , 
	})

	#
	# Excluded elements
	#
	if excluded_tagtrees:
		exclude_filter_query = _QUERY_VISIBLE_ELEMENTS_BY_USER(user, excluded_tagtrees, elementtypes=['datagroup'], table_name=table_name)
		sql_where_filters.append( """  NOT exists ( %s ) """ % ( exclude_filter_query) )

	return queryset.extra(
		where = sql_where_filters,
	)

def TAG_FILTER_BY_TIMEGRAPH(tenant, user, queryset, table_name, key_field='uuid', user_field='owner_id', excluded_tagtrees=None):

	# Gli utenti admin di default vedono tutto, a meno che non siano stati specificati dei tagtree da escludere
	if user.is_superuser:
		if not excluded_tagtrees:
			return queryset

	# Utenti normali o admin con dei filtri

	sql_where_filters = []

	#
	# Visible elements
	#
	visible_trees = TagTree.get_visible_trees_by_user(tenant, user)
	visible_filter_query = _QUERY_VISIBLE_ELEMENTS_BY_USER(user, visible_trees , elementtypes=['datagroup'], table_name='datagroup')
	sql_where_filters.append( """ 
		(
		   (
		      "timegraphserie".timegraph_id = "timegraph".uuid
		      AND
		      "timegraph".datagroup_id      = "datagroup".uuid
		      AND
 		      exists ( %(subquery)s )
		   )
		
		   OR 
		   
		   %(table_name)s.%(user_field)s = %(user_id)s
		) 
	""" % {
		"table_name": table_name, 
		"key_field" : key_field, 
		"user_field": user_field,

		"user_id"   : user.id if user else '-1',
		
		"subquery"  : visible_filter_query , 
	})

	#
	# Excluded elements
	#
	if excluded_tagtrees:
		exclude_filter_query = _QUERY_VISIBLE_ELEMENTS_BY_USER(user, excluded_tagtrees, elementtypes=['datagroup'], table_name=table_name)
		sql_where_filters.append("""  NOT exists ( %s ) """ % (exclude_filter_query) )

	return queryset.extra(
		#tables = ['"datagroup"','"timegraph"'],
		where  = sql_where_filters,
	)

# <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
# <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
# FUNZIONI TEMPORANEE FINCHE' NON AGGIUSTIAMO IL TAGGING DI TIMEGRAPH E TIMEGRAPHSERIE
# <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
# <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

#------------------------------------------------------------------------------#
# TENTANTS                                                                     #
#------------------------------------------------------------------------------#

@property
def get_tenant_child_tags(self):

	if hasattr(self, 'filter_user'):
		user = getattr(self, 'filter_user')
	else:
		user = None

	if user:
		trees = TagTree.get_visible_trees_by_user(self, user)
	else:
		trees     = TagTree.objects.filter(tenant=self).all()

	root_tags = Tag.objects.filter(tenant=self).filter(tree_id__in=[ t.uuid for t in trees], parent__isnull=True)

	return root_tags

Tenant.add_to_class('child_tags', get_tenant_child_tags)

#------------------------------------------------------------------------------#

def set_tag_visibility(self, user=None):
	self.filter_user = user

Tenant.add_to_class('set_tag_visibility', set_tag_visibility)

#------------------------------------------------------------------------------#

def get_logs_by_user(self, user=None):

	LOGS_ELEMENTTYPES = [
		TAG_ELEMENTTYPE_NODE      ,
		TAG_ELEMENTTYPE_IFACE     ,
		TAG_ELEMENTTYPE_STORAGE   ,
		TAG_ELEMENTTYPE_SERVICE   ,
		TAG_ELEMENTTYPE_DEVICE    ,
	]

	logs = self.get_logs()

	# Gli utenti superuser non filtrano in nessun modo
	if user is None or (user is not None and user.is_superuser):
		return logs

	# Dobbiamo mostrare tutti i log in base alla visibilita' dell'utente sui tagtree
	visible_trees = TagTree.get_visible_trees_by_user(self.tenant, user)

	# Troviamo tutti gli UUID degli elementi visibili dall'utente
	elements_uuid_query = _QUERY_VISIBLE_ELEMENTS_BY_USER(user, visible_trees, elementtypes=LOGS_ELEMENTTYPES, table_name='e')

	# In base agli UUID degli elementi trovo il "path" dell'elemento 
	query_paths = """
	SELECT CASE 
	          WHEN (e.parent_id = p.uuid) THEN (p.name || ':' || e.name) 
	          ELSE e.name 
	       END 
	  FROM monitoredelement e LEFT JOIN monitoredelement p ON (e.parent_id = p.uuid AND p.kind = 1)
	 WHERE exists (%s)
	""" % (elements_uuid_query)

	# Filtro sui path
	logs = logs.extra(
		where  = [ 
			"""
			event.element_path IN ( %s )
			""" % (query_paths)
		],
	)

	return logs


Tenant.add_to_class('get_logs_by_user', get_logs_by_user)

MonitoredElement.add_to_class('get_logs_by_user', get_logs_by_user)
MonitoredNode.add_to_class('get_logs_by_user', get_logs_by_user)
MonitoredIface.add_to_class('get_logs_by_user', get_logs_by_user)
MonitoredStorage.add_to_class('get_logs_by_user', get_logs_by_user)
MonitoredService.add_to_class('get_logs_by_user', get_logs_by_user)
MonitoredDevice.add_to_class('get_logs_by_user', get_logs_by_user)

DataGroup.add_to_class('get_logs_by_user', get_logs_by_user)

Condition.add_to_class('get_logs_by_user', get_logs_by_user)

#------------------------------------------------------------------------------#

def _get_all_datagroups_by_user(self, user=None, excluded_tagtrees=None):

	elements = self.get_all_datagroups()

	if user:
		if user.is_superuser and not excluded_tagtrees:
			return elements

		elements = TAG_FILTER(self, user, elements, 'datagroup' , elementtypes=['datagroup'], excluded_tagtrees=excluded_tagtrees)

	return elements


def _get_all_datasources_by_user(self, user=None, excluded_tagtrees=None):

	elements = self.get_all_datasources()

	#
	# FIX: i datasource non sono taggati
	#
	#	if user.is_superuser and not excluded_tagtrees:
	#		return elements
	#	elements = TAG_FILTER(self, user, elements, 'datasource' , elementtypes=['datasource'], excluded_tagtrees=excluded_tagtrees)

	if user:
		if user.is_superuser and not excluded_tagtrees:
			return elements

		#
		# FIX: i timegraph non sono taggati quindi abbiamo dovuto filtare partendo dai datagroup!!! Bruttissimo
		#
		elements = TAG_FILTER_BY_DATAGROUP(self, user, elements, 'datasource' , excluded_tagtrees=excluded_tagtrees, subquery_key_field='datagroup_id')

	return elements


def _get_all_conditions_by_user(self, user=None, excluded_tagtrees=None):

	elements = self.get_all_conditions()

	if user:
		if user.is_superuser and not excluded_tagtrees:
			return elements

		elements = TAG_FILTER(self, user, elements, 'condition' , elementtypes=['condition'], excluded_tagtrees=excluded_tagtrees)

	return elements

def _get_all_timegraphs_by_user(self, user=None, excluded_tagtrees=None):

	elements = self.get_all_timegraphs()

	if user:
		if user.is_superuser and not excluded_tagtrees:
			return elements

		#
		# FIX: i timegraph non sono taggati quindi abbiamo dovuto filtare partendo dai datagroup!!! Bruttissimo
		#
		elements = TAG_FILTER_BY_DATAGROUP(self, user, elements, 'timegraph' , excluded_tagtrees=excluded_tagtrees, subquery_key_field='datagroup_id')

	return elements	

def _get_all_timegraphseries_by_user(self, user=None, excluded_tagtrees=None):

	elements = self.get_all_timegraphseries()

	if user:
		if user.is_superuser and not excluded_tagtrees:
			return elements

		#
		# FIX: i timegraph non sono taggati quindi abbiamo dovuto filtare partendo dai timegraph->datagroup!!! Bruttissimo
		#
		elements = TAG_FILTER_BY_TIMEGRAPH(self, user, elements, 'timegraphserie' , excluded_tagtrees=excluded_tagtrees)

	return elements	


Tenant.add_to_class('get_all_datagroups_by_user'     , _get_all_datagroups_by_user     )
Tenant.add_to_class('get_all_datasources_by_user'    , _get_all_datasources_by_user    )
Tenant.add_to_class('get_all_conditions_by_user'     , _get_all_conditions_by_user     )
Tenant.add_to_class('get_all_timegraphs_by_user'     , _get_all_timegraphs_by_user     )
Tenant.add_to_class('get_all_timegraphseries_by_user', _get_all_timegraphseries_by_user)

#------------------------------------------------------------------------------#

MonitoredElement.add_to_class('get_all_conditions_by_user'     , _get_all_conditions_by_user     )
MonitoredElement.add_to_class('get_all_timegraphs_by_user'     , _get_all_timegraphs_by_user     )
MonitoredElement.add_to_class('get_all_timegraphseries_by_user', _get_all_timegraphseries_by_user)

#------------------------------------------------------------------------------#
# Links                                                                        #
#------------------------------------------------------------------------------#

def get_links_by_user(self, user=None):

	links = Link.objects.filter(tenant=self)

	# Gli utenti superuser non filtrano in nessun modo
	if user is None or (user is not None and user.is_superuser):
		return links

	LOGS_ELEMENTTYPES = [
		TAG_ELEMENTTYPE_IFACE     ,
	]

	# Dobbiamo mostrare tutti i link in base alla visibilita' dell'utente sui tagtree
	visible_trees = TagTree.get_visible_trees_by_user(self.tenant, user)

	# Troviamo tutti gli UUID degli elementi visibili dall'utente
	elements_uuid_query = _QUERY_VISIBLE_ELEMENTS_BY_USER(user, visible_trees, elementtypes=LOGS_ELEMENTTYPES, table_name='le', key_field='iface_id')

	# In base agli UUID degli elementi visibili, trovo i link_element di interfacce che l'utente puo' vedere
	# e da questi trovo i link.
	query_paths = """
	SELECT 
	       le.link_id 
	  FROM 
	       link_element le
	 WHERE 
	       exists (%s)
	""" % (elements_uuid_query)

	# Filtro finale sui link
	links = links.extra(
		where  = [ 
			"""
			link.uuid IN ( %s )
			""" % (query_paths)
		],
	)

	return links

Tenant.add_to_class('get_links_by_user', get_links_by_user)

#------------------------------------------------------------------------------#
# Cluster                                                                      #
#------------------------------------------------------------------------------#

def get_nodes_visible_by_user(self, user):

	elements = self.monitoredelement_set

	if user.is_superuser:
		return elements

	elements = TAG_FILTER(self.tenant, user, elements, 'monitoredelement' )

	return elements

Cluster.add_to_class('get_nodes_visible_by_user', get_nodes_visible_by_user)

#------------------------------------------------------------------------------#
# Monitored element                                                            #
#------------------------------------------------------------------------------#

@classmethod
def get_visible_by_user(cls, tenant, user):
	
	tenant.db_set_current_schema()

	if user is None or user.is_superuser:
		return cls.objects.filter(tenant=tenant)

	elements = cls.objects

	elements = TAG_FILTER(tenant, user, elements, 'monitoredelement' )

	return elements

MonitoredElement.add_to_class('get_visible_by_user', get_visible_by_user)

#------------------------------------------------------------------------------#

def get_sub_elements_by_user(self, user):

	elements = self.get_sub_elements()

	if user.is_superuser:
		return elements

	elements = TAG_FILTER(self.tenant, user, elements, 'monitoredelement' )
	
	return elements

MonitoredElement.add_to_class('get_sub_elements_by_user', get_sub_elements_by_user)

#------------------------------------------------------------------------------#

def get_datagroups_by_user(self, user):

	elements = self.datagroups

	if user.is_superuser:
		return elements

	elements = TAG_FILTER(self.tenant, user, elements, 'datagroup' )

	return elements

MonitoredElement.add_to_class('get_datagroups_by_user', get_datagroups_by_user)

#------------------------------------------------------------------------------#
# Datagroup                                                                    #
#------------------------------------------------------------------------------#

def get_conditions_by_user(self, user):
	elements = self.conditions

	if user.is_superuser:
		return elements

	elements = TAG_FILTER(self.tenant, user, elements, 'condition' )

	return elements

DataGroup.add_to_class('get_conditions_by_user', get_conditions_by_user)

#------------------------------------------------------------------------------#
# CONDITIONS                                                                   #
#------------------------------------------------------------------------------#

def get_condition_tags(self):

	uuids = [ r[0] for r in ElementTag.objects.filter(elementid=self.uuid).values_list('tag_id') ]

	return Tag.objects.filter(uuid__in=uuids)

Condition.add_to_class('tags', get_condition_tags)

@classmethod
def get_visible_condition_by_user(cls, tenant, user):
	
	tenant.db_set_current_schema()

	if user is None or user.is_superuser:
		return cls.objects.filter(tenant=tenant)

	elements = cls.objects

	elements = TAG_FILTER(tenant, user, elements, 'condition' )

	return elements

Condition.add_to_class('get_visible_by_user', get_visible_condition_by_user)

#------------------------------------------------------------------------------#
# DATASOURCE                                                                   #
#------------------------------------------------------------------------------#

@classmethod
def get_visible_datasource_by_user(cls, tenant, user):
	
	tenant.db_set_current_schema()

	if user is None or user.is_superuser:
		return cls.objects.filter(tenant=tenant)

	elements = cls.objects

	elements = TAG_FILTER(tenant, user, elements, 'datasource' )

	return elements

DataSource.add_to_class('get_visible_by_user', get_visible_datasource_by_user)


#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
# ENTITY PERMISSIONS / VISIBILITY  based on TAGGGING                           #
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#

"""
Attraverso i tag e' gestita la visibilita' di:

	* Elementi di Monitoraggio
	* Datagroup
	* Condition

Nel "core" la verifica delle autorizzazioni avviene attraverso il modello PermissionModel.

Per gestire la visibilita' basata sui Tag e' necessario che ai modelli del core 
venga aggiunto/sovrascritto il metodo:

	<modello>._verify_permission()

"""

def _check_auth(tagged_entity, user, auth_type):

	tenant = tagged_entity.tenant

	if user is None:
		# No user specified (CLI without user?)
		return True

	if user.is_superuser:
		return True

	tenant.db_set_current_schema()

	if user.id == tagged_entity.owner_id: # check ids instead of objects in order to avoid SELECT on related objects "owner"
		return True

	tree_ids = [r[0] for r in ElementTag.objects.filter(elementid=tagged_entity.uuid).values_list('tree_id')]

	groups = user.groups.all()

	n = TagTreeGroupPermission.objects.filter(tenant=tenant, dirtree_id__in=tree_ids, group__in=groups,  auth_type=auth_type).count()

	return n > 0

MonitoredElement.add_to_class('_verify_permission', _check_auth)
DataGroup       .add_to_class('_verify_permission', _check_auth)









#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
# LOW LEVEL API FOR TAGGING                                                    #
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#

# Per certe operazioni dobbiamo taggare entita' e allo stesso tempo
# avere la struttura gerarchica dell'entita' taggata (es: nodo e tutti i suoi sottoelementi).
#
# Questa struttura dati ci serve per codificare questa informazione nelle api
#
def TAGGED_STRUCTURE(elementtag, subelements_structure=None):
	return { 
		'elementtag' : elementtag, 
		'subelements': subelements_structure or []
	}

def _bulk_save_tagged_structure_and_signal(tag, tagged_structure):
	"""
	Prende tutti gli ElementTag contenuti nella struttura "tagged_structure" (TAGGED_STRUCTURE)
	e salva in blocco (bulk_create) tutti quelli che definiscono
	un attributo "created" valorizzato a "True"
	"""

	def _extract_elementtags(tagged_structure2):

		r = []

		et = tagged_structure2['elementtag']

		if hasattr(et, 'created'):
			if et.created == True:
				r.append( et )

		for e in tagged_structure2['subelements']:
			r += _extract_elementtags(e)

		return r

	all_tagged_elements = _extract_elementtags(tagged_structure)

	#
	# SAVE
	#
	ElementTag.objects.bulk_create(all_tagged_elements)

	# The bulk_create does not trigger ElementTag.post_save signal.
	# Trigger the generic signal.
	sanet_tag_post_tagging_changed.send(sender=tag.__class__, instance=tag)

	return


def _remove_duplicated_elementtags(tag, elementtype, elementid):
	"""
	Dato un tag (e il corrispondente tree), rimuove tutto il tagging 
	(tutti gli ElemenTag) associato all'elemento <elementtype, elementid>
	che non sia dentro il tag.
	"""

	uuids_to_delete = []

	# Poiche' l'elemento non puo' comparire piu' volte nello stesso tagtree, cancelliamo (se esistono) 
	# i riferimenti (ElementTag) in  altri tag dello stesso tree.
	uuids = list(ElementTag.objects.filter(tenant=tag.tenant, tree=tag.tree, elementtype=elementtype, elementid=elementid).exclude(tag=tag).values_list('uuid', flat=True))
	while len(uuids)>0:
		uuids_to_delete += uuids
		uuids = list( ElementTag.objects.filter(tagged_parent__in=uuids).values_list('uuid', flat=True) )

	# Cancelliamo tutti gli elemnttag una volta sola.
	ElementTag.objects.filter(uuid__in=uuids_to_delete).delete()

	return

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def _tag_element(tag, element, directly_tagged=True, recursive=True, parent=None, force_retagging=False):
	"""
	Funzione chiamata da cli per il tagging ottimizzato degli elementi.

	@param tag: oggetto di classe Tag
	@param element: oggetto di classe MonitoredElement
	@param directly_tagged: booleano
	@param recursive: indica se bisogna taggare solo l'elemento o ricorsivamente tutti i sotto-elementi.
	@param parent: oggetto di tipo ElementTag

	@return: True se il tagging e' stato modificato, False altrimenti.
	"""

	tagged, tagged_structure = _tag_element_and_save_and_get_structure(tag, element, directly_tagged=directly_tagged, recursive=recursive, parent=parent, force_retagging=force_retagging)

	return tagged

def _tag_element_and_save_and_get_structure(tag, element, directly_tagged=True, recursive=True, parent=None, force_retagging=False):

	tagged, tagged_structure = _tag_element_and_get_structure(tag, element, directly_tagged=directly_tagged, recursive=recursive, parent=parent, force_retagging=force_retagging)

	_bulk_save_tagged_structure_and_signal(tag, tagged_structure)

	return tagged, tagged_structure


def _tag_element_and_get_structure(tag, element, directly_tagged=True, recursive=True, parent=None, force_retagging=False):
	"""
	Restituisce:
	
		tagged			: True se le informazioni di tagging dell'elemento o del sotto-elemento sono state modificate
		TAGGED_STRUCTURE        : struttura ad albero degli elementi taggati
	
	"""

	if DEBUG: print "_tag_element", tag, element, directly_tagged, recursive, parent

	unique_params = {}
	if tag.tree.uniques:
		unique_params = { 'unique_sentinel' : ElementTag.TAGGED_UNIQUE }

	#
	# TAG ENTITY
	#

	tagged = False

	elementtype = ELEMENT_KIND_TO_TAG_ELEMENTTYPE[element.kind]
	elementid   = element.uuid

	if tag.tree.uniques:
		# Poiche' d'elemento non puo' comparire piu' volte nello stesso tagtree, cancelliamo se esistono, riferimenti in altri tag dello stesso tree
		_remove_duplicated_elementtags(tag, elementtype, elementid)

	try:
		et = ElementTag.objects.get(tenant=tag.tenant, tag=tag, tree=tag.tree, elementtype=elementtype, elementid=elementid, **unique_params)
		et.created = False
		created = False
	except ElementTag.DoesNotExist, ex:
		et = ElementTag(tenant=tag.tenant, tag=tag, tree=tag.tree, elementtype=elementtype, elementid=elementid, **unique_params)
		et.created = True
		created = True

	fields_to_update = {}

	if parent:
		if et.tagged_parent != parent.uuid:
			et.tagged_parent = parent.uuid

			#et.save()
			if not created:
				fields_to_update['tagged_parent'] = parent.uuid

	if et.directly_tagged != directly_tagged:

		et.directly_tagged = directly_tagged

		#et.save()
		if not created:

			fields_to_update['directly_tagged'] = directly_tagged

	if fields_to_update:
		ElementTag.objects.filter(tenant=tag.tenant, tree=tag.tree, tag=tag, elementtype=elementtype, elementid=elementid).update(**fields_to_update)

	#
	# TAG SUB ELEMENTS / DATAGROUPS
	#

	subelements = []

	if created or force_retagging:
		tagged = True

		if recursive:
			for dg in element.datagroups.all():
				tagged, tagged_structure = _tag_datagroup_and_get_structure(tag, dg, directly_tagged=False, recursive=True, parent=et, force_retagging=force_retagging)

				subelements.append(tagged_structure)

			for sub_element in element.get_sub_elements().all():
				tagged, tagged_structure = _tag_element_and_get_structure(tag, sub_element, directly_tagged=False, recursive=True, parent=et, force_retagging=force_retagging)

				subelements.append(tagged_structure)

	#
	# Save
	#
	tagged_structure = TAGGED_STRUCTURE(et, subelements)

	return tagged, tagged_structure



#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def _tag_datagroup(tag, datagroup, directly_tagged=True, recursive=True, parent=None, force_retagging=False):

	tagged, tagged_structure = _tag_datagroup_and_get_structure(tag, datagroup, directly_tagged=directly_tagged, recursive=recursive, parent=parent, force_retagging=force_retagging)

	_bulk_save_tagged_structure_and_signal(tag, tagged_structure)

	return tagged

def _tag_datagroup_and_get_structure(tag, datagroup, directly_tagged=True, recursive=True, parent=None, force_retagging=False):
	"""
	Restituisce:
	
		tagged			: True se le informazioni di tagging dell'elemento o del sotto-elemento sono state modificate
		TAGGED_STRUCTURE        : struttura ad albero degli elementi taggati
	
	"""

	if DEBUG: print "_tag_datagroup", tag, datagroup, directly_tagged, recursive, parent

	unique_params = {}
	if tag.tree.uniques:
		unique_params = { 'unique_sentinel' : ElementTag.TAGGED_UNIQUE }

	#
	# TAG ENTITY
	#

	tagged = False

	elementtype = TAG_ELEMENTTYPE_DATAGROUP
	elementid   = datagroup.uuid

	if tag.tree.uniques:
		_remove_duplicated_elementtags(tag, elementtype, elementid)

	try:
		et = ElementTag.objects.get(tenant=tag.tenant, tag=tag, tree=tag.tree, elementtype=elementtype, elementid=elementid, **unique_params)
		et.created = False
		created = False
	except ElementTag.DoesNotExist, ex:
		et = ElementTag(tenant=tag.tenant, tag=tag, tree=tag.tree, elementtype=elementtype, elementid=elementid, **unique_params)
		et.created = True
		created = True

	fields_to_update = {}

	if parent:
		if et.tagged_parent != parent.uuid:
			et.tagged_parent = parent.uuid

			#et.save()
			if not created:
				fields_to_update['tagged_parent'] = parent.uuid

	if et.directly_tagged != directly_tagged:
		et.directly_tagged = directly_tagged

		#et.save()
		if not created:
			fields_to_update['directly_tagged'] = directly_tagged

	if fields_to_update:
		ElementTag.objects.filter(tenant=tag.tenant, tree=tag.tree, tag=tag, elementtype=elementtype, elementid=elementid).update(**fields_to_update)

	#
	# TAG SUB ELEMENTS / DATAGROUPS
	#
	
	subelements = []

	if created or force_retagging:
		tagged = True

		if recursive:
			for c in datagroup.conditions.all():
				tagged, tagged_structure = _tag_condition_and_get_structure(tag, c, directly_tagged=False, parent=et)

				subelements.append(tagged_structure)

	#
	# Save
	#
	tagged_structure = TAGGED_STRUCTURE(et, subelements)
	
	return tagged, tagged_structure

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def _tag_condition(tag, condition, directly_tagged=True, recursive=True, parent=None, force_retagging=False):

	tagged, tagged_structure = _tag_condition_and_save_and_get_structure(tag, condition, directly_tagged=directly_tagged, recursive=recursive, parent=parent, force_retagging=force_retagging)
	
	return tagged

def _tag_condition_and_save_and_get_structure(tag, condition, directly_tagged=True, recursive=True, parent=None, force_retagging=False):
	
	tagged, tagged_structure = _tag_condition_and_get_structure(tag, condition, directly_tagged=directly_tagged, recursive=recursive, parent=parent, force_retagging=force_retagging)
	
	_bulk_save_tagged_structure_and_signal(tag, tagged_structure)
	
	return tagged, tagged_structure

def _tag_condition_and_get_structure(tag, condition, directly_tagged=True, recursive=True,  parent=None, force_retagging=False):
	"""
	Restituisce:

		tagged                  : True se le informazioni di tagging della condition sono state aggiornate
		TAGGED_STRUCTURE        : struttura ad albero degli elementi taggati

	"""

	if DEBUG: print "_tag_condition", tag, condition, directly_tagged, recursive, parent

	tagged = False

	elementtype = TAG_ELEMENTTYPE_CONDITION
	elementid   = condition.uuid

	if tag.tree.uniques:
		_remove_duplicated_elementtags(tag, elementtype, elementid)
	
	unique_params = {}
	if tag.tree.uniques:
		unique_params = { 'unique_sentinel' : ElementTag.TAGGED_UNIQUE }
	
	#
	# TAG ENTITY
	#
	
	try:
		et = ElementTag.objects.get(tenant=tag.tenant, tag=tag, tree=tag.tree, elementtype=elementtype, elementid=elementid, **unique_params)
		et.created = False
		created = False
	except ElementTag.DoesNotExist, ex:
		et = ElementTag(tenant=tag.tenant, tag=tag, tree=tag.tree, elementtype=elementtype, elementid=elementid, **unique_params)
		et.created = True
		created = True
	
	fields_to_update = {}
	
	if parent:
		if et.tagged_parent != parent.uuid:
			et.tagged_parent = parent.uuid
			
			#et.save()
			if not created:
				fields_to_update['tagged_parent'] = parent.uuid

	if et.directly_tagged != directly_tagged:
		et.directly_tagged = directly_tagged

		#et.save()
		if not created:
			fields_to_update['directly_tagged'] = directly_tagged

	if fields_to_update:
		ElementTag.objects.filter(tenant=tag.tenant, tree=tag.tree, tag=tag, elementtype=elementtype, elementid=elementid).update(**fields_to_update)

	if created:
		tagged = True

	#
	# Save
	#
	tagged_structure = TAGGED_STRUCTURE(et)

	return tagged, tagged_structure

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def _untag_element(tag, element, recursive=True):

	if DEBUG: print "_untag_element", tag, element, recursive

	elementtype = ELEMENT_KIND_TO_TAG_ELEMENTTYPE[element.kind]
	elementid   = element.uuid

	try:

		qs = ElementTag.objects.filter(tenant=tag.tenant, tag=tag, tree=tag.tree, elementtype=elementtype, elementid=elementid )

		if recursive:
			et_uuid = ElementTag.objects.filter(tenant=tag.tenant, tag=tag, tree=tag.tree, elementtype=elementtype, elementid=elementid).values_list('uuid', flat=True)

			# All sub elements
			qs = qs | ElementTag.objects.filter(tagged_parent__in=et_uuid)
			elem_uuids = qs.values_list('uuid', flat=True)

			# All datagroups of the element and subelements
			dg_qs = ElementTag.objects.filter(tagged_parent__in=elem_uuids)
			dg_uuids = dg_qs.values_list('uuid', flat=True)

			# All conditions of all datagroups
			con_qs = ElementTag.objects.filter(tagged_parent__in=dg_uuids)

			qs = qs | dg_qs | con_qs

		qs.delete()

		return True

	except ElementTag.DoesNotExist, e:
		return False

#------------------------------------------------------------------------------#

def _untag_datagroup(tag, datagroup, recursive=True):

	if DEBUG: print "_untag_datagroup", tag, datagroup, recursive

	elementtype = TAG_ELEMENTTYPE_DATAGROUP
	elementid   = datagroup.uuid

	try:
		qs = ElementTag.objects.filter(tenant=tag.tenant, tag=tag, elementtype=elementtype, elementid=elementid)

		if recursive:
			et_uuid = ElementTag.objects.filter(tenant=tag.tenant, tag=tag, elementtype=elementtype, elementid=elementid).values_list('uuid', flat=True)[0]

			qs = qs | ElementTag.objects.filter(tagged_parent=et_uuid)

		qs.delete()

		return True

	except ElementTag.DoesNotExist, e:
		return False

#------------------------------------------------------------------------------#

def _untag_condition(tag, condition, recursive=True):

	if DEBUG: print "_untag_condition", tag, condition, recursive

	elementtype = TAG_ELEMENTTYPE_CONDITION
	elementid   = condition.uuid

	try:
		et = ElementTag.objects.get(tenant=tag.tenant, tag=tag, elementtype=elementtype, elementid=elementid)
		et.delete()

		return True

	except ElementTag.DoesNotExist, e:
		return False

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

#
# Queste funzioni sono delle utility per taggare entita' senza preoccuparci
# di che tipo sono. 
#
# Vanno usate solo in casi molto particolari!!!!!
#

def _tag_raw_entity(tag, entity, directly_tagged, recursive):

	tag_functions = {
		MonitoredElement : _tag_element,
		MonitoredNode    : _tag_element,
		MonitoredIface   : _tag_element,
		MonitoredStorage : _tag_element,
		MonitoredService : _tag_element,
		MonitoredDevice  : _tag_element,
		DataGroup        : _tag_datagroup,
		Condition        : _tag_condition,
	}

	fun = tag_functions[entity.__class__]

	return fun(tag, entity, directly_tagged=directly_tagged, recursive=recursive)

def _untag_raw_entity(tag, entity, recursive):

	untag_functions = {
		MonitoredElement : _untag_element,
		MonitoredNode    : _untag_element,
		MonitoredIface   : _untag_element,
		MonitoredStorage : _untag_element,
		MonitoredService : _untag_element,
		MonitoredDevice  : _untag_element,
		DataGroup        : _untag_datagroup,
		Condition        : _untag_condition,
	}

	fun = untag_functions[entity.__class__]

	return fun(tag, entity, recursive)

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def _tagtree_recalculate_tagged_elements(tree, nodes_to_tag=None, tagged_callback=None, tagged_callback_context=None):
	"""
	Questa funzione serve per taggare automaticamente i nodi di un tree con flag 'allnodes=True'.
	
	Se non viene passato in input un elenco di nodi da taggare ("nodes_to_tag"), il sistema
	cerca di applicare le regole di tagging automatico su tutti i nodi del sistema (anche
	quelli gia' taggati).
	
	La funzioen "tagged_callback", se specificata viene chiamata ogni volta che un nodo
	viene taggato. La funzione ha la seguente signature:
	
		def tagged_callback(context, tree, tag, elementtag, node):
			...
			...
			return
			
	Dove "tree" e "tag" indicano in quale tree/tag il nodo "node" e' stato taggato ("elementtag").
	Il parametro "context" contiene il riferimento ai dati passati dal parametro "tagged_callback_context".
	
	IMPORTANTE: QUESTA FUNZIONE NON NOTIFICA IN NESSUN IL SISTEMA TAGSD DELLA MODIFICA DEL TAGGING.
	
	"""

	if not tree.allnodes:
		return 0

	root_tag = tree.root
	if root_tag is None:
		return 0

	# Se non hanno specificato un insieme di nodi da taggare nell'albero
	# allora prendiamo tutti i nodi che ancora non compaiono taggati nel "tree".
	if nodes_to_tag is None:
		nodes_to_tag = MonitoredNode.objects.filter(tenant=tree.tenant)

		#
		# TODO: sarebbe meglio prevedere ANCHE una modalita' per non "ritaggare" ricorsivamente i nodi 
		#       che sono gia' assegnati all'albero. Si potrebbe fare cosi:
		#
		#       tagged_nodes_uuid = [row[0] for row in ElementTag.objects.filter(tree=tree, elementtype=TAG_ELEMENTTYPE_NODE).values_list('elementid')]
		#
		#       nodes_to_tag = nodes_to_tag.exclude(uuid__in=tagged_nodes_uuid)

		nodes_to_tag = nodes_to_tag.order_by('name').all()

	# 
	# Troviamo le regole di inserimento definite per il tree
	#
	root_has_rule = root_tag.tagrules.count() > 0
	insert_rules  = list(TagEntryInsertRule.objects.filter(tagentry__tree=tree).all())

	count = 0

	for node in nodes_to_tag:

		# Verifichiamo in quale tag deve essere inserito automaticamente il nodo
		target_tag = None
		for rule in insert_rules:
			if rule.match(node):
				target_tag = rule.tagentry
				break

		# Nessun tag dell'albero da usare, vediamo se usare la radice.

		#print __file__, "_tagtree_recalculate_tagged_elements", root_tag, root_has_rule, target_tag

		# Non ho trovato nessun tag in base alle regole di inserimento?
		if not target_tag:
			# Se la radice non ha regole, tagghiamo il nodo nella radice
			if not root_has_rule:
				target_tag = root_tag
			else:
				# Anche la radice ha una regola, ma non ha fatto match quindi
				# questo nodo non si tagga.
				continue

		#
		# Tagghiamo
		#
		tagged, tagged_structure = _tag_element_and_get_structure(target_tag, node)
		
		_bulk_save_tagged_structure_and_signal(target_tag, tagged_structure)

		#
		#
		#
		if tagged:
			count += 1

			if tagged_callback:
				tagged_callback(tagged_callback_context, tree, target_tag, tagged_structure['elementtag'], node )

	return count

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

#
# ATTENTION: these signals can be triggered inside transactional code. 
#            DO NOT USE TRANSACTIONAL CODE INSIDE THESE SIGNAL HANDLER 
#
#

@receiver(post_save, sender=PriorityLabel)
def _priority_label_update_cache(sender, **kwargs):
	"""
	E' stato modificato un livello di priorita'. 
	Non sappiamo se hanno cambiato il range del livello o solo una descrizione.
	Poiche' i livelli di priorita' influenzano il calcolo degli allarmi
	critici, colorazione, ecc.. e' meglio invalidare la cache
	"""
	priority_label  = kwargs['instance']
	created         = kwargs['created']

	if not created:
		return

	#print __file__, "_priority_label_update_cache"

	#
	# TAG STATE CASE: 11
	#
	signal_state_refresh(priority_label.tenant.uuid)

	return

@receiver(post_delete, sender=PriorityLabel)
def _priority_label_on_delete(sender, **kwargs):

	priority_label = kwargs['instance']

	#
	# TAG STATE CASE: 12
	#
	signal_state_refresh(priority_label.tenant.uuid)

	return

#------------------------------------------------------------------------------#

@receiver(post_save, sender=TenantSetting)
def _tenantsetting_changed(sender, **kwargs):
	"""
	Quando viene aggiornato il CRITICAL_LEVEL o MINIMUM_LEVEL del tenant li stati dei tagtree (e tag) devono essere ricalcolati.
	"""
	tenant_setting  = kwargs['instance']

	if tenant_setting.name not in [ TenantSetting.SETTING_CRITICAL_LEVEL, TenantSetting.SETTING_MINIMUM_LEVEL ]:
		return

	# NOTA: questa funzione non solleva eccezioni quindi possiamo eseguirla 
	#       senza preoccuparci che un eventuale eccezione sollevata qui annulli
	#       una transazione in corso.
	signal_state_refresh(tenant_setting.tenant.uuid)

	return

#------------------------------------------------------------------------------#
# Automatic tagging inside node or elements
#------------------------------------------------------------------------------#

@receiver(post_save, sender=MonitoredIface)
@receiver(post_save, sender=MonitoredStorage)
@receiver(post_save, sender=MonitoredService)
@receiver(post_save, sender=MonitoredDevice)
def update_tagged_element(sender, **kwargs):

	element = kwargs['instance']
	created = kwargs['created']

	# Do automatic tag only when the elementi is created
	if not created:
		return

	main_element = element.parent

	# Find where the main elementi si tagged
	tagged_elements = ElementTag.objects.filter( elementtype=ELEMENT_KIND_TO_TAG_ELEMENTTYPE[ main_element.kind ], elementid=main_element.uuid).select_related('tag')
	for et in tagged_elements:
		# Tag this element:
		# - where the main element is tagged  -> parent = et
		# - as child tagged element           -> directly_tagged = False
		_tag_element(et.tag, element, directly_tagged=False, parent=et)

	return

@receiver(post_save, sender=DataGroup)
def update_tagged_datagroup(sender, **kwargs):

	datagroup  = kwargs['instance']
	created    = kwargs['created']
	
	if not created:
		return

	main_element = datagroup.element

	if DEBUG: print "update_tagged_datagroup", main_element
	
	# Find where the main elementi si tagged
	tagged_elements = ElementTag.objects.filter( elementtype=ELEMENT_KIND_TO_TAG_ELEMENTTYPE[ main_element.kind ], elementid=main_element.uuid).select_related('tag')
	for et in tagged_elements:
		# Tag this element:
		# - where the main element is tagged  -> parent = et
		# - as child tagged element           -> directly_tagged = False
		_tag_datagroup(et.tag, datagroup, directly_tagged=False, parent=et)

	return

@receiver(post_save, sender=Condition)
def _update_tagged_condition_save(sender, **kwargs):

	condition  = kwargs['instance']
	created    = kwargs['created']

	# Do automatic tag only when the elementi is created
	if not created:
		return

	main_element = condition.datagroup

	# Find where the main elementi si tagged
	tagged_elements = ElementTag.objects.filter( elementtype=TAG_ELEMENTTYPE_DATAGROUP, elementid=main_element.uuid).select_related('tag')

	for et in tagged_elements:
		# Tag this element:
		# - where the main element is tagged  -> parent = et
		# - as child tagged element           -> directly_tagged = False
		_tag_condition(et.tag, condition, directly_tagged=False, parent=et)

	return

#------------------------------------------------------------------------------#
# 
# Automatic notification
#

#
# Evento di UPDATE custom di Sanet.
#
# FIX: questo sistema e' inefficiente, bisogna trovare un modo di:
#  * ridurre il numero di notifiche
#  * ridurre il numero di "connessioni" aperte per effettuare queste operazioni
#
from tags.server.tagsd.api import signal_condition_changed

@receiver(sanet_model_update, sender=Condition)
def _update_tagged_condition_update(sender, **kwargs):

	condition      = kwargs['instance']
	updated_fields = kwargs['updated_fields']

	#
	# Dobbiamo notificare update di una condition solo quando cambia:
	#   * la priority
	#   * lo stato
	#
	if 'priority' in updated_fields or 'status' in updated_fields:
		signal_condition_changed(condition.tenant_id, condition.uuid)

	return 

#------------------------------------------------------------------------------#

@receiver(post_delete, sender=MonitoredNode)
@receiver(post_delete, sender=MonitoredService)
@receiver(post_delete, sender=MonitoredIface)
@receiver(post_delete, sender=MonitoredStorage)
@receiver(post_delete, sender=MonitoredDevice)
@receiver(post_delete, sender=DataGroup)
@receiver(post_delete, sender=Condition)
@receiver(post_delete, sender=TimeGraph)
def _tags_post_delete_cleanup_post(sender, **kwargs):

	o = kwargs['instance']

	if DEBUG: print "DEBUG: tags_post_delete_cleanup", sender.__class__, o.uuid

	try:
		ElementTag.objects.filter(elementid=o.uuid).delete()
	except ElementTag.DoesNotExist, e:
		pass

	return

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

@receiver(post_save, sender=ElementTag)
def _elementtag_post_save(sender, **kwargs):

	o = kwargs['instance']

	#tree = TagTree.objects.get(uuid=o.tree_id)
	#tree.update_state_time()

	return

#------------------------------------------------------------------------------#
# 
# Cancellazione di elementi taggati
#
# Sta cambiando la struttura del tagging. Dobbiamo notificare il sistema.

@receiver(post_delete, sender=ElementTag)
def _elementtag_post_delete_post(sender, **kwargs):

	elementtag = kwargs['instance']

	#
	# IMPORTANTE: i campi tenant_id e tree_id dovrebbero essere ancora valorizzati
	#             anche correttamente dopo la .delete() quindi
	#             possiamo usarli per notificare.
	#

	try:
		signal_state_refresh( elementtag.tenant_id , [ elementtag.tree_id ] , force=True)
	except TagTree.DoesNotExist, e:
		pass

	return

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

@receiver(post_save, sender=MonitoredNode)
def add_new_node_to_trees_with_all_nodes(sender, **kwargs):
	#
	# Automatic add nodes to trees with "all-nodes"
	#
	node    = kwargs['instance']
	created = kwargs['created']
	raw     = kwargs['raw']

	tenant = node.tenant

	# not use the object when in this state. See django doc.
	if raw: 
		return

	# Automatic tagging is valid for new nodes
	if not created:
		return 

	trees = TagTree.objects.filter(tenant=tenant, allnodes=True).all()

	tagged_count = 0

	for tree in trees:
		tagged_count += _tagtree_recalculate_tagged_elements(tree, nodes_to_tag=[node])

	if tagged_count > 0:
		signal_state_refresh(tenant.uuid, [tree.uuid for tree in trees], force=True)

	return


#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
# SANET CORE SIGNALS                                                           #
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#
#------------------------------------------------------------------------------#

@receiver(sanet_init_subapp)
def _sanet_init_subapp_tags(sender, **kwargs):

	logger = logging.getLogger('sanet.global')

	logger.info( "[tags] Init"  )

	return

@receiver(sanet_check_tenant)
def _sanet_check_tenant_tags(sender, **kwargs):

	logger = logging.getLogger('sanet.global')

	t = kwargs['instance']
	logger.info( "[tags] Check tenant %s" % (t)  )

	return

@receiver(sanet_cleanup_tenant)
def _sanet_cleanup_tenant_tags(sender, **kwargs):

	logger = logging.getLogger('sanet.global')

	t = kwargs['instance']
	logger.info( "[tags] Cleanup tenant %s" % (t)  )

	#
	# Remove old invalid data
	#
	from django.core.management import call_command
	call_command('tag_clean_dead_elementtag'    , tenant=t.name, log_level=logger.getEffectiveLevel(), interactive=False, verbosity=True, show_traceback=True)

	return

@receiver(sanet_post_check)
def _sanet_post_check_tags(sender, **kwargs):

	logger = logging.getLogger('sanet.global')

	logger.info( "[tags] Post check" )

	return

@receiver(sanet_tenant_create)
def _sanet_tenant_create_tags(sender, **kwargs):

	tenant = kwargs['instance']

	_check_tags(tenant)

	return

def _check_tags(tenant):

	from tags.api import tagtree_get, tag_create

	tenant.db_set_current_schema()

	if TagTree.objects.count() > 0:
		return

	tree = tagtree_get(tenant, 'geo', create=True)

	name = 'Global'

	tag  = tag_create(tree, 'Global', parent=None)

	return

#------------------------------------------------------------------------------#
# SANETD SERVER SIGNALS                                                        #
#------------------------------------------------------------------------------#

@receiver(sanetd_process_changed_conditions)
def _tagsd_sanetd_process_changed_conditions(sender, **kwargs):

	logger             = kwargs['log']
	tenant             = kwargs['tenant']
	changed_conditions = kwargs['changed_conditions']

	condition_ids = [c.uuid for c in changed_conditions]

	signal_conditions_changed(tenant.uuid, condition_ids)

	return 

"""

Gestione delle wildcard
============================

Il server centrale emette il segnale:

   sanetd_alarm_msg_create_template_symbols
   
per chiedere di creare i simboli da espandere nelle wildcard dei messaggi {{ }}}.

Le wildcard per gestire le informazioni dei tag iniziano con la chiave:

	{{ tags.<...> }}
"""

@receiver(sanetd_alarm_msg_create_template_symbols)
def _tags_sanetd_alarm_msg_create_template_symbols(sender, **kwargs):

	log              = kwargs['log']
	tenant           = kwargs['tenant']
	node             = kwargs['node']
	element          = kwargs['element']
	condition        = kwargs['condition']
	template_symbols = kwargs['template_symbols']

	log.log(9, "TAGS template symbols")
	log.log(9, "     log              = %s", log )
	log.log(9, "     tenant           = %s", tenant )
	log.log(9, "     node             = %s", node )
	log.log(9, "     element          = %s", element )
	log.log(9, "     condition        = %s", condition )
	log.log(9, "     template_symbols = %s", template_symbols )

	#
	# Setup tags wildcards
	#

	wildcards = {
		'tags': {
			# {{ tags.tree.<tree name> }}
			'tree': TemplateTagTreeExpander(condition)
		}
	}

	#
	# Update wildcars register
	#
	template_symbols.update(wildcards)

	return


class TemplateTagTreeExpander(object):
	"""
	Oggetto "semi-dizionario" che espande in una stringa tutti i tag 
	in cui e':
		* taggata la condition
		* e' taggato l'elemento associato alla condition
	
	Questo oggetto serve per gestire le wildcard:
	
		{{ tags.tree.<tree name> }}
	"""

	def __init__(self, condition):
		self.condition = condition

	def __getitem__(self, i):

		tagtree_name = i

		self.condition.tenant.db_set_current_schema()

		q = ElementTag.objects.filter(tree__name=tagtree_name)

		q = q.filter(Q(elementid=self.condition.element_id) | Q(elementid=self.condition.uuid))

		paths = set()
		for et in q.all():
			paths.add( et.tag.path )

		return ','.join(list(paths))


#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

@receiver(statusd_handle_request)
def _statusd_handle_request_tags(sender, **kwargs):
	
	server  = kwargs['server']
	command = kwargs['command']
	params  = kwargs['params']

	logger = server.logger
	
	logger.info("webtests: processing request: %s", command)


	if  command == 'get_tag_status':

		from tags.server.tagsd.api import get_tags_state_safe

		#
		# CHECK PARAMS
		#
		tenant_id = params['tenant_id']
		tid       = params['id']

		t = _get_tag(tenant_id, tid)

		#
		# GET STATE
		#
		state_by_uuid = get_tags_state_safe(tenant_id, [t.uuid])

		if state_by_uuid is None:
			raise Exception("Unable to get status for tag: %s. Sanet server not working properly." % (tid))

		state_data = state_by_uuid.get(t.uuid, None)
		if state_data is None:
			raise Exception("Status for tag %s not calculated yet" % (tid))

		status = state_data['visible_status']

		#
		# RESPONSE
		#
		response = { 'data': status }
		
		return response
		
	else:
		return None

def _get_tag(tenant_id, rid):

	tenant = Tenant.objects.get(uuid=tenant_id)
	tenant.db_set_current_schema()

	t = None
	try:
		t = Tag.objects.get(uuid=rid)
	except Tag.DoesNotExist, e:
		pass

	try:
		t = t or Tag.objects.get(path=rid)
	except Tag.DoesNotExist, e:
		pass

	if t == None:
		raise Exception("Tag with id '%s' not found" % (rid))

	return t
















