# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

Tagsd - API di comunicazione con
====================================================================================

.. contents:: Contenuti


Introduzione
------------------

Questo file contiene le API per comunicare via rete con il process TAGSD.

Le richieste non sono altro che strutture dati inviate al server in formato JSON.

Le risposte sono strutture dati codificate in formato Pickle.

Via rete, sia richieste che risposte vengono incapsulate in una struttura dati
binaria composta dai seguenti campi

::
	
	CAMPO    LUNGHEZZA    DESCRIZIONE
	-----    ---------    ------------
	
	len      2            Lunghezza dal campo "data" (in network order)
	data     *            Stringa di byte a lunghezza variabile


Schema di comunicazione
------------------------

::
    
           CLIENT                                      TAGSD
                                                       SERVER
           
          request
             |
             v
          header 
             + 
      encoded_request 
           (JSON)        
             |
             v
           SEND ----------------------------------->  RECEIVE
                                                         |
                                                         v
                                                      response
                                                         |
                                                         v
                                                       header
                                                         +
                                                  encoded response
                                                     (PICKLED)
                                                         |
                                                         V
                <-----------------------------------    SEND
                                                      
                                                   
"""

import os
import sys
import time
import logging
import socket
import threading
import struct
import traceback

import socket

import json
import cPickle 

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

DEBUG = os.environ.get('DEBUG_TAGSD_API', False)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

TAGSD_SERVER_IP   = '127.0.0.1'

TAGSD_SERVER_PORT = 48740           # ASSIGNED WITH: abs(binascii.crc32('tagsd') % 0x00000fff)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class TagStateSocket(object):
	"""
	Classe astratta per gestire la comunicazione tra client e TAGSD
	tramite socket.
	
	I metodi di questa classe sono sia per socket TCP che UDP.
	"""

	def __init__(self, *args, **kwargs):
		super(TagStateSocket, self).__init__(*args, **kwargs)

		self._socket = None

		self.init_socket()

	def init_socket(self):
		pass

	def __getattr__(self, name):
		"""
		Questo serve per poter accedere a metodi di self._socket  direttamente:
		
			self._socket.settimeout(10) ->  self.settimeout(10)
		"""
		return getattr(self._socket, name)

	#
	# ENCODING DATI DELLE RICHIESTE
	#
	@staticmethod
	def encode_request(request):
		return json.dumps(request)

	@staticmethod
	def decode_request(request):
		return json.loads(request)

	#
	# ENCODING DATI DELLE RISPOSTE
	#
	@staticmethod
	def encode_response(data):
		
		# IMPORTANT: il protocollo "pickle" e' schizzinoso nel codificare strutture
		#            dati se usano il __slot__.
		#            Per evitare problemi di codifica/decodifica bisogna utilizzare
		#            il protocollo piu' nuovo.
		return cPickle.dumps(data, cPickle.HIGHEST_PROTOCOL)

	@staticmethod
	def decode_response(data):
		return cPickle.loads(data)


	#
	# METODI DI COMODO PER INVIARE RICHIESTE E RISPOSTE VIA SOCKET TCP
	#
	@staticmethod
	def send_request_socket(socket, request):
		request = TagStateSocket.encode_request(request)

		buff = struct.pack('!i', len(request)) + request

		socket.send(buff)

	@staticmethod
	def read_request_socket(socket):

		data = socket.recv(4)
		if not data:
			return None
		data_len = struct.unpack('!i', data)[0]

		request = ''

		while data_len > 0:
			data = socket.recv(data_len)
			if data is None:
				break
			request += data
			data_len -= len(data)

		return TagStateSocket.decode_request(request)

	@staticmethod
	def send_response_socket(socket, data):
		data = TagStateSocket.encode_response(data)

		buff = struct.pack('!i', len(data)) + data

		return socket.send( buff )

	@staticmethod
	def read_response_socket(socket):

		header = socket.recv(4)
		if not header:
			raise Exception("Connection closed before response")
		data_len = struct.unpack('!i', header)[0]

		response = ''
		while data_len > 0:
			data = socket.recv(data_len)
			if data is None:
				break
			response += data
			data_len -= len(data)

		if len(response) == 0:
			raise Exception("Connection closed before response")

		return TagStateSocket.decode_response(response)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class ClientTagStateSocket(TagStateSocket):
	"""
	Classe da usare nel CLIENT per comunicare con TAGSD in TCP.
	"""

	CMD_SERVER_STATUS     = 'server_status'

	CMD_TREES_STATE       = 'trees_state'
	CMD_TAGS_STATE        = 'tags_state'

	CMD_REFRESH_ALL       = 'refresh_all'
	CMD_REFRESH_ALL_FORCE = 'refresh_all_force'

	CMD_RELOAD_CONDITIONS = 'reload_conditions'

	CMD_TREE_STRUCTURE    = 'tree_structure'

	def init_socket(self):
		self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_LINGER, struct.pack('ii', 1, 10))

	def connect(self):
		self._socket.connect((TAGSD_SERVER_IP, TAGSD_SERVER_PORT))

	#
	# Alcuni comandi prevedono una risposta, altri no
	#
	def _send_and_receive(self, cmd):

		TagStateSocket.send_request_socket(self._socket, cmd )

		response = TagStateSocket.read_response_socket(self._socket)

		return response

	def _send_only(self, cmd):

		TagStateSocket.send_request_socket(self._socket, cmd )

		return None

	#
	# Comandi che prevedono risposta
	#
	def get_server_status(self):
		cmd = {
			"command" : ClientTagStateSocket.CMD_SERVER_STATUS
		}
		return self._send_and_receive(cmd)

	def get_trees_state(self, tenantid, tree_ids):
		cmd = {
			"command" : ClientTagStateSocket.CMD_TREES_STATE,
			"tenantid": tenantid,
			"uuids"   : tree_ids
		}
		return self._send_and_receive(cmd)

	def get_tags_state(self, tenantid, tag_ids):
		cmd = {
			"command" : ClientTagStateSocket.CMD_TAGS_STATE,
			"tenantid": tenantid,
			"uuids"   : tag_ids
		}
		return self._send_and_receive(cmd)


	def get_tree_structure(self, tenantid, tree_id):
		cmd = {
			"command" : ClientTagStateSocket.CMD_TREE_STRUCTURE,
			"tenantid": tenantid,
			"uuid"    : tree_id
		}
		return self._send_and_receive(cmd)

	#
	# Comandi unidirezionali
	#
	def refresh_all(self, tenantid=None, tree_ids=None):
		cmd = {
			"command" : ClientTagStateSocket.CMD_REFRESH_ALL,
			"tenantid": tenantid,
			"tree_ids": tree_ids,
		}
		return self._send_only(cmd)

	def refresh_all_force(self, tenantid=None, tree_ids=None):
		cmd = {
			"command" : ClientTagStateSocket.CMD_REFRESH_ALL_FORCE,
			"tenantid": tenantid,
			"tree_ids": tree_ids,
		}
		return self._send_only(cmd)

	def reload_conditions(self, tenantid=None, conditions_ids=None):
		cmd = {
			"command"       : ClientTagStateSocket.CMD_RELOAD_CONDITIONS,
			"tenantid"      : tenantid,
			"conditions_ids": conditions_ids,
		}
		return self._send_only(cmd)

class ClientTagStateSocketUDP(ClientTagStateSocket):
	"""
	Classe da usare nel CLIENT per comunicare con TAGSD in UDP
	"""

	def init_socket(self):
		self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
		
		self._socket.settimeout(1)

	def _send_and_receive(self, cmd):

		# SEND:  header (4) + encoded request

		request = TagStateSocket.encode_request(cmd)

		buff = struct.pack('!i', len(request)) + request

		self._socket.sendto( buff, (TAGSD_SERVER_IP, TAGSD_SERVER_PORT) )

		# RECEIVE: header (4) + encoded body
		buff, address = self._socket.recvfrom(65535)

		data = buff[4:]

		response = TagStateSocket.decode_response(data)

		return response


	def _send_only(self, cmd):

		# SEND:  header (4) + encoded request

		request = TagStateSocket.encode_request(cmd)

		buff = struct.pack('!i', len(request)) + request

		self._socket.sendto( buff, (TAGSD_SERVER_IP, TAGSD_SERVER_PORT) )

		return None

#-----------------------------------------------------------------------------#
# API DI COMODO                                                               #
#-----------------------------------------------------------------------------#

def get_tagsd_server_status(timeout=2):

	try:
		# Lo stato del server lo chiediamo via UDP per fare prima (non sono tanti dati)
		client = ClientTagStateSocketUDP()

		client.settimeout(timeout)

		return client.get_server_status()

	except Exception, e:
		if DEBUG: print __file__, type(e), str(e), traceback.format_exc()
		return None
	finally:
		try:
			client.close()
		except Exception, e:
			pass

#-----------------------------------------------------------------------------#
# API DI NOTIFICA AL SERVER TAGSD                                             #
#-----------------------------------------------------------------------------#

_signal_to_server = True

_signal_socket    = None

#-----------------------------------------------------------------------------#

def init_tagsd_signaling():
	
	global _signal_socket
        
	try:
		_signal_socket = ClientTagStateSocketUDP()
		return True
	except Exception, e:
		if DEBUG: print __file__, type(e), str(e), traceback.format_exc()
		return False
	pass

#-----------------------------------------------------------------------------#

def set_signal_tagsd_server(mode):
	"""
	Abilita/Disabilita le azioni di "notifica" delle funzioni:
	
		* signal_condition_changed()
		* signal_state_refresh_safe()
	
	"""
	global _signal_to_server

	_signal_to_server = mode

	return

#-----------------------------------------------------------------------------#

def signal_conditions_changed(tenantid, condition_ids):
	for conditionid in condition_ids:
		signal_condition_changed(tenantid, conditionid)

def signal_condition_changed(tenantid, conditionid):
	"""
	Notifica a server centrale (tagsd) che la condition con UUID "conditionid" 
	deve essere ricaricata e lo stato dei tagtree deve essere ricalcolato.
	
	Restituisce True se la notifica avviene con successo, False altrimenti. 
	"""

	global _signal_socket

	if DEBUG: print __file__, "signal_condition_changed", conditionid

	if not _signal_to_server:
		return False

	try:
		if _signal_socket is not None:
			client = _signal_socket
		else:
			client = ClientTagStateSocketUDP()

		client.reload_conditions(tenantid, [conditionid])

		return True

	except Exception, e:
		if DEBUG: print __file__, type(e), str(e), traceback.format_exc()
		return False

	finally:
		try:
			if not _signal_socket:
				client.close()
		except Exception, e:
			pass

#-----------------------------------------------------------------------------#

def signal_state_refresh(tenantid, tree_ids=None, force=False):

	global _signal_socket

	if DEBUG: print __file__, "signal_state_refresh_safe", tree_ids, force

	if not _signal_to_server:
		return False

	try:
		if _signal_socket is not None:
			client = _signal_socket
		else:
			client = ClientTagStateSocketUDP()

		if force:
			client.refresh_all_force(tenantid, tree_ids=tree_ids)
		else:
			client.refresh_all(tenantid, tree_ids=tree_ids)

		return True

	except Exception, e:
		if DEBUG: print __file__, type(e), str(e), traceback.format_exc()
		return False

	finally:
		try:
			if _signal_socket is None:
				client.close()
		except Exception, e:
			pass

#-----------------------------------------------------------------------------#
# Funzioni *safe* per ottenere lo stato di TagTree/Tag/TagElement             #
#-----------------------------------------------------------------------------#

def get_tags_state_safe(tenantid, ids_for_state):
	"""
	Restituisce un dizionario con tutti gli stati di Tag/ElementTag con gli UUID indicati nella lista "ids_for_state"
	
	Gli stati sono indicizzati per UUID.
	
	Se "tagsd" non e' in esecuzione o ci sono problemi di qualunque tipo, restituisce None.
	"""

	try:
		client = ClientTagStateSocket()
		client.connect()
		state_by_uuid = client.get_tags_state(tenantid,  ids_for_state  )
		return state_by_uuid
	except Exception, e:
		if DEBUG: print __file__, type(e), str(e), traceback.format_exc()
		return None
	finally:
		try:
			client.close()
		except Exception, e:
			pass

def get_trees_state_safe(tenantid, ids_for_state):
	"""
	Restituisce un dizionario con tutti gli stati dei tree con gli UUID indicati nella lista "ids_for_state"
	
	Gli stati sono indicizzati per UUID.
	
	Se "tagsd" non e' in esecuzione o ci sono problemi di qualunque tipo, restituisce None.
	"""
	try:
		client = ClientTagStateSocket()
		client.connect()
		state_by_uuid = client.get_trees_state(tenantid,  ids_for_state  )
		return state_by_uuid
	except Exception, e:
		if DEBUG: print __file__, type(e), str(e), traceback.format_exc()
		return None
	finally:
		try:
			client.close()
		except Exception, e:
			pass



def get_tree_structure(tenantid, root_id):
	"""
	Restituisce la struttura dati TagTreeState usata dal server per calcolare lo stato del root tag root_id.
	"""
	try:
		client = ClientTagStateSocket()
		client.connect()
		structure = client.get_tree_structure(tenantid,  root_id  )
		return structure
	finally:
		try:
			client.close()
		except Exception, e:
			pass


