# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

Tagsd Server
===========================================================================

.. contents:: Contenuti

Introduzione
---------------------------------------------------------------------------

Questo processo mantiene in memoria le strutture dati per la gestione degli stati dei tagtree.

Il server accetta connessioni dall'esterno per ricevere due tipi di richiest

	* Notifica Aggioramento di stato
	* Lettura stato di un tagtree, di un tag o di un elemento taggato

Il programma e' diviso in 3 thread principali (e un thread pool):

	* Main Thread     : lancia gli altri thread e il threadpool
	* Listen Thread   : accetta connessioni dall'esterno
	* Status Thread   : aggiorna in maniera "asincrona" le strutture dati interne.


Schema thread interni e strutture dati
---------------------------------------------------------------------------

.. image:: ../_static/tagsd.png
	:width: 800

::
    
                  TagsdServer
                 +---------------------------------------------------------------------+
                 |     +------+       +------+       +------+       +------+           |   +-----+
       TCP  ---------> |Listen|       |Thread|       |ACTION|       |Status|    read   |   |SANET|
      socket     |     |Thread|-----> | Pool |-----> |QUEUE |-----> |Thread| <-------------| DB  |
                 |     +------+       +------+       +------+       +------+           |   +-----+
                 |         ^              |                             |              |
       UDP       |         |              |                             | update       |
      socket---------------+              |                             |              |
                 |                        |                             v              |
                 |                        |               +--------------------------+ |
                 |                        |       read    |                          | |
                 |                        +-------------->| TreeState Data Structure | |
                 |                                        |                          | |
                 |                                        +--------------------------+ |
                 +---------------------------------------------------------------------+

"""

import os
import sys
import time
import traceback

import logging

import threading
import Queue

import socket
import select
import struct
import json

from django.db import connection

from sanet_common.threadutils import AbstractDaemonThread


from core.api.base import db_set_current_tenant_schema
from core.models import Tenant, TenantSetting

from tags.models import TagTree

from tags.api.state2 import load_conditions_for_tagging, load_tree_tagging_data, create_tag_statetree_structure

from tags.server.tagsd.api import TagStateSocket, ClientTagStateSocket, TAGSD_SERVER_IP, TAGSD_SERVER_PORT


#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

_DEBUG = int(os.environ.get('DEBUG', 0) or 0)

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

TAGSD_SERVER_STATUS_WORKING = 'WORKING'

TAGSD_SERVER_STATUS_ERROR   = 'ERROR'

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def TENANT_DATA():
	"""
	Struttura dati per tenere in memoria lo stato dei tagtree/elementi taggati di un tenant.
	"""

	return {
		# Dizionario con tutti i dati delle condition ( indicizzati per UUID ) utilizzate per il calcolo degli stati.
		'conditions'    : None,

		# Dizionario che punta a tutti i tags.api.state.TreeNode per ogni elemento di tutti i tagtree (elementi indicizzati per tag.uuid, elementtag.uuid)
		'states_dict'   : {},

		# Dizioanrio che punta a tutti i tags.api.state.TreeNode della root di ogni tagtree (elementi indicizzati per tag.uuid)
		'states_by_tree': {},
	}

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class TagsdAbstractThread(AbstractDaemonThread):
	"""
	Tutti i thread hanno un riferimento alla classe TagServer in modo
	da accedere a:
	
		* strutture dati condivise, ecc.
		* logger
		* ecc..
	"""

	def __init__(self, server):

		super(TagsdAbstractThread, self).__init__()

		self.server = server

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#
class ThreadPool(object):
	"""
	Semplice thread pool per gestire le richieste provenienti dal thread
	di esecuzione
	
	"""

	def __init__(self, server, num_workers=5):

		self.server = server

		self.num_workers        = num_workers
		self.workers            = []
		self.work_queue         = Queue.Queue(-1)
		self.work_queue_timeout = -1

	def init_workers(self):
		for i in xrange(self.num_workers):
			t = ListenWorkerThread(i, self.server, self)
			self.workers.append(t)
			t.start()

	def assign_client_to_worker(self, handler_function, params):
		self.work_queue.put_nowait( (handler_function, params) )

	def wait_for_workers(self):
		for worker in self.workers:
			if worker.isAlive():
				worker.join(1)

	def get_status(self):
		return {
			'workers': [ worker.isAlive() for worker in self.workers ],
			'queue'  : self.work_queue.qsize(),
		}

class ListenWorkerThread(TagsdAbstractThread):

	def __init__(self, tid, server, listen_thread):

		super(ListenWorkerThread, self).__init__(server)

		self.tid           = tid
		self.listen_thread = listen_thread

	def run(self):
		while 1:
			try:
				data = self.listen_thread.work_queue.get(True, 1)
				self.server.logger.log(1, "Listen-Worker(%s): working on %s", self.tid, data )

				handler_function, params = data

				handler_function(*params)

			except Queue.Empty, er:
				data = None

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#


class ListenThread(TagsdAbstractThread):

	def __init__(self, server, thread_pool):
		super(ListenThread, self).__init__(server)

		self.thread_pool = thread_pool

		self.loop_wait = 2

	def run(self):
		while 1:
			self.listen_for_clients()

			self.server.logger.error("Listen loop ended. Restart in %s seconds.", self.loop_wait )
			time.sleep(self.loop_wait)

	def listen_for_clients(self):

		bind_address = (TAGSD_SERVER_IP, TAGSD_SERVER_PORT)

		self.server.logger.info("Opening listen socket on %s", bind_address )

		listen_socket_tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		listen_socket_tcp.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		listen_socket_tcp.bind(bind_address)
		listen_socket_tcp.listen(5)

		listen_socket_udp = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
		listen_socket_udp.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		listen_socket_udp.bind(bind_address)

		while 1:

			try:
				readable, writable, errors = select.select([listen_socket_tcp, listen_socket_udp], [], [])

				if errors:
					raise Exception("Errors while waiting for connections")

				# Nuova connessione TCP. Creiamo nuovo socket.
				if listen_socket_tcp in readable:

					client_socket, client_addr = listen_socket_tcp.accept()

					self.accept_client(client_socket, client_addr)

				# Nuovo pacchetto UDP? Leggiamolo.
				if listen_socket_udp in readable:

					data, client_address = listen_socket_udp.recvfrom(2**16)

					self.accept_client_udp(listen_socket_udp, client_address, data)

			except Exception, e:
				self.server.logger.error("ERROR IN LISTEN LOOP: %s:%s -> %s" % (type(e), str(e), traceback.format_exc()))

				try:
					listen_socket_udp.close()
					listen_socket_tcp.close()
				except Exception, e:
					pass

				return

	def accept_client(self, client_socket, client_addr):

		# Le connessioni TCP sono piu' lente.
		# La nuova connessione viente gestita interamente dal thread del pool

		self.thread_pool.assign_client_to_worker( self.handle_client_tcp, (client_socket, client_addr) )

	def accept_client_udp(self, server_socket, client_address, data):

		# Le richieste via UDP non passano dal thread pool.

		#self.thread_pool.assign_client_to_worker( self.handle_client_udp, (server_socket, client_address, data) )

		self.handle_client_udp(server_socket, client_address, data)


	def handle_client_tcp(self, conn, addr):

		try:
			while 1:
				try:
					t1 = time.time()

					data = TagStateSocket.read_request_socket(conn)
					if not data:
						return

					t2 = time.time()

					cmd = data

					response_data = self.process_command(cmd)

					t3 = time.time()

					TagStateSocket.send_response_socket(conn, response_data)

					t4 = time.time()

					self.server.logger.log(1, "%s: %s tot=%.4f  receive=%.4f  process=%.4f  send=%.4f", addr, cmd['command'], t4 -t1, t2-t1, t3-t2, t4-t3 )

				except Exception, e:
					self.server.logger.error("ERROR IN CLIENT LOOP: %s:%s -> %s" % (type(e), str(e), traceback.format_exc()))
					break

			conn.close()

		except Exception, e:
			self.server.logger.error("UNKNOWN ERROR WITH TCP CLIENT: %s:%s -> %s" % (type(e), str(e), traceback.format_exc()))


	def handle_client_udp(self, server_socket, client_address, data):

		try:
			t1 = time.time()
			#
			# REQUEST = header (4) + encoded_data
			#
			header, data = data[0:4], data[4:]

			cmd = TagStateSocket.decode_request(data)

			#
			# PROCESS
			#
			response_data = self.process_command(cmd)

			#
			# RESPONSE = header (4) + encoded_response
			#
			if response_data is not None:

				data = TagStateSocket.encode_response(response_data)

				buff = struct.pack('!i', len(data)) + data
				server_socket.sendto( buff,  client_address)

			t2 = time.time()

			#self.server.logger.log(9, "%s: %s tot=%.4f", client_address, cmd['command'], t2 -t1 )

		except Exception, e:
			self.server.logger.error("ERROR IN UDP CLIENT: %s:%s -> %s" % (type(e), str(e), traceback.format_exc()))

	def process_command(self, cmd):
		"""
		Processa un comando ricevuto da un client.
		
		Restituisce i dati da inviare al client.
		
		Restituisce None se il comando non prevede una risposta.
		"""

		if cmd['command'] == ClientTagStateSocket.CMD_SERVER_STATUS:

			response_data = { 
				'status': TAGSD_SERVER_STATUS_WORKING, 
				'info'  : '',
				'data'  : {
					'tenants': dict([ 
						(
							uuid, 
							{
							'conditions': len(tenant_data['conditions']),
							'trees': dict([ (uuid, tree_state._get_state_dict()) for uuid, tree_state in tenant_data['states_by_tree'].items() ]),
							}
						) for uuid, tenant_data in self.server.tenants_data.items() 
					]),
				},
				'threads': {
					'alive' : self.server.status_thread.isAlive(),
					'pool'  : self.thread_pool.get_status(),
				}
			}
			
			#response_data = json.dumps(response_data)

		elif cmd['command'] == ClientTagStateSocket.CMD_TREES_STATE:

			tenantid = cmd['tenantid']
			uuid     = cmd['uuids']

			tenant_data = self.server.tenants_data[tenantid]

			response_data = {}
			for uuid in cmd['uuids']:
				try:
					statetree = tenant_data['states_by_tree'][uuid]
					d = statetree._get_state_dict()
				except Exception, e:
					d = None

				response_data[uuid] = d

		elif cmd['command'] == ClientTagStateSocket.CMD_TAGS_STATE:

			tenantid = cmd['tenantid']
			uuids    = cmd['uuids']

			tenant_data = self.server.tenants_data[tenantid]

			response_data = {}
			for uuid in uuids:
				try:
					statetree = tenant_data['states_dict'][uuid]
					d = statetree.get_state_dict()
				except Exception, e:
					d = None

				response_data[uuid] = d

		elif cmd['command'] == ClientTagStateSocket.CMD_REFRESH_ALL:

			tenantid = cmd['tenantid']
			tree_ids = cmd['tree_ids']

			self.server.status_thread.signal_refresh_all(tenantid, tree_ids)

			response_data = None

		elif cmd['command'] == ClientTagStateSocket.CMD_REFRESH_ALL_FORCE:

			tenantid = cmd['tenantid']
			tree_ids = cmd['tree_ids']

			self.server.status_thread.signal_refresh_all_force(tenantid, tree_ids)

			response_data = None

		elif cmd['command'] == ClientTagStateSocket.CMD_RELOAD_CONDITIONS:

			tenantid       = cmd['tenantid']
			conditions_ids = cmd['conditions_ids']

			self.server.status_thread.signal_reload_conditions(tenantid, conditions_ids)

			response_data = None

		#
		# UTILITY (DEBUG) METHODS
		#
		elif cmd['command'] == ClientTagStateSocket.CMD_TREE_STRUCTURE:

			tenantid = cmd['tenantid']
			uuid     = cmd['uuid']

			self.server.logger.info(" STRUCTURE FOR: %s %s", tenantid, uuid)

			tenant_data = self.server.tenants_data[tenantid]

			response_data = None
			try:
				response_data = tenant_data['states_dict'][uuid]
			except Exception, e:
				# Data not found
				pass

		return response_data

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#




class StatusThread(TagsdAbstractThread):

	ACTION_REFRESH_ALL       = 'REFRESH_ALL'
	ACTION_REFRESH_ALL_FORCE = 'REFRESH_ALL_FORCE'
	ACTION_RELOAD_CONDITIONS = 'RELOAD_CONDITIONS'

	def __init__(self, server):

		super(StatusThread, self).__init__(server)

		self.action_queue = Queue.Queue(-1)

		self.action_queue_timeout       = 5

		self.action_queue_timeout_short = 2

		self.action_max_group_size      = 100

	def _push_action(self, fast_action, action, **params):
		"""
		Le azioni di refresh da fare eseguire al server hanno tutte un timestamp
		per sapere quando sono arrivate.
		"""
		self.action_queue.put_nowait( (fast_action, {'ts': time.time(), "action": action, "params": params }) )

	def signal_refresh_all(self, tenantid=None, tree_ids=None, fast_action=False):
		
		self._push_action(fast_action, StatusThread.ACTION_REFRESH_ALL      , tenantid=tenantid, tree_ids=tree_ids)

	def signal_refresh_all_force(self, tenantid=None, tree_ids=None, fast_action=False):
		
		self._push_action(fast_action, StatusThread.ACTION_REFRESH_ALL_FORCE, tenantid=tenantid, tree_ids=tree_ids)

	def signal_reload_conditions(self, tenantid=None, conditions_ids=None, fast_action=False):
		
		self._push_action(fast_action, StatusThread.ACTION_RELOAD_CONDITIONS, tenantid=tenantid, conditions_ids=conditions_ids)

	def run(self):

		connection.close()

		self.logger = self.server.logger

		grouping_actions = []

		time_to_wait = self.action_queue_timeout

		while 1:

			#self.logger.debug("Waiting for action. queued_actions=%s   grouping_actions=%s", self.action_queue.qsize(), len(grouping_actions) )

			try:
				#
				# CALC TIMEOUT
				#
				if grouping_actions:
					time_to_wait = self.action_queue_timeout_short
				else:
					time_to_wait = self.action_queue_timeout

				self.logger.debug("Waiting %s s for action (queued=%s ,  groupped=%s)", time_to_wait, self.action_queue.qsize(), len(grouping_actions) )

				#
				# WAIT
				#
				action = None
				try:
					fast_action, action = self.action_queue.get(True, time_to_wait)
				except Queue.Empty, er:
					pass

				#
				# PROCESS
				#
				if action is not None:
					grouping_actions.append( action )

					if not fast_action:
						if len(grouping_actions) < self.action_max_group_size:
							continue

				self.logger.debug("Processing actions=%s  ( queued actions pending=%s )", len(grouping_actions), self.action_queue.qsize(),  )

				self.group_actions_and_perform( grouping_actions )

				grouping_actions = []

				connection.close()

			except Exception, e:
				self.logger.error("Unkown error in state thread processing: %s:%s -> %s", type(e), str(e), traceback.format_exc())


	def group_actions_and_perform(self, actions):
		"""
		Tutte le action sono dizionari con i seguenti campi:
		
			{ ts: <timestamp> , action: tipo , tenantid: <uuid> , tree_ids: [ <uuid> , ....], condition_ids: [ <uuid> , ....] }
		
		Poiche' alcune action (ACTION_REFRESH_ALL_FORCE) sono piu' "potenti" di altre action, 
		questa funzione analizza tutte le actione e cerca di comprimere ad una sola action "significativa"
		per ogni tree.
		"""

		#print "____GROUPING____ %s" % (len(actions))

		actions_by_tenant = {}

		for action_data in actions:

			tenantid = action_data['params']['tenantid']

			#print "___PARAMS:", action_data['params']

			action_to_perform = actions_by_tenant.get( tenantid, None) 

			if action_to_perform is None:
				action_to_perform = action_data
			else:
				if action_to_perform['action'] != action_data['action'] and action_data['action'] == StatusThread.ACTION_REFRESH_ALL_FORCE:
					action_to_perform['action'] = action_data['action']

				if action_to_perform['action'] == action_data['action']:

					#print "action_to_perform", action_to_perform
					#print "action_data      ", action_data

					#
					# Aggrego gli aggiornamenti sui tree
					#
					s1 = action_to_perform['params'].get('tree_ids', None) or []
					s2 = action_data      ['params'].get('tree_ids', None) or []

					action_to_perform['params']['tree_ids'] = list(set(s1)|set(s2))

					#
					# Aggrego gli aggiornamenti sulle condition
					#
					c1 = action_to_perform['params'].get('conditions_ids', None) or []
					c2 = action_data      ['params'].get('conditions_ids', None) or []

					# sommo tutte le condition che devono essere aggiornate e le assegno alla chiave di dizionario
					action_to_perform['params']['conditions_ids'] = list(set(c1+c2))

				action_to_perform['ts'] = action_data['ts']

			actions_by_tenant[ tenantid ] = action_to_perform

		actions = actions_by_tenant.values()

		#print "____GROUPING END____"

		for action in actions:
			self.peform_action(action)

		return

	def peform_action(self, action):

		self.logger.log(9, "PERFORMING ACTION: %r", action)

		update_request_time    = action.get('ts', None)

		action_cmd, params_dict = action['action'], action['params']

		if action_cmd in [ StatusThread.ACTION_REFRESH_ALL, StatusThread.ACTION_REFRESH_ALL_FORCE, StatusThread.ACTION_RELOAD_CONDITIONS]:

			force_refresh            = action_cmd in [StatusThread.ACTION_REFRESH_ALL_FORCE ]
			force_reload_conditions  = action_cmd in [StatusThread.ACTION_RELOAD_CONDITIONS, StatusThread.ACTION_REFRESH_ALL_FORCE ]

			self.logger.debug("Refreshing")

			tenantid       = params_dict['tenantid']
			conditions_ids = params_dict.get('conditions_ids',   [])
			tree_ids       = params_dict.get('tree_ids'      , None)

			tenants = Tenant.objects
			if tenantid is not None:
				tenants = tenants.filter(uuid=tenantid)

			tenants = tenants.all()

			for tenant in tenants:
				self.update_tenant(tenant, update_request_time, force_refresh, force_reload_conditions, conditions_ids=conditions_ids, tree_ids=tree_ids)

		self.logger.log(9, "END PERFORM")

	def update_tenant(self, tenant, update_request_time=None, force_refresh=False, force_reload_conditions=False, conditions_ids=None, tree_ids=None):

		self.logger.debug("force_refresh: %s", force_refresh)

		tenant_data = self.server.tenants_data.get(tenant.uuid, self.server.tenants_data.setdefault(tenant.uuid, TENANT_DATA()))

		t_start = time.time()

		db_set_current_tenant_schema(tenant.name)

		critical_priority = tenant.get_setting(TenantSetting.SETTING_CRITICAL_LEVEL)
		minimum_priority  = tenant.get_setting(TenantSetting.SETTING_MINIMUM_LEVEL)

		self.logger.info("Tenant: %s (force_refresh=%s, force_reload_conditions=%s, conditions_ids=%s, tree_ids=%s)", tenant.name, force_refresh, force_reload_conditions, conditions_ids, tree_ids)
		self.logger.info("   critical lvel: last=%s current=%s", critical_priority, None)
		self.logger.info("   mininum  lvel: last=%s current=%s", minimum_priority , None)

		#
		# Reload tenant conditions?
		#

		current_conditions_by_uuid = tenant_data['conditions']

		if current_conditions_by_uuid is None or force_reload_conditions:

			self.logger.debug( "\t" + "Loading conditions...")

			t_condition = time.time()

			loaded_conditions_by_uuid = load_conditions_for_tagging(conditions_ids)

			t_loaded = time.time()

			self.logger.info( "\t" + "Loaded conditions %s %s", len(loaded_conditions_by_uuid), t_loaded - t_condition)

			if current_conditions_by_uuid is None:
				current_conditions_by_uuid = loaded_conditions_by_uuid
			else:
				if conditions_ids:
					current_conditions_by_uuid.update(loaded_conditions_by_uuid)
				else:
					current_conditions_by_uuid = loaded_conditions_by_uuid

		tenant_data['conditions'] = current_conditions_by_uuid

		#
		# Update tree state? Reload tree state?
		#

		self.logger.debug("\t" + "Loading trees (ids=%s)...", tree_ids)

		trees = TagTree.objects

		# Update only specific trees?
		if tree_ids:
			trees = trees.filter(uuid__in=tree_ids).filter(nostate=False)

		# FOR DEBUG: update only tree with the given name
		if self.server.debug_tagtree_names:

			names = self.server.debug_tagtree_names.split(',')

			self.logger.warning("DEBUG_TREES: %s", names)
			trees = trees.filter(name__in=names)

		trees = trees.all().order_by('-nostate', 'name')

		self.logger.info("\t" + "Loaded from DB %s trees", len(trees))

		for tree in trees:

			if tree.root is None: 
				self.logger.info("\t" + "\t" + "%20s %32s has no root tag", tree.name, tree.uuid, )
				continue

			if tree.nostate is True:
				self.logger.info("\t" + "\t" + "%20s %32s is nostate", tree.name, tree.uuid, )
				continue

			current_treestate = tenant_data['states_by_tree'].get( tree.uuid , None)

			#
			# Lo stato corrente del server e' stato calcolato dopo che e' arrivata 
			#
			# IMPORTANTE: current_treestate e' un oggetto TreeNode. Il campo "update_ts" non e' un campo ufficiale, viene aggiunto/gestito solo in questo programma
			#
			if current_treestate  and  update_request_time < current_treestate.update_ts:
					self.logger.info("\t" + "\t" + "Update tree %s skipped. Action too old %s < %s", tree.name, update_request_time, current_treestate.update_ts )
					continue

			t_reload = time.time()
			
			if current_treestate is None or force_refresh:
				current_treestate, t_load, t_create = self.reload_tagtree_structure( tenant_data, tree )
			else:
				t_load, t_create = 0, 0

			#
			# CALCULATE STATE FROM TREE + CONDITIONS DATA
			#

			t_state = time.time()

			status = current_treestate.calculate_state(current_conditions_by_uuid, critical_priority=critical_priority, minimum_priority=minimum_priority)

			# IMPORTANTE: Ricordiamoci che questo tree e' stato aggiornato in base ad una struttura dati caricata al momento t_reload.
			#             Se nel frattempo sono arrivate delle richieste di aggiornamento dobbiamo ricaricare.
			#             Se le richieste di refresh sono arrivate prima di t_reload, andranno buttate.
			current_treestate.update_ts = t_reload

			t_end = time.time()
			
			#if _DEBUG:
			#	if len(trees) == 1:
			#		current_treestate.visit()


			#
			# Update global structure
			#
			tenant_data['states_by_tree'][ tree.uuid ] = current_treestate

			#
			# Update Tree.stime on DB
			#
			tree.update_state_time( current_treestate.ts )

			t = (tree.name, tree.uuid, t_end - t_reload, t_load, t_create, t_end - t_state   , status, current_treestate.visible_status, current_treestate.alarms, current_treestate.critical_alarms)

			self.logger.info("\t" + "\t" + "%20s %32s tot=%5.3f  load=%5.3f  create=%5.3f  state=%5.3f  %5s %5s %5s %5s", *t)

		t_final = time.time()

		self.logger.debug("\t" + "Total: %s", t_final - t_start)

	def reload_tagtree_structure(self, tenant_data, tree):

		#
		# LOAD DATA FROM DB
		#
		t_load = time.time()

		tags_by_parent, tagged_by_tag, tagged_by_parent = load_tree_tagging_data(tree)

		t_create = time.time()

		#
		# CREATE TREE 
		#
		root_tag = tree.root

		new_states_dict = tenant_data['states_dict']

		statetree = create_tag_statetree_structure(new_states_dict, tags_by_parent, tagged_by_tag, tagged_by_parent, root_tag)

		tenant_data['states_dict'] = new_states_dict

		t_end = time.time()

		return statetree, t_create - t_load, t_end - t_create


#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

class TagServer(object):

	def __init__(self, logger, debug_tagtree_names=None):
		self.logger = logger

		self.thread_pool   = None
		self.listen_thread = None
		self.status_thread = None

		# Dizionario indicizzato per tenant_id che 
		self.tenants_data = {}
		
		self.debug_tagtree_names = debug_tagtree_names

	def start(self):

		self.logger.info("Starting server...")

		#
		# INIT
		#
		self.logger.debug("Init thread pool")
		self.thread_pool = ThreadPool(self)
		self.thread_pool.init_workers()

		self.logger.debug("Init listen thread")
		threads = []
		self.listen_thread = ListenThread(self, self.thread_pool)
		threads.append(self.listen_thread)

		self.logger.debug("Init status thread")
		self.status_thread = StatusThread(self)
		self.status_thread.signal_refresh_all_force(fast_action=True)
		threads.append(self.status_thread)

		#
		# START ALL
		#
		self.logger.debug("Starting all threads")
		for t in threads:
			t.start()

		#
		# WAIT
		#
		self.logger.debug("Waiting for threads")
		
		for t in threads:
			while t.isAlive():
				t.join(1)

		self.thread_pool.wait_for_workers()

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def init_logger(log_level, log_file):

	logger = logging.getLogger('tagsd')
	logger.setLevel( log_level )

	hdlr = logging.FileHandler(log_file, "a")
	hdlr.setFormatter( logging.Formatter('%(asctime)s %(levelname)s tagsd > %(message)s') )
	hdlr.setLevel( log_level )
	logger.addHandler(hdlr)

	hdlr = logging.StreamHandler()
	hdlr.setFormatter( logging.Formatter('%(asctime)s %(levelname)s tagsd > %(message)s') )
	hdlr.setLevel( log_level )
	logger.addHandler(hdlr)

	return logger

