# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

"""

show_exec_times
========================================================================

Syntax: 

	::
		
		sanet-manage show_exec_times [ --tree <name>,<name>,<name,...] 

The program will show the total execution time for a single node. The total execution time for a node is calculated by summing all node's datagroups last execution time.

Options:

	::
		
		--tree      List of comma-seprated tagtree names.
		
		            If the nodes is tagged inside anyone of the specified tagtree, the command will show the related tags' path.

**Example**:

	::
		
		# sanet-manage show_exec_times
		NODE                                                   TIME  TAGS
		internet                                              0.248  

	::
		
		# sanet-manage show_exec_times
		NODE                                                   TIME  TAGS
		internet                                              0.248  
"""


import sys
from optparse import make_option

from django.db import connection
from django.utils.translation import ugettext as _

from core.models import MonitoredNode
from core.models import DataGroup
from core.api.cache import init_cache_manager_from_conf
from core.management.commands import TenantBasedCommand

from tags.models import TagTree

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

class Command(TenantBasedCommand):

	help = _("Tempi totali di esecuzione di un nodo con informazione sul tagging")

	def add_arguments(self, parser):

		super(Command, self).add_arguments(parser)

		parser.add_argument('--trees'            , action='store'     , dest='tree_names', default=None   , help='List of comma-seprated tagtree names.'),

	def handle(self, cmd='', *args, **options):

		tenant = self.get_tenant(options)

		#
		# Calculate tags for node
		#
		tags_by_node = {}
		
		tree_names = options['tree_names']
		if tree_names:
			if tree_names == '*':
				tree_names = [t.name for t in TagTree.objects.all()]
			else:
				tree_names = tree_names.split(",")

			tags_by_node = self.get_tags_by_node(tree_names)

		#
		# Calculate times for every node
		#
		totals_by_node_id = self.get_datagroups_exec_times_by_node(tree_names)

		#
		# PRINT TABLE
		#
		nodes = MonitoredNode.objects.order_by('name').select_related('agent').values_list('uuid', 'name', 'agent__name')

		max_len = max([len(name) for uuid,name,agent_name in nodes])

		FORMAT = "%-"+str(max_len)+"s  %-20s  %7s  %3s  %7s  %-s"
		print FORMAT % ('NODE', 'AGENT', 'TIME',  'N', 'AVG', 'TAGS')
		
		for node_id, node_name, agent_name in nodes:

			tot_time,tot_dg = totals_by_node_id.get(node_id, [0,0] )
			if tot_dg > 0:
				avg = float(tot_time) / tot_dg
			else:
				avg = 0

			tags     = tags_by_node.get(node_id, [])

			print FORMAT % (node_name, agent_name, "%7.3f" % (tot_time), tot_dg, "%7.3f" % (avg), ', '.join(tags),)

		return

	def get_tags_by_node(self, tree_names):

		QUERY = """
			select   node.uuid
			       --, node.name 
			       --, et.uuid 
			       --, et.tag_id
			       , tag.path
			  from monitoredelement node  inner join tags_elementtag et    on (node.uuid  = et.elementid and et.directly_tagged=True)
			                              inner join tagtree         tree  on (et.tree_id = tree.uuid)
			                              inner join tagentry        tag   on (et.tag_id  = tag.uuid)
			 where node.kind = 1
			   and tree.name in %s
			-- order by node.name
		"""

		params = [ tuple(tree_names),]

		c = connection.cursor()

		c.execute(QUERY, params)

		data = c.fetchall()
		
		tags_by_node = {}

		for node_id, tag_name in data:
			tag_names = tags_by_node.get(node_id, tags_by_node.setdefault(node_id, []) )
			tag_names.append(tag_name)

		return tags_by_node


	def get_datagroups_exec_times_by_node(self, tree_name):
	
	
		cache_manager = init_cache_manager_from_conf()
	
		QUERY = """
			select   dg.uuid
			       , dg_element.uuid
			       , dg_element.parent_id
			  from datagroup dg inner join monitoredelement dg_element ON (dg.element_id = dg_element.uuid)
			;
		"""
		#	-- where 
		#	--       exists (
		#	--               select node.uuid
		#	--                from monitoredelement node  inner join tags_elementtag et    on (node.uuid  = et.elementid and et.directly_tagged=True)
		#	--                			inner join tagtree         tree  on (et.tree_id = tree.uuid)
		#	--               where tree.name = %s
		#	--                 and (node.uuid = dg_element.uuid 
		#	--                      or
		#	--                      node.uuid = dg_element.parent_id
		#	--                     )
		#	--       )

		params = [ tree_name ]

		c = connection.cursor()

		#print QUERY

		c.execute(QUERY)#, params)

		data = c.fetchall()

		totals_by_node_id = {}

		for dg_uuid, element_id, parent_element_id in data:

			fields = DataGroup.raw_load_fields_from_cache(cache_manager, dg_uuid, ['lastchecktotaltime'])
			
			val = fields.get('lastchecktotaltime', None)
			if not val:
				val = 0

			node_id = parent_element_id or element_id

			totals     = totals_by_node_id.get(node_id, totals_by_node_id.setdefault(node_id, [0,0]))
			totals[0] += val
			totals[1] += 1

		return  totals_by_node_id


