# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import time
import pprint
import traceback

from sanet_common.sanetdefs.constants import DIRENTRY_AUTH_READWRITE, CSTATE_UU, CSTATE_TO_CODENAME, CSTATES_UNCHECKABLE
from sanet_common.timeutils import timestamp_to_isostr

from django.conf import settings
from core.compatibility import reverse

from core.models import Tenant, MonitoredNode

from tags.models import Tag, ElementTag
from tags.server.tagsd.api import get_tags_state_safe

from webui.models import webui_elementbadge_urls

from webui.api.Data import BaseDataSet
from webui.api.constans import ICON_WARNING, ICON_UNCHECK, ICON_NOSTATE, ICON_OLD
from webui.decorators import calculate_exec_time

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def flat_keys(d):
	l = d.keys()
	for i in d.values():
		l += flat_keys(i)
	return l

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

SORT_FUN = lambda x: (x['status_order'], x['name'])

class DataSet(BaseDataSet):

	def __init__(self, id):
		super(DataSet, self).__init__(id)

	@calculate_exec_time
	def get_data(self, options): 

		tenant            = options['tenant']

		user              = options.get('user')

		capabilities      = options['user_capabilities']

		element           = self.fetch_entity(options)

		critical_priority = options['critical_priority']
		minimum_priority  = options['minimum_priority']

		open_badges       = options['params'].get('open_badges',{})
		opening           = options['params'].get('opening'    ,None)

		#
		# State 
		#

		if not opening:

			main_element = element
			open_badges  = flat_keys(open_badges)

			#print __file__, "\t", "MAIN ELEMENT", element
			#print __file__, "\t", "OPEN:"
			#for uuid in open_badges:
			#	print __file__, "\t", "\t", uuid

			ids_for_state = []
			ids_for_state += [r[0] for r in Tag.objects.filter(parent__in=open_badges).values_list('uuid')]
			ids_for_state += [r[0] for r in ElementTag.objects.filter(tag_id__in=open_badges+ids_for_state).values_list('uuid')]

			# FIX: la gestione delle icone nostate_icon fa veramente schifo cosi'

			sections = []

			if isinstance(main_element, Tenant):

				# Se stiamo guardando i badge nella pagina principale del tenanto dobbiamo filtrare per utente
				main_element.set_tag_visibility(user)

				# Prendiamo tutti i "root" tag
				tags = main_element.child_tags.select_related('tree')
			else:
				tag  = main_element
				tags = [ tag ]

			#print "\t", "TAGS:"
			#for t in tags:
			#	print "\t", "\t", t

			if isinstance(main_element, Tenant):

				#
				# LOAD STATES
				#
				state_tags   = [ t for t in tags if t.tree.nostate == False ]
				nostate_tags = [ t for t in tags if t.tree.nostate == True  ]
				ids_for_state += [t.uuid for t in state_tags ]

				state_by_uuid = get_tags_state_safe( tenant.uuid, ids_for_state  )

				# Vediamo se gli stati sono "aggiornati" 
				if state_by_uuid:
					for t in state_tags:
						state_data = state_by_uuid.get(t.uuid,None)
						if state_data:
							if t.tree.get_state_time() > state_data['_ts']:
								state_data['isold'] = 1

				#
				# CREATE BADGES
				#

				# Prima tutti i root tag di cui calcolo lo stato
				objects = []
				for elem in state_tags:
					o = create_badgeview_object(state_by_uuid, elem, open_badges=open_badges,  user=user)
					objects.append(o)

				# Ordiniamoli
				objects = sorted(objects, key=SORT_FUN)
				sections.append({'name':'Tags', 'objects': objects})

				# I tag no state no hanno bisogno di stato
				objects = []
				for elem in nostate_tags:
					o = create_badgeview_object(state_by_uuid, elem, open_badges=open_badges, user=user, nostate=True)
					objects.append(o)

				sections.append({'name':'Service tags', 'objects': objects})
			else:
				# Per un tag mostro i badge di tutti i subtag e sotto elementi taggati
				tags             = main_element.child_tags.all()
				tagged, children = main_element.get_elementtags_with_elements(only_directly_tagged=True, path_names=True)

				# HACK: La funzione create_badgeview_object() calcola l'icona da far vedere, ma gli oggetti
				#       children sono di tipo MonitoredElement. Gli oggetti "children" relativi a nodi
				#       avrebbero icona "node" e questo e' incorente con il fatto che per oggetti
				#       di tipo MonitoredNode il calcolo dell'icona e' piu' complessa.
				#       Qui carichiamo gli oggetti MonitoredNode da usare piu' sotto al posto degli oggetti MonitoreElement.
				# FIX: trovare un modo per non fare questa query in piu'.
				nodes_by_uuid = dict((n.uuid, n) for n in MonitoredNode.objects.filter(uuid__in=[c.uuid for c in children]).all())

				#
				# LOAD STATES
				#
				ids_for_state += [t.uuid for t in tags ]
				ids_for_state += [et.uuid for et in tagged ]
				state_by_uuid = get_tags_state_safe(tenant.uuid, ids_for_state )

				# Stati aggiornati?
				if state_by_uuid:
					tree_ts = main_element.tree.get_state_time()
					for uuid in ids_for_state:
						state_data = state_by_uuid.get(uuid,None)
						if state_data:
							if tree_ts > state_data['_ts']:
								state_data['isold'] = 1

				#
				# CREATE BADGES
				#

				# Sezione tag
				objects = []
				elements =  list(tags)
				for elem in elements: 

					o = create_badgeview_object(state_by_uuid, elem, open_badges=open_badges, user=user)
					objects.append(o)

				objects = sorted(objects, key=SORT_FUN)
				sections.append({'name':'Tags', 'objects': objects})

				# Sezioni elementi taggati
				objects_by_type = {}
				for c in list(children): 
					
					# HACK: MonitoredElement -> MonitoredNode
					# FIX: cambiare in maniera piu' ottimizzata.
					c = nodes_by_uuid.get(c.uuid, c)

					objects = objects_by_type.get(c.resource_type,  objects_by_type.setdefault(c.resource_type, []))

					#cerco in tagged l'element_tag associato a c
					element_tag = ''
					for i in tagged:
						if i.elementid == c.uuid:
							element_tag = i.uuid
							break

					c.cache_id = element_tag


					o = create_badgeview_object(state_by_uuid, c, open_badges=open_badges, user=user)

					o["element_tag"] = element_tag

					objects.append(o) 

				sections += create_element_sections(objects_by_type)

			data = {
				'etype': element.resource_type,
				'uuid' : element.resource_id,

				'opening'  : opening,

				'sections' : sections,
				#'can_untag' : 1 if 'tags.tag.tag_element' in capabilities else 0
			}

		else:
			elem = Tag.objects.select_related('tree').get(uuid=opening)
			#print "OPENING", elem.name

			open_badges = flat_keys(open_badges)
			#print "\t", "OPEN:"
			#for uuid in open_badges:
			#	print "\t", "\t", uuid

			#
			# LOAD STATES
			#

			# Troviamo gli UUID di tutti le entita' (tag o taggedelement) di cui farci dare lo stato
			ids_for_state = [elem.uuid]
			ids_for_state += [r[0] for r in Tag.objects.filter(parent__in=ids_for_state).values_list('uuid')]
			ids_for_state += [r[0] for r in ElementTag.objects.filter(tag_id__in=ids_for_state).values_list('uuid')]

			state_by_uuid = get_tags_state_safe(tenant.uuid, ids_for_state )

			# Stati aggiornati?
			if state_by_uuid:
				for uuid in ids_for_state:
					state_data = state_by_uuid.get(uuid,None)
					if state_data:
						if elem.tree.get_state_time() > state_data['_ts']:
							state_data['isold'] = 1

			#
			# CREATE BADGES
			#
			objects = []
			o = create_badgeview_object(state_by_uuid, elem, open_badges, user=user)
			objects.append(o)

			data = {
				'etype': element.resource_type,
				'uuid' : element.resource_id,

				'opening': opening,

				'objects': objects,
				
				#'can_untag' : 1 if "tags.tag.tag_element" in capabilities else 0
			}

		#
		# URLS
		#
		data['urls'] = {
			'reload_state': reverse('tags.views.reload_state') if "tags.tag.state.force_reload" in capabilities else '' ,
		}

		#
		# Capabilities
		#
		data.update({
			'can_untag'       : 1 if "tags.tag.tag_element"        in capabilities else 0,
			'can_force_reload': 1 if "tags.tag.state.force_reload" in capabilities else 0,
		})

		return {
			'dataset': data
		}

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#


def create_badgeview_object(state_by_uuid, element, open_badges={}, nostate_icon=ICON_NOSTATE, minimum_priority=None, critical_priority=None, user=None, nostate=False):

	# FIX: la gestione delle icone nostate_icon fa veramente schifo cosi'

	status                = CSTATE_UU
	status_code           = CSTATE_TO_CODENAME(CSTATE_UU)
	status_info           = 'State not calculated'

	alarms_count          = 0
	critical_alarms_count = 0
	exploded              = False
	icon_url              = None
	icon_overlay_url      = None

	status           = None
	levels           = []

	if not nostate:

		if state_by_uuid is not None:

			uuid_to_use = element.cache_id if hasattr(element, 'cache_id') else element.uuid

			cached_state = state_by_uuid.get(uuid_to_use, None)

			if cached_state:
				status                = cached_state['visible_status']
				alarms_count          = cached_state['alarms']
				exploded              = cached_state['exploded']
				critical_alarms_count = cached_state['critical']

				if critical_alarms_count > 0 or exploded:
					icon_overlay_url = ICON_WARNING
					status_code      = 'DN'
				#elif status in CSTATES_UNCHECKABLE:
				#	icon_overlay_url = ICON_UNCHECK

				elif status is None:
					icon_overlay_url = ICON_NOSTATE

				if critical_alarms_count > 0 or exploded:
					icon_overlay_url = ICON_WARNING
					status_code     = 'DN'

				status_info = 'Last updated %s.' % (timestamp_to_isostr(cached_state['_ts']))

				if cached_state.has_key('isold'):
					status_info      += 'Status is not up-to-date. System is probably recalculating it'

					if settings.WEBUI_STATE_SHOW_OLD_ICON:
						icon_overlay_url = ICON_OLD

			else:
				status_info = 'Status not found/calculated yet.'
				icon_overlay_url = nostate_icon
		else:
			# Non ho dati (server tagsd non attivo?), 
			icon_overlay_url = nostate_icon
			status_info      = 'Unable to get status. Sanet server not workingg properly.'
	else:
		icon_overlay_url = nostate_icon
		status_info      = 'Status disabled by configuration'


	#
	# Additional buttons for element's page sections (tabs)
	#
	section_urls = []

	apps_urls = webui_elementbadge_urls.send(None, element=element, element_page_url=element.page_url)
	for app_name, urls in apps_urls:
		if urls:
			section_urls.extend( urls )


	#
	# Dati del singolo elemento
	#
	status_code           = CSTATE_TO_CODENAME(status)

	element_data = {

		'id'             : element.uuid,
		'type'           : element.resource_type,
		'name'           : element.path if element.resource_type == 'interface' or element.resource_type == 'storage' or element.resource_type == 'service' or element.resource_type == 'device' else element.name,
		'url'            : element.page_url,
		'icon'           : element.get_icon_url_with_state(status_code),

		'nostate'        : element.tree.nostate if hasattr(element, 'tree') else False,
		'status'         : status_code,
		'status_info'    : status_info,
		'alarms'         : alarms_count,
		'icon_overlay'   : icon_overlay_url,

		'status_order'   : status if status is not None else 99,

		'sections'       : [],

		'tag'            : 1 if element.resource_type == "tag" else 0,

		'can_edit'       : None,
		'user_can_rename': None,
		'user_can_delete': None,
		
		'section_urls'   : section_urls,
	}

	if element.resource_type == "tag":

		can_edit = ''
		if user.is_superuser == True:
			can_edit = 1
		else:
			can_edit = 1 if element.can_edit(user)['auth_type__max'] == DIRENTRY_AUTH_READWRITE else 0

		capabilities = user.get_capabilities(element.tenant)

		element_data['can_edit']        = can_edit
		element_data['user_can_rename'] = 1 if "tags.tag.edit"   in capabilities else 0
		element_data['user_can_delete'] = 1 if "tags.tag.delete" in capabilities else 0

	#
	# Aperto?
	#
	is_open = element.uuid in open_badges
	if is_open:
		sections = []

		#
		# Subtags
		#
		subtags          = element.child_tags.all()

		objects = []
		for elem in list(subtags):
			o = create_badgeview_object(state_by_uuid, elem, open_badges, nostate_icon=nostate_icon, user=user, nostate=nostate)
			objects.append(o) 

		objects = sorted(objects, key=SORT_FUN)
		sections.append({'name':'Tags', 'objects':objects})

		#
		# Elements
		#
		tagged, children = element.get_elementtags_with_elements(only_directly_tagged=True)

		elementtag_by_elementid = dict([ (et.elementid, et) for et in tagged])

		objects_by_type = {}
		for elem in list(children):

			elementtag = elementtag_by_elementid.get( elem.uuid, None)

			elem.cache_id = elementtag.uuid

			objects = objects_by_type.get(elem.resource_type,  objects_by_type.setdefault(elem.resource_type, []))

			o = create_badgeview_object(state_by_uuid, elem, open_badges, nostate_icon=nostate_icon, user=user, nostate=nostate)

			#cerco in tagged l'element_tag associato a c
			element_tag = ''
			for i in tagged:
				if i.elementid == elem.uuid:
					element_tag = i.uuid
					break

			o["element_tag"] = element_tag

			objects.append(o) 

		sections += create_element_sections(objects_by_type)

		element_data['sections'] = sections

	return element_data

#------------------------------------------------------------------------------#
#                                                                              #
#------------------------------------------------------------------------------#

def create_element_sections(objects_by_type):

	SECTIONS = [
		#<resource_type>, <section title>
		( 'node'        , 'Nodes'     ), 
		( 'interface'   , 'Interfaces'), 
		( 'storage'     , 'Storages'  ), 
		( 'service'     , 'services'  ), 
		( 'device'      , 'Devices'   ), 
		( 'condition'   , 'Conditions')
	]

	sections = []

	for resource_type, section_title in SECTIONS:

		objects = objects_by_type.get(resource_type,None)
		if not objects: continue

		objects = sorted(objects, key=SORT_FUN)

		sections.append({'name':section_title, 'objects':objects})

	return sections

