# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import time
import math

from core.compatibility import reverse
from django.db.models import Q

from core.api.cache import init_cache_manager_from_conf
from core.models import DataSource, TimeGraph, TimeGraphSerie, TenantSetting, Tenant, PriorityLabel

from sanet_common.sanetdefs.constants import ELEMENT_KIND_STR, ELEMENT_KIND_VAL, ELEMENT_KIND_IFACE

from webui.api.Data import BaseDataSet

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

class DataSet(BaseDataSet):

	def __init__(self, id):

		super(DataSet, self).__init__(id)

	def get_data(self, options): 

		user   = options.get('user')
		tenant = options.get('tenant')
		
		params = options.get('params', {})

		page     = int(params.get('page'         ,    0))
		pagesize = int(params.get('pagesize'     ,   5))
		section  = params.get('section'          , None)
		search   = params.get('search_string'    ,   '')

		element = self.fetch_entity(options)

		data = produce_data(tenant, user, element, page=page, section=section, search=search, pagesize=pagesize)

		return {
			'dataset': data
		} 

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#


def produce_data(tenant, user, element, page=0, section=None, search='', pagesize=None):
	
	minimum_priority  = tenant.get_setting(TenantSetting.SETTING_MINIMUM_LEVEL)

	sections = []

	LIMIT = int(pagesize) if pagesize is not None else 5
	start = page  * LIMIT
	stop  = start + LIMIT

	LIMIT_TIMEGRAPH = LIMIT
	tstart = page  * LIMIT_TIMEGRAPH
	tstop  = start + LIMIT_TIMEGRAPH

	tot_counters = {}
	page_numbers = {}

	cache_manager = init_cache_manager_from_conf()

	entities  = {}
	datagroup = {}
	
	
	ELEMENT_SECTIONS = [
		 'node'      
		,'interface' 
		,'storage'   
		,'service'   
		,'device'    
	]
	ALL_SECTIONS = ELEMENT_SECTIONS + [ 'timegraph' ]
	
	TITLES = {
		 'node'      : 'Nodes'
		,'interface' : 'Interfaces'
		,'storage'   : 'Storages'
		,'service'   : 'Services'
		,'device'    : 'Device'
		,'timegraph' : 'Timegraph'
	}
	


	if section is not None:
		troblevalues = [ section.lower() ]
	else:
		troblevalues = ALL_SECTIONS
	

	cached_linked_elements = {}
	
	def add_linked_data(entry, element):

		linked_data = None
		
		if cached_linked_elements.has_key(element.uuid):
			linked_data = cached_linked_elements[element.uuid]
		else:
			i = element.get_specific_element()
			l_iface = i.get_linked_iface()
			if l_iface is None:
				linked_data = None
			else:
				l_node  = l_iface.get_node()
				linked_data = {
					'iface_name': l_iface.name,
					'iface_url' : l_iface.page_url,
					
					'node_name' : l_node.name,
					'node_url'  : l_node.page_url,
				};
			cached_linked_elements[element.uuid] = linked_data
			
		if linked_data != None:
			entry['linked'] = linked_data
		
		
	
	


	all_selected_conditions = element.get_all_conditions_by_user(user)

	for section in ELEMENT_SECTIONS:
		
		if section not in troblevalues:
			continue

		if not LIMIT:
			break

		kind         = ELEMENT_KIND_VAL[section]
		

		base_selected_conditions = all_selected_conditions.filter(element__kind=kind)
		selected_conditions      = base_selected_conditions.select_related('element', 'element__parent', 'datagroup')

		if search != '':
			selected_conditions  = selected_conditions.filter(Q(title__icontains=search) | Q(path__icontains=search))

		tot_counters[section] = float(selected_conditions.count())
		page_numbers[section]  = int(math.ceil(float(tot_counters[section]) / LIMIT)) # add page to round value

		selected_conditions      = selected_conditions.order_by('status', '-statuslastchange')
		
		selected_conditions      = selected_conditions[start:stop]

		priority_labels = {}

		for c in selected_conditions:

			# expanded title
			title = c.expanded_title(cache_manager=cache_manager)

			d = c.datagroup
			d.load_fields_from_cache(cache_manager=cache_manager)
			
			p_label = priority_labels.get(c.priority, priority_labels.setdefault(c.priority, PriorityLabel.get_label(c.priority)))
			
			
			entry = {
				'type'                    : 'condition',
				'id'                      : c.uuid, 
				'title'                   : title, 
				'state'	                  : c.get_state(),
				'url'                     : c.page_url, # reverse('resource_page', kwargs={'rt': c.resource_type, 'uuid': c.resource_id}),

				'datagroup'               : '',
				
				'element_kind'            : c.element.path,
				'element_kind_url'        : c.element.page_url, 

				'statuslastchange'        : time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(c.statuslastchange)       )  if c.statuslastchange        != 0 else "-",
				'statuslastnotablechange' : time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(c.statuslastnotablechange))  if c.statuslastnotablechange != 0 else "-",
				'lastdone'                : time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(d.lastdone)               )  if d.lastdone                != 0 else "-",
				'primary'                 : c.primary,
				'priority'                : c.priority,

				'plabel'                  : p_label,
				'ignored'                 : c.priority < minimum_priority
			}
			
			if c.element.kind == ELEMENT_KIND_IFACE:
				add_linked_data(entry, c.element)

			entities.setdefault(section, []).append(entry)

	if 'timegraph' in troblevalues and LIMIT_TIMEGRAPH:

		section = 'timegraph'

		selected_timegraph = element.get_all_timegraphs_by_user(user).select_related('element','element__parent','datagroup')

		if search != '':
			selected_timegraph  = selected_timegraph.filter(Q(title__icontains=search) | Q(element__name__icontains=search)| Q(element__parent__name__icontains=search))

		tot_counters[section] = selected_timegraph.count()
		page_numbers[section]  = int(math.ceil(float(tot_counters[section]) / LIMIT_TIMEGRAPH)) # add page to round value

		selected_timegraph  = selected_timegraph.order_by('element__name', 'datagroup__name')
		selected_timegraph = selected_timegraph[tstart:tstop]

		# Troviamo tutti i datasource usati dai timegraph per poter calcolare lo "stato" di ogniuno
		dg_ids = [ t.datagroup_id for t in selected_timegraph ]
		tg_ids = [ t.uuid         for t in selected_timegraph ]

		datasources = DataSource.objects.filter(datagroup__uuid__in=dg_ids)
		series = TimeGraphSerie.objects.filter(timegraph__uuid__in=tg_ids).select_related('timegraph')
		
		# Troviamo tutti i datasource dei datagroup dei timegraph
		ds_by_datagroup_id = {}
		for ds in datasources:
			dss = ds_by_datagroup_id.get(ds.datagroup_id, ds_by_datagroup_id.setdefault(ds.datagroup_id, {} ) )
			dss[ds.name] = ds

		# Troviamo tutti i datasource usati dalle serie dello stesso timegraph
		ds_by_timegraph = {}
		for s in series:
			dg_id = s.timegraph.datagroup_id
			tg_id = s.timegraph_id
			
			datasources = ds_by_datagroup_id.get(dg_id, {})
			ds = datasources.get(s.variable_value, None)
			if ds:
				dss = ds_by_timegraph.get(tg_id, ds_by_timegraph.setdefault(tg_id, set()))
				dss.add( ds )


		for t in selected_timegraph:
			
			datasources = ds_by_timegraph.get(t.uuid, [])
			
			state = t.get_state_from_datasources( datasources )

			entry = {
				'type'             : 'timegraph',
				'id'               : t.uuid,
				'state'	           : state,
				'element_kind'     : t.element.path,
				'element_kind_url' : t.element.page_url, #reverse('webui.views.index', kwargs={'rt': ELEMENT_KIND_STR[t.element.kind], 'uuid': t.element.uuid}),
				'title'            : t.expanded_title(cache_manager=cache_manager), 
				'statuslastchange' : '',
				'datagroup'        : '',
				'lastdone'         : '',
			}

			if t.element.kind == ELEMENT_KIND_IFACE:
				add_linked_data(entry, t.element)

			entities.setdefault(section, []).append(entry)


	for k, v in entities.items():

		section_title = TITLES[k]

		section_data = {
			'sectionid'   : k,
			'title'       : section_title,
			'entries'     : v,

			'tot_entries' : tot_counters[k],
			'tot_pages'   : page_numbers[k],

			'page'        : page,
			'pagesize'    : pagesize,
			# used to simplify timegraph table in view
			'istimegraph' : k == 'timegraph' or False,
		}
		sections.append(section_data)


	data = {
		'sections': sections,
		'search'  : search
	}

	return data


