# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import time
import json
import base64

from django.conf import settings

from core.compatibility import reverse
from django.http import HttpResponse
from django.template import RequestContext, loader, Context

from core.models import UserCustomSetting

from webui.models import *
from webui.models import webui_load_page_extension

from webui.api import format_html

from webui.api.Data.logs import LogOptions

from webui.api.widgets import WIDGET_CONFIG

from webui.api.widgets import FrameWidget
from webui.api.widgets import CustomDivWidget
from webui.api.widgets import AlarmsListWidget

from webui.api.extra_menu import get_external_menu_entries

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

"""
FIX: Per il momento i layout non e' altro che una struttura di dizionari con campi.

	 Sarebbe stato meglio definire delle classi con proprieta' e metodi in modo
	 da poter aggiungere metodi.
"""

PAGELAYOUT_STYLE_NORMAL       = 'box-body'               # layout visibile con sfondo scuro + bordo
PAGELAYOUT_STYLE_TRANSPARENT  = 'box-body-transparent'   # layout invisibile

def PageLayout(id, orientation, items=[], hidden=False, css_class=PAGELAYOUT_STYLE_NORMAL, min_height=None, style_params=[]):
	return {
		'id'          : id,
		'orientation' : orientation,
		'items'       : items,
		'hidden'      : hidden,
		'css_class'   : css_class,  
		'min_height'  : min_height,
		'style_params': style_params,
	}

def VerticalLayout(id, items=[], hidden=False, css_class=PAGELAYOUT_STYLE_NORMAL, min_height=None, style_params=[]): 
	return PageLayout(id, 'v', items=items, hidden=hidden, css_class=css_class, min_height=min_height, style_params=style_params)

def HorizontalLayout(id, items=[], hidden=False, css_class=PAGELAYOUT_STYLE_NORMAL, min_height=None, style_params=[]): 
	return PageLayout(id, 'h', items=items, hidden=hidden, css_class=css_class, min_height=min_height, style_params=style_params)

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

def PageTab(name='', title='', icontype='', iconvalue='', active_tab=False, enabled=True, order=0):

	t = {
		'name'       : name      , # name for tab
		'title'      : title     , # title human readable
		'icontype'   : icontype  , # choose icon: fonticon, image
		'iconvalue'  : iconvalue , # icon url, run for icon and image
		'active_tab' : active_tab, # active tab information
		'enabled'    : enabled   ,
		'order'      : order     , # Tab position 
	}
	return t

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

class BasePage(object):

	TAB_GENERAL = 'general'

	LAYOUT_INFO = 'L-info'

	PAGE_HTML_TEMPLATE = 'html/index.html'

	def __init__(self, rid, tenant, global_settings):

		user = global_settings['user']
		
		self.app_name          = 'webui'

		self.pagetype          = None             # stringa con il tipo della pagina (tenant, site, userpage, configuration, ecc..)
		self.tenant            = tenant           # tenant di riferimento

		self.rid               = rid              # id della risorsa della pagina (uuid o altro identificativo univoco)                
		self.resource          = None             # Variabile interna che condifica la risorsa della pagina               

		self.global_settings   = global_settings

		self.user_settings     = {}
		self.user_capabilities = {}

		self.admin_required    = False
		self.has_nav_bar       = True             # Show navigation bar on the left side
		self.has_search        = True

		self.show_explore_bars = False

		self.user_capabilities = user.get_capabilities(self.tenant)
		self.user_settings     = UserCustomSetting.get_user_settings(user, tenant=tenant)

		self.cache_manager     = global_settings['cache_manager']

		self.request_user      = None

		# List of page extensions from plugins/subapps
		self.extensions        = []

		return

	def _load_page_extensions(self):

		# Signal subapps to load page extension
		ret_data = webui_load_page_extension.send(None, page=self) 

		for plugin, extension_object in ret_data:
			if extension_object is not None:
				self.extensions.append(extension_object)

	def renderResponse(self, request, page_settings):
		
		self._load_page_extensions()

		error_response = self.userCanAccess(request)
		if error_response is not None:
			return error_response

		# Inizializziamo la pagina prima del render (carichiamo dati utili al rendering)
		self.initRender(request)

		# Render della pagina
		html_text = self.renderHTML(request, page_settings)

		return HttpResponse(html_text)

	def userCanAccess(self, request):

		user = request.user

		if self.admin_required:
			if not user.is_superuser:
				context = {
				}
				#c = RequestContext(request, context) django 1.11
				t = loader.get_template( 'html/user_not_admin_error.html' )
				html_text = t.render(context, request)	 # django 1.11
				return HttpResponse(html_text)

		return None

	def initRender(self, request):

		self.request_user       = request.user

		self.show_explore_bars  = self.user_settings.get('page.show_explore_bars', False) 

	def get_title(self):
		return ''

	def renderHTML(self, request, page_settings):

		user = request.user

		page_title = self.get_title()

		#
		# Carico la strutura della pagina
		#
		page_structure   = self.render()

		page_structure_encoded   = base64.b64encode(json.dumps(page_structure))

		#
		# Render della pagina vero e proprio
		#
		page_settings.update({
			'page_title'                    : page_title,

			'extra_menus'                   : self.get_additional_menus(),

			'page_structure'                : page_structure,
			'page_structure_encoded'        : page_structure_encoded,

			'has_nav_bar'                   : 1 if self.has_nav_bar                                   else 0,
			'has_search'                    : 1 if self.has_search                                    else 0,
			'show_explore_bars'             : 1 if self.has_nav_bar and self.show_explore_bars        else 0,
			"show_system_status_on_web"     : 1 if 'gui.view.system_status' in self.user_capabilities else 0,

			'additional_javascript_files'   : self.get_additional_jsfile_urls(),
			'additional_header_sections'    : self.loadAdditionalHeaderSections(),
			'additional_tenant_menu_entries': self.loadAdditionalTenantMenuEntries(),
		})

		page_settings.update( self.getAdditionalRenderSettings(request) )

		#
		# Load additional header sections (HTML) from sub-apps
		#

		#
		# Render page
		#
		#c = RequestContext(request, page_settings)   django 1.11
		t = loader.get_template( self.PAGE_HTML_TEMPLATE )
		html_text = t.render(page_settings, request)	# django 1.11

		return html_text
		
	def loadAdditionalTenantMenuEntries(self):

		entries = []

		for extension in self.extensions:
			entries += extension.get_extension_tenant_menu_entries()

		return entries

	def loadAdditionalHeaderSections(self):

		tenant   = self.tenant
		resource = self.resource or self.tenant
		#rt       = resource.resource_type
		#uuid     = resource.resource_id

		r = []

		for extension in self.extensions:
			safe_html = extension.get_extension_additional_header_section_html()

			if safe_html is not None:
				r.append( safe_html )

		return r

	def render(self):

		#
		# Load Tabs
		#
		tabs   = self.get_tabs()

		# sort tabs list before render
		tabs = sorted(tabs, cmp=lambda tab1,tab2: tab1['order'] - tab2['order'])

		#
		# Load layout
		#
		layout = self.get_page_layout()

		#
		# Load widgets
		#
		widgets = self.get_widgets()

		data = {
			'name'   : self.pagetype,   

			'layout' : layout,
			'widgets': widgets,
			'tabs'   : tabs,

			'bar'    : True,
		}

		return data

	#
	# Metodi pubblici per costruire la pagina. 
	# Le sottoclassi devono sovrascrivere questi metodi per personalizzare
	# l'aspetto/funzionamento della pagina.
	#
	def get_tabs(self):

		tabs = []

		# Default tab
		tabs.append(PageTab(**{
				'name'      : self.TAB_GENERAL,  # name for tab
				'title'     : 'General',         # title human readable
				'icontype'  : 'fonticon',        # choose icon: fonticon, image
				'iconvalue' : 'icon-home',       # icon url, run for icon and image
				'active_tab': True,              # active tab information
				'order'     : -100,
		}))

		#
		# Load tabs from plugins
		#
		for extension in self.extensions:
			tabs += extension.get_extension_tabs()

		return tabs

	def get_page_layout(self):

		tabs_layout = {}

		# Default layout/panel
		tabs_layout[self.TAB_GENERAL ] = VerticalLayout(self.LAYOUT_INFO)

		#
		# Load layouts from plugins
		#
		for extension in self.extensions:
			layouts = extension.get_extension_layouts()
			tabs_layout.update( layouts )

		return tabs_layout


	def get_widgets(self):

		widgets = []

		#
		# Load widgets from plugins
		#
		for extension in self.extensions:
			widgets += extension.get_extension_widgets()

		return widgets

	def getAdditionalRenderSettings(self, request):

		user = request.user

		show_tabs_title           = self.user_settings.get('page.tabs.show_title', True)

		return {
			"show_tab_title"           : show_tabs_title,   # Mostra il icona+ titolo nei tab. Default = no
		}

	def get_additional_jsfile_urls(self):

		urls = []

		# Default javascript for the page
		page_js_file = 'page/%s.js' % (self.pagetype)

		try:
			if self.app_name:
				# This won't work only for for "webui" sub-app, which has ".../web/" prefix and not ".../webui/".
				url = reverse('static.file.%s' % (self.app_name) , kwargs={'path': page_js_file})
			else:
				url = reverse('static.file.serve' , kwargs={'path': page_js_file})
		except Exception, e:
			url = None

		if url is None:
			url = reverse('static.file.serve' , kwargs={'path': page_js_file})

		urls.append(url) 

		#
		# Load additional urls from plugins
		#
		for extension in self.extensions:
			urls += extension.get_extension_additional_jsfile_urls()

		#for u in urls:
		#	print urls

		return urls

	def get_additional_menus(self):

		if self.resource is None:
			return []

		data = get_external_menu_entries(self.tenant, self.resource.resource_type, self.resource.resource_id)

		return data

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

class AlarmsCountPage(object):

	#
	# Classi di comodo per estendere la BasePage con metodi 
	# comuni alle pagine che devono contenere
	#
	# 	1) Tabella conteggio allarmi
	#
	# 	2) Lista allarmi in corso
	#
	def get_alarms_tab_container(self):
		raise NotImplemented()

	def create_alarms_section_widgets(self):

		widgets = []

		tab, container = self.get_alarms_tab_container()

		widget_alarms_counters_widget = self.create_alarms_counters_widget()
		widget_alarms_details_widget  = self.create_alarms_details_widget()

		rt  = self.resource.resource_type 
		rid = self.resource.resource_id   

		hideable = True
		visible  = False

		if settings.TENANT_HOMEPAGE_ALARMS_SECTION_PAGE_WIDE:
			hideable = False
			visible  = True

		wid = 'framewidget_' + rt + '_' + rid;
		widget = FrameWidget( 
			id        = wid, 

			tab       = tab,
			container = container,

			rdata = {
				"struct": [
					{
						"type"    : "widget",
						"contents": widget_alarms_counters_widget
					},
					{
						"type"     : "wframesub",
						"hideable" : hideable,
						"updateAll": True,
						"visible"  : visible,
						"buttonpos": 0,
						"contents" :
						[
							{
								"type"    : "widget",
								"contents": widget_alarms_details_widget
							},
						]
					}
				]
			},
			autorefresh_button = True,
		)
		w = widget.to_dict()
		widgets.append(w)

		return widgets

	def create_alarms_counters_widget(self):

		tab, container = self.get_alarms_tab_container()

		rt  = self.resource.resource_type 
		rid = self.resource.resource_id   

		w = CustomDivWidget(
			id          = "alarmscount_%s_%s" % (rt, rid),

			tab         = tab,
			container   = container,

			rdata       = None,

			remote      = '%s_%s.alarmscount' % (rt, rid),
			datasource  = '%s_%s.alarmscount' % (rt, rid),

			rtimer      = settings.WEBUI_REFRESH_TIMERS['alarms_count'],

			showspinner = True,
		)

		return w.to_dict()

	def create_alarms_details_widget(self):

		tab, container = self.get_alarms_tab_container()

		rt  = self.resource.resource_type 
		rid = self.resource.resource_id   

		w = AlarmsListWidget(
			id         = "alarmslist_%s_%s" % (rt, rid),

			tab        = tab,
			container  = container,

			rdata      = None,

			remote     = '%s_%s.alarmslist' % ( rt, rid),
			datasource = '%s_%s.alarmslist' % ( rt, rid),

			rtimer     = settings.WEBUI_REFRESH_TIMERS['alarms_list'],

			# reload spinner
			showspinner = True,
		)

		json = w.to_dict()

		return json

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

class TroubleshootPage(object):

	def get_troubleshoot_tab_container(self):
		raise NotImplemented()

	def create_troubleshoot_area_widget(self):

		tab, container, switch_container_id = self.get_troubleshoot_tab_container()

		rt  = self.resource.resource_type 
		rid = self.resource.resource_id   

		widgets = []

		w = CustomDivWidget(
			id = "troubleshoot-button",

			tab       = tab,
			container = container, 

			rdata = {
				'structure': {
					'template':  format_html("""
						<div class="button active" data-show="%s" data-hide="self" data-show-filter="widget"> Hide Troubleshooting View </div>
						""" % (switch_container_id))
				}
			}
		)

		widget = w.to_dict()
		widgets.append(widget)

		w = WIDGET_CONFIG({
			'id'       : 'troubleshoot-area',
			'wtype'    : ["troubleshoot"],

			'tab'      : ' '.join([tab,]),
			'container': container,

			'remote'   :  "%s_%s.%s" % (rt, rid,  'troubleshoot'),
			'dataset'  :  "%s_%s.%s" % (rt, rid,  'troubleshoot'),

			'showspinner': True,
		})
		widgets.append(w)

		return widgets

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

class LogPage(object):
	
	def get_log_tab_layout(self):
		raise NotImplemented()

	def get_additional_jsfile_urls(self):
		#page_js_file = 'page/log.js'
		#url = reverse('static.file.serve' , kwargs={'path': page_js_file})
		#return [ url, ]
		return []

	def get_tab_def(self):

		tab, container = self.get_log_tab_layout()

		return  [ 
			PageTab(**{
				'name'      : tab,  # name for tab
				'title'     : 'Logs',  # title human readable
				'icontype'  : 'fonticon', # choose icon: fonticon, image
				'iconvalue' : 'icon-list', # icon url, run for icon and image
				'active_tab': False, # active tab information
				'order'     : 10000,
			})
		]

	def get_layouts(self):

		tab, container = self.get_log_tab_layout()

		layouts = {}
		layouts[tab] = HorizontalLayout(container )

		return layouts

	def get_log_widgets(self):

		widgets = []

		tab, container = self.get_log_tab_layout()

		rt  = self.resource.resource_type 
		rid = self.resource.resource_id   

		w = WIDGET_CONFIG({
			'id'       : 'log-%s' % (self.resource.uuid),
			'wtype'    : ["log"],

			'tab'      : ' '.join([tab,]),
			'container': container,

			'remote'   :  "%s_%s.%s" % (rt, rid,  'logs'),
			'dataset'  :  "%s_%s.%s" % (rt, rid,  'logs'),
			
			'rtimer'   :  settings.WEBUI_REFRESH_TIMERS['logs'],

			'autorefresh_button': True,
			'showspinner'       : True,

			'rdata'    : {
				'options': LogOptions(), #None, # rdata,
				'init': True,
			},
		})
		widgets.append(w)

		return widgets
