# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import time

from django.conf import settings

from core.compatibility import reverse

from sanet_common.sanetdefs.constants import *

from core.models import TimeGraph 
from core.models import MonitoredElement

from core.statecolor import get_state_color

from tags.models import ElementTag, ELEMENT_KIND_TO_TAG_ELEMENTTYPE

from webui.models import *

from webui.api import format_html 
from webui.api.constans import ICON_WARNING

from webui.api.widgets import WIDGET_CONFIG

from webui.api.widgets import TimeGraph3Widget
from webui.api.widgets import FrameWidget
from webui.api.widgets import CustomDivWidget 

from webui.api.widgets import create_element_badge_widget
from webui.api.widgets import widget_list_of_datagroup

from webui.api.Pages import PageTab
from webui.api.Pages import VerticalLayout, HorizontalLayout
from webui.api.Pages import BasePage
from webui.api.Pages import AlarmsCountPage
from webui.api.Pages import TroubleshootPage
from webui.api.Pages import LogPage

from webui.api.utils import render_template

from webui.api.constans import ICON_WARNING

from webui.api.widgets import widget_resource_section_title

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

class ElementPage(BasePage, AlarmsCountPage, TroubleshootPage, LogPage):

	# La pagina di un elementi di monitoraggio deve sempre caricare almeno
	# l'lemeneto del tipo giusto. Ogni sotto classe di questa pagina
	# cambiano la classe del modello da usare, ma non cambia il caricamento
	# dell'oggetto.
	_element_class = None

	TAB_LOG             = 'log'
	LAYOUT_LOG          = 'L-log'

	LAYOUT_INFO         = 'L-info' 
	LAYOUT_ALARMS       = 'L-alarms'
	LAYOUT_PRIMARYGRAPH = 'L-primarygraph'
	LAYOUT_MONITOR      = 'L-monitoringarea'
	LAYOUT_TROUBLE      = 'L-troubleshootarea'

	TAB_CONFIG          = 'config'
	LAYOUT_CONFIG       = 'L-config'

	def __init__(self, rid, tenant, global_settings):

		super(ElementPage, self).__init__(rid, tenant, global_settings)

		# Carichiamo l'oggetto giusto per questa pagina
		self.element  = self._element_class.objects.get(uuid=self.rid)
		self.resource = self.element

		# Dati comuni utilizzati in piu' punti della pagina (cosi' ottimizziamo
		self.parent                = None
		self.conditions            = None
		self.conditions_by_element = {}
		self.primary_condition     = None


	def _prepare_common_data(self):
		"""
		Nella pagina vengono utilizzati molto spesso dati comuni:
		   * Le condition
		   * GLi elementi / datagroup collegati all'elemento visualizzato
		   * ecc.
		   
		Con questa funzione carichiamo una volta sola tutti i dati.
		
		"""
		#
		# Conditions
		#
		self.conditions = self.resource.get_all_conditions().select_related('element').all()

		for c in self.conditions:

			l = self.conditions_by_element.get(c.element_id, self.conditions_by_element.setdefault(c.element_id, []))
			l.append(c)

			if c.element_id == self.resource.uuid and c.primary == True:
				self.primary_condition = c

		if self.primary_condition:
			self.primary_condition.load_fields_from_cache(self.cache_manager)

		#
		# Parent element
		#
		if isinstance(self.element, MonitoredElement):
			parent = self.element.parent
			if parent is not None:
				self.parent = parent.get_specific_element()

	def userCanAccess(self, request):

		error_response = super(ElementPage, self).userCanAccess(request)

		#
		# FIX: verify access  self.element
		#

		return error_response

	def get_title(self):
		return self.resource.name

	def initRender(self, request):

		super(ElementPage, self).initRender(request)

		self.layout_trouble    = self.user_settings.get('page.element.troubleshoot_view', 0) == 1
		self.layout_monitor    = not self.layout_trouble

	def get_log_tab_layout(self):
		return self.TAB_LOG, self.LAYOUT_LOG

	def get_tabs(self):

		tabs = super(ElementPage, self).get_tabs()

		#
		# Config element tab
		#
		tab_enabled = 'core.element.conf' in self.user_capabilities

		tabs += [
			PageTab(**{
				'name'      : self.TAB_CONFIG, # name for tab
				'title'     : 'Configuration', # title human readable
				'icontype'  : 'fonticon',      # choose icon: fonticon, image
				'iconvalue' : 'icon-cog',      # icon url, run for icon and image
				'active_tab': False,           # active tab information
				'enabled'   : tab_enabled,
				'order'     : 1000,
			})
		]

		tabs.extend( LogPage.get_tab_def(self) )

		return tabs

	def get_page_layout(self):

		tabs_layout = super(ElementPage, self).get_page_layout()

		# Default layout per tutte le pagine degli elementi.

		alarm_h = 130
		graph_h = 200
		info_h  = alarm_h + graph_h + 10

		tabs_layout[self.TAB_GENERAL ] = VerticalLayout('L1', [
							HorizontalLayout('L1-1', [
									VerticalLayout(self.LAYOUT_INFO                  , min_height=info_h) ,  # details, notes, tags
									VerticalLayout('L1-1-X', [
										HorizontalLayout(self.LAYOUT_ALARMS      , min_height=alarm_h),  # alarms
										HorizontalLayout(self.LAYOUT_PRIMARYGRAPH, min_height=graph_h),  # primary graph
									])
							]),
							HorizontalLayout(self.LAYOUT_MONITOR, hidden=not self.layout_monitor),                   # monitoring
							HorizontalLayout(self.LAYOUT_TROUBLE, hidden=not self.layout_trouble)                    # troubleshoot
						])
		
		tabs_layout[self.TAB_CONFIG   ] = VerticalLayout(self.LAYOUT_CONFIG)

		#tabs_layout.update( LogPage.get_layouts(self) )
		#tabs_layout[self.TAB_GENERAL ] = VerticalLayout(self.LAYOUT_PRIMARYGRAPH)

		return tabs_layout

	#-------------------------------------------------------------------------------#

	# Metodi di  AlarmsCountPage
	def get_alarms_tab_container(self):
		return self.TAB_GENERAL, self.LAYOUT_ALARMS

	#-------------------------------------------------------------------------------#

	def get_widgets(self):

		self._prepare_common_data()

		widgets = []


		widgets += self.get_info_widgets()

		widgets += self.get_alarms_widgets()

		widgets += self.create_primary_graph_widget()

		widgets += self.create_monitoring_area_widget()

		widgets += self.create_troubleshoot_area_widget()

		widgets += self.get_config_widget()

		widgets += super(ElementPage, self).get_widgets()

		return widgets

	def get_config_widget(self):

		widgets = []

		w = widget_resource_section_title(self.resource, 'Configuration', tab=self.TAB_CONFIG, container=self.LAYOUT_CONFIG)
		widgets.append(w)

		#
		# Main configuration section
		#
		w = WIDGET_CONFIG({
			'id'                 : self.rid + "-configuration",
			'wtype'              : ["configelement"],

			'tab'                : self.TAB_CONFIG,
			'container'          : self.LAYOUT_CONFIG,

			'height'             : 'min',
			'rdata'              : None,

			'rtimer'             : 0,
			'autorefresh_button' : True,

			'remote'             : "%s_%s.%s" % (self.element.resource_type, self.element.resource_id, 'element_config_data'),
		})
		widgets.append(w)

		#
		# Main configuration section
		#
		#w = WIDGET_CONFIG({
		#	'id'                 : self.rid + "-configuration-customfields",
		#	'wtype'              : ["configelement_customfields"],
		#
		#	'tab'                : self.TAB_CONFIG,
		#	'container'          : self.LAYOUT_CONFIG,
		#
		#	'height'             : 'min',
		#	'rdata'              : None,
		#
		#	'rtimer'             : 0,
		#	'autorefresh_button' : True,
		#
		#	'remote'             : "%s_%s.%s" % (self.element.resource_type, self.element.resource_id, 'element_config_data_customfields'),
		#})
		#widgets.append(w)


		return widgets

	def get_advanced_details_from_extensions(self, n):

		details = []

		# infors from plugins
		for extension in self.extensions:

			extension_details = extension.get_element_advanced(n)

			details += extension_details

		return details



	def get_info_widgets(self):

		widgets = []

		#
		# INFO widgets
		#
		widget_info  = self.create_global_info_widget(self.element)

		# infors from plugins
		extensions_info_widgets = []
		for extension in self.extensions:
			extensions_info_widgets += extension.get_info_widgets()
		
		all_widget_infos = [ widget_info ] + extensions_info_widgets

		#
		# ADVANCED INFO widgets
		#
		widget_advanced_info  = self.create_advanced_info_widget(self.element)

		# infors from plugins
		advanced_extensions_info_widgets = []
		for extension in self.extensions:
			advanced_extensions_info_widgets += extension.get_advanced_info_widgets()

		# FIX: remove from here
		widget_tags  = self.create_tags_widget(self.element)

		all_advanced_widget_infos = [ widget_advanced_info ] + advanced_extensions_info_widgets + [ widget_tags ]

		#
		# SEPARATOR
		#
		widget = FrameWidget(
			id = 'a',

			tab       = self.TAB_GENERAL,
			container = self.LAYOUT_INFO,

			rdata = {
				"struct": 
					[ { "type"    : "widget", "contents": widget_def } for widget_def in all_widget_infos ]
					+
					[
						{
							"type"     : "wframesub",
							"hideable" : True,
							"visible"  : False,
							#"buttonpos": 0, # Serve?
							"contents" : [ { "type": "widget", "contents": widget_def } for widget_def in all_advanced_widget_infos ]
						}
					]
			},
		)

		widgets.append(widget.to_dict())

		return widgets

	def create_global_info_widget(self, n, wid="node_info_details_widget"):

		critical_priority = self.global_settings['critical_priority']
		minimum_priority  = self.global_settings['minimum_priority' ]

		#
		# Stato elemento
		#
		state, critical_alarms_count, minimum_dn_count, dn_count = n.get_state_and_conditions(
			critical_priority = critical_priority, 
			minimum_priority  = minimum_priority, 
			loaded_conditions = self.conditions,
		)

		ignored_dn_count = dn_count - minimum_dn_count

		icon_url = n.get_icon_url() + get_state_color(state)

		icon_overlay_url = None
		if critical_alarms_count > 0:
			icon_overlay_url = ICON_WARNING

		#
		# Dettagli elemento
		#
		details = self.get_info_details(n)

		details.append({ 'name': 'Problems', 'value': "%s (>= critical), %s (>= minimum) / %s (totals)" % (critical_alarms_count, minimum_dn_count, dn_count) })

		c = {
			'element'          : n,
			'primary_condition': self.get_primary_condition(),
			'element_state'    : state,
			'ignored_dn_count' : ignored_dn_count,
			
			'icon_url'         : icon_url,
			'icon_overlay_url' : icon_overlay_url,
			
			'details'          : details 
		}

		#
		# Stato nodo (parent)
		#
		parent = self.parent
		if parent:
			parent_conditions = parent.get_all_conditions().select_related('element').all()

			parent_state, critical_alarms_count, minimum_dn_count, dn_count = parent.get_state_and_conditions(
				critical_priority = critical_priority, 
				minimum_priority  = minimum_priority, 
				loaded_conditions = parent_conditions
			)

			c.update({
				'parent'       : parent,
				'parent_state' : parent_state,
			})

		#
		# Render widget
		#
		html = render_template( 'html/parts/element_details.html', c )
		
		w = CustomDivWidget(
			id        = wid,
			tab       = self.TAB_GENERAL,
			container = self.LAYOUT_INFO,
			width     = 8,
			height    = 2,
			rdata = {
				'structure': {
					'template': format_html( html  )
				}
			}
		)
		return w.to_dict()

	def get_primary_condition(self):
		return self.primary_condition

	def get_info_details(self, n):
		return []

	def create_advanced_info_widget(self, n):
		
		details = self.get_advanced_details(n)
		
		html = render_template( 'html/parts/advanced_details.html', {
			'details'           : details,
			'primary_condition' : self.get_primary_condition(),
		})
		
		w = CustomDivWidget(
			id        = "node_info_details_extra_widget",
			tab       = self.TAB_GENERAL,
			container = self.LAYOUT_INFO,
			width     = 8,
			height    = 2,
			rdata     = {
				'structure': {
					'template': format_html( html  )
				}
			}
		)
		return w.to_dict()

	def get_advanced_details(self, n):
		details = []
		details.append({ 'name': 'Advanced'       , 'value': 'foo'  })
		return details

	def create_tags_widget(self, n):

		# FIX: RENDERE QUESTO WIDGET AUTO-REFRESHANTE. QUESTA VA NEL CODICE DELLA REMOTE PER GESTIRE L'AGGIORNAMENTO.

		etype = ELEMENT_KIND_TO_TAG_ELEMENTTYPE[n.kind]
		eid   = n.uuid

		tags = Tag.get_visible_by_user(self.tenant, self.request_user)
		ets = ElementTag.objects.filter(elementtype=etype, elementid=eid, tag__in=tags).select_related('tag', 'tenant').all()

		tags = []
		for et in ets:

			path = []
			for path_tag in et.tag.get_tags_path():

				path.append({
					'uuid' : et.uuid,
					'tagid': path_tag.uuid, 
					'name' : path_tag.name,
				})

			entry = {
				'uuid': et.uuid,
				'path': path,
			}

			tags.append(entry)
		
		#
		# Can untag?
		#
		user = self.request_user
		
		can_edit = True
		can_edit = can_edit and ('tags.tag.tag_element' in user.get_capabilities(self.element.tenant))
		can_edit = can_edit and self.element.can_write(user)

		rdata = {
			'tags'    : tags,
			'can_edit': can_edit,
		}

		widget_test = WIDGET_CONFIG({
			'id'       : "%s_%s_%s" % (n.resource_type, n.resource_id, 'tags'),
			'wtype'    : ["taglist"],

			'tab'      : ' '.join([self.TAB_GENERAL,]),
			'container': self.LAYOUT_INFO,

			'rdata'    : rdata,
		})
		return widget_test

	#-----------------------------------------------------------------------------#

	def get_alarms_widgets(self):

		tab, container = self.get_alarms_tab_container()

		widgets = []

		#
		# TABELLA PRIORITA' E DETTAGLIO ALLARMI
		#
		widgets += self.create_alarms_section_widgets()

		return widgets

	#-----------------------------------------------------------------------------#

	def create_primary_graph_widget(self):

		widgets = []

		n = self.element
		if isinstance(n, MonitoredElement):

			#n = n.get_specific_element()

			timegraphs = TimeGraph.objects.filter(element=n, primary=True).all()

			if len(timegraphs) > 0:

				tg = timegraphs[0]


				w = TimeGraph3Widget(
					tg,
					id        = self.rid+"_primary_graph",

					tab       = ' '.join([self.TAB_GENERAL,]),
					container = self.LAYOUT_PRIMARYGRAPH,

					rtimer    = settings.WEBUI_REFRESH_TIMERS['primary_graph'], 
				)
				w = w.to_dict();

				widgets.append(w)

			else:
				w = CustomDivWidget(
					id        = self.rid+"_primary_graph",

					tab       = self.TAB_GENERAL,
					container = self.LAYOUT_PRIMARYGRAPH,

					height    = 1,

					rdata     = {
						'structure': {
							'template': "* No graph configured for this element *"
						}
					}
				)
				w = w.to_dict()
				widgets.append(w)

		return widgets

	#-----------------------------------------------------------------------------#

	def create_monitoring_area_widget(self):

		widgets = []

		w = CustomDivWidget(
			id        = "details",

			tab       = self.TAB_GENERAL,
			container = self.LAYOUT_MONITOR,

			rdata     = {
				'structure': {
					'template':  format_html("""
						<table style="width:100%;">
							<tr>
								<td>
									<div style="text-align:right;">
											<span style="font-size:x-large;">Dati Monitoraggio</span> 
									</div>
								</td>
								<td>
									<div class="button" data-show=""" + self.LAYOUT_TROUBLE+""" data-hide="self">Troubleshooting View</div>
								</td>
							</tr>
						</table>
					""")
				}
			}
		)
		widget = w.to_dict()
		widgets.append(widget)

		#
		# Widget contenente nodi, allarmi, ecc....
		#
		widgets += self._monitoring_area_widget_node()

		return widgets

	def _monitoring_area_widget_node(self):

		critical_priority = self.global_settings['critical_priority']
		minimum_priority  = self.global_settings['minimum_priority' ]

		widgets = []

		# header nodo
		w = CustomDivWidget(
			id        = "details",

			tab       = self.TAB_GENERAL,
			container = self.LAYOUT_MONITOR,

			rdata     = {
				'structure': {
					'template':  format_html("""
						<div class="section_hd_container">
							<span class="section_header">%s</span>
						</div>
						%s
					""" % (self.pagetype.title(), ""))
				}
			},
			role = 'section-divider',
		)
		widget = w.to_dict()
		widgets.append(widget)

		loaded_conditions = self.conditions

		# get list of datagroup
		widgets += widget_list_of_datagroup(self.rid, 
		                                    self.LAYOUT_MONITOR, 
		                                    self.TAB_GENERAL, 
		                                    critical_priority = critical_priority, 
		                                    minimum_priority  = minimum_priority, 
		                                    loaded_conditions = loaded_conditions, 
		                                    cache_manager     = self.cache_manager,
		                                   )

		# Widgetgs for every element
		for kind in (ELEMENT_KIND_IFACE, ELEMENT_KIND_STORAGE, ELEMENT_KIND_SERVICE, ELEMENT_KIND_DEVICE):
			# add list of element
			widgets += self._monitoring_area_widget(self.conditions_by_element, kind, minimum_priority=minimum_priority, critical_priority=critical_priority)

		return widgets

	def _monitoring_area_widget(self, conditions_by_element, kind, minimum_priority=None, critical_priority=None):

		elements = MonitoredElement.kind_to_class(kind).objects.filter(parent=self.resource).all()

		# human readable element type
		element_type_str = ELEMENT_KIND_STR[kind]

		widgets = []
		if elements:

			title = element_type_str.title()

			# header interface
			w = CustomDivWidget(
				id        = "%s_section_title" % (element_type_str),
				tab       = self.TAB_GENERAL,
				container = self.LAYOUT_MONITOR,
				width     = 'max',
				rdata     = {
					'structure': {
						'template':  format_html("""
							<div class="section_hd_container">
								<span class="section_header">%(title)s</span>
							</div>
						""" % {'title': title } )
					},
				},
				role = 'section-divider',
			)
			widget = w.to_dict()
			widgets.append(widget)

			for element in elements:

				#
				# Calcolo Stato
				#
				conditions = conditions_by_element.get(element.uuid, [])

				#status, critical_alarms_count, alarms_count, dn_count = element._get_state_and_conditions(conditions, critical_priority=critical_priority, minimum_priority=minimum_priority)
				#if (critical_alarms_count > 0):
				#	status = CSTATE_TO_CODENAME(CSTATE_DN)
				status, critical_alarms_count, alarms_count, dn_count = element.get_state_and_conditions(critical_priority=critical_priority, minimum_priority=minimum_priority, loaded_conditions=conditions)
				
				icon_url              = element.get_icon_url_with_state(status)

				icon_overlay_url      = ''
				if critical_alarms_count > 0:
					icon_overlay_url = ICON_WARNING

				#
				# Widget
				#
				widgets += create_element_badge_widget(element, 
				                                       alarms_count, 
				                                       critical_alarms_count, 
				                                       icon_url, 
				                                       icon_overlay_url, 
				                                       status, 
				                                       self.pagetype, 
				                                       self.TAB_GENERAL, 
				                                       self.LAYOUT_MONITOR, 
				                                       refreshtime=0, 
				                                       cache_manager=self.cache_manager
				)

		return widgets

	def get_troubleshoot_tab_container(self):
		return self.TAB_GENERAL, self.LAYOUT_TROUBLE, self.LAYOUT_MONITOR

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

class ElementConfigPage(object):

	TAB_CONFIG    = 'config'

	LAYOUT_CONFIG = 'L-config'

	def get_config_tab_layout(self):
		return self.TAB_CONFIG, self.LAYOUT_CONFIG

	def get_additional_jsfile_urls(self):
		page_js_file = 'page/_configelement.js'
		url = reverse('static.file.serve' , kwargs={'path': page_js_file})
		return [ url, ]

	def get_tab_def(self):

		tenant = self.global_settings['tenant']
		user   = self.global_settings['user']

		enabled = 'core.element.conf' in user.get_capabilities(tenant)

		return  [ 
			PageTab(**{
				'name'      : self.TAB_CONFIG,
				'title'     : 'Config',
				'icontype'  : 'fonticon',
				'iconvalue' : 'icon-wrench',
				'active_tab': False,
				'enabled'   : enabled,
			})
		]

	def get_layouts(self):
		layouts = {}
		layouts[self.TAB_CONFIG] = HorizontalLayout(self.LAYOUT_CONFIG )
		return layouts

	def get_config_widgets(self):

		widgets = []

		tab, container = self.get_config_tab_layout()

		rt  = self.resource.resource_type 
		rid = self.resource.resource_id   

		w = widget_resource_section_title(self.resource, 'Config', tab=tab, container=container)
		widgets.append( w )

		w = CustomDivWidget(
			tab       = tab,
			container = container,
			
			rdata     = {
				'structure': {
					'template':  self.get_config_form()
				}
			}
		)

		widget = w.to_dict()
		widgets.append(widget)

		return widgets

	def get_config_form(self):

		from webui.views.objects.configure import model_configure_forms

		element = self.resource.get_specific_element()

		action_url = reverse('webui.views.objects.configure.monitoredelement_config_fields', kwargs={'uuid': self.resource.uuid })

		form_type = model_configure_forms[ self.resource.__class__ ]
		form = form_type(element.tenant, instance=element)

		html = form.render_html(action_url)

		return html
