# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.


from django.db.models import Q

from django.core.paginator import Paginator


from core.models import MonitoredElement, Condition

from tags.models import TAG_FILTER

from webui.api import format_html

from webui.api.widgets import CustomDivWidget, create_label_widget

from webui.api.utils import render_template

from webui.api.Pages import BasePage, PageTab
from webui.api.Pages import VerticalLayout

from webui.views import search_names_by_ip

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

class Page(BasePage):

	TAB_SEARCH        = 'general'
	LAYOUT_SEARCH     = 'L-search'

	def __init__(self, rid, tenant, settings):

		super(Page, self).__init__(rid, tenant, settings)

		self.pagetype = 'searchkeys'
		
		self.admin_required = False

		self.term =  settings['request'].GET_OR_POST().get('term','').strip()

	def get_title(self):
		return "Search"

	def get_tabs(self):
		
		# define tabs element for page
		return  [
			
			PageTab(**{
				'name'      : self.TAB_SEARCH,       # name for tab
				'title'     : 'Search result', # title human readable
				'icontype'  : 'fonticon',       # choose icon: fonticon, image
				'iconvalue' : 'icon-search',  # icon url, run for icon and image
				'active_tab': True,            # active tab information
			}),
		]

	def get_page_layout(self):
		tabs_layout = {}
		tabs_layout[self.TAB_SEARCH] =  VerticalLayout(self.LAYOUT_SEARCH)
		return tabs_layout		

	def get_widgets(self):


		user   = self.global_settings['user']
		
		tenant = self.tenant
		
		widgets = []

		
		if self.term:

			#
			# Elements
			#
			elements    = MonitoredElement.objects.select_related('parent')
			elements    = TAG_FILTER(tenant, user, elements, "monitoredelement")
			
			filters = Q() | Q(name__icontains=self.term)|Q(parent__name__icontains=self.term)|Q(description__icontains=self.term)
			
			names = search_names_by_ip(tenant, self.term)
			if names:
				filters = filters | Q(name__in=names)

			elements = elements.filter(filters)
			
			elements = elements.order_by('kind', 'parent', 'name')

			p_elem = Paginator(elements, 50)

			#
			# Conditions
			#
			conditions = Condition.objects
			conditions   = TAG_FILTER(tenant, user, conditions, "condition")

			conditions = conditions.filter(Q(path__icontains=self.term))
			conditions = conditions.order_by('path')
			p_cond = Paginator(conditions, 50)

			c = {
				'term'      : self.term,
				'elements'  : p_elem.page(1),
				'conditions': p_cond.page(1),
			}

		else:
			c = {
				'term' : self.term,
				'elements' : None,
				'conditions' : None
			}


		html = render_template( 'html/parts/search_results.html', c )

		w = CustomDivWidget(
			id = "search_results",
			
			tab       = self.TAB_SEARCH,
			container = self.LAYOUT_SEARCH,

			rdata = {
				'structure': {
					'template': format_html( html )
				}
			}
		)
		
		widgets.append(w.to_dict())

		return widgets

	def render(self):

		tabs    = self.get_tabs()
		layout  = self.get_page_layout()
		widgets = self.get_widgets()

		page_struct =  {
			'name'   : self.pagetype,   
			'layout' : layout,
			'widgets': widgets,
			'tabs'   : tabs,
		}

		return page_struct

	def render2(self):

		widgets = []
		
		widgets.append( create_label_widget("SEARCH", tab=self.TAB_SEARCH, container=self.LAYOUT_SEARCH) )
		
		
		c = {
			'term': self.term
		}
		html = render_template( 'html/parts/search_results.html', c )
		
		w = CustomDivWidget(
			id = "search_results",
			
			tab       = self.TAB_SEARCH,
			container = self.LAYOUT_SEARCH,

			rdata = {
				'structure': {
					'template': format_html( html )
				}
			}
		)
		
		widgets.append(w.to_dict())		
		

		layout = {
			'general' : 
				VerticalLayout(self.LAYOUT_SEARCH, [
				]),
		}
				
		return {
			'name'   : self.pagetype,   
			'layout' : layout,
			'widgets': widgets,
			'tabs'   : self.tabs,
		}
