# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import os

from django.conf import settings

from core.models import Tenant, TenantSetting
from core.api.base import  tenant_get_media_url

from tags.models import *

from webui.models import *
from webui.api import *
from webui.api.utils import render_template

from webui.api.constans import ICON_LOADING
from webui.api.Data.badgeview import create_badgeview_object

from webui.api.widgets import WIDGET_CONFIG
from webui.api.widgets import FrameWidget
from webui.api.widgets import CustomDivWidget

from webui.api.Pages import BasePage
from webui.api.Pages import VerticalLayout, HorizontalLayout

from webui.api.Pages import AlarmsCountPage
from webui.api.Pages import TroubleshootPage
from webui.api.Pages import LogPage

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

class Page(BasePage, AlarmsCountPage, TroubleshootPage, LogPage):

	LAYOUT_INFO    = 'L-info'

	LAYOUT_ALARMS  = 'L-alarms'

	LAYOUT_TAGS    = 'L-tags'
	LAYOUT_TAGS2    = 'L-tags2'

	LAYOUT_TROUBLE = 'L-troubleshootarea'

	#TAB_CONFIGURE  = 'configure'
	#LAYOUT_CONFIG  = 'L-configure'

	TAB_LOG    = 'log'
	LAYOUT_LOG = 'L-log'

	def __init__(self, rid, tenant, global_settings):

		super(Page, self).__init__(rid, tenant, global_settings)

		self.pagetype = 'tenant'

		self.resource = tenant

		self.critical_priority = PRIORITY_MAX + 1  # by default the critical priority is not set ( )


	def initRender(self, request):
		super(Page, self).initRender(request)

		self.layout_trouble    = self.user_settings.get('page.tenant.troubleshoot_view', 0) == 1
		self.layout_monitor    = not self.layout_trouble

	def get_title(self):
		return self.resource.long_name

	def get_tabs(self):

		self.tabs = super(Page, self).get_tabs()

		user = self.global_settings['user']

		self.tabs.extend( LogPage.get_tab_def(self)        )

		return self.tabs


	def get_page_layout(self):

		layout    = super(Page, self).get_page_layout()

		first_row = []

		if settings.TENANT_HOMEPAGE_ALARMS_SECTION_PAGE_WIDE:
			first_row = [
				VerticalLayout(self.LAYOUT_ALARMS)
			]
		else:
			first_row = [
				VerticalLayout(self.LAYOUT_INFO  , min_height=170), # details
				VerticalLayout(self.LAYOUT_ALARMS, min_height=170)  # alarms
			]

		layout[self.TAB_GENERAL] = VerticalLayout('L1', [
			HorizontalLayout('L1-1', first_row),
			HorizontalLayout(self.LAYOUT_TAGS   , hidden=not self.layout_monitor),                 # tabs
			HorizontalLayout(self.LAYOUT_TROUBLE, hidden=not self.layout_trouble), 
		])

		layout.update( LogPage.get_layouts(self) )

		return layout

	def get_additional_jsfile_urls(self):
		urls = super(Page, self).get_additional_jsfile_urls()

		urls += LogPage.get_additional_jsfile_urls(self)

		return urls

	def get_widgets(self):

		user   = self.global_settings['user']

		critical_priority = self.global_settings['critical_priority']
		minimum_priority  = self.global_settings['minimum_priority']

		t1 = time.time()

		#
		# Carichiamo lo stato globale del tenant (dei root tag)
		#
		self.tenant.set_tag_visibility(user)

		tags = self.tenant.child_tags

		t2 = time.time()

		widgets = []
		
		widgets += super(Page, self).get_widgets()

		#
		# INFO widgets
		#
		widget_info = self.create_global_info_widget(self.tenant)

		# infors from plugins
		extensions_info_widgets = []
		for extension in self.extensions:
			extensions_info_widgets += extension.get_info_widgets()

		all_widget_infos = [ widget_info ] + extensions_info_widgets

		#
		# ADVANCED INFO widgets
		#
		widget_advanced_info_more = self.create_advanced_info_widget(self.tenant)

		# infors from plugins
		advanced_extensions_info_widgets = []
		for extension in self.extensions:
			advanced_extensions_info_widgets += extension.get_advanced_info_widgets()

		all_advanced_widget_infos = [ widget_advanced_info_more ] + advanced_extensions_info_widgets

		#
		# SEPARATOR
		#
		widget_expander = FrameWidget(
			tab       = self.TAB_GENERAL,
			container = self.LAYOUT_INFO,
			rdata = {
				"struct": 
					[ { "type"    : "widget", "contents": widget_def } for widget_def in all_widget_infos ]
					+
					[
						{
							"type"     : "wframesub",
							"hideable" : True,
							"visible"  : False,
							#"buttonpos": 0, # Serve?
							"contents" : [ { "type": "widget", "contents": widget_def } for widget_def in all_advanced_widget_infos ]
						}
					]
				#]
			},
		)
		widgets.append(widget_expander.to_dict())


		#
		# TABELLA PRIORITA' E DETTAGLIO ALLARMI
		#
		widgets += self.create_alarms_section_widgets()

		#
		# SEZIONE PER TROUBLE SHOOT AREA
		#
		w = CustomDivWidget(
			id = "details",
                
			tab       = self.TAB_GENERAL,
			container = self.LAYOUT_TAGS,
                
			rdata = {
				'structure': {
					'template':  format_html("""
						<div class="button" data-show="L-troubleshootarea" data-hide="self">
							Show Troubleshooting View
						</div>
					""")
				}
			}
		)
		widget = w.to_dict()
		widgets.append(widget)

		#
		# ROOT TAGS
		#
		widget_tags = self.tags_grid_widget(self.tenant, user=user)
		widgets.extend(widget_tags)

		#
		# Configure TAB
		#
		# troubleshooting area
		widgets_trouble = self.create_troubleshoot_area_widget()
		widgets.extend(widgets_trouble)

		#
		# Log Tab
		#
		widgets.extend( self.get_log_widgets() )

		t3 = time.time()

		return widgets


	def get_config_widgets(self):

		TAB       = self.TAB_CONFIGURE
		CONTAINER = self.LAYOUT_CONFIG

		widgets = []

		# AREA DATI

		configured_critical_level = self.global_settings['critical_priority']

		priority_levels = []
		for l in PriorityLabel.objects.order_by('min_level').all():
			priority_levels.append(l)
			if configured_critical_level is not None:
				if l.in_range(configured_critical_level):
					l.selected = True

		html = render_template( 'html/parts/tenant_settings.html', { 
			'tenant': self.tenant, 
			'show_all': True,
			'critical_level' : configured_critical_level,
			'priority_levels': priority_levels,
		})

		w = CustomDivWidget(
			id    = "user_data",
			tab       = TAB,
			container = CONTAINER,  
			rdata = {
				'structure': { 
					'template': format_html( html )
				}
			}
		)
		widgets.append( w.to_dict() )

		return widgets

	#-------------------------------------------------------------------------------#

	# Metodi di  AlarmsCountPage
	def get_alarms_tab_container(self):
		return self.TAB_GENERAL, self.LAYOUT_ALARMS

	def get_troubleshoot_tab_container(self):
		return self.TAB_GENERAL, self.LAYOUT_TROUBLE, self.LAYOUT_TAGS

	#-------------------------------------------------------------------------------#

	def get_log_tab_layout(self):
		return self.TAB_LOG, self.LAYOUT_LOG

	#-------------------------------------------------------------------------------#

	def create_global_info_widget(self, n):

		img_url = tenant_get_media_url(self.tenant.uuid, 'site.png')

		w = CustomDivWidget(
			id        = 'info',
			tab       = self.TAB_GENERAL,
			container = self.LAYOUT_INFO,

			rdata  = {
				"structure": {
					'template': 
						format_html('''
							<center>
								<img height="110" src="%s" title="%s" />
							</center>
						''' % (img_url, self.tenant.long_name or self.tenant.name))
				}
			}
		)

		return w.to_dict()

	def create_advanced_info_widget(self, n):

		details = []
		details.append({ 'name': 'Name'        , 'value': self.tenant.name })
		details.append({ 'name': 'Long name'   , 'value': self.tenant.long_name or ''})

		# FIX: usare una sola query con un groupby
		labels = {
			ELEMENT_KIND_NODE   : "Total nodes",
			ELEMENT_KIND_IFACE  : "Total intefaces",
			ELEMENT_KIND_STORAGE: "Total storages",
			ELEMENT_KIND_SERVICE: "Total service",
			ELEMENT_KIND_DEVICE : "Total devices",
		}
		for kind, num in self.tenant.get_counters().items():
			details.append({ 'name': '%s' % (labels[kind])       , 'value': num })

		details.append({ 'name': 'Total conditions'  , 'value': self.tenant.get_all_conditions().count() })

		critical_level = self.tenant.get_setting(TenantSetting.SETTING_CRITICAL_LEVEL)
		if critical_level:
			s = "%s (level: %s)" % (PriorityLabel.get_label(critical_level), critical_level, )
		else:
			s = None
			
		details.append({ 'name': 'Critical level'    , 'value': s })

		html = render_template( 'html/parts/advanced_details.html', {
			'details'      : details 
		} )

		w = CustomDivWidget(
			id = "tenant_%s_more_info" % (self.tenant.uuid),

			tab        = self.TAB_GENERAL,
			container  = self.LAYOUT_INFO,

			rdata = {
				'structure': {
					'template': format_html( html )
				}
			}
		)
		return w.to_dict()

	#-------------------------------------------------------------------------------#

	def tags_grid_widget(self, tenant, user=None):

		element = tenant

		#
		# Create badges without state for a quick view
		#

		fake_states = {}

		root_tags = element.child_tags.all()

		sections = []
		objects = []

		for t in root_tags:
			o = create_badgeview_object(fake_states, t, nostate_icon=ICON_LOADING, user=user) # FIX: la gestione delle icone fa veramente schifo cosi'
			objects.append(o)

		objects = sorted(objects, key=lambda x: x['status_order'])

		sections.append({'name':'Tags', 'objects': objects})

		data = {
			'etype'     : element.resource_type,
			'uuid'      : element.resource_id,
			'sections'  : sections,
			'can_untag' : False
		}

		#
		# Create widget
		#
		widgets = []

		rt  = element.resource_type
		rid = element.uuid

		tab = self.TAB_GENERAL
		container = self.LAYOUT_TAGS

		w = WIDGET_CONFIG({
			'id'       : 'tags-%s' % (self.resource.uuid),
			'wtype'    : ["badgeview"],

			'tab'      : ' '.join([tab,]),
			'container': container,

			'remote'   : "%s_%s.%s" % (rt, rid,  'badgeview'),
			'dataset'  : "%s_%s.%s" % (rt, rid,  'badgeview'),

			'rtimer'   : settings.WEBUI_REFRESH_TIMERS['tags'],

			'rdata'    : data,
		})
		widgets.append(w)

		return widgets






