# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import time

from django.http import HttpResponse, HttpResponseForbidden

from django.template import RequestContext, loader

from core.models import CapabilityException

from webui.api.response import HttpJsonErrorResponse

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def calculate_exec_time(view_fun):
	"""
	Set the current tenant context for the selected view.
	
	It looks for the 'tenant_id' variable inside  session data
	or inside the request parameters (GET or POST).
	
	If no tenant_id is defined, the default tenant is selected.
	"""

	def _calculate_exec_time(request, *args, **kwargs):

		try:
			ts_start = time.time()
			return view_fun(request, *args, **kwargs)
		finally:
			ts_end = time.time()
			#print __file__, "calculate_exec_time", "EXECUTION TIME:", view_fun.__name__, view_fun.__module__, ts_end - ts_start

	return _calculate_exec_time

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def tenant_dependent(view_fun):
	"""
	Set the current tenant context for the selected view.
	
	It looks for the 'tenant_id' variable inside  session data
	or inside the request parameters (GET or POST).
	
	If no tenant_id is defined, the default tenant is selected.
	"""

	def _tenant_dependent_wrapper(request, *args, **kwargs):

		if request.tenant is None:

			context = {
			}

			#c = RequestContext(request, context)   django 1.11
			t = loader.get_template( 'html/user_config_error.html' )
			html_text = t.render(context, request)	 # django 1.11

			return HttpResponseForbidden(html_text)

		return view_fun(request, *args, **kwargs)

	return _tenant_dependent_wrapper

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def ajax_tenant_dependent(view_fun):

	def _ajax_tenant_dependent_wrapper(request, *args, **kwargs):

		if request.tenant is None:
			r = HttpJsonErrorResponse("Tenant information missing for this session. The session is probably not authenticated")
			return r

		return view_fun(request, *args, **kwargs)

	return _ajax_tenant_dependent_wrapper

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

def user_is_admin(view_fun):

	def _user_is_admin_wrapper(request, *args, **kwargs):

		if not request.user.is_superuser:

			context = {
			}
			#c = RequestContext(request, context)  django 1.11
			t = loader.get_template( 'html/user_not_admin_error.html' )
			html_text = t.render(context, request)	 # django 1.11

			return HttpResponseForbidden(html_text)

		return view_fun(request, *args, **kwargs)

	return _user_is_admin_wrapper

#-----------------------------------------------------------------------------#
#                                                                             #
#-----------------------------------------------------------------------------#

#
# Decoratore di comodo per controllare le capability di un utente
#

def capability_required(*capabilities):

	def _capability_required(view_fun):

		def wrapped_view(request, *args, **kwargs):

			try:
				tenant = request.tenant
				user   = request.user

				for c in capabilities:
					user.assert_capability(tenant, c)

				return view_fun(request, *args, **kwargs)

			except CapabilityException, e:
				return HttpResponseForbidden("NOT ALLOWED: " + str(e))

		return wrapped_view

	return _capability_required


def ajax_capability_required(*capabilities):

	def _ajax_capability_required(view_fun):

		def wrapped_view(request, *args, **kwargs):

			try:
				tenant = request.tenant
				user   = request.user

				for c in capabilities:
					user.assert_capability(tenant, c)

				return view_fun(request, *args, **kwargs)

			except CapabilityException, e:
				return HttpJsonErrorResponse("NOT ALLOWED: " + str(e))

		return wrapped_view

	return _ajax_capability_required
