# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import time
import types
import json 

from core.compatibility import reverse

from core.models import Tenant
from core.api.base import db_set_current_tenant_schema

from webui.session import SESSION_TENANTID_KEY

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

# request.REQUEST rimosso da django 1.11
class RequestGetPostMiddleware(object):

	def process_request(self, request):

		def _return_get_or_post_data(self):
			# No request.REQUEST for Django >=1.8 
			d = {}
			d.update( dict(self.GET.items() ) )

			d.update( dict(self.POST.items()) )

			if self.method == "POST" and self.content_type == "application/json":

				data = json.loads(self.body)

				print data

				d.update( data )

			return d

		request.GET_OR_POST = types.MethodType( _return_get_or_post_data, request )

		return

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

class TenantMiddleware(object):
	"""
	Quando l'utente apre una sessione HTTP con l'interfaccia web, il *tenant* corrente
	viene associato alla sessione.
	
	Questo middleware si preoccupa di assegnare un oggetto Tenant alla request
	
		request.tenant
	
	Di fatto le request che devono essere processate sono tutte le request
	che non servono per gestire file statici/media.
	"""
	
	def __init__(self, *args, **kwargs):

		super(TenantMiddleware, self).__init__(*args, **kwargs)

		# I file statici sono tutti serviti attraverso URL/view ben preciese.
		self.BASE_STATIC_PATH = reverse('static.file.serve', kwargs={ 'path':'' })
		self.BASE_MEDIA_PATH  = reverse('media.file.serve' , kwargs={ 'path':'' })

	def process_request(self, request):
		
		#print "------------"
		#print "BASE_STATIC_PATH", self.BASE_STATIC_PATH
		#print "BASE_MEDIA_PATH ", self.BASE_MEDIA_PATH
		#print "path            ", request.path
		#print "path_info       ", request.path_info
		
		if self._requires_tenant_handling(request):
			
			self._add_session_tenant_to_request(request)
		else:
			#print __file__, request.path, "NO TENANT"
			pass
		
		# Tell to django to process the request with all the other middlewares/views.
		return None
	
	def _requires_tenant_handling(self, request):
		
		# Static files do not need tenant/session 
		if request.path.startswith(self.BASE_STATIC_PATH) or \
		   request.path.startswith(self.BASE_MEDIA_PATH): # or \
		   #request.path.find('/static/') >= 0:                  # IGNORE <SANET_INSTALL_DIR>/<subapp>/static/<path>  FIX: find a better method to check this
			   return False
		
		return True
	
	def _add_session_tenant_to_request(self, request):

		assert hasattr(request, 'session'), "The Sanet authentication middleware requires session middleware to be installed. Edit your MIDDLEWARE_CLASSES setting to insert 'django.contrib.sessions.middleware.SessionMiddleware'."
		
		request.tenant = None
		
		if request.user.is_anonymous():
			# Gli utenti anonimi sono usati per effettuare il login iniziale
			# quindi non c'e' un tenant associato, MA non dobbiamo dare errore.
			return
		
		if request.session.has_key(SESSION_TENANTID_KEY):
			
			tenantid = request.session[SESSION_TENANTID_KEY]
			
			#
			# Troviamo il tenant associato alla sessione corrente
			#
			tenant = Tenant.objects.get(uuid=tenantid)
			
			#
			# Spostiamo la sessione corrente sul tenant
			#
			#db_set_current_tenant_schema(tenant.name)
			tenant.db_set_current_schema()

			# Impostiamo la visibilita del tenant in base all'utente corrente
			tenant.set_tag_visibility(request.user)
			
			request.tenant = tenant

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

# import redis

class AnalyticsMiddleware(object):

	def __init__(self, *args, **kwargs):

		super(AnalyticsMiddleware, self).__init__(*args, **kwargs)

		# I path "statici" non hanno bisongo di essere gestiti dal middleware essere ignorati dal middleware
		self.BASE_STATIC_PATH = reverse('static.file.serve', kwargs={ 'path':'' })
		self.BASE_MEDIA_PATH  = reverse('media.file.serve' , kwargs={ 'path':'' })

		#
		# TODO: self.rediscon = redis.Redis()
		#

	def process_view(self, request, view_func, view_args, view_kwargs):

		# Ignore request for static filee (do not perform unuseful DB access)
		if request.path.startswith(self.BASE_STATIC_PATH) or request.path.startswith(self.BASE_MEDIA_PATH):
			return

		try:
			t1 = time.time()
			response = view_func(request, *view_args, **view_kwargs)
		finally:
			t2 = time.time()

			#
			# TODO: self.rediscon.incr('analytics_page_hit_' + request.path )

		return response

