# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import time
import json
import operator

from django.http import HttpResponse

from sanet_common.sanetdefs.constants import ELEMENT_KIND_STR, ELEMENT_KIND_VAL
from sanet_common.sanetdefs.constants import DIRENTRY_AUTH_READWRITE

from core.models import MonitoredNode, MonitoredElement, DataGroup, Condition
from core.api.base import db_set_current_tenant_schema
from core.api.elements import element_rename

from webui.models import *
from webui.decorators import tenant_dependent

from tags.models import *
from tags.models import Tag, _untag_element, _untag_datagroup, _untag_condition

from tags.api import tag_rename, tag_remove

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@tenant_dependent
def sub_elements(request):

	tenant = request.tenant
	user   = request.user

	db_set_current_tenant_schema(tenant.name)

	sid    = request.GET_OR_POST().get('sid', "/")
	stype  = request.GET_OR_POST().get('stype')
	
	entries = _sub_elements(tenant, stype, sid, full_path=False, user=user)
	
	return HttpResponse(json.dumps(entries))
	

def _sub_elements(tenant, stype, sid, full_path=True, user=None):
	
	r = []

	if stype == 'root':
		
		nodes = MonitoredNode.get_visible_by_user(tenant=tenant, user=user)

		for n in nodes:
			
			r.append({
				'data': {
					'stype'   : ELEMENT_KIND_STR[n.kind],
					'sid'     : str(n.uuid),
				},

				'key': n.uuid,

				'title'   : n.name, 
				'icon'    : n.get_icon_url(), 
				

				'folder': True,
				"lazy"  : True,

				'href'     : n.page_url,
			})

	elif stype in ELEMENT_KIND_STR.values():

		uuid = sid
		e = MonitoredElement.objects.get(kind=ELEMENT_KIND_VAL[stype], uuid=uuid)

		sub_elements = e.get_sub_elements_by_user(user).all()
		
		for sube in sub_elements:
			
			has_sub = (sube.get_sub_elements_by_user(user).count() + sube.get_datagroups_by_user(user).filter(deleted=False).count()) > 0
			
			r.append({
				'data': {
					'stype'   : ELEMENT_KIND_STR[sube.kind],
					'sid'     : sube.uuid,
				},

				'key': sube.uuid,

				'title'   : (e.name + ":" + sube.name) if full_path else sube.name, 
				'icon'    : sube.get_icon_url(),

				'folder': has_sub,
				"lazy"  : True,

				'href'     : sube.page_url,
			})

		datagroups = e.get_datagroups_by_user(user).filter(deleted=False).order_by('path').all()

		for dg in datagroups:

			has_sub = dg.get_conditions_by_user(user).filter(deleted=False).count() > 0

			r.append({
				'data': {
					'stype'   : "datagroup",
					'sid'     : dg.uuid,
				},

				'key': dg.uuid,

				'title'   : (e.name+":"+dg.name) if full_path else dg.name, 				
				'icon'    : dg.get_icon_url(),

				'folder': has_sub,
				"lazy"  : True,

				'href'    : e.page_url,
			})

	elif stype == 'datagroup':
		
		uuid = sid
		dg = DataGroup.active_objects.get(uuid=uuid)
		
		element = dg.element
		
		for c in dg.get_conditions_by_user(user).filter(deleted=False).order_by('path').all():
			
			r.append({
				'data': {
					'stype'   : "condition",
					'sid'     : c.uuid,
				},

				'key': c.uuid,

				'title'   : (c.element.name+":"+c.name) if full_path else c.name,
				'icon'    : c.get_icon_url(),

				'folder': False,
				"lazy"  : True,

				'href'     : element.page_url,
			})

	return r 

#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@tenant_dependent
def sub_tags(request):
	
	tenant = request.tenant
	user   = request.user
    
	db_set_current_tenant_schema(request.tenant.name)
	
	stype     = request.GET_OR_POST().get('stype', "/")
	sid       = request.GET_OR_POST().get('sid', None)
	tags_type = request.GET_OR_POST()['tags_type'] #flag per discriminare i tree/tag di servizio dai tag normali
	
	tenant = request.tenant

	r = []
	
	if stype == "root":
		
		trees = TagTree.get_visible_trees_by_user(tenant, user).all()
		
		nostate = False if tags_type == "tags" else True
		
		root_tags = Tag.objects.filter(tenant=request.tenant, parent__isnull=True).filter(tree__in=trees, tree__nostate = nostate).order_by('name').all()
		
		for tag in root_tags:
			
			can_edit = ''
			#if user.is_superuser == True:
			#	can_edit = 1
			#else:
			#	can_edit = 1 if tag.can_edit(user)['auth_type__max'] == DIRENTRY_AUTH_READWRITE else 0
			
			can_edit = tag.tree.can_edit(user)
			
			
			capabilities = user.get_capabilities(request.tenant)
			
			r.append({
				'data': {
					'stype'   : 'tag',
					'sid'     : str(tag.uuid),

					'ttype'   : 'tag',
					'tid'     : str(tag.uuid),

					"can_edit": can_edit,
					"user_can_rename": 1 if "tags.tag.edit"   in capabilities else 0,
					"user_can_delete": 1 if "tags.tag.delete" in capabilities else 0,
				},

				'key': tag.uuid,

				'title'   : tag.name,
				'icon'    : tag.get_icon_url(),

				'folder': True,
				"lazy"  : True,


				#"children": []
			})

	elif stype == 'tag':
		
		parent_tag = Tag.objects.get(uuid = sid)
		
		# Sotto tag
		tags = parent_tag.child_tags.order_by('path')
		for tag in tags:
			
			can_edit = ''
			if user.is_superuser == True:
				can_edit = 1
			else:
				can_edit = 1 if tag.can_edit(user)['auth_type__max'] == DIRENTRY_AUTH_READWRITE else 0
			
			capabilities = user.get_capabilities(request.tenant)
			
			r.append({
				'data': {
					'stype'   : 'tag',
					'sid'     : str(tag.uuid),

					'ttype'   : 'tag',
					'tid'     : str(tag.uuid),

					"can_edit": can_edit, #in base ai permessi dell'utente sul tagtree
					"user_can_rename": 1 if "tags.tag.edit"   in capabilities else 0,
					"user_can_delete": 1 if "tags.tag.delete" in capabilities else 0
				},

				'key': tag.uuid,

				'title'   : tag.name,
				'icon'    : tag.get_icon_url(),

				'folder': True,
				"lazy"  : True,

			})

		#
		# Elementi taggati direttamente. 
		# Dobbiamo trovare gli elementtag e usare i loro ID
		#
		tagged, elements = parent_tag.get_elementtags_with_elements()

		element_by_id = dict([ (e.uuid, e) for e in elements])
		
		tmp = []
	
		for t in tagged:
			
			# NOTA: Per ora i datagroup taggati direttamente non devono essere visibili nell'albero
			if t.elementtype == TAG_ELEMENTTYPE_DATAGROUP:
				continue;

			e = element_by_id[t.elementid]
			
			if isinstance(e, MonitoredElement):
				e = e.get_specific_element()
				if isinstance(e, MonitoredNode):
					title = e.name
				else:
					# monitored element, not node
					title = e.parent.name +":"+e.name
			else:
				# condition and timegraph tagged
				title = e.element.name +":"+e.name
				
			icon = e.resource_type + '.gif'

			#r.append({
			tmp.append({
				'data': {
					'stype'   : e.resource_type,
					'sid'     : str(e.uuid),

					'ttype'   : 'etag',
					'tid'     : str(t.uuid),
				},
				
				'extraClasses': 'tagged',

				'key': t.uuid,

				'title'   : title,
				'icon'    : e.get_icon_url(),

				'folder': True,
				"lazy"  : True,
				
				'href'    : e.page_url,
			})
		
		tmp = sorted(tmp, key=lambda item:item['title'])
		r = r + tmp

			
		#r = sorted(r, key=lambda item:item['title'])

	elif stype == 'etag':

		#
		# Dobbiamo espandere un oggetto taggato. Lo espandiamo ma
		# dobbiamo distinguere gli elementi interni taggati da quelli
		# esplicitamente non taggati.
		#
		
		# Troviamo l'elemento taggato e il suo tag
		et = ElementTag.objects.get(uuid=sid)
		e = et.element
		tag = et.tag

		# Troviamo tutti le entry del menu' per questo elemento
		sid = et.elementid
		if et.elementtype in [TAG_ELEMENTTYPE_NODE, TAG_ELEMENTTYPE_IFACE, TAG_ELEMENTTYPE_STORAGE, TAG_ELEMENTTYPE_SERVICE, TAG_ELEMENTTYPE_DEVICE]:
			stype = ELEMENT_KIND_STR[e.kind]
		elif et.elementtype == TAG_ELEMENTTYPE_DATAGROUP:
			stype = 'datagroup'
		else:
			stype = None

		r = _sub_elements(tenant, stype, sid, full_path=False, user=user)

		# Troviamo gli elementtag corrispondenti per capire cosa e' taggato.
		elementtag_id_by_element_id = dict([ (row[1],row[0]) for row in tag.tagged.values_list('uuid', 'elementid').all()])
		
		# Marchiamo le entry del menu come 'tagged' o 'nottagged'
		# per distinguere gli elementi realmente "visibili" da quelli 
		# non visibili.
		for e in r:
			
			et_uuid = elementtag_id_by_element_id.get(e['data']['sid'], None)
			if et_uuid is not None:
				e.update({
					'data': {
						'ttype'   : 'etag',
						'tid'     : et_uuid,
						
						'sid'     : e['data']['sid'],
						'stype'   : e['data']['stype']
					},
					'extraClasses': 'tagged',
				})
			else:
				e.update({
					'extraClasses': 'not-tagged',
				})
	else:
		r = _sub_elements(tenant, stype, sid, user=user)

	return HttpResponse(json.dumps(r))

   
#----------------------------------------------------------------------#
#                                                                      #
#----------------------------------------------------------------------#

@tenant_dependent
def addmap(request):
	
	tenant = request.tenant
	
	tagid = request.GET_OR_POST().get('tagid')
	name = request.GET_OR_POST().get('name', "No Name")
	
	if tagid is not None and tagid !="default":
		from maps.views import  add_new_map_to_tag
		from django.shortcuts import redirect
		
		m, info = add_new_map_to_tag(tenant, tagid)
		
		map_uuid = m.uuid
		
	else:
		from maps.views import  add_new_map
		from django.shortcuts import redirect
		
		m, info = add_new_map(tenant, name)
		
		map_uuid = m.uuid
		
	return redirect('resource_page', rt='map', uuid=map_uuid)

	
	

@tenant_dependent
def rename_tag_or_element(request):
	
	tenant = request.tenant
	db_set_current_tenant_schema(request.tenant.name)
	
	new_name = request.GET_OR_POST()['new_name']
	uuid     = request.GET_OR_POST()['uuid']
	etype    = request.GET_OR_POST()['type']
	
	if new_name != None and new_name != "":
		if etype == "tag":
			try:
				tag = Tag.objects.get(tenant = tenant, uuid = uuid)
				tag_rename(tag, new_name)
			except Exception, e:
				raise e
		
		else:
			try:
				element = MonitoredElement.objects.get(tenant = tenant, uuid = uuid)
				element_rename(element, new_name)
			except Exception, e:
				raise e
	
	return HttpResponse("ok")


@tenant_dependent
def delete_tag(request):
	
	tenant = request.tenant
	db_set_current_tenant_schema(request.tenant.name)
	
	uuid     = request.GET_OR_POST()['uuid']
	#etype    = request.GET_OR_POST()['type']
	
	try:
		tag = Tag.objects.get(tenant = tenant, uuid = uuid)
		tag_remove(tag, True)
	except Exception, e:
		raise e

	return HttpResponse("ok")


def untag_tree_element(request):
	
	tenant = request.tenant
	db_set_current_tenant_schema(request.tenant.name)
	
	uuid        = request.GET_OR_POST()['uuid']
	etype       = request.GET_OR_POST()['type']
	element_tag = request.GET_OR_POST()['element_tag']
	
	element_tag = ElementTag.objects.get(tenant = tenant, uuid = element_tag)

	try:
		
		if etype == "condition":
			condition = Condition.objects.get(tenant = tenant, uuid = uuid)
			_untag_condition(element_tag.tag, condition)
			
		elif etype == "datagroup":
			datagroup = DataGroup.objects.get(tenant = tenant, uuid = uuid)
			_untag_datagroup(element_tag.tag, datagroup)
			
		else:
			element = MonitoredElement.objects.get(tenant = tenant, uuid = uuid)
			_untag_element(element_tag.tag, element)
			
	except Exception, e:
		raise e
	
	return HttpResponse("ok")





