# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import time
import json
from operator import itemgetter
import urllib

import StringIO
import csv

from django.db import connection
from django.http import HttpResponse
from django.db.models import Q
from django.http import HttpResponse, HttpResponseRedirect
from django.contrib.auth.decorators import login_required

from sanet_common.sanetdefs.constants import CSTATE_FROM_CODENAME, PRIORITY_NONE, ELEMENT_KIND_VAL
from sanet_common.timeutils import timestamp_to_isostr

from core.compatibility import reverse

from core.models import MonitoredElement,MonitoredNode, Condition, DataSource, TenantSetting
from core.api.base import db_set_current_tenant_schema
from core.api.cache import init_cache_manager_from_conf
from core.api.datasource import init_datasource_store_from_conf

from tags.models import *

from webui.decorators import tenant_dependent
from webui.api import get_ds_module
from webui.api.types import MODEL_TO_WEB_ELEMENT_TYPE

# Questa funzione e' usata solo qui
from webui.api.web import _get_tags_conditions_data

from webui.api.web import get_tree_data

from webui.api.web import condition_get_data

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

@tenant_dependent
def get_data(request):

	"""
	This is the main data-collector for the whole page. It receives one only GET or POST paramenter (d), with a list of pipe(|)-separated couples of *uuid*, dot (.), *part*.  
	"""

	tenant = request.tenant

	t0 = time.time()

	db_set_current_tenant_schema(request.tenant.name)

	# Cache globale
	cache_manager = init_cache_manager_from_conf()

	# Cache con oggetti python temporanei che i diversi dataset possono scambiarsi 
	# senza ricalcolare piu' volte le stesse cose.
	dataset_cache = {}

	#
	# critical priority
	#
	critical_priority = tenant.get_setting(TenantSetting.SETTING_CRITICAL_LEVEL)
	if critical_priority is None:
		critical_priority = PRIORITY_NONE

	minimum_priority  = tenant.get_setting(TenantSetting.SETTING_MINIMUM_LEVEL)
	if minimum_priority is None:
		minimum_priority = PRIORITY_NONE

	res = {}

	datasets = request.GET_OR_POST().get('d')
	debug    = request.GET_OR_POST().get('debug')

	options = {
		'__cache_manager__'       : cache_manager,
		'__dataset_cache__'       : dataset_cache
	}

	options.update({
		'request'          : request,

		# Questi parametri servono nel 90% dei casi
		'tenant'           : request.tenant,

		'user'             : request.user,
		'user_capabilities': request.user.get_capabilities(tenant),

		'critical_priority': critical_priority,
		'minimum_priority' : minimum_priority,

		'rt'              : request.GET_OR_POST().get('rt', None),
		'rid'             : request.GET_OR_POST().get('rid', None),

		#
		# Questi parametri a che servono?
		#

		'from_ts'         : request.GET_OR_POST().get('from'),
		'to_ts'           : request.GET_OR_POST().get('to'),
		
		'remote'          : request.GET_OR_POST().get('mode')=="remote",

		'is_table'        : request.GET_OR_POST().get('requester', None) == "table",
		
		'page'            : int(request.GET_OR_POST().get('page', '0')),
		'pagesize'        : int(request.GET_OR_POST().get('pagesize', '100')),

		'filter_criteria' : request.GET_OR_POST().get('filter[criteria]', None),
		'filter_field'    : request.GET_OR_POST().get('filter[field]', None),

		'do_filter'       : Q(),
		'params'          : None,
	})
	options['offset'] = options.get('page') * options.get('pagesize')

	if options['to_ts']   is None: options['to_ts']   = time.time()
	if options['from_ts'] is None: options['from_ts'] = options['to_ts'] - 60 * 60 * 8

	if datasets is not None:
		datasets = json.loads(datasets)
	else:
		datasets = {}

	for dataset, datasetoptions in datasets.items():

		t1 = time.time()

		parts = dataset.split('.')
		uuid  = parts[0]
		part  = parts[1]

		if settings.DEBUG:
			data = _load_data_with_module(part, uuid, options, datasetoptions)
		else:
			try:
				data = _load_data_with_module(part, uuid, options, datasetoptions)
			except Exception, e:
				data = {
					'dataset': { 
						'error': str(e) 
					}
				}

		res [dataset] = data
		if options.get('is_table'): 
			res = res[dataset]['dataset']
		elif options.get('remote'): 
			res = res[dataset]['dataset']

		t2 = time.time()

	tf = time.time()

	connection.close()  # HACK: la conversione in json potrebbe richiedere tempo. Liberiamo risorse anche sul DB prima di passare il tempo a mandare la risposta al client.
	data = json.dumps(res)

	return HttpResponse(data)


def _load_data_with_module(part, uuid, options, datasetoptions):

	options['params'] = datasetoptions or {}

	# Load module
	dataset_module = get_ds_module(part)

	# Create dataset handler
	dataset_handler = dataset_module(uuid)

	# Init internal variabiles
	dataset_handler.init_global_settings(options)

	# Get data from dataset
	data = dataset_handler.get_data(options)

	return data


#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#


@login_required
@tenant_dependent
def page_by_path(request):

	path = request.GET_OR_POST()['path']
	kind = request.REQUEST.get('kind', None)
	
	if path.find(';') == -1:
		
		path = path.split(':')
		
		if len(path) == 1:
			name = path[0]
			e = MonitoredNode.objects.get(name=name)

		elif len(path) == 2:
			name = path[0]
			sub  = path[1]
			if kind == None:
				e = MonitoredElement.objects.filter(name=sub, parent__name=name)[0].get_specific_element()
			else:
				e = MonitoredElement.objects.get(name=sub, parent__name=name, kind=ELEMENT_KIND_VAL[kind]).get_specific_element()

	else:
		e = Condition.objects.get(path=path)

	url = reverse('resource_page', kwargs={'rt': MODEL_TO_WEB_ELEMENT_TYPE[e.__class__], 'uuid': e.uuid})

	return HttpResponseRedirect(url);






















#--------------------------------------------------------------------------------------#
#                                                                                      #
#--------------------------------------------------------------------------------------#

# dbilli - INUTILE NON USATA
#@tenant_dependent
#def get_last_value_node(request):
#	return _get_last_value_node(request)

def _get_last_value_node(request):
    #
    # FIX: dbilli - USATA SOLO DALLA DASHBOARD!!!!!! SPOSTARE NELLA DASHBOARD
    #
    
    """
    Get last value of dasource for a tagged node
    must filter for tag and classification
    """
    
    cache_manager = init_cache_manager_from_conf()
    
    tenant = request.tenant
    db_set_current_tenant_schema(tenant.name)
    
    taguuid        = request.GET_OR_POST().get('tagid', '')
    if taguuid != None:
        taguuid        = urllib.unquote(taguuid)
    
    classification = request.GET_OR_POST().get('classification', '')
    
    elementtype    = request.GET_OR_POST().get('element_type', '')
    element_name   = request.GET_OR_POST().get('element_name', '')
    
    dstitle        = request.GET_OR_POST().get('dstitle', '*')
    
    interval       = int(request.GET_OR_POST().get('interval', 0))
     
    top            = int(request.GET_OR_POST().get('top', 10))
    orderby        = int(request.GET_OR_POST().get('orderby', 0))
    
    f  = Q()
    ft = Q()
    
    t1 = time.time()

    if element_name != '':
        f = Q(element__name=element_name)
    else:
        if taguuid != '':
            # try to get uuid
            ft = Q(tenant= tenant, tag__uuid=taguuid) | Q(tenant = tenant, tag__path=taguuid)  

        if elementtype != '':
            ft = ft & Q(elementtype=ELEMENT_KIND_TO_TAG_ELEMENTTYPE[int(elementtype)])

        if taguuid != '' or elementtype != '':
            # get Elements for selected tags
            elements = list(ElementTag.objects.filter(ft).exclude(elementtype__in=['datagroup', 'condition']))
            elements = list([e.element for e in elements])
            f = f & Q(element__in=elements)


    # Troviamo i datasource referenziati        
    if classification:
        f = f & Q(classification=classification)
    
    if dstitle != '*':
        if "*" in dstitle:
            tmptitle = dstitle.replace("*","")
            f = f & Q(title__icontains=tmptitle)
        else:
            f = f & Q(title=dstitle)
    datasources = DataSource.active_objects.filter(f)
    
    store = init_datasource_store_from_conf()
    ret = []
    for d in datasources:
        element_name = d.element.name
            
        if elementtype != 'node':
            try:
                element_name = "%s->%s" % (d.element.parent.name, element_name)
            except:
                pass

        
        if int(interval):
            seconds = 60 * 60 * int(interval)
        else:
            seconds = 60 * 60
            
        ts_end   = int(time.time())
        ts_start = ts_end - seconds
        try:
            data = store.fetch_values(d.tenant_id, d.uuid, ts_start=ts_start, ts_end=ts_end)
        except:
            continue

        dslen = len(data['dataseries'])
        if d.max_val:
            dataseries = [dv for dv in data['dataseries'] if dv != None and dv <= d.max_val]
        else:
            dataseries = [dv for dv in data['dataseries'] if dv != None]
        

        if d.max_val:
            dataseries = [dv for dv in data['dataseries'] if dv != None and dv <= d.max_val]
        else:
            dataseries = [dv for dv in data['dataseries'] if dv != None]

        if not dataseries:
            lastvalue = 0
            media     = 0
            status    = 0
        else:
            lastvalue = _get_last_value_dataseries(dataseries)
            media     = sum(dataseries)/float(dslen)
            status    = 1
        #primary_condition = d.element.get_primary_condition()
        #primary_condition = None
        #if primary_condition:
        #    status = primary_condition.status
        #else:
        #    status = None
        
        try:
            tg = TimeGraph.active_objects.filter(datagroup=d.datagroup)
            for t in tg:
                ret.append([element_name, d.element.uuid, d.element.kind, d.title, lastvalue , media, timestamp_to_isostr(ts_end), t.ylegend, status])
        except TimeGraph.DoesNotExist:
            #	 "Datasource %s not have a timegraph" % d.name
            ylegend = ""
            
            ret.add([element_name, d.element.uuid, d.element.kind, d.title, lastvalue , media, timestamp_to_isostr(ts_end), ylegend, status])

    t2 = time.time()
 
    ret = sorted(ret, key=itemgetter(orderby))
    ret.reverse()
    result = ret[:top]
    return HttpResponse(json.dumps(result))

    
def _get_last_value_dataseries(dataseries):
    dataseries.reverse()
    for item in dataseries:
        if item:
            return item
    return 0

#--------------------------------------------------------------------------------------#
#                                                                                      #
#--------------------------------------------------------------------------------------#

#
# disabilitata da dbilli - se e' roba solo della dashboard non deve stare qui
#@tenant_dependent
#def get_conditions(request):
#    return _get_conditions(request)

def _get_conditions(request, tagtree=None):
    """
    Usata dal gadget della dashboard:
    
       * sanet3_conditions_pie
       * sanet3_conditions_table
       * sanet3_conditions_number
       
    """
    
    
    tenant = request.tenant
    db_set_current_tenant_schema(tenant.name)
    
    
    
    # compute filter for conditions
    tmp_filter = {}
    # tmp_filter['uuid'] = request.GET_OR_POST().get('uuid', '')
    element_type = request.GET_OR_POST().get('element_type', '')
    if element_type != '':
        tmp_filter['element__kind'] = int(element_type)
    
    element_name = request.GET_OR_POST().get('element_name', '')
    if element_name != '':
        tmp_filter['element__name__icontains'] = element_name
    
    element_list = request.GET_OR_POST().get('element_list', '')
    
    if element_list != '':
        tmp_filter['element__uuid__in'] = json.loads(element_list)
    
    status = request.GET_OR_POST().get('status', '')
    if status != '':
        slist = status.split(",")
        tmp_filter['status__in'] = []
        for s in slist:
            if s == 'UN':
                for UN_status in CSTATES_UNCHECKABLE:
                    tmp_filter['status__in'].append(UN_status)
            else:
                tmp_filter['status__in'].append(CSTATE_FROM_CODENAME(s))
    
    classification = request.GET_OR_POST().get('classification', '')
    if classification != '':
        tmp_filter['classification'] = classification
    
    try:
        priority = request.GET_OR_POST().get('priority', PRIORITY_NONE)
        priority = int(priority)
        
    except ValueError, e:
        level = PriorityLabel.objects.get(label = priority)
        priority = level.min_level
    except Exception, e:
        raise e
        
    tmp_filter['priority__gte'] = int(priority)
    
    #tmp_filter['priority__gte'] = request.GET_OR_POST().get('priority', PRIORITY_NONE)
    
    # keep only value that are not null
    data_filter = dict([(k,v) for k,v in tmp_filter.items() if v])
    data_filter['summary']      = request.GET_OR_POST().get('summary', '')
    data_filter['summary_data'] = request.GET_OR_POST().get('summary_data', '')
    
    
    # if tagtree is defined, get first "tagtree", else get the tagtree in the REQUEST object
    if tagtree != None:
        data_filter['tree_id'] = tagtree
    else:
        data_filter['tree_id'] = request.GET_OR_POST().get('tree', '')
    
    if data_filter['tree_id'] != None:
        data_filter['tree_id']      = urllib.unquote(data_filter['tree_id'])
    
    _list = condition_get_data(request.tenant.name, data_filter)

    return HttpResponse(json.dumps(_list))

#--------------------------------------------------------------------------------------#
#                                                                                      #
#--------------------------------------------------------------------------------------#

# FIX: disabilitata da dbilli - se e' roba usata dalla dashboard non deve stare qui
#@tenant_dependent
#def get_tags_conditions(request):
#    return _get_tags_conditions(request)

def _get_tags_conditions(request):
    
	tenant = request.tenant
	db_set_current_tenant_schema(tenant.name)
    
	# compute filter for conditions
	tmp_filter = {}
    
	# tmp_filter['uuid'] = request.GET_OR_POST().get('uuid', '')
	element_type = request.GET_OR_POST().get('element_type', '')
	if element_type:
		tmp_filter['element__kind'] = int(element_type)

	element_name = request.GET_OR_POST().get('element_name', '')
	if element_name:
		tmp_filter['element__name'] = element_name
    
	status = request.GET_OR_POST().get('status', '')
	if status:
		tmp_filter['status'] = CSTATE_FROM_CODENAME(status)
    
	classification = request.GET_OR_POST().get('classification', '')
	if classification:
		tmp_filter['classification'] = classification
    
	#tmp_filter['critical_priority'] = tenant.get_setting(TenantSetting.SETTING_CRITICAL_LEVEL)
	#tmp_filter['minimum_priority']  = tenant.get_setting(TenantSetting.SETTING_MINIMUM_LEVEL)

	# keep only value that are not null
	data_filter = dict([(k,v) for k,v in tmp_filter.items() if v != ''])

	data_filter['summary'     ] = request.GET_OR_POST().get('summary'     , '')
	data_filter['summary_data'] = request.GET_OR_POST().get('summary_data', '')
	data_filter['tree_uuid'   ] = request.GET_OR_POST().get('tree_uuid'   , '')

	if data_filter['tree_uuid'] != None:
		data_filter['tree_uuid'] = urllib.unquote(data_filter['tree_uuid'])
    
	_list = _get_tags_conditions_data(request.tenant.name, data_filter)

	return HttpResponse(json.dumps(_list))  


#--------------------------------------------------------------------------------------#
#                                                                                      #
#--------------------------------------------------------------------------------------#


# NOTA: Disabilitata da dbilli - se e' roba usata solo dalla dashboard non deve stare qui

#@tenant_dependent
#def get_tree_with_status(request):
#	#
#	# FIX: MA QUESTA VIEW E' USATA DA QUALCUNO?!!?!
#	#
#	return _get_tree_with_status(request)



def _get_tree_with_status(request):
	#
	# FIX: dbilli - QUESTA FUNZIONE E' USATA DAL CODICE DELLA DASHBOARD e da get_tree_with_status() che ho disabilitato
	#      Capire se e' possibile spostare questa view nella dashboard e basta.
	#
	
	# FIX: nome del parametro fuorviante, dovrebbe essere "uuid_or_path".
	uuid = request.GET_OR_POST().get('uuid', '')
	if uuid != None:
		uuid = urllib.unquote(uuid)

	ret = get_tree_data(request.tenant.name, uuid)

	return HttpResponse(json.dumps(ret))

#--------------------------------------------------------------------------------------#
#                                                                                      #
#--------------------------------------------------------------------------------------#

def _get_tree_list(request):

	tenant = request.tenant
	user   = request.user

	db_set_current_tenant_schema(tenant.name)

	tags = Tag.get_visible_by_user(tenant, user).order_by('-path')

	ret = {}
	for tag in tags:
		ret[tag.uuid] = tag.get_path()

	return ret

#@tenant_dependent
#def get_tree_list(request):
#
#	ret = _get_tree_list(request)
#
#	return HttpResponse(json.dumps(ret))  

#--------------------------------------------------------------------------------------#
#                                                                                      #
#--------------------------------------------------------------------------------------#

@tenant_dependent
def get_tree_list_order(request):
	"""
	Questa view e' chiamata da INDEX.HTML.
	
	Serve per selezionare i tag dove creare sotto-tag tramite interfaccia grafica.
	
	Dobbiamo selezionare tutti i tag su cui l'utente ha permesso di lettura o scrittura.
	"""

	tenant = request.tenant
	user   = request.user

	db_set_current_tenant_schema(tenant.name)

	tags = Tag.get_visible_by_user(tenant, user, DIRENTRY_AUTH_READWRITE).order_by('-path')

	l = [ (t.uuid, t.get_path()) for t in tags.all() ]

	return HttpResponse(json.dumps(l))  

#--------------------------------------------------------------------------------------#
#                                                                                      #
#--------------------------------------------------------------------------------------#

@tenant_dependent
def export_list_csv(request):

	tenant = request.tenant
	db_set_current_tenant_schema(tenant.name)

	fields   = request.GET_OR_POST().get('fields', 'name')
	order_by = request.GET_OR_POST().get('order_by', 'name')

	fields = fields.split(',')
	order_by = order_by.split(',')
	nodes = MonitoredNode.objects.all().values_list(*fields).order_by(*order_by)

	def _csv2string(data):
		si = StringIO.StringIO()
		cw = csv.writer(si)
		cw.writerows(data)
		s =  si.getvalue().strip('\r\n')
		si.close()
		return s

	data = _csv2string(nodes)

	response = HttpResponse(content_type='text/csv')
	response['Content-Disposition'] = 'attachment; filename="nodes.csv"'
	response.write(data)

	return response
