# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import os
import time

from django.http import HttpResponse
from django.shortcuts import render
from django import forms

from sanet_common.sanetdefs.constants import PRIORITY_NONE
from sanet_common.sanetdefs.constants import PRIORITY_MIN, PRIORITY_MAX

from core.compatibility import ErrorList
from core.models import Tenant, TenantSetting, PriorityLabel
from core.models import PriorityLabel
from core.api.base import tenant_get_media_url
from core.api.base import tenant_get_media_file_path
from core.api.base import tenant_get_icons_dir

from tags.models import *

from webui.api.utils import render_template

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

DEFAULT_CRITICAL = PRIORITY_MAX + 1
DEFAULT_MINIMUM  = PRIORITY_NONE

def update_settings(request, uuid):
	

	tenant = Tenant.objects.get(uuid=uuid)

	minimum = request.GET_OR_POST()['minimum_level']
	if minimum:
		l = PriorityLabel.objects.get(tenant=tenant, label=minimum)
		minimum = l.min_level
	else:
		minimum = DEFAULT_MINIMUM
		
	critical = request.GET_OR_POST()['critical_level']
	if critical:
		l = PriorityLabel.objects.get(tenant=tenant, label=critical)
		critical = l.min_level
	else:
		critical = DEFAULT_CRITICAL

	# Validate
	error = None
	if critical != DEFAULT_CRITICAL and minimum != DEFAULT_MINIMUM:
		if critical <  minimum:
			error = "Critical level can not be lower than minimum level"

	# Update
	updated = False
	if not error:
		TenantSetting.save_setting(tenant, TenantSetting.SETTING_CRITICAL_LEVEL, critical)
		TenantSetting.save_setting(tenant, TenantSetting.SETTING_MINIMUM_LEVEL , minimum )
		updated = True

	html = render_tenant_config_details(tenant, updated=updated, error=error)

	return HttpResponse(html)


def render_tenant_config_details(tenant, error=None, updated=False):

	#
	# Create Form
	#
	settings = TenantSetting.get_all_settings(tenant)

	configured_minimum_level  = settings[TenantSetting.SETTING_MINIMUM_LEVEL ]
	configured_critical_level = settings[TenantSetting.SETTING_CRITICAL_LEVEL]


	priority_levels = []
	for l in PriorityLabel.objects.order_by('min_level').all():
		priority_levels.append(l)
		if configured_critical_level is not None:
			if l.in_range(configured_critical_level):

				l.is_critical = True
				
		if configured_minimum_level is not None:
			if l.in_range(configured_minimum_level):
				l.is_minimum = True

	levels = []
	for l in xrange(PRIORITY_NONE, PRIORITY_MAX):
		levels.append({
			'value'      : l if l != PRIORITY_NONE else "" ,
			'label'      : l if l != PRIORITY_NONE else "-",
			'description': l if l != PRIORITY_NONE else "-"
		})

	html = render_template( 'html/parts/tenant_settings.html', { 
		'tenant'         : tenant, 
		
		'minimum_level'  : configured_minimum_level,
		'critical_level' : configured_critical_level,
		
		'levels'         : levels,
		'priority_levels': priority_levels,
		
		'success': updated and (error == None),
		'error'  : error,
		
		'ts': time.time()
	})

	return html


	
	



#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

def remove_priority_level(request, uuid, label):
	
	p = PriorityLabel.objects.get(label=label)
	p.delete()
	
	return HttpResponse('ok')

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

from webui.api.response import HttpJsonSuccessResponse

def update_icon(request, uuid):
	
	tenant = Tenant.objects.get(uuid=uuid)
	
	if request.method == 'POST':	
	
		uploaded_file_data = request.FILES.get('file')
		if uploaded_file_data:
			file_name = tenant_get_media_file_path(uuid, 'site.png')
			
			with open(file_name, 'wb+') as destination:
				for chunk in uploaded_file_data.chunks():
					destination.write(chunk)
			
			icon_url = tenant_get_media_url(tenant.uuid, 'site.png') + '?_ts=%s' % (time.time())
			#return HttpJsonResponResponse(icon_url)
			return HttpJsonSuccessResponse(data={'icon_url': icon_url})

	icon_url = tenant_get_media_url(tenant.uuid, 'site.png')
	html = render_template( 'html/parts/configtenant_upload_icon.html', { 
		'tenant'  : tenant, 
		'icon_url': icon_url,
		'ts': time.time()
		
	})

	return HttpResponse(html)

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

def icon_sanitize_name(icon_name):

	icon_name = icon_name.encode('ascii', 'ignore')
	icon_name = icon_name.replace(' ','-')
	return icon_name


def add_icon(request, uuid):

	tenant = Tenant.objects.get(uuid=uuid)

	icons_dir = tenant_get_icons_dir(tenant.uuid)

	request.method

	if request.method == 'POST':

		uploaded_file_data = request.FILES.get('file')
		if uploaded_file_data:

			file_name = icon_sanitize_name(uploaded_file_data.name)

			file_name = os.path.join(icons_dir, file_name) 

			with open(file_name, 'wb+') as destination:
				for chunk in uploaded_file_data.chunks():
					destination.write(chunk)

	return HttpResponse("ok")
	
#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

def remove_icon(request, uuid, icon_name):

	tenant = Tenant.objects.get(uuid=uuid)

	icons_dir = tenant_get_icons_dir(tenant.uuid)

	request.method

	if request.method == 'POST':
		file_path = os.path.join(icons_dir, icon_name + '.png') 
		os.remove(file_path)

	return HttpResponse("ok")



#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

def create_priority_choices():
	for a in xrange(PRIORITY_MIN, PRIORITY_MAX+1):
		yield (a, str(a))

class PriorityLabelForm(forms.Form):

	label       = forms.CharField(required=True)
	min_level   = forms.ChoiceField(required=True, choices=create_priority_choices())
	max_level   = forms.ChoiceField(required=True, choices=create_priority_choices())
	description = forms.CharField(required=False)

	def clean(self):

		if not self.errors:
			cleaned_data = self.cleaned_data
			label     = cleaned_data.get("label")
			min_level = int(cleaned_data.get("min_level"))
			max_level = int(cleaned_data.get("max_level"))

			fake_level = PriorityLabel(label=label, min_level=min_level, max_level=max_level)

			try:
				fake_level.validate()
			except Exception, e:
				msg = str(e)
				self._errors["label"] = ErrorList([msg])

			# Always return the full collection of cleaned data.
			return cleaned_data

		return super(PriorityLabelForm, self).clean()



def add_priority_level(request, uuid):

	tenant = Tenant.objects.get(uuid=uuid)
	
	if request.method == 'POST': 
		form = PriorityLabelForm(request.POST)
		if form.is_valid():
			
			label       = form.cleaned_data['label']
			min_level   = int(form.cleaned_data['min_level'])
			max_level   = int(form.cleaned_data['max_level'])
			description = form.cleaned_data['description']
			
			l, created = PriorityLabel.objects.get_or_create(tenant=tenant, label=label, min_level=min_level, max_level=max_level)
			l.description = description
			l.validate()
			l.save()
			
			
			return HttpResponse("OK")
	else:
		form = PriorityLabelForm() # An unbound form

	return render(request, 'html/parts/configtenant_prioritylabel.html', {
		'form': form,
		'tenant': tenant,
	})
