# Copyright (C) 2018 Laboratori Guglielmo Marconi S.p.A. <http://www.labs.it>
#
# This file is part of SANET
# SANET is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3 of the License
#
# SANET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with SANET. If not, see <http://www.gnu.org/licenses/>.

import os
import os.path
import time

import subprocess

from django.http import HttpResponse

from core.models import Tenant, TimeGraph

from core.api.base import db_set_current_tenant_schema
from core.api.base import tenant_get_media_file_path
from core.api.cache import init_cache_manager_from_conf

from core.api.datasource import init_datasource_store_from_conf

from core.api.timegraph.rrdutils import convert_expr_to_rpn

from webui.views import render_new_timegraph3

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

TIMEGRAPH_IMAGE_EXPIRE_TIME = 60

#-------------------------------------------------------------------------------#
#                                                                               #
#-------------------------------------------------------------------------------#

def timegraph_image(request, uuid):
	
	"""
	Questa view genere una immagine in formato PNG del timegraph specificato.
	
	Supporta i parametri:
	
		gtype		png | rrdgraph
		
		start		timestamp 
		end		timestamp
		interval	n secondi.
		
		width 		intero
		height		intero
		
		force		1 | 0
		
	Per evitare la generazione di troppi file, le immagini vengono generate al massimo ogni 60 secondi.
	
	I file temporanei prodotti da questa view sono contenuti nella directory:
	
		{{VAR_DIR}}/<tenant>/timegraph_images/
		
	I file in questa directory vengono rimossi dal comando:
	
		sanet-manage clean_temp_web_data
	"""

	tenant_name = request.GET_OR_POST().get('tenant', None)
	if tenant_name:
		tenant = Tenant.objects.get(name=tenant_name)
	else:
		tenant = Tenant.objects.get(is_primary=True)

	request.tenant = tenant
	db_set_current_tenant_schema(tenant.name)

	#
	# REQUEST PARAMS 
	#
	graph_type = request.GET_OR_POST().get('gtype', 'png')

	force      = int(request.REQUEST.get('force', 0))

	render_params = _get_render_params(request)

	#
	# CREATE IMAGE
	#

	dest_dir   = tenant_get_media_file_path(tenant.uuid, 'timegraph_images') 

	try:
		os.makedirs(dest_dir)
	except Exception, e:
		print __file__, "timegraph_image", "Unable to create temp dir", type(e), str(e)
		pass

	width, height, time_start, time_end = render_params

	#
	# Nome univoco
	#
	image_ts = int(time.time()) 
	image_ts = image_ts - (image_ts % TIMEGRAPH_IMAGE_EXPIRE_TIME) 

	filename = "%s_%s_%s_%s_%s_%s.png" % (uuid, graph_type, image_ts, time_end - time_start, width, height )

	#
	# Generate file
	#
	image_path = os.path.join(dest_dir, filename)

	if not os.path.exists(image_path) or force:

		if graph_type == 'png':
			_timegraph_image_png(request, uuid, image_path, render_params)

		elif graph_type == 'rrdgraph':
			_timegraph_image_rrdtool(request, uuid, image_path, render_params)

		else:
			raise Exception("INVALID GRAPH TYPE: %s" % (graph_type))

	with open(image_path, "rb") as f:
		return HttpResponse(f.read(), mimetype="image/png")

	return


def _get_render_params(request):

	time_start = request.GET_OR_POST().get('start'   , None)
	time_end   = request.GET_OR_POST().get('end'     , None)

	interval   = request.GET_OR_POST().get('interval', None)

	width      = int(request.GET_OR_POST().get('w', 400))
	height     = int(request.GET_OR_POST().get('h', 200))

	if interval:
		time_end = int(time.time())
		time_start = time_end - int(interval)
	else:
		if not time_start and not time_end:
			time_end   = int(time.time())
			time_start = time_end - (60 * 60 * 8)
		elif time_start and not time_end:
			time_start = _param_to_ts( time_start )
			time_end   = int(time.time())
		else:
			time_start = _param_to_ts( time_start )
			time_end   = _param_to_ts( time_end   )

	return width, height, time_start, time_end


def _timegraph_image_png(request, uuid, image_path, render_params):

	from reports.data import LineChartWithFullLegendReportResult

	def _param_to_ts(v):
		try:
			t = time.mktime(time.strptime(v, "%Y-%m-%d %H:%M"))
		except Exception, e:
			try:
				t = time.mktime(time.strptime(v, "%Y-%m-%d"))
			except Exception, e:
				t = int(v)
		return t

	DEFAULT_FONT = "Helvetica"
	LEGEND_FONT_SIZE = 12

	DEFAULT_COLORS = [
		'#00ff00',
		'#0000ff',
		'#ff0000',
		'#ffff00',
		'#ff00ff',
		'#00ffff',
	]
	current_default = 0


	tenant = request.tenant

	width, height, time_start, time_end = render_params

	tg = TimeGraph.objects.get(uuid=uuid)

	tg_data = render_new_timegraph3(tg, time_start, time_end, None)

	series = [ s for s in tg_data['series'] if s['format'] != 'tick' ]

	max_value = 0
	values = []
	colors = []
	labels = []

	last_used_color = 0
	for s in series:

		m = max(s['values'])
		max_value = max(max_value, m)

		values.append(s['values'])
		if s['color']:
			colors.append(s['color'])
		else:
			colors.append(DEFAULT_COLORS[last_used_color])

			last_used_color = min(len(DEFAULT_COLORS)-1, last_used_color+1)

		labels.append(s['label'])

	lcrr = LineChartWithFullLegendReportResult(

		width       = width,
		height      = height,

		report_id   = '1',
		uid         = '%s' % (tg.uuid),

		time_start  = time_start,
		time_end    = time_end,
		max_value   = max_value,
		values      = values,
		colors      = colors,
		line_labels = labels,

		values_format = "%%s%s" % (tg_data['ylegend']),

		title = tg_data['title'],

		cache = False,
	)
	lcrr.set_tenant(tg.tenant)

	lcrr.generate(width=width, height=height, force_path=image_path)

	return


def _timegraph_image_rrdtool(request, uuid, image_path, render_params):

	cache_manager = init_cache_manager_from_conf()

	DEFAULT_COLORS = [
		'#0000ff',
		'#00ff00',
		'#ff0000',
		'#ffff00',
		'#ff00ff',
		'#00ffff'
	] + ([ '#0000ff' ] * 20)

	tenant = request.tenant

	#width, height, time_start, time_end = _get_render_params(request)

	width, height, time_start, time_end = render_params

	tg = TimeGraph.objects.get(uuid=uuid)

	#
	#
	#
	rrd_params    = []
	def_params    = []
	cdef_params   = []
	tick_params   = []
	line_params   = []
	area_params   = []

	gprint_params = []
	print_params  = []

	rrd_params = [
		'--width'         , str(width),
		'--height'        , str(height),
		'--start'         , str(int(time_start)),
		'--end'           , str(int(time_end  )),

		'--title'         , "'%s'" % (tg.expanded_title(cache_manager=cache_manager)),

		'--vertical-label', tg.ylegend,

		'--only-graph' if request.GET_OR_POST().get('onlygraph', None) else '',
	]

	series            = [ [], [] ]
	series_final_vars = [ [], [] ]

	datasource_store = init_datasource_store_from_conf()

	dg = tg.datagroup
	real_series     = list( tg.get_real_series().order_by('render_order', 'style').all()     )
	computed_series = list( tg.get_computed_series().all() )

	datasources_by_name = dict([ (ds.name, ds) for ds in dg.datasources.filter(name__in=[ s.variable_value for s in real_series ]) ])

	for ds_name, ds in datasources_by_name.items():
		dirname, filename = datasource_store._get_datasource_rrd_filename(tenant.name, ds.uuid)
		ds.rrd_file = filename

	for serie in real_series + computed_series:

		ds = datasources_by_name.get(serie.variable_value, None)

		if ds:
			d = 'DEF:%s_data="%s":data:AVERAGE' % (serie.name, ds.rrd_file)
			#def_params.append(d)
			rrd_params.append(d)

			cdef = 'CDEF:%s=%s_data' % (serie.name, serie.name)
			#cdef_params.append(cdef)
			rrd_params.append(cdef)

		if serie.style == 'line':

			line = 'LINE:%s%s:"%s\t"' % (serie.name, serie.color or DEFAULT_COLORS.pop(0), serie.legend)
			#line_params.append(line)
			rrd_params.append(line)

		elif serie.style == 'area':

			line = 'AREA:%s%sa0:"%s\t"' % (serie.name, serie.color or DEFAULT_COLORS.pop(0), serie.legend)
			#line_params.append(line)
			rrd_params.append(line)

		elif serie.style == 'tick':

			rpn_expr = convert_expr_to_rpn(serie.computed_value_expr)
			cdef = 'CDEF:%s=%s' % (serie.name, rpn_expr)
			#cdef_params.append(cdef)
			rrd_params.append(cdef)

			line = 'TICK:%s%s:1:"%s\l"' % (serie.name, serie.color or '#ff0000', serie.legend)
			#tick_params.append(line)
			rrd_params.append(line)

			continue

		elif serie.style == 'invisible':
			continue

		for CF, label in [('MAX','Max'),('MIN', 'Min'), ('AVERAGE','Avg')]: #,('LAST','Current')]:

			str_format = '%(label)s %(value)s%(unit)s ' % { 'label': label, 'value': '%.2lf%s', 'unit': tg.ylegend }
			if CF == 'AVERAGE':
				str_format += '\l'

			print_param  = ('GPRINT:%(vname)s:%(CF)s:"%(format)s"' % { 'vname' : serie.name, 'CF'    : CF, 'format': str_format, }).encode('utf-8')

			rrd_params.append(print_param)

	outfile = image_path

	rrd_params = ["rrdtool", "graph", outfile] + rrd_params

	cmd = " ".join(rrd_params)

	subprocess.call(cmd, shell=True)




